type token =
  | LEFT_PARENTHESIS
  | RIGHT_PARENTHESIS
  | LEFT_BRACKET
  | RIGHT_BRACKET
  | DOT
  | NEGATION
  | COLON
  | COMMA
  | EQUALITY
  | DISEQUALITY
  | EOI
  | FOF
  | CNF
  | INCLUDE
  | SINGLE_QUOTED of (string)
  | DOLLAR_WORD of (string)
  | DOLLAR_DOLLAR_WORD of (string)
  | DISTINCT_OBJECT of (string)
  | LOWER_WORD of (string)
  | UPPER_WORD of (string)
  | UNSIGNED_INTEGER of (string)
  | SIGNED_INTEGER of (string)
  | REAL of (string)
  | DOLLAR_TRUE
  | DOLLAR_FALSE
  | DOLLAR
  | AND
  | OR
  | FORALL
  | EXISTS
  | BIJECTION
  | LEFT_IMPLICATION
  | RIGHT_IMPLICATION
  | UNKNOWN

open Parsing;;
# 22 "src/parser_tptp.mly"
  
  type var = Var.var
  type symbol = Symbol.symbol
  type term = Term.term
  type literal = Term.literal
  type clause = Term.clause

  (* these are only valid for the current clause
     and have to be invalidated with init_clause for every new clause *)

  (* the variable id counter for the currently read term/clause *)
  let var_id_counter = 
    Counter.create_with 0
      
  (* mapping of the variable names (e.g. "X") of the currently read term/clause
     to variables. *)
  let (var_map: (string * Var.var) list ref) =
    ref []

  (* the literals of the currently read clause *)
  let (literals: literal list ref) =
    ref []


  (* reset everything in order to parse a new term/clause *)
  let init_clause () =
    Counter.set var_id_counter 0;
    var_map := [];
    literals := []


	
  (* gets the variables associated with a string from the variable mapping
     creates a new mapping for a new variable *)
  let get_var (var_name: string) =
    try 
      (* way faster than List.assoc *)
      snd (
	List.find
	  (fun (var_name', _) ->
             var_name = var_name'
	  )
	  !var_map
      )
    with
      | Not_found ->
	  let new_var = 
	    Var.create_universal (Counter.value var_id_counter)
	  in
	    Counter.inc var_id_counter;
	    var_map := (var_name, new_var) :: !var_map;
	    new_var

  (* need to detect if input is fof and contains a conjecture *)
  (* is the input in fof format? *)
  let fof =
    ref false

  (* does the fof input contain a conjecture? *)
  let theorem =
    ref false

# 102 "src/parser_tptp.ml"
let yytransl_const = [|
  257 (* LEFT_PARENTHESIS *);
  258 (* RIGHT_PARENTHESIS *);
  259 (* LEFT_BRACKET *);
  260 (* RIGHT_BRACKET *);
  261 (* DOT *);
  262 (* NEGATION *);
  263 (* COLON *);
  264 (* COMMA *);
  265 (* EQUALITY *);
  266 (* DISEQUALITY *);
  267 (* EOI *);
  268 (* FOF *);
  269 (* CNF *);
  270 (* INCLUDE *);
  280 (* DOLLAR_TRUE *);
  281 (* DOLLAR_FALSE *);
  282 (* DOLLAR *);
  283 (* AND *);
  284 (* OR *);
  285 (* FORALL *);
  286 (* EXISTS *);
  287 (* BIJECTION *);
  288 (* LEFT_IMPLICATION *);
  289 (* RIGHT_IMPLICATION *);
  290 (* UNKNOWN *);
    0|]

let yytransl_block = [|
  271 (* SINGLE_QUOTED *);
  272 (* DOLLAR_WORD *);
  273 (* DOLLAR_DOLLAR_WORD *);
  274 (* DISTINCT_OBJECT *);
  275 (* LOWER_WORD *);
  276 (* UPPER_WORD *);
  277 (* UNSIGNED_INTEGER *);
  278 (* SIGNED_INTEGER *);
  279 (* REAL *);
    0|]

let yylhs = "\255\255\
\001\000\001\000\002\000\002\000\003\000\003\000\004\000\004\000\
\006\000\010\000\010\000\012\000\012\000\014\000\016\000\016\000\
\016\000\016\000\016\000\016\000\015\000\015\000\017\000\019\000\
\019\000\018\000\020\000\020\000\013\000\013\000\013\000\013\000\
\021\000\024\000\024\000\025\000\025\000\022\000\027\000\007\000\
\009\000\011\000\011\000\028\000\028\000\032\000\032\000\033\000\
\033\000\023\000\023\000\023\000\034\000\038\000\038\000\035\000\
\035\000\035\000\035\000\036\000\039\000\039\000\041\000\041\000\
\041\000\037\000\037\000\042\000\042\000\045\000\046\000\043\000\
\043\000\040\000\040\000\044\000\044\000\050\000\049\000\026\000\
\030\000\031\000\031\000\053\000\005\000\056\000\056\000\057\000\
\057\000\052\000\052\000\052\000\058\000\058\000\058\000\058\000\
\060\000\060\000\059\000\059\000\054\000\054\000\008\000\008\000\
\047\000\047\000\051\000\048\000\048\000\048\000\055\000\029\000\
\000\000"

let yylen = "\002\000\
\002\000\001\000\001\000\002\000\001\000\001\000\001\000\001\000\
\010\000\001\000\001\000\001\000\001\000\003\000\001\000\001\000\
\001\000\001\000\002\000\002\000\001\000\001\000\003\000\001\000\
\003\000\003\000\001\000\003\000\001\000\001\000\003\000\001\000\
\006\000\001\000\001\000\001\000\003\000\002\000\001\000\010\000\
\001\000\001\000\003\000\003\000\001\000\001\000\003\000\001\000\
\002\000\001\000\001\000\001\000\001\000\001\000\003\000\001\000\
\001\000\003\000\003\000\001\000\001\000\001\000\001\000\001\000\
\001\000\001\000\004\000\001\000\004\000\001\000\001\000\001\000\
\001\000\001\000\004\000\001\000\004\000\001\000\001\000\001\000\
\001\000\002\000\001\000\001\000\006\000\004\000\001\000\001\000\
\003\000\001\000\003\000\001\000\001\000\004\000\001\000\001\000\
\001\000\003\000\002\000\003\000\001\000\003\000\001\000\001\000\
\001\000\001\000\001\000\001\000\001\000\001\000\001\000\000\000\
\002\000"

let yydefred = "\000\000\
\000\000\000\000\002\000\000\000\000\000\000\000\113\000\000\000\
\000\000\005\000\006\000\007\000\008\000\000\000\000\000\000\000\
\001\000\004\000\106\000\105\000\104\000\000\000\103\000\000\000\
\111\000\000\000\000\000\000\000\000\000\087\000\000\000\041\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\085\000\000\000\039\000\107\000\073\000\080\000\110\000\109\000\
\108\000\056\000\057\000\034\000\035\000\000\000\010\000\000\000\
\012\000\013\000\021\000\022\000\029\000\030\000\032\000\000\000\
\062\000\000\000\050\000\051\000\052\000\053\000\000\000\060\000\
\061\000\063\000\064\000\065\000\000\000\000\000\000\000\072\000\
\000\000\000\000\000\000\000\000\000\000\048\000\000\000\045\000\
\000\000\000\000\086\000\000\000\000\000\000\000\042\000\000\000\
\000\000\000\000\015\000\016\000\017\000\018\000\000\000\000\000\
\038\000\000\000\000\000\000\000\000\000\000\000\049\000\000\000\
\000\000\089\000\031\000\096\000\000\000\000\000\000\000\095\000\
\081\000\000\000\092\000\000\000\020\000\019\000\000\000\026\000\
\000\000\023\000\014\000\000\000\000\000\058\000\068\000\000\000\
\076\000\000\000\059\000\000\000\000\000\000\000\044\000\000\000\
\047\000\099\000\000\000\000\000\000\000\083\000\043\000\000\000\
\000\000\009\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\040\000\000\000\100\000\082\000\084\000\
\000\000\000\000\091\000\028\000\025\000\000\000\037\000\000\000\
\000\000\055\000\102\000\000\000\094\000\033\000\069\000\077\000\
\098\000"

let yydgoto = "\002\000\
\007\000\008\000\009\000\010\000\011\000\012\000\013\000\039\000\
\033\000\054\000\094\000\055\000\056\000\057\000\058\000\103\000\
\059\000\060\000\130\000\128\000\061\000\062\000\063\000\064\000\
\132\000\065\000\066\000\087\000\095\000\118\000\151\000\088\000\
\089\000\067\000\068\000\069\000\070\000\140\000\071\000\072\000\
\073\000\074\000\075\000\076\000\077\000\078\000\079\000\080\000\
\081\000\082\000\083\000\147\000\167\000\148\000\026\000\031\000\
\040\000\122\000\123\000\170\000"

let yysindex = "\030\000\
\097\255\000\000\000\000\032\255\034\255\039\255\000\000\048\255\
\121\255\000\000\000\000\000\000\000\000\011\255\011\255\002\255\
\000\000\000\000\000\000\000\000\000\000\062\255\000\000\090\255\
\000\000\092\255\086\255\086\255\021\000\000\000\128\255\000\000\
\129\255\130\255\011\255\131\255\048\000\073\000\132\255\053\000\
\000\000\048\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\133\255\000\000\126\255\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\145\255\
\000\000\048\000\000\000\000\000\000\000\000\000\009\255\000\000\
\000\000\000\000\000\000\000\000\000\000\148\255\000\000\000\000\
\000\000\149\255\000\000\093\000\104\000\000\000\133\255\000\000\
\123\255\011\255\000\000\150\255\174\255\154\255\000\000\027\255\
\048\000\048\000\000\000\000\000\000\000\000\000\048\000\141\255\
\000\000\104\255\104\255\104\255\104\255\160\255\000\000\161\255\
\093\000\000\000\000\000\000\000\165\255\156\255\168\255\000\000\
\000\000\158\255\000\000\162\255\000\000\000\000\151\255\000\000\
\147\255\000\000\000\000\172\255\169\255\000\000\000\000\178\255\
\000\000\180\255\000\000\183\255\182\255\189\255\000\000\193\255\
\000\000\000\000\186\255\089\000\174\255\000\000\000\000\174\255\
\174\255\000\000\048\000\048\000\194\255\141\255\104\255\104\255\
\000\000\104\255\000\000\000\000\174\255\000\000\000\000\000\000\
\195\255\203\255\000\000\000\000\000\000\048\000\000\000\209\255\
\211\255\000\000\000\000\174\255\000\000\000\000\000\000\000\000\
\000\000"

let yyrindex = "\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\191\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\213\255\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\119\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\213\255\000\000\026\255\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\014\255\000\000\198\255\000\000\
\242\255\000\000\208\255\000\000\000\000\000\000\213\255\000\000\
\041\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\213\255\164\255\000\000\
\000\000\166\255\000\000\000\000\000\000\000\000\077\255\000\000\
\087\255\000\000\000\000\000\000\215\255\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\218\255\000\000\000\000\000\000\
\000\000\000\000\171\255\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\251\255\000\000\028\000\000\000\000\000\000\000\000\000\000\000\
\219\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000"

let yygindex = "\000\000\
\000\000\213\000\000\000\000\000\000\000\000\000\000\000\067\000\
\195\000\182\000\140\000\000\000\197\255\000\000\000\000\000\000\
\000\000\000\000\072\000\078\000\000\000\000\000\229\255\000\000\
\076\000\167\255\000\000\000\000\240\255\000\000\000\000\199\255\
\000\000\000\000\000\000\000\000\000\000\184\255\200\255\000\000\
\000\000\000\000\000\000\000\000\210\255\214\255\246\255\175\255\
\225\255\241\255\000\000\177\255\000\000\120\255\000\000\000\000\
\147\000\000\000\000\000\058\000"

let yytablesize = 385
let yytable = "\117\000\
\093\000\146\000\090\000\023\000\023\000\025\000\105\000\101\000\
\117\000\030\000\086\000\120\000\168\000\121\000\133\000\066\000\
\019\000\106\000\107\000\066\000\020\000\066\000\068\000\068\000\
\023\000\019\000\110\000\011\000\179\000\020\000\001\000\021\000\
\014\000\011\000\015\000\120\000\142\000\127\000\129\000\016\000\
\066\000\066\000\046\000\131\000\066\000\066\000\066\000\066\000\
\046\000\134\000\139\000\141\000\141\000\125\000\126\000\145\000\
\086\000\111\000\017\000\135\000\135\000\135\000\135\000\136\000\
\136\000\136\000\136\000\120\000\133\000\027\000\120\000\120\000\
\169\000\171\000\137\000\137\000\137\000\137\000\027\000\023\000\
\022\000\024\000\119\000\120\000\027\000\086\000\176\000\177\000\
\024\000\178\000\138\000\138\000\138\000\138\000\024\000\127\000\
\129\000\028\000\120\000\029\000\169\000\150\000\141\000\141\000\
\032\000\141\000\119\000\003\000\004\000\005\000\006\000\035\000\
\135\000\135\000\182\000\135\000\136\000\136\000\019\000\136\000\
\044\000\045\000\020\000\046\000\047\000\048\000\049\000\137\000\
\137\000\036\000\137\000\096\000\004\000\005\000\006\000\041\000\
\037\000\038\000\119\000\090\000\093\000\119\000\119\000\138\000\
\138\000\091\000\138\000\104\000\108\000\109\000\113\000\115\000\
\097\000\098\000\119\000\124\000\099\000\100\000\101\000\102\000\
\046\000\143\000\144\000\149\000\153\000\093\000\154\000\090\000\
\152\000\119\000\093\000\093\000\101\000\090\000\156\000\157\000\
\158\000\155\000\159\000\019\000\160\000\166\000\116\000\020\000\
\161\000\047\000\048\000\049\000\019\000\162\000\163\000\116\000\
\020\000\165\000\047\000\048\000\049\000\164\000\071\000\070\000\
\174\000\003\000\180\000\070\000\181\000\070\000\070\000\070\000\
\078\000\079\000\183\000\088\000\184\000\079\000\112\000\079\000\
\079\000\079\000\036\000\054\000\097\000\018\000\034\000\092\000\
\070\000\070\000\112\000\173\000\070\000\070\000\070\000\070\000\
\172\000\175\000\079\000\079\000\114\000\185\000\079\000\079\000\
\079\000\079\000\000\000\074\000\000\000\000\000\000\000\074\000\
\000\000\074\000\076\000\076\000\067\000\000\000\000\000\000\000\
\067\000\000\000\067\000\069\000\069\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\074\000\074\000\000\000\000\000\
\074\000\074\000\074\000\074\000\000\000\067\000\067\000\000\000\
\000\000\067\000\067\000\067\000\067\000\075\000\000\000\000\000\
\000\000\075\000\000\000\075\000\077\000\077\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\042\000\000\000\000\000\000\000\000\000\043\000\075\000\075\000\
\000\000\000\000\075\000\075\000\075\000\075\000\019\000\000\000\
\044\000\045\000\020\000\046\000\047\000\048\000\049\000\050\000\
\051\000\084\000\000\000\000\000\052\000\053\000\085\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\019\000\
\000\000\044\000\045\000\020\000\046\000\047\000\048\000\049\000\
\050\000\051\000\085\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\019\000\000\000\044\000\045\000\020\000\
\046\000\047\000\048\000\049\000\050\000\051\000\019\000\000\000\
\044\000\045\000\020\000\046\000\047\000\048\000\049\000\050\000\
\051\000"

let yycheck = "\091\000\
\093\000\093\000\093\000\014\000\015\000\016\000\066\000\093\000\
\091\000\026\000\038\000\093\000\149\000\093\000\104\000\002\001\
\015\001\009\001\010\001\006\001\019\001\008\001\009\001\010\001\
\035\000\015\001\084\000\002\001\165\000\019\001\001\000\021\001\
\001\001\008\001\001\001\117\000\109\000\097\000\098\000\001\001\
\027\001\028\001\002\001\103\000\031\001\032\001\033\001\034\001\
\008\001\106\000\107\000\108\000\109\000\027\001\028\001\113\000\
\084\000\085\000\011\001\106\000\107\000\108\000\109\000\106\000\
\107\000\108\000\109\000\149\000\158\000\008\001\152\000\153\000\
\152\000\153\000\106\000\107\000\108\000\109\000\002\001\090\000\
\014\000\015\000\093\000\165\000\008\001\113\000\159\000\160\000\
\002\001\162\000\106\000\107\000\108\000\109\000\008\001\155\000\
\156\000\008\001\180\000\008\001\180\000\118\000\159\000\160\000\
\019\001\162\000\117\000\011\001\012\001\013\001\014\001\091\000\
\159\000\160\000\174\000\162\000\159\000\160\000\015\001\162\000\
\017\001\018\001\019\001\020\001\021\001\022\001\023\001\159\000\
\160\000\002\001\162\000\006\001\012\001\013\001\014\001\005\001\
\008\001\008\001\149\000\008\001\008\001\152\000\153\000\159\000\
\160\000\093\000\162\000\003\001\001\001\001\001\028\001\002\001\
\027\001\028\001\165\000\002\001\031\001\032\001\033\001\034\001\
\020\001\002\001\002\001\008\001\007\001\002\001\005\001\002\001\
\001\001\180\000\007\001\008\001\002\001\008\001\028\001\004\001\
\008\001\027\001\001\001\015\001\001\001\093\000\018\001\019\001\
\002\001\021\001\022\001\023\001\015\001\008\001\002\001\018\001\
\019\001\008\001\021\001\022\001\023\001\005\001\001\001\002\001\
\007\001\011\001\008\001\006\001\002\001\008\001\009\001\010\001\
\001\001\002\001\002\001\093\000\002\001\006\001\002\001\008\001\
\009\001\010\001\004\001\002\001\002\001\009\000\028\000\042\000\
\027\001\028\001\087\000\156\000\031\001\032\001\033\001\034\001\
\155\000\158\000\027\001\028\001\090\000\180\000\031\001\032\001\
\033\001\034\001\255\255\002\001\255\255\255\255\255\255\006\001\
\255\255\008\001\009\001\010\001\002\001\255\255\255\255\255\255\
\006\001\255\255\008\001\009\001\010\001\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\027\001\028\001\255\255\255\255\
\031\001\032\001\033\001\034\001\255\255\027\001\028\001\255\255\
\255\255\031\001\032\001\033\001\034\001\002\001\255\255\255\255\
\255\255\006\001\255\255\008\001\009\001\010\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\001\001\255\255\255\255\255\255\255\255\006\001\027\001\028\001\
\255\255\255\255\031\001\032\001\033\001\034\001\015\001\255\255\
\017\001\018\001\019\001\020\001\021\001\022\001\023\001\024\001\
\025\001\001\001\255\255\255\255\029\001\030\001\006\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\015\001\
\255\255\017\001\018\001\019\001\020\001\021\001\022\001\023\001\
\024\001\025\001\006\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\015\001\255\255\017\001\018\001\019\001\
\020\001\021\001\022\001\023\001\024\001\025\001\015\001\255\255\
\017\001\018\001\019\001\020\001\021\001\022\001\023\001\024\001\
\025\001"

let yynames_const = "\
  LEFT_PARENTHESIS\000\
  RIGHT_PARENTHESIS\000\
  LEFT_BRACKET\000\
  RIGHT_BRACKET\000\
  DOT\000\
  NEGATION\000\
  COLON\000\
  COMMA\000\
  EQUALITY\000\
  DISEQUALITY\000\
  EOI\000\
  FOF\000\
  CNF\000\
  INCLUDE\000\
  DOLLAR_TRUE\000\
  DOLLAR_FALSE\000\
  DOLLAR\000\
  AND\000\
  OR\000\
  FORALL\000\
  EXISTS\000\
  BIJECTION\000\
  LEFT_IMPLICATION\000\
  RIGHT_IMPLICATION\000\
  UNKNOWN\000\
  "

let yynames_block = "\
  SINGLE_QUOTED\000\
  DOLLAR_WORD\000\
  DOLLAR_DOLLAR_WORD\000\
  DISTINCT_OBJECT\000\
  LOWER_WORD\000\
  UPPER_WORD\000\
  UNSIGNED_INTEGER\000\
  SIGNED_INTEGER\000\
  REAL\000\
  "

let yyact = [|
  (fun _ -> failwith "parser")
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'file) in
    Obj.repr(
# 131 "src/parser_tptp.mly"
      (
	let clauses =
          _1
        in
	let is_fof =
	  !fof
	and is_theorem =
	  !theorem
	in
	  (* reset for next parser run *)
	  fof := false;
	  theorem := false;

          if is_fof then
	    raise (Const.FOF is_theorem)

          else
            clauses
      )
# 443 "src/parser_tptp.ml"
               : Term.clause list))
; (fun __caml_parser_env ->
    Obj.repr(
# 152 "src/parser_tptp.mly"
      ( print_endline "empty problem specification"; raise Const.PARSE_ERROR )
# 449 "src/parser_tptp.ml"
               : Term.clause list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'tptp_input) in
    Obj.repr(
# 161 "src/parser_tptp.mly"
      ( [_1] )
# 456 "src/parser_tptp.ml"
               : 'file))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'tptp_input) in
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'file) in
    Obj.repr(
# 164 "src/parser_tptp.mly"
      ( _1 :: _2 )
# 464 "src/parser_tptp.ml"
               : 'file))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'annotated_formula) in
    Obj.repr(
# 168 "src/parser_tptp.mly"
      ( _1 )
# 471 "src/parser_tptp.ml"
               : 'tptp_input))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'include_) in
    Obj.repr(
# 171 "src/parser_tptp.mly"
      ( _1 )
# 478 "src/parser_tptp.ml"
               : 'tptp_input))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'fof_annotated) in
    Obj.repr(
# 177 "src/parser_tptp.mly"
      ( _1 )
# 485 "src/parser_tptp.ml"
               : 'annotated_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'cnf_annotated) in
    Obj.repr(
# 180 "src/parser_tptp.mly"
      ( _1 )
# 492 "src/parser_tptp.ml"
               : 'annotated_formula))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 7 : 'name) in
    let _5 = (Parsing.peek_val __caml_parser_env 5 : 'formula_role) in
    let _7 = (Parsing.peek_val __caml_parser_env 3 : 'fof_formula) in
    let _8 = (Parsing.peek_val __caml_parser_env 2 : 'annotations) in
    Obj.repr(
# 184 "src/parser_tptp.mly"
    ( fof := true; [] )
# 502 "src/parser_tptp.ml"
               : 'fof_annotated))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'binary_formula) in
    Obj.repr(
# 188 "src/parser_tptp.mly"
    ( "" )
# 509 "src/parser_tptp.ml"
               : 'fof_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'unitary_formula) in
    Obj.repr(
# 191 "src/parser_tptp.mly"
    ( "" )
# 516 "src/parser_tptp.ml"
               : 'fof_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'nonassoc_binary) in
    Obj.repr(
# 196 "src/parser_tptp.mly"
    ( "" )
# 523 "src/parser_tptp.ml"
               : 'binary_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'assoc_binary) in
    Obj.repr(
# 199 "src/parser_tptp.mly"
    ( "" )
# 530 "src/parser_tptp.ml"
               : 'binary_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'unitary_formula) in
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'binary_connective) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'unitary_formula) in
    Obj.repr(
# 203 "src/parser_tptp.mly"
    ( "" )
# 539 "src/parser_tptp.ml"
               : 'nonassoc_binary))
; (fun __caml_parser_env ->
    Obj.repr(
# 207 "src/parser_tptp.mly"
    ( "" )
# 545 "src/parser_tptp.ml"
               : 'binary_connective))
; (fun __caml_parser_env ->
    Obj.repr(
# 209 "src/parser_tptp.mly"
    ( "" )
# 551 "src/parser_tptp.ml"
               : 'binary_connective))
; (fun __caml_parser_env ->
    Obj.repr(
# 211 "src/parser_tptp.mly"
    ( "" )
# 557 "src/parser_tptp.ml"
               : 'binary_connective))
; (fun __caml_parser_env ->
    Obj.repr(
# 213 "src/parser_tptp.mly"
    ( "" )
# 563 "src/parser_tptp.ml"
               : 'binary_connective))
; (fun __caml_parser_env ->
    Obj.repr(
# 215 "src/parser_tptp.mly"
    ( "" )
# 569 "src/parser_tptp.ml"
               : 'binary_connective))
; (fun __caml_parser_env ->
    Obj.repr(
# 217 "src/parser_tptp.mly"
    ( "" )
# 575 "src/parser_tptp.ml"
               : 'binary_connective))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'or_formula) in
    Obj.repr(
# 221 "src/parser_tptp.mly"
    ( "" )
# 582 "src/parser_tptp.ml"
               : 'assoc_binary))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'and_formula) in
    Obj.repr(
# 223 "src/parser_tptp.mly"
    ( "" )
# 589 "src/parser_tptp.ml"
               : 'assoc_binary))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'unitary_formula) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'more_or_formula) in
    Obj.repr(
# 227 "src/parser_tptp.mly"
    ( "" )
# 597 "src/parser_tptp.ml"
               : 'or_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'unitary_formula) in
    Obj.repr(
# 231 "src/parser_tptp.mly"
    ( "" )
# 604 "src/parser_tptp.ml"
               : 'more_or_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'unitary_formula) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'more_or_formula) in
    Obj.repr(
# 233 "src/parser_tptp.mly"
    ( "" )
# 612 "src/parser_tptp.ml"
               : 'more_or_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'unitary_formula) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'more_and_formula) in
    Obj.repr(
# 237 "src/parser_tptp.mly"
    ( "" )
# 620 "src/parser_tptp.ml"
               : 'and_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'unitary_formula) in
    Obj.repr(
# 241 "src/parser_tptp.mly"
    ( "" )
# 627 "src/parser_tptp.ml"
               : 'more_and_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'unitary_formula) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'more_and_formula) in
    Obj.repr(
# 243 "src/parser_tptp.mly"
    ( "" )
# 635 "src/parser_tptp.ml"
               : 'more_and_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'quantified_formula) in
    Obj.repr(
# 247 "src/parser_tptp.mly"
    ( "" )
# 642 "src/parser_tptp.ml"
               : 'unitary_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'unary_formula) in
    Obj.repr(
# 249 "src/parser_tptp.mly"
    ( "" )
# 649 "src/parser_tptp.ml"
               : 'unitary_formula))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'fof_formula) in
    Obj.repr(
# 251 "src/parser_tptp.mly"
    ( "" )
# 656 "src/parser_tptp.ml"
               : 'unitary_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_formula) in
    Obj.repr(
# 253 "src/parser_tptp.mly"
    ( "" )
# 663 "src/parser_tptp.ml"
               : 'unitary_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 5 : 'quantifier) in
    let _3 = (Parsing.peek_val __caml_parser_env 3 : 'variable_list) in
    let _6 = (Parsing.peek_val __caml_parser_env 0 : 'unitary_formula) in
    Obj.repr(
# 257 "src/parser_tptp.mly"
    ( "" )
# 672 "src/parser_tptp.ml"
               : 'quantified_formula))
; (fun __caml_parser_env ->
    Obj.repr(
# 261 "src/parser_tptp.mly"
    ( "" )
# 678 "src/parser_tptp.ml"
               : 'quantifier))
; (fun __caml_parser_env ->
    Obj.repr(
# 263 "src/parser_tptp.mly"
    ( "" )
# 684 "src/parser_tptp.ml"
               : 'quantifier))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'variable) in
    Obj.repr(
# 267 "src/parser_tptp.mly"
    ( "" )
# 691 "src/parser_tptp.ml"
               : 'variable_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'variable) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'variable_list) in
    Obj.repr(
# 269 "src/parser_tptp.mly"
    ( "" )
# 699 "src/parser_tptp.ml"
               : 'variable_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'unary_connective) in
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'unitary_formula) in
    Obj.repr(
# 273 "src/parser_tptp.mly"
    ( "" )
# 707 "src/parser_tptp.ml"
               : 'unary_formula))
; (fun __caml_parser_env ->
    Obj.repr(
# 277 "src/parser_tptp.mly"
    ( "" )
# 713 "src/parser_tptp.ml"
               : 'unary_connective))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 7 : 'name) in
    let _5 = (Parsing.peek_val __caml_parser_env 5 : 'formula_role) in
    let _7 = (Parsing.peek_val __caml_parser_env 3 : 'cnf_formula) in
    let _8 = (Parsing.peek_val __caml_parser_env 2 : 'annotations) in
    Obj.repr(
# 283 "src/parser_tptp.mly"
      (
	let clause = 
	  _7
	in
	  init_clause ();
	  clause
      )
# 729 "src/parser_tptp.ml"
               : 'cnf_annotated))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 293 "src/parser_tptp.mly"
    ( let role =
        _1
      in
        if role = "conjecture" then
          theorem := true;

        _1
    )
# 743 "src/parser_tptp.ml"
               : 'formula_role))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'null) in
    Obj.repr(
# 304 "src/parser_tptp.mly"
      ( "" )
# 750 "src/parser_tptp.ml"
               : 'annotations))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'source) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'optional_info) in
    Obj.repr(
# 307 "src/parser_tptp.mly"
      ( "" )
# 758 "src/parser_tptp.ml"
               : 'annotations))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'disjunction) in
    Obj.repr(
# 313 "src/parser_tptp.mly"
      ( _2 )
# 765 "src/parser_tptp.ml"
               : 'cnf_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'disjunction) in
    Obj.repr(
# 316 "src/parser_tptp.mly"
      ( _1 )
# 772 "src/parser_tptp.ml"
               : 'cnf_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'literal) in
    Obj.repr(
# 320 "src/parser_tptp.mly"
      ( [_1] )
# 779 "src/parser_tptp.ml"
               : 'disjunction))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'literal) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'disjunction) in
    Obj.repr(
# 323 "src/parser_tptp.mly"
      ( _1 :: _3 )
# 787 "src/parser_tptp.ml"
               : 'disjunction))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_formula) in
    Obj.repr(
# 330 "src/parser_tptp.mly"
      ( _1 )
# 794 "src/parser_tptp.ml"
               : 'literal))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_formula) in
    Obj.repr(
# 333 "src/parser_tptp.mly"
      ( Term.request_negated_literal _2 )
# 801 "src/parser_tptp.ml"
               : 'literal))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'plain_atom) in
    Obj.repr(
# 337 "src/parser_tptp.mly"
      ( _1 )
# 808 "src/parser_tptp.ml"
               : 'atomic_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'defined_atom) in
    Obj.repr(
# 340 "src/parser_tptp.mly"
      ( _1 )
# 815 "src/parser_tptp.ml"
               : 'atomic_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'system_atom) in
    Obj.repr(
# 343 "src/parser_tptp.mly"
      ( _1 )
# 822 "src/parser_tptp.ml"
               : 'atomic_formula))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'plain_term_top) in
    Obj.repr(
# 347 "src/parser_tptp.mly"
      ( Term.request_literal true _1 )
# 829 "src/parser_tptp.ml"
               : 'plain_atom))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'term) in
    Obj.repr(
# 351 "src/parser_tptp.mly"
      ( [ _1 ] )
# 836 "src/parser_tptp.ml"
               : 'arguments))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'term) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arguments) in
    Obj.repr(
# 354 "src/parser_tptp.mly"
      ( _1 :: _3 )
# 844 "src/parser_tptp.ml"
               : 'arguments))
; (fun __caml_parser_env ->
    Obj.repr(
# 358 "src/parser_tptp.mly"
      ( Term.true_literal )
# 850 "src/parser_tptp.ml"
               : 'defined_atom))
; (fun __caml_parser_env ->
    Obj.repr(
# 361 "src/parser_tptp.mly"
      ( Term.false_literal )
# 856 "src/parser_tptp.ml"
               : 'defined_atom))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'term) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'term) in
    Obj.repr(
# 364 "src/parser_tptp.mly"
      ( Term.request_literal true (Term.request_func (Symbol.equality, [| _1; _3|])) )
# 864 "src/parser_tptp.ml"
               : 'defined_atom))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'term) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'term) in
    Obj.repr(
# 367 "src/parser_tptp.mly"
      ( Term.request_literal false (Term.request_func (Symbol.equality, [| _1; _3|])) )
# 872 "src/parser_tptp.ml"
               : 'defined_atom))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'system_term_top) in
    Obj.repr(
# 371 "src/parser_tptp.mly"
      ( Term.request_literal true _1 )
# 879 "src/parser_tptp.ml"
               : 'system_atom))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'function_term) in
    Obj.repr(
# 376 "src/parser_tptp.mly"
      ( _1 )
# 886 "src/parser_tptp.ml"
               : 'term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'variable) in
    Obj.repr(
# 379 "src/parser_tptp.mly"
      ( Term.request_var _1 )
# 893 "src/parser_tptp.ml"
               : 'term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'plain_term) in
    Obj.repr(
# 383 "src/parser_tptp.mly"
      ( _1 )
# 900 "src/parser_tptp.ml"
               : 'function_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'defined_term) in
    Obj.repr(
# 386 "src/parser_tptp.mly"
      ( _1 )
# 907 "src/parser_tptp.ml"
               : 'function_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'system_term) in
    Obj.repr(
# 389 "src/parser_tptp.mly"
      ( _1 )
# 914 "src/parser_tptp.ml"
               : 'function_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'constant) in
    Obj.repr(
# 393 "src/parser_tptp.mly"
      ( Term.request_const (Symbol.create_predicate _1 0) )
# 921 "src/parser_tptp.ml"
               : 'plain_term_top))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : 'functor_) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'arguments) in
    Obj.repr(
# 396 "src/parser_tptp.mly"
      ( let subterms = Array.of_list _3 in
	  Term.request_func (Symbol.create_predicate _1 (Array.length subterms), subterms)
      )
# 931 "src/parser_tptp.ml"
               : 'plain_term_top))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'constant) in
    Obj.repr(
# 402 "src/parser_tptp.mly"
      ( Term.request_const (Symbol.create_function _1 0) )
# 938 "src/parser_tptp.ml"
               : 'plain_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : 'functor_) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'arguments) in
    Obj.repr(
# 405 "src/parser_tptp.mly"
      ( let subterms = Array.of_list _3 in
	  Term.request_func (Symbol.create_function _1 (Array.length subterms), subterms)
      )
# 948 "src/parser_tptp.ml"
               : 'plain_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_word) in
    Obj.repr(
# 411 "src/parser_tptp.mly"
      ( _1 )
# 955 "src/parser_tptp.ml"
               : 'constant))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_word) in
    Obj.repr(
# 415 "src/parser_tptp.mly"
      ( _1 )
# 962 "src/parser_tptp.ml"
               : 'functor_))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'number) in
    Obj.repr(
# 419 "src/parser_tptp.mly"
      ( print_endline ("Parser_tptp: <defined_term: number> not supported: " ^ _1); raise Const.PARSE_ERROR )
# 969 "src/parser_tptp.ml"
               : 'defined_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 422 "src/parser_tptp.mly"
      ( print_endline ("Parser_tptp: <defined_term: distinct_object> not supported: " ^ _1); raise Const.PARSE_ERROR )
# 976 "src/parser_tptp.ml"
               : 'defined_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'system_constant) in
    Obj.repr(
# 426 "src/parser_tptp.mly"
      ( Term.request_const (Symbol.create_predicate _1 0) )
# 983 "src/parser_tptp.ml"
               : 'system_term_top))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : 'system_functor) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'arguments) in
    Obj.repr(
# 429 "src/parser_tptp.mly"
      ( let subterms = Array.of_list _3 in
	  Term.request_func (Symbol.create_predicate _1 (Array.length subterms), subterms)
      )
# 993 "src/parser_tptp.ml"
               : 'system_term_top))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'system_constant) in
    Obj.repr(
# 435 "src/parser_tptp.mly"
      ( Term.request_const (Symbol.create_function _1 0) )
# 1000 "src/parser_tptp.ml"
               : 'system_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : 'system_functor) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'arguments) in
    Obj.repr(
# 438 "src/parser_tptp.mly"
      ( let subterms = Array.of_list _3 in
	  Term.request_func (Symbol.create_function _1 (Array.length subterms), subterms)
      )
# 1010 "src/parser_tptp.ml"
               : 'system_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_system_word) in
    Obj.repr(
# 444 "src/parser_tptp.mly"
      ( _1 )
# 1017 "src/parser_tptp.ml"
               : 'system_functor))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_system_word) in
    Obj.repr(
# 448 "src/parser_tptp.mly"
      ( _1 )
# 1024 "src/parser_tptp.ml"
               : 'system_constant))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 454 "src/parser_tptp.mly"
      ( get_var _1 )
# 1031 "src/parser_tptp.ml"
               : 'variable))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'general_term) in
    Obj.repr(
# 460 "src/parser_tptp.mly"
      ( "" )
# 1038 "src/parser_tptp.ml"
               : 'source))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'useful_info) in
    Obj.repr(
# 464 "src/parser_tptp.mly"
      ( "" )
# 1045 "src/parser_tptp.ml"
               : 'optional_info))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'null) in
    Obj.repr(
# 467 "src/parser_tptp.mly"
      ( "" )
# 1052 "src/parser_tptp.ml"
               : 'optional_info))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'general_term_list) in
    Obj.repr(
# 471 "src/parser_tptp.mly"
      ( "" )
# 1059 "src/parser_tptp.ml"
               : 'useful_info))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 3 : 'file_name) in
    let _4 = (Parsing.peek_val __caml_parser_env 2 : 'formula_selection) in
    Obj.repr(
# 476 "src/parser_tptp.mly"
      ( print_endline "Parser_tptp: <include> not supported."; raise Const.PARSE_ERROR )
# 1067 "src/parser_tptp.ml"
               : 'include_))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'name_list) in
    Obj.repr(
# 480 "src/parser_tptp.mly"
      ( _3 )
# 1074 "src/parser_tptp.ml"
               : 'formula_selection))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'null) in
    Obj.repr(
# 483 "src/parser_tptp.mly"
      ( [] )
# 1081 "src/parser_tptp.ml"
               : 'formula_selection))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'name) in
    Obj.repr(
# 487 "src/parser_tptp.mly"
      ( [_1] )
# 1088 "src/parser_tptp.ml"
               : 'name_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'name) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'name_list) in
    Obj.repr(
# 490 "src/parser_tptp.mly"
      ( _1 :: _3 )
# 1096 "src/parser_tptp.ml"
               : 'name_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'general_data) in
    Obj.repr(
# 496 "src/parser_tptp.mly"
      ( "" )
# 1103 "src/parser_tptp.ml"
               : 'general_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'general_data) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'general_term) in
    Obj.repr(
# 499 "src/parser_tptp.mly"
      ( "" )
# 1111 "src/parser_tptp.ml"
               : 'general_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'general_list) in
    Obj.repr(
# 502 "src/parser_tptp.mly"
      ( "" )
# 1118 "src/parser_tptp.ml"
               : 'general_term))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_word) in
    Obj.repr(
# 506 "src/parser_tptp.mly"
      ( "" )
# 1125 "src/parser_tptp.ml"
               : 'general_data))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : 'atomic_word) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'general_arguments) in
    Obj.repr(
# 509 "src/parser_tptp.mly"
      ( "" )
# 1133 "src/parser_tptp.ml"
               : 'general_data))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'number) in
    Obj.repr(
# 512 "src/parser_tptp.mly"
      ( "" )
# 1140 "src/parser_tptp.ml"
               : 'general_data))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 515 "src/parser_tptp.mly"
      ( "" )
# 1147 "src/parser_tptp.ml"
               : 'general_data))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'general_term) in
    Obj.repr(
# 519 "src/parser_tptp.mly"
      ( [_1] )
# 1154 "src/parser_tptp.ml"
               : 'general_arguments))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'general_term) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'general_arguments) in
    Obj.repr(
# 522 "src/parser_tptp.mly"
      ( _1 :: _3 )
# 1162 "src/parser_tptp.ml"
               : 'general_arguments))
; (fun __caml_parser_env ->
    Obj.repr(
# 526 "src/parser_tptp.mly"
      ( [] )
# 1168 "src/parser_tptp.ml"
               : 'general_list))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'general_term_list) in
    Obj.repr(
# 529 "src/parser_tptp.mly"
      ( _2 )
# 1175 "src/parser_tptp.ml"
               : 'general_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'general_term) in
    Obj.repr(
# 533 "src/parser_tptp.mly"
      ( [_1] )
# 1182 "src/parser_tptp.ml"
               : 'general_term_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'general_term) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'general_term_list) in
    Obj.repr(
# 536 "src/parser_tptp.mly"
      ( _1 :: _3 )
# 1190 "src/parser_tptp.ml"
               : 'general_term_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_word) in
    Obj.repr(
# 541 "src/parser_tptp.mly"
      ( _1 )
# 1197 "src/parser_tptp.ml"
               : 'name))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 544 "src/parser_tptp.mly"
      ( _1 )
# 1204 "src/parser_tptp.ml"
               : 'name))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 548 "src/parser_tptp.mly"
      ( _1 )
# 1211 "src/parser_tptp.ml"
               : 'atomic_word))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 551 "src/parser_tptp.mly"
      ( _1 )
# 1218 "src/parser_tptp.ml"
               : 'atomic_word))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 555 "src/parser_tptp.mly"
      ( print_endline ("Parser_tptp: <$$word> not supported: " ^ _1); raise Const.PARSE_ERROR )
# 1225 "src/parser_tptp.ml"
               : 'atomic_system_word))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 559 "src/parser_tptp.mly"
      ( _1 )
# 1232 "src/parser_tptp.ml"
               : 'number))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 562 "src/parser_tptp.mly"
      ( _1 )
# 1239 "src/parser_tptp.ml"
               : 'number))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 565 "src/parser_tptp.mly"
      ( _1 )
# 1246 "src/parser_tptp.ml"
               : 'number))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_word) in
    Obj.repr(
# 569 "src/parser_tptp.mly"
      ( _1 )
# 1253 "src/parser_tptp.ml"
               : 'file_name))
; (fun __caml_parser_env ->
    Obj.repr(
# 572 "src/parser_tptp.mly"
      ( )
# 1259 "src/parser_tptp.ml"
               : 'null))
(* Entry parse_file *)
; (fun __caml_parser_env -> raise (Parsing.YYexit (Parsing.peek_val __caml_parser_env 0)))
|]
let yytables =
  { Parsing.actions=yyact;
    Parsing.transl_const=yytransl_const;
    Parsing.transl_block=yytransl_block;
    Parsing.lhs=yylhs;
    Parsing.len=yylen;
    Parsing.defred=yydefred;
    Parsing.dgoto=yydgoto;
    Parsing.sindex=yysindex;
    Parsing.rindex=yyrindex;
    Parsing.gindex=yygindex;
    Parsing.tablesize=yytablesize;
    Parsing.table=yytable;
    Parsing.check=yycheck;
    Parsing.error_function=parse_error;
    Parsing.names_const=yynames_const;
    Parsing.names_block=yynames_block }
let parse_file (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =
   (Parsing.yyparse yytables 1 lexfun lexbuf : Term.clause list)
;;
# 575 "src/parser_tptp.mly"




# 1289 "src/parser_tptp.ml"
