%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% 
%%%% Some Preprocessing For KRH
%%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%:- use_module('../contrib/pl/swilib/pretty').
%:- use_module('../contrib/pl/swilib/fromonto').
:- use_module(rrtrafo_ecl).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- op(500, fx, #).

preprocess_argv :-
%	current_prolog_flag(argv, Argv),
%	get_flag(argv, Argv), 
	argv(all, Argv),
	append(_, [_|Args], Argv),
	!,
	( memberchk("-rr", Args) ->
		(memberchk("-ehyper", Args) ->
			Mode = ehyper
		  ; Mode = rr)
	  ; (memberchk("-ehyper", Args) ->
			Mode = ehyper
		  ; Mode = standard)
	),
	( memberchk("-fixed_dom", Args) ->
	  FixedDom = true
	; FixedDom = false
	),
	( memberchk("-input_dom", Args) ->
	  InputDom = true
	; InputDom = false
	),
	Options = [mode=Mode, fixed_dom=FixedDom, input_dom=InputDom],
	get_regular_arg(Args, InputFile),
	preprocess_file(Options, InputFile).

main :-
	fix_ops,
%        catch(preprocess_argv, E, (print_message(error, E), fail)),
	block(
		preprocess_argv,
		E,
		(printf("Exception: %w%n", [E]), fail)),
% 	preprocess_argv,
       halt.                                           
main :-
        halt.

from_file(Goal, File) :-
        open(File, read, Stream),
%        current_input(In),
%	block( (call(Goal) -> Result = true ; Result = false),
%	       E,
%	       ( set_input(In),
%                close(Stream),
%		 throw(E)
%	       )
%            ),
%        set_input(In),
%        close(Stream),
%        Result = true.
	block(
		(call(Goal) -> Result = true ; Result = false),
		E,
		(close(Stream), exit_block(E))
	),
%	call(Goal),
	close(Stream),
	Result = true.


get_regular_arg(Args, Arg) :-
	reverse(Args, [Arg|_]),
%	\+ sub_atom(Arg, 0, 1, _, '-'),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

preprocess_file(Options, File) :-
	write('% Reading...'),
	nl,
	file_clauses(File, Cs),
	length(Cs, Len),
	write('% '),
	printf('%d clauses.', [Len]),
	nl,
	domain_restrict_clauses(Cs, Options, Cs1),
	write('% Writing...'),
	nl,
	( member(C, Cs1),
%	  write_term(C, [quoted(true),ignore_ops(true),portray(true)]),
	write_canonical(C),	    
	  writeln('.'),  
	  % pp_clause(C),
	  fail
	; true
	),
	write('% Done.'),
	nl.

file_clauses(File, Clauses) :-
       open(File, read, Stream),
	findall(T,
		(	repeat,
			read(Stream, T),
			(T== end_of_file -> !, fail;
			true)),
		Clauses),
	close(Stream).
%	write(Clauses).


%	from_file(findall(T,
%			  ( repeat,
%			    read(T),write(T),
%			    ( T == end_of_file ->
%				!,
%		              fail
%			    ; true
%			    )
%			  ),
%			  Clauses),
%		  File).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%msg1(Msg) :-
%	msg1(Msg, []).
%msg1(Msg, Args) :-
%	write(user_error, '% Pre: '),
%	format(user_error, Msg, Args),
%	nl(user_error),
%	flush_output(user_error).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%  
%%%% SWI-prolog prints operators >= 1000 in brackets (also if
%%%% ignore_ops is true). portray is used here to patch this.
%%%%  
fix_ops :-
	findall(O, (current_op(P,_,O), P >= 1000), Os),
	sort(Os, Os1),
	( member(O1, Os1),
	  asserta(portray(O1) :- write_term(O1, [quoted(true)])),
	  fail
	; true
	).
	
