(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

val key_term : Opt.indexargs -> Term.term -> Term.term

(** Objects through which predicate extents are accessed. *)

(** {6 Support of Modifiable States } *)

module Store :
    sig
      type store
      type state
      val obtain_state : store -> state
      val install_state : state -> unit
      val add_layer : state -> state
      val new_layer : store -> state
    end

(** {6 Access Objects } *)

module Writer :
  sig
    type writer
    val make_writer : Opt.indextype -> Opt.indexargs -> int -> writer
    val insert : writer -> Term.term -> Fact.entry -> unit
    val remove_variant : writer -> Term.term -> unit
    val clear : writer -> unit
    val get_store : writer -> Store.store
    val size : writer -> int
  end

module Reader :
  sig
    type reader
    val make_readers : Writer.writer -> reader list
    val null_reader : unit -> reader
    val applicability : reader -> Opt.binding_pattern ->
      Opt.applicability_value
    val iter_unification_candidates :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_instance_candidates :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_variant_candidates :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_generalization_candidates :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_unification_candidates_ml :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_instance_candidates_ml :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_variant_candidates_ml :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_generalization_candidates_ml :
      (Fact.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit      
    val iter_all : (Fact.entry -> unit) -> reader -> unit
    val get_indexargs : reader -> Opt.indexargs
(** DEBUG *)
    val print_info : reader -> unit
  end


