(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


type cfg = {
    mutable current_directory: string;
    mutable echo_flag: bool;
    mutable debug_flag: bool;
    mutable back_subsumption_flag: bool;
    mutable back_subsumption_within_news_flag: bool;
    mutable sort_heads_flag: bool;
    mutable check_disjunctions_are_ground_flag: bool;
    mutable proof_terms_flag: bool;
    mutable level_cut_flag: bool;
    mutable negative_units_flag: bool;
    mutable merge_inf_input_flag: bool;
    mutable wp_semantics_flag: bool;
    mutable show_derived_literals_flag: bool;
    mutable max_weight_initial_param: int;
    mutable max_weight_increment_param: int;
    mutable body_sorting_method_param: int;
    mutable weighing_method_param: int;
    mutable sup_left_flag: bool;
    mutable unit_sup_right_flag: bool;    
    mutable ehyper: bool;
    mutable full_branch_redundancy_flag: bool;
    mutable hyperextension_flag: bool;
    mutable szs_output_flag: bool;
    mutable szs_problem_name: string
  }

let get_current_directory cfg =
  cfg.current_directory

let set_current_directory cfg directory =
  cfg.current_directory <- directory

let get_echo_flag cfg = cfg.echo_flag
let set_echo_flag cfg bool = cfg.echo_flag <- bool
let get_debug_flag cfg = cfg.debug_flag
let set_debug_flag cfg bool = cfg.debug_flag <- bool
let get_back_subsumption_flag cfg = cfg.back_subsumption_flag
let set_back_subsumption_flag cfg bool =
  cfg.back_subsumption_flag <- bool
let get_back_subsumption_within_news_flag cfg = 
  cfg.back_subsumption_within_news_flag
let set_back_subsumption_within_news_flag cfg bool =
  cfg.back_subsumption_within_news_flag <- bool
let get_sort_heads_flag cfg = cfg.sort_heads_flag
let set_sort_heads_flag cfg bool = cfg.sort_heads_flag <- bool
let get_check_disjunctions_are_ground_flag cfg =
  cfg.check_disjunctions_are_ground_flag
let set_check_disjunctions_are_ground_flag cfg bool =
  cfg.check_disjunctions_are_ground_flag <- bool
let get_proof_terms_flag cfg = cfg.proof_terms_flag
let set_proof_terms_flag cfg bool = cfg.proof_terms_flag <- bool
let get_level_cut_flag cfg = cfg.level_cut_flag
let set_level_cut_flag cfg bool = cfg.level_cut_flag <- bool
let get_negative_units_flag cfg = cfg.negative_units_flag
let set_negative_units_flag cfg bool = cfg.negative_units_flag <- bool
let get_merge_inf_input_flag cfg = cfg.merge_inf_input_flag
let set_merge_inf_input_flag cfg bool = cfg.merge_inf_input_flag <- bool
let get_wp_semantics_flag cfg = cfg.wp_semantics_flag
let set_wp_semantics_flag cfg bool = cfg.wp_semantics_flag <- bool
let get_show_derived_literals_flag cfg = cfg.show_derived_literals_flag
let set_show_derived_literals_flag cfg bool =
  cfg.show_derived_literals_flag <- bool
let get_max_weight_initial_param cfg = cfg.max_weight_initial_param
let set_max_weight_initial_param cfg n = cfg.max_weight_initial_param <- n
let get_max_weight_increment_param cfg =
  cfg.max_weight_increment_param
let set_max_weight_increment_param cfg n =
  cfg.max_weight_increment_param <- n
let get_body_sorting_method_param cfg = cfg.body_sorting_method_param
let set_body_sorting_method_param cfg n = cfg.body_sorting_method_param <- n
let get_weighing_method_param cfg = cfg.weighing_method_param
let set_weighing_method_param cfg n = cfg.weighing_method_param <- n
let get_sup_left_flag cfg = cfg.sup_left_flag
let set_sup_left_flag cfg bool = cfg.sup_left_flag <- bool
let get_unit_sup_right_flag cfg = cfg.unit_sup_right_flag
let set_unit_sup_right_flag cfg bool = cfg.unit_sup_right_flag <- bool
let get_ehyper cfg = cfg.ehyper
let set_ehyper cfg bool = cfg.ehyper <- bool
let get_full_branch_redundancy_flag cfg = cfg.full_branch_redundancy_flag
let set_full_branch_redundancy_flag cfg bool = cfg.full_branch_redundancy_flag <- bool
let get_hyperextension_flag cfg = cfg.hyperextension_flag
let set_hyperextension_flag cfg bool = cfg.hyperextension_flag <- bool
let set_szs_output_flag cfg bool = cfg.szs_output_flag <-  bool
let get_szs_output_flag cfg = cfg.szs_output_flag
let set_szs_problem_name cfg name = cfg.szs_problem_name <- name
let get_szs_problem_name cfg = cfg.szs_problem_name

let flags =
  [("back_subsumption",
    ("Use back subsumption - i.e. remove derived or input facts\n\
     if they are subsumed by a newly derived clause. [Works not with\n\
     disjunctions]",
       false,
       get_back_subsumption_flag, set_back_subsumption_flag));
   ("back_subsumption_within_news",
    ("Use back subsumption within the set of newly derived clauses.",
     false,
     get_back_subsumption_within_news_flag,
     set_back_subsumption_within_news_flag));
   ("sort_heads",
    ("Sort disjunctive heads heuristically.",
     true,
     get_sort_heads_flag, set_sort_heads_flag));
   ("check_disjunctions_are_ground",
    ("Verify at runtime for each derived disjunction that is it ground.",
     true,
     get_check_disjunctions_are_ground_flag,
     set_check_disjunctions_are_ground_flag));
   ("negative_units",
    ("Maintain negative units to detect refutations sooner and enable\n\
      complement splitting.",
      true,
     get_negative_units_flag, set_negative_units_flag));
   ("merge_inf_input",
    ("Merge facts inferred before any disjunctive branching into the input\n\
      facts. These derived facts then serve as lemmas in subsequent\n\
      tasks. This flag can be switched on and off around a call to 'run'\n\
      to generate certain lemmas.",
      false,
     get_merge_inf_input_flag, set_merge_inf_input_flag));
   ("wp_semantics",
    ("[Experimental] Enforce the `weak perfect' KRHyper semantics of\n\
      disjunctions. Overrides effects of other flags conflicting with\n\
      that semantics.",
     false,
     get_wp_semantics_flag, set_wp_semantics_flag));
   ("proof_terms",
    ("Generate proof terms for derived facts.",
     false,
     get_proof_terms_flag, set_proof_terms_flag));
   ("level_cut",
    ("Apply the level cut reduction. This is only effective if proof_terms\n\
      is also set.\n\
      Note: No level cut is applied at a node whose proof involves\n\
      stratified negation as failure or findall.",
     true,
     get_level_cut_flag, set_level_cut_flag));
   ("echo", 
    ("Print each input statement before it is processed.",
     false,
     get_echo_flag, set_echo_flag));
   ("debug",
    ("Debug mode. Print extra information about clause compilation.",
     false,
     get_debug_flag, set_debug_flag));
   ("sup_left_flag",
    ("Use superposition left. If set to false, ordered paramodulation left will be used.",
     true,
     get_sup_left_flag, set_sup_left_flag)); 
   ("unit_sup_right_flag",
    ("Use unit superposition right. If set to false, ordered paramodulation right will be used.",
     true,
     get_unit_sup_right_flag, set_unit_sup_right_flag));          
   ("full_branch_redundancy_flag",
    ("Use full branch redundancy handling. If set to true, any facts and clauses within the\n
    current branch may be declared redundant, disregarding the E-Hyper calculus rule about\n
    not crossing decision nodes.\n
    If set to false, the calculus will be performed strictly as defined.\n
    Note that there is no efficiency advantage in setting the flag to false, the option is merely\n
    included to allow for a strict calculus implementation. The default value is true.",
     true,
     get_full_branch_redundancy_flag, set_full_branch_redundancy_flag));
   ("hyperextension_flag",
    ("Use hyperextension. If set to true, the hyperextension inference of the (non-equational)\n
    Hyper-tableaux calculus will be used for resolving away non-equational literals, even when\n
    the program is running in E-Hyper mode (i.e. there are equational literals present). The\n
    hyperextension inference is equivalent to a series of sup-left and ref-inferences in the\n
    E-Hyper calculus, so this has no effect on soundness and completeness, and it is faster\n
    than treating all literals as equations. However, to allow for a strict E-Hyper calculus\n
    implementation, usage of the hyperextension can be switched off by setting this flag to false.\n
    In this case, all non-equational literals will be converted to equations, and only the inferences\n
    detailed in the E-Ehyper calculus description will be used.\n
    Note that in this case you should ensure that the input contains no predicates and functions sharing\
    arity and name, because this distinction will be lost during the conversion to equations.",
     true,
     get_hyperextension_flag, set_hyperextension_flag));   
   ("szs_output_flag",
    ("Display SZS-conforming output. If set to true, each result printout will include\n\
      one line displaying the result in SZS-syntax. If the input clause set is found to\n\
      be satisfiable, then the line will read:\n\
      'SZS status Satisfiable'\n\
      Otherwise the line will read:\n\
      'SZS status Unsatisfiable'\n\n\
      Notes: When enumerating models, a 'Closed' result indicates that no further models\n\
      exist. The SZS status may still be satisfiable, provided that at least one model\n\
      has been found.\n\
      The SZS status line is not commented out like other E-KRHyper output, to conform with\n\
      CASC-regulations. The output of E-KRHyper will therefore no longer be Prolog-readable\n\
      if this flag is set to true.",
     false,
     get_szs_output_flag, set_szs_output_flag));      
       
   ("show_derived_literals",
    ("Show facts as they are derived. May be useful for debugging.\n\
      Output are Prolog readable statements, one per line:\n\n  \
      derived_fact(Fact).  -- Fact has been derived in an extension step.\n  \
      attached_disjunction(DisjunctionId). -- The tableau has been\n      \
          extended  with a disjunction.\n  \
      selected_disjunct(Fact, DisjunctionId). -- Fact has been\n      \
       selected from an attached disjunction.\n\n\
     Notes: DisjunctionIds are currently not implemented.\n\
       Negative facts (see flag negative_units) are currently not considered.",
     false,
     get_show_derived_literals_flag, set_show_derived_literals_flag))]


let parameters =
  [("max_weight_initial", 
    ("The initial weight limit for inferred literals. If -1, weight\n\
      handling is not used. For infinite Herbrand domains, weight handling\n\
      can be required to ensure refutational completeness and\n\
      `finite-branch-model' completeness.",
     20,
     get_max_weight_initial_param, set_max_weight_initial_param));
   ("max_weight_increment", 
    ("The minimum amount by which the weight limit for inferred literals\n\
      is increased in a weight reset step. If max_weight_increment >= 0,\n\
      then the actual number used is the maximum of this number and the\n\
      smallest number by which an additional literal can possibly be\n\
      inferred. If max_weight_increment = -1, max_weight_initial is\n\
      used as weight limit and not incremented, yielding an obviously\n\
      incomplete calculus.",
      1,
      get_max_weight_increment_param,
      set_max_weight_increment_param));
   ("body_sorting_method", 
    ("Determines the heuristics for sorting body literals.\n\
      The following values can be used:\n\
      0 - A heuristics that ensures permissibility for builtins is used.\n\
      1 - The literal order of the input is used, except that delta\n    \
	  predicates are preferred. This may be useful on inputs\n    \
          resulting from program transformations.\n\
      2 - The literal order of the input is used.",
      0, 
      get_body_sorting_method_param,
      set_body_sorting_method_param));
   ("weighing_method", 
    ("Determines how the weight of a term is computed.\n\
      The following values can be used:\n\
      0 - The size of the term, i.e. the number of occurences of\n    \
	  predicate, function and constant symbols as well as variables\n    \
          in it is used.\n\
      1 - The depth of the term is used.",
     0,
     get_weighing_method_param,
     set_weighing_method_param))]
    
let set_to_defaults cfg =
  List.iter
    (fun (_, (_,default, _, writer)) ->
      (writer cfg default))
    flags;
  List.iter
    (fun (_, (_,default, _, writer)) ->
      (writer cfg default))
    parameters

let make_cfg () =
  let cfg =
    { current_directory=Sys.getcwd ();
      echo_flag=false;
      debug_flag=false;
      back_subsumption_flag=false;
      back_subsumption_within_news_flag=false;
      sort_heads_flag=false;
      check_disjunctions_are_ground_flag=false;
      negative_units_flag=false;
      merge_inf_input_flag=false;
      proof_terms_flag=false;
      level_cut_flag=false;
      wp_semantics_flag=false;
      show_derived_literals_flag=false;
      sup_left_flag=false;
      unit_sup_right_flag=false;      
      ehyper=false;
      max_weight_initial_param=0;
      max_weight_increment_param=0;
      body_sorting_method_param=0;
      weighing_method_param=0;
      full_branch_redundancy_flag=false;
      hyperextension_flag=false;
      szs_output_flag=false;
      szs_problem_name=""
    } in
  set_to_defaults cfg;
  cfg
   
let print_smallheader x =
  print_string "----------------------------------------";
  print_string "--------------------------------------\n";
  print_string x;
  print_string "\n----------------------------------------";
  print_string "--------------------------------------\n"

let print_settings cfg =
  List.iter
    (fun (flag, (_, _, reader, _)) ->
      print_string "set_flag(";
      print_string flag;
      print_string ", ";
      print_string (if reader cfg then "true" else "false");
      print_endline ").")
    flags;
  List.iter
    (fun (flag, (_, _, reader, _)) ->
      print_string "set_parameter(";
      print_string flag;
      print_string ", ";
      print_int (reader cfg);
      print_endline ").")
    parameters

let print_flags_doc cfg =
  List.iter
    (fun (flag, (doc, default, _, _)) ->
      print_smallheader (Printf.sprintf "Flag: %s, default: %b" flag default);
      Printf.printf "%s\n\n" doc)
    flags;
  flush stdout

let print_parameters_doc cfg =
  List.iter
    (fun (param, (doc, default, _, _)) ->
      print_smallheader (Printf.sprintf "Parameter: %s, default: %d" 
			   param default);
      Printf.printf "%s\n\n" doc)
    parameters;
  flush stdout

let set_flag cfg flag newval =
  try
    let (_, _, _, writer) = List.assoc flag flags in
    writer cfg newval
  with Not_found ->
    raise (Err.Error "Unknown flag specified.")

let set_parameter cfg param newval =
  try
    let (_, _, _, writer) = List.assoc param parameters in
    writer cfg newval
  with Not_found ->
    raise (Err.Error "Unknown parameter specified.")


(****************************************************************************)

type verbosity = int

let current_verbosity = ref 3

let get_verbosity () = !current_verbosity

let set_verbosity n = current_verbosity := n

let current_msg_output = ref stderr

let get_msg_output () = !current_msg_output

let set_msg_output s =
  match s with
    "stdout" -> current_msg_output := stdout
  | _ -> current_msg_output := stderr
