(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

let st = Term.make_symbol_table ()

let string_to_term symtable string =
  Stream.next 
    (PrologRead.prolog_parser symtable (Stream.of_string (string^".")))

let tst s1 s2 =
  let c = Term.make_context 0 in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2 in
  StdOrder.term_compare t1 c t2 c

let test_orient s1 =
	let t = string_to_term st s1 in
		let t2 = Equality.orient st t in
			PrologPrint.print t2

let test_sub s1 s2 =
  let c = Term.make_context 0 in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2 in
  LrpOrder.is_subterm t1 t2

let test_lrpo s1 s2 =
  let c = Term.make_context 0 in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2 in
  LrpOrder.lrpo_greater t1 t2

  
let test_lrpo_c s1 s2 =
  let c = Term.make_context 0 in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2 in
  LrpOrder.lrpo_greater_c t1 c t2 c  
  
let vtest s1 s2 =
    let c = Term.make_context 0 in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2 in
  Term.variant_chk t1 t2
  
let susutest s1 s2 =
    let c = Term.make_context 0 in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2 in
  Term.np_subsumes_chk t1 t2  
  
let test_lrpo2 s=
	let t=string_to_term st s in
		match t with
			Term.Compound c ->
				let t1=(Term.compound_args c).(0)
				and t2=(Term.compound_args c).(1)
				in
					LrpOrder.lrpo_greater t1 t2
			|_ -> raise (Failure "Incompatible input error.")
			
let gc_test s =
	let t=string_to_term st s in
		match t with
			Term.Compound c ->
				let t1=(Term.compound_args c).(2)
				and t2=(Term.compound_args c).(3)
				in
					LrpOrder.ground_like_lrpo_greater t1 t2
			|_ -> raise (Failure "Incompatible input error.")


    
(* This shows that renum is wrong if variables would be shared, try e.g.
   test_renum "p(f(X,Y), g(Y,X,Y))"
   [PrologRead.prolog_parser does no longer return shared vars,
   so this works correctly]
*)
let test_renum string =
  let st = Term.make_symbol_table ()
  and a1 = Array.make Term.max_vars (-1) in
  let t1 = (Convert.term_nth_arg (string_to_term st string) 1) in
  PrologPrint.print t1;
  print_newline ();
  Term.renumber_variables_term a1 t1;
  PrologPrint.print t1;
  print_newline ();
  let a2 = Array.make Term.max_vars (-1) in
  let t2 = Term.apply 
      (Convert.term_nth_arg (string_to_term st string) 1)
      (Term.make_context 0) in
  PrologPrint.print t2;
  print_newline ();
  Term.renumber_variables_term a2 t2;
  PrologPrint.print t2;
  print_newline ()

let u2 s1 s2 n =
  let st = Term.make_symbol_table ()
  and trail = Term.make_trail () in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2 in
  for i = 1 to n do
    let c1 = Term.make_context 0
    and c2 = Term.make_context 1
    and saved_trail = Term.save_trail trail in
    let trail1 = Term.unify t1 c1 t2 c2 trail in
    if Term.failed trail1 then
      begin
	()
      end
    else
      begin
	let r = Term.apply t1 c1 in ()
	  (* renumber ... *)
      end;
    let r = Term.undo_bindings trail saved_trail in ()
  done

(** Match two terms t1 >= t2 given as strings and print the result.
 Also test subsumes_chk. *)
let m1 s1 s2 =
  let st = Term.make_symbol_table ()
  and trail = Term.make_trail () in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2
  and c1 = Term.make_context 0 in
  if Term.subsumes_chk t1 t2 then
    print_string "Subsumes_chk succeeded\n"
  else
    print_string "Subsumes_chk failed\n";
  print_string "Matching\n";
  let trail1 = Term.match_terms t1 c1 t2 trail in
      if Term.failed trail1 then
	print_string "Failed"
      else
	begin
	  print_string "succ\n";
	  let t3 = Term.apply t1 c1 in
	  print_string "1: ";
	  PrologPrint.print t3;
	  print_newline ();
	  Term.renumber_variables_term (Array.make Term.max_vars (-1)) t3;
	  print_string "3: ";
	  PrologPrint.print t3;
	end

let u1 s1 s2 =
  let st = Term.make_symbol_table ()
  and trail = Term.make_trail () in
  let t1 = string_to_term st s1
  and t2 = string_to_term st s2
  and c1 = Term.make_context 0
  and c2 = Term.make_context 1 in
  let trail1 = Term.unify t1 c1 t2 c2 trail in
      if Term.failed trail1 then
	print_string "Failed"
      else
	begin
	  let t3 = Term.apply t1 c1 in
	  print_string "1: ";
	  PrologPrint.print t3;
	  print_newline ();
	  let t4 = Term.apply t2 c2 in
	  print_string "2: ";
	  PrologPrint.print t4;
	  print_newline ();
	  Term.renumber_variables_term (Array.make Term.max_vars (-1)) t3;
	  print_string "3: ";
	  PrologPrint.print t3;
	end
	

(** Read in a prolog file and print its contents in Prolog syntax. *)
let dump_prolog_file filename =
  let st = Term.make_symbol_table () 
  and input = open_in_bin filename in
  let p = PrologRead.prolog_parser st (Stream.of_channel input) in
  try 
    Stream.iter
      (fun term -> 
	PrologPrint.print term ;
	print_string ".";
	print_newline ())
      p;
    close_in input
  with e ->
    close_in input;
    raise e

let print_symbols symbol_table =
  Term.symbol_table_iter
    (fun symbol ->
      print_string (Term.symbol_name symbol);
      print_string "/";
      print_int (Term.symbol_arity symbol);
      print_newline ()
    )
    symbol_table

(** Parse Prolog terms in the given string and print them. Finally print
   the symbol table. *)
let print_terms string =
  let st = Term.make_symbol_table () in
  let p = PrologRead.prolog_parser st (Stream.of_string string) in
  Stream.iter 
    (fun term -> 
      PrologPrint.print term ;
      print_string ".";
      print_newline ())
    p;
  print_string "---------- Symbol Table ----------";
  print_newline ();
  print_symbols st


let rwl s =
  let p = PrologGenlex.prolog_lexer (Stream.of_string s) in p
