(*
* Copyright (C) 2006 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

exception Failed

(** Miscellaneous equality related operations. **)

val multi_ident : Term.term -> Term.term -> Term.term -> Term.term -> bool

type equation

val get_is_oriented : equation -> bool

val get_term : equation -> Term.term
	
val get_left : equation -> Term.term

val get_right : equation -> Term.term

val make_unchecked_equation : Term.term -> Term.term -> Term.term -> equation

val make_checked_equation : Term.term -> Term.term -> Term.term -> equation

val make_oriented_equation : Term.term -> Term.term -> Term.term -> equation

val set_is_oriented : equation -> bool -> unit
	
val set_left : equation -> Term.term -> unit
	
val set_right : equation -> Term.term -> unit

val set_is_ground : equation -> unit

val get_is_ground : equation -> bool

val copy_equation : equation -> equation

type place =
	ArrayPlace of (Term.term array * int)
	| EqLeftPlace of equation
	| EqRightPlace of equation

val get_place : place -> Term.term

val set_place : place -> Term.term -> unit

val make_array_place : Term.term -> place

val null_place : place
	
(** [iter_nonvar_subterms_place f term place] applies [f] to all subterms of [term], including [term]
itself. *)
val iter_nonvar_subterms_place : (Term.term -> place -> unit) -> Term.term -> place -> unit

(** Like [iter_nonvar_subterms_place], but does not apply [f] to [term] itself, only to proper subterms
of [term].*)
val iter_nonvar_subterms_place_no_top : (Term.term -> place -> unit) -> Term.term -> place -> unit

val iter_nonvar_subterms : (Term.term -> place -> unit) -> Term.term -> unit



(**[sup_right l r s t] attempts to perform a superposition right inference from one equation l=r into
another equation s=t. Specifically it is attempted to substitute an occurrence of l in s with r.
The various required conditions are checked.
The function returns a tuple.
If no substitution is possible, then the tuple consists of s=t and an empty context.
Otherwise, the tuple consists of the rewritten version of s=t and the unifier of l and s. If s=t
is part of a clause, this context can be applied to the remaining literals of the clause.

t can be of the type [Term.Unbound], this allows to perform a superposition inference into a
non-equation term. **)  
val sup_right : Term.term -> Term.term -> Term.term -> Term.term -> (Term.term * Term.term * Term.context) 

val sup_left : Term.term -> Term.term -> Term.term -> Term.term -> Term.term -> bool -> (Term.term * Term.term * Term.context) 

(* 
The folowing functions each return a list of tuples.
If no substitution is possible, then the list is empty.
Otherwise, each tuple represents the result of a successful substitution:
The first value is the rewritten version of s=t,
the second value is a context, which represents the unifier of l and s. If s=t
is part of a clause, this context can be applied to the remaining literals of the clause. *)
val unit_sup_right_lr : Term.term -> Term.term -> Term.term -> Term.symbol -> (Term.term  * Term.context) list

val unit_sup_right_lr_into_term : Term.term -> Term.term -> Term.term -> (Term.term * Term.context) list

val unit_sup_right_lr_into_eq : Term.term -> Term.term -> Term.term -> Term.term -> Term.symbol -> (Term.term * Term.context) list

val unit_sup_right_lr_into_eq_place : Term.term -> Term.term -> Term.term -> place -> Term.symbol -> bool -> bool -> (Term.term * Term.context)

val unit_sup_right_lr_into_term_place : Term.term -> Term.term -> Term.term -> place -> Term.symbol -> bool -> (Term.term * Term.context)

val sup_left_lr_into_term : Term.term -> Term.term -> Term.term -> Term.term -> bool -> (Term.term * Term.context) list

val sup_left_lr_into_eq : Term.term -> Term.term -> Term.term -> Term.term -> Term.symbol ->  Term.term -> bool -> (Term.term * Term.context) list

val sup_left_lr_into_term_place : Term.term -> Term.term -> Term.term -> place -> Term.symbol -> Term.term -> bool -> (Term.term * Term.context)

val sup_left_lr_into_eq_place : Term.term -> Term.term -> Term.term -> place -> Term.symbol -> Term.term -> bool -> bool -> (Term.term * Term.context)


(**[term_to_equation term] converts a term into an [equation]. Caution: there are no
checks regarding term; it is assumed that its main functor is the equality symbol and
that it has two argument subterms. The function will work with any two-argument term. *)
val term_to_equation : Term.term -> equation