(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Arithmetic evaluation. *)

(** Arithmetic operations correspond to those of ISO-Prolog.
   (For now only a few exemplary operations are implemented.)
 *)

(** Initialize the association of between evaluable functions and
    symbols for the given symbol_table. This function has to be called
    on a symbol table, before arithmetic evaluation can be used.
 *)
val init_evaluables : Term.symbol_table -> unit

(** Remove the association of between evaluable functions and
    symbols for the given symbol_table.
*)
val clear_evaluables : Term.symbol_table -> unit

(** [eval term context] returns the result of evaluating the given
   term in the given context. 
 *)
val eval : Term.term -> Term.context -> Term.term

(** {6 Extending the System with Evaluable Functions} *)

(** [register_evaluable symbol_table name arity implementation] effects that
   the function [name]/[arity] becomes available as evaluable function.
   See module term for more information. *)
val register_evaluable : Term.symbol_table -> string -> int -> Term.f -> unit
