(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)
(*
let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let sp = print_string " "
let ps s = print_string s
let pi i = print_int i
*)

type fact =
    InputFact of Term.term * bool * fact ref
  | DerivedFact of Term.term * bool * fact ref 	
  | DerivedFactWithProof of Term.term * proof * bool * fact ref
  | InputEquationFact of Equality.equation * bool * bool * fact ref
  | DerivedEquationFact of Equality.equation * bool * bool * fact ref
  | DerivedEquationFactWithProof of Equality.equation * bool * proof * bool * fact ref  
  | FactWithUnknownProof
  | NullFact
  
      
and proof =
    BuiltinProof
  | InputFactProof
  | InputClauseProof
  | Proof of Term.term * fact array
  | DisjProof of int * Term.term * fact array * fact list
  | NegProof of Term.term * fact array * fact
  | UnknownProof
  | UnitSupRightProof of Term.term * Term.term * fact * fact
  | NegUnitSupRightProof of Term.term * Term.term * fact * fact * fact
  | SupLeftProof of Term.term * Term.term * fact * fact
  | EqRefProof of Term.term * fact
  | HyperPureProof  of Term.term * fact array
	
  
let null_fact = NullFact

let fact_with_unknown_proof = FactWithUnknownProof  

  
let get_term fact = 
  match fact with
    InputFact (t, _, _) -> t
  | DerivedFact (t, _, _) -> t
  | InputEquationFact (eq, _, _, _) -> Equality.get_term eq
  | DerivedEquationFact (eq, _, _, _) -> Equality.get_term eq
  | DerivedFactWithProof (t, _, _, _) -> t
  | DerivedEquationFactWithProof (eq, _, _, _, _) -> Equality.get_term eq  
  | FactWithUnknownProof -> Term.Unbound
  | NullFact -> Term.Unbound

let get_equation fact =
	match fact with
		InputEquationFact (eq, _, _, _) -> eq
		|DerivedEquationFact (eq, _, _, _) -> eq
		|DerivedEquationFactWithProof (eq, _, _, _, _) -> eq
		|_ -> (raise (Err.Error ("Trying to get equation from non-equation fact (Fact.get_equation)."))) 
  
let get_proof fact =
  match fact with
    InputFact _ -> InputFactProof
  | InputEquationFact _ -> InputFactProof  
  | DerivedFact _ -> BuiltinProof
  | DerivedEquationFact _ -> BuiltinProof
  | DerivedFactWithProof (_, p, _, _) -> p
  | DerivedEquationFactWithProof (_, _, p, _, _) -> p  
  | FactWithUnknownProof -> UnknownProof
  | NullFact -> BuiltinProof
  
  
(*
let rec print_proof p =
	match p with
  | InputFactProof -> ps "InputFactProof"
  | InputClauseProof -> ps "InputClauseProof"
  | Proof (t, f) -> ps "Proof: ";ppp t
  | DisjProof (i, t, fa, fl) -> ps "DisjProof: ";pi i;ps " ";ppp t;ps " | ";Array.iter (fun fae -> print_proof (get_proof fae)) fa
	|_ -> failwith "Bad proof for this function!"
    | NegProof of Term.term * fact array * fact
  | UnknownProof
  | UnitSupRightProof of Term.term * Term.term * fact * fact
  | NegUnitSupRightProof of Term.term * Term.term * fact * fact * fact
  | SupLeftProof of Term.term * Term.term * fact * fact
  | EqRefProof of Term.term * fact
  | HyperPureProof  of Term.term * fact array	  
*)  
    
let is_redundant fact =
	match fact with
    InputFact (_, _, b) -> not(!b == null_fact)
  | DerivedFact (_, _, b) -> not(!b == null_fact)
  | InputEquationFact (_, _, _, b) -> not(!b == null_fact)
  | DerivedEquationFact (_, _, _, b) -> not(!b == null_fact)
  | DerivedFactWithProof (_, _, _, b) -> not(!b == null_fact)
  | DerivedEquationFactWithProof (_, _, _, _, b) -> not(!b == null_fact)  
  | FactWithUnknownProof -> false
  | NullFact -> false
  
let get_reducer fact =
	match fact with
    InputFact (_, _, b) -> !b
  | DerivedFact (_, _, b) -> !b
  | InputEquationFact (_, _, _, b) -> !b
  | DerivedEquationFact (_, _, _, b) -> !b
  | DerivedFactWithProof (_, _, _, b) -> !b
  | DerivedEquationFactWithProof (_, _, _, _, b) -> !b  
  | FactWithUnknownProof -> null_fact
  | NullFact -> null_fact  
  
let is_ground fact =
	match fact with
    InputFact (_, g, _) -> g
  | DerivedFact (_, g, _) -> g
  | InputEquationFact (_, _, g, _) -> g
  | DerivedEquationFact (_, _, g, _) -> g
  | DerivedFactWithProof (_, _, g, _) -> g
  | DerivedEquationFactWithProof (_, _, _, g, _) -> g  
  | FactWithUnknownProof -> false
  | NullFact -> false  
  
let declare_redundant fact = 
	match fact with
    InputFact (_, _, b) -> b := fact_with_unknown_proof
  | DerivedFact (_, _, b) -> b := fact_with_unknown_proof
  | InputEquationFact (_, _, _, b) -> b := fact_with_unknown_proof
  | DerivedEquationFact (_, _, _, b) -> b := fact_with_unknown_proof
  | DerivedFactWithProof (_, _, _, b) -> b := fact_with_unknown_proof
  | DerivedEquationFactWithProof (_, _, _, _, b) -> b := fact_with_unknown_proof  
  | FactWithUnknownProof -> ()
  | NullFact -> ()
  
let declare_redundant_by_fact fact reduced_fact = 
	match fact with
    InputFact (_, _, b) -> b := reduced_fact
  | DerivedFact (_, _, b) -> b := reduced_fact
  | InputEquationFact (_, _, _, b) -> b := reduced_fact
  | DerivedEquationFact (_, _, _, b) -> b := reduced_fact
  | DerivedFactWithProof (_, _, _, b) -> b := reduced_fact
  | DerivedEquationFactWithProof (_, _, _, _, b) -> b := reduced_fact  
  | FactWithUnknownProof -> ()
  | NullFact -> ()  
  
let undeclare_redundant fact =	
	match fact with
    InputFact (_, _, b) -> b := null_fact
  | DerivedFact (_, _, b) -> b := null_fact
  | InputEquationFact (_, _, _, b) -> b := null_fact
  | DerivedEquationFact (_, _, _, b) -> b := null_fact
  | DerivedFactWithProof (_, _, _, b) -> b := null_fact
  | DerivedEquationFactWithProof (_, _, _, _, b) -> b := null_fact  
  | FactWithUnknownProof -> ()
  | NullFact -> ()  
  
type entry = fact      


    
let entry_print fact =
  PrologPrint.print (get_term fact)

let make_input_fact term = 
	let b = ref null_fact
	and g = Termutils.is_ground term
	in
		InputFact (term, g, b)

let make_input_equation_fact eq left_side_indexed =
	let b = ref null_fact
	and g = Equality.get_is_ground eq	
	in
		InputEquationFact (eq, left_side_indexed, g, b)

let make_subproofs n =
  Array.make n NullFact

let make_derived_fact term =
	let b = ref null_fact
	and g = Termutils.is_ground term	
	in
  		DerivedFact (term, g, b)

let make_derived_equation_fact eq left_side_indexed =
	let b = ref null_fact
	and g = Equality.get_is_ground eq	
	in
		DerivedEquationFact (eq, left_side_indexed, g, b)  
  
let make_derived_equation_fact_with_proof eq left_side_indexed proof =
	let b = ref null_fact
	and g = Equality.get_is_ground eq	
	in
		DerivedEquationFactWithProof (eq, left_side_indexed, proof, g, b)  		
		
let make_derived_fact_with_proof term proof =
	let b = ref null_fact
	and g = Termutils.is_ground term	
	in
  		DerivedFactWithProof (term, proof, g, b)
  
let make_fact_like fact term =
	let b = ref null_fact
	and g = Termutils.is_ground term	
	in
  match fact with
    InputFact _ -> InputFact (term, g, b)
  | InputEquationFact (_, left_side_indexed, _, _) -> 
  		InputEquationFact ((Equality.term_to_equation term), left_side_indexed, g, b)
  | DerivedFact _ -> DerivedFact (term, g, b)
  | DerivedEquationFact (_, left_side_indexed, _, _) -> 
  		DerivedEquationFact ((Equality.term_to_equation term), left_side_indexed, g, b)
  | DerivedEquationFactWithProof (_, left_side_indexed, proof, _, _) -> 
  		DerivedEquationFactWithProof ((Equality.term_to_equation term), left_side_indexed, proof, g, b)
  | DerivedFactWithProof (_, proof, _, _) -> DerivedFactWithProof (term, proof, g, b)
  | FactWithUnknownProof -> FactWithUnknownProof
  | NullFact -> NullFact

let is_equation fact =
	match fact with
		InputEquationFact _ -> true
		|DerivedEquationFact _ -> true
		|DerivedEquationFactWithProof _ -> true
		|_ -> false
	    
(****************************************************************************)

let make_ref_proofterm st nodenum =
  Convert.array_to_compound
    st
    "ref"
    [|Convert.string_to_term st ("node"^string_of_int nodenum)|]

let make_disj_ref_proofterm st nodenum =
  Convert.array_to_compound
    st
    "ref"
    [|Convert.string_to_term st ("disj"^string_of_int nodenum)|]
  
let make_input_fact_proofterm st nodenum term =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.string_to_term st "input_fact";
      Convert.string_to_term st "[]";
      Convert.string_to_term st "[]"|]

let make_input_clause_proofterm st nodenum term =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.string_to_term st "input_clause";
      Convert.string_to_term st "[]";
      Convert.string_to_term st "[]"|]      
      
let make_builtin_fact_proofterm st nodenum term =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.string_to_term st "builtin";
      Convert.string_to_term st "[]";
      Convert.string_to_term st "[]"|]

let make_unknown_fact_proofterm st nodenum term =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.string_to_term st "unknown";
      Convert.string_to_term st "[]";
      Convert.string_to_term st "[]"|]
    
let make_hyper_proofterm st nodenum term clause subproofs neg_subproofs =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.array_to_compound st "hyper" [|clause|];
      Convert.list_of_term_to_term st (Array.to_list subproofs);
      Convert.list_of_term_to_term st neg_subproofs|]
      
let make_hyper_pure_proofterm st nodenum term clause subproofs neg_subproofs =
  Convert.array_to_compound
    st
    "node"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      Convert.array_to_compound st "hyper-purify" [|clause|];
      Convert.list_of_term_to_term st (Array.to_list subproofs);
      Convert.list_of_term_to_term st neg_subproofs|]
      
let make_unit_sup_right_proofterm st nodenum term eq_clause into_clause into_subproof eq_subproof neg_subproofs =
	Convert.array_to_compound
		st
		"node"
		[|Convert.string_to_term st ("node"^string_of_int nodenum);
		  term;
		  Convert.array_to_compound st "unit-sup-right" [|into_clause; eq_clause|];       
    	  Convert.list_of_term_to_term st (Array.to_list [|into_subproof; eq_subproof|]);
      	  Convert.list_of_term_to_term st neg_subproofs|]
    	  
let make_sup_left_proofterm st nodenum term eq_clause into_clause into_subproof eq_subproof =
	Convert.array_to_compound
		st
		"node"
		[|Convert.string_to_term st ("node"^string_of_int nodenum);
		  term;
		  Convert.array_to_compound st "sup-left" [|into_clause; eq_clause|];       
    	  Convert.list_of_term_to_term st (Array.to_list [|into_subproof; eq_subproof|]);
      	  Convert.string_to_term st "[]"|]	  
  
    	  
let make_eq_ref_proofterm st nodenum term clause subproof  =
	Convert.array_to_compound
		st
		"node"
		[|Convert.string_to_term st ("node"^string_of_int nodenum);
		  term;
		  Convert.array_to_compound st "ref" [|clause|];       
    	  Convert.list_of_term_to_term st [subproof];
      	  Convert.string_to_term st "[]"|]    	  
    	  
let make_disj_proofterm st id clause subproofs neg_subproofs =
  Convert.array_to_compound
    st
    "split"
    [|Convert.string_to_term st ("disj"^string_of_int id);
      Convert.array_to_compound st "hyper" [|clause|];
      Convert.list_of_term_to_term st (Array.to_list subproofs);
      Convert.list_of_term_to_term st neg_subproofs|]

let make_branch_proofterm st nodenum term subproof =
  Convert.array_to_compound
    st
    "branch"
    [|Convert.string_to_term st ("node"^string_of_int nodenum);
      term;
      subproof|]

(* hmmm - what should be in the table is the "proof_id"
   - shared for diff goals in case of disjunction!
   - issues: space & explicitly different handling for input facts
   - perhaps just handle disjunctions specially
   - units: identify by their fact (including particular ancestors)
   - disjunction: identify by the proof term
*)

module Table = Hashtbl.Make(
  struct
    type t = fact
    let equal = (=) (* ### Fact-identity: = (and not ==) *)
    let hash = Hashtbl.hash
  end)

module DisjTable = Hashtbl.Make(
  struct
    type t = int
    let equal = (=)
    let hash = Hashtbl.hash
  end)

type sign = Pos | Neg

let negate st term =
  Convert.array_to_compound st "~" [|term|]

let rec proof_to_term_1 st sign table dtable node fact =
  try 
    let refnum = Table.find table fact in 
    make_ref_proofterm st refnum
  with Not_found ->
    node := !node + 1;
    Table.add table fact !node;
    proof_to_term_2 st table dtable node 
      (match sign with Pos -> get_term fact | Neg -> negate st (get_term fact))
      (get_proof fact)

and proof_to_term_2 st table dtable node goal_term proof =
  match proof with
    BuiltinProof ->
      make_builtin_fact_proofterm st !node goal_term
  | UnknownProof ->
      make_unknown_fact_proofterm st !node goal_term
  | InputFactProof ->
      make_input_fact_proofterm st !node goal_term
  | InputClauseProof ->
  	  make_input_clause_proofterm st !node goal_term    
  | Proof (clause, subfacts) ->
      let nodeval = !node in
      let subproofs =
	Array.map
	  (fun subfact -> proof_to_term_1 st Pos table dtable node subfact)
	  subfacts in
      make_hyper_proofterm st nodeval goal_term clause subproofs []
  | NegProof (clause, subfacts, negfact) ->
      let nodeval = !node in
      let subproofs =
	Array.map
	  (fun subfact -> proof_to_term_1 st Pos table dtable node subfact)
	  subfacts in
      let negproof = proof_to_term_1 st Neg table dtable node negfact in
      make_hyper_proofterm st nodeval goal_term clause subproofs [negproof]
  | NegUnitSupRightProof (into_term, eq_term, into_subfact, eq_subfact, negfact) ->	  
	  let nodeval = !node
	  and negproof = proof_to_term_1 st Neg table dtable node negfact
	  and into_proofterm = proof_to_term_1 st Pos table dtable node into_subfact
	  and eq_proofterm = proof_to_term_1 st Pos table dtable node eq_subfact	  
	  in
      	make_unit_sup_right_proofterm st nodeval goal_term into_term eq_term into_proofterm eq_proofterm [negproof]
  | DisjProof (id, clause, subfacts, negfacts) -> 
      let nodeval = !node in
      make_branch_proofterm
	st 
	nodeval
	goal_term
	(try 
	  let dnum = DisjTable.find dtable id in
	  make_disj_ref_proofterm st dnum
	with Not_found ->
	  DisjTable.add dtable id id;
	  let subproofs =
	    Array.map
	      (fun subfact -> proof_to_term_1 st Pos table dtable node subfact)
	      subfacts in
	  let negproofs =
	    List.map
	      (fun negfact -> proof_to_term_1 st Neg table dtable node negfact)
	      negfacts in
	  make_disj_proofterm st id clause subproofs negproofs)
  | UnitSupRightProof (into_term, eq_term, into_subfact, eq_subfact) ->	  
	  let nodeval = !node
	  and into_proofterm = proof_to_term_1 st Pos table dtable node into_subfact
	  and eq_proofterm = proof_to_term_1 st Pos table dtable node eq_subfact	  
	  in
      	make_unit_sup_right_proofterm st nodeval goal_term into_term eq_term into_proofterm eq_proofterm []
  | SupLeftProof (into_term, eq_term, into_subfact, eq_subfact) ->	  
	  let nodeval = !node
	  and into_proofterm = proof_to_term_1 st Pos table dtable node into_subfact
	  and eq_proofterm = proof_to_term_1 st Pos table dtable node eq_subfact	  
	  in
      	make_sup_left_proofterm st nodeval goal_term into_term eq_term into_proofterm eq_proofterm     
  | EqRefProof (term, subfact) ->	  
	  let nodeval = !node
	  and proofterm = proof_to_term_1 st Pos table dtable node subfact	  
	  in
      	make_eq_ref_proofterm st nodeval goal_term term proofterm  
  | HyperPureProof (clause, subfacts) ->
      let nodeval = !node in
      let subproofs =
	Array.map
	  (fun subfact -> proof_to_term_1 st Pos table dtable node subfact)
	  subfacts in
      make_hyper_pure_proofterm st nodeval goal_term clause subproofs []          
      
      
let proof_term_for_fact symbol_table fact =
  let table = Table.create 64 
  and dtable = DisjTable.create 64 
  and node = ref 0 in
  proof_to_term_1 symbol_table Pos table dtable node fact

type l_proof = proof list

type proof_term_aux =
    {table:int Table.t;
     dtable:int DisjTable.t;
     node:int ref}

let make_proof_term_aux () =
  {table=Table.create 64;
   dtable=DisjTable.create 64;
   node=ref 0}
    
let proof_term_for_l_proof symbol_table proofs =
  let aux = make_proof_term_aux () in
  let table = aux.table
  and dtable = aux.dtable
  and node = aux.node in
  match proofs with
    [proof] ->
      proof_to_term_2 
	symbol_table
	table
	dtable
	node
	(Term.Constant (Term.get_false_symbol symbol_table))
	proof
  | _ ->
      let proof_terms =
	(List.map
	   (fun proof ->
	     node := !node + 1;
	     proof_to_term_2 
	       symbol_table
	       table
	       dtable
	       node
	       (Term.Constant (Term.get_false_symbol symbol_table))
	       proof)
	   proofs) in
      Convert.array_to_compound
	symbol_table
	"node"
	[|Convert.string_to_term symbol_table "node_goal";
	  Convert.string_to_term symbol_table "goal";
	  Convert.string_to_term symbol_table "all_closed";
	  Convert.list_of_term_to_term symbol_table proof_terms;
	  Convert.list_of_term_to_term symbol_table []|]

let proof_term_for_r_proof symbol_table aux proof =
  let table = aux.table
  and dtable = aux.dtable
  and node = aux.node in
  node := !node + 1;
  proof_to_term_2 
    symbol_table
    table
    dtable
    node
    (Term.Constant (Term.get_false_symbol symbol_table))
    proof
