(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Graph operations. This module is similar to the graph library module
   common in Prolog systems.

   The P-representation of a graph is a list of (from, to) vertex
   pairs, where the pairs can be in any old order.

   The S-representation of a graph is a list of [(vertex, neighbours)]
   pairs, where the pairs are ordered according to their vertex and
   the neighbours of each vertex are also orderered. 

   Structural equality (i.e. [=]) and [Pervasives.compare] are used
   for identity and comparison of vertices. *)

(** Graphs in P-representation. *)
type 'a p_graph = ('a * 'a) list

(** Graphs in S-representation. *)
type 'a s_graph = ('a * ('a list)) list

(** Return the S-graph representation of the given P-graph. *)
val p_to_s_graph : 'a p_graph -> 'a s_graph

(** Return a P-graph representation of the given S-graph. *)
val s_to_p_graph : 'a s_graph -> 'a p_graph

(** Return the transposed P-graph. *)
val p_transpose : 'a p_graph -> 'a p_graph

(** Returns the transitive closure of the given S-graph as S-graph. *)
val warshall : 'a s_graph -> 'a s_graph

(** Returns a list of vertices of the given S-graph in topological ordering.
   Raises [Not_found] if the graph is cyclic. *)
val top_sort : 'a s_graph -> 'a list

 
