(*
* Copyright (C) 2006 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let sp = print_string " "
let ps s = print_string s
let pi i = print_int i



exception Found

(** [is_subterm term1 term2] tests whether term1 is a subterm of term2 and returns
* true or false accordingly. *)							
(*let rec is_subterm t1 t2 =
	if Term.term_ident t1 t2 then true
	else match t2 with
		Term.Compound k ->
			let no_match_yet = ref true
			and index = ref 0
			and len = Array.length (Term.compound_args k)
			in 
				while (!no_match_yet && (!index < len)) do
					if (is_subterm t1 (Term.compound_args k).(!index)) then
						no_match_yet := false;
					index := (!index + 1)
				done;
				(!no_match_yet = false)
		|_ -> false 
*)

let rec is_subterm t1 t2 =
	if (Term.term_ident t1 t2) then 
		(true)
	else 
		(match t2 with
			Term.Compound k ->
				(try
					(Array.iter
						(fun arg ->
							if (is_subterm t1 arg) then
								(raise Found))
						(Term.compound_args k);
					false)
				with Found -> true)
			|_ -> false)

			
let clip_comp number =
	if (number=0) then 0
	else
		(if (number>0) then 1
		else (-1))		
		
(** [functor_symbol_precedence_compare term1 term2] compares term1 and term2
	according to the symbol precedence order of the leading symbols (functors in case
	of compounds, names in case of constants and the terms themselves otherwise). It returns:
    -  -1 if the functor of term1 is less than the functor of term2.

    -  0 if the functor of term1 and the functor of term2 are equal.

    -  1 if the functor of term1 is greater than the functor of term2.
	Symbols are ordered by arity, then by name (for example: f/3 > a/2 > f/2 > g/2 > g/1).
*)		

let functor_symbol_precedence_compare term1 term2=
	match term1, term2 with
		Term.Float n1, Term.Float n2 ->
			Pervasives.compare n1 n2
    	| Term.Float _, _ ->
			-1
    	| _, Term.Float _ ->
			1
    	| Term.Integer n1, Term.Integer n2 ->
			Pervasives.compare n1 n2
    	| Term.Integer _, _ ->
			-1
    	| _, Term.Integer _ ->
			1
    	| Term.Constant s1, Term.Constant s2 ->
			Pervasives.compare (Term.symbol_name s1) (Term.symbol_name s2)
    	| Term.Constant s1, _ ->
			-1
    	| _, Term.Constant _ ->
			1
		| Term.Compound k1, Term.Compound k2 ->
		    let arity_comp = 
			    	(Pervasives.compare 
			    		(Term.symbol_arity (Term.compound_functor k1)) 
			    		(Term.symbol_arity (Term.compound_functor k2)))
			in
				if arity_comp>0 then 1
				else 
					(if arity_comp<0 then (-1)
					else 
						(Pervasives.compare 
							(Term.symbol_name (Term.compound_functor k1)) 
							(Term.symbol_name (Term.compound_functor k2))))
    	| Term.Variable _, _ | _, Term.Variable _->
			raise (Failure "Variable in symbol comparison in LrpOrder.functor_symbol_precedence_compare.") 
    	| Term.Unbound, _ | _, Term.Unbound ->
			raise (Failure "Unbound in symbol comparison in LrpOrder.functor_symbol_precedence_compare.")

exception Failed

							
(** [lrpo_greater term1 term2] tests whether term1 is greater than term2 in the LRPO
and returns true or false. *)	


				
let rec lrpo_greater t1 t2 =
	if Term.term_ident t1 t2 then false	(* lrpo_greater is irreflexive *)
	else
		match t1, t2 with
			Term.Unbound, _ -> false
			|_, Term.Unbound -> true
			|Term.Variable v1, _ ->
				false								(* a variable is never greater than another term *)
			|_, Term.Variable v2 ->
				if is_subterm t2 t1 then true		
				else false							(* t2 is a variable and a proper subterm of t1 *)
			|Term.Compound k1, Term.Compound k2 ->
				(try
					(Array.iter
						(fun t ->
							if (Term.term_ident t t2)||(lrpo_greater t t2) then 
								(raise Found)
							else ())
						(Term.compound_args k1);
					match (clip_comp (functor_symbol_precedence_compare t1 t2)) with
						1 ->
							(Array.iter
								(fun t ->
									if lrpo_greater t1 t then ()
									else (raise Failed))
								(Term.compound_args k2);
							true)
	 					(* t1 has greater functor than t2 and is greater than each argument of t2 *)
						|0 -> 
							(let exclusive_1 = ref []
							and exclusive_2 = ref []
							and check_1 = Array.make (Array.length (Term.compound_args k1)) true
							and check_2 = Array.make (Array.length (Term.compound_args k2)) true
							in
								Array.iteri
									(fun i1 a1 ->
										if check_1.(i1) then (
											try
												(Array.iteri
													(fun i2 a2 ->
														if Term.term_ident a1 a2 then
															(check_1.(i1)<-false;
															check_2.(i2)<-false;
															raise Failed))
													(Term.compound_args k2);
												exclusive_1 := a1 :: !exclusive_1)
											with Failed -> ()))
									(Term.compound_args k1);
								Array.iteri
									(fun i2 a2 ->
										if check_2.(i2) then
											(exclusive_2 := a2 :: !exclusive_2))
									(Term.compound_args k2);
								if ((List.length !exclusive_1) > 0) then
									(List.for_all
										(fun a2 ->
											List.exists
												(fun a1 ->
													lrpo_greater a1 a2)
												!exclusive_1)
										!exclusive_2)
								else (false))
(* t1 and t2 have the same functor and the argument list of t1 is multiset-greater than that of t2 *)
						|(-1) -> false
						|_ -> (raise (Err.Error ("Bad comparison result in LrpOrder.lrpo_greater."))))
					with Found -> true
						|Failed -> false
					)
			|Term.Compound k1, _ ->
				true									(* a compound functor has higher arity than non-compound term,
   														   thus second condition in NR01, p.13 is always met *)
   			|_, Term.Compound k2 ->
				false									(* a non-compound term is never greater than a compound *)
			|_, _ ->
				(functor_symbol_precedence_compare t1 t2)=1		(* non-compound terms are compared using the symbol precedence order *)
				
					
and lrpo_greater_shared_vars t1 t2 =
	if (Termutils.contains_vars_of t1 t2) then
		(lrpo_greater t1 t2)
	else false
	
and lrpo_greater_exp t1 t2 =
	if Term.term_ident t1 t2 then false	(* lrpo_greater is irreflexive *)
	else
		match t1, t2 with
			Term.Unbound, _ -> false
			|_, Term.Unbound -> true
			|Term.Variable v1, _ ->
				false								(* a variable is never greater than another term *)
			|_, Term.Variable v2 ->
				if is_subterm t2 t1 then true		(* t2 is a variable and a proper subterm of t1 *)
				else false
			|Term.Compound k1, Term.Compound k2 ->
				let check =
					(try
						(Array.iter
							(fun t ->
								if (Term.term_ident t t2)||(lrpo_greater_shared_vars t t2) then 
									(raise Found)
								else ())
							(Term.compound_args k1);
						false)
					with Found -> true)
				in
					if check then true	(* t1 contains subterm equal to or greater than t2 *)
					else 
						(match (clip_comp (functor_symbol_precedence_compare t1 t2)) with
							1 ->
								(try
									(Array.iter
										(fun t ->
											if lrpo_greater_shared_vars t1 t then ()
											else (raise Failed))
										(Term.compound_args k2);
									true)
								with Failed -> false)	 
									(* t1 has greater functor than t2 and is greater than each argument of t2 *)
							|0 -> 
								(let exclusive_1 = ref []
								and exclusive_2 = ref []
								and check_1 = Array.make (Array.length (Term.compound_args k1)) true
								and check_2 = Array.make (Array.length (Term.compound_args k2)) true
								in
									Array.iteri
										(fun i1 a1 ->
											if check_1.(i1) then (
												try
													(Array.iteri
														(fun i2 a2 ->
															if Term.term_ident a1 a2 then
																(check_1.(i1)<-false;
																check_2.(i2)<-false;
																raise Failed))
														(Term.compound_args k2);
													exclusive_1 := a1 :: !exclusive_1)
												with Failed -> ()))
										(Term.compound_args k1);
									Array.iteri
										(fun i2 a2 ->
											if check_2.(i2) then
												(exclusive_2 := a2 :: !exclusive_2))
										(Term.compound_args k2);
									if ((List.length !exclusive_1) > 0) then
										(List.for_all
											(fun a2 ->
												List.exists
													(fun a1 ->
														lrpo_greater a1 a2)
													!exclusive_1)
											!exclusive_2)
									else (false))
(* t1 and t2 have the same functor and the argument list of t1 is multiset-greater than that of t2 *)
							|(-1) -> false
							|_ -> (raise (Err.Error ("Bad comparison result in LrpOrder.lrpo_greater.")))
						)
			|Term.Compound k1, _ ->
				true									(* a compound functor has higher arity than non-compound term,
   														   thus second condition in NR01, p.13 is always met *)
   			|_, Term.Compound k2 ->
				false									(* a non-compound term is never greater than a compound *)
			|_, _ ->
				(functor_symbol_precedence_compare t1 t2)=1		(* non-compound terms are compared using the symbol precedence order *)
	
	
	
				
let rec lrpo_greater_unshared_vars t1 t2 =
	if Term.term_ident t1 t2 then false	(* lrpo_greater is irreflexive *)
	else
		match t1, t2 with
			Term.Unbound, _ -> false
			|_, Term.Unbound -> true
			|Term.Variable v1, _ ->
				false								(* a variable is never greater than another term *)
			|_, Term.Variable v2 ->
				false
			|Term.Compound k1, Term.Compound k2 ->
				let check =
					(try
						(Array.iter
							(fun t ->
								if (Term.term_ident t t2)||(lrpo_greater_unshared_vars t t2) then 
									(raise Found)
								else ())
							(Term.compound_args k1);
						false)
					with Found -> true)
				in
					if check then true	(* t1 contains subterm equal to or greater than t2 *)
					else 
						(match (clip_comp (functor_symbol_precedence_compare t1 t2)) with
							1 ->
								(try
									(Array.iter
										(fun t ->
											if lrpo_greater_unshared_vars t1 t then ()
											else (raise Failed))
										(Term.compound_args k2);
									true)
								with Failed -> false)	 
									(* t1 has greater functor than t2 and is greater than each argument of t2 *)
							|0 -> 
								(let exclusive_1 = ref []
								and exclusive_2 = ref []
								and check_1 = Array.make (Array.length (Term.compound_args k1)) true
								and check_2 = Array.make (Array.length (Term.compound_args k2)) true
								in
									Array.iteri
										(fun i1 a1 ->
											if check_1.(i1) then (
												try
													(Array.iteri
														(fun i2 a2 ->
															if Term.term_ident a1 a2 then
																(check_1.(i1)<-false;
																check_2.(i2)<-false;
																raise Failed))
														(Term.compound_args k2);
													exclusive_1 := a1 :: !exclusive_1)
												with Failed -> ()))
										(Term.compound_args k1);
									Array.iteri
										(fun i2 a2 ->
											if check_2.(i2) then
												(exclusive_2 := a2 :: !exclusive_2))
										(Term.compound_args k2);
									if ((List.length !exclusive_1) > 0) then
										(List.for_all
											(fun a2 ->
												List.exists
													(fun a1 ->
														lrpo_greater a1 a2)
													!exclusive_1)
											!exclusive_2)
									else (false))
(* t1 and t2 have the same functor and the argument list of t1 is multiset-greater than that of t2 *)
							|(-1) -> false
							|_ -> (raise (Err.Error ("Bad comparison result in LrpOrder.lrpo_greater_unshared_vars.")))							
						)
			|Term.Compound k1, _ ->
				true									(* a compound functor has higher arity than non-compound term,
   														   thus second condition in NR01, p.13 is always met *)
   			|_, Term.Compound k2 ->
				false									(* a non-compound term is never greater than a compound *)
			|_, _ ->
				(functor_symbol_precedence_compare t1 t2)=1		(* non-compound terms are compared using the symbol precedence order *)
					
				
let ground_compare_lrpo_greater t1 t2 =
	if Term.term_ident t1 t2 then 0
	else
		if lrpo_greater t1 t2 then 1
		else -1
(*			if lrpo_greater t2 t1 then -1
			else raise (Err.Error ("Error in LrpOrder-ground_compare_lrpo_greater - non-ground input?"))*)

(*******************************************************************************
variant treating variables as lower-class constants, comparing them by variable
numbers - useful for clause-internal literal sorting
*******************************************************************************)
			
let ground_like_functor_symbol_precedence_compare term1 term2=
	match term1, term2 with
	    Term.Variable v1, Term.Variable v2 -> 
	    	Pervasives.compare (Term.variable_number v1) (Term.variable_number v2)
		| Term.Variable _, _ -> -1
		| _, Term.Variable _-> 1 
		| Term.Float n1, Term.Float n2 ->
			Pervasives.compare n1 n2
    	| Term.Float _, _ ->
			-1
    	| _, Term.Float _ ->
			1
    	| Term.Integer n1, Term.Integer n2 ->
			Pervasives.compare n1 n2
    	| Term.Integer _, _ ->
			-1
    	| _, Term.Integer _ ->
			1
    	| Term.Constant s1, Term.Constant s2 ->
			Pervasives.compare (Term.symbol_name s1) (Term.symbol_name s2)
    	| Term.Constant s1, _ ->
			-1
    	| _, Term.Constant _ ->
			1
		| Term.Compound k1, Term.Compound k2 ->
		    let arity_comp=
		    	Pervasives.compare 
		    		(Term.symbol_arity (Term.compound_functor k1)) 
		    		(Term.symbol_arity (Term.compound_functor k2))
			in
				if arity_comp>0 then 1
				else
					(if (arity_comp<0) then (-1)
					else
			 			(Pervasives.compare 
			 				(Term.symbol_name (Term.compound_functor k1)) 
			 				(Term.symbol_name (Term.compound_functor k2))))
    	| Term.Unbound, _ | _, Term.Unbound ->
			raise (Failure "Unbound in symbol comparison.")			
			
			
let rec ground_like_lrpo_greater t1 t2 =
	if Term.term_ident t1 t2 then false	
	else
		match t1, t2 with
			Term.Unbound, _ -> false
			|Term.Compound k1, Term.Compound k2 ->
				(try
					(Array.iter
						(fun t ->
							if (Term.term_ident t t2)||(ground_like_lrpo_greater t t2) then 
								(raise Found)
							else ())
						(Term.compound_args k1);
					match (clip_comp (functor_symbol_precedence_compare t1 t2)) with
						1 ->
							(try
								(Array.iter
									(fun t ->
										if ground_like_lrpo_greater t1 t then ()
										else (raise Failed))
									(Term.compound_args k2);
								true)
							with Failed -> false)	 
								(* t1 has greater functor than t2 and is greater than each argument of t2 *)
						|0 -> 
							(try
								(Array.iteri
									(fun i k1_arg ->
										let k2_arg = (Term.compound_args k2).(i)
										in
											if (Term.term_ident k1_arg k2_arg) then
												()
											else
												(if (ground_like_lrpo_greater k1_arg k2_arg) then
													(raise Found)
												else
													(raise Failed)))
									(Term.compound_args k1);
								false)
							with 
								 Found -> true
								|Failed -> false)
						|(-1) -> (false)
						|_ -> (raise (Err.Error ("Bad comparison result in LrpOrder.ground_like_lrpo_greater."))))
				with Found -> true
					|Failed -> false
				)
			|Term.Compound k1, _ ->
				true									(* a compound functor has higher arity than non-compound term,
   														   thus second condition in NR01, p.13 is always met *)
   			|_, Term.Compound k2 ->
				false									(* a non-compound term is never greater than a compound *)						
			|_, _ ->
				(clip_comp (ground_like_functor_symbol_precedence_compare t1 t2))=1		(* non-compound terms are compared using the symbol precedence order *)




					
let ground_like_compare_lrpo_greater t1 t2 =							
	if Term.term_ident t1 t2 then 0
	else
		if ground_like_lrpo_greater t1 t2 then 1
		else
			if ground_like_lrpo_greater t2 t1 then -1
			else 
				(pe " ";ppp t1;ps " ";ppp t2;pe ".";
				raise (Err.Error ("Error in LrpOrder.ground_like_compare_lrpo_greater")))



(** [lrpo_greater_c term1 context1 term2 context2] works the same as [lrpo_greater] except for
taking non-dereferenced terms. *)	
	
let rec lrpo_greater_c t1 c1 t2 c2 =
	let t1, c1 = Term.dereference t1 c1
	and t2, c2 = Term.dereference t2 c2
	in
		match t1, t2 with
			Term.Variable v1, _ ->  let vn1 = Term.variable_number v1 in
									let t1sub = (Term.context_terms c1).(vn1) in
										if t1sub = Term.Unbound then false
										else lrpo_greater_c t1sub (Term.context_contexts c1).(vn1) t2 c2
			|_, Term.Variable v2 -> let vn2 = Term.variable_number v2 in
									let t2sub = (Term.context_terms c2).(vn2) in
										if t2sub = Term.Unbound then not (Term.occur_check vn2 c2 t1 c1)
										else lrpo_greater_c t1 c1 t2sub (Term.context_contexts c2).(vn2)
			|Term.Compound k1, Term.Compound k2 ->
				let check=ref false
				and index=ref 0
				and len1=Array.length (Term.compound_args k1)
				in
					while not !check && !index<len1 do
						if (lrpo_greater_c (Term.compound_args k1).(!index) c1 t2 c2)||(Term.term_non_ident (Term.compound_args k1).(!index) c1 t2 c2) then
							check:=true;
						index:=!index+1
					done;
					if !check then true	(* t1 contains subterm equal to or greater than t2 *)
					else let len2=Array.length (Term.compound_args k2)
						in
						(match (functor_symbol_precedence_compare t1 t2) with
							1 -> index:=0;
								while not !check && !index<len2 do
									if not (lrpo_greater_c t1 c1 (Term.compound_args k2).(!index) c2) then
										check:=true;
									index:=!index+1
								done;
								not !check	(* t1 has greater functor than t2 and is greater than each argument of t2 *)
							|0 -> index:=0;
								let fail=ref false
								and safe=ref false
								in
								while not !fail && not !check && !index<len2 do
									if not (lrpo_greater_c t1 c1 (Term.compound_args k2).(!index) c2) then
										check:=true;
									if not !safe && not (Term.term_non_ident (Term.compound_args k1).(!index) c1 (Term.compound_args k2).(!index) c2) then
										if (lrpo_greater_c (Term.compound_args k1).(!index) c1 (Term.compound_args k2).(!index) c2) then
											safe:=true
										else fail:=true;
									index:=!index+1
								done;
								!safe && not !check && not !fail (* t1 and t2 have the same functor and the argument list of t1 is lexicographically greater than that of t2 *)
							|_ -> false
						)
			|Term.Compound k1, _ ->
				true									(* a compound functor has higher arity than non-compound term,
   														   thus second condition in NR01, p.13 is always met *)
   			|_, Term.Compound k2 ->
				false									(* a non-compound term is never greater than a compound *)
			|_, _ ->
				(functor_symbol_precedence_compare t1 t2)=1		(* non-compound terms are compared using the symbol precedence order *)
	
				
				
				
				
				
				
let unit_lrpo_greater t1 t2 equals =
	(not (Term.term_ident t1 t2)) && (
		if equals == (Convert.term_functor_symbol t1) then
			let (l1, r1) = Convert.term_pair t1
			in
			if equals == (Convert.term_functor_symbol t2) then
				let (l2, r2) = Convert.term_pair t2
				in
					let l1eql2 = Term.term_ident l1 l2
					and r1eql2 = Term.term_ident r1 l2
					and l1eqr2 = Term.term_ident l1 r2
					and r1eqr2 = Term.term_ident r1 r2
					in
					let l2int1 = l1eql2 || r1eql2
					and r2int1 = l1eqr2 || r1eqr2
					and l1int2 = l1eql2 || l1eqr2
					and r1int2 = r1eql2 || r1eqr2
					in
						if l2int1 && r2int1 then false
						else
					let l1gl2 = (not l1eql2) && (lrpo_greater l1 l2)
					and r1gl2 = (not r1eql2) && (lrpo_greater r1 l2)
					and l1gr2 = (not l1eqr2) && (lrpo_greater l1 r2)
					and r1gr2 = (not r1eqr2) && (lrpo_greater r1 r2)
					in
					let t1gl2 = ((not l1int2) && l1gl2) || ((not r1int2) && r1gl2)
					and t1gr2 = ((not l1int2) && l1gr2) || ((not r1int2) && r1gr2)
					in
					let l2ok = l2int1 || t1gl2
					and r2ok = r2int1 || t1gr2
					in
					l2ok && r2ok && ((not l2int1) || (not r2int1))
			else
				let l1eqt2 = Term.term_ident l1 t2
				and r1eqt2 = Term.term_ident r1 t2
				in
					if l1eqt2 || r1eqt2 then false
					else
				let l1gt2 = (not l1eqt2) && (lrpo_greater l1 t2)
				and r1gt2 = (not r1eqt2) && (lrpo_greater r1 t2)
				in
				l1gt2 || r1gt2
		else
			if equals == (Convert.term_functor_symbol t2) then
				let (l2, r2) = Convert.term_pair t2
				in
					let t1eql2 = Term.term_ident t1 l2
					and t1eqr2 = Term.term_ident t1 r2
					in
						if t1eql2 && t1eqr2 then false
						else
					let t1gl2 = (not t1eql2) && (lrpo_greater t1 l2)
					and t1gr2 = (not t1eqr2) && (lrpo_greater t1 r2)
					in
					let l2ok = t1eql2 || (t1gl2 && (not t1eqr2))
					and r2ok = t1eqr2 || (t1gr2 && (not t1eql2))
					in
					l2ok && r2ok && ((not t1eql2) || (not t1eqr2))
			else
				lrpo_greater t1 t2
	)
	
	
let eq_lrpo_greater l1 r1 l2 r2 =
	let l1eql2 = Term.term_ident l1 l2
	and r1eql2 = Term.term_ident r1 l2
	and l1eqr2 = Term.term_ident l1 r2
	and r1eqr2 = Term.term_ident r1 r2
	in
	let l2int1 = l1eql2 || r1eql2
	and r2int1 = l1eqr2 || r1eqr2
	and l1int2 = l1eql2 || l1eqr2
	and r1int2 = r1eql2 || r1eqr2
	in
		if l2int1 && r2int1 then false
		else
	let l1gl2 = (not l1eql2) && (lrpo_greater l1 l2)
	and r1gl2 = (not r1eql2) && (lrpo_greater r1 l2)
	and l1gr2 = (not l1eqr2) && (lrpo_greater l1 r2)
	and r1gr2 = (not r1eqr2) && (lrpo_greater r1 r2)
	in
	let t1gl2 = ((not l1int2) && l1gl2) || ((not r1int2) && r1gl2)
	and t1gr2 = ((not l1int2) && l1gr2) || ((not r1int2) && r1gr2)
	in
	let l2ok = l2int1 || t1gl2
	and r2ok = r2int1 || t1gr2
	in
	l2ok && r2ok && ((not l2int1) || (not r2int1))