(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Messages to the user. *)

(** [msg verbosity message] outputs the given string as message to
   stderr, trailed by a newline and flush, if the current verbosity
   setting is >= [verbosity]. *)
val msg : Cfg.verbosity -> string -> unit

(** Variant of [msg] that takes a format string with a single argument.

  Note: There seems no way in pure OCaml to use the curried formatting on
  arbitrary arguments such that:
  - Flushing is performed after printing (this would work with
    the development version of kprintd (Nov 2002)).
  - The print string is only constructed when required by [verbosity].
*)
val msg1 : Cfg.verbosity -> ('a -> 'b, out_channel, unit) format -> 'a -> unit

(** Variant of [msg] that takes a format string with two arguments.
*)
val msg2 : Cfg.verbosity -> 
  ('a -> 'b -> 'c, out_channel, unit) format -> 'a -> 'b -> unit

(** Variant of [msg] that takes a format string with three arguments.
*)
val msg3 : Cfg.verbosity ->
  ('a -> 'b -> 'c -> 'd, out_channel, unit) format -> 'a -> 'b -> 'c -> unit

(** Variant of [msg] that takes a format string with four arguments.
*)
val msg4 : Cfg.verbosity ->
  ('a -> 'b -> 'c -> 'd -> 'e, out_channel, unit) format ->
    'a -> 'b -> 'c -> 'd -> unit

(** Variant of [msg] that takes a format string with five arguments.
*)
val msg5 : Cfg.verbosity ->
  ('a -> 'b -> 'c -> 'd -> 'e -> 'f, out_channel, unit) format ->
    'a -> 'b -> 'c -> 'd -> 'e -> unit

(** Variant of [msg] that takes a format string with eight arguments.
*)
val msg8 : Cfg.verbosity ->
  ('a -> 'b -> 'c -> 'd -> 'e -> 'f -> 'g -> 'h -> 'i, out_channel, unit)
    format -> 'a -> 'b -> 'c -> 'd -> 'e -> 'f -> 'g -> 'h -> unit
