(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)
let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let sp = print_string " "
let ps s = print_string s
let pi i = print_int i

module FactEntry =
  struct
    type entry = Fact.fact * bool
    let get_fact entry =
    	match entry with
    		(fact, _) -> fact
    let get_term entry = 
    	match entry with
    		(fact, _) -> Fact.get_term fact
    let get_lr entry = 
    	match entry with
    		(_, lr) -> lr
  end    

module FactTable =
  LayeredIndex.Make(DtIndex.Make(FactEntry)(DtIndex.AlistTableImplementation))

type store = {
    table:FactTable.index
  }
      
let make_store () =
  { table=FactTable.make_index ()
  }

let clear store =
  Msg.msg 3 "Clearing negative units store.";
  FactTable.clear store.table

let null_store =
  make_store ()
    
let insert store fact = 
  FactTable.insert store.table (Fact.get_term fact) (fact, true);
  if (Fact.is_equation fact) then
  	(let flip_term =
  		Term.binary_flip (Fact.get_term fact)
  	in
  		FactTable.insert store.table flip_term (fact, false))

exception Found of Fact.fact

let lookup_ctx1 = Term.make_context 0
let lookup_ctx2 = Term.make_context 1
let lookup_trail = Term.make_trail ()

(*
let lookupOLD store fact =
  try
    let term = Fact.get_term fact in	
    let saved_trail = Term.save_trail lookup_trail in
    FactTable.iter_unification_candidates  
      (fun neg_fact ->	
	let trail1 =
	  Term.unify
	    term 
	    lookup_ctx1 
	    (FactEntry.get_term neg_fact)
	    lookup_ctx2
	    lookup_trail in
	if Term.failed trail1 then ()
	else
	  let term1 = (Term.apply (FactEntry.get_term neg_fact) lookup_ctx2) in
	  Term.renumber_variables_term (Array.make Term.max_vars (-1)) term1;
	  let neg_fact_1 =
	    (* BUG? Fact.make_fact_like neg_fact term1 in *)
	    Fact.make_fact_like neg_fact term1 in
	  ignore (Term.undo_bindings trail1 saved_trail);
	  raise (Found neg_fact_1)
      )
      store.table
      term
      lookup_ctx1;
      
      if (Fact.is_equation fact) then (
      let flip_term = Term.binary_flip term
      in 
          FactTable.iter_unification_candidates  
      		(fun neg_fact ->
				let trail1 =
				  Term.unify
				    flip_term 
				    lookup_ctx1 
				    (Fact.get_term neg_fact)
				    lookup_ctx2
				    lookup_trail in
	if Term.failed trail1 then
	  ()
	else
	  let term1 = (Term.apply (Fact.get_term neg_fact) lookup_ctx2) in
	  Term.renumber_variables_term (Array.make Term.max_vars (-1)) term1;
	  let neg_fact_1 =
	    (* BUG? Fact.make_fact_like neg_fact term1 in *)
	    Fact.make_fact_like neg_fact term1 in
	  ignore (Term.undo_bindings trail1 saved_trail);
	  raise (Found neg_fact_1)
      )
      store.table
      flip_term
      lookup_ctx1);
     (*print_endline "FAILURE"; *)
   raise Not_found
  with Found neg_fact ->
    neg_fact
     (* |Not_found -> (raise Not_found)*)
 *)     
     
     
     
     
(* under construction *)     
let lookup store fact =
  try
    let term = Fact.get_term fact 
    in	
    let saved_trail = Term.save_trail lookup_trail 
    and is_eq = (Fact.is_equation fact)
    in
    let flip_term =
    	if is_eq then 
    		(Term.binary_flip term)
    	else
    		(term)
    in
    FactTable.iter_unification_candidates  
      (fun neg_fact ->	
	let trail1 =
		if (FactEntry.get_lr neg_fact) then
		  (Term.unify
		    term 
		    lookup_ctx1 
		    (FactEntry.get_term neg_fact)
		    lookup_ctx2
		    lookup_trail)
	    else
	    	(if is_eq then
			  (Term.unify
			    flip_term 
			    lookup_ctx1 
			    (FactEntry.get_term neg_fact)
			    lookup_ctx2
			    lookup_trail)
			else
				(Term.fail))
	     in
	if Term.failed trail1 then
		()
	else
	  (	(* let term1 = (Term.apply (FactEntry.get_term neg_fact) lookup_ctx2) in 
	  Term.renumber_variables_term (Array.make Term.max_vars (-1)) term1; *)
	  let neg_fact_1 = (FactEntry.get_fact neg_fact)
	    (* BUG? Fact.make_fact_like neg_fact term1 in *)
	    (*Fact.make_fact_like (FactEntry.get_fact neg_fact) term1*) in
	  ignore (Term.undo_bindings trail1 saved_trail);
	  raise (Found neg_fact_1)
      ))
      store.table
      term
      lookup_ctx1;
     (*print_endline "FAILURE"; *)
   raise Not_found
  with Found neg_fact ->
    neg_fact    
      
     
let iter_all f store =
  FactTable.iter_all f store.table
    
type state = FactTable.state
let obtain_state store = FactTable.obtain_state store.table
let install_state state = FactTable.install_state state
let add_layer state = FactTable.add_layer state
