(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Store for negative unit literals. *)

(** Used to represent negative unit facts, which is required for

 - complement splitting
 - improved handling of negative input unit fact, which often are
   goals of refuation problems
*)

module FactEntry :
	sig
		type entry = Fact.fact * bool
		val get_term : entry -> Term.term
	end

type store

val make_store : unit -> store

val null_store : store
    
val insert : store -> Fact.fact -> unit

(** If the input fact is unifiable with a fact that has been previously
    inserted, that entry is returned. Otherwise Not_found is raised.
 *)
val lookup : store -> Fact.fact -> Fact.fact

(** [iter_all f store] applies [f] to all terms in the store. *)
val iter_all : (FactEntry.entry -> unit) -> store -> unit

val clear : store -> unit

type state
val obtain_state : store -> state
val install_state : state -> unit
val add_layer : state -> state



