(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


type bound =
    Free
  | Bound
      
type binding_pattern =
    bound array

type mode =
    InputGround
  | OutputGround
  | OutputFree

type mode_specifier =
    mode array

type discrimination_value =
   float

type applicability_value =
    float

type proof =
    BuiltinProof
  | UnknownProof

(** Array of discrimination values corresponding to the arguments
   of a predicate. *)
type discrimination_values =
    discrimination_value array

let binding_pattern_covers bp1 bp2 =
  try
    for i = 0  to (Array.length bp1) - 1 do
      match bp1.(i), bp2.(i) with
	Free, Bound ->
	  raise Not_found
      | _ ->
	  ();
    done;
    true;
  with Not_found ->
    false

type predicate_kind =
    Edb
  | Idb
  | BuiltIn

type indexargs =
    WholeTerm
  | Argument of int
  | Arguments of int array

type indextype =
    HashDt
  | AlistDt
  | NoIndex

type declaration =
    DiscriminationHintsDecl of discrimination_values
  | IndexDecl of indextype * indexargs
  | ModeDecl of mode_specifier

let compute_discrimination_hint binding_pattern discrimination_values =
  (* ### formula? *)
  let n = ref 1.0 in
  for i = 0 to (Array.length binding_pattern) - 1 do
    match binding_pattern.(i) with
      Bound ->
	n := !n *. (1.0 -. discrimination_values.(i))
    | Free ->
	()
  done;
  1.0 -. !n

(** Heuristic. Idea: assume arguments in lower positions are more
   discriminating. That is compatible with forward paths through
   property(object, value) predicates. It also coinides with
   discrimination tree indexing on the whole predicate term or on the
   first argument, in that bindings on lower positions are preferred.
 *)
let make_default_discrimination_hints arity =
  let a = Array.make arity 0.0 in
  for i = 0 to arity - 1 do
    a.(i) <- (1.0 /. (float(i) +. 2.0))
  done;
  a
