(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** A lexical analyzer for ISO-Prolog syntax.

   This module is similar to the Genlex module of the OCaml standard
   library, but uses ISO-Prolog instead of OCaml syntax.

   Current Restrictions:

   - Input should be ISO-Prolog which uses ISO-Latin-1 encoding.

   - Characters are classified as capitalized and small according to the
     ISO-Latin-1 encoding.

   - No operator support. Input must be produced by Prolog's
     write_canonical predicate.

   - Both /* */ and % comments are allowed. /* */ comments may not be
     nested.

   - Prolog strings are not supported.

   - Integers and floats are currently read in using their OCaml print
     representation (which may be the same as Prolog's).

   - The following escape codes are supported within quoted atoms:
     \n, \r, \t, and \DDD, where DDD is the decimal ASCII code of
     a character.

*)

(** The type of tokens. *)
type prolog_token =
    PunctuationPrologToken of string
  | VarPrologToken of string
  | ConstPrologToken of string
  | IntPrologToken of int
  | FloatPrologToken of float

(** The lexer function. *)  
val prolog_lexer : char Stream.t -> prolog_token Stream.t

         


        
