(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


module TermEntry =
  struct
    type entry = Term.term
    let get_term entry = entry
  end    

module TermTable =
  DtIndex.Make(TermEntry)(DtIndex.HashTableImplementation)

type table = TermTable.index

let make_table () =
  TermTable.make_index ()

let null_table =
  TermTable.make_index ()
    
exception Found of Term.term

let lookup_term table term =
  try
    TermTable.iter_instance_candidates
      (* We cannot directly return the found term if we use the
	 index.ml interface, since this is allowed to return a
	 superset of terms.
       *)
      (fun term1 -> 
	if Term.term_ident term1 term then
	  raise (Found term1)
	else
	  ())
      table
      term
      Term.null_context;
    raise Not_found
  with Found term1 ->
    term1

let rec share_term table term =
  match term with 
    Term.Compound k ->
      if Termutils.is_ground term then
	try
	  lookup_term table term
	with Not_found ->
	  let term1 =
	    Term.Compound
	      (Term.make_compound
		 (Term.compound_functor k)
		 (Array.map
		    (fun term2 -> share_term table term2)
		    (Term.compound_args k))) in
	  TermTable.insert table term1 term1;
	  term1
      else
	Term.Compound
	  (Term.make_compound
	     (Term.compound_functor k)
	     (Array.map
		(fun term2 -> share_term table term2)
		(Term.compound_args k)))
  | _ ->
      term
	
let share_subterms table term =
  match term with 
    Term.Compound k ->
      Term.Compound
	(Term.make_compound
	   (Term.compound_functor k)
	   (Array.map
	      (fun term2 -> share_term table term2)
	      (Term.compound_args k)))
  | _ -> 
      term
