(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


let rec term_compare t1 c1 t2 c2 =
  let t1, c1 = Term.dereference t1 c1
  and t2, c2 = Term.dereference t2 c2 in
    match t1, t2 with
      Term.Variable v1, Term.Variable v2 ->
	if c1 == c2 then
	  Pervasives.compare 
	    (Term.variable_number v1)
	    (Term.variable_number v2)
	else 
	  Pervasives.compare
	    (Term.context_multiplier c1)
	    (Term.context_multiplier c2)
    | Term.Variable v1, _ ->
	-1
    | _, Term.Variable v2 ->
	1
    | Term.Float n1, Term.Float n2 ->
	Pervasives.compare n1 n2
    | Term.Float _, _ ->
	-1
    | _, Term.Float _ ->
	1
    | Term.Integer n1, Term.Integer n2 ->
	Pervasives.compare n1 n2
    | Term.Integer _, _ ->
	-1
    | _, Term.Integer _ ->
	1
    | Term.Constant s1, Term.Constant s2 ->
	Pervasives.compare (Term.symbol_name s1) (Term.symbol_name s2)
    | Term.Constant s1, _ ->
	-1
    | _, Term.Constant _ ->
	1
    | Term.Compound k1, Term.Compound k2 ->
	let s1 = Term.compound_functor k1
	and s2 = Term.compound_functor k2 in
	let v1 = 
	  Pervasives.compare
	    (Term.symbol_arity s1)
	    (Term.symbol_arity s2) in
	if v1 = 0 then
	  let v2 = 
	    Pervasives.compare
	      (Term.symbol_name s1)
	      (Term.symbol_name s2) in
	  if v2 = 0 then
	    map_term_compare
	      (Term.compound_args k1) c1 (Term.compound_args k2) c2 0
	  else
	    v2
	else
	  v1
    | Term.Unbound, _ | _, Term.Unbound ->
	raise (Failure "Unbound in term comparison.")


and map_term_compare ts1 c1 ts2 c2 i =
  if i >= Array.length ts1 then
    0
  else
    let v = term_compare ts1.(i) c1 ts2.(i) c2 in
    if v = 0 then
      map_term_compare ts1 c1 ts2 c2 (i + 1)
    else
      v
	  


