(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Comparison of terms according to the ISO-Prolog standard total
ordering of terms. *)

(** The ISO-Prolog standard total ordering of terms orders terms as
follows (from the smallest term to the greatest):

- variables, sytem dependent

- floating point numbers, in numeric order. 

- integers, in numeric order. 

- atoms, in alphabetical order. 

- compound terms, ordered first by arity,
  then by the name of the principal functor and by the arguments in
  left-to-right order.

- A list is treated as a compound term (whose principal functor is '.'/2).
*)


(** [term_compare term1 context1 term2 context2] compares term1 in context1
    and term2 in context2 according to the ISO-Prolog standard term ordering
   and returns

    -  -1 if term1 is less than term2.

    -  0 if term1 and term2 are equal.

    -  1 if term1 is greater than term2.
   *)
val term_compare : 
   Term.term -> Term.context -> Term.term -> Term.context -> int
