(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

let pe s = print_endline s
let pl l = print_int (List.length l)
let ps s = print_string s
let sp = print_string " "

let rec iter_vars f term =
  match term with
    Term.Variable v ->
      f v
  | Term.Compound k ->
      let args = Term.compound_args k in
      for i = 0 to (Array.length args) - 1 do
	iter_vars f args.(i)
      done
  | _ ->
      ()

let get_max_varnum term =
	let max_varnum = ref (-1)
	in
	iter_vars
		(fun v ->
			if (Term.variable_number v) > !max_varnum then
				(max_varnum := (Term.variable_number v))
			else ())
		term;
	!max_varnum      
      
let is_ground term =
  try
    iter_vars (fun v -> raise Not_found) term;
    true
  with Not_found ->
    false

let is_nonvar term =
  match term with
    Term.Variable _ ->
      false
  | _ ->
      true

let varnum term =
  match term with
    Term.Variable v ->
      Term.variable_number v
  | _ ->
      raise (Failure "Variable expected.")

let contains_vars_of t1 t2 =
	let not_contained = Array.make Term.max_vars true
	in
		iter_vars
			(fun v ->
				not_contained.(Term.variable_number v) <- false)
			t1;
		try (
			iter_vars
				(fun v ->
					if not_contained.(Term.variable_number v) then 
						(raise Not_found))
				t2;
			true)
		with Not_found -> false
		
			
let rec get_shared_varnums terms =
	let vars = Array.make Term.max_vars (-1)
	and shared_varnums = ref []
	in 
	get_shared_varnums_1 terms vars shared_varnums 1
	
and	get_shared_varnums_1 terms vars shared_varnums i =
	match terms with
		[] -> !shared_varnums
		|term :: rest_terms -> 
			iter_vars
				(fun v ->
					let vnum = Term.variable_number v
					in	
						match vars.(vnum) with
							(-1) -> vars.(vnum)<-i
							|0 -> ()
							|_ ->
								if vars.(vnum)<i then
									(shared_varnums := vnum :: !shared_varnums)
								else ())
				term;
			get_shared_varnums_1 rest_terms vars shared_varnums (i+1)
			
			
let split_context term context special_varnums =
	let normal_bindings = Array.make Term.max_vars Term.Unbound
	and special_bindings = Array.make Term.max_vars Term.Unbound
	in
		iter_vars
			(fun v ->
				let vnum = Term.variable_number v
				in
					if (List.mem vnum special_varnums) then
						(if (special_bindings.(vnum) == Term.Unbound) then ()
						else (special_bindings.(vnum) <- Term.apply (Term.Variable v) context))
					else (if (normal_bindings.(vnum) == Term.Unbound) then ()
						else (normal_bindings.(vnum) <- Term.apply (Term.Variable v) context)))
			term;
			(Term.make_bound_context normal_bindings 0,
			Term.make_bound_context special_bindings 0)