(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Utility functions for terms. *)

(** [iter_vars f term] effects that [f] is applied to all variables
   in [term]. If a variable has multiple occurences in [term],
   [f] may be called on each occurence.
 *)
val iter_vars : (Term.variable -> unit) -> Term.term -> unit

(** [is_ground term] returns true iff [term] contains no variables.
 *)
val is_ground : Term.term -> bool

(** [is_nonvar term] returns true iff [term] is not a variable. *)
val is_nonvar : Term.term -> bool

(** [term] is assumed to be term of shape [Variable variable]. The
   variable number of [variable] is returned. *) 
val varnum : Term.term -> int

(** [contains_vars_of term1 term2] returns true iff all variables of
[term2] occur in [term1]. *)
val contains_vars_of : Term.term -> Term.term -> bool

(** [get_shared_varnums terms] returns a list of the variable numbers
of all variables that occur in more than one term of [terms]. *)
val get_shared_varnums : Term.term list -> int list

(** experimental *)
val split_context : Term.term -> Term.context -> int list -> (Term.context * Term.context)