(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


type term_weighing_method =
    SymbolCount
  | Depth

let rec symbol_count term w =
  match term with
    Term.Compound c ->
      let args = Term.compound_args c in
      map_symbol_count 
	args ((Array.length args) - 1) (w + 1)
  | _ -> w + 1

and map_symbol_count args i w =
  if i < 0 then
    w
  else
    map_symbol_count args (i - 1) (symbol_count args.(i) w)

let rec depth term w m =
  match term with
    Term.Compound c ->
      let args = Term.compound_args c in
      map_depth args ((Array.length args) - 1) (w + 1) m
  | _ -> max m (w + 1)

and map_depth args i w m =
  if i < 0 then
    m
  else
    map_depth args (i - 1) w (depth args.(i) w m)

let term_weight term_weight_method term =
  match term_weight_method with
    SymbolCount ->
      symbol_count term 0
  | Depth ->
      (* useful e.g. for TPTP Problem NUM284-1.014 *) 
      depth term 0 0


