
/* Uit logica voor informatici, blz. 202.   */

%predicates
   N1/1, N2/1, N3/1.
   N4/1, N5/1, N6/1.
   N7/1, N8/1, N9/1.

   I/1.    // I(x) : x is a valid index.
   D/1.    // D(x) : x is a valid digit. (cell contents)

   R/2.    // R(x,y) : x and y are in the same group of
           //   (1,2,3), (4,5,6), (7,8,9).

   C/3.    // C(x,y,z) : position (x,y) contains z.

%firstorder

   // We introduce the predicates N1 ..., N9 representing
   // the digits 1, ..., 9.

   <x> N1(x).
   <x> N2(x) /\ ! N1(x).
   <x> N3(x) /\ ! N1(x) /\ ! N2(x). 
   <x> N4(x) /\ ! N1(x) /\ ! N2(x) /\ ! N3(x). 
   <x> N5(x) /\ ! N1(x) /\ ! N2(x) /\ ! N3(x) /\ ! N4(x). 
   <x> N6(x) /\ ! N1(x) /\ ! N2(x) /\ ! N3(x) /\ ! N4(x) /\ ! N5(x).
   <x> N7(x) /\ ! N1(x) /\ ! N2(x) /\ ! N3(x) /\ ! N4(x) /\ ! N5(x) 
             /\ ! N6(x).
   <x> N8(x) /\ ! N1(x) /\ ! N2(x) /\ ! N3(x) /\ ! N4(x) /\ ! N5(x) 
             /\ ! N6(x) /\ ! N7(x).
   <x> N9(x) /\ ! N1(x) /\ ! N2(x) /\ ! N3(x) /\ ! N4(x) /\ ! N5(x)
             /\ ! N6(x) /\ ! N7(x) /\ ! N8(x). 


   // We define I(x) as x = 1, x = 2, ... x = 9.

   [x] I(x) <-> ( N1(x) \/ N2(x) \/ N3(x) \/ N4(x) \/ N5(x) \/
                  N6(x) \/ N7(x) \/ N8(x) \/ N9(x) ).


   // Each of N1, ..., N9 has at most one inhabitant: 

   [x1,x2] N1(x1) /\ N1(x2) -> x1 = x2.
   [x1,x2] N2(x1) /\ N2(x2) -> x1 = x2.
   [x1,x2] N3(x1) /\ N3(x2) -> x1 = x2.
   [x1,x2] N4(x1) /\ N4(x2) -> x1 = x2.
   [x1,x2] N5(x1) /\ N5(x2) -> x1 = x2.
   [x1,x2] N6(x1) /\ N6(x2) -> x1 = x2.
   [x1,x2] N7(x1) /\ N7(x2) -> x1 = x2.
   [x1,x2] N8(x1) /\ N8(x2) -> x1 = x2.
   [x1,x2] N9(x1) /\ N9(x2) -> x1 = x2.




   // Every valid entry should have a contents: 

   [x,y] I(x) /\ I(y) -> <z> I(z) /\ C(x,y,z).


   // No distinct entries in the same column should have the same
   // contents:

   [x,y1,y2,z] 
        I(x) /\ I(y1) /\ I(y2) /\ y1 != y2 /\ C(x,y1,z) /\ C(x,y2,z) -> FALSE.


   // No distinct entries in the same row should have the same contents:

   [x1,x2,y,z]
      I(x1) /\ I(x2) /\ x1 != x2 /\ I(y) /\ C(x1,y,z) /\ C(x2,y,z) -> FALSE.


   // A similar condition holds for blocks, but it is a bit more tricky
   // to formulate it. R(x,y) means that x and y are in the same
   // group of (1,2,3)  (4,5,6)   (7,8,9). 

   [x1,x2] ( N1(x1) \/ N2(x1) \/ N3(x1) ) /\ 
           ( N1(x2) \/ N2(x2) \/ N3(x2) ) -> R(x1,x2).
   [x1,x2] ( N4(x1) \/ N5(x1) \/ N6(x1) ) /\
           ( N4(x2) \/ N5(x2) \/ N6(x2) ) -> R(x1,x2).
   [x1,x2] ( N7(x1) \/ N8(x1) \/ N9(x1) ) /\
           ( N7(x2) \/ N8(x2) \/ N9(x2) ) -> R(x1,x2).

   [x1,x2,y1,y2,z]
         R(x1,x2) /\ R(y1,y2) /\ ( x1 != x2 \/ y1 != y2 ) /\ 
         C(x1,y1,z) /\ C(x2,y2,z) -> FALSE.


   /////////////////////////////////


   [x,y,z] N1(x) /\ N1(y) /\ N1(z) -> C(x,y,z).
   [x,y,z] N1(x) /\ N2(y) /\ N7(z) -> C(x,y,z).
   [x,y,z] N1(x) /\ N3(y) /\ N4(z) -> C(x,y,z).
   [x,y,z] N1(x) /\ N4(y) /\ N8(z) -> C(x,y,z).
   [x,y,z] N1(x) /\ N6(y) /\ N3(z) -> C(x,y,z).

   [x,y,z] N2(x) /\ N5(y) /\ N1(z) -> C(x,y,z).
   [x,y,z] N2(x) /\ N6(y) /\ N2(z) -> C(x,y,z).
   [x,y,z] N2(x) /\ N7(y) /\ N4(z) -> C(x,y,z).

   [x,y,z] N3(x) /\ N1(y) /\ N8(z) -> C(x,y,z).
   [x,y,z] N3(x) /\ N2(y) /\ N6(z) -> C(x,y,z).
   [x,y,z] N3(x) /\ N4(y) /\ N9(z) -> C(x,y,z).
   [x,y,z] N3(x) /\ N7(y) /\ N1(z) -> C(x,y,z).

   [x,y,z] N4(x) /\ N3(y) /\ N8(z) -> C(x,y,z).
   [x,y,z] N4(x) /\ N8(y) /\ N1(z) -> C(x,y,z).
   [x,y,z] N4(x) /\ N9(y) /\ N9(z) -> C(x,y,z).

   [x,y,z] N5(x) /\ N1(y) /\ N2(z) -> C(x,y,z).
   [x,y,z] N5(x) /\ N2(y) /\ N9(z) -> C(x,y,z).
   [x,y,z] N5(x) /\ N4(y) /\ N5(z) -> C(x,y,z).

   [x,y,z] N6(x) /\ N6(y) /\ N8(z) -> C(x,y,z).
   [x,y,z] N6(x) /\ N7(y) /\ N2(z) -> C(x,y,z).
   
   [x,y,z] N7(x) /\ N1(y) /\ N4(z) -> C(x,y,z).
   [x,y,z] N7(x) /\ N2(y) /\ N2(z) -> C(x,y,z).
   [x,y,z] N7(x) /\ N6(y) /\ N1(z) -> C(x,y,z).
   [x,y,z] N7(x) /\ N8(y) /\ N6(z) -> C(x,y,z).

   [x,y,z] N8(x) /\ N3(y) /\ N5(z) -> C(x,y,z).
   [x,y,z] N8(x) /\ N4(y) /\ N7(z) -> C(x,y,z).
   [x,y,z] N8(x) /\ N5(y) /\ N4(z) -> C(x,y,z).

   [x,y,z] N9(x) /\ N1(y) /\ N3(z) -> C(x,y,z).
   [x,y,z] N9(x) /\ N3(y) /\ N9(z) -> C(x,y,z).
   [x,y,z] N9(x) /\ N4(y) /\ N2(z) -> C(x,y,z).


