

// Formalization of SuDoKus.


%functions 
   x1/0, x2/0, x3/0, x4/0, x5/0, x6/0, x7/0, x8/0, x9/0.
   y1/0, y2/0, y3/0, y4/0, y5/0, y6/0, y7/0, y8/0, y9/0.
   d1/0, d2/0, d3/0, d4/0, d5/0, d6/0, d7/0, d8/0, d9/0.

   val/2.

%predicates

   H/1.    // H(x) : x is a horizontal index.
   V/1.    // V(x) : x is a vertical index.
   D/1.    // D(x) : x is a valid digit. (cell contents)

   RH/2,RV/2.   // R(x,y) : x and y are in the same group of
                //   (1,2,3), (4,5,6), (7,8,9).

   
%firstorder

   // The x-coordinates are distinct: 

   x2!=x1.
   x3!=x2 /\ x3!=x1.
   x4!=x3 /\ x4!=x2 /\ x4!=x1.
   x5!=x4 /\ x5!=x3 /\ x5!=x2 /\ x5!=x1.
   x6!=x5 /\ x6!=x4 /\ x6!=x3 /\ x6!=x2 /\ x6!=x1.
   x7!=x6 /\ x7!=x5 /\ x7!=x4 /\ x7!=x3 /\ x7!=x2 /\ x7!=x1.
   x8!=x7 /\ x8!=x6 /\ x8!=x5 /\ x8!=x4 /\ x8!=x3 /\ x8!=x2 /\ x8!=x1.
   x9!=x8 /\ x9!=x7 /\ x9!=x6 /\ x9!=x5 /\ x9!=x4 /\ x9!=x3 /\ x9!=x2 /\ x9!=x1.


   [z] H(z) <-> 
         z=x1 \/ z=x2 \/ z=x3 \/ z=x4 \/ z=x5 \/ z=x6 \/ z=x7 \/ z=x8 \/ z=x9.


   // The y-coordinates are distinct:

   y2!=y1.
   y3!=y2 /\ y3!=y1.
   y4!=y3 /\ y4!=y2 /\ y4!=y1.
   y5!=y4 /\ y5!=y3 /\ y5!=y2 /\ y5!=y1.
   y6!=y5 /\ y6!=y4 /\ y6!=y3 /\ y6!=y2 /\ y6!=y1.
   y7!=y6 /\ y7!=y5 /\ y7!=y4 /\ y7!=y3 /\ y7!=y2 /\ y7!=y1.
   y8!=y7 /\ y8!=y6 /\ y8!=y5 /\ y8!=y4 /\ y8!=y3 /\ y8!=y2 /\ y8!=y1.
   y9!=y8 /\ y9!=y7 /\ y9!=y6 /\ y9!=y5 /\ y9!=y4 /\ y9!=y3 /\ y9!=y2 /\ y9!=y1.

   [z] V(z) <->
         z=y1 \/ z=y2 \/ z=y3 \/ z=y4 \/ z=y5 \/ z=y6 \/ z=y7 \/ z=y8 \/ z=y9.


   // The digits are distinct:

   d2!=d1.
   d3!=d2 /\ d3!=d1.
   d4!=d3 /\ d4!=d2 /\ d4!=d1.
   d5!=d4 /\ d5!=d3 /\ d5!=d2 /\ d5!=d1.
   d6!=d5 /\ d6!=d4 /\ d6!=d3 /\ d6!=d2 /\ d6!=d1.
   d7!=d6 /\ d7!=d5 /\ d7!=d4 /\ d7!=d3 /\ d7!=d2 /\ d7!=d1.
   d8!=d7 /\ d8!=d6 /\ d8!=d5 /\ d8!=d4 /\ d8!=d3 /\ d8!=d2 /\ d8!=d1.
   d9!=d8 /\ d9!=d7 /\ d9!=d6 /\ d9!=d5 /\ d9!=d4 /\ d9!=d3 /\ d9!=d2 /\ d9!=d1.


   [z] D(z) <-> 
         z=d1 \/ z=d2 \/ z=d3 \/ z=d4 \/ z=d5 \/ z=d6 \/ z=d7 \/ z=x8 \/ z=d9.


   // Each entry should be a digit:

   // [x,y] H(x) /\ V(y) -> D( val(x,y) ).


   // No distinct entries in the same column should have the same
   // contents:

   [x,y1,y2] 
        H(x) /\ V(y1) /\ V(y2) /\ y1 != y2 -> val(x,y1) != val(x,y2). 


   // No distinct entries in the same row should have the same contents:

   [x1,x2,y]
      H(x1) /\ H(x2) /\ V(y) /\ x1 != x2 -> val(x1,y) != val(x2,y).


   // A similar condition holds for blocks, but it is a bit more tricky
   // to formulate it. RV(x,y) and RH(x,y) mean that x and y are in the same
   // group of (1,2,3)  (4,5,6)  (7,8,9). 


   [z1,z2] ( z1 = x1 \/ z1 = x2 \/ z1 = x3 ) /\ 
           ( z2 = x1 \/ z2 = x2 \/ z2 = x3 ) -> RH(z1,z2).
   [z1,z2] ( z1 = x4 \/ z1 = x5 \/ z1 = x6 ) /\
           ( z2 = x4 \/ z2 = x5 \/ z2 = x6 ) -> RH(z1,z2).
   [z1,z2] ( z1 = x7 \/ z1 = x8 \/ z1 = x9 ) /\
           ( z2 = x7 \/ z2 = x8 \/ z2 = x9 ) -> RH(z1,z2).


   [z1,z2] ( z1 = y1 \/ z1 = y2 \/ z1 = y3 ) /\ 
           ( z2 = y1 \/ z2 = y2 \/ z2 = y3 ) -> RV(z1,z2).
   [z1,z2] ( z1 = y4 \/ z1 = y5 \/ z1 = y6 ) /\
           ( z2 = y4 \/ z2 = y5 \/ z2 = y6 ) -> RV(z1,z2).
   [z1,z2] ( z1 = y7 \/ z1 = y8 \/ z1 = y9 ) /\
           ( z2 = y7 \/ z2 = y8 \/ z2 = y9 ) -> RV(z1,z2).

   [x1,x2,y1,y2] 
         RH(x1,x2) /\ RV(y1,y2) /\ ( x1 != x2 \/ y1 != y2 ) 
	           -> val(x1,y1) != val(x2,y2). 


   /////////////////////////////////


   // Accelerating axioms:
  
   [x,z] H(x) /\ D(z) ->
      val(x,y1) = z \/ val(x,y2) = z \/ val(x,y3) = z \/ 
      val(x,y4) = z \/ val(x,y5) = z \/ val(x,y6) = z \/
      val(x,y7) = z \/ val(x,y8) = z \/ val(x,y9) = z.


   [y,z] V(y) /\ D(z) ->
      val(x1,y) = z \/ val(x2,y) = z \/ val(x3,y) = z \/
      val(x4,y) = z \/ val(x5,y) = z \/ val(x6,y) = z \/
      val(x7,y) = z \/ val(x8,y) = z \/ val(x9,y) = z.




   val(x1,y1) = d1.
   val(x1,y2) = d7.
   val(x1,y3) = d4.
   val(x1,y4) = d8.
   val(x1,y6) = d3.

   val(x2,y5) = d1.
   val(x2,y6) = d2. 
   val(x2,y7) = d4.

   val(x3,y1) = d8.
   val(x3,y2) = d6. 
   val(x3,y4) = d9.
   val(x3,y7) = d1.

   val(x4,y3) = d8.
   val(x4,y8) = d1.
   val(x4,y9) = d9.

   val(x5,y1) = d2.
   val(x5,y2) = d9.
   val(x5,y4) = d5.

   val(x6,y6) = d8.
   val(x6,y7) = d2.

   val(x7,y1) = d4.
   val(x7,y2) = d2.
   val(x7,y6) = d1.
   val(x7,y8) = d6.

   val(x8,y3) = d5.
   val(x8,y4) = d7.
   val(x8,y5) = d4.

   val(x9,y1) = d3.
   val(x9,y3) = d9.
   val(x9,y4) = d2.


