

// Code written by Hans de Nivelle, February 2006. 



namespace gottlob_leitsch
{



   namespace
   {

      // Number of occurrences of v in va. 

      unsigned int nroccurrences( variable v, varatom va )
      {
         unsigned int nr = 0;  
         for( std::vector< variable > :: const_iterator 
                 p = va. a1;
	         p != va. a2;
	         ++ p )
         {
            if( *p == v ) 
               nr ++ ;  
         }
         return nr; 
      }


      // Number of atoms in prem containing v. 

      unsigned int nratoms( variable v, const premisseset& prem )
      {
         unsigned int nr = 0;  
         for( premisseset::const_iterator 
                 p = prem. begin( ); 
	         p != prem. end( ); 
	         ++ p )
         {
            if( nroccurrences( v, *(*p)) != 0 ) 
               nr ++ ;
         }
         return nr; 
      }
 

      // Number of positions in prem containing v. 

      unsigned int nroccurrences( variable v, const premisseset& prem ) 
      {
         unsigned int nr = 0; 
         for( premisseset::const_iterator 
                 p = prem. begin( ); 
	         p != prem. end( ); 
	         ++ p )
         {
            nr += nroccurrences( v, *(*p));  
         }
         return nr;
      }


      // Set of non-bound variables in va. 

      varset nonboundvariables( const groundsubst& fixed, 
		                const premisseset& prem ) 
      {
         varset res1;
         varset res2;

         for( premisseset::const_iterator p = prem. begin( );
                 p != prem. end( );
	         ++ p )
         {
            for( std::vector< variable > :: const_iterator
                    q = (*(*p)). a1;
	            q != (*(*p)). a2;
	            ++ q )
            {
               res2. insert( *q ); 
            }
        }

        while( res2. size( ) != 0 )
        {
           variable v = res2. select( );
	   res2. remove(v);
	   if( fixed. lookup(v) == 0 )
              res1. insert( v );
        }

        return res1; 
      }


      unsigned int sharevariable( varatom va1, varatom va2,
                                  const groundsubst& fixed )
      {
         for( std::vector< variable > :: const_iterator
                 p = va1. a1;
	         p != va1. a2; 
                 ++ p )
         {
            for( std::vector< variable > :: const_iterator
                    q = va2. a1;
		    q != va2. a2;
		    ++ q )
            {
               if( *p == *q && fixed. lookup(*p) == 0 )
                  return true;
            }
         }
         return false;
      }


      unsigned int sumofdistances( const premisseset& prem,
		                   const groundsubst& fixed,
				   variable v )
      {
         std::vector< unsigned int > dist;
         for( premisseset::const_iterator
                 q = prem. begin( );
	         q != prem. end( );
	         ++ q )
         {
            if( nroccurrences( v, (*(*q)) ) != 0 )
               dist. push_back( 0 );
	    else
               dist. push_back( 999999 );
         }

         bool change = true;
         while( change )
         {
            change = false;
	 
	    std::vector< unsigned int > :: iterator r1 = dist. begin( );
	    for( premisseset::const_iterator
                    q1 = prem. begin( );
		    q1 != prem. end( );
		    ++ q1, ++ r1 )
            {
               std::vector< unsigned int > :: iterator r2 = dist. begin( );
	       for( premisseset::const_iterator
                       q2 = prem. begin( );
		       q2 != prem. end( );
		       ++ q2, ++ r2 )
               {
                  if( q1 != q2 && sharevariable( *(*q1), *(*q2), fixed ))
                  {
                     if( *r2 > *r1 + 1 )
                     {
                        *r2 = *r1 + 1;
		        change = true;
                     }

		     if( *r1 > *r2 + 1 )
                     {
                        *r1 = *r2 + 1;
		        change = true;
                     }
                  }
               }
            }
         }

         unsigned int res = 0;
         for( std::vector< unsigned int > :: const_iterator 
	         r = dist. begin( );
                 r != dist. end( );
	         ++ r )
         {
            res += *r;
         }
         return res;
      }



      bool sharevariable( varatom va, const varset& vs )
      {
         for( std::vector< variable > :: const_iterator 
                 p = va. a1;
	         p != va. a2;
	         ++ p )
         {
            if( vs. contains( *p ))
               return true;
         }
         return false;
      }


   }





   variable select( const premisseset& prem,
	            const groundsubst& fixed )
   {

      varset vars = nonboundvariables( fixed, prem );
      ASSERT( vars. size( ) != 0 );

      variable best_v = vars. select( );
      vars. remove( best_v );

      unsigned int best_occ = nroccurrences( best_v, prem ); 
      unsigned int best_sum = sumofdistances( prem, fixed, best_v );

      while( vars. size( ) != 0 )
      {
         variable v = vars. select( );
         vars. remove(v);

         unsigned int occ = nroccurrences( v, prem );
         unsigned int sum = sumofdistances( prem, fixed, v );

         if( occ > best_occ || ( occ == best_occ && sum < best_sum ))
         {
            best_occ = occ;
	    best_sum = sum; 
	    best_v = v;
         }
      }

      return best_v;
   }



   namespace 
   {

      struct matchedatom
      {
         groundsubst subst;
         unsigned int atom_nr;

         matchedatom( const groundsubst& subst, unsigned int atom_nr )
            : subst( subst ), atom_nr( atom_nr )
         {
         }

      };



      std::vector< matchedatom > 
      all_matchings( groundsubst& subst,
                  varatom va,
		  const model& m,
		  unsigned int upperbound )
      {
         std::vector< matchedatom > res;

         if( upperbound == 0 )
            return res;

         unsigned int subst_size = subst. size( );

         if( va. p != predicate::diseq )
         { 
            unsigned int atom_nr = 0;  
            for( model::atom_iterator atom = m. atoms_begin( );
                    atom != m. atoms_end( );
	            ++ atom, ++ atom_nr )
            {
               groundatom ga = m. getatom( atom );
	       if( va. p == ga. p && subst. match( va, ga ))
                  res. push_back( matchedatom::matchedatom( subst, atom_nr ));
               subst. backtrack( subst_size );
	       if( res. size( ) == upperbound )
                  return res;
            }
         }
         else
         {
            std::vector< variable > :: const_iterator a = va. a1;
	    variable v1 = *(a++);
	    variable v2 = *a;

	    const element* e1 = subst. lookup(v1);
	    const element* e2 = subst. lookup(v2);

	    // If e2 is defined while e1 is not defined, then
	    // it is better to exchange:

	    if( e1 == 0 && e2 != 0 )
            {
               variable v3 = v2;
	       v2 = v1;
	       v1 = v3;

	       const element* e3 = e2;
	       e2 = e1;
	       e1 = e3; 
            }

            unsigned int atom_nr1 = 0; 
	    for( model::atom_iterator atom1 = m. atoms_begin( );
                    atom1 != m. atoms_end( );
	            ++ atom1, ++ atom_nr1 ) 
            {
               groundatom ga1 = m. getatom( atom1 );
	       if( ga1. p == predicate::create )
               {
                  element x1 = *( ga1. a1 ); 
                  if( !e1 || *e1 == x1 )
                  {
                     unsigned int atom_nr2 = 0; 
	             for( model::atom_iterator atom2 = m. atoms_begin( );
                             atom2 != m. atoms_end( );
		             ++ atom2, ++ atom_nr2 )
                     {
                        groundatom ga2 = m. getatom( atom2 ); 
                        if( ga2. p == predicate::create )
                        {
                           element x2 = *( ga2. a1 ); 
                           {
                              if( x1 != x2 )
                              { 
                                 if( !e2 || *e2 == x2 )
                                 {
                                    if( e1 == 0 )
	                               subst. assign( v1, x1 );
			            if( e2 == 0 )
                                       subst. assign( v2, x2 );
		
			            unsigned int nr = atom_nr1;
			            if( atom_nr2 > atom_nr1 )
                                       nr = atom_nr2; 

			            res. push_back( 
				      matchedatom::matchedatom( subst, nr ));

			            subst. backtrack( subst_size );
			            if( res. size( ) == upperbound )
                                       return res;
                                 }
                              }
                           }
                        }
                     }
                  }
               }
            }
         }    
         return res;
      }


      // this is off course a tremendeously inefficient implementation.

      void cleanup_matchings( std::vector< std::vector< matchedatom > > & all )
      {
         unsigned int nriterations = 0;

      restart: 
         for( std::vector< std::vector< matchedatom > > :: iterator
                 p1 = all. begin( );
	         p1 != all. end( );
	         ++ p1 )
         {
            for( std::vector< matchedatom > :: iterator
                    q1 = p1 -> begin( );
		    q1 != p1 -> end( );
		    ++ q1 )
            {
               for( std::vector< std::vector< matchedatom > > :: const_iterator
                       p2 = all. begin( );
		       p2 != all. end( );
		       ++ p2 )
               {
                  if( p1 != p2 )
                  {
                     for( std::vector< matchedatom > :: const_iterator
                             q2 = p2 -> begin( );
			     q2 != p2 -> end( );
			     ++ q2 )
                     {
                        if( q1 -> subst. consistentwith( q2 -> subst ))
                           goto found;
                     }

		     // this means that q1 is not concistent with
		     // any of the elements in p2.

                     *q1 = p1 -> back( );
		     p1 -> pop_back( );
		     ++ nriterations ;
                     goto restart;
                  }
               }
            found:
	       ;
            }
         }
         std::cout << "nriterations " << nriterations << "\n\n"; 
      }
   };





   partial< application::match >
   application::findmatch1( premmatching& match,
		            groundsubst& fixed,
			    premisseset component,
			    const rule& r,
			    const model& m,
			    statistics_collector& stat,
			    unsigned int level,
			    bool showing )
   {
      stat. findmatch1_calls ++ ;

#if 0
      std::cout << "Entering findmatch1 on level " << level << ":\n"; 
      std::cout << "Component: " << component << "\n";
      std::cout << "Match: " << match << "\n";
      std::cout << "Rule: " << r << "\n"; 
      std::cout << "Fixed: " << fixed << "\n";
      // std::cout << "Model: " << m << "\n";
      std::cout << "\n";
#endif

      variable selected = select( component, fixed );

      if( showing ) 
      {
         if( level == 0 ) 
         {
            std::cout << "--------------------------------------------\n";
            std::cout << "Application Split Statistics:\n";
            std::cout << "model: " << m << "\n";
            std::cout << "rule: " << r << "\n";

	    std::cout << "possible matches: \n";
	    for( premisseset::const_iterator 
		    p = component. begin( ); 
                    p != component. end( ); 
	            ++ p )
            {
               std::vector< matchedatom > all = 
		       all_matchings( fixed, *(*p), m, 99999999 );

	       std::cout << all. size( ) << ": "; 
	       std::cout << *(*p) << " ==>\n"; 
	       std::cout << "       "; 
	       for( std::vector< matchedatom > :: const_iterator 
                       q = all. begin( );
		       q != all. end( );
		       ++ q )
               {
                  std::cout << ( q -> subst ). apply_on( *(*p) ) << "  ";
               }
	       std::cout << "\n"; 
            }
	    std::cout << "\n"; 
         }
         std::cout << "fixed: " << fixed << "\n";
         std::cout << "component: " << component << "\n";
         std::cout << "level: " << level << "\n";

         std::cout << "\n";

         std::cout << "free var: " << nonboundvariables( fixed, component );
         std::cout << "\n";

         varset vs = nonboundvariables( fixed, component );
         std::cout << "nroccurrences [ ";
         while( vs. size( ))
         {
            variable v = vs. select( );
	    vs. remove(v);
	    std::cout << nroccurrences( v, component );

	    if( vs. size( ))
               std::cout << ", ";
	    else
               std::cout << " ]\n\n";
         }

         vs = nonboundvariables( fixed, component );
         std::cout << "sum of distances [ ";
         while( vs. size( ))
         {
            variable v = vs. select( );
	    vs. remove(v);
	    std::cout << sumofdistances( component, fixed, v );

	    if( vs. size( ))
               std::cout << ", ";
	    else
               std::cout<< " ]\n\n";
         }

         std::cout << "selecting: " << selected << "\n\n\n\n"; 

      }

      partial< application::match > res;
      unsigned int fixed_size = fixed. size( );

      unsigned int nr_elements = m. nrelements( ); 
      element e;
      for( unsigned int k = 0; k != nr_elements; ++ k, ++ e ) 
      {
         stat. nrguesses ++ ; 

         fixed. assign( selected, e ); 

         varset vv;
         vv. insert(selected );

         partial< application::match > 
            res2 = findmatch3( match, fixed, component, vv, r, m, 
			       stat, level + 1, showing ); 
   
         if( ! res. isdefined( ) ||
             ( res2. isdefined( ) && 
                  res. contents( ). m. sup_atoms( ) > 
	          res2. contents( ). m. sup_atoms( )))
         {
            res = res2;
         }

         fixed. backtrack( fixed_size ); 

         if( res. isdefined( ) && 
             r. isdisjunctive( ) && r. nr_conclusions( ) == 0 )
         {
            return res;
         }
      }

      return res; 
   }


   partial< application::match >
   application::findmatch2( premmatching& match,
		            groundsubst& fixed,
			    premisseset component,
			    const rule& r,
			    const model& m,
			    statistics_collector& stat,
			    unsigned int level,
			    bool showing ) 
   {
      stat. findmatch2_calls ++ ;

#if 0
      std::cout << "Entering findmatch2 at level " << level << ":\n"; 
      std::cout << "Component: " << component << "\n";
      std::cout << "Match: " << match << "\n";
      std::cout << "Fixed: " << fixed << "\n";
      std::cout << "Model: " << m << "\n";

      std::cout << "\n";
#endif

      ASSERT( component. size( ) != 0 );

      // We attempt to split component,
      // but only if r is a lemma. 
 
      if( r. isdisjunctive( ) && r. nr_conclusions( ) == 0 )
      {

         premisseset component2;
         component2. insert( * ( component. begin( )) );

         component2. close( r, fixed );

         if( component2. size( ) < component. size( ))
         { 
#if 0 
            std::cout << "splitting: " << component << "\n";
	    std::cout << component2 << "\n";
	    std::cout << component - component2 << "\n\n";
#endif
            stat. nrsplits ++ ;

            partial< application::match > res1 =
               findmatch1( match, fixed, component2, r, m, stat, 
			   level, showing );

            if( ! res1. isdefined( ))
               return res1;

            partial< application::match > res2 =
               findmatch2( match, fixed, component - component2,
                           r, m, stat, level, showing );

            if( ! res2. isdefined( ))
               return res2;

            return match::match( res1. contents( ). m + res2. contents( ). m,
		                 res1. contents( ). s + res2. contents( ). s );
         }
      }

      return findmatch1( match, fixed, component, r, m, stat, level, showing );
   }


   partial< application::match > 
   application::findmatch3( premmatching& matched,
		            groundsubst& fixed,
                            premisseset component,
			    varset changed,
			    const rule& r,
			    const model& m,
			    statistics_collector& stat,
			    unsigned int level,
			    bool showing )
   {
      stat. findmatch3_calls ++ ;

#if 0
      std::cout << "Entering findmatch3 at level " << level << ":\n"; 
      std::cout << "Component: " << component << "\n";
      std::cout << "Changed: " << changed << "\n";
      std::cout << "Matched: " << matched << "\n"; 
      std::cout << "Fixed: " << fixed << "\n";
      std::cout << "Model: " << m << "\n"; 
#endif

      unsigned int fixed_size = fixed. size( );
      unsigned int matched_size = matched. size( );

      // We look for atoms that can be matched in 0 or 1 ways,
      // and which contain a variable from changed.

      while( changed. size( ) != 0 && component. size( ) != 0 )
      {
         unsigned int fixed_size_this_round = fixed. size( );

         premisseset oneway;

         for( premisseset::const_iterator
                 p = component. begin( );
	         p != component. end( );
	         ++ p )
         {
            varatom va = * (*p);
	    if( sharevariable( va, changed ) || va. p. getarity( ) == 0 )
            {
               std::vector< matchedatom > match = 
		       all_matchings( fixed, va, m, 2 );

               if( match. size( ) == 0 )
               {
                  stat. nrnomatch ++ ;

                  fixed. backtrack( fixed_size );
	          matched. backtrack( matched_size );
	          return partial< application::match > :: partial( );
               }

               if( match. size( ) == 1 )
               {
                  stat. nrforward ++ ; 
                  oneway. insert( *p );
	          fixed = match. front( ). subst;
                  matched. assign( *p, match. front( ). atom_nr ); 
               }
            }
         }

         component = component - oneway;
         changed = fixed. getaddedvars( fixed_size_this_round );
      }

      // If we managed to get the component empty, we have in principle found 
      // a complete matching, but we need to check that the conclusion is
      // false, in case the rule is not a lemma.

      if( component. size( ) == 0 )
      {
         if( conclusionistrue( m, fixed, r, stat ))
         {
            fixed. backtrack( fixed_size );
	    matched. backtrack( matched_size );
            return partial< match > :: partial( );
         }
         else
         {
            match result = match::match( matched, fixed );
            fixed. backtrack( fixed_size );
            matched. backtrack( matched_size );
            return result;
         }
      }

      partial< match > 
         result = findmatch2( matched, fixed, component, r, m, stat, 
		              level, showing );

      fixed. backtrack( fixed_size );
      matched. backtrack( matched_size );
      return result;
   }



   void test_gottlob_leitsch( const model& m,
		              const rulesystem& rs )
   {
      ASSERT( rs. size( ) == 1 );
      rule r = rs. rules [0]. r;

      std::cout << m << "\n";
      std::cout << r << "\n";

      ASSERT( r. isdisjunctive( ));
      ASSERT( r. nr_conclusions( ) == 0 );

      statistics_collector stat;
      std::vector< unsigned int > split_distr;
      split_distr. push_back(0);
         // This vector counts how the splits are distributed.


      premisseset premisses = premisseset::premisseset( rule );
      varset vars = rule. getvariables( rule. prems_begin( ),
			                rule. prems_end( ));

      premmatching pp;
      groundsubst ss;

      partial< match > bestthisrule = 
                        findmatch3( pp, ss, premisses, vars, rule -> r, m, 
				    stat, 0, false ); 

      if( bestthisrule. isdefined( ))
      {
         std::cout << "found solution: ";
	 std::cout << bestthisrule. contents( ). s << "\n";
      }
      else
         std::cout << "found no solution\n";
     

      // Show the statistics:

      std::cout << stat << "\n";

      std::cout << "distr: "; 
      for( unsigned int i = 0; i < split_distr. size( ); ++ i )
      {
         std::cout << split_distr [i] << " < " << ( 1 << i );
	 if( i < split_distr. size( ) - 1 )
            std::cout << ", ";
         else
            std::cout << "."; 
      }
      std::cout << "\n\n\n";

   }



}


