

// This file contains the Gottlob-Leitsch algorithm for matching.
// It is based on forward reasoning, splitting and guessing. 

// The algorithm in DPLL is better, but we keep the Gottlob-Leitsch
// algorithm for the purpose of comparing.


#ifndef GL_FINDMATCH_INCLUDED
#define GL_FINDMATCH_INCLUDED   1


#include "../partial.h"
#include "../model.h" 
#include "premmatching.h"


#include "statistics_collector.h"


namespace gottlob_leitsch
{


   struct match 
   {
      premmatching m;
      groundsubst s;

      match( const premmatching& m, const groundsubst& s )
         : m(m), s(s)
      {
      }
   }; 



   static partial< match > 
   findmatch1( premmatching& matched,
	       groundsubst& fixed, 
               premisseset component,
	       const rule& r,
	       const model& m, 
	       statistics_collector& stat,
	       unsigned int level, 
	       bool showing ); 

   static partial< match >
   findmatch2( premmatching& matched,
	       groundsubst& fixed,
               premisseset component,
	       const rule& r,
	       const model& m,
	       statistics_collector& stat,
	       unsigned int level,
	       bool showing ); 

   static partial< match > 
   findmatch3( premmatching& matched,
	       groundsubst& fixed,
               premisseset component,
	       varset changed,
	       const rule& r,
	       const model& m,
	       statistics_collector& stat,
	       unsigned int level,
	       bool showing ); 


   static void test_gottlob_leitsch( const model& m, const rulesystem& rs );

};



#endif


