

// Written by Hans de Nivelle, February 2006.


#include "premisseset.h"


namespace gottlob_leitsch
{

   premisseset::premisseset( ) 
   {
   }


   premisseset::premisseset( const rule& r )
   {
      for( rule::prem_iterator 
              p = r. prems_begin( );
              p != r. prems_end( );
	      ++ p )
      {
         repr. push_back(p);
      }
   }


   premisseset::premisseset( const premisseset& p )
      : repr( p. repr )
   {
   }


   void premisseset::operator = ( const premisseset& p )
   {
      repr = p. repr;
   }


   bool premisseset::insert( rule::prem_iterator p )
   {
      for( const_iterator q = begin( );
              q != end( );
	      ++ q )
      {
         if( *q == p )
            return false;
      }
      repr. push_back(p);
      return true;
   }


   bool premisseset::contains( rule::prem_iterator p ) const 
   {
      for( const_iterator q = begin( );
              q != end( );
	      ++ q )
      {
         if( *q == p )
            return true;
      }
      return false;
   }


   bool premisseset::remove( rule::prem_iterator p )
   {
      for( std::vector< rule::prem_iterator > :: iterator 
              q = repr. begin( );
	      q != repr. end( );
	      ++ q ) 
      {
         if( *q == p )
         {
            repr. erase(q);
	    return true;
         }
      }
      return false;
   }


   void premisseset::close( const rule& r ) 
   {
      unsigned int pos = 0;
      while( pos < repr. size( ))
      {
         for( rule::prem_iterator p = r. prems_begin( );
                 p !=  r. prems_end( );
	         ++ p )
         {
            if( sharevariables( *( repr [ pos ] ), *p ))
               insert( p );  
         }
         ++ pos; 
      }
   }


   void premisseset::close( const rule& r, const varset& avoid ) 
   {
      unsigned int pos = 0;

      while( pos < repr. size( ))
      {
         for( rule::prem_iterator p = r. prems_begin( );
                 p !=  r. prems_end( );
	         ++ p )
         {
            if( sharevariables( *( repr [ pos ] ), *p, avoid ))
               insert( p );  
         }
         ++ pos; 
      }
   }


   void premisseset::close( const rule& r, const varmatching& avoid )
   {
      unsigned int pos = 0;

      while( pos < repr. size( ))
      {
         for( rule::prem_iterator p = r. prems_begin( );
                 p !=  r. prems_end( );
	         ++ p )
         {
            if( sharevariables( *( repr [ pos ] ), *p, avoid ))
               insert( p );  
         }
         ++ pos; 
      }
   }


   void premisseset::close( const rule& r, const groundsubst& avoid )
   {
      unsigned int pos = 0;

      while( pos < repr. size( ))
      {
         for( rule::prem_iterator p = r. prems_begin( );
                 p != r. prems_end( );
	         ++ p )
         {
            if( sharevariables( *( repr [ pos ] ), *p, avoid ))
               insert( p );
         }
         ++ pos; 
      }
   }


   bool premisseset::containsvariable( varatom va, variable v )
   {
      for( std::vector< variable > :: const_iterator p = va. a1;
              p != va. a2;
	      ++ p )
      {
         if( *p == v )
            return true;
      }
      return false; 
   }


   bool premisseset::containsvariable( varatom va, const varset& vs )
   {
      for( std::vector< variable > :: const_iterator p = va. a1;
              p != va. a2;
	      ++ p )
      {
         if( vs. contains( *p ))
            return true;
      }
      return false; 
   }


   bool premisseset::sharevariables( varatom va1, varatom va2 )
   {
      for( std::vector< variable > :: const_iterator
	      p = va1. a1;
	      p != va1. a2;
	      ++ p )
      {
         for( std::vector< variable > :: const_iterator
                 q = va2. a1;
                 q != va2. a2;
                 ++ q )
         {
	    if( *p == *q )
	       return true;
         }
      }
      return false;
   }


   bool premisseset::sharevariables( varatom va1, varatom va2,
		                     const varset& avoid )
   {
      for( std::vector< variable > :: const_iterator
	      p = va1. a1;
	      p != va1. a2;
	      ++ p )
      {
         for( std::vector< variable > :: const_iterator
                 q = va2. a1;
                 q != va2. a2;
                 ++ q )
         {
	    if( *p == *q && ! avoid. contains( *p ))
	       return true;
         }
      }
      return false;
   }


   bool premisseset::sharevariables( varatom va1, varatom va2,
		                     const varmatching& avoid )
   {
      for( std::vector< variable > :: const_iterator
	      p = va1. a1;
	      p != va1. a2;
	      ++ p )
      {
         for( std::vector< variable > :: const_iterator
                 q = va2. a1;
                 q != va2. a2;
                 ++ q )
         {
	    if( *p == *q && avoid. lookup(*p) == 0 )
	       return true;
         }
      }
      return false;
   }


   bool premisseset::sharevariables( varatom va1, varatom va2,
		                     const groundsubst& avoid )
   {
      for( std::vector< variable > :: const_iterator
	      p = va1. a1;
	      p != va1. a2;
	      ++ p )
      {
         for( std::vector< variable > :: const_iterator
                 q = va2. a1;
                 q != va2. a2;
                 ++ q )
         {
	    if( *p == *q && avoid. lookup(*p) == 0 )
	       return true;
         }
      }
      return false;
   }


   premisseset operator + ( const premisseset& p1, 
		            const premisseset& p2 )
   {
      premisseset res = p1;
      for( premisseset::const_iterator p = p2. begin( );
              p != p2. end( );
	      ++ p )
      {
         res. insert(*p);
      }
      return res;
   }


   premisseset operator - ( const premisseset& p1,
		            const premisseset& p2 )
   {
      premisseset res;

      for( premisseset::const_iterator p = p1. begin( );
              p != p1. end( );
	      ++ p )
      {
         if( ! p2. contains( *p ))
            res. repr. push_back( *p );
      }
      return res;
   }


   std::ostream& operator << ( std::ostream& stream, const premisseset& s )
   {
      stream << "{ ";
      for( premisseset::const_iterator p = s. begin( );
              p != s. end( );
	      ++ p )
      {
         if( p != s. begin( ))
            stream << ", ";

         stream << *(*p);
      }
      stream << " }";
      return stream;
   }

}


