

// Written by Hans de Nivelle,
// February 2006.

// A premisseset is a set of premisses of a rule. 


#ifndef GL_PREMISSESET_INCLUDED
#define GL_PREMISSESET_INCLUDED   1


#include <vector>
#include <iostream>


#include "rule.h"
#include "varset.h"
#include "varmatching.h"
#include "groundsubst.h"


namespace gottlob_leitsch
{

   class premisseset
   {

      std::vector< rule::prem_iterator > repr;

   public:

      premisseset( );
         // Constructs empty premisse set.

      premisseset( const rule& r ); 
         // Constructs premisseset of the premisses of r.

      premisseset( const premisseset& p );
         // Copy constructor.

      void operator = ( const premisseset& p );
         // Copying assignment.

      bool insert( rule::prem_iterator p );
         // Insert p. True is returned if p is really added,
         // it did not occur already.

      bool contains( rule::prem_iterator p ) const;
         // True if p is contained. 

      bool remove( rule::prem_iterator p );
         // True if p was contained and really removed.

      void close( const rule& r ); 
         // Close the premisse set by iteratively adding all 
         // premisses that are connected through a shared variable. 

      void close( const rule& r, const varset& avoid );
         // Close the premisse set by iteratively adding all
         // premisses that are connected through a shared variable
         // which is not in avoid. 

      void close( const rule& r, const varmatching& avoid ); 
         // Same as previous, but avoiding variables in the
         // domain of avoid. 

      void close( const rule& r, const groundsubst& avoid );
         // Same as previous, but avoiding variables in the
         // domain of avoid.

      typedef std::vector< rule::prem_iterator > :: const_iterator 
	      const_iterator;

      const_iterator begin( ) const { return repr. begin( ); } 
      const_iterator end( ) const { return repr. end( ); }

      unsigned int size( ) const { return repr. size( ); }


      friend premisseset operator + ( const premisseset& p1, 
		                      const premisseset& p2 );

      friend premisseset operator - ( const premisseset& p1,
		                      const premisseset& p2 );

      friend std::ostream& operator << ( std::ostream& , const premisseset& );


   private:
      static bool containsvariable( varatom va, variable v );
      static bool containsvariable( varatom va, const varset& vs );

      static bool sharevariables( varatom va1, varatom va2 );
      static bool sharevariables( varatom va1, varatom va2, 
		                  const varset& avoid );
   
      static bool sharevariables( varatom va1, varatom va2, 
		                  const varmatching& avoid ); 
      static bool sharevariables( varatom va1, varatom va2,
		                  const groundsubst& avoid );

   };


   premisseset operator + ( const premisseset& p1,
		            const premisseset& p2 );

   premisseset operator - ( const premisseset& p1,
		            const premisseset& p2 );

   std::ostream& operator << ( std::ostream& , const premisseset& );

}


#endif


