

// Code written by Hans de Nivelle, March 2006. 


#include "premmatching.h"


namespace gottlob_leitsch
{

   const unsigned int* premmatching::lookup( rule::prem_iterator p ) const
   {

      for( std::vector< assignment > :: const_iterator 
	      q = repr. begin( );
              q != repr. end( );
	      q ++ )
      {
         if( ( q -> p ) == p )
         {
	    return & ( q -> val );
         }
      }
      return 0;
   }


   void premmatching::assign( rule::prem_iterator p, unsigned int val )
   {
      ASSERT( lookup(p) == 0 );
      repr. push_back( assignment::assignment( p, val ));
   }


   unsigned int premmatching::sup_atoms( ) const
   {
      unsigned int res = 0;
      for( std::vector< assignment > :: const_iterator
              q = repr. begin( );
	      q != repr. end( );
	      ++ q )
      {
         if( ( q -> val ) + 1 > res )
            res = ( q -> val ) + 1;
      }
      return res;
   }


   premisseset premmatching::getdomain( ) const
   {
      premisseset res;

      for( std::vector< assignment > :: const_iterator
              q = repr. begin( );
	      q != repr. end( );
	      ++ q )
      {
         res. insert( q -> p );
      }
      return res;
   }


   premisseset premmatching::getaddedpremisses( unsigned int x ) const 
   {
      premisseset res;

      ASSERT( x <= size( )); 

      for( std::vector< assignment > :: const_iterator
              q = repr. begin( ) + x;
	      q != repr. end( );
	      ++ q )
      {
         res. insert( q -> p );
      }
      return res;
   }


   bool premmatching::consistentwith( const premmatching& other ) const
   {
      for( std::vector< assignment > :: const_iterator 
              q = repr. begin( );
	      q != repr. end( );
	      ++ q )
      {
         const unsigned int* val = other. lookup( q -> p );  
         if( val && *val != ( q -> val ))
            return false; 
      }
      return true; 
   }


   premmatching operator + ( const premmatching& m1,
		             const premmatching& m2 ) 
   {
      premmatching res = m1;

      for( std::vector< premmatching::assignment > :: const_iterator
              q = m2. repr. begin( );
	      q != m2. repr. end( );
	      ++ q )
      {
         const unsigned int* val = m1. lookup( q -> p );
         ASSERT( ! val || *val == ( q -> val ));
         if( !val )
         {
            res. repr. push_back( 
		     premmatching::assignment::assignment( q -> p, q -> val ));
         }
      }
      return res; 
   }


   std::ostream& operator << ( std::ostream& stream,
		               const premmatching& s )
   {
      stream << "PremisseMatching( ";
      for( std::vector< premmatching::assignment > :: const_iterator 
	      q = s. repr. begin( );
              q !=  s. repr. end( );
	      ++ q )
      {
         if( q != s. repr. begin( ))
            stream << ",  "; 
         stream << *( q -> p ) << " ==> " << ( q -> val );
      }
      stream << " )";
      return stream;
   }

}

