

// Code written by Hans de Nivelle, March 2006.


// A premmisse matching assigns integers to the premisses
// of a rule. 


#ifndef GL_PREMMATCHING_INCLUDED
#define GL_PREMMATCHING_INCLUDED    1


#include <iostream>
#include <vector>


#include "rule.h"
#include "premisseset.h"


namespace gottlob_leitsch
{

   class premmatching
   {

      struct assignment
      {
         rule::prem_iterator p; 
         unsigned int val; 

         assignment( rule::prem_iterator p, unsigned int val ) 
            : p( p ), val( val ) 
         {
         }

      };


   private:
      std::vector< assignment > repr; 

   public:

      void clear( );

      unsigned int size( ) const;
         // Returns number of assigned premisses. 
         // the result can be used for backtracking. 

      const unsigned int* lookup( rule::prem_iterator ) const; 
         // Lookup returns 0 if there is no value. The 
         // pointer remains valid until the next non-const 
         // operation. 

      void backtrack( unsigned int );
         // One can take the size( ), and later backtrack 
         // to size( ). 

      void assign( rule::prem_iterator p, unsigned int val );
         // Assigns p := val. 
         // One should not reassign to a premisse that already 
         // has a value. 

      unsigned int sup_atoms( ) const;
         // Returns the supremum of the (numbers of) the atoms
         // into which a premisse is matched.

      premisseset getdomain( ) const;
         // Returns the domain of the premmatching as premisseset.

      premisseset getaddedpremisses( unsigned int ) const;
         // Get the premisses that have been added since last size( ). 

      bool consistentwith( const premmatching& other ) const; 
         // True if this premisse matching is consistent with the other
         // premisse matching. 

      friend premmatching operator + ( const premmatching& , 
		                       const premmatching& );

      friend std::ostream& operator << ( std::ostream&, const premmatching& );

   };


   inline void premmatching::clear( )
   {
      repr. clear( ); 
   }
   

   inline unsigned int premmatching::size( ) const
   {
      return repr. size( ); 
   }


   inline void premmatching::backtrack( unsigned int x ) 
   {
      ASSERT( x <= repr. size( ));
      while( x < repr. size( ))
      repr. pop_back( );
   }


   premmatching operator + ( const premmatching& ,
		             const premmatching& );

}


#endif


