val () = app load ["Moscow", "Problem", "Meson", "Metis"];

infixr |-> ::> @> oo;

val () = installPP Term.pp_term;
val () = installPP Term.pp_formula;
val () = installPP Subst.pp_subst;
val () = installPP Thm.pp_thm;
val () = installPP Clause.pp_clause;
val () = installPP Clause.pp_rewrs;
val () = installPP Clauseset.pp_clauseset;
val () = installPP Support.pp_sos;
val () = installPP Resolution.pp_resolution;
val () = quotation := true;

Useful.trace_level := 2;

val old_quietdec = !quietdec;
val () = quietdec := true;
open Useful Term Canon Thm Problem Meter Solver Meson Resolution Metis;
val () = quietdec := old_quietdec;

val time = Moscow.time;

(* ------------------------------------------------------------------------- *)
(* Helper functions.                                                         *)
(* ------------------------------------------------------------------------- *)

fun SAY s =
  print
  ("-------------------------------------" ^
   "-------------------------------------\n" ^ s ^ "\n\n");

fun printval p x = (print (PP.pp_to_string 79 p x ^ "\n\n"); x);

(* ------------------------------------------------------------------------- *)
val () = SAY "Peano's axioms";
(* ------------------------------------------------------------------------- *)

val peano = (axiomatize o parse_formula)
  `(!y. ~(suc y = 0)) /\
   (!x y. suc x = suc y ==> x = y) /\
   (!x. x + 0 = x) /\
   (!x y. x + suc y = suc (x + y)) /\
   (!y. y = 0 \/ y = suc (pre y)) /\
   (!y. pre (suc y) = y)`;

(* ------------------------------------------------------------------------- *)
val () = SAY "Integers";
(* ------------------------------------------------------------------------- *)

val integers = (axiomatize o list_mk_conj o map parse_formula)
  [(* Addition *)
   `(!x. x + 0 = x) /\
    (!x. x + ~x = 0) /\
    (!x y. x + y = y + x) /\
    (!x y z. x + y + z = x + (y + z))`,
   (* Multiplication *)
   `(!x. x * 1 = x) /\
    (!x. x * 0 = 0) /\
    (!x y. x * ~y = ~(x * y)) /\
    (!x y. x * y = y * x) /\
    (!x y z. x * y * z = x * (y * z)) /\
    (!x y z. x * (y + z) = x * y + x * z)`,
   (* Less than or equal to *)
   `~(1 <= 0) /\
    (!x. x <= x) /\
    (!x y. x <= y /\ y <= x ==> x = y) /\
    (!x y z. x <= y /\ y <= z ==> x <= z) /\
    (!x y. x <= y \/ y <= x) /\
    (!x y. x <= y ==> ~y <= ~x) /\
    (!x y v w. x <= y /\ v <= w ==> x + v <= y + w) /\
    (!x y z. 0 <= z /\ x <= y ==> x * z <= y * z)`];

(* ------------------------------------------------------------------------- *)
val () = SAY "Automatically generating theorems";
(* ------------------------------------------------------------------------- *)

val clause_parm =
  {literal_order = false,
   term_order = false,
   termorder_parm = Termorder.defaults};

val AX = map (Clause.mk_clause clause_parm) o axiomatize o parse_formula;

val integers' = map (Clause.mk_clause clause_parm) integers;

val pure = Resolution.new
  {restart = SOME 0,
   clause_parm = clause_parm,
   sos_parm = Support.defaults,
   set_parm = Clauseset.defaults};

val ints = Resolution.calibrate (fn n => n * n) (Resolution.add integers' pure);

datatype status =
  Satisfiable
| Unsatisfiable of clause
| Unknown of Resolution.resolution;

fun advance res =
  if not (Resolution.check res) then Unknown res else
    case Resolution.select res of NONE => Satisfiable
    | SOME (dcl_res as ((_,cl),_)) =>
      if Clause.is_empty cl then Unsatisfiable cl
      else advance (Resolution.deduce dcl_res);

fun disco res =
  let
    val () = print "\n---\n"
    val Unknown res = advance res
    val () = print "|"
    val {set,sos,...} = Resolution.dest res
    val sos = Support.reset sos
    val xres = Resolution.overlay {iter = 0, set = set, sos = sos} res
    val xres = Resolution.add (AX `0 <= n`) xres
    val () = print "\n"
    val Unknown xres = advance xres
  in
    case advance xres of Unsatisfiable cl => cl
    | _ => disco (Resolution.calibrate (fn n => 2 * n) (Resolution.restart res))
  end;

(*
Useful.trace_level := 4;
*)

disco ints;

val Unknown res = advance ints;
val {iter,set,sos,...} = Resolution.dest res;
Clauseset.rewrites set;
Clauseset.clauses set;
Support.to_list sos;

(*
val () = print (proof_to_string (proof (hd (Option.valOf (hd l)))));
*)
