------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Utils where

{-

This modules defines a number of general-purpose functions.

-}

-------------------------------------------------------------------------
-- searching for a 0 solution

search0 :: (Integer -> Integer) -> Integer -> Integer -> Integer
search0 f mn mx = srch mn mx
 where
  srch mn mx
    | mn+1 >= mx = mn
    | y < 0      = srch x mx
    | y > 0      = srch mn x
    | y == 0     = x
   where
    x = (mn+mx) `div` 2
    y = f x

-------------------------------------------------------------------------
-- justification

ljust, rjust :: Int -> String -> String
ljust k s = s ++ replicate (k - length s) ' '
rjust k s = replicate (k - length s) ' ' ++ s
 
-------------------------------------------------------------------------
-- 3-tuples

fst3 (x,_,_) = x
snd3 (_,x,_) = x
thd3 (_,_,x) = x

-------------------------------------------------------------------------
-- lists

cross :: [[a]] -> [[a]]
cross []       = [[]]
cross (xs:xss) = [ x:ys | x <- xs, ys <- cross xss ]

holes :: [a] -> [[a]]
holes []     = []
holes (x:xs) = xs : [ x : hxs | hxs <- holes xs ]

number :: (a -> Bool) -> [a] -> Int
number p = length . filter p

-------------------------------------------------------------------------
-- the end.
