/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Atom.hpp"
#include "Tracer.hpp"
#include "ShellReader.hpp"
#include "Signature.hpp"
#include "XML.hpp"
using namespace Shell;
Atom::Atom (const PredicateSymbol& p, const TermList& args)
  :
  _data (new Data (p,args))
{
} 
Atom::Atom (const Term& lhs, const Term& rhs)
  :
  _data (new Data (lhs,rhs))
{
} 
Atom::Atom (const PredicateSymbol& p)
  :
  _data (new Data (p))
{
} 
AtomList::AtomList (const Atom& a)
  : Lst<Atom> ()
{
} 
Atom::Data::Data (const Term& l, const Term& r) 
  :
  _counter (1),
  _functor (Signature::sig.equality()),
  _args (TermList (l, TermList(r)))
{
} 
void Atom::operator = (const Atom& t)
{
  if (t._data) {
    t._data->ref ();
  }
  if (_data) {
    _data->deref ();
  }
  _data = t._data;
} 
void Atom::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;
  if (_counter == 0) {
    delete this;
  }
} 
bool Atom::equals (const Atom& a) const
{
  return functor() == a.functor() &&
         args().equals(a.args());
} 
XML::Element* Atom::toXML() const
{
  if (args().isEmpty()) {
    XML::Element* a = new XML::EmptyElement("propositional_constant");
    a->addAttribute(new XML::TextAttribute("name",functor().name()));
    return a;
  }
  XML::DeepElement* a = new XML::DeepElement("atom");
  a->addAttribute(new XML::TextAttribute("functor",functor().name()));
  Iterator<Term> ts(args());
  while ( ts.hasNext() ) {
    a->addChild(ts.next().toXML());
  }
  return a;
} 
string Atom::toString () const
{
  Reader reader;
  return reader.toString(reinterpret_cast<const Vampire::Atom*>(this));
} 
