/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "AtomIterator.hpp"
#include "Tracer.hpp"
#include "Formula.hpp"
namespace Shell {
class AtomIterator::Element 
{
public:
  Element (const FormulaList& list, Element* rest) 
    : _isList(true), _list(const_cast<FormulaList*>(&list)), _rest(rest) {}
  Element (const Formula& f, Element* rest) 
    : _isList(false), _formula(const_cast<Formula*>(&f)), _rest(rest) {}
  bool _isList;
  union{ 
    FormulaList* _list;
    Formula* _formula;
  };
  Element* _rest;
};
void AtomIterator::popReserve ()
{
  TRACER("AtomIterator::popReserve()");
  ASS(_reserve);
  Element* next = _reserve->_rest;
  delete _reserve;
  _reserve = next;
} 
AtomIterator::AtomIterator (const Formula& t)
  : _current(const_cast<Formula*>(&t)),
    _reserve(0)
{
} 
AtomIterator::AtomIterator (const FormulaList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 
bool AtomIterator::hasNext ()
{
  TRACER("AtomIterator::hasNext");
  for (;;) {
    if (! _current) {
      if (! _reserve) {
	return false;
      }
      if (_reserve->_isList) {
	FormulaList* first = _reserve->_list;
	if (first->isEmpty()) {
	  popReserve();
	  continue;
	}
	else { 
	  _current = const_cast<Formula*>(&(first->head()));
	  _reserve->_list = const_cast<FormulaList*>(&(first->tail()));
	}
      }
      else { 
	_current = _reserve->_formula;
	popReserve();
      }
    }
    switch (_current->connective()) {
    case ATOM:
      return true;
    case AND:
    case OR:
      _reserve = new Element(_current->args().tail(),_reserve);
      _current = const_cast<Formula*>(&_current->args().head());
      break;
    case IMP:
    case IFF:
    case XOR:
      _reserve = new Element(_current->right(),_reserve);
      _current = const_cast<Formula*>(&_current->left());
      break;
    case NOT:
      _current = const_cast<Formula*>(&_current->uarg());
      break;
    case FORALL:
    case EXISTS:
      _current = const_cast<Formula*>(&_current->qarg());
      break;
    case FALSE:
    case TRUE:
      _current = 0;
      break;
#if DEBUG_SHELL
    default:
      ASS(false);
#endif
    }
  }
} 
const Atom& AtomIterator::next ()
{
  ASS(_current->connective() == ATOM);
  const Atom& a = _current->atom();
  _current = 0;
  return a;
} 
AtomIterator::~AtomIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
}
