/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BIT_STRING_HPP
#define BIT_STRING_HPP
#include <cstddef>                                   
#include <climits>                                   
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "BitWord.hpp" 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BIT_STRING
#  define DEBUG_NAMESPACE "BitString<class Alloc>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc>
class BitString
{
 private:
  BitString() {};
 public:
  BitString(size_t s) : sz(s) {};
  ~BitString() {};
  void* operator new(size_t,size_t s) 
  {
   CALL("operator new(size_t,size_t s)");
    return Alloc::allocate(sizeof(BitString) + sizeof(BitWord)*s);
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)");
    Alloc::deallocate(obj,sizeof(BitString) + sizeof(BitWord)*((static_cast<BitString*>(obj))->size()));
  };
  void operator delete(void* obj,size_t) 
  {
   CALL("operator delete(void* obj,size_t)");
    Alloc::deallocate(obj,sizeof(BitString) + sizeof(BitWord)*((static_cast<BitString*>(obj))->size()));
  };
  size_t size() const { return sz; }; 
  static ulong minSize() { return sizeof(BitString); }; 
  static ulong maxSize() { return ULONG_MAX; }; 
  const BitWord& word(size_t wn) const { return *(begin() + wn); };
  BitWord& word(size_t wn) { return *(begin() + wn); };
  const BitWord* begin() const 
  { 
    return 
      static_cast<const BitWord*>(static_cast<const void*>(this + 1)); 
  };  
  BitWord* begin() { return static_cast<BitWord*>(static_cast<void*>(this + 1)); }; 
 private:   
  void* operator new(size_t) { return 0; };
  BitString(const BitString&) {};
  BitString& operator=(const BitString&) { return *this; };
 private:
  size_t sz;
}; 
}; 
#endif
