/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef COMPACT_NODE_SMALL_HEADER_SKIP_LIST_HPP
#define COMPACT_NODE_SMALL_HEADER_SKIP_LIST_HPP 
#include <climits>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPACT_NODE_SMALL_HEADER_SKIP_LIST
 #define DEBUG_NAMESPACE "CNSHSkList<class Alloc,class ValType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class ValType,class KeyType>
class CNSHSkList
{
 public:
  class Surfer;
  class Node
  {
   public:
    ~Node() {};
    KeyType key() const 
    {
     CALL("key() const"); 
     ASSERT(checkObj());
     return _key; 
    };
    const ValType& value() const { return _value; };
    ValType& value() { return _value; };
    Node* next() { return stripJump(*(jump())); };
    const Node* next() const { return stripJump(*(jump())); };    
   private:
    static bool isHighestJump(const Node* ptr) 
    {
      return (reinterpret_cast<ulong>(ptr) << 30) != 0UL;
    };
    static Node* markAsHighestJump(Node* ptr)
    {
      return 
	reinterpret_cast<Node*>(reinterpret_cast<ulong>(ptr) | 0x00000001);
    };
    static const Node* markAsHighestJump(const Node* ptr)
    {
      return 
	reinterpret_cast<const Node*>(reinterpret_cast<ulong>(ptr) | 0x00000001);
    };
    static Node* stripJump(Node* ptr)
    {
      return 
        reinterpret_cast<Node*>((reinterpret_cast<ulong>(ptr) >> 1) << 1);
    };
    static const Node* stripJump(const Node* ptr)
    {
      return 
        reinterpret_cast<const Node*>((reinterpret_cast<ulong>(ptr) >> 1) << 1);
    };
    ulong computeHeight() const
    {
      CALL("Node::computeHeight() const");
      const Node* const * jmp = jump();
      while (!isHighestJump(*jmp))
	{
	  ++jmp;
	};
      return jmp - jump();
    }; 
    Node** jump() { return static_cast<Node**>(static_cast<void *>(this + 1)); }; 
    const Node* const * jump() const { return static_cast<const Node* const *>(static_cast<const void *>(this + 1)); };
    Node(const KeyType& k) : _key(k) {};
    static const Node* node(const Node * const * jmp)  
    {
     return (static_cast<const Node*>(static_cast<const void*>(jmp))) - 1;
    };
    static Node* node(Node ** jmp)           
    {
      return (static_cast<Node*>(static_cast<void*>(jmp))) - 1;
    };
    void* operator new(size_t size,ulong ht) 
    {    
     CALL("operator new(size_t size,ulong ht)"); 
      return Alloc::allocate(size + (ht+1)*sizeof(Node*));
    };
    void operator delete(void* obj)
    {
     CALL("operator delete(void* obj)");
      Alloc::deallocate(obj,sizeof(Node) + ((static_cast<Node*>(obj))->computeHeight() + 1)*sizeof(Node*));
    };   
    void operator delete(void* obj,ulong ht)
    {
     CALL("operator delete(void* obj,ulong ht)");
      Alloc::deallocate(obj,sizeof(Node) + (ht + 1)*sizeof(Node*));
    };   
    #ifndef NO_DEBUG 
     bool checkObj() const
     {
      #ifndef DEBUG_ALLOC
       return true;
      #else 
       ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<const void*>(this));
       if (!objDesc) 
        { 
         DMSG("CNSHSkList<..>::Node::checkObj()> Null pointer to object descriptor.\n");  
         return false; 
        }; 
       bool res = true;   
       #ifdef DEBUG_ALLOC_OBJ_SIZE
        if (objDesc->size !=  (sizeof(Node) + (computeHeight() + 1)*sizeof(Node*)))
         {
          DMSG("CNSHSkList<..>::Node::checkObj()> objDesc->size != (sizeof(Node) + (computeHeight() + 1)*sizeof(Node*)) \n");
          res = false;
         };
       #endif
       #ifdef DEBUG_ALLOC_OBJ_STATUS
        if (objDesc->status != ObjDesc::Persistent) 
         {
          DMSG("CNSHSkList<..>::Node::checkObj()> objDesc->status != ObjDesc::Persistent\n");
          res = false;
         };
       #endif
       if (!res) DOP(DMSG("CNSHSkList<..>::Node::checkObj()> host object ") << *objDesc << "\n"); 
       return res;
      #endif
     };
    #endif
   private:  
    KeyType _key;
    ValType _value;
   friend class CNSHSkList<Alloc,ValType,KeyType>;
   friend class Surfer;
  }; 
  class Surfer
  {
   public:
    Surfer() {};
    Surfer(const CNSHSkList& sl) { reset(sl); };
    ~Surfer() {};
    void reset(const CNSHSkList& sl) 
    {
     CALL("Surfer::reset(const CNSHSkList& sl)");
     ASSERT(sl.checkIntegrity());
     _currMaxLevel = sl._currMaxLevel;
     const Node* const * top = sl._topIndex;
     _currNode = *top;
     ASSERT(!Node::isHighestJump(_currNode));
     if (_currNode)
      {	
	for (ulong lv = 1UL; lv <= _currMaxLevel; ++lv)
	  _jump[lv] = top[lv];
      };
     ASSERT(checkIntegrity());
    }; 
    const Node* currNode() const { return _currNode; };
    const Node* const * currNodeAddr() const { return &_currNode; }; 
    bool next()
    {
     CALL("Surfer::next()");
     ASSERT_IN(checkIntegrity(),"START");
     _currNode = _currNode->next();
     ASSERT_IN(checkIntegrity(),"END");
     return (_currNode != 0);
    }; 
    bool find(const KeyType searchKey)
    {
      CALL("Surfer::find(const KeyType& searchKey)");
      ASSERT(checkIntegrity());
      if (_currNode->key() <= searchKey) 
	{
	  return _currNode->key() == searchKey;
	};
      const Node* node;
      const Node* nextNode;
      ulong lv = _currMaxLevel;
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_no_overjumps_yet:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      if (!nextNode)
		{
		  --lv;
		  _currMaxLevel = lv;
		  if (lv) goto jump_no_overjumps_yet;
		  _currNode = node;
		  goto level_zero_no_overjumps_yet;
		};
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_no_overjumps_yet;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return true;
	}
      else 
	{
	level_zero_no_overjumps_yet:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  if ((!_currNode) || (_currNode->key() < searchKey)) 	    
	    return false;
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  goto level_zero_no_overjumps_yet;
	};
    check_lv_overjump_exists:
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_overjump_exists:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      ASSERT(nextNode);
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_overjump_exists;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return true;
	}
      else 
	{
	level_zero_overjump_exists:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  ASSERT(_currNode);
	  if (_currNode->key() < searchKey) 	    
	    return false;
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  goto level_zero_overjump_exists;
	};
    }; 
    void getTo(const KeyType searchKey)
    {
      CALL("Surfer::getTo(const KeyType& searchKey)");
      ASSERT(checkIntegrity());
      ASSERT(_currNode->key() >= searchKey);
      if (_currNode->key() == searchKey) 
	{
	  return;
	};
      const Node* node;
      const Node* nextNode;
      ulong lv = _currMaxLevel;
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_no_overjumps_yet:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      if (!nextNode)
		{
		  --lv;
		  _currMaxLevel = lv;
		  if (lv) goto jump_no_overjumps_yet;
		  _currNode = node;
		  goto level_zero_no_overjumps_yet;
		};
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_no_overjumps_yet;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return;
	}
      else 
	{
	level_zero_no_overjumps_yet:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() >= searchKey);
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return;
	    };
	  goto level_zero_no_overjumps_yet;
	};
    check_lv_overjump_exists:
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_overjump_exists:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      ASSERT(nextNode);
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_overjump_exists;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return;
	}
      else 
	{
	level_zero_overjump_exists:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() >= searchKey);
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return;
	    };
	  goto level_zero_overjump_exists;
	};
    }; 
    #ifdef DEBUG_NAMESPACE
     bool checkIntegrity()
     {
      if (!_currNode) return true;
      ulong currNodeHeight = _currNode->computeHeight();
      if (currNodeHeight > _currMaxLevel) 
       {
        DMSG("CNSHSkList<..>::Surfer::checkIntegrity()> currNodeHeight > _currMaxLevel\n");
        return false;  
       };
      for (ulong lv = 1UL; lv <= _currMaxLevel; ++lv)
	{
	  if (Node::isHighestJump(_jump[lv]))
	    {
	      DMSG("CNSHSkList<..>::Surfer::checkIntegrity()> Node::isHighestJump(_jump[lv])\n");
	      return false;
	    };
	  if (!_jump[lv])
	    {
	      DMSG("CNSHSkList<..>::Surfer::checkIntegrity()> !_jump[lv]\n");
	      return false;
	    };
	  if (lv <= currNodeHeight)
	    {
	      if (_jump[lv]->key() < _currNode->key())
		{
		  DMSG("CNSHSkList<..>::Surfer::checkIntegrity()> _jump[lv]->key() < _currNode->key()\n");
		  return false;
		};
	    }
	  else
	    {
	      if (lv > 1UL)
		{
		  if (_jump[lv]->key() > _jump[lv-1]->key())
		    {
		      DMSG("CNSHSkList<..>::Surfer::checkIntegrity()> _jump[lv]->key() > _jump[lv-1]->key()\n");
		      return false;
		    };
		};
	    };
	};
      return true;
     };
    #endif    
   private:
    ulong _currMaxLevel;
    const Node* _jump[28]; 
    const Node* _currNode;
  }; 
 public:
  CNSHSkList() 
   {
     CALL("constructor CNSHSkList()"); 
     _currMaxLevel = 0;
     _sizeOfTopIndex = 1;
     _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*));
     _topIndex[0] = (Node*)0;
     _insertCount = 0;
     _size = 0;
     ASSERT(checkIntegrity());
   }; 
  ~CNSHSkList() 
   {
    CALL("destructor ~CNSHSkList()");
    ASSERT(checkIntegrity());
    if (DestructionMode::isThorough())
      {
	Node* node = _topIndex[0];
	Node* tmp;
	while (node) 
	  { 
	    tmp = node;   
	    node = node->next(); 
	    delete tmp;
	  };  
	Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
      };
    BK_CORRUPT(_topIndex);
   }; 
  void* operator new(size_t) 
  {
   CALL("operator new(size_t)"); 
    return Alloc::allocate(sizeof(CNSHSkList));
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)");
    Alloc::deallocate(obj,sizeof(CNSHSkList));
  };
  void reset() 
  {
   CALL("reset()");
   ASSERT(checkIntegrity());
   Node* node = _topIndex[0];
   Node* tmp;
   while (node) 
     { 
       tmp = node;   
       node = node->next(); 
       delete tmp;
     };  
   Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
   _currMaxLevel = 0;
   _sizeOfTopIndex = 1;
   _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*));
   _topIndex[0] = (Node*)0;
   _insertCount = 0;
   _size = 0;
   ASSERT(checkIntegrity());	
  }; 
  operator bool() const 
  {
   CALL("operator bool() const"); 
   return _topIndex[0] != (Node*)0; 
  }; 
  const Node* first() const 
  {
   CALL("first() const"); 
   return _topIndex[0]; 
  };
  Node* first() 
  {
   CALL("first()");
   return _topIndex[0]; 
  };
  ulong size() const { return _size; }; 
#ifndef NO_DEBUG
  ulong count() const 
  {   
   ulong res = 0; 
   for (const Node* n = first(); n; n = n->next()) 
    { 
     res++;
    };
   return res;
  };
#endif
  Node* insert(const KeyType& searchKey,bool& newNode);
  bool remove(const KeyType& searchKey,ValType& val);
  bool remove(const KeyType& searchKey);
  #ifndef NO_DEBUG
   bool checkIntegrity() const
   {
    if (!_topIndex) return false; 
    #ifdef DEBUG_ALLOC
     bool res = true;
     ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<void*>(_topIndex));          
     if (!objDesc) 
      { 
       DMSG("CNSHVoidSkList<..>::checkIntegrity()> Null pointer to object descriptor for *_topIndex.\n");  
       return false; 
      };     
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size !=  _sizeOfTopIndex * sizeof(Node*)) 
       {
        DMSG("CNSHVoidSkList<..>::checkIntegrity()> Wrong size of *_topIndex.\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent) 
       {
        DMSG("CNSHVoidSkList<..>::checkIntegrity()> objDesc->status != ObjDesc::Persistent for *_topIndex.\n");
        res = false; 
       };
     #endif      
     if (!res) return false;
    #endif
     for (ulong lv = 1UL; lv <= _currMaxLevel; ++lv)
       {
	 if (Node::isHighestJump(_topIndex[lv]))
	   {
	     DMSG("CNSHVoidSkList<..>::checkIntegrity()> Node::isHighestJump(_topIndex[lv])\n");
	     return false;
	   };
	 if (!_topIndex[lv])
	   {
	     DMSG("CNSHVoidSkList<..>::checkIntegrity()> !_topIndex[lv]\n");
	     return false;
	   };
	 if (_topIndex[lv]->key() > _topIndex[lv-1]->key())
	   {
	     DMSG("CNSHVoidSkList<..>::checkIntegrity()> _topIndex[lv]->key() > _topIndex[lv-1]->key()\n");
	     return false;
	   };
       };
     if (count() != size())
       {
	 DMSG("CNSHVoidSkList<..>::checkIntegrity()> count() != size()\n");
	 return false;
       };
    return true; 
   }; 
  #endif
  ostream& output(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     node->output(ostr,_currMaxLevel) << "\n";
    };
   return ostr;
  };
  ostream& outputKeys(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     ostr << node->key() << "\n";
    };
   return ostr;
  };
 private:
  CNSHSkList(const CNSHSkList& sl);
  CNSHSkList& operator=(const CNSHSkList& sl);
  ulong randomHeight() const
  { 
   ulong res = 0;
   ulong tmp = _insertCount;
   while (!(tmp % 2))
    { 
     res++;
     tmp = tmp/2;
    };
   return res;
  }; 
 private:
  ulong _sizeOfTopIndex;
  Node** _topIndex;
  ulong _currMaxLevel;
  ulong _insertCount;
  ulong _size;
 friend class Surfer;   
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPACT_NODE_SMALL_HEADER_SKIP_LIST
 #define DEBUG_NAMESPACE "CNSHSkList<class Alloc,class ValType,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class ValType,class KeyType>
inline
typename
CNSHSkList<Alloc,ValType,KeyType>::Node* 
CNSHSkList<Alloc,ValType,KeyType>::insert(const KeyType& searchKey,bool& newNode)  
{
 CALL("insert(const KeyType& searchKey,bool& newNode)");
 ASSERT_IN(checkIntegrity(),"START");
 Node* update[32];
 Node* currNode = Node::node(_topIndex);      
 DOP(Node* debugPseudonode = Node::node(_topIndex));
 ulong currLevel = _currMaxLevel;
 Node* nextNode = currNode->jump()[currLevel];  
 ASSERT(!Node::isHighestJump(nextNode));
 KeyType nodeKey;
 check_next_node:
  if (nextNode)
   {
     ASSERT(nextNode->computeHeight() <= currLevel);
    nodeKey = nextNode->key();
    if (nodeKey > searchKey) 
     {
      currNode = nextNode;
      nextNode = Node::stripJump(nextNode->jump()[currLevel]);        
      goto check_next_node;
     };
    if (nodeKey == searchKey) 
      { 
	newNode = false;
	ASSERT(checkIntegrity()); 
	return nextNode; 
      };
    update[currLevel] = currNode;
    while (currLevel) 
     { 
      currLevel--;
      nextNode = currNode->jump()[currLevel]; 
      ASSERT(!Node::isHighestJump(nextNode));
      nodeKey = nextNode->key();
      while (nodeKey > searchKey)
       {
	ASSERT(nextNode->computeHeight() == currLevel);
	currNode = nextNode;
        nextNode = Node::stripJump(nextNode->jump()[currLevel]);
        nodeKey = nextNode->key();
       };
      if (nodeKey == searchKey) 
	{ 
	  newNode = false;
	  ASSERT(checkIntegrity()); 
	  return nextNode; 
	};
      update[currLevel] = currNode;
     };
    ASSERT(nextNode->key() < searchKey);
   }
  else 
   {
    update[currLevel] = currNode;
    if (currLevel) 
     {
      currLevel--;
      nextNode = currNode->jump()[currLevel];
      ASSERT(!Node::isHighestJump(nextNode));
      goto check_next_node;
     };
   };
  ASSERT(!currLevel);    
  ASSERT(_insertCount < ULONG_MAX);
  _insertCount++;  
  _size++;
  currLevel = randomHeight();
  currNode = new (currLevel) Node(searchKey); 
  Node** nodeJump = currNode->jump();  
  Node** jumpAddr; 
  Node* jump;
  if (currLevel > _currMaxLevel)
   {
     Node** newTopIndex = (Node**)Alloc::allocate((currLevel + 1) * sizeof(Node*));
    ulong lev;
    for (lev = 0UL; lev <= _currMaxLevel; ++lev)
      newTopIndex[lev] = _topIndex[lev];
     Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));  
    Node* pseudonode = Node::node(_topIndex);      
    for (lev = 0UL; lev <= _currMaxLevel; ++lev)
      {
        ASSERT(!Node::isHighestJump(update[lev]));
	if (update[lev] == pseudonode)      
	  {
	    ASSERT((!newTopIndex[lev]) || (newTopIndex[lev]->key() < searchKey));
	    nodeJump[lev] = newTopIndex[lev];
	    newTopIndex[lev] = currNode;
	  }
	else
	  {
	    jumpAddr = update[lev]->jump() + lev;
	    jump = *jumpAddr;	
	    if (Node::isHighestJump(jump))
	      {
		ASSERT(update[lev]->computeHeight() == lev);
		*jumpAddr = Node::markAsHighestJump(currNode);	
		nodeJump[lev] = Node::stripJump(jump);
	      }
	    else
	      {
		ASSERT(update[lev]->computeHeight() > lev);
		*jumpAddr = currNode;	
		nodeJump[lev] = jump;
	      };
	  };
      }; 
    _sizeOfTopIndex = currLevel + 1;
    _topIndex = newTopIndex;
    ++_currMaxLevel;
    while (_currMaxLevel < currLevel)
     {
      _topIndex[_currMaxLevel] = currNode;
      nodeJump[_currMaxLevel] = 0;
      ++_currMaxLevel;
     };
    _topIndex[_currMaxLevel] = currNode;
    nodeJump[_currMaxLevel] = Node::markAsHighestJump(static_cast<Node*>(0));
    ASSERT(currNode == debugPseudonode || currNode->computeHeight() == _currMaxLevel);
   }
  else 
   {       
     ulong lev;
     for (lev = 0UL; lev < currLevel; ++lev)
       {
	 ASSERT(!Node::isHighestJump(update[lev]));
	 jumpAddr = update[lev]->jump() + lev;
	 jump = *jumpAddr;		
	 if (Node::isHighestJump(jump))
	   {
	     ASSERT(update[lev] != debugPseudonode);
	     ASSERT(update[lev]->computeHeight() == lev);
	     *jumpAddr = Node::markAsHighestJump(currNode);
	     nodeJump[lev] = Node::stripJump(jump);
	   }
	 else
	   {
	     ASSERT(update[lev] == debugPseudonode || update[lev]->computeHeight() > lev);
	     *jumpAddr = currNode;
	     nodeJump[lev] = jump;
	   };
       };
     ASSERT(!Node::isHighestJump(update[lev]));
     jumpAddr = update[lev]->jump() + lev;
     jump = *jumpAddr;		
     if (Node::isHighestJump(jump))
       {
	 ASSERT(update[lev] != debugPseudonode);
	 ASSERT(update[lev]->computeHeight() == lev);
	 *jumpAddr = Node::markAsHighestJump(currNode);
	 nodeJump[lev] = jump;
       }
     else
       {
	 ASSERT(update[lev] == debugPseudonode || update[lev]->computeHeight() > lev);
	 *jumpAddr = currNode;
	 nodeJump[lev] = Node::markAsHighestJump(jump);
       };
     ASSERT(currNode == debugPseudonode || currNode->computeHeight() == currLevel);
   }; 
  newNode = true;
  ASSERT_IN(checkIntegrity(),"END");
  return currNode;
}; 
template <class Alloc,class ValType,class KeyType>
inline
bool CNSHSkList<Alloc,ValType,KeyType>::remove(const KeyType& searchKey,ValType& val) 
{
  CALL("remove(const KeyType& searchKey,ValType& val)");
 ASSERT(checkIntegrity());
 Node** jumpAddr = _topIndex;
 bool highest = false;
 ulong currLevel = _currMaxLevel;  
 Node* node;
 next_node:   
  node = jumpAddr[currLevel];
  ASSERT(highest == Node::isHighestJump(node));
  if (highest) node = Node::stripJump(node);
  if (node)
   {
    ASSERT(node->computeHeight() >= currLevel);  
    if (node->key() == searchKey)
     { 
       if (highest)
	 {
	   jumpAddr[currLevel] = Node::markAsHighestJump(node->jump()[currLevel]);
	 }
       else
	 {
	   jumpAddr[currLevel] = Node::stripJump(node->jump()[currLevel]);
	 };
      if (currLevel) 
	{ 
	  currLevel--;
	  highest = false;
	  goto next_node; 
	};
      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel)
	_currMaxLevel--; 
      val = node->value();
      delete node;
      _size--;
      ASSERT(checkIntegrity());
      return true;
     };
    if (node->key() > searchKey) 
     {
      jumpAddr = node->jump();
      highest = true;
      goto next_node;
     };
    if (currLevel) 
      { 
	currLevel--;	
	highest = false;
	goto next_node; 
      }; 
    ASSERT(checkIntegrity());   
    return false;
   };
  if (currLevel) 
    { 
      currLevel--;      
      highest = false;
      goto next_node; 
    };
  ASSERT(checkIntegrity());
  return false;
}; 
template <class Alloc,class ValType,class KeyType>
inline
bool CNSHSkList<Alloc,ValType,KeyType>::remove(const KeyType& searchKey) 
{
 CALL("remove(const KeyType& searchKey)");
 ASSERT(checkIntegrity());
 Node** jumpAddr = _topIndex;
 bool highest = false;
 ulong currLevel = _currMaxLevel; 
 Node* node;
 next_node:   
  node = jumpAddr[currLevel];
  ASSERT(highest == Node::isHighestJump(node));
  if (highest) node = Node::stripJump(node);
  if (node)
   {
    ASSERT(node->computeHeight() >= currLevel);  
    if (node->key() == searchKey)
     { 
       if (highest)
	 {
	   jumpAddr[currLevel] = Node::markAsHighestJump(node->jump()[currLevel]);
	 }
       else
	 {
	   jumpAddr[currLevel] = Node::stripJump(node->jump()[currLevel]);
	 };
      if (currLevel) 
	{ 
	  currLevel--;
	  highest = false;
	  goto next_node; 
	};
      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel)
	_currMaxLevel--; 
      delete node;
      _size--;
      ASSERT(checkIntegrity());
      return true;
     };
    if (node->key() > searchKey) 
     {
      jumpAddr = node->jump();
      highest = true;
      goto next_node;
     };
    if (currLevel) 
      { 
	currLevel--;	
	highest = false;
	goto next_node; 
      }; 
    ASSERT(checkIntegrity());    
    return false;
   };
  if (currLevel) 
    { 
      currLevel--;      
      highest = false;
      goto next_node; 
    };
  ASSERT(checkIntegrity());
  return false;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPACT_NODE_SMALL_HEADER_SKIP_LIST
 #define DEBUG_NAMESPACE "CNSHVoidSkList<class Alloc,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class KeyType>
class CNSHVoidSkList
{
 public:
  class Surfer;
  class Node
  {
   public:
    ~Node() {};
    KeyType key() const 
    {
     CALL("key() const"); 
     ASSERT(checkObj());
     return _key; 
    };
    Node* next() { return stripJump(*(jump())); };
    const Node* next() const { return stripJump(*(jump())); };  
   private:
    static bool isHighestJump(const Node* ptr) 
    {
      return (reinterpret_cast<ulong>(ptr) << 30) != 0UL;
    };
    static Node* markAsHighestJump(Node* ptr)
    {
      return 
	reinterpret_cast<Node*>(reinterpret_cast<ulong>(ptr) | 0x00000001);
    };
    static const Node* markAsHighestJump(const Node* ptr)
    {
      return 
	reinterpret_cast<const Node*>(reinterpret_cast<ulong>(ptr) | 0x00000001);
    };
    static Node* stripJump(Node* ptr)
    {
      return 
        reinterpret_cast<Node*>((reinterpret_cast<ulong>(ptr) >> 1) << 1);
    };
    static const Node* stripJump(const Node* ptr)
    {
      return 
        reinterpret_cast<const Node*>((reinterpret_cast<ulong>(ptr) >> 1) << 1);
    };
    ulong computeHeight() const
    {
      CALL("Node::computeHeight() const");  
      const Node* const * jmp = jump();
      while (!isHighestJump(*jmp))
	{
	  ++jmp;
	};
      return jmp - jump();
    };
    Node** jump() { return (Node**)(this + 1); }; 
    const Node* const * jump() const { return (const Node* const *)(this + 1); };
    Node(const KeyType& k) :  _key(k) {};
    static const Node* node(const Node * const * jmp)  
    {
     return ((Node*)jmp) - 1;
    };
    static Node* node(Node ** jmp)           
    {
     return ((Node*)jmp) - 1;
    };
    void* operator new(size_t,ulong ht) 
    {    
     CALL("operator new(size_t,ulong ht)");  
      void* res = Alloc::allocate(sizeof(Node) + (ht+1)*sizeof(Node*));
      return res;
    };
    void operator delete(void* obj)
    {
     CALL("operator delete(void* obj)");
      Alloc::deallocate(obj,sizeof(Node) + ((static_cast<Node*>(obj))->computeHeight() + 1)*sizeof(Node*));
    };   
    void operator delete(void* obj,ulong ht)
    {
     CALL("operator delete(void* obj,ulong ht)");
      Alloc::deallocate(obj,sizeof(Node) + (ht + 1)*sizeof(Node*));
    };   
    #ifndef NO_DEBUG 
     bool checkObj() const
     {
      #ifndef DEBUG_ALLOC
       return true;
      #else 
       ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<const void*>(this));
       if (!objDesc) 
        { 
         DMSG("CNSHVoidSkList<..>::Node::checkObj()> Null pointer to object descriptor.\n");  
         return false; 
        }; 
       bool res = true;   
       #ifdef DEBUG_ALLOC_OBJ_SIZE
        if (objDesc->size !=  (sizeof(Node) + (computeHeight() + 1)*sizeof(Node*)))
         {
          DMSG("CNSHVoidSkList<..>::Node::checkObj()> objDesc->size != (sizeof(Node) + (computeHeight() + 1)*sizeof(Node*)) \n");
          res = false;
         };
       #endif
       #ifdef DEBUG_ALLOC_OBJ_STATUS
        if (objDesc->status != ObjDesc::Persistent) 
         {
          DMSG("CNSHVoidSkList<..>::Node::checkObj()> objDesc->status != ObjDesc::Persistent\n");
          res = false;
         };
       #endif
       if (!res) DOP(DMSG("CNSHVoidSkList<..>::Node::checkObj()> host object ") << *objDesc << "\n"); 
       return res;
      #endif
     };
    #endif
   private:
    KeyType _key;  
   friend class CNSHVoidSkList<Alloc,KeyType>;
   friend class Surfer;
  }; 
  class Surfer
  {
   public:
    Surfer() {};
    Surfer(const CNSHVoidSkList& sl) { reset(sl); };
    ~Surfer() {};
    void reset(const CNSHVoidSkList& sl) 
    {
     CALL("Surfer::reset(const CNSHVoidSkList& sl)");
     ASSERT(sl.checkIntegrity());
     _currMaxLevel = sl._currMaxLevel;
     const Node* const * top = sl._topIndex;
     _currNode = *top;
     ASSERT(!Node::isHighestJump(_currNode));
     if (_currNode)
      {	
	for (ulong lv = 1UL; lv <= _currMaxLevel; ++lv)
	  _jump[lv] = top[lv];
      };
     ASSERT(checkIntegrity());
    }; 
    const Node* currNode() const { return _currNode; };
    const Node* const * currNodeAddr() const { return &_currNode; }; 
    bool next()
    {
     CALL("Surfer::next()");
     ASSERT_IN(checkIntegrity(),"START");
     _currNode = _currNode->next();
     ASSERT_IN(checkIntegrity(),"END");
     return (_currNode != 0);
    }; 
    bool find(const KeyType searchKey)
    {
      CALL("Surfer::find(const KeyType& searchKey)");
      ASSERT(checkIntegrity());
      if (_currNode->key() <= searchKey) 
	{
	  return _currNode->key() == searchKey;
	};
      const Node* node;
      const Node* nextNode;
      ulong lv = _currMaxLevel;
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_no_overjumps_yet:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      if (!nextNode)
		{
		  --lv;
		  _currMaxLevel = lv;
		  if (lv) goto jump_no_overjumps_yet;
		  _currNode = node;
		  goto level_zero_no_overjumps_yet;
		};
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_no_overjumps_yet;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return true;
	}
      else 
	{
	level_zero_no_overjumps_yet:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  if ((!_currNode) || (_currNode->key() < searchKey)) 	    
	    return false;
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  goto level_zero_no_overjumps_yet;
	};
    check_lv_overjump_exists:
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_overjump_exists:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      ASSERT(nextNode);
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_overjump_exists;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return true;
	}
      else 
	{
	level_zero_overjump_exists:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  ASSERT(_currNode);
	  if (_currNode->key() < searchKey) 	    
	    return false;
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return true;
	    };
	  goto level_zero_overjump_exists;
	};
    }; 
    void getTo(const KeyType searchKey)
    {
      CALL("Surfer::getTo(const KeyType& searchKey)");
      ASSERT(checkIntegrity());
      ASSERT(_currNode->key() >= searchKey);
      if (_currNode->key() == searchKey) 
	{
	  return;
	};
      const Node* node;
      const Node* nextNode;
      ulong lv = _currMaxLevel;
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_no_overjumps_yet:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      if (!nextNode)
		{
		  --lv;
		  _currMaxLevel = lv;
		  if (lv) goto jump_no_overjumps_yet;
		  _currNode = node;
		  goto level_zero_no_overjumps_yet;
		};
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_no_overjumps_yet;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return;
	}
      else 
	{
	level_zero_no_overjumps_yet:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() >= searchKey);
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return;
	    };
	  goto level_zero_no_overjumps_yet;
	};
    check_lv_overjump_exists:
      if (lv)
	{
	  node = _jump[lv];
	  ASSERT(!Node::isHighestJump(node));
	  ASSERT(node);
	  if (node->key() < searchKey)
	    {
	      --lv;
	      goto check_lv_overjump_exists;
	    };
	  if (node->key() > searchKey)
	    {
	    jump_overjump_exists:
	      nextNode = Node::stripJump(node->jump()[lv]);
	      ASSERT(nextNode);
	      if (nextNode->key() < searchKey)
		{
		  _jump[lv] = nextNode;
		  --lv;
		  if (lv) goto jump_overjump_exists;		  
		  _currNode = node;
		  goto level_zero_overjump_exists;
		};
	      if (nextNode->key() > searchKey)
		{
		  node = nextNode;
		  goto jump_overjump_exists;
		};
	      _currNode = nextNode;
	      ASSERT(_currNode->key() == searchKey);
	      ASSERT(checkIntegrity());
	      return;
	    };
	  _currNode = node;
	  ASSERT(_currNode->key() == searchKey);
	  ASSERT(checkIntegrity());
	  return;
	}
      else 
	{
	level_zero_overjump_exists:
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() > searchKey);
	  _currNode = _currNode->next();
	  ASSERT(_currNode);
	  ASSERT(_currNode->key() >= searchKey);
	  ASSERT(_currNode->computeHeight() == 0UL);
	  if (_currNode->key() == searchKey) 
	    {
	      ASSERT(checkIntegrity());
	      return;
	    };
	  goto level_zero_overjump_exists;
	};
    }; 
    #ifdef DEBUG_NAMESPACE
     bool checkIntegrity()
     {
      if (!_currNode) return true;
      ulong currNodeHeight = _currNode->computeHeight();
      if (currNodeHeight > _currMaxLevel) 
       {
        DMSG("CNSHVoidSkList<..>::Surfer::checkIntegrity()> currNodeHeight > _currMaxLevel\n");
        return false;  
       };
      for (ulong lv = 1UL; lv <= _currMaxLevel; ++lv)
	{
	  if (Node::isHighestJump(_jump[lv]))
	    {
	      DMSG("CNSHVoidSkList<..>::Surfer::checkIntegrity()> Node::isHighestJump(_jump[lv])\n");
	      return false;
	    };
	  if (!_jump[lv])
	    {
	      DMSG("CNSHVoidSkList<..>::Surfer::checkIntegrity()> !_jump[lv]\n");
	      return false;
	    };
	  if (lv <= currNodeHeight)
	    {
	      if (_jump[lv]->key() < _currNode->key())
		{
		  DMSG("CNSHVoidSkList<..>::Surfer::checkIntegrity()> _jump[lv]->key() < _currNode->key()\n");
		  return false;
		};
	    }
	  else
	    {
	      if (lv > 1UL)
		{
		  if (_jump[lv]->key() > _jump[lv-1]->key())
		    {
		      DMSG("CNSHVoidSkList<..>::Surfer::checkIntegrity()> _jump[lv]->key() > _jump[lv-1]->key()\n");
		      return false;
		    };
		};
	    };
	};
      return true;
     };
    #endif    
   private:
    ulong _currMaxLevel;
    const Node* _jump[28]; 
    const Node* _currNode;
  }; 
public:
  CNSHVoidSkList() 
  {
    CALL("constructor CNSHVoidSkList()"); 
    _currMaxLevel = 0;
    _sizeOfTopIndex = 1;
    _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*));
    _topIndex[0] = (Node*)0;
    _insertCount = 0;
    _size = 0;
    ASSERT(checkIntegrity());
  }; 
  ~CNSHVoidSkList() 
  {
    CALL("destructor ~CNSHVoidSkList()");
    ASSERT(checkIntegrity());
    if (DestructionMode::isThorough())
      {
	Node* node = _topIndex[0];
	Node* tmp;
	while (node) 
	  { 
	    tmp = node;   
	    node = node->next(); 
	    delete tmp;
	  };  
	Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
      };
  }; 
  void* operator new(size_t) 
  {
   CALL("operator new(size_t)"); 
    return Alloc::allocate(sizeof(CNSHVoidSkList));
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)");
    Alloc::deallocate(obj,sizeof(CNSHVoidSkList));
  };
  void reset() 
  {
    CALL("reset()");
    ASSERT(checkIntegrity());
    Node* node = _topIndex[0];
    Node* tmp;
    while (node) 
      { 
	tmp = node;   
	node = node->next(); 
	delete tmp;
      };  
    Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
    _currMaxLevel = 0;
    _sizeOfTopIndex = 1;
    _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*)); 
    _topIndex[0] = (Node*)0;
    _insertCount = 0;
    _size = 0;
    ASSERT(checkIntegrity());
  }; 
  operator bool() const 
  {
   CALL("operator bool() const"); 
   return _topIndex[0] != 0; 
  }; 
  const Node* first() const 
  {
   CALL("first() const"); 
   return _topIndex[0]; 
  };
  Node* first() 
  {
   CALL("first()");
   return _topIndex[0]; 
  };
  ulong size() const { return _size; }; 
#ifndef NO_DEBUG
  ulong count() const 
  {   
   ulong res = 0; 
   for (const Node* n = first(); n; n = n->next()) 
    { 
     res++;
    };
   return res;
  };
#endif
  bool insert(const KeyType& searchKey);
  bool remove(const KeyType& searchKey);
  #ifndef NO_DEBUG
   bool checkIntegrity() const
   {
     CALL("checkIntegrity() const");
    if (!_topIndex) return false; 
    #ifdef DEBUG_ALLOC
     bool res = true;
     ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<void*>(_topIndex));          
     if (!objDesc) 
      { 
       DMSG("CNSHVoidSkList<..>::checkIntegrity()> Null pointer to object descriptor for *_topIndex.\n");  
       return false; 
      };     
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size !=  _sizeOfTopIndex * sizeof(Node*)) 
       {
        DMSG("CNSHVoidSkList<..>::checkIntegrity()> Wrong size of *_topIndex.\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent) 
       {
        DMSG("CNSHVoidSkList<..>::checkIntegrity()> objDesc->status != ObjDesc::Persistent for *_topIndex.\n");
        res = false; 
       };
     #endif      
     if (!res) return false;
    #endif
     for (ulong lv = 1UL; lv <= _currMaxLevel; ++lv)
       {
	 if (Node::isHighestJump(_topIndex[lv]))
	   {
	     DMSG("CNSHVoidSkList<..>::checkIntegrity()> Node::isHighestJump(_topIndex[lv])\n");
	     return false;
	   };
	 if (!_topIndex[lv])
	   {
	     DMSG("CNSHVoidSkList<..>::checkIntegrity()> !_topIndex[lv]\n");
	     return false;
	   };
	 if (_topIndex[lv]->key() > _topIndex[lv-1]->key())
	   {
	     DMSG("CNSHVoidSkList<..>::checkIntegrity()> _topIndex[lv]->key() > _topIndex[lv-1]->key()\n");
	     return false;
	   };
       };
     if (count() != size())
       {
	 DMSG("CNSHVoidSkList<..>::checkIntegrity()> count() != size()\n");
	 return false;
       };
    return true; 
   }; 
  #endif
  ostream& output(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     node->output(ostr,_currMaxLevel) << "\n";
    };
   return ostr;
  };
  ostream& outputKeys(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     ostr << node->key() << "\n";
    };
   return ostr;
  };
 private:
  CNSHVoidSkList(const CNSHVoidSkList& sl);
  CNSHVoidSkList& operator=(const CNSHVoidSkList& sl);
  ulong randomHeight() const
  { 
   ulong res = 0;
   ulong tmp = _insertCount;
   while (!(tmp % 2))
    { 
     res++;
     tmp = tmp/2;
    };
   return res;
  }; 
 private:
  ulong _sizeOfTopIndex;
  Node** _topIndex;
  ulong _currMaxLevel;
  ulong _insertCount;
  ulong _size;
 friend class Surfer; 
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPACT_NODE_SMALL_HEADER_SKIP_LIST
 #define DEBUG_NAMESPACE "CNSHVoidSkList<class Alloc,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class KeyType>
inline
bool CNSHVoidSkList<Alloc,KeyType>::insert(const KeyType& searchKey)  
{
 CALL("insert(const KeyType& searchKey)");
 ASSERT_IN(checkIntegrity(),"START");
 Node* update[32];
 Node* currNode = Node::node(_topIndex);      
 DOP(Node* debugPseudonode = Node::node(_topIndex));
 ulong currLevel = _currMaxLevel;
 Node* nextNode = currNode->jump()[currLevel];  
 ASSERT(!Node::isHighestJump(nextNode));
 KeyType nodeKey;
 check_next_node:
  if (nextNode)
   {
     ASSERT(nextNode->computeHeight() <= currLevel);
    nodeKey = nextNode->key();
    if (nodeKey > searchKey) 
     {
      currNode = nextNode;
      nextNode = Node::stripJump(nextNode->jump()[currLevel]);        
      goto check_next_node;
     };
    if (nodeKey == searchKey) 
      { 
	ASSERT(checkIntegrity()); 
	return false; 
      };
    update[currLevel] = currNode;
    while (currLevel) 
     { 
      currLevel--;
      nextNode = currNode->jump()[currLevel]; 
      ASSERT(!Node::isHighestJump(nextNode));
      nodeKey = nextNode->key();
      while (nodeKey > searchKey)
       {
	ASSERT(nextNode->computeHeight() == currLevel);
	currNode = nextNode;
        nextNode = Node::stripJump(nextNode->jump()[currLevel]);
        nodeKey = nextNode->key();
       };
      if (nodeKey == searchKey) { ASSERT(checkIntegrity()); return false; };
      update[currLevel] = currNode;
     };
    ASSERT(nextNode->key() < searchKey);
   }
  else 
   {
    update[currLevel] = currNode;
    if (currLevel) 
     {
      currLevel--;
      nextNode = currNode->jump()[currLevel];
      ASSERT(!Node::isHighestJump(nextNode));
      goto check_next_node;
     };
   };
  ASSERT(!currLevel);    
  ASSERT(_insertCount < ULONG_MAX);
  _insertCount++;  
  _size++;
  currLevel = randomHeight();
  currNode = new (currLevel) Node(searchKey); 
  Node** nodeJump = currNode->jump();  
  Node** jumpAddr; 
  Node* jump;
  if (currLevel > _currMaxLevel)
   {
     Node** newTopIndex = (Node**)Alloc::allocate((currLevel + 1) * sizeof(Node*));
    ulong lev;
    for (lev = 0UL; lev <= _currMaxLevel; ++lev)
      newTopIndex[lev] = _topIndex[lev];
     Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
    Node* pseudonode = Node::node(_topIndex);      
    for (lev = 0UL; lev <= _currMaxLevel; ++lev)
      {
        ASSERT(!Node::isHighestJump(update[lev]));
	if (update[lev] == pseudonode)      
	  {
	    ASSERT((!newTopIndex[lev]) || (newTopIndex[lev]->key() < searchKey));
	    nodeJump[lev] = newTopIndex[lev];
	    newTopIndex[lev] = currNode;
	  }
	else
	  {
	    jumpAddr = update[lev]->jump() + lev;
	    jump = *jumpAddr;	
	    if (Node::isHighestJump(jump))
	      {
		ASSERT(update[lev]->computeHeight() == lev);
		*jumpAddr = Node::markAsHighestJump(currNode);	
		nodeJump[lev] = Node::stripJump(jump);
	      }
	    else
	      {
		ASSERT(update[lev]->computeHeight() > lev);
		*jumpAddr = currNode;	
		nodeJump[lev] = jump;
	      };
	  };
      }; 
    _sizeOfTopIndex = currLevel + 1;
    _topIndex = newTopIndex;
    ++_currMaxLevel;
    while (_currMaxLevel < currLevel)
     {
      _topIndex[_currMaxLevel] = currNode;
      nodeJump[_currMaxLevel] = 0;
      ++_currMaxLevel;
     };
    _topIndex[_currMaxLevel] = currNode;
    nodeJump[_currMaxLevel] = Node::markAsHighestJump(static_cast<Node*>(0));
    ASSERT(currNode == debugPseudonode || currNode->computeHeight() == _currMaxLevel);
   }
  else 
   {       
     ulong lev;
     for (lev = 0UL; lev < currLevel; ++lev)
       {
	 ASSERT(!Node::isHighestJump(update[lev]));
	 jumpAddr = update[lev]->jump() + lev;
	 jump = *jumpAddr;		
	 if (Node::isHighestJump(jump))
	   {
	     ASSERT(update[lev] != debugPseudonode);
	     ASSERT(update[lev]->computeHeight() == lev);
	     *jumpAddr = Node::markAsHighestJump(currNode);
	     nodeJump[lev] = Node::stripJump(jump);
	   }
	 else
	   {
	     ASSERT(update[lev] == debugPseudonode || update[lev]->computeHeight() > lev);
	     *jumpAddr = currNode;
	     nodeJump[lev] = jump;
	   };
       };
     ASSERT(!Node::isHighestJump(update[lev]));
     jumpAddr = update[lev]->jump() + lev;
     jump = *jumpAddr;		
     if (Node::isHighestJump(jump))
       {
	 ASSERT(update[lev] != debugPseudonode);
	 ASSERT(update[lev]->computeHeight() == lev);
	 *jumpAddr = Node::markAsHighestJump(currNode);
	 nodeJump[lev] = jump;
       }
     else
       {
	 ASSERT(update[lev] == debugPseudonode || update[lev]->computeHeight() > lev);
	 *jumpAddr = currNode;
	 nodeJump[lev] = Node::markAsHighestJump(jump);
       };
     ASSERT(currNode == debugPseudonode || currNode->computeHeight() == currLevel);
   }; 
  ASSERT_IN(checkIntegrity(),"END");
  return true;
}; 
template <class Alloc,class KeyType>
inline
bool CNSHVoidSkList<Alloc,KeyType>::remove(const KeyType& searchKey) 
{
 CALL("remove(const KeyType& searchKey)");
 ASSERT(checkIntegrity());
 Node** jumpAddr = _topIndex;
 bool highest = false;
 ulong currLevel = _currMaxLevel; 
 Node* node;
 next_node:   
  node = jumpAddr[currLevel];
  ASSERT(highest == Node::isHighestJump(node));
  if (highest) node = Node::stripJump(node);
  if (node)
   {
    ASSERT(node->computeHeight() >= currLevel);  
    if (node->key() == searchKey)
     { 
       if (highest)
	 {
	   jumpAddr[currLevel] = Node::markAsHighestJump(node->jump()[currLevel]);
	 }
       else
	 {
	   jumpAddr[currLevel] = Node::stripJump(node->jump()[currLevel]);
	 };
      if (currLevel) 
	{ 
	  currLevel--;
	  highest = false;
	  goto next_node; 
	};
      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel)
	_currMaxLevel--; 
      delete node;
      _size--;
      ASSERT(checkIntegrity());
      return true;
     };
    if (node->key() > searchKey) 
     {
      jumpAddr = node->jump();
      highest = true;
      goto next_node;
     };
    if (currLevel) 
      { 
	currLevel--;	
	highest = false;
	goto next_node; 
      }; 
    ASSERT(checkIntegrity());     
    return false;
   };
  if (currLevel) 
    { 
      currLevel--;      
      highest = false;
      goto next_node; 
    };
  ASSERT(checkIntegrity());
  return false;
}; 
}; 
#endif
