/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <cstdlib>
#include "Defragmentation.hpp"
using namespace BK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_DEFRAGMENTATION 
#  define DEBUG_NAMESPACE "Defragmentation"
#endif
#include "debugMacros.hpp"
Defragmentation::Defragmentation() :
  _nextEligibleHashSizeNumber(0UL),
  _currentHashTableSize(0UL),
  _currentLoad(0UL),
  _hashTableForBeginnings(0),
  _hashTableForEnds(0)
{
  CALL("constructor Defragmentation()");
  _eligibleHashSizes[0] = 93187UL; 
  _eligibleHashSizes[1] = 187973UL;
  _eligibleHashSizes[2] = 389173UL; 
  _eligibleHashSizes[3] = 814309UL; 
  _eligibleHashSizes[4] = 1712369UL; 
  _eligibleHashSizes[5] = 3451241UL; 
  _eligibleHashSizes[6] = 7368791UL; 
  _eligibleHashSizes[7] = 17144507UL;
  _eligibleHashSizes[8] = 37667713UL;
  _eligibleHashSizes[9] = 76918277UL;	     
};
Defragmentation::~Defragmentation()
{
  CALL("destructor ~Defragmentation()");
};
void Defragmentation::reset()
{
  CALL("reset()");
  _currentLoad = 0UL;
  for (ulong i = 0UL; i < _currentHashTableSize; ++i)
    {
      _hashTableForBeginnings[i] = freeCellValue();
      _hashTableForEnds[i] = freeCellValue();
    };
}; 
bool Defragmentation::merge(void** piece,
			    ulong size)
{
  CALL("merge(..)");
  if (!_currentHashTableSize) return false;
  ASSERT(_currentLoad <= _currentHashTableSize); 
  void** newPiece;
  ulong newSize;
  void*** prefixEndLocation = tryToFindInEnds(piece);
  void*** suffixBeginningLocation = tryToFindInBeginnings(piece + size);
  if (prefixEndLocation)
    {
      void** prefix = *prefixEndLocation;
      newPiece = prefix;
      ulong prefixSize = reinterpret_cast<ulong>(prefix[0]);
      *prefixEndLocation = deletedCellValue();
      if (suffixBeginningLocation)
	{
	  ASSERT(_currentLoad >= 2UL);
	  void** suffix = *suffixBeginningLocation;
	  ulong suffixSize = reinterpret_cast<ulong>(suffix[0]);
	  newSize = prefixSize + size + suffixSize;
	  *suffixBeginningLocation = deletedCellValue();
	  *(findInEnds(suffix + suffixSize,suffix)) = newPiece;
	  --_currentLoad;
	}
      else 
	{
	  ASSERT(_currentLoad >= 1UL);
	  newSize = prefixSize + size;
	  *(locationFor(piece + size,_hashTableForEnds)) = newPiece;
	};
      newPiece[0] = reinterpret_cast<void*>(newSize);
      return true;
    }
  else 
    {      
      if (suffixBeginningLocation)
	{
	  ASSERT(_currentLoad >= 1UL);
	  void** suffix = *suffixBeginningLocation;
	  newPiece = piece;
	  ulong suffixSize = reinterpret_cast<ulong>(suffix[0]);
	  newSize = size + suffixSize;
	  *suffixBeginningLocation = deletedCellValue();
	  *(findInEnds(suffix + suffixSize,suffix)) = newPiece;
	  *(locationFor(newPiece,_hashTableForBeginnings)) = newPiece;
	  newPiece[0] = reinterpret_cast<void*>(newSize);
	  return true;
	}
      else
	return false;
    };
}; 
bool Defragmentation::integrate(void** piece,ulong size) 
{
  CALL("integrate(void** piece,ulong size)");
  ASSERT(_currentLoad <= _currentHashTableSize);
  ASSERT(!tryToFindInBeginnings(piece));
  ASSERT(!tryToFindInEnds(piece));
  ASSERT(!tryToFindInBeginnings(piece + size));
  ASSERT(!tryToFindInEnds(piece + size));
  if (_currentLoad == _currentHashTableSize)
    {
      if (!rehash()) return false;
    }
  else
    if (currentLoadFactor() >= criticalLoadFactor())
      {
	rehash(); 
      };
  ASSERT(_currentLoad < _currentHashTableSize); 
  *(locationFor(piece,_hashTableForBeginnings)) = piece;
  *(locationFor(piece + size,_hashTableForEnds)) = piece;
  ++_currentLoad;
  piece[0] = reinterpret_cast<void*>(size);
  return true;
}; 
void Defragmentation::resetResultCollection()
{
  CALL("resetResultCollection()");
  _nextResult = 0UL;
}; 
void** Defragmentation::collectNextPiece(ulong& size)
{
  CALL("collectNextPiece(ulong& size)");
 check_cell:
  if (_nextResult >= _currentHashTableSize) 
    return static_cast<void**>(0);
  void** piece = _hashTableForBeginnings[_nextResult];
  ++_nextResult;
  if (piece != freeCellValue() &&
      piece != deletedCellValue())
    {
      size = reinterpret_cast<ulong>(piece[0]);
      return piece;
    };
  goto check_cell;
}; 
inline
void*** Defragmentation::findInBeginnings(void** ptr)
{
  CALL("findInBeginnings(void** ptr)");
  ulong index = hashFunction(ptr);
  DOP(ulong debugOriginalIndex = index);
  ASSERT(index < _currentHashTableSize);
  void*** currentLocation = _hashTableForBeginnings + index;
  while ((*currentLocation) != ptr)
    {
      ASSERT((*currentLocation) != freeCellValue());
      ++index;
      if (index == _currentHashTableSize) index = 0UL;
      ASSERT(index != debugOriginalIndex);
      currentLocation = _hashTableForBeginnings + index;
    };
  return currentLocation;
}; 
void*** Defragmentation::findInEnds(void** keyPtr,void** valPtr)
{
  CALL("findInEnds(void** keyPtr,void** valPtr)");
  ulong index = hashFunction(keyPtr);
  DOP(ulong debugOriginalIndex = index);
  ASSERT(index < _currentHashTableSize);
  void*** currentLocation = _hashTableForEnds + index;
  while ((*currentLocation) != valPtr)
    {
      ASSERT((*currentLocation) != freeCellValue());
      ++index;
      if (index == _currentHashTableSize) index = 0UL;
      ASSERT(index != debugOriginalIndex);
      currentLocation = _hashTableForEnds + index;
    };
  return currentLocation;
}; 
inline
void*** Defragmentation::tryToFindInBeginnings(void** ptr)
{
  CALL("tryToFindInBeginnings(void** ptr)");
  if (!_currentHashTableSize) return static_cast<void***>(0);
  ulong index = hashFunction(ptr);
  ulong originalIndex = index;
  ASSERT(index < _currentHashTableSize);
  void*** currentLocation = _hashTableForBeginnings + index;
 check_cell:
  if ((*currentLocation) == ptr) return currentLocation;
  if ((*currentLocation) == freeCellValue()) return static_cast<void***>(0);
  ++index;
  if (index == _currentHashTableSize) index = 0UL;
  if (index == originalIndex) return static_cast<void***>(0);
  currentLocation = _hashTableForBeginnings + index;
  goto check_cell;
}; 
inline
void*** Defragmentation::tryToFindInEnds(void** ptr)
{
  CALL("tryToFindInEnds(void** ptr)");
  if (!_currentHashTableSize) return static_cast<void***>(0);
  ulong index = hashFunction(ptr);
  ulong originalIndex = index;
  ASSERT(index < _currentHashTableSize);
  void*** currentLocation = _hashTableForEnds + index;
 check_cell:
  if ((*currentLocation) == freeCellValue()) return static_cast<void***>(0);
  if ((*currentLocation) != deletedCellValue())
    {
      void** piece = *currentLocation;
      ulong size = reinterpret_cast<ulong>(piece[0]);
      if (piece + size == ptr) return currentLocation;
    };
  ++index;
  if (index == _currentHashTableSize) index = 0UL;
  if (index == originalIndex) return static_cast<void***>(0);
  currentLocation = _hashTableForEnds + index;
  goto check_cell;
}; 
inline
void*** Defragmentation::locationFor(void** ptr,void*** hashTable)
{
  CALL("locationFor(void** ptr,void*** hashTable)");
  ulong index = hashFunction(ptr);
  DOP(ulong debugOriginalIndex = index);
  ASSERT(index < _currentHashTableSize);
  void*** currentLocation = hashTable + index;
  while ((*currentLocation) != freeCellValue() &&
	 (*currentLocation) != deletedCellValue())
    {
      ASSERT((*currentLocation) != ptr);
      ASSERT((*currentLocation) + reinterpret_cast<ulong>((*currentLocation)[0]) != ptr);
      ++index;
      if (index == _currentHashTableSize) index = 0UL;
      ASSERT(index != debugOriginalIndex);
      currentLocation = hashTable + index;
    };
  return currentLocation;
}; 
bool Defragmentation::rehash()
{
  CALL("rehash()");
  if (_nextEligibleHashSizeNumber > lastEligibleHashSizeNumber())
    return false;
  ulong newHashTableSize = _eligibleHashSizes[_nextEligibleHashSizeNumber];
  ASSERT(newHashTableSize >= 2UL);
  void*** newMemory = 
    static_cast<void***>(malloc(2 * newHashTableSize * sizeof(void*)));
  if (!newMemory) return false;
  ++_nextEligibleHashSizeNumber;
  ulong oldHashTableSize = _currentHashTableSize;
  void*** oldHashTableForBeginnings = _hashTableForBeginnings;
  void*** oldMemory = _hashTableForBeginnings;
  _currentHashTableSize = newHashTableSize; 
  _hashTableForBeginnings = newMemory;
  _hashTableForEnds = newMemory + newHashTableSize;
  for (ulong i = 0UL; i < _currentHashTableSize; ++i)
    {
      _hashTableForBeginnings[i] = freeCellValue();
      _hashTableForEnds[i] = freeCellValue();
    };
  for (ulong j = 0UL; j < oldHashTableSize; ++j)
    {
      void** piece = oldHashTableForBeginnings[j];
      if (piece != freeCellValue() &&
	  piece != deletedCellValue())
	{
	  ulong size = reinterpret_cast<ulong>(piece[0]);
	  *(locationFor(piece,_hashTableForBeginnings)) = piece;
	  *(locationFor(piece + size,_hashTableForEnds)) = piece;
	};
    };
  free(static_cast<void*>(oldMemory));
  return true;
}; 
