/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef DEFRAGMENTATION_HPP
#define DEFRAGMENTATION_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
using namespace std;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_DEFRAGMENTATION
#  define DEBUG_NAMESPACE "Defragmentation"
#endif
#include "debugMacros.hpp"
namespace BK 
{
  class Defragmentation
  {
  public:
    Defragmentation();
    ~Defragmentation();
    void reset();
    bool merge(void** piece,
	       ulong size);
    bool integrate(void** piece,
                   ulong size); 
    void resetResultCollection();
    void** collectNextPiece(ulong& size);
  private: 
    static void** freeCellValue() { return reinterpret_cast<void**>(0); };
    static void** deletedCellValue() { return reinterpret_cast<void**>(1); };
    static float criticalLoadFactor() { return 0.75; };
    float currentLoadFactor() const 
    { 
      return static_cast<float>(_currentLoad) / _currentHashTableSize;
    };
    ulong hashFunction(void** ptr)
    {
      CALL("hashFunction(void** ptr)");
      ASSERT(_currentHashTableSize);
      return
	(reinterpret_cast<ulong>(ptr) >> 2) % _currentHashTableSize;
    };
    void*** findInBeginnings(void** ptr);
    void*** findInEnds(void** keyPtr,void** valPtr);
    void*** tryToFindInBeginnings(void** ptr);
    void*** tryToFindInEnds(void** ptr);
    void*** locationFor(void** ptr,void*** hashTable);
    bool rehash();
  private:
    ulong _eligibleHashSizes[10]; 
    static ulong lastEligibleHashSizeNumber() 
    {
      return 9UL;
    };
    ulong _nextEligibleHashSizeNumber;
    ulong _currentHashTableSize;
    ulong _currentLoad;
    void*** _hashTableForBeginnings;
    void*** _hashTableForEnds;
    ulong _nextResult;
  }; 
}; 
#endif
