/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef EQUIVALENCE_RELATION_HPP
#define EQUIVALENCE_RELATION_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_EQUIVALENCE_RELATION
#  define DEBUG_NAMESPACE "EquivalenceRelation<ulong SizeOfCarrier>"
#endif
#include "debugMacros.hpp" 
namespace BK 
{
template <ulong SizeOfCarrier>
 class EquivalenceRelation
  {
   public:
    class Iterator 
     {
      public:
       Iterator() {};
       Iterator(const EquivalenceRelation& er) 
        : next(er.registered), 
          end(er.next_registered),
          can_rep(er.can_rep) 
        {};
       void Reset(const EquivalenceRelation& er) 
        {
         next = er.registered;
         end = er.next_registered;
         can_rep = er.can_rep;
        };    
       bool Next(ulong& x,ulong& y)
        {
	 CALL("EquivalenceRelation<ulong SizeOfCarrier>::Iterator");
         if (next == end) return false;
         x = *next;
         y = can_rep[x];
         ASSERT(x > y);
         next++;
         return true;
        }; 
      private:
       const ulong* next;
       const ulong* end;
       const ulong* can_rep;
     }; 
    class EqClass 
     {
      public:
       EqClass() {};
       EqClass(EquivalenceRelation& rel,const ulong& x) 
       { 
	 CALL("EqClass::constructor EqClass(EquivalenceRelation& rel,const ulong& x)");
	 ASSERT(rel.Normalized());
        _rel = &rel;
	_frame = rel.can_rep;
        _canRep = _frame[x];
        _next = rel.registered;
        _end = rel.next_registered;
       }; 
       void reset(EquivalenceRelation& rel,const ulong& x)
       {
	 CALL("reset(EquivalenceRelation& rel,const ulong& x)");
	 ASSERT(rel.Normalized());
        _rel = &rel;
	_frame = rel.can_rep;
        _canRep = _frame[x];
        _next = rel.registered;
        _end = rel.next_registered;
       };
       void reset() 
       {
        _next = _rel->registered;
        _end = _rel->next_registered;
       };     
       bool nextElement(ulong& x) 
       {
	if (_next > _end) return false;
	while (_next < _end)
	 {
	  x = *_next;
          _next++;
	  if (_frame[x] == _canRep) return true;
	 };  
        x = _canRep;
        _next++;
        return true; 
       };            
      private:
       EquivalenceRelation* _rel;
       const ulong* _frame;
       ulong _canRep;
       const ulong* _next;
       const ulong* _end;
     }; 
    class MergedPairs 
    {
     public:
      MergedPairs() {};
      MergedPairs(EquivalenceRelation& rel,const ulong& x,const ulong& y) 
      { 
	reset(rel,x,y); 
      };
      void reset(EquivalenceRelation& rel,const ulong& x,const ulong& y)
      {
       CALL("MergedPairs::reset(EquivalenceRelation& rel,const ulong& x,const ulong& y)");
       ASSERT(!rel.Equivalent(x,y));
       rel.Normalize();
       _xClass.reset(rel,x);
       _yClass.reset(rel,y);
       ALWAYS(_xClass.nextElement(_currX)); 
      };
      bool nextPair(ulong& x,ulong& y)
      {
       CALL("nextPair(ulong& x,ulong& y)");
       if (_yClass.nextElement(y)) 
        {
	 if (_currX < y) { x = _currX; }
         else 
	  { 
	   x = y;
           y = _currX;
          };
         return true; 
        };
       if (_xClass.nextElement(_currX))  
	{
         _yClass.reset();
         ALWAYS(_yClass.nextElement(y));
         if (_currX < y) { x = _currX; }
         else 
	  { 
	   x = y;
           y = _currX;
          };
	 return true;
        };
       return false;
      }; 
     private:
      EqClass _xClass; 
      EqClass _yClass;
      ulong _currX;
    }; 
   public: 
    EquivalenceRelation() : next_registered(registered), name("?")
     {
      for (ulong i = 0; i < SizeOfCarrier; i++) can_rep[i] = i;
     };
    EquivalenceRelation(const char* nm) : next_registered(registered), name(nm)
     {
      for (ulong i = 0; i < SizeOfCarrier; i++) can_rep[i] = i;
     }; 
    ~EquivalenceRelation() 
    { 
      CALLM(name,"destructor ~EquivalenceRelation()"); 
      BK_CORRUPT(*this);
    };
    EquivalenceRelation& operator=(const EquivalenceRelation& er)
    {
     CALL("operator=(const EquivalenceRelation& er)");
     Reset();     
     for (const ulong* x = er.registered; x < er.next_registered; x++)
      {
       ASSERT(*x > er.can_rep[*x]);
       can_rep[*x] = er.can_rep[*x];
       *next_registered = *x; 
       next_registered++;
      };
     return *this;
    };  
    void Reset() 
     {
      CALLM(name,"Reset()");
      while (next_registered != registered) 
       {
        next_registered--;
        ASSERT(inCarrier(*next_registered));
        can_rep[*next_registered] = *next_registered;
       };
     };
    bool Identity() const { return next_registered == registered; };
    ulong CanRep(ulong x) const 
     {
      CALLM(name,"CanRep(ulong x) const");
      while (can_rep[x] != x) 
       {
        x = can_rep[x]; 
       };
      return x;
     };
    bool Equivalent(const ulong& x,const ulong& y) const
     {
      CALLM(name,"Equivalent(ulong x,ulong y) const");
      ASSERT(inCarrier(x));
      ASSERT(inCarrier(y));
      return CanRep(x) == CanRep(y);
     };
    void MakeEquivalent(ulong x,ulong y)
     {
      CALLM(name,"MakeEquivalent(ulong x,ulong y)");
      ASSERT(inCarrier(x));
      ASSERT(inCarrier(y));
      ulong can_x = CanRep(x);
      ulong can_y = CanRep(y);
      if (can_x < can_y) { Register(can_y,can_x); }
      else 
       { 
        if (can_x > can_y) Register(can_x,can_y);  
       }; 
     };
    bool Split(ulong& x,ulong& y)
     { 
      CALLM(name,"Split(ulong& x,ulong& y)");
      if (next_registered == registered) return false; 
      next_registered--;
      x = *next_registered;
      y = can_rep[x];
      ASSERT(y == can_rep[y]);
      ASSERT(x > y);
      ASSERT(inCarrier(x));      
      can_rep[x] = x;
      return true;
     };
    void Subtract(EquivalenceRelation& eqRel) 
     {
      CALLM(name,"Subtract(EquivalenceRelation& eqRel)");
      Normalize();
      DOP(static EquivalenceRelation debugER1);
      DOP(static EquivalenceRelation debugER2);
      DOP(static EquivalenceRelation debugER3);
      DOP(static EquivalenceRelation debugER4);
      DOP(debugER1 = eqRel);
      DOP(debugER2 = *this);
      ulong x;
      ulong can_x;
      ulong* i;
      for (i = next_registered - 1 ; i>= registered; i--)
       {
        x = *i;
        can_x = can_rep[x];
        if (eqRel.Equivalent(x,can_x)) { Unregister(x); }
        else
         {
          eqRel.MakeEquivalent(x,can_x);
         };
       }; 
      DOP(debugER3.MakeComposition(*this,debugER1));
      DOP(debugER4.MakeComposition(debugER2,debugER1));       
      ASSERT(debugER3.Implies(debugER4));
      ASSERT(debugER4.Implies(debugER3));
      #ifdef DEBUG_NAMESPACE
       for (i = registered; i < next_registered; i++) 
        {
	 ASSERT(!debugER1.Equivalent(*i,can_rep[*i]));  
         debugER1.MakeEquivalent(*i,can_rep[*i]);
        };   
      #endif
     }; 
    void MakeComposition(const EquivalenceRelation& er1,
			 const EquivalenceRelation& er2)
    {
     CALL("MakeComposition(const EquivalenceRelation& er1,const EquivalenceRelation& er2)");
     Reset();
     const ulong* x;
     for (x = er1.registered; x < er1.next_registered; x++)
      MakeEquivalent(*x,er1.can_rep[*x]);
     for (x = er2.registered; x < er2.next_registered; x++)
      MakeEquivalent(*x,er2.can_rep[*x]);
    };
    bool Implies(const EquivalenceRelation& er) const 
    {
     for (const ulong* x = registered; x < next_registered; x++)
       if (!er.Equivalent(*x,can_rep[*x])) return false;
     return true;      
    };
    void Normalize()
     {
      CALL("Normalize()");
      CALLM(name,"Normalize()");  
      for (ulong* i = next_registered - 1; i >= registered; i--)
       {
        can_rep[*i] = CanRep(*i);
        ASSERT(inCarrier(can_rep[*i]));
       }; 
     };   
    #ifndef NO_DEBUG
     bool Normalized() const
     {
      for (ulong* i = next_registered - 1; i >= registered; i--)       
       if (can_rep[can_rep[*i]] != can_rep[*i]) return false;
      return true;
     };
    #endif
   public: 
    #ifndef NO_DEBUG_VIS
     ostream& output(ostream& str) const
     {
      str << name << " : ";
      for (const ulong* i = registered; i < next_registered; i++)
       {
        str << *i << " -> " << can_rep[*i];
        if (i + 1 < next_registered) str << " | ";
       };
      return str;
     }; 
    #endif   
    #ifdef DEBUG_NAMESPACE
     static bool inCarrier(ulong x) { return (x < SizeOfCarrier); }
    #endif
   private:
    void Register(ulong x,ulong can_x) 
     {
      CALLM(name,"Register(ulong x,ulong can_x)");
      ASSERT(x > can_x);
      ASSERT(x == can_rep[x]);
      ASSERT(inCarrier(x));
      ASSERT(inCarrier(can_x));
      *next_registered = x;
      next_registered++;
      can_rep[x] = can_x; 
     }; 
    void Unregister(ulong x)
     {
      CALLM(name,"Unregister(ulong x)");
      ASSERT(inCarrier(x));
      ASSERT(inCarrier(can_rep[x])); 
      can_rep[x] = x;
      next_registered--;
      if ((*next_registered) != x)
       {
        for (ulong* i = registered; i < next_registered; i++)
         if ((*i) == x)
          {
           *i = *next_registered;
           return;
          };
       };
     }; 
   private:
    ulong can_rep[SizeOfCarrier];
    ulong registered[SizeOfCarrier];
    ulong* next_registered;
    const char* name;
   friend class Iterator;
   friend class EqClass;
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
 template <ulong SizeOfCarrier>
 ostream& operator<<(ostream& str,const BK::EquivalenceRelation<SizeOfCarrier>& er)
 {
  return er.output(str);
 };
}; 
#endif
#endif
