/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef GLOBAL_CLOCK_H
#define GLOBAL_CLOCK_H
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ARRAY
#  define DEBUG_NAMESPACE "GlobalClock"
#endif
#include "debugMacros.hpp"
namespace BK
{
  class GlobalClock
    {
    public:
      class InitStatic
	{
	public:
	  InitStatic();
	private:
	  static long _count;
	}; 
      class AlarmHandler
	{
	public:
	  AlarmHandler(void (*handlingFun)(int)) : 
	    handlingFunction(handlingFun),
	    _previous(GlobalClock::_currentAlarmHandler)
	    {
	      GlobalClock::_currentAlarmHandler = this;
	    };
	  ~AlarmHandler() 
	    {
	      if (GlobalClock::_currentAlarmHandler == this)
		{
		  GlobalClock::_currentAlarmHandler = _previous;
		};
	    };
	public:
	  void (*handlingFunction)(int);
	private:
	  AlarmHandler() {};
	private:
	  AlarmHandler* _previous;
	  friend class GlobalClock;
	}; 
      class Timer
	{
	public:
	  Timer() { reset(); };
	  ~Timer() {};
	  void reset() { _stopped = true; _elapsedDeciseconds = 0; };
	  void start() 
	    {
	      if (_stopped)
		{
		  _stopped = false;
		  _startTime = GlobalClock::elapsedDeciseconds();
		};
	    };
	  void stop() 
	    {
	      if (!_stopped)
		{
		  _stopped = true;
		  _elapsedDeciseconds += (GlobalClock::elapsedDeciseconds() - _startTime);
		};
	    };
	  int elapsedDeciseconds() const
	    { 
	      if (_stopped) return _elapsedDeciseconds;
	      return _elapsedDeciseconds + 
		(GlobalClock::elapsedDeciseconds() - _startTime);
	    };
	  float elapsedSeconds() const { return ((float)elapsedDeciseconds())/10; };
	private:
	  bool _stopped;
	  int _startTime;
	  int _elapsedDeciseconds;
	}; 
    public:
      static int elapsedDeciseconds() { return _elapsedDeciseconds; };
      static float elapsedSeconds() { return ((float)_elapsedDeciseconds)/10; };
    private: 
#if (BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE == BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_UNIX)
      static void alarmSignalHandler(int);
#elif (BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE == BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_MSWIN32)
	  class _Private; 
#endif
    private:
      static AlarmHandler* _currentAlarmHandler;
      static int _elapsedDeciseconds;
      friend class InitStatic;
      friend class AlarmHandler;
#if (BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE == BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_MSWIN32)
	  friend class _Private; 
#endif
    }; 
  static GlobalClock::InitStatic globalClockInitStatic;
}; 
#endif
