/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef OBJECT_POOL_WITH_MEMORY_HPP
#define OBJECT_POOL_WITH_MEMORY_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
#include "ObjectPool.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_POOL_WITH_MEMORY
#  define DEBUG_NAMESPACE "ObjectPoolWithMemory<class Alloc,class C>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class C>
class ObjectPoolWithMemory : public ObjectPool<C>
{
public:
  ObjectPoolWithMemory() : _free(0) 
  {
  };
  ~ObjectPoolWithMemory();
  void prepareNObjects(ulong n)
  {
    CALL("prepareNObjects(ulong n)");
    while (n)
      {
	Entry* e = 
	  ::new (Alloc::allocate(sizeof(Entry))) Entry();
	e->nextFree() = _free;
	_free = e;
	--n;
      };
  };
private:
  typedef typename ObjectPool<C>::Entry Entry;
private:  
  Entry* reserveEntry()
  {
    CALL("reserveEntry()");
    Entry* res;
    if (_free)
      {
	res = _free;
	_free = _free->nextFree();
      }
    else
      {
	res = 
	  ::new (Alloc::allocate(sizeof(Entry))) Entry();
      };
    return res;
  };
  Entry* tryToReserveEntry()
  {
    CALL("reserveEntry()");
    Entry* res;
    if (_free)
      {
	res = _free;
	_free = _free->nextFree();
      }
    else
      {
	void* mem = Alloc::tryToAllocate(sizeof(Entry));
	if (!mem) return static_cast<Entry*>(0);
	res = ::new (mem) Entry();
      };
    return res;
  };
  void recycleEntry(Entry* e)
  {
    CALL("recycleEntry(Entry* e)");
    e->nextFree() = _free;
    _free = e;
  };
private:
  Entry* _free;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_POOL_WITH_MEMORY
#  define DEBUG_NAMESPACE "ObjectPoolWithMemory<class Alloc,class C>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
  template <class Alloc,class C>
  inline
  ObjectPoolWithMemory<Alloc,C>::~ObjectPoolWithMemory()
  {
    CALL("destructor ~ObjectPoolWithMemory()");
    if (DestructionMode::isThorough())
      {
	while (_free)
	  {
	    Entry* e = _free;
	    _free = e->nextFree();
	    e->~Entry();
	    Alloc::deallocate(static_cast<void*>(e),sizeof(Entry));
	  };
      };
  };
}; 
#endif
