/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */
         
#ifndef SKIP_LIST_HPP
#define SKIP_LIST_HPP
#include <cstddef>
#include <climits>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SKIP_LIST
 #define DEBUG_NAMESPACE "SkipList<class Alloc,class ValType,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template < class Alloc, class ValType, class KeyType >
class SkipList
{
 public:
  class Surfer;
  class Destruction;
  class Node
    {
    public:
      ~Node() {};
      KeyType key() const { return _key; };
      const ValType& value() const { return _value; };
      ValType& value() { return _value; };
      Node* next() { return *(jump()); };
      const Node* next() const { return *(jump()); };    
    private:
      int level() const { return _level; };
    private: 
      Node** jump() { return static_cast<Node**>(static_cast<void*>(this + 1)); }; 
      const Node* const * jump() const 
      { 
	return static_cast<const Node* const *>(static_cast<const void*>(this + 1)); 
      };
      Node** jump(int lev) { return (static_cast<Node**>(this + 1)) + lev; }; 
      const Node* const * jump(int lev) const { return (static_cast<const Node* const *>(this + 1)) + lev; };
      Node(int l,const KeyType& k) : _level(l), _key(k) {};
      Node* next(int lev) 
      { 
	return (static_cast<Node**>(static_cast<void*>(this + 1)))[lev]; 
      }; 
      const Node* next(int lev) const 
      { 
	return (static_cast<const Node* const *>(static_cast<const void*>(this + 1)))[lev]; 
      };
      static const Node* node(const Node * const * jmp)  
	{
	  return (static_cast<Node*>(jmp)) - 1;
	};
      static Node* node(Node ** jmp)           
	{
	  return (static_cast<Node*>(static_cast<void*>(jmp))) - 1;
	};
      void* operator new(size_t size,int lv) 
	{    
	  CALL("operator new(size_t size,int lv)");
	  return Alloc::allocate(size + (lv+1)*sizeof(Node*));
	};
      void operator delete(void* obj)
	{
	  CALL("operator delete(void* obj)");
	  Alloc::deallocate(obj,sizeof(Node) + ((static_cast<Node*>(obj))->level() + 1)*sizeof(Node*));
	};   
void operator delete(void* obj,int lv)
	{
	  CALL("operator delete(void* obj,int lv)");
	  Alloc::deallocate(obj,sizeof(Node) + (lv + 1)*sizeof(Node*));
	};   
    private: 
      int _level;
      ValType _value;
      KeyType _key;    
      friend class SkipList<Alloc,ValType,KeyType>;
      friend class Surfer;
      friend class Destruction;
    }; 
  class Surfer
    {
    public:
      Surfer() {};
      Surfer(const SkipList& sl) { reset(sl); };
      ~Surfer() {};
      void reset(const SkipList& sl) 
	{
	  CALL("Surfer::reset(const SkipList& sl)");
	  ASSERT(sl.checkIntegrity());
	  _currMaxLevel = sl._currMaxLevel;
	  _jump = sl._topIndex;
	  _currNode = *_jump;
	};
      const Node* currNode() const { return _currNode; };
      const Node* const * currNodeAddr() const { return &_currNode; }; 
      bool next()
	{
	  CALL("next()");
	  _currNode = _currNode->next();
	  return (bool)_currNode;
	}; 
      bool find(const KeyType& searchKey) { return find(searchKey,_currNode); };
      bool find(const KeyType& searchKey,const Node*& node)
	{
	  CALL("Surfer::find(const KeyType& searchKey,const Node*& node)");
	ml_next_node:
	  node = _jump[_currMaxLevel]; 
	  if (node) 
	    {	  
	      ASSERT(node->level() == _currMaxLevel); 
	      if (node->key() < searchKey)
		{
		  if (!_currMaxLevel) return false;
		  int currLevel = _currMaxLevel - 1;
		  const Node* const * jump_ = _jump;        
		next_node:
		  node = jump_[currLevel];
		  ASSERT(node->level() >= currLevel);
		  if (node->key() == searchKey) return true;
		  if (node->key() < searchKey)
		    { 
		      if (!currLevel) return false;   
		      currLevel--;  
		    }
		  else 
		    { 
		      jump_ = node->jump();
		    };
		  goto next_node;
		};
	      _jump = node->jump();
	      if (node->key() == searchKey) return true;
	      goto ml_next_node;
	    }
	  else 
	    {
	      if (!_currMaxLevel) return false;
	      _currMaxLevel--;
	      goto ml_next_node;
	    };      
	}; 
    private:
      int _currMaxLevel;
      const Node* const * _jump;
      const Node* _currNode;
    }; 
  class Destruction
    {
    public:
      Destruction() {};
      Destruction(SkipList& sl) { reset(sl); }; 
      ~Destruction() {};
      void reset(SkipList& sl) 
	{ 
	  _node = sl.first();
	  sl._currMaxLevel = 0;
	  sl._insertCount = 0;
	  sl._size = 0;
	  *sl._topIndex = 0;
	};
      bool next(ValType& val) 
	{
	  if (!_node) return false;
	  val = _node->value();
	  Node* tmp = _node;
	  _node = _node->next();
	  delete tmp;
	  return true;
	};
    private: 
      Node* _node;
    };
 public:
  SkipList() :
    _currMaxLevel(0),
    _insertCount(0),
    _size(0)
    {
      CALL("constructor SkipList()");
      *_topIndex = 0; 
    };
  ~SkipList() 
    {
      CALL("destructor ~SkipList()");
      if (DestructionMode::isThorough())
	{
	  Node* node = *_topIndex;
	  Node* tmp;
	  while (node) 
	    { 
	      tmp = node;   
	      node = node->next(); 
	      delete tmp;
	    }; 
	};
    }; 
  void* operator new(size_t) 
    {
      CALL("operator new(size_t)");
      return Alloc::allocate(sizeof(SkipList));
    };
  void operator delete(void* obj) 
    {
      CALL("operator delete(void* obj)"); 
      Alloc::deallocate(obj,sizeof(SkipList));
    };
  void reset() 
    {
      CALL("reset()");
      Node* node = *_topIndex;
      Node* tmp;
      while (node) 
	{ 
	  tmp = node;   
	  node = node->next(); 
	  delete tmp;
	}; 
      _currMaxLevel = 0;
      _insertCount = 0;
      _size = 0;
      *_topIndex = 0; 
    }; 
  operator bool() const { return (*_topIndex) != 0;  }; 
  const Node* first() const { return *_topIndex; };
  Node* first() { return *_topIndex; };
  ulong size() const { return _size; }; 
#ifndef NO_DEBUG
  ulong count() const 
    {
      ulong res = 0; 
      for (const Node* n = first(); n; n = n->next()) res++;
      return res;
    };
#endif
  bool contains(const KeyType& searchKey) const
    {
      Surfer surfer(*this);
      const Node* dummy;
      return surfer.find(searchKey,dummy);
    }; 
  Node* insert(const KeyType& searchKey,bool& newNode)  
    {
      CALL("insert(const KeyType& searchKey,bool& newNode)");
      ASSERT(checkIntegrity());
      Node* update[28];
      Node* currNode = Node::node(_topIndex);      
      int currLevel = _currMaxLevel;
      Node* nextNode = currNode->next(currLevel);  
      KeyType nodeKey;
    check_next_node:
      if (nextNode)
	{
	  ASSERT(nextNode->level() <= currLevel);
	  nodeKey = nextNode->key();
	  if (nodeKey > searchKey) 
	    {
	      currNode = nextNode;
	      nextNode = nextNode->next(currLevel);        
	      goto check_next_node;
	    };
	  if (nodeKey == searchKey) { newNode = false; ASSERT(checkIntegrity()); return nextNode; };
	  update[currLevel] = currNode;
	  while (currLevel) 
	    { 
	      currLevel--;
	      nextNode = currNode->next(currLevel);
	      nodeKey = nextNode->key();
	      while (nodeKey > searchKey)
		{
		  currNode = nextNode;
		  nextNode = nextNode->next(currLevel); 
		  nodeKey = nextNode->key();
		};
	      if (nodeKey == searchKey) { newNode = false; ASSERT(checkIntegrity()); return nextNode; };
	      update[currLevel] = currNode;
	    };
	  ASSERT(nextNode->key() < searchKey);
	}
      else 
	{
	  update[currLevel] = currNode;
	  if (currLevel) 
	    {
	      currLevel--;
	      nextNode = currNode->next(currLevel);
	      goto check_next_node;
	    };
	};
      ASSERT(!currLevel);    
      newNode = true;
      ASSERT(_insertCount < INT_MAX);
      _insertCount++;  
      _size++;
      currLevel = randomLevel();
      currNode = new (currLevel) Node(currLevel,searchKey); 
      Node** nodeJump = currNode->jump();  
      Node** jump; 
      if (currLevel > _currMaxLevel)
	{
	  for (int lev = _currMaxLevel; lev>=0; lev--)
	    { 
	      jump = update[lev]->jump() + lev; 
	      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true);
	      nodeJump[lev] = *jump;       
	      *jump = currNode;
	    };
	  do
	    {
	      _currMaxLevel++;
	      _topIndex[_currMaxLevel] = currNode;
	      nodeJump[_currMaxLevel] = 0;
	    }
	  while (_currMaxLevel < currLevel);
	}
      else 
	{
	  for (int lev = currLevel; lev>=0; lev--)
	    { 
	      jump = update[lev]->jump() + lev; 
	      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true); 
	      nodeJump[lev] = *jump;       
	      *jump = currNode;
	    };
	}; 
      ASSERT(checkIntegrity());
      return currNode;
    }; 
  bool remove(const KeyType& searchKey,ValType& val) 
    {
      CALL("remove(const KeyType& searchKey,ValType& val)");
      ASSERT(checkIntegrity());
      Node** jump = _topIndex;  
      int currLevel = _currMaxLevel;
      Node* node;
    next_node:   
      node = jump[currLevel]; 
      if (node)
	{
	  ASSERT(node->level() >= currLevel);  
	  if (node->key() == searchKey)
	    { 
	      jump[currLevel] = node->jump()[currLevel]; 
	      if (currLevel) { currLevel--; goto next_node; };
	      val = node->value();
	      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel) _currMaxLevel--;
	      delete node;
	      _size--;
	      ASSERT(checkIntegrity());
	      ASSERT(!contains(searchKey));
	      return true;
	    };
	  if (node->key() > searchKey) 
	    {
	      jump = node->jump();        
	      goto next_node;
	    };
	  if (currLevel) { currLevel--; goto next_node; }; 
	  ASSERT(checkIntegrity());    
	  ASSERT(!contains(searchKey)); 
	  return false;
	};
      if (currLevel) { currLevel--; goto next_node; };
      ASSERT(checkIntegrity());
      ASSERT(!contains(searchKey));
      return false;
    }; 
#ifndef NO_DEBUG
  bool checkIntegrity() const
    {
      if (_currMaxLevel)
	for (int lev = 0; lev <= _currMaxLevel; lev++)
	  {
	    if (!(_topIndex[lev])) return false;     
	  };
      return (count() == size()); 
    }; 
#endif
 private:
  SkipList(const SkipList&);
  SkipList& operator=(const SkipList&);
  int randomLevel() const
    { 
      int res = 0;
      ulong tmp = _insertCount;
      while (!(tmp % 2))
	{ 
	  res++;
	  tmp = tmp/2;
	};
      return res;
    }; 
 private:
  Node* _topIndex[28];
  int _currMaxLevel;
  ulong _insertCount;
  ulong _size;
  friend class Surfer; 
  friend class Destruction;   
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SKIP_LIST
#define DEBUG_NAMESPACE "VoidSkipList<class Alloc,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class KeyType>
class VoidSkipList
{
 public:
  class Surfer;  
  class Node
    {
    public:
      ~Node() {};
      const KeyType& key() const { return _key; };
      Node* next() { return *(jump()); };
      const Node* next() const { return *(jump()); };  
    private: 
      Node(int l,const KeyType& k) : _level(l), _key(k) {};
      int level() const { return _level; };
      Node** jump() 
      { 
	return static_cast<Node**>(static_cast<void*>(this + 1)); 
      }; 
      const Node* const * jump() const 
      { 
	return static_cast<const Node* const *>(static_cast<const void*>(this + 1)); 
      };
      Node** jump(int lev) { return (static_cast<Node**>(static_cast<void*>(this + 1))) + lev; }; 
      const Node* const * jump(int lev) const 
      { 
	return (static_cast<const Node* const *>(static_cast<const void*>(this + 1))) + lev; 
      };
      Node* next(int lev) 
      { 
	return (static_cast<Node**>(static_cast<void*>(this + 1)))[lev]; 
      }; 
      const Node* next(int lev) const 
      { 
	return (static_cast<const Node* const *>(static_cast<const void*>(this + 1)))[lev]; 
      };
      static const Node* node(const Node * const * jmp)
	{
	  return (static_cast<Node*>(static_cast<void*>(jmp))) - 1;
	};
      static Node* node(Node ** jmp)
	{
	  return (static_cast<Node*>(static_cast<void*>(jmp))) - 1;
	};
      void* operator new(size_t size,int lv) 
	{ 
	  CALL("operator new(size_t size,int lv)");
	  return Alloc::allocate(size + (lv+1)*sizeof(Node*));
	};
      void operator delete(void* obj)
	{
	  CALL("operator delete(void* obj)");
	  Alloc::deallocate(obj,sizeof(Node) + ((static_cast<Node*>(obj))->level() + 1)*sizeof(Node*));
	};   
      void operator delete(void* obj,int lv)
	{
	  CALL("operator delete(void* obj,int lv)");
	  Alloc::deallocate(obj,sizeof(Node) + (lv + 1)*sizeof(Node*));
	};   
    private:
      int _level;
      KeyType _key;
      friend class VoidSkipList<Alloc,KeyType>;
      friend class Surfer;
    }; 
  class Surfer
    {
    public:
      Surfer() {};
      Surfer(const VoidSkipList& sl) { reset(sl); };
      ~Surfer() {};
      void reset(const VoidSkipList& sl) 
	{
	  CALL("Surfer::reset(const VoidSkipList& sl)");
	  ASSERT(sl.checkIntegrity());
	  _currMaxLevel = sl._currMaxLevel;
	  _jump = sl._topIndex;
	  _currNode = *_jump;
	};
      const Node* currNode() const { return _currNode; };
      const Node* const * currNodeAddr() const { return &_currNode; }; 
      bool next()
	{
	  CALL("next()");
	  _currNode = _currNode->next();
	  return (bool)_currNode;
	}; 
      bool find(const KeyType& searchKey) { return find(searchKey,_currNode); };
      bool find(const KeyType& searchKey,const Node*& node)
	{
	  CALL("Surfer::find(const KeyType& searchKey,const Node*& node)");
	ml_next_node:
	  node = _jump[_currMaxLevel];   
	  if (node) 
	    {
	      ASSERT(node->level() == _currMaxLevel); 
	      if (node->key() < searchKey)
		{
		  if (!_currMaxLevel) return false;
		  int currLevel = _currMaxLevel - 1;
		  const Node* const * jump_ = _jump;        
		next_node:
		  node = jump_[currLevel];
		  ASSERT(node->level() >= currLevel);
		  if (node->key() == searchKey) return true;
		  if (node->key() < searchKey)
		    { 
		      if (!currLevel) return false;   
		      currLevel--;  
		    }
		  else 
		    jump_ = node->jump();
		  goto next_node;
		};
	      _jump = node->jump();
	      if (node->key() == searchKey) return true;
	      goto ml_next_node;
	    }
	  else 
	    {
	      if (!_currMaxLevel) return false;
	      _currMaxLevel--;
	      goto ml_next_node;
	    };      
	}; 
    private:
      int _currMaxLevel;
      const Node* const * _jump;
      const Node* _currNode;
    }; 
 public:
  VoidSkipList() :
    _currMaxLevel(0),
    _insertCount(0),
    _size(0)
    {
      CALL("constructor VoidSkipList()");
      *_topIndex = 0; 
    };
  ~VoidSkipList() 
    {
      CALL("destructor ~VoidSkipList()");
      if (DestructionMode::isThorough())
	{
	  Node* node = *_topIndex;
	  Node* tmp;
	  while (node) 
	    { 
	      tmp = node;   
	      node = node->next(); 
	      delete tmp;
	    };  
	};
    }; 
  void* operator new(size_t) 
    {
      CALL("operator new(size_t)");
      return Alloc::allocate(sizeof(VoidSkipList)); 
    };
  void operator delete(void* obj) 
    {
      CALL("operator delete(void* obj)");
      Alloc::deallocate(obj,sizeof(VoidSkipList));
    };
  void reset() 
    {
      CALL("reset()");
      Node* node = *_topIndex;
      Node* tmp;
      while (node) 
	{ 
	  tmp = node;   
	  node = node->next(); 
	  delete tmp;
	};  
      _currMaxLevel = 0;
      _insertCount = 0;
      _size = 0;
      *_topIndex = 0; 
    }; 
  operator bool() const 
    {
      return (*_topIndex) != 0; 
    };
  bool isNonempty() const { return *_topIndex; };
  ulong size() const { return _size; };
#ifndef NO_DEBUG
  ulong count() const 
    {
      ulong res = 0; 
      for (const Node* n = first(); n; n = n->next()) res++;
      return res;
    };
#endif 
  const Node* first() const { return *_topIndex; };
  Node* first() { return *_topIndex; };
  bool contains(const KeyType& searchKey) const
    {
      Surfer surfer(*this);
      const Node* dummy;
      return surfer.find(searchKey,dummy);
    }; 
  bool insert(const KeyType& searchKey)  
    {
      CALL("insert(const KeyType& searchKey)");
      ASSERT(checkIntegrity());
      Node* update[28];
      Node* currNode = Node::node(_topIndex);
      int currLevel = _currMaxLevel;
      Node* nextNode = currNode->next(currLevel);
      KeyType nodeKey;
    check_next_node:
      if (nextNode)
	{
	  ASSERT(nextNode->level() <= currLevel);
	  nodeKey = nextNode->key();
	  if (nodeKey > searchKey) 
	    {
	      currNode = nextNode;
	      nextNode = nextNode->next(currLevel);        
	      goto check_next_node;
	    };
	  if (nodeKey == searchKey) { ASSERT(checkIntegrity()); return false; };
	  update[currLevel] = currNode;
	  while (currLevel) 
	    { 
	      currLevel--;
	      nextNode = currNode->next(currLevel);
	      nodeKey = nextNode->key();
	      while (nodeKey > searchKey)
		{
		  currNode = nextNode;
		  nextNode = nextNode->next(currLevel); 
		  nodeKey = nextNode->key();
		};
	      if (nodeKey == searchKey) { ASSERT(checkIntegrity()); return false; };
	      update[currLevel] = currNode;
	    };
	  ASSERT(nextNode->key() < searchKey);
	}
      else 
	{
	  update[currLevel] = currNode;
	  if (currLevel) 
	    {
	      currLevel--;
	      nextNode = currNode->next(currLevel);
	      goto check_next_node;
	    };
	};
      ASSERT(!currLevel);    
      ASSERT(_insertCount < INT_MAX);
      _insertCount++;  
      _size++;
      currLevel = randomLevel();
      currNode = new (currLevel) Node(currLevel,searchKey); 
      Node** nodeJump = currNode->jump();  
      Node** jump; 
      if (currLevel > _currMaxLevel)
	{
	  for (int lev = _currMaxLevel; lev>=0; lev--)
	    { 
	      jump = update[lev]->jump() + lev; 
	      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true);
	      nodeJump[lev] = *jump;       
	      *jump = currNode;
	    };
	  do
	    {
	      _currMaxLevel++;
	      _topIndex[_currMaxLevel] = currNode;
	      nodeJump[_currMaxLevel] = 0;
	    }
	  while (_currMaxLevel < currLevel);
	}
      else 
	{
	  for (int lev = currLevel; lev>=0; lev--)
	    { 
	      jump = update[lev]->jump() + lev; 
	      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true); 
	      nodeJump[lev] = *jump;       
	      *jump = currNode;
	    };
	}; 
      ASSERT(checkIntegrity());
      return true;
    }; 
  bool insert(const KeyType& searchKey,Node*& node)  
    {
      CALL("insert(const KeyType& searchKey,Node*& node)");
      ASSERT(checkIntegrity());
      Node* update[28];
      Node* currNode = Node::node(_topIndex);
      int currLevel = _currMaxLevel;
      Node* nextNode = currNode->next(currLevel);
      KeyType nodeKey;
    check_next_node:
      if (nextNode)
	{
	  ASSERT(nextNode->level() <= currLevel);
	  nodeKey = nextNode->key();
	  if (nodeKey > searchKey) 
	    {
	      currNode = nextNode;
	      nextNode = nextNode->next(currLevel);        
	      goto check_next_node;
	    };
	  if (nodeKey == searchKey) 
	    { 
	      ASSERT(checkIntegrity()); 
	      node = nextNode;
	      return false; 
	    };
	  update[currLevel] = currNode;
	  while (currLevel) 
	    { 
	      currLevel--;
	      nextNode = currNode->next(currLevel);
	      nodeKey = nextNode->key();
	      while (nodeKey > searchKey)
		{
		  currNode = nextNode;
		  nextNode = nextNode->next(currLevel); 
		  nodeKey = nextNode->key();
		};
	      if (nodeKey == searchKey) 
		{ 
		  ASSERT(checkIntegrity()); 
		  node = nextNode;
		  return false; 
		};
	      update[currLevel] = currNode;
	    };
	  ASSERT(nextNode->key() < searchKey);
	}
      else 
	{
	  update[currLevel] = currNode;
	  if (currLevel) 
	    {
	      currLevel--;
	      nextNode = currNode->next(currLevel);
	      goto check_next_node;
	    };
	};
      ASSERT(!currLevel);    
      ASSERT(_insertCount < INT_MAX);
      _insertCount++;  
      _size++;
      currLevel = randomLevel();
      currNode = new (currLevel) Node(currLevel,searchKey); 
      Node** nodeJump = currNode->jump();  
      Node** jump; 
      if (currLevel > _currMaxLevel)
	{
	  for (int lev = _currMaxLevel; lev>=0; lev--)
	    { 
	      jump = update[lev]->jump() + lev; 
	      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true);
	      nodeJump[lev] = *jump;       
	      *jump = currNode;
	    };
	  do
	    {
	      _currMaxLevel++;
	      _topIndex[_currMaxLevel] = currNode;
	      nodeJump[_currMaxLevel] = 0;
	    }
	  while (_currMaxLevel < currLevel);
	}
      else 
	{
	  for (int lev = currLevel; lev>=0; lev--)
	    { 
	      jump = update[lev]->jump() + lev; 
	      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true); 
	      nodeJump[lev] = *jump;       
	      *jump = currNode;
	    };
	}; 
      ASSERT(checkIntegrity());
      node = currNode;
      return true;
    }; 
  bool remove(const KeyType& searchKey) 
    {
      CALL("remove(const KeyType& searchKey)");
      ASSERT(checkIntegrity());
      Node** jump = _topIndex;  
      int currLevel = _currMaxLevel;
      Node* node;
    next_node:   
      node = jump[currLevel]; 
      if (node)
	{
	  ASSERT(node->level() >= currLevel);  
	  if (node->key() == searchKey)
	    { 
	      jump[currLevel] = node->jump()[currLevel]; 
	      if (currLevel) { currLevel--; goto next_node; };
	      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel) _currMaxLevel--;
	      delete node;
	      _size--;
	      ASSERT(checkIntegrity());
	      ASSERT(!contains(searchKey));
	      return true;
	    };
	  if (node->key() > searchKey) 
	    {
	      jump = node->jump();        
	      goto next_node;
	    };
	  if (currLevel) { currLevel--; goto next_node; }; 
	  ASSERT(checkIntegrity());
	  ASSERT(!contains(searchKey));     
	  return false;
	};
      if (currLevel) { currLevel--; goto next_node; };
      ASSERT(checkIntegrity());
      ASSERT(!contains(searchKey));
      return false;
    }; 
#ifndef NO_DEBUG
  bool checkIntegrity() const
    {
      if (_currMaxLevel)
	for (int lev = 0; lev <= _currMaxLevel; lev++)
	  {
	    if (!(_topIndex[lev])) return false;
	  };
      return  (count() == size()); 
    }; 
#endif
 private:  
  VoidSkipList(const VoidSkipList&);  
  VoidSkipList& operator=(const VoidSkipList&);
  int randomLevel() const
    { 
      int res = 0;
      ulong tmp = _insertCount;
      while (!(tmp % 2))
	{ 
	  res++;
	  tmp = tmp/2;
	};
      return res;
    }; 
 private:
  Node* _topIndex[28];
  int _currMaxLevel;
  ulong _insertCount;
  ulong _size;
  friend class Surfer;    
}; 
}; 
#endif
