/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef STATIC_CELL_ARRAY_HPP
#define STATIC_CELL_ARRAY_HPP 
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "Array.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATIC_CELL_ARRAY
#  define DEBUG_NAMESPACE "StaticCellArray<class Alloc,class ElType,ulong MinIncrementSize>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class ElType,ulong MinIncrementSize>
class StaticCellArray
{
 private:
  class WrappedElement;
 public:
  StaticCellArray() 
   : _redirectTable(DOP("StaticCellArray::_redirectTable"))
  {
   #ifdef DEBUG_NAMESPACE
    _name = "?";
   #endif
    for (ulong i = 0UL; i < _redirectTable.size(); ++i)
      _redirectTable[i] = new WrappedElement(); 
  }; 
  StaticCellArray(const char* nm) 
   : _redirectTable(DOP("StaticCellArray::_redirectTable")) 
  {
   #ifdef DEBUG_NAMESPACE
    _name = nm;
   #endif
    for (ulong i = 0UL; i < _redirectTable.size(); ++i)
      _redirectTable[i] = new WrappedElement(); 
  }; 
  ~StaticCellArray() 
    { 
      CALLM(_name,"destructor ~StaticCellArray()"); 
      if (DestructionMode::isThorough())
	for (ulong i = 0; i < size(); i++)
	  delete (_redirectTable[i]);
    };
  void* operator new(size_t) 
  {
   CALL("operator new (size_t)");
   return Alloc::allocate(sizeof(StaticCellArray));
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)");
   Alloc::deallocate(obj,sizeof(StaticCellArray)); 
  };
  ElType& operator[](ulong i) 
  {
   CALLM(_name,"operator[](ulong i)");
   ASSERT(i < size());  
   return _redirectTable[i]->content;
  };
  ElType& operator[](ulong i) const 
  {
   CALLM(_name,"operator[](ulong i) const"); 
   ASSERT(i < size());  
   return _redirectTable[i]->content;
  }; 
  const ElType& sub(ulong i) const 
  {
   CALLM(_name,"const ElType& sub(ulong i) const");
   if (i >= size()) expand(i);
   return _redirectTable[i]->content;
  };
  ElType& sub(ulong i) 
  {
   CALLM(_name,"ElType& sub(ulong i)");
   if (i >= size()) expand(i);
   return _redirectTable[i]->content;
  };
  void expand(ulong i)
  {
   CALLM(_name,"expand(ulong i)");
   ulong oldSize = size(); 
   _redirectTable.expand(i);
   for (ulong j = oldSize; j < size(); j++)
     _redirectTable[j] = new WrappedElement();
  }; 
  bool tryToExpand(ulong i)
  {
    CALLM(_name,"tryToExpand(ulong i)");
    ulong oldSize = size(); 
    if (_redirectTable.tryToExpand(i))
      {
	for (ulong j = oldSize; j < size(); j++)
	  _redirectTable[j] = new WrappedElement(); 
	return true;
      };
    return false;
  };
  void expand()
  {
   CALLM(_name,"expand()");
   ulong oldSize = size();
   _redirectTable.expand();  
   ASSERT(size() > oldSize);
   for (ulong i = oldSize; i < size(); i++)
     _redirectTable[i] = new WrappedElement(); 
  }; 
  ulong size() const { return _redirectTable.size(); };
  #ifndef NO_DEBUG
   void freeze() { _redirectTable.freeze(); };
   void unfreeze() { _redirectTable.unfreeze(); };
  #endif  
 private:
  class WrappedElement 
  {
   public:
    WrappedElement() {};
    ~WrappedElement() {};
    void* operator new(size_t) 
    {
     CALL("WrappedElement::operator new(size_t)");
     return Alloc::allocate(sizeof(WrappedElement));
    };
    void operator delete(void* obj)
    {
     CALL("WrappedElement::operator delete(void* obj)");
     Alloc::deallocate(obj,sizeof(WrappedElement));
    };
   public:
    ElType content;
  };
 private: 
  StaticCellArray(const StaticCellArray& a);
  StaticCellArray& operator=(const StaticCellArray& a);
 private:
  Array<Alloc,WrappedElement*,DefaultRelocation<WrappedElement*>,MinIncrementSize> _redirectTable;
  #ifdef DEBUG_NAMESPACE
   const char* _name;
  #endif
}; 
}; 
#endif
