/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TAGGED_PTR_HPP
#define TAGGED_PTR_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TAGGED_PTR
#  define DEBUG_NAMESPACE "TaggedPtr<class C>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
enum PtrTag
{
  PtrTag00 = 0UL,
  PtrTag01 = 1UL,
  PtrTag10 = 2UL,
  PtrTag11 = 3UL
};
template <class C>
class TaggedPtr
{
public:
  TaggedPtr() : _content(0) {};
  explicit TaggedPtr(C* ptr) : _content(ptr) {};
  TaggedPtr(C* ptr,PtrTag tag) {
    _content = reinterpret_cast<C*>(reinterpret_cast<ulong>(ptr) | tag);
  };
  TaggedPtr(C* ptr,ulong tag) : _content(ptr) {
    CALL("constructor TaggedPtr(C* ptr,ulong tag)");
    ASSERT(tag <= 3UL);
    _content = reinterpret_cast<C*>(reinterpret_cast<ulong>(ptr) | tag);
  };
  TaggedPtr& operator=(C* const & ptr) {
    _content = ptr;
    return *this;
  };
  PtrTag tag() const {
    return 
      static_cast<PtrTag>((reinterpret_cast<ulong>(_content) << 30) >> 30);
  };
  bool clean() const {
    return !(reinterpret_cast<ulong>(_content) << 30);
  };
  void setTag(PtrTag tag) {
    _content = 
      reinterpret_cast<C*>(((reinterpret_cast<ulong>(_content) >> 2) << 2) | tag);
  };
  void setTag(ulong tag) {
    CALL("setTag(ulong tag)");
    ASSERT(tag <= 3UL);
    _content = 
      reinterpret_cast<C*>(((reinterpret_cast<ulong>(_content) >> 2) << 2) | tag);
  };
  C* strip() const {
    return reinterpret_cast<C*>((reinterpret_cast<ulong>(_content) >> 2) << 2);
  };
  C* strip(PtrTag& t) const {
    t = tag();
    return strip();
  };
  C* rawContent() const { return _content; };
private:
  C* _content;
}; 
}; 
#endif
