/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TIMER_HPP
#define TIMER_HPP
#include <ctime>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TIMER
#  define DEBUG_NAMESPACE "Timer"
#endif
#include "debugMacros.hpp"
namespace BK 
{
class Timer
{
 public:
  Timer() 
  {
   CALL("constructor Timer()"); 
   reset(); 
  };
  ~Timer() 
  {
    CALL("destructor ~Timer()");
    BK_CORRUPT(*this);
  }; 
  void reset()
  {
   stopped = true;
   elapsedSec = 0;
  };
  void start()
  {
   if (stopped)   
    {
     stopped = false;
     startTime = clock();     
    };
  };
  void stop()
  {
   if (!stopped)
    {
     stopped = true;
     clock_t currTime = clock();
     if (currTime > startTime)      
       elapsedSec += (((float)(currTime - startTime))/CLOCKS_PER_SEC);
    };
  };
  float elapsed()
  {
   if (stopped) return elapsedSec; 
   clock_t currTime = clock();     
   if (currTime > startTime) 
    return elapsedSec + (((float)(currTime - startTime))/CLOCKS_PER_SEC); 
   return elapsedSec;
  };
 private:   
  bool stopped;
  float elapsedSec;
  clock_t startTime;
}; 
}; 
#endif
