/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "jargon.hpp"
#include "Debug.hpp"
#ifdef DEBUG_OUT
#  define DEBUG_OUT_STR std::cout
#else
#  define DEBUG_OUT_STR BK::nullStream
#endif
#ifdef DEBUG_NAMESPACE
#  undef DCOMMA
#  define DCOMMA ,
#  undef CALL
#  ifdef DEBUG_NO_TRACING
#    define CALL(fname)
#  else
#    define CALL(fname) BK::Debug::Call debugFunctionCall(DEBUG_NAMESPACE "::" fname);
#  endif 
#  undef CALLM
#  ifdef DEBUG_NO_TRACING
#    define CALLM(objName,fname)
#  else
#    define CALLM(objName,fname) BK::Debug::Call debugMethodCall(DEBUG_NAMESPACE "::" fname,objName);
#  endif
#  undef OPEN_BR_DECL
#  ifdef DEBUG_NO_TRACING
#    define OPEN_BR_DECL(id) BK::Debug::DummyOpenDataBracket debugOpenDataBracket##id
#  else
#    define OPEN_BR_DECL(id) BK::Debug::OpenDataBracket debugOpenDataBracket##id
#  endif
#  undef OPEN_BR
#  ifdef DEBUG_NO_TRACING
#    define OPEN_BR(id,name) debugOpenDataBracket##id ()
#  else
#    define OPEN_BR(id,name) debugOpenDataBracket##id ( DEBUG_NAMESPACE "::" name)
#  endif
#  undef CLOSE_BR_DECL
#  ifdef DEBUG_NO_TRACING
#    define CLOSE_BR_DECL(id) BK::Debug::DummyCloseDataBracket debugCloseDataBracket##id
#  else
#    define CLOSE_BR_DECL(id) BK::Debug::CloseDataBracket debugCloseDataBracket##id
#  endif
#  undef COP
#  ifdef DEBUG_NO_TRACING
#    define COP(cop)
#  else
#    define COP(cop) BK::Debug::cp(cop);
#  endif
#  undef ASSERT
#  ifndef DEBUG_NO_ASSERT
#    define ASSERT(cond) { if (!( cond )) BK::Debug::error(DEBUG_OUT_STR,"Assertion " #cond " violated."); }
#  else 
#    define ASSERT(cond)
#  endif
#  undef ASSERT_IN
#  ifndef DEBUG_NO_ASSERT
#    ifndef DEBUG_NO_TRACING
#      define ASSERT_IN(cond,cop) { if (!( cond )) { BK::Debug::cp(cop); BK::Debug::error(DEBUG_OUT_STR,"Assertion " #cond " violated."); }}
#    else
#      define ASSERT_IN(cond,cop) { if (!( cond )) { BK::Debug::error(DEBUG_OUT_STR,"Assertion " #cond " violated."); }}
#    endif
#  else 
#    define ASSERT_IN(cond,cop)
#  endif
#  undef ALWAYS
#  ifndef DEBUG_NO_ASSERT
#    define ALWAYS(cond)  { if (!( cond )) BK::Debug::error(DEBUG_OUT_STR,"Assertion ALWAYS(" #cond ") violated."); }
#  else 
#    define ALWAYS(cond) cond
#  endif
#  undef NEVER
#  ifndef DEBUG_NO_ASSERT
#    define NEVER(cond)  { if ( cond ) BK::Debug::error(DEBUG_OUT_STR,"Assertion NEVER(" #cond ") violated."); }
#  else 
#    define NEVER(cond) cond
#  endif
#  undef ICP
#  ifdef DEBUG_NO_TRACING
#    define ICP(cop) { BK::Debug::error(DEBUG_OUT_STR,"Illegal control point"); }
#  else
#    define ICP(cop) { BK::Debug::cp(cop); BK::Debug::error(DEBUG_OUT_STR,"Illegal control point"); }
#  endif
#  undef LC_INIT
#  define LC_INIT(name,limit) BK::Debug::LimitedCounter<0UL> debugLimitedCounter##name(limit)
#  undef LC_RESET
#  define LC_RESET(name) debugLimitedCounter##name.reset()
#  undef LC_INC
#  define LC_INC(name) \
   { \
     debugLimitedCounter##name.inc(); \
     if (debugLimitedCounter##name.overflow()) \
       BK::Debug::error(DEBUG_OUT_STR,"Overflow in limited counter " #name "."); \
   }
#  undef DLC_INIT
#  define DLC_INIT(limit) BK::Debug::LimitedCounter<0UL> debugDefaultLimitedCounter(limit)
#  undef DLC_RESET
#  define DLC_RESET debugDefaultLimitedCounter.reset()
#  undef DLC_INC
#  define DLC_INC \
   { \
     debugDefaultLimitedCounter.inc(); \
     if (debugDefaultLimitedCounter.overflow()) \
       BK::Debug::error(DEBUG_OUT_STR,"Overflow in the default limited counter."); \
   }
#  undef DOP
#  define DOP(e) e
#  undef BK_CORRUPT
#  define BK_CORRUPT(obj) \
   { \
     unsigned long* debugObjEnd = (unsigned long*)((&obj) + 1); \
     for (unsigned long* debugPtr = (unsigned long*)(&obj); \
          debugPtr + 1 <= debugObjEnd; \
	  ++debugPtr) \
       *debugPtr = 66666666UL; \
   }
#  undef BK_CORRUPTMEM
#  define BK_CORRUPTMEM(ptr,size) \
   { \
     for (size_t debugIntVar = 0; debugIntVar < size/sizeof(unsigned long); ++debugIntVar) \
       ((unsigned long*)ptr)[debugIntVar] = 777777777UL; \
   }
#  undef DMSG
#  define DMSG(m) DEBUG_OUT_STR << m
#  undef DWARN
#  ifdef DEBUG_WARN
#    define DWARN(m) DEBUG_OUT_STR << "BK::Debug warning:" << m << "\n"
#  else
#    define DWARN(m)
#  endif
#  undef REPCPH
#  define REPCPH { BK::Debug::reportCP(DEBUG_OUT_STR); BK::Debug::history(DEBUG_OUT_STR); }
#  undef TRACE
#  define TRACE(traceCP,traceAterError) { BK::Debug::setTracing(&(DEBUG_OUT_STR),traceCP,traceAterError); }
#  undef TRACINGOFF
#  define TRACINGOFF { BK::Debug::tracingOff(); }
#else
#  undef DCOMMA
#  define DCOMMA 
#  undef CALL
#  define CALL(fname)
#  undef CALLM
#  define CALLM(objName,fname)
#  undef OPEN_BR_DECL
#  define OPEN_BR_DECL(id) 
#  undef OPEN_BR
#  define OPEN_BR(id,name)      
#  undef CLOSE_BR_DECL
#  define CLOSE_BR_DECL(id)
#  undef COP
#  define COP(cop)
#  undef ASSERT
#  define ASSERT(cond)
#  undef ASSERT_IN
#  define ASSERT_IN(cond,cop)
#  undef ALWAYS
#  define ALWAYS(cond) cond
#  undef NEVER
#  define NEVER(cond) cond
#  undef ICP
#  define ICP(cop)
#  undef LC_INIT
#  define LC_INIT(name,limit) 
#  undef LC_RESET
#  define LC_RESET(name) 
#  undef LC_INC
#  define LC_INC(name)
#  undef DLC_INIT
#  define DLC_INIT(limit)
#  undef DLC_RESET
#  define DLC_RESET
#  undef DLC_INC
#  define DLC_INC 
#  undef DOP
#  define DOP(e) 
#  undef BK_CORRUPT
#  define BK_CORRUPT(obj)
#  undef BK_CORRUPTMEM
#  define BK_CORRUPTMEM(ptr,size)
#  undef DMSG
#  define DMSG(m)
#  undef DWARN
#  define DWARN(m)
#  undef REPCPH
#  define REPCPH 
#  undef TRACE
#  define TRACE(traceCP,traceAterError)
#  undef TRACINGOFF
#  define TRACINGOFF
#endif
#ifndef NO_DEBUG 
#  undef SETDEBUGEVENTHOOK
#  define SETDEBUGEVENTHOOK(hookFun) { BK::Debug::setNextEventHook(hookFun); }
#else 
#  undef SETDEBUGEVENTHOOK
#  define SETDEBUGEVENTHOOK(hookFun) { Debugging_event_hook_uncommented }
#endif 
#undef DF
#ifdef NO_DEBUG
#  define DF { Debugging_feature_uncommented } 
#else
#  define DF 
#endif
