/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <fstream>
#include "CodeGenerator.hpp"
#include "Syntax.hpp"
#include "Unit.hpp"
using namespace Shell;
using namespace std;
void CodeGenerator::generate (const UnitList& units)
{
  Iterator<Unit> us(units);
  while (us.hasNext()) {
    generate(us.next());
  }
} 
void CodeGenerator::generate (const Unit& unit)
{
  ASS(unit.unitType() == CLAUSE);
  int c = generate(unit.clause());
  _stream << "  _metaTheory.push(Unit(\"ax" << c
	  << "\",AXIOM,clause" << c
	  << "));\n";
} 
CodeGenerator::CodeGenerator (std::ostream& stream)
  : _stream(stream),
    _counter(0)
{
} 
void CodeGenerator::generateCodeForTheories ()
{
  ifstream in("theories.tptp");
  UnitList units;
  Syntax::readUnits(in,units,IS_TPTP);
  _stream << "void TheoryFinder::addTheory ()\n"
             "{\n"
             "  TermList termList0;\n"
             "  LiteralList literalList0;\n";
  CodeGenerator generator(cout);
  generator.generate(units);
  _stream << "}\n";
} 
int CodeGenerator::generate (const Clause& clause)
{
  int lits = generate(clause.literals());
  _stream << "  Clause clause" << ++_counter 
	  << "(literalList" << lits << ");\n";
  return _counter;
} 
int CodeGenerator::generate (const LiteralList& literals)
{
  if (literals.isEmpty()) {
    return 0;
  }
  int u = generate(literals.head());
  int us = generate(literals.tail());
  _stream << "  LiteralList literalList" << ++_counter
	  << "(literal" << u
	  << ",literalList" << us
	  << ");\n";
  return _counter;
} 
int CodeGenerator::generate (const Literal& literal)
{
  int atm = generate(literal.atom());
  _stream << "  Literal literal" << ++_counter 
	  << "(" << (literal.sign() ? "true" : "false")
	  << ",atom" << atm << ");\n";
  return _counter;
} 
int CodeGenerator::generate (const TermList& terms)
{
  if (terms.isEmpty()) {
    return 0;
  }
  int u = generate(terms.head());
  int us = generate(terms.tail());
  _stream << "  TermList termList" << ++_counter
	  << "(term" << u
	  << ",termList" << us
	  << ");\n";
  return _counter;
} 
int CodeGenerator::generate (const Term& term)
{
  switch (term.tag()) {
  case VAR:
    _stream << "  Term term" << ++_counter
	    << "(Var(" << term.var().toInt() << "));\n";
    return _counter;
  case COMPOUND:
    {
      const FunctionSymbol& functor = term.functor();
      int args = generate(term.args());
      int fun = functor.number();
      if (! _funs.member(fun)) { 
	_funs.insert(fun);
	_stream << "  FunctionSymbol& fun" << fun
		<< " = _sig.createFun(\"" 
		<< functor.name() << "\","
		<< functor.arity() << ");\n";
      }
      _stream << "  Term term" << ++_counter
	      << "(fun" << fun
	      << ",termList" << args
	      << ");\n";
      return _counter;
    }
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
} 
int CodeGenerator::generate (const Atom& atom)
{
  const PredicateSymbol& functor = atom.functor();
  int args = generate(atom.args());
  int pred = functor.number();
  if (! _preds.member(pred)) { 
    _preds.insert(pred);
    _stream << "  PredicateSymbol& pred" << pred
	    << " = _sig.createPred(\"" 
	    << functor.name() << "\","
	    << functor.arity() << ");\n";
  }
  _stream << "  Atom atom" << ++_counter
	  << "(pred" << pred
	  << ",termList" << args
	  << ");\n";
  return _counter;
} 
