/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __DoubleSubstitution__
#define __DoubleSubstitution__
#include <string>
namespace Shell {
class Term;
class TermList;
class Atom;
class Literal;
class LiteralList;
class DoubleSubstitution
{
public:
  DoubleSubstitution () : _nextVariable(0) {}
  bool unify (const Atom& a1,const Atom& a2);
  bool resolve (const Literal& l1,const Literal& l2);
  void apply (LiteralList&,int index);
  string toString () const;
public: 
  class Binding;
private:	
  class VariableBank
  {
  public:
    VariableBank();
    ~VariableBank();
    Binding* get (Var v) const;
    bool bind (Var v,const Term& t,VariableBank* vb);
    void set (Var v,const Term& t,VariableBank* b);
    string toString (const VariableBank*) const;
  private:
    int _capacity;
    Binding** _bindings;
    void extend(int index);
    bool occurs (Var v,const Term& t,const VariableBank* vb) const;
    bool occurs (Var v,const TermList& t,const VariableBank* vb) const;
  }; 
  bool unify (const Term& t1,
	      const Term& t2,
	      VariableBank* vb1,
	      VariableBank* vb2);
  bool unify (const TermList& ts1,
	      const TermList& ts2,
	      VariableBank* vb1,
	      VariableBank* vb2);
  void apply (LiteralList&,VariableBank*);
  void apply (Literal&,VariableBank*);
  void apply (Atom&,VariableBank*);
  void apply (TermList&,VariableBank*);
  void apply (Term&,VariableBank*);
  VariableBank _bank[2];
  int _nextVariable;
}; 
}
#endif
