/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Environment.hpp"
#include "NameArray.hpp"
#include "Options.hpp"
#include "Signature.hpp"
#include "Stack.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
namespace Shell {
InputSyntax Environment::inputSyntax = IS_TPTP;
OutputSyntax Environment::outputSyntax = OS_VAMPIRE;
Mode Environment::mode = MODE_CASC_J3;
string Environment::include("/home/voronkov/TPTP-v3.1.1/");
Kernel* Environment::kernel = 0;
class Environment::Constants {
public:
  static NameArray parameters;
  static NameArray modes;
  static NameArray functions;
  static NameArray predicates;
  static const int funArities[];
  static const int predArities[];
private:
  static const char* par[];
  static const char* md[];
  static const char* evalfun[];
  static const char* evalpred[];
};
const char* Environment::Constants::par[] = {
  "include",
  "input_syntax",
  "mode",
  "output_syntax"};
NameArray Environment::Constants::parameters(par,
					     sizeof(par)/sizeof(char*));
const char* Environment::Constants::md[] = {
  "casc",
  "kif",
  "spider",
  "vampire"};
NameArray Environment::Constants::modes(md,
					sizeof(md)/sizeof(char*));
const char* Environment::Constants::evalfun[] = {
  "general:to_string",
  "int:abs",
  "int:compare",
  "int:div",
  "int:from_string",
  "int:max",
  "int:min",
  "int:minus",
  "int:mod",
  "int:multiply",
  "int:negate",
  "int:plus",
  "int:random",
  "int:sign",
  "int:to_string",
  "real:abs",
  "real:acos",
  "real:asin",
  "real:atan",
  "real:atan2",
  "real:ceil",
  "real:compare",
  "real:cos",
  "real:cosh",
  "real:div",
  "real:exp",
  "real:floor",
  "real:from_string",
  "real:ln",
  "real:log10",
  "real:max",
  "real:min",
  "real:minus",
  "real:multiply",
  "real:negate",
  "real:plus",
  "real:power",
  "real:random",
  "real:round",
  "real:sign",
  "real:sin",
  "real:sinh",
  "real:sqrt",
  "real:tan",
  "real:tanh",
  "real:to_string",
  "real:trunc",
  "string:compare",
  "string:concat",
  "string:size"
};
const int Environment::Constants::funArities[] = {
  1, 
  1, 
  2, 
  2, 
  1, 
  2, 
  2, 
  2, 
  2, 
  2, 
  1, 
  2, 
  1, 
  1, 
  1, 
  1, 
  1, 
  1, 
  1, 
  2, 
  1, 
  2, 
  1, 
  1, 
  2, 
  1, 
  1, 
  1, 
  1, 
  1, 
  2, 
  2, 
  2, 
  2, 
  1, 
  2, 
  2, 
  1, 
  1, 
  1, 
  1, 
  1, 
  1, 
  1, 
  1, 
  1, 
  1, 
  2, 
  2, 
  1 
};
NameArray Environment::Constants::functions(evalfun,
					    sizeof(evalfun)/sizeof(char*));
const char* Environment::Constants::evalpred[] = {
  "general:is_int",
  "general:is_real",
  "general:is_string",
  "integer:geq",
  "integer:greater",
  "integer:leq",
  "integer:less",
  "real:geq",
  "real:greater",
  "real:leq",
  "real:less",
  "string:geq",
  "string:greater",
  "string:leq",
  "string:less"
};
const int Environment::Constants::predArities[] = {
  1,
  1,
  1,
  2,
  2,
  2,
  2,
  2,
  2,
  2,
  2,
  2,
  2,
  2,
  2
};
NameArray Environment::Constants::predicates(evalpred,
					    sizeof(evalpred)/sizeof(char*));
Options* Environment::options (const string* name)
{
  TRACER("Environment::options");
  if (! name) {
    return OptionsMap::active->options;
  }
  for (OptionsMap* map = OptionsMap::map; map; map = map->next) {
    if (map->name == *name) {
      return map->options;
    }
  }
  USER_ERROR((string)"options with name " + *name + " not found");
} 
void Environment::setOptions (const string* name,Options* options)
{
  TRACER("Environment::setOptions");
  OptionsMap* mapEntry = 0;
  if (! name) {
    mapEntry = OptionsMap::active;
  }
  else {
    for (OptionsMap* map = OptionsMap::map; map; map = map->next) {
      if (map->name == *name) {
	mapEntry = map;
	break;
      }
    }
  }
  if (mapEntry) { 
    delete mapEntry->options;
    mapEntry->options = options;
    OptionsMap::active = mapEntry;
    return;
  }
  OptionsMap::map = new OptionsMap(*name,options,OptionsMap::map);
  OptionsMap::active = OptionsMap::map;
} 
Environment::OptionsMap::OptionsMap (const string& nm,
				     Options* opt,
				     OptionsMap* nxt)
  : name(nm),
    options(opt),
    next(nxt)
{
} 
int Environment::dropKB (const string& name)
{
  TRACER("Environment::dropKB");
  KnowledgeBase* previous = 0;
  KnowledgeBase* found = 0;
  for (KnowledgeBase* kb = KnowledgeBase::all;
       kb;
       kb = kb->next) {
    if (kb->name == name) {
      found = kb;
      break;
    }
    previous = kb;
  }
  if (! found) { 
    return 0; 
  }
  if (previous) {
    previous->next = found->next;
  }
  else {
    KnowledgeBase::all = found->next;
  }
  int result = found->units.length();
  delete found;
  return result;
} 
void Environment::setKBStatus (const string& kbName,bool act)
{
  TRACER("Environment::setKBStatus");
  KnowledgeBase* kb = findByName(kbName,false);
  if (! kb) { 
    USER_ERROR((string)"no knowledge base with name " + kbName);
  }
  kb->active = act;
} 
int Environment::loadKB (const string& name,const string& fileName,
			  bool* active, const string* syn)
{
  TRACER("Environment::loadKB/1");
  InputSyntax syntax = syn ? Syntax::stringToInputSyntax(*syn) : inputSyntax;
  ifstream in(fileName.c_str());
  if (! in) {
    USER_ERROR((string)"cannot open file" + fileName);
  }
  KnowledgeBase* kb = findByName(name,true);
  if (active) {
    kb->active = *active;
  }
  return loadKB(*kb,in,syntax);
} 
int Environment::loadKB (KnowledgeBase& kb,istream& in,InputSyntax syntax)
{
  TRACER("Environment::loadKB/2");
  UnitList units;
  Syntax::readUnits(in,units,syntax);
  UnitList newUnits(units);
  newUnits.append(kb.units);
  kb.units = newUnits;
  return units.length();
} 
void Environment::collectUnits (UnitList& units)
{
  TRACER("Environment::collectUnits");
  bool kifMode = (mode == MODE_KIF);
  Stack<Unit> us(units);
  for (KnowledgeBase* kb = KnowledgeBase::all;
       kb;
       kb = kb->next) {
    if (! kb->active) {
      continue;
    }
    bool sos = false;
    if (kifMode && kb->name != "$" && kb->name != "$query") {
      sos = true;
    }
    Iterator<Unit> it(kb->units);
    while (it.hasNext()) {
      Unit& u = it.next();
      us.push(u);
    }
  }
} 
Environment::KnowledgeBase* Environment::findByName (const string& name,
						     bool addIfNotFound)
{
  TRACER("Environment::findByName");
  for (KnowledgeBase* kb = KnowledgeBase::all;
       kb;
       kb = kb->next) {
    if (kb->name == name) {
      return kb;
    }
  }  
  if (! addIfNotFound) {
    return 0;
  }
  KnowledgeBase::all = new KnowledgeBase(name,KnowledgeBase::all);
  return KnowledgeBase::all;
} 
Environment::KnowledgeBase::KnowledgeBase (const string& nm,
					   KnowledgeBase* nxt)
  : name(nm),
    active(true),
    next(nxt)
{
} 
void Environment::setInputSyntax (const string& syn)
{
  inputSyntax = Syntax::stringToInputSyntax(syn);
} 
const char* Environment::modeName ()
{
  return Constants::modes[mode];
} 
bool Environment::set (const char* parameter,const char* value)
{
  TRACER("Environment::set");
  int index = Constants::parameters.find(parameter);
  switch (index) {
  case INCLUDE:
    include = value;
    return true;
  case INPUT_SYNTAX:
    inputSyntax = Syntax::stringToInputSyntax(value);
    return true;
  case MODE:
    index = Constants::modes.find(value);
    if (index == -1) {
      break;
    }
    mode = (Mode)index;
    return true;
  case OUTPUT_SYNTAX:
    outputSyntax = Syntax::stringToOutputSyntax(value);
    return true;
  case -1: 
    return false;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
  USER_ERROR((string)"wrong value (" + value + ") for " + parameter);
}
bool Environment::set (const string& parameter,const string& value)
{
  return set(parameter.c_str(),value.c_str());
} 
void Environment::bindFunction (const string& name,const string& function)
{
  int index = Constants::functions.find(function.c_str());
  if (index == -1) {
    USER_ERROR((string)"no evaluable function with name " + function);
  }
  Signature::sig.createFun(name,Constants::funArities[index]).
    bindToEvaluableSymbol(index);
} 
void Environment::bindRelation (const string& name,const string& relation)
{
  int index = Constants::predicates.find(relation.c_str());
  if (index == -1) {
    USER_ERROR((string)"no evaluable predicate with name " + relation);
  }
  Signature::sig.createPred(name,Constants::predArities[index]).
    bindToEvaluableSymbol(index);
} 
void Environment::ass (const string* kbname,
		       UnitList& units,
		       bool addIfNotFound)
{
  string name(kbname ? *kbname : "$");
  KnowledgeBase* kb = findByName(name,addIfNotFound);
  if (! kb) {
    USER_ERROR((string)"no knowledge base with name " + name);
  }
  UnitList newUnits(units);
  newUnits.append(kb->units);
  kb->units = newUnits;
} 
}
