/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Environment__
#define __Environment__
#include <string>
#include <iostream>
#include "Allocator.hpp"
#include "Syntax.hpp"
#include "Unit.hpp"
using namespace std;
namespace Shell {
class Kernel;
class Options;
class UnitList;
enum Mode {
  MODE_CASC_J3 = 0,
  MODE_KIF = 1,
  MODE_SPIDER = 2,
  MODE_VAMPIRE = 3
};
enum Parameter {
  INCLUDE = 0,
  INPUT_SYNTAX = 1,
  MODE = 2,
  OUTPUT_SYNTAX = 3
};
class Environment
{
public:
  static bool set(const string& parameter,const string& value);
  static bool set(const char* parameter,const char* value);
  static void setInputSyntax(const string& syntax);
  static void setOutputSyntax(const string& syntax);
  static void setMode(const string& mode);
  static const char* modeName();
  static const char* inputSyntaxName();
  static const char* outputSyntaxName();
  static Options* options(const string* name);
  static void setOptions(const string* name,Options* options);
  static int loadKB(const string& name,const string& fileName,
		    bool* active, const string* syntax);
  static int dropKB(const string& name);
  static void setKBStatus(const string& kb,bool active);
  static void collectUnits(UnitList& units);
  static void ass(const string* kb,UnitList& units,bool addIfNotFound);
  static void bindFunction(const string& name,const string& function);
  static void bindRelation(const string& name,const string& relation);
  static InputSyntax inputSyntax;
  static OutputSyntax outputSyntax;
  static Mode mode;
  static string include;
  static Kernel* kernel;
private:
  class KnowledgeBase;
  class Constants;
  class OptionsMap {
  public:
    string name;
    Options* options;
    OptionsMap* next;
    OptionsMap(const string& name,Options* options,OptionsMap* next);
    static OptionsMap* map;
    static OptionsMap* active;
    CLASS_NAME("Environment::OptionsMap");
    USE_ALLOCATOR(OptionsMap);
  };
  class KnowledgeBase {
  public:
    KnowledgeBase (const string& name,KnowledgeBase* next);
    string name;
    UnitList units;
    bool active;
    KnowledgeBase* next;
    static KnowledgeBase* all;
    CLASS_NAME("Environment::KnowledgeBase");
    USE_ALLOCATOR(KnowledgeBase);
  };
  static KnowledgeBase* findByName(const string& name,bool addIfNotFound);
  static int loadKB(KnowledgeBase& kb,istream& in,InputSyntax syntax);
}; 
}
#endif
