/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __EqualityAxioms__
#define __EqualityAxioms__
#include "Var.hpp"
namespace Shell {
class Term;
class TermList;
class Atom;
class Clause;
class Formula;
class Unit;
class EqualityAxioms
{
public:
  static bool isEqualityAxiom (const Unit&);
private:
  static bool isEqualityAxiom (const Formula&);
  static bool isEqualityAxiom (const Clause&);
  static bool isReflexivityAxiom (const Formula&);
  static bool isReflexivityAxiom (const Clause&);
  static bool isFunctionReflexivityAxiom (const Formula&);
  static bool isFunctionReflexivityAxiom (const Clause&);
  static bool isPredicateReflexivityAxiom (const Formula&);
  static bool isPredicateReflexivityAxiom (const Clause&);
  static bool isSymmetryAxiom (const Formula&);
  static bool isSymmetryAxiom (const Clause&);
  static bool isTransitivityAxiom (const Formula&);
  static bool isTransitivityAxiom (const Clause&);
  static bool isTautology (const Atom&);
  static bool isTransitivity (const Atom& a1, const Atom& a2, const Atom& a3);
  static bool isFunctionMonotonicity (const Atom& a1, const Atom& a2);
  static bool isPredicateMonotonicity (const Atom& a1, const Atom& a2, 
				       const Atom& a3);
  static bool isSwapOf (const Atom&, const Atom&);
  static bool equalsUpTo (const Term& t1,const Term& t2,Var x,Var y);
  static bool equalsUpTo (const TermList& t1,const TermList& t2,Var x,Var y);
};
}
#endif 
