/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "EqualityProxy.hpp"
#include "Signature.hpp"
#include "Problem.hpp"
#include "Inference.hpp"
using namespace Shell;
EqualityProxy::EqualityProxy (Signature& signature)
  : _equality(signature.equality())
{
  string proxy("E");
  for (;;) {
    PredicateSymbol* symb = signature.existsPred(proxy,2);
    if (symb) { 
      if (symb->isEqualityProxy()) { 
	_proxy = symb;
	return;
      }
      proxy += "E";
    }
    else { 
      _proxy = &signature.createPred(proxy,2);
      _proxy->makeEqualityProxy();
      return;
    }
  }
} 
void EqualityProxy::apply (Problem& problem)
{
  UnitChain::DelIterator us(problem.giveUnits());
  while (us.hasNext()) {
    Unit u(us.next());
    LiteralList ls(u.clause().literals());
    apply(ls);
    if (ls != u.clause().literals()) { 
      Unit result(IR_EQUALITY_PROXY_REPLACEMENT,Clause(ls),u);
      us.replace(result);
    }
  }
  Term x(Var(0));
  Literal lit(true,Atom(*_proxy,TermList(x,TermList(x))));
  LiteralList lits(lit);
  Clause reflexivity(lits);
  problem.addUnit(Unit(reflexivity,
		       new Inference(IR_EQUALITY_PROXY_AXIOM1)));
  Term y(Var(1));
  TermList xy(x,TermList(y));
  Literal lft(false,Atom(*_proxy,xy));
  Literal rht(true,Atom(_equality,xy));
  LiteralList lits2(lft,LiteralList(rht));
  Clause proxy(lits2);
  problem.addUnit(Unit(proxy,
		       new Inference(IR_EQUALITY_PROXY_AXIOM2)));
} 
void EqualityProxy::apply (LiteralList& ls)
{
  if (ls.isEmpty()) {
    return;
  }
  Literal hd(ls.head());
  apply(hd);
  LiteralList tl(ls.tail());
  apply(tl);
  if (hd != ls.head() || tl != ls.tail()) {
    ls = LiteralList(hd,tl);
  }
} 
void EqualityProxy::apply (Literal& l)
{
  Atom a(l.atom());
  if (! a.isEquality()) {
    return;
  }
  TermList ts(a.args());
  if (l.isPositive()) {
    if (ts.head().tag() != VAR ||
	ts.second().tag() != VAR) {
      return;
    }
  }
  else {
    if (ts.head().tag() == VAR ||
	ts.second().tag() == VAR) {
      return;
    }
  }
  l = Literal(l.sign(),Atom(*_proxy,ts));
} 
