/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Flattening.hpp"
#include "Unit.hpp"
#include "Tracer.hpp"
#include "Position.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
using namespace Shell;
void Flattening::flatten (Unit& u)
{
  TRACER("Flattening::flatten (Unit&)");
  ASS(u.unitType() == FORMULA);
  Formula f(u.formula());
  Flattening flat;
  flat.flatten(f,Position());
  if (! f.equals(u.formula())) {
    Inference* inf = new Inference(IR_FLATTEN,UnitList(u));
    inf->setInfo(flat._info);
    u = Unit(f,inf);
  }
} 
void Flattening::flatten (Formula& f, const Position& pos)
{
  TRACER("Flattening::flatten (Formula&, const Position&)");
  switch (f.connective()) {
  case ATOM:
  case TRUE:
  case FALSE:
    return;
  case AND:
  case OR: 
    {
      FormulaList flattenedArgs (f.args());
      if (flatten(flattenedArgs,f.connective(),pos,0)) {
	_info = new InferenceInfo(IR_FLATTEN,pos,_info);
      }
      else if (flattenedArgs == f.args()) { 
	return;
      }
      f = Formula (f.connective(), flattenedArgs);
      return;
    }
  case IMP:
  case IFF:
  case XOR: 
    {
      Formula flattenedLeft (f.left());
      flatten(flattenedLeft,Position(0,pos));
      Formula flattenedRight (f.right());
      flatten(flattenedRight,Position(1,pos));
      if (flattenedLeft == f.left() && flattenedRight == f.right()) {
	return;
      }
      f = Formula (f.connective(), flattenedLeft, flattenedRight);
      return;
    }
  case NOT: 
    {
      Formula flattenedArg (f.uarg());
      flatten(flattenedArg,Position(0,pos));
      if (flattenedArg == f.uarg()) {
	return;
      }
      f = Formula (NOT, flattenedArg);
      return;
    }
  case FORALL:
  case EXISTS: 
    {
      Formula flattenedArg (f.qarg());
      flatten(flattenedArg,Position(0,pos));
      if (flattenedArg.connective() == f.connective()) {
	VarList vs (f.vars());
	vs.append (flattenedArg.vars());
	f = Formula (f.connective(), vs, flattenedArg.qarg());
	_info = new InferenceInfo(IR_FLATTEN,pos,_info);
	return;
      }
      if (flattenedArg == f.qarg()) {
	return;
      }
      f = Formula (f.connective(), f.vars(), flattenedArg);
      return;
    }
  default:
    ASS( false );
    return;
  }
} 
bool Flattening::flatten (FormulaList& fs, 
			  Connective c, 
			  const Position& pos,
			  int k)
{
  TRACER("Flattening::flatten (FormulaList&...)");
  if (fs.isEmpty()) {
    return false;
  }
  Formula flattenedHead (fs.head()); 
  flatten(flattenedHead,Position(k,pos));
  FormulaList flattenedTail(fs.tail());
  bool flattened = flatten(flattenedTail,c,pos,k+1);
  if (flattenedHead.connective() == c) { 
    fs = flattenedHead.args();
    fs.append(flattenedTail);
    return true;
  }
  if (flattenedHead == fs.head() && flattenedTail == fs.tail()) {
    return false;
  }
  fs = FormulaList (flattenedHead, flattenedTail);
  return flattened;
} 
