/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <string>
#include "IntNameTable.hpp"
#include "VampireException.hpp"
using namespace Shell;
IntNameTable::IntNameTable ()
  :
  _buckets (new EntryList* [_noOfBuckets]),
  _nextNumber (0)
{
  if ( ! _buckets ) {
    ALLOCATION_FAILURE;
  }
  for ( int i = _noOfBuckets - 1; i >= 0; i-- ) 
    _buckets [i] = EntryList::empty ();
} 
int IntNameTable::insert (const string& str)
{
  int bucket = hash (str);
  for (EntryList* l = _buckets [bucket]; ! l->isEmpty (); l = l->tail()) {
    if (str == l->head()->right()) { 
      return l->head()->left();
    }
  }
  int result = _nextNumber;
  _nextNumber ++;
  Entry* newEntry = new Entry (result,str);
  EntryList::push (newEntry, _buckets [bucket]);
  return result;
} 
IntNameTable::~IntNameTable ()
{
  for ( int i = _noOfBuckets - 1; i >= 0; i-- ) {
    for (EntryList* lst = _buckets [i];
         ! lst->isEmpty ();
         lst = lst->tail()) {
      delete lst->head ();
    }
    _buckets[i] ->destroy ();
  }
  delete [] _buckets;
} 
int IntNameTable::hash (string str)
{
  int result = 0;
  for (const char* s = str.c_str() ; *s != 0; s++) {
    result = ( result * 10 + *s ) % _noOfBuckets;
  }
  return result;
} 
string IntNameTable::operator[] (int num) const
{
  for ( int i = _noOfBuckets - 1; i >= 0; i-- ) {
    for (EntryList* lst = _buckets [i]; ! lst->isEmpty (); lst = lst->tail()) {
      if (lst->head()->left() == num) { 
	return lst->head()->right();
      }
    }
  }
  ASSERTION_VIOLATION;
} 
