/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Assertion.hpp"
#include "IntegerSet.hpp"
#include "VampireException.hpp"
using namespace Vampire;
bool IntegerSet::member (int n) const
{
  ASS(n >= 0);
  int index = n/BITS_PER_INT;
  if (index >= _words) {
    return false;
  }
  return _set[index] & (1u << (n % BITS_PER_INT));
} 
void IntegerSet::remove (int n)
{
  ASS(n >= 0);
  int index = n/BITS_PER_INT;
  if (index >= _words) {
    return;
  }
  _set[index] &= ~(1u << (n % BITS_PER_INT));
} 
void IntegerSet::insert (int n)
{
  ASS(n >= 0);
  int index = n/BITS_PER_INT;
  if (index >= _words) { 
    int words = _words*2;
    if (words < index+1) {
      words = index+1;
    }
    unsigned int* set = new unsigned[words];
    if (! set) {
      ALLOCATION_FAILURE;
    }
    for (int i = _words-1;i >= 0;i--) {
      set[i] = _set[i];
    }
    for (int j = _words;j < words;j++) {
      set[j] = 0u;
    }
    delete [] _set;
    _set = set;
    _words = words;
  }
  _set[index] |= (1u << (n % BITS_PER_INT));
} 
