/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Kernel__
#define __Kernel__
#include "RunResult.hpp"
#include "VampireKernel.hpp"
#include "XML.hpp"
namespace Shell {
class Problem;
class Options;
class Term;
class Timer;
class TermList;
class RenamingToNew;
class Signature;
class Clause;
class Formula;
class Literal;
class LiteralList;
class Kernel
{
public:
  Kernel(Options& options,Signature& signature);
  int passClauses(const Problem& problem);
  void passTerm(const Term& t, 
		RenamingToNew& renaming);
  void passOptions();
  void passSignature();
  void getClause(const VampireKernel::Clause*, Clause& clause) const;
  void getDefinition(const VampireKernel::Clause*, Formula& formula) const;
  void getTerm(const VampireKernel::Subterm*, Term& term) const;
  const VampireKernel::Clause* refutation() const 
    { return _kernel.refutation(); }
  const VampireKernel& kernel() const { return _kernel; }
  RunResult run(long timeLimitInDeciseconds,
		Problem& problem,
		Timer& timer,
		XML::DeepElement* log);
  RunResult rerun(long timeLimit);
  void closeSession() { _kernel.closeSession(); }
  const Options& options() const { return _options; }
  VampireKernel::TerminationReason terminationReason() const
    {return _kernel.terminationReason();}
private:
  void getLiteralList(const VampireKernel::Literal*, 
		      LiteralList& literals) const;
  void getLiteral(const VampireKernel::Literal*, Literal& literal) const;
  void getTermList(const VampireKernel::Subterm*, TermList& terms) const;
  VampireKernel _kernel;
  Options& _options;
  VampireKernel::Options _kernelOptions;
  Signature& _signature;
  class Register;
};
}
#endif
