/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Literal.hpp"
#include "Signature.hpp"
#include "ShellReader.hpp"
#include "XML.hpp"
using namespace Shell;
LiteralList:: LiteralList (const Literal& hd, const LiteralList& tl)
  : Lst<Literal> (hd,tl) 
{
}
void Literal::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;
  if (_counter == 0) {
    delete this;
  }
} 
void Literal::operator = (const Literal& t)
{
  if (t._data) {
    t._data->ref ();
  }
  if (_data) {
    _data->deref ();
  }
  _data = t._data;
} 
XML::Element* Literal::toXML() const
{
  XML::DeepElement* lit = new XML::DeepElement("literal");
  lit->addAttribute(new XML::TextAttribute("sign",sign() ? "+" : "-"));
  lit->addChild(atom().toXML());
  return lit;
} 
string Literal::toString () const
{
  Reader reader;
  return reader.toString(reinterpret_cast<const Vampire::Literal*>(this));
} 
bool Literal::equals (const Literal& l) const
{
  return sign() == l.sign() &&
         atom().equals(l.atom()); 
} 
bool LiteralList::equals (const LiteralList& lits) const
{
  if (isEmpty()) {
    return lits.isEmpty();
  }
  return head().equals(lits.head()) &&
         tail().equals(lits.tail());
} 
