/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Literal__
#define __Literal__
#include "Atom.hpp"
namespace Shell {
class Literal 
{
public: 
 private:
  class Data;
 public:
  Literal () : _data (0) {}
  Literal (bool sign, const Atom& atom);
  Literal (const Literal& t);
  ~Literal ();
  void operator= (const Literal& rhs);
  bool operator == (const Literal& rhs) const { return _data == rhs._data; }
  bool operator != (const Literal& rhs) const { return _data != rhs._data; }
  bool isPositive () const;
  bool isNegative () const;
  bool sign () const;
  const Atom& atom () const;
  bool isNull () const { return _data == 0; }
  XML::Element* toXML () const;
  bool equals (const Literal&) const;
  string toString () const;
 private:
  Data* _data;
}; 
class Literal::Data 
{
 public:
  Data ();
  Data (bool sign, const Atom& atom)  
    : _counter(1), 
      _sign(sign ? 1u : 0u),
      _atom(atom) {}
  bool isPositive () const { return _sign ; }
  bool isNegative () const  { return ! _sign; }
  bool sign () const  { return _sign; }
  const Atom& atom () const  { return _atom; }
  void ref () { ASS (this); _counter++; }
  void deref ();
  CLASS_NAME("Literal");
  USE_ALLOCATOR(Data);
 protected:
  unsigned int _counter : 30;
  unsigned int _sign : 1;
  unsigned int _selected : 1;
  Atom _atom;
}; 
class LiteralList 
: public Lst<Literal>
{
 public:
  LiteralList () : Lst<Literal> () {}
  LiteralList (const LiteralList& ts) : Lst<Literal> (ts) {}
  explicit LiteralList (const Literal& l) : Lst<Literal> (l) {}
  LiteralList (const Literal& hd, const LiteralList& tl);
  explicit LiteralList (LstData<Literal>* d) : Lst<Literal> (d) {}
  bool equals (const LiteralList&) const;
  const LiteralList& tail () const
    { return static_cast<const LiteralList&>(Lst<Literal>::tail()); }
  string toString (string separator) const;
}; 
inline
Literal::~Literal () 
{
  if (_data) {
    _data->deref ();
  }
} 
inline
Literal::Literal (const Literal& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 
inline
Literal::Literal (bool sign, const Atom& atom)
  :
  _data (new Data(sign,atom))
{
} 
inline
bool Literal::isPositive () const
{
  return _data->sign();
} 
inline
bool Literal::isNegative () const
{
  return ! _data->sign();
} 
inline
bool Literal::sign () const
{
  return _data->sign();
} 
inline
const Atom& Literal::atom () const
{
  return _data->atom();
} 
}
#endif 
