/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Miniscope__
#define __Miniscope__
#include "Var.hpp"
#include "Position.hpp"
namespace Shell {
class VarCounter;
class InferenceInfo;
class Position;
class Miniscope 
{
 public:
  static void miniscope (Unit& u);
 private:
  Connective _quant;
  const Formula& _formula;
  const int _nfs;
  int _realNfs;
  int _nvs;
  bool** _occ;    
  bool* _mem;
  bool** _index;
  Var* _vars;
  int* _N;
  int _maxV;
  Formula* _formulas;
  Position _position;
  InferenceInfo*& _inferences;
  static void miniscope (Formula& f, InferenceInfo*&, const Position&);
  static void miniscope (FormulaList&, Connective c, InferenceInfo*&,
			 const Position& pos, int argNumber,
			 const FormulaList& firstArg);
  static void findOccurring (const Formula& f, const VarList& in, 
			     VarList& out, VarList& removedVars);
  static void markOccurring (const Formula&, VarCounter&);
  static void markOccurring (const TermList&, VarCounter&);
  static void markOccurring (const Term&, VarCounter&);
  static void topMiniscope (Formula& f, Connective Q, const VarList& vs, 
			    InferenceInfo*&, const Position& pos);
  static void topMiniscope (FormulaList& fs, Connective Q, 
			    const VarList& vs, InferenceInfo*&, 
			    const Position& pos, int index);
  static void appendN (const FormulaList& fst, const FormulaList& snd, 
		       int N, FormulaList& result);
  static void splitVarList (const VarList& in, VarList& out, 
			    VarList& removed, VarCounter&);
  Miniscope (Connective Q, const VarList& vars, const Formula&,
	     const Position& pos, InferenceInfo*& infs);
  ~Miniscope ();
  bool noMiniscoping () const;
  void pushOnes ();
  bool pushMany ();
  void putResult (Formula& res);
  void count (const Formula&, int formNumber);
  void count (const Term&, int formNumber);
  void pushOne (int v, int k);
  void occurInSameFormulas (int v, IntList& vs) const; 
  bool occurInSameFormulas (int v, int w) const; 
  void decodeVars (const IntList& in, VarList& out) const;
  void decodeFormulas (const IntList& in, IntList& out) const;
  void decodeFormulas (const IntList& in, IntList& out, 
  		       int inIndex, int outIndex) const;
  void deleteVars (const IntList& vs);
  void collectVars (VarList& vs);
  int maxVar (const VarList& vars);
}; 
}
#endif 
