/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <string>
#include <climits>
#include <cstdlib>
#include "Tracer.hpp"
#include "Problem.hpp"
#include "XML.hpp"
using namespace Shell;
Problem::Problem (const string& name, const UnitList& lst)
  : 
  _hasFormulas (false),
  _name (name)
{
  Iterator<Unit> us (lst);
  while ( us.hasNext() ) {
    Unit u (us.next());
    switch ( u.unitType() ) {
      case FORMULA:
        _hasFormulas = true;
        break;
      default:
        break;
    }
    _units.addLast (u);
  }
} 
Problem::~Problem ()
{
  while (! _units.isEmpty() ) {
    _units.delFirst ();
  }
} 
void Problem::addUnit (Unit u)
{
  TRACER ( "Problem::addUnit" );
  _units.addLast(u);
} 
XML::Element* Problem::toXML () const
{
  XML::DeepElement* problem = new XML::DeepElement("problem");
  UnitChain::Iterator us (_units);
  while (us.hasNext()) {
    problem->addChild(us.next().toXML());
  }
  return problem;
} 
string Problem::toString () const
{
  string problem("");
  UnitChain::Iterator us (_units);
  while (us.hasNext()) {
    problem += us.next().toString() + '\n';
  }
  return problem;
} 
