/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __ProofChecker__
#define __ProofChecker__
#include <string>
#include <iostream>
#include "Unit.hpp"
#include "Signature.hpp"
#include "XML.hpp"
namespace Shell {
class Refutation;
class InferenceInfo;
class Position;
class Substitution;
class VarCounter;
class ProofChecker
{
public:
  static void check (const Refutation& refutation);
  static bool isRectified (const Formula&);
private:
  void checkInference (const Unit& unit);
  static bool hasSinglePremise (const Inference*);
  static bool checkReplace (const Formula& F, const Position& p,
			    const Formula& g, const Formula& G);
  static void isFormulaInference (const Unit& conclusion, 
				  Formula& premise,
				  Position& position);
  static void isFormulaInference (const Unit& conclusion, 
				  Formula& premise1,
				  Formula& premise2,
				  Position& position);
  static void isFormulaInference (const Unit& conclusion, 
				  FormulaList& premises,
				  Position& position);
  static bool isNegationOf (const FormulaList& fs, const FormulaList nfs);
  static bool obtainedBySubstitution(const Formula&, const Formula&,
				     Substitution&);
  static bool obtainedBySubstitution(const Formula&, const Formula&,
				     Substitution&, VarCounter& bound);
  static bool obtainedBySubstitution(const FormulaList&, const FormulaList&,
				     Substitution&, VarCounter& bound);
  static bool obtainedBySubstitution(const Term&, const Term&,
				     Substitution&, VarCounter& bound);
  static bool obtainedBySubstitution(const TermList&, const TermList&,
				     Substitution&, VarCounter& bound);
  static void checkRectify (const Unit& conclusion, const Formula& premise);
  static void checkClosure (const Unit& conclusion, const Formula& premise);
  void checkChoiceAxiom (const Unit& conclusion);
  void checkInput (const Unit& conclusion);
  static void checkFlattening (const Unit& conclusion, 
			       const Formula& premise,
			       const Position& position);
  static void checkNotAnd (const Unit& conclusion, 
			   const Formula& premise,
			   const Position& position,
			   bool isAnd);
  static void checkNotImp (const Unit& conclusion, 
			   const Formula& premise,
			   const Position& position);
  static void checkNotIff (const Unit& conclusion, 
			   const Formula& premise,
			   const Position& position,
			   bool isIff);
  static void checkNotNot (const Unit& conclusion, 
			   const Formula& premise,
			   const Position& position);
  static void checkNotForall (const Unit& conclusion, 
			      const Formula& premise,
			      const Position& position,
			      bool isForall);
  static void checkForallAnd (const Unit& conclusion, 
			      const Formula& premise,
			      const Position& position,
			      bool isAnd);
  static void checkForallOr (const Unit& conclusion, 
			     const Formula& premise,
			     const Position& position,
			     bool isForall);
  static void checkImpToOr (const Unit& conclusion, 
			    const Formula& premise,
			    const Position& position);
  static void checkIffToAnd (const Unit& conclusion, 
			     const Formula& premise,
			     const Position& position);
  static void checkXorToAnd (const Unit& conclusion, 
			     const Formula& premise,
			     const Position& position);
  static void checkForallElimination (const Unit& conclusion, 
				      const Formula& premise,
				      const Position& position);
  static void checkMonotoneReplacement (const Unit& conclusion, 
					const Formula& premise1,
					const Formula& premise2,
					const Position& position);
  static void checkEquivalenceToImplication (const Unit& conclusion, 
					     const Formula& premise,
					     const Position& position);
  static void checkDummyQuantifierRemoval (const Unit& conclusion, 
					   const Formula& premise,
					   const Position& position);
  static void checkQuantifierSwap (const Unit& conclusion, 
				   const Formula& premise,
				   const Position& position);
  static void checkClausify (const Unit& conclusion, const Unit& premise);
  static bool isGoal (const Unit& unit);
  static bool areVariants (const Formula& f1, const Formula& f2);
  bool checkInput (const Formula&);
  bool checkInput (const Atom&);
  bool checkInput (const Clause&);
  static bool scanFlattenedFormOf (Connective, 
				   const FormulaList&,
				   FormulaList&);
  static bool isDisjunctIn (LiteralList& lits, 
			    const Formula& f,
			    Position& pos);
  Signature _input;
  Signature _fresh;
};
class ProofCheckingException 
{
 public:                                
  ProofCheckingException (const string& file,
			  int line,
			  const Unit& unit,
			  bool isPremise,
			  const InferenceInfo* info);
  ProofCheckingException (const string& file,
			  int line,
			  const string& message);
  ~ProofCheckingException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
 private:
  bool _hasUnit;
  string _file;
  int _line;
  bool _isPremise;
  Unit _unit;
  string _message;
  const InferenceInfo* _info;
}; 
}
#define PROOF_CHECKING_ERROR(unit,isPremise,info) \
      throw ProofCheckingException(__FILE__,__LINE__,unit,isPremise,info)
#define PROOF_CHECKING_ERR(msg) \
      throw ProofCheckingException(__FILE__,__LINE__,msg)
#endif
