/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Property__
#define __Property__
#include <string>
using namespace std;
namespace Shell {
class Problem;
class Unit;
class Clause;
class Literal;
class Atom;
class Term;
class Formula;
class VarCounter;
enum Category {
  NEQ,  
  HEQ,
  PEQ,
  HNE,
  NNE,
  FEQ,
  FNE, 
  EPR,
  UEQ
};
enum Prop {
  PR_HAS_X_EQUALS_Y = 1u,
  PR_HAS_FUNCTION_DEFINITIONS = 2u,
  PR_HAS_SUBSET = 4u,
  PR_HAS_EXTENSIONALITY = 8u,
  PR_GROUP = 16u,
  PR_RING = 32u,
  PR_ROBBINS_ALGEBRA = 64u,
  PR_NA_RING = 128u,
  PR_BOOLEAN_ALGEBRA = 256u,
  PR_LATTICE = 512u,
  PR_LO_GROUP = 1024u,
  PR_COMBINATOR_B = 2048u,
  PR_COMBINATOR = 4096u,
  PR_HAS_CONDENSED_DETACHMENT1 = 8192u,
  PR_HAS_CONDENSED_DETACHMENT2 = 16384u,
  PR_HAS_FLD1                  = 32768u,
  PR_HAS_FLD2                  = 65536u
}; 
class Property 
{
 public:
  explicit Property (const Problem&);
  void scanProblem ();
  Category category() const { return _category;}
  string categoryString() const;
  string toString () const;
  string toSpider () const;
  int clauses () const { return _goalClauses + _axiomClauses; }
  int formulas () const { return _goalFormulas + _axiomFormulas; }
  int unitClauses () const { return _unitGoals + _unitAxioms; }
  int hornClauses () const { return _hornGoals + _hornAxioms; }
  bool hasProp (Prop p) const { return _props & p; }
  void addProp (Prop p) { _props |= p; }
  unsigned int props () const { return _props; }
 private:
  static bool hasXEqualsY (const Clause&);
  static bool isXEqualsY (const Atom&);
  static bool hasXEqualsY (const Formula&, VarCounter&, int polarity);
  void scan (const Unit&);
  void scan (const Formula&);
  void scan (const Clause&, bool isAxiom);
  void scan (const Atom&, bool& isGround);
  void scan (const Term&, bool& isGround);
  char axiomTypes () const;
  char goalTypes () const;
  char equalityContent () const;
  char nonGroundUnitContent () const;
  char groundPositiveContent () const;
  char goalsAreGround () const;
  char setClauseSize () const;
  char setLiteralSize () const;
  char setTermSize () const;
  char maxFunArity () const;
  char maxPredArity () const;
  const Problem& _problem;
  int _goalClauses;
  int _axiomClauses;
  int _equalityAtoms;
  int _atoms;
  int _goalFormulas;
  int _axiomFormulas;
  int _subformulas;
  int _terms;
  int _unitGoals;
  int _unitAxioms;
  int _hornGoals;
  int _hornAxioms;
  int _equationalClauses;
  int _pureEquationalClauses;
  int _groundUnitAxioms;
  int _positiveAxioms;
  int _groundPositiveAxioms;
  int _groundGoals;
  int _maxFunArity;
  int _maxPredArity;
  int _variablesInThisClause;
  int _totalNumberOfVariables;
  int _maxVariablesInClause;
  unsigned _props;
  Category _category;
}; 
}
#endif 
