/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Problem.hpp"
#include "Signature.hpp"
#include "PureLiteral.hpp"
#include "VampireException.hpp"
namespace Shell {
class PureLiteral::Counter
{
public:
  Counter();
  int pos;
  int neg;
}; 
PureLiteral::Counter::Counter()
  :
  pos(0),
  neg(0)
{
}
PureLiteral::PureLiteral (Problem& problem,Signature& signature)
  : _problem(problem),
    _counter (new Counter[signature.noOfPreds()]),
    _pure(0)
{
  if (! _counter) {
    ALLOCATION_FAILURE;   
  }
} 
int PureLiteral::findAndRemove ()
{
  int removed = 0;
  UnitChain::Iterator units (_problem.units());
  while (units.hasNext()) {
    const Unit& u = units.next();
    ASS (u.unitType() == CLAUSE);
    count(u.clause().literals());
  }
  while (_pure > 0) {
    UnitChain::DelIterator del (_problem.giveUnits());
    while (del.hasNext()) {
      const Unit& u = del.next();
      ASS (u.unitType() == CLAUSE);
      const LiteralList& ls (u.clause().literals());
      Iterator<Literal> lits(ls);
      while (lits.hasNext()) {
	const Literal& lit = lits.next();
	const PredicateSymbol& pred = lit.atom().functor();
	if (pred.isEquality() || 
	    pred.isAnswer() ||
	    pred.isEvaluable()) {
	  continue;
	}
	const Counter& c = _counter[pred.number()];
	if (c.pos == 0 || c.neg == 0) {
	  uncount(ls);
	  del.del();
	  removed++;
	  break;
	}
      }
    }
  }
  return removed;
} 
void PureLiteral::count (const LiteralList& ls)
{
  Iterator<Literal> lits(ls);
  while (lits.hasNext()) {
    const Literal& lit = lits.next();
    const PredicateSymbol& pred = lit.atom().functor();
    if (pred.isEquality() || 
	pred.isAnswer() ||
	pred.isEvaluable()) {
      continue;
    }
    int p = pred.number();
    Counter& c = _counter[p];
    if (lit.sign()) {
      if (c.pos == 0) { 
	if (c.neg == 0) {
	  _pure++;
	}
	else {
	  _pure--;
	}
      }
      c.pos++;
    }
    else {
      if (c.neg == 0) { 
	if (c.pos == 0) {
	  _pure++;
	}
	else {
	  _pure--;
	}
      }
      c.neg++;
    }
  }
} 
void PureLiteral::uncount (const LiteralList& ls)
{
  Iterator<Literal> lits(ls);
  while (lits.hasNext()) {
    const Literal& lit = lits.next();
    const PredicateSymbol& pred = lit.atom().functor();
    if (pred.isEquality() || 
	pred.isAnswer() ||
	pred.isEvaluable()) {
      continue;
    }
    int p = pred.number();
    Counter& c = _counter[p];
    if (lit.sign()) {
      c.pos--;
      ASS(c.pos >= 0);
      if (c.pos == 0) { 
	if (c.neg == 0) {
	  _pure--;
	}
	else {
	  _pure++;
	}
      }
    }
    else {
      c.neg--;
      ASS(c.neg >= 0);
      if (c.neg == 0) { 
	if (c.pos == 0) {
	  _pure--;
	}
	else {
	  _pure++;
	}
      }
    }
  }
} 
}
