/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Reader__
#define __Reader__
#include <string>
#include "VampireKernelEnvironment.hpp"
namespace Vampire 
{
using namespace std;
enum TermTag {
  VAR = 0u,
  INTERPRETED = 1u,
  COMPOUND = 2u
};
class Background {
public:
  Background ();
  enum StorageKind {
    SHELL = 0,
    PASSIVE = 1,
    ACTIVE = 2,
    TEMPORARY = 3,
    PACKED = 4,
  }; 
  int isClause : 1;
  int isDefinition : 1;
  StorageKind storageKind : 3;
  int derivedFromAxioms : 1;
  int isInput : 1;
  int isObtainedDuringPreprocessing : 1;
  int binaryResolutionUsed : 1;
  int forwardTransitiveChainingUsed : 1;
  int backwardTransitiveChainingUsed : 1;
  int forwardSuperpositionUsed : 1;
  int backwardSuperpositionUsed : 1;
  int equalityResolutionUsed : 1;
  int equalityFactoringUsed : 1;
  int forwardDemodulationUsed : 1;
  int backwardDemodulationUsed : 1;
  int simplificationByEqualityResolutionUsed : 1;
  int forwardSubsumptionResolutionUsed : 1;
  int backwardSubsumptionResolutionUsed : 1;
  int isReanimatedPassive : 1;
  int evaluationUsed : 1;
  int foldingUsed : 1;
  int clausificationUsed : 1;
  int choiceAxiom : 1;
  int monotoneReplacement : 1;
  int forallElimination : 1;
  int rectify : 1;
  int notAnd : 1;
  int notOr : 1;
  int notImp : 1;
  int notIff : 1;
  int notXor : 1;
  int notNot : 1;
  int notForall : 1;
  int notExists : 1;
  int impToOr : 1;
  int iffToAnd : 1;
  int xorToEnd : 1;
  int closure : 1;
  int flattening : 1;
  int normalisation : 1;
  int ennf : 1;
  int nnf : 1;
  int dummyQuantifierRemoval : 1;
  int forallAnd : 1;
  int existsOr : 1;
  int quantifierSwap : 1;
  int forallOr : 1;
  int existsAnd : 1;
  int permutation : 1;
  int reorderEquailities : 1;
  int halfEquivalence : 1;
  int miniscope : 1;
  int skolemise : 1;
  int clausify : 1;
  int removeDuplicateLiterals : 1;
  int kernelInference : 1;
  int equalityProxyReplacement : 1;
  int equalityProxyAxiom1 : 1;
  int equalityProxyAxiom2 : 1;
  int definitionUnfolding : 1;
  int rowVariableExpansion : 1;
  int predicateDefinition : 1;
  int reduceFalseTrue : 1;
};
class FunctionSymbol {};
class PredicateSymbol {};
class Term {};
class TermList {};
class Atom {};
class Literal {};
class LiteralList {};
class Clause {};
class ClauseList {};
class Formula {};
class Unit {};
class UnitList {};
class Reader {
protected:
  virtual TermTag tag(const Term*) const = 0;
  virtual int var(const Term*) const = 0;
  virtual const FunctionSymbol* functor(const Term*) const = 0;
  virtual string name(const FunctionSymbol*) const = 0;
  virtual const TermList* args(const Term*) const = 0;
  virtual const TermList* endOfArgs(const Term*) const;
  virtual const VampireKernelEnvironment::InterpretedConstant* value(const Term*) const = 0;
  virtual bool isEmpty(const TermList* start,const TermList* end) const = 0;
  virtual const Term* head(const TermList*) const = 0;
  virtual const TermList* tail(const TermList*,const TermList* end) const = 0;
  virtual bool isEquality(const Atom*) const = 0;
  virtual const TermList* args(const Atom*) const = 0;
  virtual const TermList* endOfArgs(const Atom*) const;
  virtual const PredicateSymbol* functor(const Atom*) const = 0;
  virtual string name(const PredicateSymbol*) const = 0;
  virtual bool isPositive(const Literal*) const = 0;
  virtual const Atom* atom(const Literal*) const = 0;
  virtual bool isEmpty(const LiteralList* start,
		       const LiteralList* end) const = 0;
  virtual const Literal* head(const LiteralList*) const = 0;
  virtual const LiteralList* tail(const LiteralList* start,
				  const LiteralList* end) const = 0;
  virtual const LiteralList* literals(const Clause*) const = 0;
  virtual const LiteralList* endOfLiterals(const Clause*) const;
  virtual int number(const Unit*) const = 0;
  virtual const Clause* clause(const Unit*) const = 0;
  virtual const Formula* formula(const Unit*) const = 0;
  virtual const UnitList* premises(const Unit*) const = 0;
  virtual const UnitList* endOfPremises(const Unit*) const;
  virtual bool isEmpty(const UnitList* start,
		       const UnitList* end) const = 0;
  virtual const Unit* head(const UnitList*) const = 0;
  virtual const UnitList* tail(const UnitList* start,
			       const UnitList* end) const = 0;
  virtual void saveBackground(const Unit*,Background&) const = 0;
  virtual Unit* origin(const Unit*) const;
public:
  string toString(const Term*) const;
  string toString(const TermList* start,const TermList* end) const;
  string toString(const Atom*) const;
  string toString(const Literal*) const;
  string toString(const LiteralList* start,
		  const LiteralList* end) const;
  string toString(const Clause*) const;
  string toString(const Unit*) const;
  string toString(const Formula*) const;
  string varToString(int var) const;
};
}
#endif
