/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Int.hpp"
#include "Clause.hpp"
#include "Renaming.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
using namespace Shell;
Renaming::Renaming (bool backtrack)
  : _direct(0),
    _inverse(0),
    _nextDirect(0),
    _nextInverse(0),
    _backtrack(backtrack)
{
  TRACER("Renaming::Renaming");
  if (_backtrack) {
    _stack = 0;
    _stackCapacity = 0;
    _stackPos = 0;
  }
} 
Renaming::~Renaming ()
{
  TRACER("Renaming::~Renaming");
  delete [] _direct;
  delete[] _inverse;
  if (_backtrack) {
    delete _stack;
  }
} 
bool Renaming::bind (int v1,int v2)
{
  TRACER("Renaming::bind");
  if (v1 >= _nextDirect) {
    expandToFit(v1,_direct,_nextDirect);
  }
  else if (_direct[v1] == v2) { 
    return true;
  }
  else if (_direct[v1] != -1) { 
    return false;
  }
  _direct[v1] = v2;
  if (_backtrack && _stackPos >= _stackCapacity) { 
    ASS(_stackPos == _stackCapacity);
    _stackCapacity += 16;
    int* newStack = new int[_stackCapacity];
    if (! newStack) {
      ALLOCATION_FAILURE;
    }
    for (int i = _stackPos-1;i >= 0;i--) {
      newStack[i] = _stack[i];
    }
    delete [] _stack;
    _stack = newStack;
  }
  if (v2 >= _nextInverse) {
    expandToFit(v2,_inverse,_nextInverse);
    _inverse[v2] = v1;
    if (_backtrack) {
      _stack[_stackPos++] = v1;
    }
    return true;
  }
  if (_inverse[v2] != -1) { 
    return false;
  }
  _inverse[v2] = v1;
    if (_backtrack) {
      _stack[_stackPos++] = v1;
    }
  return true;
} 
void Renaming::expandToFit (int v,int*& varArray,int& ln)
{
  TRACER("Renaming::expandToFit");
  int newLength = max(ln+8,v+1);
  int* newArray = new int[newLength];
  if (! newArray) {
    ALLOCATION_FAILURE;
  }
  for (int i = 0;i < ln;i++) {
    newArray[i] = varArray[i];
  }
  for (int k = ln;k < newLength;k++) {
    newArray[k] = -1;
  }
  delete [] varArray;
  varArray = newArray;
  ln = newLength;
} 
void Renaming::backtrack (int point)
{
  TRACER("Renaming::backtrack");
  ASS(_backtrack);
  ASS(_stackPos >= point);
  while (_stackPos > point) {
    _stackPos--;
    int v1 = _stack[_stackPos];
    int v2 = _direct[v1];
    _direct[v1] = -1;
    _inverse[v2] = -1;
  }
} 
