/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Run__
#define __Run__
#include <string>
#include "Timer.hpp"
#include "RunResult.hpp"
#include "VampireKernel.hpp"
#include "Var.hpp"
#include "XML.hpp"
using namespace std;
namespace Shell {
class Options;
class Problem;
class FormulaList;
class UnitList;
class Run
{
public:
  Run(Options& options,const VarList& answerVariables);
  virtual ~Run();
  virtual void run() = 0;
  void addToLog(XML::Element*);
  static Run* currentRun() { return _current; }
  void saveLog();
  void terminateBySignal(string signalDescription,int result);
  static string signalToString (int sigNum);
  static void setSignalHandlers();
  void logResult(const string terminationReason,RunResult res);
  static string toString(RunResult res);
protected:
  virtual void reportAllocationFailure() = 0;
  bool _saved;
  Options& _options;
  Timer _timer;
  static Run* _current;
  XML::DeepElement* _log;
  bool _silent;
  VarList _answerVariables;
  static void cancelSignalHandling();
  XML::Element* generalProofAttemptInfo();
  friend class RunSaver;
  class RunSaver {
  public:
    RunSaver(Run* run)
    { _saved = Run::_current;
      Run::_current = run; }
    ~RunSaver() { Run::_current = _saved; };
  private:
    Run* _saved;
  }; 
}; 
}
#endif
