/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Literal.hpp"
#include "ShapeComparator.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
using namespace Shell;
Comparison ShapeComparator::compare (const Literal& l1,const Literal& l2)
{
  TRACER("ShapeComparator::compare (const Literal&...)");
  switch (Compare::compare(l1.sign(),l2.sign())) {
  case LESS:
    return LESS;
  case GREATER:
    return GREATER;
  case EQUAL:
    return compare(l1.atom(),l2.atom());
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
Comparison ShapeComparator::compare (const Atom& a1,const Atom& a2)
{
  TRACER("ShapeComparator::compare (const Atom& a1...)");
  switch (Compare::compare(a1.functor().number(),a2.functor().number())) {
  case LESS:
    return LESS;
  case GREATER:
    return GREATER;
  case EQUAL:
    break;
  }
  if (! a1.isEquality()) { 
    return compare(a1.args(),a2.args());
  }
  const TermList& args1 = a1.args();
  const Term* s1 = &args1.head();
  const Term* s2 = &args1.second();
  const TermList& args2 = a2.args();
  const Term* t1 = &args2.head();
  const Term* t2 = &args2.second();
  Comparison ss = compare(*s1,*s2);
  if (ss == GREATER) { 
    const Term* s = s1;
    s1 = s2;
    s2 = s;
  }
  Comparison tt = compare(*t1,*t2);
  if (tt == GREATER) { 
    const Term* t = t1;
    t1 = t2;
    t2 = t;
  }
  if (ss == EQUAL) { 
    if (tt == EQUAL) { 
      return compare(*s1,*t1);
    }
    switch (compare(*s1,*t2)) {
    case LESS: 
      return LESS;
    case EQUAL: 
      return GREATER;
    case GREATER: 
      return GREATER;
    }
  }
  if (tt != EQUAL) { 
    switch (compare(*s2,*t2)) {
    case LESS: 
      return LESS;
    case EQUAL: 
      return compare(*s1,*t1);
    case GREATER: 
      return GREATER;
    }
  }
  switch (compare(*t1,*s2)) {
  case LESS: 
    return GREATER;
  case EQUAL: 
    return LESS;
  case GREATER: 
    return LESS;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
Comparison ShapeComparator::compare (const TermList& ts1,const TermList& ts2)
{
  TRACER("ShapeComparator::compare (const TermList& ts1...)");
  if (ts1.isEmpty()) {
    ASS(ts2.isEmpty());
    return EQUAL;
  }
  switch (compare(ts1.head(),ts2.head())) {
  case LESS:
    return LESS;
  case EQUAL:
    return compare(ts1.tail(),ts2.tail());
  case GREATER:
    return GREATER;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
Comparison ShapeComparator::compare (const Term& t1,const Term& t2)
{
  TRACER("ShapeComparator::compare (const Term& t1...)");
  switch (Compare::compare((int)t1.tag(),(int)t2.tag())) {
  case LESS:
    return LESS;
  case GREATER:
    return GREATER;
  case EQUAL:
    break;
  }
  switch (t1.tag()) {
  case VAR:
    return EQUAL;
  case COMPOUND:
    switch (Compare::compare(t1.functor().number(),t2.functor().number())) {
    case LESS:
      return LESS;
    case GREATER:
      return GREATER;
    case EQUAL:
      return compare(t1.args(),t2.args());
    }
  case INTERPRETED:
    return (Comparison)t1.value().compare(t2.value());
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
