/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __signature__
#define __signature__
#define NO_OF_BUCKETS 2039
#include <string>
#include "Atom.hpp"
#include "List.hpp"
using namespace std;
namespace Shell {
class Symbol;
class FunctionSymbol;
class PredicateSymbol;
class SymbolList;
class Problem;
class Signature 
{ 
 public:
  class Iterator {
   public:
    explicit Iterator(SymbolList** buckets);
    bool hasNext();
    Symbol& next();
   private:
    SymbolList** _firstBucket;
    SymbolList** _currentBucket;
    SymbolList* _currentList;
  }; 
  class FunIterator 
    : public Iterator {
   public:
    explicit FunIterator(Signature& s) :
      Iterator(s._fbuckets)
      {}
  };
  class PredIterator 
    : public Iterator {
   public:
    explicit PredIterator(Signature& s) :
      Iterator(s._pbuckets)
      {}
  };
  Signature();
  ~Signature();
  FunctionSymbol& createFun(const string& name, int arity)
    { return static_cast<FunctionSymbol&>(create(name, arity, 
				        	 _fbuckets, false)); }
  PredicateSymbol& createPred(const string& name, int arity)
    { return static_cast<PredicateSymbol&>(create(name, arity, 
						  _pbuckets, true)); }
  PredicateSymbol& createAnswerPredicate(int arity);
  void registerKernelNumber(Symbol& symbol, unsigned long number);
  const FunctionSymbol& getFunctionByItsKernelNumber(unsigned long number);
  const FunctionSymbol& getFunctionByItsKernelNumber(unsigned long number,
	 					      int arity);
  const PredicateSymbol& getPredicateByItsKernelNumber(unsigned long number);
  const PredicateSymbol& getPredicateByItsKernelNumber(unsigned long number,
						        int arity);
  PredicateSymbol& equality() { return *_equality; }
  int noOfPreds() const { return _noOfPreds; }
  int noOfFuns() const { return _noOfFuns; }
  string summary();
  XML::Element* toXML() const;
  FunctionSymbol& newSkolemFunction(int arity);
  PredicateSymbol& newName(int arity);
  void setSkolemPrefix(const string& prefix);
  FunctionSymbol* existsFun(const string& name, int arity)
  { return static_cast<FunctionSymbol*>(exists(name,arity,_fbuckets)); }
   PredicateSymbol* existsPred(const string& name, int arity)
  { return static_cast<PredicateSymbol*>(exists(name,arity,_pbuckets)); }
  void resetSkolemFunctionNumbers() {_lastSkolem = 0;}
  static Signature sig;
 private:  
  SymbolList * _fbuckets [NO_OF_BUCKETS];
  SymbolList * _pbuckets [NO_OF_BUCKETS];
  int _noOfPreds;
  int _noOfFuns;
  int _constants;
  int _propositionalSymbols;
  int _maxFunArity;
  int _maxPredArity;
  PredicateSymbol* _equality;
  Symbol** _kernelSymbols;
  int _maxNumberOfKernelSymbols;
  int _lastSkolem;
  int _lastName;
  string _skolemPrefix;
  string _namePrefix;
  static int hash(const string& name,int arity);  
  Symbol& create(const string& str, 
		 int arity, 
		 SymbolList** buckets, 
		 bool isPred);
  static Symbol* exists(const string& name, int arity, SymbolList** lst);
  void expandKernelSymbolsToFit(unsigned long symbolKernelNumber);
 friend class SymCounter;
 friend class FunIterator;
 friend class PredIterator;
}; 
}
#endif 
