/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "SubformulaIterator.hpp"
#include "Tracer.hpp"
#include "Formula.hpp"
using namespace Shell;
SubformulaIterator::SubformulaIterator (const Formula& t)
  : _current(const_cast<Formula*>(&t)),
    _reserve(0)
{
} 
SubformulaIterator::SubformulaIterator (const FormulaList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 
bool SubformulaIterator::hasNext ()
{
  TRACER("SubformulaIterator::hasNext");
  if (_current) {
    return true;
  }
  while (_reserve) {
    if (_reserve->_isList) {
      FormulaList* first = _reserve->_list;
      if (first->isEmpty()) {
	Element* rest = _reserve->_rest;
	delete _reserve;
	_reserve = rest;
      }
      else { 
	_current = const_cast<Formula*>(&(first->head()));
	_reserve->_list = const_cast<FormulaList*>(&(first->tail()));
	return true;
      }
    }
    else { 
      _current = _reserve->_formula;
      Element* rest = _reserve->_rest;
      delete _reserve;
      _reserve = rest;
      return true;
    }
  }
  return false;
} 
const Formula& SubformulaIterator::next ()
{
  Formula& result = *_current;
  switch (result.connective()) {
  case ATOM:
  case TRUE:
  case FALSE:
    _current = 0;
    break;
  case AND:
  case OR:
    _reserve = new Element(result.args(),_reserve);
    _current = 0;
    break;
  case IMP:
  case IFF:
  case XOR:
    _current = const_cast<Formula*>(&result.left());
    _reserve = new Element(result.right(),_reserve);
    break;
  case NOT:
    _current = const_cast<Formula*>(&result.uarg());
    break;
  case FORALL:
  case EXISTS:
    _current = const_cast<Formula*>(&result.qarg());
    break;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
  return result;
} 
SubformulaIterator::~SubformulaIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
