/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Symbol__
#define __Symbol__
#include "List.hpp"
#include "XML.hpp"
using namespace std;
namespace Shell {
class Symbol 
{
public:
  int arity () const { return _arity; }
  const string& name () const { return _name;}
  int number () const { return _number; }
  void setNumber (int number) { _number = number; }
  bool operator == (const Symbol& symbol) const 
  { return this == &symbol; }
  bool operator != (const Symbol& symbol) const 
  { return this != &symbol; }
  unsigned long kernelNumber () const { return _kernelNumber; }
  void setKernelNumber (unsigned long n) { _kernelNumber = n; }
  int weight () const { return _weight; }
  int precedence () const { return _precedence; }
  void setWeight (int weight) { _weight = weight; }
  void setPrecedence (int precedence) { _precedence = precedence; }
  bool isFunction () const { return _isFunction; }
  bool isName () const { return _isName; }
  void makeName () { _isName = 1; }
  bool isEqualityProxy () const { return _isEqualityProxy; }
  void makeEqualityProxy () { _isEqualityProxy = 1; }
  bool isSkolemFunction () const { return _isSkolem; }
  void makeSkolem () { _isSkolem = 1; }
  bool isEvaluable () const { return _isEvaluable; }
  int evaluatorNumber () const
  {
    ASS(isEvaluable());
    return _evaluatorNumber;
  }
  void bindToEvaluableSymbol(int evaluatorNumber)
  {
    _isEvaluable = 1;
    _evaluatorNumber = evaluatorNumber;
  }
  XML::Element* toXML() const;
  CLASS_NAME("Symbol");
  USE_ALLOCATOR(Symbol);
protected:
  Symbol (const string& name, int arity, int& number);
  Symbol (const Symbol&);
  string _name;
  int _arity;
  int _number;
  unsigned long _kernelNumber;
  int _weight;
  int _precedence;
  unsigned int _isFunction : 1;
  unsigned int _isEqualityProxy : 1;
  unsigned int _isSkolem : 1;
  unsigned int _isEvaluable : 1;
  unsigned int _isName : 1;
  unsigned int _isAnswer : 1;
  int _evaluatorNumber;
}; 
class SymbolList :
  public List <Symbol*> 
{
public:
  ~SymbolList() { delete _head; }
  static SymbolList* empty ()
  { return static_cast<SymbolList*>(List<Symbol*>::empty()); }
  SymbolList* tail () const
  { return static_cast<SymbolList*>(List<Symbol*>::tail()); }
  SymbolList (Symbol* head, SymbolList* tail) :
    List<Symbol*>(head,tail) {}
};
class FunctionSymbol :
  public Symbol
{
public:
  FunctionSymbol (const string& name, int arity, int& count);
  USE_ALLOCATOR(FunctionSymbol);
}; 
class PredicateSymbol :
  public Symbol
{
public:
  PredicateSymbol(const string& name, int arity, int& count);
  bool isEquality () const { return _number == 0; }
  USE_ALLOCATOR(PredicateSymbol);
  int level() const { return _level; }
  int negativeLevel() const { return _negativeLevel; }
  void setLevel(int level) { _level = level; }
  void setNegativeLevel(int level) { _negativeLevel = level; }
  void makeAnswer () { _isAnswer = 1; }
  bool isAnswer () const { return _isAnswer; }
private:
  int _level;
  int _negativeLevel;
}; 
ostream& operator << (ostream& str, const Symbol* p);
}
#endif
