/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "TermVarIterator.hpp"
#include "Tracer.hpp"
#include "Literal.hpp"
using namespace Shell;
TermVarIterator::TermVarIterator (const Term& t)
  : _current(const_cast<Term*>(&t)),
    _reserve(0)
{
} 
TermVarIterator::TermVarIterator (const TermList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 
TermVarIterator::TermVarIterator (const Atom& a)
  : _current(0),
    _reserve(new Element(a.args(),0))
{
} 
TermVarIterator::TermVarIterator (const Literal& l)
  : _current(0),
    _reserve(new Element(l.atom().args(),0))
{
} 
bool TermVarIterator::hasNext ()
{
  TRACER("TermVarIterator::hasNext");
  for (;;) { 
    if (! _current) { 
      while (_reserve) {
	TermList* first = _reserve->_list;
	if (first->isEmpty()) {
	  Element* rest = _reserve->_rest;
	  delete _reserve;
	  _reserve = rest;
	}
	else { 
	  _current = const_cast<Term*>(&(first->head()));
	  _reserve->_list = const_cast<TermList*>(&(first->tail()));
	  goto current_exists;
	}
      }
      return false;
    }
  current_exists:
    switch (_current->tag()) {
    case VAR:
      return true;
    case INTERPRETED:
      _current = 0;
      break;
    case COMPOUND:
      _reserve = new Element(_current->args(), _reserve);
      _current = 0;
      break;
    }
  }
} 
Var TermVarIterator::next ()
{
  Var result = _current->var();
  _current = 0;
  return result;
} 
TermVarIterator::~TermVarIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
