/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __MyTracer__
#  define __MyTracer__
#if DEBUG_SHELL
#include <iostream>
#include <string>
#include "XML.hpp"
using namespace std;
namespace Shell {
enum ControlPointKind {
  CP_ENTRY,
  CP_EXIT,
  CP_MID
}; 
class MyTracer {
 public:
  explicit MyTracer (const string& fun);
  virtual ~MyTracer ();
  static void printStack (ostream&);
  static XML::Element* stackToXML();
  static void controlPoint (const string&);
  static long passedControlPoints () { return _passedControlPoints; }
 private:
  string _fun;
  MyTracer* _previous;
  static MyTracer* _current;
  void printStack (ostream&, int& depth);
  static string _lastControlPoint;
  static long _passedControlPoints;
  static ControlPointKind _lastPointKind;
  static void controlPoint (const string&, ControlPointKind);
  static void outputLastControlPoint (ostream& str);
};
}
#  define TRACER(Fun) MyTracer tmp (Fun)
#  define CONTROL(description) MyTracer::controlPoint(description)
#  define AFTER(number,command) \
      if (MyTracer::passedControlPoints() >= number) { command };
#  define BETWEEN(number1,number2,command) \
      if (MyTracer::passedControlPoints() >= number1 && \
	  MyTracer::passedControlPoints() <= number2) \
      { command };
#else
#  define TRACER(Fun) 
#  define CONTROL(description)
#endif
#endif 
