/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BS_INDEX_HPP
#define BS_INDEX_HPP 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "SkipList.hpp"
#include "GList.hpp"
#include "Tuple.hpp"
#include "PathIndexWithOpenPaths.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX
#define DEBUG_NAMESPACE "BSIndex"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BSIndex
  {
  public:
    class Integrator;
    class Retrieval;
    class Removal;
    typedef BK::Pair<ulong,Term> LTPair;
    typedef BK::Pair<ulong,ulong> LAPair;
    typedef BK::Triple<ulong,ulong,Term> LATTriple;
    typedef BK::GList<BK::GlobAlloc,ulong,BSIndex> LList; 
    typedef BK::GList<BK::GlobAlloc,LTPair,BSIndex> LTList; 
    typedef BK::GList<BK::GlobAlloc,LAPair,BSIndex> LAList; 
    typedef BK::GList<BK::GlobAlloc,LATTriple,BSIndex> LATList; 
    typedef BK::VoidSkipList<BK::GlobAlloc,ulong> CSkipList; 
    typedef BK::SkipList<BK::GlobAlloc,LList*,ulong> CLSkipList; 
    typedef BK::SkipList<BK::GlobAlloc,LTList*,ulong> CLTSkipList; 
    typedef BK::SkipList<BK::GlobAlloc,LAList*,ulong> CLASkipList; 
    typedef BK::SkipList<BK::GlobAlloc,LATList*,ulong> CLATSkipList; 
    typedef PathIndexWithOpenPaths<CLTSkipList,CLSkipList> 
    OrdLitPITree;
    typedef PathIndexWithOpenPaths<CLATSkipList,CLASkipList> 
    SymLitPITree;  
    typedef PathIndexWithOpenPathsDestruction<CLTSkipList,CLSkipList> 
    OrdLitPITreeDestruction;
    typedef PathIndexWithOpenPathsDestruction<CLATSkipList,CLASkipList> 
    SymLitPITreeDestruction;  
  public:
    BSIndex();
    ~BSIndex();
    void reset();
    ulong maxlNumOfNonproplLits() const { return _maxlNumOfNonproplLits; };
  private: 
    static void remove(const ulong& clNum,CLSkipList& lst)
    {
      LList* llist;
      if (lst.remove(clNum,llist)) LList::destroyList(llist); 
    };
    static void remove(const ulong& clNum,CLTSkipList& lst)
    {
      LTList* ltlist;
      if (lst.remove(clNum,ltlist)) LTList::destroyList(ltlist); 
    }; 
    static void remove(const ulong& clNum,CLASkipList& lst)
    {
      LAList* lalist;
      if (lst.remove(clNum,lalist)) LAList::destroyList(lalist); 
    };  
    static void remove(const ulong& clNum,CLATSkipList& lst)
    {
      LATList* latlist;
      if (lst.remove(clNum,latlist)) LATList::destroyList(latlist); 
    };
    static void recycle(CLSkipList& lst)
    { 
      LList* llist;
      CLSkipList::Destruction destr(lst);
      while (destr.next(llist)) LList::destroyList(llist);      
    };
    static void recycle(CLTSkipList& lst)
    { 
      LTList* ltlist;
      CLTSkipList::Destruction destr(lst);
      while (destr.next(ltlist)) LTList::destroyList(ltlist);      
    }; 
    static void recycle(CLASkipList& lst)
    { 
      LAList* lalist;
      CLASkipList::Destruction destr(lst);
      while (destr.next(lalist)) LAList::destroyList(lalist);      
    };
    static void recycle(CLATSkipList& lst)
    { 
      LATList* latlist;
      CLATSkipList::Destruction destr(lst);
      while (destr.next(latlist)) LATList::destroyList(latlist);      
    };
    void insertPropLit(const Term& lit,ulong clauseNum);
    void removePropLit(const Term& lit,ulong clauseNum);
    const CSkipList* propLitOccurences(const Term& lit)
    {
      CALL("propLitOccurences(const Term& lit)");
      DOP(_propLitOccurences.unfreeze());
      CSkipList* occList = _propLitOccurences.sub(lit.HeaderNum());
      DOP(_propLitOccurences.freeze());
      if (!occList) 
	{
	  occList = new CSkipList();
	  _propLitOccurences[lit.HeaderNum()] = occList;
	};
      return occList;
    };
  private: 
    ulong _maxlNumOfNonproplLits;
    BK::Array<BK::GlobAlloc,CSkipList*,BK::DefaultRelocation<CSkipList*>,512UL> _propLitOccurences;
    OrdLitPITree::Trie _ordLitInd[1];
    SymLitPITree::Trie _symLitInd[1];
    friend class Integrator;
    friend class Retrieval;
    friend class Removal; 
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX
#define DEBUG_NAMESPACE "BSIndex::Integrator"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BSIndex::Integrator
  { 
  public:
    void init(BSIndex* index)
    {
      CALL("init(BSIndex* index)");
      _index = index;
      BK::init(_ordLitIntegrator);
      BK::init(_symLitIntegrator);
    };
    void destroy()
    {
      CALL("destroy()");
      BK::destroy(_symLitIntegrator);
      BK::destroy(_ordLitIntegrator);
      BK_CORRUPT(*this);
    };
    void reset(BSIndex* index) { _index = index; }; 
    void clause(const ulong& clNum) { 
      _clauseNum = clNum; 
      _litNum = 0; 
    };
    void endOfClause()
    {
      if (_litNum > _index->_maxlNumOfNonproplLits)
	_index->_maxlNumOfNonproplLits = _litNum;
    };
    void propLit(const Term& lit);
    void litHeader(const Term& hd)
    {
      _ordLitIntegrator.reset(_index->_ordLitInd[0]);
      _ordLitIntegrator.nonconstFunc(hd);   
      ++_litNum;
    };
    void subterm(const Term& t);
    void symLitHeader(const Term& hd)
    {
      _symLitIntegrator.reset(_index->_symLitInd[0]);
      _symLitIntegrator.nonconstFunc(hd);   
      ++_litNum;
      _symLitArgNum = 0;
      _symLitHeader = hd;
    };
    void symLitSecondArg() 
    { 
      _symLitIntegrator.reset(_index->_symLitInd[0]);
      _symLitIntegrator.nonconstFunc(_symLitHeader);   
      _symLitArgNum = 1; 
    };
    void symLitSubterm(const Term& t);
  private:
    BSIndex* _index;
    OrdLitPITree::Integrator _ordLitIntegrator;
    SymLitPITree::Integrator _symLitIntegrator;
    ulong _clauseNum;
    ulong _litNum;
    ulong _symLitArgNum;
    Term _symLitHeader;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX
#define DEBUG_NAMESPACE "BSIndex::Removal"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BSIndex::Removal
  {
  public:
    void init(BSIndex* index)
    {
      CALL("init(BSIndex* index)");
      _index = index;
      BK::init(_ordLitRemoval);
      BK::init(_symLitRemoval);
    };
    void destroy()
    {
      CALL("destroy()");
      BK::destroy(_symLitRemoval);
      BK::destroy(_ordLitRemoval);
      BK_CORRUPT(*this);
    };
    void reset(BSIndex* index) { _index = index; }; 
    void clause(const ulong& clNum) { _clauseNum = clNum; };
    void propLit(const Term& lit);
    bool litHeader(const Term& hd)
    {
      CALL("litHeader(const Term& hd)");
      _ordLitRemoval.reset(_index->_ordLitInd[0]);
      return _ordLitRemoval.nonconstFunc(hd);
    };
    bool nonconstFunc(const Term& f)
    {
      CALL("nonconstFunc(const Term& f)");
      if (_ordLitRemoval.occurences()) BSIndex::remove(_clauseNum,_ordLitRemoval.indexedObj());
      return _ordLitRemoval.nonconstFunc(f);
    };
    bool constant(const Term& c)
    {
      CALL("constant(const Term& c)");
      if (_ordLitRemoval.occurences()) BSIndex::remove(_clauseNum,_ordLitRemoval.indexedObj());
      if (_ordLitRemoval.constant(c))
	{
	  BSIndex::remove(_clauseNum,_ordLitRemoval.constIndObj());
	  return true;
	}; 
      return false;
    };
    void variable()
    {
      CALL("variable()");
      if (_ordLitRemoval.occurences()) BSIndex::remove(_clauseNum,_ordLitRemoval.indexedObj());
      _ordLitRemoval.after();
    };
    void after() { _ordLitRemoval.after(); };
    bool symbol(const Term& s) 
    {
      CALL("symbol(const Term& s)");
      if (s.isVariable()) { variable(); return true; };
      if (s.arity()) return nonconstFunc(s); 
      return constant(s);
    };
    void endOfLit() { CALL("endOfLit()"); _ordLitRemoval.recycleObsoleteNodes(); };
    bool symLitHeader(const Term& hd)
    {
      CALL("symLitHeader(const Term& hd)");
      _symLitHeader = hd;
      _symLitRemoval.reset(_index->_symLitInd[0]);
      return _symLitRemoval.nonconstFunc(hd);
    };
    bool symLitSecondArg()
    {
      CALL("symLitSecondArg()");
      _symLitRemoval.recycleObsoleteNodes();
      _symLitRemoval.reset(_index->_symLitInd[0]);
      return _symLitRemoval.nonconstFunc(_symLitHeader);  
    };
    bool symLitNonconstFunc(const Term& f)
    {
      CALL("symLitNonconstFunc(const Term& f)");
      if (_symLitRemoval.occurences()) BSIndex::remove(_clauseNum,_symLitRemoval.indexedObj());
      return _symLitRemoval.nonconstFunc(f);
    };
    bool symLitConstant(const Term& c)
    {
      CALL("symLitConstant(const Term& c)");
      if (_symLitRemoval.occurences()) BSIndex::remove(_clauseNum,_symLitRemoval.indexedObj());
      if (_symLitRemoval.constant(c))
	{
	  BSIndex::remove(_clauseNum,_symLitRemoval.constIndObj());
	  return true;
	}; 
      return false;
    };
    void symLitVariable()
    {
      CALL("symLitVariable()");
      if (_symLitRemoval.occurences()) BSIndex::remove(_clauseNum,_symLitRemoval.indexedObj());
      _symLitRemoval.after();
    };
    void symLitAfter() { _symLitRemoval.after(); };
    bool symLitSymbol(const Term& s) 
    {
      CALL("symLitSymbol(const Term& s)");
      if (s.isVariable()) { symLitVariable(); return true; };
      if (s.arity()) return symLitNonconstFunc(s); 
      return symLitConstant(s);
    };
    void endOfSymLit() { CALL("endOfSymLit()"); _symLitRemoval.recycleObsoleteNodes(); };
  private:
    BSIndex* _index;
    OrdLitPITree::Removal _ordLitRemoval;
    SymLitPITree::Removal _symLitRemoval;
    Term _symLitHeader;
    ulong _clauseNum;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX
#define DEBUG_NAMESPACE "BSIndex::Retrieval"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BSIndex::Retrieval
  {
  public:
    void init(BSIndex* index)
    {
      CALL("init(BSIndex* index)");
      _index = index;
      BK::init(_ordLitRetrieval);
      BK::init(_symLitRetrieval);
    };
    void destroy()
    {
      CALL("destroy()");
      BK::destroy(_symLitRetrieval);
      BK::destroy(_ordLitRetrieval);
      BK_CORRUPT(*this);
    };
    void reset(BSIndex* index) { _index = index; }; 
    bool propLit(const Term& lit,const CSkipList*& occList)
    {
      CALL("propLit(const Term& lit,const CSkipList*& occList)");
      occList = _index->propLitOccurences(lit);
      return (bool)(*occList);
    };
    bool litHeader(const Term& hd)
    {
      return ((_ordLitRetrieval.reset(_index->_ordLitInd[0])) && (_ordLitRetrieval.nonconstFunc(hd)));
    };
    bool nonconstFunc(const Term& f)
    {
      return _ordLitRetrieval.nonconstFunc(f); 
    };
    bool constant(const Term& c,const CLSkipList*& occList)
    {
      if (_ordLitRetrieval.constant(c))
	{
	  occList = &(_ordLitRetrieval.constIndObj());
	  return (bool)(*occList);
	}; 
      return false;
    };
    bool variable(const CLTSkipList*& occList)
    {
      occList = &(_ordLitRetrieval.indexedObj());  
      _ordLitRetrieval.after();
      return (bool)(*occList);
    };
    bool symLitHeader(const Term& hd)
    {
      _symLitHeader = hd;
      return ((_symLitRetrieval.reset(_index->_symLitInd[0])) && (_symLitRetrieval.nonconstFunc(hd)));
    };
    bool symLitSecondArg()
    {
      return ((_symLitRetrieval.reset(_index->_symLitInd[0])) && (_symLitRetrieval.nonconstFunc(_symLitHeader)));   
    };
    bool symLitNonconstFunc(const Term& f)
    {
      return _symLitRetrieval.nonconstFunc(f); 
    };
    bool symLitConstant(const Term& c,const CLASkipList*& occList)
    {
      if (_symLitRetrieval.constant(c))
	{
	  occList = &(_symLitRetrieval.constIndObj());
	  return (bool)(*occList);
	}; 
      return false;
    };
    bool symLitVariable(const CLATSkipList*& occList)
    {
      occList = &(_symLitRetrieval.indexedObj());  
      _symLitRetrieval.after();
      return (bool)(*occList);
    };
  private:
    BSIndex* _index;
    OrdLitPITree::Retrieval _ordLitRetrieval;
    SymLitPITree::Retrieval _symLitRetrieval;
    Term _symLitHeader;
  }; 
}; 
#endif
