/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BS_INDEX_FRONT_END_HPP
#define BS_INDEX_FRONT_END_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "BSIndex.hpp"
#include "Clause.hpp"
#include "Term.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX_FRONT_END
#  define DEBUG_NAMESPACE "BSIndexFrontEnd"
#endif
#include "debugMacros.hpp"
namespace VK
{
class BSIndexFrontEnd 
{
 public: 
  void init()
  {
    CALL("init(ClauseNumIndex* clauseNumIndex)");
    _isLocked = 0UL;
    BK::init(_index);
    _integrator.init(&_index);
    _removal.init(&_index);
#ifdef DEBUG_NAMESPACE
    BK::init(_termTraversal,"BackwardSubsumption::_termTraversal");
#else
    BK::init(_termTraversal);
#endif
  };
  void destroy()
  {
    CALL("destroy()");
    BK::destroy(_termTraversal);
    _removal.destroy();
    _integrator.destroy();
    BK::destroy(_index);
    BK_CORRUPT(*this);
  };
  void reset()
  {
    CALL("reset()");
    ASSERT(!_isLocked);
    _index.reset();
  };
  void lock() { 
    CALL("lock()");
    ++_isLocked; 
  };
  void unlock() {
    CALL("unlock()");
    ASSERT(_isLocked > 0UL);
    --_isLocked; 
  };
  bool isLocked() const { return _isLocked != 0UL; };
  void integrate(Clause* cl);
  void remove(Clause* cl);
  BSIndex* index() { return &_index; };
 private:
  ulong _isLocked;
  BSIndex _index;
  BSIndex::Integrator _integrator;
  BSIndex::Removal _removal;
  Term::Traversal _termTraversal;
}; 
}; 
#endif
