/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BS_MACHINE_HPP
#define BS_MACHINE_HPP 
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "GlobalStopFlag.hpp"
#include "Array.hpp"
#include "GIncCountingSort.hpp"
#include "Stack.hpp"
#include "ExpandingStack.hpp"
#include "StaticCellArray.hpp"
#include "BSIndex.hpp"
#include "Term.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BSMachine
  {
  public:
    enum Tag { PropLit, Const, Var, SymLitConst, SymLitVar};
    typedef BSIndex::LTPair LTPair;  
    typedef BSIndex::LAPair LAPair;
    typedef BSIndex::LATTriple LATTriple;
    typedef BSIndex::LList LList;
    typedef BSIndex::LTList LTList;
    typedef BSIndex::LAList LAList;
    typedef BSIndex::LATList LATList;
    typedef BSIndex::CSkipList CSkipList;
    typedef BSIndex::CLSkipList CLSkipList;
    typedef BSIndex::CLTSkipList CLTSkipList;
    typedef BSIndex::CLASkipList CLASkipList;
    typedef BSIndex::CLATSkipList CLATSkipList;
  public:
    BSMachine();
    ~BSMachine();
    bool subsumeNextMultisetMode(ulong& subsumedClNum);
    bool resetQuery(ulong numOfLiterals,const BSIndex* index);
    bool endOfQuery();
    bool queryPropLit(const CSkipList* occList);
    void queryNextLit();
    bool queryEndOfLit();
    bool queryConstant(const Term& c,const CLSkipList* occList);
    bool queryVariable(const Term& v,const CLTSkipList* occList,bool primary);
    void queryNextSymLit();
    bool queryEndOfSymLit();
    void querySymLitSecondArg() { _symLitEqualizer[_nextSymLit].secondArg(); };
    bool querySymLitConstant(const Term& c,const CLASkipList* occList);
    bool querySymLitVariable(const Term& v,const CLATSkipList* occList,bool primary);
    void resetSubstitutionCollection() { _substJoin.resetResultCollection(); }; 
    bool substNextPair(ulong& var,Term& term) { 
      return _substJoin.resultNextPair(var,term); 
    };
    bool litIsSubsumedByNonGround(ulong litNum) const {
      return _substJoin.litIsSubsumed(litNum);
    };
#ifndef NO_DEBUG_VIS
    ostream& outputSubst(ostream& str) { return _substJoin.outputResult(str); };  
#endif 
#ifndef NO_DEBUG_VIS
    ostream& outputLitEqualizers(ostream& str) const;
#endif
  private:
    class Stream
    {
    public:
      Stream() 
	: _C(new CSkipList::Surfer)
      {
	CALL("constructor Stream::Stream()");
      }
      ~Stream()
      {
	delete _C;
      }
      const Tag tag() const { return _tag; };
      void reset(const CSkipList* sl) { 
	_tag = PropLit;
	_C->reset(*sl);
      };
      void reset(const CLSkipList* sl) { 
	_tag = Const;
	_CL->reset(*sl);
      }; 
      void reset(const CLTSkipList* sl) { 
	_tag = Var;
	_CLT->reset(*sl);
      };
      void reset(const CLASkipList* sl) { 
	_tag = SymLitConst;
	_CLA->reset(*sl);
      }; 
      void reset(const CLATSkipList* sl)  { 
	_tag = SymLitVar;
	_CLAT->reset(*sl);
      };
      bool equalize(ulong& clNum,bool& notEnd);
      bool equalize(ulong clNum);
      const LList* litList() const;
      const LTList* litTermList() const;
      const LAList* litArgList() const;
      const LATList* litArgTermList() const;
      bool topClauseNum(ulong& clNum) const;
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const;
#endif
    private:
      Tag _tag;
      union {
	BSIndex::CSkipList::Surfer* _C;
	BSIndex::CLSkipList::Surfer* _CL;
	BSIndex::CLTSkipList::Surfer* _CLT;
	BSIndex::CLASkipList::Surfer* _CLA;
	BSIndex::CLATSkipList::Surfer* _CLAT;
      };
    };   
    class ClauseEqualizer
    {
    public:
      ClauseEqualizer() {};
      ~ClauseEqualizer() {};
      ulong currClauseNum() const { return _currClauseNum; };
      void reset()
      {
	_currClauseNum = ULONG_MAX;
	_firstPrim = 0;
	_nextUnsortedPrim = 0;
	_sorting.reset(); 
      };
      bool next();
      Stream* const * map() const { return _map; };
      void initEqualizing()
      {
	for (ulong i = 0; i < _nextUnsortedPrim; i++) initPrimary(i);
	_endOfPrim = _firstPrim + _sorting.size();
      };
      bool primary(int num,const CSkipList* occList)
      {
	PrimWrapper* wrapper = _unsortedPrim + _nextUnsortedPrim;
	wrapper->num = num;
	wrapper->tag = PropLit;
	wrapper->list = static_cast<const void*>(occList);
	_nextUnsortedPrim++;
	_sorting.push(occList->size());
	return (bool)(*occList);
      };
      bool primary(int num,const CLSkipList* occList)
      {
	PrimWrapper* wrapper = _unsortedPrim + _nextUnsortedPrim;
	wrapper->num = num;
	wrapper->tag = Const;
	wrapper->list = static_cast<const void*>(occList);
	_nextUnsortedPrim++;
	_sorting.push(occList->size());
	return (bool)(*occList);
      };
      bool primary(int num,const CLTSkipList* occList)
      {
	PrimWrapper* wrapper = _unsortedPrim + _nextUnsortedPrim;
	wrapper->num = num;
	wrapper->tag = Var;
	wrapper->list = static_cast<const void*>(occList);
	_nextUnsortedPrim++;
	_sorting.push(occList->size());
	return (bool)(*occList);
      };
      bool primary(int num,const CLASkipList* occList)
      {
	PrimWrapper* wrapper = _unsortedPrim + _nextUnsortedPrim;
	wrapper->num = num;
	wrapper->tag = SymLitConst;
	wrapper->list = static_cast<const void*>(occList);
	_nextUnsortedPrim++;
	_sorting.push(occList->size());
	return (bool)(*occList);
      };
      bool primary(int num,const CLATSkipList* occList)
      {
	PrimWrapper* wrapper = _unsortedPrim + _nextUnsortedPrim;
	wrapper->num = num;
	wrapper->tag = SymLitVar;
	wrapper->list = static_cast<const void*>(occList);
	_nextUnsortedPrim++;
	_sorting.push(occList->size());
	return (bool)(*occList);
      };
      bool secondary(int num,const CSkipList* occList)
      {
	_stream[_firstPrim].reset(occList);
	_map[num] = _stream + _firstPrim;
	_firstPrim++;
	return (bool)(*occList);
      };
      bool secondary(int num,const CLSkipList* occList)
      {
	_stream[_firstPrim].reset(occList);
	_map[num] = _stream + _firstPrim;
	_firstPrim++;
	return (bool)(*occList);
      };
      bool secondary(int num,const CLTSkipList* occList)
      {
	_stream[_firstPrim].reset(occList);
	_map[num] = _stream + _firstPrim;
	_firstPrim++;
	return (bool)(*occList);
      }; 
      bool secondary(int num,const CLASkipList* occList)
      {
	_stream[_firstPrim].reset(occList);
	_map[num] = _stream + _firstPrim;
	_firstPrim++;
	return (bool)(*occList);
      };
      bool secondary(int num,const CLATSkipList* occList)
      {
	_stream[_firstPrim].reset(occList);
	_map[num] = _stream + _firstPrim;
	_firstPrim++;
	return (bool)(*occList);
      };
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const
      {
	str << "PRIM ";
	ulong i;
	for (i = _firstPrim; i < _endOfPrim; i++)
	  _stream[i].output(str) << " ";
	str << "  SEC ";
	for (i = 0; i < _firstPrim; i++)
	  _stream[i].output(str) << " "; 
	return str;
      }; 
#endif
    private:
      bool equalize();
      bool equalizePrimary();
      void equalizeSecondary();
#ifdef DEBUG_NAMESPACE
      bool properlyEqualized() 
      {
	ulong clNum;
	for (ulong i = 0; i < _endOfPrim; i++)   
	  if (_stream[i].topClauseNum(clNum))
	    {
	      if (clNum != _currClauseNum) return false;
	    }
	  else return false;
	return true;
      };
#endif
    private:
      struct PrimWrapper 
      {
      public:
	ulong num;
	Tag tag;
	const void* list;
      };
    private:
      void initPrimary(ulong i) 
      {
	CALL("ClauseEqualizer::initPrimary(ulong i)"); 
	ASSERT(i < _nextUnsortedPrim);
	ulong pos = _firstPrim + _sorting.num()[i];
	switch (_unsortedPrim[i].tag)
	  { 
	  case PropLit: 
	    _stream[pos].reset(static_cast<const CSkipList*>(_unsortedPrim[i].list)); break;
	  case Const: 
	    _stream[pos].reset(static_cast<const CLSkipList*>(_unsortedPrim[i].list)); break; 
	  case Var: 
	    _stream[pos].reset(static_cast<const CLTSkipList*>(_unsortedPrim[i].list)); break;        
	  case SymLitConst: 
	    _stream[pos].reset(static_cast<const CLASkipList*>(_unsortedPrim[i].list)); break;
	  case SymLitVar: 
	    _stream[pos].reset(static_cast<const CLATSkipList*>(_unsortedPrim[i].list)); break;
	  };
	_map[_unsortedPrim[i].num] = _stream + pos;
      };
    private:
      Stream _stream[VampireKernelConst::MaxClauseSize];
      Stream* _map[VampireKernelConst::MaxClauseSize];   
      ulong _currClauseNum;
      ulong _firstPrim;
      ulong _endOfPrim;
      BK::GSimpleIncCountingSort<ulong,VampireKernelConst::MaxClauseSize> _sorting;  
      PrimWrapper _unsortedPrim[VampireKernelConst::MaxClauseSize];   
      ulong _nextUnsortedPrim; 
    }; 
    class LitSubst
    {
    public:
      LitSubst() 
	: _litNum(DOP("BSMachine::LitSubst::_litNum")),
	  _matrix(DOP("BSMachine::LitSubst::_matrix"))
      {
	DOP(_litNum.freeze());
	DOP(_matrix.freeze());
      };
      ~LitSubst() {};
      bool adjustSizes(ulong maxlNumOfLitsInIndex,
		       ulong numOfVarOccs)
      {
	CALL("adjustSizes(..)");
	_currentRowSize = numOfVarOccs;
	ulong maxRowNum = (maxlNumOfLitsInIndex) 
	  ? maxlNumOfLitsInIndex - 1
	  : 0UL;
	DOP(_litNum.unfreeze());
	if (!_litNum.tryToExpand(maxRowNum)) return false;
	DOP(_litNum.freeze());
	ulong matrixSize = maxlNumOfLitsInIndex * numOfVarOccs;
	if (matrixSize) matrixSize--;
	DOP(_matrix.unfreeze());
	if (!_matrix.tryToExpand(matrixSize)) return false;
	DOP(_matrix.freeze());
	return true;
      };
      void reset() { _nextRow = _matrix.memory(); _nextLitNum = _litNum.memory(); };
      void resetCursor() 
      { 
	CALL("LitSubst::resetCursor()"); 
	ASSERT(_nextRow > _matrix.memory()); 
	_cursor = _matrix.memory();
	_litNumCursor = _litNum.memory(); 
      };   
      void endOfRow() { _nextRow += _currentRowSize; _nextLitNum++; };  
      bool moveCursor() { _cursor += _currentRowSize; _litNumCursor++; return (_cursor != _nextRow); }; 
      void set(ulong col,const Term& t) { _nextRow[col] = t; }; 
      const Term& get(ulong col) const { return _cursor[col]; };  
      void setLitNum(ulong ln) { *_nextLitNum = ln; };
      ulong getLitNum() const { return *_litNumCursor; };
#ifndef NO_DEBUG_VIS    
      ostream& output(ostream& str,ulong width) const 
      {
	const ulong* ln = _litNum.memory();
	for (const Term* row = _matrix.memory(); 
	     row != _nextRow; 
	     row += _currentRowSize) 
	  { 
	    str << "row " << (row - _matrix.memory())/_currentRowSize;
	    str << " lit[" << *ln << "] ";
	    for (ulong col = 0; col < width; col++) 
	      str << "  v" << col << " -> " << row[col];      
	    str << '\n';
	    ln++;
	  };
	return str;
      };
#endif
    private:
      ulong _currentRowSize;
      BK::Array<BK::GlobAlloc,ulong,BK::DefaultRelocation<ulong>,32UL> _litNum; 
      ulong* _nextLitNum;
      BK::Array<BK::GlobAlloc,Term,BK::DefaultRelocation<Term>,128UL> _matrix;
      Term* _nextRow;
      const Term* _cursor;
      ulong* _litNumCursor;
    }; 
    class SubstJoin
    {
    public: 
      SubstJoin();
      ~SubstJoin();
      bool reset(ulong numOfQueryLits, ulong maxlNumOfLitsInIndex);
      void nextLit(LitSubst* ls);
      void endOfLit();
      void variable(ulong v);
      bool joinInMultisetMode();
      void resetResultCollection();
      bool resultNextPair(ulong& var,Term& term);
      bool litIsSubsumed(ulong litNum) const;
#ifndef NO_DEBUG_VIS
      ostream& outputResult(ostream& str);
#endif
#ifndef NO_DEBUG_VIS
      ostream& outputCode(ostream& str) const;
#endif
    private:
      struct VarOcc
      {
	bool registered;
	ulong var;
	long subst;
	ulong queryVar;
      };
      struct Command
      {
	bool nextSubst;
	ulong var1;
	LitSubst* subst; 
	long substNum; 
	ulong var2;
	Command* backtrackCommand;      
      };
    private: 
      void resetLitRegister() {
	while (_registeredLiterals.nonempty()) 
	  _litRegistered[_registeredLiterals.pop()] = false;
      };
      void registerLit(ulong ln) { 
	_litRegistered[ln] = true; _registeredLiterals.push(ln); 
      };  
      void unregisterLastLit() {
	CALL("unregisterLastLit()");
	ASSERT(_registeredLiterals.nonempty());
	_litRegistered[_registeredLiterals.pop()] = false;
      };
    private:
      BK::Array<BK::GlobAlloc,LitSubst*,BK::DefaultRelocation<LitSubst*>,32UL> _subst;   
      BK::Array<BK::GlobAlloc,Command*,BK::DefaultRelocation<Command*>,32UL> _firstSubstCom;
      long _nextSubst;
      ulong _litVariables;
      long _maxBacktrackLit;
      BK::Array<BK::GlobAlloc,Command,BK::DefaultRelocation<Command>,128UL> _code;
      Command* _endOfCode;
      VarOcc _firstOcc[VampireKernelConst::MaxNumOfVariables];
      BK::Stack<VarOcc*,VampireKernelConst::MaxNumOfVariables> _registeredVars;
      const VarOcc* const* _nextRegistered;  
      BK::ExpandingStack<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,32UL> _registeredLiterals;
      BK::Array<BK::GlobAlloc,bool,BK::DefaultRelocation<bool>,32UL> _litRegistered;
    }; 
    class LitEqualizer
    {
    public:
      LitEqualizer() : subst() {}; 
      void reset() 
      {
	_nextConst = 0;
	_nextVar = 0;
      };  
      void constant(ulong strNum) 
      {
	_const[_nextConst].streamNum = strNum;
	_nextConst++;
      };
      void variable(ulong strNum) 
      {
	_var[_nextVar].streamNum = strNum;
	_nextVar++;     
      };  
      bool endOfLit(const BSIndex* index)
      {
	CALL("endOfLit(const BSIndex* index)");
	return subst.adjustSizes(index->maxlNumOfNonproplLits(),_nextVar);
      }; 
      void connect(Stream* const * map)
      {
	ulong i;	 
	for (i = 0; i < _nextConst; i++) _const[i].stream = map[_const[i].streamNum];
	for (i = 0; i < _nextVar; i++) _var[i].stream = map[_var[i].streamNum];     
      };
      void load()
      {
	ulong i;		 
	for (i = 0; i < _nextConst; i++) _const[i].load();
	for (i = 0; i < _nextVar; i++) _var[i].load(); 
      };   
      bool equalize();
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const;
#endif
    public:
      LitSubst subst;
    private:
      class ConstStream
      {
      public:
	ConstStream() {};
	~ConstStream() {};
	void load() 
	{
	  CALL("LitEqualizer::ConstStream::load()");
	  llist = stream->litList();
	  ASSERT(llist);       
	};
	ulong topLitNum() const { return llist->hd(); };
	const LList* list() const { return llist; };
	bool equalize(ulong& litNum,bool& notEnd);
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str) const { return stream->output(str) << " {" << llist << '}'; };
#endif
      public:
	union { Stream* stream; ulong streamNum; };
	const LList* llist;  
      }; 
      class VarStream
      {
      public:
	VarStream() {};
	~VarStream() {};
	void load() 
	{
	  CALL("LitEqualizer::VarStream::load()");
	  ltlist = stream->litTermList(); 
	  ASSERT(ltlist);       
	};
	ulong topLitNum() const { return ltlist->hd().first; };
	const Term& topTerm() const { return ltlist->hd().second; }; 
	const LTList* list() const { return ltlist; };
	bool equalize(ulong& litNum,bool& notEnd); 
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str) const { return stream->output(str) << " {" << ltlist << '}'; };
#endif     
      public:
	union { Stream* stream; ulong streamNum; };
	const LTList* ltlist;
      }; 
    private:
#ifdef DEBUG_NAMESPACE
      bool constStreamsNonempty() 
      {
	for (ulong i = 0; i < _nextConst; i++) if (!(_const[i].list())) return false;
	return true;
      };
      bool constProperlyEqualized() 
      {
	for (ulong i = 0; i < _nextConst; i++) if (_const[i].topLitNum() != _currLitNum) return false;
	return true; 
      };
      bool varStreamsNonempty()
      {
	for (ulong i = 0; i < _nextVar; i++) if (!(_var[i].list())) return false;
	return true;
      }; 
      bool varProperlyEqualized() 
      {
	for (ulong i = 0; i < _nextVar; i++) if (_var[i].topLitNum() != _currLitNum) return false;
	return true; 
      };
#endif
      void collectSubst() 
      {
	for (ulong i = 0; i < _nextVar; i++) subst.set(i,_var[i].topTerm());      
	subst.setLitNum(_currLitNum);
	subst.endOfRow();
      };
      bool equalizeFirst();
      bool equalizeNext();
      bool equalizeAll();
    private: 
      ConstStream _const[VampireKernelConst::MaxTermSize];
      VarStream _var[VampireKernelConst::MaxTermSize]; 
      ulong _nextConst;
      ulong _nextVar;
      ulong _currLitNum;
    }; 
    class SymLitEqualizer
    {
    public:
      SymLitEqualizer() : subst() {}; 
      ~SymLitEqualizer() {};
      void reset() 
      {
	_nextConst1 = 0;
	_nextConst2 = 0;
	_nextVar1 = 0;
	_nextVar2 = 0;
	_secondArg = false;
      };
      void secondArg() { _secondArg = true; };
      void constant(ulong strNum) 
      {
	if (_secondArg) 
	  {
	    _const2[_nextConst2].streamNum = strNum;
	    _nextConst2++;
	  }
	else
	  {
	    _const1[_nextConst1].streamNum = strNum;
	    _nextConst1++;
	  };
      };
      void variable(ulong strNum) 
      {
	if (_secondArg) 
	  {
	    _var2[_nextVar2].streamNum = strNum;
	    _nextVar2++; 
	  }
	else
	  {
	    _var1[_nextVar1].streamNum = strNum;
	    _nextVar1++; 
	  };    
      };
      bool endOfLit(const BSIndex* index)
      {
	CALL("endOfLit(const BSIndex* index)");
	return 
	  subst.adjustSizes(2*index->maxlNumOfNonproplLits(),
			    _nextVar1 + _nextVar2);   
      }; 
      void connect(Stream* const * map)
      {
	ulong i;
	for (i = 0; i < _nextConst1; i++) _const1[i].stream = map[_const1[i].streamNum];
	for (i = 0; i < _nextConst2; i++) _const2[i].stream = map[_const2[i].streamNum];
	for (i = 0; i < _nextVar1; i++) _var1[i].stream = map[_var1[i].streamNum];     
	for (i = 0; i < _nextVar2; i++) _var2[i].stream = map[_var2[i].streamNum];  
      };
      void load()
      {
	ulong i;
	for (i = 0; i < _nextConst1; i++) _const1[i].load();
	for (i = 0; i < _nextConst2; i++) _const2[i].load();
	for (i = 0; i < _nextVar1; i++) _var1[i].load(); 
	for (i = 0; i < _nextVar2; i++) _var2[i].load(); 
      };
      bool equalize();
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const;
#endif
    public:
      LitSubst subst;
    private:
      class ConstStream
      {
      public:
	ConstStream() {};
	~ConstStream() {};
	void load() 
	{
	  CALL("SymLitEqualizer::ConstStream::load()");
	  lalist = stream->litArgList();
	  ASSERT(lalist);       
	};
	ulong topLitNum() const { return lalist->hd().first; };
	ulong topArgNum() const { return lalist->hd().second; };
	const LAList* list() const { return lalist; };
	bool equalize(ulong& litNum,ulong& argNum,bool& notEnd);
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str) const { return stream->output(str) << " {" << lalist << '}'; };
#endif
      public:
	union { Stream* stream; ulong streamNum; };
	const LAList* lalist;  
      }; 
      class VarStream
      {
      public:
	VarStream() {};
	~VarStream() {};
	void load() 
	{
	  CALL("SymLitEqualizer::VarStream::load()");
	  latlist = stream->litArgTermList(); 
	  ASSERT(latlist);       
	};
	ulong topLitNum() const { return latlist->hd().first; };
	ulong topArgNum() const { return latlist->hd().second; };
	const Term& topTerm() const { return latlist->hd().third; }; 
	const Term& secondTerm() const { return latlist->tl()->hd().third; }; 
	const LATList* list() const { return latlist; };
	bool equalize(ulong& litNum,ulong& argNum,bool& notEnd);
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str) const { return stream->output(str) << " {" << latlist << '}'; };
#endif
      public:
	union { Stream* stream; ulong streamNum; };
	const LATList* latlist;
      }; 
    private:
#ifdef DEBUG_NAMESPACE
      bool constStreamsNonempty() 
      {
	ulong i;		  
	for (i = 0; i < _nextConst1; i++) if (!(_const1[i].list())) return false;
	for (i = 0; i < _nextConst2; i++) if (!(_const2[i].list())) return false; 
	return true;
      };
      bool constProperlyEqualized() 
      {
	ulong i;
	for (i = 0; i < _nextConst1; i++) 
	  if ((_const1[i].topLitNum() != _currLitNum) || (_const1[i].topArgNum() != _currArgNum1))
	    return false;
	for (i = 0; i < _nextConst2; i++) 
	  if ((_const2[i].topLitNum() != _currLitNum) || (_const2[i].topArgNum() != _currArgNum2)) 
	    return false;
	return true; 
      };
      bool varStreamsNonempty()
      {
	ulong i;
	for (i = 0; i < _nextVar1; i++) if (!(_var1[i].list())) return false;
	for (i = 0; i < _nextVar2; i++) if (!(_var2[i].list())) return false;
	return true;
      }; 
      bool varProperlyEqualized() 
      {
	ulong i;	  
	for (i = 0; i < _nextVar1; i++) 
	  if ((_var1[i].topLitNum() != _currLitNum) || (_var1[i].topArgNum() != _currArgNum1))       
	    return false;
	for (i = 0; i < _nextVar2; i++) 
	  if ((_var2[i].topLitNum() != _currLitNum) || (_var2[i].topArgNum() != _currArgNum2))  
	    return false;     
	return true; 
      };
#endif  
      void collectSubst();
      bool equalizeFirst();
      bool equalizeNext();
      bool alternativeArg1();
      bool alternativeArg2();
      bool equalizeAll();
    private:
      ConstStream _const1[VampireKernelConst::MaxTermSize];
      ConstStream _const2[VampireKernelConst::MaxTermSize];
      VarStream _var1[VampireKernelConst::MaxTermSize]; 
      VarStream _var2[VampireKernelConst::MaxTermSize]; 
      ulong _nextConst1;
      ulong _nextConst2;
      ulong _nextVar1;   
      ulong _nextVar2;
      bool _secondArg;
      ulong _currLitNum;
      ulong _currArgNum1;
      ulong _currArgNum2;
      bool _altArg1;
      bool _altArg2;
    }; 
  private:
    bool equalizeLiterals();
#ifndef NO_DEBUG_VIS
    static const char* spell(const Tag& tag);
#endif
  private:
    const BSIndex* _currentIndex;
    ClauseEqualizer _clauseEqualizer;
    ulong _nextStream;
    BK::StaticCellArray<BK::GlobAlloc,LitEqualizer,32UL> _litEqualizer;
    BK::StaticCellArray<BK::GlobAlloc,SymLitEqualizer,32UL> _symLitEqualizer;
    ulong _nextLit;
    ulong _nextSymLit;
    SubstJoin _substJoin;
#ifndef NO_DEBUG_VIS 
    friend class Stream;
#endif
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_MACHINE
#define DEBUG_NAMESPACE "BSMachine"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline 
  bool 
  BSMachine::resetQuery(ulong numOfLiterals,const BSIndex* index) 
  {
    CALL("resetQuery(ulong numOfLiterals,const BSIndex* index)");
    ASSERT(numOfLiterals > 0UL);
    _currentIndex = index; 
    _clauseEqualizer.reset();
    _nextStream = 0;
    _nextLit = 0;
    _nextSymLit = 0;
    if (!_substJoin.reset(numOfLiterals,
			  index->maxlNumOfNonproplLits()))
      return false;
    DOP(_litEqualizer.unfreeze());
    if (!_litEqualizer.tryToExpand(numOfLiterals - 1)) return false;
    DOP(_litEqualizer.freeze());
    DOP(_symLitEqualizer.unfreeze());
    if (!_symLitEqualizer.tryToExpand(numOfLiterals - 1)) return false;
    DOP(_symLitEqualizer.freeze());
    return true;
  }; 
  inline bool BSMachine::endOfQuery() 
  {
    CALL("endOfQuery()");
    _clauseEqualizer.initEqualizing();
    ulong i;
    for (i = 0; i < _nextLit; i++) _litEqualizer[i].connect(_clauseEqualizer.map()); 
    for (i = 0; i < _nextSymLit; i++) _symLitEqualizer[i].connect(_clauseEqualizer.map());
    return true;
  }; 
  inline bool BSMachine::queryPropLit(const CSkipList* occList) 
  {  
    CALL("queryPropLit(const CSkipList* occList)");
    if (!_clauseEqualizer.primary(_nextStream,occList)) return false;
    _nextStream++;
    return true;
  }; 
  inline void BSMachine::queryNextLit() 
  { 
    CALL("queryNextLit()");
    _litEqualizer[_nextLit].reset(); 
    _substJoin.nextLit(&(_litEqualizer[_nextLit].subst)); 
  }; 
  inline bool BSMachine::queryEndOfLit() 
  {    
    CALL("queryEndOfLit()");
    if (!_litEqualizer[_nextLit].endOfLit(_currentIndex))
      return false;
    _nextLit++; 
    _substJoin.endOfLit(); 
    return true; 
  }; 
  inline bool BSMachine::queryConstant(const Term& c,const CLSkipList* occList) 
  { 
    CALL("queryConstant(const Term& c,const CLSkipList* occList)");
    ASSERT(c.IsConstant());
    if (!_clauseEqualizer.primary(_nextStream,occList)) return false;
    _litEqualizer[_nextLit].constant(_nextStream); 
    _nextStream++;
    return true;
  }; 
  inline bool BSMachine::queryVariable(const Term& v,const CLTSkipList* occList,bool primary)
  {
    CALL("queryVariable(const Term& v,const CLTSkipList* occList,bool primary)");
    ASSERT(v.isVariable());
    if (primary)
      {
	if (!_clauseEqualizer.primary(_nextStream,occList)) return false; 
      }  
    else if (!_clauseEqualizer.secondary(_nextStream,occList)) return false;
    _litEqualizer[_nextLit].variable(_nextStream); 
    _nextStream++;
    _substJoin.variable(v.var());
    return true;
  }; 
  inline void BSMachine::queryNextSymLit() 
  { 
    CALL("queryNextSymLit()");
    _symLitEqualizer[_nextSymLit].reset(); 
    _substJoin.nextLit(&(_symLitEqualizer[_nextSymLit].subst));  
  }; 
  inline bool BSMachine::queryEndOfSymLit() 
  { 
    CALL("queryEndOfSymLit()");
    if (!_symLitEqualizer[_nextSymLit].endOfLit(_currentIndex))
      return false;
    _nextSymLit++; 
    _substJoin.endOfLit(); 
    return true; 
  }; 
  inline bool BSMachine::querySymLitConstant(const Term& c,const CLASkipList* occList)
  {
    CALL("querySymLitConstant(const Term& c,const CLASkipList* occList)");
    ASSERT(c.IsConstant());
    if (!_clauseEqualizer.primary(_nextStream,occList)) return false;
    _symLitEqualizer[_nextSymLit].constant(_nextStream); 
    _nextStream++;
    return true;
  }; 
  inline bool BSMachine::querySymLitVariable(const Term& v,const CLATSkipList* occList,bool primary)
  {  
    CALL("querySymLitVariable(const Term& v,const CLATSkipList* occList,bool primary)");
    ASSERT(v.isVariable());
    if (primary)
      {
	if (!_clauseEqualizer.primary(_nextStream,occList)) return false; 
      }  
    else if (!_clauseEqualizer.secondary(_nextStream,occList)) return false;
    _symLitEqualizer[_nextSymLit].variable(_nextStream); 
    _nextStream++;
    _substJoin.variable(v.var());
    return true;
  }; 
#ifndef NO_DEBUG_VIS
  inline ostream& BSMachine::outputLitEqualizers(ostream& str) const 
  {
    ulong i;
    for (i = 0; i < _nextLit; i++)
      _litEqualizer[i].output(str << "LITERAL " << i << '\n') << '\n';
    for (i = 0; i < _nextSymLit; i++)
      _symLitEqualizer[i].output(str << "SYM. LIT. " << i << '\n') << '\n';
    return str << '\n';
  }; 
#endif
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_MACHINE
#define DEBUG_NAMESPACE "BSMachine::Stream"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline bool BSMachine::Stream::equalize(ulong& clNum,bool& notEnd)
  {
    CALL("Stream::equalize(ulong& clNum,bool& notEnd)");
    switch (tag())
      {
      case PropLit: 
	if (_C->find(clNum)) 
	  { 
	    ASSERT(_C->currNode()->key() == clNum); 
	    return true;
	  };
	if ((notEnd = (_C->currNode() != 0))) 
	  clNum = _C->currNode()->key();
	return false;         
      case Const: 
	if (_CL->find(clNum))
	  {
	    ASSERT(_CL->currNode()->key() == clNum); 
	    return true;
	  };
	if ((notEnd = (_CL->currNode() != 0))) 
	  clNum = _CL->currNode()->key();
	return false; 
      case Var:
	if (_CLT->find(clNum))
	  {
	    ASSERT(_CLT->currNode()->key() == clNum); 
	    return true;
	  };
	if ((notEnd = (_CLT->currNode() != 0))) 
	  clNum = _CLT->currNode()->key();
	return false;
      case SymLitConst: 
	if (_CLA->find(clNum))
	  {
	    ASSERT(_CLA->currNode()->key() == clNum); 
	    return true;
	  };
	if ((notEnd = (_CLA->currNode() != 0))) 
	  clNum = _CLA->currNode()->key();
	return false;
      case SymLitVar:
	if (_CLAT->find(clNum))
	  {
	    ASSERT(_CLAT->currNode()->key() == clNum); 
	    return true;
	  };
	if ((notEnd = (_CLAT->currNode() != 0))) 
	  clNum = _CLAT->currNode()->key();
	return false;
      };
    ICP("ICP0");
    return false;
  }; 
  inline bool BSMachine::Stream::equalize(ulong clNum)
  {
    CALL("Stream::equalize(ulong clNum)");
    switch (tag())
      {
      case PropLit: return _C->find(clNum);
      case Const: return _CL->find(clNum);
      case Var: return _CLT->find(clNum);
      case SymLitConst: return _CLA->find(clNum);
      case SymLitVar: return _CLAT->find(clNum);
      };
    ICP("ICP0");
    return false;
  }; 
  inline const BSMachine::LList* BSMachine::Stream::litList() const 
  {
    CALL("Stream::litList() const");
    ASSERT(tag() == Const); 
    return _CL->currNode()->value();
  }; 
  inline const BSMachine::LTList* BSMachine::Stream::litTermList() const 
  {
    CALL("Stream::litTermList() const");
    ASSERT(tag() == Var); 
    return _CLT->currNode()->value();
  }; 
  inline const BSMachine::LAList* BSMachine::Stream::litArgList() const 
  {
    CALL("Stream::litArgList() const");
    ASSERT(tag() == SymLitConst); 
    return _CLA->currNode()->value();
  }; 
  inline const BSMachine::LATList* BSMachine::Stream::litArgTermList() const 
  {
    CALL("Stream::litArgTermList() const");
    ASSERT(tag() == SymLitVar); 
    return _CLAT->currNode()->value();
  }; 
  inline bool BSMachine::Stream::topClauseNum(ulong& clNum) const
  { 
    CALL("Stream::topClauseNum(ulong& clNum)");
    switch (tag())
      {
      case PropLit: 
	if (_C->currNode()) 
	  { 
	    clNum = _C->currNode()->key();
	    return true; 
	  };
	return false;
      case Const: 
	if (_CL->currNode())
	  {
	    clNum = _CL->currNode()->key();  
	    return true;
	  };
	return false;
      case Var: 
	if (_CLT->currNode())
	  {
	    clNum = _CLT->currNode()->key(); 
	    return true;
	  };
	return false;
      case SymLitConst: 
	if (_CLA->currNode())
	  {
	    clNum = _CLA->currNode()->key();  
	    return true;
	  };
	return false;
      case SymLitVar:
	if (_CLAT->currNode())
	  {
	    clNum = _CLAT->currNode()->key();  
	    return true;
	  };
      };
    ICP("ICP0");
    return false;
  }; 
#ifndef NO_DEBUG_VIS
  inline ostream& BSMachine::Stream::output(ostream& str) const 
  {
    str << spell(tag()) << '{';
    ulong clNum;
    if (topClauseNum(clNum)) str << clNum << ",..";
    return str << "}";
  }; 
#endif
}; 
#endif
