/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BACK_PARA_CANDIDATES_HPP
#define BACK_PARA_CANDIDATES_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "SkipList.hpp"
#include "GList.hpp"
#include "SimplificationOrdering.hpp"
#include "DestructionMode.hpp"
namespace VK
{
  class Term;
  class Clause;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACK_PARA_CANDIDATES
#  define DEBUG_NAMESPACE "BackParaCandidates"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BackParaCandidates
  {
  public:
    class LitRedexPair
    {
    public:
      LitRedexPair(ulong ln,
		   const Term* red,
		   SimplificationOrdering::StoredConstraint* cns,
		   LitRedexPair* nxt) : 
	_litNum(ln),
	_redex(red),
	_next(nxt)
      {
	CALL("LitRedexPair::constructor LitRedexPair(ulong ln,Term* red,SimplificationOrdering::StoredConstraint* cns,LitRedexPair* nxt)");
	if (cns)
	  {
	    _constraint = SimplificationOrdering::current()->copy(cns); 
	  }
	else
	  _constraint = 0;
      };  
      LitRedexPair(ulong ln,
		   const Term* red,
		   SimplificationOrdering::StoredConstraint* cns) :      
	_litNum(ln),
	_redex(red),
	_next(0)
      {
	CALL("LitRedexPair::constructor LitRedexPair(ulong ln,Term* red,SimplificationOrdering::StoredConstraint* cns)");
	if (cns)
	  {
	    _constraint = SimplificationOrdering::current()->copy(cns); 
	  }
	else
	  _constraint = 0;
      };
      ~LitRedexPair() 
      {
	CALL("destructor ~LitRedexPair()");
	if (BK::DestructionMode::isThorough())
	  {
	    if (_constraint)
	      {
		SimplificationOrdering::current()->releaseConstraint(_constraint);
	      };
	  };
      };
      void* operator new(size_t) 
      { 
	CALL("LitRedexPair::operator new(size_t)");
	return BK::GlobAlloc::allocate(sizeof(LitRedexPair));
      };
      void  operator delete(void* obj) 
      {
	CALL("LitRedexPair::operator delete(void* obj)"); 
	BK::GlobAlloc::deallocate(obj,sizeof(LitRedexPair));
      };
      ulong litNum() const { return _litNum; };
      const Term* redex() const { return _redex; };
      SimplificationOrdering::StoredConstraint* constraint() const 
      {
	CALL("constraint() const");
	return _constraint; 
      };
      LitRedexPair* next() const { return _next; }; 
      static void destroy(LitRedexPair* lst) 
      {
	CALL("destroy(LitRedexPair* lst)");
	LitRedexPair* rest; 
	while (lst) { rest = lst->next(); delete lst; lst = rest; };
      };
    private:
      ulong _litNum;
      const Term* _redex;
      SimplificationOrdering::StoredConstraint* _constraint;
      LitRedexPair* _next;
    }; 
    typedef BK::SkipList<BK::GlobAlloc,LitRedexPair*,Clause*> SkList;
  public:      
    class Iterator     
    {
    public:
      Iterator() {};
      Iterator(const BackParaCandidates& lst) { reset(lst); };
      ~Iterator() {};
      void reset(const BackParaCandidates& lst) 
      { 
	_currNode = lst._skList.first(); 
	if (_currNode) { _currLRPair = _currNode->value(); }
	else _currLRPair = 0;
      };
      bool next(Clause*& cl,LitRedexPair*& lrlist)
      {
	CALL("next(Clause*& cl,LitRedexPair*& lrlist)");
	if (_currLRPair) 
	  { 
	    cl = _currNode->key();       
	    lrlist = _currLRPair;
	    _currLRPair = _currLRPair->next();
	    return true;
	  };
	if (_currNode) 
	  {
	  next_node:
	    _currNode = _currNode->next();
	    if (_currNode)
	      {
		lrlist = _currNode->value();
		if (lrlist) 
		  {
		    cl = _currNode->key();  
		    _currLRPair = lrlist->next();
		    return true;
		  }
		else goto next_node;
	      };
	  };
	return false;
      }; 
    private:
      const SkList::Node* _currNode;
      LitRedexPair* _currLRPair; 
    }; 
  public:
    BackParaCandidates() : _skList() {};
    ~BackParaCandidates() 
    {
      CALL("destructor ~BackParaCandidates()");
      if (BK::DestructionMode::isThorough()) 
	{
	  SkList::Destruction destr(_skList); 
	  LitRedexPair* lrlist;
	  while (destr.next(lrlist)) LitRedexPair::destroy(lrlist); 
	};
    };
    void* operator new(size_t) 
    { 
      return BK::GlobAlloc::allocate(sizeof(BackParaCandidates)); 
    };
    void operator delete(void* obj) 
    { 
      BK::GlobAlloc::deallocate(obj,sizeof(BackParaCandidates)); 
    };
    void add(Clause* cl,ulong ln,const Term* red,SimplificationOrdering::StoredConstraint* cns);
    void remove(Clause* cl);
    bool empty() const { return !_skList; };
#ifndef NO_DEBUG
    bool contains(Clause* cl) const { return _skList.contains(cl); }; 
    ulong size() const { return _skList.size(); };
    Clause* first() const { return _skList.first()->key(); };
    Clause* second() const { return _skList.first()->next()->key(); };
#endif
  private:
    SkList _skList;
    friend class Iterator;
  }; 
  inline 
  void 
  BackParaCandidates::add(Clause* cl,
			  ulong ln,
			  const Term* red,
			  SimplificationOrdering::StoredConstraint* cns)
  {
    CALL("add(Clause* cl,ulong ln,const Term* red,SimplificationOrdering::StoredConstraint* cns)");
    bool newNode;
    SkList::Node* node = _skList.insert(cl,newNode);
    if (newNode) 
      { 
	node->value() = new LitRedexPair(ln,red,cns);
      }
    else
      { 
	LitRedexPair* lrlist = node->value();
	ASSERT(lrlist);
	if ((lrlist->litNum() != ln) || (lrlist->redex() != red))
	  {
	    node->value() = new LitRedexPair(ln,red,cns,lrlist);     
	  };
      };
    ASSERT(_skList.contains(cl));
  }; 
  inline void BackParaCandidates::remove(Clause* cl)
  {
    CALL("remove(Clause* cl)");
    LitRedexPair* lrlist; 
    if (_skList.remove(cl,lrlist)) LitRedexPair::destroy(lrlist); 
  }; 
}; 
#endif
