/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BACKWARD_FOLDING_HPP
#define BACKWARD_FOLDING_HPP
#include <queue>
#include <list>
#include "jargon.hpp"
#include "Receiver.hpp"
#include "Multiset.hpp"
#include "ExpandingStack.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "BSIndexFrontEnd.hpp"
#include "BSMachine.hpp"
#include "BSSubstitutionCache.hpp"
#include "ClauseNumIndex.hpp"
#include "Clause.hpp"
#include "Term.hpp"
namespace VK
{
  class TemporaryClauseCollector;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_FOLDING
#define DEBUG_NAMESPACE "BackwardFolding"
#endif
#include "debugMacros.hpp"
namespace VK {
  class BackwardFolding : private BK::Receiver<VK::Clause*>
  {
  public:
    BackwardFolding(BSIndexFrontEnd* index,
		    ClauseNumIndex* clauseNumIndex,
		    TemporaryClauseCollector* clauseCollector) :
      _isLocked(false),
      _index(index),
      _clauseNumIndex(clauseNumIndex),
      _clauseCollector(clauseCollector),
      _useCurrentDefForNegations(false),
      _inversePolarity(false)
    {
      CALL("constructor BackwardFolding(BSIndexFrontEnd* index,ClauseNumIndex* clauseNumIndex)");
      _retrieval.init(_index->index());
    };
    ~BackwardFolding() {};
    void reset() 
    {
      CALL("reset()");
      _isLocked = false;
      _useCurrentDefForNegations = false;
      _inversePolarity = false;
    }; 
    void setFoldForSplitting(bool fl) { _foldForSplitting = fl; };
    void setFoldSingleComponentsForPropSpl(bool fl) {
      _foldSingleComponentsForPropSpl = fl;
    };
    void setReusePropSplNamesForNegations(bool fl) {
      _reusePropSplNamesForNegations =  fl; };
    bool isLocked() const { return _isLocked; };
    bool queueIsEmpty() const { return _definitionQueue.empty(); };
    BK::Receiver<Clause*>* asNameIntroReceiver() {
      return static_cast<BK::Receiver<Clause*>*>(this);
    };
    void receive(Clause* const & cl);
    bool nextDefinition();
    void abandonCurrentDefinition();
    bool foldNext();
    Clause* folded() { return _currentFolded; };
    bool collectResult();
 public:
    typedef BSIndex::LTPair LTPair;
    typedef BSIndex::LAPair LAPair;
    typedef BSIndex::LATTriple LATTriple;
    typedef BSIndex::LList LList;
    typedef BSIndex::LTList LTList;
    typedef BSIndex::LAList LAList;
    typedef BSIndex::LATList LATList;
    typedef BSIndex::CSkipList CSkipList;
    typedef BSIndex::CLSkipList CLSkipList;
    typedef BSIndex::CLTSkipList CLTSkipList;
    typedef BSIndex::CLASkipList CLASkipList;
    typedef BSIndex::CLATSkipList CLATSkipList;
  private:
    void lock() { _isLocked = true; _index->lock(); };
    void unlock() { _isLocked = false; _index->unlock(); };
    bool preproCurrentDef();
    bool preproCurrentDefForNegations();
    bool preproQueryLit(const Term* lit,bool inversePolarity);
    bool preproQuerySymLit(const Term* lit,bool inversePolarity);
    bool preproQueryComplexTerm(const Term* t);
    bool preproQueryComplexTermArguments(const Term* t);
    bool preproQuerySymLitComplexTerm(const Term* t);
    bool preproQuerySymLitComplexTermArguments(const Term* t);
    bool preliminarySubstitutionCheck();
    bool separateRemainderLiterals();
    bool isSubsumedByGround(const Term* lit) const;
  private:
    bool _foldForSplitting;
    bool _foldSingleComponentsForPropSpl;
    bool _reusePropSplNamesForNegations;
    bool _isLocked;
    BSIndexFrontEnd* _index;
    BSIndex::Retrieval _retrieval;
    BSMachine _machine; 
    ClauseNumIndex* _clauseNumIndex;
    TemporaryClauseCollector* _clauseCollector;
    std::queue<Clause*,list<Clause*> > _definitionQueue;
    Clause* _currentDefinition;
    bool _useCurrentDefForNegations;
    bool _inversePolarity;
    BK::ExpandingStack<BK::GlobAlloc,Term*,BK::FastRelocation<Term*>,32UL> _groundQueryLiterals;
    const Term* _definedLit;
    BSSubstitutionCache::Instance _definedLitInstance;
    BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> _freeVariables;
    BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> _boundVariables;
    BSSubstitutionCache _subst;
    Clause* _currentFolded;
    BK::ExpandingStack<BK::GlobAlloc,const LiteralList::Element*,BK::FastRelocation<const LiteralList::Element*>,32UL> _remainderLiterals;
  }; 
}; 
#endif
