/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CHAINING_HPP
#define CHAINING_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "PremiseWithFactoring.hpp"
#include "SizeFormula.hpp"
#include "SimplificationOrdering.hpp"
#include "IUMachine.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class Term;
  class Clause;
  class SymbolInfo;
  class UnifIndex;
  class UnifIndex::IndexedObject;
  class ForwardProcessing;
  class Statistics;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CHAINING
#  define DEBUG_NAMESPACE "Chaining"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Chaining
  {
  public:
    Chaining(UnifIndex* unifIndex);
    ~Chaining();
    void setStatistics(Statistics* stat)
    {
      _dummyClauseCollector.setStatistics(stat);
      _statistics = stat;
    };
    void setClauseCollector(TemporaryClauseCollector* cc) 
    {
      _clauseCollector = cc;
    };
    void setForwardProcessing(ForwardProcessing* fp)
    {
      _forwardProcessing = fp;
    };
    void setTransitivity(const SymbolInfo* table) 
    {
      _transitivity = table;
    };
    void setChainingThroughVariablesFlag(bool fl)
    {
      _chainingThroughVariables = fl;
    };
    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };  
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };
    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };
    void setClauseSizeLimit(long lim)
    {
      _clauseSizeLimit = lim;
    };
    bool loadQuery(Clause* cl);
    bool nextInference();
  private:
    enum OuterContinuation
    {
      LoadForwardPositive,
      NextForwardPositive,
      NextBackwardPositive,
      NextForwardNegative1,
      NextBackwardNegative1,
      NextForwardNegative2,
      NextBackwardNegative2
    }; 
    enum InnerContinuation
    {
      FirstLit,
      FirstFactoring,
      NextFactoring
    }; 
  private:
    bool nextForwardPositive();
    bool nextBackwardPositive();
    bool nextForwardNegative1();
    bool nextBackwardNegative1();
    bool nextForwardNegative2();
    bool nextBackwardNegative2();
    bool posChainingLitSelectedIn(Clause* cl);
    bool negChainingLitSelectedIn(Clause* cl);
    bool buildResolvent(UnifTerm* newLitArg1,
			const Term& newLitArg2,
			bool inheritFromMain);
    bool buildResolvent(const Term& newLitArg1,
			UnifTerm* newLitArg2,
			bool inheritFromMain);
    bool nextFactoring();
    bool subsumeMainPremise();
    bool collectMainPremiseInstance();
    bool subsumeSidePremise();
    bool collectSidePremiseInstance();
    bool demodulateSelectedLits();
    bool collectMainPremiseSubtermsAvailableForParamodulationInto();
    bool collectSidePremiseSubtermsAvailableForParamodulationInto();
  private:  
    OPEN_BR_DECL(1);
    Statistics* _statistics;
    const SymbolInfo* _transitivity;
    bool _chainingThroughVariables;
    Clause* _query;
    TemporaryClauseCollector* _clauseCollector;
    ForwardProcessing* _forwardProcessing;
    UnifIndex* _unifIndex;
    BK::Comparison _headTailComparison;
    OuterContinuation _outerContinuation;
    InnerContinuation _innerContinuation;
    PremiseWithFactoring _mainPremise;
    PremiseWithFactoring _sidePremise;
    ulong _mainVarIndex;
    ulong _sideVarIndex;
    SizeFormula _mainPremResidualSizeFormula;
    long _mainPremResidualSize;
    long _clauseSizeLimit;
    ulong _currentMainLit;
    SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
    SimplificationOrdering::CompiledConstraintForUnifiers* _constraintOnMain;
    SimplificationOrdering::ConstraintCheckOnUnifier* _sideConstraintCheck;
    SimplificationOrdering::StoredConstraint* _sideConstraint;
    Term _pseudoHeader;
    IUMachine _unificationMachine;
    const Clause* _currentTransitivityLaw;
    UnifTerm* _mainHead;
    UnifTerm* _mainTail;
    UnifIndex::IndexedObject* _currentCandidate;
    Term _newLitHeader;
    TemporaryClause::Extras::Handle _dummyClause;
    TemporaryClauseCollector _dummyClauseCollector;
    bool _forwardMode;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;   
    bool _mainSelectedLitDemodulationAttempted;
    bool _mainSelectedLitDemodulated;
    CLOSE_BR_DECL(1);
  }; 
}; 
#endif
