/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CLAUSE_HPP
#define CLAUSE_HPP
#include <iostream>
#include <cstddef>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "GList.hpp"
#include "BitWord.hpp"
#include "Term.hpp"
#include "ClauseBackground.hpp"
#include "LiteralList.hpp"
#include "ExpandingStack.hpp"
#include "SymbolInfo.hpp"
using namespace std;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.hpp"
namespace VK
{                           
  class Clause     
  {
  public:
    class Context;
    class Traversal;
    class OpenUnshared; 
    class NewNumber;
  private:
    class ClauseList {};  
  public: 
    enum MainSet
    {
      MainSetFree = 0UL, 
      MainSetNew = 1UL,
      MainSetPassive = 2UL,
      MainSetReservedPassive = 3UL, 
      MainSetActive = 4UL,
      MainSetInputQuery = 5UL,
      MainSetInputSupport = 6UL,
      MainSetInputHistory = 7UL,
      MainSetNameIntro = 8UL,
      MainSetTrash = 10UL,
      MainSetHistory = 11UL,
      MainSetAnswers = 12UL
    };
    enum FlagNumber
    {
      MainSetPieceBegin = 0UL,
      MainSetPieceEnd = 3UL, 
      FlagLiteralsShared = 4UL,
      FlagInForwardSubsumptionIndex = 5UL,
      FlagInBackwardSubsumptionIndex = 6UL,  
      FlagInForwardDemodulationIndex = 7UL,  
      FlagInBackDemodulationIndex = 8UL, 
      FlagInInferenceDispatcherAsActive = 9UL,
      FlagInClauseNumIndex = 10UL, 
      FlagInQueueForBackSimplification = 11UL,
      FlagInQueueForIntegrationIntoBackwardDemodIndex = 12UL,
      FlagInQueueForIntegrationIntoBSIndex = 13UL,
      FlagHasSubsumedPremiseInstance = 14UL,
      FlagBlockedForSubsumption = 15UL,
      FlagBlockedForDemodulationFrom = 16UL,
      FlagParamodulationsIntoProtected = 17UL,
      FlagIsNameIntro = 18UL,
      FlagContainsAnswerLiterals = 19UL, 
      FlagHasChildren = 20UL,
      FlagAllNameIntroPartsDeployed = 21UL,
      ProtectionCounterBegin = 27UL, 
      ProtectionCounterEnd = 31UL
    };
  public:
    Clause(LiteralList* lits);
    Clause();
    ~Clause() {};
    void* operator new(size_t);
    void operator delete(void* obj);
    void clearFlags();
    void setMainSet(const MainSet& ms);
    MainSet mainSet() const { return (MainSet)_flags.piece(MainSetPieceBegin,MainSetPieceEnd); };
    bool flag(const FlagNumber& fn) const { return _flags.bit(fn); };
    void setFlag(const FlagNumber& fn) { _flags.set(fn); };
    void clearFlag(const FlagNumber& fn) { _flags.clear(fn); };
    Clause* Next() const { return next; };
    void SetNext(Clause* c) { next = c; };
    Clause* Previous() const { return previous; };
    void SetPrevious(Clause* p) { previous = p; };
    LiteralList* LitList() const;
    const ClauseBackground& background() const;
    ClauseBackground& background();
    void setBackground(const ClauseBackground& a) { _background = a; };
    void* inputClauseOrigin() const;
    void setInputClauseOrigin(void* orig);
    ulong Number() const;
    void SetNumber(long n) { number = n; };
    long inferenceDepth() const { return _background.inferenceDepth(); };
    long weight() const;
    long weightBase() const { return _weightBase; };
    long size() const;  
    void AssignMaxPriority();
    bool MaxPriority() const;
    ulong NumOfEqLits() const;
    bool containsEquality() const;
    long practicallyMinResidualSize() const;
    void setPracticallyMinResidualSize(long mcs) {  _practicallyMinResidualSize = mcs; };
    ulong numOfAllLiterals() const;
    ulong numOfNonanswerLiterals() const;
    void setSize(long s) { _size = s; };
    void setWeightBase(long w) { _weightBase = w; };
    bool isInputClause() const;
    bool isInputOrPreprocessedInputClause() const;
    bool isSubgoal() const { return _background.isSubgoal(); };
    bool isSupportClause() const { return _background.isSupport(); };
    bool isNonSubsuming() const { return _background.isNonSubsuming(); };
    bool isNonSubsumable() const { return _background.isNonSubsumable(); };
    bool isPartOfNameIntroForPropSplitting() const {
      return _background.isPartOfNameIntroForPropSplitting();
    };
    bool isProducedByPropSplitting() const {
      return _background.isProducedByPropSplitting();
    };
    void markAsProducedByPropSplitting() {
      return _background.markAsProducedByPropSplitting();
    };
    void markAsQuery() { _background.markAsQuery(); };
    bool isPersistent() const;
    bool isProtected() const;
    void protect();
    void unprotect();
    bool hasSubsumedPremiseInstance() const;
    void registerPremiseInstanceSubsumption();
    bool paramodulationsIntoProtected() const;
    void protectParamodulationsInto();
    bool isBlockedForSubsumption() const;
    void blockForSubsumption();
    void unblockForSubsumption();
    bool isBlockedForDemodulationFrom() const;
    void blockForDemodulationFrom();
    void unblockForDemodulationFrom();
    void MakeVIP() { _background.MakeVIP(); };
    bool IsVIP() const { return _background.IsVIP(); }; 
    void markAsNameIntro() { setFlag(FlagIsNameIntro); };
    bool isNameIntro() const { return flag(FlagIsNameIntro); };
    void markAsNameIntroWithAllPartsDeployed() {
      setFlag(FlagAllNameIntroPartsDeployed); };
    bool isNameIntroWithAllPartsDeployed() {
      return flag(FlagAllNameIntroPartsDeployed); };
    ulong NumberOfLiteral(const Term* lit) const;
    bool Empty() const;
    bool isAnswer() const;
    bool Unit() const;
    bool PositiveUnitEquality() const;
    bool posEqSelected() const;
    bool IsRule() const; 
    bool isSymmetryLaw() const;
    bool isCommutativityLaw(ulong& func) const;
    bool isTransitivityLaw() const;
    bool isTransitivityLaw(ulong& pred) const;
    bool ContainsEliminated(const SymbolInfo* predElimTable) const;
    bool singletonVar(ulong var) const;
    void Destroy();
    void DestroyUnshared();
    static void Remove(Clause* c);
    static Clause* MakeEmptyClause();
    bool IsOrphan() const;
    bool allMaximalSelected() const;
    ostream& outputBackground(ostream& str) const;
    ostream& output(ostream& str) const;   
  private:  
    Clause* next;
    Clause* previous;
    LiteralList* litList; 
    ClauseBackground _background;
    long number;
    long _size;
    long _weightBase;
    long _practicallyMinResidualSize;
    BK::BitWord _flags;   
  }; 
}; 
namespace std
{
  inline ostream& operator<<(ostream& str,const VK::Clause& cl)
  {
    return cl.output(str);
  }; 
  inline ostream& operator<<(ostream& str,const VK::Clause* cl)
  {
    return cl->output(str);
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause::NewNumber"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class Clause::NewNumber
  {
  public:
    NewNumber() : _nextNumber(0UL) {};
    NewNumber(ulong n) : _nextNumber(n) {};
    void reset(ulong n) { _nextNumber = n; };
    ulong generate() { ++_nextNumber; return _nextNumber - 1; };
  private:
    ulong _nextNumber;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause::Context"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class Clause::Context
  {
  public:
    class AutoActivator
    {
    public:
      AutoActivator(Context* c) : _previous(Context::getCurrent())
      {
	Context::setCurrent(c);
      };
      ~AutoActivator()
      {
	Context::setCurrent(_previous);
      };
    private:
      Context* _previous;
    }; 
  public:
    Context() :
      _selectedPosEqPenaltyCoefficient(1.0),
      _useSelectedPosEqPenaltyCoefficient(false),
      _nongoalPenaltyCoefficient(1.0),
      _useNongoalPenaltyCoefficient(false),
      _showLiteralProperties(false)
    {
    };
    ~Context() {};
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
    void setSelectedPosEqPenaltyCoefficient(const float& c) 
    { 
      _selectedPosEqPenaltyCoefficient = c;
      _useSelectedPosEqPenaltyCoefficient = (c != 1.0);
    };
    void setNongoalPenaltyCoefficient(const float& c) 
    { 
      _nongoalPenaltyCoefficient = c; 
      _useNongoalPenaltyCoefficient = (c != 1.0);
    };
    void setShowLiteralProperties(bool fl) { _showLiteralProperties = fl; };
    long computeWeight(long weightBase,
		       bool subgoal,
		       bool posEqSelected) const;
  private:
    float _selectedPosEqPenaltyCoefficient; 
    bool _useSelectedPosEqPenaltyCoefficient;
    float _nongoalPenaltyCoefficient;
    bool _useNongoalPenaltyCoefficient;
    bool _showLiteralProperties;
    static Context* _current;
    friend class Clause;
  }; 
}; 
#endif
