/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CLAUSE_INL
#  define CLAUSE_INL
#include "Clause.hpp"
#include "LiteralProperties.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#  define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.hpp"
namespace VK
{ 
  inline
  Clause::Clause(LiteralList* lits) : 
    litList (lits)
  {
    setMainSet(MainSetFree);      
  };
  inline
  Clause::Clause()
  {
    next = this;
    previous = this;  
    litList = 0;
    setMainSet(MainSetFree);
  };
  inline
  void* Clause::operator new(size_t)
  {
    CALL("operator new(size_t)");
    return BK::GlobAlloc::allocate(sizeof(Clause));
  };
  inline
  void Clause::operator delete(void* obj)
  { 
    CALL("operator delete(void* obj)");
    BK::GlobAlloc::deallocate(obj,sizeof(Clause));
  };
  inline
  void Clause::clearFlags()
  {
    _flags.clear();
  }; 
  inline
  void Clause::setMainSet(const MainSet& ms) 
  {
    CALL("setMainSet(const MainSet& ms)");
    ASSERT(!(ms >> 4));   
    _flags.setPiece(MainSetPieceBegin,MainSetPieceEnd,ms);      
    ASSERT(mainSet() == ms);
  };
  inline 
  void Clause::protect()
  {
    CALL("Clause::protect()");
    ulong protCounter = 
      _flags.piece(ProtectionCounterBegin,ProtectionCounterEnd);
    ASSERT(protCounter < 31UL);
    ++protCounter;
    _flags.setPiece(ProtectionCounterBegin,ProtectionCounterEnd,protCounter);
  };
  inline 
  void Clause::unprotect()
  {
    CALL("Clause::protect()");
    ulong protCounter = 
      _flags.piece(ProtectionCounterBegin,ProtectionCounterEnd);
    ASSERT(protCounter > 0UL);
    --protCounter;
    _flags.setPiece(ProtectionCounterBegin,ProtectionCounterEnd,protCounter);
  };
  inline 
  bool Clause::isProtected() const
  {
    CALL("isProtected() const");
    ulong protCounter = 
      _flags.piece(ProtectionCounterBegin,ProtectionCounterEnd);
    return protCounter != 0UL;
  };
  inline
  LiteralList* Clause::LitList() const 
  {
    CALL("LitList() const");
    return litList; 
  };
  inline
  const ClauseBackground& Clause::background() const 
  {
    CALL("background() const"); 
    return _background; 
  }; 
  inline
  ClauseBackground& Clause::background()
  {
    CALL("background()"); 
    return _background; 
  };  
  inline
  void* Clause::inputClauseOrigin() const 
  {
    CALL("inputClauseOrigin() const");
    ASSERT(isInputClause());
    return _background.inputClauseOrigin();
  };
  inline
  void Clause::setInputClauseOrigin(void* orig)
  {
    CALL("setInputClauseOrigin(void* orig)");
    ASSERT(isInputClause());
    _background.setInputClauseOrigin(orig);
  };
  inline
  ulong Clause::Number() const 
  {
    return number;
  };
  inline
  long Clause::weight() const 
  {
    return 
      Context::getCurrent()->computeWeight(_weightBase,
                                           isSubgoal(),
                                           posEqSelected());
  }; 
  inline
  long Clause::size() const  
  {
    if (_size < 0) return ((-1)*_size);
    return _size;
  };
  inline
  void Clause::AssignMaxPriority() 
  { 
    if (_size > 0) _size = (-1)*_size; 
  };
  inline
  bool Clause::MaxPriority() const 
  {
    CALL("MaxPriority() const"); 
    return (_size < 0); 
  };    
  inline
  ulong Clause::NumOfEqLits() const 
  {
    CALL("NumOfEqLits() const"); 
    ulong res = 0;
    for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
      if (iter.currentElement()->isEquality()) res++;
    return res;       
  };
  inline
  bool Clause::containsEquality() const 
  {
    CALL("containsEquality() const"); 
    for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
      if (iter.currentElement()->isEquality()) return true;
    return false;
  }; 
  inline
  long Clause::practicallyMinResidualSize() const 
  {
    return _practicallyMinResidualSize; 
  };
  inline
  ulong Clause::numOfAllLiterals() const 
  { 
    return LitList()->length();
  };	
  inline
  ulong Clause::numOfNonanswerLiterals() const
  {
    CALL("numOfNonanswerLiterals() const");
    if (flag(FlagContainsAnswerLiterals))
      {
	ulong res = 0;
	for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
	  if (!iter.currentLiteral()->isAnswerLiteral()) res++; 
	ASSERT(res < numOfAllLiterals());
	return res;  
      }
    else
      return numOfAllLiterals();
  };
  inline
  bool Clause::isInputClause() const
  {
    CALL("isInputClause()");
    return _background.isInput();
  };
  inline
  bool Clause::isInputOrPreprocessedInputClause() const
  {
    CALL("isInputOrPreprocessedInputClause() const");
    return _background.isInput() || _background.isPrepro();
  };
  inline
  bool Clause::isPersistent() const
  {
    CALL("isPersistent() const");
    return  mainSet() < MainSetTrash; 
  };
  inline
  bool Clause::hasSubsumedPremiseInstance() const
  {
    CALL("hasSubsumedPremiseInstance() const");
    return flag(FlagHasSubsumedPremiseInstance);
  };
  inline
  void Clause::registerPremiseInstanceSubsumption()
  {
    CALL("registerPremiseInstanceSubsumption()");
    setFlag( FlagHasSubsumedPremiseInstance);
  };
  inline
  bool Clause::paramodulationsIntoProtected() const
  {
    CALL("paramodulationsIntoProtected() const");
    return flag(FlagParamodulationsIntoProtected);
  };
  inline
  void Clause::protectParamodulationsInto()
  {
    CALL("protectParamodulationsInto()");
    setFlag(FlagParamodulationsIntoProtected);
  };
  inline
  bool Clause::isBlockedForSubsumption() const
  {
    CALL("isBlockedForSubsumption() const");
    return flag(FlagBlockedForSubsumption);
  };
  inline
  void Clause::blockForSubsumption()
  { 
    CALL("blockForSubsumption()");
    ASSERT(!isBlockedForSubsumption());
    setFlag(FlagBlockedForSubsumption);
  };
  inline
  void Clause::unblockForSubsumption()
  {
    CALL("unblockForSubsumption()");
    ASSERT(isBlockedForSubsumption());
    clearFlag(FlagBlockedForSubsumption);
  };
  inline
  bool Clause::isBlockedForDemodulationFrom() const
  {
    CALL("isBlockedForDemodulationFrom() const");
    return flag(FlagBlockedForDemodulationFrom);
  };
  inline
  void Clause::blockForDemodulationFrom()
  {
    CALL("blockForDemodulationFrom()");
    setFlag(FlagBlockedForDemodulationFrom);
  };
  inline
  void Clause::unblockForDemodulationFrom()
  {
    CALL("unblockForDemodulationFrom()");
    clearFlag(FlagBlockedForDemodulationFrom);
  };
  inline
  ulong Clause::NumberOfLiteral(const Term* lit) const
  {
    CALL("NumberOfLiteral(Term* lit) const");
    ulong i = 0;
    for (LiteralList::Iterator iter(LitList());;iter.next())
      {
	ASSERT(i < LitList()->length());
	if (iter.currentLiteral() == lit) return i;
	i++; 
      };
    ICP("0");
  }; 
  inline
  bool Clause::Empty() const 
  { 
    return !numOfAllLiterals(); 
  };
  inline
  bool Clause::isAnswer() const 
  {
    CALL("isAnswer() const");
    ASSERT(LitList()->numOfSelectedLits() || LitList()->containsOnlyAnswerLiterals());
    return !LitList()->numOfSelectedLits();
  };
  inline
  bool Clause::Unit() const 
  { 
    CALL("Unit() const"); 
    if (flag(FlagContainsAnswerLiterals)) 
      {
	ulong n = 0UL;
	for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
	  if (!iter.currentLiteral()->isAnswerLiteral())
	    {
	      if (n) return false;
	      n++;                
	    };
	ASSERT((n == 0UL) || (n == 1UL)); 
	return n != 0UL; 
      } 
    else 
      return (numOfAllLiterals() == 1);
  };
  inline
  bool Clause::PositiveUnitEquality() const
  { 
    if (!Unit()) return false;    
    const Term* lit = LitList()->firstElem()->literal();
    return (lit->Positive()) && (lit->IsEquality());
  };
  inline
  bool Clause::posEqSelected() const 
  {
    CALL("posEqSelected() const");
    ulong numOfSel = LitList()->numOfSelectedLits();
    for (LiteralList::Iterator iter(LitList()); numOfSel; iter.next())
      {
	if (iter.currentElement()->isEquality() && 
	    iter.currentElement()->positive() &&
	    iter.currentElement()->properties().useEqForParamodulationFrom())
	  return true; 
	numOfSel--;
      };
    return false;
  };
  inline
  bool Clause::isTransitivityLaw() const
  {
    ulong dummy;
    return isTransitivityLaw(dummy);
  };
  inline
  bool Clause::IsOrphan() const 
  {
    CALL("IsOrphan() const");
    bool parent;
    Clause* cl;
    for (const ClauseBackground::AncestorList* anc = _background.ancestors(); 
	 anc; 
	 anc = anc->tl())
     {
       cl = anc->hd().strip(parent);
       if (parent && !cl->isPersistent()) return true;
     };
    return false;
  };
inline void Clause::Remove(Clause* c)
{
  Clause* previous_cl = c->previous;
  Clause* next_cl = c->next;
  previous_cl->next = next_cl;
  next_cl->previous = previous_cl;
};  
inline Clause* Clause::MakeEmptyClause()   
{
  return new Clause(new (0UL) LiteralList(0UL,0UL));
};
inline bool Clause::IsRule() const
{
  CALL("IsRule() const");
  ASSERT(!Empty());
  if (Unit()) return false;
  LiteralList::Iterator iter(LitList());
  while (iter.notEnd() && 
	 (iter.currentLiteral()->IsSplitting() ||
	  iter.currentLiteral()->isAnswerLiteral()))
    {
      iter.next();
    };
  if (iter.end()) return false; 
  Term::Polarity firstLitPol = iter.currentLiteral()->polarity();
  iter.next();
  while (iter.notEnd())
    {
      if (!iter.currentLiteral()->IsSplitting()
	  && !iter.currentLiteral()->isAnswerLiteral()
	  && iter.currentLiteral()->polarity() != firstLitPol) 
	return true; 
      iter.next();   
    };
  return false; 
}; 
inline bool Clause::singletonVar(ulong var) const
{
  CALL("singletonVar(ulong var) const");
  Term::Iterator litIter;
  ulong count = 0UL;
  for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
    {
      litIter.Reset(iter.currentLiteral());
      if (litIter.NextVar())
	{
	  do
	    {
	      if (litIter.symbolRef().var() == var)
		{
		  ++count;
		  if (count > 1UL) return false;
		};
	    }
	  while (litIter.NextVar());
	};
    };
  return true;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#  define DEBUG_NAMESPACE "Clause::Context"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  inline
  long Clause::Context::computeWeight(long weightBase,
				      bool subgoal,
				      bool posEqSelected) const
  {
    CALL("computeWeight(..)");
    long wt = weightBase; 
    if (_useSelectedPosEqPenaltyCoefficient && posEqSelected)
      {
	wt = (long)(wt * _selectedPosEqPenaltyCoefficient);
      };
    if (_useNongoalPenaltyCoefficient && !subgoal)
      {
	wt = (long)(wt * _nongoalPenaltyCoefficient);	  
      };
    if (wt > (long)VampireKernelConst::MaxClauseSize) 
      {
	wt = VampireKernelConst::MaxClauseSize;
      }
    else 
      if (wt < 1)
	{
	  wt = 1;
	};
    return wt;
  }; 
}; 
#endif
