/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ClauseBackground.hpp"
using namespace VK;
using namespace BK;
#ifdef DEBUG_CLAUSE_BACKGROUND
#undef DEBUG_NAMESPACE
#define DEBUG_NAMESPACE "ClauseBackground"
#endif
#include "debugMacros.hpp"
void ClauseBackground::copyTo(ClauseBackground& copy) const 
{
  CALL("ClauseBackground& copy");
  if (isInput() || isNameIntro())
    {
      copy = *this;
    }
  else
    {
      AncestorList* orgClauses = _ancestors;
      AncestorList* targClauses = 0;
      AncestorList** targClausesAddr = &targClauses;      
      while (orgClauses)
	{
	  *targClausesAddr = 
	    new AncestorList(orgClauses->hd(),static_cast<AncestorList*>(0));
	  targClausesAddr = &((*targClausesAddr)->tl());
	  orgClauses = orgClauses->tl();
	};
      copy._rules = _rules;
      copy._ancestors = targClauses;
      copy._inferenceDepth = _inferenceDepth;
    };
}; 
ostream& ClauseBackground::outputRules(ostream& str,const BitWord& _rules) 
{
  if (_rules.bit(ClauseBackground::Input)) str << "in ";
  if (_rules.bit(ClauseBackground::Prepro)) str << "pp ";
  if (_rules.bit(ClauseBackground::Subgoal)) str << "sg ";	
  if (_rules.bit(ClauseBackground::Support)) str << "sprt ";  
  if (_rules.bit(ClauseBackground::NonSubsuming)) str << "nsmng "; 
  if (_rules.bit(ClauseBackground::NonSubsumable)) str << "nsmbl ";  
  if (_rules.bit(ClauseBackground::BinRes)) str << "br "; 
  if (_rules.bit(ClauseBackground::ForwTransChain)) str << "ftc ";
  if (_rules.bit(ClauseBackground::BackTransChain)) str << "btc ";
  if (_rules.bit(ClauseBackground::ForwSup)) str << "fs ";
  if (_rules.bit(ClauseBackground::BackSup)) str << "bs ";
  if (_rules.bit(ClauseBackground::EqRes)) str << "er ";
  if (_rules.bit(ClauseBackground::EqFact)) str << "ef ";
  if (_rules.bit(ClauseBackground::ForwDemod)) str << "fd ";
  if (_rules.bit(ClauseBackground::BackDemod)) str << "bd ";
  if (_rules.bit(ClauseBackground::EqResSimp)) str << "ers ";
  if (_rules.bit(ClauseBackground::ForwSubsRes)) str << "fsr ";
  if (_rules.bit(ClauseBackground::BackSubsRes)) str << "bsr ";
  if (_rules.bit(ClauseBackground::Reanimation)) str << "rea ";
  if (_rules.bit(ClauseBackground::ExternalEvaluation)) str << "eeval ";
  if (_rules.bit(ClauseBackground::NameIntro)) str << "ni ";
  if (_rules.bit(ClauseBackground::Folding)) str << "fld ";
  if (_rules.bit(ClauseBackground::Clausification)) str << "cl ";
  if (_rules.bit(ClauseBackground::PartOfNameIntroForPropSplitting)) str << "psplnm ";
  if (_rules.bit(ClauseBackground::PropSplitting)) str << "spl ";
  return str;
}; 
#ifdef DEBUG_CLAUSE_BACKGROUND
#undef DEBUG_NAMESPACE
#define DEBUG_NAMESPACE "OpenClauseBackground"
#endif
#include "debugMacros.hpp"
void OpenClauseBackground::writeNewBackgroundTo(ClauseBackground& b) 
{
  CALL("writeNewBackgroundTo(ClauseBackground& b)");
  ClauseBackground::AncestorList* ancestors = 0;
  for (const Ancestor* anc = _ancestors.begin(); 
       anc < _ancestors.end(); 
       anc++)
    {
      ancestors = new ClauseBackground::AncestorList(*anc,ancestors);      
    };
  b._rules = _rules;
  b._ancestors = ancestors;
  b._inferenceDepth = _inferenceDepth;
}; 
