/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ClauseSet.hpp"
#include "Clause.inl"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_SET
#define DEBUG_NAMESPACE "ClauseSet"
#endif
#include "debugMacros.hpp"
ClauseSet::ClauseSet(const Clause::MainSet& tag) :
  _tag(tag),
  numOfClauses(0L), 
  removed(0L),
  _mainIteratorIsEngaged(false)
#ifdef DEBUG_NAMESPACE
  , name("?")
#endif
{
};
ClauseSet::ClauseSet(const Clause::MainSet& tag,const char* nm) : 
  _tag(tag),
  numOfClauses(0L), 
  removed(0L),
  _mainIteratorIsEngaged(false)
#ifdef DEBUG_NAMESPACE
  , name(nm)
#endif
{
};
ClauseSet::~ClauseSet() 
{
};
void ClauseSet::Insert(Clause* c) 
{ 
  c->setMainSet(_tag);
  numOfClauses++; 
  Clause* next_cl = clauses.Next();
  c->SetNext(next_cl);
  next_cl->SetPrevious(c);
  clauses.SetNext(c);
  c->SetPrevious(&clauses);
};
void ClauseSet::Remove(Clause* c) 
{
  CALLM(name,"Remove(Clause* c)"); 
  numOfClauses--; 
  ASSERT(numOfClauses >= 0L);
  removed++;
  if (c == current)
    {
      current = current->Next();
      if (current == (&clauses)) current = 0; 
    };
  Clause::Remove(c); 
  c->setMainSet(Clause::MainSetFree);
}; 
void ClauseSet::receive(Clause* const & cl)
{ 
  CALLM(name,"receive(Clause* const & cl)");
  Insert(cl);
}; 
Clause* ClauseSet::Next() 
{
  CALLM(name,"Next()");
  Clause* res = current;
  if (current)
    {
      current = current->Next();
      if (current == (&clauses)) current = 0; 
    }
  else
    _mainIteratorIsEngaged = false;
  return res;
}; 
Clause* ClauseSet::Top() 
{
  CALLM(name,"Top()");
  if (Empty()) return 0;     
  return clauses.Next();
};
Clause* ClauseSet::Pop()
{
  CALLM(name,"Pop()");
  ASSERT(Nonempty());
  Clause* c = clauses.Next();
  Remove(c);
  return c;
}; 
void ClauseSet::Purge()
{
  CALL("Purge()");
  while (!Empty())
    {
      Pop();
    };
  ASSERT(numOfClauses == 0UL);
  removed = 0L;
}; 
void ClauseSet::DestroyAllClauses()
{
  CALLM(name,"DestroyAllClauses()");
  while (!Empty())
    {
      Pop()->Destroy();
    };
  removed = 0L;
}; 
#ifndef NO_DEBUG_VIS
ostream& ClauseSet::outputClauses(ostream& str)
{
  resetMainIterator();
  for (Clause* c = Next(); c; c = Next()) str << c << '\n';
  return str;
};
#endif
