/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef COMPONENT_NAME_HPP
#define COMPONENT_NAME_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Receiver.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "ForwardSubsumptionCode.hpp"
#include "ForwardSubsumptionCommand.hpp"
#include "ExpandingStack.hpp"
#include "ObjectCache.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "ClauseAssembler.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
#define DEBUG_NAMESPACE "ComponentName"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class ComponentName 
  {
  public:
    ComponentName(ClauseAssembler* nameIntroAssembler,
		  BK::Receiver<Clause*>* nameIntroReceiver);
    ~ComponentName();
    bool findName(BasicTempLitList& component,
		  Clause*& nameIntroClause);
    Clause* registerNameIntro(const BasicTempLitList& component,
			      const TemporaryLiteral& definedLit);
    bool findName(const Flatterm* term,Clause*& nameIntroClause)
    {
      CALL("findName(const Flatterm* term,Clause*& nameIntroClause)");
      makeDummyLit(term);
      return findName(dummyComponent,nameIntroClause);
    };
    Clause* registerNameIntro(const Flatterm* term,
			      const TemporaryLiteral& definedLit,
			      const BasicTempLitList& definitionBody);
#ifndef NO_DEBUG_VIS 
    ostream& outputQuery(ostream& str) { return query.output(str); };
#endif
  private:
    class Query
    {
    public:
      class Literal
      {    
      public:  
	Literal() {};
	~Literal() {};
	void* operator new(size_t)
	{
	  CALL("Query::Literal::operator new(size_t)");
	  return BK::GlobAlloc::allocate(sizeof(Literal));
	};
	void operator delete(void* obj)
	{
	  CALL("Query::Literal::operator delete(void* obj)");
	  BK::GlobAlloc::deallocate(obj,sizeof(Literal));
	};
	const Flatterm* Args() const { return arg1; };  
	const Flatterm* Arg1() const { return arg1; }; 
	const Flatterm* Arg2() const { return arg2; };
	Literal* Next() const { return next; };
	void SetNext(Literal* nl) { next = nl; };
	void Set(const Flatterm* a,Literal* nl) { arg1 = a; next = nl; };
	void Set(const Flatterm* a1,const Flatterm* a2,Literal* nl) 
	{ arg1 = a1; arg2 = a2; next = nl; };
	void Capture() { captured = true; };
	void Release() { captured = false; };
	bool Captured() const { return captured; };
      public:
	ostream& output(ostream& str) const { return (Args() - 1)->outputTerm(str); };
      private:
	const Flatterm* arg1;
	const Flatterm* arg2;
	bool captured;
	Literal* next;
      }; 
    public:
      Query();
      ~Query() { CALL("destructor ~Query()"); };
      void ResetQuery();
      void EndOfQuery();
      bool PushLit(const Flatterm* lit);
      bool PushUnordEq(Term header,const Flatterm* arg1,const Flatterm* arg2);
      bool PushOrdEq(Term header,const Flatterm* arg1,const Flatterm* arg2);
      Literal* FirstLit(ulong header_num)  
      {
	CALL("FirstLit(ulong header_num)");
	if (header_num >= _litsWithHeader.size()) return 0;
	return _litsWithHeader[header_num];
      }; 
      ulong NumOfDiffVars() const { return numOfDiffVars; };
      ulong NumOfLits() const { return numOfLits; }
      bool Load(BasicTempLitList& component,ulong numOfVars);
    private:
      void LinkEquations();
    public: 
      ostream& output(ostream& str) const
      {
	for (ulong p = 0; p < _presentHeaders.size(); p++)
	  for (const Literal* lit = _litsWithHeader[_presentHeaders.nth(p)]; lit; lit = lit->Next())
	    lit->output(str) << '\n';
	return str;
      };
    private: 
      BK::ObjectCache<BK::GlobAlloc,64UL,Literal> _litCache;
      BK::Array<BK::GlobAlloc,Literal*,BK::DefaultRelocation<Literal*>,512UL> _litsWithHeader;
      ulong _posOrdEqHeaderNum;
      ulong _negOrdEqHeaderNum;
      BK::ExpandingStack<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,32UL> _presentHeaders;
      Literal* last_pos_unord_eq;
      Literal* last_neg_unord_eq;
      ulong numOfDiffVars;
      ulong numOfLits;
    }; 
  private: 
    bool findVariant(void*& indexedObject);  
    void makeDummyLit(const Flatterm* term)
    {
      CALL("makeDummyLit(const Flatterm* term)");
      LiteralProperties prop; 
      dummyTmpLiteral->pushNonpropHeader(Term::dummyPositiveHeader1(),
					 prop);
      dummyTmpLiteral->pushTerm(term); 
      dummyTmpLiteral->endOfLiteral();
    };
  private:
    ClauseAssembler* _nameIntroAssembler;
    BK::Receiver<Clause*>* _nameIntroReceiver;
    const Flatterm* subst[VampireKernelConst::MaxClauseSize]; 
    ForwardSubsumptionCode variantIndex;
    Query query;
    TemporaryLiteral::Extras::Handle dummyTmpLiteral;
    BasicTempLitList dummyComponent;
    BK::ExpandingStack<BK::GlobAlloc,Query::Literal*,BK::FastRelocation<Query::Literal*>,32UL> backtrack_lit;
    BK::Stack<ForwardSubsumptionCommand*,ForwardSubsumptionCommand::MaxCodeDepth> backtrack_instr;
    BK::Stack<const Flatterm*,ForwardSubsumptionCommand::MaxCodeDepth> backtrack_cursor; 
  }; 
}; 
#endif
