/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "EqualityFactoring.hpp"
#include "TemporaryClauseCollector.hpp"
#include "Clause.inl"
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_EQUALITY_FACTORING
#  define DEBUG_NAMESPACE "EqualityFactoring"
#endif
#include "debugMacros.hpp"
EqualityFactoring::EqualityFactoring(long ind) : 
  OPEN_BR(1,"br 1") DCOMMA 
_positiveEqualities(DOP("EqualityFactoring::_positiveEqualities")),
  _posEqProperties(DOP("EqualityFactoring::_posEqProperties")),
  _otherLiterals(DOP("EqualityFactoring::_otherLiterals")),
  _otherLitProperties(DOP("EqualityFactoring::_otherLitProperties")),
  _varIndex(ind),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize)
{
  CALL("constructor EqualityFactoring()");
  DOP(_positiveEqualities.freeze());
  DOP(_posEqProperties.freeze());
  DOP(_otherLiterals.freeze());
  DOP(_otherLitProperties.freeze());
};
EqualityFactoring::~EqualityFactoring() {};
bool EqualityFactoring::Load(Clause* cl) 
{
  CALL("Load(Clause* cl)");    
  DOP(_positiveEqualities.unfreeze());  
  if (!_positiveEqualities.tryToExpand(cl->numOfAllLiterals()))
    return false;
  DOP(_positiveEqualities.freeze());   
  DOP(_posEqProperties.unfreeze());  
  if (!_posEqProperties.tryToExpand(cl->numOfAllLiterals()))
    return false;
  DOP(_posEqProperties.freeze());   
  DOP(_otherLiterals.unfreeze());  
  if (!_otherLiterals.tryToExpand(cl->numOfAllLiterals()))
    return false;
  DOP(_otherLiterals.freeze());     
  DOP(_otherLitProperties.unfreeze());  
  if (!_otherLitProperties.tryToExpand(cl->numOfAllLiterals()))
    return false;
  DOP(_otherLitProperties.freeze());     
  _positiveEqualities.reset();
  _posEqProperties.reset();
  _otherLiterals.reset();
  _otherLitProperties.reset();
  _freePosEqMemory = _posEqMemory;
  _lastPosEqNum = -1;
  _lastSelectedPosEqNum = -1;
  _freshPremise = true;
  _premise = cl;
  LiteralList* lits = cl->LitList();
  long num_of_max = lits->numOfSelectedLits();
  if (num_of_max == 1) return false;
  LiteralList::Iterator iter(lits);
  do
    {
      loadSelected(iter.currentLiteral(),iter.currentElement()->properties());
      iter.next();       
      num_of_max--; 
    }
  while (num_of_max); 
  if (_lastPosEqNum == -1) return false;
#ifdef DEBUG_NAMESPACE
  COP("max loaded"); 
  long i;
  for (i = 0; i<= _lastPosEqNum; i++) 
    { 
      ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(i))); 
    }; 
#endif 
  while (iter.notEnd())
    {               
      loadUnselected(iter.currentLiteral(),iter.currentElement()->properties());
      iter.next();
    };
#ifdef DEBUG_NAMESPACE
  for (i = 0; i<= _lastPosEqNum; i++) 
    { 
      COP("nonmax loaded"); 
      ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(i))); 
    }; 
#endif
  return (_lastPosEqNum > 0);
}; 
bool EqualityFactoring::NextInference()
{
  CALL("NextInference()");
  if (_freshPremise)
    {
      _freshPremise = false;
#ifdef DEBUG_NAMESPACE
      for (long i = 0; i<= _lastPosEqNum; i++) 
	{ 
	  ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(i))); 
	}; 
#endif       
      if (!FirstInf()) return false;
    }
  else 
    {
    try_next_inf:
#ifdef DEBUG_NAMESPACE
      for (long i = 0; i<= _lastPosEqNum; i++) 
	{ 
	  ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(i))); 
	}; 
#endif       
      if (!NextInf()) return false;
    };
  _clauseCollector->openNewClause(ClauseBackground::EqFact,_premise);
  _clauseCollector->announceNewClause();
  if (CollectResult()) return true;
  goto try_next_inf;
}; 
void 
EqualityFactoring::loadSelected(Term* lit,const LiteralProperties& prop)
{
  CALL("loadSelected(Term* lit,const LiteralProperties& prop)");
  if (lit->Positive() && lit->IsEquality())
    {
      _lastPosEqNum++;
      _lastSelectedPosEqNum++;
      _positiveEqualities.push(_freePosEqMemory);
      _posEqProperties.push(prop);
      DOP(UnifTerm* debugMem = _freePosEqMemory);
      UnifTerm::StandardTermToUnifTerm(_varIndex,lit,_freePosEqMemory);   
      ASSERT(_freePosEqMemory > debugMem);
      ASSERT(_freePosEqMemory - debugMem <= (long)VampireKernelConst::MaxTermSize);
      ASSERT(_freePosEqMemory - _posEqMemory <= (long)VampireKernelConst::MaxClauseSize);  
      ASSERT(debugMem->Bracket() == _freePosEqMemory);
      ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(_lastPosEqNum))); 
      ASSERT(_lastPosEqNum >= 0);       
      ASSERT((!_lastPosEqNum) || (UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(_lastPosEqNum - 1))));
    }
  else 
    { 
      _otherLiterals.push(lit);
      _otherLitProperties.push(prop);
    }; 
}; 
void 
EqualityFactoring::loadUnselected(Term* lit,const LiteralProperties& prop)
{
  CALL("loadUnselected(Term* lit,const LiteralProperties& prop)");
  if (lit->Positive() && lit->IsEquality())
    {
      _lastPosEqNum++;
      ASSERT(_lastPosEqNum > 0); 
      _positiveEqualities.push(_freePosEqMemory);
      _posEqProperties.push(prop);
      DOP(UnifTerm* debugMem = _freePosEqMemory);
      UnifTerm::StandardTermToUnifTerm(_varIndex,lit,_freePosEqMemory);   
      ASSERT(_freePosEqMemory > debugMem);
      ASSERT(_freePosEqMemory - debugMem <= (long)VampireKernelConst::MaxTermSize);
      ASSERT(_freePosEqMemory - _posEqMemory <= (long)VampireKernelConst::MaxClauseSize);  
      ASSERT(debugMem->Bracket() == _freePosEqMemory);
      ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(_lastPosEqNum)));
      ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(_lastPosEqNum - 1)));
    }
  else 
    { 
      _otherLiterals.push(lit);
      _otherLitProperties.push(prop);
    };
}; 
bool EqualityFactoring::FirstInf()
{
  CALL("FirstInf()");
  ASSERT(_lastSelectedPosEqNum >= 0);
  ASSERT(_lastPosEqNum >= _lastSelectedPosEqNum);
  ASSERT(_lastPosEqNum > 0);
  _currST = 0;
  _currS_T_ = 1;
  COP("COP0");
  ASSERT(UnifTerm::CorrectComplexTerm(_positiveEqualities.nth(_currST)));
  LoadST(posEqLHS(_currST),
	 posEqRHS(_currST),
	 posEqIsOriented(_currST));
 try_current_s_t_:
  COP("try_current_s_t_");
  LoadS_T_(_positiveEqualities.nth(_currS_T_),_posEqProperties.nth(_currS_T_)); 
  Unifier::current()->UnbindAll();
  if (performInference()) return true;
  _currS_T_++;
  if (_currS_T_ == _currST) _currS_T_++;
  if (_currS_T_ <= _lastPosEqNum) goto try_current_s_t_;
  _currST++;
  if (_currST > _lastSelectedPosEqNum) return false; 
  COP("COP10"); 
  LoadST(posEqLHS(_currST),
	 posEqRHS(_currST),
	 posEqIsOriented(_currST));
  _currS_T_ = 0; 
  goto try_current_s_t_;
}; 
bool EqualityFactoring::NextInf()
{
  CALL("NextInf()");
  Unifier::current()->UnbindAll();
  if (performInference()) return true;
 try_next_s_t_:
  _currS_T_++;
  if (_currS_T_ == _currST) _currS_T_++;
  if (_currS_T_ <= _lastPosEqNum)
    {
    try_curr_s_t_:
      LoadS_T_(_positiveEqualities.nth(_currS_T_),_posEqProperties.nth(_currS_T_)); 
      Unifier::current()->UnbindAll();
      if (performInference()) return true;
      goto try_next_s_t_;
    };
  _currST++;
  if (_currST > _lastSelectedPosEqNum) return false;  
  LoadST(posEqLHS(_currST),
	 posEqRHS(_currST),
	 posEqIsOriented(_currST));
  _currS_T_ = 0;
  goto try_curr_s_t_; 
}; 
bool EqualityFactoring::CollectOtherPosEq()
{
  CALL("CollectOtherPosEq()");
  LiteralProperties prop;
  for (long ln = 0; ln <= _lastPosEqNum; ln++)
    if ((ln != _currST) && (ln != _currS_T_))
      {
	prop.inheritFrom(_posEqProperties.nth(ln));
	if (!_clauseCollector->pushLiteralWithUnifier(_positiveEqualities.nth(ln),
						      prop)) 
	  return false;
      };
  return true; 
}; 
bool EqualityFactoring::CollectOtherLiterals()
{
  CALL("CollectOtherLiterals()");
  LiteralProperties prop;
  ASSERT(_otherLiterals.size() == _otherLitProperties.size());
  for (ulong n = 0UL; n < _otherLiterals.size(); ++n)
    {
      prop.inheritFrom(_otherLitProperties.nth(n));
      if (!_clauseCollector->pushLiteralWithUnifier(_varIndex,
						    _otherLiterals.nth(n),
						    prop)) 
	return false;
    };
  return true; 
}; 
bool EqualityFactoring::CollectResult()
{
  CALL("CollectResult()");
  return 
    _premise->isPersistent()
    && CollectOtherPosEq()  
    && CollectNewLits() 
    && CollectOtherLiterals() 
    && _clauseCollector->finaliseClause();
}; 
void EqualityFactoring::LoadST(UnifTerm* lhs,UnifTerm* rhs,bool oriented) 
{
  CALL("LoadST(UnifTerm* lhs,UnifTerm* rhs,bool oriented)");
  _STIsOriented = !oriented;
  _termS = lhs;
  _termT = rhs;
};
void EqualityFactoring::LoadS_T_(UnifTerm* eq_lit,const LiteralProperties& prop) 
{
  CALL("LoadS_T_(UnifTerm* eq_lit,const LiteralProperties& prop)");
  ASSERT(UnifTerm::CorrectComplexTerm(eq_lit));
  s_t_ = eq_lit;
  _termS_ = eq_lit + 1;
  _termT_ = _termS_->Bracket();
  _freshS_T_ = true;
  _propertiesOfs_t_ = prop;
};
bool EqualityFactoring::performInference()
{
  CALL("performInference()");
  if (_freshS_T_)
    {
      _freshS_T_ = false;
      return performFirstInf();
    };
  return performNextInf(); 
};   
bool EqualityFactoring::CollectNewLits()
{
  CALL("CollectResult()");  
  LiteralProperties prop1;
  prop1.inheritFrom(_propertiesOfs_t_);
  LiteralProperties prop2; 
  return
    _clauseCollector->pushLiteralWithUnifier(s_t_,prop1) &&
    _clauseCollector->pushEqLiteralWithUnifier(Term::negativeUnordEqHeader(),
					       _termT,
					       _termT_,
					       prop2);
};   
#ifndef NO_DEBUG_VIS
ostream& EqualityFactoring::outputActiveLits(ostream& str) const
{
  str << "S = "; UnifTerm::outputTerm(str,_termS) << '\n';  
  str << "T = "; UnifTerm::outputTerm(str,_termT) << '\n';
  str << "S_ = "; UnifTerm::outputTerm(str,_termS_) << '\n';  
  str << "T_ = "; UnifTerm::outputTerm(str,_termT_) << '\n';
  return str;
}; 
#endif
bool EqualityFactoring::performFirstInf()
{
  CALL("FirstInf()");
  _swapST = _STIsOriented;
  swap_s_t_ = true;
  if (Unification::Unify(_termS,_termS_)) return true;
  swap_s_t_ = false;
  UnifTerm* tmp = _termS_;
  _termS_ = _termT_;
  _termT_ = tmp;
  if (Unification::Unify(_termS,_termS_)) return true;
  if (!_swapST) return false;
  _swapST = false;     
  tmp = _termS;
  _termS = _termT;
  _termT = tmp;
  if (Unification::Unify(_termS,_termS_)) { swap_s_t_ = true; return true; };
  tmp = _termS_;
  _termS_ = _termT_;
  _termT_ = tmp;
  ASSERT(!swap_s_t_);
  ASSERT(!_swapST);
  return Unification::Unify(_termS,_termS_);
}; 
bool EqualityFactoring::performNextInf()
{
  CALL("NextInf()");
  UnifTerm* tmp;
  if (swap_s_t_)
    {
      swap_s_t_ = false;
      tmp = _termS_;
      _termS_ = _termT_;
      _termT_ = tmp;
      if (Unification::Unify(_termS,_termS_)) return true;
    };
  if (_swapST)
    {
      _swapST = false; 
      tmp = _termS;
      _termS = _termT;
      _termT = tmp;
      if (Unification::Unify(_termS,_termS_)) { swap_s_t_ = true; return true; };
      tmp = _termS_;
      _termS_ = _termT_;
      _termT_ = tmp;
      ASSERT(!swap_s_t_);
      ASSERT(!_swapST);
      return Unification::Unify(_termS,_termS_); 
    }
  else return false;
}; 
