/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef EQUALITY_FACTORING_HPP
#define EQUALITY_FACTORING_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "ExpandingStack.hpp"
#include "UnifTerm.hpp"
#include "LiteralProperties.hpp"
#include "Unification.hpp"
namespace VK
{
  class TemporaryClauseCollector;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_EQUALITY_FACTORING
#  define DEBUG_NAMESPACE "EqualityFactoring"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class EqualityFactoring
  {
  public:
    EqualityFactoring(long ind);
    ~EqualityFactoring();
    void SetIndex(long ind) { _varIndex = ind; };  
    void setClauseCollector(TemporaryClauseCollector* cc)
    {
      _clauseCollector = cc;
    };
    void setClauseSizeLimit(long lim)
    {
      _clauseSizeLimit = lim;
    };
    bool Load(Clause* cl);
    bool NextInference();
  private:   
    void Reset();  
    UnifTerm* posEqLHS(long n) {
      CALL("posEqLHS(long n)");
      if (_posEqProperties.nth(n).isReverseOrientedEq())
	return _positiveEqualities.nth(n)->Arg2(); 
      return _positiveEqualities.nth(n)->Arg1();
    };
    UnifTerm* posEqRHS(long n) {
      CALL("posEqRHS(long n)");
      if (_posEqProperties.nth(n).isReverseOrientedEq())
	return _positiveEqualities.nth(n)->Arg1(); 
      return _positiveEqualities.nth(n)->Arg2();
    };
    bool posEqIsOriented(long n) {
      CALL("posEqIsOriented(long n)");
      return 
	_posEqProperties.nth(n).isOrientableEq() &&
	(_positiveEqualities.nth(n)->symbol().Func().IsOrderedEq() ||
	 _posEqProperties.nth(n).isPreorientedEq());
    };
    void loadSelected(Term* lit,const LiteralProperties& prop);
    void loadUnselected(Term* lit,const LiteralProperties& prop);
    bool FirstInf();
    bool NextInf();
    bool CollectOtherPosEq();
    bool CollectOtherLiterals();
    bool CollectResult();
    void LoadST(UnifTerm* lhs,UnifTerm* rhs,bool oriented);
    void LoadS_T_(UnifTerm* eq_lit,const LiteralProperties& prop);
    bool performInference();
    bool CollectNewLits();   
    bool performFirstInf();
    bool performNextInf();
#ifndef NO_DEBUG_VIS
    ostream& outputActiveLits(ostream& str) const; 
#endif
  private:
    OPEN_BR_DECL(1);
    TemporaryClauseCollector* _clauseCollector;
    UnifTerm _posEqMemory[VampireKernelConst::MaxClauseSize]; 
    UnifTerm* _freePosEqMemory;
    BK::ExpandingStack<BK::GlobAlloc,UnifTerm*,BK::FastRelocation<UnifTerm*>,32UL> _positiveEqualities;
    BK::ExpandingStack<BK::GlobAlloc,LiteralProperties,BK::FastRelocation<LiteralProperties>,32UL> _posEqProperties;
    long _lastPosEqNum;
    long _lastSelectedPosEqNum;
    BK::ExpandingStack<BK::GlobAlloc,Term*,BK::FastRelocation<Term*>,32UL> _otherLiterals;
    BK::ExpandingStack<BK::GlobAlloc,LiteralProperties,BK::FastRelocation<LiteralProperties>,32UL> 
    _otherLitProperties;
    Clause* _premise;
    long _varIndex;
    bool _freshPremise;
    long _currST;
    long _currS_T_;
    bool _STIsOriented;
    bool _swapST;  
    UnifTerm* _termS;
    UnifTerm* _termT;
    bool swap_s_t_;
    UnifTerm* _termS_;
    UnifTerm* _termT_;
    bool _freshS_T_;
    UnifTerm* s_t_;
    LiteralProperties _propertiesOfs_t_;
    long _clauseSizeLimit;
    CLOSE_BR_DECL(1);
  }; 
}; 
#endif
