/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FLATTERM_HPP
#define FLATTERM_HPP
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "Unifier.hpp"
#include "VariableRenaming.hpp"
#include "Stack.hpp"
#include "Comparison.hpp"
#include "Receiver.hpp"
#include "SymbolInfo.hpp"
namespace VK
{
  class Clause;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Flatterm    
  { 
  public: 
    class Iterator;
    class ConstIterator;
    class Open;
    class OpenWithMemory;
  public:
    Flatterm() : 
      _isSubstitutionPosition(false),
      _isInputNonBasicPosition(false)
    {
    };
    ~Flatterm() {};
    Term& symbolRef() { return _symbol; };
    const Term& symbolRef() const { return _symbol; };
    bool isVariable() const { return _symbol.isVariable(); };
    bool isComplex() const 
    {
      CALL("isComplex() const");
      return _symbol.isComplex(); 
    };
    ulong Var() const 
    {
      CALL("Var() const");
      ASSERT(isVariable());
      return _symbol.var(); 
    };
    ulong functor() const 
    { 
      CALL("functor() const");
      ASSERT(isComplex());
      return _symbol.functor(); 
    };
    ulong arity() const { return _symbol.arity(); };
    bool isSubstitutionPosition() const { return _isSubstitutionPosition; };
    bool isInputNonBasicPosition() const { return _isInputNonBasicPosition; };
    bool isNonBasicPosition() const 
    { 
      return isSubstitutionPosition() || isInputNonBasicPosition();
    };
    void SetSymbol(const Term& sym) { _symbol = sym; };
    void SetSymbol(const Term& sym,bool substPos) 
    { 
      _symbol = sym; 
      setSubstitutionPositionFlag(substPos); 
    };
    void SetVar(ulong var) { _symbol.MakeVar(var); _after = this + 1; };
    void SetVar(ulong var,bool substPos) 
    { 
      _symbol.MakeVar(var); 
      _after = this + 1; 
      setSubstitutionPositionFlag(substPos); 
    };
    void SetVar(const Term& var) { _symbol = var; _after = this + 1; };
    void SetVar(const Term& var,bool substPos) 
    { 
      _symbol = var; 
      _after = this + 1; 
      setSubstitutionPositionFlag(substPos); 
    };
    void SetConst(const Term& csym) { _symbol = csym; _after = this + 1; }; 
    void SetConst(const Term& csym,bool substPos) 
    { 
      _symbol = csym; 
      _after = this + 1;
      setSubstitutionPositionFlag(substPos); 
    }; 
    void setSubstitutionPositionFlag(bool substPos)
    {
      _isSubstitutionPosition = substPos;
    };
    void setInputNonBasicPositionFlag(bool inpNonBasPos)
    {
      _isInputNonBasicPosition = inpNonBasPos;
    };
    const Flatterm* after() const { return _after; };
    Flatterm* after() { return _after; };
    const Flatterm* next() const { return (this + 1); }; 
    Flatterm* next() { return (this + 1); };
    const Flatterm* previous() const { return (this - 1); };     
    Flatterm* previous() { return (this - 1); };
    Flatterm* arg1() { return next(); };
    const Flatterm* arg1() const { return next(); };
    Flatterm* arg2() { return arg1()->after(); };
    const Flatterm* arg2() const { return arg1()->after(); };
    const Flatterm* lastSymbol() const { return after() - 1; };
    Flatterm* lastSymbol() { return after() - 1; };
    ulong subtermPosition(const Flatterm* subterm) const
    {
      return subterm - this;
    };
    const Flatterm* subtermIn(ulong position) const 
    {
      return this + position;
    };
    bool containsPosition(ulong position) const
    {
      return (this + position) <= _after; 
    };
    bool containsSubterm(const Flatterm* subterm) const 
    {
      return (subterm >= this) && (subterm < _after);
    };
    bool properlyContainsSubterm(const Flatterm* subterm) const 
    {
      return (subterm > this) && (subterm < _after);
    };
    void setAfter(Flatterm* nt) { _after = nt; };
    long size() const { return (_after - this); };
    ulong Depth() const;
    bool Nonground() const;
    bool containsSymbol(const Term& sym) const;
    bool equal(const Flatterm* t) const;
    bool equal(const Flatterm* t,
	       const SymbolInfo* commutativityTable,
	       bool& commutativityUsed) const;
    void commutativityLawsForEqual(const Flatterm* t,
				   const SymbolInfo* commutativityTable,
				   BK::Receiver<const Clause*>* laws) const;
    bool EqualArguments(const Flatterm* t) const;
    bool EqualArguments(const Flatterm* t,
			const SymbolInfo* commutativityTable,
			bool& commutativityUsed) const;
    void commutativityLawsForEqualArguments(const Flatterm* t,
					    const SymbolInfo* commutativityTable,
					    BK::Receiver<const Clause*>* laws) const;
    bool Complementary(const Flatterm* t) const;
    bool Complementary(const Flatterm* t,
		       const SymbolInfo* commutativityTable,
		       bool& commutativityUsed) const;
    void commutativityLawsForComplementary(const Flatterm* t,
					   const SymbolInfo* commutativityTable,
					   BK::Receiver<const Clause*>* laws) const;
    void renameVariables(VariableRenaming& var_renaming);
    bool IsBackJump() const { return _after < this; };
    void SetBackJump(Flatterm* jmp) 
    {
      CALL("SetBackJump(Flatterm* jmp)");
      ASSERT(jmp);
      ASSERT(jmp < this);
      _after = jmp; 
    };
    void BinSwapArgs();  
    float resistanceToUnification(ulong depthCoeff,
				  float constCoeff,
				  float varCoeff) const;
    ulong markInputNonBasicPositions(); 
    void unmarkAllSubstitutionPositions();
    void unmarkSubstitutionPositionsAbove(Flatterm* subterm);
    bool canBeEvaluated() const;
    ostream& output(ostream& str) const { return str << _symbol; };
    ostream& outputTerm(ostream& str) const;
  private:
    static 
    void 
    copyWithOffset(const Flatterm* org,Flatterm* dest,long offset);
  private:
    Flatterm* _after;
    Term _symbol;
    bool _isSubstitutionPosition;
    bool _isInputNonBasicPosition;
  }; 
}; 
namespace std
{
  inline ostream& operator<<(ostream& str,const VK::Flatterm* t) 
  {    
    return t->outputTerm(str);
  };
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::Open"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Flatterm::Open
  {
  public:
    Open() 
      : open_parentheses("open_parentheses"),
	bracket_counter(),
	depth_counter()
    {};
    Open(Flatterm* mem) 
      : open_parentheses("open_parentheses"),
	bracket_counter()
    { 
      Reset(mem);
    };
    ~Open() {};
    void Reset(Flatterm* mem) 
    {
      CALL("Reset(Flatterm* mem)");
      DOP(debug_fst_pos = mem);
      curr_pos = mem; 
      open_parentheses.reset();
      bracket_counter.Reset();
      depth_counter.Reset(1UL);
    };
    ulong Depth() const { return depth_counter.Depth(); };
    void PushVar(ulong var)
    {
      CALL("PushVar(ulong var)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 
    void PushVar(ulong var,bool substPos)
    {
      CALL("PushVar(ulong var,bool substPos)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var,substPos);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    };
    void PushVar(const Term& var)
    {
      CALL("PushVar(const Term& var)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    };
    void PushVar(const Term& var,bool substPos)
    {
      CALL("PushVar(const Term& var,bool substPos)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var,substPos);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    };
    void PushConst(const Term& csym)
    {
      CALL("PushConst(const Term& csym)");     
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetConst(csym);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 
    void PushConst(const Term& csym,bool substPos)
    {
      CALL("PushConst(const Term& csym,bool substPos)");     
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetConst(csym,substPos);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 
    void PushNonconstFun(const Term& fsym)
    {
      CALL("PushNonconstFun(const Term& fsym)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetSymbol(fsym);
      bracket_counter.Open(fsym.arity());
      open_parentheses.push(curr_pos);
      depth_counter.Down();
      curr_pos++;
    }; 
    void PushNonconstFun(const Term& fsym,bool substPos)
    {
      CALL("PushNonconstFun(const Term& fsym,bool substPos)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetSymbol(fsym,substPos);
      bracket_counter.Open(fsym.arity());
      open_parentheses.push(curr_pos);
      depth_counter.Down();
      curr_pos++;
    }; 
    void Push(const Term& sym) 
    {
      if (sym.isVariable()) { PushVar(sym); }
      else if (sym.arity()) { PushNonconstFun(sym); }
      else PushConst(sym);
    };
    void Push(const Term& sym,bool substPos) 
    {
      if (sym.isVariable()) { PushVar(sym,substPos); }
      else if (sym.arity()) { PushNonconstFun(sym,substPos); }
      else PushConst(sym,substPos);
    };
    Flatterm* CurrPos() const { return curr_pos; };
  private: 
    void CloseNParentheses(unsigned n)
    {
      depth_counter.Up(n);
      while (n)
	{
	  open_parentheses.pop()->setAfter(curr_pos);
	  n--;
	};
    }; 
  private:
    Flatterm* curr_pos;
    BK::Stack<Flatterm*,VampireKernelConst::MaxTermSize> open_parentheses;
    BracketCounter<VampireKernelConst::MaxTermDepth> bracket_counter;
    DepthCounter depth_counter;
    DOP(Flatterm* debug_fst_pos);
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::OpenWithMemory"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Flatterm::OpenWithMemory : public Flatterm::Open
  {
  public:
    OpenWithMemory() : Open(memory) {};
    void Reset() { Open::Reset(memory); };
    const Flatterm* Term() const { return memory; }; 
    Flatterm* Term() { return memory; }; 
    long currentSize() const { return CurrPos() - memory; };
  private:
    OpenWithMemory(const OpenWithMemory& orig);
    OpenWithMemory& operator=(const OpenWithMemory& orig);
  private:
    Flatterm memory[VampireKernelConst::MaxTermSize+2];
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::Iterator"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Flatterm::Iterator
  {
  public:
    Iterator() {};
    Iterator(Flatterm* t) { Reset(t); };
    void Reset(Flatterm* t)
    {
      curr_pos = t;
      end_of_term = t->after();
    };
    Flatterm* CurrPos() const { return curr_pos; };
    bool EndOfTerm() const { return (curr_pos == end_of_term); };
    bool NotEndOfTerm() const { return (curr_pos < end_of_term); };
    void NextSym() { curr_pos++; };
    void after() { curr_pos = curr_pos->after(); }; 
  private:
    Flatterm* curr_pos;
    Flatterm* end_of_term;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::ConstIterator"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Flatterm::ConstIterator
  {
  public:
    ConstIterator() {};
    ConstIterator(const Flatterm* t) { Reset(t); };
    void Reset(const Flatterm* t)
    {
      curr_pos = t;
      end_of_term = t->after();
    };
    const Flatterm* CurrPos() const { return curr_pos; };
    bool EndOfTerm() const { return (curr_pos == end_of_term); };
    bool NotEndOfTerm() const { return (curr_pos < end_of_term); };
    void NextSym() { curr_pos++; };
    void after() { curr_pos = curr_pos->after(); }; 
  private:
    const Flatterm* curr_pos;
    const Flatterm* end_of_term;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline bool Flatterm::equal(const Flatterm* t) const
  {
    CALL("equal(const Flatterm* t) const");
    if (symbolRef() != t->symbolRef()) return false; 
    register long n = size();
    if (n != t->size()) return false;
    n--;
    while (n)
      {
	if ((this[n].symbolRef()) != (t[n].symbolRef())) return false;   
	n--;  
      };
    return true;
  }; 
  inline bool Flatterm::equal(const Flatterm* t,
			      const SymbolInfo* commutativityTable,
			      bool& commutativityUsed) const
  {
    CALL("equal(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const");
    if (size() != t->size()) return false;
    ConstIterator iter1(this);
    ConstIterator iter2(t);
    const Flatterm* arg11;
    const Flatterm* arg21;
    const Flatterm* arg12;
    const Flatterm* arg22;
    commutativityUsed = false; 
    do
      {
	if (iter1.CurrPos()->symbolRef() != iter2.CurrPos()->symbolRef()) 
	  return false;
	if ((iter1.CurrPos()->isComplex()) && 
	    commutativityTable->commutativityLaw(iter1.CurrPos()->functor()))
	  {    
	    arg11 = iter1.CurrPos() + 1;
	    arg21 = iter2.CurrPos() + 1;
	    bool commUsed;
	    if (arg11->equal(arg21,commutativityTable,commUsed))
	      {      
		COP("equal 1st args");
		commutativityUsed = commutativityUsed || commUsed;
		iter1.NextSym();
		iter1.after();
		iter2.NextSym();
		iter2.after();      
	      }
	    else 
	      {
		arg12 = arg11->after();
		arg22 = arg21->after();
		if (arg12->equal(arg21,commutativityTable,commUsed)
		    && arg22->equal(arg11,commutativityTable,commUsed))
		  {
		    commutativityUsed = true;
		    iter1.after();         
		    iter2.after();
		  }
		else return false;
	      };
	  }
	else
	  {
	    iter1.NextSym();
	    iter2.NextSym();
	  };
      }
    while (iter1.NotEndOfTerm());
    return true;
  }; 
  inline 
  void 
  Flatterm::commutativityLawsForEqual(const Flatterm* t,
				      const SymbolInfo* commutativityTable,
				      BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForEqual(..) const");
    ConstIterator iter1(this);
    ConstIterator iter2(t);
    const Flatterm* arg11;
    const Flatterm* arg21;
    const Flatterm* arg12;
    const Flatterm* arg22;      
    const Clause* commutativityLaw;
    do
      {
	ASSERT(iter1.CurrPos()->symbolRef() == iter2.CurrPos()->symbolRef());
	if ((iter1.CurrPos()->isComplex()) && 
	    (commutativityLaw = commutativityTable->commutativityLaw(iter1.CurrPos()->functor())))
	  {    
	    arg11 = iter1.CurrPos() + 1;
	    arg21 = iter2.CurrPos() + 1;
	    bool commUsed;
	    if (arg11->equal(arg21,commutativityTable,commUsed))
	      {      
		COP("equal 1st args");
		if (commUsed)
		  arg11->commutativityLawsForEqual(arg21,commutativityTable,laws);
		iter1.NextSym();
		iter1.after();
		iter2.NextSym();
		iter2.after();      
	      }
	    else 
	      {
		arg12 = arg11->after();
		arg22 = arg21->after();
		arg12->commutativityLawsForEqual(arg21,commutativityTable,laws);
		arg22->commutativityLawsForEqual(arg11,commutativityTable,laws);
		laws->receive(commutativityLaw);
		iter1.after();         
		iter2.after();
	      };
	  }
	else
	  {
	    iter1.NextSym();
	    iter2.NextSym();
	  };
      }
    while (iter1.NotEndOfTerm());
  }; 
  inline bool Flatterm::EqualArguments(const Flatterm* t) const
  {
    CALL("EqualArguments(const Flatterm* t) const");
    const Flatterm* sym1 = this + 1;
    const Flatterm* sym2 = t + 1;
    while (sym1 != _after)
      {
	if (sym1->symbolRef() != sym2->symbolRef()) return false;
	sym1++;
	sym2++;
      };
    return true; 
  }; 
  inline bool Flatterm::EqualArguments(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const
  {
    CALL("EqualArguments(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const");
    commutativityUsed = false;
    bool commUsed;
    ConstIterator iter1(this);
    ConstIterator iter2(t);
    iter1.NextSym();
    iter2.NextSym();
    while (iter1.NotEndOfTerm())
      {
	if (!iter1.CurrPos()->equal(iter2.CurrPos(),commutativityTable,commUsed)) 
	  return false;
	commutativityUsed = commutativityUsed || commUsed;
	iter1.after();
	iter2.after();
      };
    return true;
  }; 
  inline
  void 
  Flatterm::commutativityLawsForEqualArguments(const Flatterm* t,
					       const SymbolInfo* commutativityTable,
					       BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForEqualArguments(..) const");
    ConstIterator iter1(this);
    ConstIterator iter2(t);
    iter1.NextSym();
    iter2.NextSym();
    while (iter1.NotEndOfTerm())
      {
	iter1.CurrPos()->commutativityLawsForEqual(iter2.CurrPos(),
						   commutativityTable,
						   laws);
	iter1.after();
	iter2.after();
      };
  }; 
  inline bool Flatterm::Complementary(const Flatterm* lit) const
  {
    return (symbolRef().ComplementaryHeader(lit->symbolRef()))
      && (EqualArguments(lit));
  }; 
  inline bool Flatterm::Complementary(const Flatterm* t,
				      const SymbolInfo* commutativityTable,
				      bool& commutativityUsed) const
  {
    return (symbolRef().ComplementaryHeader(t->symbolRef()))
      && (EqualArguments(t,commutativityTable,commutativityUsed));
  }; 
  inline 
  void 
  Flatterm::commutativityLawsForComplementary(const Flatterm* t,
					      const SymbolInfo* commutativityTable,
					      BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForComplementary(..) const");
    ASSERT(symbolRef().ComplementaryHeader(t->symbolRef()));
    commutativityLawsForEqualArguments(t,commutativityTable,laws);
  }; 
  inline float Flatterm::resistanceToUnification(ulong depthCoeff,
						 float constCoeff,
						 float varCoeff) const
  {
    CALL("resistanceToUnification(ulong depthCoeff,float constCoeff,float varCoeff) const");
    if (isVariable()) return varCoeff;
    if (!arity()) return constCoeff;
    float res = 0.0;
    const Flatterm* end = after();
    for (const Flatterm* arg = next(); arg != end; arg = arg->after())
      res += arg->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
    return 1.0 + (res/depthCoeff);
  }; 
}; 
#endif
