/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_MATCHING_INDEX_TREE_HPP
#define FORWARD_MATCHING_INDEX_TREE_HPP
#ifndef NO_DEBUG_VIS 
#include <iostream>
#endif
#include <climits>
#include "jargon.hpp"
#ifdef _SUPPRESS_WARNINGS_
#include "RuntimeError.hpp"
#endif
#include "GlobalStopFlag.hpp"
#include "GlobAlloc.hpp"
#include "Stack.hpp"
#include "EquivalenceRelation.hpp"
#include "Multiset.hpp"
#include "DestructionMode.hpp"
#include "RuntimeError.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "Flatterm.hpp"
using namespace std;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_INDEX_TREE
#define DEBUG_NAMESPACE "FMCodeTree<LeafInfo>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class LeafInfo> 
  class FMCodeTree
  {
  public:
    class InstructionWithNext;
    class InstructionWithFork;
    class Func;
    class FuncNF;
    class Var;
    class VarNF;
    class Compare;
    class CompareNF;
    class Leaf;
    class Instruction
    {
    public:
      enum Tag 
      { 
	FuncTag, 
	FuncNFTag,
	VarTag, 
	VarNFTag, 
	CompareTag,
	CompareNFTag, 
	LeafTag 
      };
    public:
      Instruction(Tag t)  : _tag(t) {};
      ~Instruction() {};
      Tag tag() const { return _tag; };
#ifndef NO_DEBUG
      bool isLeaf() const { return _tag == LeafTag; };
      bool isCompare() const { return (_tag == CompareTag) || (_tag == CompareNFTag); };
#endif  
#if (defined DEBUG_NAMESPACE) || (!(defined NO_DEBUG_VIS)) 
      bool hasNext() const { return !isLeaf(); };
      bool hasFork() const 
      { 
	return (_tag == FuncTag) ||
	  (_tag == VarTag) ||
	  (_tag == CompareTag);
      };
#endif
      Instruction* next() const
      {
	CALL("next() const");
	ASSERT(hasNext());
	return (static_cast<const InstructionWithNext*>(this))->next();
      };
      Instruction** nextAddr() 
      {
	CALL("nextAddr()");
	ASSERT(hasNext());
	return (static_cast<InstructionWithNext*>(this))->nextAddr();
      };
      void setNext(Instruction* instr) 
      {
	CALL("setNext(Instruction* instr)");
	ASSERT(hasNext());
	(static_cast<InstructionWithNext*>(this))->setNext(instr);
      };
      Instruction* fork() const
      {
	CALL("fork() const");
	ASSERT(hasFork());
	return (static_cast<const InstructionWithFork*>(this))->fork();
      };
      Instruction** forkAddr() 
      {
	CALL("forkAddr()");
	ASSERT(hasFork());
	return (static_cast<InstructionWithFork*>(this))->forkAddr();
      };
      void setFork(Instruction* instr) 
      {
	CALL("setFork(Instruction* instr)");
	ASSERT(hasFork());
	(static_cast<InstructionWithFork*>(this))->setFork(instr);
      };
      InstructionWithFork* asInstrWithFork()
      {
	CALL("asInstrWithFork()");
	ASSERT(hasFork());
	return static_cast<InstructionWithFork*>(this);
      };
      const InstructionWithFork* asInstrWithFork() const 
      {
	CALL("asInstrWithFork() const");
	ASSERT(hasFork());
	return static_cast<const InstructionWithFork*>(this);
      };
      Func* asFunc() 
      { 
	CALL("asFunc()");
	ASSERT(tag() == FuncTag);
	return static_cast<Func*>(this); 
      };  
      const Func* asFunc() const  
      { 
	CALL("asFunc() const");
	ASSERT(tag() == FuncTag);
	return static_cast<const Func*>(this); 
      };
      FuncNF* asFuncNF() 
      { 
	CALL("asFuncNF()");
	ASSERT(tag() == FuncNFTag);
	return static_cast<FuncNF*>(this); 
      };
      const FuncNF* asFuncNF() const 
      { 
	CALL("asFuncNF() const");
	ASSERT(tag() == FuncNFTag);
	return static_cast<const FuncNF*>(this); 
      };
      Var* asVar()  
      {  
	CALL("asVar()");
	ASSERT(tag() == VarTag);
	return static_cast<Var*>(this);  
      };
      const Var* asVar() const  
      {  
	CALL("asVar() const");
	ASSERT(tag() == VarTag);
	return static_cast<const Var*>(this);  
      };
      VarNF* asVarNF()  
      {  
	CALL("asVarNF()");
	ASSERT(tag() == VarNFTag);
	return static_cast<VarNF*>(this);  
      };
      const VarNF* asVarNF() const
      {  
	CALL("asVarNF() const");
	ASSERT(tag() == VarNFTag);
	return static_cast<const VarNF*>(this);  
      };
      Compare* asCompare()  
      {  
	CALL("asCompare()");
	ASSERT(tag() == CompareTag);
	return static_cast<Compare*>(this);  
      };
      const Compare* asCompare() const 
      {  
	CALL("asCompare() const");
	ASSERT(tag() == CompareTag);
	return static_cast<const Compare*>(this);  
      };
      CompareNF* asCompareNF()  
      {  
	CALL("asCompareNF()");
	ASSERT(tag() == CompareNFTag);
	return static_cast<CompareNF*>(this);  
      };
      const CompareNF* asCompareNF()  const 
      {  
	CALL("asCompareNF() const");
	ASSERT(tag() == CompareNFTag);
	return static_cast<const CompareNF*>(this);  
      };
      Leaf* asLeaf()  
      {  
	CALL("asLeaf()");
	ASSERT(tag() == LeafTag);
	return static_cast<Leaf*>(this);  
      };
      const Leaf* asLeaf()  const 
      {  
	CALL("asLeaf() const");
	ASSERT(tag() == LeafTag);
	return static_cast<const Leaf*>(this);  
      };
    public:
#ifndef NO_DEBUG_VIS 
      ostream& output(ostream& str,const Flatterm* const * subst) const
      {
	CALL("output(ostream& str,const Flatterm* const * subst) const");
	str << '[' << (ulong)this << "] ";
	switch (tag())
	  { 
	  case FuncTag: str << "Func " << asFunc()->symbol(); break; 
	  case FuncNFTag: str << "FuncNF " << asFuncNF()->symbol(); break; 
	  case VarTag: str << "Var " << asVar()->var() - subst; break;
	  case VarNFTag: str << "VarNF " << asVarNF()->var() - subst; break;
	  case CompareTag: str << "Compare " 
			       << asCompare()->var1() - subst << ' ' 
			       << asCompare()->var2() - subst; break;
	  case CompareNFTag: str << "CompareNF " 
				 << asCompareNF()->var1() - subst << ' ' 
				 << asCompareNF()->var2() - subst; break;
	  case LeafTag: str << "Leaf "; break;
	  };
	if (hasNext()) str << " n[" << (ulong)next() << "] ";
	if (hasFork()) str << " f[" << (ulong)fork() << ']'; 
	return str;
      }; 
#endif
    private:
      Instruction();
    private: 
      Tag _tag;
    }; 
    class InstructionWithNext : 
      public Instruction
    {
    public:
      InstructionWithNext(typename Instruction::Tag t)  :
	Instruction(t), 
	_next(0) 
      {
      };
      Instruction** nextAddr() { return &_next; };
      Instruction* next() const { return _next; };
      void setNext(Instruction* instr) { _next = instr; };
    private:
      InstructionWithNext();
    private: 
      Instruction* _next;
    }; 
    class InstructionWithFork : 
      public InstructionWithNext
    {
    public:
      InstructionWithFork(typename Instruction::Tag t)  : 
	InstructionWithNext(t), 
	_fork(0) 
      {
      };
      Instruction** forkAddr() { return &_fork; };
      Instruction* fork() const { return _fork; };
      void setFork(Instruction* instr) { _fork = instr; };
    private:
      InstructionWithFork();
    private: 
      Instruction* _fork;
    }; 
    class Func : public InstructionWithFork
    {
    public:
      Func(const Term& f) : 
	InstructionWithFork(Instruction::FuncTag), 
	_symbol(f) 
      {};
      const Term& symbol() const { return _symbol; };
      void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(Func));
      };
      void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(Func));
      };
    private:
      Func();
    private:
      Term _symbol;
    }; 
    class FuncNF : public InstructionWithNext
    {
    public:
      FuncNF(const Term& f) : 
	InstructionWithNext(Instruction::FuncNFTag), 
	_symbol(f) 
      {};
      const Term& symbol() const { return _symbol; };
      void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(FuncNF));
      };
      void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(FuncNF));
      };
      Func* copyWithFork(Instruction* frk) 
      {
	CALL("copyWithFork(Instruction* frk)");
	Func* res = new Func(_symbol);
	res->setNext(InstructionWithNext::next());
	res->setFork(frk);
	return res;
      };
    private:
      FuncNF();
    private:
      Term _symbol;
    }; 
    class Var : public InstructionWithFork
    {
    public:
      Var(const Flatterm** v) : 
	InstructionWithFork(Instruction::VarTag), 
	_var(v) 
      {};
      const Flatterm** var() const { return _var; };
      void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(Var));
      };
      void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(Var));
      };
    private:
      Var();
    private:
      const Flatterm** _var;
    }; 
    class VarNF : public InstructionWithNext
    {
    public:
      VarNF(const Flatterm** v) : 
	InstructionWithNext(Instruction::VarNFTag), 
	_var(v) 
      {};
      const Flatterm** var() const { return _var; };
      void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(VarNF));
      };
      void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(VarNF));
      };
      Var* copyWithFork(Instruction* frk) 
      {
	CALL("copyWithFork(Instruction* frk)");
	Var* res = new Var(_var);
	res->setNext(Instruction::next());
	res->setFork(frk);
	return res;
      };
    private:
      VarNF();
    private:
      const Flatterm** _var;
    }; 
    class Compare : public InstructionWithFork
    {
    public:
      Compare(const Flatterm** v1,const Flatterm** v2) : 
	InstructionWithFork(Instruction::CompareTag), 
	_var1(v1), 
	_var2(v2)
      {
      };
      const Flatterm** var1() const { return _var1; };
      const Flatterm** var2() const { return _var2; };
      void setVar1(const Flatterm** v) { _var1 = v; };
      void setVar2(const Flatterm** v) { _var2 = v; };
      void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(Compare));
      };
      void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(Compare));
      };
    private:
      Compare();
    private:
      const Flatterm** _var1;
      const Flatterm** _var2;
    }; 
    class CompareNF : public InstructionWithNext
    {
    public:
      CompareNF(const Flatterm** v1,const Flatterm** v2) : 
	InstructionWithNext(Instruction::CompareNFTag), 
	_var1(v1), 
	_var2(v2)
      {
      };
      const Flatterm** var1() const { return _var1; };
      const Flatterm** var2() const { return _var2; };
      void setVar1(const Flatterm** v) { _var1 = v; };
      void setVar2(const Flatterm** v) { _var2 = v; };
      void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(CompareNF));
      };
      void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(CompareNF));
      };
      Compare* copyWithFork(Instruction* frk) 
      {
	CALL("copyWithFork(Instruction* frk)");
	Compare* res = new Compare(_var1,_var2);
	res->setNext(Instruction::next());
	res->setFork(frk);
	return res;
      };
    private:
      CompareNF();
    private:
      const Flatterm** _var1;
      const Flatterm** _var2;
    }; 
    class Leaf : public Instruction
    {
    public:
      Leaf() : Instruction(Instruction::LeafTag) {};
      ~Leaf() {};
      const LeafInfo& info() const { return _info; };
      LeafInfo& info() { return _info; };
      void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(Leaf));
      };
      void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(Leaf));
      };
    private:
      LeafInfo _info;
    }; 
    class Integrator 
    {
    public:
      Integrator(const Flatterm** subst)
	: _constrToInsert("Integrator::_constrToInsert"),
	  _constrToShift("Integrator::_constrToShift"),
	  _constrPassed("Integrator::_constrPassed"),
	  _constrPassedInShift("Integrator::_constrPassedInShift"),
	  _constrIntegrated("Integrator::_constrIntegrated"),
	  _constrOld("Integrator::_constrOld"),
	  _matchingIndexSubstitution(subst)
      {
	CALL("constructor Integrator(const Flatterm** subst)");
      };
      ~Integrator() 
      {
	CALL("destructor ~Integrator()");
      };
      void Reset(Instruction** n)
      {
	_nodeAddr = n;
	_constrToInsert.Reset();
	_constrToShift.Reset();
	_constrPassed.Reset();
	_constrIntegrated.Reset();
	_constrOld.Reset();
	_registeredVariables.reset();
	_currentMatchingIndexVar = _matchingIndexSubstitution;
	_readModeFlag = ((*_nodeAddr) != 0);
	_currentDepth = 0UL;
	DOP(_debugNonlinearity = 0UL);
	DOP(_debugNumOfPassedCompares = 0UL);
      };
      BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables>& Vars() 
      { 
	return _registeredVariables; 
      };
      const Flatterm** MapVar(ulong v) const { 
	return _firstVariableOccurence[v]; 
      }; 
      void Push(const Term& sym)
      { 
	if (sym.isVariable()) { PushVar(sym); } else PushFunc(sym); 
      };  
      void PushVar(const Term& sym)
      {  
	CALL("PushVar(const Term& sym)");
	ulong var = sym.var();
	ASSERT(var < VampireKernelConst::MaxNumOfVariables); 
	if (_readModeFlag)
	  {
	    _readModeFlag = ReadVar();
	    ASSERT(debugInSubst(_currentMatchingIndexVar));
	    if (_readModeFlag)
	      {
		if (_registeredVariables.scoreOf(var)) 
		  { 
		    DOP(++_debugNonlinearity);
		    ReadCompare(_firstVariableOccurence[var],_currentMatchingIndexVar); 
		  }
		else 
		  {
		    _registeredVariables.add(var);
		    _firstVariableOccurence[var] = _currentMatchingIndexVar;
		  };        
	      } 
	    else 
	      {
		InsertVar(_currentMatchingIndexVar);
		if (_registeredVariables.scoreOf(var))
		  {
		    DOP(++_debugNonlinearity);
		    EnterWriteMode();
		    WriteCompare(_firstVariableOccurence[var],_currentMatchingIndexVar);
		  } 
		else 
		  {
		    _registeredVariables.add(var);
		    _firstVariableOccurence[var] = _currentMatchingIndexVar;
		    EnterWriteMode();
		  };     
	      };
	    _currentMatchingIndexVar++;
	  }
	else 
	  WriteVar(var);          
      }; 
      void PushFunc(const Term& sym)
      {
	CALL("PushFunc(const Term& sym)");
	if (_readModeFlag)
	  {
	    _readModeFlag = ReadFunc(sym);
	    if (!_readModeFlag) 
	      {
		InsertFunc(sym);
		EnterWriteMode(); 
	      }; 
	  }
	else 
	  {
	    WriteFunc(sym);
	  };   
      }; 
      Leaf* GetLeaf(bool& newLeaf)
      {
	CALL("GetLeaf()");
	_readModeFlag = _readModeFlag && ReadCollectedComps(); 
	newLeaf = !_readModeFlag;       
	if (_readModeFlag) 
	  {
	    ASSERT((*_nodeAddr)->isLeaf());
	    return (*_nodeAddr)->asLeaf();
	  };
	ASSERT(_debugNonlinearity == _debugNumOfPassedCompares);
	ASSERT(!(*_nodeAddr));
	*_nodeAddr = new Leaf();
	return (*_nodeAddr)->asLeaf();
      };
    private:
      typedef 
      BK::EquivalenceRelation<VampireKernelConst::MaxTermSize> 
      EqualityConstraint;
    private:
      void MoveDown() 
      { 
	_nodeAddr = (*_nodeAddr)->nextAddr(); 
	++_currentDepth;
      };
      void MoveRightFrom(InstructionWithFork* instr)
      {
	CALL("MoveRightFrom(InstructionWithFork* instr)");
	ASSERT(instr->next());
	_nodeAddr = instr->forkAddr();
      };
      void MoveRightFrom(FuncNF* instr)
      {
	CALL("MoveRightFrom(FuncNF* instr)");
	ASSERT(instr->next());
	*_nodeAddr = instr->copyWithFork(0);
	_nodeAddr = (*_nodeAddr)->forkAddr();
	delete instr;
      };
      void MoveRightFrom(VarNF* instr)
      {
	CALL("MoveRightFrom(VarNF* instr)");
	ASSERT(instr->next());
	*_nodeAddr = instr->copyWithFork(0);
	_nodeAddr = (*_nodeAddr)->forkAddr();
	delete instr;
      };
      void MoveRightFrom(CompareNF* instr)
      {
	CALL("MoveRightFrom(CompareNF* instr)");
	ASSERT(instr->next());
	*_nodeAddr = instr->copyWithFork(0);
	_nodeAddr = (*_nodeAddr)->forkAddr();
	delete instr;
      };
      ulong VarNum(const Flatterm** v) const { 
	return v - _matchingIndexSubstitution; 
      };
      const Flatterm** nativeVar(ulong var_num) const 
      {
	CALL("nativeVar(ulong var_num) const");
	ASSERT(var_num < VampireKernelConst::MaxTermSize); 
	return _matchingIndexSubstitution + var_num; 
      };
#ifdef DEBUG_NAMESPACE
      bool debugInSubst(const Flatterm** var) const 
      { 
	return (var >= _matchingIndexSubstitution) && (var < (_matchingIndexSubstitution + VampireKernelConst::MaxTermSize));
      }; 
#endif
      void MakeShifted(ulong v1,ulong v2) 
      {
	CALL("MakeShifted(ulong v1,ulong v2)");
	ASSERT(v1 < v2);
	ASSERT(!_constrIntegrated.Equivalent(v1,v2));
	_constrToShift.MakeEquivalent(v1,v2);
      };
      void MakeIntegrated(ulong v1,ulong v2) 
      {
	CALL("MakeIntegrated(ulong v1,ulong v2)");
	ASSERT(v1 < v2);
	ASSERT(!_constrIntegrated.Equivalent(v1,v2));
	ASSERT(!_constrPassed.Equivalent(v1,v2));
	ASSERT(!_constrOld.Equivalent(v1,v2));
	_constrIntegrated.MakeEquivalent(v1,v2);
      };    
#ifdef DEBUG_NAMESPACE
      void CheckConstrIntegrity()
      { 
	CALL("CheckConstrIntegrity()");
	ASSERT(_constrToInsert.Implies(_constrIntegrated));      
	ASSERT(_constrPassed.Implies(_constrIntegrated));   
	ASSERT(_constrToShift.Implies(_constrOld));        
	ASSERT(_constrPassed.Implies(_constrOld));
	static EqualityConstraint debugER("debugER");  
	debugER.MakeComposition(_constrToShift,_constrPassed);
	ASSERT(_constrOld.Implies(debugER));
	debugER.MakeComposition(_constrToInsert,_constrPassed);
	ASSERT(_constrIntegrated.Implies(debugER));
      }; 
#endif 
      void PassUpperComp(CompareNF* instr)
      {
	CALL("PassUpperComp(CompareNF* instr)");
	DOP(CheckConstrIntegrity());
	const Flatterm** var1 = instr->var1();
	const Flatterm** var2 = instr->var2();
	const ulong var_num1 = VarNum(var1);
	const ulong var_num2 = VarNum(var2);
	ASSERT(!_constrToShift.Equivalent(var_num1,var_num2));
	ASSERT(!_constrOld.Equivalent(var_num1,var_num2));
	ASSERT(!_constrPassed.Equivalent(var_num1,var_num2));
	bool shared = _constrIntegrated.Equivalent(var_num1,var_num2);
	if (shared)
	  {
	    _constrPassed.MakeEquivalent(var_num1,var_num2);
	    DOP(++_debugNumOfPassedCompares);
	    _nodeAddr = instr->nextAddr();
	    ++_currentDepth;
	  }
	else 
	  {
	    *_nodeAddr = instr->next();
	    delete instr;
	  };
	Instruction* tail = *_nodeAddr;
	ASSERT(tail);
	typename EqualityConstraint::MergedPairs 
	  mergedPairs(_constrOld,var_num1,var_num2);
	ulong merged1;
	ulong merged2;
	while (mergedPairs.nextPair(merged1,merged2))
	  {
	    ASSERT(merged1 < merged2);     
	    ASSERT(!_constrOld.Equivalent(merged1,merged2));
	    if (_constrIntegrated.Equivalent(merged1,merged2) && (!_constrPassed.Equivalent(merged1,merged2)))
	      {
		shared = true;
		_constrPassed.MakeEquivalent(merged1,merged2);
		DOP(++_debugNumOfPassedCompares);
		*_nodeAddr = new CompareNF(nativeVar(merged1),nativeVar(merged2));
		(*_nodeAddr)->setNext(tail);
		MoveDown();
	      };
	  };    
	_constrOld.MakeEquivalent(var_num1,var_num2);
	if (!shared) MakeShifted(var_num1,var_num2);
	ASSERT((*_nodeAddr) == tail);
	DOP(CheckConstrIntegrity());
      }; 
      bool PassCompWF(Compare* instr)
      {
	CALL("PassCompWF(Compare* instr)");
	ASSERT(_readModeFlag);
	const Flatterm** var1 = instr->var1();
	const Flatterm** var2 = instr->var2();
	ulong var_num1 = VarNum(var1);
	ulong var_num2 = VarNum(var2);
	ASSERT(!_constrToShift.Equivalent(var_num1,var_num2));
	ASSERT(!_constrOld.Equivalent(var_num1,var_num2)); 
	ASSERT(!_constrPassed.Equivalent(var_num1,var_num2));
	ASSERT((*_nodeAddr) == instr);
	ASSERT(instr->next());
	if (_constrIntegrated.Equivalent(var_num1,var_num2))
	  {
	    goto shift_and_insert_shared;      
	  }
	else 
	  {
	    typename EqualityConstraint::MergedPairs 
	      mergedPairs(_constrOld,var_num1,var_num2);
	    ulong merged1;
	    ulong merged2; 
	    while (mergedPairs.nextPair(merged1,merged2))
	      {
		ASSERT(merged1 < merged2);     
		ASSERT(!_constrOld.Equivalent(merged1,merged2));
		if (_constrIntegrated.Equivalent(merged1,merged2) && (!_constrPassed.Equivalent(merged1,merged2)))
		  {
		    var_num1 = merged1;
		    var_num2 = merged2;
		    instr->setVar1(nativeVar(var_num1));
		    instr->setVar2(nativeVar(var_num2));        
		    goto shift_and_insert_shared;
		  };
	      };
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed;
		_constrPassedInShift.MakeEquivalent(var_num1,var_num2);
		CopyCompsToShift(instr->nextAddr());
	      };
	    MoveRightFrom(instr->asInstrWithFork());
	    DOP(CheckConstrIntegrity());
	    return false;
	  };
      shift_and_insert_shared:
	COP("shift_and_insert_shared");
	if (instr->fork()) 
	  {
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed; 
		CopyCompsToShift(instr->forkAddr());  
	      };      
	  };
	ASSERT(instr == (*_nodeAddr));
	_constrPassed.MakeEquivalent(var_num1,var_num2);
	DOP(++_debugNumOfPassedCompares);
	MoveDown();
	Instruction* tail = *_nodeAddr; 
	ASSERT(tail);
	typename EqualityConstraint::MergedPairs 
	  mergedPairs(_constrOld,var_num1,var_num2);
	ulong merged1;
	ulong merged2; 
	while (mergedPairs.nextPair(merged1,merged2))
	  {
	    ASSERT(merged1 < merged2);     
	    ASSERT(!_constrOld.Equivalent(merged1,merged2));
	    if (_constrIntegrated.Equivalent(merged1,merged2) && (!_constrPassed.Equivalent(merged1,merged2)))
	      {
		_constrPassed.MakeEquivalent(merged1,merged2);	  
		DOP(++_debugNumOfPassedCompares);
		*_nodeAddr = new CompareNF(nativeVar(merged1),nativeVar(merged2));
		(*_nodeAddr)->setNext(tail);
		MoveDown();
	      };
	  };    
	ASSERT((*_nodeAddr) == tail);
	_constrOld.MakeEquivalent(var_num1,var_num2); 
	COP("end");
	DOP(CheckConstrIntegrity());
	ASSERT(*_nodeAddr);
	return true;
      }; 
      bool PassCompNF(CompareNF* instr)
      {
	CALL("PassCompNF(CompareNF* instr)");
	ASSERT(_readModeFlag);
	const Flatterm** var1 = instr->var1();
	const Flatterm** var2 = instr->var2();
	ulong var_num1 = VarNum(var1);
	ulong var_num2 = VarNum(var2);
	ASSERT(!_constrToShift.Equivalent(var_num1,var_num2));
	ASSERT(!_constrOld.Equivalent(var_num1,var_num2)); 
	ASSERT(!_constrPassed.Equivalent(var_num1,var_num2));
	ASSERT((*_nodeAddr) == instr);
	ASSERT(instr->next());
	if (_constrIntegrated.Equivalent(var_num1,var_num2))
	  {
	    goto shift_and_insert_shared;      
	  }
	else 
	  {
	    typename EqualityConstraint::MergedPairs 
	      mergedPairs(_constrOld,var_num1,var_num2);
	    ulong merged1;
	    ulong merged2; 
	    while (mergedPairs.nextPair(merged1,merged2))
	      {
		ASSERT(merged1 < merged2);     
		ASSERT(!_constrOld.Equivalent(merged1,merged2));
		if (_constrIntegrated.Equivalent(merged1,merged2) && 
		    !_constrPassed.Equivalent(merged1,merged2))
		  {
		    var_num1 = merged1;
		    var_num2 = merged2;
		    instr->setVar1(nativeVar(var_num1));
		    instr->setVar2(nativeVar(var_num2));        
		    goto shift_and_insert_shared;
		  };
	      };
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed;
		_constrPassedInShift.MakeEquivalent(var_num1,var_num2);
		CopyCompsToShift(instr->nextAddr());
	      };
	    MoveRightFrom(instr);
	    DOP(CheckConstrIntegrity());
	    return false;
	  };
      shift_and_insert_shared:
	COP("shift_and_insert_shared");
	ASSERT(instr == (*_nodeAddr));
	_constrPassed.MakeEquivalent(var_num1,var_num2);	  
	DOP(++_debugNumOfPassedCompares);
	MoveDown();
	Instruction* tail = *_nodeAddr;
	ASSERT(tail);
	typename EqualityConstraint::MergedPairs 
	  mergedPairs(_constrOld,var_num1,var_num2);
	ulong merged1;
	ulong merged2; 
	while (mergedPairs.nextPair(merged1,merged2))
	  {
	    ASSERT(merged1 < merged2);     
	    ASSERT(!_constrOld.Equivalent(merged1,merged2));
	    if (_constrIntegrated.Equivalent(merged1,merged2) && (!_constrPassed.Equivalent(merged1,merged2)))
	      {
		DOP(++_debugNumOfPassedCompares);
		_constrPassed.MakeEquivalent(merged1,merged2);
		*_nodeAddr = new CompareNF(nativeVar(merged1),nativeVar(merged2));
		(*_nodeAddr)->setNext(tail);
		MoveDown();
	      };
	  };    
	ASSERT((*_nodeAddr) == tail);
	_constrOld.MakeEquivalent(var_num1,var_num2); 
	COP("end");
	DOP(CheckConstrIntegrity()); 
	ASSERT(*_nodeAddr);    
	return true;
      }; 
    private:
      bool ReadFunc(const Term& sym)
      { 
	CALL("ReadFunc(const Term& sym)")
	  Instruction* curr_node;  
      check_tag:   
	curr_node = *_nodeAddr; 
	ASSERT(curr_node);
	switch (curr_node->tag())
	  {
	  case Instruction::FuncTag:  
	    if (curr_node->asFunc()->symbol() == sym)
	      {
		DOP(CheckConstrIntegrity());   
		if (curr_node->fork()) 
		  {
		    if (!_constrToShift.Identity())  
		      {       
			_constrPassedInShift = _constrPassed; 
			CopyCompsToShift(curr_node->forkAddr());
		      };
		  };
		MoveDown();
		return true;
	      };
	    if (curr_node->asFunc()->symbol() > sym)
	      {
		if (!_constrToShift.Identity())  
		  {
		    _constrPassedInShift = _constrPassed;
		    InsertCompsToShift(_nodeAddr);
		  };
		return false;
	      };
	    DOP(CheckConstrIntegrity());
	    if (curr_node->fork())      
	      {
		if (!_constrToShift.Identity())
		  {
		    _constrPassedInShift = _constrPassed;  
		    CopyCompsToShift(curr_node->nextAddr());
		  };
		MoveRightFrom(curr_node->asInstrWithFork());
		goto check_tag;
	      };
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed; 
		InsertCompsToShift(curr_node->nextAddr());
	      };
	    MoveRightFrom(curr_node->asInstrWithFork());
	    return false;
	  case Instruction::FuncNFTag:  
	    if (curr_node->asFuncNF()->symbol() == sym)
	      {
		MoveDown();
		return true;
	      };
	    if (curr_node->asFuncNF()->symbol() > sym)
	      {
		if (!_constrToShift.Identity())  
		  {
		    _constrPassedInShift = _constrPassed;
		    InsertCompsToShift(_nodeAddr);
		  };
		return false;
	      };
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed; 
		InsertCompsToShift(curr_node->nextAddr());
	      };
	    MoveRightFrom(curr_node->asFuncNF());
	    return false; 
	  case Instruction::VarTag:
	    DOP(CheckConstrIntegrity());
	    if (curr_node->fork())      
	      {
		if (!_constrToShift.Identity())
		  {
		    _constrPassedInShift = _constrPassed;  
		    CopyCompsToShift(curr_node->nextAddr());
		  };
		MoveRightFrom(curr_node->asInstrWithFork());
		goto check_tag;
	      };
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed; 
		InsertCompsToShift(curr_node->nextAddr());
	      };
	    MoveRightFrom(curr_node->asInstrWithFork());
	    return false;
	  case Instruction::VarNFTag:
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed; 
		InsertCompsToShift(curr_node->nextAddr());
	      };
	    MoveRightFrom(curr_node->asVarNF());
	    return false;
	  case Instruction::CompareTag:
	    ICP("ICP0");
	    return false;
	  case Instruction::CompareNFTag:
	    PassUpperComp(curr_node->asCompareNF());
	    goto check_tag;
#ifdef DEBUG_NAMESPACE
	  default: ICP("X0"); return false;
#else 
#ifdef _SUPPRESS_WARNINGS_
	  default: return false;
#endif
#endif
	  };
	  ICP("END");
	  return false;
      }; 
      bool ReadVar()
      {
	CALL("ReadVar()");
	Instruction* curr_node;
      check_tag:
	curr_node = *_nodeAddr;        
	switch (curr_node->tag())
	  {
	  case Instruction::FuncTag: 
	  case Instruction::FuncNFTag:
	    if (!_constrToShift.Identity())
	      {
		_constrPassedInShift = _constrPassed;
		InsertCompsToShift(_nodeAddr);
	      };   
	    return false; 
	  case Instruction::VarTag:
	    DOP(CheckConstrIntegrity());
	    if (curr_node->fork()) 
	      {
		if (!_constrToShift.Identity())
		  {
		    _constrPassedInShift = _constrPassed; 
		    CopyCompsToShift(curr_node->forkAddr()); 
		  };
	      };              
	    MoveDown();
	    return true;
	  case Instruction::VarNFTag:
	    MoveDown();
	    return true;
	  case Instruction::CompareTag:
	    ICP("ICP0");
	    return false;
	  case Instruction::CompareNFTag:
	    PassUpperComp(curr_node->asCompareNF());
	    goto check_tag;
#ifdef DEBUG_NAMESPACE
	  default: ICP("X0"); return false;
#else 
#ifdef _SUPPRESS_WARNINGS_
	  default: return false;
#endif
#endif
	  };  
      }; 
      void ReadCompare(const Flatterm** v1,const Flatterm** v2)
      {
	CALL("ReadCompare(const Flatterm** v1,const Flatterm** v2)");
	ASSERT(v1<v2);  
	ASSERT(debugInSubst(v1));
	ASSERT(debugInSubst(v2));
	ulong var_num1 = VarNum(v1);
	ulong var_num2 = VarNum(v2);
	MakeIntegrated(var_num1,var_num2);
	_constrToInsert.MakeEquivalent(var_num1,var_num2);  
	DOP(CheckConstrIntegrity());
      }; 
      bool ReadCollectedComps()
      {
	CALL("ReadCollectedComps()");
	ASSERT(_readModeFlag);
	ASSERT(*_nodeAddr);
	ASSERT((*_nodeAddr)->isLeaf() || (*_nodeAddr)->next());
	Instruction* curr_node;
      check_node:
	curr_node = *_nodeAddr;
	ASSERT(curr_node);
	if (curr_node->tag() == Instruction::CompareTag) 
	  {
	    ASSERT(curr_node->next());
	    if (PassCompWF(curr_node->asCompare()) || (*_nodeAddr)) 
	      {
		goto check_node;
	      };
	    DOP(CheckConstrIntegrity());
	    MoveCompsToInsert(_nodeAddr);
	    _constrToShift.Reset();
	    ASSERT(!(*_nodeAddr));
	    return false;
	  }
	else
	  if (curr_node->tag() == Instruction::CompareNFTag) 
	    {
	      ASSERT(curr_node->next());
	      if (PassCompNF(curr_node->asCompareNF()) || (*_nodeAddr)) 
		{
		  goto check_node;
		};
	      DOP(CheckConstrIntegrity());
	      MoveCompsToInsert(_nodeAddr);
	      _constrToShift.Reset();
	      ASSERT(!(*_nodeAddr));
	      return false;
	    };
	if (_constrToShift.Identity())
	  {
	    Instruction** _nodeAddrBackup = _nodeAddr;
	    *_nodeAddr = 0;
	    DOP(CheckConstrIntegrity());
	    if (MoveCompsToInsert(_nodeAddr))
	      {
		ASSERT(!(*_nodeAddr));
		ASSERT((*_nodeAddrBackup)->tag() == Instruction::CompareNFTag);
		CompareNF* tmp = static_cast<CompareNF*>(*_nodeAddrBackup);
		(*_nodeAddrBackup) = tmp->copyWithFork(curr_node);
		if (_nodeAddr == tmp->nextAddr())
		  _nodeAddr = (*_nodeAddrBackup)->nextAddr();
		delete tmp;
		return false;
	      }
	    else 
	      {
		ASSERT(_nodeAddr == _nodeAddrBackup);
		ASSERT(!(*_nodeAddr));
		*_nodeAddr = curr_node;          
		return true; 
	      };      
	  }
	else 
	  { 
	    _constrPassedInShift = _constrPassed;
	    InsertCompsToShift(_nodeAddr);
	    ASSERT((*_nodeAddr)->tag() == Instruction::CompareNFTag);       
	    ASSERT((!_constrIntegrated.Equivalent(VarNum((*_nodeAddr)->asCompareNF()->var1()),VarNum((*_nodeAddr)->asCompareNF()->var2()))));
	    MoveRightFrom((*_nodeAddr)->asCompareNF());
	    MoveCompsToInsert(_nodeAddr);
	    ASSERT(!(*_nodeAddr));
	    return false; 
	  };
      }; 
    private:
      void InsertVar(const Flatterm** v)
      {
	CALL("InsertVar(const Flatterm** v)");
	ASSERT(debugInSubst(v));
	Instruction* tmp = *_nodeAddr;
	if (tmp || (_currentDepth < _startDepthForSkeletonNF))
	  {
	    *_nodeAddr = new Var(v);
	    (*_nodeAddr)->setFork(tmp);
	  }
	else
	  {
	    *_nodeAddr = new VarNF(v);
	  };
	MoveDown(); 
      }; 
      void InsertFunc(const Term& func)
      {
	CALL("InsertFunc(const Term& func)");
	Instruction* tmp = *_nodeAddr;
	if (tmp || (_currentDepth < _startDepthForSkeletonNF))
	  {
	    *_nodeAddr = new Func(func);
	    (*_nodeAddr)->setFork(tmp);
	  }
	else
	  *_nodeAddr = new FuncNF(func);
	MoveDown();
      }; 
      void EnterWriteMode()
      {
	CALL("EnterWriteMode()");
	ASSERT(!(*_nodeAddr));
	ASSERT(_constrToShift.Identity());
	ASSERT(_constrToInsert.Implies(_constrIntegrated));
	ASSERT(_constrPassed.Implies(_constrIntegrated));
	ASSERT(_constrPassed.Implies(_constrOld));     
	MoveCompsToInsert(_nodeAddr);
	ASSERT(!(*_nodeAddr));  
      }; 
      void WriteVar(const ulong& v)
      {
	CALL("WriteVar(const ulong& v)");
	ASSERT(debugInSubst(_currentMatchingIndexVar));
	if (_currentDepth < _startDepthForSkeletonNF)
	  {
	    *_nodeAddr = new Var(_currentMatchingIndexVar);
	  }
	else
	  *_nodeAddr = new VarNF(_currentMatchingIndexVar);
	MoveDown();
	if (_registeredVariables.scoreOf(v)) 
	  {
	    DOP(++_debugNonlinearity);
	    WriteCompare(_firstVariableOccurence[v],_currentMatchingIndexVar); 
	  }
	else 
	  {
	    _registeredVariables.add(v);
	    _firstVariableOccurence[v] = _currentMatchingIndexVar;
	  };   
	_currentMatchingIndexVar++;
      }; 
      void WriteFunc(const Term& func)
      {
	CALL("WriteFunc(const Term& func)");
	if (_currentDepth < _startDepthForSkeletonNF)
	  {
	    *_nodeAddr = new Func(func);
	  }
	else
	  {
	    *_nodeAddr = new FuncNF(func);
	  };
	MoveDown();
      }; 
      void WriteCompare(const Flatterm** v1,const Flatterm** v2)
      {
	CALL("WriteCompare(const Flatterm** v1,const Flatterm** v2)");
	ASSERT(debugInSubst(v1));
	ASSERT(debugInSubst(v2));
	*_nodeAddr = new CompareNF(v1,v2); 
	MoveDown();	  
	DOP(++_debugNumOfPassedCompares);
      }; 
    private:
      void CopyCompsToShift(Instruction** addr)
      {
	CALL("CopyCompsToShift(Instruction** addr)");
	DOP(bool debugAtLeastOneShifted = _constrToShift.Identity());  
	ASSERT(_constrPassed.Implies(_constrPassedInShift));
	ulong vn1,vn2;
	typename EqualityConstraint::Iterator iter(_constrToShift);
	Instruction* old_node = *addr;
	ASSERT(old_node);
	Instruction* new_node; 
	while (iter.Next(vn2,vn1))
	  {
	    ASSERT(vn1<vn2);
	    ASSERT(debugInSubst(nativeVar(vn1)));
	    ASSERT(debugInSubst(nativeVar(vn2)));
	    ASSERT(!_constrIntegrated.Equivalent(vn1,vn2));
	    ASSERT(_constrOld.Equivalent(vn1,vn2)); 
	    if (!_constrPassedInShift.Equivalent(vn1,vn2))
	      {
		DOP(debugAtLeastOneShifted = true);
		_constrPassedInShift.MakeEquivalent(vn1,vn2);  
		new_node = new CompareNF(nativeVar(vn1),nativeVar(vn2));
		*addr = new_node;
		addr = new_node->nextAddr(); 
	      };
	  };
	*addr = old_node;
	ASSERT(debugAtLeastOneShifted);
      }; 
      void InsertCompsToShift(Instruction** addr)
      {
	CALL("InsertCompsToShift(Instruction** addr)");
	DOP(bool debugAtLeastOneShifted = _constrToShift.Identity());  
	DOP(CheckConstrIntegrity());     
	ASSERT(_constrPassed.Implies(_constrPassedInShift));
	ulong vn1,vn2;
	Instruction* old_node = *addr;
	ASSERT(old_node);
	Instruction* new_node; 
	while (_constrToShift.Split(vn2,vn1))
	  {    
	    ASSERT(vn1<vn2);
	    ASSERT(debugInSubst(nativeVar(vn1)));
	    ASSERT(debugInSubst(nativeVar(vn2)));
	    ASSERT(!_constrIntegrated.Equivalent(vn1,vn2));
	    ASSERT(_constrOld.Equivalent(vn1,vn2));
	    if (!_constrPassedInShift.Equivalent(vn1,vn2))
	      {
		DOP(debugAtLeastOneShifted = true);  
		_constrPassedInShift.MakeEquivalent(vn1,vn2);
		new_node = new CompareNF(nativeVar(vn1),nativeVar(vn2));
		*addr = new_node;
		addr = new_node->nextAddr(); 
	      };
	  };
	ASSERT(_constrToShift.Identity());
	*addr = old_node;
	ASSERT(debugAtLeastOneShifted);
      }; 
      bool MoveCompsToInsert(Instruction**& addr)
      {
	CALL("MoveCompsToInsert(Instruction**& addr)");
	ulong vn1,vn2;
	ASSERT(!(*addr));
	Instruction* new_node;
	bool res = false;
	while (_constrToInsert.Split(vn2,vn1))
	  {
	    ASSERT(vn1<vn2);
	    ASSERT(debugInSubst(nativeVar(vn1)));
	    ASSERT(debugInSubst(nativeVar(vn2)));
	    ASSERT(_constrIntegrated.Equivalent(vn1,vn2));
	    if (!_constrPassed.Equivalent(vn1,vn2))
	      {
		ASSERT(!_constrOld.Equivalent(vn1,vn2));
		_constrPassed.MakeEquivalent(vn1,vn2);	  
		DOP(++_debugNumOfPassedCompares);
		new_node = new CompareNF(nativeVar(vn1),nativeVar(vn2));
		*addr = new_node;
		addr = new_node->nextAddr(); 
		++_currentDepth;
		res = true;   
	      };
	  };
	ASSERT(_constrToInsert.Identity());
	*addr = 0;
	ASSERT(_constrToInsert.Identity());
	ASSERT(_constrPassed.Implies(_constrIntegrated));
	ASSERT(_constrIntegrated.Implies(_constrPassed));
	return res;
      }; 
    private: 
#ifndef NO_DEBUG_VIS
      ostream& outputComps(ostream& str) 
      {
	_constrIntegrated.Normalize();
	_constrOld.Normalize();
	_constrPassed.Normalize();
	_constrToShift.Normalize();
	_constrToInsert.Normalize();
	return str << _constrIntegrated << '\n'
		   << _constrOld << '\n'
		   << _constrPassed << '\n'
		   << _constrToShift << '\n'
		   << _constrToInsert << '\n';      
      };
#endif
    private: 
      Instruction** _nodeAddr;
      EqualityConstraint _constrToInsert;
      EqualityConstraint _constrToShift;
      EqualityConstraint _constrPassed;
      EqualityConstraint _constrPassedInShift;
      EqualityConstraint _constrIntegrated;
      EqualityConstraint _constrOld;   
      const Flatterm** _matchingIndexSubstitution;
      const Flatterm** _firstVariableOccurence[VampireKernelConst::MaxNumOfVariables];
      BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> _registeredVariables; 
      const Flatterm** _currentMatchingIndexVar;
      bool _readModeFlag;
      ulong _currentDepth;
      enum 
      {
	_startDepthForSkeletonNF = 14UL
      };
#ifdef DEBUG_NAMESPACE 
      ulong _debugNonlinearity;
      ulong _debugNumOfPassedCompares;
#endif  
    }; 
    class Removal 
    {  
    public:
      Removal(const Flatterm** subst)
	: _matchingIndexSubstitution(subst)
      {};
      ~Removal() {};
      void Reset(Instruction** n)
      {
	CALL("Reset(Instruction** n)");
	ASSERT(*n);
	_nodeAddr = n;
	_branchToDestroy = n;
	_constr.Reset();
	_constrPassed.Reset();
	_registeredVariables.reset();
	_currentMatchingIndexVar = _matchingIndexSubstitution; 
      };
      bool Skip(const Term& sym)
      {
	if (sym.isVariable()) { return SkipVar(sym.var()); }
	else return SkipFunc(sym);
      };    
      bool SkipFunc(const Term& func)
      {
	CALL("SkipFunc(const Term& func)");
	ASSERT(*_nodeAddr);
	const Flatterm** var1;
	const Flatterm** var2;
	ulong var_num1;
	ulong var_num2;
	Instruction* curr_node;  
      check_node:
	curr_node = *_nodeAddr;
	switch (curr_node->tag())
	  {
	  case Instruction::FuncTag:
	    if (curr_node->asFunc()->symbol() == func) 
	      {
		if (curr_node->fork()) _branchToDestroy = _nodeAddr;
		MoveDown();
		return true;
	      };
	  try_fork:
	    if (curr_node->fork())
	      {
		MoveRight();
		_branchToDestroy = _nodeAddr;
		goto check_node;
	      }        
	    else return false; 
	  case Instruction::FuncNFTag:
	    if (curr_node->asFuncNF()->symbol() == func) 
	      {
		MoveDown();
		return true;
	      };
	    return false;
	  case Instruction::VarTag: goto try_fork;
	  case Instruction::VarNFTag: return false;
	  case Instruction::CompareTag:
	    ICP("ICP0");
	    return false;
	  case Instruction::CompareNFTag: 
	    var1 = curr_node->asCompareNF()->var1();
	    var2 = curr_node->asCompareNF()->var2();     
	    var_num1 = VarNum(var1);
	    var_num2 = VarNum(var2);
	    if (_constr.Equivalent(var_num1,var_num2))
	      {
		_constrPassed.MakeEquivalent(var_num1,var_num2);
		MoveDown();
		goto check_node;
	      }       
	    else return false;  
#ifdef DEBUG_NAMESPACE
	  default: ICP("X0"); return false;
#else 
#ifdef _SUPPRESS_WARNINGS_
	  default: return false;
#endif
#endif        
	  }; 
	ICP("END");
	return false;
      }; 
      bool SkipComps()
      {
	CALL("SkipComps()");
	ASSERT(*_nodeAddr);
	const Flatterm** var1;
	const Flatterm** var2;
	ulong var_num1;
	ulong var_num2;
	Instruction* curr_node;
      check_tag:
	curr_node = *_nodeAddr;
	if (curr_node->tag() == Instruction::CompareTag)
	  {
	    var1 = curr_node->asCompare()->var1();
	    var2 = curr_node->asCompare()->var2();
	    var_num1 = VarNum(var1);
	    var_num2 = VarNum(var2);
	    if (_constr.Equivalent(var_num1,var_num2))
	      {
		_constrPassed.MakeEquivalent(var_num1,var_num2);
		if (curr_node->fork()) _branchToDestroy = _nodeAddr;       
		MoveDown();       
		goto check_tag;
	      };
	    if (curr_node->fork())
	      {
		MoveRight();
		_branchToDestroy = _nodeAddr;
		goto check_tag;
	      };
	    return false;
	  }
	else
	  if (curr_node->tag() == Instruction::CompareNFTag)
	    {
	      var1 = curr_node->asCompareNF()->var1();
	      var2 = curr_node->asCompareNF()->var2();
	      var_num1 = VarNum(var1);
	      var_num2 = VarNum(var2);
	      if (_constr.Equivalent(var_num1,var_num2))
		{
		  _constrPassed.MakeEquivalent(var_num1,var_num2);      
		  MoveDown();       
		  goto check_tag;
		};
	      return false;
	    };
	_constr.Subtract(_constrPassed);  
	return _constr.Identity();  
      }; 
      bool SkipVar(const Term& var) { return SkipVar(var.var()); };
      Leaf* GetLeaf() const 
      {
	CALL("GetLeaf()");
	ASSERT((*_nodeAddr)->isLeaf());
	return (*_nodeAddr)->asLeaf();
      };
      Instruction** BranchToDestroy() const { return _branchToDestroy; }; 
    private: 
      typedef BK::EquivalenceRelation<VampireKernelConst::MaxTermSize> EqualityConstraint;
    private:
      void MoveDown() { _nodeAddr = (*_nodeAddr)->nextAddr(); };
      void MoveRight() { _nodeAddr = (*_nodeAddr)->forkAddr(); }; 
      ulong VarNum(const Flatterm** v) const { return v - _matchingIndexSubstitution; };   
      bool SkipVar(ulong var)
      {
	CALL(" FMCodeTree<LeafInfo>::Removal::SkipVar(ulong var)");
	if (_registeredVariables.scoreOf(var))
	  {
	    _constr.MakeEquivalent(VarNum(_firstVariableOccurence[var]),VarNum(_currentMatchingIndexVar)); 
	  }
	else 
	  {
	    _registeredVariables.add(var);
	    _firstVariableOccurence[var] = _currentMatchingIndexVar;
	  };
	_currentMatchingIndexVar++;
	return SkipVar();    
      }; 
      bool SkipVar()
      {
	CALL("SkipVar()");
	ASSERT(*_nodeAddr);
	const Flatterm** var1;
	const Flatterm** var2;
	ulong var_num1;
	ulong var_num2;
	Instruction* curr_node;  
      check_node:
	curr_node = *_nodeAddr;
	switch (curr_node->tag())
	  {
	  case Instruction::FuncTag: return false; 
	  case Instruction::FuncNFTag: return false; 
	  case Instruction::VarTag:    
	    if (curr_node->fork()) _branchToDestroy = _nodeAddr;
	    MoveDown();
	    return true;     
	  case Instruction::VarNFTag:    
	    MoveDown();
	    return true;     
	  case Instruction::CompareTag:
	    ICP("ICP0");
	    return false;
	  case Instruction::CompareNFTag: 
	    var1 = curr_node->asCompareNF()->var1();
	    var2 = curr_node->asCompareNF()->var2();
	    var_num1 = VarNum(var1);
	    var_num2 = VarNum(var2);
	    if (_constr.Equivalent(var_num1,var_num2))
	      {
		_constrPassed.MakeEquivalent(var_num1,var_num2);
		MoveDown();
		goto check_node;
	      }       
	    else return false;
#ifdef DEBUG_NAMESPACE
	  default:   
	    ICP("ICP0"); return false;
#else 
#ifdef _SUPPRESS_WARNINGS_
	  default: return false;
#endif
#endif 
	  };
      }; 
    private:
      Instruction** _nodeAddr;
      Instruction** _branchToDestroy;
      EqualityConstraint _constr; 
      EqualityConstraint _constrPassed;
      const Flatterm** _matchingIndexSubstitution;  
      const Flatterm** _firstVariableOccurence[VampireKernelConst::MaxNumOfVariables];
      BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> _registeredVariables; 
      const Flatterm** _currentMatchingIndexVar;
    }; 
    class Retrieval
    {   
    public:
      Retrieval() 
	: backtrack_nodes("Retrieval::backtrack_nodes"),
	  backtrack_cursors("Retrieval::backtrack_cursors")
      {
	CALL("constructor Retrieval()");
      };
      ~Retrieval() 
      {
	CALL("destructor ~Retrieval()");
      };
      Leaf* Match(const Flatterm* query,Instruction* tree)
      {
	CALL("Match(const Flatterm* query,Instruction* tree)");
	if (!tree) return 0;
	backtrack_nodes.reset();  
	backtrack_cursors.reset();
	const Flatterm* cursor = query;
	Instruction* curr_node = tree;
	Term curr_sym;
	COP("before check_node");
      check_node:
	COP("check_node");
	ASSERT(curr_node); 
	switch (curr_node->tag()) 
	  {
	  case Instruction::FuncTag:
	    COP("Func");
	    curr_sym = cursor->symbolRef();                 
	    if (curr_sym.isVariable()) goto backtrack;   
	  func_compare_sym:
	    if (curr_sym == (static_cast<Func*>(curr_node))->symbol())
	      {
		cursor = cursor->next();
		curr_node = curr_node->next();
		goto check_node;  
	      };      
	    if (curr_sym < (static_cast<Func*>(curr_node))->symbol()) goto backtrack;
	    curr_node = curr_node->fork();
	    if (!curr_node) goto backtrack;
	  func_check_node:
	    if (curr_node->tag() == Instruction::FuncNFTag)
	      goto funcNF_compare_sym;
	    if (curr_node->tag() == Instruction::FuncTag)
	      goto func_compare_sym;
	    ASSERT(curr_node->tag() == Instruction::CompareNFTag);
	    if ((*((static_cast<CompareNF*>(curr_node))->var1()))->equal(*((static_cast<CompareNF*>(curr_node))->var2())))
	      {
		curr_node = curr_node->next();
		goto func_check_node;
	      };
	    goto backtrack;
	  case Instruction::FuncNFTag:
	    COP("FuncNF");
	    curr_sym = cursor->symbolRef();                 
	    if (curr_sym.isVariable()) goto backtrack;   
	  funcNF_compare_sym:
	    if (curr_sym == (static_cast<FuncNF*>(curr_node))->symbol())
	      {
		cursor = cursor->next();
		curr_node = curr_node->next();
		goto check_node;  
	      };      
	    goto backtrack;
	  case Instruction::VarTag:
	    COP("Var");
	    *((static_cast<Var*>(curr_node))->var()) = cursor;  
	    if (curr_node->fork() && !cursor->symbolRef().isVariable())
	      {
		backtrack_nodes.push(curr_node->fork());
		backtrack_cursors.push(cursor);
	      };
	    cursor = cursor->after();
	    curr_node = curr_node->next();
	    goto check_node;     
	  case Instruction::VarNFTag:
	    COP("VarNF");
	    *((static_cast<VarNF*>(curr_node))->var()) = cursor; 
	    cursor = cursor->after();
	    curr_node = curr_node->next();
	    goto check_node;     
	  case Instruction::CompareTag:
	    COP("Compare");
	    if ((*((static_cast<Compare*>(curr_node))->var1()))->equal(*((static_cast<Compare*>(curr_node))->var2())))
	      {
		if (curr_node->fork())
		  {
		    backtrack_nodes.push(curr_node->fork());
		    backtrack_cursors.push(cursor);
		  };
		curr_node = curr_node->next();
		goto check_node;
	      }
	    else 
	      {
		curr_node = curr_node->fork();
		if (curr_node) goto check_node;
		goto backtrack;
	      };
	  case Instruction::CompareNFTag:
	    COP("CompareNF");
	    if ((*((static_cast<CompareNF*>(curr_node))->var1()))->equal(*((static_cast<CompareNF*>(curr_node))->var2())))
	      {
		curr_node = curr_node->next();
		goto check_node;
	      };
	    goto backtrack;
	  case Instruction::LeafTag:
	    COP("Leaf"); 
	    return (static_cast<Leaf*>(curr_node)); 
#ifdef DEBUG_NAMESPACE
	  default: COP("default1"); ICP("X0"); return 0;
#else
#ifdef _SUPPRESS_WARNINGS_
	  default: COP("default2"); return 0;  
#endif
#endif
	  };
	COP("before bakctrack");
      backtrack:
	COP("backtrack");
	BK::GlobalStopFlag::check();
	if (backtrack_nodes.empty())
	  {   
	    return 0;
	  };
	curr_node = backtrack_nodes.pop(); 
	cursor = backtrack_cursors.pop();
	goto check_node;
      }; 
      Leaf* MatchAgain()
      {
	CALL("MatchAgain()");
	const Flatterm* cursor;
	Instruction* curr_node;
	Term curr_sym;
	goto backtrack;
      check_node:
	COP("check_node");
	ASSERT(curr_node); 
	switch (curr_node->tag()) 
	  {
	  case Instruction::FuncTag:
	    COP("Func");
	    curr_sym = cursor->symbolRef();                 
	    if (curr_sym.isVariable()) goto backtrack;   
	  func_compare_sym:
	    if (curr_sym == (static_cast<Func*>(curr_node))->symbol())
	      {
		cursor = cursor->next();
		curr_node = curr_node->next();
		goto check_node;  
	      };      
	    if (curr_sym < (static_cast<Func*>(curr_node))->symbol()) goto backtrack;
	    curr_node = curr_node->fork();
	    if (!curr_node) goto backtrack;
	  func_check_node:
	    if (curr_node->tag() == Instruction::FuncNFTag)
	      goto funcNF_compare_sym;
	    if (curr_node->tag() == Instruction::FuncTag)
	      goto func_compare_sym;
	    ASSERT(curr_node->tag() == Instruction::CompareNFTag);
	    if ((*((static_cast<CompareNF*>(curr_node))->var1()))->equal(*((static_cast<CompareNF*>(curr_node))->var2())))
	      {
		curr_node = curr_node->next();
		goto func_check_node;
	      };
	    goto backtrack;
	  case Instruction::FuncNFTag:
	    COP("FuncNF");
	    curr_sym = cursor->symbolRef();                 
	    if (curr_sym.isVariable()) goto backtrack;   
	  funcNF_compare_sym:
	    if (curr_sym == (static_cast<FuncNF*>(curr_node))->symbol())
	      {
		cursor = cursor->next();
		curr_node = curr_node->next();
		goto check_node;  
	      };      
	    goto backtrack;
	  case Instruction::VarTag:
	    COP("Var");
	    *((static_cast<Var*>(curr_node))->var()) = cursor;  
	    if (curr_node->fork() && !cursor->symbolRef().isVariable())
	      {
		backtrack_nodes.push(curr_node->fork());
		backtrack_cursors.push(cursor);
	      };
	    cursor = cursor->after();
	    curr_node = curr_node->next();
	    goto check_node;     
	  case Instruction::VarNFTag:
	    COP("VarNF");
	    *((static_cast<VarNF*>(curr_node))->var()) = cursor; 
	    cursor = cursor->after();
	    curr_node = curr_node->next();
	    goto check_node;     
	  case Instruction::CompareTag:
	    COP("Compare");
	    if ((*((static_cast<Compare*>(curr_node))->var1()))->equal(*((static_cast<Compare*>(curr_node))->var2())))
	      {
		if (curr_node->fork())
		  {
		    backtrack_nodes.push(curr_node->fork());
		    backtrack_cursors.push(cursor);
		  };
		curr_node = curr_node->next();
		goto check_node;
	      }
	    else 
	      {
		curr_node = curr_node->fork();
		if (curr_node) goto check_node;
		goto backtrack;
	      };
	  case Instruction::CompareNFTag:
	    COP("CompareNF");
	    if ((*((static_cast<CompareNF*>(curr_node))->var1()))->equal(*((static_cast<CompareNF*>(curr_node))->var2())))
	      {
		curr_node = curr_node->next();
		goto check_node;
	      };
	    goto backtrack;
	  case Instruction::LeafTag:
	    COP("Leaf"); 
	    return (static_cast<Leaf*>(curr_node)); 
#ifdef DEBUG_NAMESPACE
	  default: COP("default1"); ICP("X0"); return 0;
#else
#ifdef _SUPPRESS_WARNINGS_
	  default: COP("default2"); return 0;  
#endif
#endif
	  };
	COP("before bakctrack");
      backtrack:
	COP("backtrack");
	if (backtrack_nodes.empty())
	  {
	    return 0;
	  };
	curr_node = backtrack_nodes.pop(); 
	cursor = backtrack_cursors.pop();
	goto check_node;
      }; 
    private:
      BK::Stack<Instruction*,VampireKernelConst::MaxTermSize> backtrack_nodes;
      BK::Stack<const Flatterm*,VampireKernelConst::MaxTermSize> backtrack_cursors;
    }; 
  public:
    static void Destroy(Instruction* tree);  
    static void DestroyBranch(Instruction** branch);
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_INDEX_TREE
#define DEBUG_NAMESPACE "FMCodeTree<LeafInfo>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class LeafInfo> 
  void FMCodeTree<LeafInfo>::Destroy(Instruction* tree)
  {
    CALL("Destroy(Instruction* tree)");
    if (!tree) return;
    switch (tree->tag())
      {
      case Instruction::FuncTag:
	Destroy(tree->fork());
	Destroy(tree->next());
	delete tree->asFunc();
	break;
      case Instruction::FuncNFTag:
	Destroy(tree->next());
	delete tree->asFuncNF();
	break;      
      case Instruction::VarTag:
	Destroy(tree->fork());
	Destroy(tree->next());
	delete tree->asVar();
	break;
      case Instruction::VarNFTag: 
	Destroy(tree->next());
	delete tree->asVarNF();
	break;
      case Instruction::CompareTag:
	Destroy(tree->fork());
	Destroy(tree->next());
	delete tree->asCompare();
	break;
      case Instruction::CompareNFTag: 
	Destroy(tree->next());
	delete tree->asCompareNF();
	break;
      case Instruction::LeafTag:
	delete tree->asLeaf();
	break;
#ifdef DEBUG_NAMESPACE
      default: ICP("ICP0"); return;
#else
#  ifdef _SUPPRESS_WARNINGS_
      default: BK::RuntimeError::report("Bad tag in FMCodeTree<..>::Destroy(Instruction* tree)");
#  endif
#endif
      };
  }; 
  template <class LeafInfo> 
  void FMCodeTree<LeafInfo>::DestroyBranch(Instruction** branch)
  {
    CALL("DestroyBranch(Instruction** branch)");  
    Instruction* fork;
    switch ((*branch)->tag())
      {
      case Instruction::FuncTag:
	fork = (*branch)->fork();
	Destroy((*branch)->next());
	delete (*branch)->asFunc();
	*branch = fork;
	break;
      case Instruction::FuncNFTag:
	Destroy(*branch);
	*branch = 0;
	break;
      case Instruction::VarTag:
	fork = (*branch)->fork();
	Destroy((*branch)->next());
	delete (*branch)->asVar();
	*branch = fork;
	break;
      case Instruction::VarNFTag: 
	Destroy(*branch);
	*branch = 0;
	break;
      case Instruction::CompareTag:
	fork = (*branch)->fork();
	Destroy((*branch)->next());
	delete (*branch)->asCompare();
	*branch = fork;
	break;
      case Instruction::CompareNFTag:
	Destroy(*branch);
	*branch = 0;
	break;
      case Instruction::LeafTag:
	delete (*branch)->asLeaf();
	*branch = 0;
	break;
#ifdef DEBUG_NAMESPACE
      default: ICP("ICP0"); return;
#else
#  ifdef _SUPPRESS_WARNINGS_
      default: BK::RuntimeError::report("Bad tag in FMCodeTree<..>::DestroyBranch(Instruction** branch)");
#  endif
#endif
      };
  }; 
}; 
#endif
