/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_SUBSUMPTION_HPP
#define FORWARD_SUBSUMPTION_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "ForwardSubsumptionCode.hpp"
#include "Stack.hpp"
#include "ExpandingStack.hpp"
#include "ObjectCache.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
#include "Clause.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION
#define DEBUG_NAMESPACE "ForwardSubsumption"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class ForwardSubsumptionCommand;
  class ForwardSubsumption : public ForwardSubsumptionCode 
  {
  public:
    class Query
    {
    public:
      class Literal
      {    
      public:  
	Literal() {};
	~Literal() {};
	void* operator new(size_t)
	{
	  CALL("Literal::operator new(size_t)");
	  return BK::GlobAlloc::allocate(sizeof(Literal));
	};
	void operator delete(void* obj)
	{
	  CALL("Literal::operator delete(void* obj)");
	  BK::GlobAlloc::deallocate(obj,sizeof(Literal));
	};
	const Flatterm* Args() const { return arg1; };   
	const Flatterm* Arg1() const { return arg1; };
	const Flatterm* Arg2() const { return arg2; };
	Literal* Next() const { return next; };
	void SetNext(Literal* nl) { next = nl; };
	void Set(const Flatterm* a,Literal* nl) { arg1 = a; captured = 0L; next = nl; };
	void Set(const Flatterm* a1,const Flatterm* a2,Literal* nl) 
	{ arg1 = a1; arg2 = a2; captured = 0L; next = nl; };
	void Capture() { captured++; };
	void Release() { captured--; };
	long Captured() { return captured; };
      public:
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str) const { return (Args() - 1)->outputTerm(str); };
#endif
      private:
	const Flatterm* arg1;
	const Flatterm* arg2;
	long captured;
	Literal* next;
      }; 
    public:
      Query();
      ~Query();
      void reset();
      void LinkEquations(); 
      void EndOfQuery(); 
      bool PushLit(const Flatterm* lit);  
      bool PushUnordEq(Term header,const Flatterm* arg1,const Flatterm* arg2);
      bool PushOrdEq(Term header,const Flatterm* arg1,const Flatterm* arg2);
      Literal* FirstLit(ulong header_num)  
      {
	CALL("FirstLit(ulong header_num)");
	if (header_num >= _litsWithHeader.size()) return 0;
	return _litsWithHeader[header_num];
      }; 
    public:
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const;
#endif 
    private: 
      BK::ObjectCache<BK::GlobAlloc,64UL,Literal> _litCache;
      BK::Array<BK::GlobAlloc,Literal*,BK::FastRelocation<Literal*>,512UL> _litsWithHeader;
      ulong _posOrdEqHeaderNum;
      ulong _negOrdEqHeaderNum;
      BK::ExpandingStack<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,32UL> _presentHeaders;
      Literal* last_pos_unord_eq;
      Literal* last_neg_unord_eq;
    }; 
  public:
    ForwardSubsumption();
    ~ForwardSubsumption();
    void reset();
    Query& query() { return _query; };
    Clause* Subsume();
#ifndef NO_DEBUG_VIS 
    ostream& outputQuery(ostream& str) { return _query.output(str); };
#endif
  private:
  private:
    Query _query;
    const Flatterm* subst[VampireKernelConst::MaxClauseSize];
    BK::ExpandingStack<BK::GlobAlloc,Query::Literal*,BK::FastRelocation<Query::Literal*>,32UL> backtrack_lit;
    BK::Stack<ForwardSubsumptionCommand*,ForwardSubsumptionCommand::MaxCodeDepth> backtrack_instr;
    BK::Stack<const Flatterm*,ForwardSubsumptionCommand::MaxCodeDepth> backtrack_cursor;
  }; 
}; 
#endif
