/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ForwardSubsumptionCommand.hpp" 
#include "Stack.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_COMMAND
 #define DEBUG_NAMESPACE "ForwardSubsumptionCommand"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
void ForwardSubsumptionCommand::DestroyTree(ForwardSubsumptionCommand* tree) 
{
  CALL("DestroyTree(ForwardSubsumptionCommand* tree)");
  ASSERT(tree);
  if (tree->Fork()) DestroyTree(tree->Fork());
  if (tree->IsSuccess()) 
    {
      delete tree; 
    }
  else
    {
      DestroyTree(tree->Next());
      delete tree; 
    };
}; 
void ForwardSubsumptionCommand::DestroyTree(ForwardSubsumptionCommand* tree,void (*destroyIndexedObj)(void*)) 
{
  CALL("DestroyTree(ForwardSubsumptionCommand* tree,void (*destroyIndexedObj)(void*))"); 
  ASSERT(tree);
  if (tree->Fork()) DestroyTree(tree->Fork());
  if (tree->IsSuccess()) 
    {
      destroyIndexedObj(tree->IndexedObject());
      delete tree; 
    }
  else
    {
      DestroyTree(tree->Next());
      delete tree; 
    };
}; 
#ifndef NO_DEBUG_VIS
ostream& ForwardSubsumptionCommand::outputTree(ostream& str,ForwardSubsumptionCommand* tree,const Flatterm** subst)
{
  long fork_num = 0;
  if (!tree) { return str; };
  Stack<ForwardSubsumptionCommand*,MaxCodeDepth> forks("forks");
  Stack<long,MaxCodeDepth> fork_nums("fork_nums");
  ForwardSubsumptionCommand* node = tree;
 next_branch:
  while (node)
    {
      node->output(str,subst);
      if (node->Fork()) 
	{
	  forks.push(node); 
	  fork_nums.push(fork_num); 
	  str << " [" << fork_num << ']';
	  fork_num++;
	};
      str << '\n';
      node = node->Next();
    };
  if (forks.empty()) { return str; };
  node = forks.pop();
  str << "\nFORK [" << fork_nums.pop() << "]\n";
  node = node->Fork();
  goto next_branch;
}; 
ostream& ForwardSubsumptionCommand::output(ostream& str,const Flatterm** vars) const
{
  str << '[' << (ulong)this << ']';
  Term header;
  switch (Tag())
    {
    case LitHeader: header.MakeHeader(HeaderNum()); str << "LitHeader " << header; break;
    case PropLit: header.MakeHeader(HeaderNum()); str << "PropLit " << header; break;   
    case UnordEq: header.MakeHeader(HeaderNum()); str << "UNOrdEq " << header; break; 
    case OrdEq: header.MakeHeader(HeaderNum()); str << "OrdEq " << header; break; 
    case SecondEqArg: str << "SecondEqArg "; break;
    case Func: str << "Func " << func(); break;
    case Var: str << "Var " << var() - vars; break;
    case Compare: str << "Compare " << Var1() - vars << ' ' << Var2() - vars; break;
    case CompareNF: str << "CompareNF " << Var1() - vars << ' ' << Var2() - vars; break;
    case Success: str << "Success " << TheClause(); break;
    };
  return str;
}; 
#endif
