/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_SUBSUMPTION_COMMAND_HPP
#define FORWARD_SUBSUMPTION_COMMAND_HPP
#ifndef NO_DEBUG_VIS
 #include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "DestructionMode.hpp"
#include "Term.hpp"
#include "BitSet.hpp" 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_COMMAND
 #define DEBUG_NAMESPACE "ForwardSubsumptionCommand"
#endif
#include "debugMacros.hpp"
namespace VK
{
class Clause;
class Flatterm;
class ForwardSubsumptionCommand
 {
  public:
   enum {
     MaxCodeDepth = VampireKernelConst::MaxClauseSize*2 + 1
   };
   enum TAG 
    {
     LitHeader,
     PropLit,
     UnordEq, 
     OrdEq, 
     SecondEqArg, 
     Func, 
     Var, 
     Compare,
     CompareNF, 
     Success 
    }; 
  public:
   ForwardSubsumptionCommand(TAG t) : fork(0), next(0), tag(t) 
   {
    CALL("constructor ForwardSubsumptionCommand(TAG t)");
   };
   ~ForwardSubsumptionCommand() 
    {
     CALL("destructor ~ForwardSubsumptionCommand()");
    };
   inline void* operator new(size_t)
    {
     CALL("operator new(size_t)");
      return BK::GlobAlloc::allocate(sizeof(ForwardSubsumptionCommand)); 
    };
   inline void operator delete(void* obj)
    {
     CALL("operator delete(void* obj)"); 
      BK::GlobAlloc::deallocate(obj,sizeof(ForwardSubsumptionCommand));
    };
   inline TAG Tag() const { return tag; };
   inline void SetTag(TAG t) { tag = t; };
   inline bool IsFunc() const { return (tag == Func); };
   inline bool IsCompare() const { return (tag == Compare); };
   inline bool IsCompareNF() const { return (tag == CompareNF); };  
   inline bool IsSuccess() const { return (tag == Success); };
   inline ForwardSubsumptionCommand* Next() const { return next; };
   inline void SetNext(ForwardSubsumptionCommand* c) { next = c; };
   inline ForwardSubsumptionCommand** NextAddr() { return &(next); };
   inline ForwardSubsumptionCommand* Fork() const { return fork; };
   inline void SetFork(ForwardSubsumptionCommand* c) { fork = c; };
   inline ForwardSubsumptionCommand** ForkAddr() { return &(fork); };
   inline ulong HeaderNum() const { return arg1.header_num; };
   inline void SetHeader(Term h) { arg1.header_num = h.HeaderNum(); }; 
   static inline ForwardSubsumptionCommand* NewLitHeader(Term header)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(LitHeader);
     res->SetHeader(header);
     return res;
    };
   static inline ForwardSubsumptionCommand* NewPropLit(Term header)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(PropLit);
     res->SetHeader(header);
     return res;
    }; 
   static inline ForwardSubsumptionCommand* NewUnordEq(Term header)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(UnordEq);
     res->SetHeader(header);
     return res;
    }; 
   static inline ForwardSubsumptionCommand* NewOrdEq(Term header)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(OrdEq);
     res->SetHeader(header);
     return res;
    };
   inline bool SwapFlag() const { return arg2.swap_flag; };
   inline void SwapOn() { arg2.swap_flag = true; };
   inline void SwapOff() { arg2.swap_flag = false; };    
   static inline ForwardSubsumptionCommand* NewSecondEqArg()
    {
     return new ForwardSubsumptionCommand(SecondEqArg);
    };
   inline Term func() const { return Term(arg1.func); };
   inline void SetFunc(Term f) { arg1.func = f.Content(); };
   static inline ForwardSubsumptionCommand* NewFunc(Term f)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(Func);
     res->SetFunc(f);
     return res;
    };
   inline const Flatterm** var() const { return arg1.var1; };
   inline void SetVar(const Flatterm** v) { arg1.var1 = v; }; 
   static inline ForwardSubsumptionCommand* NewVar(const Flatterm** v)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(Var);
     res->SetVar(v);
     return res; 
    };  
   inline const Flatterm** Var1() const { return arg1.var1; };
   inline void SetVar1(const Flatterm** v) { arg1.var1 = v; }; 
   inline const Flatterm** Var2() const { return arg2.var2; };
   inline void SetVar2(const Flatterm** v) { arg2.var2 = v; };
   static inline ForwardSubsumptionCommand* NewCompare(const Flatterm** v1,const Flatterm** v2)
    {
     CALL("NewCompare(const Flatterm** v1,const Flatterm** v2)");
     ASSERT(v1 < v2);
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(Compare);
     res->SetVar1(v1);
     res->SetVar2(v2);
     return res;
    }; 
   static inline ForwardSubsumptionCommand* NewCompareNF(const Flatterm** v1,const Flatterm** v2)
    {
     CALL("NewCompareNF(const Flatterm** v1,const Flatterm** v2)");     
     ASSERT(v1 < v2);
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(CompareNF);
     res->SetVar1(v1);
     res->SetVar2(v2);
     return res;
    };
   inline Clause* TheClause() const { return arg1.clause; };
   inline void* IndexedObject() const { return static_cast<void*>(arg1.clause); };
   inline void SetClause(Clause* cl) { arg1.clause = cl; };
   inline void SetIndexedObject(void* obj) { arg1.clause = static_cast<Clause*>(obj); };
   inline ulong NumOfDiffVars() const { return arg2.numOfDiffVars; };
   inline void SetNumOfDiffVars(ulong n) { arg2.numOfDiffVars = n; };
   ulong NumOfLits() const { return (ulong)next; };
   void SetNumOfLits(ulong n) { next = reinterpret_cast<ForwardSubsumptionCommand*>(n); }; 
   static inline ForwardSubsumptionCommand* NewSuccess(Clause* cl)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(Success);
     res->SetClause(cl);
     return res;
    };  
   static inline ForwardSubsumptionCommand* NewSuccess(void* indObj)
    {
     ForwardSubsumptionCommand* res = new ForwardSubsumptionCommand(Success);
     res->SetIndexedObject(indObj);
     return res;
    };  
   static void DestroyTree(ForwardSubsumptionCommand* tree); 
   static void DestroyTree(ForwardSubsumptionCommand* tree,void (*destroyIndexedObj)(void*));
  public: 
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str,const Flatterm** vars) const;
    static ostream& outputTree(ostream& str,ForwardSubsumptionCommand* tree,const Flatterm** subst);
   #endif
  private: 
   union 
    {
     ulong header_num;
     ulong func; 
     const Flatterm** var1;
     Clause* clause;
    } arg1;
   union
    { 
     const Flatterm** var2;
     bool swap_flag;
     ulong numOfDiffVars; 
    } arg2;
   ForwardSubsumptionCommand* fork; 
   ForwardSubsumptionCommand* next; 
   TAG tag;
 }; 
}; 
#endif
