/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_SUPERPOSITION_HPP
#define FORWARD_SUPERPOSITION_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Stack.hpp"
#include "IUMachine.hpp"
#include "SizeFormula.hpp"
#include "Array.hpp"
#include "ExpandingStack.hpp"
#include "GlobAlloc.hpp"
#include "SimplificationOrdering.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class ForwardProcessing;
  class Statistics;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "ForwardSuperposition"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ForwardSuperposition
  {
  public:
    ForwardSuperposition(UnifIndex* unifIndex);
    ~ForwardSuperposition();
    void setStatistics(Statistics* stat)
    {
      _dummyClauseCollector.setStatistics(stat);
      _statistics = stat;
    };
    void setClauseCollector(TemporaryClauseCollector* cc) 
    {
      _clauseCollector = cc;
    };
    void setForwardProcessing(ForwardProcessing* fp)
    {
      _forwardProcessing = fp;
    };
    void setNoDuplicateRedexes(bool fl)
    {
      _noDuplicateRedexesFlag = fl;
    };
    void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;   
    };
    void setUseInputNonBasicness(bool fl)
    {
      _useInputNonBasicness = fl;
    };
    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };
    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };
    void setClauseSizeLimit(long lim)
    {
      _clauseSizeLimit = lim;
    };
    bool Load(Clause* cl);
    bool NextInference();
  private: 
    enum Continuation { ContFSup, ContEqRes };
    enum ParamodulationMode {
      Init,
      ParaFromComplex,
      ParaFromVar 
    };
  private:
    void compileMainPremiseResidualSizeFormula();
    bool NextFSupInference();
    bool NextEqRes();
    UnifTerm* nextSelectedMainLit(LiteralProperties& prop)
    {
      if (_nextMainSelectedLitNum)
	{
	  _nextMainSelectedLitNum--;
	  prop = _mainSelectedLitProperties[_nextMainSelectedLitNum];
	  return _mainSelectedLiterals.nth(_nextMainSelectedLitNum);
	};
      return 0;
    }; 
    bool LoadLiteral(UnifTerm* lit,const LiteralProperties& prop);
    bool LoadRestOfLit();
    void loadEligibleSubtermPositions(UnifTerm* term);
    bool LoadTermToRewrite(UnifTerm* t)
    {
      CALL("LoadTermToRewrite(UnifTerm* t)");
      _currentParamodulationTargetTerm = t;
      if (!_retrievalOfUnifiable.CompileLiteral(_currentParamodulationTargetTerm))
	return false;
      _paramodulationMode = Init;
      return true;
    };
    bool Paramodulate();
    UnifIndex::IndexedObject* NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr);  
    bool CollectResult();
    bool CollectEqResResult();
    bool collectSidePremiseLiterals(const LiteralProperties::SLModifier& slmod); 
    bool collectMainPremiseLiterals(const LiteralProperties::SLModifier& slmod);
    bool collectRewrittenLiteral(const LiteralProperties::SLModifier& slmod);
    bool subsumeMainPremise();
    bool collectMainPremiseInstance();
    bool subsumeSidePremise();
    bool collectSidePremiseInstance();
    bool demodulateSelectedLits();
    bool collectMainPremiseSubtermsAvailableForParamodulationInto();
    bool collectSidePremiseSubtermsAvailableForParamodulationInto();
  private:
    OPEN_BR_DECL(1);
    Statistics* _statistics;
    TemporaryClauseCollector* _clauseCollector;
    ForwardProcessing* _forwardProcessing;
    UnifIndex* _unifIndex;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;
    bool _noDuplicateRedexesFlag;
    bool _noInfBetweenRulesFlag;
    bool _useInputNonBasicness;
    long _clauseSizeLimit;
    Clause* _mainPremise;
    ulong _mainPremiseVarIndex;
    LiteralList* _allMainPremiseLiterals;
    LiteralList::Element* _firstNonselectedMainLit;
    ulong _numOfSelectedMainLits;
    bool _equalityResolutionApplicable;
    UnifTerm _mainSelectedLiteralMemory[VampireKernelConst::MaxClauseSize];
    UnifTerm* _freeMainSelectedLitMemory;
    BK::ExpandingStack<BK::GlobAlloc,UnifTerm*,BK::FastRelocation<UnifTerm*>,32UL> _mainSelectedLiterals; 
    BK::Array<BK::GlobAlloc,LiteralProperties,BK::FastRelocation<LiteralProperties>,32UL> _mainSelectedLitProperties;  
    SizeFormula _mainPremiseResidualSizeFormula;
    long _mainPremiseResidualSize;
    bool _mainSelectedLitDemodulationAttempted;
    bool _mainSelectedLitDemodulated;
    ulong _nextMainSelectedLitNum;
    UnifTerm* _currentMainLiteral;
    LiteralProperties _currentMainLitProperties;
    BK::Stack<UnifTerm*,VampireKernelConst::MaxTermSize> _eligibleSubtermPositions;
    SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
    SimplificationOrdering::CompiledConstraintForUnifiers* _currentCompiledConstraint;   
    UnifTerm* _eqArg1;
    UnifTerm* _eqArg2;
    bool _swapArguments;
    Clause* _sidePremise;    
    ulong _sidePremiseVarIndex;
    SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;
    IUMachine _retrievalOfUnifiable;
    UnifTerm* _currentParamodulationTargetTerm;
    Term _pseudoHeader;
    UnifIndex::IndexedObject* _currentCandidate;
    bool _noRulesAsSidePremises;
    ParamodulationMode _paramodulationMode;
    Continuation _continuation;
    TemporaryClause::Extras::Handle _dummyClause;
    TemporaryClauseCollector _dummyClauseCollector;
    CLOSE_BR_DECL(1);
  }; 
}; 
#endif
