/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "InferenceDispatcher.hpp"
#include "Clause.inl"
#include "SymbolInfo.hpp"
VK::InferenceDispatcher::InferenceDispatcher() : 
  binResFlag(false),
  OPEN_BR(1,"br 1") DCOMMA  
  _paramodulationFlag(false),
  _paramodulateFromVariables(false),
  _noInfBetweenRulesFlag(false),
  _binRes(&_ordResIndex),  
  _forwardSuperposition(&_ordResIndex), 
  _equalityFactoring(0L), 
  _backwardSuperposition(&_ordResIndex), 
  _chaining(&_ordResIndex),
  _query(0),
  _clauseCollector(0)
{
  CALL("constructor InferenceDispatcher()");
  unlock();
}; 
VK::InferenceDispatcher::~InferenceDispatcher() 
{
};
void VK::InferenceDispatcher::reset()
{
  CALL("reset()");
  BK::destroy(*this);
  BK::init(*this);
}; 
void VK::InferenceDispatcher::newQuery(Clause* cl) 
{ 
  CALL("newQuery(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(!cl->isAnswer());
  _query = cl; 
  if (builtInTransitivityLaw(_query))
    {
      _noMoreInferences = true;
    }
  else
    {
      _noMoreInferences = false;
      _continuation = FirstBinRes;
    }; 
}; 
void VK::InferenceDispatcher::abandonCurrentQuery()
{
  CALL("abandonCurrentQuery()");
  ASSERT(isLocked());
  _query->unprotect();
  unlock();
}; 
bool VK::InferenceDispatcher::nextInference()
{
  CALL("nextInference()");
  ASSERT(isLocked());
  if (_noMoreInferences) return false;
  switch (_continuation)
    { 
    case FirstBinRes:
      COP("FirstBinRes");
      if (!binResFlag) 
	goto first_forw_sup;   
      ASSERT(Unifier::current()->NoBoundVariables());
      if (!_binRes.Load(_query))
	goto first_chaining;  
      _continuation = NextBinRes;  
    case NextBinRes:
      COP("NextBinRes");
      if (_binRes.NextInference()) return true;
      Unifier::current()->UnbindAll(); 
    case FirstChaining:
      COP("FirstChaining");
    first_chaining:
      ASSERT(Unifier::current()->NoBoundVariables()); 
      if ((!_chainingTable) ||
	  (!_chaining.loadQuery(_query)))
	goto first_forw_sup; 
      _continuation = NextChaining;
    case NextChaining:
      COP("NextChaining");
      ASSERT(_chainingTable);
      if (_chaining.nextInference())
	{
	  return true;
	};
    case FirstForwSup:
    first_forw_sup:
    COP("FirstForwSup");
    ASSERT(Unifier::current()->NoBoundVariables());
    if (!_paramodulationFlag) return false; 
    if ((!_forwardSuperposition.Load(_query))
	|| (!_forwardSuperposition.NextInference()))
      goto first_eq_factor;
    _continuation = NextForwSup;
    return true; 
    case NextForwSup:
      COP("NextForwSup");
      ASSERT(_paramodulationFlag);
      if (_forwardSuperposition.NextInference()) return true;
      Unifier::current()->UnbindAll(); 
    case FirstEqFactor:
    first_eq_factor:
    COP("first_eq_factor");
    ASSERT(_paramodulationFlag);
    ASSERT(Unifier::current()->NoBoundVariables());
    if (!_equalityFactoring.Load(_query)) goto first_back_sup; 
    _continuation = NextEqFactor;            
    case NextEqFactor: 
      COP("NextEqFactor");
      ASSERT(_paramodulationFlag);
      if (_equalityFactoring.NextInference()) return true;
      Unifier::current()->UnbindAll();   
    case FirstBackSup:
    first_back_sup:
    COP("FirstBackSup");
    ASSERT(_paramodulationFlag);
    ASSERT(Unifier::current()->NoBoundVariables());
    if (!_backwardSuperposition.SetQuery(_query)) 
      {       
	return false; 
      };     
    _continuation = NextBackSup;      
    case NextBackSup:
      COP("NextBackSup");
      ASSERT(_paramodulationFlag);
      if (_backwardSuperposition.NextInference()) return true;
      Unifier::current()->UnbindAll();
      return false; 
    };
  ICP("ICP1"); 
  return false;   
}; 
void VK::InferenceDispatcher::makeActive(Clause* cl)
{
  CALL("makeActive(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(!cl->flag(Clause::FlagInInferenceDispatcherAsActive));
  if (!builtInTransitivityLaw(cl))
    {
      _ordResIndex.Integrate(cl);
    };
  cl->setFlag(Clause::FlagInInferenceDispatcherAsActive);  
  cl->unprotect();
  unlock();
}; 
void VK::InferenceDispatcher::remove(Clause* cl) 
{
  CALL("remove(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ASSERT(cl->flag(Clause::FlagInInferenceDispatcherAsActive));
  if (!builtInTransitivityLaw(cl))
    {
      _ordResIndex.Remove(cl); 
    };
  cl->clearFlag(Clause::FlagInInferenceDispatcherAsActive);   
  cl->unprotect();
  unlock();
}; 
inline
bool VK::InferenceDispatcher::builtInTransitivityLaw(const Clause* cl)
{
  CALL("builtInTransitivityLaw(const Clause* cl)");
  if (!_chainingTable) return false;
  ulong transPred;
  if (!cl->isTransitivityLaw(transPred) ||
      VampireKernelConst::isEquality(transPred))
    return false;
  return (_chainingTable->transitivityLaw(transPred) != 0);
}; 
