/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef INFERENCE_DISPATCHER_HPP
#define INFERENCE_DISPATCHER_HPP
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "UnifIndex.hpp"
#include "BinaryResolution.hpp"
#include "ForwardSuperposition.hpp"
#include "EqualityFactoring.hpp"
#include "BackwardSuperposition.hpp"
#include "Chaining.hpp"
#include "TemporaryClause.hpp"
namespace VK
{
  class Clause;
  class SymbolInfo;
  class TemporaryClauseCollector;
  class ForwardProcessing;
  class Statistics;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INFERENCE_DISPATCHER
 #define DEBUG_NAMESPACE "InferenceDispatcher"
#endif
#include "debugMacros.hpp"
namespace VK
{
class InferenceDispatcher
{
public:
  InferenceDispatcher();
  ~InferenceDispatcher();
  void reset();
  bool isLocked() const { return _isLocked; };
  void setStatistics(Statistics* stat)
  {
    _ordResIndex.setStatistics(stat);
    _binRes.setStatistics(stat);
    _forwardSuperposition.setStatistics(stat);
    _backwardSuperposition.setStatistics(stat);
    _chaining.setStatistics(stat);
  };
  void setClauseCollector(TemporaryClauseCollector* cc) 
  {
    _clauseCollector = cc;
    _binRes.setClauseCollector(cc);
    _forwardSuperposition.setClauseCollector(cc);
    _equalityFactoring.setClauseCollector(cc);
    _backwardSuperposition.setClauseCollector(cc);
    _chaining.setClauseCollector(cc);
  };
  void setForwardProcessing(ForwardProcessing* fp)
  {
    _binRes.setForwardProcessing(fp);
    _forwardSuperposition.setForwardProcessing(fp);
    _backwardSuperposition.setForwardProcessing(fp);
    _chaining.setForwardProcessing(fp);
  };
  void setParamodulationFlag(bool fl) 
    {
      _paramodulationFlag = fl;
      _ordResIndex.SetParamodulation(fl); 
    }; 
  void setParamodulateFromVariablesFlag(bool fl) 
  {
    _paramodulateFromVariables = fl;
    _ordResIndex.setParamodulateFromVariablesFlag(fl);
    _backwardSuperposition.setParamodulateFromVariablesFlag(fl);
  };
  void setSimBackSup(long degree)
    {
      _backwardSuperposition.setSimultaneous(degree);
    };
  void setForwSupNoDuplicateRedexes(bool fl)
  {
    _forwardSuperposition.setNoDuplicateRedexes(fl);
  };
  void setBackSupNoDuplicateRedexes(bool fl)
  {
    _ordResIndex.setBackSupNoDuplicateRedexes(fl);
  };
  void setUseInputNonBasicness(bool fl)
  {
    _ordResIndex.setUseInputNonBasicness(fl);
    _forwardSuperposition.setUseInputNonBasicness(fl);
  };
  void setChainingTable(const SymbolInfo* table)
  {
    _chainingTable = table;
    _chaining.setTransitivity(table);
    _ordResIndex.setChainingTable(table);
  };
  void setChainingThroughVariablesFlag(bool fl)
  {
    _chainingThroughVariables = fl;
    _chaining.setChainingThroughVariablesFlag(fl);
    _ordResIndex.setChainingThroughVariablesFlag(fl);
  };
  void setMainPremiseSubsumption(bool fl)
  {
    _binRes.setMainPremiseSubsumption(fl);
    _forwardSuperposition.setMainPremiseSubsumption(fl);
    _backwardSuperposition.setMainPremiseSubsumption(fl);
    _chaining.setMainPremiseSubsumption(fl);
  };
  void setSidePremiseSubsumption(bool fl)
  {
    _binRes.setSidePremiseSubsumption(fl);
    _forwardSuperposition.setSidePremiseSubsumption(fl);
    _backwardSuperposition.setSidePremiseSubsumption(fl);
    _chaining.setSidePremiseSubsumption(fl);
  };
  void setMainSelectedLitDemodulation(bool fl)
  {
    _binRes.setMainSelectedLitDemodulation(fl);
    _forwardSuperposition.setMainSelectedLitDemodulation(fl);
    _backwardSuperposition.setMainSelectedLitDemodulation(fl);
    _chaining.setMainSelectedLitDemodulation(fl);
  };
  void setSideSelectedLitDemodulation(bool fl)
  {
    _binRes.setSideSelectedLitDemodulation(fl);
    _forwardSuperposition.setSideSelectedLitDemodulation(fl);
    _backwardSuperposition.setSideSelectedLitDemodulation(fl);
    _chaining.setSideSelectedLitDemodulation(fl);
  };
  void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;
      _binRes.setNoInfBetweenRulesFlag(fl);
      _forwardSuperposition.setNoInfBetweenRulesFlag(fl);
      _backwardSuperposition.setNoInfBetweenRulesFlag(fl);
    };
  void setOrientEqualitiesInSupport(bool orientPos,bool orientNeg)
  {
    _ordResIndex.setOrientEqualitiesInSupport(orientPos,orientNeg);
  };
  void setClauseSizeLimit(long lim)
  {
    _binRes.setClauseSizeLimit(lim);
    _forwardSuperposition.setClauseSizeLimit(lim);
    _equalityFactoring.setClauseSizeLimit(lim);
    _backwardSuperposition.setClauseSizeLimit(lim);
    _chaining.setClauseSizeLimit(lim);
  };
  void newQuery(Clause* cl);
  void abandonCurrentQuery();
  bool nextInference();
  void makeActive(Clause* cl);
  void remove(Clause* cl);
 public:
  bool binResFlag;
 private:
  enum Continuation 
  { 
    FirstBinRes, 
    NextBinRes, 
    FirstChaining,
    NextChaining,
    FirstForwSup, 
    NextForwSup,
    FirstEqFactor,
    NextEqFactor,
    FirstBackSup,
    NextBackSup 
  };
private:
  bool builtInTransitivityLaw(const Clause* cl);
  void lock() { _isLocked = true; };
  void unlock() { _isLocked = false; };
private:
  OPEN_BR_DECL(1);
  bool _paramodulationFlag;
  bool _paramodulateFromVariables;
  bool _noInfBetweenRulesFlag;
  const SymbolInfo* _chainingTable;
  bool _chainingThroughVariables;
  bool _noMoreInferences;
  UnifIndex _ordResIndex;
  BinaryResolution _binRes;
  ForwardSuperposition _forwardSuperposition;
  EqualityFactoring _equalityFactoring;
  BackwardSuperposition _backwardSuperposition;
  Chaining _chaining;
  Clause* _query;
  TemporaryClauseCollector* _clauseCollector;
  Continuation _continuation;
  bool _isLocked;
  CLOSE_BR_DECL(1);
}; 
}; 
#endif
