/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef INSTANCE_RETRIEVAL_FOR_DEMODULATION_HPP
#define INSTANCE_RETRIEVAL_FOR_DEMODULATION_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif   
#include "jargon.hpp"
#include "VampireKernelConst.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "Term.hpp"
#include "GlobAlloc.hpp"
#include "PathIndexingForInstanceRetrieval.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INSTANCE_RETRIEVAL_FOR_DEMODULATION
 #define DEBUG_NAMESPACE "InstRet"
#endif
#include "debugMacros.hpp"
namespace VK
{
class InstRet
{ 
 public:
  class Substitution
    {
    public:
      class Instance;
    public:
      const Term& map(ulong var) const 
	{ 
	  return ((const PathIndexingForInstanceRetrieval*)this)->subst(var);
	};
      const Term* nativeVar(ulong varNum) const
	{ 
	  return (static_cast<const PathIndexingForInstanceRetrieval*>(static_cast<const void*>(this)))->nativeVar(varNum);
	};
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const
	{
	  return (static_cast<const PathIndexingForInstanceRetrieval*>(static_cast<const void*>(this)))->outputSubst(str);
	};
#endif      
    private:
      Substitution() {};
      ~Substitution() {};
      friend class InstRet;
    };
 public:
  InstRet() {};
  ~InstRet() {};
  void integrate(const Term* complexTerm);
  void remove(const Term* complexTerm);
  bool newQuery(const Term* query);
  bool nextMatch(const Term*& match);
  const Substitution* substitution() const
    {
      CALL("substitution() const");
      ASSERT(!_noMoreMatches);
      return static_cast<const Substitution*>(static_cast<const void*>(&_index));
    };
 private:
#ifdef DEBUG_NAMESPACE
  bool isValidMatch(const Term* term) const;
#endif
 private:
  PathIndexingForInstanceRetrieval _index;
  bool _noMoreMatches;
  DOP(const Term* _debugQuery);
}; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline 
  ostream& operator<<(ostream& str,const VK::InstRet::Substitution* subst)
  {
    return subst->output(str);
  };
};
#endif 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INSTANCE_RETRIEVAL_FOR_DEMODULATION
# define DEBUG_NAMESPACE "InstRet::Substitution::Instance"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class InstRet::Substitution::Instance
  {
  public:
    Instance() : 
      _dummyTerm0((Term::Arity)0,(Term::Functor)0,Signature::PositivePolarity)  
    {
      _dummyTerm1[0].MakeComplex((Term::Arity)1,(Term::Functor)0,Signature::PositivePolarity);
    };
    ~Instance() {};
    void reset(const Term& t,const InstRet::Substitution* subst);
    void reset(const Term* t,const InstRet::Substitution* subst);
    const Term& symbol() const;
    bool next();
    static 
    bool 
    equal(const Term& t1,const Term& t2,const InstRet::Substitution* subst);
    static 
    bool 
    equal(const Term* t1,const Term* t2,const InstRet::Substitution* subst);
  private:
    const InstRet::Substitution* _subst;
    Term::Iterator _topLevel;
    Term::Iterator _substIter;
    bool _onTop;
    Term _dummyTerm0;
    Term _dummyTerm1[2];
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INSTANCE_RETRIEVAL_FOR_DEMODULATION
#define DEBUG_NAMESPACE "InstRet::Substitution::Instance"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  void 
  InstRet::Substitution::Instance::reset(const Term& t,
					 const InstRet::Substitution* subst)
  {
    CALL("reset(const Term& t,const InstRet::Substitution* subst)");
    if (t.isVariable())
      {
	_subst = subst;
	if (subst->map(t.var()).isVariable())
	  {
	    _dummyTerm1[1] = subst->map(t.var());
	    _substIter.Reset(_dummyTerm1);
	    ALWAYS(_substIter.Next());
	  }
	else
	  {
	    ASSERT(subst->map(t.var()).IsReference());
	    _substIter.Reset(subst->map(t.var()).deref());
	  };
	_topLevel.Reset(&_dummyTerm0);
	NEVER(_topLevel.Next());
	_onTop = false;
      }
    else
      {
	ASSERT(t.IsReference());
	reset(t.deref(),subst);
      };
  }; 
  inline
  void 
  InstRet::Substitution::Instance::reset(const Term* t,
					 const InstRet::Substitution* subst)
  {
    CALL("reset(const Term* t,const InstRet::Substitution* subst)");
    ASSERT(t->isComplex());
    _subst = subst;
    _topLevel.Reset(t);
    _onTop = true;
  }; 
  inline const Term& InstRet::Substitution::Instance::symbol() const
  {
    CALL("symbol() const");
    if (_onTop) return _topLevel.symbolRef();
    return _substIter.symbolRef();
  }; 
  inline bool InstRet::Substitution::Instance::next()
  {
    CALL("next()");
    if (_onTop)
      {
      next_on_top:
	if (!_topLevel.Next()) 
	  {
	    return false;
	  };
	if (_topLevel.CurrentSymIsVar())
	  {
	    _onTop = false;
	    if (_subst->map(_topLevel.symbolRef().var()).isVariable())
	      {
		_dummyTerm1[1] = _subst->map(_topLevel.symbolRef().var());
		_substIter.Reset(_dummyTerm1);
		ALWAYS(_substIter.Next());
	      }
	    else
	      {
		ASSERT(_subst->map(_topLevel.symbolRef().var()).IsReference());
		_substIter.Reset(_subst->map(_topLevel.symbolRef().var()).deref());
	      };
	  };
	return true;
      }
    else 
      {
	if (_substIter.Next())
	  {
	    return true;
	  };
	_onTop = true;
	goto next_on_top;
      };
  }; 
  inline
  bool 
  InstRet::Substitution::Instance::equal(const Term& t1,
					 const Term& t2,
					 const InstRet::Substitution* subst)
  {
    CALL("equal(const Term& t1,..)");
    Instance iter1;
    Instance iter2;
    iter1.reset(t1,subst);
    iter2.reset(t2,subst);
    do
      {
	if (iter1.symbol() != iter2.symbol()) return false;
	iter2.next();
      }
    while (iter1.next());
    return true;
  }; 
  inline
  bool 
  InstRet::Substitution::Instance::equal(const Term* t1,
					 const Term* t2,
					 const InstRet::Substitution* subst)
  {
    CALL("equal(const Term* t1,..)");
    ASSERT(t1->isComplex() && t2->isComplex()); 
    Instance iter1;
    Instance iter2;
    iter1.reset(t1,subst);
    iter2.reset(t2,subst);
    do
      {
	if (iter1.symbol() != iter2.symbol()) return false;
	iter2.next();
      }
    while (iter1.next());
    return true;
  }; 
}; 
#endif
