/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */
#ifndef INSTANCE_WITH_UNIFIER_HPP
#define INSTANCE_WITH_UNIFIER_HPP
#ifndef NO_DEBUG_VIS 
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "Unifier.hpp"
#include "UnifTerm.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INSTANCE_WITH_UNIFIER
#define DEBUG_NAMESPACE "InstanceWithUnifier"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class InstanceWithUnifier
  {
  public:
    InstanceWithUnifier() 
    {
      dummyTerm[0].MakeComplex((Term::Arity)1,(Term::Functor)VampireKernelConst::ReservedPred1,Signature::PositivePolarity);	  
    };
    ~InstanceWithUnifier() {};
    void Reset(ulong index,const Term* t) 
    {
      CALL("Reset(ulong index,Term* t)");
      ASSERT(t->isComplex());
      top.Reset(t);
      var_offset = Unifier::variableOffset(index);
      _currentPositionIsInSubstitution = false;
      empty = false;
      curr_sym.MkFunc(top.symbolRef());
    };
    void Reset(ulong index,const Term& t)
    {
      CALL("Reset(ulong index,Term& t)");
      ASSERT(t.isVariable() || t.IsReference());
      dummyTerm[1] = t;
      Reset(index,dummyTerm);
      SkipSym();
      ASSERT(!Empty());
    }; 
    bool Empty() const { return empty; };
    const UnifTerm::Symbol& CurrSym() const { return curr_sym; };
    bool currentPositionIsInSubstitution() const { 
      return _currentPositionIsInSubstitution; 
    };
    const Term* CurrPosInBase() const { return top.CurrPos(); }; 
    void SkipTerm()
    {
      CALL("SkipTerm()");
      if (_currentPositionIsInSubstitution)
	{
	  word.SkipTerm();
	  if (word.Empty())
	    {
	      _currentPositionIsInSubstitution = false;
	      if (!top.Next()) return;
	      goto top_check_sym;
	    }
	  else
	    {
	      if (word.CurrVar())
		{
		  curr_sym.MkVar(word.CurrVar());
		}
	      else
		{
		  curr_sym = word.CurrSubterm()->symbol();
		};
	    };
	}
      else 
	{
	  if ((empty = !top.SkipTerm())) return;
	top_check_sym:
	  if (top.CurrentSymIsVar())
	    {
	      bool free_var;
	      void* bind = (Unifier::current()->variableBase(top.symbolRef().var()) + var_offset)->Unref(free_var);
	      if (free_var)
		{
		  curr_sym.MkVar((Unifier::Variable*)bind);
		}
	      else
		{
		  _currentPositionIsInSubstitution = true;
		  word.Reset((UnifTerm*)bind);
		  curr_sym = word.CurrSubterm()->symbol();
		};
	    }
	  else 
	    {
	      curr_sym.MkFunc(top.symbolRef());
	    };
	};
    };
    void SkipSym()
    {
      CALL("SkipSym()");
      bool free_var;
      void* bind;
      if (_currentPositionIsInSubstitution)
	{
	  word.SkipSym();
	  if (word.Empty())
	    {
	      _currentPositionIsInSubstitution = false;
	      goto top_next_sym;
	    }
	  else
	    {
	      if (word.CurrVar())
		{
		  curr_sym.MkVar(word.CurrVar());
		}
	      else
		{
		  curr_sym = word.CurrSubterm()->symbol();
		};
	    };
	}
      else 
	{
	top_next_sym:
	  if (top.Next())
	    {
	      if (top.CurrentSymIsVar())
		{
		  bind = (Unifier::current()->variableBase(top.symbolRef().var()) + var_offset)->Unref(free_var);
		  if (free_var)
		    {
		      curr_sym.MkVar((Unifier::Variable*)bind);
		    }
		  else
		    {
		      _currentPositionIsInSubstitution = true;
		      word.Reset((UnifTerm*)bind);
		      curr_sym = word.CurrSubterm()->symbol();
		    };
		}
	      else 
		{
		  curr_sym.MkFunc(top.symbolRef());
		};
	    }
	  else { empty = true; };
	};
    };
    static bool equal(ulong index1,const Term& t1,ulong index2,const Term& t2)
    {
      CALL("equal(ulong index1,const Term& t1,ulong index2,const Term& t2)");
      InstanceWithUnifier iter1;
      InstanceWithUnifier iter2;
      iter1.Reset(index1,t1);
      iter2.Reset(index2,t2);
      do
	{
	  if (iter1.CurrPosInBase() == iter2.CurrPosInBase()) { iter1.SkipTerm(); iter2.SkipTerm(); }
	  else  
	    if (iter1.CurrSym() == iter2.CurrSym()) { iter1.SkipSym(); iter2.SkipSym(); } 
	    else return false;             
	}
      while (!iter1.Empty());
      return true;
    }; 
    static bool equal(ulong index1,const Term* t1,ulong index2,const Term* t2)
    {
      CALL("equal(ulong index1,const Term* t1,ulong index2,const Term* t2)");
      ASSERT(t1->isComplex());
      ASSERT(t2->isComplex());
      InstanceWithUnifier iter1;
      InstanceWithUnifier iter2;
      iter1.Reset(index1,t1);
      iter2.Reset(index2,t2);
      do
	{
	  if (iter1.CurrPosInBase() == iter2.CurrPosInBase()) { iter1.SkipTerm(); iter2.SkipTerm(); }
	  else  
	    if (iter1.CurrSym() == iter2.CurrSym()) { iter1.SkipSym(); iter2.SkipSym(); } 
	    else return false;             
	}
      while (!iter1.Empty());
      return true;
    };
  private:
    unsigned var_offset;
    UnifTerm::Iter word;
    Term::Iterator top;
    bool _currentPositionIsInSubstitution;
    bool empty;
    UnifTerm::Symbol curr_sym;
    Term dummyTerm[2];
  }; 
}; 
#endif 
