/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LPO_HPP
#define LPO_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Array.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "SimplificationOrdering.hpp"
#include "Flatterm.hpp"
#include "Term.hpp"
#include "UnifTerm.hpp"
#include "Comparison.hpp"
#include "GlobAlloc.hpp"
#include "ObjectPoolWithMemory.hpp"
#include "LinearCombination.hpp"
#include "ForwardMatchingSubstitutionCache.hpp"
#include "InstanceRetrievalForDemodulation.hpp"
#include "LocalUnification.hpp"
#include "VariableSet.hpp"
#include "InstanceWithUnifier.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class LPO : public SimplificationOrdering
  {
  public:
    LPO();
    ~LPO();
    void* operator new(size_t);
    void operator delete(void* obj);
    BK::Comparison compare(const Flatterm* term1,
			   const WeightPolynomial& weight1,
			   const Flatterm* term2,
			   const WeightPolynomial& weight2);
    BK::Comparison compareAtoms(const Flatterm* lit1,
				const WeightPolynomial& weight1,
				const Flatterm* lit2,
				const WeightPolynomial& weight2);
    BK::Comparison compareOnCurrentUnifier(UnifTerm* term1,
					   UnifTerm* term2);
    BK::Comparison compareOnCurrentUnifier(const Term& term1,
					   const Term& term2,
					   ulong variableIndex);
    BK::Comparison compare(const Term* complexTerm1,
			   const Term* complexTerm2,
			   const InstRet::Substitution* subst);
    BK::Comparison compare(const Term& term1,
			   const Term& term2,
			   const ForwardMatchingSubstitutionCache* subst);
    SimplificationOrdering::SpecialisedComparison* reserveSpecialisedComparison();
    void releaseSpecialisedComparison(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintGreater(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintGreaterOrEqual(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintLess(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintLessOrEqual(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintNotEqual(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint* copy(SimplificationOrdering::StoredConstraint* constr);
    void releaseConstraint(SimplificationOrdering::StoredConstraint* constr);
    SimplificationOrdering::ConstraintCheckOnUnifier* reserveConstraintCheckOnUnifier();
    void releaseConstraintCheckOnUnifier(SimplificationOrdering::ConstraintCheckOnUnifier* ccu);
    SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* 
    reserveConstraintCheckOnForwardMatchingSubst();
    void 
    releaseConstraintCheckOnForwardMatchingSubst(SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* ccfms);
    SimplificationOrdering::CompiledConstraintForUnifiers* reserveCompiledConstraintForUnifiers();
    void releaseCompiledConstraintForUnifiers(SimplificationOrdering::CompiledConstraintForUnifiers* cc);
    SimplificationOrdering::CompiledConstraintForInstanceRetrieval* reserveCompiledConstraintForInstanceRetrieval();
    void releaseCompiledConstraintForInstanceRetrieval(SimplificationOrdering::CompiledConstraintForInstanceRetrieval* cc);
    BK::Comparison compareWithTransposition(const Flatterm* term);
#ifdef NO_DEBUG
  private:
#else
  public:
#endif
    class SpecialisedComparison;
    class ConstraintCheckOnUnifier;
    class ConstraintCheckOnForwardMatchingSubst;
    class StoredConstraint;
    class CompiledConstraintForUnifiers;
    class CompiledConstraintForInstanceRetrieval;
  private:
    BK::Comparison compare(const Flatterm* term1,
			   const Flatterm* term2);
    BK::Comparison compareComplex(const Flatterm* complexTerm1,
				  const Flatterm* complexTerm2);
    bool greater(const Flatterm* term1,
		 const Flatterm* term2);
    bool greaterOrEqual(const Flatterm* term1,
			const Flatterm* term2);
    bool complexGreater(const Flatterm* complexTerm1,
			const Flatterm* complexTerm2);
    bool complexGreaterOrEqual(const Flatterm* complexTerm1,
			       const Flatterm* complexTerm2);
    bool complexGreaterThanProperSubterms(const Flatterm* complexTerm1,
					  const Flatterm* complexTerm2);
    BK::Comparison cMA(const Flatterm* complexTerm1,
		       const Flatterm* args2,
		       const Flatterm* endOfArgs2);
    BK::Comparison cLMA(const Flatterm* complexTerm1,
			const Flatterm* complexTerm2,
			const Flatterm* args1,
			const Flatterm* endOfArgs1,
			const Flatterm* args2,
			const Flatterm* endOfArgs2);
    bool alpha(const Flatterm* args1,
	       const Flatterm* endOfArgs1,
	       const Flatterm* complexTerm2);
    bool groundWithAllPrecBelow(const Flatterm* complexTerm,
				long prec);
    bool containsEqOrGrPrec(const Flatterm* complexTerm,
			    long prec);
    BK::Comparison compareWithConst(const Flatterm* complexTerm,
				    long constPrec);
    BK::Comparison cMA(UnifTerm* complexTerm1,
		       UnifTerm* args2,
		       UnifTerm* endOfArgs2);
    BK::Comparison cMA(const Term* complexTerm1,
		       const Term* args2,
		       ulong variableIndex);
    BK::Comparison cMA(UnifTerm* complexTerm1,
		       const Term* args2,
		       ulong variableIndex);
   BK::Comparison cMA(const Term* complexTerm1,
		      UnifTerm* args2,
		      UnifTerm* endOfArgs2,
		      ulong variableIndex);
    BK::Comparison cLMA(UnifTerm* complexTerm1,
			UnifTerm* complexTerm2,
			UnifTerm* args1,
			UnifTerm* endOfArgs1,
			UnifTerm* args2,
			UnifTerm* endOfArgs2);
    BK::Comparison cLMA(const Term* complexTerm1,
			const Term* complexTerm2,
			const Term* args1,
			const Term* args2,
			ulong variableIndex);
    BK::Comparison cLMA(UnifTerm* complexTerm1,
			const Term* complexTerm2,
			UnifTerm* args1,
			UnifTerm* endOfArgs1,
			const Term* args2,
			ulong variableIndex);
    bool alpha(UnifTerm* args1,
	       UnifTerm* endOfArgs1,
	       UnifTerm* complexTerm2);
    bool alpha(const Term* args1,
	       const Term* complexTerm2,
	       ulong variableIndex);
    bool alpha(UnifTerm* args1,
	       UnifTerm* endOfArgs1,
	       const Term* complexTerm2,
	       ulong variableIndex);
    bool alpha(const Term* args1,
	       UnifTerm* complexTerm2,
	       ulong variableIndex);
    bool lexMAOrEqual(UnifTerm* complexTerm1,
		      UnifTerm* complexTerm2,
		      UnifTerm* args1,
		      UnifTerm* endOfArgs1,
		      UnifTerm* args2,
		      UnifTerm* endOfArgs2);
    bool lexMA(UnifTerm* complexTerm1,
	       UnifTerm* complexTerm2,
	       UnifTerm* args1,
	       UnifTerm* endOfArgs1,
	       UnifTerm* args2,
	       UnifTerm* endOfArgs2);
    bool majo(UnifTerm* complexTerm1,
	      UnifTerm* args2,
	      UnifTerm* endOfArgs2);
    bool termWithUnifierContainsVariable(UnifTerm* complexTerm,
					 Unifier::Variable* var);
    bool termWithUnifierContainsVariable(const Term* complexTerm,
					 ulong variableIndex,
					 Unifier::Variable* var);
    bool termWithUnifierContainsVariable(const Term& term,
					 ulong variableIndex,
					 Unifier::Variable* var);
    bool 
    complexGreaterOrEqualOnCurrentUnifier(UnifTerm* complexTerm1,
					  UnifTerm* complexTerm2);
    bool 
    complexGreaterOnCurrentUnifier(UnifTerm* term1,
				   UnifTerm* term2);
    bool 
    greaterOrEqualOnCurrentUnifier(UnifTerm* term1,
				   UnifTerm* term2);
    bool 
    greaterOnCurrentUnifier(UnifTerm* term1,
			    UnifTerm* term2);
    bool 
    complexGreaterThanProperSubtermsOnCurrentUnifier(UnifTerm* complexTerm1,
						     UnifTerm* complexTerm2);
    BK::Comparison compareComplexOnCurrentUnifier(const Term* complexTerm1,
						  const Term* complexTerm2,
						  ulong variableIndex);
    bool 
    complexGreaterThanProperSubtermsOnCurrentUnifier(const Term* complexTerm1,
						     const Term* complexTerm2,
						     ulong variableIndex);
    BK::Comparison compareComplexOnCurrentUnifier(UnifTerm* complexTerm1,
						  const Term* complexTerm2,
						  ulong variableIndex);
    bool 
    complexGreaterThanProperSubtermsOnCurrentUnifier(const Term* complexTerm1,
						     ulong variableIndex,
						     UnifTerm* complexTerm2);
    bool 
    complexGreaterThanProperSubtermsOnCurrentUnifier(UnifTerm* complexTerm1,
						     const Term* complexTerm2,
						     ulong variableIndex);
    BK::Comparison compareOnCurrentUnifier(UnifTerm* term1,
					   const Term* complexTerm2,
					   ulong variableIndex);
    BK::Comparison compareOnCurrentUnifier(UnifTerm* term1,
					   const Term& term2,
					   ulong variableIndex);
    bool 
    groundWithAllPrecBelowOnCurrentUnifier(UnifTerm* complexTerm,
					   long prec);
    bool 
    containsEqOrGrPrecOnCurrentUnifier(UnifTerm* complexTerm,long prec);
    BK::Comparison
    compareWithConstOnCurrentUnifier(UnifTerm* complexTerm,
				     long constPrec);
    bool 
    groundWithAllPrecBelowOnCurrentUnifier(const Term* complexTerm,
					   ulong variableIndex,
					   long prec);
    bool 
    containsEqOrGrPrecOnCurrentUnifier(const Term* complexTerm,
				       ulong variableIndex,
				       long prec);
    BK::Comparison
    compareWithConstOnCurrentUnifier(const Term* complexTerm,
				     ulong variableIndex,
				     long constPrec);
    bool complexGreater(const Term* complexTerm1,
			const InstRet::Substitution* subst,
			const Term* complexTerm2);
    bool complexGreaterThanProperSubterms(const Term* complexTerm1,
					  const InstRet::Substitution* subst,
					  const Term* complexTerm2);
    bool complexGreater(const InstRet::Substitution* subst,
			const Term* complexTerm1,
			const Term* complexTerm2);
    bool complexGreater(const Term* complexTerm1,
			const Term* complexTerm2,
			const InstRet::Substitution* subst);
    bool greater(const Term* complexTerm1,
		 const InstRet::Substitution* subst,
		 const Term& term2);
    bool greater(const Term& term1,
		 const InstRet::Substitution* subst,
		 const Term& term2);
    bool greater(const InstRet::Substitution* subst,
		 const Term* complexTerm1,
		 const Term& term2);
    bool greater(const InstRet::Substitution* subst,
		 const Term& term1,
		 const Term& term2);
    bool greater(const Term& term1,
		 const Term& term2,
		 const InstRet::Substitution* subst);
    bool greater(const Term* complexTerm1,
		 const Term& term2,
		 const InstRet::Substitution* subst);
    BK::Comparison cMA(const Term* complexTerm1,
		       const InstRet::Substitution* subst,
		       const Term* args2);
    BK::Comparison cMA(const InstRet::Substitution* subst,
		       const Term* complexTerm1,
		       const Term* args2);
    BK::Comparison cMA(const Term* complexTerm1,
		       const Term* args2,
		       const InstRet::Substitution* subst);
    BK::Comparison cLMA(const Term* complexTerm1,
			const InstRet::Substitution* subst,
			const Term* complexTerm2,
			const Term* args1,
			const Term* args2);
    BK::Comparison cLMA(const InstRet::Substitution* subst,
			const Term* complexTerm1,
			const Term* complexTerm2,
			const Term* args1,
			const Term* args2);
    bool lexMA(const Term* complexTerm1,
	       const InstRet::Substitution* subst,
	       const Term* complexTerm2,
	       const Term* args1,
	       const Term* args2);
    bool lexMA(const InstRet::Substitution* subst,
	       const Term* complexTerm1,
	       const Term* complexTerm2,
	       const Term* args1,
	       const Term* args2);
    bool lexMA(const Term* complexTerm1,
	       const Term* complexTerm2,
	       const InstRet::Substitution* subst,
	       const Term* args1,
	       const Term* args2);
    bool majo(const Term* complexTerm1,
	      const InstRet::Substitution* subst,
	      const Term* args2);
    bool majo(const InstRet::Substitution* subst,
	      const Term* complexTerm1,
	      const Term* args2);
    bool majo(const Term* complexTerm1,
	      const Term* args2,
	      const InstRet::Substitution* subst);
    bool alpha(const Term* args1,
	       const InstRet::Substitution* subst,
	       const Term* complexTerm2);
    bool alpha(const Term* args1,
	       const Term* complexTerm2,
	       const InstRet::Substitution* subst);
    bool alpha(const InstRet::Substitution* subst,
	       const Term* args1,
	       const Term* complexTerm2);
    BK::Comparison compare(const InstRet::Substitution* subst,
			   const Term* complexTerm1,
			   const Term& term2);
    BK::Comparison compare(const InstRet::Substitution* subst,
			   const Term* complexTerm1,
			   const Term* complexTerm2);
    BK::Comparison compare(const InstRet::Substitution* subst,
			   const Term& term1,
			   const Term* complexTerm2);
    BK::Comparison compare(const InstRet::Substitution* subst,
			   const Term& term1,
			   const Term& term2);
    BK::Comparison compare(const Term& term1,
			   const Term& term2,
			   const InstRet::Substitution* subst);
    BK::Comparison compare(const Term& term1,
			   const Term* complexTerm2,
			   const InstRet::Substitution* subst);
    BK::Comparison compare(const Term& term1,
			   const InstRet::Substitution* subst,
			   const Term& term2);
    bool 
    groundWithAllPrecBelow(const Term* complexTerm,
			   const InstRet::Substitution* subst,
			   long prec);
    bool 
    containsEqOrGrPrec(const Term* complexTerm,
		       const InstRet::Substitution* subst,
		       long prec);
    BK::Comparison
    compareWithConst(const Term* complexTerm,
		     const InstRet::Substitution* subst,
		     long constPrec);
    bool complexContainsVar(const Term* complexTerm,
			    const InstRet::Substitution* subst,
			    const Term& var);
    BK::Comparison compare(const Term* complexTerm1,const Term* complexTerm2);
    BK::Comparison compare(const Term& term1,const Term& term2);
    bool greater(const Term* complexTerm1,const Term* complexTerm2); 
    bool greater(const Term* complexTerm1,const Term& term2); 
    bool greater(const Term& term1,const Term& term2); 
    BK::Comparison cMA(const Term* complexTerm1,
		       const Term* args2);
    BK::Comparison cLMA(const Term* complexTerm1,
			const Term* complexTerm2,
			const Term* args1,
			const Term* args2);
    bool alpha(const Term* args1,
	       const Term* complexTerm2);
    bool lexMA(const Term* complexTerm1,
	       const Term* complexTerm2,
	       const Term* args1,
	       const Term* args2);
    bool majo(const Term* complexTerm1,const Term* args2);
    bool 
    groundWithAllPrecBelow(const Term* complexTerm,
			   long prec);
    bool 
    containsEqOrGrPrec(const Term* complexTerm,
		       long prec);
    BK::Comparison
    compareWithConst(const Term* complexTerm,
		     long constPrec);
    BK::Comparison compare(const Term& term1,
			   const ForwardMatchingSubstitutionCache* subst,
			   const Flatterm* term2);
    bool complexContainsVar(const Term* complexTerm,
			    const ForwardMatchingSubstitutionCache* subst,
			    const Term& var);
    bool
    complexGreaterThanProperSubterms(const Term* complexTerm1,
				     const Term* complexTerm2,
				     const ForwardMatchingSubstitutionCache* subst);
    BK::Comparison cMA(const Term* complexTerm1,
		       const Term* args2,
		       const ForwardMatchingSubstitutionCache* subst);
    BK::Comparison cMA(const Term* complexTerm1,
		       const ForwardMatchingSubstitutionCache* subst,
		       const Flatterm* args2,
		       const Flatterm* endOfArgs2);
    BK::Comparison cMA(const Flatterm* complexTerm1,
		       const Term* args2,
		       const ForwardMatchingSubstitutionCache* subst);
    BK::Comparison cLMA(const Term* complexTerm1,
			const Term* complexTerm2,
			const Term* args1,
			const Term* args2,
			const ForwardMatchingSubstitutionCache* subst);
    BK::Comparison cLMA(const Term* complexTerm1,
			const ForwardMatchingSubstitutionCache* subst,
			const Flatterm* complexTerm2,
			const Term* args1,
			const Flatterm* args2,
			const Flatterm* endOfArgs2);
    bool alpha(const Term* args1,
	       const Term* complexTerm2,
	       const ForwardMatchingSubstitutionCache* subst);
    bool alpha(const Flatterm* args1,
	       const Flatterm* endOfArgs1,
	       const Term* complexTerm2,
	       const ForwardMatchingSubstitutionCache* subst);
    bool alpha(const Term* args1,
	       const ForwardMatchingSubstitutionCache* subst,
	       const Flatterm* complexTerm2);
    bool 
    groundWithAllPrecBelow(const Term* complexTerm,
			   const ForwardMatchingSubstitutionCache* subst,
			   long prec);
    bool 
    containsEqOrGrPrec(const Term* complexTerm,
		       const ForwardMatchingSubstitutionCache* subst,
		       long prec);
    BK::Comparison
    compareWithConst(const Term* complexTerm,
		     const ForwardMatchingSubstitutionCache* subst,
		     long constPrec);
  private:
    SpecialisedComparison* _specialisedComparisonPool;
    ConstraintCheckOnUnifier* _constraintCheckOnUnifierPool;
    ConstraintCheckOnForwardMatchingSubst* _constraintCheckOnForwardMatchingSubstPool;
    CompiledConstraintForUnifiers* _compiledConstraintOnUnifiersPool;
    CompiledConstraintForInstanceRetrieval* _compiledConstraintForInstanceRetrievalPool;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,UnifTerm::Iter> _prefixIterPool;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,InstanceWithUnifier> _stTermWithSubstPool;
    ulong _numberOfCompiledConstraintsForInstanceRetrieval;
    ulong _numberOfConstraintChecksForInstanceRetrieval;
    ulong _numberOfCompiledConstraintsOnUnifiers;
    ulong _numberOfConstraintChecksOnUnifiers;
    friend class SpecialisedComparison;
    friend class ConstraintCheckOnUnifier;
    friend class ConstraintCheckOnForwardMatchingSubst;
    friend class CompiledConstraintForUnifiers;
    friend class CompiledConstraintForInstanceRetrieval;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO::SpecialisedComparison"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LPO::SpecialisedComparison :
    public SimplificationOrdering::SpecialisedComparison
  {
  public:
    void load(const Term& lhs,const Term& rhs);
    bool lhsVarInstMayOrder();
    bool rhsVarInstMayOrder();
    void specialiseGreater();
    void specialiseGreaterOrEqual();
    void specialiseLess();
    void specialiseLessOrEqual();
    void specialiseNotEqual();
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
#endif
  private:
    SpecialisedComparison(LPO* ord) : 
      _ordering(ord),
      _nextInPool(0),
      _constraintGreater(0),
      _constraintGreaterOrEqual(0),
      _constraintLess(0),
      _constraintLessOrEqual(0),
      _constraintNotEqual(0)
    {
    };
    ~SpecialisedComparison();
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(SpecialisedComparison));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(SpecialisedComparison));
    };
    SpecialisedComparison* getNextInPool() { return _nextInPool; };
    void setNextInPool(LPO::SpecialisedComparison* sc) { _nextInPool = sc; };
    void specialise();
    LPO::StoredConstraint* storedConstraintGreater();
    LPO::StoredConstraint* storedConstraintGreaterOrEqual();
    LPO::StoredConstraint* storedConstraintLess();
    LPO::StoredConstraint* storedConstraintLessOrEqual();
    LPO::StoredConstraint* storedConstraintNotEqual();
  private:
    LPO* _ordering;
    Term _lhs;
    Term _rhs;
    VariableSet _lhsVars;
    VariableSet _rhsVars;
    bool _specialised;
    SpecialisedComparison* _nextInPool;
    LPO::StoredConstraint* _constraintGreater;
    LPO::StoredConstraint* _constraintGreaterOrEqual;
    LPO::StoredConstraint* _constraintLess;
    LPO::StoredConstraint* _constraintLessOrEqual;
    LPO::StoredConstraint* _constraintNotEqual;
    friend class LPO;
    friend class LPO::CompiledConstraintForUnifiers;
    friend class LPO::CompiledConstraintForInstanceRetrieval;
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline
  ostream& operator<<(ostream& str,
		      const VK::LPO::SpecialisedComparison& sc)
  {
    return sc.output(str);
  };
  inline
  ostream& operator<<(ostream& str,
		      const VK::LPO::SpecialisedComparison* sc)
  {
    return sc->output(str);
  };
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#  define DEBUG_NAMESPACE "LPO::ConstraintCheckOnUnifier"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LPO::ConstraintCheckOnUnifier :
    public SimplificationOrdering::ConstraintCheckOnUnifier
  {
  public:
    bool holds(const SimplificationOrdering::StoredConstraint* constr,
	       ulong variableIndex);
  private:
    ConstraintCheckOnUnifier(LPO* ord) : 
      _ordering(ord),
      _nextInPool(0) 
    {
    };
    ~ConstraintCheckOnUnifier();   
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(ConstraintCheckOnUnifier));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(ConstraintCheckOnUnifier));
    };  
    ConstraintCheckOnUnifier* getNextInPool() { return _nextInPool; };
    void setNextInPool(LPO::ConstraintCheckOnUnifier* ccu) { _nextInPool = ccu; };
    bool holdsGreater(const StoredConstraint* constr);
    bool holdsGreaterOrEqual(const StoredConstraint* constr);
    bool holdsLess(const StoredConstraint* constr);
    bool holdsLessOrEqual(const StoredConstraint* constr);
    bool holdsNotEqual(const StoredConstraint* constr);
  private:
    ConstraintCheckOnUnifier();
  private:
    LPO* _ordering;
    ConstraintCheckOnUnifier* _nextInPool;
    ulong _currentVariableIndex;
    InstanceWithUnifier _lhsIter;
    InstanceWithUnifier _rhsIter;
    friend class LPO;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#  define DEBUG_NAMESPACE "LPO::ConstraintCheckOnForwardMatchingSubst"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LPO::ConstraintCheckOnForwardMatchingSubst :
    public SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst
  {
  public:
    bool holds(const SimplificationOrdering::StoredConstraint* constr,
	       const ForwardMatchingSubstitutionCache* subst);
  private:
    ConstraintCheckOnForwardMatchingSubst(LPO* ord) : 
      _ordering(ord),
      _nextInPool(0)
    {
    };
    ~ConstraintCheckOnForwardMatchingSubst();
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(ConstraintCheckOnForwardMatchingSubst));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(ConstraintCheckOnForwardMatchingSubst));
    };  
    ConstraintCheckOnForwardMatchingSubst* getNextInPool() { return _nextInPool; };
    void setNextInPool(LPO::ConstraintCheckOnForwardMatchingSubst* ccfms) 
    { 
      _nextInPool = ccfms; 
    };
    bool holdsGreater(const StoredConstraint* constr);
    bool holdsGreaterOrEqual(const StoredConstraint* constr);
    bool holdsLess(const StoredConstraint* constr);
    bool holdsLessOrEqual(const StoredConstraint* constr);
    bool holdsNotEqual(const StoredConstraint* constr);
  private:
    ConstraintCheckOnForwardMatchingSubst();
  private:
    LPO* _ordering;
    ConstraintCheckOnForwardMatchingSubst* _nextInPool;
    const ForwardMatchingSubstitutionCache* _subst;
    ForwardMatchingSubstitutionCache::Instance _lhsInstance;
    ForwardMatchingSubstitutionCache::Instance _rhsInstance;
    friend class LPO;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO::StoredConstraint"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LPO::StoredConstraint
  {
  public:
    enum TargetRelation { Greater, GreaterOrEqual, Less, LessOrEqual, NotEqual };
  public:
    StoredConstraint(TargetRelation tr,
		     const Term& lhs,
		     const Term& rhs) :
      _referenceCounter(0L),
      _targetRelation(tr),
      _lhs(lhs),
      _rhs(rhs)
    {
      CALL("constructor StoredConstraint(TargetRelation tr,..)");
    };
    ~StoredConstraint() 
    {
      CALL("destructor ~StoredConstraint()");
      ASSERT(!referenceCounter());
    };
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(StoredConstraint));
    };  
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(StoredConstraint));
    };
    const long& referenceCounter() const { return _referenceCounter; };
    void incReferenceCounter() 
    {
      CALL("incReferenceCounter()");
      ++_referenceCounter; 
    };
    void decReferenceCounter() 
    {
      CALL("decReferenceCounter()");
      --_referenceCounter;
      ASSERT(_referenceCounter >= 0L);
    };
    const TargetRelation& targetRelation() const { return _targetRelation; };
    const Term& lhs() const { return _lhs; };
    const Term& rhs() const { return _rhs; };
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
      switch (targetRelation())
	{
	case Greater:        str << "Greater        "; break;
	case GreaterOrEqual: str << "GreaterOrEqual "; break;
	case Less:           str << "Less           "; break;
	case LessOrEqual:    str << "LessOrEqual    "; break;
	case NotEqual:       str << "NotEqual       "; break;
	default:             str << "?????????????? ";
	};
      str << lhs() << " vs " << rhs();
      return str;
    };
#endif
  private:
    StoredConstraint();
    StoredConstraint(const StoredConstraint&);
    StoredConstraint& operator=(const StoredConstraint&);
  private:
    long _referenceCounter;
    TargetRelation _targetRelation;
    Term _lhs;
    Term _rhs;
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline
  ostream& operator<<(ostream& str,
		      const VK::LPO::StoredConstraint& constr)
  {
    return constr.output(str);
  };
  inline
  ostream& operator<<(ostream& str,
		      const VK::LPO::StoredConstraint* constr)
  {
    return constr->output(str);
  };
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO::CompiledConstraintForUnifiers"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LPO::CompiledConstraintForUnifiers :
    public SimplificationOrdering::CompiledConstraintForUnifiers
  {
  public:
    CompiledConstraintForUnifiers(LPO* ord) :
      _ordering(ord),
      _nextInPool(0)
    {
    };
    ~CompiledConstraintForUnifiers();  
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(CompiledConstraintForUnifiers));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(CompiledConstraintForUnifiers));
    }; 
    CompiledConstraintForUnifiers* getNextInPool() const { return _nextInPool; };
    void setNextInPool(CompiledConstraintForUnifiers* cc) { _nextInPool = cc; };
    void load(UnifTerm* lhs,UnifTerm* rhs);
    bool lhsVarInstMayOrder();
    bool rhsVarInstMayOrder();
    void compileGreaterOrEqual();
    void compileGreater();
    bool holdsOnCurrentUnifier();
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
#endif
  private:
    enum TargetRelation
    {
      GreaterOrEqual,
      Greater
    };
  private:
    CompiledConstraintForUnifiers();
  private:
    LPO* _ordering;
    LPO::CompiledConstraintForUnifiers* _nextInPool;
    TargetRelation _targetRelation;
    UnifTerm* _lhs;
    UnifTerm* _rhs;
    UnifTerm::Iter _lhsIter;
    UnifTerm::Iter _rhsIter;
    VariableSet _lhsVars;
    VariableSet _rhsVars;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO::CompiledConstraintForInstanceRetrieval"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LPO::CompiledConstraintForInstanceRetrieval :
    public SimplificationOrdering::CompiledConstraintForInstanceRetrieval
  {
  public:
    CompiledConstraintForInstanceRetrieval(LPO* ord) :
      _ordering(ord),
      _nextInPool(0)
    {
    };
    ~CompiledConstraintForInstanceRetrieval();  
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(CompiledConstraintForInstanceRetrieval));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(CompiledConstraintForInstanceRetrieval));
    }; 
    CompiledConstraintForInstanceRetrieval* getNextInPool() const { return _nextInPool; };
    void setNextInPool(CompiledConstraintForInstanceRetrieval* cc) { _nextInPool = cc; };
    void loadGreater(const Term* lhs,const Term* rhs);
    void loadNotEqual(const Term* lhs,const Term* rhs);
    bool lhsVarInstMayOrder();
    bool rhsVarInstMayOrder();
    void compile(const InstRet::Substitution* subst);
    bool holds();
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
#endif
  private:
    enum TargetRelation
    {
      Greater,
      NotEqual
    };
  private:
    CompiledConstraintForInstanceRetrieval();
  private:
    LPO* _ordering;
    LPO::CompiledConstraintForInstanceRetrieval* _nextInPool;
    TargetRelation _targetRelation;
    const InstRet::Substitution* _subst;
    Term _lhs;
    Term _rhs;
    VariableSet _lhsVars;
    VariableSet _rhsVars;
  }; 
}; 
#endif
