/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LITERAL_PROPERTIES_HPP
#  define LITERAL_PROPERTIES_HPP
#include <iostream>
#include <climits>                          
#include "jargon.hpp"
#include "Comparison.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernel.hpp"
using namespace std;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_PROPERTIES
#  define DEBUG_NAMESPACE "LiteralProperties"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class LiteralProperties
  {
  private:
    enum 
    {
      _SizeOfULong = sizeof(ulong)*CHAR_BIT,
      _SizeOfSelectionLayer = 6,
      _SizeOfDeploymentMode = 2,
      _SizeOfSLModifier = 2,
      _SizeOfEqualityMode = 2,
      _SizeOfEqualityOrientationMode = 2,
      _SizeOfFlagSelected = 1,
      _SizeOfSelectabilityCategory = 2,
      _SizeOfReserved = 
      _SizeOfULong -
      (_SizeOfSelectionLayer + 
       _SizeOfDeploymentMode + 
       _SizeOfSLModifier +
       _SizeOfFlagSelected +
       _SizeOfSelectabilityCategory)
    };
    enum
    {
      _BeforeSelectionLayer = 0,
      _AfterSelectionLayer = _SizeOfULong - _SizeOfSelectionLayer,
      _WithoutSelectionLayer = _BeforeSelectionLayer + _AfterSelectionLayer,
      _BeforeDeploymentMode = _SizeOfSelectionLayer,
      _AfterDeploymentMode = _SizeOfULong - (_BeforeDeploymentMode + _SizeOfDeploymentMode),
      _WithoutDeploymentMode = _BeforeDeploymentMode + _AfterDeploymentMode,
      _BeforeSLModifier = _BeforeDeploymentMode + _SizeOfDeploymentMode,
      _AfterSLModifier = _SizeOfULong - (_BeforeSLModifier + _SizeOfSLModifier),
      _WithoutSLModifier = _BeforeSLModifier + _AfterSLModifier,
      _BeforeEqualityMode = _BeforeSLModifier + _SizeOfSLModifier,
      _AfterEqualityMode = _SizeOfULong - (_BeforeEqualityMode + _SizeOfEqualityMode),
      _WithoutEqualityMode = _BeforeEqualityMode + _AfterEqualityMode,
      _BeforeEqualityOrientationMode = _BeforeEqualityMode + _SizeOfEqualityMode,
      _AfterEqualityOrientationMode = _SizeOfULong - (_BeforeEqualityOrientationMode + _SizeOfEqualityOrientationMode),
      _WithoutEqualityOrientationMode = _BeforeEqualityOrientationMode + _AfterEqualityOrientationMode,
      _BeforeFlagSelected = _BeforeEqualityOrientationMode + _SizeOfEqualityOrientationMode,
      _AfterFlagSelected = _SizeOfULong - (_BeforeFlagSelected + _SizeOfFlagSelected),
      _WithoutFlagSelected = _BeforeFlagSelected + _AfterFlagSelected,
      _BeforeSelectabilityCategory = _SizeOfULong - _SizeOfSelectabilityCategory,
      _AfterSelectabilityCategory = 0,
      _WithoutSelectabilityCategory = _BeforeSelectabilityCategory + _AfterSelectabilityCategory
    };
  public:
    enum DeploymentMode
    {
      DefaultDeploymentMode = 0UL,
      RuleHead = 1UL,    
      Subquery = 2UL
    };
    enum SLModifier 
    {
      DefaultSLModifier = 0UL,
      PostponeSelection = 1UL,
      AccelerateSelection = 2UL
    };
    enum EqualityMode
    {
      DefaultEqualityMode = 0UL,
      UseEqForResolution = 1UL,
      UseEqForResolutionOnly = 2UL
    };
    enum EqualityOrientationMode 
    {
      DefaultEqualityOrientation = 0UL,
      GivenEqualityOrientation = 1UL,
      ReverseGivenEqualityOrientation = 2UL, 
      ProhibitEqualityOrientation = 3UL
    };
    enum SelectabilityCategory
    {
      DefaultSelectability = 0UL,
      LowSelectability = 1UL,
      HighSelectability = 2UL
    };
    enum 
    {
      MaximalSelectionLayer = 63UL
    };
  public:
    LiteralProperties() : _data(0UL) {};
    explicit LiteralProperties(ulong data) : _data(data) {};
    void makeDefault() { _data = 0UL; };
    void make(ulong data) { _data = data; };
    ulong content() const { return _data; };
    void setSelectionLayer(ulong l);
    void setDeploymentMode(const DeploymentMode& m);
    void resetDeploymentMode() { setDeploymentMode(DefaultDeploymentMode); };
    void makeRuleHead() { setDeploymentMode(RuleHead); };
    void makeSubquery() { setDeploymentMode(Subquery); };
    void setEqualityMode(const EqualityMode& m);
    void setEqualityOrientationMode(const EqualityOrientationMode& m);
    void setDefaultEqualityOrientationMode() {
      setEqualityOrientationMode(DefaultEqualityOrientation);
    };
    void reverseEqualityOrientationMode();
    void setSelectionLayerModifier(const SLModifier& m);
    void setFlagSelected(bool fl);
    void makeSelected();
    void makeUnselected();
    void setSelectabilityCategory(const SelectabilityCategory& c);
    bool isDefault() const 
    { 
      return 
	(_data == 0UL) || (_data == (1UL << _AfterFlagSelected)); 
    };
    ulong selectionLayer() const;
    DeploymentMode deploymentMode() const;
    EqualityMode equalityMode() const;
    EqualityOrientationMode equalityOrientationMode() const;
    bool isOrientableEq() const {   
      return 
	equalityOrientationMode() != ProhibitEqualityOrientation;
    }; 
    bool isReverseOrientedEq() const {
      return 
	equalityOrientationMode() == ReverseGivenEqualityOrientation;
    };
    bool isPreorientedEq() const { 
      return 
	equalityOrientationMode() == GivenEqualityOrientation ||
	equalityOrientationMode() == ReverseGivenEqualityOrientation;
    };
    bool useEqForResolution() const;
    bool useEqForParamodulationFrom() const;
    SLModifier selectionLayerModifier() const;
    bool isSelected() const;
    SelectabilityCategory selectabilityCategory() const;
    void inheritFrom(const LiteralProperties& prop);
    void inheritFrom(const LiteralProperties& prop,
		     const SLModifier& slmod);
    void factoriseWith(const LiteralProperties& prop,bool symmetryUsed);
    ostream& output(ostream& str) const;
  private:
    BK::Comparison compareSelectabilityForFactoring(const LiteralProperties& prop) const;
  private: 
    ulong _data;
  }; 
}; 
namespace std
{
  inline ostream& operator<<(ostream& str,const VK::LiteralProperties& prop)
  {
    return prop.output(str);
  }; 
};
#endif
