/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LITERAL_PROPERTIES_INL
#  define LITERAL_PROPERTIES_INL
#include "LiteralProperties.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_PROPERTIES
#  define DEBUG_NAMESPACE "LiteralProperties"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  inline
  void LiteralProperties::setSelectionLayer(ulong l)
  {
    CALL("setSelectionLayer(ulong l)");
    if (l > MaximalSelectionLayer) l = MaximalSelectionLayer;
    _data = 
      (_data << (_SizeOfSelectionLayer + _BeforeSelectionLayer)) >>
      (_SizeOfSelectionLayer + _BeforeSelectionLayer);
    _data |= (l << _AfterSelectionLayer);
  }; 
  inline
  void LiteralProperties::setDeploymentMode(const DeploymentMode& m)
  {
    CALL("setDeploymentMode(const DeploymentMode& m)");
    ASSERT(m == DefaultDeploymentMode ||
	   m == RuleHead ||
	   m == Subquery);
    ulong before = 
      (_data >> (_SizeOfDeploymentMode + _AfterDeploymentMode)) <<
      (_SizeOfDeploymentMode + _AfterDeploymentMode);
    _data = 
      (_data << (_SizeOfDeploymentMode + _BeforeDeploymentMode)) >>
      (_SizeOfDeploymentMode + _BeforeDeploymentMode);
    _data |= before;
    _data |= (m << _AfterDeploymentMode);
  }; 
  inline
  void LiteralProperties::setEqualityMode(const EqualityMode& m)
  {
    CALL("setEqualityMode(const EqualityMode& m)");
    ASSERT(m == DefaultEqualityMode ||
	   m == UseEqForResolution ||
	   m == UseEqForResolutionOnly);
    ulong before = 
      (_data >> (_SizeOfEqualityMode + _AfterEqualityMode)) <<
      (_SizeOfEqualityMode + _AfterEqualityMode);
    _data = 
      (_data << (_SizeOfEqualityMode + _BeforeEqualityMode)) >>
      (_SizeOfEqualityMode + _BeforeEqualityMode);
    _data |= before;
    _data |= (m << _AfterEqualityMode);
  }; 
  inline
  void 
  LiteralProperties::setEqualityOrientationMode(const EqualityOrientationMode& m)
  {
    CALL("setEqualityOrientationMode(const EqualityOrientationMode& m)");
    ASSERT(m == DefaultEqualityOrientation ||
	   m == GivenEqualityOrientation ||
	   m == ReverseGivenEqualityOrientation ||
	   m == ProhibitEqualityOrientation);    
    ulong before = 
      (_data >> (_SizeOfEqualityOrientationMode + _AfterEqualityOrientationMode)) <<
      (_SizeOfEqualityOrientationMode + _AfterEqualityOrientationMode);
    _data = 
      (_data << (_SizeOfEqualityOrientationMode + _BeforeEqualityOrientationMode)) >>
      (_SizeOfEqualityOrientationMode + _BeforeEqualityOrientationMode);
    _data |= before;
    _data |= (m << _AfterEqualityOrientationMode);
  }; 
  inline
  void LiteralProperties::reverseEqualityOrientationMode()
  {
    CALL("reverseEqualityOrientationMode()");
    if (equalityOrientationMode() == GivenEqualityOrientation)
      {
	setEqualityOrientationMode(ReverseGivenEqualityOrientation);
      }
    else 
      if (equalityOrientationMode() == ReverseGivenEqualityOrientation)
	setEqualityOrientationMode(GivenEqualityOrientation);
  }; 
  inline
  void LiteralProperties::setSelectionLayerModifier(const SLModifier& m)
  {
    CALL("setSelectionLayerModifier(const SLModifier& m)");
    ASSERT(m == DefaultSLModifier ||
	   m == AccelerateSelection ||
	   m == PostponeSelection);
    ulong before = 
      (_data >> (_SizeOfSLModifier + _AfterSLModifier)) <<
      (_SizeOfSLModifier + _AfterSLModifier);
    _data = 
      (_data << (_SizeOfSLModifier + _BeforeSLModifier)) >>
      (_SizeOfSLModifier + _BeforeSLModifier);
    _data |= before;
    _data |= (m << _AfterSLModifier);
    ASSERT(selectionLayerModifier() == m);
  }; 
  inline
  void LiteralProperties::setFlagSelected(bool fl)
  {
    CALL("setFlagSelected(bool fl)");
    ulong before = 
      (_data >> (_SizeOfFlagSelected + _AfterFlagSelected)) <<
      (_SizeOfFlagSelected + _AfterFlagSelected);
    _data = 
      (_data << (_SizeOfFlagSelected + _BeforeFlagSelected)) >>
      (_SizeOfFlagSelected + _BeforeFlagSelected);
    _data |= before;
    _data |= ((fl) ? (1UL << _AfterFlagSelected) : 0UL);
  }; 
  inline
  void LiteralProperties::makeSelected()
  {
    CALL("makeSelected()");
    setFlagSelected(true);
  }; 
  inline
  void LiteralProperties::makeUnselected()
  {
    CALL("makeUnselected()");
    setFlagSelected(false);
  }; 
  inline
  void LiteralProperties::setSelectabilityCategory(const SelectabilityCategory& c)
  {
    CALL("setSelectabilityCategory(const SelectabilityCategory& c)");
    ASSERT(c == DefaultSelectability ||
	   c == LowSelectability ||
	   c == HighSelectability);
    ((_data >>= _SizeOfSelectabilityCategory) <<= 
     _SizeOfSelectabilityCategory) |= c;
  }; 
  inline
  ulong LiteralProperties::selectionLayer() const
  {
    CALL("selectionLayer() const");
    return (_data << _BeforeSelectionLayer) >> _WithoutSelectionLayer;
  }; 
  inline
  LiteralProperties::DeploymentMode LiteralProperties::deploymentMode() const
  {
    CALL("deploymentMode() const");
    ulong res = (_data << _BeforeDeploymentMode) >> _WithoutDeploymentMode;
    ASSERT(res == DefaultDeploymentMode ||
	   res == RuleHead ||
	   res == Subquery);
    return static_cast<DeploymentMode>(res);
  }; 
  inline
  LiteralProperties::EqualityMode LiteralProperties::equalityMode() const
  {
    CALL("equalityMode() const");
    ulong res = (_data << _BeforeEqualityMode) >> _WithoutEqualityMode;
    ASSERT(res == DefaultEqualityMode ||
	   res == UseEqForResolution ||
	   res == UseEqForResolutionOnly);
    return static_cast<EqualityMode>(res);
  }; 
  inline
  LiteralProperties::EqualityOrientationMode 
  LiteralProperties::equalityOrientationMode() const
  {
    CALL("equalityOrientationMode() const");
    ulong res = (_data << _BeforeEqualityOrientationMode) >> _WithoutEqualityOrientationMode;
    ASSERT(res == DefaultEqualityOrientation ||
	   res == GivenEqualityOrientation ||
	   res == ReverseGivenEqualityOrientation ||
	   res == ProhibitEqualityOrientation);    
    return static_cast<EqualityOrientationMode>(res);
  }; 
  inline
  bool LiteralProperties::useEqForResolution() const
  {
    CALL("useEqForResolution() const");
    return equalityMode() != DefaultEqualityMode;
  };
  inline
  bool LiteralProperties::useEqForParamodulationFrom() const
  {
    CALL("useEqForParamodulationFrom() const");
    return equalityMode() != UseEqForResolutionOnly;
  };
  inline
  LiteralProperties::SLModifier LiteralProperties::selectionLayerModifier() const
  {
    CALL("selectionLayerModifier() const");
    ulong res = (_data << _BeforeSLModifier) >> _WithoutSLModifier;
    ASSERT(res == DefaultSLModifier ||
	   res == AccelerateSelection ||
	   res == PostponeSelection);
    return static_cast<SLModifier>(res);
  }; 
  inline
  bool LiteralProperties::isSelected() const
  {
    CALL("isSelected() const");
    return 
      ((_data << _BeforeFlagSelected) >> _WithoutFlagSelected) != 0UL;
  }; 
  inline
  LiteralProperties::SelectabilityCategory
  LiteralProperties::selectabilityCategory() const
  {
    CALL("selectabilityCategory() const");
    ulong res = 
      (_data << _BeforeSelectabilityCategory) >> _WithoutSelectabilityCategory;
    ASSERT(res == DefaultSelectability ||
	   res == LowSelectability ||
	   res == HighSelectability);
    return static_cast<SelectabilityCategory>(res);
  }; 
  inline
  BK::Comparison
  LiteralProperties::compareSelectabilityForFactoring(const LiteralProperties& prop) const
  {
    CALL("compareSelectabilityForFactoring(const LiteralProperties& prop) const");
    if (selectabilityCategory() > prop.selectabilityCategory()) return BK::Greater;
    if (selectabilityCategory() < prop.selectabilityCategory()) return BK::Less;
    return BK::Equal;
  }; 
}; 
#endif
