/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "VampireKernelDebugFlags.hpp"
#include "LiteralSelection.hpp"
#include "Term.hpp"
#include "SymbolInfo.hpp"
#include "Signature.hpp"
#include "TemporaryLiteral.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "LiteralSelection"
#endif 
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
LiteralSelection::LiteralSelection() :
  _oneOfMaxSizeFilter(CustomLiteralSelection::literalSize),
  _allOfMaxSizeFilter(CustomLiteralSelection::literalSize),
  _oneOfMinSizeFilter(CustomLiteralSelection::literalSize),
  _allOfMinSizeFilter(CustomLiteralSelection::literalSize),
  _allNegEqOrAllFilter(CustomLiteralSelection::isNegativeEquality),
  _oneNegEqOrAnyoneFilter(CustomLiteralSelection::isNegativeEquality),
  _allExceptPosEqOrAllFilter(CustomLiteralSelection::isNotAvailableForParamodFrom),
  _allWithMaxNumOfDiffVarsFilter(CustomLiteralSelection::numOfDifferentVariables),
  _allWithMostShallowVarsFilter(&_inverseVarDepthProfile1,&_inverseVarDepthProfile2),
  _complexFilter3(&_complexCharacteristic1_1,&_complexCharacteristic1_2),
  _complexFilter5(&_complexCharacteristic1_1,&_complexCharacteristic1_2),
  _complexLiteralOrdering1_1(10UL,1.1f,0.0f),
  _complexLiteralOrdering1_2(10UL,2.0f,0.0f),
  _complexLiteralOrdering1_3(10UL,2.0f,-10.0f)
{      
  _mainLiteralOrdering = 0;
  _splittingIsEnabled = false;
  _oneOfEachPolarityFilter.setSelectOne(false);
  _oneOfMaxSizeFilter.setSelectOne(true);
  _allOfMaxSizeFilter.setSelectOne(false);
  _oneOfMinSizeFilter.setSelectOne(true);
  _allOfMinSizeFilter.setSelectOne(false);
  _allNegEqOrAllFilter.setSelectOne(false);
  _oneNegEqOrAnyoneFilter.setSelectOne(true);
  _allExceptPosEqOrAllFilter.setSelectOne(false);
  _oneBigNegEqOrOneBigFilter.setSelectOne(true);
  _oneBigNegEqOrOneBigFilter.resetFilters();
  _oneBigNegEqOrOneBigFilter.addFilter(&_allNegEqOrAllFilter);
  _oneBigNegEqOrOneBigFilter.addFilter(&_oneOfMaxSizeFilter);
  _oneSmallPreferablyNegEqFilter.setSelectOne(true);
  _oneSmallPreferablyNegEqFilter.resetFilters();
  _oneSmallPreferablyNegEqFilter.addFilter(&_allOfMinSizeFilter);
  _oneSmallPreferablyNegEqFilter.addFilter(&_oneNegEqOrAnyoneFilter);
  _allWithMaxNumOfDiffVarsFilter.setSelectOne(false);
  _allWithMostShallowVarsFilter.setSelectOne(false);
  _oneOfMaxSizeMaxNongroundFilter.setSelectOne(true);
  _oneOfMaxSizeMaxNongroundFilter.resetFilters();
  _oneOfMaxSizeMaxNongroundFilter.addFilter(&_allWithMaxNumOfDiffVarsFilter);
  _oneOfMaxSizeMaxNongroundFilter.addFilter(&_allWithMostShallowVarsFilter);
  _oneOfMaxSizeMaxNongroundFilter.addFilter(&_oneOfMaxSizeFilter);
  _complexFilter1.setSelectOne(true);
  _complexFilter1.resetFilters();
  _complexFilter1.addFilter(&_allExceptPosEqOrAllFilter);
  _complexFilter1.addFilter(&_allWithMaxNumOfDiffVarsFilter);
  _complexFilter1.addFilter(&_allWithMostShallowVarsFilter);
  _complexFilter1.addFilter(&_oneOfMaxSizeFilter);
  _complexFilter2.setSelectOne(true);
  _complexFilter2.resetFilters();
  _complexFilter2.addFilter(&_allExceptPosEqOrAllFilter);
  _complexFilter2.addFilter(&_allOfMaxSizeFilter);
  _complexFilter2.addFilter(&_allWithMaxNumOfDiffVarsFilter);
  _complexFilter2.addFilter(&_allWithMostShallowVarsFilter);
  _complexFilter3.setSelectOne(false);
  _complexFilter4.setSelectOne(true);
  _complexFilter4.resetFilters();
  _complexFilter4.addFilter(&_allExceptPosEqOrAllFilter);
  _complexFilter4.addFilter(&_complexFilter3);
  _complexFilter4.addFilter(&_oneOfMaxSizeFilter);
  _complexFilter5.setSelectOne(true);
  _maxSelectionAcceptability1.resetPrimitives();
  _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::oneLiteralSelected);
  _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::noBigNegativeUnselected);
  _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::noNegativeEqualitiesUnselected);
  _maxSelectionAcceptability2.resetPrimitives();
  _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::oneLiteralSelected);
  _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::noNegativeWithManyDiffVarsUnselected);
  _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::noNegativeEqualitiesUnselected);
}; 
void 
LiteralSelection::setFunNum(long num)
{
  switch (num) 
    {
    case 0L : 
      _customSelection = &_totalSelection;
      return;
    case 1L : 
      _customSelection = &_maximalSelection;
      return;
    case 2L : 
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeFilter);
      return;
    case 3L :    
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundFilter);
      return;
    case 4L : 
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneBigNegEqOrOneBigFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability1);
      return;
    case 5L : 
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_1);
      return;
    case 6L : 
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_2);
      return;
    case 7L : 
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_3);
      return;
    case 1001L : 
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter1);
      return;
    case 1002L :
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter2);
      return;
    case 1003L : 
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter4);
      return;
    case 1004L :
      _customSelection = &_incompleteBidirectSelection;
      _incompleteBidirectSelection.setFilterForPositiveInRules(&_oneOfMaxSizeFilter);
      _incompleteBidirectSelection.setFilterForNegativeInRules(&_oneOfMaxSizeFilter);
      _incompleteBidirectSelection.setFilterForPositiveFacts(&_oneOfMaxSizeFilter);
      _incompleteBidirectSelection.setFilterForNegativeFacts(&_oneOfMaxSizeFilter);
      return;
    case -2L :    
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeFilter);
      return;
    case -3L :      
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundFilter);
      return;
    case -4L :        
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneBigNegEqOrOneBigFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability1);
      return;
    case -5L : 
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_1);
      return;
    case -6L : 
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_2);
      return;
    case -7L : 
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1_3);
      return;
    case -1003L : 
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter4);
      return;
    default: BK::RuntimeError::report("Wrong selection number.");
    };
}; 
void LiteralSelection::setMainLiteralOrdering(StandardLiteralOrdering* ord) 
{ 
  _mainLiteralOrdering = ord;
  _totalSelection.setMainLiteralOrdering(ord);
  _maximalSelection.setMainLiteralOrdering(ord);
  _hyperSelection.setMainLiteralOrdering(ord);
  _maxWithCorrectionSelection.setMainLiteralOrdering(ord);
  _arbitrarySelection.setMainLiteralOrdering(ord);
  _incompleteBidirectSelection.setMainLiteralOrdering(ord);
  _bestOnlySelection.setMainLiteralOrdering(ord);
};
void 
LiteralSelection::select(LitList& lits,
			 LitList& selectedLits,
			 LitList& unselectedLits,
			 bool& allMaximalSelected) const
{
  CALL("select(LitList& lits,LitList& selectedLits,LitList& unselectedLits,bool& allMaximalSelected) const");
  LitListStatistics stat; 
  allMaximalSelected = false;
  TempLitListExtras::adjustSelectionLayers(lits);
  BasicTempLitList nonselectableLits(unselectedLits.elementPool());
  bool allRemainingAreSelected;
  unselectNonselectable(lits,nonselectableLits,allRemainingAreSelected);
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  if (allRemainingAreSelected) 
    {
      selectedLits.collectAllFrom(lits);
      goto mark_and_return;  
    };
  stat.collectFor(lits);  
  if (_splittingIsEnabled)
    {      
      selectWithNonblockingSplitting(lits,
				     selectedLits,
				     unselectedLits,
				     allMaximalSelected,
				     stat); 
      goto mark_and_return;
    };
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  if (_stepResolutionLiteralSelectionFlag)
    {
      bool containsNextStateLiterals = false;
      bool containsCurrentStateLiterals = false;
      LitListElement* el = lits.first();
    check_lit:
      if (el)
	{
	  if (isNextStepLiteral(*(el->literal()))) 
	    {
	      containsNextStateLiterals = true;
	    }
	  else
	    if (isCurrentStepLiteral(*(el->literal()))) 
	      containsCurrentStateLiterals = true;
	  if (containsNextStateLiterals && containsCurrentStateLiterals)
	    {
	      unselectCurrentStateLiterals(lits,selectedLits,unselectedLits);
	      ASSERT(!lits.isEmpty());
	      if (lits.length() == 1) 
		{ 
		  lits.popTo(selectedLits); 
		  goto mark_and_return; 
		};
	      stat.collectFor(lits);
	    }
	  else
	    {
	      el = el->next();
	      goto check_lit;
	    };
	};	
    };
#endif
  _customSelection->select(lits,selectedLits,unselectedLits,allMaximalSelected,stat);
 mark_and_return:
  if (!nonselectableLits.isEmpty())
    {
      allMaximalSelected = false;
      unselectedLits.collectAllFrom(nonselectableLits);
    };
  mark(selectedLits,unselectedLits);
}; 
void 
LiteralSelection::selectForSupport(LitList& lits,
				   LitList& selectedLits,
				   LitList& unselectedLits) const
{
  CALL("selectForSupport(..)");
  LitListStatistics stat; 
  TempLitListExtras::adjustSelectionLayers(lits);
  BasicTempLitList nonselectableLits(unselectedLits.elementPool());
  bool allRemainingAreSelected;
  unselectNonselectable(lits,nonselectableLits,allRemainingAreSelected);
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  if (allRemainingAreSelected) 
    {
      selectedLits.collectAllFrom(lits);
      goto mark_and_return;  
    };
  stat.collectFor(lits);
  if (stat.numOfNegNonanswerLits() || stat.numOfNoneqNonanswerLits())
    {
      while (!lits.isEmpty())
	{
	  if (lits.last()->literal()->isEquality() && 
	      lits.last()->literal()->isPositive() &&
	      lits.last()->literal()->properties().useEqForParamodulationFrom())
	    {
	      lits.popTo(unselectedLits);
	    }
	  else
	    lits.popTo(selectedLits);
	};
    }
  else 
    {
      bool dummyBool;
      _customSelection->select(lits,selectedLits,unselectedLits,dummyBool,stat);
    };
 mark_and_return:
  unselectedLits.collectAllFrom(nonselectableLits);
  mark(selectedLits,unselectedLits);
}; 
inline 
void 
LiteralSelection::selectWithNonblockingSplitting(LitList& lits,
						 LitList& selectedLits,
						 LitList& unselectedLits,
						 bool& allMaximalSelected,
						 LitListStatistics& stat) const
{
  CALL("selectWithNonblockingSplitting(..) const");
  ASSERT(!TempLitListExtras::containsAnswerLiterals(lits));
  ASSERT(lits.length() > 1);
  allMaximalSelected = false;
  if (lits.length() > (stat.numOfNegSplittingBranches() + stat.numOfPosSplittingBranches()))
    {
      DLC_INIT(lits.length());
      for (LitListElement* el = lits.first(); el; )
	{
	  DLC_INC;
	  if (el->literal()->header().IsSplitting())     
	    {	      
	      LitListElement* tmp = el->next();
	      lits.passTo(el,unselectedLits);
	      el = tmp;
	    }
	  else
	    el = el->next();
	};
      ASSERT(!lits.isEmpty());
      if (lits.length() == 1) 
	{
	  lits.popTo(selectedLits);
	  allMaximalSelected = true;
	}
      else 
	{
	  LitListStatistics newStat;
	  newStat.collectFor(lits);
	  _customSelection->select(lits,selectedLits,unselectedLits,allMaximalSelected,newStat); 
	};
    }
  else 
    {
      LitListElement* youngestBranch = lits.last();
      ASSERT(youngestBranch);
      ASSERT(youngestBranch->literal()->isSplittingBranch());
      DLC_INIT(lits.length() - 1);
      for (LitListElement* el = youngestBranch->previous();
	   el;
	   el = youngestBranch->previous())
	{
	  DLC_INC;
	  ASSERT(el->literal()->isSplittingBranch());
	  if (el->literal()->header().functor() > 
	      youngestBranch->literal()->header().functor())
	    {
	      lits.popTo(unselectedLits);
	      youngestBranch = el;
	    }
	  else
	    {
	      lits.passTo(el,unselectedLits);
	    };
	};
      ASSERT(lits.length() == 1UL);
      ASSERT(lits.last() == youngestBranch);
      lits.popTo(selectedLits);
    };
}; 
inline 
void 
LiteralSelection::unselectNonselectable(LitList& lits,
					LitList& unselectedLits,
					bool& allRemainingShouldBeSelected)
{
  CALL("unselectNonselectable(..)");
  if (lits.isEmpty()) { allRemainingShouldBeSelected = true; return; };
  LitListElement* el = lits.last();
  while (el &&
	 (el->literal()->isAnswerLiteral() || 
	  el->literal()->properties().selectionLayer() != 0UL))
    {
      el = el->previous();
      lits.popTo(unselectedLits);
    };
  if (lits.length() <= 1UL) 
    {
      allRemainingShouldBeSelected = true; 
      return; 
    };
  ASSERT(el);
  LiteralProperties maxSelectability = el->literal()->properties();
  el = el->previous();
  ASSERT(el);
  do
    {
      if (el->literal()->isAnswerLiteral() || 
	  el->literal()->properties().selectionLayer() != 0UL)
	{
	  LitListElement* tmp = el->previous();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	{
	  Comparison cmp = 
	    compareSelectability(el->literal()->properties(),maxSelectability);
	  if (cmp == Greater)
	    {
	      maxSelectability = el->literal()->properties();
	      do
		{
		  lits.popTo(unselectedLits);
		}
	      while (lits.last() != el);
	      el = el->previous();
	    }
	  else
	    if (cmp == Less)
	      {
		LitListElement* tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
	      }
	    else
	      {
		ASSERT(cmp == Equal);
		el = el->previous();
	      };
	};
    }
  while (el);
  allRemainingShouldBeSelected = 
    lits.length() <= 1UL ||
    maxSelectability.selectabilityCategory() == LiteralProperties::HighSelectability; 
}; 
inline
BK::Comparison
LiteralSelection::compareSelectability(const LiteralProperties& prop1,
				       const LiteralProperties& prop2)
{
  CALL("compareSelectability(..)");
  LiteralProperties::SelectabilityCategory cat1 =
    prop1.selectabilityCategory();
  LiteralProperties::SelectabilityCategory cat2 =
    prop2.selectabilityCategory();  
  if ((cat1 == LiteralProperties::DefaultSelectability) &&
      (cat2 == LiteralProperties::DefaultSelectability))
    return Equal;
  if (cat1 == LiteralProperties::DefaultSelectability)
    {
      ASSERT(cat2 != LiteralProperties::DefaultSelectability);
      if (cat2 == LiteralProperties::LowSelectability)
	return Greater;
      ASSERT(cat2 == LiteralProperties::HighSelectability);
      return Less;
    }
  else if (cat1 == LiteralProperties::LowSelectability)
    {
      if (cat2 != LiteralProperties::LowSelectability)
	return Less;
      return Equal;
    }
  else
    {
      ASSERT(cat1 == LiteralProperties::HighSelectability);
      if (cat2 != LiteralProperties::HighSelectability)
	return Greater;
      return Equal;
    };
}; 
inline 
void 
LiteralSelection::mark(LitList& selectedLits,LitList& unselectedLits)
{
  CALL("mark(LitList& selectedLits,LitList& unselectedLits)");
  LitListElement* el;
  for (el = selectedLits.first();
       el;
       el = el->next())
    {
      el->literal()->markAsSelected();
    };
  for (el = unselectedLits.first();
       el;
       el = el->next())
    {
      el->literal()->markAsUnselected();
    };	 
}; 
#ifdef VKERNEL_FOR_STEP_RESOLUTION
inline
bool 
LiteralSelection::isNextStepLiteral(const TemporaryLiteral& lit)
{
  return lit.arity() && 
    (!lit.isEquality()) && 
    lit.flatterm()->next()->arity();
};
inline
bool 
LiteralSelection::isCurrentStepLiteral(const TemporaryLiteral& lit)
{
  return lit.arity() && 
    (!lit.isEquality()) && 
    (!lit.flatterm()->next()->arity());
};
void 
inline
LiteralSelection::unselectCurrentStateLiterals(LitList& lits,
					       LitList& selectedLits,
					       LitList& unselectedLits) const
{
  CALL("unselectCurrentStateLiterals(LitList& lits,...) const");
  LitListElement* el = lits.first();
  while (el)
    {
      if (isCurrentStepLiteral(*(el->literal())))
	{	  
	  LitListElement* tmp = el->next();
	  lits.passTo(el,unselectedLits);
	  el = tmp;
	}
      else
	el = el->next();
    };
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "LiteralSelection::ComplexLiteralCharacteristic1"
#endif 
#include "debugMacros.hpp"
LiteralSelection::ComplexLiteralCharacteristic1::~ComplexLiteralCharacteristic1()
{
};
void 
LiteralSelection::ComplexLiteralCharacteristic1::computeFor(TemporaryLiteral& lit)
{
  CALL("computeFor(TemporaryLiteral& lit)");
  _positive = lit.isPositiveForSelection();
  _numOfDifferentVars = lit.numOfDifferentVariables();
  _numOfVarOccurences = lit.numberOfVariableOccurences();
}; 
Comparison 
LiteralSelection::ComplexLiteralCharacteristic1::compareWith(CustomLiteralSelection::ComplexLiteralCharacteristic* c)
{
  CALL("compareWith(ComplexLiteralCharacteristic* c)");
  if (_positive != (static_cast<ComplexLiteralCharacteristic1*>(c))->_positive)
    return Incomparable;
  if (_positive)
    {
      if (_numOfDifferentVars > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Greater;
      if (_numOfDifferentVars < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Less;
      if (_numOfVarOccurences > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Greater;
      if (_numOfVarOccurences < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Less;
    }
  else 
    {      
      if (_numOfDifferentVars > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Less;
      if (_numOfDifferentVars < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfDifferentVars)
	return Greater;
      if (_numOfVarOccurences > 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Less;
      if (_numOfVarOccurences < 
	  (static_cast<ComplexLiteralCharacteristic1*>(c))->_numOfVarOccurences)
	return Greater;
    };
  return Equal;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "LiteralSelection::ComplexLiteralOrdering1"
#endif 
#include "debugMacros.hpp"
LiteralSelection::ComplexLiteralOrdering1::~ComplexLiteralOrdering1()
{
};
Comparison 
LiteralSelection::ComplexLiteralOrdering1::compare(TemporaryLiteral& lit1,
						   TemporaryLiteral& lit2) const
{
  CALL("compare(..) const"); 
  if (lit1.isEquality())
    {
      if (lit2.isEquality())
	{
	  if (lit1.isNegative())
	    {
	      if (lit2.isNegative())
		{
		  if (lit1.size() == 3UL) 
		    return Greater;
		  if (lit2.size() == 3UL) 
		    return Less;
		  float resDiff = 
		    lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) - 
		    lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff);
		  if (resDiff > 0.0) return Greater;
		  if (resDiff < 0.0) return Less;
		  if (lit1.numberOfVariableOccurences() >
		      lit2.numberOfVariableOccurences())
		    return Less;
		  if (lit1.numberOfVariableOccurences() <
		      lit2.numberOfVariableOccurences())
		    return Greater;
		  if (lit1.numOfDifferentVariables() >
		      lit2.numOfDifferentVariables())
		    return Less;
		  if (lit1.numOfDifferentVariables() <
		      lit2.numOfDifferentVariables())
		    return Greater;
		  return Incomparable;
		}
	      else
		return Greater;
	    }
	  else
	    if (lit2.isNegative())
	      {
		return Less;
	      }
	    else 
	      {
		float resDiff = 
		  lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) - 
		  lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff);
		if (resDiff > 0.0) return Greater;
		if (resDiff < 0.0) return Less;
		if (lit1.numOfDifferentVariables() >
		    lit2.numOfDifferentVariables())
		  return Greater;
		if (lit1.numOfDifferentVariables() <
		    lit2.numOfDifferentVariables())
		  return Less;
		return Incomparable;	
	      };
	}
      else 
	{
	  if (lit1.isNegative())
	    {
	      return Greater;
	    }
	  else
	    {
	      if (lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) > 
		  lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff)) 
		return Greater;
	      return Less;
	    };
	};
    }
  else
    if (lit2.isEquality())
      {
	if (lit2.isNegative())
	  {
	    return Less;
	  }
	else
	  {
	    if (lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) < 
		lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff)) 
	      return Less;
	    return Greater;
	  };
      }
    else 
      {
	float resDiff = 
	  lit1.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff) - 
	  lit2.resistance(_resistanceDepthCoeff,_resistanceConstCoeff,_resistanceVarCoeff);
	if (resDiff > 0.0) return Greater;
	if (resDiff < 0.0) return Less;
	if (lit1.isPositiveForSelection())
	  {
	    if (lit2.isPositiveForSelection())
	      {
		if (lit1.numOfDifferentVariables() >
		    lit2.numOfDifferentVariables())
		  return Greater;
		if (lit1.numOfDifferentVariables() <
		    lit2.numOfDifferentVariables())
		  return Less;		  
		return Incomparable;	
	      }
	    else 
	      return Greater;	      
	  }
	else
	  if (lit2.isPositiveForSelection())
	    {
	      return Less;
	    }
	  else 
 	    {
	      if (lit1.numOfDifferentVariables() >
		  lit2.numOfDifferentVariables())
		return Less;
	      if (lit1.numOfDifferentVariables() <
		  lit2.numOfDifferentVariables())
		return Greater;
	      return Incomparable;
	    };	
      };
}; 
Comparison
LiteralSelection::ComplexLiteralOrdering1::compareLayers(const TemporaryLiteral& lit1,
							 const TemporaryLiteral& lit2) const
{
  CALL("compareLayers(..) const");
  if (lit1.isEquality() && lit1.isNegative())
    {
      if ((!lit2.isEquality()) || lit2.isPositive())
	return Greater;
    }
  else
    if (lit2.isEquality() && lit2.isNegative())
      return Less;
  return Equal;
}; 
