/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LITERAL_SHARING_HPP
#define LITERAL_SHARING_HPP 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "DiscTree.hpp"
#include "Array.hpp"
#include "Term.hpp"
#include "TermSharing.hpp"
namespace VK
{
class Flatterm;
class DiscTree::Node;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SHARING
#  define DEBUG_NAMESPACE "LiteralSharing"
#endif
#include "debugMacros.hpp"
namespace VK
{
class LiteralSharing
{
 public:
  LiteralSharing(TermSharing* si);
  ~LiteralSharing();
  bool isLocked() const { return _isLocked; };
  void lock() 
  {
    CALL("lock()");
    ASSERT(!_isLocked);
    _isLocked = true;
  };
  void unlock() 
  {
    CALL("unlock()");
    ASSERT(_isLocked);
    _isLocked = false;
  };
  void reset();
  DiscTree::Node** Tree(const Term& header);
  DiscTree::Node* Integrate(Term* lit); 
  DiscTree::Node* Integrate(const Flatterm* lit); 
  DiscTree::Node* IntegrateEq(const Term& header,
                          const Flatterm* arg1,
                          const Flatterm* arg2); 
  void InfoForDeletion(Term* lit,
                       DiscTree::Node**& branch_to_destroy,
                       DiscTree::Node*& leaf,
                       DiscTree::Node*& last_fork); 
  void destroyLiteralCopy(Term* lit);
private:
  static void DestroyLeafContent(DiscTree::Node* leaf);
  Term* ShareProp(const Term& header);
 private:
  bool _isLocked;
  BK::Array<BK::GlobAlloc,Term*,BK::DefaultRelocation<Term*>,128UL> _propositional;  
  TermSharing* _termSharing;
  BK::Array<BK::GlobAlloc,DiscTree::Node*,BK::DefaultRelocation<DiscTree::Node*>,512UL> _discTree;
}; 
}; 
#endif
