/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "GlobalStopFlag.hpp"
#include "Master.hpp"
#include "PagedPassiveClauseStorage.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
void Master::inputClause3(Clause* cl)
{
  CALL("inputClause3(Clause* cl)");
  if (cl->isSupportClause())
    {
      inputSupportClause3(cl);
    }
  else
    inputQueryClause3(cl);
}; 
void Master::inputQueryClause3(Clause* cl)
{
  CALL("inputQueryClause3(Clause* cl)");  
  Clause* sharedRetainedAnswer;
  ASSERT(!cl->isSupportClause());
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause3_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    case Continuation_InputQueryClause3_BackFold:
      cancelTermination();
      goto back_fold;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputQueryClause3(Clause* cl)");
    };
  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;
  if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro,cl))
    {
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{	  
	  ASSERT(!_newInputClause.isNull());
	  _pagedPassive.maximalPriorityOn();
	  _forwardProcessing.load(_newInputClause);
	process_collected_input:
	  _continuation = Continuation_Init;
	  while (_forwardProcessing.anotherRetained(_retainedClause))
	    {
	      if (_retainedClause->isAnswer())
		{
		  sharedRetainedAnswer = 
		    _clauseAssembler.assembleShared(*_retainedClause);
		  if (!sharedRetainedAnswer)
		    {
		      ++_statistics.clauses.generated.discardedTooLarge;
		    }
		  else
		    {
		      ++_statistics.clauses.retained.total;
		      ASSERT(sharedRetainedAnswer);
		      ASSERT(sharedRetainedAnswer->isAnswer());
		      sharedRetainedAnswer->setMainSet(Clause::MainSetNew);
		      sharedRetainedAnswer->SetNumber(_clauseNumberGenerator->generate());
		      _answers.Insert(sharedRetainedAnswer);
		      _proof = sharedRetainedAnswer;
		      terminate(Proof);   
		      disposeObsolete(); 
		      _timer.stop();
		      _timerIsOn = false; 
		      _continuation = Continuation_InputQueryClause3_ProcessCollectedInput;
		      return;
		    };
		}
	      else 
		{
		  long clauseNumber = _clauseNumberGenerator->generate();
                  ostream* outputStreamForPagedPassive = 
                    (_parameters.showPassive)? (&cout) : 0;
		  if (!_clauseAssembler.assembleInPagedPassiveClauseStorage(_pagedPassive,
									    *_retainedClause,
									    clauseNumber,
									    outputStreamForPagedPassive))
		    {
		      ++_statistics.clauses.generated.discardedTooLarge;
		    }
		  else
		    ++_statistics.clauses.retained.total;
		};
	    }; 
	  _pagedPassive.maximalPriorityOff(); 
	}; 
    back_fold:
      backFold3();
      if (_proof) 
	{
	  terminate(Proof);  
	  _timer.stop();
	  _timerIsOn = false; 
	  _continuation = Continuation_InputQueryClause3_BackFold;
	  return; 
	};    
    }; 
  _continuation = Continuation_Init;
  _timer.stop();
  _timerIsOn = false;
}; 
void Master::inputSupportClause3(Clause* cl)
{
  CALL("inputSupportClause3(Clause* cl)");
  ASSERT(cl->isSupportClause());
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputSupportClause3_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputSupportClause3(Clause* cl)");
    };
  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;
  if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro,cl))
    {
      _inputClauseCollector.markAsSupport();
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{
	  ASSERT(!_newInputClause.isNull());
	  _forwardProcessingForSupport.load(_newInputClause);
	process_collected_input:
	  while (_forwardProcessingForSupport.anotherRetained(_retainedClause))
	    {
	      sharedRetainedClause = 
		_clauseAssembler.assembleShared(*_retainedClause);
	      if (!sharedRetainedClause)
		{
		  ++_statistics.clauses.generated.discardedTooLarge;
		}
	      else
		{
		  ++_statistics.clauses.retained.total;
		  ASSERT(sharedRetainedClause);
		  sharedRetainedClause->setMainSet(Clause::MainSetNew);
		  if (sharedRetainedClause->isAnswer()) 
		    {
		      rememberAnswer0(sharedRetainedClause);
		      _proof = sharedRetainedClause;
		      terminate(Proof); 
		      _timer.stop();
		      _timerIsOn = false;
		      _continuation = Continuation_InputSupportClause3_ProcessCollectedInput;
		      return;
		    };
		  sharedRetainedClause->SetNumber(_clauseNumberGenerator->generate());      
		  if (_parameters.useClauseNumIndex) 
		    {
		      sharedRetainedClause->protect();
		      _clauseNumIndex.insert(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		    };
		  if (_parameters.showPassive) _parameters.showPassive(sharedRetainedClause);
		  integratePassive(sharedRetainedClause);
		  makeActive0(sharedRetainedClause);
		};
	    }; 
	  _continuation = Continuation_Init;
	  integrateIntoBackDemIndexFromQueue(); 
	}; 
    }; 
}; 
void Master::mainLoop3()
{
  CALL("mainLoop3()");
  PagedPassiveClauseStorage::StoredClause selectedClause;
  Clause* recentlyReanimatedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_MainLoop3_SelectCurrent:
      cancelTermination();
      _continuation = Continuation_Init;
      goto select_current;
    case Continuation_MainLoop3_BackSimplify:
      cancelTermination();
      goto back_simplify;
    case Continuation_MainLoop3_BackFold:
      cancelTermination();
      goto back_fold;
    case Continuation_MainLoop3_ProcessNewInference:       
      cancelTermination();
      _continuation = Continuation_Init;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::mainLoop3()");
    };
  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;
  _startMemory = (float)GlobAlloc::occupiedByObjects(); 
  ASSERT(!_terminated);
 select_current: 
  GlobalStopFlag::check(); 
  if (!_pagedPassive.select(selectedClause)) 
    { 
      terminate(EmptyPassive); 
      _timer.stop();
      _timerIsOn = false; 
      return; 
    };
  if (_active.Total() >= _parameters.maxActive) 
    { 
      terminate(MaxActive); 
      _timer.stop();
      _timerIsOn = false; 
      return; 
    };
  if (_parameters.orphan_murder
      && (!selectedClause.isVIP())
      && (selectedClause.isOrphan()))
    {   
      ++_statistics.clauses.retained.passiveOrphansMurdered;
      _pagedPassive.popSelected();
      goto select_current;
    };
  _sharedReanimatedClause = reanimate3(selectedClause);
  _pagedPassive.popSelected();
  if (_sharedReanimatedClause)
    {
      ++_statistics.clauses.retained.total;
      _sharedReanimatedClause->SetNumber(_clauseNumberGenerator->generate());
      _sharedReanimatedClause->setMainSet(Clause::MainSetFree); 
      if (_parameters.useClauseNumIndex) 
	{
	  _sharedReanimatedClause->protect();
	  _clauseNumIndex.insert(_sharedReanimatedClause);
	  _sharedReanimatedClause->unprotect();
	};
      if (_parameters.showPassive) _parameters.showPassive(_sharedReanimatedClause);
    }
  else 
    {
      goto select_current; 
    };
  internalTick3();
  if (_sharedReanimatedClause->isAnswer()) 
    { 
      _answers.Insert(_sharedReanimatedClause);
      _proof = _sharedReanimatedClause;
      _sharedReanimatedClause = 0;
      terminate(Proof);  
      _timer.stop();
      _timerIsOn = false; 
      _continuation = Continuation_MainLoop3_SelectCurrent;
      return; 
    };
  if (_parameters.backward_subsumption) 
    if (backSubsumeBy(_sharedReanimatedClause)) 
      if (_parameters.orphan_murder) killActiveOrphans3();      
  disposeObsolete();
  internalTick3();
 back_simplify:
  backSimplifyBy3(_sharedReanimatedClause);
  if (_proof) 
    {
      terminate(Proof);  
      _timer.stop();
      _timerIsOn = false; 
      _continuation = Continuation_MainLoop3_BackSimplify;
      return; 
    };    
  disposeObsolete();
  makeActive3(_sharedReanimatedClause);
  recentlyReanimatedClause = _sharedReanimatedClause;
  _sharedReanimatedClause = 0; 
  if (_parameters.outputStreamForPassiveProfiles &&
      (_active.Total() % _parameters.showPassiveProfileInterval == 0))
    {
      _pagedPassive.outputProfile(*_parameters.outputStreamForPassiveProfiles);
    };
  eliminateAllocationBufferDeficit3();  
  _infDispatcher.newQuery(recentlyReanimatedClause);
 process_new_inference:
  processCollected3();
  if (_proof) 
    { 
      terminate(Proof); 
      _timer.stop();
      _timerIsOn = false; 
      _continuation = Continuation_MainLoop3_ProcessNewInference;
      return; 
    };
  if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
    { 
      terminate(MaxPassive); 
      _infDispatcher.abandonCurrentQuery();
      _timer.stop();
      _timerIsOn = false; 
      return; 
    };
  internalTick3();
  GlobalStopFlag::check();
  if (_infDispatcher.nextInference()) goto process_new_inference;
  _infDispatcher.abandonCurrentQuery();
  GlobalStopFlag::check();
 back_fold:
  backFold3();
  if (_proof) 
    {
      terminate(Proof);  
      _timer.stop();
      _timerIsOn = false; 
      _continuation = Continuation_MainLoop3_BackFold;
      return; 
    };    
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_backFolding.queueIsEmpty());
  disposeObsolete();
  ASSERT(_trash.Empty());
  GlobalStopFlag::check();
  goto select_current; 
};
void Master::backSimplifyBy3(Clause* simplifier)
{
  CALL("backSimplifyBy3(Clause* simplifier)");
  Clause* backSimplified; 
  bool somethingSimplified;
  bool onBranch;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_MainLoop3_BackSimplify:
      somethingSimplified = true;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::backSimplifyBy3(..)");
    };
  somethingSimplified = false;
  if (_parameters.backward_demodulation > 0)
    {
      if (_backDemodulation.query(simplifier))
	{ 
	  while (_backDemodulation.nextSimplified(onBranch))
	    {      
	      simplifier->MakeVIP();
	      if (_backDemodulation.collectInference())
		{ 
		  backSimplified = _backDemodulation.simplified();
		  ASSERT(backSimplified->mainSet() == Clause::MainSetActive); 
		  if (onBranch)
		    {
		      ++_statistics.inferences.simplifying.onBranchBackwardDemodulation;
		      ++_statistics.clauses.retained.activeSimplifiedByOnBranchDemodulation;
		    }
		  else
		    {
		      ++_statistics.inferences.simplifying.backwardDemodulation;
		      ++_statistics.clauses.retained.activeSimplifiedByDemodulation;
		    };
		  backSimplified->setFlag(Clause::FlagHasChildren);
		  enqueueForDisposal(backSimplified);  
		  somethingSimplified = true;
		process_new_inference:
		  processCollected3(); 
		  if (_proof) 
		    { 
		      terminate(Proof);  
		      return; 
		    };
		  _continuation = Continuation_Init;
		  if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
		    { 
		      terminate(MaxPassive); 
		      _backDemodulation.abandonCurrentQuery();
		      return; 
		    }; 
		};
	      GlobalStopFlag::check();  
	      internalTick3();
	    }; 
	  _backDemodulation.abandonCurrentQuery();
	}; 
      if (_parameters.orphan_murder && somethingSimplified) killActiveOrphans3();
    };
}; 
void Master::backFold3()
{
  CALL("backFold3()");
  Clause* backFolded; 
  bool somethingFolded;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_MainLoop3_BackFold:
    case Continuation_InputQueryClause3_BackFold:
      somethingFolded = true;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::backFold3(..)");
    };
  somethingFolded = false;
  if (_parameters.back_fold_in_spl)
    {
    next_definition:
      if (_backFolding.nextDefinition())
	{
	fold_next:
	  if (_backFolding.foldNext())
	    {
	      if (_backFolding.collectResult())
		{ 
		  backFolded = _backFolding.folded();
		  ASSERT(backFolded->mainSet() == Clause::MainSetActive);
		  ++_statistics.inferences.simplifying.backwardFolding;
		  ++_statistics.clauses.retained.activeBackFolded;
		  backFolded->setFlag(Clause::FlagHasChildren);
		  enqueueForDisposal(backFolded);        
		  somethingFolded = true;
		process_new_inference:
		  processCollected3(); 
		  if (_proof) 
		    { 
		      terminate(Proof); 
		      return; 
		    };
		  _continuation = Continuation_Init;
		  if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
		    { 
		      terminate(MaxPassive); 
		      _backDemodulation.abandonCurrentQuery();
		      return; 
		    }; 
		  GlobalStopFlag::check();  
		  internalTick3();
		}; 
	      goto fold_next;
	    }; 
	  _backFolding.abandonCurrentDefinition();
	  goto next_definition;
	}; 
      if (_parameters.orphan_murder && somethingFolded) killActiveOrphans3();
    }; 
}; 
Clause* Master::reanimate3(PagedPassiveClauseStorage::StoredClause& cl)
{
  CALL("reanimate3(PagedPassiveClauseStorage::StoredClause& cl)");
  ASSERT(_parameters.main_alg == 3); 
  if (!_reanimatedClauseCollector.loadClause(cl.literals(),
					     ClauseBackground::Reanimation))
    return 0;
  _reanimatedClauseCollector.addRules(cl.rules());
  cl.resetAncestors();
  Clause* ancestor;
  bool parent;
  while (cl.nextAncestor(ancestor,parent))
    {
      if (parent)
	{
	  _reanimatedClauseCollector.addParent(ancestor);
	}
      else
	_reanimatedClauseCollector.addSimplifier(ancestor);
    };
  _reanimatedClauseCollector.announceNewClause();
  if (!_reanimatedClauseCollector.finaliseClause(_reanimatedClause))
    return 0;
  _forwardProcessing.load(_reanimatedClause);
  if (!_forwardProcessing.anotherRetained(_retainedClause))
    return 0;
  Clause* res = 
    _clauseAssembler.assembleShared(*_retainedClause);
  if (!res)
    ++_statistics.clauses.retained.passiveRecycled;
  return res;
}; 
void Master::processCollected3()
{
  CALL("processCollected3()");
  Clause* sharedRetainedAnswer;
 process_another_retained:
  while (_forwardProcessing.anotherRetained(_retainedClause))
    {
      if (_retainedClause->isAnswer())
	{
	  sharedRetainedAnswer = 
	    _clauseAssembler.assembleShared(*_retainedClause);
	  if (!sharedRetainedAnswer)
	    {
	      ++_statistics.clauses.generated.discardedTooLarge;
	    }
	  else
	    {
	      ++_statistics.clauses.retained.total;
	      ASSERT(sharedRetainedAnswer->isAnswer());
	      sharedRetainedAnswer->setMainSet(Clause::MainSetNew);
	      sharedRetainedAnswer->SetNumber(_clauseNumberGenerator->generate());
	      _answers.Insert(sharedRetainedAnswer);
	      _proof = sharedRetainedAnswer;
	      return;
	    };
	}
      else 
	{
	  long clauseNumber = _clauseNumberGenerator->generate();
	  ostream* outputStreamForPagedPassive = 
	    (_parameters.showPassive)? (&cout) : 0;
	  if (!_clauseAssembler.assembleInPagedPassiveClauseStorage(_pagedPassive,
								    *_retainedClause,
								    clauseNumber,
								    outputStreamForPagedPassive))
	    {
	      ++_statistics.clauses.generated.discardedTooLarge;
	    }
	  else
	    ++_statistics.clauses.retained.total;
	  eliminateAllocationBufferDeficit3();
	}; 
    }; 
  if (!_newClauseQueue.isEmpty())
    {
      _forwardProcessing.load(_newClauseQueue);
      goto process_another_retained;
    };
}; 
void Master::makeActive3(Clause* cl)
{
  CALL("makeActive3(Clause* cl)");
  ++(_statistics.clauses.retained.selected);
  cl->setFlag(Clause::FlagHasChildren); 
  _active.Insert(cl);
  _infDispatcher.makeActive(cl);
  if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
  if (_parameters.backward_subsumption) 
    { 
      integrateIntoBackwardSubsumptionIndex(cl);
    };
  if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
  if (_parameters.backward_demodulation > 0) integrateIntoBackDemodulationIndex(cl);
  if (_parameters.showActive) _parameters.showActive(cl);
}; 
bool Master::releaseMemory3(ulong size) 
{
  CALL("releaseMemory3(ulong size)");
  if (size > PagedPassiveClauseStorage::pageSize()) 
    return false;
 try_free_page:
  if (_pagedPassive.recyclePage()) return true;
  if (_pagedPassive.cancelWorstClause()) 
    {
      ++_statistics.clauses.retained.passiveRecycled;
      goto try_free_page; 
    };
  return false;
}; 
bool Master::releaseALotOfMemory3()
{
  CALL("releaseALotOfMemory3()");
  bool res = false;
  ulong numOfClausesToRecycle =
    _pagedPassive.numberOfCurrentlyKeptClauses() / 10;
  bool somePassiveClauseRecycled = false;
  while (numOfClausesToRecycle &&
	 _pagedPassive.cancelWorstClause())
    {
      ++_statistics.clauses.retained.passiveRecycled;
      --numOfClausesToRecycle;
      somePassiveClauseRecycled = true;
      BK::GlobalStopFlag::check();
    };
  while (_pagedPassive.recyclePage()) 
    {
      res = true;
    };
#ifdef CHANGE_WEIGHT_LIMIT_AFTER_MAJOR_RECYCLING
  if (somePassiveClauseRecycled)
    {
      _parameters.maxMaxWeight = _pagedPassive.maxPresentWeight();
      updateWeightLimit(_pagedPassive.maxPresentWeight());
    };
#endif
#ifdef DONT_RECYCLE_ACTIVE
  numOfClausesToRecycle = _active.Size() / 10;
  while (numOfClausesToRecycle)
    {
      Clause* cand = goodActiveCandidateForDisposal();
      if (!cand) return res;
      --numOfClausesToRecycle;
      res = true;
      enqueueForDisposal(cand);
      ++_statistics.clauses.retained.activeRecycled;
      BK::GlobalStopFlag::check();
    };
#endif
  return res;
}; 
void Master::killActiveOrphans3()
{
  CALL("killActiveOrphans3()");
  killActiveOrphans0();
}; 
void Master::eliminateAllocationBufferDeficit3()
{
  CALL("eliminateAllocationBufferDeficit3()");
  if (BK::GlobAlloc::activeBufferDeficit())
    {
      long disposalBound = 
	(long)(BK::GlobAlloc::occupiedByObjects() - 
	       (_parameters.allocation_buffer_price * 
		BK::GlobAlloc::activeBufferDeficit()));
      do
	{
	  while (!_pagedPassive.recyclePage())
	    {
	      if (_pagedPassive.cancelWorstClause()) 
		{
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  if (BK::GlobAlloc::activeBufferDeficit()) 
		    BK::GlobAlloc::annulBufferDeficit();
		  return;
		};
	    };
	}
      while (BK::GlobAlloc::activeBufferDeficit() && 
	     BK::GlobAlloc::occupiedByObjects() > disposalBound);
      if (BK::GlobAlloc::activeBufferDeficit()) 
	BK::GlobAlloc::annulBufferDeficit();
    };
}; 
