/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef NEGATIVE_EQUALITY_SPLITTING_H
#define NEGATIVE_EQUALITY_SPLITTING_H
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
namespace VK
{
  class TemporaryClause;
  class TemporaryClauseCollector;
  class ComponentName;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEGATIVE_EQUALITY_SPLITTING
#  define DEBUG_NAMESPACE "NegativeEqualitySplitting"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class NegativeEqualitySplitting
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
  public:
    NegativeEqualitySplitting(ComponentName* componentNaming);
    ~NegativeEqualitySplitting();
    void reset();
    void setMinimalEligibleArgSize(ulong n)
    {
      _minimalEligibleArgSize = n;
    };
    void setMaximalEligibleVarCount(ulong n)
    {
      _maximalEligibleVarCount = n;
      if (_maximalEligibleVarCount >= VampireKernelConst::MaxNumOfVariables)
	{
	  _maximalEligibleVarCount = VampireKernelConst::MaxNumOfVariables - 1;
	};
    };
    bool
    splitOffAppropriateNegEqArguments(TemporaryClause& cl,
				      TemporaryClauseCollector* newClauseCollector,
				      ulong& numOfSplitOffArgs,
				      ulong& numOfNewNames);
  private:
    NegativeEqualitySplitting() {};
    bool splitOffAppropriateArgument(LitHandle& lit,
				     TemporaryClauseCollector* newClauseCollector,
				     bool& newNameWasIntroduced,
				     Clause*& nameIntroClause);
    static ulong variableOutside(const WeightPolynomial::Monomials& vars);
  private:
    long _minimalEligibleArgSize;
    long _maximalEligibleVarCount;
    ComponentName* _componentNaming;
    Flatterm _newVariable[1];   
    LitHandle _definedLiteral;
    LitHandle _definitionBodyLiteral;
    BasicTempLitList _definitionBody;
  }; 
}; 
#endif
