/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PAGED_PASSIVE_CLAUSE_STORAGE_INL
#  define PAGED_PASSIVE_CLAUSE_STORAGE_INL
#include "PagedPassiveClauseStorage.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage"
#endif 
#include "debugMacros.hpp"
namespace VK
{
inline
void PagedPassiveClauseStorage::reset() 
{ 
  CALL("reset()"); 
  BK::destroy(*this); 
  BK::init(*this); 
}; 
inline
void PagedPassiveClauseStorage::openLiteral(const LiteralProperties& prop)
{
  CALL("openLiteral(const LiteralProperties& prop)");
  ASSERT(_debugClauseOpenForWriting); 
  ASSERT(_cursor + 2*sizeInWords(sizeof(Term)) < _debugEndOfCurrentClause);
  ASSERT(_debugNumberOfLiteralsToWrite);
  _currentLiteral = (Term*)_cursor;
  _currentLiteralSize = 0UL;
  _cursor += sizeInWords(sizeof(Term));
  ((Term*)_cursor)->Make(prop.content());
  _cursor += sizeInWords(sizeof(Term));
}; 
inline
void PagedPassiveClauseStorage::writeSymbol(const Term& sym) 
{
  CALL("writeSymbol(const Term& sym)");
  ASSERT(_debugNumberOfLiteralsToWrite);
  ASSERT(_cursor + sizeInWords(sizeof(Term)) <= _debugEndOfCurrentClause);
  *((Term*)_cursor) = sym;
  _cursor += sizeInWords(sizeof(Term));
  _currentLiteralSize++;
}; 
inline
void PagedPassiveClauseStorage::closeLiteral()
{ 
  CALL("closeLiteral()");
  ASSERT(_debugNumberOfLiteralsToWrite);
  DOP(_debugNumberOfLiteralsToWrite--);
  _currentLiteral->Make(_currentLiteralSize);       
}; 
inline
void PagedPassiveClauseStorage::writeAncestor(const Clause* ancestor,bool parent)
{
  CALL("writeAncestor(const Clause* ancestor,bool parent)");
  ASSERT(_debugClauseOpenForWriting); 
  ASSERT(!_debugNumberOfLiteralsToWrite);
  ASSERT(_debugNumberOfAncestorsToWrite);
  ASSERT(_cursor + 1 <= _debugEndOfCurrentClause);
  if (parent)
    {
      *_cursor = 
	static_cast<void*>(Ancestor(const_cast<Clause*>(ancestor),true).rawPtr());
    }
  else
    {
      *_cursor = static_cast<void*>(Ancestor(const_cast<Clause*>(ancestor)).rawPtr());
    };
  _cursor++;
  DOP(_debugNumberOfAncestorsToWrite--);
}; 
inline
void PagedPassiveClauseStorage::endOfClause()
{
  CALL("endOfClause()");
  ASSERT(_debugClauseOpenForWriting);  
  DOP(_debugClauseOpenForWriting = false);
  ASSERT(!_debugNumberOfLiteralsToWrite);
  ASSERT(!_debugNumberOfAncestorsToWrite);
  ASSERT(!empty()); 
  ASSERT(_category[_minNonemptyCatNum].nonempty());
  ASSERT(_category[_maxNonemptyCatNum].nonempty());
}; 
inline
void 
PagedPassiveClauseStorage::loadClause(const StoredClauseHandle& handle,
	                              StoredClause& clause)
{
  CALL("loadClause(const StoredClauseHandle& handle,StoredClause& clause)");
  ASSERT((bool)handle);
  readFromMemory(handle,
		 clause.profile(),
		 clause.literals(),
		 clause.ancestors());
}; 
inline
long PagedPassiveClauseStorage::categoryNum(long clauseWeight) 
{
  if (clauseWeight > (long)MaximalCategoryNumber) return MaximalCategoryNumber;
  if (clauseWeight < 0L) return 0L;
  return clauseWeight;
};
inline
ulong PagedPassiveClauseStorage::sizeInWords(ulong sizeInBytes) 
{
  return (sizeInBytes + sizeof(void*) - 1) / sizeof(void*);
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::ClauseLiterals"
#endif 
#include "debugMacros.hpp"
namespace VK
{ 
inline
bool PagedPassiveClauseStorage::ClauseLiterals::nextLiteral() 
{
  CALL("nextLiteral()");
  ASSERT(!_numOfRemainingSymbols);
  if (_numOfRemainingLiterals)
    {
      _numOfRemainingLiterals--;
      _numOfRemainingSymbols = _cursor->Content();
      _cursor++;
      _currentLiteralProperties.make(_cursor->Content());
      _cursor++;
      return true;
    }
  else return false;   
};
inline
bool PagedPassiveClauseStorage::ClauseLiterals::nextSymbol(Term& sym) 
{
  CALL("nextSymbol(Term& sym)");
  if (_numOfRemainingSymbols) 
    {
      _numOfRemainingSymbols--;
      sym = *_cursor;
      _cursor++;
      return true;
    }
  else return false;
};
inline
void PagedPassiveClauseStorage::ClauseLiterals::reset() 
{
  CALL("reset()");    
  _cursor = _beginning;
  _numOfRemainingLiterals = _numberOfLiterals;
  _numOfRemainingSymbols = 0;
};
inline
const LiteralProperties& 
PagedPassiveClauseStorage::ClauseLiterals::currentLiteralProperties() const
{
  CALL("currentLiteralProperties() const");
  return _currentLiteralProperties;
};
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::ClauseAncestors"
#endif 
#include "debugMacros.hpp"
namespace VK
{
inline
bool 
PagedPassiveClauseStorage::ClauseAncestors::nextAncestor(Clause*& anc,
	                                                 bool& isParent) 
{
  CALL("nextAncestor(Clause*& anc,boo)"); 
 if (_numOfRemainingAncestors)
    {
      _numOfRemainingAncestors--;
      Ancestor tmp((Clause*)(*_cursor));
      anc = tmp.strip(isParent);
      _cursor++;
      return true;
    }
  else return false;
};
inline
void PagedPassiveClauseStorage::ClauseAncestors::reset()
{
  CALL("reset()");    
  _cursor = _beginning;
  _numOfRemainingAncestors = _numberOfAncestors;
};
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage::StoredClause"
#endif 
#include "debugMacros.hpp"
namespace VK
{
inline
bool PagedPassiveClauseStorage::StoredClause::isOrphan() 
{
  CALL("isOrphan()");
  resetAncestors();
  Clause* ancestor;
  bool parent;
  while (nextAncestor(ancestor,parent))
    {
      if (parent && !ancestor->isPersistent()) return true;
    };
  return false;  
};
}; 
#endif
