/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "PassiveClauseSet.hpp"
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PASSIVE_CLAUSE_SET
#  define DEBUG_NAMESPACE "PassiveClauseSet"
#endif
#include "debugMacros.hpp"
void PassiveClauseSet::Reset()
{
  _maxPriorityFlag = false;
  num_of_currently_kept = 0L;
  num_of_used = 0L;
  num_of_removed = 0L;
  num_of_cancelled = 0L;
  numOfReachable = LONG_MAX;
  for (long cat = 0; cat < NumberOfCategories; cat++) StatReset(cat);
  _minNonemptyCatNum = 0;
  _maxNonemptyCatNum = 0;
  max_accepted_cat = NumberOfCategories-1;
  _totalSelectedByAgeOrWeight = 0;
  _totalSelectedByWeight = 0;
}; 
void 
PassiveClauseSet::setAgeWeightRatio(long numerator,
				    long denominator) 
{
  _ageWeightRatioNumerator = numerator;
  _ageWeightRatioDenominator = denominator;
  if (_ageWeightRatioNumerator < 0) _ageWeightRatioNumerator = 0;
  if (_ageWeightRatioDenominator < 0) _ageWeightRatioDenominator = 0;
  if (_ageWeightRatioNumerator == 0 && _ageWeightRatioDenominator == 0) 
    _ageWeightRatioNumerator = 1;
  if (_ageWeightRatioNumerator != 0 && _ageWeightRatioDenominator != 0) 
    {
      long gcd = 
	BK::Math::greatestCommonDivisor(_ageWeightRatioNumerator,_ageWeightRatioDenominator);
      _ageWeightRatioNumerator = _ageWeightRatioNumerator/gcd;
      _ageWeightRatioDenominator = _ageWeightRatioDenominator/gcd;
    };
  if (_ageWeightRatioDenominator > 0) 
    _magicCoefficient = 
      1.0 + 
      (((float)_ageWeightRatioNumerator)/((float)_ageWeightRatioDenominator));
}; 
void PassiveClauseSet::setBestGoodWeightRatio(long numerator,
					      long denominator)
{
  _bestGoodWeightNumerator = numerator;
  _bestGoodWeightDenominator = denominator;
  if (_bestGoodWeightNumerator <= 0L) _bestGoodWeightNumerator = 1L;
  if (_bestGoodWeightDenominator < 0L) _bestGoodWeightDenominator = 0L;
};
Clause* PassiveClauseSet::Next() 
{
  if (!(EmptyCat(0)))
    {
      Clause* res = _category[0].Next();
      Clause::Remove(res);
      StatUsed(0); 
      return res;
    };
  if (_ageWeightRatioDenominator > 0)
    {
      if (EmptyCat(_minNonemptyCatNum))
	{
	  return 0; 
	};
      if ((_totalSelectedByAgeOrWeight % 
	   (_ageWeightRatioNumerator + _ageWeightRatioDenominator)) <
	  _ageWeightRatioDenominator)
	{
	  long concession = 0UL;
	  long tmp = _totalSelectedByWeight;
	  while ((tmp % (_bestGoodWeightNumerator + _bestGoodWeightDenominator) >=
		  _bestGoodWeightNumerator) &&
		 _minNonemptyCatNum + concession + 1 < NumberOfCategories)
	    {
	      ++concession;
	      tmp = tmp / (_bestGoodWeightNumerator + _bestGoodWeightDenominator);
	    };
	  long cat = _minNonemptyCatNum + concession;
	  while (EmptyCat(cat))
	    {
	      ASSERT(cat > _minNonemptyCatNum);
	      --cat;
	    };
	  ASSERT(!EmptyCat(cat));
	  Clause* res = _category[cat].Next();
	  Clause::Remove(res);
	  StatUsed(cat); 
	  ++_totalSelectedByAgeOrWeight;
	  ++_totalSelectedByWeight;
	  return res;
	}
      else 
	{
	  unsigned long oldestNum = _category[_minNonemptyCatNum].Next()->Number();
	  long oldestCat = _minNonemptyCatNum;
	  for (long cat = _minNonemptyCatNum + 1; cat<=_maxNonemptyCatNum; cat++)
	    {
	      if ((!(EmptyCat(cat))) && (_category[cat].Next()->Number() < oldestNum))
		{
		  oldestNum = _category[cat].Next()->Number();
		  oldestCat = cat;
		};
	    };
	  ++_totalSelectedByAgeOrWeight;
	  Clause* res = _category[oldestCat].Next();
	  Clause::Remove(res);
	  StatUsed(oldestCat);
	  return res;
	};
    }
  else 
    {
      if (EmptyCat(0)) { return 0; };
      Clause* res = _category[0].Next();
      Clause::Remove(res);
      StatUsed(0);
      return res;
    };
}; 
Clause* PassiveClauseSet::TheWorstUnblockedClause()
{
  CALL("TheWorstUnblockedClause()");
  for (long cat = _maxNonemptyCatNum; cat >= _minNonemptyCatNum; cat--) 
    if (!(EmptyCat(cat))) 
      for (Clause* cand = _category[cat].Previous(); 
	   cand != &(_category[cat]); 
	   cand = cand->Previous())
	if (!cand->isProtected()) 
	  {
	    return cand;
	  };
  return 0;   
}; 
Clause* PassiveClauseSet::theWorstUnblockedClauseAfter(Clause* cl)
{
  CALL("theWorstUnblockedClauseAfter(Clause* cl)");
  long cat = Category(cl);
  Clause* cand = cl;
 previous:
  cand = cand->Previous();
  if (cand == &(_category[cat]))
    {
    try_smaller_cat:
      if (cat <= _minNonemptyCatNum) return 0;
      --cat;
      if (EmptyCat(cat)) goto try_smaller_cat;
      cand = _category[cat].Previous();
      ASSERT(cand != &(_category[cat]));
    };
  if (!cand->isProtected()) 
    {
      return cand;
    };
  goto previous;
}; 
Clause* PassiveClauseSet::TheMostUnreachableClause()
{
  CALL("TheMostUnreachableClause()");
  if (!UnreachableKept()) return 0;
  long cat;
  ulong youngestNumber = 0UL;
  for (cat = _maxNonemptyCatNum;
       (cat >= _minNonemptyCatNum) && (!definitelyReachable(cat));
       --cat)
    {
      if (!(EmptyCat(cat))) 
	{
	  if (_category[cat].Previous()->Number() > youngestNumber)
	    youngestNumber = _category[cat].Previous()->Number();
	};
    };
  ++cat;
  if (!youngestNumber) return 0;
  while (cat <= _maxNonemptyCatNum)
    {
      if (!(EmptyCat(cat))) 
	{
	  if (_category[cat].Previous()->Number() == youngestNumber)
	    return _category[cat].Previous();
	};
      ++cat;
    };
  ICP("END");
  return 0;
}; 
void PassiveClauseSet::NormalizeMinMaxCat(long lb,long rb)
{
  _minNonemptyCatNum = lb;
  _maxNonemptyCatNum = rb;
  while ((EmptyCat(_minNonemptyCatNum)) && (_minNonemptyCatNum < _maxNonemptyCatNum)) _minNonemptyCatNum++;
  while ((EmptyCat(_maxNonemptyCatNum)) && (_minNonemptyCatNum < _maxNonemptyCatNum)) _maxNonemptyCatNum--;
};
void PassiveClauseSet::StatReset(long cat) 
{
  currently_kept[cat] = 0L;
  used[cat] = 0L;
  removed[cat] = 0L;
};
void PassiveClauseSet::StatInserted(long cat)
{
  currently_kept[cat]++;
  num_of_currently_kept++;    
  if (cat > _maxNonemptyCatNum) { NormalizeMinMaxCat(_minNonemptyCatNum,cat); }
  else
    if (cat < _minNonemptyCatNum) NormalizeMinMaxCat(cat,_maxNonemptyCatNum);
};
void PassiveClauseSet::StatUsed(long cat)
{
  used[cat]++;
  currently_kept[cat]--;
  num_of_used++;     
  num_of_currently_kept--;
  numOfReachable--; 
  if (EmptyCat(cat) && ((cat == _minNonemptyCatNum) || (cat == _maxNonemptyCatNum)))
    NormalizeMinMaxCat(_minNonemptyCatNum,_maxNonemptyCatNum);
};
void PassiveClauseSet::StatRemoved(long cat)
{
  removed[cat]++;
  currently_kept[cat]--;
  num_of_removed++;
  num_of_currently_kept--;
  numOfReachable--; 
  if (EmptyCat(cat) && ((cat == _minNonemptyCatNum) || (cat == _maxNonemptyCatNum)))
    NormalizeMinMaxCat(_minNonemptyCatNum,_maxNonemptyCatNum);
};
void PassiveClauseSet::StatCancelled(long cat)
{
  num_of_cancelled++;
  currently_kept[cat]--;
  num_of_currently_kept--;
  if (EmptyCat(cat) && ((cat == _minNonemptyCatNum) || (cat == _maxNonemptyCatNum)))
    NormalizeMinMaxCat(_minNonemptyCatNum,_maxNonemptyCatNum);     
};
long PassiveClauseSet::NumOfLighter(long cat)
{
  long res = 0L;
  for (long c = 0; c < cat; c++) res += currently_kept[c]; 
  return res;
};
bool PassiveClauseSet::Reachable(long cat)
{
  CALL("Reachable(long cat)");
  if (numOfReachable > num_of_currently_kept) return true;   
  long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
  if (!num_of_lighter_or_same_weight) return true; 
  if (numOfReachable <= num_of_lighter_or_same_weight) return false;
  long magic_number = (long)(num_of_lighter_or_same_weight * _magicCoefficient);
  return (numOfReachable - (num_of_currently_kept - numOfReachable) > magic_number);
}; 
bool PassiveClauseSet::definitelyReachable(long cat)
{
  CALL("definitelyReachable(long cat)");   
  if (numOfReachable > num_of_currently_kept) return true; 
  long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
  if (!num_of_lighter_or_same_weight) return true; 
  if (numOfReachable <= num_of_lighter_or_same_weight) return false;
  long magic_number = (long)(num_of_lighter_or_same_weight * _magicCoefficient);
  DOP(bool debugRes = (magic_number < numOfReachable));
  ASSERT((!debugRes) || potentiallyReachable(cat));
  return (magic_number < numOfReachable); 
}; 
bool PassiveClauseSet::potentiallyReachable(long cat)
{
  CALL("potentiallyReachable(long cat)");   
  if (numOfReachable > num_of_currently_kept) return true; 
  long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
  if (!num_of_lighter_or_same_weight) return true; 
  if (numOfReachable <= num_of_lighter_or_same_weight) return false;
  return true; 
}; 
void PassiveClauseSet::CorrectMaxAcceptedCat1()
{
  if (num_of_currently_kept < numOfReachable) 
    {
      max_accepted_cat = NumberOfCategories-1;
      return;
    };
  if (definitelyReachable(max_accepted_cat))
    {
      while (definitelyReachable(max_accepted_cat + 1))
	{
	  max_accepted_cat++;
	  if (max_accepted_cat > _maxNonemptyCatNum) 
	    {
	      max_accepted_cat = NumberOfCategories-1;
	      return;
	    };
	};
    }
  else
    {
      do
	{
	  max_accepted_cat--;
	}
      while ((max_accepted_cat > 0) && (!(definitelyReachable(max_accepted_cat))));
    };     
}; 
void PassiveClauseSet::CorrectMaxAcceptedCat2()
{
  if (num_of_currently_kept < numOfReachable) 
    {
      max_accepted_cat = NumberOfCategories-1;
      return;
    };
  if (potentiallyReachable(max_accepted_cat))
    {
      while (potentiallyReachable(max_accepted_cat + 1))
	{
	  max_accepted_cat++;
	  if (max_accepted_cat > _maxNonemptyCatNum) 
	    {
	      max_accepted_cat = NumberOfCategories-1;
	      return;
	    };
	};
    }
  else
    {
      do
	{
	  max_accepted_cat--;
	}
      while ((max_accepted_cat > 0) && (!(potentiallyReachable(max_accepted_cat))));
    };     
}; 
void PassiveClauseSet::CorrectMaxAcceptedCat3()
{
  if (num_of_currently_kept < numOfReachable) 
    {
      max_accepted_cat = NumberOfCategories - 1;
      return;
    };
  if (Reachable(max_accepted_cat))
    {
      while (Reachable(max_accepted_cat + 1))
	{
	  max_accepted_cat++;
	  if (max_accepted_cat > _maxNonemptyCatNum) 
	    {
	      max_accepted_cat = NumberOfCategories - 1;
	      return;
	    };
	};
    }
  else
    {
      do
	{
	  max_accepted_cat--;
	}
      while ((max_accepted_cat > 0) && (!(Reachable(max_accepted_cat))));
    };     
}; 
void PassiveClauseSet::ResetIterator()
{
  if (EmptyCat(_minNonemptyCatNum)) { next_clause = 0; return; };
  iter_cat = _minNonemptyCatNum;
  next_clause = _category[iter_cat].Next();
};
Clause* PassiveClauseSet::IterNext() 
{
  Clause* res = next_clause;     
  if (res)
    {
      next_clause = res->Next();
      if (next_clause == &(_category[iter_cat]))
	{
	  iter_cat++;       
	  while ((iter_cat <= _maxNonemptyCatNum) && (EmptyCat(iter_cat))) { iter_cat++; };
	  if (iter_cat > _maxNonemptyCatNum) 
	    { 
	      next_clause = 0; 
	    }
	  else { next_clause = _category[iter_cat].Next(); };
	};
    };
  return res;
}; 
ostream& PassiveClauseSet::outputProfile(ostream& str) const
{
  CALL("outputProfile(ostream& str) const");
  str << "%============ Profile of Passive (weight:stored:oldest):";
  ulong columns = 0UL;
  for (long cat = _minNonemptyCatNum; cat <= _maxNonemptyCatNum; cat++)
    {
      if (!EmptyCat(cat))
	{
	  if (columns == 0UL) 
	    {
	      str << "\n% ";
	      columns = 5UL;
	    };
	  str << cat << ":" << currently_kept[cat] 
	      << ":" << _category[cat].Next()->Number() << "   ";
	  --columns;
	};
    };
  return
    str << "\n%======================================================\n";
}; 
void PassiveClauseSet::DestroyAllClauses()
{
  CALL("DestroyAllClauses()");
  for (long cat = _minNonemptyCatNum; cat <= _maxNonemptyCatNum; cat++)
    {
      while (!(EmptyCat(cat)))
	{
	  Clause* c = _category[cat].Next();
	  Clause::Remove(c);
	  c->Destroy();
	};
    };
  Reset();  
}; 
