/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PASSIVE_CLAUSE_SET_HPP
#define PASSIVE_CLAUSE_SET_HPP
#include <climits>
#include <iostream>
#include "jargon.hpp"
#include "Math.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Clause.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PASSIVE_CLAUSE_SET
#define DEBUG_NAMESPACE "PassiveClauseSet"
#endif
#include "debugMacros.hpp"
namespace VK
{
class PassiveClauseSet
{
 public:
  PassiveClauseSet() : 
    _ageWeightRatioNumerator(1),
    _ageWeightRatioDenominator(0),
    _bestGoodWeightNumerator(1),
    _bestGoodWeightDenominator(0),
    _magicCoefficient(0.0)
  { 
    Reset(); 
  };
  ~PassiveClauseSet() { CALL("destructor ~PassiveClauseSet()"); }; 
  void Reset();
  void DestroyAllClauses();
  void setAgeWeightRatio(long numerator,long denominator);
  void setBestGoodWeightRatio(long numerator,long denominator);
  void maxPriorityOn() { _maxPriorityFlag = true; };
  void maxPriorityOff() { _maxPriorityFlag = false; }; 
  bool isEmpty() const { return !num_of_currently_kept; };
  long Size() const { return num_of_currently_kept; };
  long Total() const { return num_of_currently_kept + num_of_used + num_of_removed + num_of_cancelled; };
  long CurrentlyKept() const { return num_of_currently_kept; };
  long UsedAndRemoved() const { return (num_of_used + num_of_removed); };
  long UnreachableKept() const
    {
      return (num_of_currently_kept > numOfReachable)? (num_of_currently_kept - numOfReachable) : 0; 
    };
  void SetNumOfReachable1(long n) 
    {
      CALL("SetNumOfReachable1(const long n)");
      numOfReachable = n; 
      CorrectMaxAcceptedCat1();
    };  
  void SetNumOfReachable2(long n) 
    {
      CALL("SetNumOfReachable2(const long n)");
      numOfReachable = n; 
      CorrectMaxAcceptedCat2();
    };
  void SetNumOfReachable3(long n) 
    {
      CALL("SetNumOfReachable3(const long n)");
      numOfReachable = n; 
      CorrectMaxAcceptedCat3();
    };
  long MaxAcceptedWeight() const
    {
      if (num_of_currently_kept < numOfReachable) return VampireKernelConst::MaxClauseSize; 
      if (_ageWeightRatioDenominator > 0) { return Weight(max_accepted_cat); }
      else return 0L; 
    };
  long MaxPresentWeight() const { return Weight(_maxNonemptyCatNum); };
  long MinPresentWeight() const { return Weight(_minNonemptyCatNum); };
  void Cancel(Clause* c) 
  { 
    CALL("Cancel(Clause* c)");
    Clause::Remove(c); 
    StatCancelled(Category(c)); 
  };
  void Remove(Clause* c) 
    { 
      CALL("Remove(Clause* c)");
      Clause::Remove(c); 
      StatRemoved(Category(c)); 
    };
  void Insert(Clause* c) 
    { 
      if (_maxPriorityFlag) 
	{ 
	  c->AssignMaxPriority();
	};
      Clause* previous_cl = _category[Category(c)].Previous();
      previous_cl->SetNext(c);
      c->SetPrevious(previous_cl);
      c->SetNext(&(_category[Category(c)]));
      _category[Category(c)].SetPrevious(c);
      StatInserted(Category(c));
    };
  Clause* Next();
  Clause* TheWorstUnblockedClause();
  Clause* theWorstUnblockedClauseAfter(Clause* cl);
  Clause* TheMostUnreachableClause();
  void ResetIterator();
  Clause* IterNext();
  ostream& outputProfile(ostream& str) const;
private:
  enum 
  {
    NumberOfCategories = VampireKernelConst::MaxClauseSize + 1
  };
 private:   
  void NormalizeMinMaxCat(long lb,long rb);
  void StatReset(long cat);
  void StatInserted(long cat);
  void StatUsed(long cat);
  void StatRemoved(long cat);
  void StatCancelled(long cat);
  long Total(long cat) const { 
    return currently_kept[cat] + used[cat] + removed[cat]; 
  };
  long NumOfLighter(long cat);
  bool Reachable(long cat);
  bool definitelyReachable(long cat);
  bool potentiallyReachable(long cat);
  void CorrectMaxAcceptedCat1();
  void CorrectMaxAcceptedCat2();
  void CorrectMaxAcceptedCat3();
  long Category(long weight) 
    { 
      if (_ageWeightRatioDenominator > 0)
	{
	  if (weight < 0L) return 0L;
	  if (weight >= NumberOfCategories) return NumberOfCategories - 1;
          return weight;
	};
      return 0L;
    };
  long Category(Clause* c) 
    {
      if (c->MaxPriority()) return 0; 
      return Category(c->weight());
    };
  long Weight(long cat) const { return cat; };
  bool EmptyCat(long cat) const { 
    return (_category[cat].Next() == &(_category[cat])); };
 private: 
  bool _maxPriorityFlag;
  long _ageWeightRatioNumerator;   
  long _ageWeightRatioDenominator;
  long _bestGoodWeightNumerator;
  long _bestGoodWeightDenominator;
  float _magicCoefficient; 
  long _totalSelectedByAgeOrWeight;
  long _totalSelectedByWeight;
  Clause _category[NumberOfCategories];
  long _minNonemptyCatNum;
  long _maxNonemptyCatNum;
  long max_accepted_cat;
 private: 
  long currently_kept[NumberOfCategories];
  long used[NumberOfCategories];
  long removed[NumberOfCategories];
  long num_of_currently_kept;
  long num_of_used;
  long num_of_removed;
  long num_of_cancelled;
  long numOfReachable;
  Clause* next_clause;
  long iter_cat; 
}; 
}; 
#endif
