/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PATH_INDEX_HPP
#define PATH_INDEX_HPP 
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "GlobAlloc.hpp"
#include "Stack.hpp"
#include "DestructionMode.hpp"
#include "VampireKernelDebugFlags.hpp" 
#include "VampireKernelConst.hpp" 
#include "Term.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEX
#  define DEBUG_NAMESPACE "PathIndex<FuncIndObj,ConstIndObj>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class FuncIndObj,class ConstIndObj>
  class PathIndex
  {
  public: 
    class Integrator;
    class Retrieval;
    class Removal;
    class Trie;
  public:
    class Node
    {
    public:
      Node(const Term& f) : _func(f), _altern(0) 
      {
	CALL("Node::(constructor Node(const Term& f))");
	ASSERT(_func.isComplex());
	ulong arity = f.arity();
	ASSERT(arity);
	Node** deeperPos = deeper();
	for (ulong argn = 0; argn < arity; argn++) deeperPos[argn] = 0;
	BK::init(indObj());
      }; 
      Node(const Term& c,bool) 
	: _func(c), _altern(0) 
      {
	CALL("Node::(constructor Node(const Term& c))");
	ASSERT(_func.isComplex());
	ASSERT(!(_func.arity()));
	BK::init(constIndObj());
      };
      ~Node() 
      {
	CALL("Node::destructor ~Node()");
	if (BK::DestructionMode::isThorough())
	  {
	    if (isConstant())
	      {
		BK::destroy(constIndObj());
	      }
	    else
	      BK::destroy(indObj());
	  };
      };
      void* operator new(size_t,ulong arity) 
      {
	CALL("operator new(size_t,ulong arity)");
	ASSERT(arity);
	return static_cast<void*>((static_cast<FuncIndObj*>(BK::GlobAlloc::allocate(sizeof(FuncIndObj) + sizeof(Node) + arity*sizeof(Node*)))) + 1);
      }; 
      void* operator new(size_t) 
      { 
	CALL("operator new(size_t)");
	return static_cast<void*>((static_cast<ConstIndObj*>(BK::GlobAlloc::allocate(sizeof(ConstIndObj) + sizeof(Node)))) + 1); 
      }; 
      void operator delete(void* obj)
      {
	CALL("operator delete(void* obj)");
	ulong arity = (static_cast<Node*>(obj))->func().arity();
	if (arity)
	  {
	    BK::GlobAlloc::deallocate(static_cast<void*>((static_cast<FuncIndObj*>(obj)) - 1),sizeof(FuncIndObj) + sizeof(Node) + arity*sizeof(Node*));
	  }
	else
	  {
	    BK::GlobAlloc::deallocate(static_cast<void*>((static_cast<ConstIndObj*>(obj)) - 1),sizeof(ConstIndObj) + sizeof(Node));
	  };
      }; 
      void operator delete(void* obj,ulong arity)
      {
	CALL("operator delete(void* obj,ulong arity)");
	ASSERT(arity);
        BK::GlobAlloc::deallocate(static_cast<void*>((static_cast<FuncIndObj*>(obj)) - 1),sizeof(FuncIndObj) + sizeof(Node) + arity*sizeof(Node*));
      }; 
#ifndef NO_DEBUG
      static ulong funcMinSize() { return sizeof(FuncIndObj) + sizeof(Node) + sizeof(Node*); };
      static ulong funcMaxSize() { return sizeof(FuncIndObj) + sizeof(Node) + Term::maxArity()*sizeof(Node*); };
      static ulong constMinSize() { return sizeof(ConstIndObj) + sizeof(Node); };
      static ulong constMaxSize() { return sizeof(ConstIndObj) + sizeof(Node); };
#endif    
      const Term& func() const { return _func; }; 
      Node*& altern() { return _altern; };
      const Node* const & altern() const { return _altern; };
      Node** alternAddr() { return &_altern; };
      const Node* const * alternAddr() const { return &_altern; };
      FuncIndObj& indObj() 
      {
	CALL("Node::indObj()");
	ASSERT(!isConstant());
	return *((static_cast<FuncIndObj*>(static_cast<void*>(this))) - 1);
      };
      ConstIndObj& constIndObj() 
      {
	CALL("Node::constIndObj()");
	ASSERT(isConstant());
	return *((static_cast<ConstIndObj*>(static_cast<void*>(this))) - 1);
      };    
      const FuncIndObj& indObj() const 
      {
	CALL("Node::indObj() const");
	ASSERT(!isConstant());
	return *((static_cast<const FuncIndObj*>(static_cast<const void*>(this))) - 1);
      };
      const ConstIndObj& constIndObj() const
      {
	CALL("Node::constIndObj() const");
	ASSERT(isConstant());
	return *((static_cast<const ConstIndObj*>(static_cast<const void*>(this))) - 1);
      };
      const FuncIndObj* indObjAddr() const 
      {
	CALL("Node::indObjAddr() const");
	ASSERT(!isConstant());
	return (static_cast<const FuncIndObj*>(static_cast<const void*>(this))) - 1;
      };
      static Node* node(const FuncIndObj* indObjAddr)
      {
	return static_cast<Node*>(static_cast<void*>(indObjAddr + 1));
      };
      const ConstIndObj* constIndObjAddr() const 
      {
	CALL("Node::constIndObjAddr() const");
	ASSERT(isConstant());
	return (static_cast<const ConstIndObj*>(static_cast<const void*>(this))) - 1;
      };
      static const Node* node(const ConstIndObj* constIndObjAddr)
      {
	return static_cast<const Node*>(static_cast<void*>(constIndObjAddr + 1));
      };
      Node** deeper() 
      {
	CALL("Node::deeper()");
	ASSERT(isNonconst());
	return static_cast<Node**>(static_cast<void*>(this + 1));
      };   
      const Node* const * deeper() const 
      {
	CALL("Node::deeper() const");
	ASSERT(isNonconst());
	return static_cast<const Node* const *>(static_cast<const void*>(this + 1));
      };
      ulong arity() const { return func().arity(); };
      bool isConstant() const { return !(func().arity()); };
      bool isNonconst() const { return func().arity() != 0UL; };
#ifndef NO_DEBUG
      Node*& debugPreviousClosedPath() { return _debugPreviousClosedPath; };
      const Node* debugPreviousClosedPath() const { return _debugPreviousClosedPath; };
      ulong& debugPreviousArgNum() { return _debugPreviousArgNum; };
      ulong debugPreviousArgNum() const { return _debugPreviousArgNum; };
#endif    
    private:
      Term _func;
      Node* _altern;
#ifndef NO_DEBUG
      Node* _debugPreviousClosedPath;
      ulong _debugPreviousArgNum;
#endif    
    }; 
  public:
    class Trie
    {
    public:
      Trie() : _content(0) {};
      ~Trie() 
      {
	CALL("destructor ~Trie()");
	if (BK::DestructionMode::isThorough())
	  {
	    destroyTree(_content);
	  };
      };
      operator bool() const { return (bool)_content; };
    private:
      Node** contentAddr() { return &_content; };   
    private:
      Node* _content;
      friend class Integrator;
      friend class Retrieval;
      friend class Removal;
    }; 
    class Integrator
    {
    public:
      Integrator() :
	_backtrackNodeAddr("PathIndex::Integrator::_backtrackNodeAddr")
#ifndef NO_DEBUG
	, _debugBacktrackPrevNode("PathIndex::Integrator::_debugBacktrackPrevNode"),
	_debugBacktrackPrevArgNum("PathIndex::Integrator::_debugBacktrackPrevArgNum")
#endif
      {
	CALL("constructor Integrator()");
      };
      ~Integrator() 
      {
	CALL("destructor ~Integrator()");
      };
      void reset(Trie& tree)
      { 
	_nodeAddr = tree.contentAddr();
	_backtrackNodeAddr.reset();
#ifndef NO_DEBUG
	_debugPreviousNode = 0;
	_debugBacktrackPrevNode.reset();
	_debugBacktrackPrevArgNum.reset();
#endif
      }; 
      void nonconstFunc(const Term& f)
      {
	CALL("nonconstFunc(const Term& f)");
	ASSERT(f.isComplex());
	ulong arity = f.arity();
	ASSERT(arity);
      check_node:
	if (!(*_nodeAddr)) 
	  {
	    _lastNode = new (arity) Node(f);
#ifndef NO_DEBUG
	    _lastNode->debugPreviousClosedPath() = _debugPreviousNode;
	    _lastNode->debugPreviousArgNum() = _debugPreviousArgNum;
#endif
	    *_nodeAddr = _lastNode;
	  node_found:
	    _nodeAddr = _lastNode->deeper();
	    for (ulong argn = arity - 1; argn; argn--)
	      {
		_backtrackNodeAddr.push(_lastNode->deeper() + argn);
#ifndef NO_DEBUG
		_debugBacktrackPrevNode.push(_lastNode);
		_debugBacktrackPrevArgNum.push(argn);
#endif
	      };
#ifndef NO_DEBUG
	    _debugPreviousNode = _lastNode;
	    _debugPreviousArgNum = 0UL;
#endif
	    return;
	  };   
	_lastNode = *_nodeAddr;
	ASSERT(_lastNode->debugPreviousClosedPath() == _debugPreviousNode);
	if (_lastNode->func() == f) goto node_found;
	_nodeAddr = _lastNode->alternAddr();  
	goto check_node;
      }; 
      void constant(const Term& c)
      {
	CALL("constant(const Term& c)");    
	ASSERT(c.isComplex());
	ASSERT(!c.arity());
      check_node:
	if (!(*_nodeAddr)) 
	  {
	    _lastNode = new Node(c,true); 
#ifndef NO_DEBUG
	    _lastNode->debugPreviousClosedPath() = _debugPreviousNode;
	    _lastNode->debugPreviousArgNum() = _debugPreviousArgNum;
#endif   
	    *_nodeAddr = _lastNode;
	  node_found: 
	    backtrack();
	    return;
	  };
	_lastNode = *_nodeAddr;
	if (_lastNode->func() == c) goto node_found;
	ASSERT(_lastNode->debugPreviousClosedPath() == _debugPreviousNode);  
	_nodeAddr = _lastNode->alternAddr();  
	goto check_node;
      }; 
      void variable()
      {
	CALL("variable()");
	backtrack();  
      };
      FuncIndObj& indexedObj() const { return _lastNode->indObj(); };
      ConstIndObj& constIndObj() const { return _lastNode->constIndObj(); };
    private:
      void backtrack()
      {
	CALL("backtrack()");
	if (_backtrackNodeAddr) 
	  {
	    _nodeAddr = _backtrackNodeAddr.pop();
#ifndef NO_DEBUG
	    _debugPreviousNode = _debugBacktrackPrevNode.pop();
	    _debugPreviousArgNum = _debugBacktrackPrevArgNum.pop();
#endif
	  };
      }; 
    private:
      Node** _nodeAddr;
      Node* _lastNode;
      BK::Stack<Node**,VampireKernelConst::MaxTermSize> _backtrackNodeAddr;
#ifndef NO_DEBUG
      Node* _debugPreviousNode;
      ulong _debugPreviousArgNum;
      BK::Stack<Node*,VampireKernelConst::MaxTermSize> _debugBacktrackPrevNode;
      BK::Stack<ulong,VampireKernelConst::MaxTermSize> _debugBacktrackPrevArgNum;
#endif
    }; 
    class Retrieval
    {
    public:
      Retrieval() :
	_backtrackNodeAddr("PathIndex::Retrieval::_backtrackNodeAddr")
      {
	CALL("Retrieval::constructor Retrieval()");
      };
      ~Retrieval() 
      {
	CALL("Retrieval::destructor ~Retrieval()");
      };
      bool reset(Trie& tree)
      {     
	CALL("Retrieval::reset(Trie& tree)");
	_nodeAddr = tree.contentAddr();
	_backtrackNodeAddr.reset();
	DOP(debugEndOfTerm = false);
	return (*_nodeAddr) != 0;
      }; 
      bool nonconstFunc(const Term& f)
      {
	CALL("Retrieval::nonconstFunc(const Term& f)");
	ASSERT(!debugEndOfTerm);
	ASSERT(f.arity());
      check_node: 
	if (!(*_nodeAddr)) return false;
	_lastNode = *_nodeAddr;  
	if (_lastNode->func() == f)
	  {
	    _nodeAddr = _lastNode->deeper();
	    ulong arity = f.arity();
	    for (ulong argn = arity - 1; argn; argn--) _backtrackNodeAddr.push(_lastNode->deeper() + argn); 
	    return true; 
	  };
	_nodeAddr = _lastNode->alternAddr();
	goto check_node;
      }; 
      bool constant(const Term& c)
      {
	CALL("Retrieval::constant(const Term& c)"); 
	ASSERT(!debugEndOfTerm);   
	ASSERT(!c.arity());
      check_node: 
	if (!(*_nodeAddr)) return false;
	_lastNode = *_nodeAddr;  
	if (_lastNode->func() == c)
	  {
	    backtrack();
	    return true; 
	  };
	_nodeAddr = _lastNode->alternAddr();
	goto check_node;
      }; 
      void after() { CALL("Retrieval::after()"); ASSERT(!debugEndOfTerm); backtrack(); };
      const FuncIndObj& indObj() const { return _lastNode->indObj(); };
      const ConstIndObj& constIndObj() const { return _lastNode->constIndObj(); };  
      const FuncIndObj* indObjAddr() const { return _lastNode->indObjAddr(); };
      const ConstIndObj* constIndObjAddr() const { return _lastNode->constIndObjAddr(); };  
    private:
      void backtrack()
      {
	CALL("Retrieval::backtrack()");
	if (_backtrackNodeAddr) { _nodeAddr = _backtrackNodeAddr.pop(); }
	else 
	  {
	    DOP(debugEndOfTerm = true);
	  };     
      }; 
    private:
      Node** _nodeAddr;
      Node* _lastNode;
      BK::Stack<Node**,VampireKernelConst::MaxTermSize> _backtrackNodeAddr;
#ifdef DEBUG_NAMESPACE
      bool debugEndOfTerm;  
#endif
    }; 
    class Removal
    {
    public:
      Removal() :
	_backtrackNodeAddr("PathIndex::Removal::_backtrackNodeAddr"),
	_candidatesForRecycling("PathIndex::Removal::_candidatesForRecycling")
      {
	CALL("Removal::constructor Removal()");
      };
      ~Removal() 
      {
	CALL("Removal::destructor ~Removal()");
      };
      void reset(Trie& tree)
      {     
	CALL("Removal::reset(Trie& tree)");
	_nodeAddr = tree.contentAddr();
	_backtrackNodeAddr.reset();
	_candidatesForRecycling.reset(); 
	DOP(debugEndOfTerm = false);
      }; 
      bool nonconstFunc(const Term& f)
      {
	CALL("Removal::nonconstFunc(const Term& f)");
	ASSERT(!debugEndOfTerm);
	ASSERT(f.arity());
      check_node: 
	if (!(*_nodeAddr)) return false;
	_lastNode = *_nodeAddr;  
	if (_lastNode->func() == f)
	  {
	    _candidatesForRecycling.push(_nodeAddr);
	    _nodeAddr = _lastNode->deeper();
	    ulong arity = f.arity();
	    for (ulong argn = arity - 1; argn; argn--) _backtrackNodeAddr.push(_lastNode->deeper() + argn); 
	    return true; 
	  };
	_nodeAddr = _lastNode->alternAddr();
	goto check_node;
      }; 
      bool constant(const Term& c)
      {
	CALL("Removal::constant(const Term& c)"); 
	ASSERT(!debugEndOfTerm);   
	ASSERT(!c.arity());
      check_node: 
	if (!(*_nodeAddr)) return false;
	_lastNode = *_nodeAddr;  
	if (_lastNode->func() == c)
	  {
	    _candidatesForRecycling.push(_nodeAddr);
	    backtrack();
	    return true; 
	  };
	_nodeAddr = _lastNode->alternAddr();
	goto check_node;
      }; 
      void after() { CALL("Removal::after()"); ASSERT(!debugEndOfTerm); backtrack(); }; 
      FuncIndObj& indexedObj() const 
      { 
	CALL("Removal::indexedObj() const");   
	return _lastNode->indObj(); 
      };
      ConstIndObj& constIndObj() const 
      { 
	CALL("Removal::constIndObj() const"); 
	return _lastNode->constIndObj(); 
      };  
      void recycleObsoleteNodes()
      {
	CALL("Removal::recycleObsoleteNodes()"); 
	while (_candidatesForRecycling.nonempty()) recycleObsolete(*(_candidatesForRecycling.pop()));
      };
    private:
      void backtrack()
      {
	CALL("Removal::backtrack()");
	if (_backtrackNodeAddr) { _nodeAddr = _backtrackNodeAddr.pop(); }
	else 
	  {
	    DOP(debugEndOfTerm = true);
	  };
      }; 
    private:
      Node** _nodeAddr;
      Node* _lastNode;
      BK::Stack<Node**,VampireKernelConst::MaxTermSize> _backtrackNodeAddr;
      BK::Stack<Node**,VampireKernelConst::MaxTermSize> _candidatesForRecycling;
#ifdef DEBUG_NAMESPACE
      bool debugEndOfTerm;  
#endif
    }; 
  public: 
#ifndef NO_DEBUG_VIS
    static ostream& outputPath(ostream& str,const Node* nd);
#endif
  private: 
    static void destroyTree(Node*& tree);
    static void recycleObsolete(Node*& node);
    friend class Removal;
    friend class Trie;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEX
#define DEBUG_NAMESPACE "PathIndex<FuncIndObj,ConstIndObj>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class FuncIndObj,class ConstIndObj>
  inline void PathIndex<FuncIndObj,ConstIndObj>::destroyTree(Node*& tree)
  {   
    CALL("destroyTree(Node*& tree)");
    if (tree)
      {
	destroyTree(tree->altern());   
	if (tree->isNonconst())
	  {             
	    for (ulong a = 0; a < tree->arity(); a++) 
	      {
		destroyTree(tree->deeper()[a]);
	      };
	  };
	delete tree;
	tree = 0;
      };
  }; 
  template <class FuncIndObj,class ConstIndObj>
  inline void PathIndex<FuncIndObj,ConstIndObj>::recycleObsolete(Node*& node)
  {
    CALL("recycleObsolete(Node*& node)");
    Node* tmp;
    if (node) 
      {
	if (node->isConstant()) 
	  {
	    if (!node->constIndObj()) 
	      {
		tmp = node;       
		node = tmp->altern();
		delete tmp;
	      };
	  }
	else 
	  { 
	    if (!node->indObj())
	      {
		ulong arity = node->arity();
		for (ulong argn = 0; argn < arity; argn++) 
		  if (node->deeper()[argn]) return;  
		tmp = node;       
		node = tmp->altern();
		delete tmp; 
	      };
	  };
      };
  }; 
#ifndef NO_DEBUG_VIS
  template <class FuncIndObj,class ConstIndObj>
  inline 
  ostream& PathIndex<FuncIndObj,ConstIndObj>::outputPath(ostream& str,const Node* nd)
  {
    if (!nd) return str;
    if (nd->debugPreviousClosedPath())
      {
	outputPath(str,nd->debugPreviousClosedPath());
	str << "." << nd->debugPreviousArgNum() << ".";
      }
    str << nd->func();
    return str;
  }; 
#endif
}; 
#endif
