/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "PremiseWithFactoring.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PREMISE_WITH_FACTORING
 #define DEBUG_NAMESPACE "PremiseWithFactoring"
#endif
#include "debugMacros.hpp"
using namespace VK;
PremiseWithFactoring::PremiseWithFactoring() :
  _factoredLits(DOP("_factoredLits")),
  _rotationRequired(DOP("_rotationRequired"))
{
  CALL("constructor PremiseWithFactoring()"); 
  DOP(_factoredLits.freeze());
  DOP(_rotationRequired.freeze());
}; 
PremiseWithFactoring::~PremiseWithFactoring() 
{ 
  CALL("destructor ~PremiseWithFactoring()"); 
};
bool PremiseWithFactoring::GetFirstFactor(ulong first_candidate)
{
  CALL("GetFirstFactor(ulong first_candidate)");
  if (lastSelected() != 0UL)
    {
      bool rotationRequired;
      _factoredLits.reset();
      _rotationRequired.reset();
      if (FindFirstFactor(first_candidate,rotationRequired))
	{
	  _lastFactoredLitNum = first_candidate;
	  _rotationRequiredForLastFactored = rotationRequired;
	  was_factored = true;
	  factoriseLiteralProperties(rotationRequired);
	  ASSERT(_rotationRequired.size() == _factoredLits.size());
	  return true;
	};
    };
  return false;
}; 
bool PremiseWithFactoring::GetNextFactor()
{
  CALL("GetNextFactor()");
  ulong candidate = _lastFactoredLitNum + 1;
  bool rotationRequired;
 next_factor:
  if (FindFirstFactor(candidate,rotationRequired))
    {
      _factoredLits.push(_lastFactoredLitNum);
      _rotationRequired.push(_rotationRequiredForLastFactored);
      _lastFactoredLitNum = candidate;
      _rotationRequiredForLastFactored = rotationRequired;
      factoriseLiteralProperties(rotationRequired);
      ASSERT(_rotationRequired.size() == _factoredLits.size());
      return true;
    }
  else 
    {
      Unifier::current()->UnbindLastSection();
      if (_rotationRequiredForLastFactored)
	{
	  _rotationRequiredForLastFactored = false;
	  if (_mainLiteralRotation)
	    {
	      if (Unification::UnifyComplex(RulePremise::_selectedLiteral[_lastFactoredLitNum],
					    _mainLiteralRotation))
		{
		  ASSERT(_rotationRequired.size() == _factoredLits.size());
		  return true;
		};
	    }
	  else
	    {
	      ASSERT(RulePremise::_selectedLiteralRotation[_lastFactoredLitNum]);
	      if (Unification::UnifyComplex(RulePremise::_selectedLiteralRotation[_lastFactoredLitNum],
					    _mainLiteral))
		{
		  ASSERT(_rotationRequired.size() == _factoredLits.size());
		  return true;
		};
	    };
	};
      RulePremise::_litIsMarked[_lastFactoredLitNum] = false;
      candidate = _lastFactoredLitNum + 1;
      if (_factoredLits.empty())
	{
	  ASSERT(_rotationRequired.empty());
	  if (FindFirstFactor(candidate,rotationRequired))
	    {
	      _lastFactoredLitNum = candidate;
	      _rotationRequiredForLastFactored = rotationRequired;
	      factoriseLiteralProperties(rotationRequired);
	      return true;
	    }
	  else 
	    return false;
	}
      else 
	{
	  _lastFactoredLitNum = _factoredLits.pop();
	  _rotationRequiredForLastFactored = _rotationRequired.pop();
	  goto next_factor;        
	};
    };
}; 
bool PremiseWithFactoring::FindFirstFactor(ulong& candidate,bool& rotationRequired)
{
  CALL("FindFirstFactor(ulong& candidate,bool& rotationRequired)");
  while (candidate <= _lastSelected)
    {
      if (candidate != cut_lit)
	{
	  if (Unification::UnifyComplex(RulePremise::_selectedLiteral[candidate],
					_mainLiteral))  
	    {
	      RulePremise::_litIsMarked[candidate] = true;
	      rotationRequired = 
		_mainLiteralRotation ||
		RulePremise::_selectedLiteralRotation[candidate];
	      return true;
	    };
	  if (_mainLiteralRotation)
	    {
	      if (Unification::UnifyComplex(RulePremise::_selectedLiteral[candidate],
					    _mainLiteralRotation))
		{		  
		  RulePremise::_litIsMarked[candidate] = true;
		  rotationRequired = false;
		  return true;
		};
	    }
	  else if (RulePremise::_selectedLiteralRotation[candidate])
	    {
	      if (Unification::UnifyComplex(RulePremise::_selectedLiteralRotation[candidate],
					    _mainLiteral))
		{		  
		  RulePremise::_litIsMarked[candidate] = true;
		  rotationRequired = false;
		  return true;
		};
	    };
	};
      ++candidate;
    };
  return false;
}; 
inline
void PremiseWithFactoring::factoriseLiteralProperties(bool symmetryUsed)
{
  CALL("factoriseLiteralProperties(bool symmetryUsed)");
  _resolvedLiteralProperties = _selectedLiteralProperties[_lastFactoredLitNum];
  for (ulong* lit = _factoredLits.begin();
       lit != _factoredLits.end();
       ++lit)
    _resolvedLiteralProperties.factoriseWith(_selectedLiteralProperties[*lit],symmetryUsed);
}; 
