/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PREMISE_WITH_FACTORING_HPP
#define PREMISE_WITH_FACTORING_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "RulePremise.hpp"
#include "ExpandingStack.hpp"
#include "GlobAlloc.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PREMISE_WITH_FACTORING
 #define DEBUG_NAMESPACE "PremiseWithFactoring"
#endif
#include "debugMacros.hpp"
namespace VK
{
class PremiseWithFactoring : public RulePremise
 {
  public:
   PremiseWithFactoring();
   ~PremiseWithFactoring();
   bool Load(ulong ind,Clause* c)
   {
    CALL("Load(ulong ind,Clause* c)");
    if (!RulePremise::load(ind,c)) return false;
    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(c->numOfAllLiterals()))
      return false;
    DOP(_factoredLits.freeze());
    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(c->numOfAllLiterals()))
      return false;
    DOP(_rotationRequired.freeze());
    return true;
   };
   bool LoadOptimized(ulong ind,Clause* c)
   {
    CALL("LoadOptimized(ulong ind,Clause* c)");
    if (!RulePremise::loadOptimized(ind,c)) return false;
    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(c->numOfAllLiterals()))
      return false;    
    DOP(_factoredLits.freeze());
    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(c->numOfAllLiterals()))
      return false;    
    DOP(_rotationRequired.freeze());
    return true;
   };
   bool Load(ulong ind,UnifIndex::IndexedObject* res_pair)
   {
    CALL("Load(ulong ind,UnifIndex::IndexedObject* res_pair)");
    if (!RulePremise::load(ind,res_pair)) return false;
    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_factoredLits.freeze());
    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_rotationRequired.freeze()); 
    return true;   
   };
   bool LoadOptimized(ulong ind,UnifIndex::IndexedObject* res_pair)
   {
    CALL("LoadOptimized(ulong ind,UnifIndex::IndexedObject* res_pair)");
    if (!RulePremise::loadOptimized(ind,res_pair)) return false;
    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_factoredLits.freeze());
    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_rotationRequired.freeze());  
    return true; 
   };
   void ResetFactoring()
    {
     CALL("ResetFactoring()");
     was_factored = false;
     if (lastSelected() != 0UL)
      {
       _factoredLits.reset();
       _rotationRequired.reset();
       _mainLiteral = RulePremise::_selectedLiteral[cut_lit]; 
       _mainLiteralRotation = RulePremise::_selectedLiteralRotation[cut_lit]; 
      };
    };
   bool WasFactored() { return was_factored; };
   bool GetFirstFactor(ulong first_candidate);
   bool GetNextFactor();
  private:
   bool FindFirstFactor(ulong& candidate,bool& rotationRequired);
   void factoriseLiteralProperties(bool symmetryUsed);  
 private:
   BK::ExpandingStack<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,32UL> _factoredLits;
   BK::ExpandingStack<BK::GlobAlloc,bool,BK::FastRelocation<bool>,32UL> _rotationRequired;
   UnifTerm* _mainLiteral; 
   UnifTerm* _mainLiteralRotation;
   ulong _lastFactoredLitNum;
   bool _rotationRequiredForLastFactored;
   bool was_factored;
 }; 
}; 
#endif
