/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <cstdlib>
#include "Signature.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIGNATURE
#  define DEBUG_NAMESPACE "Signature"
#endif 
#include "debugMacros.hpp"
using namespace VK;
Signature* Signature::_current;
Signature::Signature() 
  : _inputVariableName(DOP("Signature::_inputVariableName")),
    _nextInputVariableNumber(0UL),
    _symbolName(static_cast<char*>(0),"Signature::_symbolName"),
    _interpretedConstantValue(DOP("Signature::_interpretedConstantValue")),
    _isInterpretedConstant(false,"Signature::_isInterpretedConstant"),
    _isEvaluableSymbol(false,"Signature::_isEvaluableSymbol"),
    _isInputSymbol(false,"Signature::_isInputSymbol"),
    _isPredicate(false,"Signature::_isPredicate"),
    _isCommutative(false,"Signature::_isCommutative"),
    _arity(ULONG_MAX,"Signature::_arity"),
    _isAnswerPredicate(false,"Signature::_isAnswerPredicate"),
    _positiveNonBasicArguments(static_cast<bool*>(0),"Signature::_positiveNonBasicArguments"),
    _negativeNonBasicArguments(static_cast<bool*>(0),"Signature::_negativeNonBasicArguments"),
    _numOfNonBasicArguments(0UL),
    _endOfInputFlag(false),
    _nextSymbolNumber(0UL),
    _numberOfSkolemSymbols(0L),
    _maxlNumOfSkolemSymbols(LONG_MAX),
    _nextSkolemConstantSuffix(0UL),
    _nextSkolemFunctionSuffix(0UL),
    _nextSkolemPredicateSuffix(0UL),
    _weight(1,"Signature::_weight"),
    _symbolPrecedence(0L,"Signature::_symbolPrecedence"),  
    _positiveHeaderRank(DefaultPositiveHeaderRank,"Signature::_positiveHeaderRank"),  
    _negativeHeaderRank(DefaultNegativeHeaderRank,"Signature::_negativeHeaderRank"),
    _minimalSymbolPrecedence(LONG_MAX),
    _maximalSymbolPrecedence(LONG_MIN),
    _someSymbolsAreEvaluable(false),
    _someConstantsAreInterpreted(false)
{
  CALL("constructor Signature()");
  ASSERT(PositivePolarity == 0UL);
  ASSERT(NegativePolarity == 1UL);
  _symbolName.sub(VampireKernelConst::UnordEqNum) = "equal"; 
  _isInputSymbol.sub(VampireKernelConst::UnordEqNum) = true;
  _isPredicate.sub(VampireKernelConst::UnordEqNum) = true;
  _isCommutative.sub(VampireKernelConst::UnordEqNum) = true;           
  _arity.sub(VampireKernelConst::UnordEqNum) = 2UL;
  setSymbolWeightConstantPart(VampireKernelConst::UnordEqNum,1);
  if (VampireKernelConst::UnordEqNum >= _nextSymbolNumber) _nextSymbolNumber = VampireKernelConst::UnordEqNum + 1;
  _symbolName.sub(VampireKernelConst::OrdEqNum) = "Equal"; 
  _isInputSymbol.sub(VampireKernelConst::OrdEqNum) = true;
  _isPredicate.sub(VampireKernelConst::OrdEqNum) = true; 
  _isCommutative.sub(VampireKernelConst::OrdEqNum) = true; 
  _arity.sub(VampireKernelConst::OrdEqNum) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::OrdEqNum,1);
  if (VampireKernelConst::OrdEqNum >= _nextSymbolNumber) 
    _nextSymbolNumber = VampireKernelConst::OrdEqNum + 1; 
  _symbolName.sub(VampireKernelConst::BuiltInTrueNum) = "True"; 
  _isInputSymbol.sub(VampireKernelConst::BuiltInTrueNum) = true;
  _isPredicate.sub(VampireKernelConst::BuiltInTrueNum) = true; 
  _isCommutative.sub(VampireKernelConst::BuiltInTrueNum) = false; 
  _arity.sub(VampireKernelConst::BuiltInTrueNum) = 0UL;
  setSymbolWeightConstantPart(VampireKernelConst::BuiltInTrueNum,1);
  if (VampireKernelConst::BuiltInTrueNum >= _nextSymbolNumber) 
    _nextSymbolNumber = VampireKernelConst::BuiltInTrueNum + 1; 
  _symbolName.sub(VampireKernelConst::ReservedPred0) = "ReservedPred0"; 
  _isPredicate.sub(VampireKernelConst::ReservedPred0) = true; 
  _arity.sub(VampireKernelConst::ReservedPred0) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::ReservedPred0,1);
  if (VampireKernelConst::ReservedPred0 >= _nextSymbolNumber) _nextSymbolNumber = VampireKernelConst::ReservedPred0 + 1; 
  _symbolName.sub(VampireKernelConst::ReservedPred1) = "ReservedPred1"; 
  _isPredicate.sub(VampireKernelConst::ReservedPred1) = true; 
  _arity.sub(VampireKernelConst::ReservedPred1) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::ReservedPred1,1);
  if (VampireKernelConst::ReservedPred1 >= _nextSymbolNumber) _nextSymbolNumber = VampireKernelConst::ReservedPred1 + 1; 
  _symbolName.sub(VampireKernelConst::ReservedPred2) = "ReservedPred2"; 
  _isPredicate.sub(VampireKernelConst::ReservedPred2) = true;
  _arity.sub(VampireKernelConst::ReservedPred2) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::ReservedPred2,1);
  if (VampireKernelConst::ReservedPred2 >= _nextSymbolNumber) _nextSymbolNumber = VampireKernelConst::ReservedPred2 + 1; 
  _firstInputOrSkolemSymbolNumber = _nextSymbolNumber;
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++) 
    if (!reservedSymbolNumber(fn)) 
      {
	_symbolName.sub(fn) = "NoSymbol"; 
	_arity.sub(fn) = 0UL;
      };
  for (ulong vn = 0; vn < VampireKernelConst::MaxNumOfVariables; vn++)
    {
      _canonicalVariableName[vn] = createCanonicalVariableName(vn);
    };
}; 
Signature::~Signature() 
{
  CALL("destructor ~Signature()");
  if (BK::DestructionMode::isThorough())
    {
      for (ulong m = 0UL; m < _positiveNonBasicArguments.size(); ++m)
	if (_positiveNonBasicArguments[m])
	  BK::GlobAlloc::deallocate(static_cast<void*>(_positiveNonBasicArguments[m]),
				    arity(m) * sizeof(bool));
      for (ulong n = 0UL; n < _negativeNonBasicArguments.size(); ++n)
	if (_negativeNonBasicArguments[n])
	  BK::GlobAlloc::deallocate(static_cast<void*>(_negativeNonBasicArguments[n]),
				    arity(n) * sizeof(bool));
    };
}; 
void Signature::resetInputVariables() 
{ 
  CALL("resetInputVariables()");
  while (_nextInputVariableNumber)  
    {
      _nextInputVariableNumber--; 
      delete [] (_inputVariableName.sub(_nextInputVariableNumber));
    };
}; 
ulong Signature::inputVariableNumber(const char* varName)
{
  CALL("inputVariableNumber(const char* varName)");
  for (ulong vn = 0; vn < _nextInputVariableNumber; vn++)
    if (!(strcmp(varName,_inputVariableName.sub(vn)))) return vn;
  _inputVariableName.sub(_nextInputVariableNumber) = new char[strlen(varName) + 1];
  strcpy(_inputVariableName.sub(_nextInputVariableNumber),varName);
  _nextInputVariableNumber++;
  return _nextInputVariableNumber - 1; 
}; 
ulong Signature::inputSymbolNumber(bool predicate,const char* symName,ulong ar)
{
  CALL("inputSymbolNumber(bool predicate,const char* symName,ulong ar)");
  _symbolName.sub(_nextSymbolNumber) = new char[strlen(symName) + 1];
  _arity.sub(_nextSymbolNumber) = ar;
  strcpy(_symbolName.sub(_nextSymbolNumber),symName);
  _isInputSymbol.sub(_nextSymbolNumber) = true;
  _isPredicate.sub(_nextSymbolNumber) = predicate;   
  _isCommutative.sub(_nextSymbolNumber) = false; 
  _nextSymbolNumber++;
  ASSERT(_nextSymbolNumber <= _symbolName.size());
  ASSERT(_nextSymbolNumber <= _isInputSymbol.size());
  ASSERT(_nextSymbolNumber <= _isPredicate.size());
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
  return _nextSymbolNumber - 1;  
}; 
ulong 
Signature::interpretedConstantNumber(const InterpretedConstantValue& value)
{
  CALL("interpretedConstantNumber(..)");
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (isInterpretedConstant(fn) &&
	  _interpretedConstantValue[fn].equals(value))
	return fn;
    };
  _interpretedConstantValue.sub(_nextSymbolNumber) = value;
  _arity.sub(_nextSymbolNumber) = 0UL;
  _isInterpretedConstant.sub(_nextSymbolNumber) = true;
  _isInputSymbol.sub(_nextSymbolNumber) = !_endOfInputFlag;
  _isPredicate.sub(_nextSymbolNumber) = false;   
  _isCommutative.sub(_nextSymbolNumber) = false; 
  assignUniqueSymbolPrecedence(_nextSymbolNumber);
  _nextSymbolNumber++;
  ASSERT(_nextSymbolNumber <= _interpretedConstantValue.size());
  ASSERT(_nextSymbolNumber <= _isInputSymbol.size());
  ASSERT(_nextSymbolNumber <= _isPredicate.size());
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
  _someConstantsAreInterpreted = true;
  return _nextSymbolNumber - 1;  
}; 
void Signature::makeCommutative(ulong symId)
{
  CALL("makeCommutative(ulong symId)");
  _isCommutative.sub(symId) = true;
}; 
void Signature::makeNoncommutative(ulong symId)
{
  CALL("makeNoncommutative(ulong symId)");
  _isCommutative.sub(symId) = false;
};
void Signature::setCommutativity(const SymbolInfo* commutativityTable,
				 const SymbolInfo* symmetryTable)
{
  CALL("setCommutativity(const Clause* const *commutativityTable,const Clause* const *symmetryTable)");
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)     
    _isCommutative[fn] = 
      (commutativityTable && commutativityTable->commutativityLaw(fn)) || 
      (symmetryTable && symmetryTable->symmetryLaw(fn));
};
void Signature::markBasicArgument(bool positive,ulong symId,ulong argNum)
{
  CALL("markBasicArgument(bool positive,ulong symId,ulong argNum)");
  ASSERT(argNum < arity(symId));
  if (!reservedSymbolNumber(symId))
    {
      if (positive)
	{
	  bool* bits = _positiveNonBasicArguments.sub(symId);
	  if (bits)
	    {
	      bits[argNum] = false;
	    };
	}
      else 
	{
	  bool* bits = _negativeNonBasicArguments.sub(symId);
	  if (bits)
	    {
	      bits[argNum] = false;
	    };
	};
    };
}; 
void Signature::markNonBasicArgument(bool positive,ulong symId,ulong argNum)
{
  CALL("markNonBasicArgument(bool positive,ulong symId,ulong argNum)");  
  ASSERT(argNum < arity(symId));
  if (!reservedSymbolNumber(symId))
    {
      bool* bits;
      if (positive)
	{
	  bits = _positiveNonBasicArguments.sub(symId);
	}
      else 
	{
	  bits = _negativeNonBasicArguments.sub(symId);
	};
      if (bits)
	{
	  bits[argNum] = true;
	}
      else
	{
	  bits = 
	    static_cast<bool*>(BK::GlobAlloc::allocate(arity(symId) * sizeof(bool)));
	  for (ulong i = 0UL; i < arity(symId); ++i) bits[i] = false;
	  bits[argNum] = true;
	  if (positive)
	    {  
	      _positiveNonBasicArguments[symId] = bits;
	    }
	  else 
	    {
	      _negativeNonBasicArguments[symId] = bits;
	    };
	};
    };
}; 
void Signature::makeEvaluable(ulong num) {
  CALL("makeEvaluable(ulong num)");
  _isEvaluableSymbol.sub(num) = true;
  _someSymbolsAreEvaluable = true;
};
bool Signature::createNewSkolemConstantNumber(ulong& res)
{
  CALL("createNewSkolemConstantNumber()");
  static char tmpString[256]; 
  if (_numberOfSkolemSymbols >= _maxlNumOfSkolemSymbols) 
    return false;
  strcpy(tmpString,VampireKernelConst::SkolemConstantNamePrefix());
  char* endOfPrefix = tmpString + strlen(VampireKernelConst::SkolemConstantNamePrefix());
  char* suffix;
 try_suffix:
  suffix = endOfPrefix; 
  print(_nextSkolemConstantSuffix,suffix);
  suffix[0] = '\0';
  if (isRegisteredSymbolName(tmpString)) 
    {
      _nextSkolemConstantSuffix++;
      goto try_suffix;
    };
  _symbolName.sub(_nextSymbolNumber) = new char[strlen(tmpString) + 1];
  strcpy(_symbolName.sub(_nextSymbolNumber),tmpString);
  _isInputSymbol.sub(_nextSymbolNumber) = false;
  _isPredicate.sub(_nextSymbolNumber) = false;   
  _isCommutative.sub(_nextSymbolNumber) = false; 
  _arity.sub(_nextSymbolNumber) = 0UL;
  setSymbolWeightConstantPart(_nextSymbolNumber,1UL);
  res = _nextSymbolNumber;
  _nextSymbolNumber++;
  ASSERT(_nextSymbolNumber <= _symbolName.size());
  ASSERT(_nextSymbolNumber <= _isInputSymbol.size());
  ASSERT(_nextSymbolNumber <= _isPredicate.size());
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
  _numberOfSkolemSymbols++;
  return true; 
}; 
bool Signature::createNewSkolemFunctionNumber(ulong arity,ulong& res)
{
  CALL("createNewSkolemFunctionNumber(arity,ulong& res)");
  static char tmpString[256]; 
  if (_numberOfSkolemSymbols >= _maxlNumOfSkolemSymbols) 
    return false;
  strcpy(tmpString,VampireKernelConst::SkolemFunctionNamePrefix());
  char* endOfPrefix = tmpString + strlen(VampireKernelConst::SkolemFunctionNamePrefix());
  char* suffix;
 try_suffix:
  suffix = endOfPrefix; 
  print(_nextSkolemFunctionSuffix,suffix);
  suffix[0] = '\0';
  if (isRegisteredSymbolName(tmpString)) 
    {
      _nextSkolemFunctionSuffix++;
      goto try_suffix;
    };
  _symbolName.sub(_nextSymbolNumber) = new char[strlen(tmpString) + 1];
  strcpy(_symbolName.sub(_nextSymbolNumber),tmpString);
  _isInputSymbol.sub(_nextSymbolNumber) = false;
  _isPredicate.sub(_nextSymbolNumber) = false;   
  _isCommutative.sub(_nextSymbolNumber) = false; 
  _arity.sub(_nextSymbolNumber) = arity;
  setSymbolWeightConstantPart(_nextSymbolNumber,1UL);
  res = _nextSymbolNumber;
  _nextSymbolNumber++;
  ASSERT(_nextSymbolNumber <= _symbolName.size());
  ASSERT(_nextSymbolNumber <= _isInputSymbol.size());
  ASSERT(_nextSymbolNumber <= _isPredicate.size());
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
  _numberOfSkolemSymbols++; 
  return true;
}; 
bool Signature::createNewSkolemPredicateNumber(ulong arity,ulong& res)
{
  CALL("createNewSkolemPredicateNumber(arity,ulong& res)");
  static char tmpString[256]; 
  if (_numberOfSkolemSymbols >= _maxlNumOfSkolemSymbols) 
    return false;
  strcpy(tmpString,VampireKernelConst::SkolemPredicateNamePrefix());
  char* endOfPrefix = tmpString + strlen(VampireKernelConst::SkolemPredicateNamePrefix());
  char* suffix;
 try_suffix:
  suffix = endOfPrefix; 
  print(_nextSkolemPredicateSuffix,suffix);
  suffix[0] = '\0';
  if (isRegisteredSymbolName(tmpString)) 
    {
      _nextSkolemPredicateSuffix++;
      goto try_suffix;
    }; 
  _symbolName.sub(_nextSymbolNumber) = new char[strlen(tmpString) + 1];
  strcpy(_symbolName.sub(_nextSymbolNumber),tmpString);
  _isInputSymbol.sub(_nextSymbolNumber) = false;
  _isPredicate.sub(_nextSymbolNumber) = true;   
  _isCommutative.sub(_nextSymbolNumber) = false; 
  _arity.sub(_nextSymbolNumber) = arity;
  setSymbolWeightConstantPart(_nextSymbolNumber,1UL);
  res = _nextSymbolNumber;
  _nextSymbolNumber++;
  ASSERT(_nextSymbolNumber <= _symbolName.size());
  ASSERT(_nextSymbolNumber <= _isInputSymbol.size());
  ASSERT(_nextSymbolNumber <= _isPredicate.size());
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
  _numberOfSkolemSymbols++;  
  return true;
}; 
void Signature::assignUniqueSymbolPrecedence(ulong symId)
{
  CALL("assignUniqueSymbolPrecedence(ulong symId)");
  if (_minimalSymbolPrecedence > LONG_MIN)
    {
      setSymbolPrecedence(symId,_minimalSymbolPrecedence - 1);
      return;
    };
  if (_maximalSymbolPrecedence < LONG_MAX)
    {
      setSymbolPrecedence(symId,_maximalSymbolPrecedence + 1);
      return;
    };
  for (long i = LONG_MIN + 1; i < LONG_MAX; ++i)
    {
      ulong dummy;
      if (!symbolPrecedenceIsReserved(i,dummy))
	{
	  setSymbolPrecedence(symId,i);
	  return;
	};
    };
  BK::RuntimeError::report("All precedences are reserved?!?!?!");
}; 
ostream& Signature::output(ostream& str)
{ 
  str << "%=================== Signature: =====================\n";
  ASSERT((_nextSymbolNumber <= _symbolName.size()) || 
	 (_nextSymbolNumber <= _interpretedConstantValue.size()));
  ASSERT(_nextSymbolNumber <= _isInputSymbol.size());
  ASSERT(_nextSymbolNumber <= _isPredicate.size());
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
  ulong fn;
  for (fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (!reservedSymbolNumber(fn) || 
	  fn == VampireKernelConst::UnordEqNum)
	{
	  str << fn << ": " << symbolName(fn) << "/";
	  ASSERT(fn < _arity.size());
	  str << _arity[fn]; 
	  if (_isPredicate[fn]) 
	    { 
	      const bool* posNonBasicArgs = positiveNonBasicArguments(fn);
	      const bool* negNonBasicArgs = negativeNonBasicArguments(fn);
	      if (posNonBasicArgs || negNonBasicArgs)
		{
		  str << " (";
		  if (posNonBasicArgs)
		    {
		      str << "+";
		      for (ulong i = 0UL; i < _arity[fn]; ++i)
			if (posNonBasicArgs[i])
			  {
			    str << "N";
			  }
			else
			  str << "b";
		    };
		  if (negNonBasicArgs)
		    {
		      if (posNonBasicArgs) str << ",";
		      str << "-";
		      for (ulong i = 0UL; i < _arity[fn]; ++i)
			if (negNonBasicArgs[i])
			  {
			    str << "N";
			  }
			else
			  str << "b";
		    };
		  str << ")";
		};
	    }
	  else 
	    {
	      const bool* nonBasicArgs = positiveNonBasicArguments(fn);
	      if (nonBasicArgs)
		{
		  str << " (";
		  for (ulong i = 0UL; i < _arity[fn]; ++i)
		    if (nonBasicArgs[i])
		      {
			str << "N";
		      }
		    else
		      str << "b";
		  str << ")";
		};
	    };
	  if (_isInputSymbol[fn]) { str << " input  "; } else str << " skolem ";
	  if (_isPredicate[fn]) 
	    { 
	      str << "predicate ";
	      if (_isCommutative[fn]) { str << "symmetric      "; } else str << "nonsymmetric  "; 
	    } 
	  else
	    {
	      str << "function  ";
	      if (_isCommutative[fn]) str << "commutative    ";  
	    };
	  str << "\n";
	};
    };
  for (fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if ((!reservedSymbolNumber(fn)) || 
	  (fn == VampireKernelConst::UnordEqNum))
	{
	  if (_isPredicate[fn]) 
	    { 
	      str << "header_precedence(++" 
		  << symbolName(fn)
		  << "/" << arity(fn) 
		  << ") = " << headerRank(PositivePolarity,fn) << ",     "
		  << "header_precedence(--"
		  << symbolName(fn)
		  << "/" << arity(fn) 
		  << ") = " << headerRank(NegativePolarity,fn) << "\n";
	    };
	  str << "symbol_precedence("
	      << ((_isPredicate[fn])? "predicate" : "function") << " "
	      << symbolName(fn)
	      << "/" << arity(fn) 
	      << ") = " << symbolPrecedence(fn) << "\n";
	  str << "weight("
	      << ((_isPredicate[fn])? "predicate" : "function") << " "
	      << symbolName(fn)
	      << "/" << arity(fn) 
	      << ") = " << weight(fn) << "\n";
	};
    };
  return str << "%============ End of signature. =================\n";
}; 
ostream& Signature::outputNonBasicArguments(ostream& str)
{
  CALL("outputNonBasicArguments(ostream& str)");
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (!reservedSymbolNumber(fn) && arity(fn))
	{
	  if (_isPredicate[fn]) 
	    { 
	      const bool* posNonBasicArgs = positiveNonBasicArguments(fn);
	      const bool* negNonBasicArgs = negativeNonBasicArguments(fn);
	      if (posNonBasicArgs || negNonBasicArgs)
		{
		  str << "% " << fn << ": " 
		      << symbolName(fn) << "/";
		  ASSERT(fn < _arity.size());
		  str << _arity[fn]; 
		  str << " (";
		  if (posNonBasicArgs)
		    {
		      str << "+";
		      for (ulong i = 0UL; i < _arity[fn]; ++i)
			if (posNonBasicArgs[i])
			  {
			    str << "N";
			  }
			else
			  str << "b";
		    };
		  if (negNonBasicArgs)
		    {
		      if (posNonBasicArgs) str << ",";
		      str << "-";
		      for (ulong i = 0UL; i < _arity[fn]; ++i)
			if (negNonBasicArgs[i])
			  {
			    str << "N";
			  }
			else
			  str << "b";
		    };
		  str << ")";
		  if (_isInputSymbol[fn]) { str << " input  "; } else str << " skolem ";
		  str << "predicate ";
		  if (_isCommutative[fn]) str << "symmetric      "; 
		  str << "\n";
		};
	    }
	  else 
	    {
	      const bool* nonBasicArgs = positiveNonBasicArguments(fn);
	      if (nonBasicArgs)
		{	      
		  str << "% " << fn << ": "
		      << symbolName(fn) << "/";
		  ASSERT(fn < _arity.size());
		  str << _arity[fn]; 
		  str << " (";
		  for (ulong i = 0UL; i < _arity[fn]; ++i)
		    if (nonBasicArgs[i])
		      {
			str << "N";
		      }
		    else
		      str << "b";
		  str << ")";
		  if (_isInputSymbol[fn]) { str << " input  "; } else str << " skolem ";
		  str << "function  ";
		  if (_isCommutative[fn]) str << "commutative    ";  
		  str << "\n";
		};
	    };
	};
    };
  return str;
}; 
bool Signature::isRegisteredSymbolName(const char* str)
{
  CALL("isRegisteredSymbolName(const char* str)");
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (!isInterpretedConstant(fn) && 
	  !(strcmp(str,_symbolName.sub(fn)))) 
	return true;
    };
  return false;  
}; 
const char* Signature::createCanonicalVariableName(ulong varNum)
{
  CALL("createCanonicalVariableName(ulong varNum)");
  static char tmpString[256];
  strcpy(tmpString,VampireKernelConst::CanonicalVariableNamePrefix());
  char* cursor = tmpString + strlen(VampireKernelConst::CanonicalVariableNamePrefix());
  print(varNum,cursor);
  cursor[0] = '\0';
  char* res = new char[strlen(tmpString) + 1]; 
  strcpy(res,tmpString);
  return res;
}; 
bool 
Signature::symbolPrecedenceIsReserved(long precedence,ulong& symId) const
{
  CALL("symbolPrecedenceIsReserved(long precedence,ulong& symId) const");
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (precedence == symbolPrecedence(fn))
	{
	  symId = fn;
	  return true;
	};
    };
  return false;
}; 
void Signature::setSymbolPrecedence(ulong symId,long prec)
{
  CALL("setSymbolPrecedence(ulong symId,long prec)");
#ifdef DEBUG_NAMESPACE
  for (ulong fn = _firstInputOrSkolemSymbolNumber; fn < _nextSymbolNumber; fn++)
    {
      if (fn != symId && symbolPrecedence(fn) == prec)
	{
	  if (isPredicate(symId))
	    {
	      if (isPredicate(fn))
		{
		  ASSERT(headerRank(PositivePolarity,symId) !=
			 headerRank(PositivePolarity,fn));
		  ASSERT(headerRank(PositivePolarity,symId) !=
			 headerRank(NegativePolarity,fn));
		  ASSERT(headerRank(NegativePolarity,symId) !=
			 headerRank(PositivePolarity,fn));
		  ASSERT(headerRank(NegativePolarity,symId) !=
			 headerRank(NegativePolarity,fn));
		};
	    }
	  else
	    {
	      ASSERT(isPredicate(fn));
	    };
	};
    };
#endif
  _symbolPrecedence.sub(symId) = prec;
  if (symId == VampireKernelConst::UnordEqNum)
    {
      _symbolPrecedence.sub(VampireKernelConst::OrdEqNum) = prec;
    }
  else
    if (symId == VampireKernelConst::OrdEqNum)
      _symbolPrecedence.sub(VampireKernelConst::UnordEqNum) = prec;
  if (prec > _maximalSymbolPrecedence)
    _maximalSymbolPrecedence = prec;
  if (prec < _minimalSymbolPrecedence)
    _minimalSymbolPrecedence = prec;
}; 
void Signature::endOfInput()
{
  CALL("endOfInput()");
  _endOfInputFlag = true;
  ASSERT((_nextSymbolNumber <= _symbolName.size()) || 
	 (_nextSymbolNumber <= _interpretedConstantValue.size()));
  ASSERT(_nextSymbolNumber <= _isInputSymbol.size());
  ASSERT(_nextSymbolNumber <= _isPredicate.size());
  ASSERT(_nextSymbolNumber <= _isCommutative.size());
}; 
void Signature::resetAllArgumentsAsNonBasic()
{
  CALL("resetAllArgumentsAsNonBasic()");
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (!reservedSymbolNumber(fn) && arity(fn))
	{
	  bool* bits = _positiveNonBasicArguments.sub(fn);
	  if (!bits) 
	    bits = 
	      static_cast<bool*>(BK::GlobAlloc::allocate(arity(fn) * sizeof(bool)));
	  for (ulong i = 0UL; i < arity(fn); ++i)
	    bits[i] = true;
	  _positiveNonBasicArguments[fn] = bits;
	  if (_isPredicate[fn]) 
	    {  
	      bool* bits = _negativeNonBasicArguments.sub(fn);
	      if (!bits) 
		bits = 
		  static_cast<bool*>(BK::GlobAlloc::allocate(arity(fn) * sizeof(bool)));
	      for (ulong i = 0UL; i < arity(fn); ++i)
		bits[i] = true;
	      _negativeNonBasicArguments[fn] = bits;
	    };
	};
    };
}; 
void Signature::finaliseArgumentBasicness()
{
  CALL("finaliseArgumentBasicness()");
  _numOfNonBasicArguments = 0UL;
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (!reservedSymbolNumber(fn) && arity(fn))
	{
	  bool* bits = _positiveNonBasicArguments.sub(fn);
	  if (bits)
	    {
	      ulong numOfNonBasicArgs = 0UL;
	      for (ulong i = 0UL; i < arity(fn); ++i)
		{
		  if (bits[i]) ++numOfNonBasicArgs;
		};
	      if (numOfNonBasicArgs)
		{
		  _numOfNonBasicArguments += numOfNonBasicArgs;
		}
	      else
		{
		  BK::GlobAlloc::deallocate(static_cast<void*>(bits),
					    arity(fn) * sizeof(bool));
		  _positiveNonBasicArguments[fn] = 0;
		};
	    };
	  if (_isPredicate[fn]) 
	    {  
	      bool* bits = _negativeNonBasicArguments.sub(fn);
	      if (bits)
		{
		  ulong numOfNonBasicArgs = 0UL;
		  for (ulong i = 0UL; i < arity(fn); ++i)
		    {
		      if (bits[i]) ++numOfNonBasicArgs;
		    };
		  if (numOfNonBasicArgs)
		    {
		      _numOfNonBasicArguments += numOfNonBasicArgs;
		    }
		  else
		    {
		      BK::GlobAlloc::deallocate(static_cast<void*>(bits),
						arity(fn) * sizeof(bool));
		      _negativeNonBasicArguments[fn] = 0;
		    };
		};
	    };
	};
    };
}; 
bool Signature::symbolPropertiesPreserveCompleteness() {
  CALL("symbolPropertiesPreserveCompleteness()");
  for (ulong fn = 0; fn < _nextSymbolNumber; fn++)
    {
      if (!reservedSymbolNumber(fn))
	{
	  if (_isPredicate[fn])
	    {
	      if (headerRank(NegativePolarity,fn) < headerRank(PositivePolarity,fn))
		return false;
	    }
	  else 
	    {
	      if (arity(fn) == 1 && 
		  weight(fn) == 0 &&
		  symbolPrecedence(fn) != _maximalSymbolPrecedence)
		return false;
	    };
	};
    };
  return true;
}; 
