/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef SIGNATURE_HPP  
#define SIGNATURE_HPP
#include <iostream>
#include <cstring>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "VampireKernel.hpp"
#include "VampireKernelEnvironment.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
#include "Comparison.hpp"
#include "SymbolInfo.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIGNATURE
 #define DEBUG_NAMESPACE "Signature"
#endif 
#include "debugMacros.hpp"
namespace VK
{
class Clause;
class Signature
{ 
 public:
  enum Polarity
  {
    PositivePolarity = 0UL,
    NegativePolarity = 1UL
  };
  typedef VampireKernelEnvironment::InterpretedConstant InterpretedConstantValue;
 public:
  Signature();
  ~Signature();
  static Signature* current() { return _current; };
  static void setCurrent(Signature* sig) { _current = sig; };
  inline 
  static 
  Polarity inverse(const Polarity& pol) 
  { 
    if (pol == PositivePolarity) return NegativePolarity;
    return PositivePolarity;
  };
  void resetInputVariables(); 
  ulong inputVariableNumber(const char* varName);
  ulong inputSymbolNumber(bool predicate,const char* funName,ulong arity);
  static ulong equalityId() { return VampireKernelConst::UnordEqNum; };
  ulong 
  interpretedConstantNumber(const InterpretedConstantValue& value);
  bool isPredicate(ulong symId) 
  {
    CALL("isPredicate(ulong symId)");
    return _isPredicate.sub(symId); 
  }; 
  ulong arity(ulong symId) const
    {
      CALL("arity(ulong symId) const");
      if (symId >= _arity.size()) return 0UL;
      return _arity[symId]; 
    };  
  bool isSkolem(ulong symId) const 
    {
      CALL("isSkolem(ulong symId) const");
      if (symId >= _isInputSymbol.size()) return false;
      return !_isInputSymbol[symId]; 
    };   
  bool isCommutative(ulong symId) 
    {
      CALL("isCommutative(ulong symId)");
      if (symId >= _isCommutative.size()) return false;
      return _isCommutative[symId]; 
    };
  bool isInterpretedConstant(ulong symId) const {
    CALL("isInterpretedConstant(ulong symId) const");
    return  symId < _isInterpretedConstant.size() &&
	_isInterpretedConstant[symId];
  };
  bool isEvaluableSymbol(ulong num) const {
    CALL("isEvaluableSymbol(ulong num) const");
    return num < _isEvaluableSymbol.size() &&
      _isEvaluableSymbol[num];
  };
  const char* symbolName(ulong symId) const
    { 
      CALL("symbolName(ulong symId) const");
      if (isInterpretedConstant(symId))
	{
	  ASSERT(symId < _interpretedConstantValue.size());
	  return _interpretedConstantValue[symId].toString().c_str();
	};
      ASSERT(symId < _symbolName.size());
      return _symbolName[symId]; 
    };  
  const InterpretedConstantValue& 
  interpretedConstantValue(ulong symId) const 
  {
    CALL("interpretedConstantValue(..)");
    ASSERT(isInterpretedConstant(symId));
    ASSERT(symId < _interpretedConstantValue.size());
    return _interpretedConstantValue[symId];
  };
  bool isAnswerPredicate(ulong num) const
    {
      CALL("isAnswerPredicate(ulong num) const");
      return (num < _isAnswerPredicate.size()) &&
	_isAnswerPredicate[num];
    };
  const bool* positiveNonBasicArguments(ulong num) const
  {
    CALL("positiveNonBasicArguments(ulong num) const");
    if (num < _positiveNonBasicArguments.size())
      {
	return _positiveNonBasicArguments[num];
      }
    else
      return 0;
  };
  const bool* negativeNonBasicArguments(ulong predId) const
  {
    CALL("negativeNonBasicArguments(ulong predId) const");
    if (predId < _negativeNonBasicArguments.size())
      {
	return _negativeNonBasicArguments[predId];
      }
    else
      return 0;
  };
  void makeCommutative(ulong symId);
  void makeNoncommutative(ulong symId);
  void setCommutativity(const SymbolInfo* commutativityTable,
                        const SymbolInfo* symmetryTable);
  void makeAnswerPredicate(ulong predId)
    {
      CALL("makeAnswerPredicate(ulong predId)");
      ASSERT((predId < _isPredicate.size()) && _isPredicate[predId]);
      _isAnswerPredicate.sub(predId) = true;
    };
  void markBasicArgument(bool positive,ulong symId,ulong argNum);
  void markNonBasicArgument(bool positive,ulong symId,ulong argNum);
  void makeEvaluable(ulong num);
  void endOfInput();
  void resetAllArgumentsAsNonBasic();
  void finaliseArgumentBasicness();
  bool nonBasicArgumentExists() const
  {
    return _numOfNonBasicArguments != 0UL;
  };
  void setMaxlNumOfSkolemSymbols(long n) { _maxlNumOfSkolemSymbols = n; };
  ulong nextSymbolNumber() const { return _nextSymbolNumber; };
  bool canCreateNSkolemSymbols(ulong n) 
    {  
      return _numberOfSkolemSymbols  <= (long)(_maxlNumOfSkolemSymbols - n);  
    };
  bool createNewSkolemConstantNumber(ulong& res);
  bool createNewSkolemFunctionNumber(ulong arity,ulong& res);
  bool createNewSkolemPredicateNumber(ulong arity,ulong& res);
  const char* canonicalVariableName(ulong varNum)
    {
      CALL("canonicalVariableName(ulong varNum)");
      return _canonicalVariableName[varNum];
    };
  long weight(ulong symId)
    {
      CALL("weight(ulong symId)");
      if (symId >= _weight.size()) return 1;
      return _weight[symId]; 
    };
  void setSymbolWeightConstantPart(ulong symId,const long& weight)
    {
      _weight.sub(symId) = weight;
    };
  void setInputSymbolPrecedence(ulong symId,long prec)
    {
      CALL("setPrecedence(ulong symId,long prec)");
      setSymbolPrecedence(symId,prec);
    };
  void setInputHeaderRank(Polarity polarity,
			  ulong predId,
			  long rank)
    {
      CALL("setInputHeaderRank(..)");
      if (polarity == PositivePolarity)
	{
	  _positiveHeaderRank.sub(predId) = rank;
	}
      else
	{
	  _negativeHeaderRank.sub(predId) = rank;
	};
    };
  void assignUniqueSymbolPrecedence(ulong symId);
  void assignDefaultHeaderRank(Polarity polarity,ulong predId) {
    CALL("assignDefaultHeaderRank(Polarity polarity,ulong predId)");
    if (polarity == PositivePolarity)
      {
	_positiveHeaderRank.sub(predId) = DefaultPositiveHeaderRank;
      }
    else
      {
	_negativeHeaderRank.sub(predId) = DefaultNegativeHeaderRank;
      };
  };
  long symbolPrecedence(ulong symId) const
    {
      CALL("symbolPrecedence(ulong symId) const");
      if (symId >= _symbolPrecedence.size()) return 0L;
      return _symbolPrecedence[symId];
    };
  long headerRank(ulong polarity,ulong predId)
    {
      CALL("headerRank(ulong polarity,ulong predId)");
      ASSERT(polarity == PositivePolarity || 
	     polarity == NegativePolarity);
      if (polarity == NegativePolarity)
	{
	  if (predId >= _negativeHeaderRank.size()) return 0L;
	  return _negativeHeaderRank[predId];
	}
      else
	{
	  if (predId >= _positiveHeaderRank.size())  return 0L;
	  return _positiveHeaderRank[predId];
	};
    }; 
  BK::Comparison 
  compareHeaderRanks(ulong polarity1,ulong predId1,ulong polarity2,ulong predId2)
    {
      CALL("compareHeaderRanks(ulong polarity1,ulong predId1,ulong polarity2,ulong predId2)");
      ASSERT((polarity1 == PositivePolarity) || (polarity1 == NegativePolarity));
      ASSERT((polarity2 == PositivePolarity) || (polarity2 == NegativePolarity));  
      long precedenceDifference = 
	headerRank(polarity1,predId1) - headerRank(polarity2,predId2);
      if (precedenceDifference)
	{ 
	  return (precedenceDifference > 0) ? BK::Greater : BK::Less; 
	}
      else 
	return BK::Equal;
    };      
  bool someSymbolsAreEvaluable() const { return _someSymbolsAreEvaluable; };
  bool someConstantsAreInterpreted() const { return _someConstantsAreInterpreted; };
  bool symbolPropertiesPreserveCompleteness();
  ostream& output(ostream& str);
  ostream& outputNonBasicArguments(ostream& str);
 public:
  enum
  {
    DefaultPositiveHeaderRank = VampireKernel::DefaultPositiveHeaderRank,
    DefaultNegativeHeaderRank = VampireKernel::DefaultNegativeHeaderRank
  };
 private:
  typedef BK::Array<BK::GlobAlloc,char*,BK::FastRelocation<char*>,512UL> StrArray;
  typedef BK::Array<BK::GlobAlloc,double,BK::FastRelocation<double>,512UL> DoubleArray;
  typedef BK::Array<BK::GlobAlloc,bool,BK::FastRelocation<bool>,512UL> BoolArray;
  typedef BK::Array<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,512UL> ULongArray;
  typedef BK::Array<BK::GlobAlloc,long,BK::FastRelocation<long>,512UL> LongArray;
  typedef BK::Array<BK::GlobAlloc,bool*,BK::FastRelocation<bool*>,512UL> BoolPtrArray;
  class InterpretedConstValueArrayId {};
  typedef BK::Array<BK::GlobAlloc,InterpretedConstantValue,BK::DefaultRelocation<InterpretedConstantValue>,512UL> InterpretedConstValueArray;
  class WeightArrayId {};
  typedef BK::Array<BK::GlobAlloc,long,BK::FastRelocation<long>,512UL> WeightArray;
  class WeightPtrArrayId {};
 private: 
  bool symbolPrecedenceIsReserved(long precedence,ulong& symId) const;
  void setSymbolPrecedence(ulong symId,long precedence);
  bool reservedSymbolNumber(ulong symId) 
    {
      return (symId == VampireKernelConst::UnordEqNum) 
	|| (symId == VampireKernelConst::OrdEqNum) 
	|| (symId == VampireKernelConst::BuiltInTrueNum)
	|| (symId == VampireKernelConst::ReservedPred0)  
	|| (symId == VampireKernelConst::ReservedPred1)  
	|| (symId == VampireKernelConst::ReservedPred2) 
	;                   
    };
  static void print(ulong num,char*& target)
    {
      if (!num) { target[0] = '0'; target++; return; }; 
      char lastDigit;
      switch (num % 10) 
	{   
	case 0UL: lastDigit = '0'; break;
	case 1UL: lastDigit = '1'; break;
	case 2UL: lastDigit = '2'; break;   
	case 3UL: lastDigit = '3'; break;
	case 4UL: lastDigit = '4'; break;
	case 5UL: lastDigit = '5'; break;   
	case 6UL: lastDigit = '6'; break;
	case 7UL: lastDigit = '7'; break;
	case 8UL: lastDigit = '8'; break;
	case 9UL: lastDigit = '9'; break;
#ifdef DEBUG_NAMESPACE
	default: ICP("ICP0"); return; 
#else
#ifdef _SUPPRESS_WARNINGS_
	default: return;  
#endif
#endif 
	};
      num = num / 10;     
      if (num) print(num,target);
      target[0] = lastDigit;
      target++;
    }; 
  bool isRegisteredSymbolName(const char* str);
  static const char* createCanonicalVariableName(ulong varNum);
 private:
  StrArray _inputVariableName;
  ulong _nextInputVariableNumber; 
  StrArray _symbolName;
  InterpretedConstValueArray _interpretedConstantValue;
  BoolArray _isInterpretedConstant;
  BoolArray _isEvaluableSymbol;
  BoolArray _isInputSymbol;
  BoolArray _isPredicate;
  BoolArray _isCommutative;
  ULongArray _arity;
  BoolArray _isAnswerPredicate;
  BoolPtrArray _positiveNonBasicArguments;
  BoolPtrArray _negativeNonBasicArguments;
  ulong _numOfNonBasicArguments;
  bool _endOfInputFlag;
  ulong _nextSymbolNumber;
  ulong _firstInputOrSkolemSymbolNumber;
  long _numberOfSkolemSymbols; 
  long _maxlNumOfSkolemSymbols;
  ulong _nextSkolemConstantSuffix;
  ulong _nextSkolemFunctionSuffix;
  ulong _nextSkolemPredicateSuffix;  
  const char* _canonicalVariableName[VampireKernelConst::MaxNumOfVariables];
  WeightArray _weight;
  LongArray _symbolPrecedence;
  LongArray _positiveHeaderRank;  
  LongArray _negativeHeaderRank;
  long _minimalSymbolPrecedence;
  long _maximalSymbolPrecedence;
  bool _someSymbolsAreEvaluable;
  bool _someConstantsAreInterpreted;
  static Signature* _current;
}; 
}; 
namespace std
{
inline ostream& operator<<(ostream& str,VK::Signature& sig) 
{ 
  return sig.output(str); 
};
};
#endif
