/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef SYMBOL_INFO_HPP
#define SYMBOL_INFO_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
namespace VK
{
  class Clause;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SYMBOL_INFO
#  define DEBUG_NAMESPACE "SymbolInfo"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class SymbolInfo
  {
  public: 
    SymbolInfo() :
      _table(Entry(),"SymbolInfo::_table")
    {
      CALL("constructor SymbolInfo()");
      DOP(_table.freeze());
      _numOfCommutative = 0UL;
      _numOfSymmetric = 0UL;
      _numOfTransitive = 0UL;
    };
    ~SymbolInfo() {};
    void reset() 
    {
      CALL("reset()");
      for (ulong n = 0; n < _table.size(); n++)
	_table[n].reset();
      _numOfCommutative = 0UL;
      _numOfSymmetric = 0UL;
      _numOfTransitive = 0UL;
    };
    ulong numOfCommutative() const { return _numOfCommutative; };
    ulong numOfSymmetric() const { return _numOfSymmetric; }; 
    ulong numOfTransitive() const { return _numOfTransitive; }; 
    const Clause* commutativityLaw(ulong funNum) const
    {
      CALL("commutativityLaw(ulong funNum) const");
      if (funNum >= _table.size()) return 0;
      return _table[funNum].commutativityLaw;
    };
    const Clause* symmetryLaw(ulong predNum) const
    {
      CALL("symmetryLaw(ulong predNum) const");
      if (predNum >= _table.size()) return 0;
      return _table[predNum].symmetryLaw;
    };
    const Clause* transitivityLaw(ulong predNum) const
    {
      CALL("transitivityLaw(ulong predNum) const");
      if (predNum >= _table.size()) return 0;
      return _table[predNum].transitivityLaw;
    };
    void registerCommutativity(ulong funNum,const Clause* axiom)
    {
      CALL("registerCommutativity(ulong funNum,const Clause* axiom)");
      ASSERT(axiom);
      DOP(_table.unfreeze());
      if (!_table.sub(funNum).commutativityLaw) _numOfCommutative++;
      DOP(_table.freeze());
      _table[funNum].commutativityLaw = axiom;
    };  
    void registerSymmetry(ulong predNum,const Clause* axiom)
    {
      CALL("registerSymmetry(ulong predNum,const Clause* axiom)");
      ASSERT(axiom);
      DOP(_table.unfreeze());
      if (!_table.sub(predNum).symmetryLaw) _numOfSymmetric++;
      DOP(_table.freeze());
      _table[predNum].symmetryLaw = axiom;
    };
    void registerTransitivity(ulong predNum,const Clause* axiom)
    {
      CALL("registerTransitivity(ulong predNum,const Clause* axiom)");
      ASSERT(axiom);
      DOP(_table.unfreeze());
      if (!_table.sub(predNum).transitivityLaw) _numOfTransitive++;
      DOP(_table.freeze());
      _table[predNum].transitivityLaw = axiom;
    };
  private: 
    class Entry
    {
    public:
      Entry() 
      {
	commutativityLaw = 0; 
	symmetryLaw = 0;
	transitivityLaw = 0;
      };
      ~Entry() {};
      void reset() 
      {
	commutativityLaw = 0; 
	symmetryLaw = 0; 
	transitivityLaw = 0;
      };  
    public:
      const Clause* commutativityLaw;
      const Clause* symmetryLaw;
      const Clause* transitivityLaw;
    }; 
  private:
    BK::Array<BK::GlobAlloc,Entry,BK::FastRelocation<Entry>,128UL> _table;
    ulong _numOfCommutative;
    ulong _numOfSymmetric;
    ulong _numOfTransitive;
  }; 
}; 
#endif
