/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TEMP_LIT_LIST_EXTRAS_HPP
#define TEMP_LIT_LIST_EXTRAS_HPP
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "ObjectPool.hpp"
#include "ObjectCache.hpp"
#include "Receiver.hpp"
#include "ComparableCharacteristic.hpp"
#include "CharacteristicComparisonMaker.hpp"
#include "ComplexComparisonMaker.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.inl"
#include "BasicTempLitList.hpp"
#include "Term.hpp"
#include "VariableRenaming.hpp"
#include "VariableWeighting.hpp"
namespace VK
{  
  class SymbolInfo;
  class Clause;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class TempLitListExtras
  {
  public:
    typedef BasicTempLitList::Element Element;
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef BK::ObjectPool<BasicTempLitList> Pool;
    typedef BK::ObjectPool<BasicTempLitList>::EntryHandle Handle;
    class StdCharacteristics;
    typedef BK::ComparableCharacteristic<const BasicTempLitList*> ComparableCharacteristic;
    typedef BK::CMaker<const BasicTempLitList*> ComparisonMaker;
    class Size;
    class NumOfVarOccurences;
    class NumOfLiterals;
    class CharacteristicComparisonMaker;
    class ComplexComparisonMaker;
    class Context;
  public:
    static long computeWeightBase(const BasicTempLitList& lst);
    static long size(const BasicTempLitList& lst);
    static 
    void
    computeSizeAndWeightBase(const BasicTempLitList& lst,
			     long& sz,
			     long& weightBase);
    static 
    void
    computeSizeAndWeightBaseAndNumOfPosEq(const BasicTempLitList& lst,
					  long& sz,
					  long& weightBase,
					  long& numOfPosEq);
    static
    long minimalPosEqLHSSize(const BasicTempLitList& lst);
    static 
    bool 
    instantiationIsPossible(const BasicTempLitList& lst,
			    ulong var,
			    ulong instanceDepth,
			    long instanceSize,
			    bool checkInputNonBasicness);
    static
    void
    passInOrder(BasicTempLitList& src,
		BasicTempLitList& dest,
		const VariableWeighting& varWeighting);
    static 
    void 
    renameVariables(const BasicTempLitList& lst,
		    VariableRenaming& varRenaming);
    static
    void
    weed(BasicTempLitList& lst,
	 bool& isTautology,
	 bool& isEqTautology);
    static
    void
    weed(BasicTempLitList& lst,
	 bool& isTautology,
	 bool& isEqTautology,
	 const SymbolInfo* commutativityTableForEqTaut,
	 const SymbolInfo* commutativityTableForPropTaut,
	 const SymbolInfo* commutativityTableForTrivialEqRes,
	 const SymbolInfo* commutativityTableForIdenticalLitFactoring,
	 BK::Receiver<const Clause*>* usedCommutativityLaws);
    static 
    bool 
    containsOnlyAnswerLiterals(const BasicTempLitList& lst);
    static 
    bool 
    containsAnswerLiterals(const BasicTempLitList& lst);
    static 
    bool 
    containsSplittingBranches(const BasicTempLitList& lst);
    static 
    bool
    containsHeader(const BasicTempLitList& lst,const Term& header);
    static 
    bool
    isGroundSingleton(const BasicTempLitList& lst) {
      return 
	lst.length() == 1UL && lst.first()->literal()->isGround();
    };
    static 
    ulong
    countDifferentVariables(const BasicTempLitList& lst);
    static
    void 
    collectVariableOccurences(const BasicTempLitList& lst,
			      WeightPolynomial::Monomials& vars);
    static
    void 
    preorientEqualities(const BasicTempLitList& lst);
    static 
    void adjustSelectionLayers(BasicTempLitList& lst);
    static 
    void freshList(Handle& h,Pool* pool);
    static 
    void freshList(Handle& h);
  private:
    static
    bool complementary(const LitHandle& lit1,
		       const LitHandle& lit2,
		       const SymbolInfo* commutativityTable,
		       BK::Receiver<const Clause*>* usedCommutativityLaws);
    static
    bool equal(const LitHandle& lit1,
	       const LitHandle& lit2,
	       const SymbolInfo* commutativityTable,
	       BK::Receiver<const Clause*>* usedCommutativityLaws);
    static
    bool complementaryEq(const LitHandle& lit1,
			 const LitHandle& lit2,
			 const SymbolInfo* commutativityTable,
			 BK::Receiver<const Clause*>* usedCommutativityLaws);
    static
    bool equalEq(const LitHandle& lit1,
		 const LitHandle& lit2,
		 const SymbolInfo* commutativityTable,
		 BK::Receiver<const Clause*>* usedCommutativityLaws,
		 bool& eqSymmetryUsed);
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::Size"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TempLitListExtras::Size : 
    public ComparableCharacteristic
  {
  public:
    Size();
    ~Size();
    void* operator new(size_t);
    void operator delete(void*);
    void computeFor(const BasicTempLitList* const & lst);
    BK::Comparison compare(ComparableCharacteristic& c);
    BK::Comparison inverseCompare(ComparableCharacteristic& c);
  private:
    long _value;
  };  
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::NumOfVarOccurences"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TempLitListExtras::NumOfVarOccurences : 
    public ComparableCharacteristic
  {
  public:
    NumOfVarOccurences();
    ~NumOfVarOccurences();
    void* operator new(size_t);
    void operator delete(void*);
    void computeFor(const BasicTempLitList* const & lst);
    BK::Comparison compare(ComparableCharacteristic& c);
    BK::Comparison inverseCompare(ComparableCharacteristic& c);
  private:
    long _value;
  };  
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::NumOfLiterals"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TempLitListExtras::NumOfLiterals : 
    public ComparableCharacteristic
  {
  public:
    NumOfLiterals();
    ~NumOfLiterals();
    void* operator new(size_t);
    void operator delete(void*);
    void computeFor(const BasicTempLitList* const & lst);
    BK::Comparison compare(ComparableCharacteristic& c);
    BK::Comparison inverseCompare(ComparableCharacteristic& c);
  private:
    ulong _value;
  };  
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::CharacteristicComparisonMaker"
#endif
#include "debugMacros.hpp"
namespace BK {
  template <class T>
  class CCM : public CMaker<T>
  {
  public:
    CCM() {};
    CCM(CMaker<T>* comp)
    {
      makeInverse(comp);
    };
    CCM(CMaker<T>* comp1,
			   CMaker<T>* comp2) 
    {
      makeComposition(comp1,comp2);
    };
    void makeInverse(CMaker<T>* comp)
    {
      _state = Inverse_Clean;
      _comp1 = comp;
    };
    void makeComposition(CMaker<T>* comp1,
			 CMaker<T>* comp2)
    {
      _state = Composition_0_0_0_0;
      _comp1 = comp1;
      _comp2 = comp2;
    };
    void loadLHS(const T& obj);
    void loadRHS(const T& obj);
    Comparison compare();
    Comparison inverseCompare();
    void swapSides();
  protected:
    ~CCM();
  private:
    enum State
    {
      Inverse_Clean,
      Inverse_LHSLoaded,
      Inverse_RHSLoaded,
      Inverse_LHSRHSLoaded,
      Inverse_Sw_LHSLoaded,
      Inverse_Sw_RHSLoaded,
      Inverse_Sw_LHSRHSLoaded,
      Composition_0_0_0_0,
      Composition_0_1_0_0,
      Composition_0_1_0_1,
      Composition_1_0_0_0,
      Composition_1_0_1_0,
      Composition_1_1_0_0,
      Composition_1_1_0_1,
      Composition_1_1_1_0,
      Composition_1_1_1_1,
      Composition_Sw_0_1_0_0,
      Composition_Sw_0_1_0_1,
      Composition_Sw_1_0_0_0,
      Composition_Sw_1_0_1_0,
      Composition_Sw_1_1_0_0,
      Composition_Sw_1_1_0_1,
      Composition_Sw_1_1_1_0,
      Composition_Sw_1_1_1_1
    };
  private:
    State _state;
    T _lhsBackup;
    T _rhsBackup;
    CMaker<T>* _comp1;
    CMaker<T>* _comp2;
  }; 
  template <class T>
  CCM<T>::~CCM()
  {
  };
  template <class T>
  void 
  CCM<T>::loadLHS(const T& obj)
  {
    CALL("loadLHS(const T& obj)");
    switch (_state)
      {
      case Inverse_Clean:
	_comp1->loadLHS(obj);
	_state = Inverse_LHSLoaded;
	break;
      case Inverse_LHSLoaded:
	_comp1->loadLHS(obj);
	break;      
      case Inverse_RHSLoaded:
	_comp1->loadLHS(obj);
	_state = Inverse_LHSRHSLoaded;
	break;
      case Inverse_LHSRHSLoaded:
	_comp1->loadLHS(obj);
	break;      
      case Inverse_Sw_LHSLoaded:
	_comp1->loadRHS(obj);
	_state = Inverse_Sw_LHSRHSLoaded;
	break;
      case Inverse_Sw_RHSLoaded:
	_comp1->loadRHS(obj);
	break;
      case Inverse_Sw_LHSRHSLoaded:
	_comp1->loadRHS(obj);
	break;
      case Composition_0_0_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_0_0_0;
	break;
      case Composition_0_1_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_1_0_0;
	break;
      case Composition_0_1_0_1:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_1_0_1;
	break;
      case Composition_1_0_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	break;
      case Composition_1_0_1_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_0_0_0;
	break;
      case Composition_1_1_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	break;
      case Composition_1_1_0_1:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	break;
      case Composition_1_1_1_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_1_0_0;
	break;
      case Composition_1_1_1_1:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_1_0_1;
	break;
      case Composition_Sw_0_1_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_0_0_0;
	break;
      case Composition_Sw_0_1_0_1:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_1_0_0_0;
	break;
      case Composition_Sw_1_0_0_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_Sw_1_1_0_0;
	break;
      case Composition_Sw_1_0_1_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_Sw_1_1_1_0;
	break;
      case Composition_Sw_1_1_0_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	break;
      case Composition_Sw_1_1_0_1:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_Sw_1_1_0_0;
	break;
      case Composition_Sw_1_1_1_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	break;
      case Composition_Sw_1_1_1_1:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_Sw_1_1_1_0;
	break;      
      };
  }; 
  template <class T>
  void CCM<T>::loadRHS(const T& obj)
  {
    CALL("loadRHS(const T& obj)");
    switch (_state)
      {
      case Inverse_Clean:
	_comp1->loadRHS(obj);
	_state = Inverse_RHSLoaded;
	break;
      case Inverse_LHSLoaded:
	_comp1->loadRHS(obj);
	_state = Inverse_LHSRHSLoaded;
	break;
      case Inverse_RHSLoaded:
	_comp1->loadRHS(obj);
	break;
      case Inverse_LHSRHSLoaded:
	_comp1->loadRHS(obj);
	break;
      case Inverse_Sw_LHSLoaded:
	_comp1->loadLHS(obj);
	break;
      case Inverse_Sw_RHSLoaded:
	_comp1->loadLHS(obj);
	_state = Inverse_Sw_LHSRHSLoaded;
	break;
      case Inverse_Sw_LHSRHSLoaded:
	_comp1->loadLHS(obj);
	break;
      case Composition_0_0_0_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_0_1_0_0;
	break;
      case Composition_0_1_0_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	break;
      case Composition_0_1_0_1:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_0_1_0_0;
	break;
      case Composition_1_0_0_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_1_1_0_0;
	break;
      case Composition_1_0_1_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_1_1_1_0;
	break;
      case Composition_1_1_0_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	break;
      case Composition_1_1_0_1:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_1_1_0_0;
	break;
      case Composition_1_1_1_0:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	break;
      case Composition_1_1_1_1:
	_rhsBackup = obj;
	_comp1->loadRHS(obj);
	_state = Composition_1_1_1_0;
	break;
      case Composition_Sw_0_1_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_Sw_1_1_0_0;
	break;
      case Composition_Sw_0_1_0_1:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_Sw_1_1_0_1;
	break;
      case Composition_Sw_1_0_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	break;
      case Composition_Sw_1_0_1_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_Sw_1_0_0_0;
	break;
      case Composition_Sw_1_1_0_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	break;
      case Composition_Sw_1_1_0_1:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	break;
      case Composition_Sw_1_1_1_0:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_Sw_1_1_0_0;
	break;
      case Composition_Sw_1_1_1_1:
	_lhsBackup = obj;
	_comp1->loadLHS(obj);
	_state = Composition_Sw_1_1_0_1;
	break;
      };
  }; 
  template <class T>
  Comparison CCM<T>::compare()
  {
    CALL("compare()");
    Comparison cmp;
    switch (_state)
      {       
      case Inverse_Clean: return Incomparable;
      case Inverse_LHSLoaded: return Incomparable;
      case Inverse_RHSLoaded: return Incomparable;
      case Inverse_LHSRHSLoaded: return _comp1->inverseCompare();
      case Inverse_Sw_LHSLoaded: return Incomparable;
      case Inverse_Sw_RHSLoaded: return Incomparable;
      case Inverse_Sw_LHSRHSLoaded: return _comp1->compare();
      case Composition_0_0_0_0: return Incomparable;
      case Composition_0_1_0_0: return Incomparable;
      case Composition_0_1_0_1: return Incomparable;
      case Composition_1_0_0_0: return Incomparable;
      case Composition_1_0_1_0: return Incomparable;
      case Composition_1_1_0_0:
	cmp = _comp1->compare();
	if (isStrict(cmp)) return cmp;
	_comp2->loadLHS(_lhsBackup);
	_comp2->loadRHS(_rhsBackup);
	_state = Composition_1_1_1_1;
	return _comp2->compare();
      case Composition_1_1_0_1:
	cmp = _comp1->compare();
	if (isStrict(cmp)) return cmp;
	_comp2->loadLHS(_lhsBackup);
	_state = Composition_1_1_1_1;
	return _comp2->compare();
      case Composition_1_1_1_0:
	cmp = _comp1->compare();
	if (isStrict(cmp)) return cmp;
	_comp2->loadRHS(_rhsBackup);
	_state = Composition_1_1_1_1;
	return _comp2->compare();
      case Composition_1_1_1_1:
	cmp = _comp1->compare();
	if (isStrict(cmp)) return cmp;
	return _comp2->compare();
      case Composition_Sw_0_1_0_0: return Incomparable;
      case Composition_Sw_0_1_0_1: return Incomparable;
      case Composition_Sw_1_0_0_0: return Incomparable;
      case Composition_Sw_1_0_1_0: return Incomparable;
      case Composition_Sw_1_1_0_0:
	cmp = _comp1->inverseCompare();
	if (isStrict(cmp)) return cmp;
	_comp2->loadLHS(_lhsBackup);
	_comp2->loadRHS(_rhsBackup);
	_state = Composition_Sw_1_1_1_1;
	return _comp2->inverseCompare();
      case Composition_Sw_1_1_0_1:
	cmp = _comp1->inverseCompare();
	if (isStrict(cmp)) return cmp;
	_comp2->loadLHS(_lhsBackup);
	_state = Composition_Sw_1_1_1_1;
	return _comp2->inverseCompare();
      case Composition_Sw_1_1_1_0:
	cmp = _comp1->inverseCompare();
	if (isStrict(cmp)) return cmp;
	_comp2->loadRHS(_rhsBackup);
	_state = Composition_Sw_1_1_1_1;
	return _comp2->inverseCompare();
      case Composition_Sw_1_1_1_1:
	cmp = _comp1->inverseCompare();
	if (isStrict(cmp)) return cmp;
	return _comp2->inverseCompare();
      }; 
    return Incomparable;
  }; 
    template <class T>
      Comparison CCM<T>::inverseCompare()
      {
	CALL("inverseCompare()");
	Comparison cmp;
	switch (_state)
	  {       
	  case Inverse_Clean: return Incomparable;
	  case Inverse_LHSLoaded: return Incomparable;
	  case Inverse_RHSLoaded: return Incomparable;
	  case Inverse_LHSRHSLoaded: return _comp1->compare();
	  case Inverse_Sw_LHSLoaded: return Incomparable;
	  case Inverse_Sw_RHSLoaded: return Incomparable;
	  case Inverse_Sw_LHSRHSLoaded: return _comp1->inverseCompare();
	  case Composition_0_0_0_0: return Incomparable;
	  case Composition_0_1_0_0: return Incomparable;
	  case Composition_0_1_0_1: return Incomparable;
	  case Composition_1_0_0_0: return Incomparable;
	  case Composition_1_0_1_0: return Incomparable;
	  case Composition_1_1_0_0:
	    cmp = _comp1->inverseCompare();
	    if (isStrict(cmp)) return cmp;
	    _comp2->loadLHS(_lhsBackup);
	    _comp2->loadRHS(_rhsBackup);
	    _state = Composition_1_1_1_1;
	    return _comp2->inverseCompare();
	  case Composition_1_1_0_1:
	    cmp = _comp1->inverseCompare();
	    if (isStrict(cmp)) return cmp;
	    _comp2->loadLHS(_lhsBackup);
	    _state = Composition_1_1_1_1;
	    return _comp2->inverseCompare();
	  case Composition_1_1_1_0:
	    cmp = _comp1->inverseCompare();
	    if (isStrict(cmp)) return cmp;
	    _comp2->loadRHS(_rhsBackup);
	    _state = Composition_1_1_1_1;
	    return _comp2->inverseCompare();
	  case Composition_1_1_1_1:
	    cmp = _comp1->inverseCompare();
	    if (isStrict(cmp)) return cmp;
	    return _comp2->inverseCompare();
	  case Composition_Sw_0_1_0_0: return Incomparable;
	  case Composition_Sw_0_1_0_1: return Incomparable;
	  case Composition_Sw_1_0_0_0: return Incomparable;
	  case Composition_Sw_1_0_1_0: return Incomparable;
	  case Composition_Sw_1_1_0_0:
	    cmp = _comp1->compare();
	    if (isStrict(cmp)) return cmp;
	    _comp2->loadLHS(_lhsBackup);
	    _comp2->loadRHS(_rhsBackup);
	    _state = Composition_Sw_1_1_1_1;
	    return _comp2->compare();
	  case Composition_Sw_1_1_0_1:
	    cmp = _comp1->compare();
	    if (isStrict(cmp)) return cmp;
	    _comp2->loadLHS(_lhsBackup);
	    _state = Composition_Sw_1_1_1_1;
	    return _comp2->compare();
	  case Composition_Sw_1_1_1_0:
	    cmp = _comp1->compare();
	    if (isStrict(cmp)) return cmp;
	    _comp2->loadRHS(_rhsBackup);
	    _state = Composition_Sw_1_1_1_1;
	    return _comp2->compare();
	  case Composition_Sw_1_1_1_1:
	    cmp = _comp1->compare();
	    if (isStrict(cmp)) return cmp;
	    return _comp2->compare();
	  };
	return Incomparable;
      }; 
    template <class T>
      void CCM<T>::swapSides()
      {
	CALL("swapSides()");
	switch (_state)
	  {       
	  case Inverse_Clean: break;
	  case Inverse_LHSLoaded: _state = Inverse_Sw_RHSLoaded; break;
	  case Inverse_RHSLoaded: _state = Inverse_Sw_LHSLoaded; break;
	  case Inverse_LHSRHSLoaded: _state = Inverse_Sw_LHSRHSLoaded; break;
	  case Inverse_Sw_LHSLoaded: _state = Inverse_RHSLoaded; break;
	  case Inverse_Sw_RHSLoaded: _state = Inverse_LHSLoaded; break;
	  case Inverse_Sw_LHSRHSLoaded: _state = Inverse_LHSRHSLoaded; break;
	  case Composition_0_0_0_0: break;
	  case Composition_0_1_0_0: _state = Composition_Sw_0_1_0_0; break;
	  case Composition_0_1_0_1: _state = Composition_Sw_0_1_0_1; break;
	  case Composition_1_0_0_0: _state = Composition_Sw_1_0_0_0; break;
	  case Composition_1_0_1_0: _state = Composition_Sw_1_0_1_0; break;
	  case Composition_1_1_0_0: _state = Composition_Sw_1_1_0_0; break;
	  case Composition_1_1_0_1: _state = Composition_Sw_1_1_0_1; break;
	  case Composition_1_1_1_0: _state = Composition_Sw_1_1_1_0; break;
	  case Composition_1_1_1_1: _state = Composition_Sw_1_1_1_1; break;
	  case Composition_Sw_0_1_0_0: _state = Composition_0_1_0_0; break;
	  case Composition_Sw_0_1_0_1: _state = Composition_0_1_0_1; break;
	  case Composition_Sw_1_0_0_0: _state = Composition_1_0_0_0; break;
	  case Composition_Sw_1_0_1_0: _state = Composition_1_0_1_0; break;
	  case Composition_Sw_1_1_0_0: _state = Composition_1_1_0_0; break;
	  case Composition_Sw_1_1_0_1: _state = Composition_1_1_0_1; break;
	  case Composition_Sw_1_1_1_0: _state = Composition_1_1_1_0; break;
	  case Composition_Sw_1_1_1_1: _state = Composition_1_1_1_1; break;      
	  };
      }; 
}
namespace VK {
  class TempLitListExtras::CharacteristicComparisonMaker : 
    public BK::CCMaker<const BasicTempLitList*,ComparableCharacteristic>
  {
  public:
    CharacteristicComparisonMaker() {};
    CharacteristicComparisonMaker(ComparableCharacteristic* c1,
				  ComparableCharacteristic* c2) :
      BK::CCMaker<const BasicTempLitList*,ComparableCharacteristic>(c1,c2)
    {
    };
    ~CharacteristicComparisonMaker() {};
    void* operator new(size_t);
    void operator delete(void*);
  };  
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::ComplexComparisonMaker"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TempLitListExtras::ComplexComparisonMaker : 
    public BK::CCM<const BasicTempLitList*>
  {
  public:
    ComplexComparisonMaker() {};
    ComplexComparisonMaker(BK::CMaker<const BasicTempLitList*>* comp) :
      BK::CCM<const BasicTempLitList*>(comp)
    {
    };
    ComplexComparisonMaker(BK::CMaker<const BasicTempLitList*>* comp1,
			   BK::CMaker<const BasicTempLitList*>* comp2) :
      BK::CCM<const BasicTempLitList*>(comp1,comp2)
    {
    };
    ~ComplexComparisonMaker() {};
    void* operator new(size_t);
    void operator delete(void*);
  };  
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::Context"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TempLitListExtras::Context
  {
  public:
    class AutoActivator
    {
    public:
      AutoActivator(Context* c) : _previous(Context::getCurrent())
      {
	Context::setCurrent(c);
      };
      ~AutoActivator()
      {
	Context::setCurrent(_previous);
      };
    private:
      Context* _previous;
    }; 
  public:
    Context() {};
    ~Context() {};
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
    BK::ObjectPool<BasicTempLitList>* defaultPool()
    {
      return static_cast<BK::ObjectPool<BasicTempLitList>*>(&_defaultPool);
    };
  private:
    BK::ObjectPoolWithMemory<BK::GlobAlloc,BasicTempLitList> _defaultPool;
    BK::ObjectCache<BK::GlobAlloc,16UL,Size> _sizeAlloc;
    BK::ObjectCache<BK::GlobAlloc,16UL,NumOfVarOccurences> _numOfVarOccurencesAlloc;
    BK::ObjectCache<BK::GlobAlloc,16UL,NumOfLiterals> _numOfLiteralsAlloc;
    BK::ObjectCache<BK::GlobAlloc,16UL,CharacteristicComparisonMaker> _charactCompMakerAlloc;
    BK::ObjectCache<BK::GlobAlloc,16UL,ComplexComparisonMaker> _complexCompMakerAlloc;
    static Context* _current;
    friend class Size;
    friend class NumOfVarOccurences;
    friend class NumOfLiterals;
    friend class CharacteristicComparisonMaker;
    friend class ComplexComparisonMaker;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::StdCharacteristics"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TempLitListExtras::StdCharacteristics
  {
  public:
    StdCharacteristics() {};
    ~StdCharacteristics() {};
    void reset();
    ulong numOfAllLits() const { return _numOfNonanswerLits + _numOfAnswerLits; };
    ulong numOfNonanswerLits() const { return _numOfNonanswerLits; };
    ulong numOfAnswerLits() const { return _numOfAnswerLits; };
    ulong numOfEqLits() const { return _numOfEqLits; };
    ulong numOfNoneqNonanswerLits() const { return _numOfNonanswerLits - _numOfEqLits;  };
    ulong numOfNegNonanswerLits() const { return _numOfNegNonanswerLits; };
    ulong numOfPosNonanswerLits() const { return _numOfNonanswerLits - _numOfNegNonanswerLits; };
    ulong numOfNegSplittingBranches() const { return _numOfNegSplittingBranches; };
    ulong numOfPosSplittingBranches() const { return _numOfPosSplittingBranches; };
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    ulong numOfNextStateLits() const { return _numOfNextStateLits; };
    ulong numOfCurrStateLits() const { return _numOfCurrStateLits; };
#endif
    void collectFor(const BasicTempLitList& lst)
    {
      CALL("collectFor(const BasicTempLitList& lst)");
      reset();
      addFor(lst);
    };
    void addFor(const BasicTempLitList& lst);
    void addFor(const LitHandle& lit);
  private:
    ulong _numOfNonanswerLits;
    ulong _numOfAnswerLits;
    ulong _numOfEqLits;
    ulong _numOfNegNonanswerLits;
    ulong _numOfNegSplittingBranches;
    ulong _numOfPosSplittingBranches;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    ulong _numOfNextStateLits;
    ulong _numOfCurrStateLits;
#endif
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  bool 
  TempLitListExtras::containsOnlyAnswerLiterals(const BasicTempLitList& lst)
  {
    CALL("containsOnlyAnswerLiterals(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (!el->literal()->isAnswerLiteral())
	return false;
    return true;
  }; 
  inline
  bool 
  TempLitListExtras::containsAnswerLiterals(const BasicTempLitList& lst)
  {
    CALL("containsAnswerLiterals(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (el->literal()->isAnswerLiteral())
	return true;
    return false;
  }; 
  inline
  bool 
  TempLitListExtras::containsSplittingBranches(const BasicTempLitList& lst)
  {
    CALL("containsSplittingBranches(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (el->literal()->isSplittingBranch())
	return true;
    return false;
  }; 
  inline
  void 
  TempLitListExtras::preorientEqualities(const BasicTempLitList& lst)
  {
    CALL("preorientEqualities(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (el->literal()->isEquality())
	el->literal()->orderEqArgs();
  };
  inline
  void TempLitListExtras::adjustSelectionLayers(BasicTempLitList& lst)
  {
    CALL("adjustSelectionLayers(BasicTempLitList& lst)");
    if (lst.isEmpty()) return;
    if (lst.length() == 1UL)
      {
	LiteralProperties& prop = lst.first()->literal()->properties();
	prop.setSelectionLayerModifier(LiteralProperties::DefaultSLModifier);
	prop.setSelectionLayer(0UL);
	return;
      };
    bool defExists = false;
    ulong minLayerWithDef = ULONG_MAX;
    ulong maxLayerWithDef = 0UL;
    bool postExists = false;
    ulong minLayerWithPost = ULONG_MAX;
    ulong maxLayerWithPost = 0UL;
    bool accelExists = false;
    ulong minLayerWithAccel = ULONG_MAX;
    ulong maxLayerWithAccel = 0UL;
	const Element* el;
    for (el = lst.first(); el; el = el->next())
      {
	LiteralProperties& prop = el->literal()->properties();
	ulong layer = prop.selectionLayer();
	switch (prop.selectionLayerModifier())
	  {
	  case LiteralProperties::DefaultSLModifier: 
	    defExists = true;
	    if (layer < minLayerWithDef) minLayerWithDef = layer;
	    if (layer > maxLayerWithDef) maxLayerWithDef = layer;
	    break;	
	  case LiteralProperties::PostponeSelection:
	    postExists = true;
	    if (layer < minLayerWithPost) minLayerWithPost = layer;
	    if (layer > maxLayerWithPost) maxLayerWithPost = layer;
	    break;	
	  case LiteralProperties::AccelerateSelection:
	    accelExists = true;
	    if (layer < minLayerWithAccel) minLayerWithAccel = layer;
	    if (layer > maxLayerWithAccel) maxLayerWithAccel = layer;
	    break;	
	  };
      };
    if (!postExists && !accelExists && minLayerWithDef == 0UL)
      return; 
    ulong incForDef = 
      (accelExists) ? ((maxLayerWithAccel - minLayerWithAccel) + 1) : 0UL;
    ulong incForPost = 
      (defExists) ? ((maxLayerWithDef - minLayerWithDef) + 1 + incForDef) : incForDef;
    for (el = lst.first(); el; el = el->next())
      {
	LiteralProperties& prop = el->literal()->properties();
	ulong layer = prop.selectionLayer();
	switch (prop.selectionLayerModifier())
	  {
	  case LiteralProperties::DefaultSLModifier:
	    prop.setSelectionLayer(layer - minLayerWithDef + incForDef);
	    break;	
	  case LiteralProperties::PostponeSelection:
	    prop.setSelectionLayer(layer - minLayerWithPost + incForPost);
	    break;	
	  case LiteralProperties::AccelerateSelection:
	    prop.setSelectionLayer(layer - minLayerWithAccel);
	    break;	
	  };
	prop.setSelectionLayerModifier(LiteralProperties::DefaultSLModifier);
      };
  }; 
  inline
  void TempLitListExtras::freshList(Handle& h,Pool* pool)
  {
    CALL("freshList(Handle& h,Pool* pool)");
    ASSERT(pool);
    pool->reserveObject(h);
    h->reset();
  };
  inline
  void TempLitListExtras::freshList(Handle& h)
  {
    CALL("freshList(Handle& h)");
    ASSERT(Context::getCurrent());
    freshList(h,Context::getCurrent()->defaultPool());
  };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::StdCharacteristics"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  void TempLitListExtras::StdCharacteristics::reset()
  {
    CALL("reset()");
    _numOfNonanswerLits = 0UL;
    _numOfAnswerLits = 0UL;
    _numOfEqLits = 0UL;
    _numOfNegNonanswerLits = 0UL;
    _numOfNegSplittingBranches = 0UL;
    _numOfPosSplittingBranches = 0UL;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    _numOfNextStateLits = 0UL;
    _numOfCurrStateLits = 0UL;
#endif
  }; 
  inline
  void 
  TempLitListExtras::StdCharacteristics::addFor(const BasicTempLitList& lst)
  {
    CALL("addFor(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      addFor(el->literal());
  }; 
  inline
  void TempLitListExtras::StdCharacteristics::addFor(const LitHandle& lit)
  {
    CALL("addFor(const LitHandle& lit)");
    if (lit->isAnswerLiteral())
      {
	ASSERT(!lit->isEquality());      
	++_numOfAnswerLits;
      }
    else 
      {
	++_numOfNonanswerLits;
	if (lit->isEquality()) 
	  {
	    ++_numOfEqLits;
	  }
	else
	  {
	    if (lit->isSplittingBranch())
	      if (lit->isNegativeForSelection())
		{
		  ++_numOfNegSplittingBranches;
		}
	      else 
		++_numOfPosSplittingBranches;
	  };
	if (lit->isNegativeForSelection()) 
	  ++_numOfNegNonanswerLits;
      };
  }; 
}; 
#endif
