/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TEMPORARY_CLAUSE_COLLECTOR_HPP
#define TEMPORARY_CLAUSE_COLLECTOR_HPP
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "ObjectPool.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "TemporaryLiteral.hpp"
#include "LiteralProperties.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "ClauseBackground.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseQueue.hpp"
#include "Term.hpp"
#include "Clause.hpp"
#include "UnifTerm.hpp"
#include "Unifier.hpp"
#include "PagedPassiveClauseStorage.hpp"
#include "Statistics.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_COLLECTOR
#  define DEBUG_NAMESPACE "TemporaryClauseCollector"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class TemporaryClauseCollector
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef BasicTempLitList::Element LitListElement;
    typedef BK::ObjectPool<TemporaryClause>::EntryHandle ClauseHandle;
  public:
    enum RejectionReason
    {
      None,
      PropositionalTautology,
      EquationalTautology,
      ClauseTooLarge,
      LiteralTooLarge,
      TooManyLiterals,
      TooManyVariables,
      LiteralTooDeep
    };
  public:
    TemporaryClauseCollector();
#ifndef NO_DEBUG
    TemporaryClauseCollector(const char* name);
#endif
    ~TemporaryClauseCollector();
    void reset();
    void setMarkSubstitutionPositions(bool fl)
    {
      _markSubstitutionPositions = fl;
    };
    void setShowAnnouncedClauses(ostream* str)
    {
      _showAnnouncedClauses = str;
    };
    void setShowCompleteClauses(ostream* str)
    {
      _showCompleteClauses = str;
    };
    void setPreprocessingInProgress(bool fl)
    {
      _preprocessingInProgress = fl;
    };
    void setOutputQueue(TemporaryClauseQueue* q) { _outputQueue = q; };
    void setStatistics(Statistics* stat) { _statistics = stat; };
    void setTautologyCheck(bool fl) { _tautologyCheck = fl; };
    void setLiteralFactoring(bool fl) { _literalFactoring = fl; };
    void setSizeLimit(ulong l) 
    { 
      _sizeLimit = l; 
      if (_sizeLimit > VampireKernelConst::MaxClauseSize)
	_sizeLimit = VampireKernelConst::MaxClauseSize;
    };
    void setLiteralSizeLimit(ulong l) 
    { 
      _literalSizeLimit = l; 
      if (_literalSizeLimit > VampireKernelConst::MaxTermSize)
	_literalSizeLimit = VampireKernelConst::MaxTermSize;
    };
    void setLengthLimit(ulong l) 
    { 
      _lengthLimit = l;
      if (_lengthLimit > VampireKernelConst::MaxlNumOfLits)
	_lengthLimit = VampireKernelConst::MaxlNumOfLits;
    };
    RejectionReason rejectionReason() const { return _rejectionReason; };
    bool rejectedAsRedundant() const
    {
      return 
	(_rejectionReason == PropositionalTautology) ||
	(_rejectionReason == EquationalTautology);
    };
    void openNewClause(ClauseBackground::Rule rule);
    void openNewClause(ClauseBackground::Rule rule,Clause* parent)
    {
      CALL("openNewClause(ClauseBackground::Rule rule,Clause* parent)");
      openNewClause(rule);
      addParent(parent);
    };
    void openNewClause(ClauseBackground::Rule rule,
		       Clause* parent1,
		       Clause* parent2)
    {
      CALL("openNewClause(ClauseBackground::Rule rule,Clause* parent1,Clause* parent2)");
      openNewClause(rule);
      addParent(parent1);
      addParent(parent2);
    };
    void openNewClause(ClauseBackground::Rule rule,
		       Clause* parent1,
		       Clause* parent2,
		       Clause* parent3)
    {
      CALL("openNewClause(ClauseBackground::Rule rule,..,Clause* parent3)");
      openNewClause(rule);
      addParent(parent1);
      addParent(parent2);
      addParent(parent3);
    };
    void addParent(Clause* cl) 
    { 
      CALLM(_name,"addParent(Clause* cl)");
      if (cl->isSubgoal()) _newClause->background().markAsSubgoal();
      _newClause->background().pushParent(cl,cl->inferenceDepth());
    };
    void announceNewClause();
    bool finaliseClause();
    bool finaliseClause(ClauseHandle& recipient);
    bool finalisePropNameIntroPart(const TemporaryClause& parentCl);
    bool openNewNonpropLiteral(const Term& header,
			       const LiteralProperties& prop);
    bool pushPropLiteral(const Term& header,
			 const LiteralProperties& prop);
    bool openNewLiteral(const Term& header,
			const LiteralProperties& prop);
    LiteralProperties& currLitProperties() const {
      CALL("currLitProperties() const");
      return _currentLiteral->properties();
    };
    bool collectLiteral(LitHandle& lit);
    bool collectLiteralList(BasicTempLitList& lst);
    bool closeNonpropLiteral();
    bool closeLiteral();
    bool pushVar(ulong v);
    bool pushVar(ulong v,bool substPos);
    bool pushVar(const Term& var);
    bool pushVar(const Term& var,bool substPos);
    bool pushConst(const Term& c);
    bool pushConst(const Term& c,bool substPos);
    bool pushNonconstFunc(const Term& f);
    bool pushNonconstFunc(const Term& f,bool substPos);
    bool pushFunc(const Term& f);
    bool pushFunc(const Term& f,bool substPos);
    bool pushSymbol(const Term& sym);
    bool pushSymbol(const Term& sym,bool substPos);
    void addRule(ClauseBackground::Rule r) 
    { 
      _newClause->background().UsedRule(r); 
    };
    void addRules(const ClauseBackground::Rules& r)
    {
      _newClause->background().UsedRules(r); 
    };
    void makeVIP() { _newClause->background().MakeVIP(); };
    void markAsSupport() { _newClause->background().markAsSupport(); };
    void markAsSubgoal() { _newClause->background().markAsSubgoal(); };
    void markAsNonSubsuming() { _newClause->background().markAsNonSubsuming(); };
    void markAsNonSubsumable() { _newClause->background().markAsNonSubsumable(); };
    void addSimplifier(Clause* cl) 
    { 
      CALLM(_name,"addSimplifier(Clause* cl)");
      if (cl->isSubgoal()) _newClause->background().markAsSubgoal();
      _newClause->background().pushSimplifier(cl,cl->inferenceDepth());
    };
    void addSimplified(Clause* cl) 
    { 
      CALLM(_name,"addSimplified(Clause* cl)");
      if (cl->isSubgoal()) _newClause->background().markAsSubgoal();
      _newClause->background().pushSimplified(cl,cl->inferenceDepth());
    };
    bool pushLiteral(const Term* lit,
		     const LiteralProperties& prop);
    bool pushLiteral(const Term& header,
		     const Flatterm* arg1,
		     const WeightPolynomial::Monomials& varsForPadding,
		     const LiteralProperties& prop);
    bool loadClause(const Clause* c,
		    ClauseBackground::Rule rule,
		    Clause* parent);
    bool loadClause(PagedPassiveClauseStorage::ClauseLiterals& literals,
		    ClauseBackground::Rule rule);
    bool pushLiteralWithUnifier(ulong index,
				const Term* lit,
				const LiteralProperties& prop);
    bool pushLiteralWithUnifier(UnifTerm* lit,
				const LiteralProperties& prop); 
    bool pushEqLiteralWithUnifier(const Term& header,
				  UnifTerm* arg1,
				  UnifTerm* arg2,
				  const LiteralProperties& prop);
    bool pushTermWithUnifier(ulong index,const Term& t);
    bool pushTermWithUnifier(ulong index,const Term& t,bool inSubst);
    bool pushComplexTermWithUnifier(ulong index,const Term* t);
    bool pushComplexTermWithUnifier(ulong index,const Term* t,bool inSubst);
    bool pushTermWithUnifier(UnifTerm* t);
    bool pushTermWithUnifier(UnifTerm* t,bool inSubst);
    bool pushComplexTermWithUnifier(UnifTerm* t);
    bool pushComplexTermWithUnifier(UnifTerm* t,bool inSubst);
    bool pushLiteralWithReplacement(ulong index,
				    const Term* lit,
				    const Term* redex,
				    UnifTerm* replacingTerm,
				    const LiteralProperties& prop);
    bool pushLiteralWithReplacement(UnifTerm* lit,
				    UnifTerm* redex,
				    const Term& replacingTerm,
				    ulong index,
				    const LiteralProperties& prop);
    bool pushPseudoLiteralWithUnifier(ulong index,
				      const Term& header,
				      const Term& arg);
  private:
    void reject(const RejectionReason& reason);
    void registerGenerated(ClauseBackground::Rule rule);
  private:
    OPEN_BR_DECL(1);
    bool _markSubstitutionPositions;
    ostream* _showAnnouncedClauses;
    ostream* _showCompleteClauses;
    bool _preprocessingInProgress;
    TemporaryClauseQueue* _outputQueue;
    RejectionReason _rejectionReason;
    ClauseHandle _newClause;
    LitHandle _newLiteral;
    LitHandle _currentLiteral;
    Statistics* _statistics;
    Unifier::Variable::RenamingToInt _unifierVarRenaming;
    ulong _nonanswerSize;
    ulong _totalSize;
    bool _tautologyCheck;
    bool _literalFactoring;
    ulong _sizeLimit;
    ulong _literalSizeLimit;
    ulong _lengthLimit;
    bool _collectingNonanswerLiteral;    
    bool _collectingNonpropLiteral;
    CLOSE_BR_DECL(1);
    DOP(bool _debugTrash);
#ifndef NO_DEBUG
    const char* _name;
#endif 
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_COLLECTOR
#  define DEBUG_NAMESPACE "TemporaryClauseCollector"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  bool TemporaryClauseCollector::pushFunc(const Term& f) 
  {
    CALLM(_name,"pushFunc(const Term& f)");
    ASSERT(f.isComplex());
    return (f.arity())? pushNonconstFunc(f) : pushConst(f);
  };
  inline
  bool TemporaryClauseCollector::pushFunc(const Term& f,bool substPos) 
  {
    CALLM(_name,"pushFunc(const Term& f,bool substPos)");
    ASSERT(f.isComplex());
    return (f.arity())? pushNonconstFunc(f,substPos) : pushConst(f,substPos);
  };
  inline
  bool TemporaryClauseCollector::pushSymbol(const Term& sym)
  {
    CALLM(_name,"pushSymbol(const Term& sym)");
    if (sym.isVariable()) return pushVar(sym);
    return (sym.arity())? pushNonconstFunc(sym) : pushConst(sym);
  };
  inline
  bool TemporaryClauseCollector::pushSymbol(const Term& sym,bool substPos)
  {
    CALLM(_name,"pushSymbol(const Term& sym,bool substPos)");
    if (sym.isVariable()) return pushVar(sym,substPos);
    return (sym.arity())? pushNonconstFunc(sym,substPos) : pushConst(sym,substPos);
  };
  inline
  bool TemporaryClauseCollector::pushVar(ulong v)
  {
    CALLM(_name,"pushVar(ulong v)");    
    ASSERT(!_debugTrash);
    if (v >= VampireKernelConst::MaxNumOfVariables) 
      {
	reject(TooManyVariables);
	return false;
      };    
    _newLiteral->pushVar(v);  
    return true;
  }; 
  inline
  bool TemporaryClauseCollector::pushVar(ulong v,bool substPos)
  {
    CALLM(_name,"pushVar(ulong v,bool substPos)");    
    ASSERT(!_debugTrash);
    if (v >= VampireKernelConst::MaxNumOfVariables) 
      {
	reject(TooManyVariables);
	return false;
      };    
    _newLiteral->pushVar(v,substPos);  
    return true;
  }; 
  inline
  bool TemporaryClauseCollector::pushVar(const Term& var)
  {
    CALLM(_name,"pushVar(const Term& var)");
    ASSERT(!_debugTrash);
    _newLiteral->pushVar(var); 
    return true;
  }; 
  inline
  bool TemporaryClauseCollector::pushVar(const Term& var,bool substPos)
  {
    CALLM(_name,"pushVar(const Term& var,bool substPos)");
    ASSERT(!_debugTrash);
    _newLiteral->pushVar(var,substPos); 
    return true;
  }; 
  inline
  bool TemporaryClauseCollector::pushConst(const Term& c)
  {
    CALLM(_name,"pushConst(const Term& c)");
    ASSERT(!_debugTrash);
    _newLiteral->pushConst(c);
    return true;
  }; 
  inline
  bool TemporaryClauseCollector::pushConst(const Term& c,bool substPos)
  {
    CALLM(_name,"pushConst(const Term& c,bool substPos)");
    ASSERT(!_debugTrash);
    _newLiteral->pushConst(c,substPos);
    return true;
  }; 
}; 
#endif
