/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TEMPORARY_LITERAL_INL
#define TEMPORARY_LITERAL_INL
#include "TemporaryLiteral.hpp"
#include "LiteralProperties.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_LITERAL
#  define DEBUG_NAMESPACE "TemporaryLiteral"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  const WeightPolynomial& TemporaryLiteral::weight()
  {
    CALL("weight()");
    if (!_weightCollected) collectWeight();
    return _weight; 
  }; 
  inline
  const WeightPolynomial::Monomials& TemporaryLiteral::variableOccurences() 
  {
    CALL("variableOccurences()");
    if (!_variableOccurencesCollected) collectVariableOccurences();
    return _variableOccurences;
  }; 
  inline
  long TemporaryLiteral::numberOfVariableOccurences()
  {
    CALL("numberOfVariableOccurences()");
    return variableOccurences().sumOfCoefficients();
  }; 
  inline
  void TemporaryLiteral::variables(const ulong*& begin,const ulong*& end)
  {
    CALL("variables(const ulong*& begin,const ulong*& end)");
    if (_weightCollected)
      {
	begin = _weight.monomials().begin();
	end = _weight.monomials().end();
      }
    else
      if (_variableOccurencesCollected)
	{
	  begin = _variableOccurences.begin();
	  end = _variableOccurences.end();
	}
      else
	{
	  collectWeight();
	  begin = _weight.monomials().begin();
	  end = _weight.monomials().end();
	};
  }; 
  inline
  ulong TemporaryLiteral::numOfDifferentVariables() 
  {
    CALL("numOfDifferentVariables()");
    ASSERT(_debugLiteralComplete);
    if (_weightCollected)
      {
	return _weight.monomials().length();
      }
    else
      if (_variableOccurencesCollected)
	{
	  return _variableOccurences.length();
	}
      else
	{
	  collectWeight();
	  return _weight.monomials().length();
	};
  }; 
  inline
  bool TemporaryLiteral::containsVariable(ulong var)
  {
    CALL("containsVariable(ulong var)");
    ASSERT(_debugLiteralComplete);
    if (_weightCollected)
      {
	return _weight.monomials().coefficient(var) != 0;
      }
    else
      if (_variableOccurencesCollected)
	{
	  return _variableOccurences.coefficient(var) != 0;
	}
      else
	{
	  collectWeight();
	  return _weight.monomials().coefficient(var) != 0;
	};
  }; 
  inline
  const WeightPolynomial& TemporaryLiteral::arg1Weight()
  {
    CALL("arg1Weight()");
    ASSERT(!isAnswerLiteral());
    if (!_arg1Weight) collectArg1Weight();
    return *_arg1Weight; 
  };
  inline
  const WeightPolynomial& TemporaryLiteral::arg2Weight()
  { 
    CALL("arg2Weight()");
    ASSERT(!isAnswerLiteral());
    if (!_arg2Weight) collectArg2Weight();
    return *_arg2Weight; 
  };
  inline
  bool TemporaryLiteral::arg1ContainsVariable(ulong var)
  {
    CALL("arg1ContainsVariable(ulong var)");
    return arg1Weight().monomials().coefficient(var) != 0;
  };
  inline
  bool TemporaryLiteral::arg2ContainsVariable(ulong var)
  {
    CALL("arg2ContainsVariable(ulong var)");
    return arg2Weight().monomials().coefficient(var) != 0;
  };
  inline
  bool TemporaryLiteral::arg1IsGround()
  {
    CALL("arg1IsGround()");
    return arg1Weight().monomials().isZero();
  };
  inline
  bool TemporaryLiteral::arg2IsGround()
  {
    CALL("arg2IsGround()");
    return arg2Weight().monomials().isZero();
  };
  inline
  ulong TemporaryLiteral::numOfDifferentVariablesInArg1()
  {
    CALL("numOfDifferentVariablesInArg1()");
    return arg1Weight().numberOfVariables();
  };
  inline
  ulong TemporaryLiteral::numOfDifferentVariablesInArg2()
  {
    CALL("numOfDifferentVariablesInArg2()");
    return arg2Weight().numberOfVariables();
  }; 
  inline
  bool TemporaryLiteral::isPositiveForSelection() const
  {
    CALL("isPositiveForSelection() const");
    if ((!Context::getCurrent()->_inversePolarityForSelection) ||
	isEquality() || 
	isSplittingBranch() ||
	(Context::getCurrent()->_chainingTable && 
	 Context::getCurrent()->_chainingTable->transitivityLaw(_header.functor()))) 
      return isPositive();
    return isNegative();
  };
  inline
  bool TemporaryLiteral::isNegativeForSelection() const
  {
    CALL("isNegativeForSelection() const");
    if ((!Context::getCurrent()->_inversePolarityForSelection) ||
	isEquality() || 
	isSplittingBranch() ||
	(Context::getCurrent()->_chainingTable &&
	 Context::getCurrent()->_chainingTable->transitivityLaw(_header.functor()))) 
      return isNegative();
    return isPositive();
  };
  inline
  const Flatterm* TemporaryLiteral::arg1() const 
  {
    CALL("arg1() const");
    ASSERT(_debugLiteralComplete);
    return _arg1;
  };
  inline
  Flatterm* TemporaryLiteral::arg1() 
  {
    CALL("arg1()");
    ASSERT(_debugLiteralComplete);
    return _arg1;
  };
  inline
  const Flatterm* TemporaryLiteral::arg2() const 
  { 
    CALL("arg2() const");
    ASSERT(_debugLiteralComplete);
    return _arg2;
  };
  inline
  Flatterm* TemporaryLiteral::arg2()
  { 
    CALL("arg2()");
    ASSERT(_debugLiteralComplete);
    return _arg2;
  };
  inline
  const Flatterm* TemporaryLiteral::equalityLHS() const {
    CALL("equalityLHS() const");
    if (_properties.equalityOrientationMode() == 
	LiteralProperties::ReverseGivenEqualityOrientation)
      return arg2();
    return arg1();
  };
  inline
  const Flatterm* TemporaryLiteral::equalityRHS() const {
    CALL("equalityRHS() const");
    if (_properties.equalityOrientationMode() == 
	LiteralProperties::ReverseGivenEqualityOrientation)
      return arg1();
    return arg2();
  };
  inline
  void TemporaryLiteral::linkEqArgs() 
  { 
    CALL("linkEqArgs()");
    ASSERT(_debugLiteralComplete);
    ASSERT(flatterm() < (flatterm()->after()));
    flatterm()->after()->SetBackJump(flatterm() + 1); 
  };
  inline
  void TemporaryLiteral::makeEqUnordered()
  {
    CALL("makeEqUnordered()");
    _header.MakeUnordered();
    flatterm()->symbolRef() = _header;
    if (argumentsSwapped())
      {
	_arg1 = rawArg1();
	_arg2 = rawArg2();
	_properties.reverseEqualityOrientationMode();
      };
    _arg1Weight = 0;
    _arg2Weight = 0;
  };
  inline
  void TemporaryLiteral::pushVar(const ulong& v)
  {
    CALL("pushVar(const ulong& v)");
    ASSERT(!_debugLiteralComplete);
    _openFlatterm.PushVar(Term(TermVar,v));
  };
  inline
  void TemporaryLiteral::pushVar(const ulong& v,bool substPos)
  {
    CALL("pushVar(const ulong& v,bool substPos)");
    ASSERT(!_debugLiteralComplete);
    _openFlatterm.PushVar(Term(TermVar,v),substPos);
  };
  inline
  void TemporaryLiteral::pushVar(const Term& v) 
  {
    CALL("pushVar(const Term& v)");
    ASSERT(!_debugLiteralComplete);
    _openFlatterm.PushVar(v); 
  };
  inline
  void TemporaryLiteral::pushVar(const Term& v,bool substPos) 
  {
    CALL("pushVar(const Term& v,bool substPos)");
    ASSERT(!_debugLiteralComplete);
    _openFlatterm.PushVar(v,substPos); 
  };
  inline
  void TemporaryLiteral::pushConst(const Term& c) 
  {
    CALL("pushConst(const Term& c)");
    ASSERT(!_debugLiteralComplete);
    _openFlatterm.PushConst(c); 
  };
  inline
  void TemporaryLiteral::pushConst(const Term& c,bool substPos) 
  {
    CALL("pushConst(const Term& c,bool substPos)");
    ASSERT(!_debugLiteralComplete);
    _openFlatterm.PushConst(c,substPos); 
  };
  inline
  bool TemporaryLiteral::pushNonconstFun(const Term& f) 
  {
    CALL("pushNonconstFun(const Term& f)");
    ASSERT(!_debugLiteralComplete);
    if (_openFlatterm.Depth() >= VampireKernelConst::MaxTermDepth)
      return false;
    _size += f.arity();
    if (_size > VampireKernelConst::MaxTermSize)
      return false;
    _openFlatterm.PushNonconstFun(f); 
    return true;   
  };
  inline
  bool TemporaryLiteral::pushNonconstFun(const Term& f,bool substPos) 
  {
    CALL("pushNonconstFun(const Term& f,bool substPos)");
    ASSERT(!_debugLiteralComplete);
    if (_openFlatterm.Depth() >= VampireKernelConst::MaxTermDepth)
      return false;
    _size += f.arity();
    if (_size > VampireKernelConst::MaxTermSize)
      return false;
    _openFlatterm.PushNonconstFun(f,substPos); 
    return true;  
  };
  inline
  bool TemporaryLiteral::pushSym(const Term& sym)
  {
    CALL("pushSym(const Term& sym)");
    if (sym.isVariable()) { pushVar(sym); return true; };
    if (sym.arity()) return pushNonconstFun(sym);
    pushConst(sym);
    return true;   
  };
  inline
  bool TemporaryLiteral::pushSym(const Term& sym,bool substPos)
  {
    CALL("pushSym(const Term& sym,bool substPos)");
    if (sym.isVariable()) { pushVar(sym,substPos); return true; };
    if (sym.arity()) return pushNonconstFun(sym,substPos);
    pushConst(sym,substPos);
    return true;   
  };
  inline
  void TemporaryLiteral::pushNonpropHeader(const Term& hd,
					   const LiteralProperties& prop)
  {
    CALL("pushNonpropHeader(const Term& hd,..)");    
    _openFlatterm.Reset();
    _size = hd.arity() + 1;
    _weightCollected = false;
    _variableOccurencesCollected = false;
    _resistanceComputed = false;
    DOP(_debugLiteralComplete = false);
    DOP(_arg1 = 0);
    DOP(_arg2 = 0);
    _openFlatterm.PushNonconstFun(hd);
    _header = hd;
    _properties = prop;
  }; 
  inline
  void TemporaryLiteral::pushPropLit(const Term& lit,
				     const LiteralProperties& prop)
  {
    CALL("pushPropLit(const Term& lit,..)");   
    _openFlatterm.Reset();
    _size = 1;
    _weightCollected = false;
    _variableOccurencesCollected = false;
    _resistance = false;
    DOP(_arg1 = 0);
    DOP(_arg2 = 0);
    _openFlatterm.PushConst(lit);
    _header = lit;
    _properties = prop;
    DOP(_debugLiteralComplete = true);
  };
  inline
  bool 
  TemporaryLiteral::pushPseudoLit(const Term& header,
				  const Flatterm* arg1,
				  const WeightPolynomial::Monomials& varsForPadding,
				  const LiteralProperties& prop)
  {
    CALL("pushPseudoLit(const Term& header,const Flatterm* arg1,..)");    
    ASSERT(header.arity() == 1 + varsForPadding.length());
    _openFlatterm.Reset();
    _size = header.arity() + 1;
    _weightCollected = false;
    _variableOccurencesCollected = false;
    _resistance = false;
    DOP(_debugLiteralComplete = false);
    DOP(_arg1 = 0);
    DOP(_arg2 = 0);
    _openFlatterm.PushNonconstFun(header);
    Flatterm::ConstIterator iter(arg1); 
    do
      {
	if (!pushSym(iter.CurrPos()->symbolRef(),false)) return false;
	iter.NextSym();
      }
    while (iter.NotEndOfTerm());
    const ulong* endOfVars = varsForPadding.end();
    for (const ulong* var = varsForPadding.begin(); var != endOfVars; ++var)
      {
	pushVar(*var,true);
      };
    _header = header;
    _properties = prop;
    DOP(_debugLiteralComplete = true);
    return true;
  }; 
  inline
  void TemporaryLiteral::endOfLiteral()
  {
    CALL("endOfLiteral()");
    DOP(_debugLiteralComplete = true);
    ASSERT(!_weightCollected);
    if (header().IsEquality())
      {
	_arg1 = rawArg1();
	_arg2 = rawArg2();
        _arg1Weight = 0;
        _arg2Weight = 0;
	if (header().Positive())
	  {
	    LiteralProperties::EqualityMode mode = 
	      Context::getCurrent()->allPosEqMode();
	    if (mode != LiteralProperties::DefaultEqualityMode)
	      _properties.setEqualityMode(mode);
	  }
	else
	  {
	    LiteralProperties::EqualityMode mode = 
	      Context::getCurrent()->allNegEqMode();
	    if (mode != LiteralProperties::DefaultEqualityMode)
	      _properties.setEqualityMode(mode);
	  };
      };
  }; 
  inline
  bool TemporaryLiteral::argsAreEqual() const
  {
    CALL("argsAreEqual() const");
    ASSERT(isEquality());
    return arg1()->equal(arg2());
  };
  inline
  bool TemporaryLiteral::argsAreEqual(const SymbolInfo* commutativityTable,
				      bool& commutativityUsed) const
  {
    CALL("argsAreEqual(const SymbolInfo* commutativityTable) const");
    ASSERT(isEquality());
    return arg1()->equal(arg2(),commutativityTable,commutativityUsed);
  };
  inline
  void 
  TemporaryLiteral::commutativityLawsForArgsAreEqual(const SymbolInfo* commutativityTable,
						     BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForArgsAreEqual(..)");
    ASSERT(isEquality());
    arg1()->commutativityLawsForEqual(arg2(),commutativityTable,laws);
  }; 
  inline
  bool TemporaryLiteral::equal(const TemporaryLiteral& lit) const 
  {
    CALL("equal(const TemporaryLiteral& lit) const");
    return flatterm()->equal(lit.flatterm()); 
  };
  inline
  bool TemporaryLiteral::equal(const TemporaryLiteral& lit,
			       const SymbolInfo* commutativityTable,
			       bool& commutativityUsed) const 
  { 
    CALL("equal(const TemporaryLiteral& lit,const SymbolInfo* commutativityTable,bool& commutativityUsed) const");
    return flatterm()->equal(lit.flatterm(),commutativityTable,commutativityUsed); 
  };
  inline
  void 
  TemporaryLiteral::commutativityLawsForEqual(const TemporaryLiteral& lit,
					      const SymbolInfo* commutativityTable,
					      BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForEqual(..) const");
    flatterm()->commutativityLawsForEqual(lit.flatterm(),commutativityTable,laws);
  }; 
  inline
  bool TemporaryLiteral::isComplementary(const TemporaryLiteral& lit) const 
  {
    CALL("isComplementary(const TemporaryLiteral& lit) const"); 
    return flatterm()->Complementary(lit.flatterm()); 
  };
  inline
  bool TemporaryLiteral::isComplementary(const TemporaryLiteral& lit,
					 const SymbolInfo* commutativityTable,
					 bool& commutativityUsed) const 
  { 
    CALL("isComplementary(const TemporaryLiteral& lit,const SymbolInfo* commutativityTable,bool& commutativityUsed) const");
    return 
      flatterm()->Complementary(lit.flatterm(),
				commutativityTable,
				commutativityUsed); 
  };
  inline
  void 
  TemporaryLiteral::commutativityLawsForComplementary(const TemporaryLiteral& lit,
						      const SymbolInfo* commutativityTable,
						      BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForComplementary(..) const");
    flatterm()->commutativityLawsForComplementary(lit.flatterm(),
						  commutativityTable,
						  laws);
  }; 
  inline 
  bool TemporaryLiteral::complementaryEqLit(const TemporaryLiteral& lit) const 
  {
    CALL("complementaryEqLit(const TemporaryLiteral& lit) const");
    ASSERT(isEquality() && lit.isEquality());
    if (polarity() == lit.polarity()) return false;
    if (arg1()->equal(lit.arg1()))
      {
	return arg2()->equal(lit.arg2());
      }
    else
      {
	if ((isOrderedEquality()) && (lit.isOrderedEquality()))
	  {
	    return false;
	  }
	else return (arg1()->equal(lit.arg2())) && (arg2()->equal(lit.arg1()));
      };
  }; 
  inline 
  bool TemporaryLiteral::complementaryEqLit(const TemporaryLiteral& lit,
					    const SymbolInfo* commutativityTable,
					    bool& commutativityUsed) const 
  {
    CALL("complementaryEqLit(const TemporaryLiteral& lit,const SymbolInfo* commutativityTable,bool& commutativityUsed) const");
    ASSERT(isEquality() && lit.isEquality());
    if (polarity() == lit.polarity()) return false;
    if (arg1()->equal(lit.arg1(),commutativityTable,commutativityUsed))
      {
	bool commUsed;
	if (arg2()->equal(lit.arg2(),commutativityTable,commUsed))
	  {
	    commutativityUsed = commutativityUsed || commUsed;
	    return true;
	  };
	return false;
      }
    else
      {
	if ((isOrderedEquality()) && (lit.isOrderedEquality()))
	  {
	    return false;
	  }
	else 
	  {
	    if (arg1()->equal(lit.arg2(),commutativityTable,commutativityUsed))
	      {
		bool commUsed;
		if (arg2()->equal(lit.arg1(),commutativityTable,commUsed))
		  {
		    commutativityUsed = commutativityUsed || commUsed;
		    return true;
		  };
	      };
	    return false;
	  };
      };
  }; 
  inline
  void 
  TemporaryLiteral::commutativityLawsForComplementaryEqLit(const TemporaryLiteral& lit,
							   const SymbolInfo* commutativityTable,
							   BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForComplementaryEqLit(..) const");
    ASSERT(isEquality() && lit.isEquality());
    ASSERT(polarity() != lit.polarity());
    bool commUsed;
    if (arg1()->equal(lit.arg1(),commutativityTable,commUsed))
      {
	if (commUsed)
	  arg1()->commutativityLawsForEqual(lit.arg1(),commutativityTable,laws);
	arg2()->commutativityLawsForEqual(lit.arg2(),commutativityTable,laws);
      }
    else
      {
	ASSERT(isUnorderedEquality() || lit.isUnorderedEquality());
	arg1()->commutativityLawsForEqual(lit.arg2(),commutativityTable,laws);
	arg2()->commutativityLawsForEqual(lit.arg1(),commutativityTable,laws);
      };
  }; 
  inline 
  bool TemporaryLiteral::equalEqLit(const TemporaryLiteral& lit,
				    bool& eqSymmetryUsed) const 
  {
    CALL("equalEqLit(const TemporaryLiteral& lit,bool& eqSymmetryUsed) const");
    ASSERT(isEquality() && lit.isEquality());
    if (polarity() != lit.polarity()) return false;
    if (arg1()->equal(lit.arg1()))
      {   
	eqSymmetryUsed = false;
	return arg2()->equal(lit.arg2());
      }
    else
      {
	if ((isUnorderedEquality()) || (lit.isUnorderedEquality()))
	  {
	    eqSymmetryUsed = true;
	    return (arg1()->equal(lit.arg2())) && (arg2()->equal(lit.arg1()));
	  }
	else return false; 
      }; 
  }; 
  inline 
  bool TemporaryLiteral::equalEqLit(const TemporaryLiteral& lit,
				    const SymbolInfo* commutativityTable,
				    bool& commutativityUsed,
				    bool& eqSymmetryUsed) const 
  {
    CALL("equalEqLit(const TemporaryLiteral& lit,const SymbolInfo* commutativityTable,bool& commutativityUsed,bool& eqSymmetryUsed) const");
    ASSERT(isEquality() && lit.isEquality());
    if (polarity() != lit.polarity()) return false;
    if (arg1()->equal(lit.arg1(),commutativityTable,commutativityUsed))
      {
	bool commUsed;
	if (arg2()->equal(lit.arg2(),commutativityTable,commUsed))
	  {
	    eqSymmetryUsed = false;
	    commutativityUsed = commutativityUsed || commUsed;
	    return true;
	  };
	return false;
      }
    else
      {
	if ((isUnorderedEquality()) || (lit.isUnorderedEquality()))
	  {
	    if (arg1()->equal(lit.arg2(),commutativityTable,commutativityUsed))
	      {
		bool commUsed;
		if (arg2()->equal(lit.arg1(),commutativityTable,commUsed))
		  {
		    eqSymmetryUsed = true;
		    commutativityUsed = commutativityUsed || commUsed;
		    return true;
		  };
	      };
	    return false;
	  }
	else 
	  return false; 
      }; 
  }; 
  inline
  void 
  TemporaryLiteral::commutativityLawsForEqualEqLit(const TemporaryLiteral& lit,
						   const SymbolInfo* commutativityTable,
						   BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForEqualEqLit(..) const");    
    ASSERT(isEquality() && lit.isEquality());
    ASSERT(polarity() == lit.polarity());
    bool commUsed;
    if (arg1()->equal(lit.arg1(),commutativityTable,commUsed))
      {
	if (commUsed)
	  arg1()->commutativityLawsForEqual(lit.arg1(),commutativityTable,laws);
	arg2()->commutativityLawsForEqual(lit.arg2(),commutativityTable,laws);
      }
    else
      {
	ASSERT(isUnorderedEquality() || lit.isUnorderedEquality());
	arg1()->commutativityLawsForEqual(lit.arg2(),commutativityTable,laws);
	arg2()->commutativityLawsForEqual(lit.arg1(),commutativityTable,laws);
      };
  }; 
  inline 
  bool TemporaryLiteral::instantiationPossible(ulong var,
					       ulong substDepth,
					       long substSize,
					       bool checkInputNonBasicness)
  {
    CALL("instantiationPossible(ulong var,..)");
    ASSERT(substDepth);
    ulong occs = (ulong)(variableOccurences().coefficient(var));
    if (occs)
      {     
	ulong newSize = size() + (substSize * occs) - occs;
	if (newSize > VampireKernelConst::MaxTermSize ||
	    depth() + substDepth > VampireKernelConst::MaxTermDepth)
	  return false;
	if (!checkInputNonBasicness) return true;
	Flatterm::Iterator iter(flatterm());
	do
	  {
	    if (iter.CurrPos()->isVariable() &&
		iter.CurrPos()->Var() == var &&
		iter.CurrPos()->isInputNonBasicPosition())
	      return false;
	    iter.NextSym();
	  }
	while (!iter.EndOfTerm());
	return true;
      }
    else 
      return true;
  }; 
  inline 
  void TemporaryLiteral::orderEqArgs()
  {
    CALL("orderEqArgs()");
    ASSERT(isEquality());
    if (isOrderedEquality()) return; 
    ASSERT(arg1() == rawArg1());
    ASSERT(arg2() == rawArg2());
    switch (SimplificationOrdering::current()->compare(arg1(),arg1Weight(),
						       arg2(),arg2Weight()))
      {
      case BK::Less:
	swapArguments();
        _header.MakeOrdered();
        flatterm()->symbolRef() = _header;
	return;
      case BK::Equal: return;
      case BK::Greater:
        _header.MakeOrdered();
        flatterm()->symbolRef() = _header;
	return;
      case BK::Incomparable: return;
      };
    ICP("ICP0");  
    return;  
  }; 
  inline 
  const Flatterm* TemporaryLiteral::rawArg1() const
  {
    CALL("rawArg1() const");
    ASSERT(_debugLiteralComplete);
    return flatterm() + 1;
  };  
  inline 
  Flatterm* TemporaryLiteral::rawArg1()
  {
    CALL("rawArg1()");
    ASSERT(_debugLiteralComplete);
    return flatterm() + 1;
  };
  inline 
  const Flatterm* TemporaryLiteral::rawArg2() const
  {
    CALL("rawArg2() const");
    ASSERT(_debugLiteralComplete);
    return rawArg1()->after();
  };   
  inline 
  Flatterm* TemporaryLiteral::rawArg2()
  {
    CALL("rawArg2()");
    ASSERT(_debugLiteralComplete);
    return rawArg1()->after();
  };   
  inline 
  void TemporaryLiteral::swapArguments() 
  { 
    CALL("swapArguments()");
    ASSERT(_arg1 == rawArg1());
    ASSERT(_arg2 == rawArg2());
    _arg1 = rawArg2(); 
    _arg2 = rawArg1(); 
    ASSERT((!_arg1Weight) || (_arg1Weight == &_rawArg1Weight));
    ASSERT((!_arg2Weight) || (_arg2Weight == &_rawArg2Weight));
    WeightPolynomial* tmp = _arg1Weight;
    _arg1Weight = _arg2Weight;
    _arg2Weight = tmp;
    _properties.reverseEqualityOrientationMode();
  }; 
  inline 
  void TemporaryLiteral::collectWeight() 
  {
    CALL("collectWeight()");
    ASSERT(!_weightCollected);
    _weight.reset();
    TermWeighting::current()->collectWeight(_weight,flatterm());
    _weightCollected = true;
  };
  inline 
  void TemporaryLiteral::collectVariableOccurences()
  {
    CALL("collectVariableOccurences()");
    ASSERT(!_variableOccurencesCollected);
    _variableOccurences.reset();
    for (Flatterm::Iterator iter(flatterm());
	 iter.NotEndOfTerm();
	 iter.NextSym())
      {
	if (iter.CurrPos()->symbolRef().isVariable())
	  {
	    _variableOccurences.add1(iter.CurrPos()->symbolRef().var());
	  };
      };
    _variableOccurencesCollected = true;
  };
  inline 
  void TemporaryLiteral::collectArg1Weight()
  {
    CALL("collectArg1Weight()");
    ASSERT(!_arg1Weight);
    if (argumentsSwapped()) { _arg1Weight = &_rawArg2Weight; }
    else
      _arg1Weight = &_rawArg1Weight;
    _arg1Weight->reset();
    TermWeighting::current()->collectWeight(*_arg1Weight,arg1());
  };  
  inline 
  void TemporaryLiteral::collectArg2Weight()
  {
    CALL("collectArg2Weight()");
    ASSERT(!_arg2Weight);
    if (argumentsSwapped()) { _arg2Weight = &_rawArg1Weight; }
    else
      _arg2Weight = &_rawArg2Weight;
    _arg2Weight->reset();
    TermWeighting::current()->collectWeight(*_arg2Weight,arg2());
  };
  inline 
  ulong TemporaryLiteral::markInputNonBasicPositions()
  {
    CALL("markInputNonBasicPositions()");
    return flatterm()->markInputNonBasicPositions();
  }; 
  inline 
  void TemporaryLiteral::unmarkAllSubstitutionPositions()
  {
    CALL("unmarkAllSubstitutionPositions()");
    flatterm()->unmarkAllSubstitutionPositions();
  };
}; 
#endif
