/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TERM_HPP
#define TERM_HPP
#include <cstddef>
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "Comparison.hpp"
#include "Stack.hpp"
#include "ExpandingStack.hpp"
#include "BitSet.hpp"
#include "SmallHeaderSkipList.hpp"
#include "Signature.hpp"
#include "VariableSet.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Clause;
  const ulong TermArityMask         = 0x00003FFC;
  const ulong TermArityInvMask     = 0xFFFF9003;
  const ulong TermSplittingMask     = 0x00004000;
  const ulong TermPolarityMask      = 0x00008000;
  const ulong TermPolarityInvMask  = 0xFFFF7FFF;
  const ulong TermVarMask           = 0xFFFF0000;
  const ulong TermVarInvMask       = 0x0000FFFF;   
  const ulong TermFunctorMask       = 0xFFFF0000;
  const ulong TermFunctorInvMask   = 0x0000FFFF;
  enum TermTypeTag 
  { 
    TermRef      = 0x00000000, 
    TermVar      = 0x00000001,
    TermComplex  = 0x00000002 
  };
  class Term
  {
  public:
    typedef ulong Arity;
    typedef ulong Functor;
    typedef Signature::Polarity Polarity;
    static Polarity positivePolarity() { return Signature::PositivePolarity; };
    static Polarity negativePolarity() { return Signature::NegativePolarity; };
    class OpenCompact;
    class Iterator;
    class IteratorWithDepth;
    class Traversal;
    class Arguments;
    class OpenUnshared;
    class FastOpenUnshared;
    typedef BK::BitSet<BK::GlobAlloc> LitNumMask;
    typedef BK::SHSkList<BK::GlobAlloc,LitNumMask*,Clause*> OccList;
    class AssociatedInfo
    {
    public:
      AssociatedInfo() {};
      ~AssociatedInfo() {};
      OccList* occurences() { return _occurences; };
      const OccList* occurences() const { return _occurences; };
      void setOccurences(OccList* l) { _occurences = l; };
    private:
      OccList* _occurences;
    }; 
  public: 
    Term() {};
    explicit Term(const ulong& cont) : content(cont) {};
    explicit Term(Term* const & r) : ref(r) {};
    Term(const Term& t) : content(t.content) {};
    Term(TermTypeTag,ulong var) 
      : content(var) 
    {
      (content <<= 16) |= TermVar;
    };
    Term(Arity a,Functor f,Polarity p) : content(f)
    { 
      CALL("costructor Term(Arity a,Functor f,Polarity p)");
      ASSERT(a <= Term::maxArity());
      ASSERT(((ulong)p) == 0UL || ((ulong)p) == 1UL);
      (((((content <<= 1) |= p) <<= 13) |= a) <<= 2) |= TermComplex;  
    }; 
    ~Term() {};
    void Make(const ulong& c) { content = c; };
    void MakeVar(ulong var)
    {   
      ((content = var) <<= 16) |= TermVar;   
    };
    void MakeComplex(Arity a,Functor f,Polarity p)
    {
      CALL("MakeComplex(Arity a,Functor f,Polarity p)");
      ASSERT(a <= Term::maxArity());
      ((((((content = f) <<= 1) |= p) <<= 13) |= a) <<= 2) |= TermComplex;  
    }; 
    void makeReference(Term* r) { ref = r; };
    void MakeHeader(ulong hdNum)
    {
      ulong fun = hdNum/2;
      ulong pol = hdNum%2;
      MakeComplex((Term::Arity)0,(Term::Functor)fun,(Polarity)pol); 
    };
    void MakeSplitting() 
    {
      content |= TermSplittingMask;
    }; 
    TermTypeTag typeTag() const { return TermTypeTag((content << 30) >> 30); };
    Arity arity() const { return (content << 18) >> 20; };
    static Arity maxArity() 
    {
      return VampireKernelConst::MaxSymbolArity; 
    };
    static Functor maxFunctor()
    {
      CALL("maxFunctor()");
      ASSERT(VampireKernelConst::MaxlSymbolNum <= 65535);
      return VampireKernelConst::MaxlSymbolNum;
    };
    static ulong maxHeaderNum()
    {
      CALL("maxHeaderNum()");
      ASSERT(VampireKernelConst::MaxlHeaderNum <= maxFunctor()*2 + 1);
      return maxFunctor()*2 + 1;
    };
    Polarity polarity() const { return static_cast<Polarity>((content << 16) >> 31); };
    bool Positive() const { return polarity() == Signature::PositivePolarity; };
    bool Negative() const { return polarity() != Signature::PositivePolarity; };
    bool IsSplitting() const 
    {
      return ((content << 17) >> 31) != 0UL;
    };
    Functor functor() const { return content >> 16; };
    ulong HeaderNum() const 
    {
      return functor()*2 + polarity();
    };
    ulong var() const { return content >> 16; }; 
    const ulong& Content() const { return content; };
    ulong& Content() { return content; };
    void SetPolarity(Polarity p)
    {
      (content &= TermPolarityInvMask) |= 
	(static_cast<ulong>(p) << 15);   
    };   
    void SetVar(ulong v) 
    {
      (content &= TermVarInvMask) |= (v << 16);    
    };
    void makePositive() { SetPolarity(Signature::PositivePolarity); };
    void flipPolarity() 
    { 
      if (Positive()) 
	{ 
	  SetPolarity(Signature::NegativePolarity); 
	} 
      else 
	SetPolarity(Signature::PositivePolarity);
    };
    void SetFunctor(ulong f)
    {
      (content &= TermFunctorInvMask) |= (f << 16);
    };
    void MakeUnordered() { SetFunctor(VampireKernelConst::UnordEqNum); };
    void MakeOrdered() { SetFunctor(VampireKernelConst::OrdEqNum); };  
    void* operator new(size_t,ulong arity) 
    {    
      CALL("operator new(size_t,ulong arity)");
      return BK::GlobAlloc::allocate((arity + 1)*sizeof(Term) + sizeof(AssociatedInfo));
    };
    void operator delete(void* obj) 
    {
      CALL("operator delete(void* obj)");
      BK::GlobAlloc::deallocate(obj,((static_cast<Term*>(obj))->arity() + 1)*sizeof(Term) + sizeof(AssociatedInfo));
    }; 
#if (BK_COMPILER_FAMILY != BK_COMPILER_FAMILY_GCC_3)
    void operator delete(void* obj,ulong arity)
    {
      CALL("operator delete(void* obj,ulong arity)");
      BK::GlobAlloc::deallocate(obj,(arity + 1)*sizeof(Term) + sizeof(AssociatedInfo));
    };
#endif  
    Term& operator=(const Term& t) { content = t.content; return *this; };
    bool operator==(const Term& t) const { return content == t.content; };
    bool operator!=(const Term& t) const { return content != t.content; };
    bool operator<(const Term& t) const { return content < t.content; };
    bool operator<=(const Term& t) const { return content <= t.content; }; 
    bool operator>(const Term& t) const { return content > t.content; };
    bool operator>=(const Term& t) const { return content >= t.content; };
    bool hasGreaterPrecedenceThan(const Term& f) const 
    {
      return 
	Signature::current()->symbolPrecedence(functor()) > 
	Signature::current()->symbolPrecedence(f.functor());
    };
    BK::Comparison comparePrecedence(const Term& f) const
    {
      CALL("comparePrecedence(const Term& f) const");
      long diff = 
	Signature::current()->symbolPrecedence(functor()) - 
	Signature::current()->symbolPrecedence(f.functor());
      if (diff > 0L) return BK::Greater;
      if (diff < 0L) return BK::Less;
      ASSERT(functor() == f.functor());
      return BK::Equal;
    };
    long symbolPrecedence() const
    {
      CALL("symbolPrecedence() const");
      return Signature::current()->symbolPrecedence(functor());
    };
    long symbolWeight() const
    {
      return Signature::current()->weight(functor());
    };
    bool ComplementaryHeader(const Term& hd) const 
    {
      return (functor() == hd.functor()) && (polarity() != hd.polarity());
    };
    static bool EqualComplexTerms(const Term* t1,const Term* t2); 
    bool IsReference() const { return !typeTag(); };
    bool isVariable() const { return typeTag() == TermVar; };
    bool isComplex() const { return typeTag() == TermComplex; };
    bool IsConstant() const { return isComplex() && (!arity()); };
    bool InNonconstFunc() const { return isComplex() && arity(); };
    bool isNotArg() const { return typeTag() == TermComplex; };
    bool IsEquality() const { return VampireKernelConst::isEquality(functor()); };
    bool IsOrderedEq() const 
    {
      CALL("IsOrderedEq()");
      ASSERT(IsEquality());
      return VampireKernelConst::isOrderedEq(functor());
    };
    bool IsUnorderedEq() const 
    {
      CALL("IsUnorderedEq()");
      ASSERT(IsEquality());  
      return VampireKernelConst::isUnorderedEq(functor());
    };
    bool isBuiltInTruthValue() const
    {
      return VampireKernelConst::isBuiltInTrue(functor());
    };
    bool isPXYZ() const; 
    bool isAnswerHeader() const
    {
      return Signature::current()->isAnswerPredicate(functor());
    };
    bool isAnswerLiteral() const
    {
      return isAnswerHeader();
    };
    bool isEvaluable() const {
      CALL("isEvaluable() const");
      return
	IsEquality() ||
	Signature::current()->isEvaluableSymbol(functor());
    };
    bool isInterpretedConstant() const {
      CALL("isInterpretedConstant() const");
      return 
	Signature::current()->isInterpretedConstant(functor());
    };
    static bool Symmetry(const Term* t1,const Term* t2);
    Term* deref() const { return ref; }; 
    Term* Args() { return this + arity(); };
    const Term* Args() const { return this + arity(); };
    Term* Next() { return this - 1; };
    const Term* Next() const { return this - 1; };
    Term* NthArg(ulong n) { return Args() - n; }; 
    const Term* NthArg(ulong n) const { return Args() - n; };
    Term* Arg1() { return Args(); };
    const Term* Arg1() const { return Args(); };
    Term* Arg2() { return Args()->Next(); };
    const Term* Arg2() const { return Args()->Next(); };
    const bool* nonBasicArguments() const
    {
      CALL("nonBasicArguments() const");
      ASSERT(isComplex());
      if (Positive())
	{
	  return Signature::current()->positiveNonBasicArguments(functor());
	}
      else
	{
	  return Signature::current()->negativeNonBasicArguments(functor());
	};
    };
    bool containsSymbol(const Term& sym) const;
    bool containsComplexSubterm(const Term* t) const;
    bool isGround() const;
    void collectVariables(VariableSet& varSet) const;
    void destroyUnsharedComplex();
    void destroyUnsharedLit();
    AssociatedInfo* associatedInfo() const 
    { 
      return (AssociatedInfo*)(this + arity() + 1); 
    };
    OccList* Occurences() const
    {
      return associatedInfo()->occurences();
    }; 
    void InitOccurences() const { associatedInfo()->setOccurences((OccList*)0); };
    void DestroyOccurences() const;
    void RegisterOccurence(Clause* cl,ulong litNum) const;
    void RemoveOccurences(Clause* cl) const;
    void markBasicArgumentsInSignature() const;
    static ulong UnordHeaderNum(ulong header_num)
    {
      CALL("UnordHeaderNum(ulong header_num)");
      ASSERT(header_num/2 == VampireKernelConst::UnordEqNum || 
	     header_num/2 == VampireKernelConst::OrdEqNum);
      return VampireKernelConst::UnordEqNum*2 + header_num%2; 
    };
    const VampireKernelEnvironment::InterpretedConstant&
    interpretedConstantValue() const {
      CALL("interpretedConstantValue() const");
      return Signature::current()->interpretedConstantValue(functor());
    };
    static ostream& output(ostream& str,const Term& t);
    static ostream& output(ostream& str,const Term* t);
    ostream& output(ostream& str) const { return output(str,this); };
    static const Term& dummyPositiveHeader1() {
      static Term header((Term::Arity)1,(Term::Functor)VampireKernelConst::ReservedPred1,Signature::PositivePolarity);
      return header;
    };
    static const Term& positiveOrdEqHeader() { 
      static Term header((Term::Arity)2,(Term::Functor)VampireKernelConst::OrdEqNum,Signature::PositivePolarity);
      return header;
    };
    static const Term& negativeOrdEqHeader() { 
      static Term header((Term::Arity)2,(Term::Functor)VampireKernelConst::OrdEqNum,Signature::NegativePolarity);
      return header;
    };
    static const Term& positiveUnordEqHeader() { 
      static Term header((Term::Arity)2,(Term::Functor)VampireKernelConst::UnordEqNum,Signature::PositivePolarity);
      return header;
    };
    static const Term& negativeUnordEqHeader() { 
      static Term header((Term::Arity)2,(Term::Functor)VampireKernelConst::UnordEqNum,Signature::NegativePolarity);
      return header;
    };
    static const Term& builtInTrue() { 
      static Term header((Term::Arity)0,(Term::Functor)VampireKernelConst::BuiltInTrueNum,Signature::PositivePolarity);
      return header;
    };
    static const Term& builtInFalse() { 
      static Term header((Term::Arity)0,(Term::Functor)VampireKernelConst::BuiltInTrueNum,Signature::NegativePolarity);
      return header;
    };
  private:
    Term(const bool&);
    Term& operator=(const bool&);
    Term(const long&);
    Term& operator=(const long&);
  private:
    union 
    {
      ulong content;
      Term* ref;
    };
  }; 
}; 
#undef DEBUG_NAMESPACE   
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term::Traversal"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class Term::Traversal 
  {
  public:
    enum State { Func, Var, End };
  public:
    Traversal() 
    {
      CALLM("?","constructor Traversal()");
      DOP(_debugName = "??");
      wrapper[0] = Term::dummyPositiveHeader1();   
      ASSERT_IN(wrapper->isNotArg(),"end");
    }; 
    Traversal(const char* name)
    {
      CALLM(name,"constructor Traversal()");
      DOP(_debugName = name);
      wrapper[0] = Term::dummyPositiveHeader1();  
      ASSERT_IN(wrapper->isNotArg(),"end");
    }; 
    Traversal(const Term& t)
    {
      CALLM("?","constructor Traversal(const Term& t)");
      DOP(_debugName = "???");
      wrapper[0] = Term::dummyPositiveHeader1();  
      wrapper[1] = t;
      trm = wrapper + 1;
      if (t.isVariable()) 
	{
	  currState = Var;
	  sym = trm;
	}
      else 
	{
	  currState = Func;
	  sym = trm->deref();
	}; 
      ASSERT_IN(wrapper->isNotArg(),"end");
    };
    Traversal(const char* name,const Term& t)
    {
      CALLM(name,"constructor Traversal(const char* name,const Term& t)");
      DOP(_debugName = name);
      wrapper[0] = Term::dummyPositiveHeader1();  
      wrapper[1] = t;
      trm = wrapper + 1;
      if (t.isVariable()) 
	{
	  currState = Var;
	  sym = trm;
	}
      else 
	{
	  currState = Func;
	  sym = trm->deref();
	}; 
      ASSERT_IN(wrapper->isNotArg(),"end");
    };
    Traversal(const Term* t) 
    {
      CALLM("?","constructor Traversal(const Term* t)");
      DOP(_debugName = "???");
      ASSERT(t->isComplex());
      wrapper[0] = Term::dummyPositiveHeader1();  
      wrapper[1].makeReference((Term*)t);
      trm = wrapper + 1;
      currState = Func;
      sym = trm->deref();
      ASSERT_IN(wrapper->isNotArg(),"end");
    };
    Traversal(const char* name,const Term* t)
    {
      CALLM(name,"constructor Traversal(const char* name,const Term* t)");
      DOP(_debugName = name);
      ASSERT(t->isComplex());
      wrapper[0] = Term::dummyPositiveHeader1();  
      wrapper[1].makeReference((Term*)t);
      trm = wrapper + 1;
      currState = Func;
      sym = trm->deref();
      ASSERT_IN(wrapper->isNotArg(),"end");
    };
    void reset(const Term& t)
    {
      CALLM(_debugName,"reset(const Term& t)");
      ASSERT(wrapper->isNotArg());
      subterms.reset();
      wrapper[1] = t;
      trm = wrapper + 1;
      if (t.isVariable()) 
	{
	  currState = Var;
	  sym = trm;
	}
      else
	{
	  currState = Func;
	  sym = trm->deref();
	}; 
      ASSERT_IN(wrapper->isNotArg(),"end");
    };
    void reset(const Term* t)
    {
      CALLM(_debugName,"reset(const Term& t)");
      ASSERT(wrapper->isNotArg());
      ASSERT(t->isComplex());
      subterms.reset();
      wrapper[1].makeReference((Term*)t);
      trm = wrapper + 1;
      currState = Func;
      sym = trm->deref();
      ASSERT_IN(wrapper->isNotArg(),"end");
    };
    const State& state() const { return currState; };
    operator bool() const { return currState != End; };
    void next()
    {
      CALLM(_debugName,"next()");
      ASSERT(wrapper->isNotArg());
      const Term* nextTrm = trm->Next();
      if (currState == Func && sym->arity())
	{
	  if (!nextTrm->isNotArg()) subterms.push(nextTrm);
	  trm = sym->Args();
	}
      else 
	{
	  if (nextTrm->isNotArg())
	    { 
	      if (subterms.empty()) { currState = End; ASSERT(wrapper->isNotArg()); return; }
	      else 
		{          
		  trm = subterms.pop();
		};
	    }
	  else 
	    {
	      trm = nextTrm;
	    };
	};
      if (trm->isVariable())
	{
	  currState = Var;
	  sym = trm;
	}
      else 
	{
	  currState = Func;
	  sym = trm->deref();
	};
      ASSERT(wrapper->isNotArg());
    };
    void after()
    {
      CALLM(_debugName,"after()");
      ASSERT(wrapper->isNotArg());
      trm = trm->Next();
      if (trm->isNotArg()) 
	if (subterms.empty()) { currState = End; ASSERT(wrapper->isNotArg()); return; } 
	else trm = subterms.pop();
      if (trm->isVariable())
	{
	  currState = Var;
	  sym = trm;
	}
      else 
	{
	  currState = Func;
	  sym = trm->deref();
	};   
      ASSERT(wrapper->isNotArg());
    };
    const Term& symbol() const { return *sym; };
    const Term& term() const { return *trm; };
  private:
    Term wrapper[2];
    State currState;
    BK::Stack<const Term*,VampireKernelConst::MaxTermSize> subterms;
    const Term* trm;
    const Term* sym;
#ifndef NO_DEBUG
    const char* _debugName;
#endif
  }; 
}; 
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term::Arguments"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Term::Arguments
  {
  public:
    Arguments() {};
    Arguments(const Term& t) : _args((Term*)(t.deref()->Args())) {};
    Arguments(const Term* t) : _args((Term*)(t->Args())) {};
    Arguments(const Arguments& args) : _args(args._args) {};
    Arguments& operator=(const Arguments& args) { 
      _args = args._args; 
      return *this; 
    };
    void reset(const Term& t) 
    {
      CALL(" reset(const Term& t)");
      ASSERT(t.IsReference());
      ASSERT(t.deref()->arity());
      _args = (Term*)(t.deref()->Args()); 
    };
    void reset(const Term* t)
    {
      CALL("reset(const Term* t)");
      ASSERT(t->isComplex());
      ASSERT(t->arity());
      _args = (Term*)(t->Args());
    };
    Term* nth(ulong n) { return _args - n; };
    const Term* nth(ulong n) const { return _args - n; };
  private:
    Term* _args;
  }; 
}; 
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term::Iterator"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Term::Iterator 
  {
  public:
    Iterator() {};   
    Iterator(const Term* t) 
    { 
      sym = t;
      isVar = false;
    };     
    void Reset(const Term* t) 
    { 
      subterms.reset(); 
      sym = t;
      isVar = false;
    };                   
    const Term& symbolRef() const { return *sym; };
    const Term* CurrPos() const { return sym; };
    bool CurrentSymIsVar() const { return isVar; };
    bool Next()
    { 
      if (isVar)
	{
	  sym = sym->Next();
	}
      else
	{
	  sym = sym->Args();  
	};
      if (sym->isNotArg()) 
	{
	  if (subterms.empty()) { return false; };
	  sym = subterms.pop();
	};
      if (!((isVar = sym->isVariable()))) 
	{
	  if (!sym->Next()->isNotArg()) { subterms.push(sym->Next()); };
	  sym = sym->deref(); 
	};
      return true;
    };
    bool NextVar()
    {
      while (Next())
	{
	  if (isVar) { return true; };
	};
      return false;
    };
    bool SkipTerm()
    {
      if (isVar)
	{
	  sym = sym->Next();
	  if (sym->isNotArg()) 
	    {
	      if (subterms.empty()) { return false; };
	      sym = subterms.pop();
	    };
	}
      else
	{
	  if (subterms.empty()) { return false; };
	  sym = subterms.pop();
	}; 
      if (!((isVar = sym->isVariable()))) 
	{
	  if (!(sym->Next()->isNotArg())) { subterms.push(sym->Next()); };
	  sym = sym->deref(); 
	};
      return true; 
    }; 
  private:
    BK::Stack<const Term*,VampireKernelConst::MaxTermSize> subterms;   
    const Term* sym;
    bool isVar;  
  }; 
}; 
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "BracketCounter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <ulong MaxDepth>
  class BracketCounter
  {
  public:
    BracketCounter() : curr_holes(0) {  next_holes = holes; };
    ~BracketCounter() {};
    BracketCounter& operator=(const BracketCounter& bc)
    {
      CALL("operator=(const BracketCounter& bc)");
      next_holes = holes + (bc.next_holes - bc.holes);
      curr_holes = bc.curr_holes;
      int* orig = bc.next_holes;
      int* targ = next_holes;
      while (targ != holes)
	{
	  *targ = *orig;
	  --orig;
	  --targ;
	};
      return *this;
    };
    void Reset() { curr_holes = 0; next_holes = holes; };
    bool IsOpen() const { return curr_holes != 0; };
    void Open(int arity) 
    {
      CALL("Open(int arity)");  
      ASSERT(arity > 0); 
      ASSERT(next_holes < (holes + MaxDepth));
      *next_holes = curr_holes-1;
      next_holes++;
      curr_holes = arity;
    };
    unsigned long Close()
    {
      CALL("Close()");
      curr_holes--;
      if (curr_holes) return 0; 
      unsigned closed = 1;
    pop_holes:
      if (next_holes == (holes + 1)) return closed;
      next_holes--;
      ASSERT(next_holes >= holes); 
      curr_holes = (*next_holes);
      if (curr_holes) return closed;
      closed++;
      goto pop_holes;
    };
  private:
    int holes[MaxDepth];
    int* next_holes;
    int curr_holes;  
  }; 
}; 
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "ExpandingBracketCounter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ExpandingBracketCounter
  {
  public:
    ExpandingBracketCounter()
      : _allHoles(DOP("ExpandingBracketCounter::_allHoles"))
    {
    };
    ~ExpandingBracketCounter() {};
    void reset() { _currentNumOfHoles = 0; _allHoles.reset(); };
    bool isOpen() const { return _currentNumOfHoles != 0L; };
    void openSafe(ulong arity)
    {
      CALL("open(ulong arity)");
      ASSERT(arity > 0); 
      ASSERT(_currentNumOfHoles >= -1);
      _allHoles.pushSafe(_currentNumOfHoles - 1);
      _currentNumOfHoles = (long)arity;
    };
    ulong close() 
    {
      CALL("close()");
      _currentNumOfHoles--;
      if (_currentNumOfHoles) return 0UL; 
      ulong res = 1UL; 
    pop_holes: 
      if (_allHoles.empty()) return res;
      _currentNumOfHoles = _allHoles.pop();
      if (_currentNumOfHoles) return res;
      res++;         
      goto pop_holes;
    }; 
  private: 
    BK::ExpandingStack<BK::GlobAlloc,long,BK::FastRelocation<long>,128UL> _allHoles; 
    long _currentNumOfHoles;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "DepthCounter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class DepthCounter
  {
  public: 
    DepthCounter() {};
    DepthCounter& operator=(const DepthCounter& dc)
    {
      CALL("operator=(const DepthCounter& dc)");
      depth = dc.depth;
      return *this;
    };
    void Reset(ulong d) 
    {
      CALL("Reset(ulong d)");
      ASSERT(d);
      depth = d; 
    };
    ulong Depth() const { return depth; };
    void Down() 
    {
      depth++; 
    };
    void Up(ulong n) 
    { 
      CALL("Up(ulong n)");
      depth -= n; 
    };
  private:
    ulong depth;
  };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term::IteratorWithDepth"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Term::IteratorWithDepth  
    : private Iterator,
      private BracketCounter<VampireKernelConst::MaxTermDepth>,
      private DepthCounter 
  {
  public:
    IteratorWithDepth() {};
    ~IteratorWithDepth() {};
    void Reset(const Term* t,ulong d) 
    {
      CALL("Reset(const Term* t,ulong d)");
      ASSERT(d);
      Iterator::Reset(t);
      BracketCounter<VampireKernelConst::MaxTermDepth>::Reset();
      DepthCounter::Reset(d);
    };
    const Term& symbolRef() const { return Iterator::symbolRef(); };
    bool CurrentSymIsVar() const { return Iterator::CurrentSymIsVar(); };
    bool Next()
    {
      if (symbolRef().arity())
	{
	  BracketCounter<VampireKernelConst::MaxTermDepth>::Open(symbolRef().arity());
	  DepthCounter::Down();
	}
      else DepthCounter::Up(BracketCounter<VampireKernelConst::MaxTermDepth>::Close());
      return Iterator::Next();
    };
    ulong Depth() const { return DepthCounter::Depth(); }; 
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "SmartDepthCounter<ulong MaxDepth>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <ulong MaxDepth>
  class SmartDepthCounter 
    : private BracketCounter<MaxDepth>, private DepthCounter
  {
  public:  
    SmartDepthCounter() {};
    ~SmartDepthCounter() {};
    void Reset(const unsigned long d)
    {
      CALL("Reset(const unsigned long d)");
      BracketCounter<MaxDepth>::Reset();
      DepthCounter::Reset(d);
    };
    void NonconstantFunctor(int arity) 
    {
      CALL("NonconstantFunctor(int arity)");
      ASSERT(arity > 0);
      ASSERT(DepthCounter::Depth() < VampireKernelConst::MaxTermDepth);
      BracketCounter<MaxDepth>::Open(arity);
      DepthCounter::Down();
    };
    void VarOrConst() { DepthCounter::Up(BracketCounter<MaxDepth>::Close()); };
    void NextSym(const Term& sym)
    {
      if (sym.arity()) { NonconstantFunctor(sym.arity()); }
      else VarOrConst();       
    }; 
    unsigned long Depth() { return DepthCounter::Depth(); };
  }; 
}; 
#undef DEBUG_NAMESPACE   
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term::OpenCompact"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class Term::OpenCompact
  {
  public:
    OpenCompact() 
      : _holes("_holes"), 
    _numOfHoles("_numOfHoles")
#ifdef DEBUG_NAMESPACE
      , _debugDepthCounter()
#endif
    {
    };
    OpenCompact(Term* mem) 
      : _holes("_holes"), 
      _numOfHoles("_numOfHoles")
#ifdef DEBUG_NAMESPACE
      , _debugDepthCounter()
#endif
    {   
      _currHole = &_term;
      _currNumOfHoles = 1;
      _freeMem = mem;
      DOP(_debugDepthCounter.Reset(1));
      DOP(_debugSizeCounter = 1);
    };
    void reset(Term* mem)
    {
      _holes.reset();
      _numOfHoles.reset();
      _currHole = &_term;
      _currNumOfHoles = 1;
      _freeMem = mem;
      DOP(_debugDepthCounter.Reset(1));
      DOP(_debugSizeCounter = 1);
    };
    void pushVar(ulong var) 
    {
      CALL("pushVar(ulong var)");
      DOP(_debugDepthCounter.VarOrConst());
      ASSERT(_debugDepthCounter.Depth() && (_debugDepthCounter.Depth() <= VampireKernelConst::MaxTermDepth));
      ASSERT(_currHole);
      _currHole->MakeVar(var);
      _currNumOfHoles--;
      if (_currNumOfHoles) 
	{
	  _currHole = _currHole->Next();
	}
      else
	if (_holes.nonempty())
	  {
	    _currHole = _holes.pop();  
	    _currNumOfHoles = _numOfHoles.pop();
	    ASSERT(_currNumOfHoles);
	  }
	else
	  {
	    DOP(_currHole = (Term*)0);     
	  };
    };
    void pushVar(const Term& var)
    {
      CALL("pushVar(const Term& var)");
      ASSERT(var.isVariable());
      DOP(_debugDepthCounter.VarOrConst());
      ASSERT(_debugDepthCounter.Depth() && (_debugDepthCounter.Depth() <= VampireKernelConst::MaxTermDepth));
      ASSERT(_currHole);
      *_currHole = var;
      _currNumOfHoles--;
      if (_currNumOfHoles) 
	{
	  _currHole = _currHole->Next();
	}
      else
	if (_holes.nonempty())
	  {
	    _currHole = _holes.pop();  
	    _currNumOfHoles = _numOfHoles.pop();
	    ASSERT(_currNumOfHoles);
	  }
	else
	  {
	    DOP(_currHole = (Term*)0);
	  };
    };
    void pushConst(const Term& c) 
    {
      CALL("pushConst(const Term& c)");
      ASSERT(c.isComplex());
      ASSERT(!c.arity());
      DOP(_debugDepthCounter.VarOrConst());
      ASSERT(_debugDepthCounter.Depth() && (_debugDepthCounter.Depth() <= VampireKernelConst::MaxTermDepth));
      ASSERT(_currHole);
      _currHole->makeReference(_freeMem);   
      _currNumOfHoles--;
      *_freeMem = c;
      _freeMem++;
      if (_currNumOfHoles) 
	{
	  _currHole = _currHole->Next();
	}
      else
	if (_holes.nonempty())
	  {
	    _currHole = _holes.pop();  
	    _currNumOfHoles = _numOfHoles.pop();
	    ASSERT(_currNumOfHoles);
	  }
	else
	  {
	    DOP(_currHole = (Term*)0);
	  };    
    };
    void pushNonconstFun(const Term& f)
    {
      CALL("pushNonconstFun(const Term& f)");
      ASSERT(f.isComplex());
      ASSERT(f.arity());
      DOP(_debugDepthCounter.NonconstantFunctor(f.arity()));
      ASSERT(_debugDepthCounter.Depth() && (_debugDepthCounter.Depth() <= VampireKernelConst::MaxTermDepth));
      DOP(_debugSizeCounter += f.arity());
      ASSERT(_debugSizeCounter <= VampireKernelConst::MaxTermSize);
      ASSERT(_currHole);
      _currHole->makeReference(_freeMem);
      _currNumOfHoles--;
      if (_currNumOfHoles) 
	{
	  _holes.push(_currHole->Next());
	  _numOfHoles.push(_currNumOfHoles);
	};
      *_freeMem = f;
      _currHole = _freeMem->Args();
      _freeMem += (f.arity() + 1);
      _currNumOfHoles = f.arity();
    };
    void push(const Term& sym)
    {
      CALL("push(const Term& sym)");
      if (sym.isVariable()) { pushVar(sym); }
      else 
	if (sym.arity()) { pushNonconstFun(sym); } 
	else pushConst(sym);
    };
    Term* freeMemory() const { return _freeMem; };
  private:
    BK::Stack<Term*,VampireKernelConst::MaxTermDepth> _holes;
    BK::Stack<ulong,VampireKernelConst::MaxTermDepth> _numOfHoles;
    Term* _currHole;
    ulong _currNumOfHoles;
    Term _term;
    Term* _freeMem;
#ifdef DEBUG_NAMESPACE
    SmartDepthCounter<VampireKernelConst::MaxTermDepth + 1> _debugDepthCounter;
    ulong _debugSizeCounter;
#endif
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline bool Term::EqualComplexTerms(const Term* t1,const Term* t2)
  {
    CALL("EqualComplexTerms(const Term* t1,const Term* t2)");
    if (*t1 == *t2)
      {
	Iterator iter1;
	Iterator iter2;
	iter1.Reset(t1);
	iter2.Reset(t2);
	while (iter1.Next())
	  {
	    iter2.Next();
	    if (iter1.symbolRef() != iter2.symbolRef()) { return false; };
	  };
	return true;
      }
    else 
      {
	return false; 
      };
  }; 
  inline bool Term::containsSymbol(const Term& sym) const
  {
    CALL("containsSymbol(const Term& sym) const");
    ASSERT(isComplex());
    Iterator iter(this);
    do
      {
	if (iter.symbolRef() == sym) return true;
      }
    while (iter.Next());
    return false;
  }; 
}; 
namespace std
{
  inline ostream& operator<<(ostream& str,const VK::Term& t)
  {
    return VK::Term::output(str,t);
  }; 
  inline ostream& operator<<(ostream& str,const VK::Term* t)
  {
    return VK::Term::output(str,t);
  }; 
}; 
#endif
