/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef UNIF_INDEX_HPP
#define UNIF_INDEX_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "UnifTerm.hpp"
#include "Unifier.hpp"
#include "Multiset.hpp"
#include "Unification.hpp"
#include "BackParaCandidates.hpp"
#include "SimplificationOrdering.hpp"
#include "DestructionMode.hpp"
#include "Statistics.hpp"
#include "LiteralList.hpp"
namespace VK
{
  class Clause;
  class BackParaCandidates;
  class SymbolInfo;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex"
#endif 
#include "debugMacros.hpp"
namespace VK
{
class UnifIndex
{
public:
  class TreeNode;
  class IndexedObject;
  class Retrieval;
 public:
  UnifIndex();
  ~UnifIndex();
  void reset();
  void setStatistics(Statistics* stat) { _statistics = stat; };
  void SetParamodulation(bool fl) { paramodulation = fl; }; 
  void setParamodulateFromVariablesFlag(bool fl) 
  {
    _paramodulateFromVariables = fl;
  };
  void setUseInputNonBasicness(bool fl)
  {
    _useInputNonBasicness = fl;
  };
  void setBackSupNoDuplicateRedexes(bool fl)
  {
    _backSupNoDuplicateRedexes = fl;
  };
  void setChainingTable(const SymbolInfo* table) { _chainingTable = table; };
  void setChainingThroughVariablesFlag(bool fl)
  {
    _chainingThroughVariables = fl;
  };
  void setOrientEqualitiesInSupport(bool orientPos,bool orientNeg)
  {
    _orientPosEqualitiesInSupport = orientPos;
    _orientNegEqualitiesInSupport = orientNeg;
  };
  TreeNode** Tree(const Term& hd) 
    { 
      CALL("Tree(const Term& hd)");
      DOP(_literalIndex.unfreeze());
      TreeNode** res = &(_literalIndex.sub(hd.HeaderNum()));   
      DOP(_literalIndex.freeze());
      return res;
    };
  TreeNode** FParaTree(const Term& fun) 
    {
      CALL("FParaTree(const Term& fun)");
      DOP(_literalIndex.unfreeze());
      TreeNode** res = &(_literalIndex.sub(fun.HeaderNum()));   
      DOP(_literalIndex.freeze());
      return res;
    };
  TreeNode** FParaFromVarTree() { return &forward_para_from_var_tree; };
  TreeNode** BParaTree() { return &backward_para_tree; };
  TreeNode** pos1ChainingTree(ulong pred) 
  { 
    return &(_pos1ChainingTree.sub(pred)); 
  };
  TreeNode** pos2ChainingTree(ulong pred) 
  { 
    return &(_pos2ChainingTree.sub(pred)); 
  };
  TreeNode** neg1ChainingTree(ulong pred) 
  { 
    return &(_neg1ChainingTree.sub(pred)); 
  };
  TreeNode** neg2ChainingTree(ulong pred) 
  { 
    return &(_neg2ChainingTree.sub(pred)); 
  };
  void Integrate(Clause* c);
  void Remove(Clause* c);
 private:
  TreeNode* CreateBranch(TreeNode*& leaf);
  void Integrate(const LiteralList::Element* lit,Clause* cl,unsigned lit_num);
  TreeNode* IntegratePropLit(Term lit); 
  TreeNode* IntegrateArgs(TreeNode** tree); 
  void integrateForBackwardSuperposition(const LiteralList::Element* lit,
					 Clause* cl,
					 unsigned lit_num);
  void 
  integrateComplexSubtermForBackwardSuperposition(const Term* subterm,
						  Clause* cl,
						  unsigned litNum,
						  SimplificationOrdering::StoredConstraint* constr);
  void integrateBackSupRedex(const Term* redex,
			     Clause* cl,
			     unsigned lit_num,
			     SimplificationOrdering::StoredConstraint* constr);
  void integrateForForwardSuperposition(const LiteralList::Element* lit,Clause* cl,unsigned lit_num);
  void 
  integrateOrientedForForwardSuperposition(const LiteralList::Element* lit,
					   Clause* cl,
					   unsigned lit_num); 
  void integrateConstrainedForForwardSuperposition(const Term& lhs,const Term& rhs,Clause* cl,unsigned lit_num,
					  SimplificationOrdering::StoredConstraint* constr);
  void integrateForChaining(const Term* lit,Clause* cl,unsigned litNum);
  void integratePosForChaining(const Term* lit,Clause* cl,unsigned litNum);
  void integrateNegForChaining(const Term* lit,Clause* cl,unsigned litNum);
  void integratePos1ForChaining(ulong predNum,
				const Term& head,
				const Term& tail,
				Clause* cl,
				unsigned litNum,
				SimplificationOrdering::StoredConstraint* constraint);
  void integratePos2ForChaining(ulong predNum,
				const Term& head,
				const Term& tail,
				Clause* cl,
				unsigned litNum,
				SimplificationOrdering::StoredConstraint* constraint);
  void integrateNeg1ForChaining(ulong predNum,
				const Term& head,
				const Term& tail,
				Clause* cl,
				unsigned litNum,
				SimplificationOrdering::StoredConstraint* constraint);
  void integrateNeg2ForChaining(ulong predNum,
				const Term& head,
				const Term& tail,
				Clause* cl,
				unsigned litNum,
				SimplificationOrdering::StoredConstraint* constraint);
  TreeNode* IntegrateNonpropLit(const Term* lit); 
  TreeNode* integrateEqForResolution(const Term* lit); 
  TreeNode* LeafAndBranch(const Term* t,TreeNode** tree,TreeNode**& branch);
  void RemovePropLit(const Term* lit,Clause* cl);
  void Remove(const Term* t,Clause* cl,TreeNode** tree);
  void removeEqFromResolutionIndex(const Term* eqLit,Clause* cl);
  void RemoveBPCandidates(const Term* t,Clause* cl,TreeNode** tree);
  void RemoveBPRedex(const Term* t,Clause* cl);
  void removeComplexSubtermFromBPI(const Term* subterm,Clause* cl);
  void RemoveFromBPI(const LiteralList::Element* lit,Clause* cl);
  void RemoveFromFPI(const LiteralList::Element* lit,Clause* cl);
  void removeFromChainingIndex(const Term* lit,Clause* cl);
  void removePosFromChainingIndex(const Term* lit,Clause* cl);
  void removeNegFromChainingIndex(const Term* lit,Clause* cl);
  void Remove(const LiteralList::Element* lit,Clause* cl);
 private:
  Statistics* _statistics;
  Term::IteratorWithDepth l_iter;
  BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> var_counter;
  Term pseudo_literal[2];
  BK::Stack<const Term*,VampireKernelConst::MaxTermSize> _previousBackSupRedexes;
  BK::Array<BK::GlobAlloc,TreeNode*,BK::FastRelocation<TreeNode*>,512UL> _literalIndex;
  TreeNode* forward_para_from_var_tree;  
  TreeNode* backward_para_tree; 
  BK::Array<BK::GlobAlloc,TreeNode*,BK::FastRelocation<TreeNode*>,128UL> _pos1ChainingTree;
  BK::Array<BK::GlobAlloc,TreeNode*,BK::FastRelocation<TreeNode*>,128UL> _pos2ChainingTree;
  BK::Array<BK::GlobAlloc,TreeNode*,BK::FastRelocation<TreeNode*>,128UL> _neg1ChainingTree;
  BK::Array<BK::GlobAlloc,TreeNode*,BK::FastRelocation<TreeNode*>,128UL> _neg2ChainingTree;
  bool paramodulation;
  bool _paramodulateFromVariables;
  bool _useInputNonBasicness;
  bool _backSupNoDuplicateRedexes;
  const SymbolInfo* _chainingTable;
  bool _chainingThroughVariables;
  bool _orientPosEqualitiesInSupport;
  bool _orientNegEqualitiesInSupport; 
  SimplificationOrdering::SpecialisedComparison* _specialisedArgComparison;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex::IndexedObject"
#endif 
#include "debugMacros.hpp"
namespace VK
{
class UnifIndex::IndexedObject
{
 private: 
  unsigned lit_num;
  Clause* clause;
  Term term; 
  SimplificationOrdering::StoredConstraint* _orderingConstraint; 
  IndexedObject* next_pair;
 public: 
  IndexedObject(unsigned ln,Clause* cl,IndexedObject* np) : 
    lit_num(ln), 
    clause(cl), 
    _orderingConstraint(0),
    next_pair(np) 
    {
    };
  IndexedObject(unsigned ln,Clause* cl,const Term& t,SimplificationOrdering::StoredConstraint* cns,IndexedObject* np) 
    : lit_num(ln), 
    clause(cl), 
    term(t),
    _orderingConstraint(cns),
    next_pair(np) 
    {        
      CALL("IndexedObject(unsigned ln,Clause* cl,const Term& t,SimplificationOrdering::StoredConstraint* cns,IndexedObject* np)");
      ASSERT(term.isVariable() || term.IsReference());
      ASSERT((!term.isVariable()) || (term.var() < VampireKernelConst::MaxNumOfVariables));
      if (cns) cns = SimplificationOrdering::current()->copy(cns);
    }; 
  ~IndexedObject() 
    { 
      CALL("destructor ~IndexedObject()"); 
      if (BK::DestructionMode::isThorough())
	{
	  if (_orderingConstraint)
	    SimplificationOrdering::current()->releaseConstraint(_orderingConstraint);
	};
    };
  void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(IndexedObject));
    };
  void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(IndexedObject));
    };
 public:
  unsigned LitNum() const { return lit_num; };
  Clause* TheClause() const { return clause; };
  SimplificationOrdering::StoredConstraint*  orderingConstraint() { 
    return _orderingConstraint; 
  };
  const Term& ReplacingTerm() const { return term; }; 
  const Term& chainingTail() const { return term; };
  IndexedObject* Next() { return next_pair; };
  IndexedObject** Tail() { return &next_pair; };
  static void Remove(Clause* c,IndexedObject*& lst)
    {
      CALL("Remove(Clause* c,IndexedObject*& lst)");
      IndexedObject* pair_to_delete;
      while ((lst) && (lst->TheClause() == c))
	{
	  pair_to_delete = lst;
	  lst = lst->Next();
	  delete pair_to_delete;
	};
      if (!lst) return;
      IndexedObject** tail = lst->Tail();
      while (*tail) 
	{
	  if ((*tail)->TheClause() == c)
	    {
	      pair_to_delete = *tail;
	      *tail = (*tail)->Next();
	      delete pair_to_delete;  
	    }
	  else tail = (*tail)->Tail(); 
	}; 
    }; 
  static void Destroy(IndexedObject* lst)
    {
      CALL("Destroy(IndexedObject* lst)");
      IndexedObject* tmp;
      while (lst) 
	{
	  tmp = lst->Next();    
	  delete lst;      
	  lst = tmp;
	};
    }; 
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex::TreeNode"
#endif 
#include "debugMacros.hpp"
namespace VK
{
class UnifIndex::TreeNode
{
 public:
  enum Tag
  {
    Func,
    Const,
    Var,
    VarFirstOcc,
    Leaf
  };
 private:
  struct Intermediate
  {
    Tag tag;
    unsigned long depth;
    UnifTerm::Symbol symbol;
    TreeNode* below;
    TreeNode* alternative;
  };
  struct LeafNode
  {
    Tag tag;
    unsigned long depth;
    Term term;
    union
    {
      UnifIndex::IndexedObject* _indObjects;
      BackParaCandidates* _indObjectsForBP;
    };
  };
 public: 
  TreeNode() {};
  ~TreeNode() {};
  static TreeNode* NewNode(Tag t,UnifTerm::Symbol s,unsigned d)
    {
      Intermediate* res = static_cast<Intermediate*>(BK::GlobAlloc::allocate(sizeof(Intermediate)));
      res->tag = t;
      res->depth = d;
      res->symbol = s;
      res->below = 0;
      res->alternative = 0;
      return static_cast<TreeNode*>(static_cast<void*>(res));
    };
  void DeleteNode()
    {
      BK::GlobAlloc::deallocate(static_cast<void*>(this),sizeof(Intermediate));
    };
  static TreeNode* NewLeaf()
    {
      LeafNode* res = static_cast<LeafNode*>(BK::GlobAlloc::allocate(sizeof(LeafNode)));
      res->tag = Leaf;
      res->depth = 0;
      res->term.makeReference(0);
      res->_indObjects = 0; 
      return static_cast<TreeNode*>(static_cast<void*>(res));
    }; 
  void DeleteLeaf()
    {
      CALL("DeleteLeaf()");
      if ((static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects) UnifIndex::IndexedObject::Destroy((static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects);
      BK::GlobAlloc::deallocate(static_cast<void*>(this),sizeof(LeafNode));
    };
  void DeleteBPILeaf()
    {
      CALL("DeleteBPILeaf()");
      if ((static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjectsForBP) delete (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjectsForBP;
      BK::GlobAlloc::deallocate(static_cast<void*>(this),sizeof(LeafNode));
    };  
  const Tag& tag() const { return (static_cast<const Intermediate*>(static_cast<const void*>(this)))->tag; }; 
  bool IsLeaf() const { return (tag() == Leaf); };
  bool IsVar() const { return tag() == Var || tag() == VarFirstOcc; };
  const UnifTerm::Symbol& Symbol() const { return (static_cast<const Intermediate*>(static_cast<const void*>(this)))->symbol; };
  TreeNode* Below() const { return (static_cast<const Intermediate*>(static_cast<const void*>(this)))->below; };
  TreeNode** BelowAddr() { return &((static_cast<Intermediate*>(static_cast<void*>(this)))->below); };
  void SetBelow(TreeNode* b) { (static_cast<Intermediate*>(static_cast<void*>(this)))->below = b; };
  TreeNode* Alternative() const { return (static_cast<const Intermediate*>(static_cast<const void*>(this)))->alternative; };
  TreeNode** AlternativeAddr() { return &((static_cast<Intermediate*>(static_cast<void*>(this)))->alternative); };
  void SetAlternative(TreeNode* a) { (static_cast<Intermediate*>(static_cast<void*>(this)))->alternative = a; };
  unsigned long Depth() const { return (static_cast<const Intermediate*>(static_cast<const void*>(this)))->depth; }; 
  bool SameSymbol(const Term& s) const
    {
      if (s.isVariable())
	{
	  return (Symbol().var() == Unifier::current()->variableBase(s.var()));
	}
      else 
	{ 
	  return (Symbol().Func() == s);
	};
    }; 
  const Term& indexedTerm() const { 
    return (static_cast<const LeafNode*>(static_cast<const void*>(this)))->term; 
  };
  void SetTerm(const Term* t) { 
    (static_cast<LeafNode*>(static_cast<void*>(this)))->term.makeReference(const_cast<Term*>(t)); 
  };
  void SetTerm(const Term& t) { (static_cast<LeafNode*>(static_cast<void*>(this)))->term = t; };
  UnifIndex::IndexedObject* indObjects() const { return (static_cast<const LeafNode*>(static_cast<const void*>(this)))->_indObjects; };
  BackParaCandidates* BPCandidates() const { return (static_cast<const LeafNode*>(static_cast<const void*>(this)))->_indObjectsForBP; };
  void AddPair(unsigned ln,Clause* cl)
    {
      (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects = new UnifIndex::IndexedObject(ln,cl,(static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects);
    };
  void addForwardSupTriple(unsigned ln,Clause* cl,Term t,SimplificationOrdering::StoredConstraint* constr)
    {
      (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects = 
	new UnifIndex::IndexedObject(ln,cl,t,constr,(static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects);
    }; 
  void addBackSupTriple(unsigned ln,Clause* cl,const Term* t,SimplificationOrdering::StoredConstraint* constr)
    {
      CALL("addBackSupTriple(unsigned ln,Clause* cl,const Term* t,SimplificationOrdering::StoredConstraint* constr)");
      BackParaCandidates* candidates = (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjectsForBP;
      if (!candidates)         
	{
	  candidates = new BackParaCandidates();
	  (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjectsForBP = candidates;
	};    
      candidates->add(cl,ln,t,constr);
    }; 
  void addChainingQuadruple(unsigned litNum,
			    Clause* cl,
			    const Term& tail,
			    SimplificationOrdering::StoredConstraint* constr)
  {
    CALL("addChainingQuadruple(..)");    
    (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects = 
      new UnifIndex::IndexedObject(litNum,cl,tail,constr,(static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects);
  };
  void Remove(Clause* c)
    {
      UnifIndex::IndexedObject::Remove(c,(static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjects);
    };
  void RemoveBPCandidates(Clause* c)      
    {
      CALL("RemoveBPCandidates(Clause* c)");
      ASSERT((static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjectsForBP);
      BackParaCandidates* candidates = (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjectsForBP;
      candidates->remove(c); 
      if (candidates->empty()) 
	{
	  delete candidates;
	  (static_cast<LeafNode*>(static_cast<void*>(this)))->_indObjectsForBP = 0;
	};     
    };
 public: 
  static void Destroy(TreeNode* tree);
  static void DestroyBPI(TreeNode* tree);
 public: 
#ifndef NO_DEBUG_VIS
  ostream& output(ostream& str) const { return output(str,0); };
  ostream& output(ostream& str,unsigned var_offset) const;
#endif
}; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  ostream& operator<<(ostream& str,const VK::UnifIndex::TreeNode::Tag& tag);
};
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex::Retrieval"
#endif 
#include "debugMacros.hpp"
namespace VK
{
class UnifIndex::Retrieval
{  
 public: 
  enum Situation
  {
    StartOfExtraction,
    VarSameVar,
    VarAnything,
    FuncFunc,
    FuncFreeVar,
    FuncBoundVar
  };
 public:
  Retrieval()
    {
      for (ulong i = 0; i < VampireKernelConst::MaxTermSize; i++)
	{
	  term[i] = 0;
	};
    };
  void SetIndex(ulong index) 
    {
      CALL("SetIndex(ulong index)");
      var_offset = Unifier::variableOffset(index); 
    };
  ulong GetIndex() const { return Unifier::variableIndex(var_offset); };
  UnifIndex::TreeNode* CurrNode() const { return curr_node; };
  void Below()  
    {
      CALL("Below()");
      ASSERT(!(curr_node->IsLeaf()));
      *next = curr_node; 
      next++;
      curr_node = curr_node->Below();
    };
  void Alternative() 
    { 
      CALL("!(curr_node->IsLeaf())");
      ASSERT(!(curr_node->IsLeaf())); 
      curr_node = curr_node->Alternative();
    };
  void RestoreCurrent() { next--; curr_node = *next; };
  Unifier::Variable* Shift(Unifier::Variable* v) { 
    return (v + var_offset); 
  }; 
  void Start(UnifIndex::TreeNode* tree) 
    {
      free_memory = memory;
      curr_sit = backtrack_situation;
      next_term = term;
      ASSERT(tree);
      next = path;
      curr_node = tree;
      next_bookmark = bookmark;
    };
  void Bind(Unifier::Variable* v);
  bool AlterBinding();
  bool BindWithOccCheck(Unifier::Variable* v,bool& same_var);
  bool AlterBindingWithOccCheck(Unifier::Variable* v,bool& same_var);
  bool ExtractUnifiable(UnifTerm* word);
  bool AlterUnifiable();
#ifndef NO_DEBUG_VIS
  ostream& outputPath(ostream& str) const;  
  friend ostream& operator<<(ostream& str,const Retrieval::Situation& sit);
#endif  
 private:
  void ExtractTerm()
    {
      *next_bookmark = next; 
      next_bookmark++; 
      if (curr_node->IsVar())
	{
	  var_extracted = Shift(curr_node->Symbol().var());
	  Below();
	}
      else 
	{
	  var_extracted = 0;
	  unsigned long init_depth = curr_node->Depth();
	  do 
	    {
	      Below();
	    }
	  while (init_depth < curr_node->Depth());
	};
    }; 
  bool ModifyTerm();
  bool ExtractTermWithout(Unifier::Variable* var,bool& same_var);
  bool ModifyTermWithout(Unifier::Variable* var,
			 bool& same_var);
  void ToUnifTerm(UnifIndex::TreeNode** pos,UnifTerm*& memory);
  void NextIter() 
    {
      CALL("NextIter()");
      next_term++;
      curr_iter = *next_term;
      if (!curr_iter) 
	{
	  curr_iter = new UnifTerm::IterWithBacktracking(); 
	  *next_term = curr_iter;
	};
      ASSERT(next_term - term <= (long)VampireKernelConst::MaxTermSize);
    };
  void ForgetIter() 
    {
      CALL("ForgetIter()");
      next_term--; 
      ASSERT(next_term >= term);
      curr_iter = *next_term;
    };
  bool BacktrackExtraction();
  bool CompleteExtraction();
 private:
  UnifIndex::TreeNode** bookmark[VampireKernelConst::MaxTermSize];
  UnifIndex::TreeNode*** next_bookmark;
  UnifIndex::TreeNode* path[VampireKernelConst::MaxTermSize];
  UnifIndex::TreeNode** next;
  UnifIndex::TreeNode** changed_suffix;
  UnifIndex::TreeNode* curr_node;
  unsigned var_offset;
  Unifier::Variable* var_extracted;
  UnifTerm memory[VampireKernelConst::MaxTermSize];
  UnifTerm* free_memory;
  UnifTerm::IterWithBacktracking* term[VampireKernelConst::MaxTermSize];
  UnifTerm::IterWithBacktracking** next_term;
  UnifTerm::IterWithBacktracking* curr_iter;
  Situation backtrack_situation[VampireKernelConst::MaxTermSize*2];
  Situation* curr_sit; 
}; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  ostream& operator<<(ostream& str,const VK::UnifIndex::Retrieval::Situation& sit);
};
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex::Retrieval"
#endif 
#include "debugMacros.hpp"
namespace VK
{
inline bool UnifIndex::Retrieval::ModifyTerm()
{
  CALL("ModifyTerm()");
  UnifIndex::TreeNode** begin = (*(next_bookmark-1));
  ASSERT(begin < next);
 rollback:
  RestoreCurrent();
  changed_suffix = next;
  if (next == begin)  
    {
      Alternative();
      if (curr_node) { next_bookmark--; ExtractTerm(); return true; }
      else { next_bookmark--; return false; };
    };
  Alternative();
  if (curr_node)
    {
      unsigned long init_depth = (*begin)->Depth();
      do 
	{
	  Below();
	}
      while (init_depth < curr_node->Depth());
      var_extracted = 0;
      return true;
    }
  else { goto rollback; };
}; 
inline bool UnifIndex::Retrieval::ExtractTermWithout(Unifier::Variable* var,bool& same_var)
{
  CALL("ExtractTermWithout(Unifier::Variable* const var,bool& same_var)");
  *next_bookmark = next; 
  next_bookmark++;
  same_var = false;
 check_node:
  switch (curr_node->tag())
    {
    case UnifIndex::TreeNode::Func:
      {
	var_extracted = 0;
	unsigned long init_depth = curr_node->Depth();
	UnifIndex::TreeNode** init_pos = next;
	Below();
      check_depth:
        if (curr_node->Depth() <= init_depth) { return true; };
      intern_check_node:
	switch (curr_node->tag())
          {
	  case UnifIndex::TreeNode::Func: Below(); goto intern_check_node;
	  case UnifIndex::TreeNode::Const: Below(); goto check_depth;
	  case UnifIndex::TreeNode::Var: 
            {
	      bool free_var;
	      void* bind = Shift(curr_node->Symbol().var())->Unref(free_var);
	      if (free_var)
		{
		  if (static_cast<Unifier::Variable*>(bind) == var) { goto try_alternative; } 
		  else { Below(); goto check_depth; };
		}
	      else 
		{
		  if (UnifTerm::Occurs(var,static_cast<UnifTerm*>(bind)))
		    { goto try_alternative; }
		  else { Below(); goto check_depth; };
		};
            };
	  case UnifIndex::TreeNode::VarFirstOcc: Below(); goto check_depth;
#ifdef DEBUG_NAMESPACE
	  default: ICP("ICP1"); return false;
#elif defined _SUPPRESS_WARNINGS_
	  default: return false;
#endif
          };
      try_alternative:
        Alternative();
        if (curr_node) { goto intern_check_node; }
        else 
	  {
	    RestoreCurrent();
	    if (next == init_pos) 
	      {
		Alternative();
		if (curr_node) { goto check_node; }
		else { next_bookmark--; return false; };
	      }
	    else { goto try_alternative; };
	  };
      }; 
    case UnifIndex::TreeNode::Const: 
      Below(); 
      var_extracted = 0;
      return true;
    case UnifIndex::TreeNode::Var: 
      {
	bool free_var;
	void* bind = Shift(curr_node->Symbol().var())->Unref(free_var);
	if (free_var)
	  {
	    var_extracted = Shift(curr_node->Symbol().var());
	    Below();
	    same_var = ((static_cast<Unifier::Variable*>(bind)) == var);
	    return true;
	  }
	else
	  {
	    if (UnifTerm::Occurs(var,static_cast<UnifTerm*>(bind)))
	      {
		Alternative();
		if (curr_node) { goto check_node; }
		else { next_bookmark--; return false; };
	      }
	    else
	      {
		var_extracted = Shift(curr_node->Symbol().var());
		Below();
		return true;
	      };
	  };
      };
    case UnifIndex::TreeNode::VarFirstOcc:
      var_extracted = Shift(curr_node->Symbol().var());
      Below(); 
      return true;
#ifdef DEBUG_NAMESPACE
    default: ICP("ICP2"); return false;
#elif defined _SUPPRESS_WARNINGS_
    default: return false;
#endif
    };
  ICP("ICP3");
  return false;  
}; 
inline bool UnifIndex::Retrieval::ModifyTermWithout(Unifier::Variable* var,bool& same_var)
{
  CALL("ModifyTermWithout(Unifier::Variable* var,bool& same_var)");
  UnifIndex::TreeNode** begin = (*(next_bookmark-1));
  ASSERT(begin < next);
  unsigned long init_depth = (*begin)->Depth();
  same_var = false; 
  changed_suffix = next-1;
 rollback:
  RestoreCurrent();
  if (next < changed_suffix) { changed_suffix = next; };
  if (next == begin)  
    {
      Alternative();
      next_bookmark--;
      return ((curr_node) && (ExtractTermWithout(var,same_var)));
    };
 try_alternative:
  Alternative();
  if (curr_node)
    {
    check_node:
      switch (curr_node->tag())
        {
	case UnifIndex::TreeNode::Func: Below(); goto check_node;
	case UnifIndex::TreeNode::Const: Below(); goto check_depth;
	case UnifIndex::TreeNode::Var:
          {
	    bool free_var;
	    void* bind = Shift(curr_node->Symbol().var())->Unref(free_var);
	    if (free_var)
	      {
		if (static_cast<Unifier::Variable*>(bind) == var) { goto try_alternative; } 
		else { Below(); goto check_depth; };
	      }
	    else 
	      {
		if (UnifTerm::Occurs(var,static_cast<UnifTerm*>(bind)))
		  { goto try_alternative; }
		else { Below(); goto check_depth; };
	      };
          };
	case UnifIndex::TreeNode::VarFirstOcc: Below(); goto check_depth;
#ifdef DEBUG_NAMESPACE
	default: ICP("ICP1"); return false;
#elif defined _SUPPRESS_WARNINGS_
	default: return false;
#endif
        };
      ICP("ICP2");
      return false;
    check_depth:
      if (curr_node->Depth() <= init_depth) { var_extracted = 0; return true; }
      else 
        {
	  goto check_node; 
        };
    }
  else { goto rollback; };
}; 
inline void UnifIndex::Retrieval::ToUnifTerm(UnifIndex::TreeNode** pos,UnifTerm*& memory)
{
  CALL("ToUnifTerm(UnifIndex::TreeNode** pos,UnifTerm*& memory)");
  UnifIndex::TreeNode* node;
  while (pos < next)
    {
      node = *pos;
      switch (node->tag())
	{
	case UnifIndex::TreeNode::Func:
	  memory->symbol() = node->Symbol();
	  memory->ResetBracket();
	  break;
	case UnifIndex::TreeNode::Const:
	  memory->symbol() = node->Symbol();
	  memory->SetBracket(memory+1);
	  break; 
	case UnifIndex::TreeNode::Var: 
	case UnifIndex::TreeNode::VarFirstOcc: 
	  memory->symbol().MkVar(Shift(node->Symbol().var()));
	  memory->SetBracket(memory+1); 
	  break;
#ifdef DEBUG_NAMESPACE
	default: ICP("ICP1"); return; 
#elif defined _SUPPRESS_WARNINGS_
	default: return; 
#endif
	};
      pos++;
      memory++;   
    }; 
}; 
inline void UnifIndex::Retrieval::Bind(Unifier::Variable* v)
{
  ExtractTerm();
  if (var_extracted)
    {
      Unifier::current()->Bind(v,static_cast<void*>(var_extracted));
    }
  else 
    {
      UnifTerm* word = free_memory;
      ToUnifTerm(*(next_bookmark-1),free_memory);
      word->SetBracket(free_memory);
      Unifier::current()->Bind(v,static_cast<void*>(word));
    };
}; 
inline bool UnifIndex::Retrieval::AlterBinding()
{
  CALL("AlterBinding()");
  Unifier::Variable* v = Unifier::current()->LastBound();
  void* bind = v->Binding();
  ASSERT(bind);
  Unifier::current()->UnbindLast();
  if (ModifyTerm())
    {
      if (var_extracted)
	{
	  Unifier::current()->Bind(v,static_cast<void*>(var_extracted));
	  if (!(Unifier::current()->isVariable(bind))) 
	    {
	      free_memory = static_cast<UnifTerm*>(bind); 
	    };
	}
      else 
	{
	  if (Unifier::current()->isVariable(bind))
	    {
	      UnifTerm* word = free_memory;
	      ToUnifTerm(*(next_bookmark-1),free_memory);
	      word->SetBracket(free_memory);
	      Unifier::current()->Bind(v,static_cast<void*>(word));
	    }
	  else
	    {
	      UnifTerm* ch_suffix = static_cast<UnifTerm*>(bind) + (changed_suffix - (*(next_bookmark-1)));
	      free_memory = ch_suffix;
	      ASSERT(free_memory >= (static_cast<UnifTerm*>(bind)));
	      UnifTerm* old_end_of_word = (static_cast<UnifTerm*>(bind))->Bracket();
	      ToUnifTerm(changed_suffix,free_memory);
	      ASSERT(free_memory > (static_cast<UnifTerm*>(bind)));
	      (static_cast<UnifTerm*>(bind))->SetBracket(free_memory);
	      UnifTerm::resetCorruptBrackets((static_cast<UnifTerm*>(bind))+1,ch_suffix,old_end_of_word,free_memory);
	      Unifier::current()->Bind(v,bind);
	    };
	}
      return true;
    }
  else 
    {
      if (!(Unifier::current()->isVariable(bind))) { free_memory = static_cast<UnifTerm*>(bind); };
      return false; 
    };
}; 
inline bool UnifIndex::Retrieval::BindWithOccCheck(Unifier::Variable* v,bool& same_var)
{
  CALL("BindWithOccCheck(Unifier::Variable* v,bool& same_var)");
  if (ExtractTermWithout(v,same_var))
    {
      if (same_var) { return true; };
      if (var_extracted)
	{
	  Unifier::current()->Bind(v,static_cast<void*>(var_extracted)); 
	}
      else 
	{
	  UnifTerm* word = free_memory;
	  ASSERT(free_memory);
	  ToUnifTerm(*(next_bookmark-1),free_memory);
	  word->SetBracket(free_memory);
	  ASSERT(UnifTerm::CorrectComplexTerm(word));
	  ASSERT(!(UnifTerm::Occurs(v,word)));
	  Unifier::current()->Bind(v,static_cast<void*>(word));
	};
      return true;
    }
  else 
    {
      return false;
    };
}; 
inline bool UnifIndex::Retrieval::AlterBindingWithOccCheck(Unifier::Variable* v,bool& same_var)
{
  CALL("AlterBindingWithOccCheck(Unifier::Variable* v,bool& same_var)");
  void* bind = v->Binding();
  bool was_var = true;
  if (bind) 
    {
      ASSERT(v == Unifier::current()->LastBound());
      Unifier::current()->UnbindLast(); 
      if (!(Unifier::current()->isVariable(bind))) { was_var = false; free_memory = static_cast<UnifTerm*>(bind); };
    };
  if (ModifyTermWithout(v,same_var))
    {
      if (var_extracted)
	{
	  if (!(same_var)) { Unifier::current()->Bind(v,static_cast<void*>(var_extracted)); };
	}
      else 
	{
	  if (was_var)
	    {
	      UnifTerm* word = free_memory;
	      ToUnifTerm(*(next_bookmark-1),free_memory);
	      ASSERT(free_memory);
	      word->SetBracket(free_memory);
	      ASSERT(UnifTerm::CorrectComplexTerm(word));
	      ASSERT(!(UnifTerm::Occurs(v,word)));
	      Unifier::current()->Bind(v,static_cast<void*>(word));
	    }
	  else
	    {
	      UnifTerm* ch_suffix = static_cast<UnifTerm*>(bind) + (changed_suffix - (*(next_bookmark-1)));
	      free_memory = ch_suffix;
	      ASSERT(free_memory);
	      UnifTerm* old_end_of_word = (static_cast<UnifTerm*>(bind))->Bracket();
	      ToUnifTerm(changed_suffix,free_memory);
	      ASSERT(free_memory);
	      (static_cast<UnifTerm*>(bind))->SetBracket(free_memory);
	      UnifTerm::resetCorruptBrackets((static_cast<UnifTerm*>(bind))+1,ch_suffix,old_end_of_word,free_memory);
	      ASSERT(UnifTerm::CorrectComplexTerm(static_cast<UnifTerm*>(bind)));
	      ASSERT(!(UnifTerm::Occurs(v,static_cast<UnifTerm*>(bind))));
	      Unifier::current()->Bind(v,bind);
	    };
	}
      return true;
    }
  else 
    {
      return false; 
    };
}; 
}; 
#endif
