/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef UNIF_TERM_HPP
#define UNIF_TERM_HPP
#ifndef NO_DEBUG_VIS 
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "Unifier.hpp"
#include "Stack.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class UnifTerm
  {
  public:
    class Symbol 
    {
    public:
      Symbol() {};
      Symbol(const Term& s) 
      {
	if (s.isVariable())
	  {
	    MkVar(Unifier::current()->variableBase(s.var()));
	  }
	else 
	  {
	    sym = s;
	  };
      };
      ~Symbol() {};
      bool operator==(const Symbol& s) const { return sym == s.sym; };
      bool operator!=(const Symbol& s) const { return sym != s.sym; };
      bool IsFunctor() const { return !isVariable(); };
      bool isVariable() const { return sym.typeTag() == TermRef; };
      const Term& Func() const { return sym; };
      Term& Func() { return sym; };
      Unifier::Variable* var() const 
      { 
	return static_cast<Unifier::Variable*>(static_cast<void*>(sym.deref())); 
      };
      void MkFunc(const Term& f) { sym = f; };
      void flipPolarity() { sym.flipPolarity(); };
      void MkVar(const Unifier::Variable* v) 
      { 
	sym.makeReference(const_cast<Term*>(static_cast<const Term*>(static_cast<const void*>(v)))); 
      };
    public: 
      ostream& output(ostream& str) const { return output(str,0); };
      ostream& output(ostream& str,unsigned var_offset) const
      {
	if (IsFunctor()) return str << Func();
	return (var() + var_offset)->output(str);
      };
    private:
      Term sym;
    }; 
  public: 
    UnifTerm() {};
    ~UnifTerm() {};
    UnifTerm::Symbol& symbol() { return head; };
    const UnifTerm::Symbol& symbol() const { return head; };
    void SetBracket(UnifTerm* br) { bracket = br; };
    void ResetBracket() { bracket = 0; };
    UnifTerm* Bracket() const { return bracket; };
    UnifTerm* Arg1() { return this + 1; };
    UnifTerm* Arg2() { return Arg1()->Bracket(); };
    ostream& output(ostream& str) const
    {
      return symbol().output(str) << " [" << (ulong)Bracket() << ']'; 
    };
    long size() const { return Bracket() - this; };
    static
    void StandardTermToUnifTerm(ulong index,const Term* t,UnifTerm*& memory);
    static
    void StandardTermToUnifTerm(ulong index,const Term& t,UnifTerm*& memory);
    static
    void binaryStandardTermToUnifTerm(ulong index,
				      const Term& top,
				      const Term& arg1,
				      const Term& arg2,
				      UnifTerm*& memory);
  public:
    class SimpleIter;
    class SimpleConstIter;
    class Iter;
    class ConstIter;
    class IterWithBacktracking;
  public:
    static bool equal(UnifTerm* term1,UnifTerm* term2);
    static bool equalComplex(UnifTerm* term1,UnifTerm* term2);
    static bool identical(const UnifTerm* term1,const UnifTerm* term2);
    static bool containsSubterm(const UnifTerm* term1,
				const UnifTerm* term2);
#ifndef NO_DEBUG
    static bool CorrectComplexTerm(const UnifTerm* term);
#endif
#ifndef NO_DEBUG_VIS 
    static ostream& outputTerm(ostream& str,const UnifTerm* term);
#endif
  public:
    static void CorrectBracket(UnifTerm* term);
  public: 
    static void resetCorruptBrackets(UnifTerm* term,
				       UnifTerm* changed_suffix,
				       UnifTerm* old_end_of_term,
				       UnifTerm* new_end_of_term);
    static bool Occurs(Unifier::Variable* v,UnifTerm* term);
#ifndef NO_DEBUG
    static bool OccursIneff(Unifier::Variable* v,UnifTerm* term);
#endif
  private: 
    UnifTerm::Symbol head;
    UnifTerm* bracket;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm::SimpleIter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class UnifTerm::SimpleIter
  {
  public:
    SimpleIter() {};
    SimpleIter(UnifTerm* term) 
    {
      CALL("SimpleIter::constructor SimpleIter(UnifTerm* term)");
      Reset(term);
    };
    void Reset(UnifTerm* term)
    {
      CALL("SimpleIter::Reset(UnifTerm* term)");
      ASSERT(UnifTerm::CorrectComplexTerm(term)); 
      curr_pos = term;
      end_of_term = term->Bracket();
    };
    void MemorizeState(SimpleIter& memo) const
    {
      memo.curr_pos = curr_pos;
      memo.end_of_term = end_of_term;
    };
    void RestoreState(const SimpleIter& memo)
    {
      curr_pos = memo.curr_pos;
      end_of_term = memo.end_of_term;
    }; 
    UnifTerm* CurrPos() const { return curr_pos; };
    bool Empty() const { return (curr_pos == end_of_term); };
    bool NotEmpty() const { return end_of_term != curr_pos; };
    void SkipSym()
    {
      CALL("SimpleIter::SkipSym()");
      curr_pos++; 
      ASSERT(curr_pos <= end_of_term);
    };
    void SkipTerm()
    {
      CALL("SimpleIter::SkipTerm()");
      ASSERT((curr_pos < curr_pos->Bracket()) 
	     && (curr_pos->Bracket() - curr_pos <= (long)VampireKernelConst::MaxTermSize));
      curr_pos = curr_pos->Bracket();
      ASSERT(curr_pos);
    };
  private:
    UnifTerm* curr_pos;
    UnifTerm* end_of_term;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm::SimpleConstIter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class UnifTerm::SimpleConstIter
  {
  public:
    SimpleConstIter() {};
    SimpleConstIter(const UnifTerm* term)
    {
      CALL("SimpleConstIter::constructor SimpleConstIter(const UnifTerm* term)");
      Reset(term);
    };
    void Reset(const UnifTerm* term) 
    {
      CALL("SimpleConstIter::Reset(UnifTerm* term)");
      ASSERT(UnifTerm::CorrectComplexTerm(term)); 
      curr_pos = term;
      end_of_term = term->Bracket();
    };    
    void MemorizeState(SimpleConstIter& memo) const
    {
      memo.curr_pos = curr_pos;
      memo.end_of_term = end_of_term;
    };
    void RestoreState(const SimpleConstIter& memo)
    {
      curr_pos = memo.curr_pos;
      end_of_term = memo.end_of_term;
    }; 
    const UnifTerm* CurrPos() const { return curr_pos; };
    bool Empty() const { return (curr_pos == end_of_term); };
    bool NotEmpty() const { return end_of_term != curr_pos; };
    void SkipSym()
    {
      CALL("SimpleConstIter::SkipSym()");
      curr_pos++; 
      ASSERT(curr_pos <= end_of_term);
    };
    void SkipTerm()
    {
      CALL("SimpleConstIter::SkipTerm()");
      ASSERT((curr_pos < curr_pos->Bracket()) 
	     && (curr_pos->Bracket() - curr_pos <= (long)VampireKernelConst::MaxTermSize));
      curr_pos = curr_pos->Bracket();
      ASSERT(curr_pos);
    };
  private:
    const UnifTerm* curr_pos;
    const UnifTerm* end_of_term;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm::Iter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class UnifTerm::Iter : private UnifTerm::SimpleIter
  {
  public: 
    Iter() {};
    Iter(UnifTerm* term) { Reset(term); };
    void Reset(UnifTerm* term) 
    {
      CALL("Iter::Reset(UnifTerm* term)");
      SimpleIter::Reset(term);
      ASSERT(term);
      ASSERT(term->symbol().IsFunctor());
      ASSERT(term->Bracket());
      _term = term;
      _endOfTerm = term->Bracket();
      next_continuation = continuation;
      curr_var = (Unifier::Variable*)0;
    };
    bool Empty() const
    {
      CALL("Iter::Empty() const"); 
#ifdef DEBUG_NAMESPACE
      if (SimpleIter::Empty()) 
	{
	  ASSERT(next_continuation == continuation);
	  return true;
	}
      else 
	return false; 
#else
      return SimpleIter::Empty(); 
#endif
    };
    bool NotEmpty() const { return SimpleIter::NotEmpty(); };
    Unifier::Variable* CurrVar() const { return curr_var; };
    UnifTerm* CurrSubterm() const { return SimpleIter::CurrPos(); };
    bool currentPositionIsInSubstitution() const 
    {
      return 
	CurrVar() ||
	CurrSubterm() < _term ||
	CurrSubterm() >= _endOfTerm;
    };
    void SkipSym() 
    {
      CALL("Iter::SkipSym()");
      ASSERT(!SimpleIter::Empty());
      SimpleIter::SkipSym();
      if (SimpleIter::Empty()) 
	{
	  if (next_continuation != continuation)
	    {
	      PopContinuation();
	    }
	  else
	    {
	      return; 
	    };
	};
      if (SimpleIter::CurrPos()->symbol().isVariable())
	{
	  curr_var = SimpleIter::CurrPos()->symbol().var();
	  void* bind = curr_var->Binding();
	  if (bind) 
	    {
	    check_bind_type: 
	      if (Unifier::current()->isVariable(bind))
		{
		  curr_var = (Unifier::Variable*)bind;
		  bind = curr_var->Binding();
		  if (bind) { goto check_bind_type; };
		}
	      else 
		{
		  curr_var = (Unifier::Variable*)0;
		  SimpleIter::SkipSym();
		  if (SimpleIter::NotEmpty())
		    {
		      PushContinuation();
		    };
		  SimpleIter::Reset((UnifTerm*)bind);
		};
	    };
	}
      else
	{
	  curr_var = (Unifier::Variable*)0;
	};
    }; 
    void SkipTerm()
    {
      CALL("Iter::SkipTerm()");
      ASSERT(!SimpleIter::Empty());
      SimpleIter::SkipTerm(); 
      if (SimpleIter::Empty()) 
	{
	  if (next_continuation - continuation)
	    {
	      PopContinuation();
	    }
	  else
	    {
	      return; 
	    };
	};
      if (SimpleIter::CurrPos()->symbol().isVariable())
	{
	  curr_var = SimpleIter::CurrPos()->symbol().var();
	  void* bind = curr_var->Binding();
	  if (bind) 
	    {
	    check_bind_type: 
	      if (Unifier::current()->isVariable(bind))
		{
		  curr_var = (Unifier::Variable*)bind;
		  bind = curr_var->Binding();
		  if (bind) { goto check_bind_type; };
		}
	      else 
		{
		  curr_var = (Unifier::Variable*)0;
		  SimpleIter::SkipSym();
		  if (SimpleIter::NotEmpty())
		    {
		      PushContinuation();
		    };
		  SimpleIter::Reset((UnifTerm*)bind);
		};
	    };
	}
      else
	{
	  curr_var = (Unifier::Variable*)0;
	};
    }; 
  private:
    void PopContinuation()
    {
      next_continuation--;
      SimpleIter::RestoreState(*next_continuation);
    };
    void PushContinuation()
    {
      CALL("PushContinuation()");
      ASSERT(next_continuation < continuation + VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices);
      SimpleIter::MemorizeState(*next_continuation); 
      next_continuation++;
    };
  private:
    UnifTerm* _term;
    UnifTerm* _endOfTerm;
    SimpleIter* next_continuation;
    SimpleIter continuation[VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices];
    Unifier::Variable* curr_var;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm::ConstIter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class UnifTerm::ConstIter : private SimpleConstIter
  {
  public: 
    ConstIter() {};
    ConstIter(const UnifTerm* term) { Reset(term); };
    void Reset(const UnifTerm* term) 
    {
      CALL("ConstIter::Reset(const UnifTerm* term)");
      SimpleConstIter::Reset(term);
      ASSERT(term);
      ASSERT(term->symbol().IsFunctor());
      ASSERT(term->Bracket());
      _term = term;
      _endOfTerm = term->Bracket();
      next_continuation = continuation;
      curr_var = (Unifier::Variable*)0;
    };
    bool Empty() const
    {
      CALL("ConstIter::Empty() const"); 
#ifdef DEBUG_NAMESPACE
      if (SimpleConstIter::Empty()) 
	{
	  ASSERT(next_continuation == continuation);
	  return true;
	}
      else { return false; };
#else
      return SimpleConstIter::Empty(); 
#endif
    };
    bool NotEmpty() const { return SimpleConstIter::NotEmpty(); };
    Unifier::Variable* CurrVar() const { return curr_var; };
    const UnifTerm* CurrSubterm() const { return SimpleConstIter::CurrPos(); };
    bool currentPositionIsInSubstitution() const 
    {
      return 
	CurrVar() ||
	CurrSubterm() < _term ||
	CurrSubterm() >= _endOfTerm;
    };
    void SkipSym() 
    {
      CALL("ConstIter::SkipSym()");
      ASSERT(!(SimpleConstIter::Empty()));
      SimpleConstIter::SkipSym();
      if (SimpleConstIter::Empty()) 
	{
	  if (next_continuation - continuation)
	    {
	      PopContinuation();
	    }
	  else
	    {
	      return; 
	    };
	};
      if (SimpleConstIter::CurrPos()->symbol().isVariable())
	{
	  curr_var = SimpleConstIter::CurrPos()->symbol().var();
	  void* bind = curr_var->Binding();
	  if (bind) 
	    {
	    check_bind_type: 
	      if (Unifier::current()->isVariable(bind))
		{
		  curr_var = (Unifier::Variable*)bind;
		  bind = curr_var->Binding();
		  if (bind) { goto check_bind_type; };
		}
	      else 
		{
		  curr_var = (Unifier::Variable*)0;
		  SimpleConstIter::SkipSym();
		  if (SimpleConstIter::NotEmpty())
		    {
		      PushContinuation();
		    };
		  SimpleConstIter::Reset((UnifTerm*)bind);
		};
	    };
	}
      else
	{
	  curr_var = (Unifier::Variable*)0;
	};
    }; 
    void SkipTerm() 
    {
      CALL("ConstIter::SkipTerm()");
      ASSERT(!(SimpleConstIter::Empty()));
      SimpleConstIter::SkipTerm(); 
      if (SimpleConstIter::Empty()) 
	{
	  if (next_continuation - continuation)
	    {
	      PopContinuation();
	    }
	  else
	    {
	      return; 
	    };
	};
      if (SimpleConstIter::CurrPos()->symbol().isVariable())
	{
	  curr_var = SimpleConstIter::CurrPos()->symbol().var();
	  void* bind = curr_var->Binding();
	  if (bind) 
	    {
	    check_bind_type: 
	      if (Unifier::current()->isVariable(bind))
		{
		  curr_var = (Unifier::Variable*)bind;
		  bind = curr_var->Binding();
		  if (bind) { goto check_bind_type; };
		}
	      else 
		{
		  curr_var = (Unifier::Variable*)0;
		  SimpleConstIter::SkipSym();
		  if (SimpleConstIter::NotEmpty())
		    {
		      PushContinuation();
		    };
		  SimpleConstIter::Reset((UnifTerm*)bind);
		};
	    };
	}
      else
	{
	  curr_var = (Unifier::Variable*)0;
	};
    }; 
  private:
    void PopContinuation()
    {
      next_continuation--;
      SimpleConstIter::RestoreState(*next_continuation);
    };
    void PushContinuation()
    {
      CALL("PushContinuation()");
      ASSERT(next_continuation < continuation + VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices);
      SimpleConstIter::MemorizeState(*next_continuation); 
      next_continuation++;
    };
  private:
    const UnifTerm* _term;
    const UnifTerm* _endOfTerm;
    SimpleConstIter* next_continuation;
    SimpleConstIter continuation[VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices];
    Unifier::Variable* curr_var;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm::IterWithBacktracking"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class UnifTerm::IterWithBacktracking : private UnifTerm::SimpleIter
  {
  public:
#ifdef DEBUG_NAMESPACE
    UnifTerm* TopLevel() { return top_level; };
#endif  
  public: 
    IterWithBacktracking() {};
    void Reset(UnifTerm* term) 
    {
      CALL("IterWithBacktracking::Reset(UnifTerm* term)");
      ASSERT(term);
      ASSERT(term->symbol().IsFunctor());
      ASSERT(term->Bracket());
      SimpleIter::Reset(term);
#ifdef DEBUG_NAMESPACE
      top_level = term;
#endif
      next_continuation = continuation;
      curr_var = (Unifier::Variable*)0;
      next_curr_var = current_var;
      undo_num = 0;
    };
    bool Empty() const { return SimpleIter::Empty(); };
    bool NotEmpty() const { return SimpleIter::NotEmpty(); };
    Unifier::Variable* CurrVar() const { return curr_var; };
    UnifTerm* CurrSubterm() const { return SimpleIter::CurrPos(); };
    void SkipSym() 
    {
      CALL("IterWithBacktracking::SkipSym()");
      ASSERT(!(SimpleIter::Empty()));
      SimpleIter::MemorizeState(state[undo_num]);
      PushCurrVar();
      unsigned undo_action = NOTHING;
      SimpleIter::SkipSym();
      if (SimpleIter::Empty()) 
	{
	  if (next_continuation - continuation)
	    {
	      undo_action += POP;
	      PopContinuation();
	      SimpleIter::MemorizeState(used_continuation[undo_num]);
	    }
	  else
	    {
	      undo[undo_num] = (ActionToUndo)undo_action;
	      undo_num++;
	      return; 
	    };
	};
      if (SimpleIter::CurrPos()->symbol().isVariable())
	{
	  curr_var = SimpleIter::CurrPos()->symbol().var();
	  void* bind = curr_var->Binding();
	  if (bind) 
	    {
	    check_bind_type: 
	      if (Unifier::current()->isVariable(bind))
		{
		  curr_var = (Unifier::Variable*)bind;
		  bind = curr_var->Binding();
		  if (bind) { goto check_bind_type; };
		}
	      else 
		{
		  curr_var = (Unifier::Variable*)0;
		  SimpleIter::SkipSym();
		  if (SimpleIter::NotEmpty())
		    {
		      undo_action += PUSH;
		      PushContinuation();
		    };
		  SimpleIter::Reset((UnifTerm*)bind);
		};
	    };
	}
      else
	{
	  curr_var = (Unifier::Variable*)0;
	};
      undo[undo_num] = (ActionToUndo)undo_action;
      undo_num++;
    }; 
    void SkipTerm() 
    {
      CALL("IterWithBacktracking::SkipTerm()");
      ASSERT(!(SimpleIter::Empty()));
      SimpleIter::MemorizeState(state[undo_num]);
      PushCurrVar();
      unsigned undo_action = NOTHING;
      SimpleIter::SkipTerm(); 
      if (SimpleIter::Empty()) 
	{
	  if (next_continuation - continuation)
	    {
	      undo_action += POP;
	      PopContinuation();
	      SimpleIter::MemorizeState(used_continuation[undo_num]);
	    }
	  else
	    {
	      undo[undo_num] = (ActionToUndo)undo_action;
	      undo_num++;
	      return; 
	    };
	};
      if (SimpleIter::CurrPos()->symbol().isVariable())
	{
	  curr_var = SimpleIter::CurrPos()->symbol().var();
	  void* bind = curr_var->Binding();
	  if (bind) 
	    {
	    check_bind_type: 
	      if (Unifier::current()->isVariable(bind))
		{
		  curr_var = (Unifier::Variable*)bind;
		  bind = curr_var->Binding();
		  if (bind) { goto check_bind_type; };
		}
	      else 
		{
		  curr_var = (Unifier::Variable*)0;
		  SimpleIter::SkipSym();
		  if (SimpleIter::NotEmpty())
		    {
		      undo_action += PUSH;
		      PushContinuation();
		    };
		  SimpleIter::Reset((UnifTerm*)bind);
		};
	    };
	}
      else
	{
	  curr_var = (Unifier::Variable*)0;
	};
      undo[undo_num] = (ActionToUndo)undo_action;
      undo_num++;
    }; 
    void Backtrack()
    {
      CALL("IterWithBacktracking::Backtrack()");
      ASSERT(undo_num);
      undo_num--;
      SimpleIter::RestoreState(state[undo_num]);
      PopCurrVar();
      switch (undo[undo_num])
	{
	case NOTHING: break;
	case POP:
	  next_continuation->RestoreState(used_continuation[undo_num]);
	  next_continuation++;
	  break;
	case PUSH:
	  next_continuation--;
	  break;
	case POP_AND_PUSH: 
	  (next_continuation-1)->RestoreState(used_continuation[undo_num]);
	  break;
#ifdef DEBUG_NAMESPACE
	default: ICP("default"); 
#endif 
	};
    }; 
  private:
    enum ActionToUndo
    {
      NOTHING = 0,
      POP = 1, 
      PUSH = 2, 
      POP_AND_PUSH = 3 
    };
  private:
    void PushCurrVar() 
    {
      CALL("IterWithBacktracking::PushCurrVar()");
      *next_curr_var = curr_var;
      next_curr_var++;
      ASSERT(next_curr_var - current_var <= (long)VampireKernelConst::MaxTermSize);
    };   
    void PopCurrVar()
    {
      CALL("IterWithBacktracking::PopCurrVar()");
      next_curr_var--;
      ASSERT(next_curr_var >= current_var);
      curr_var = *next_curr_var;
    };   
    void PopContinuation()
    {
      next_continuation--;
      SimpleIter::RestoreState(*next_continuation);
    };
    void PushContinuation()
    {
      CALL("PushContinuation()");
      ASSERT(next_continuation < continuation + VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices);
      SimpleIter::MemorizeState(*next_continuation); 
      next_continuation++;
    };
  private:
    SimpleIter* next_continuation;
    SimpleIter continuation[VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices];
    Unifier::Variable* current_var[VampireKernelConst::MaxTermSize];
    Unifier::Variable** next_curr_var;
    Unifier::Variable* curr_var;
#ifdef DEBUG_NAMESPACE
    UnifTerm* top_level;
#endif
    ActionToUndo undo[VampireKernelConst::MaxTermSize];
    unsigned undo_num;
    SimpleIter state[VampireKernelConst::MaxTermSize];
    SimpleIter used_continuation[VampireKernelConst::MaxTermSize];
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline 
  void 
  UnifTerm::StandardTermToUnifTerm(ulong index,
				   const Term& t,
				   UnifTerm*& memory) 
  {
    CALL("StandardTermToUnifTerm(ulong index,const Term& t,UnifTerm*& memory)");
    if (t.isVariable())
      {
	memory->symbol().MkVar(Unifier::current()->variable(t.var(),index));
	memory->SetBracket(memory + 1);
	memory++;
      }
    else StandardTermToUnifTerm(index,t.deref(),memory); 
  }; 
  inline
  bool UnifTerm::containsSubterm(const UnifTerm* term1,
				 const UnifTerm* term2)
  {
    CALL("containsSubterm(..)");
    ASSERT(term1->symbol().isVariable() || CorrectComplexTerm(term1));
    return term1 <= term2 && term2 < term1->Bracket();
  }; 
}; 
#endif 
