/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef UNIFORM_TERM_WEIGHTING_H
#define UNIFORM_TERM_WEIGHTING_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TermWeighting.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIFORM_TERM_WEIGHTING
#define DEBUG_NAMESPACE "UniformTermWeighting"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ForwardMatchingSubstitutionCache;
  class LocalUnification;
  class UniformTermWeighting : public TermWeighting
    {
    public: 
      UniformTermWeighting() 
	{
	  setScheme(Uniform);
	};
      ~UniformTermWeighting();
      void* operator new(size_t);
      void operator delete(void* obj);
      void collectWeight(WeightPolynomial& weight,const Flatterm* term);
      void collectWeight(WeightPolynomial& weight,long coefficient,const Flatterm* term);
      long computeMinimalInstanceWeight(const Flatterm* term,
							       bool& containsVariables);
      long computeWeightIfGroundAndLessOrEqual(long coefficient,
								      const Flatterm* term,
								      long weightLimit);
      void collectWeight(WeightPolynomial& weight,const Term& term);
      void collectWeight(WeightPolynomial& weight,long coefficient,const Term& term);
      void collectWeight(WeightPolynomial& weight,long coefficient,const UnifTerm* term);
      void collectWeight(WeightPolynomial& weight,const UnifTerm* term);
      void collectWeight(WeightPolynomial& weight,const Term& term,ulong variableIndex);
      void 
	collectWeight(WeightPolynomial& weight,
		      long coefficient,
		      const Term* complexTerm,
		      const InstRet::Substitution* subst);
      void 
	collectWeight(WeightPolynomial& weight,
		      const Term* complexTerm,
		      const InstRet::Substitution* subst);
      void collectWeight(WeightPolynomial& weight,
			 const Term& term,
			 const ForwardMatchingSubstitutionCache* subst);
      void
      collectWeight(WeightPolynomial& weight,
		    const Term& term,
		    const LocalUnification* unifier);
      long computeMinimalInstanceWeight(const UnifTerm* term,bool& containsVariables);
      long computeWeightIfGroundAndLessOrEqual(long coefficient,const UnifTerm* term,long weightLimit);
      long computeWeightIfGroundAndLessOrEqual(const UnifTerm* term,long weightLimit);
      long computeMinimalInstanceWeightIfLessOrEqual(long coefficient,
									    const UnifTerm* term,
									    long weightLimit,
									    bool& ground);
      long computeMinimalInstanceWeightIfLessOrEqual(const UnifTerm* term,
									    long weightLimit,
									    bool& ground);
      long computeMinimalInstanceWeight(const Term* term,bool& containsVariables);
      long computeWeightIfGroundAndLessOrEqual(long coefficient,
								      const Term* term,
								      long weightLimit);
      long computeWeightIfGroundAndLessOrEqual(const Term* term,long weightLimit);
      long computeMinimalInstanceWeightIfLessOrEqual(long coefficient,
									    const Term* term,
									    long weightLimit,
									    bool& ground);
      long computeMinimalInstanceWeightIfLessOrEqual(const Term* term,
									    long weightLimit,
									    bool& ground);
    private:
      InstRet::Substitution::Instance _instance;
      Term::Iterator _iter;
    }; 
}; 
#endif
