/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef HELL_CONST_HPP
#define HELL_CONST_HPP
#include <climits>
#include <cstddef>
#include "jargon.hpp"
class VampireKernelConst
{
 public:
  static const char* SkolemConstantNamePrefix() { return "c__"; };
  static const char* SkolemFunctionNamePrefix() { return "f__"; };
  static const char* SkolemPredicateNamePrefix() { return "p__"; };
  static const char* CanonicalVariableNamePrefix() { return "X"; };
  enum 
  {
    MaxlNumOfLits = 65535UL, 
    MaxSymbolArity = 1023UL, 
#ifdef LARGE_TERMS
    MaxNumOfVariables = 128UL,
    MaxTermDepth = 48UL,
    MaxTermSize = 256UL,  
    MaxClauseSize = 512UL,
#else 
#ifdef HUGE_TERMS
    MaxNumOfVariables = 256UL,
    MaxTermDepth = 1024UL,
    MaxTermSize = 1024UL, 
    MaxClauseSize = 2048UL,
#else
    MaxNumOfVariables = 64UL,
    MaxTermDepth = 24UL,  
    MaxTermSize = 128UL,   
    MaxClauseSize = 256UL,
#endif
#endif
    MaxlSymbolNum = 65535UL, 
    MinMaxWeight = 1UL,
    MaxNumOfIndices = 2UL,
    UnordEqNum = 0UL,
    OrdEqNum = 1UL,
    FirstNoneqFun = 2UL,
    BuiltInTrueNum = 2UL,
    ReservedPred0 = 3UL,
    ReservedPred1 = 4UL,
    ReservedPred2 = 5UL,
    MaxlHeaderNum = MaxlSymbolNum*2 + 1
  }; 
  static bool isEquality(const ulong& num) { return (num < FirstNoneqFun); }; 
  static bool isOrderedEq(const ulong& num) { return num != 0UL; }; 
  static bool isUnorderedEq(const ulong& num) { return num == 0UL; }; 
  static bool isBuiltInTrue(const ulong& num) { return num == BuiltInTrueNum; };
  static bool isReservedPred(const ulong& num) { return (num > 1) && (num < 5); };
}; 
#endif
