/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef VARIABLE_RENAMING_HPP
#define VARIABLE_RENAMING_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLE_RENAMING
#define DEBUG_NAMESPACE "VariableRenaming"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class VariableRenaming
    {
    public: 
      VariableRenaming() 
	{
	  for (ulong v = 0; v < VampireKernelConst::MaxNumOfVariables; v++)
	    {
	      renaming[v] = 0;
	    };
	  next_var = 1;
	  next = renamed;
	};
      void Reset()
	{
	  next_var = 1;
	  while (next != renamed)
	    {
	      next--; 
	      renaming[*next] = 0;
	    };
	};
      ulong Rename(ulong var)
	{
	  ulong res = renaming[var];
	  if (!res) 
	    {
	      res = next_var;
	      renaming[var] = res;
	      next_var++;
	      *next = var;
	      next++;
	    };
	  return res - 1;
	};
    public: 
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const;
#endif
    private:
      ulong renaming[VampireKernelConst::MaxNumOfVariables];
      ulong next_var;
      ulong renamed[VampireKernelConst::MaxNumOfVariables]; 
      ulong* next;
    }; 
#ifndef NO_DEBUG_VIS
  inline
  ostream& VariableRenaming::output(ostream& str) const
  {
    str << "VARIABLE RENAMING:\n";
    for (const ulong* r = renamed; r < next; r++)
      str << 'X' << *r << " -> " << 'X' << renaming[*r] << '\n';
    return str << "END OF VARIABLE RENAMING.\n";
  }; 
#endif
}; 
#endif 
