/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "VarCounter.hpp"
#include "Int.hpp"
#include "Tracer.hpp"
#include "XML.hpp"
#include "VampireException.hpp"
using namespace Shell;
VarCounter::VarCounter ()
  :  _top(0),
     _counts(0)
{
}
VarCounter::~VarCounter ()
{
  TRACER("VarCounter::~VarCounter");
  delete [] _counts;
} 
void VarCounter::incCount (Var w)
{
  TRACER("VarCounter::incCount");
  int v = w.toInt();
  if (v >= _top) { 
    expandToFit(v);
  }
  _counts[v] ++;
} 
void VarCounter::decCount (Var w)
{
  TRACER("VarCounter::decCount");
  int v = w.toInt();
  if (v >= _top) { 
    expandToFit(v);
  }
  _counts[v] --;
} 
void VarCounter::expandToFit (int v)
{
  TRACER("VarCounter::expandToFit");
  int newTop = max(_top+16,v+1);
  int* newCounts = new int[newTop];
  if (! newCounts) {
    ALLOCATION_FAILURE;
  }
  for (int i = 0;i < _top;i++) {
    newCounts[i] = _counts[i];
  }
  for (int k = _top;k < newTop;k++) {
    newCounts[k] = 0;
  }
  delete [] _counts;
  _counts = newCounts;
  _top = newTop;
} 
int VarCounter::getCount (Var w) const 
{ 
  TRACER("VarCounter::getCount");
  int v = w.toInt();
  return v < _top ? _counts[v] : 0; 
} 
void VarCounter::setCount (Var w, int c)
{
  TRACER("VarCounter::setCount");
  int v = w.toInt();
  if (v >= _top) { 
    expandToFit(v);
  }
  _counts[v] = c;
} 
XML::Element* VarCounter::toXML () const
{
  XML::DeepElement* vc = new XML::DeepElement("var_counter");
  vc->addChild(new XML::IntegerElement("top",_top));
  XML::DeepElement* counters = new XML::DeepElement("counters");
  vc->addChild(counters);
  for (int i = 0;i < _top;i++) {
    if (_counts[i] != 0) {
      XML::EmptyElement* counter = new XML::EmptyElement("counter");
      counters->addChild(counter);
      counter->addAttribute(new XML::IntegerAttribute("var",i));
      counter->addAttribute(new XML::IntegerAttribute("value",_counts[i]));
    }
  }
  return vc;
} 
