/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Allocator__
#define __Allocator__
#include <cstdlib>
#include <string>
#include "Assertion.hpp"
#include "Debugger.hpp"
namespace Shell {
  class MyTracer;
class GlobalAllocator
{
public:
#if DEBUG_SHELL
  static void* allocate (int words, const string& className);
#else
  static void* allocate (int words);
#endif
  static void setMemoryLimitInMegabytes (int lim)
  { 
    _overflown = false;
    _memoryLimitInBytes = size_t(lim) * size_t(1000000); 
  }
  static size_t getUsedMemoryInKilobytes ()
  { return _totalNumberOfBytesAllocated / 1000; }
  static size_t getUsedMemoryInBytes ()
  { return _totalNumberOfBytesAllocated; }
#if DEBUG_SHELL
#endif
public:
  class Dummy 
  {
  private:
    void* _dummy;
  }; 
private:
  static int _remainingPieces;
  static Dummy* _free;
  static size_t _totalNumberOfBytesAllocated;
  static size_t _memoryLimitInBytes;
  static bool _overflown;
}; 
template <int WORDS>
class Allocator
{
public:
#if DEBUG_SHELL
  static void* allocate (const string& className)
#else
  static void* allocate ()
#endif
  {
    if (_freeList) {
      ASS (Debugger::reallocated(_freeList,WORDS,className));
      void* _result = _freeList;
      _freeList = _freeList->_next;
      return _result;
    }
#if DEBUG_SHELL
    void* result = GlobalAllocator::allocate(WORDS,className);
#else
    void* result = GlobalAllocator::allocate(WORDS);
#endif
    ASS (Debugger::allocated(result,WORDS,className));
    return result;
   } 
#if DEBUG_SHELL
  static void release (void* obj, const string& className)
#else
  static void release (void* obj)
#endif
  {
    ASS (Debugger::released(obj,WORDS,className));
    Allocator* m = reinterpret_cast <Allocator*>(obj);
    m->_next = _freeList;
    _freeList = m;
  }
private:
  static Allocator* _freeList;
  Allocator* _next;
  friend class MyTracer;
}; 
}
#define PIECES(C) (sizeof(C) + sizeof(GlobalAllocator::Dummy) - 1) / \
                   sizeof(GlobalAllocator::Dummy)
#if DEBUG_SHELL
#define USE_ALLOCATOR(C)                                         \
  void* operator new (size_t)                                    \
  { return Allocator<PIECES(C)>::allocate(className()); }        \
  void operator delete (void* obj)                               \
  { if (obj) Allocator<PIECES(C)>::release(obj,className()); }
#else
#define USE_ALLOCATOR(C)                              \
  void* operator new (size_t)                         \
  { return Allocator<PIECES(C)>::allocate(); }        \
  void operator delete (void* obj)                    \
  { if (obj) Allocator<PIECES(C)>::release(obj); }
#endif
#if DEBUG_SHELL
#define CLASS_NAME(name) static string className () { return name; }
#else
#define CLASS_NAME(name)
#endif
#endif 
