/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BitWord.hpp"
#ifdef DEBUG_BIT_WORD
 #define DEBUG_NAMESPACE "BitWord"
#else
 #undef DEBUG_NAMESPACE
#endif
#include "debugMacros.hpp"
using namespace BK;
long BitWord::InitStatic::_count; 
ulong BitWord::baseValAllSet;
ulong BitWord::baseValAllClear;
ulong BitWord::singleBitMask[sizeof(ulong)*CHAR_BIT];
ulong BitWord::allButOneMask[sizeof(ulong)*CHAR_BIT];
ulong BitWord::pieceMask[sizeof(ulong)*CHAR_BIT][sizeof(ulong)*CHAR_BIT];
ulong BitWord::inversePieceMask[sizeof(ulong)*CHAR_BIT][sizeof(ulong)*CHAR_BIT];
void BitWord::initStatic()
{
 baseValAllSet = makeBaseValAllSet(); 
 baseValAllClear = 0ul;
 for (size_t n = 0; n < sizeof(ulong)*CHAR_BIT; n++)
  {
   singleBitMask[n] = makeSingleBitMask(n); 
   allButOneMask[n] = makeAllButOneMask(n);
  };
 for (size_t begin = 0; begin < sizeof(ulong)*CHAR_BIT; begin++)
  for (size_t end = begin; end < sizeof(ulong)*CHAR_BIT; end++)
   {
    pieceMask[begin][end] = makePieceMask(begin,end);
    inversePieceMask[begin][end] = makeInversePieceMask(begin,end);
   };
}; 
ulong BitWord::makeBaseValAllSet()
{
 ulong x;
 ulong y = 1ul;
 while (y) { x = y; y = y << 1; };
 y = x;
 for (size_t n = 1; n < sizeof(ulong)*CHAR_BIT; n++) { x = x >> 1; y |= x; };
 return y;
};
ulong BitWord::makeSingleBitMask(size_t n)
{
 ulong x;
 ulong y = 1ul;
 while (y) { x = y; y = y << 1; };
 while (n) { x = x >> 1; n--; };
 return x;
};
ulong BitWord::makeAllButOneMask(size_t n)
{
 ulong result = makeSingleBitMask(n);
 result ^= makeBaseValAllSet();
 return result;
};
ulong BitWord::makePieceMask(size_t begin,size_t end)
{
 return ((makeBaseValAllSet() << begin) >> (sizeInBits() - end + begin - 1)) << (sizeInBits() - end - 1);
}; 
ulong BitWord::makeInversePieceMask(size_t begin,size_t end)
{
 return (makePieceMask(begin,end) ^ makeBaseValAllSet());
}; 
