/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CHARACTERISTIC_COMPARISON_MAKER_HPP
#define CHARACTERISTIC_COMPARISON_MAKER_HPP
#include "Comparison.hpp"
#include "ComparisonMaker.hpp"
#include "ComparableCharacteristic.hpp"
#ifdef DEBUG_CHARACTERISTIC_COMPARISON_MAKER
#  undef DEBUG_NAMESPACE
#  define DEBUG_NAMESPACE "CCMaker<..>"
#endif
#include "debugMacros.hpp"
namespace BK
{
  template <class T,class ComparableCharacteristic = ComparableCharacteristic<T> >
  class CCMaker : public CMaker<T>
  {
  public:
    CCMaker() {};
    CCMaker(ComparableCharacteristic* c1,
				  ComparableCharacteristic* c2) 
    {
      reset(c1,c2);
    };
    ~CCMaker();
    void reset(ComparableCharacteristic* c1,
	       ComparableCharacteristic* c2) 
    { 
      _lhs = c1;
      _rhs = c2;
    };
    void loadLHS(const T& obj) { _lhs->computeFor(obj); };
    void loadRHS(const T& obj) { _rhs->computeFor(obj); };
    Comparison compare() { return _lhs->compare(*_rhs); };
    Comparison inverseCompare() { return _lhs->inverseCompare(*_rhs); };
    void swapSides() { ComparableCharacteristic* c = _lhs; _lhs = _rhs; _rhs = c; };
  private:
    ComparableCharacteristic* _lhs;
    ComparableCharacteristic* _rhs;
  }; 
}; 
#ifdef DEBUG_CHARACTERISTIC_COMPARISON_MAKER
#  undef DEBUG_NAMESPACE
#  define DEBUG_NAMESPACE "CCMaker<..>"
#endif
#include "debugMacros.hpp"
namespace BK
{
  template <class T,class ComparableCharacteristic>
  CCMaker<T,ComparableCharacteristic>::~CCMaker()
  {
  };
}; 
#endif
