/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef NO_DEBUG
#include <iostream>
#include <csignal>
#include <climits>
#include "Debug.hpp"
#include "Error.hpp"
#include "DestructionMode.hpp"
#include "debugMacros.hpp"
using  namespace std;
using namespace BK;
long Debug::InitStatic::_count; 
const char* Debug::job;  
ulong Debug::eventId;
Debug::Call* Debug::currentCall;
Debug::Call* Debug::bottomCall;
long Debug::callDepth;
const char* Debug::errorDescription; 
ostream* Debug::trace;
bool Debug::traceCP;
bool Debug::traceAfterError;
void (*Debug::nextEventHook)(ulong);
void (*Debug::errorHook)() = 0;
bool Debug::errorFlag;
Debug::InitStatic::InitStatic() 
{
  if (!_count) 
    { 
#ifdef DEBUG_ERROR_SIGNALS
      Debug::initSigHandlers();
#endif
      Debug::callDepth = 0L;
      Debug::job = "unknown job";
      Debug::eventId = 0UL;
      Debug::errorDescription = "unknown";
      Debug::trace = 0;
      Debug::traceCP = false;   
      Debug::traceAfterError = false;
      Debug::nextEventHook = 0;
      Debug::errorHook = 0;
      Debug::bottomCall = new Debug::Call("<global>","<global>");
      Debug::errorFlag = false;
    };    
  ++_count;
}; 
Debug::InitStatic::~InitStatic() 
{
  --_count;
  if (!_count)
    {      
      delete Debug::bottomCall;
    };
}; 
void Debug::reportCP(std::ostream& str,
		     const char* cn,
		     const char* cp,
		     const char* on)
{
  str << cn << '/' << cp << " [" << on << ']' << flush;
};
void Debug::error(std::ostream& str)
{
 str << "Job " << job << " aborted.\n"; 
 str << "Error : " << errorDescription << " in\n";
 history(str); 
 str << "The job was " << job << '\n';
 if (!traceAfterError) tracingOff();
 if (errorHook) 
   {
     void (*tmpErrorHook)();
     tmpErrorHook = errorHook;
     errorHook = 0;
     tmpErrorHook();
   };
 errorFlag = true;
 DestructionMode::makeFast();
 restoreSigHandlers();
 throw Error(errorDescription);
};
void Debug::error(std::ostream& str,const char* msg)
{
 errorDescription = msg;
 error(str);
};
void Debug::debugFeatureInternalFault(std::ostream& str,const char* msg)
{
 str << "Job " << job << " aborted.\n"; 
 str << "Error : debug feature internal fault : " << msg << " in \n";
 history(str); 
 str << "The job was " << job << '\n';
 if (!traceAfterError) tracingOff();
 errorFlag = true;
 DestructionMode::makeFast();
 restoreSigHandlers();
 throw Error(msg);
};
void Debug::reportCP(std::ostream& str)
{
  str << eventId << " ";
  reportCP(str,
	   currentCall->_callName,
	   currentCall->_controlPoint,
	   currentCall->_objName); 
  str << '\n';
};
void Debug::history(std::ostream& str)
{
  str << eventId << " ";
  int indent = 0;
  for (Call* call = currentCall; call; call = call->_outerCall)
    {
      for (OpenDataBracket* bracket = currentCall->_dataBrackets;
	   bracket;
	   bracket = bracket->_outerBracket)
	{
	  for (int i = 0; i < indent; i++) str << ' ';
	  if (indent) str << "|_ ";
	  ++indent;
	  switch (bracket->_status)
	    {
	    case OpenDataBracket::InitOpen:
	      str << "-> INIT BRACKET ";
	      break;
	    case OpenDataBracket::InitClose:
	      str << "<- INIT BRACKET ";
	      break;
	    case OpenDataBracket::DestrOpen:
	      str << "-> DESTR BRACKET ";
	      break;
	    case OpenDataBracket::DestrClose:
	      str << "<- DESTR BRACKET ";
	      break;
	    };
	  str << bracket->_name << '\n' << flush;
	};
      for (int i = 0; i < indent; i++) str << ' ';
      if (indent) str << "|_ ";
      ++indent;
      reportCP(str,
	       call->_callName,
	       call->_controlPoint,
	       call->_objName);      
      str << '\n';
    };
}; 
void Debug::initSigHandlers()
{
#ifdef BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE
#  if (BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE == BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_UNIX)
  signal(SIGFPE,handleSIGFPE);
  signal(SIGILL,handleSIGILL);
  signal(SIGSEGV,handleSIGSEGV);
  signal(SIGBUS,handleSIGBUS);  
  signal(SIGTRAP,handleSIGTRAP);
  signal(SIGABRT,handleSIGABRT);
#  elif (BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE == BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_MSWIN32 )
  signal(SIGFPE,handleSIGFPE);
  signal(SIGILL,handleSIGILL);
  signal(SIGSEGV,handleSIGSEGV); 
#  else
#    error "Unsupported target operational environment type, check BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE"
#  endif
#else
#  error "Undefined macro BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE"
#endif
}; 
void Debug::restoreSigHandlers()
{
  #ifdef BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE
#  if (BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE == BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_UNIX)
  signal(SIGFPE,SIG_DFL);
  signal(SIGILL,SIG_DFL);
  signal(SIGSEGV,SIG_DFL);
  signal(SIGBUS,SIG_DFL);  
  signal(SIGTRAP,SIG_DFL);
  signal(SIGABRT,SIG_DFL);
#  elif (BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE == BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_MSWIN32 )
  signal(SIGFPE,SIG_DFL);
  signal(SIGILL,SIG_DFL);
  signal(SIGSEGV,SIG_DFL); 
#  else
#    error "Unsupported target operational environment type, check BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE"
#  endif
#else
#  error "Undefined macro BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE"
#endif
}; 
void Debug::handleSIGFPE(int)
{
 restoreSigHandlers();
 error(DEBUG_OUT_STR,"Program Error Signal SIGFPE (arithmetic error)");
};
void Debug::handleSIGILL(int)
{
 restoreSigHandlers();
 error(DEBUG_OUT_STR,"Program Error Signal SIGILL (illegal instruction)");
};
void Debug::handleSIGSEGV(int)
{
 restoreSigHandlers();
 error(DEBUG_OUT_STR,"Program Error Signal SIGSEGV (segmentation violation)");
};
void Debug::handleSIGBUS(int) 
{
 restoreSigHandlers();
 error(DEBUG_OUT_STR,"Program Error Signal SIGBUS (bus error)");
};
void Debug::handleSIGABRT(int)
{
 restoreSigHandlers();
 error(DEBUG_OUT_STR,"Program Error Signal SIGABRT");
};
void Debug::handleSIGTRAP(int)
{
 restoreSigHandlers();
 error(DEBUG_OUT_STR,"Program Error Signal SIGTRAP");
};
#endif
