/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef EXPANDING_MULTISET_HPP
#define EXPANDING_MULTISET_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "ExpandingStack.hpp"
#include "Array.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_EXPANDING_MULTISET
#  define DEBUG_NAMESPACE "ExpandingMultiset<Alloc,CarrierType,MinIncrementSize>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class CarrierType,ulong MinIncrementSize>
class ExpandingMultiset
{
 public:
  ExpandingMultiset() :
    _score(0UL,"ExpandingMultiset<..>::_score"), 
    _registered(DOP("ExpandingMultiset<..>::_registered"))
   #ifdef DEBUG_NAMESPACE
    , _name("?")
   #endif
  {
   CALL("constructor ExpandingMultiset()");
  };
  ExpandingMultiset(const char* name)  :
    _score(0UL,"ExpandingMultiset<..>::_score"), 
    _registered(DOP("ExpandingMultiset<..>::_registered"))
   #ifdef DEBUG_NAMESPACE
    , _name(name)
   #endif
  {
   CALL("constructor ExpandingMultiset(const char* name)");
   COP("END");
  };
  ~ExpandingMultiset() { CALL("destructor ~ExpandingMultiset()"); };
  void expand(const CarrierType& i)
  {
   CALLM(_name,"expand(const CarrierType& i)");
   _score.expand(i);
   _registered.expand(i);
  };
  bool tryToExpand(const CarrierType& i)
  {
    CALLM(_name,"tryToExpand(const CarrierType& i)");
    return 
      _score.tryToExpand(i) && _registered.tryToExpand(i);
  };
  void reset() 
  {
   CALLM(_name,"reset()");
   CarrierType el;
   while (_registered.nonempty())
    {
     el = _registered.pop();
     _score[el] = 0L;
    };
  };
  ulong scoreOf(const CarrierType& el) const  
  {
   CALLM(_name,"scoreOf(const CarrierType& el) const");  
   return _score[el]; 
  };
  ulong safeScoreOf(const CarrierType& el) const  
  {
   CALLM(_name,"safeScoreOf(const CarrierType& el) const");
   if (el >= _score.size()) return false;  
   return _score[el]; 
  };
  void add(const CarrierType& el) 
  {
   CALLM(_name,"add(const CarrierType& el)");
   if (!(_score[el])) _registered.push(el); 
   _score[el]++;
  };
  void add(const CarrierType& el,ulong num) 
  {
   CALLM(_name,"add(const CarrierType& el,ulong num)");
   if (!(_score[el])) _registered.push(el); 
   _score[el] += num;
  };
  void addSafe(const CarrierType& el) 
  {
   CALLM(_name,"addSafe(const CarrierType& el)");
   if (!(_score.sub(el))) _registered.pushSafe(el); 
   _score[el]++;
  };
  void addSafe(const CarrierType& el,ulong num) 
  {
   CALLM(_name,"addSafe(const CarrierType& el,ulong num)");
   if (!(_score.sub(el))) _registered.pushSafe(el); 
   _score[el] += num;
  };
  void erase(const CarrierType& el)
  {
    CALLM(_name,"erase(const CarrierType& el)");
    for (CarrierType* p = _registered.begin(); p != _registered.end(); ++p)
      {
	if (*p == el)
	  {
	    _score[el] = 0L;
	    *p = _registered.top();
	    _registered.skipTop();
	    return;
	  };
      };
  };
  ulong numOfDiffElements() const 
  { 
   return _registered.size(); 
  };
  const CarrierType& nth(ulong n) const  
  {
   CALLM(_name,"nth(ulong n) const");
   return _registered.nth(n); 
  };
  #ifndef NO_DEBUG_VIS
   ostream& output(ostream& str) const
   {
    CALLM(_name,"output(ostream& str) const");
    str << '{';
    for (const CarrierType* el = _registered.begin(); el < _registered.end(); el++)
     {
      str << scoreOf(*el) << '*' << *el;
      if (el != _registered.last()) str << ' ';
     };  
    return str << '}';
   };
  #endif
  #ifndef NO_DEBUG
   void freeze() { _score.freeze(); _registered.freeze(); };
   void unfreeze() { _score.unfreeze(); _registered.unfreeze(); };
  #endif  
 private:
  Array<Alloc,ulong,DefaultRelocation<ulong>,MinIncrementSize> _score;
  ExpandingStack<Alloc, CarrierType, DefaultRelocation<CarrierType>, MinIncrementSize> _registered; 
  #ifdef DEBUG_NAMESPACE
   const char* _name;
  #endif
}; 
}; 
#endif
