/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef G_INC_COUNTING_SORT_H
#define G_INC_COUNTING_SORT_H
#ifndef NO_DEBUG_VIS
 #include <iostream>
#endif
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_G_INC_COUNTING_SORT
 #define DEBUG_NAMESPACE "GIncCountingSort<KeyType,ElType,MaxNumOfElements>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class KeyType,class ElType,ulong MaxNumOfElements>
class GIncCountingSort
{
 public:
  GIncCountingSort() : _next(0UL)
  {    
  };
  ~GIncCountingSort() {};
  void reset() 
  {
   _next = 0UL;
  };
  ElType* next() { return _elem + _next; };
  void push()
  {
   CALL("push()");
   ASSERT(_next < MaxNumOfElements);
   KeyType key = _elem[_next].key();
   ulong num = 0;
   for (ulong i = 0; i < _next; i++)
    if (_elem[i].key() < key) { num++; }
    else _num[i]++;
   _num[_next] = num;
   _next++;     
  }; 
  void push(const ElType& el)
  {
   CALL("push(const KeyType& key)");
   ASSERT(_next < MaxNumOfElements);
   _elem[_next] = el;
   KeyType key = el.key();
   ulong num = 0;
   for (ulong i = 0; i < _next; i++)
    if (_elem[i].key() < key) { num++; }
    else _num[i]++;
   _num[_next] = num;
   _next++; 
  };
  const ElType* elem() const { return _elem; };
  const ulong* num() const { return _num; };
  ulong size() const { return _next; };
  #ifndef NO_DEBUG_VIS
   ostream& output(ostream& str) const 
   {
    for (ulong i = 0; i < _next; i++) str << _elem[i].key() << " [" << _num[i] << "] "; 
    return str << '\n';
   };
  #endif
 private:
  ElType _elem[MaxNumOfElements];
  ulong _num[MaxNumOfElements];
  ulong _next;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_G_INC_COUNTING_SORT
 #define DEBUG_NAMESPACE "GSimpleIncCountingSort<KeyType,MaxNumOfElements>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class KeyType,ulong MaxNumOfElements>
class GSimpleIncCountingSort
{
 public:
  GSimpleIncCountingSort() : _next(0UL) {  };
  ~GSimpleIncCountingSort() {};
  void reset() 
  {
   _next = 0UL;
  };
  KeyType* next() { return _key + _next; };
  void push()
  {   
   CALL("push()");
   ASSERT(_next < MaxNumOfElements);
   KeyType key = _key[_next];
   ulong num = 0;
   for (ulong i = 0; i < _next; i++)
    if (_key[i] < key) { num++; }
    else _num[i]++;
   _num[_next] = num;
   _next++;    
  }; 
  void push(const KeyType& key)
  {
   CALL("push(const KeyType& key)");
   ASSERT(_next < MaxNumOfElements);
   _key[_next] = key;
   ulong num = 0;
   for (ulong i = 0; i < _next; i++)
    if (_key[i] < key) { num++; }
    else _num[i]++;
   _num[_next] = num;
   _next++;   
  };
  const ulong* num() const { return _num; };
  ulong size() const { return _next; }; 
  #ifndef NO_DEBUG_VIS
   ostream& output(ostream& str) const 
   {
    for (ulong i = 0; i < _next; i++) str << _key[i] << " [" << _num[i] << "] "; 
    return str << '\n';
   };
  #endif
 private:
  KeyType _key[MaxNumOfElements];
  ulong _num[MaxNumOfElements];
  ulong _next;
}; 
}; 
#undef DEBUG_NAMESPACE
#endif
