/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "GlobAlloc.hpp"
#include "DestructionMode.hpp"
using namespace BK;
long GlobAlloc::InitStatic::_count; 
GlobAlloc::Alloc GlobAlloc::_allocator;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ALLOC
#  define DEBUG_NAMESPACE "GlobAlloc::InitStatic"
#endif 
#include "debugMacros.hpp"
GlobAlloc::InitStatic::InitStatic()
{ 
  CALL("constructor InitStatic()");
  if (!_count) 
    {      
      GlobAlloc::basicAllocationFailureHandling().activate();
    };  
  _count++;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ALLOC
#  define DEBUG_NAMESPACE "GlobAlloc"
#endif 
#include "debugMacros.hpp"
void GlobAlloc::reset()
{
  CALL("reset()");
  allocator().reset();
}; 
void* GlobAlloc::allocate(size_t size) 
{ 
  CALL("allocate(size_t size)"); 
  ulong nsize = size/sizeof(void*);
  if (size % sizeof(void*)) nsize++;
#ifndef DEBUG_ALLOC
  return allocator().allocate(nsize);
#else
  void* res = allocator().allocate(nsize);
  if (!res) return 0;
  void** resMirror = static_cast<void**>(allocator().mirror(res));
  ObjDesc* objDesc = static_cast<ObjDesc*>(*resMirror);
  ASSERT(objDesc);
#ifdef DEBUG_ALLOC_OBJ_PTR  
  ASSERT(objDesc->objPtr == res);       
#endif
#ifdef DEBUG_ALLOC_OBJ_STATUS
  ASSERT(objDesc->status = ObjDesc::Persistent);
#endif  
#ifdef DEBUG_ALLOC_OBJ_SIZE
  objDesc->size = size;
#endif
  ASSERT(allocator().chunk(res));
  return res;
#endif
}; 
void* GlobAlloc::tryToAllocate(size_t size) 
{ 
  CALL("tryToAllocate(size_t size)"); 
  ulong nsize = size/sizeof(void*);
  if (size % sizeof(void*)) nsize++;
#ifndef DEBUG_ALLOC
  return allocator().tryToAllocateNonagressively(nsize);
#else
  void* res = allocator().tryToAllocateNonagressively(nsize);
  if (!res) return 0;
  void** resMirror = static_cast<void**>(allocator().mirror(res));
  ObjDesc* objDesc = static_cast<ObjDesc*>(*resMirror);
  ASSERT(objDesc);
#ifdef DEBUG_ALLOC_OBJ_PTR  
  ASSERT(objDesc->objPtr == res);       
#endif
#ifdef DEBUG_ALLOC_OBJ_STATUS
  ASSERT(objDesc->status = ObjDesc::Persistent);
#endif  
#ifdef DEBUG_ALLOC_OBJ_SIZE
  objDesc->size = size;
#endif
  ASSERT(allocator().chunk(res));
  return res;
#endif
}; 
void GlobAlloc::deallocate(void* obj,size_t size) 
{
  CALL("deallocate(void* obj,size_t size)");
  ulong nsize = size/sizeof(void*);
  if (size % sizeof(void*)) nsize++;
  ASSERT(allocator().chunk(obj)); 
#ifdef DEBUG_ALLOC           
  ObjDesc* objDesc = *(static_cast<ObjDesc**>(allocator().mirror(obj))); 
  ASSERT(objDesc);    
#ifdef DEBUG_ALLOC_OBJ_PTR
  ASSERT(objDesc->objPtr == obj); 
#endif
#ifdef DEBUG_ALLOC_OBJ_STATUS
  if (objDesc->status != ObjDesc::Persistent)
    {
      DMSG("Deallocating ") << *objDesc << '\n';
      ASSERT(objDesc->status == ObjDesc::Persistent); 
    };
#endif
#ifdef DEBUG_ALLOC_OBJ_SIZE
  if (objDesc->size != size) 
    { 
      DMSG("size = ") << size << "\nobjDesc:\n" << *objDesc << '\n';
      ASSERT(objDesc->size == size);
    };
#endif
#endif    
  allocator().deallocate(obj,nsize);
}; 
void GlobAlloc::getDefragmentationStatistics(ulong& calls,
					      float& averageInputSize,
					      ulong& numOfInputPieces,
					      float& averageOutputSize,
					      ulong& numOfOutputPieces)
{
  allocator().getDefragmentationStatistics(calls,
					   averageInputSize,
					   numOfInputPieces,
					   averageOutputSize,
					   numOfOutputPieces);
  averageInputSize = sizeof(void*) * averageInputSize;
  averageOutputSize = sizeof(void*) * averageOutputSize;
}; 
