/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef GLOB_ALLOC_HPP
#define GLOB_ALLOC_HPP   
#include <cstdlib>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"  
#include "ObjDesc.hpp"
#include "AllocationFailureHandling.hpp"
#include "Allocator.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ALLOC
#  define DEBUG_NAMESPACE "GlobAlloc"
#endif 
#include "debugMacros.hpp"
#define GLOB_ALLOC_CHUNK_SIZE 4194296UL
#define GLOB_ALLOC_MAX_OBJECT_SIZE 1023UL
namespace BK 
{
class GlobAlloc  
{
 public:
  class AllocationFailureHandlingId {};
  typedef AllocationFailureHandling<AllocationFailureHandlingId> AllocationFailureHandling;
 public:
  class InitStatic
  {
   public: 
    InitStatic();
   private:
    static long _count;
  };
 public:
  static void reset();
  static AllocationFailureHandling& basicAllocationFailureHandling() 
  { 
    static AllocationFailureHandling _basicAllocationFailureHandling;
    return _basicAllocationFailureHandling;
  }; 
  static void setHardMemoryLimit(long lim)
  { 
   CALL("setHardMemoryLimit(long lim)");
   allocator().setHardMemoryLimit(lim/(sizeof(void*)));
  };  
  static void setAllocationBufferSize(long abs) 
  { 
   allocator().setBufferSize(abs);  
  };
  static long occupiedByObjects() { return sizeof(void*)*allocator().occupiedByObjects(); };
  static long occupiedByChunks() { return sizeof(void*)*allocator().occupiedByChunks(); }; 
  static long totalBufferDeficit() { return sizeof(void*)*allocator().totalBufferDeficit(); };
  static long activeBufferDeficit() { return sizeof(void*)*allocator().activeBufferDeficit(); };
  static long suspendedBufferDeficit() { return sizeof(void*)*allocator().suspendedBufferDeficit(); };
  static void annulBufferDeficit() { allocator().annulBufferDeficit(); };
  static void getDefragmentationStatistics(ulong& calls,
					   float& averageInputSize,
					   ulong& numOfInputPieces,
					   float& averageOutputSize,
					   ulong& numOfOutputPieces);
  static void* allocate(size_t size);
  static void* tryToAllocate(size_t size); 
  static void deallocate(void* obj,size_t size); 
  #ifdef DEBUG_ALLOC
   static ObjDesc* hostObj(const void* ptr) 
   { 
    return *(static_cast<ObjDesc**>(allocator().mirror(ptr)));
   };
  #endif
#ifndef NO_DEBUG
   static ostream& outputInternalStatistics(ostream& str)
     {
       return allocator().outputInternalStatistics(str);
     };
#endif
#ifdef DEBUG_ALLOC_OBJ_STATUS
  static ostream& outputPersistent(ostream& str,ulong maxNumOfObjects)
    {
      return allocator().outputPersistent(str,maxNumOfObjects);
    };
#endif
public: 
  typedef Allocator<AllocationFailureHandling,GLOB_ALLOC_MAX_OBJECT_SIZE,GLOB_ALLOC_CHUNK_SIZE> Alloc;
private:
  static Alloc& allocator()
  {
    return _allocator;
  }; 
private:
  static Alloc _allocator;
  friend class InitStatic;
}; 
static GlobAlloc::InitStatic globAllocInitStatic;
}; 
#undef GLOB_ALLOC_CHUNK_SIZE
#undef GLOB_ALLOC_MAX_OBJECT_SIZE 
#endif
