/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef GLOBAL_STOP_FLAG_HPP
#define GLOBAL_STOP_FLAG_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_GLOBAL_STOP_FLAG
#  define DEBUG_NAMESPACE "GlobalStopFlag"
#endif
#include "debugMacros.hpp"
namespace BK 
{
class GlobalStopFlag
{
 public:
  class DefaultException 
    {
    public:
      DefaultException() {};
    };
  class HandlerActivator
    {
    public:
      HandlerActivator(void (*handler)())  : _handler(handler) 
	{
	  _previous = GlobalStopFlag::_currentHandlerActivator;
	  GlobalStopFlag::_currentHandlerActivator = this;
	};
      ~HandlerActivator()
	{
	  CALL("destructor ~HandlerActivator()");
	  if (DestructionMode::isThorough())
	    {
	      ASSERT(GlobalStopFlag::_currentHandlerActivator == this);
	      GlobalStopFlag::_currentHandlerActivator = _previous;
	    };
	};
    private:
      HandlerActivator();
      void handle();
    private:
      void (*_handler)();
      HandlerActivator* _previous;
      friend class GlobalStopFlag;
    };
 public:
  static void reset() { _stop = false; };
  static void raise() { _stop = true; };
  static void check()
    {
      CALL("check()");
      if (_stop)
	{
	  _currentHandlerActivator->handle();
	};
    };
 private:
  static bool _stop;
  static HandlerActivator* _currentHandlerActivator;
  friend class HandlerActivator;
}; 
}; 
#endif
