/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LINEAR_POLYNOMIAL_HPP
#define LINEAR_POLYNOMIAL_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "LinearCombination.hpp"
#include "Math.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LINEAR_POLYNOMIAL
#  define DEBUG_NAMESPACE "LinearPolynomial<CoeffType,MaxNumOfVariables>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class CoeffType,ulong MaxNumOfVariables>
class LinearPolynomial
{
 public:
  LinearPolynomial() :
    _freeMember(static_cast<CoeffType>(0))
    {
    };
  ~LinearPolynomial() 
  {
    CALL("destructor ~LinearPolynomial()");
    BK_CORRUPT(_freeMember);
  };
  void reset() { _monomials.reset(); _freeMember = static_cast<CoeffType>(0); };
  LinearCombination<CoeffType,MaxNumOfVariables>& monomials() 
    { 
      return _monomials; 
    };
  const LinearCombination<CoeffType,MaxNumOfVariables>& monomials() const 
    { 
      return _monomials; 
    };
  CoeffType& freeMember() { return _freeMember; };
  const CoeffType& freeMember() const { return _freeMember; };
  bool containsVariableWhichIsNotIn(const LinearPolynomial& lp) const
    {
      return monomials().containsVariableWhichIsNotIn(lp.monomials());
    };
  void add(const CoeffType& coeff,const LinearPolynomial& lp)
  {
    _freeMember += (coeff * lp._freeMember);
    _monomials.add(coeff,lp._monomials);
  };
  void divideBy(const CoeffType& c) 
  {
    CALL("divideBy(const CoeffType& c)");
    ASSERT(c);
    _freeMember = _freeMember / c;
    _monomials.divideBy(c);
  };
 private:
  LinearCombination<CoeffType,MaxNumOfVariables> _monomials;
  CoeffType _freeMember;
}; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
template <class CoeffType,ulong MaxNumOfVariables>
inline
ostream& operator<<(ostream& str,
		    const BK::LinearPolynomial<CoeffType,MaxNumOfVariables>& lp)
{
  str << lp.freeMember();
  if (!lp.monomials().isZero())
    str << " + " << lp.monomials();
  return str;
};
}; 
#endif
#endif
