/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef ONE_MAX_FINDER_HPP
#define ONE_MAX_FINDER_HPP
#include "Comparison.hpp"
#ifdef DEBUG_ONE_MAX_FINDER
#  undef DEBUG_NAMESPACE
#  define DEBUG_NAMESPACE "OneMaxFinder<..>"
#endif
#include "debugMacros.hpp"
namespace BK
{
  template <class EntryHandle,
    class EntryId,
    class ComparisonMaker>
  class OneMaxFinder {
  public:
    OneMaxFinder() {};
    OneMaxFinder(ComparisonMaker* comp) : _comparison(comp) {};
    ~OneMaxFinder() {};
    void switchTo(ComparisonMaker* comp) { _comparison = comp; };
    void reset() { _clean = true; };
    bool isClean() const { return _clean; };
    void anotherEntry(const EntryHandle& entry,
		      const EntryId& id);
    const EntryId& max() { 
      CALL("max()");
      ASSERT(!_clean);
      return _currentId; 
    };
  private:
    bool _clean;
    ComparisonMaker* _comparison;
    EntryId _currentId;
  }; 
}; 
#ifdef DEBUG_ONE_MAX_FINDER
#  undef DEBUG_NAMESPACE
#  define DEBUG_NAMESPACE "OneMaxFinder<..>"
#endif
#include "debugMacros.hpp"
namespace BK
{
  template <class EntryHandle,class EntryId,class ComparisonMaker>
  void 
  OneMaxFinder<EntryHandle,EntryId,ComparisonMaker>::anotherEntry(const EntryHandle& entry,
								  const EntryId& id)
  {
    CALL("anotherEntry(..)");
    if (_clean)
      {
	_clean = false;
	_currentId = id;
	_comparison->loadLHS(entry);
      }
    else
      {
	_comparison->loadRHS(entry);
	if (_comparison->compare() == Less)
	  {
	    _comparison->swapSides();
	    _currentId = id;
	  };
      };
  }; 
}; 
#endif
