/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef SMALL_HEADER_SKIP_LIST_HPP
#define SMALL_HEADER_SKIP_LIST_HPP 
#include <climits>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SMALL_HEADER_SKIP_LIST
#  define DEBUG_NAMESPACE "SHSkList<class Alloc,class ValType,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class ValType,class KeyType>
class SHSkList
{
 public:
  class Surfer;
  class Destruction;
  class Node
  {
   public:
    ~Node() {};
    KeyType key() const 
    {
     CALL("key() const"); 
     ASSERT(checkObj());
     return _key; 
    };
    const ValType& value() const { return _value; };
    ValType& value() { return _value; };
    Node* next() { return *(jump()); };
    const Node* next() const { return *(jump()); }; 
   private:
    int level() const { return _level; };
   private: 
    Node** jump() { return static_cast<Node**>(static_cast<void *>(this + 1)); }; 
    const Node* const * jump() const { return static_cast<const Node* const *>(static_cast<const void *>(this + 1)); };
    Node** jump(int lev) { return jump() + lev; };
    const Node* const * jump(int lev) const { return jump() + lev; };
    Node(int l,const KeyType& k) : _key(k), _level(l) {};
    Node* next(int lev) { return jump()[lev]; }; 
    const Node* next(int lev) const { return jump()[lev]; };
    static const Node* node(const Node * const * jmp)  
    {
     return (static_cast<const Node*>(static_cast<const void*>(jmp))) - 1;
    };
    static Node* node(Node ** jmp)           
    {
      return (static_cast<Node*>(static_cast<void*>(jmp))) - 1;
    };
    void* operator new(size_t size,int lv) 
    {    
     CALL("operator new(size_t size,int lv)"); 
     return Alloc::allocate(size + (lv+1)*sizeof(Node*));
    };
    void operator delete(void* obj)
    {
     CALL("operator delete(void* obj)");
     Alloc::deallocate(obj,sizeof(Node) + ((static_cast<Node*>(obj))->level() + 1)*sizeof(Node*));
    };   
    void operator delete(void* obj,int lv)
    {
     CALL("operator delete(void* obj,int lv)");
      Alloc::deallocate(obj,sizeof(Node) + (lv + 1)*sizeof(Node*));
    };   
    #ifndef NO_DEBUG 
     bool checkObj() const
     {
      #ifndef DEBUG_ALLOC
       return true;
      #else 
       ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<const void*>(this));
       if (!objDesc) 
        { 
         DMSG("SHSkList<..>::Node::checkObj()> Null pointer to object descriptor.\n");  
         return false; 
        }; 
       bool res = true;   
       #ifdef DEBUG_ALLOC_OBJ_SIZE
        if (objDesc->size !=  (sizeof(Node) + (level() + 1)*sizeof(Node*)))
         {
          DMSG("SHSkList<..>::Node::checkObj()> objDesc->size != (sizeof(Node) + (level() + 1)*sizeof(Node*)) \n");
          res = false;
         };
       #endif
       #ifdef DEBUG_ALLOC_OBJ_STATUS
        if (objDesc->status != ObjDesc::Persistent) 
         {
          DMSG("SHSkList<..>::Node::checkObj()> objDesc->status != ObjDesc::Persistent\n");
          res = false;
         };
       #endif
       if (!res) DOP(DMSG("SHSkList<..>::Node::checkObj()> host object ") << *objDesc << "\n"); 
       return res;
      #endif
     };
    #endif
   private: 
    KeyType _key;    
    int _level;
    ValType _value; 
   friend class SHSkList<Alloc,ValType,KeyType>;
   friend class Surfer;
   friend class Destruction;
  }; 
  class Surfer
  {
   public:
    Surfer() {};
    Surfer(const SHSkList& sl) { reset(sl); };
    ~Surfer() {};
    void reset(const SHSkList& sl) 
    {
     CALL("Surfer::reset(const SHSkList& sl)");
     ASSERT(sl.checkIntegrity());
     _currMaxLevel = sl._currMaxLevel;
     _jump = sl._topIndex;
     _currNode = _jump[0];
    };
    const Node* currNode() const { return _currNode; };
    const Node* const * currNodeAddr() const { return &_currNode; }; 
    bool next()
    {
     CALL("next()");
     _currNode = _currNode->next();
     return (bool)_currNode;
    }; 
    bool find(const KeyType& searchKey)
    {
      return find(searchKey,_currNode); 
    }; 
    void getTo(const KeyType& searchKey)
    {
      CALL("getTo(const KeyType& searchKey)");
      ALWAYS(find(searchKey,_currNode)); 
    }; 
    bool find(const KeyType searchKey,const Node*& node)
    {
     CALL("Surfer::find(const KeyType& searchKey,const Node*& node)");
     ml_next_node:
      node = _jump[_currMaxLevel];   
      if (node) 
       {
    	ASSERT(node->level() == _currMaxLevel); 
	if (node->key() < searchKey)
	 {
          if (!_currMaxLevel) return false;
          int currLevel = _currMaxLevel - 1;
          const Node* const * jump_ = _jump;        
	  next_node:
	   node = jump_[currLevel];
           ASSERT(node->level() >= currLevel);
           if (node->key() == searchKey) return true;
           if (node->key() < searchKey)
	    { 
	     if (!currLevel) return false; 
             currLevel--;  
            }
           else 
            jump_ = node->jump();
           goto next_node;
         };
        _jump = node->jump();
        if (node->key() == searchKey) return true;
        goto ml_next_node;
       }
      else 
       {
	 if (!_currMaxLevel) return false;
	 _currMaxLevel--;
	 goto ml_next_node;
       };      
    }; 
  private:
    int _currMaxLevel;
    const Node* const * _jump;
    const Node* _currNode;
  }; 
 public:
  SHSkList() 
   {
    CALL("constructor SHSkList()"); 
    _currMaxLevel = 0;
    _sizeOfTopIndex = 1;
    _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*)); 
    _topIndex[0] = (Node*)0;
    _insertCount = 0;
    _size = 0;
    ASSERT(checkIntegrity());
   }; 
  ~SHSkList() 
   {
    CALL("destructor ~SHSkList()");
    if (DestructionMode::isThorough()) 
      {
	Node* node = _topIndex[0];
	Node* tmp;
	while (node) 
	  { 
	    tmp = node;   
	    node = node->next(); 
	    delete tmp;
	  };  
	Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
      };
    BK_CORRUPT(_topIndex);
   };
  void* operator new(size_t) 
  {
   CALL("operator new(size_t)"); 
   return Alloc::allocate(sizeof(SHSkList));
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)");
    Alloc::deallocate(obj,sizeof(SHSkList));
  };
  void reset() 
  {
   CALL("reset()");
   ASSERT(checkIntegrity());
   Node* node = _topIndex[0];
   Node* tmp;
   while (node) 
     { 
       tmp = node;   
       node = node->next(); 
       delete tmp;
     };  
   Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
    _currMaxLevel = 0;
    _sizeOfTopIndex = 1;
    _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*)); 
    _topIndex[0] = (Node*)0;
    _insertCount = 0;
    _size = 0;
    ASSERT(checkIntegrity());
  }; 
  operator bool() const 
  {
   CALL("operator bool() const"); 
   return _topIndex[0] != (Node*)0; 
  }; 
  const Node* first() const 
  {
   CALL("first() const"); 
   return _topIndex[0]; 
  };
  Node* first() 
  {
   CALL("first()");
   return _topIndex[0]; 
  };
  ulong size() const { return _size; }; 
#ifndef NO_DEBUG
  ulong count() const 
  {   
   ulong res = 0; 
   for (const Node* n = first(); n; n = n->next()) 
    { 
     res++;
    };
   return res;
  };
#endif
  bool contains(const KeyType& searchKey) const
  {
   Surfer surfer(*this);
   const Node* dummy;
   return surfer.find(searchKey,dummy);
  }; 
  Node* insert(const KeyType& searchKey,bool& newNode);
  bool remove(const KeyType& searchKey,ValType& val);
  bool remove(const KeyType& searchKey);
  #ifndef NO_DEBUG
   bool checkIntegrity() const
   {
    if (!_topIndex) return false; 
    #ifdef DEBUG_ALLOC
     bool res = true;
     ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<void*>(_topIndex));          
     if (!objDesc) 
      { 
       DMSG("SHSkList<..>::checkIntegrity()> Null pointer to object descriptor for *_topIndex.\n");  
       return false; 
      };     
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size !=  _sizeOfTopIndex * sizeof(Node*)) 
       {
        DMSG("SHSkList<..>::checkIntegrity()> Wrong size of *_topIndex.\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent) 
       {
        DMSG("SHSkList<..>::checkIntegrity()> objDesc->status != ObjDesc::Persistent for *_topIndex.\n");
        res = false; 
       };
     #endif      
     if (!res) return false;
    #endif
    return (count() == size()); 
   }; 
  #endif
  ostream& output(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     node->output(ostr,_currMaxLevel) << "\n";
    };
   return ostr;
  };
  ostream& outputKeys(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     ostr << node->key() << "\n";
    };
   return ostr;
  };
 private:
  SHSkList(const SHSkList& sl);
  SHSkList& operator=(const SHSkList& sl);
  int randomLevel() const
  { 
   int res = 0;
   ulong tmp = _insertCount;
   while (!(tmp % 2))
    { 
     res++;
     tmp = tmp/2;
    };
   return res;
  }; 
 private:
  ulong _sizeOfTopIndex;
  Node** _topIndex;
  int _currMaxLevel;
  ulong _insertCount;
  ulong _size;
 friend class Surfer; 
 friend class Destruction;   
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SMALL_HEADER_SKIP_LIST
 #define DEBUG_NAMESPACE "SHSkList<class Alloc,class ValType,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class ValType,class KeyType>
inline
typename SHSkList<Alloc,ValType,KeyType>::Node* SHSkList<Alloc,ValType,KeyType>::insert(const KeyType& searchKey,bool& newNode)  
{
 CALL("insert(const KeyType& searchKey,bool& newNode)");
 ASSERT_IN(checkIntegrity(),"START");
 Node* update[32];
 Node* currNode = Node::node(_topIndex);      
 int currLevel = _currMaxLevel;
 Node* nextNode = currNode->next(currLevel);  
 KeyType nodeKey;
 check_next_node:
  if (nextNode)
   {
    ASSERT(nextNode->level() <= currLevel);
    nodeKey = nextNode->key();
    if (nodeKey > searchKey) 
     {
      currNode = nextNode;
      nextNode = nextNode->next(currLevel);        
      goto check_next_node;
     };
    if (nodeKey == searchKey) { newNode = false; ASSERT(checkIntegrity()); return nextNode; };
    update[currLevel] = currNode;
    while (currLevel) 
     { 
      currLevel--;
      nextNode = currNode->next(currLevel);
      nodeKey = nextNode->key();
      while (nodeKey > searchKey)
       {
        currNode = nextNode;
        nextNode = nextNode->next(currLevel); 
        nodeKey = nextNode->key();
       };
      if (nodeKey == searchKey) { newNode = false; ASSERT(checkIntegrity()); return nextNode; };
      update[currLevel] = currNode;
     };
    ASSERT(nextNode->key() < searchKey);
   }
  else 
   {
    update[currLevel] = currNode;
    if (currLevel) 
     {
      currLevel--;
      nextNode = currNode->next(currLevel);
      goto check_next_node;
     };
   };
  ASSERT(!currLevel);    
  newNode = true;
  ASSERT(_insertCount < INT_MAX);
  _insertCount++;  
  _size++;
  currLevel = randomLevel();
  currNode = new (currLevel) Node(currLevel,searchKey); 
  Node** nodeJump = currNode->jump();  
  Node** jump; 
  if (currLevel > _currMaxLevel)
   {
     Node** newTopIndex = (Node**)Alloc::allocate((currLevel + 1) * sizeof(Node*));
	 int lev;
    for (lev = _currMaxLevel; lev>=0; lev--)
     newTopIndex[lev] = _topIndex[lev];
     Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));   
    Node* pseudonode = Node::node(_topIndex);      
    for (lev = _currMaxLevel; lev>=0; lev--)
     { 
      if (update[lev] == pseudonode)      
       {
        jump = newTopIndex + lev;
       }
      else 
       jump = update[lev]->jump() + lev;
      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true);
      nodeJump[lev] = *jump;       
      *jump = currNode;
     };
    _topIndex = newTopIndex;
    _sizeOfTopIndex = currLevel + 1;
    do
     {
      _currMaxLevel++;
      _topIndex[_currMaxLevel] = currNode;
      nodeJump[_currMaxLevel] = (Node*)0;
     }
    while (_currMaxLevel < currLevel);
   }
  else 
   {       
    for (int lev = currLevel; lev>=0; lev--)
     { 
      jump = update[lev]->jump() + lev;  
      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true); 
      nodeJump[lev] = *jump;       
      *jump = currNode;
     };
   }; 
  ASSERT_IN(checkIntegrity(),"END");
  return currNode;
}; 
template <class Alloc,class ValType,class KeyType>
inline
bool SHSkList<Alloc,ValType,KeyType>::remove(const KeyType& searchKey,ValType& val) 
{
 CALL("remove(const KeyType& searchKey,ValType& val)");
 ASSERT(checkIntegrity());
 Node** jump = _topIndex;  
 int currLevel = _currMaxLevel; 
 Node* node;
 next_node:   
  node = jump[currLevel]; 
  if (node)
   {
    ASSERT(node->level() >= currLevel);  
    if (node->key() == searchKey)
     { 
      jump[currLevel] = node->jump()[currLevel]; 
      if (currLevel) { currLevel--; goto next_node; };
      val = node->value();
      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel) 
	_currMaxLevel--; 
      delete node;
      _size--;
      ASSERT(checkIntegrity());
      ASSERT(!contains(searchKey));
      return true;
     };
    if (node->key() > searchKey) 
     {
      jump = node->jump();        
      goto next_node;
     };
    if (currLevel) { currLevel--; goto next_node; }; 
    ASSERT(checkIntegrity());    
    ASSERT(!contains(searchKey)); 
    return false;
   };
  if (currLevel) { currLevel--; goto next_node; };
  ASSERT(checkIntegrity());
  ASSERT(!contains(searchKey));
  return false;
}; 
template <class Alloc,class ValType,class KeyType>
inline
bool SHSkList<Alloc,ValType,KeyType>::remove(const KeyType& searchKey) 
{
 CALL("remove(const KeyType& searchKey)");
 ASSERT(checkIntegrity());
 Node** jump = _topIndex;  
 int currLevel = _currMaxLevel; 
 Node* node;
 next_node:   
  node = jump[currLevel]; 
  if (node)
   {
    ASSERT(node->level() >= currLevel);  
    if (node->key() == searchKey)
     { 
      jump[currLevel] = node->jump()[currLevel]; 
      if (currLevel) { currLevel--; goto next_node; };
      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel) 
	_currMaxLevel--; 
      delete node;
      _size--;
      ASSERT(checkIntegrity());
      ASSERT(!contains(searchKey));
      return true;
     };
    if (node->key() > searchKey) 
     {
      jump = node->jump();        
      goto next_node;
     };
    if (currLevel) { currLevel--; goto next_node; }; 
    ASSERT(checkIntegrity());    
    ASSERT(!contains(searchKey)); 
    return false;
   };
  if (currLevel) { currLevel--; goto next_node; };
  ASSERT(checkIntegrity());
  ASSERT(!contains(searchKey));
  return false;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SMALL_HEADER_SKIP_LIST
 #define DEBUG_NAMESPACE "SHVoidSkList<class Alloc,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class KeyType>
class SHVoidSkList
{
 public:
  class Surfer;
  class Node
  {
   public:
    ~Node() {};
    KeyType key() const 
    {
     CALL("key() const"); 
     ASSERT(checkObj());
     return _key; 
    };
    Node* next() { return *(jump()); };
    const Node* next() const { return *(jump()); };   
   private:
    int level() const { return _level; };
    Node** jump() { return (Node**)(this + 1); }; 
    const Node* const * jump() const { return (const Node* const *)(this + 1); };
    Node** jump(int lev) { return ((Node**)(this + 1)) + lev; }; 
    const Node* const * jump(int lev) const { return ((const Node* const *)(this + 1)) + lev; };
    Node(int l,const KeyType& k) : _key(k), _level(l) {};
    Node* next(int lev) { return ((Node**)(this + 1))[lev]; }; 
    const Node* next(int lev) const { return ((const Node* const *)(this + 1))[lev]; };
    static const Node* node(const Node * const * jmp)  
    {
     return ((Node*)jmp) - 1;
    };
    static Node* node(Node ** jmp)           
    {
     return ((Node*)jmp) - 1;
    };
    void* operator new(size_t size,int lv) 
    {    
     CALL("operator new(size_t size,int lv)");
      return Alloc::allocate(size + (lv+1)*sizeof(Node*));
    };
    void operator delete(void* obj)
    {
     CALL("operator delete(void* obj)");
      Alloc::deallocate(obj,sizeof(Node) + ((static_cast<Node*>(obj))->level() + 1)*sizeof(Node*));
    };   
    void operator delete(void* obj,int lv)
    {
     CALL("operator delete(void* obj,int lv)");
      Alloc::deallocate(obj,sizeof(Node) + (lv + 1)*sizeof(Node*));
    };   
    #ifndef NO_DEBUG 
     bool checkObj() const
     {
      #ifndef DEBUG_ALLOC
       return true;
      #else 
       ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<const void*>(this));
       if (!objDesc) 
        { 
         DMSG("SHVoidSkList<..>::Node::checkObj()> Null pointer to object descriptor.\n");  
         return false; 
        }; 
       bool res = true;   
       #ifdef DEBUG_ALLOC_OBJ_SIZE
        if (objDesc->size !=  (sizeof(Node) + (level() + 1)*sizeof(Node*)))
         {
          DMSG("SHVoidSkList<..>::Node::checkObj()> objDesc->size != (sizeof(Node) + (level() + 1)*sizeof(Node*)) \n");
          res = false;
         };
       #endif
       #ifdef DEBUG_ALLOC_OBJ_STATUS
        if (objDesc->status != ObjDesc::Persistent) 
         {
          DMSG("SHVoidSkList<..>::Node::checkObj()> objDesc->status != ObjDesc::Persistent\n");
          res = false;
         };
       #endif
       if (!res) DOP(DMSG("SHVoidSkList<..>::Node::checkObj()> host object ") << *objDesc << "\n"); 
       return res;
      #endif
     };
    #endif
   private:  
    KeyType _key;
    int _level;  
   friend class SHVoidSkList<Alloc,KeyType>;
   friend class Surfer;
  }; 
  class Surfer
  {
   public:
    Surfer() {};
    Surfer(const SHVoidSkList& sl) { reset(sl); };
    ~Surfer() {};
    void reset(const SHVoidSkList& sl) 
    {
     CALL("Surfer::reset(const SHVoidSkList& sl)");
     ASSERT(sl.checkIntegrity());
     _currMaxLevel = sl._currMaxLevel;
     _jump = sl._topIndex;
     _currNode = _jump[0];
    };
    const Node* currNode() const { return _currNode; };
    const Node* const * currNodeAddr() const { return &_currNode; }; 
    bool next()
    {
     CALL("next()"); 
     _currNode = _currNode->next();
     return (bool)_currNode;
    }; 
    bool find(const KeyType& searchKey)
    {
      return find(searchKey,_currNode); 
    }; 
    void getTo(const KeyType& searchKey) 
    { 
      CALL("getTo(const KeyType& searchKey)");
      ALWAYS(find(searchKey,_currNode));       
    }; 
    bool find(const KeyType searchKey,const Node*& node)
    {
     CALL("Surfer::find(const KeyType& searchKey,const Node*& node)");
     ml_next_node:
      node = _jump[_currMaxLevel];   
      if (node) 
       {
    	ASSERT(node->level() == _currMaxLevel); 
        if (node->key() < searchKey)
	 {
          if (!_currMaxLevel) return false;
          int currLevel = _currMaxLevel - 1;
          const Node* const * jump_ = _jump;        
	  next_node:
	   node = jump_[currLevel];
           ASSERT(node->level() >= currLevel);
           if (node->key() == searchKey) return true;
           if (node->key() < searchKey)
	    { 
	     if (!currLevel) return false;   
             currLevel--;  
            }
           else 
            jump_ = node->jump();
           goto next_node;
         };
        _jump = node->jump();
        if (node->key() == searchKey) return true;
        goto ml_next_node;
       }
      else 
       {
	if (!_currMaxLevel) return false;
        _currMaxLevel--;
        goto ml_next_node;
       };      
    }; 
   private:
    int _currMaxLevel;
    const Node* const * _jump;
    const Node* _currNode;
  }; 
public:
  SHVoidSkList() 
  {
    CALL("constructor SHVoidSkList()"); 
    _currMaxLevel = 0;
    _sizeOfTopIndex = 1;
    _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*));
    _topIndex[0] = (Node*)0;
    _insertCount = 0;
    _size = 0;
    ASSERT(checkIntegrity());
  }; 
  ~SHVoidSkList() 
   {
    CALL("destructor ~SHVoidSkList()");
    if (DestructionMode::isThorough()) 
      {
	Node* node = _topIndex[0];
	Node* tmp;
	while (node) 
	  { 
	    tmp = node;   
	    node = node->next(); 
	    delete tmp;
	  };  
	Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
      };
   }; 
  void* operator new(size_t) 
  {
   CALL("operator new(size_t)"); 
    return Alloc::allocate(sizeof(SHVoidSkList));
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)"); 
    Alloc::deallocate(obj,sizeof(SHVoidSkList));
  };
  void reset() 
  {
   CALL("reset()");
   ASSERT(checkIntegrity());
   Node* node = _topIndex[0];
   Node* tmp;
   while (node) 
     { 
       tmp = node;   
       node = node->next(); 
       delete tmp;
     };  
   Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*));
   _currMaxLevel = 0;
    _sizeOfTopIndex = 1;
    _topIndex = (Node**)Alloc::allocate(_sizeOfTopIndex * sizeof(Node*)); 
    _topIndex[0] = (Node*)0;
    _insertCount = 0;
    _size = 0;
    ASSERT(checkIntegrity());
  }; 
  operator bool() const 
  {
   CALL("operator bool() const"); 
   return _topIndex[0] != 0; 
  }; 
  const Node* first() const 
  {
   CALL("first() const"); 
   return _topIndex[0]; 
  };
  Node* first() 
  {
   CALL("first()");
   return _topIndex[0]; 
  };
  ulong size() const { return _size; }; 
#ifndef NO_DEBUG
  ulong count() const 
  {   
   ulong res = 0; 
   for (const Node* n = first(); n; n = n->next()) 
    { 
     res++;
    };
   return res;
  };
#endif
  bool contains(const KeyType& searchKey) const
  {
   Surfer surfer(*this);
   const Node* dummy;
   return surfer.find(searchKey,dummy);
  }; 
  bool insert(const KeyType& searchKey);
  bool remove(const KeyType& searchKey);
  #ifndef NO_DEBUG
   bool checkIntegrity() const
   {
    if (!_topIndex) return false; 
    #ifdef DEBUG_ALLOC
     bool res = true;
     ObjDesc* objDesc = GlobAlloc::hostObj(static_cast<void*>(_topIndex));          
     if (!objDesc) 
      { 
       DMSG("SHVoidSkList<..>::checkIntegrity()> Null pointer to object descriptor for *_topIndex.\n");  
       return false; 
      };     
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size !=  _sizeOfTopIndex * sizeof(Node*)) 
       {
        DMSG("SHVoidSkList<..>::checkIntegrity()> Wrong size of *_topIndex.\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent) 
       {
        DMSG("SHVoidSkList<..>::checkIntegrity()> objDesc->status != ObjDesc::Persistent for *_topIndex.\n");
        res = false; 
       };
     #endif      
     if (!res) return false;
    #endif
    return (count() == size()); 
   }; 
  #endif
  ostream& output(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     node->output(ostr,_currMaxLevel) << "\n";
    };
   return ostr;
  };
  ostream& outputKeys(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     ostr << node->key() << "\n";
    };
   return ostr;
  };
 private:
  SHVoidSkList(const SHVoidSkList& sl);
  SHVoidSkList& operator=(const SHVoidSkList& sl);
  int randomLevel() const
  { 
   int res = 0;
   ulong tmp = _insertCount;
   while (!(tmp % 2))
    { 
     res++;
     tmp = tmp/2;
    };
   return res;
  }; 
 private:
  ulong _sizeOfTopIndex;
  Node** _topIndex;
  int _currMaxLevel;
  ulong _insertCount;
  ulong _size;
 friend class Surfer; 
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SMALL_HEADER_SKIP_LIST
 #define DEBUG_NAMESPACE "SHVoidSkList<class Alloc,class KeyType>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class KeyType>
inline
bool SHVoidSkList<Alloc,KeyType>::insert(const KeyType& searchKey)  
{
 CALL("insert(const KeyType& searchKey)");
 ASSERT_IN(checkIntegrity(),"START");
 Node* update[32];
 Node* currNode = Node::node(_topIndex);      
 int currLevel = _currMaxLevel;
 Node* nextNode = currNode->next(currLevel);  
 KeyType nodeKey;
 check_next_node:
  if (nextNode)
   {
    ASSERT(nextNode->level() <= currLevel);
    nodeKey = nextNode->key();
    if (nodeKey > searchKey) 
     {
      currNode = nextNode;
      nextNode = nextNode->next(currLevel);        
      goto check_next_node;
     };
    if (nodeKey == searchKey) 
      { 
	ASSERT(checkIntegrity()); 
	return false; 
      };
    update[currLevel] = currNode;
    while (currLevel) 
     { 
      currLevel--;
      nextNode = currNode->next(currLevel);
      nodeKey = nextNode->key();
      while (nodeKey > searchKey)
       {
	currNode = nextNode;
        nextNode = nextNode->next(currLevel); 
        nodeKey = nextNode->key();
       };
      if (nodeKey == searchKey) { ASSERT(checkIntegrity()); return false; };
      update[currLevel] = currNode;
     };
    ASSERT(nextNode->key() < searchKey);
   }
  else 
   {
    update[currLevel] = currNode;
    if (currLevel) 
     {
      currLevel--;
      nextNode = currNode->next(currLevel);
      goto check_next_node;
     };
   };
  ASSERT(!currLevel);    
  ASSERT(_insertCount < INT_MAX);
  _insertCount++;  
  _size++;
  currLevel = randomLevel();
  currNode = new (currLevel) Node(currLevel,searchKey); 
  Node** nodeJump = currNode->jump();  
  Node** jump; 
  if (currLevel > _currMaxLevel)
   {
     Node** newTopIndex = (Node**)Alloc::allocate((currLevel + 1) * sizeof(Node*));
    int lev;
    for (lev = _currMaxLevel; lev>=0; lev--)
     newTopIndex[lev] = _topIndex[lev];
     Alloc::deallocate(_topIndex,_sizeOfTopIndex * sizeof(Node*)); 
    Node* pseudonode = Node::node(_topIndex);      
    for (lev = _currMaxLevel; lev>=0; lev--)
     { 
      if (update[lev] == pseudonode)      
       {
	jump = newTopIndex + lev;
       }
      else 
       jump = update[lev]->jump() + lev; 
      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true);
      nodeJump[lev] = *jump;       
      *jump = currNode;
     };
    _sizeOfTopIndex = currLevel + 1;
    _topIndex = newTopIndex;
    do
     {
      _currMaxLevel++;
      _topIndex[_currMaxLevel] = currNode;
      nodeJump[_currMaxLevel] = (Node*)0;
     }
    while (_currMaxLevel < currLevel);
   }
  else 
   {       
    for (int lev = currLevel; lev>=0; lev--)
     { 
      jump = update[lev]->jump() + lev;  
      ASSERT((*jump) ? ((*jump)->key() < searchKey) : true); 
      nodeJump[lev] = *jump;       
      *jump = currNode;
     };
   }; 
  ASSERT_IN(checkIntegrity(),"END");
  return true;
}; 
template <class Alloc,class KeyType>
inline
bool SHVoidSkList<Alloc,KeyType>::remove(const KeyType& searchKey) 
{
 CALL("remove(const KeyType& searchKey)");
 ASSERT(checkIntegrity());
 Node** jump = _topIndex;  
 int currLevel = _currMaxLevel; 
 Node* node;
 next_node:   
  node = jump[currLevel]; 
  if (node)
   {
    ASSERT(node->level() >= currLevel);  
    if (node->key() == searchKey)
     { 
      jump[currLevel] = node->jump()[currLevel]; 
      if (currLevel) { currLevel--; goto next_node; };
      while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel)
	_currMaxLevel--; 
      delete node;
      _size--;
      ASSERT(checkIntegrity());
      ASSERT(!contains(searchKey));
      return true;
     };
    if (node->key() > searchKey) 
     {
      jump = node->jump();        
      goto next_node;
     };
    if (currLevel) { currLevel--; goto next_node; }; 
    ASSERT(checkIntegrity());    
    ASSERT(!contains(searchKey)); 
    return false;
   };
  if (currLevel) { currLevel--; goto next_node; };
  ASSERT(checkIntegrity());
  ASSERT(!contains(searchKey));
  return false;
}; 
}; 
#endif
