/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef STACK_HPP
#define STACK_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STACK
#  define DEBUG_NAMESPACE "Stack<class ElementT,ulong maxSize>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class ElementT,ulong maxSize>
class Stack
{
 public:
  Stack() : cursor(static_cast<ElementT*>(static_cast<void*>(memory))) 
  {
   #ifdef DEBUG_STACK
    name = "?";
   #endif
  };
  Stack(const char* nm) 
   : cursor((static_cast<ElementT*>(static_cast<void*>(memory)))) 
  {
   #ifdef DEBUG_STACK
    name = nm;
   #endif
  };
  Stack(const Stack& s) 
   : cursor((static_cast<ElementT*>(static_cast<void*>(memory))))
  {
   for (ElementT* p = static_cast<ElementT*>(static_cast<void*>(s.memory)); p < s.cursor; p++) push(*p); 
   #ifdef DEBUG_STACK
    name = "?";
   #endif        
  }; 
  Stack(const Stack& s,const char* nm) 
   : cursor((static_cast<ElementT*>(static_cast<void*>(memory))))
  {
   for (ElementT* p = static_cast<ElementT*>(static_cast<void*>(s.memory)); p < s.cursor; p++) push(*p); 
   #ifdef DEBUG_STACK
    name = nm;
   #endif        
  };
  ~Stack() 
  { 
    CALLM(name,"destructor ~Stack()");
    BK_CORRUPT(*this);
  };
  Stack& operator=(const Stack& s) { copy(s); return *this; };     
  void reset() { cursor = (static_cast<ElementT*>(static_cast<void*>(memory))); };
  operator bool() const { return cursor != (static_cast<const ElementT*>(static_cast<const void*>(memory))); };
  bool empty() const { return cursor ==  (static_cast<const ElementT*>(static_cast<const void*>(memory))); };
  bool nonempty() const { return cursor != (static_cast<const ElementT*>(static_cast<const void*>(memory))); };
  ulong size() const { return (cursor - (static_cast<const ElementT*>(static_cast<const void*>(memory)))); }; 
  ElementT* begin() { return (static_cast<ElementT*>(static_cast<void*>(memory))); };
  ElementT* end() { return cursor; };
  ElementT* last() { return cursor - 1; };
  const ElementT* begin() const { return (static_cast<const ElementT*>(static_cast<const void*>(memory))); };
  const ElementT* end() const { return cursor; };
  const ElementT* last() const { return cursor - 1; };
  ElementT& top() 
  {
   CALLM(name,"(ElementT& top())"); 
   ASSERT(cursor > static_cast<ElementT*>(static_cast<void*>(memory)));
   return *(cursor-1); 
  };
  const ElementT& top() const 
  {
   CALLM(name,"(const ElementT& top())"); 
   ASSERT(cursor > static_cast<const ElementT*>(static_cast<const void*>(memory)));
   return *(cursor-1); 
  };
  ElementT& next() 
  {
   CALLM(name,"(ElementT& next())");
   ASSERT(nonfull());
   return *cursor; 
  };
  const ElementT& next() const 
  {
   CALLM(name,"(const ElementT& next())");
   ASSERT(nonfull()); 
   return *cursor; 
  };
  void push(const ElementT& el) 
  {
   CALLM(name,"push(const ElementT& el)");
   ASSERT(nonfull());
   *cursor = el; 
   cursor++; 
  };
  void push() 
  { 
   CALLM(name,"push()");
   ASSERT(nonfull());
   cursor++; 
  };
  const ElementT& pop() 
  {
   CALLM(name,"const ElementT& pop()");
   ASSERT(nonempty());
   cursor--; 
   return *cursor; 
  };
  void skipTop() 
  {
   CALLM(name,"skipTop()");
   ASSERT(nonempty());
   cursor--; 
  };
  const ElementT& nth(ulong n) const
  {
   return (static_cast<const ElementT*>(static_cast<const void*>(memory)))[n]; 
  };
  ElementT& nth(ulong n)
  {
   return (static_cast<ElementT*>(static_cast<void*>(memory)))[n]; 
  };  
 private:
  bool nonfull() const { return size() < maxSize; }; 
 private:
  char memory[sizeof(ElementT)*maxSize];
  ElementT* cursor;  
  #ifdef DEBUG_NAMESPACE
   const char* name;
  #endif   
}; 
}; 
#endif
