/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TUPLE_HPP
#define TUPLE_HPP
#include <iostream>
#include "jargon.hpp"
namespace BK 
{
  template <class T1,class T2>
  struct Pair
  {
  public:
    typedef T1 first_type;
    typedef T2 second_type;
    T1 first;
    T2 second;
    Pair() {};
    Pair(const T1& x,const T2& y) : first(x), second(y) {};
    template <class U, class V> Pair(const Pair<U,V> &p) :
      first(p.first),
      second(p.second)
    {  
    };
    bool operator<(const Pair& t)
    {
      return (first < t.first) || ((first == t.first) && (second < t.second));
    };
    bool operator>(const Pair& t)
    {
      return (first > t.first) || ((first == t.first) && (second > t.second));
    };
    bool operator<=(const Pair& t)
    {
      return (first < t.first) || ((first == t.first) && (second <= t.second));
    };
    bool operator>=(const Pair& t)
    {
      return (first > t.first) || ((first == t.first) && (second >= t.second));
    };
    bool operator==(const Pair& t)
    {
      return (first == t.first) && (second == t.second);
    };   
  }; 
}; 
namespace BK 
{
  template <class T1,class T2,class T3>
  struct Triple
  {
  public:
    typedef T1 first_type;
    typedef T2 second_type;
    typedef T3 third_type;
    T1 first;
    T2 second;
    T3 third;
    Triple() {};
    Triple(const T1& x,const T2& y,const T3& z) : 
      first(x), 
      second(y), 
      third(z) 
    {
    };
    template <class U, class V, class W> Triple(const Triple<U,V,W> &t) :
      first(t.first),
      second(t.second),
      third(t.fourth)
    {  
    };
    bool operator<(const Triple& t)
    {
      return 
	first < t.first || 
	(first == t.first && 
	 (second < t.second ||
	  (second == t.second && third < t.third)));
    };
    bool operator>(const Triple& t)
    {
      return 
	first > t.first || 
	(first == t.first && 
	 (second > t.second ||
	  (second == t.second && third > t.third)));
    };
    bool operator<=(const Triple& t)
    {    
      return 
	first < t.first || 
	(first == t.first && 
	 (second < t.second ||
	  (second == t.second && third <= t.third)));
    };
    bool operator>=(const Triple& t)
    {
      return 
	first > t.first || 
	(first == t.first && 
	 (second > t.second ||
	  (second == t.second && third >= t.third)));
    };
    bool operator==(const Triple& t)
    {
      return 
	(first == t.first) && (second == t.second) && (third == t.third);
    };   
  }; 
}; 
namespace std
{
  template <class T1,class T2>
  ostream& operator<<(ostream& str,const BK::Pair<T1,T2>& t)
  {
    return str << t.first << " " << t.second;
  };
  template <class T1,class T2,class T3>
  ostream& operator<<(ostream& str,const BK::Triple<T1,T2,T3>& t)
  {
    return str << t.first << " " << t.second << " " << t.third;
  };
};
#endif
