/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "CNF.hpp"
#include "Problem.hpp"
#include "Position.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
using namespace Shell;
void CNF::clausify (Problem& problem)
{
  if (! problem.hasFormulas()) {
    return;
  }
  UnitChain::DelIterator us (problem.giveUnits());
  while (us.hasNext()) {
    Unit u (us.next());
    if (u.unitType() == FORMULA) {
      FormulaList fs (u.formula());
      LiteralList emptyList;
      clausify(problem, fs, emptyList, u, Position());
      us.del();
    }
  }
} 
void CNF::clausify (Problem& problem, 
		    const FormulaList& formulas, 
		    const LiteralList& literals,
		    const Unit& parent,
		    const Position& position)
{
  if (formulas.isEmpty()) {
    Inference* inf = new Inference(IR_CLAUSIFY,parent);
    inf->setInfo(new InferenceInfo(IR_CLAUSIFY,position,0));
    Clause c (literals);
    problem.giveUnits().addFirst(Unit(c, inf));
    return;
  }
  Formula f(formulas.head());
  FormulaList fs(formulas.tail());
  switch (f.connective()) {
  case ATOM: 
    clausify(problem, fs, 
	     LiteralList(Literal(true, f.atom()),literals), 
	     parent, position);
    return;
  case NOT: 
    clausify(problem, fs, 
	     LiteralList(Literal(false, f.uarg().atom()), literals), 
	     parent, position);
    return;
  case OR: 
    {
      FormulaList newFs(f.args());
      newFs.append(fs);
      clausify (problem, newFs, literals, parent, position);
      return;
    }
  case AND:
    {
      Iterator<Formula> args(f.args());
      int choice = 0;
      while (args.hasNext()) {
	FormulaList newFs(args.next(), fs);
	clausify(problem,newFs,literals,parent,Position(choice++,position));
      }
      return;
    }
  case TRUE:
    return;
  case FALSE:
    clausify(problem,fs,literals,parent,position);
    return;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
} 
