/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Clause.hpp"
#include "ShellReader.hpp"
#include "XML.hpp"
using namespace Shell;
Clause::Clause (const LiteralList& literals)
  :
  _data (new Data(literals))
{
} 
void Clause::operator = (const Clause& t)
{
  if (t._data) {
    t._data->ref ();
  }
  if (_data) {
    _data->deref ();
  }
  _data = t._data;
} 
void Clause::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;
  if (_counter == 0) {
    delete this;
  }
} 
XML::Element* Clause::toXML() const
{
  if (literals().isEmpty()) {
    return new XML::EmptyElement("empty_clause");
  }
  XML::DeepElement* cls = new XML::DeepElement("clause");
  Iterator<Literal> ls(literals());
  while ( ls.hasNext() ) {
    cls->addChild(ls.next().toXML());
  }
  return cls;
} 
string Clause::toString () const
{
  Reader reader;
  return reader.toString(reinterpret_cast<const Vampire::Clause*>(this));
} 
bool Clause::equals (const Clause& c) const
{
  Iterator<Literal> ls1(literals());
  Iterator<Literal> ls2(c.literals());
  while (ls1.hasNext()) {
    if (! ls2.hasNext()) {
      return false;
    }
    if (! ls1.next().equals(ls2.next())) {
      return false;
    }
  }
  return ! ls2.hasNext();
} 
