/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Compare__
#define __Compare__
namespace Shell {
class PredicateSymbol;
class FunctionSymbol;
class Term;
class Atom;
class TermList;
class Literal;
class Clause;
class Formula;
class FormulaList;
class Unit;
class Problem;
class SymCounter;
enum Comparison
{
  LESS = -1,
  EQUAL = 0,
  GREATER = 1
};
class Compare
{
public:
  Compare () {}
  bool lessThan (const Unit&, const Unit&);
  bool lessThan (const Literal&, const Literal&);
  bool lessThan (const Formula& f1, const Formula& f2);
  static Comparison compare (bool b1,bool b2);
  static Comparison compare (int i1, int i2);
  static Comparison compare (const Term&, const Term&);
  static void createSymbolCounter (const Problem&);
  static Comparison compare (const PredicateSymbol&, const PredicateSymbol&);
  static Comparison compare (const FunctionSymbol&, const FunctionSymbol&);
private:
  static bool lessThan (const Clause& c1, const Clause& c2);   
  static Comparison compare (const Atom&, const Atom&);
  static Comparison compare (const Formula&, const Formula&);
  static Comparison compare (const Literal&, const Literal&);
  static Comparison compare (const TermList&, const TermList&);
  static Comparison compare (const FormulaList&, const FormulaList&);
  static SymCounter* _symbolCounter;
}; 
inline 
Comparison Compare::compare (int i1, int i2)
{
  if (i1 > i2)
    return GREATER;
  if (i1 == i2)
    return EQUAL;
  return LESS;
} 
inline 
Comparison Compare::compare (bool b1, bool b2)
{
  if (b1) {
    return b2 ? EQUAL : LESS;
  }
  return b2 ? GREATER : EQUAL;
} 
}
#endif
