/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#if DEBUG_SHELL
#include "Debugger.hpp"
#include "Map.hpp"
#include "VampireException.hpp"
using namespace Shell;
class Debugger::Memory 
  : public Map<void*,Info> 
{
}; 
bool Debugger::allocated (void* memory, 
			  int words,
			  const string& className)
{
  ASS(!_allocatedPieces->find(memory));
  _allocatedPieces->insert(memory,Info(2*words,className));
  return true;
} 
bool Debugger::released (void* memory, 
			 int words,
			 const string& className)
{
  Info found(-1,"");
  ASS (_allocatedPieces->find(memory,found));
  ASS (found.size % 2 == 0);
  ASS (!(found.size > 2*words));
  ASS (!(found.size < 2*words));
  ASS (found.className == className);
  _allocatedPieces->replace(memory,Info(2*words+1,""));
  return true;
} 
bool Debugger::reallocated (void* memory, 
			    int words, 
			    const string& className)
{
  Info found(-1,"");
  if (! _allocatedPieces->find(memory,found)) {
    cout << "Memory address " 
	 << memory 
	 << " of size "
	 << words
	 << " requested for allocation by " 
	 << className
	 << " has not been previously allocated\n";
    ASSERTION_VIOLATION;
  }
  ASS (found.size % 2 != 0);
  ASS (found.size == 2*words+1);
  _allocatedPieces->replace(memory,Info(2*words,className));
  return true;
} 
void Debugger::reportLeaks (ostream& str)
{
  int words = 0;
  str << "Memory leaks:" << "\n";
  for (int i = _allocatedPieces->_capacity-1; i >= 0; i--) {
    Memory::Entry& entry(_allocatedPieces->_entries[i]);
    if (entry.isOccupied()) {
      int size = entry.value().size;
      if (size % 2 == 0) {
	str << entry.key() << ": " << entry.value().className << "\n";
	words += size / 2;
      }
    }
  }
  str << "Total: " << words << " words\n";
} 
Debugger::Memory* Debugger::initialiseMemory()
{
  return new Debugger::Memory;
}
#endif
