/**
 * @file Hash.hpp
 * Defines hash functions for various types.
 * @since 26/02/2006 Bellevue
 * @since 31/03/2006 Redmond, reimplemented usig FNV hashing
 */

#ifndef __Hash__
#define __Hash__

#include <string>

namespace Shell {

/**
 * Hash functions for various types.
 */
class Hash {
public:
  static unsigned hash(const char* str);
  /** Hash function for strings */
  static unsigned hash(const std::string& str)
  { return hash(str.c_str()); }

  /** Hash function for pointers */
  static unsigned hash(const void* ptr)
  { return hashFNV(reinterpret_cast<const unsigned char*>(&ptr),sizeof(ptr)); }

  /** hash function for integers */
  static unsigned hash(int i)
  { return hashFNV(reinterpret_cast<const unsigned char*>(&i),sizeof(int)); }

  /** hash function for unsigned integers */
  static unsigned long hash(unsigned i)
  { return hashFNV(reinterpret_cast<const unsigned char*>(&i),sizeof(unsigned)); }

  /** hash function for long integers */
  static unsigned long hash(long i)
  { return hashFNV(reinterpret_cast<const unsigned char*>(&i),sizeof(long)); }

  /** hash function for floats */
  static unsigned long hash(float f)
  { return hashFNV(reinterpret_cast<const unsigned char*>(&f),sizeof(float)); }

  /** hash function for doubles */
  static unsigned long hash(double d)
  { return hashFNV(reinterpret_cast<const unsigned char*>(&d),sizeof(double)); }

private:
  static unsigned hashFNV(const unsigned char*,size_t length);
};

}

#endif
