/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __InferenceInfo__
#define __InferenceInfo__
#include <string>
#include "Position.hpp"
#include "Term.hpp"
#include "InferenceRule.hpp"
#include "VampireKernel/VampireKernel.hpp"
#include "XML.hpp"
namespace Shell {
class InferenceInfo
{
public:
  InferenceInfo (InferenceRule rule, 
		 const Position& position, 
		 InferenceInfo* next);
  InferenceInfo (InferenceRule rule, 
		 InferenceInfo* next);
  ~InferenceInfo () {}
  CLASS_NAME("InferenceInfo");
  USE_ALLOCATOR(InferenceInfo);
  InferenceInfo* next () const { return _next; }
  InferenceRule rule () const { return _rule; }
  const Position& position () const { return _position; }
  void destroy ();
  void destroyList ();
  XML::Element* toXML () const;
  string toString() const;
private:
  InferenceRule _rule;
  Position _position;
  InferenceInfo* _next;
}; 
class InferenceInfoTermList
  : public InferenceInfo
{
public:
  const TermList& terms () const { return _terms; }
  InferenceInfoTermList (InferenceRule rule,
			 const Position& position, 
			 InferenceInfo* next,
			 const TermList& terms);
  CLASS_NAME("InferenceInfoTermList");
  USE_ALLOCATOR(InferenceInfoTermList);
private:
  TermList _terms;
}; 
enum KernelRule {
  KR_IsInput = 1,
  KR_Preprocessing = 2,
  KR_BinaryResolution = 4,
  KR_ForwardTransitiveChaining = 8,
  KR_BackwardTransitiveChaining = 16,
  KR_ForwardSuperposition = 32,
  KR_BackwardSuperposition = 64,
  KR_EqualityResolution = 128,
  KR_EqualityFactoring = 256,
  KR_ForwardDemodulation = 512,
  KR_BackwardDemodulation = 1024,
  KR_SimplificationByEqualityResolution = 2048,
  KR_ForwardSubsumptionResolution = 4096,
  KR_BackwardSubsumptionResolution = 8192,
  KR_IsReanimatedPassive = 16384,
  KR_Evaluation = 32768,
  KR_IsNameIntro = 65536,
  KR_Folding = 131072,
  KR_Clausification = 262144
}; 
class KernelInferenceInfo
  : public InferenceInfo
{
public:
  explicit KernelInferenceInfo (VampireKernel::Clause::BackgroundFlagIterator&);
  CLASS_NAME("KernelInferenceInfo");
  USE_ALLOCATOR(KernelInferenceInfo);
  std::string rulesAsString () const;
private:
  unsigned _inferenceCodes;
  static std::string toString (KernelRule);
}; 
}
#endif 
