/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __InterpretedConstant__
#define __InterpretedConstant__
#define MAX_ARITY_FOR_INTERPRETED_SYMBOLS 3
#include "Allocator.hpp"
#include "VampireKernel/VampireKernelEnvironment.hpp"
using namespace VampireKernelEnvironment;
namespace Shell {
enum EvaluableFunction {
  GENERAL_TO_STRING = 0,
  INTEGER_ABS = 1,
  INTEGER_COMPARE = 2,
  INTEGER_DIV = 3,
  INTEGER_FROM_STRING = 4,
  INTEGER_MAX = 5,
  INTEGER_MIN = 6,
  INTEGER_MINUS = 7,
  INTEGER_MOD = 8,
  INTEGER_MULTIPLY = 9,
  INTEGER_NEGATE = 10,
  INTEGER_PLUS = 11,
  INTEGER_RANDOM = 12,
  INTEGER_SIGN = 13,
  INTEGER_TO_STRING = 14,
  REAL_ABS = 15,
  REAL_ACOS = 16,
  REAL_ASIN = 17,
  REAL_ATAN = 18,
  REAL_ATAN2 = 19,
  REAL_CEIL = 20,
  REAL_COMPARE = 21,
  REAL_COS = 22,
  REAL_COSH = 23,
  REAL_DIV = 24,
  REAL_EXP = 25,
  REAL_FLOOR = 26,
  REAL_FROM_STRING = 27,
  REAL_LN = 28,
  REAL_LOG10 = 29,
  REAL_MAX = 30,
  REAL_MIN = 31,
  REAL_MINUS = 32,
  REAL_MULTIPLY = 33,
  REAL_NEGATE = 34,
  REAL_PLUS = 35,
  REAL_POWER = 36,
  REAL_RANDOM = 37,
  REAL_ROUND = 38,
  REAL_SIGN = 39,
  REAL_SIN = 40,
  REAL_SINH = 41,
  REAL_SQRT = 42,
  REAL_TAN = 43,
  REAL_TANH = 44,
  REAL_TO_STRING = 45,
  REAL_TRUNC = 46,
  STRING_COMPARE = 47,
  STRING_CONCAT = 48,
  STRING_SIZE = 49
};
enum EvaluablePredicate {
  GENERAL_IS_INTEGER = 0,
  GENERAL_IS_REAL = 1,
  GENERAL_IS_STRING = 2,
  INTEGER_GEQ = 3,
  INTEGER_GREATER = 4,
  INTEGER_LEQ = 5,
  INTEGER_LESS = 6,
  REAL_GEQ = 7,
  REAL_GREATER = 8,
  REAL_LEQ = 9,
  REAL_LESS = 10,
  STRING_GEQ = 11,
  STRING_GREATER = 12,
  STRING_LEQ = 13,
  STRING_LESS = 14
};
enum ICSort {
  IC_INTEGER = 0,
  IC_REAL = 1,
  IC_STRING = 2
};
class InterpretedConstantValue 
{
public:
  InterpretedConstantValue() {}
  virtual ~InterpretedConstantValue() {}
  virtual ICSort tag() const = 0;
  virtual std::string toString() const = 0;
  virtual double realValue() const = 0;
  static InterpretedConstantValue*  evaluate(EvaluableFunction fun);
  static bool evaluate(EvaluablePredicate pred,bool& result);
  int compare (const InterpretedConstantValue* v) const;
public: 
  static InterpretedConstantValue* _args[MAX_ARITY_FOR_INTERPRETED_SYMBOLS];
  static int _nextArgumentIndex;
  friend class InterpretedConstant;
}; 
class IntegerConstantValue 
  : public InterpretedConstantValue 
{
 public:
  explicit IntegerConstantValue(long value) : _value(value) {}
  const long value() const {return _value;}
  double realValue() const {return (double)_value;}
  std::string toString() const;
  ICSort tag() const { return IC_INTEGER; }
  CLASS_NAME("IntegerConstantValue");
  USE_ALLOCATOR(IntegerConstantValue);
 private:
  long _value;
}; 
class RealConstantValue 
  : public InterpretedConstantValue 
{
 public:
  explicit RealConstantValue(double value) : _value(value) {}
  const double value () const {return _value;}
  std::string toString() const;
  ICSort tag() const { return IC_REAL; }
  double realValue() const {return _value;}
  CLASS_NAME("RealConstantValue");
  USE_ALLOCATOR(RealConstantValue);
 private:
  double _value;
}; 
class StringConstantValue 
  : public InterpretedConstantValue 
{
 public:
  explicit StringConstantValue(const std::string& value)
    : _value(value) {}
  const std::string& value () const {return _value;}
  ICSort tag() const { return IC_STRING; }
  std::string toString() const;
  double realValue() const;
  CLASS_NAME("StringConstantValue");
  USE_ALLOCATOR(StringConstantValue);
 private:
  std::string _value;
}; 
}
#endif
