/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Lexer__
#define __Lexer__
#include <iostream>
#include "Token.hpp"
#include "VampireException.hpp"
using namespace std;
#define CHARACTER_BUFFER_SIZE 65500
namespace Shell {
class Lexer;
class LexerException 
  : public VampireException
{
 public:                                
  LexerException (string message,const Lexer&);
  void cry (ostream&);
  XML::Element* toXML () const;
  ~LexerException () {}
 protected:
  string _message;
}; 
class Lexer 
{
public:
  Lexer(istream& in);
  bool isAtEndOfFile () const { return _eof; }
  int lastCharacter () const { return _lastCharacter; }
  int lineNumber () const { return _lineNumber; }
  virtual void readToken (Token&) = 0;
  virtual ~Lexer () {}
protected:
  int  _lastCharacter;
  char _charBuffer[CHARACTER_BUFFER_SIZE+1]; 
  int _charCursor;
  std::istream& _stream;
  bool _eof;
  int _lineNumber;
  bool readNextChar();
  void readNumber(Token&);
  void readUnsignedInteger();
  void saveLastChar();
  void saveChar(int character);
  void saveTokenText(Token&);
  static bool isDigit (int charCode) 
  { return charCode >= '0' && charCode <= '9'; }
  static bool isLetter (int charCode) 
  { return (charCode >= 'A' && charCode <= 'Z') || 
           (charCode >= 'a' && charCode <= 'z'); }
  void readSequence(const char* chars);
}; 
}
#endif
