/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <cmath>
#include <sstream>
#include "Int.hpp"
#include "NameArray.hpp"
#include "Options.hpp"
#include "Random.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "XML.hpp"
namespace Shell {
class Options::Constants {
public:
  static const char* _optionNames[];
  static const char* _shortNames[];
  static const char* _statisticsValues[];
  static const char* _demodulationValues[];
  static const char* _fdeValues[];
  static const char* _lcmValues[];
  static const char* _mainAlgValues[];
  static const char* _equalityProxyValues[];
  static const char* _splittingValues[]; 
  static const char* _symbolPrecedenceValues[];
  static const char* _tcValues[];
  static const char* _proofValues[];
  static int shortNameIndexes[];
  static NameArray optionNames;
  static NameArray shortNames;
  static NameArray statisticsValues;
  static NameArray demodulationValues;
  static NameArray fdeValues;
  static NameArray lcmValues;
  static NameArray mainAlgValues;
  static NameArray equalityProxyValues;
  static NameArray splittingValues; 
  static NameArray symbolPrecedenceValues;
  static NameArray tcValues;
  static NameArray proofValues;
}; 
const char* Options::Constants::_optionNames[] = {
  "age_weight_ratio",
  "allocation_buffer_price",
  "allocation_buffer_size",
  "backward_demodulation",
  "backward_subsumption",
  "basic_strategy",
  "decode",
  "equality_proxy",
  "equality_resolution_with_deletion",
  "fast_demodulation_by_commutativity",
  "forward_demodulation",
  "forward_subsumption",
  "forward_subsumption_resolution",
  "function_definition_elimination",
  "inequality_splitting",
  "inherited_negative_selection",
  "latex_output",
  "literal_comparison_mode",
  "log_file",
  "lrs_first_time_check",
  "main_alg",
  "main_premise_demodulation",
  "main_premise_subsumption",
  "max_active",
  "max_answers",
  "max_inference_depth",
  "max_passive",
  "max_skolem",
  "max_weight",
  "memory_limit",
  "miniscope",
  "naming",
  "no_inferences_between_rules",
  "nongoal_weight_coefficient",
  "normalize",
  "orphan_deletion",
  "output_messages",
  "problem_name",
  "proof",
  "proof_checking",
  "random_seed",
  "row_variable_max_length",
  "selected_equality_weight_coefficient",
  "selection",
  "show_active",
  "show_generated",
  "show_input",
  "show_options",
  "show_passive",
  "show_preprocessed",
  "show_profile",
  "show_weight_limit_changes",
  "side_premise_demodulation",
  "side_premise_subsumption",
  "simulated_time_limit",
  "simultaneous_backward_superposition",
  "sos",
  "splitting",
  "statistics",
  "superposition_from_variables",
  "symbol_precedence",
  "test_id",
  "test_number",
  "time_limit",
  "transitivity_chaining",
  "unused_predicate_definition_removal",
  "weight_increment",
  "xml_output"};
NameArray Options::Constants::optionNames(_optionNames,
					  sizeof(_optionNames)/sizeof(char*));
const char* Options::Constants::_shortNames[] = {
  "awr",
  "bas",
  "bd",
  "bs",
  "ep",
  "fde",
  "fsr",
  "l",
  "lcm",
  "m",
  "ma",
  "n",
  "nwc",
  "p",
  "pc",
  "s",
  "sos",
  "sp",
  "stl",
  "t",
  "wi"};
NameArray Options::Constants::shortNames(_shortNames,
					  sizeof(_shortNames)/sizeof(char*));
int Options::Constants::shortNameIndexes[] = {
  AGE_WEIGHT_RATIO,
  BASIC_STRATEGY,
  BACKWARD_DEMODULATION,
  BACKWARD_SUBSUMPTION,
  EQUALITY_PROXY,
  FUNCTION_DEFINITION_ELIMINATION,
  FORWARD_SUBSUMPTION_RESOLUTION,
  LOG_FILE,
  LITERAL_COMPARISON_MODE,
  MEMORY_LIMIT,
  MAIN_ALG,
  NORMALIZE,
  NONGOAL_WEIGHT_COEFFICIENT,
  PROOF,
  PROOF_CHECKING,
  SELECTION,
  SOS,
  SPLITTING,
  SIMULATED_TIME_LIMIT,
  TIME_LIMIT,
  WEIGHT_INCREMENT};
const char* Options::Constants::_statisticsValues[] = {
  "brief",
  "full",
  "none"};
NameArray Options::Constants::statisticsValues(_statisticsValues,
					       sizeof(_statisticsValues)/sizeof(char*));
const char* Options::Constants::_demodulationValues[] = {
  "all",
  "off",
  "preordered"};
NameArray Options::Constants::demodulationValues(_demodulationValues,
						 sizeof(_demodulationValues)/sizeof(char*));
const char* Options::Constants::_fdeValues[] = {
  "all",
  "none",
  "unused"};
NameArray Options::Constants::fdeValues(_fdeValues,
					sizeof(_fdeValues)/sizeof(char*));
const char* Options::Constants::_lcmValues[] = {
  "kinky",
  "predicate",
  "standard"
};
NameArray Options::Constants::lcmValues(_lcmValues,
					sizeof(_lcmValues)/sizeof(char*));
const char* Options::Constants::_mainAlgValues[] = {
  "discount",
  "lrs",
  "otter"};
NameArray Options::Constants::mainAlgValues(_mainAlgValues,
					    sizeof(_mainAlgValues)/sizeof(char*));
const char* Options::Constants::_equalityProxyValues[] = {
  "exp1",
  "off",
  "on"};
NameArray Options::Constants::equalityProxyValues(_equalityProxyValues,
						  sizeof(_equalityProxyValues)/sizeof(char*));
const char* Options::Constants::_splittingValues[] = { 
  "input_only",
  "off",
  "on"};
NameArray Options::Constants::splittingValues(_splittingValues,
					      sizeof(_splittingValues)/sizeof(char*));
const char* Options::Constants::_symbolPrecedenceValues[] = {
  "arity",
  "occurrence",
  "reverse_arity"};
NameArray Options::Constants::symbolPrecedenceValues(_symbolPrecedenceValues,
						     sizeof(_symbolPrecedenceValues)/sizeof(char*));
const char* Options::Constants::_tcValues[] = {
  "full",
  "none",
  "only_nonvariables"};
NameArray Options::Constants::tcValues(_tcValues,
				       sizeof(_tcValues)/sizeof(char*));
const char* Options::Constants::_proofValues[] = { 
  "succinct",
  "off",
  "on"};
NameArray Options::Constants::proofValues(_proofValues,
					  sizeof(_proofValues)/sizeof(char*));
Options::Options ()
  :
  _ageRatio(1),
  _weightRatio(1),
  _allocationBufferPrice(2.0),
  _allocationBufferSize(10),
  _backwardDemodulation(DEMODULATION_ALL),
  _backwardSubsumption(true),
  _basicStrategy(false),
  _equalityProxy(EP_OFF),
  _equalityResolutionWithDeletion(false),
  _fastDemodulationByCommutativity(true),
  _forwardDemodulation(DEMODULATION_ALL),
  _forwardSubsumption(true),
  _forwardSubsumptionResolution(true),
  _functionDefinitionElimination(FDE_ALL),
  _inequalitySplitting(3),
  _inheritedNegativeSelection(false),
  _inputFile(""),
  _latexOutput("off"),
  _literalComparisonMode(LCM_STANDARD),
  _logFile("off"),
  _lrsFirstTimeCheck(5),
  _mainAlg(LRS),
  _mainPremiseDemodulation(0),
  _mainPremiseSubsumption(0),
  _maxActive(0),
  _maxAnswers(1),
  _maxInferenceDepth(0),
  _maxPassive(0),
  _maxSkolem(32000),
  _maxWeight(0),
  _memoryLimit(300),
  _miniscope(false),
  _naming(4),
  _noInferencesBetweenRules(false),
  _nongoalWeightCoefficient(1.0),
  _normalize(false),
  _orphanDeletion(false),
  _outputMessages(true),
  _problemName(""),
  _proof(PROOF_ON),
  _proofChecking(false),
  _randomSeed(Random::getSeed()),
  _rowVariableMaxLength(-1),
  _selectedEqualityWeightCoefficient(1.0),
  _selection(10),
  _showActive(false),
  _showGenerated(false),
  _showInput(false),
  _showOptions(false),
  _showPassive(false),
  _showPreprocessed(false),
  _showProfile(false),
  _showWeightLimitChanges(false),
  _sidePremiseDemodulation(0),
  _sidePremiseSubsumption(0),
  _simulatedTimeLimit(0),
  _simultaneousBackwardSuperposition(false),
  _sos(false),
  _splitting(SPLIT_INPUT_ONLY),
  _statistics(STATISTICS_FULL),
  _superpositionFromVariables(true),
  _symbolPrecedence(BY_ARITY),
  _testId ("unspecified_test"),
  _testNumber(Random::getInteger()),
  _timeLimitInDeciseconds(0),
  _transitivityChaining(TC_NONE),
  _unusedPredicateDefinitionRemoval(true),
  _weightIncrement(false),
  _xmlOutput("off")
{
} 
void Options::set (const char* name,const char* value)
{
  TRACER ("Options::set/2");
  set(name,value,Constants::optionNames.find(name));
} 
void Options::set (const string& name,const string& value)
{
  TRACER ("Options::set/3");
  set(name.c_str(),value.c_str());
} 
void Options::set (const char* name,const char* value, int index)
{
  int intValue;
  float floatValue;
  switch (index) {
  case AGE_WEIGHT_RATIO:
    readAgeWeightRatio(value);
    return;
  case ALLOCATION_BUFFER_PRICE:
    if (Int::stringToInt(value,intValue)) {
      _allocationBufferPrice = intValue;
      return;
    }
    break;
  case ALLOCATION_BUFFER_SIZE:
    if (Int::stringToInt(value,intValue)) {
      _allocationBufferSize = intValue;
      return;
    }
    break;
  case BACKWARD_DEMODULATION:
    _backwardDemodulation = (Demodulation)Constants::demodulationValues.find(value);
    if (_backwardDemodulation == -1) {
      break;
    }
    return;
  case BACKWARD_SUBSUMPTION:
    _backwardSubsumption = onOffToBool(value,name);
    return;
  case BASIC_STRATEGY:
    _basicStrategy = onOffToBool(value,name);
    return;
  case DECODE:
    readFromTestId(value);
    return;
  case EQUALITY_PROXY:
    _equalityProxy = (EqualityProxy)Constants::equalityProxyValues.find(value);
    if (_equalityProxy == -1) {
      break;
    }
    return;
  case EQUALITY_RESOLUTION_WITH_DELETION:
    _equalityResolutionWithDeletion = onOffToBool(value,name);
    return;
  case FAST_DEMODULATION_BY_COMMUTATIVITY:
    _fastDemodulationByCommutativity = onOffToBool(value,name);
    return;
  case FORWARD_DEMODULATION:
    _forwardDemodulation = 
      (Demodulation)Constants::demodulationValues.find(value);
    if (_forwardDemodulation == -1) {
      break;
    }
    return;
  case FORWARD_SUBSUMPTION:
    _forwardSubsumption = onOffToBool(value,name);
    return;
  case FORWARD_SUBSUMPTION_RESOLUTION:
    _forwardSubsumptionResolution = onOffToBool(value,name);
    return;
  case FUNCTION_DEFINITION_ELIMINATION:
    _functionDefinitionElimination = 
      (FunctionDefinitionElimination)Constants::fdeValues.find(value);
    if (_functionDefinitionElimination == -1) {
      break;
    }
    return;
  case INEQUALITY_SPLITTING:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _inequalitySplitting = intValue;
      return;
    }
    break;
  case INHERITED_NEGATIVE_SELECTION:
    _inheritedNegativeSelection = onOffToBool(value,name);
    return;
  case LATEX_OUTPUT:
    _latexOutput = value;
    return;
  case LITERAL_COMPARISON_MODE:
    _literalComparisonMode = 
      (LiteralComparisonMode)Constants::lcmValues.find(value);
    if (_literalComparisonMode == -1) {
      break;
    }
    return;
  case LOG_FILE:
    _logFile = value;
    return;
  case LRS_FIRST_TIME_CHECK:
    if (Int::stringToInt(value,intValue) && 
	setLrsFirstTimeCheck(intValue)) {
      return;
    }
    break;
  case MAIN_ALG:
    _mainAlg = (MainAlg)Constants::mainAlgValues.find(value);
    if (_mainAlg == -1) {
      break;
    }
    return;
  case MAIN_PREMISE_DEMODULATION:
    _mainPremiseDemodulation = onOffToBool(value,name);
    return;
  case MAIN_PREMISE_SUBSUMPTION:
    _mainPremiseSubsumption = onOffToBool(value,name);
    return;
  case MAX_ACTIVE:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _maxActive = intValue;
      return;
    }
    break;
  case MAX_ANSWERS:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _maxAnswers = intValue;
      return;
    }
    break;
  case MAX_INFERENCE_DEPTH:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _maxInferenceDepth = intValue;
      return;
    }
    break;
  case MAX_PASSIVE:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _maxPassive = intValue;
      return;
    }
    break;
  case MAX_SKOLEM:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _maxSkolem = intValue;
      return;
    }
    break;
  case MAX_WEIGHT:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _maxWeight = intValue;
      return;
    }
    break;
  case MEMORY_LIMIT:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _memoryLimit = intValue;
      return;
    }
    break;
  case MINISCOPE:
    _miniscope = onOffToBool(value,name);
    return;
  case NAMING:
    if (Int::stringToUnsignedInt(value,intValue) &&
	setNaming(intValue)) {
      return;
    }
    break;
  case NO_INFERENCES_BETWEEN_RULES:
    _noInferencesBetweenRules = onOffToBool(value,name);
    return;
  case NONGOAL_WEIGHT_COEFFICIENT:
    if (Int::stringToFloat(value,floatValue) &&
	setNongoalWeightCoefficient(floatValue)) {
      return;
    }
    break;
  case NORMALIZE:
    _normalize = onOffToBool(value,name);
    return;
  case ORPHAN_DELETION:
    _orphanDeletion = onOffToBool(value,name);
    return;
  case OUTPUT_MESSAGES:
    _outputMessages = onOffToBool(value,name);
    return;
  case PROOF:
    _proof = (Proof)Constants::proofValues.find(value);
    if (_proof == -1) {
      break;
    }
    return;
  case PROOF_CHECKING:
    _proofChecking = onOffToBool(value,name);
    return;
  case RANDOM_SEED:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _randomSeed = intValue;
      return;
    }
    break;
  case ROW_VARIABLE_MAX_LENGTH:
    if (Int::stringToUnsignedInt(value,intValue)) {
      _rowVariableMaxLength = intValue;
      return;
    }
    break;
  case SELECTED_EQUALITY_WEIGHT_COEFFICIENT:
    if (Int::stringToFloat(value,floatValue) &&
	setSelectedEqualityWeightCoefficient(floatValue)) {
      return;
    }
    break;
  case SELECTION:
    if (Int::stringToInt(value,intValue) &&
	setSelection(intValue) ) {
      return;
    }
    break;
  case SHOW_ACTIVE:
    _showActive = onOffToBool(value,name);
    return;
  case SHOW_GENERATED:
    _showGenerated = onOffToBool(value,name);
    return;
  case SHOW_INPUT:
    _showInput = onOffToBool(value,name);
    return;
  case SHOW_OPTIONS:
    _showOptions = onOffToBool(value,name);
    return;
  case SHOW_PASSIVE:
    _showPassive = onOffToBool(value,name);
    return;
  case SHOW_PREPROCESSED:
    _showPreprocessed = onOffToBool(value,name);
    return;
  case SHOW_PROFILE:
    _showProfile = onOffToBool(value,name);
    return;
  case SHOW_WEIGHT_LIMIT_CHANGES:
    _showWeightLimitChanges = onOffToBool(value,name);
    return;
  case SIDE_PREMISE_DEMODULATION:
    _sidePremiseDemodulation = onOffToBool(value,name);
    return;
  case SIDE_PREMISE_SUBSUMPTION:
    _sidePremiseSubsumption = onOffToBool(value,name);
    return;
  case SIMULATED_TIME_LIMIT:
    _simulatedTimeLimit = readTimeLimit(value);
    return;
  case SIMULTANEOUS_BACKWARD_SUPERPOSITION:
    _simultaneousBackwardSuperposition = onOffToBool(value,name);
    return;
  case SOS:
    _sos = onOffToBool(value,name);
    return;
  case SPLITTING:
    _splitting = (Splitting)Constants::splittingValues.find(value);
    if (_splitting == -1) {
      break;
    }
    return;
  case STATISTICS:
    _statistics = (Statistics)Constants::statisticsValues.find(value);
    if (_statistics == -1) {
      break;
    }
    return;
  case SUPERPOSITION_FROM_VARIABLES:
    _superpositionFromVariables = onOffToBool(value,name);
    return;
  case SYMBOL_PRECEDENCE:
    _symbolPrecedence = 
      (SymbolPrecedence)Constants::symbolPrecedenceValues.find(value);
    if (_symbolPrecedence == -1) {
      break;
    }
    return;
  case TEST_ID:
    _testId = value;
    return;
  case TEST_NUMBER:
    if (Int::stringToInt(value,intValue)) {
      _testNumber = intValue;
      return;
    }
    break;
  case TIME_LIMIT:
    _timeLimitInDeciseconds = readTimeLimit(value);
    return;
  case TRANSITIVITY_CHAINING:
    _transitivityChaining =
      (TransitivityChaining)Constants::tcValues.find(value);
    if (_transitivityChaining == -1) {
      break;
    }
    return;
  case UNUSED_PREDICATE_DEFINITION_REMOVAL:
    _unusedPredicateDefinitionRemoval = onOffToBool(value,name);
    return;
  case WEIGHT_INCREMENT:
    _weightIncrement = onOffToBool(value,name);
    return;
  case XML_OUTPUT:
    _xmlOutput = value;
    return;
  case -1: 
    USER_ERROR((string)name + " is not a valid option");
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
  USER_ERROR((string)"wrong value (" + value + ") for " + name);
} 
void Options::setShort (const char* name,const char* value)
{
  TRACER ("Options::setShort");
  int found = Constants::shortNames.find(name);
  if (found == -1) {
    found = Constants::optionNames.find(name);
  }
  else {
    found = Constants::shortNameIndexes[found];
  }
  set(name,value,found);
} 
bool Options::onOffToBool (const char* onOff,const char* option)
{
  TRACER("Options::onOffToBool");
  if (! strcmp(onOff,"on")) {
    return true;
  }
  if (! strcmp(onOff,"off")) {
    return false;
  }
  USER_ERROR((string)"wrong value for " + option + ": " + onOff);
} 
string Options::boolToOnOff (bool val)
{
  TRACER("Options::boolToOnOff");
  static string on ("on");
  static string off ("off");
  return val ? on : off;
} 
bool Options::setSelection(int sel)
{
  switch (sel) {
  case -11:
  case -16:
  case -1012:
  case 8:
  case 10:
  case 11:
  case 13:
  case 16:
  case 17:
  case 1010:
  case 1011:
  case 1012:
  case 1013:
    _selection = sel;
    return true;
  default:
    return false;
  }
} 
bool Options::setNongoalWeightCoefficient (float newVal)
{
  if (newVal <= 0.0) {
    return false;
  }
  _nongoalWeightCoefficient = newVal;
  return true;
} 
bool Options::setNaming (int newVal)
{
  if (newVal > 32767) {
    return false;
  }
  _naming = newVal;
  return true;
} 
bool Options::setSelectedEqualityWeightCoefficient (float newVal)
{
  if (newVal <= 0.0) {
    return false;
  }
  _selectedEqualityWeightCoefficient = newVal;
  return true;
} 
bool Options::setLrsFirstTimeCheck (int newVal)
{
  if (newVal < 0 && newVal >= 100) {
    return false;
  }
  _lrsFirstTimeCheck = newVal;
  return true;
} 
void Options::output (ostream& str) const
{
  TRACER("Options::output");
  if (! showOptions()) {
    return;
  }
  str << "=========== Options ==========\n";
  for (int i = 0;i < Constants::optionNames.length;i++) {
    str << Constants::optionNames[i] << '=';
    outputValue(str,i);
    str << '\n';
  }
  str << "======= End of options =======\n";
} 
void Options::outputValue (ostream& str,int optionTag) const
{
  switch (optionTag) {
  case AGE_WEIGHT_RATIO:
    str << _ageRatio << ':' << _weightRatio;
    return;
  case ALLOCATION_BUFFER_PRICE:
    str << _allocationBufferPrice;
    return;
  case ALLOCATION_BUFFER_SIZE:
    str << _allocationBufferSize;
    return;
  case BACKWARD_DEMODULATION:
    str << Constants::demodulationValues[_backwardDemodulation];
    return;
  case BACKWARD_SUBSUMPTION:
    str << boolToOnOff(_backwardSubsumption);
    return;
  case BASIC_STRATEGY:
    str << boolToOnOff(_basicStrategy);
    return;
  case DECODE: 
    return;
  case EQUALITY_PROXY:
    str << Constants::equalityProxyValues[_equalityProxy];
    return;
  case EQUALITY_RESOLUTION_WITH_DELETION:
    str << boolToOnOff(_equalityResolutionWithDeletion);
    return;
  case FAST_DEMODULATION_BY_COMMUTATIVITY:
    str << boolToOnOff(_fastDemodulationByCommutativity);
    return;
  case FORWARD_DEMODULATION:
    str << Constants::demodulationValues[_forwardDemodulation];
    return;
  case FORWARD_SUBSUMPTION:
    str << boolToOnOff(_forwardSubsumption);
    return;
  case FORWARD_SUBSUMPTION_RESOLUTION:
    str << boolToOnOff(_forwardSubsumptionResolution);
    return;
  case FUNCTION_DEFINITION_ELIMINATION:
    str << Constants::fdeValues[_functionDefinitionElimination];
    return;
  case INEQUALITY_SPLITTING:
    str << _inequalitySplitting;
    return;
  case INHERITED_NEGATIVE_SELECTION:
    str << boolToOnOff(_inheritedNegativeSelection);
    return;
  case LATEX_OUTPUT:
    str << _latexOutput;
    return;
  case LITERAL_COMPARISON_MODE:
    str << Constants::lcmValues[_literalComparisonMode];
    return;
  case LOG_FILE:
    str << _logFile;
    return;
  case LRS_FIRST_TIME_CHECK:
    str << _lrsFirstTimeCheck;
    return;
  case MAIN_ALG:
    str << Constants::mainAlgValues[_mainAlg];
    return;
  case MAIN_PREMISE_DEMODULATION:
    str << boolToOnOff(_mainPremiseDemodulation);
    return;
  case MAIN_PREMISE_SUBSUMPTION:
    str << boolToOnOff(_mainPremiseSubsumption);
    return;
  case MAX_ACTIVE:
    str << _maxActive;
    return;
  case MAX_ANSWERS:
    str << _maxAnswers;
    return;
  case MAX_INFERENCE_DEPTH:
    str << _maxInferenceDepth;
    return;
  case MAX_PASSIVE:
    str << _maxPassive;
    return;
  case MAX_SKOLEM:
    str << _maxSkolem;
    return;
  case MAX_WEIGHT:
    str << _maxWeight;
    return;
  case MEMORY_LIMIT:
    str << _memoryLimit;
    return;
  case MINISCOPE:
    str << boolToOnOff(_miniscope);
    return;
  case NAMING:
    str << _naming;
    return;
  case NO_INFERENCES_BETWEEN_RULES:
    str << boolToOnOff(_noInferencesBetweenRules);
    return;
  case NONGOAL_WEIGHT_COEFFICIENT:
    str << _nongoalWeightCoefficient;
    return;
  case NORMALIZE:
    str << boolToOnOff(_normalize);
    return;
  case ORPHAN_DELETION:
    str << boolToOnOff(_orphanDeletion);
    return;
  case OUTPUT_MESSAGES:
    str << boolToOnOff(_outputMessages);
    return;
  case PROBLEM_NAME:
    str << _problemName;
    return;
  case PROOF:
    str << boolToOnOff(_proof);
    return;
  case PROOF_CHECKING:
    str << boolToOnOff(_proofChecking);
    return;
  case RANDOM_SEED:
    str << _randomSeed;
    return;
  case ROW_VARIABLE_MAX_LENGTH:
    str << _rowVariableMaxLength;
    return;
  case SELECTED_EQUALITY_WEIGHT_COEFFICIENT:
    str << _selectedEqualityWeightCoefficient;
    return;
  case SELECTION:
    str << _selection;
    return;
  case SHOW_ACTIVE:
    str << boolToOnOff(_showActive);
    return;
  case SHOW_GENERATED:
    str << boolToOnOff(_showGenerated);
    return;
  case SHOW_INPUT:
    str << boolToOnOff(_showInput);
    return;
  case SHOW_OPTIONS:
    str << boolToOnOff(_showOptions);
    return;
  case SHOW_PASSIVE:
    str << boolToOnOff(_showPassive);
    return;
  case SHOW_PREPROCESSED:
    str << boolToOnOff(_showPreprocessed);
    return;
  case SHOW_PROFILE:
    str << boolToOnOff(_showProfile);
    return;
  case SHOW_WEIGHT_LIMIT_CHANGES:
    str << boolToOnOff(_showWeightLimitChanges);
    return;
  case SIDE_PREMISE_DEMODULATION:
    str << boolToOnOff(_sidePremiseDemodulation);
    return;
  case SIDE_PREMISE_SUBSUMPTION:
    str << boolToOnOff(_sidePremiseSubsumption);
    return;
  case SIMULATED_TIME_LIMIT:
    str << _simulatedTimeLimit/10;
    if (_simulatedTimeLimit % 10) {
      str << '.' << _simulatedTimeLimit % 10;
    }
    return;
  case SIMULTANEOUS_BACKWARD_SUPERPOSITION:
    str << boolToOnOff(_simultaneousBackwardSuperposition);
    return;
  case SOS:
    str << boolToOnOff(_sos);
    return;
  case SPLITTING:
    str << Constants::splittingValues[_splitting];
    return;
  case STATISTICS:
    str << Constants::statisticsValues[_statistics];
    return;
  case SUPERPOSITION_FROM_VARIABLES:
    str << boolToOnOff(_superpositionFromVariables);
    return;
  case SYMBOL_PRECEDENCE:
    str << Constants::symbolPrecedenceValues[_symbolPrecedence];
    return;
  case TEST_ID:
    str << _testId;
    return;
  case TEST_NUMBER:
    str << _testNumber;
    return;
  case TIME_LIMIT:
    str << _timeLimitInDeciseconds/10;
    if (_timeLimitInDeciseconds % 10) {
      str << '.' << _timeLimitInDeciseconds % 10;
    }
    return;
  case TRANSITIVITY_CHAINING:
    str << Constants::tcValues[_transitivityChaining];
    return;
  case UNUSED_PREDICATE_DEFINITION_REMOVAL:
    str << boolToOnOff(_unusedPredicateDefinitionRemoval);
    return;
  case WEIGHT_INCREMENT:
    str << boolToOnOff(_weightIncrement);
    return;
  case XML_OUTPUT:
    str << _xmlOutput;
    return;
#if DEBUG_SHELL
  default:
    cerr << "AAA " << optionTag << "\n";
    ASS(false);
#endif
  }
} 
string Options::problemName () const
{
  TRACER("Options::problemName");
  if (_problemName != "") {
    return _problemName;
  }
  string inputFileName = inputFile();
  if (inputFileName == "") {
    return "unknown";
  }
  int length = inputFileName.length() - 1;
  const char* name = inputFileName.c_str();
  int b = length - 1; 
  while (b >= 0 && name[b] != '/') {
    b--;
  }
  b++;
  int e = length-1; 
  while (e >= b && name[e] != '.') {
    e--;
  }
  if (e < b) {
    e = length-1;
  }
  return inputFileName.substr(b,e-b);
} 
XML::Element* Options::toXML () const
{
  TRACER("Options::toXML");
  XML::DeepElement* options = new XML::DeepElement("options");
  for (int i = 0;i < Constants::optionNames.length;i++) {
    ostringstream str;
    outputValue(str,i);
    XML::EmptyElement* option = new XML::EmptyElement("option");
    options->addChild(option);
    option->addAttribute(new XML::TextAttribute("name",
						Constants::optionNames[i]));
    option->addAttribute(new XML::TextAttribute("value",str.str()));
  }
  return options;
} 
void Options::readAgeWeightRatio(const char* val)
{
  TRACER("Options::readAgeWeightRatio");
  bool found = false;
  int colonIndex = 0;
  while (val[colonIndex]) {
    if (val[colonIndex] == ':') {
      found = true;
      break;
    }
    colonIndex++;
  }
  if (found) {
    if (strlen(val) > 127) {
      USER_ERROR((string)"wrong value for age-weight ratio: " + val);
    }
    char copy[128];
    strcpy(copy,val);
    copy[colonIndex] = 0;
    int age;
    if (! Int::stringToInt(copy,age)) {
      USER_ERROR((string)"wrong value for age-weight ratio: " + val);
    }
    _ageRatio = age;
    int weight;
    if (! Int::stringToInt(copy+colonIndex+1,weight)) {
      USER_ERROR((string)"wrong value for age-weight ratio: " + val);
    }
    _weightRatio = weight;
    return;
  }
  _ageRatio = 1;
  int weight;
  if (! Int::stringToInt(val,weight)) {
    USER_ERROR((string)"wrong value for age-weight ratio: " + val);
  }
  _weightRatio = weight;
} 
int Options::readTimeLimit(const char* val)
{
  TRACER("Options::readTimeLimit");
  int length = strlen(val);
  if (length == 0 || length > 127) {
    USER_ERROR((string)"wrong value for time limit: " + val);
  }
  char copy[128];
  strcpy(copy,val);
  char* end = copy;
  while (*end) {
    end++;
  }
  end--;
  float multiplier = 10.0;
  switch (*end) {
  case 's': 
    multiplier = 10.0;
    *end = 0;
    break;
  case 'm': 
    multiplier = 600.0;
    *end = 0;
    break;
  case 'h': 
    multiplier = 36000.0;
    *end = 0;
    break;
  case 'd': 
    multiplier = 864000.0;
    *end = 0;
    break;
  default:
    break;
  }
  float number;
  if (! Int::stringToFloat(copy,number)) {
    USER_ERROR((string)"wrong value for time limit: " + val);
  }
  #ifdef _MSC_VER
  return (int)floor(number * multiplier);
  #else
  return (int)round(number * multiplier);
  #endif
} 
void Options::readFromTestId (string testId)
{
  _normalize = true;
  _testId = testId;
  string ma(testId,0,3); 
  if (ma == "dis") {
    _mainAlg = DISCOUNT;
  }
  else if (ma == "lrs") {
    _mainAlg = LRS;
  }
  else if (ma == "ott") {
    _mainAlg = OTTER;
  }
  else {
  error: USER_ERROR("bad test id " + testId);
  }
  unsigned index = testId.find_last_of('_');
  if (index == string::npos) { 
    goto error;
  }
  string timeString = testId.substr(index+1);
  _timeLimitInDeciseconds = readTimeLimit(timeString.c_str()) / 10;
  testId = testId.substr(3,index-2);
  switch (testId[0]) {
  case '+':
    testId = testId.substr(1);
    break;
  case '-':
    break;
  default:
    goto error;
  }
  index = testId.find('_');
  int selection;
  string sel = testId.substr(0,index);
  Int::stringToInt(sel,selection);
  setSelection(selection);
  testId = testId.substr(index+1);
  if (testId == "") {
    goto error;
  }
  index = testId.find('_');
  string awr = testId.substr(0,index);
  readAgeWeightRatio(awr.c_str());
  testId = testId.substr(index+1);
  while (testId != "") {
    unsigned index1 = testId.find('=');
    if (index1 == string::npos) {
      goto error;
    }
    index = testId.find('_');
    if (index1 > index) {
      goto error;
    }
    string param = testId.substr(0,index1);
    string value = testId.substr(index1+1,index-index1-1);
    setShort(param.c_str(),value.c_str());
    testId = testId.substr(index+1);
  }
} 
bool Options::outputSuppressed () const
{
  return _xmlOutput == "on" ||
         _latexOutput == "on";
} 
}
