/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Options__
#define __Options__
#include "Allocator.hpp"
#include "XML.hpp"
namespace Shell {
class Options 
{
public:
  enum Tag {
    AGE_WEIGHT_RATIO = 0,
    ALLOCATION_BUFFER_PRICE = 1,
    ALLOCATION_BUFFER_SIZE = 2,
    BACKWARD_DEMODULATION = 3,
    BACKWARD_SUBSUMPTION = 4,
    BASIC_STRATEGY = 5,
    DECODE = 6,
    EQUALITY_PROXY = 7,
    EQUALITY_RESOLUTION_WITH_DELETION = 8,
    FAST_DEMODULATION_BY_COMMUTATIVITY = 9,
    FORWARD_DEMODULATION = 10,
    FORWARD_SUBSUMPTION = 11,
    FORWARD_SUBSUMPTION_RESOLUTION = 12,
    FUNCTION_DEFINITION_ELIMINATION = 13,
    INEQUALITY_SPLITTING = 14,
    INHERITED_NEGATIVE_SELECTION = 15,
    LATEX_OUTPUT = 16,
    LITERAL_COMPARISON_MODE = 17,
    LOG_FILE = 18,
    LRS_FIRST_TIME_CHECK = 19,
    MAIN_ALG = 20,
    MAIN_PREMISE_DEMODULATION = 21,
    MAIN_PREMISE_SUBSUMPTION = 22,
    MAX_ACTIVE = 23,
    MAX_ANSWERS = 24,
    MAX_INFERENCE_DEPTH = 25,
    MAX_PASSIVE = 26,
    MAX_SKOLEM = 27,
    MAX_WEIGHT = 28,
    MEMORY_LIMIT = 29,
    MINISCOPE = 30,
    NAMING = 31,
    NO_INFERENCES_BETWEEN_RULES = 32,
    NONGOAL_WEIGHT_COEFFICIENT = 33,
    NORMALIZE = 34,
    ORPHAN_DELETION = 35,
    OUTPUT_MESSAGES = 36,
    PROBLEM_NAME = 37,
    PROOF = 38,
    PROOF_CHECKING = 39,
    RANDOM_SEED = 40,
    ROW_VARIABLE_MAX_LENGTH = 41,
    SELECTED_EQUALITY_WEIGHT_COEFFICIENT = 42,
    SELECTION = 43,
    SHOW_ACTIVE = 44,
    SHOW_GENERATED = 45,
    SHOW_INPUT = 46,
    SHOW_OPTIONS = 47,
    SHOW_PASSIVE = 48,
    SHOW_PREPROCESSED = 49,
    SHOW_PROFILE = 50,
    SHOW_WEIGHT_LIMIT_CHANGES = 51,
    SIDE_PREMISE_DEMODULATION = 52,
    SIDE_PREMISE_SUBSUMPTION = 53,
    SIMULATED_TIME_LIMIT = 54,
    SIMULTANEOUS_BACKWARD_SUPERPOSITION = 55,
    SOS = 56,
    SPLITTING = 57,
    STATISTICS = 58,
    SUPERPOSITION_FROM_VARIABLES = 59,
    SYMBOL_PRECEDENCE = 60,
    TEST_ID = 61,
    TEST_NUMBER = 62,
    TIME_LIMIT = 63,
    TRANSITIVITY_CHAINING = 64,
    UNUSED_PREDICATE_DEFINITION_REMOVAL = 65,
    WEIGHT_INCREMENT = 66,
    XML_OUTPUT = 67,
    NUMBER_OF_OPTIONS = 68 
  };
public:
  class StringInt;
  class StringIntMap;
  enum FunctionDefinitionElimination {
    FDE_ALL = 0,
    FDE_NONE = 1,
    FDE_UNUSED = 2
  };
  enum Statistics {
    STATISTICS_BRIEF = 0,
    STATISTICS_FULL = 1,
    STATISTICS_NONE = 2
  };
   enum MainAlg {
     DISCOUNT = 0,
     LRS = 1,
     OTTER = 2
   };
  enum Splitting {
    SPLIT_INPUT_ONLY = 0,
    SPLIT_OFF = 1,
    SPLIT_ON = 2
  };
  enum TransitivityChaining {
    TC_FULL = 0,
    TC_NONE = 1,
    TC_ONLY_NONVARIABLES
  };
  enum LiteralComparisonMode {
    LCM_REVERSE = 0,
    LCM_PREDICATE = 1,
    LCM_STANDARD = 2
  };
  enum Demodulation {
    DEMODULATION_ALL = 0,
    DEMODULATION_OFF = 1,
    DEMODULATION_PREORDERED = 2
  };
  enum SymbolPrecedence {
    BY_ARITY = 0,
    BY_OCCURRENCE = 1,
    BY_REVERSE_ARITY = 2
  };
  enum Proof {
    PROOF_SUCCINCT = 0,
    PROOF_OFF = 1,
    PROOF_ON = 2
  };
  enum EqualityProxy {
    EP_EXP1 = 0,
    EP_OFF = 1,
    EP_ON = 2
  };
public:
  Options ();
  void output (ostream&) const;
  void readFromTestId (string testId);
  string problemName () const;
  string testId() const { return _testId; }
  int testNumber() const { return _testNumber; }
  Statistics statistics() const { return _statistics; }
  Proof proof() const { return _proof; }
  bool proofChecking() const { return _proofChecking; }
  int naming() const { return _naming; }
  bool setNaming(int newVal);
  bool miniscope() const { return _miniscope; }
  void setMiniscope(bool newVal) { _miniscope = newVal; }
  bool normalize() const { return _normalize; }
  string include() const { return _include; }
  string logFile() const { return _logFile; }
  string inputFile() const { return _inputFile; }
  int randomSeed() const { return _randomSeed; }
  int rowVariableMaxLength() const { return _rowVariableMaxLength; }
  void setRowVariableMaxLength(int newVal) { _rowVariableMaxLength = newVal; }
  bool outputMessages() const { return _outputMessages; }
  void setOutputMessages(bool newVal) { _outputMessages = newVal; }
  bool showGenerated() const { return _showGenerated; }
  bool showPassive() const { return _showPassive; }
  bool showActive() const { return _showActive; }
  bool showProfile() const { return _showProfile; }
  bool showPreprocessed() const { return _showPreprocessed; }
  bool showInput() const { return _showInput; }
  bool showOptions() const { return _showOptions; }
  bool showWeightLimitChanges() const { return _showWeightLimitChanges; }
  bool unusedPredicateDefinitionRemoval() const { return _unusedPredicateDefinitionRemoval; }
  void setUnusedPredicateDefinitionRemoval(bool newVal) { _unusedPredicateDefinitionRemoval = newVal; }
  bool weightIncrement() const { return _weightIncrement; }
  MainAlg mainAlg() const { return _mainAlg; }
  Splitting splitting() const { return _splitting; }
  int selection() const { return _selection; }
  bool setSelection(int newValue);
  string latexOutput() const { return _latexOutput; }
  LiteralComparisonMode literalComparisonMode() const { return _literalComparisonMode; }
  bool inheritedNegativeSelection() const { return _inheritedNegativeSelection; }
  bool forwardSubsumptionResolution() const { return _forwardSubsumptionResolution; }
  void setForwardSubsumptionResolution(bool newVal) { _forwardSubsumptionResolution = newVal; }
  Demodulation forwardDemodulation() const { return _forwardDemodulation; }
  Demodulation backwardDemodulation() const { return _backwardDemodulation; }
  void setBackwardDemodulation(Demodulation newVal) { _backwardDemodulation = newVal; }
  bool basicStrategy() const { return _basicStrategy; }
  bool backwardSubsumption() const { return _backwardSubsumption; }
  void setBackwardSubsumption(bool newVal) { _backwardSubsumption = newVal; }
  bool forwardSubsumption() const { return _forwardSubsumption; }
  bool orphanDeletion() const { return _orphanDeletion; }
  int lrsFirstTimeCheck() const { return _lrsFirstTimeCheck; }
  bool setLrsFirstTimeCheck(int newVal);
  int simulatedTimeLimit() const { return _simulatedTimeLimit; }
  void setSimulatedTimeLimit(int newVal) { _simulatedTimeLimit = newVal; }
  int maxInferenceDepth() const { return _maxInferenceDepth; }
  SymbolPrecedence symbolPrecedence() const { return _symbolPrecedence; }
  bool noInferencesBetweenRules() const { return _noInferencesBetweenRules; }
  int timeLimitInDeciseconds() const { return _timeLimitInDeciseconds; }
  static int readTimeLimit(const char* val);
  int memoryLimit() const { return _memoryLimit; }
  int allocationBufferSize() const { return _allocationBufferSize; }
  float allocationBufferPrice() const { return _allocationBufferPrice; }
  int inequalitySplitting() const { return _inequalitySplitting; }
  int maxSkolem() const { return _maxSkolem; }
  long maxActive() const { return _maxActive; }
  long maxAnswers() const { return _maxAnswers; }
  void setMaxAnswers(int newVal) { _maxAnswers = newVal; }
  long maxPassive() const { return _maxPassive; }
  int maxWeight() const { return _maxWeight; }
  int ageRatio() const { return _ageRatio; }
  int weightRatio() const { return _weightRatio; }
  bool superpositionFromVariables() const { return _superpositionFromVariables; }
  EqualityProxy equalityProxy() const { return _equalityProxy; }
  bool equalityResolutionWithDeletion() const { return _equalityResolutionWithDeletion; }
  float nongoalWeightCoefficient() const { return _nongoalWeightCoefficient; }
  bool setNongoalWeightCoefficient(float newVal);
  float selectedEqualityWeightCoefficient() const { return _selectedEqualityWeightCoefficient; }
  bool setSelectedEqualityWeightCoefficient(float newValue);
  bool sos() const { return _sos; }
  void setSos(bool newVal) { _sos = newVal; }
  bool fastDemodulationByCommutativity() const { return _fastDemodulationByCommutativity; }
  FunctionDefinitionElimination functionDefinitionElimination() const { return _functionDefinitionElimination; }
  bool mainPremiseDemodulation() const { return _mainPremiseDemodulation != 0; }
  bool mainPremiseSubsumption() const { return _mainPremiseSubsumption != 0; }
  bool sidePremiseDemodulation() const { return _sidePremiseDemodulation !=0; }
  bool sidePremiseSubsumption() const { return _sidePremiseSubsumption != 0; }
  bool simultaneousBackwardSuperposition() const { return _simultaneousBackwardSuperposition; }
  TransitivityChaining transitivityChaining() const { return _transitivityChaining; }
  string xmlOutput() const { return _xmlOutput; }
  void setMemoryLimit(int newVal) { _memoryLimit = newVal; }
  void setInputFile(const string& newVal) { _inputFile = newVal; }
  void setTimeLimitInSeconds(int newVal) { _timeLimitInDeciseconds = 10*newVal; }
  void setTimeLimitInDeciseconds(int newVal) { _timeLimitInDeciseconds = newVal; }
  XML::Element* toXML() const;
  bool outputSuppressed() const;
  void set(const string& name, const string& value);
  void set(const char* name, const char* value);
  void setShort(const char* name, const char* value);
private:
  void set(const char* name, const char* value, int index);
private:
  class Constants;
  int _ageRatio;
  int _weightRatio;
  float _allocationBufferPrice;
  int _allocationBufferSize;
  Demodulation _backwardDemodulation;
  bool _backwardSubsumption;
  bool _basicStrategy;
  EqualityProxy _equalityProxy;
  bool _equalityResolutionWithDeletion;
  bool _fastDemodulationByCommutativity;
  Demodulation _forwardDemodulation;
  bool _forwardSubsumption;
  bool _forwardSubsumptionResolution;
  FunctionDefinitionElimination _functionDefinitionElimination;
  string _include;
  int _inequalitySplitting;
  bool _inheritedNegativeSelection;
  string _inputFile;
  string _latexOutput;
  LiteralComparisonMode _literalComparisonMode;
  string _logFile;
  int _lrsFirstTimeCheck;
  MainAlg _mainAlg;
  int _mainPremiseDemodulation;
  int _mainPremiseSubsumption;
  long _maxActive;
  int _maxAnswers;
  int _maxInferenceDepth;
  long _maxPassive;
  int _maxSkolem;
  int _maxWeight;
  int _memoryLimit;
  bool _miniscope;
  int _naming;
  bool _noInferencesBetweenRules;
  float _nongoalWeightCoefficient;
  bool _normalize;
  bool _orphanDeletion;
  bool _outputMessages;
  string _problemName;
  Proof _proof;
  bool _proofChecking;
  int _randomSeed;
  int _rowVariableMaxLength;
  float _selectedEqualityWeightCoefficient;
  int _selection;
  bool _showActive;
  bool _showGenerated;  
  bool _showInput;
  bool _showOptions;
  bool _showPassive;  
  bool _showPreprocessed;
  bool _showProfile;
  bool _showWeightLimitChanges;  
  int _sidePremiseDemodulation;
  int _sidePremiseSubsumption;
  int _simulatedTimeLimit;
  bool _simultaneousBackwardSuperposition;
  bool _sos;
  Splitting _splitting;
  Statistics _statistics;
  bool _superpositionFromVariables;
  SymbolPrecedence _symbolPrecedence;
  string _testId;
  int _testNumber;
  int _timeLimitInDeciseconds;
  TransitivityChaining _transitivityChaining;
  bool _unusedPredicateDefinitionRemoval;
  bool _weightIncrement;
  string _xmlOutput;
  void readAgeWeightRatio(const char* val);
  static string boolToOnOff(bool);
  void outputValue(ostream& str,int optionTag) const;
public:
  static bool onOffToBool(const char* onOff,const char* option);
}; 
}
#endif
