/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Int.hpp"
#include "IntNameTable.hpp"
#include "Lexer.hpp"
#include "Parser.hpp"
#include "Signature.hpp"
#include "Tracer.hpp"
#include "Unit.hpp"
using namespace Shell;
Parser::Parser (Lexer& lexer)
  :
  _lexer(lexer),
  _tokenCursor(0),
  _noOfTokens(0)
{
} 
ParserException::ParserException (string message,const Token& token)
  : _message (message)
{
  _message += " in line ";
  _message += Int::toString(token.line);
  _message += " at ";
  _message += token.text;
} 
void ParserException::cry (ostream& out)
{
  out << "Parser exception: " << _message << '\n';
} 
void Parser::consumeToken (TokenType tt)
{
  TRACER("Parser::consumeToken/1");
  if (currentToken().tag != tt) {
    throw ParserException(Token::toString(tt) + " expected",
			  _tokens[_tokenCursor]);
  }
  consumeToken();
} 
void Parser::consumeToken ()
{
  TRACER("Parser::consumeToken/0");
  ASS (_noOfTokens > 0);
  ASS (_tokenCursor < TOKEN_BUFFER_SIZE);
  _tokenCursor = (_tokenCursor+1) % TOKEN_BUFFER_SIZE;
  _noOfTokens--;
} 
void Parser::readToken ()
{
  TRACER("Parser::readToken");
  ASS(_noOfTokens <= TOKEN_BUFFER_SIZE);
  if (_noOfTokens == TOKEN_BUFFER_SIZE) {
    _lexer.readToken(_tokens[0]);
    throw ParserException("token buffer capacity exceeded",_tokens[0]);
  }
  int tokenIndex = (_tokenCursor + _noOfTokens) % TOKEN_BUFFER_SIZE;
  _noOfTokens++;
  _lexer.readToken(_tokens[tokenIndex]);
} 
Token& Parser::lookAhead (int index)
{
  TRACER("Parser::lookahead");
  ASS (index < TOKEN_BUFFER_SIZE);
  while (_noOfTokens <= index) {
    readToken();
  }
  return _tokens[(_tokenCursor+index) % TOKEN_BUFFER_SIZE];
} 
int Parser::var (const Token& t)
{
  return IntNameTable::vars.insert(t.text);
} 
double Parser::real (const Token& token)
{
  double d;
  if (! Int::stringToDouble(token.text.c_str(),d)) {
    throw ParserException("incorrect floating point number",token);
  }
  return d;
} 
int Parser::integer (const Token& token)
{
  int result;
  if (Int::stringToInt(const_cast<string&>(token.text),result)) {
    return result;
  }
  throw ParserException("incorrect integer",token);
} 
