/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Inference.hpp"
#include "ShellReader.hpp"
#include "Unit.hpp"
#include "VampireException.hpp"
using namespace Shell;
Vampire::TermTag Reader::tag (const Vampire::Term* t) const
{
  switch (reinterpret_cast<const Shell::Term*>(t)->tag()) {
  case VAR:
    return Vampire::VAR;
  case COMPOUND:
    return Vampire::COMPOUND;
  case INTERPRETED:
    return Vampire::INTERPRETED;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
int Reader::var (const Vampire::Term* t) const
{
  return reinterpret_cast<const Term*>(t)->var().toInt();
}
const Vampire::FunctionSymbol* Reader::functor (const Vampire::Term* t) const
{
  const FunctionSymbol* f =  &reinterpret_cast<const Term*>(t)->functor();
  return reinterpret_cast<const Vampire::FunctionSymbol*>(f);
} 
string Reader::name (const Vampire::FunctionSymbol* f) const
{
  return reinterpret_cast<const FunctionSymbol*>(f)->name();
}
const Vampire::TermList* Reader::args (const Vampire::Term* t) const
{
  const TermList* ts =  &reinterpret_cast<const Term*>(t)->args();
  return reinterpret_cast<const Vampire::TermList*>(ts);
}
const VampireKernelEnvironment::InterpretedConstant* Reader::value (const Vampire::Term* t) const
{
  return &reinterpret_cast<const Term*>(t)->value();
}
bool Reader::isEmpty (const Vampire::TermList* ts,
		      const Vampire::TermList*) const
{
  return reinterpret_cast<const TermList*>(ts)->isEmpty();
}
const Vampire::Term* Reader::head (const Vampire::TermList* ts) const
{
  const Term* t = &reinterpret_cast<const TermList*>(ts)->head();
  return reinterpret_cast<const Vampire::Term*>(t);
}
const Vampire::TermList* Reader::tail (const Vampire::TermList* ts,
				       const Vampire::TermList*) const
{
  const TermList* tl = &reinterpret_cast<const TermList*>(ts)->tail();
  return reinterpret_cast<const Vampire::TermList*>(tl);
}
bool Reader::isEquality (const Vampire::Atom* a) const
{
  return reinterpret_cast<const Atom*>(a)->isEquality();
}
const Vampire::TermList* Reader::args (const Vampire::Atom* a) const
{
  const TermList* ts =  &reinterpret_cast<const Atom*>(a)->args();
  return reinterpret_cast<const Vampire::TermList*>(ts);
}
const Vampire::PredicateSymbol* Reader::functor (const Vampire::Atom* a) const
{
  const PredicateSymbol* p =  &reinterpret_cast<const Atom*>(a)->functor();
  return reinterpret_cast<const Vampire::PredicateSymbol*>(p);
}
string Reader::name (const Vampire::PredicateSymbol* p) const
{
  return reinterpret_cast<const PredicateSymbol*>(p)->name();
}
bool Reader::isPositive (const Vampire::Literal* l) const
{
  return reinterpret_cast<const Literal*>(l)->isPositive();
}
const Vampire::Atom* Reader::atom (const Vampire::Literal* l) const
{
  const Atom* a = &reinterpret_cast<const Literal*>(l)->atom();
  return reinterpret_cast<const Vampire::Atom*>(a);
}
bool Reader::isEmpty (const Vampire::LiteralList* ls,
		      const Vampire::LiteralList*) const
{
  return reinterpret_cast<const LiteralList*>(ls)->isEmpty();
}
const Vampire::Literal* Reader::head (const Vampire::LiteralList* ls) const
{
  const Literal* l = &reinterpret_cast<const LiteralList*>(ls)->head();
  return reinterpret_cast<const Vampire::Literal*>(l);
}
const Vampire::LiteralList* Reader::tail (const Vampire::LiteralList* ls,
					  const Vampire::LiteralList*) const
{
  const LiteralList* tl = &reinterpret_cast<const LiteralList*>(ls)->tail();
  return reinterpret_cast<const Vampire::LiteralList*>(tl);
}
const Vampire::LiteralList* Reader::literals (const Vampire::Clause* c) const
{
  const LiteralList* ls = &reinterpret_cast<const Clause*>(c)->literals();
  return reinterpret_cast<const Vampire::LiteralList*>(ls);
}
int Reader::number (const Vampire::Unit* u) const
{
  return reinterpret_cast<const Unit*>(u)->number();
}
const Vampire::Clause* Reader::clause (const Vampire::Unit* u) const
{
  const Clause* c = &reinterpret_cast<const Unit*>(u)->clause();
  return reinterpret_cast<const Vampire::Clause*>(c);
}
const Vampire::Formula* Reader::formula (const Vampire::Unit* u) const
{
  const Formula* f = &reinterpret_cast<const Unit*>(u)->formula();
  return reinterpret_cast<const Vampire::Formula*>(f);
}
const Vampire::UnitList* Reader::premises (const Vampire::Unit* u) const
{
  const UnitList* us = &reinterpret_cast<const Unit*>(u)->inference()->premises();
  return reinterpret_cast<const Vampire::UnitList*>(us);
}
bool Reader::isEmpty (const Vampire::UnitList* ts,
		      const Vampire::UnitList*) const
{
  return reinterpret_cast<const UnitList*>(ts)->isEmpty();
}
const Vampire::Unit* Reader::head (const Vampire::UnitList* ts) const
{
  const Unit* t = &reinterpret_cast<const UnitList*>(ts)->head();
  return reinterpret_cast<const Vampire::Unit*>(t);
}
const Vampire::UnitList* Reader::tail (const Vampire::UnitList* ts,
				       const Vampire::UnitList*) const
{
  const UnitList* tl = &reinterpret_cast<const UnitList*>(ts)->tail();
  return reinterpret_cast<const Vampire::UnitList*>(tl);
}
void Reader::saveBackground (const Vampire::Unit* u,
			     Vampire::Background& b) const
{
  const Unit& unit = *reinterpret_cast<const Unit*>(u);
  b.isClause = unit.unitType() == CLAUSE ? 1 : 0;
  b.storageKind = Vampire::Background::SHELL;
  b.derivedFromAxioms = unit.inputType() == AXIOM ? 1 : 0;
} 
