/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Substitution__
#define __Substitution__
#include "Term.hpp"
#include "List.hpp"
namespace Shell {
class Atom;
class Formula;
class FormulaList;
class Substitution
{
public:
  Substitution ();
  ~Substitution ();
  void bind (Var v, const Term& t);
  bool isBound (Var v, Term& t) const;
  bool isBound (Var v) const;
  void undoBinding (Var v);
  void getDomain (VarList& vs) const;
  void getRange (TermList& ts) const;
  bool isInRange (const Term& t) const;
  void apply (Term&) const;
  void apply (TermList&) const;
  void apply (Atom&) const;
  void apply (Formula&);
  void apply (FormulaList&);
  Var freshVariable () { return (Var)++_lastNewVar; }
  void setLastFreshVariable (int var) { _lastNewVar = var; }
  XML::Element* toXML () const;
protected:
  typedef List<Term> Binding;
  int _top;
  Binding** _bindings;
  int _lastNewVar;
  void extend (int minimalNewCapacity);
}; 
}
#endif 
