/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "SubtermIterator.hpp"
#include "Tracer.hpp"
#include "Literal.hpp"
using namespace Shell;
SubtermIterator::SubtermIterator (const Term& t)
  : _current(const_cast<Term*>(&t)),
    _reserve(0)
{
} 
SubtermIterator::SubtermIterator (const TermList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 
SubtermIterator::SubtermIterator (const Atom& a)
  : _current(0),
    _reserve(new Element(a.args(),0))
{
} 
SubtermIterator::SubtermIterator (const Literal& l)
  : _current(0),
    _reserve(new Element(l.atom().args(),0))
{
} 
bool SubtermIterator::hasNext ()
{
  TRACER("SubtermIterator::hasNext");
  if (_current) {
    return true;
  }
  while (_reserve) {
    TermList* first = _reserve->_list;
    if (first->isEmpty()) {
      Element* rest = _reserve->_rest;
      delete _reserve;
      _reserve = rest;
    }
    else { 
      _current = const_cast<Term*>(&(first->head()));
      _reserve->_list = const_cast<TermList*>(&(first->tail()));
      return true;
    }
  }
  return false;
} 
const Term& SubtermIterator::next ()
{
  Term* result = _current;
  if (_current->tag() == COMPOUND) {
    _reserve = new Element(result->args(),_reserve);
  }
  _current = 0;
  return *result;
} 
SubtermIterator::~SubtermIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
