/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Symbol.hpp"
#include "SymCounter.hpp"
#include "XML.hpp"
using namespace Shell;
XML::Element* Symbol::toXML() const
{
  XML::DeepElement* symbol = new XML::DeepElement(isFunction() ? 
						  "function_symbol" :
						  "predicate_symbol");
  symbol->addAttribute(new XML::TextAttribute("name",_name));
  symbol->addAttribute(new XML::IntegerAttribute("arity",_arity));
  symbol->addAttribute(new XML::IntegerAttribute("number",_number));
  symbol->addChild(new XML::IntegerElement("weight",_weight));
  symbol->addChild(new XML::IntegerElement("precedence",_precedence));
  if (! isFunction()) {
    const PredicateSymbol* pred = static_cast<const PredicateSymbol*>(this);
    symbol->addChild(new XML::IntegerElement("level",
					     pred->level()));
    symbol->addChild(new XML::IntegerElement("negative_level",
					     pred->negativeLevel()));
  }
  return symbol;
} 
Symbol::Symbol (const string& name, int arity, int& number) 
  : _name (name),
    _arity (arity),
    _number (number++),
    _kernelNumber (0),
    _weight(1),
    _precedence(0),
    _isEqualityProxy(0),
    _isSkolem(0),
    _isEvaluable(0),
    _isName(0),
    _isAnswer(0)
{
}
ostream& operator << (ostream& str, const Symbol* p)
{
  str << p->name ();
  return str;
} 
FunctionSymbol::FunctionSymbol (const string& name, int arity, int& count) 
  : Symbol (name,arity,count)
{
  _isFunction = 1;
}
PredicateSymbol::PredicateSymbol (const string& name, int arity, int& count) 
  : Symbol (name,arity,count),
    _level(0),
    _negativeLevel(0)
{
  _isFunction = 0;
} 
