/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <sstream>
#include "NameArray.hpp"
#include "Syntax.hpp"
#include "TPTPLexer.hpp"
#include "TPTPParser.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
namespace Shell {
class Syntax::Constants {
public:
  static NameArray inputSyntaxes;
  static NameArray outputSyntaxes;
private:
  static const char* is[];
  static const char* os[];
};
const char* Syntax::Constants::is[] = {
  "kif",
  "tptp"};
NameArray Syntax::Constants::inputSyntaxes(is,
					   sizeof(is)/sizeof(char*));
const char* Syntax::Constants::os[] = {
  "kif",
  "tptp",
  "vampire"};
NameArray Syntax::Constants::outputSyntaxes(os,
					    sizeof(os)/sizeof(char*));
void Syntax::readUnits (const string& text,UnitList& units,InputSyntax syntax)
{
  TRACER("Syntax::readUnits (const string&...)");
  istringstream in(text);
  return readUnits(in,units,syntax);
}
void Syntax::readUnits (istream& in,UnitList& units,InputSyntax syntax)
{
  TRACER("Syntax::readUnits (istream&...)");
  switch (syntax) {
  case IS_TPTP: 
    {
      TPTPLexer lexer(in); 
      TPTPParser parser (lexer);
      parser.units(units);
      break;
    }
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
 }
InputSyntax Syntax::stringToInputSyntax (const string& syn)
{
  TRACER("Syntax::stringToInputSyntax");
  int index = Constants::inputSyntaxes.find(syn.c_str());
  if (index == -1) {
    USER_ERROR((string)"unknown name for input syntax " + syn);
  }
  return (InputSyntax)index;
} 
OutputSyntax Syntax::stringToOutputSyntax (const string& syn)
{
  TRACER("Syntax::stringToOutputSyntax");
  int index = Constants::outputSyntaxes.find(syn.c_str());
  if (index == -1) {
    USER_ERROR((string)"unknown name for output syntax " + syn);
  }
  return (OutputSyntax)index;
} 
void Syntax::readFormula (const string& text,
			  Formula& formula,
			  InputSyntax syntax)
{
  TRACER("Syntax::readFormula");
  istringstream in(text);
  switch (syntax) {
  case IS_TPTP: 
    {
      TPTPLexer lexer(in); 
      TPTPParser parser (lexer);
      parser.formula(formula);
      break;
    }
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
} 
}
