/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __TheoryFinder__
#define __TheoryFinder__
class Symbol;
class FunctionSymbol;
class PredicateSymbol;
class Problem;
class Property;
#include "Array.hpp"
#include "Signature.hpp"
#include "Unit.hpp"
namespace Shell {
class TheoryFinder {
public:
  TheoryFinder(const Problem&,Property* property);
  ~TheoryFinder();
  int search();
private:
  class MatcherState;
  class Matcher;
  class AtomMatcher;
  class LiteralListMatcher;
  class ClauseMatcher;
  class QuantifiedFormulaMatcher;
  class FormulaPairMatcher;
  class FormulaListMatcher;
  void addTheory();
  bool matchAll(const Unit& unit);
  bool matchAll(const LiteralList&);
  bool matchAll(const Formula&);
  bool matchAll(const Atom&);
  bool matchCode(const Atom&,const char* code,
		 const char* name,int arity,int prop);
  bool matchCode(const LiteralList& c,const char* code,
		 const char* name,int arity,int prop);
  bool matchCode(const Formula& f,const char* code,
		 const char* name,int arity,int prop);
  bool matchC(const Atom&);
  bool matchA(const Atom&);
  bool matchLeftInverse(const Atom&);
  bool matchLeftIdentity(const Atom&);
  bool matchRightInverse(const Atom&);
  bool matchRightIdentity(const Atom&);
  bool matchLeftDistributivity(const Atom&);
  bool matchRightDistributivity(const Atom&);
  bool matchRobbins(const Atom&);
  bool matchIdempotence(const Atom&);
  bool matchAssociator(const Atom&);
  bool matchCommutator(const Atom&);
  bool matchAlternative(const Atom&);
  bool matchAbsorption(const Atom&);
  bool matchCombinatorS(const Atom&);
  bool matchCombinatorB(const Atom&);
  bool matchCombinatorT(const Atom&);
  bool matchCombinatorO(const Atom&);
  bool matchCombinatorQ(const Atom&);
  bool matchExtensionality(const Formula&);
  bool matchExtensionality(const LiteralList&);
  bool matchCondensedDetachment1(const LiteralList&);
  bool matchCondensedDetachment2(const LiteralList&);
  bool matchFLD1(const LiteralList&);
  bool matchFLD2(const LiteralList&);
  bool matchSubset(const Formula&);
  bool matchSubset(const LiteralList&);
  void stripForall(Formula& f);
  void ass(const char* name,
	   int arity,
	   const Symbol** theorySymbols,
	   Unit& result);
  void analyse(const Clause&);
  enum Code {
    NEWVAR = 0,
    OLDVAR = 1,
    NEWSYM = 2,
    OLDSYM = 3,
    CAND = 4,
    COR = 5,
    CIMP = 6,
    CNOT = 7,
    CIFF = 8,
    CXOR = 9,
    CFORALL = 10,
    CEXISTS = 11,
    EQL = 12
  };
  const Problem& _problem;
  Unit* _currentUnit;
  Signature _sig;
  Property* _property;
  UnitList _metaTheory;
};
}
#endif
