/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#if DEBUG_SHELL
#include <climits>
#include "Tracer.hpp"
#include "XML.hpp"
#include <iostream>
#define CP_FIRST_POINT INT_MAX
#define CP_LAST_POINT -1
using namespace Shell;
MyTracer::MyTracer (const string& fun) 
  :
  _fun (fun),
  _previous (_current)
{
  _current = this;
  controlPoint(_fun,CP_ENTRY);
} 
MyTracer::~MyTracer () 
{
  _current = _previous;
  controlPoint(_fun,CP_EXIT);
} 
void MyTracer::controlPoint (const string& description, ControlPointKind cpk)
{
  _lastPointKind = cpk;
  _passedControlPoints++;
  _lastControlPoint = description;
  if (_passedControlPoints <= CP_LAST_POINT &&
      _passedControlPoints >= CP_FIRST_POINT) {
    outputLastControlPoint(cout);
  }
} 
void MyTracer::controlPoint (const string& description)
{
  controlPoint (description,CP_MID);
} 
void MyTracer::outputLastControlPoint (ostream& str)
{
  str << _passedControlPoints 
      << ": " 
      << (_lastPointKind == CP_ENTRY ? '+' : 
          _lastPointKind == CP_EXIT ? '-' : ' ')
      << _lastControlPoint 
      << "\n";
} 
void MyTracer::printStack (ostream& str)
{
  int depth = 0;
  str << "Control points passed: " << _passedControlPoints << "\n"
      << "last control point:\n";
  outputLastControlPoint(str);
  _current->printStack (str, depth);
} 
XML::Element* MyTracer::stackToXML ()
{
  XML::DeepElement* xml = new XML::DeepElement("tracer_stack");
  for (MyTracer* s = _current; s; s = s->_previous) {
    xml->addChild(new XML::TextElement("function",s->_fun));
  }
  return xml;
} 
void MyTracer::printStack (ostream& str, int& depth)
{
  if (! this) { 
    return;
  }
  _previous->printStack (str, depth);
  for (int i = depth; i > 0; i--) {
    str << ' ';
  }
  str << _fun << "\n";
  depth ++;
} 
#endif 
