/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __VampireException__
#define __VampireException__
#include <string>
#include <iostream>
#include "RunResult.hpp"
#include "XML.hpp"
using namespace std;
namespace Shell {
class VampireException 
{
 public:
  VampireException () {}
  VampireException (const string& file, int line);
  virtual XML::Element* toXML () const = 0;
  virtual void cry (ostream&) = 0;
  virtual ~VampireException () {}
 protected:
  void outputFileAndLine (ostream&) const;
  void addFileAndLine (XML::DeepElement&) const;
  string _file;
  int _line;
}; 
class AllocationFailureException 
  : public VampireException
{
 public:                                
  AllocationFailureException (const string& file, int line);
  ~AllocationFailureException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
}; 
class AssertionViolationException 
  : public VampireException
{
public:                                
  AssertionViolationException (const string& file, int line);
  ~AssertionViolationException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
#if DEBUG_SHELL
private:
  XML::Element* _stack;
#endif
}; 
class UserErrorException 
  : public VampireException
{
 public:                                
  UserErrorException (const string& _message);
  ~UserErrorException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
 private:                               
  string _message;
}; 
class VampireTerminationException 
  : public VampireException
{
 public:                                
  VampireTerminationException (const string& file, 
			       int line, 
			       string reason,
			       RunResult result);
  ~VampireTerminationException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
  RunResult getResult () const { return _result; }
  const string& getReason () const { return _reason; }
 private:                               
  string _reason;
  RunResult _result;
}; 
}
#define VAMPIRE_EXCEPTION \
  throw Shell::VampireException(__FILE__,__LINE__)
#define ALLOCATION_FAILURE \
  throw Shell::AllocationFailureException(__FILE__,__LINE__)
#define ASSERTION_VIOLATION \
  throw Shell::AssertionViolationException(__FILE__,__LINE__)
#define VAMPIRE_TERMINATION(reason,result) \
  throw Shell::VampireTerminationException(__FILE__,__LINE__,reason,result)
#define USER_ERROR(msg) \
  throw Shell::UserErrorException(msg)
#endif 
