/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BSIndex.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX
#  define DEBUG_NAMESPACE "VK::BSIndex"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
BSIndex::BSIndex() : 
  _maxlNumOfNonproplLits(0),
  _propLitOccurences(static_cast<CSkipList*>(0),"BSIndex::_propLitOccurences")
{
  CALL("constructor BSIndex()");
  DOP(_propLitOccurences.freeze());
}; 
BSIndex::~BSIndex() 
{ 
  CALL("destructor ~BSIndex()"); 
  if (DestructionMode::isThorough()) 
    {
      for (ulong i = 0; i < _propLitOccurences.size(); i++) 
	if (_propLitOccurences[i])
	  delete (_propLitOccurences[i]);
      bool constant;
      OrdLitPITreeDestruction _ordLitIndDestruction;
      if (_ordLitIndDestruction.reset(_ordLitInd[0],constant))
	{ 
	  do 
	    {
	      if (constant) { recycle(_ordLitIndDestruction.constIndObj()); }
	      else recycle(_ordLitIndDestruction.indexedObj());
	    }
	  while (_ordLitIndDestruction.next(constant));
	  _ordLitIndDestruction.recycle(); 
	};
      SymLitPITreeDestruction _symLitIndDestruction;
      if (_symLitIndDestruction.reset(_symLitInd[0],constant))
	{ 
	  do 
	    {
	      if (constant) 
		{ 
		  recycle(_symLitIndDestruction.constIndObj()); 
		  ASSERT(!_symLitIndDestruction.constIndObj());
		}
	      else 
		{
		  recycle(_symLitIndDestruction.indexedObj());
		  ASSERT(!_symLitIndDestruction.indexedObj());
		};
	    }
	  while (_symLitIndDestruction.next(constant));
	  _symLitIndDestruction.recycle(); 
	};
    };
}; 
void BSIndex::reset()
{
  CALL("reset()");
  _maxlNumOfNonproplLits = 0UL;
  for (ulong i = 0; i < _propLitOccurences.size(); i++) 
    if (_propLitOccurences[i])
      {
	delete (_propLitOccurences[i]);
	_propLitOccurences[i] = static_cast<CSkipList*>(0);
      };
  DOP(_propLitOccurences.freeze());
  bool constant;
  OrdLitPITreeDestruction _ordLitIndDestruction;
  if (_ordLitIndDestruction.reset(_ordLitInd[0],constant))
    { 
      do 
	{
	  if (constant) { recycle(_ordLitIndDestruction.constIndObj()); }
	  else recycle(_ordLitIndDestruction.indexedObj());
	}
      while (_ordLitIndDestruction.next(constant));
      _ordLitIndDestruction.recycle(); 
    };
  SymLitPITreeDestruction _symLitIndDestruction;
  if (_symLitIndDestruction.reset(_symLitInd[0],constant))
    { 
      do 
	{
	  if (constant) 
	    { 
	      recycle(_symLitIndDestruction.constIndObj()); 
	      ASSERT(!_symLitIndDestruction.constIndObj());
	    }
	  else 
	    {
	      recycle(_symLitIndDestruction.indexedObj());
	      ASSERT(!_symLitIndDestruction.indexedObj());
	    };
	}
      while (_symLitIndDestruction.next(constant));
      _symLitIndDestruction.recycle(); 
    };
}; 
inline
void BSIndex::insertPropLit(const Term& lit,ulong clauseNum)
{
  CALL("insertPropLit(const Term& lit,ulong clauseNum)");
  DOP(_propLitOccurences.unfreeze());
  CSkipList* occList = _propLitOccurences.sub(lit.HeaderNum());
  DOP(_propLitOccurences.freeze());
  if (!occList) 
    {
      occList = new CSkipList();
      _propLitOccurences[lit.HeaderNum()] = occList;
    };
  occList->insert(clauseNum);
};
inline
void BSIndex::removePropLit(const Term& lit,ulong clauseNum)
{
  CALL("removePropLit(const Term& lit,ulong clauseNum)");
  if (_propLitOccurences[lit.HeaderNum()])
    _propLitOccurences[lit.HeaderNum()]->remove(clauseNum);
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX
#define DEBUG_NAMESPACE "BSIndex::Integrator"
#endif
#include "debugMacros.hpp"
void BSIndex::Integrator::propLit(const Term& lit)
{
  CALL("propLit(const Term& lit)");
  _index->insertPropLit(lit,_clauseNum);
};
void BSIndex::Integrator::subterm(const Term& t)
{
  CALL("subterm(const Term& t)");
  bool newNode;
  CLTSkipList::Node* node = _ordLitIntegrator.indexedObj().insert(_clauseNum,newNode);  
  if (newNode)
    {
      node->value() = new LTList(LTPair(_litNum,t)); 
    }
  else
    {
      ASSERT(node->value());
      ASSERT(_litNum > node->value()->hd().first);
      node->value() = new LTList(LTPair(_litNum,t),node->value()); 
    };
  if (t.isVariable())
    {
      _ordLitIntegrator.variable();
    } 
  else 
    {
      ASSERT(t.IsReference());
      Term sym = *(t.deref());
      if (sym.arity())
	{
	  _ordLitIntegrator.nonconstFunc(sym);
	}
      else 
	{
	  _ordLitIntegrator.constant(sym);
	  CLSkipList::Node* cnode = _ordLitIntegrator.constIndObj().insert(_clauseNum,newNode); 
	  if (newNode) 
	    { 
	      cnode->value() = new LList(_litNum);
	    }    
	  else 
	    {
	      ASSERT(cnode->value());
	      ASSERT(_litNum > cnode->value()->hd());
	      cnode->value() = new LList(_litNum,cnode->value());
	    };
	};
    };
}; 
void BSIndex::Integrator::symLitSubterm(const Term& t)
{
  CALL("symLitSubterm(const Term& t)");
  bool newNode;
  CLATSkipList::Node* node = 
    _symLitIntegrator.indexedObj().insert(_clauseNum,newNode);
  ASSERT(newNode || LATTriple(_litNum,_symLitArgNum,t) > node->value()->hd());
  if (newNode) node->value() = 0;  
  node->value() = new LATList(LATTriple(_litNum,_symLitArgNum,t),node->value());
  if (t.isVariable())
    {   
      _symLitIntegrator.variable();  
    } 
  else 
    {    
      ASSERT(t.IsReference()); 
      Term sym = *(t.deref());
      if (sym.arity())
	{  
	  _symLitIntegrator.nonconstFunc(sym);
	}
      else 
	{  
	  _symLitIntegrator.constant(sym); 
	  CLASkipList::Node* cnode = _symLitIntegrator.constIndObj().insert(_clauseNum,newNode);
	  ASSERT(newNode || LAPair(_litNum,_symLitArgNum) > cnode->value()->hd());
	  if (newNode) cnode->value() = 0; 
	  cnode->value() = new LAList(LAPair(_litNum,_symLitArgNum),cnode->value());
	};
    };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX
#define DEBUG_NAMESPACE "BSIndex::Removal"
#endif
#include "debugMacros.hpp"
void BSIndex::Removal::propLit(const Term& lit)
{
  CALL("propLit(const Term& lit)");
  _index->removePropLit(lit,_clauseNum); 
};
