/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BSIndexFrontEnd.hpp"
#include "Clause.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_INDEX_FRONT_END
#  define DEBUG_NAMESPACE "BSIndexFrontEnd"
#endif
#include "debugMacros.hpp"
namespace VK
{
  void BSIndexFrontEnd::integrate(Clause* cl)
  {
    CALL("integrate(Clause* cl)");
    ASSERT(!isLocked());
    lock();
    cl->protect();
    _integrator.clause(cl->Number());
    Term header;
    for(LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
      { 
	header = *(iter.currentLiteral());
	if (header.arity())
	  {
	    if (header.IsEquality())
	      {
		_integrator.symLitHeader(header);
		_termTraversal.reset(*(iter.currentLiteral()->Args()));
		ASSERT(((bool)(_termTraversal)));
		while (_termTraversal)       
		  {
		    _integrator.symLitSubterm(_termTraversal.term());  
		    _termTraversal.next();
		  };
		_integrator.symLitSecondArg();
		_termTraversal.reset(*(iter.currentLiteral()->Args()->Next()));
		ASSERT(((bool)(_termTraversal)));
		while (_termTraversal)       
		  {
		    _integrator.symLitSubterm(_termTraversal.term());         
		    _termTraversal.next();
		  };
	      }
	    else 
	      {
		_integrator.litHeader(header);
		_termTraversal.reset(iter.currentElement()->literalRef());
		_termTraversal.next();
		ASSERT(((bool)(_termTraversal)));
		while (_termTraversal)       
		  {
		    _integrator.subterm(_termTraversal.term());         
		    _termTraversal.next();
		  };
	      };
	  } 
	else 
	  _integrator.propLit(header);
      };
    _integrator.endOfClause();
    cl->unprotect();
    unlock();
  }; 
  void BSIndexFrontEnd::remove(Clause* cl)
  {
    CALL("remove(Clause* cl)"); 
    ASSERT(!isLocked());
    lock();
    cl->protect(); 
    _removal.clause(cl->Number());
    Term header;
    for(LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
      { 
	header = *(iter.currentLiteral());   
	if (!header.isAnswerHeader())
	  {  
	    if (header.arity())
	      {
		if (header.IsEquality())
		  {
		    if (_removal.symLitHeader(header))  
		      {     
			_termTraversal.reset(*(iter.currentLiteral()->Args()));
			ASSERT(((bool)(_termTraversal)));
			while (_termTraversal)       
			  if (_removal.symLitSymbol(_termTraversal.symbol())) { _termTraversal.next(); }
			  else { _removal.symLitAfter(); _termTraversal.after(); };
			_termTraversal.reset(*(iter.currentLiteral()->Args()->Next()));
			_removal.symLitSecondArg();
			ASSERT(((bool)(_termTraversal)));
			while (_termTraversal)       
			  if (_removal.symLitSymbol(_termTraversal.symbol())) { _termTraversal.next(); }
			  else { _removal.symLitAfter(); _termTraversal.after(); }; 
			_removal.endOfSymLit();  
		      };    
		  }
		else 
		  if (_removal.litHeader(header))
		    { 
		      _termTraversal.reset(iter.currentElement()->literalRef());
		      _termTraversal.next();
		      ASSERT(((bool)(_termTraversal)));
		      while (_termTraversal)
			if (_removal.symbol(_termTraversal.symbol())) { _termTraversal.next(); }
			else { _removal.after(); _termTraversal.after(); };
		      _removal.endOfLit();  
		    };
	      } 
	    else 
	      _removal.propLit(header);
	  };
      };
    cl->unprotect();
    unlock();
  }; 
}; 
