/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BSMachine.hpp"
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
BSMachine::BSMachine() 
  : _clauseEqualizer(),
    _litEqualizer(DOP("BSMachine::_litEqualizer")),
    _symLitEqualizer(DOP("BSMachine::_symLitEqualizer")), 
    _substJoin() 
{
  DOP(_litEqualizer.freeze());
  DOP(_symLitEqualizer.freeze());
};
BSMachine::~BSMachine() { CALL("destructor ~BSMachine()"); };
bool BSMachine::subsumeNextMultisetMode(ulong& subsumedClNum)
{
  CALL("subsumeNextMultisetMode(ulong& subsumedClNum)");
  while (_clauseEqualizer.next())
    {
      ulong i;	  
      for (i = 0; i < _nextLit; i++) _litEqualizer[i].load(); 
      for (i = 0; i < _nextSymLit; i++) _symLitEqualizer[i].load();
      if (equalizeLiterals())
	{
	  if (_substJoin.joinInMultisetMode()) 
	    { 
	      subsumedClNum = _clauseEqualizer.currClauseNum(); 
	      return true; 
	    };
	};
    };
  return false;         
}; 
bool BSMachine::equalizeLiterals()
{
  CALL("equalizeLiterals()");
  ulong i;
  for (i = 0; i < _nextLit; i++)
    {
      if (!_litEqualizer[i].equalize()) return false;
    };
  for (i = 0; i < _nextSymLit; i++)  
    {
      if (!_symLitEqualizer[i].equalize()) return false;
    };
  return true;
}; 
#ifndef NO_DEBUG_VIS
const char* BSMachine::spell(const Tag& tag) 
{ 
  switch (tag)
    {
    case PropLit: return "PropLit";
    case Const: return "Const";
    case Var: return "Var";
    case SymLitConst: return "SymLitConst";
    case SymLitVar: return "SymLitVar";
    default: return "UnknownTag";
    };
};
#endif
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::ClauseEqualizer"
#endif
#include "debugMacros.hpp"
bool BSMachine::ClauseEqualizer::next() 
{ 
  CALL("next()");
  _currClauseNum--; 
  return equalize();
};
inline
bool BSMachine::ClauseEqualizer::equalize() 
{ 
  CALL("ClauseEqualizer::equalize()");
  if (equalizePrimary()) 
    { 
      equalizeSecondary(); 
      ASSERT(properlyEqualized());
      return true; 
    };
  return false;
};    
inline
bool BSMachine::ClauseEqualizer::equalizePrimary()
{ 
  CALL("ClauseEqualizer::equalizePrimary()");
  ASSERT(_firstPrim < _endOfPrim); 
  ulong str = _firstPrim;
  ulong secondPrim = _firstPrim + 1;
  bool notEmpty;
  do
    {
      if (_stream[str].equalize(_currClauseNum,notEmpty)) { str++; }
      else
	{
	  GlobalStopFlag::check();
	  if (notEmpty)
	    {
	      _stream[_firstPrim].equalize(_currClauseNum,notEmpty);
	      if (notEmpty) { str = secondPrim; }
	      else return false;   
	    }
	  else return false;
	};
    }
  while (str != _endOfPrim);
  return true;
};
inline
void BSMachine::ClauseEqualizer::equalizeSecondary()
{
  CALL("ClauseEqualizer::equalizeSecondary()");
  for (ulong i = 0; i != _firstPrim; i++) ALWAYS(_stream[i].equalize(_currClauseNum));
};
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::SubstJoin"
#endif
#include "debugMacros.hpp"
BSMachine::SubstJoin::SubstJoin() 
  : _subst(DOP("BSMachine::SubstJoin::_subst")), 
  _firstSubstCom(DOP("BSMachine::SubstJoin::_firstSubstCom")),
  _code(DOP("BSMachine::SubstJoin::_code")),
  _registeredVars(), 
  _registeredLiterals(DOP("BSMachine::SubstJoin::_registeredLiterals")),
  _litRegistered(false,"BSMachine::SubstJoin::_litRegistered")  
{
  for (ulong v = 0; v < VampireKernelConst::MaxNumOfVariables; v++) _firstOcc[v].registered = false;
  DOP(_subst.freeze());
  DOP(_firstSubstCom.freeze());
  DOP(_code.freeze());
  DOP(_registeredLiterals.freeze());
  DOP(_litRegistered.freeze());
}; 
BSMachine::SubstJoin::~SubstJoin() {};
bool BSMachine::SubstJoin::reset(ulong numOfQueryLits, ulong maxlNumOfLitsInIndex)
{
  CALL("reset(ulong numOfQueryLits, ulong maxlNumOfLitsInIndex)");
  DOP(_subst.unfreeze());
  if (!_subst.tryToExpand(numOfQueryLits)) return false; 
  DOP(_subst.freeze());
  DOP(_firstSubstCom.unfreeze());
  if (!_firstSubstCom.tryToExpand(numOfQueryLits)) return false;  
  DOP(_firstSubstCom.freeze());
  DOP(_code.unfreeze());
  if (!_code.tryToExpand(VampireKernelConst::MaxClauseSize + numOfQueryLits)) return false;
  DOP(_code.freeze());
  DOP(_registeredLiterals.unfreeze());
  if (!_registeredLiterals.tryToExpand(maxlNumOfLitsInIndex)) return false;
  DOP(_registeredLiterals.freeze());
  DOP(_litRegistered.unfreeze());
  if (!_litRegistered.tryToExpand(maxlNumOfLitsInIndex)) return false;
  DOP(_litRegistered.freeze());
  _nextSubst = 0;
  _endOfCode = _code.memory();
  while (_registeredVars) 
    {
#ifdef DEBUG_NAMESPACE
      VarOcc* debugVarOcc = _registeredVars.pop();
      debugVarOcc->registered = false; 
#else
      _registeredVars.pop()->registered = false;
#endif
    };
  return true;
}; 
void BSMachine::SubstJoin::nextLit(LitSubst* ls) 
{ 
  CALL("nextLit(LitSubst* ls)");
  _subst[_nextSubst] = ls;
  _endOfCode->nextSubst = true;
  _endOfCode->subst = ls;
  _endOfCode->substNum = _nextSubst; 
  _endOfCode->backtrackCommand = 0;
  _firstSubstCom[_nextSubst] = _endOfCode;
  _endOfCode++;
  _litVariables = 0;
  _maxBacktrackLit = -1;
};
void BSMachine::SubstJoin::endOfLit() 
{ 
  CALL("endOfLit()");
  if (_litVariables) 
    { 
      ASSERT(_nextSubst != _maxBacktrackLit);
      if (_maxBacktrackLit != -1) 
        {
	  ASSERT(_maxBacktrackLit >= 0);
	  _firstSubstCom[_nextSubst]->backtrackCommand = _firstSubstCom[_maxBacktrackLit];
	}
      else { ASSERT(!_firstSubstCom[_nextSubst]->backtrackCommand); };
      _nextSubst++; 
    } 
  else _endOfCode--; 
};
void BSMachine::SubstJoin::variable(ulong v)
{
  CALL("variable(ulong v)");
  ASSERT(v < VampireKernelConst::MaxNumOfVariables);
  if (_firstOcc[v].registered)     
    {
      _endOfCode->nextSubst = false;
      _endOfCode->var1 = _litVariables;
      _endOfCode->subst = _subst[_firstOcc[v].subst];
      _endOfCode->substNum = _firstOcc[v].subst;
      _endOfCode->var2 = _firstOcc[v].var;
      _endOfCode->backtrackCommand = _firstSubstCom[_nextSubst];
      _endOfCode++;
      if (_firstOcc[v].subst != _nextSubst && 
	  _maxBacktrackLit < _firstOcc[v].subst) 
        _maxBacktrackLit = _firstOcc[v].subst; 
    }
  else 
    {
      _firstOcc[v].registered = true;
      _firstOcc[v].var = _litVariables;
      _firstOcc[v].subst = _nextSubst;
#ifdef DEBUG_NAMESPACE
      _firstOcc[v].queryVar = v;
#endif
      _registeredVars.push(_firstOcc + v);
    };
  _litVariables++;
};         
bool BSMachine::SubstJoin::joinInMultisetMode()    
{
  CALL("joinInMultisetMode()");
  const Command* com = _code.memory();
  LitSubst* currSubst;
#ifdef DEBUG_NAMESPACE
  currSubst = 0; 
#else
#ifdef _SUPPRESS_WARNINGS_
  currSubst = 0; 
#endif 
#endif
  resetLitRegister();
  ulong litNum;
 perform_command:
  if (com == _endOfCode) return true;
 perform_command_wo_check:
  if (com->nextSubst) 
    {
      currSubst = com->subst; 
      currSubst->resetCursor();  
    check_lit_num:
      litNum = currSubst->getLitNum();
      if (_litRegistered[litNum]) 
	{ 
	  if (currSubst->moveCursor()) goto check_lit_num;   
	  goto long_backtrack;
	};
      registerLit(litNum);
      com++;
      goto perform_command;
    } 
  else 
    {
    perform_comparison:
      if (currSubst->get(com->var1) == com->subst->get(com->var2))
	{
	  com++;
	  goto perform_command;          
	};
      com = com->backtrackCommand;         
      ASSERT(com && com->nextSubst && (com->subst == currSubst));
      unregisterLastLit();
    try_next_row:
      if (currSubst->moveCursor()) 
	{
	  litNum = currSubst->getLitNum();
	  if (_litRegistered[litNum]) 
	    {
	      goto try_next_row;
	    };
	  registerLit(litNum);
	  com++;
	  ASSERT((com < _endOfCode) && (!com->nextSubst)); 
	  goto perform_comparison;
	}
      else 
	{
	long_backtrack:                  
	  com = com->backtrackCommand;
	  if (!com) return false;      
	  ASSERT(com->nextSubst && (com->subst != currSubst));  
	  unregisterLastLit();
	  currSubst = com->subst;
	try_another_row:
	  if (currSubst->moveCursor()) 
	    {
	      litNum = currSubst->getLitNum();
	      if (_litRegistered[litNum]) 
		{
		  goto try_another_row;
		};
	      registerLit(litNum);
	      com++; 
	      ASSERT(com < _endOfCode);
	      goto perform_command_wo_check; 
	    };          
	  goto long_backtrack;                                  
	};
    };
}; 
void BSMachine::SubstJoin::resetResultCollection() 
{
  _nextRegistered = _registeredVars.begin();
};
bool BSMachine::SubstJoin::resultNextPair(ulong& var,Term& term)
{
  CALL("resultNextPair(ulong& var,Term& term)");
  if (_nextRegistered == _registeredVars.end()) return false;
  var = (*_nextRegistered)->queryVar;
  term = _subst[(*_nextRegistered)->subst]->get((*_nextRegistered)->var);
  _nextRegistered++;
  return true; 
};
bool BSMachine::SubstJoin::litIsSubsumed(ulong litNum) const
{
  CALL("litIsSubsumed(ulong litNum) const");
  return _litRegistered[litNum];
}; 
#ifndef NO_DEBUG_VIS
ostream& BSMachine::SubstJoin::outputResult(ostream& str) 
{
  resetResultCollection();
  ulong v;
  Term t;
  while (resultNextPair(v,t)) str << "[X" << v << " -> " << t << "] ";      
  return str;
};
#endif
#ifndef NO_DEBUG_VIS
ostream& BSMachine::SubstJoin::outputCode(ostream& str) const 
{
  for (const Command* com = _code.memory(); com < _endOfCode; com++)
    {
      if (com->nextSubst)
	{
	  str << "SUBST (" << com->substNum << ", backtrack(";                 
	}
      else
	{       
	  str << "COMP([" << com->var1 << "]," << com->substNum << '[' << com->var2 << "],backtrack(";        
	};
      if (com->backtrackCommand)
	{
	  ulong i = 0;
	  for (i = 0; _firstSubstCom[i] != com->backtrackCommand;i++) 
	    {
	      ASSERT(_firstSubstCom[i]);
	    };        
	  str << i; 
	}
      else str << "STOP";      
      str << ")) ";
    };
  return str;  
}; 
#endif
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::LitEqualizer"
#endif
#include "debugMacros.hpp"
inline
bool BSMachine::LitEqualizer::equalize() 
{
  CALL("equalize()");
  if (equalizeFirst())
    {
      subst.reset();
      do
	{
	  collectSubst();
	}
      while (equalizeNext()); 
      return true; 
    };
  return false;
}; 
inline
bool BSMachine::LitEqualizer::equalizeFirst()
{
  CALL("equalizeFirst()");
  ASSERT(_nextConst || _nextVar); 
  _currLitNum = (_nextConst) ? (_const->topLitNum()) : (_var->topLitNum());
  return equalizeAll();
};
inline
bool BSMachine::LitEqualizer::equalizeNext()
{
  CALL("equalizeNext()");
  if (!_currLitNum) return false;
  _currLitNum--;
  return equalizeAll();
};  
inline
bool BSMachine::LitEqualizer::equalizeAll()
{
  CALL("equalizeAll()");
  bool notEnd;
  ulong str;
 equalize_const:
  if (_nextConst)
    {         
      str = 0;
      do
	{
	  if (_const[str].equalize(_currLitNum,notEnd)) { str++; }
          else 
	    if (notEnd) { str = 0; }
	    else return false;
	}
      while (str != _nextConst);        
    };
  ASSERT(constStreamsNonempty());
  ASSERT(constProperlyEqualized());
  str = 0;
  while (str != _nextVar)
    if (_var[str].equalize(_currLitNum,notEnd)) { str++; }
    else 
      if (notEnd) { goto equalize_const; }
      else return false;       
  ASSERT(varStreamsNonempty());
  ASSERT(varProperlyEqualized());
  return true; 
}; 
#ifndef NO_DEBUG_VIS
ostream& BSMachine::LitEqualizer::output(ostream& str) const
{
  str << "CONST\n";
  ulong i; 
  for (i = 0; i < _nextConst; i++) _const[i].output(str << "    ") << '\n';
  str << "VAR\n";
  for (i = 0; i < _nextVar; i++) _var[i].output(str << "    ") << '\n';
  if (_nextVar) subst.output(str << "SUBSTITUTION:\n",_nextVar) << "\nEND OF SUBSTITUTION\n";
  return str;
};
#endif
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::LitEqualizer::ConstStream"
#endif
#include "debugMacros.hpp"
inline
bool BSMachine::LitEqualizer::ConstStream::equalize(ulong& litNum,bool& notEnd)
{
  CALL("equalize(ulong& litNum,bool& notEnd)");
  ASSERT(llist);
  do
    {
      if (llist->hd() == litNum) return true;
      if (llist->hd() < litNum) { notEnd = true; litNum = llist->hd(); return false; }; 
      llist = llist->tl(); 
    }
  while (llist);
  notEnd = false;
  return false;  
}; 
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::LitEqualizer::VarStream"
#endif
#include "debugMacros.hpp"
inline
bool BSMachine::LitEqualizer::VarStream::equalize(ulong& litNum,bool& notEnd)
{
  CALL("equalize(ulong& litNum,bool& notEnd)");
  ASSERT(ltlist);
  do
    {
      if (ltlist->hd().first == litNum) return true;
      if (ltlist->hd().first < litNum) { notEnd = true; litNum = ltlist->hd().first; return false; }; 
      ltlist = ltlist->tl(); 
    }
  while (ltlist);
  notEnd = false;
  return false;  
}; 
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::SymLitEqualizer"
#endif
#include "debugMacros.hpp"
inline
bool BSMachine::SymLitEqualizer::equalize() 
{
  CALL("equalize()");
  if (equalizeFirst())
    {
      subst.reset();
      do
	{
	  collectSubst();
	}
      while (equalizeNext());
      return true;  
    };
  return false;
}; 
inline
bool BSMachine::SymLitEqualizer::equalizeFirst()
{
  CALL("equalizeFirst()");
  ASSERT(_nextConst1 || _nextVar1); 
  _currLitNum = (_nextConst1) ? (_const1->topLitNum()) : (_var1->topLitNum());
  _currArgNum1 = 1;
  _currArgNum2 = 1;
  return equalizeAll();
};
inline
bool BSMachine::SymLitEqualizer::equalizeNext()
{
  CALL("equalizeNext()");
  if (!_currLitNum) return false;
  _currLitNum--;
  _currArgNum1 = 1;
  _currArgNum2 = 1;
  return equalizeAll();
};  
inline
bool BSMachine::SymLitEqualizer::equalizeAll()
{
  CALL("equalizeAll()");
  bool notEnd;
  ulong str;
 equalize_const:
  if (_nextConst1)
    {         
      str = 0;
      do
	{
	  if (_const1[str].equalize(_currLitNum,_currArgNum1,notEnd)) { str++; }
          else 
	    if (notEnd) { str = 0; }
	    else return false;
	}
      while (str != _nextConst1);        
    };
  if (_nextConst2)
    {         
      str = 0;
      do
	{
	  if (_const2[str].equalize(_currLitNum,_currArgNum2,notEnd)) { str++; }
          else 
	    if (notEnd) { goto equalize_const;  }
	    else return false;
	}
      while (str != _nextConst2);        
    };
  ASSERT(constStreamsNonempty());
  ASSERT(constProperlyEqualized());
  str = 0;
  while (str != _nextVar1)
    if (_var1[str].equalize(_currLitNum,_currArgNum1,notEnd)) { str++; }
    else 
      if (notEnd) { goto equalize_const; }
      else return false;    
  str = 0;
  while (str != _nextVar2)
    if (_var2[str].equalize(_currLitNum,_currArgNum2,notEnd)) { str++; }
    else 
      if (notEnd) { goto equalize_const; }
      else return false;        
  ASSERT(varStreamsNonempty());
  ASSERT(varProperlyEqualized());
  if (_currArgNum1 + _currArgNum2)
    {
      _altArg1 = alternativeArg1();
      _altArg2 = alternativeArg2();
      if (_currArgNum1 != _currArgNum2) return true;
      if (_altArg1 || _altArg2) return true;                  
    };
  if (!_currLitNum) return false;
  _currLitNum--;        
  _currArgNum1 = 1;
  _currArgNum2 = 1;
  goto equalize_const;
}; 
inline
void BSMachine::SymLitEqualizer::collectSubst() 
{ 
  CALL("collectSubst()");
  ulong i;
  if ((!_currArgNum1) || (!_currArgNum2))
    {
      for (i = 0; i < _nextVar1; i++) subst.set(i,_var1[i].topTerm());
      for (i = 0; i < _nextVar2; i++) subst.set(_nextVar1 + i,_var2[i].topTerm()); 
    }
  else 
    {
      if (_altArg2) 
        {
	  for (i = 0; i < _nextVar1; i++) subst.set(i,_var1[i].topTerm());
	  for (i = 0; i < _nextVar2; i++) subst.set(_nextVar1 + i,_var2[i].secondTerm());
	  if (_altArg1)
	    { 
	      subst.setLitNum(_currLitNum);
	      subst.endOfRow(); 
	      for (i = 0; i < _nextVar1; i++) subst.set(i,_var1[i].secondTerm());
	      for (i = 0; i < _nextVar2; i++) subst.set(_nextVar1 + i,_var2[i].topTerm()); 
	    };
        }
      else
        {
	  ASSERT(_altArg1);
	  for (i = 0; i < _nextVar1; i++) subst.set(i,_var1[i].secondTerm());
	  for (i = 0; i < _nextVar2; i++) subst.set(_nextVar1 + i,_var2[i].topTerm()); 
        };
    };
  subst.setLitNum(_currLitNum);
  subst.endOfRow();  
}; 
inline
bool BSMachine::SymLitEqualizer::alternativeArg1()
{
  CALL("alternativeArg1()");
  ulong i;
  for (i = 0; i < _nextConst1; i++) 
    if ((!(_const1[i].list()->tl())) || (_const1[i].list()->tl()->hd().first != _currLitNum)) 
      return false;
  for (i = 0; i < _nextVar1; i++) 
    if ((!(_var1[i].list()->tl())) || (_var1[i].list()->tl()->hd().first != _currLitNum)) 
      return false;
  return true;
};
inline
bool BSMachine::SymLitEqualizer::alternativeArg2()
{
  CALL("alternativeArg2()");
  ulong i;
  for (i = 0; i < _nextConst2; i++) 
    if ((!(_const2[i].list()->tl())) || (_const2[i].list()->tl()->hd().first != _currLitNum)) 
      return false;
  for (i = 0; i < _nextVar2; i++) 
    if ((!(_var2[i].list()->tl())) || (_var2[i].list()->tl()->hd().first != _currLitNum)) 
      return false;
  return true;
};
#ifndef NO_DEBUG_VIS
ostream& BSMachine::SymLitEqualizer::output(ostream& str) const
{
  ulong i;	  
  str << "CONST1\n";
  for (i = 0; i < _nextConst1; i++) _const1[i].output(str << "    ") << '\n';
  str << "VAR1\n";
  for (i = 0; i < _nextVar1; i++) _var1[i].output(str << "    ") << '\n';
  str << "CONST2\n";
  for (i = 0; i < _nextConst2; i++) _const2[i].output(str << "    ") << '\n';
  str << "VAR2\n";
  for (i = 0; i < _nextVar2; i++) _var2[i].output(str << "    ") << '\n';
  str << "CURR.LIT. " << _currLitNum << ", ARG1 " << _currArgNum1 << ", ARG2 " << _currArgNum2 << '\n';
  if (_nextVar1 + _nextVar2) subst.output(str << "SUBSTITUTION:\n",_nextVar1 + _nextVar2) << "\nEND OF SUBSTITUTION\n";
  return str;
};
#endif
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::SymLitEqualizer::ConstStream"
#endif
#include "debugMacros.hpp"
inline
bool BSMachine::SymLitEqualizer::ConstStream::equalize(ulong& litNum,ulong& argNum,bool& notEnd)
{
  CALL("equalize(ulong& litNum,ulong& argNum,bool& notEnd)");
  ASSERT(lalist);
  do
    {
      if (lalist->hd().first == litNum) 
	{
	  if (lalist->hd().second == argNum) return true;
	  if (lalist->hd().second < argNum) 
            { 
	      notEnd = true; 
	      argNum = lalist->hd().second;
	      return false;            
            };
	}
      else
	{
	  if (lalist->hd().first < litNum) 
            {
	      notEnd = true; 
	      litNum = lalist->hd().first; 
	      argNum = lalist->hd().second;
	      return false; 
            }; 
	}; 
      lalist = lalist->tl();
    }
  while (lalist);
  notEnd = false;
  return false;  
}; 
#undef DEBUG_NAMESPACE 
#ifdef DEBUG_BS_MACHINE
#  define DEBUG_NAMESPACE "BSMachine::SymLitEqualizer::VarStream"
#endif
#include "debugMacros.hpp"
inline
bool BSMachine::SymLitEqualizer::VarStream::equalize(ulong& litNum,ulong& argNum,bool& notEnd)
{
  CALL("equalize(ulong& litNum,ulong& argNum,bool& notEnd)");
  ASSERT(latlist);
  do
    {
      if (latlist->hd().first == litNum) 
	{
	  if (latlist->hd().second == argNum) return true;
	  if (latlist->hd().second < argNum) 
            { 
	      notEnd = true; 
	      argNum = latlist->hd().second;
	      return false;            
            };
	}
      else
	{
	  if (latlist->hd().first < litNum) 
            {
	      notEnd = true; 
	      litNum = latlist->hd().first; 
	      argNum = latlist->hd().second;
	      return false; 
            }; 
	}; 
      latlist = latlist->tl();
    }
  while (latlist);
  notEnd = false;
  return false;  
};  
