/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BACK_DEMODULATION_HPP
#define BACK_DEMODULATION_HPP
#include "jargon.hpp"
#include "GlobAlloc.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "ExpandingMultiset.hpp"
#include "Term.hpp"
#include "InstanceRetrievalForDemodulation.hpp"
#include "LiteralList.hpp"
#include "RuntimeError.hpp"
#include "SimplificationOrdering.hpp"
namespace VK
{
  class TemporaryClauseCollector;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACK_DEMODULATION
#define DEBUG_NAMESPACE "BackDemodulation"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Clause;
  class BackDemodulation
  {
  public:
    typedef 
    BK::ExpandingMultiset<BK::GlobAlloc,ulong,32UL> 
    SetOfSplittingBranches; 
  public:
    BackDemodulation();
    ~BackDemodulation();
    void reset();
    bool isLocked() const { return _isLocked; };
    void setClauseCollector(TemporaryClauseCollector* cc)
    {
      _clauseCollector = cc;
    };
    void setDegree(long d) { _degree = d; };
    void setBDOnSplittingBranchesFlag(bool fl) 
    {
      if (_bDOnSplittingBranchesFlag && (!fl)) BK::RuntimeError::report("Switching _bDOnSplittingBranchesFlag off.");
      _bDOnSplittingBranchesFlag = fl; 
    };
    void setDemodulateFromSupport(bool fl)
    {
      _demodulateFromSupport = fl;
    };
    void setProhibitDemodulationIntoInputNonBasic(bool fl)
    {
      _prohibitDemodulationIntoInputNonBasic = fl;
    };
    void setCheckSubstSafetyWRTInputNonBasicness(bool fl)
    {
      _checkSubstSafetyWRTInputNonBasicness = fl;
    };
    bool eligibleAsQuery(const Clause* cl) const;
    bool query(Clause* cl);
    void abandonCurrentQuery();
    bool nextSimplified(bool& onBranch);
    bool collectInference();
    Clause* simplified() { return _recentlySimplified; }; 
    void integrate(Clause* cl);
    void remove(Clause* cl);
    ulong statBDOnSplittingBranches() const { return _numOfBDOnSplittingBranches; };
  private:
    void integrateLiteral(Term* lit,ulong litNum,Clause* cl);
    void integrateComplexTerm(Term* term,ulong litNum,Clause* cl);
    void removeLiteral(Term* lit,Clause* cl);
    void removeComplexTerm(Term* term,Clause* cl);
    bool query(const Term* lhsref,const Term* rhsref);
    bool safeRuleWRTInputNonBasicness(const Term& lhs,const Term& rhs);
    void collectInputNonBasicVars(const Term& term,bool alreadyInputNonBasic);
    void cancelInputNonBasicVars(const Term& term,bool alreadyInputNonBasic); 
    bool 
    appropriateDegree(const SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Category& cat)
    {
      switch (cat)
	{	  
	case SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyTautology:
	  return  (_degree >= 2);
	case SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile:
	  return (_degree >= 3);
	case SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyContradiction:
	  return (_degree >= 4);
	default: return false;
	};
    };
    bool rewrite(const Term::OccList::Node* occ);
    bool rewriteLit(const LiteralList::Element* lit);
    bool collectRHSInstance();
  private:
    void lock() { _isLocked = true; };
    void unlock() { _isLocked = false; };
  private:
    bool _isLocked;
    bool _swapQuerySides;
    long _degree;
    bool _bDOnSplittingBranchesFlag;
    bool _demodulateFromSupport;
    bool _prohibitDemodulationIntoInputNonBasic;
    bool _checkSubstSafetyWRTInputNonBasicness;
    TemporaryClauseCollector* _clauseCollector;
    InstRet _index;
    InstRet::Substitution::Instance _rhsInstance;
    SimplificationOrdering::CompiledConstraintForInstanceRetrieval* _compiledConstraint;
    SimplificationOrdering::CompiledConstraintForInstanceRetrieval* _currentConstraint;
    Clause* _rule;
    const Term* _ruleLHS;
    Term _ruleRHS;
    Term::Iterator _termIter;
    const Term::OccList::Node* _candidates; 
    Clause* _recentlySimplified;
    const Term* _redex;
    Term::Iterator _rewrittenLitIter;
    Term::Iterator _RHSIter;
    SetOfSplittingBranches _splittingBranches;
    ulong _numOfBDOnSplittingBranches;
    BK::LinearCombination<long,VampireKernelConst::MaxNumOfVariables> _inputNonBasicVarsInRHS;
  }; 
}; 
#endif
