/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BackwardSubsumption.hpp"
#include "ClauseNumIndex.hpp"
#include "Clause.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUBSUMPTION
#  define DEBUG_NAMESPACE "BackwardSubsumption"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
inline 
bool 
BackwardSubsumption::preproQuery(const Clause* cl)
{
  CALL("preproQuery(const Clause* cl)");
  ASSERT(isLocked() && _index->isLocked());
  if (!_machine.resetQuery(cl->numOfAllLiterals(),_index->index())) 
    return false; 
  for (LiteralList::Iterator iter(cl->LitList()); iter.notEnd(); iter.next())
    {
      Term* lit = iter.currentLiteral();
      if (!preproQueryLit(lit)) return false;
    };
  return _machine.endOfQuery();
}; 
inline 
bool 
BackwardSubsumption::preproQueryLit(const Term* lit)
{
  CALL("preproQueryLit(const Term* lit)");
  if (lit->arity())
    { 
      if (lit->IsEquality()) 
	{ 
	  return (preproQuerySymLit(lit) && _machine.queryEndOfSymLit()); 
	}
      else 
	{
	  if (!_retrieval.litHeader(*lit)) return false;
	  _machine.queryNextLit();     
	  return  (preproQueryComplexTermArguments(lit) && _machine.queryEndOfLit());  
	};
    }
  else 
    {
      const CSkipList* plcands;
      return ((_retrieval.propLit(*lit,plcands)) && (_machine.queryPropLit(plcands)));
    };
}; 
inline 
bool 
BackwardSubsumption::preproQueryComplexTermArguments(const Term* t) 
{
  CALL("preproQueryComplexTermArguments(const Term* t)");
  ASSERT(t->isComplex());
  ASSERT(t->arity());
  const CLTSkipList* vcands;
  bool allVars = true;
  const Term* argRef;
  for (argRef = t->Args(); ((!argRef->isNotArg()) && allVars); argRef = argRef->Next())
    allVars = argRef->isVariable();
  argRef = t->Args();
  if (argRef->isVariable())
    {
      if ((!_retrieval.variable(vcands)) || (!_machine.queryVariable(*argRef,vcands,allVars)))
	return false;  
    }
  else 
    if (!preproQueryComplexTerm(argRef->deref())) return false;
  argRef = argRef->Next();
  while (!argRef->isNotArg())
    {
      if (argRef->isVariable())
	{
	  if ((!_retrieval.variable(vcands)) || (!_machine.queryVariable(*argRef,vcands,false)))
	    return false;  
	}
      else 
	if (!preproQueryComplexTerm(argRef->deref())) return false; 
      argRef = argRef->Next();    
    };   
  return true;
}; 
inline 
bool 
BackwardSubsumption::preproQueryComplexTerm(const Term* t)
{
  CALL("preproQueryComplexTerm(const Term* t)");
  ASSERT(t->isComplex());
  const CLSkipList* ccands;
  if (t->arity())
    {
      return _retrieval.nonconstFunc(*t) && preproQueryComplexTermArguments(t);
    }
  else 
    return _retrieval.constant(*t,ccands) && _machine.queryConstant(*t,ccands);
}; 
inline 
bool 
BackwardSubsumption::preproQuerySymLit(const Term* lit)
{
  CALL("preproQuerySymLit(const Term* lit)");
  ASSERT(lit->IsEquality());
  if (!_retrieval.symLitHeader(*lit)) return false;
  _machine.queryNextSymLit();
  const Term* argRef1 = lit->Args();
  const Term* argRef2 = argRef1->Next(); 
  const CLATSkipList* slvcands;
  if (argRef1->isVariable()) 
    {   
      if ((!_retrieval.symLitVariable(slvcands)) 
	  || (!_machine.querySymLitVariable(*argRef1,slvcands,argRef2->isVariable())))
	return false;      
    }
  else 
    if (!preproQuerySymLitComplexTerm(argRef1->deref())) return false;
  if (!_retrieval.symLitSecondArg()) return false;
  _machine.querySymLitSecondArg();
  if (argRef2->isVariable()) 
    {   
      return (_retrieval.symLitVariable(slvcands)) 
	&& (_machine.querySymLitVariable(*argRef2,slvcands,false)); 
    }
  else return preproQuerySymLitComplexTerm(argRef2->deref()); 
}; 
inline 
bool 
BackwardSubsumption::preproQuerySymLitComplexTerm(const Term* t)
{
  CALL("preproQuerySymLitComplexTerm(const Term* t)");
  ASSERT(t->isComplex());
  const CLASkipList* slccands;
  if (t->arity())
    {
      return _retrieval.symLitNonconstFunc(*t) && preproQuerySymLitComplexTermArguments(t);
    }
  else 
    return _retrieval.symLitConstant(*t,slccands) && _machine.querySymLitConstant(*t,slccands);
}; 
inline 
bool 
BackwardSubsumption::preproQuerySymLitComplexTermArguments(const Term* t) 
{
  CALL("preproQuerySymLitComplexTermArguments(const Term* t)");
  ASSERT(t->isComplex());
  ASSERT(t->arity());
  const CLATSkipList* slvcands;
  bool allVars = true;
  const Term* argRef;
  for (argRef = t->Args(); ((!argRef->isNotArg()) && allVars); argRef = argRef->Next())
    allVars = argRef->isVariable();
  argRef = t->Args();          
  if (argRef->isVariable())
    {
      if ((!_retrieval.symLitVariable(slvcands)) || (!_machine.querySymLitVariable(*argRef,slvcands,allVars)))
	return false;  
    }
  else 
    if (!preproQuerySymLitComplexTerm(argRef->deref())) return false;
  argRef = argRef->Next();
  while (!argRef->isNotArg())
    {
      if (argRef->isVariable())
	{
	  if ((!_retrieval.symLitVariable(slvcands)) || (!_machine.querySymLitVariable(*argRef,slvcands,false)))
	    return false;  
	}
      else 
	if (!preproQuerySymLitComplexTerm(argRef->deref())) return false; 
      argRef = argRef->Next();    
    };   
  return true;
}; 
bool BackwardSubsumption::query(Clause* cl) 
{
  CALL("query(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  _queryClause = cl;
  bool res = preproQuery(cl); 
  if (!res) { cl->unprotect(); unlock(); };
  return res;
}; 
void BackwardSubsumption::abandonCurrentQuery()
{
  CALL("abandonCurrentQuery()");
  ASSERT(isLocked() && _index->isLocked());
  _queryClause->unprotect();
  unlock();
}; 
bool BackwardSubsumption::subsumeNext() 
{ 
  CALL("subsumeNextMultisetMode()");
  ASSERT(isLocked() && _index->isLocked());
  ulong subsumedClNum;
  while (_machine.subsumeNextMultisetMode(subsumedClNum)) 
    {
      _recentlySubsumedCl = _clauseNumIndex->find(subsumedClNum);
      ASSERT(_recentlySubsumedCl); 
      if (_recentlySubsumedCl->isPersistent())
	return true;
    }; 
  return false; 
}; 
