/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BACKWARD_SUBSUMPTION_HPP
#define BACKWARD_SUBSUMPTION_HPP 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "BSIndexFrontEnd.hpp"
#include "BSMachine.hpp"
#include "ClauseNumIndex.hpp"
#include "Clause.hpp"
#include "Term.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUBSUMPTION
#  define DEBUG_NAMESPACE "BackwardSubsumption"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BackwardSubsumption
  {
  public:
    typedef BSIndex::LTPair LTPair;  
    typedef BSIndex::LAPair LAPair;
    typedef BSIndex::LATTriple LATTriple;
    typedef BSIndex::LList LList;
    typedef BSIndex::LTList LTList;
    typedef BSIndex::LAList LAList;
    typedef BSIndex::LATList LATList;
    typedef BSIndex::CSkipList CSkipList;
    typedef BSIndex::CLSkipList CLSkipList;
    typedef BSIndex::CLTSkipList CLTSkipList;
    typedef BSIndex::CLASkipList CLASkipList;
    typedef BSIndex::CLATSkipList CLATSkipList;
  public:
    void init(BSIndexFrontEnd* index,ClauseNumIndex* clauseNumIndex)
    {
      CALL("init(ClauseNumIndex* clauseNumIndex)");
      _index = index;
      _retrieval.init(index->index());
      BK::init(_machine);
#ifdef DEBUG_NAMESPACE
      BK::init(_termTraversal,"BackwardSubsumption::_termTraversal");
#else
      BK::init(_termTraversal);
#endif
      _clauseNumIndex = clauseNumIndex;
      _isLocked = false;
    };
    void destroy()
    {
      CALL("destroy()");
      BK::destroy(_termTraversal);
      BK::destroy(_machine);
      _retrieval.destroy();
      BK_CORRUPT(*this);
    };
    void reset()
    {
      CALL("reset()");
      _isLocked = false;
    }; 
    bool isLocked() const { return _isLocked; };
    bool query(Clause* cl);
    void abandonCurrentQuery();
    bool subsumeNext();
    Clause* subsumed() const 
    { 
      CALL("subsumed() const");
      ASSERT(isLocked());
      return _recentlySubsumedCl; 
    };
  private:
    void lock() { _isLocked = true; _index->lock(); };
    void unlock() { _isLocked = false; _index->unlock(); };
    bool preproQuery(const Clause* cl);
    bool preproQueryLit(const Term* lit);
    bool preproQuerySymLit(const Term* lit);
    bool preproQueryComplexTerm(const Term* t);
    bool preproQueryComplexTermArguments(const Term* t);
    bool preproQuerySymLitComplexTerm(const Term* t);
    bool preproQuerySymLitComplexTermArguments(const Term* t);
  private:
    BSIndexFrontEnd* _index;
    BSIndex::Retrieval _retrieval;
    BSMachine _machine; 
    Term::Traversal _termTraversal;
    ClauseNumIndex* _clauseNumIndex;
    Clause* _recentlySubsumedCl;
    Clause* _queryClause; 
    bool _isLocked;
  }; 
}; 
#endif
