/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BACKWARD_SUPERPOSITION_HPP
#define BACKWARD_SUPERPOSITION_HPP
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"   
#include "DestructionMode.hpp"
#include "IUMachine.hpp"
#include "SimplificationOrdering.hpp"
#include "SizeFormula.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class ForwardProcessing;
  class Statistics;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "BackwardSuperposition"
#endif
#include "debugMacros.hpp"
namespace VK
{
class BackwardSuperposition
{
 public:
  BackwardSuperposition(UnifIndex* unifIndex); 
  ~BackwardSuperposition(); 
  void setStatistics(Statistics* stat)
  {
    _dummyClauseCollector.setStatistics(stat);
    _statistics = stat;
  };
  void setClauseCollector(TemporaryClauseCollector* cc) 
  {
    _clauseCollector = cc;
  };
  void setForwardProcessing(ForwardProcessing* fp)
  {
    _forwardProcessing = fp;
  };
  void setSimultaneous(long degree) { _simultaneousness = degree; };
  void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;     
    };
  void setParamodulateFromVariablesFlag(bool fl) 
  {
    _paramodulateFromVariables = fl;
  };
  void setMainPremiseSubsumption(bool fl)
  {
    _mainPremiseSubsumption = fl;
  };
  void setSidePremiseSubsumption(bool fl)
  {
    _sidePremiseSubsumption = fl;
  };
  void setMainSelectedLitDemodulation(bool fl)
  {
    _mainSelectedLitDemodulation = fl;
  };
  void setSideSelectedLitDemodulation(bool fl)
  {
    _sideSelectedLitDemodulation = fl;
  };
  void setClauseSizeLimit(long lim)
  {
    _clauseSizeLimit = lim;
  };
  bool SetQuery(Clause* cl); 
  bool CollectResult();
  bool NextInference(); 
private:
  void compileMainPremiseResidualSizeFormula();
  void recompileMainPremiseSelectedLitResidualSizeFormula();
  bool findNextMainSelectedPosEq();
  bool collectMainPremiseLiterals();
  bool loadEq(const LiteralList::Element* eq_lit,ulong eq_ind);
  bool Paramodulate();
  bool collectSidePremiseLiterals();
  bool LoadLHS(UnifTerm* t);
  bool NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr,
		     Clause*& candCl,
		     BackParaCandidates::LitRedexPair*& candLitRedexPair); 
  bool subsumeMainPremise(); 
  bool collectMainPremiseInstance();
  bool subsumeSidePremise(); 
  bool collectSidePremiseInstance();
  bool demodulateSelectedLits();
  bool collectMainPremiseSubtermsAvailableForParamodulationInto();
  bool collectSidePremiseSubtermsAvailableForParamodulationInto();
private:
  OPEN_BR_DECL(1);
  Statistics* _statistics;
  TemporaryClauseCollector* _clauseCollector;
  ForwardProcessing* _forwardProcessing;
  UnifIndex* _unifIndex; 
  long _simultaneousness;
  bool _noInfBetweenRulesFlag;
  bool _paramodulateFromVariables;
  bool _mainPremiseSubsumption;
  bool _sidePremiseSubsumption;
  bool _mainSelectedLitDemodulation;
  bool _sideSelectedLitDemodulation;
  long _clauseSizeLimit;
  ulong _mainPremiseVarIndex;
  Clause* _mainPremise;
  LiteralList* _allMainPremiseLiterals;
  LiteralList::Element* _firstNonselectedMainLit;
  LiteralList::Iterator _mainSelectedLitIter;
  SizeFormula _mainPremiseResidualSizeFormula;
  SizeFormula _mainPremiseSelectedLitResidualSizeFormula;
  long _mainPremiseResidualSize;
  const LiteralList::Element* _currentMainPosEqLit;
  UnifTerm _currentMainPosEqArg1Mem[VampireKernelConst::MaxTermSize];
  UnifTerm _currentMainPosEqArg2Mem[VampireKernelConst::MaxTermSize];
  UnifTerm* _currentMainPosEqRHS;
  SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
  SimplificationOrdering::CompiledConstraintForUnifiers* _currentCompiledConstraint;
  Clause* _sidePremise;
  ulong _sidePremiseVarIndex;
  SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;
  IUMachine _retrievalOfUnifiable;
  BackParaCandidates::Iterator _candidates;
  BackParaCandidates::LitRedexPair* _candLitRedexPair;
  bool _newRetrievalAttempt;
  bool _isFromVariable;
  bool _swapArguments;
  bool _constraintRequired;
  bool _noRulesAsSidePremises;
  bool _mainSelectedLitDemodulationAttempted;
  bool _mainSelectedLitDemodulated;
  TemporaryClause::Extras::Handle _dummyClause;
  TemporaryClauseCollector _dummyClauseCollector;
  CLOSE_BR_DECL(1);
}; 
}; 
#endif 
