/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BasicTempLitList.hpp"
using namespace BK;
using namespace VK;
BasicTempLitList::Context* BasicTempLitList::Context::_current; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BASIC_TEMP_LIT_LIST
#  define DEBUG_NAMESPACE "BasicTempLitList"
#endif
#include "debugMacros.hpp"
BasicTempLitList::BasicTempLitList() :
  _first(0),
  _last(0),
  _length(0UL),
  _elementPool(Context::getCurrent()->defaultElementPool())
{    
};
BasicTempLitList::BasicTempLitList(ElementPool* elemPool) :
  _first(0),
  _last(0),
  _length(0UL),
  _elementPool(elemPool)
{    
};
BasicTempLitList::BasicTempLitList(const BasicTempLitList& lst) :
  _first(0),
  _last(0),
  _length(lst._length),
  _elementPool(lst._elementPool)
{
  for (const Element* el = lst.first(); el; el = el->next())
    push(el->literal());
}; 
BasicTempLitList::~BasicTempLitList()
{
  CALL("destructor ~BasicTempLitList()");
  if (BK::DestructionMode::isThorough())
    {
      while (_first)
	{
	  Element* tmp = _first;
	  _first = _first->next();
	  ElementPool::returnObject(tmp);
	};
    };
}; 
void BasicTempLitList::relocateTo(void* target)
{
  CALL("relocateTo(void* target)");
  ASSERT(static_cast<void*>(this) != target);
  ASSERT(checkIntegrity());
  static_cast<BasicTempLitList*>(target)->_first = _first;
  static_cast<BasicTempLitList*>(target)->_last  = _last;
  static_cast<BasicTempLitList*>(target)->_elementPool  = _elementPool;
  static_cast<BasicTempLitList*>(target)->_length  = _length;
  for (Element* el = _first; el; el = el->next())
    el->_list = static_cast<BasicTempLitList*>(target);
  BK_CORRUPT(*this);
  ASSERT(static_cast<BasicTempLitList*>(target)->checkIntegrity());
}; 
void BasicTempLitList::collectAllFrom(BasicTempLitList& lst)
{
  CALL("collectAllFrom(BasicTempLitList& lst)");
  ASSERT(this != &lst);
  ASSERT(checkIntegrity()); 
  ASSERT(lst.checkIntegrity());
  if (lst.isEmpty()) return;
  if (lst._elementPool == _elementPool)
    {
      if (isEmpty())
	{
	  _first = lst._first;
	  _last = lst._last;
	}
      else 
	{
	  lst._first->_previous = _last;
	  _last->_next = lst._first;
	  _last = lst._last;
	};
      for (Element* el = lst._first; el; el = el->_next)
	el->_list = this;
    }
  else 
    {	
      Element* obsoleteElem = lst._first;
      do
	{
	  Element* newElem = _elementPool->lendObject();
	  newElem->literal().collectFrom(obsoleteElem->literal());
	  newElem->_list = this;
	  if (_last)
	    {
	      _last->_next = newElem;
	      newElem->_previous = _last;
	    }
	  else
	    {
	      _first = newElem;
	      newElem->_previous = 0;
	    };
	  _last = newElem;
	  Element* tmp = obsoleteElem->_next;
	  ElementPool::returnObject(obsoleteElem);
	  obsoleteElem = tmp;
	}
      while (obsoleteElem);
      _last->_next = 0;
    }; 
  _length += lst._length;
  lst._first = 0;
  lst._last = 0;
  lst._length = 0UL;
  COP("END");
  ASSERT(checkIntegrity());
  ASSERT(lst.checkIntegrity());
  ASSERT(lst.isEmpty());
}; 
ostream& BasicTempLitList::output(ostream& str) const
{
  CALL("output(ostream& str) const");
  for (const Element* el = first(); el; el = el->next())
    {
      el->literal()->output(str);
      if (el->next()) str << " % ";
    };
  return str;
}; 
