/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BINARY_RESOLUTION_HPP
#define BINARY_RESOLUTION_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "PremiseWithFactoring.hpp"
#include "IUMachine.hpp"
#include "SizeFormula.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class Clause;
  class ForwardProcessing;
  class Statistics;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
#define DEBUG_NAMESPACE "BinaryResolution"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BinaryResolution
  {
  public:
    BinaryResolution(UnifIndex* resIndex);
    ~BinaryResolution() { CALL("destructor ~BinaryResolution()"); };
    void setStatistics(Statistics* stat)
    {
      _dummyClauseCollector.setStatistics(stat);
      _statistics = stat;
    };
    void setClauseCollector(TemporaryClauseCollector* cc) 
    {
      _clauseCollector = cc;
    };
    void setForwardProcessing(ForwardProcessing* fp)
    {
      _forwardProcessing = fp;
    };
    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };
    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };
    void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;
    };
    void setClauseSizeLimit(long lim)
    {
      _clauseSizeLimit = lim;
    };
    bool Load(Clause* c);
    bool NextInference();  
  private:
    enum Continuation { ContFirstLit, ContFirstFactor, ContNextFactor }; 
  private: 
    bool NextFactor();
    void compileResidualSize(const Clause* c);
    inline bool CollectResult();
    bool subsumeMainPremise();
    bool subsumeSidePremise();
    bool demodulateMainSelectedLits();
    bool demodulateSideSelectedLits(bool includeMarked);
    static bool eligible(const Clause* cl); 
  private:
    OPEN_BR_DECL(1);
    Statistics* _statistics;
    PremiseWithFactoring _mainPremise, _sidePremise;
    ulong _mainPremiseVarIndex;
    ulong _sidePremiseVarIndex;
    ulong _currentMainLit;
    bool _rotationPossible;
    IUMachine ORMachine;
    UnifIndex* _unificationIndex;
    SizeFormula _residualSizeFormula;
    long _residualMainPremiseSize;
    long _clauseSizeLimit;
    TemporaryClauseCollector* _clauseCollector;
    ForwardProcessing* _forwardProcessing;
    TemporaryClause::Extras::Handle _dummyClause;
    TemporaryClauseCollector _dummyClauseCollector;
    Continuation _continuation;
    bool _noInfBetweenRulesFlag;
    bool _sideClauseCanNotBeRule;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;
    CLOSE_BR_DECL(1);
  }; 
}; 
#endif
