/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CLAUSE_ASSEMBLER_HPP
#define CLAUSE_ASSEMBLER_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "LiteralSharing.hpp"
#include "Unifier.hpp"
#include "VariableRenaming.hpp"
#include "VariableWeighting.hpp"
#include "TemporaryLiteral.hpp"
#include "LiteralProperties.hpp"
namespace VK
{
  class DiscTree::Node;
  class LiteralSelection;
  class BasicTempLitList;
  class Clause;
  class OpenClauseBackground;
  class PagedPassiveClauseStorage;
  class SymbolInfo;
  class TemporaryClause;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_ASSEMBLER
#  define DEBUG_NAMESPACE "ClauseAssembler"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ClauseAssembler
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
  public:
    ClauseAssembler(LiteralSharing* litSharing,const LiteralSelection* litSelection);
    ~ClauseAssembler() { CALL("destructor ~ClauseAssembler()"); };
    void reset();
    bool sharingIsLocked() const 
    {
      return _litSharing->isLocked();
    };
    void setInheritNegativeSelection(bool flag) { _inheritNegativeSelection = flag; };
    void setEqualityIsPossibleFlag(bool flag) 
    {
      _equalityIsPossible = flag;
    };
    void setMaxMaxWeight(long w)
    {
      CALL("setMaxMaxWeight(long w)");
      _maxMaxWeight = w;
    };
    void setMaxWeight(long w)
    {
      CALL("setMaxWeight(long w)");
      _maxWeight = w;
      if (_maxWeight < 1L)
	{
	  _maxWeight = 1L;
	}
      else if (_maxWeight > _maxMaxWeight)
	_maxWeight = _maxMaxWeight;
    };
    Clause* assembleShared(TemporaryClause& cl);
    Clause* assembleSharedWithoutSelection(TemporaryClause& cl);
    bool assembleInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
					     TemporaryClause& cl,
					     long clauseNum,
					     ostream* showNewClause);
    Clause* assembleNameIntro(const TemporaryLiteral& definedLit,
			      const BasicTempLitList& body);
    void removeFromSharing(Clause* cl);
  private:
    Clause* assembleEmptyClause(OpenClauseBackground& ancestors) const;
    Clause* assembleUnitClause(TemporaryLiteral& lit,
			       OpenClauseBackground& ancestors);
    Clause* assembleClause(BasicTempLitList& lits,
			   OpenClauseBackground& ancestors);
    Clause* assembleClauseWithoutSelection(BasicTempLitList& lits,
					   OpenClauseBackground& ancestors);
    bool assembleClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						   BasicTempLitList& lits,
						   OpenClauseBackground& ancestors,
						   long clauseNum,
						   ostream* showNewClause);
    Term* shareTmpLiteral(const TemporaryLiteral& lit);
    void writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
					    BasicTempLitList& lits);  
    void writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
					    const LitHandle& lit);
    Term* Integrate(const Flatterm* lit);
    Term* IntegrateEq(const Term& header,const Flatterm* arg1,const Flatterm* arg2);
    void removeFromSharing(Term* lit);
  private:
    LiteralSharing* _litSharing;
    VariableRenaming _intVarRenaming;
    VariableWeighting _varWeighting;
    const LiteralSelection* _litSelection;
    bool _inheritNegativeSelection;
    bool _equalityIsPossible;
    long _maxMaxWeight;
    long _maxWeight;
  }; 
}; 
#endif
