/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CLAUSE_BACKGROUND_HPP
#define CLAUSE_BACKGROUND_HPP
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#ifndef NO_DEBUG
#  include "ObjDesc.hpp"
#endif
#include "GlobAlloc.hpp"
#include "ExpandingStack.hpp"
#include "BitWord.hpp"
#include "GList.hpp"
#include "DestructionMode.hpp"
#include "TaggedPtr.hpp"
namespace VK
{
  class Clause;
}; 
using namespace std;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_BACKGROUND
#define DEBUG_NAMESPACE "ClauseBackground"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Ancestor {
  public:
    Ancestor() {};
    explicit Ancestor(Clause* cl) : _content(cl) {};
    Ancestor(Clause* cl,bool dummy) : _content(cl,BK::PtrTag01) {};
    Clause* strip() const {
      return _content.strip();
    };
    Clause* strip(bool& parent) const {
      BK::PtrTag tag;
      Clause* res = _content.strip(tag);
      parent = (tag == BK::PtrTag01);
      return res;
    };
    Clause* rawPtr() const { return _content.rawContent(); };
  private:
    BK::TaggedPtr<Clause> _content;
  };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_BACKGROUND
#  define DEBUG_NAMESPACE "ClauseBackground"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ClauseBackground
    {
    public:
      enum Rule 
      { 
	Input = 0, 
	Prepro = 1, 
	BinRes = 2, 
	ForwTransChain = 3,
	BackTransChain = 4,
	ForwSup = 5,
	BackSup = 6,
	EqRes = 7,
	EqFact = 8, 
	ForwDemod = 9,
	BackDemod = 10,
	EqResSimp = 11,
	ForwSubsRes = 12,
	BackSubsRes = 13,  
	Reanimation = 14,
	ExternalEvaluation = 15,
	NameIntro = 16,
        Folding = 17,
	Clausification = 18,
	PartOfNameIntroForPropSplitting = 19,
	PropSplitting = 20,
	Unspecified = 21,
	NonSubsuming = 27,
	NonSubsumable = 28,
	Support = 29,
	Subgoal = 30,
	VIP = 31
      };
      typedef BK::BitWord Rules;
      typedef BK::GList<BK::GlobAlloc,Ancestor,ClauseBackground> AncestorList;
    public:
      ClauseBackground() : 
	_ancestors(0),
	_inferenceDepth(0L)
	{
	};
      ~ClauseBackground()
	{
	  CALL("desctructor ~ClauseBackground()");
	  if (BK::DestructionMode::isThorough()) 
	    {
	      if (!isInput()) AncestorList::destroyList(_ancestors);
	    };
	}; 
      const Rules& properties() const { CALL("properties() const");  return _rules; };
      void SetRules(const Rules& r) { CALL("SetRules(Rules r");  _rules = r; };
      void MakeVIP() { CALL("MakeVIP()");  _rules.set(VIP); };
      bool IsVIP() const { CALL("MakeVIP()");  return _rules.bit(VIP); };
      bool isInput() const { return _rules.bit(Input); };
      bool isPrepro() const { return _rules.bit(Prepro); };
      bool isSubgoal() const { return _rules.bit(Subgoal); };
      void markAsSubgoal() { _rules.set(Subgoal); };
      bool isSupport() const { return _rules.bit(Support); };
      void markAsSupport() { _rules.set(Support); };
      void markAsQuery() { _rules.clear(Support); };
      bool isNonSubsuming() const { return _rules.bit(NonSubsuming); };
      void markAsNonSubsuming() { _rules.set(NonSubsuming); };
      bool isNonSubsumable() const { return _rules.bit(NonSubsumable); };
      void markAsNonSubsumable() { _rules.set(NonSubsumable); };
      bool isNameIntro() const { return _rules.bit(NameIntro); };
      void markAsNameIntro() { _rules.set(NameIntro); };
      bool isPartOfNameIntroForPropSplitting() const {
	return _rules.bit(PartOfNameIntroForPropSplitting);
      };
      bool isProducedByPropSplitting() const {
	return _rules.bit(PropSplitting);
      };
      void markAsProducedByPropSplitting() {
        _rules.set(PropSplitting);
      };
      AncestorList* ancestors() 
	{
	  CALL("ancestors()");
	  ASSERT(!isInput());
	  return _ancestors; 
	};
      const AncestorList* ancestors() const 
	{
	  CALL("ancestors() const");
	  ASSERT(!isInput());
	  return _ancestors; 
	};
      void* inputClauseOrigin() const
	{
	  CALL("inputClauseOrigin() const");
	  ASSERT(isInput());
	  return _inputClauseOrigin; 
	};
      void setInputClauseOrigin(void* orig)
	{
	  CALL("setInputClauseOrigin()");
	  ASSERT(isInput());
	  _inputClauseOrigin = orig; 
	};
      long inferenceDepth() const { return _inferenceDepth; };
      void setInferenceDepth(long d) { _inferenceDepth = d; };
      void copyTo(ClauseBackground& copy) const;
      static ostream& outputRules(ostream& str,const BK::BitWord& _rules);
    private:
      BK::BitWord _rules;
      union
      {
	AncestorList* _ancestors;
	void* _inputClauseOrigin;
      };
      long _inferenceDepth;
      friend class OpenClauseBackground;
    }; 
}; 
#ifdef DEBUG_CLAUSE_BACKGROUND
#undef DEBUG_NAMESPACE
#  define DEBUG_NAMESPACE "OpenClauseBackground"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class OpenClauseBackground
    {
    public:
      OpenClauseBackground() 
	: _ancestors(DOP("OpenClauseBackground::_ancestors"))
	{
	  DOP(_ancestors.freeze());
	};
      ~OpenClauseBackground() {};
      void Reset() 
	{ 
	  _rules.clear();
	  _ancestors.reset(); 
	  _inferenceDepth = 0L;
	};
      void MakeVIP() { _rules.set(ClauseBackground::VIP); };
      void UsedRule(ClauseBackground::Rule r) { _rules.set(r); };
      void UsedRules(const ClauseBackground::Rules& rs) { _rules |= rs; };
      void CancelRule(ClauseBackground::Rule r) { _rules.clear(r); };
      void markAsSubgoal() { _rules.set(ClauseBackground::Subgoal); };
      bool isSubgoal() const { return _rules.bit(ClauseBackground::Subgoal); };
      void markAsSupport() { _rules.set(ClauseBackground::Support); };
      bool isSupport() const { return _rules.bit(ClauseBackground::Support); };
      bool isNonSubsuming() const { return _rules.bit(ClauseBackground::NonSubsuming); };
      void markAsNonSubsuming() { _rules.set(ClauseBackground::NonSubsuming); };
      bool isNonSubsumable() const { return _rules.bit(ClauseBackground::NonSubsumable); };
      void markAsNonSubsumable() { _rules.set(ClauseBackground::NonSubsumable); };
      bool isPartOfNameIntroForPropSplitting() const {
	return _rules.bit(ClauseBackground::PartOfNameIntroForPropSplitting);
      };
      void pushParent(Clause* cl,long depth) 
	{
	  CALL("pushParent(Clause* cl)");
	  DOP(_ancestors.unfreeze());
	  _ancestors.pushSafe(Ancestor(cl,true));
	  DOP(_ancestors.freeze()); 
	  if (depth + 1 > _inferenceDepth) _inferenceDepth = depth + 1;
	};
      void pushSimplifier(Clause* cl,long depth) 
	{
	  CALL("pushSimplifier(Clause* cl)");
	  DOP(_ancestors.unfreeze());
	  _ancestors.pushSafe(Ancestor(cl));
	  DOP(_ancestors.freeze()); 
	  if (depth + 1 > _inferenceDepth) _inferenceDepth = depth + 1;
	};
      void pushSimplified(Clause* cl,long depth) 
	{
	  CALL("pushSimplified(Clause* cl)");
	  DOP(_ancestors.unfreeze());
	  _ancestors.pushSafe(Ancestor(cl));
	  DOP(_ancestors.freeze()); 
	  if (depth + 1 > _inferenceDepth) _inferenceDepth = depth + 1;
	};
      void writeNewBackgroundTo(ClauseBackground& b);
      const BK::BitWord& Rules() const { return _rules; };
      ulong numberOfAncestors() const { return _ancestors.size(); }; 
      const Ancestor* begin() const { return _ancestors.begin(); };
      const Ancestor* end() const { return _ancestors.end(); };
      const Ancestor* last() const { return _ancestors.last(); };
      long inferenceDepth() const { return _inferenceDepth; };
    private:
      ClauseBackground::Rules _rules;
      BK::ExpandingStack<BK::GlobAlloc,Ancestor,BK::FastRelocation<Ancestor>,128UL> _ancestors;
      long _inferenceDepth;
    }; 
}; 
#endif
