/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ClauseNumIndex.hpp"
#include "Clause.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_NUM_INDEX
 #define DEBUG_NAMESPACE "ClauseNumIndex"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
void ClauseNumIndex::insert(Clause* cl)
{
 CALL("insert(Clause* cl)");
 ASSERT(!isLocked());
 lock();
 bool newNode; 
 List::Node* node = _list.insert(cl->Number(),newNode);
 ASSERT(newNode);
 node->value() = cl;
 ASSERT(!cl->flag(Clause::FlagInClauseNumIndex));
 cl->setFlag(Clause::FlagInClauseNumIndex);
 unlock();
}; 
void ClauseNumIndex::remove(Clause* cl)
{
 CALL("remove(Clause* cl)");
 ASSERT(!isLocked());
 lock();
 ASSERT(cl->flag(Clause::FlagInClauseNumIndex));
 Clause* removedClause;
 ALWAYS(_list.remove(cl->Number(),removedClause));
 ASSERT(removedClause == cl);
 ASSERT(cl->flag(Clause::FlagInClauseNumIndex));
 cl->clearFlag(Clause::FlagInClauseNumIndex);
 unlock();
}; 
Clause* ClauseNumIndex::find(ulong num) 
{
 CALL("find(ulong num)");
 ASSERT(!isLocked());
 lock();
 _surfer.reset(_list);
 const List::Node* node;
 if (_surfer.find(num,node)) 
   {
     unlock();
     return node->value();
   };
 unlock();
 return 0;
}; 
