/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CLAUSE_SET_HPP
#define CLAUSE_SET_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Receiver.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "Clause.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_SET
#define DEBUG_NAMESPACE "ClauseSet"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ClauseSet : public BK::Receiver< Clause* >
  {
  public:
    class ConstIter
    {
    public:
      ConstIter(const ClauseSet& set)
      {
	CALL("ConstIter::constructor ConstIter(const ClauseSet& set)");
	_current = (set.Empty()) ? 0 : set.clauses.Next();
	_end = &(set.clauses);
      };
      bool isEnd() const { return _current == 0; };
      void next()
      {
	CALL("ConstIter::next()");
	if (_current)
	  {
	    _current = _current->Next();
	    if (_current == _end) _current = 0;
	  };
      };
      const Clause* clause() const { return _current; };
    private:
      const Clause* _current;
      const Clause* _end;
    }; 
    class Iter
    {
    public:
      Iter(ClauseSet& set)
      {
	CALL("Iter::constructor Iter(ClauseSet& set)");
	_current = (set.Empty()) ? 0 : set.clauses.Next();
	_end = &(set.clauses);
      };
      bool isEnd() const { return _current == 0; };
      void next()
      {
	CALL("Iter::next()");
	if (_current)
	  {
	    _current = _current->Next();
	    if (_current == _end) _current = 0;
	  };
      };
      Clause* clause() const { return _current; };
    private:
      Clause* _current;
      Clause* _end;
    }; 
  public:
    ClauseSet(const Clause::MainSet& tag);
    ClauseSet(const Clause::MainSet& tag,const char* nm);
    ~ClauseSet();
    bool Empty() const { return (clauses.Next() == (&clauses)); };
    bool Nonempty() const { return (clauses.Next() != (&clauses)); };
    void Insert(Clause* c);
    void Remove(Clause* c);
    void receive(Clause* const & cl);
    long Size() const { return numOfClauses; };
    long Removed() const { return removed; };
    long Total() const { return (Size() + Removed()); };
    bool mainIteratorIsEngaged() const 
    {
      return _mainIteratorIsEngaged;
    };
    void resetMainIterator() 
    {
      CALL("resetMainIterator()");
      ASSERT(!_mainIteratorIsEngaged);
      _mainIteratorIsEngaged = true;
      current = Empty() ? 0 : clauses.Next(); 
    };
    void releaseMainIterator()
    {
      _mainIteratorIsEngaged = false;
    };
    Clause* Next();
    Clause* Top();
    Clause* Pop();
    void Purge();
    void DestroyAllClauses();
#ifndef NO_DEBUG_VIS
    ostream& outputClauses(ostream& str);
#endif
  private: 
    Clause::MainSet _tag;
    Clause clauses;
    Clause* current;
    long numOfClauses;
    long removed;
    bool _mainIteratorIsEngaged;
#ifdef DEBUG_NAMESPACE
    const char* name;
#endif
    friend class ConstIter;
    friend class Iter;
  }; 
}; 
#endif
