/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ComponentName.hpp"
#include "Signature.hpp"
#include "GlobalStopFlag.hpp"
#include "TempLitListExtras.hpp"
#include "TemporaryLiteral.inl"
#include "ClauseAssembler.hpp"
#include "Clause.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
#  define DEBUG_NAMESPACE "ComponentName"
#endif 
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
ComponentName::ComponentName(ClauseAssembler* nameIntroAssembler,
			     BK::Receiver<Clause*>* nameIntroReceiver) : 
  _nameIntroAssembler(nameIntroAssembler),
  _nameIntroReceiver(nameIntroReceiver),
  variantIndex(subst), 
  backtrack_lit(DOP("backtrack_lit")),
  backtrack_instr(DOP("backtrack_instr")),
  backtrack_cursor(DOP("backtrack_cursor"))
{
  CALL("constructor ComponentName()");
  TemporaryLiteral::Extras::freshLiteral(dummyTmpLiteral);
  dummyComponent.push(dummyTmpLiteral);
  DOP(backtrack_lit.freeze());
};
ComponentName::~ComponentName() 
{ 
  CALL("destructor ~ComponentName()");
};
bool 
ComponentName::findName(BasicTempLitList& component,Clause*& nameIntroClause)
{
  CALL("findName(BasicTempLitList& component,Clause*& nameIntroClause)");
  if (!query.Load(component,
		  TempLitListExtras::countDifferentVariables(component)))
    return false;
  void* indexedObject;
  bool res = findVariant(indexedObject);
  nameIntroClause = static_cast<Clause*>(indexedObject);
  ASSERT(!res || nameIntroClause);
  return res;
}; 
Clause* 
ComponentName::registerNameIntro(const BasicTempLitList& component,
				 const TemporaryLiteral& definedLit)
{
  CALL("registerNameIntro(const BasicTempLitList& component,..)");
  Clause* res = 
    _nameIntroAssembler->assembleNameIntro(definedLit,component);
  res->markAsProducedByPropSplitting();
  _nameIntroReceiver->receive(res);
  variantIndex.Integrate(component,
			 static_cast<void*>(res));
  return res;
}; 
Clause* ComponentName::registerNameIntro(const Flatterm* term,
					 const TemporaryLiteral& definedLit,
					 const BasicTempLitList& definitionBody)
{
  CALL("registerNameIntro(const Flatterm* term,..)");  
  Clause* res = 
    _nameIntroAssembler->assembleNameIntro(definedLit,definitionBody);
  _nameIntroReceiver->receive(res);
  makeDummyLit(term);
  variantIndex.Integrate(dummyComponent,
			 reinterpret_cast<void*>(res));
  return res;
}; 
inline 
bool 
ComponentName::findVariant(void*& indexedObject)
{ 
  CALL("ComponentName::findVariant(void*& indexedObject)");
  ForwardSubsumptionCommand* curr_instr = variantIndex.Tree();
  if (!curr_instr) return false;
  query.EndOfQuery();
  const Flatterm* cursor = 0;
  Term curr_sym;
  Term curr_sym2;
  Query::Literal* curr_lit;
  Query::Literal* next_lit;
  DOP(backtrack_lit.unfreeze());
  if (!backtrack_lit.tryToExpand(variantIndex.maxlNumOfLits()))
    return false;
  DOP(backtrack_lit.freeze());
  backtrack_lit.reset();
  backtrack_instr.reset();
  backtrack_cursor.reset();
 check_tag:
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::LitHeader :
      curr_lit = query.FirstLit(curr_instr->HeaderNum());
    check_lit:
      if (curr_lit)
        {
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_lit;
	    }
	  else
	    { 
	      curr_lit->Capture();
	      backtrack_lit.push(curr_lit); 
	      backtrack_instr.push(curr_instr);
	      cursor = curr_lit->Args();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    };
        }
      else 
        goto try_fork;
    case ForwardSubsumptionCommand::OrdEq :
      curr_lit = query.FirstLit(curr_instr->HeaderNum());
    check_ord_eq_lit:
      if (curr_lit)
        { 
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_ord_eq_lit;
	    }
	  else
	    { 
	      curr_lit->Capture();
	      backtrack_lit.push(curr_lit); 
	      backtrack_instr.push(curr_instr);
	      cursor = curr_lit->Arg1();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    };
        }
      else 
        goto try_fork;
    case ForwardSubsumptionCommand::UnordEq : 
      curr_lit = query.FirstLit(curr_instr->HeaderNum());
    check_unord_eq_lit:
      if (curr_lit)
        {
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_unord_eq_lit;
	    }
	  else
	    {
	      curr_lit->Capture(); 
	      backtrack_lit.push(curr_lit);
	      curr_instr->SwapOn();
	      backtrack_instr.push(curr_instr);
	      cursor = curr_lit->Arg1();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    };
        }
      else 
        goto try_fork; 
    case ForwardSubsumptionCommand::PropLit :
      if (query.FirstLit(curr_instr->HeaderNum())) 
	{
	  if (curr_instr->Fork()) 
	    {           
	      backtrack_instr.push(curr_instr);
	    }; 
	  curr_instr = curr_instr->Next(); 
	  goto check_tag;
	}
      else 
	goto try_fork; 
    case ForwardSubsumptionCommand::SecondEqArg :
      if (cursor->IsBackJump()) 
	{
	  cursor = cursor->after();
	}; 
      curr_instr = curr_instr->Next(); 
      goto check_tag; 
    case ForwardSubsumptionCommand::Func :
      curr_sym = cursor->symbolRef();
      if (curr_sym.isVariable()) goto backtrack;
    compare_func:
      if (curr_sym == curr_instr->func()) 
        {
	  cursor++;
	  curr_instr = curr_instr->Next();
	  goto check_tag; 
        };
      if (curr_sym < curr_instr->func()) goto backtrack;
      curr_instr = curr_instr->Fork();
      if (!curr_instr) goto backtrack;
      if (curr_instr->IsFunc()) 
	{
          goto compare_func;
	};
      goto check_tag;
    case ForwardSubsumptionCommand::Var :
      if (!cursor->isVariable()) goto try_fork;
      *(curr_instr->var()) = cursor;
      cursor = cursor->after();
      curr_instr = curr_instr->Next();
      goto check_tag;
    case ForwardSubsumptionCommand::CompareNF :
      ASSERT((*(curr_instr->Var1()))->isVariable());
      ASSERT((*(curr_instr->Var2()))->isVariable());
      if ((*(curr_instr->Var1()))->symbolRef() == 
          (*(curr_instr->Var2()))->symbolRef())
	{
	  curr_instr = curr_instr->Next(); 
	  goto check_tag;     
	}
      else goto backtrack;
    case ForwardSubsumptionCommand::Compare :
      ASSERT((*(curr_instr->Var1()))->isVariable());
      ASSERT((*(curr_instr->Var2()))->isVariable());
      if ((*(curr_instr->Var1()))->symbolRef() == 
          (*(curr_instr->Var2()))->symbolRef())
	{
	  if (curr_instr->Fork()) 
	    {
	      backtrack_instr.push(curr_instr);
	      backtrack_cursor.push(cursor);
	    };
	  curr_instr = curr_instr->Next(); 
	  goto check_tag; 
	}
      else goto try_fork;
    case ForwardSubsumptionCommand::Success : 
      if ((curr_instr->NumOfDiffVars() == query.NumOfDiffVars())
	  && (curr_instr->NumOfLits() == query.NumOfLits()))
	{
	  indexedObject = curr_instr->IndexedObject();
	  return true;
	}
      else goto try_fork; 
#ifdef DEBUG_NAMESPACE  
    default: ICP("ICP1"); return false; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: return false; 
#endif
#endif  
    }; 
 try_fork:
  curr_instr = curr_instr->Fork();
  if (curr_instr) goto check_tag;
 backtrack: 
  if (backtrack_instr.empty())
    {
      return false;
    }; 
  curr_instr = backtrack_instr.pop();
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::LitHeader :
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();
      ASSERT(curr_lit); 
      ASSERT(curr_lit->Captured());
      curr_lit->Release();
      curr_lit = curr_lit->Next();       
      goto check_lit;
    case ForwardSubsumptionCommand::OrdEq :
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();
      ASSERT(curr_lit); 
      ASSERT(curr_lit->Captured());
      curr_lit->Release();
      curr_lit = curr_lit->Next();       
      goto check_ord_eq_lit;      
    case ForwardSubsumptionCommand::UnordEq :
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();
      ASSERT(curr_lit);
      ASSERT(curr_lit->Captured());
      if (curr_instr->SwapFlag()) 
	{        
	  next_lit = curr_lit->Next();
	  backtrack_lit.push(curr_lit);
	  curr_instr->SwapOff();
	  backtrack_instr.push(curr_instr);
	  cursor = curr_lit->Arg2();
	  curr_instr = curr_instr->Next();
	  goto check_tag;
	}
      else
	{       
	  curr_lit->Release();        
	  curr_lit = curr_lit->Next(); 
	  goto check_unord_eq_lit;
	}; 
    case ForwardSubsumptionCommand::PropLit :
      curr_instr = curr_instr->Fork();
      goto check_tag; 
    case ForwardSubsumptionCommand::Compare :
      curr_instr = curr_instr->Fork();
      cursor = backtrack_cursor.pop();
      goto check_tag;
#ifdef DEBUG_NAMESPACE  
    default: ICP("ICP2"); return false; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: return false; 
#endif
#endif 
    };
  ICP("ICP3");
    return false;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
#define DEBUG_NAMESPACE "ComponentName::Query"
#endif 
#include "debugMacros.hpp"
ComponentName::Query::Query() : 
  _litsWithHeader(static_cast<Literal*>(0),"ComponentName::Query::_litsWithHeader"),
  _posOrdEqHeaderNum(Term::positiveOrdEqHeader().HeaderNum()),
  _negOrdEqHeaderNum(Term::negativeOrdEqHeader().HeaderNum()),
  _presentHeaders(DOP("ComponentName::Query::_presentHeaders"))
{
  CALL("constructor ComponentName::Query()");
  DOP(_litsWithHeader.freeze());
};
void ComponentName::Query::ResetQuery()
{
  CALL("ResetQuery()");
  _litCache.reset();
  while (_presentHeaders.nonempty())
    {
      _litsWithHeader[_presentHeaders.pop()] = 0;
    };
  last_pos_unord_eq = 0;
  last_neg_unord_eq = 0;
  numOfLits = 0;
}; 
inline void ComponentName::Query::LinkEquations()
{
  CALL("LinkEquations()") 
    if (last_pos_unord_eq) 
      last_pos_unord_eq->SetNext(FirstLit(_posOrdEqHeaderNum));
  if (last_neg_unord_eq) 
    last_neg_unord_eq->SetNext(FirstLit(_negOrdEqHeaderNum));
}; 
void ComponentName::Query::EndOfQuery()
{
  LinkEquations();
}; 
bool ComponentName::Query::PushLit(const Flatterm* lit)
{
  CALL("PushLit(const Flatterm* lit)");
  ASSERT(!lit->symbolRef().IsEquality());
  DOP(_litsWithHeader.unfreeze());
  Literal** lit_lst_ptr = &(_litsWithHeader.sub(lit->symbolRef().HeaderNum()));
  DOP(_litsWithHeader.freeze());
  Literal* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;
  if (*lit_lst_ptr)
    {
      currLit->Set(lit + 1,*lit_lst_ptr);
    }
  else 
    {
      _presentHeaders.pushSafe(lit->symbolRef().HeaderNum());
      currLit->Set(lit + 1,static_cast<Literal*>(0));
    };
  *lit_lst_ptr = currLit; 
  currLit->Release(); 
  return true;
}; 
bool ComponentName::Query::PushUnordEq(Term header,const Flatterm* arg1,const Flatterm* arg2)
{
  CALL("PushUnordEq(Term header,const Flatterm* arg1,const Flatterm* arg2)");
  ASSERT(header.IsUnorderedEq());
  DOP(_litsWithHeader.unfreeze());
  Literal** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
  DOP(_litsWithHeader.freeze());
  Literal* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;
  if (*lit_lst_ptr)
    {
      currLit->Set(arg1,arg2,*lit_lst_ptr);
    }
  else 
    {
      _presentHeaders.pushSafe(header.HeaderNum());
      currLit->Set(arg1,arg2,static_cast<Literal*>(0));
      if (header.Negative()) { last_neg_unord_eq = currLit; }
      else 
	last_pos_unord_eq = currLit;
    };
  *lit_lst_ptr = currLit;
  currLit->Release();
  return true;
}; 
bool ComponentName::Query::PushOrdEq(Term header,const Flatterm* arg1,const Flatterm* arg2)
{
  CALL("PushOrdEq(Term header,const Flatterm* arg1,const Flatterm* arg2)");
  ASSERT(header.IsEquality() && header.IsOrderedEq());
  DOP(_litsWithHeader.unfreeze());
  Literal** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
  DOP(_litsWithHeader.freeze());
  Literal* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;
  if (*lit_lst_ptr)
    {
      currLit->Set(arg1,arg2,*lit_lst_ptr);
    }
  else 
    {
      _presentHeaders.pushSafe(header.HeaderNum());
      currLit->Set(arg1,arg2,static_cast<Literal*>(0));
    };
  *lit_lst_ptr = currLit;
  currLit->Release(); 
  return true;
}; 
bool 
ComponentName::Query::Load(BasicTempLitList& component,ulong numOfVars)
{
  CALL("Load(BasicTempLitList& component,ulong numOfVars)");
  ResetQuery(); 
  numOfDiffVars = numOfVars;
  numOfLits = component.length();
  for (BasicTempLitList::Element* el = component.first();
       el;
       el = el->next())
    {
      if (el->literal()->isEquality())
	{	      
	  el->literal()->linkEqArgs();
	  if (el->literal()->isOrderedEquality())   
	    {
	      if (!PushOrdEq(el->literal()->header(),
			     el->literal()->arg1(),
			     el->literal()->arg2()))
		return false;
	    }      
	  else 
	    {
	      if (!PushUnordEq(el->literal()->header(),
			       el->literal()->arg1(),
			       el->literal()->arg2()))
		return false;
	    };
	} 
      else 
	{
	  if (!PushLit(el->literal()->flatterm()))
	    return false;
	};
    };
  return true;
}; 
