/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef CUSTOM_LITERAL_SELECTION_HPP
#define CUSTOM_LITERAL_SELECTION_HPP 
#include "jargon.hpp"
#include "RuntimeError.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "StandardLiteralOrdering.hpp"
#include "ExpandingStack.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef BasicTempLitList LitList;
    typedef BasicTempLitList::Element LitListElement;
    typedef TempLitListExtras::StdCharacteristics LitListStatistics;
  public:
    typedef long (*NumericLiteralCharacteristic)(TemporaryLiteral&);
    typedef bool (*LiteralProperty)(TemporaryLiteral&);
    class ComplexLiteralCharacteristic;
    class InverseVarDepthProfile;
    class Filter;
    class MaxValFilter;
    class MinValFilter;
    class MaxComplexValFilter;
    class AllPOrAllFilter;
    class CompositionFilter;
    class OneOfEachPolarityFilter;
    class SelectionAcceptability;
    class ConjOfPrimitiveSelectionAcceptability;
  public:
    CustomLiteralSelection() {};
    virtual ~CustomLiteralSelection();
    void setMainLiteralOrdering(StandardLiteralOrdering* ord) { 
      _mainLiteralOrdering = ord; };
    void select(LitList& lits,
		LitList& selectedLits,
		LitList& unselectedLits,
		bool& allMaximalSelected,
		const LitListStatistics& stat); 
    static long literalSize(TemporaryLiteral& lit);
    static long numOfDifferentVariables(TemporaryLiteral& lit);
    static bool isGround(TemporaryLiteral& lit);
    static bool isNegativeEquality(TemporaryLiteral& lit);
    static bool isNotPositiveEquality(TemporaryLiteral& lit);
    static bool isNotAvailableForParamodFrom(TemporaryLiteral& lit);
    static bool oneLiteralSelected(LitListElement* firstSelected,
				   LitListElement* firstUnselected); 
    static bool noBigNegativeUnselected(LitListElement* firstSelected,
					LitListElement* firstUnselected); 
    static bool noGroundNegativeUnselected(LitListElement* firstSelected,
					   LitListElement* firstUnselected); 
    static bool noNegativeEqualitiesUnselected(LitListElement* firstSelected,
					       LitListElement* firstUnselected); 
    static bool noNegativeWithManyDiffVarsUnselected(LitListElement* firstSelected,
						      LitListElement* firstUnselected);
  protected:
    void selectMaximal(LitList& lits,
		       LitList& selectedLits,
		       LitList& unselectedLits);
    void selectOneMaximal(LiteralOrdering* litOrd,
		       LitList& lits,
		       LitList& selectedLits,
		       LitList& unselectedLits);
    void unselectPositive(LitList& lits,LitList& unselectedLits,ulong numOfPositive);
  protected:
    StandardLiteralOrdering* _mainLiteralOrdering;
  private:
    virtual
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat) = 0;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ComplexLiteralCharacteristic"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::ComplexLiteralCharacteristic
  {
  public:
    ComplexLiteralCharacteristic() {};
    virtual ~ComplexLiteralCharacteristic();
    virtual void computeFor(TemporaryLiteral& lit) = 0;
    virtual BK::Comparison compareWith(ComplexLiteralCharacteristic* c) = 0;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::InverseVarDepthProfile"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::InverseVarDepthProfile : public ComplexLiteralCharacteristic
  {
  public:
    InverseVarDepthProfile() {};
    ~InverseVarDepthProfile();
    void computeFor(TemporaryLiteral& lit);
    BK::Comparison compareWith(ComplexLiteralCharacteristic* c);
  private:
    TemporaryLiteral::VarDepthProfile _data;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::Filter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::Filter
  {
  public:
    Filter() : _alwaysSelectOne(false) {};
    virtual ~Filter();
    void setSelectOne(bool fl) { _alwaysSelectOne = fl; };
    void select(LitList& lits,
		LitList& selectedLits,
		LitList& unselectedLits)
    {
      CALL("select(..)");
      ASSERT(!lits.isEmpty());
      DOP(ulong debugNumOfLits = lits.length());
      DOP(ulong debugNumOfSelectedLits = selectedLits.length());
      DOP(ulong debugNumOfUnselected = unselectedLits.length());
      if (_alwaysSelectOne)
	{
	  selectOne(lits,selectedLits,unselectedLits);
	  ASSERT(selectedLits.length() - debugNumOfSelectedLits == 1UL);
	}
      else
	{
	  selectAll(lits,selectedLits,unselectedLits);
	  ASSERT(selectedLits.length() > debugNumOfSelectedLits);
	};
      ASSERT(lits.isEmpty());
      ASSERT(selectedLits.length() + 
	     unselectedLits.length() 
	     ==
	     debugNumOfLits + 
	     debugNumOfSelectedLits + 
	     debugNumOfUnselected);
    };
  private:
    virtual void selectOne(LitList& lits,
			   LitList& selectedLits,
			   LitList& unselectedLits) = 0;
    virtual void selectAll(LitList& lits,
			   LitList& selectedLits,
			   LitList& unselectedLits) = 0;
  private:
    bool _alwaysSelectOne;
    friend class CustomLiteralSelection::CompositionFilter;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxValFilter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::MaxValFilter : 
    public Filter
  {
  public:
    MaxValFilter(NumericLiteralCharacteristic val) : _val(val)
    {
    };
    ~MaxValFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    NumericLiteralCharacteristic _val;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MinValFilter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::MinValFilter : 
    public Filter
  {
  public:
    MinValFilter(NumericLiteralCharacteristic val) : _val(val)
    {
    };
    ~MinValFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    NumericLiteralCharacteristic _val;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxComplexValFilter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::MaxComplexValFilter : 
    public Filter
  {
  public:
    MaxComplexValFilter(ComplexLiteralCharacteristic* tmpVal1,
			ComplexLiteralCharacteristic* tmpVal2) : 
      _tmpVal1(tmpVal1),
      _tmpVal2(tmpVal2)
    {
    };
    ~MaxComplexValFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    ComplexLiteralCharacteristic* _tmpVal1;
    ComplexLiteralCharacteristic* _tmpVal2;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::AllPOrAllFilter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::AllPOrAllFilter : 
    public Filter
  {
  public:
    AllPOrAllFilter(LiteralProperty prop) : _prop(prop)
    {
    };
    ~AllPOrAllFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    LiteralProperty _prop;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::CompositionFilter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::CompositionFilter : 
    public Filter
  {
  public:
    CompositionFilter() 
    {
    };
    ~CompositionFilter();
    void resetFilters() { _filters.reset(); };
    void addFilter(Filter* f) { _filters.pushSafe(f); };
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    BK::ExpandingStack<BK::GlobAlloc,Filter*,BK::FastRelocation<Filter*>,10UL> _filters;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::OneOfEachPolarityFilter"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::OneOfEachPolarityFilter : 
    public Filter
  {
  public:
    OneOfEachPolarityFilter() 
    {
    };
    ~OneOfEachPolarityFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::SelectionAcceptability"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::SelectionAcceptability
  {
  public:
    SelectionAcceptability() {};
    virtual ~SelectionAcceptability();
    virtual bool accepts(LitListElement* firstSelected,
			 LitListElement* firstUnselected) = 0;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability :
    public SelectionAcceptability
  {
  public:
    ConjOfPrimitiveSelectionAcceptability() 
    {
    };
    ~ConjOfPrimitiveSelectionAcceptability();
    bool accepts(LitListElement* firstSelected,
		 LitListElement* firstUnselected);
    void resetPrimitives() { _primitives.reset(); };
    void addPrimitive(bool (*fun)(LitListElement*,LitListElement*)) 
    {
      _primitives.pushSafe(fun);
    };
  private:
    typedef bool (*Primitive)(LitListElement*,LitListElement*);
  private:
    BK::ExpandingStack<BK::GlobAlloc,Primitive,BK::FastRelocation<Primitive>,10UL> _primitives;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "TotalLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TotalLiteralSelection : public CustomLiteralSelection
  {
  public:
    TotalLiteralSelection() {};
    ~TotalLiteralSelection() {};
  private:    
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaximalLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class MaximalLiteralSelection : public CustomLiteralSelection
  {
  public:
    MaximalLiteralSelection() {};
    ~MaximalLiteralSelection() {};
  private:    
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "HyperLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class HyperLiteralSelection : public CustomLiteralSelection
  {
  public:
    HyperLiteralSelection() {};
    ~HyperLiteralSelection() {};
    void setFilterForNegative(CustomLiteralSelection::Filter* f)
    {
      _filterForNegative = f;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForNegative;    
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaxWithCorrectionLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class MaxWithCorrectionLiteralSelection : public CustomLiteralSelection
  {
  public:
    MaxWithCorrectionLiteralSelection() 
    {
      CALL("constructor MaxWithCorrectionLiteralSelection()");
      DOP(_filterForNegative = 0);
      DOP(_maxSelectionAcceptability = 0);
    };
    ~MaxWithCorrectionLiteralSelection() {};
    void setFilterForNegative(CustomLiteralSelection::Filter* f)
    {
      _filterForNegative = f;
    };
    void setMaxSelectionAcceptability(CustomLiteralSelection::SelectionAcceptability* test)
    {
      _maxSelectionAcceptability = test;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::SelectionAcceptability* _maxSelectionAcceptability;
    CustomLiteralSelection::Filter* _filterForNegative;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "BestOnlyLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BestOnlyLiteralSelection : public CustomLiteralSelection
  {
  public:
    BestOnlyLiteralSelection() {};
    ~BestOnlyLiteralSelection() {};
    void setLiteralQualityOrdering(LiteralOrdering* o)
    {
      _litOrdering = o;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat); 
  private:
    LiteralOrdering* _litOrdering;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "ArbitraryLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ArbitraryLiteralSelection : public CustomLiteralSelection
  {
  public:
    ArbitraryLiteralSelection() {};
    ~ArbitraryLiteralSelection() {};
    void setFilter(CustomLiteralSelection::Filter* f)
    {
      _filter = f;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filter;    
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "IncompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class IncompleteBidirectionalLiteralSelection : public CustomLiteralSelection
  {
  public:
    IncompleteBidirectionalLiteralSelection() {};
    ~IncompleteBidirectionalLiteralSelection() {};
    void setFilterForPositiveInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveInRules = f;
    };
    void setFilterForNegativeInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeInRules = f;
    };
    void setFilterForPositiveFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveFacts = f;
    };
    void setFilterForNegativeFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeFacts = f;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForPositiveInRules; 
    CustomLiteralSelection::Filter* _filterForNegativeInRules;    
    CustomLiteralSelection::Filter* _filterForPositiveFacts;      
    CustomLiteralSelection::Filter* _filterForNegativeFacts;    
    LitList _buffer;
  }; 
}; 
#endif
