/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef DIV_INTO_COMP_HPP
#define DIV_INTO_COMP_HPP
#include "jargon.hpp"
#include "Multiset.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_DIV_INTO_COMP
#  define DEBUG_NAMESPACE "DivIntoComp"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  template <ulong MaxNumOfVars>
  class DivIntoComp
  {
  public:
    DivIntoComp() :
      _singleGroundLitCompMinSize(LONG_MAX),
      _comp(DOP("DivIntoComp::_comp")), 
      _reg()
    {
      CALL("constructor DivIntoComp()"); 
      reset(); 
      DOP(_comp.freeze());
    };
    ~DivIntoComp() {};
    void setSingleGroundLitCompMinSize(long sz) {
      _singleGroundLitCompMinSize = sz; };
    bool adjustSizes(ulong newNumOfLits) {
      CALL("adjustSizes(ulong newNumOfLits)");
      DOP(_comp.unfreeze());
      if (!_comp.tryToExpand(newNumOfLits)) return false; 
      DOP(_comp.freeze());
      return true;
    };
    void reset() {
      CALL("reset()");
      _currLit = -1; 
      _reg.reset();  
      _numOfComps = 0;       
      _smallGroundLitComponent = -1; 
    };
    ulong nextLit(bool ground,ulong groundWeight) {
      CALL("nextLit(bool ground,ulong groundWeight)");
      _currLit++;
      _comp[_currLit] = _currLit; 
      if (ground)
	{     
	  if ((long)groundWeight < _singleGroundLitCompMinSize)
	    {
	      if (_smallGroundLitComponent < 0)
		{
		  _smallGroundLitComponent = _currLit;
		  _numOfComps++;
		}
	      else
		{
		  _comp[_smallGroundLitComponent] = _currLit;
		  _smallGroundLitComponent = _currLit;
		};
	    }
	  else 
	    _numOfComps++;
	} 
      else 
	_numOfComps++;
      return (ulong)_currLit; 
    };
    void nextVar(const ulong& var) {
      CALL("nextVar(const ulong& var)");
      if (registeredVar(var)) 
	{
	  long lt = _lit[var];
	  long tmp;
	  while ((tmp = _comp[lt]) != lt)
	    {  
	      _comp[lt] = _currLit;              
	      lt = tmp;
	    };
	  if (_comp[lt] != _currLit)
	    {
	      _comp[lt] = _currLit;
	      _numOfComps--;
	    };
	  _lit[var] = _currLit; 
	}
      else
	{
	  _lit[var] = _currLit;
	  registerVar(var);
	};
    };
    bool litCompPair(ulong& litNum,ulong& compNum) {
      CALL("litCompPair(ulong& litNum,ulong& compNum)");
      if (_currLit < 0) return false;
      if (_comp[_currLit] == _currLit) 
	{ 
	  litNum = _currLit;
	  compNum = _currLit;
	}
      else 
	{
	  _comp[_currLit] = _comp[_comp[_currLit]];      
	  litNum = _currLit;
	  compNum = _comp[_currLit];
	};
      _currLit--; 
      return true;
    };
    ulong numOfComponents() const { return _numOfComps; };
  private:
    bool registeredVar(const ulong& var) const { return _reg.scoreOf(var) != 0L; };
    void registerVar(const ulong& var) { _reg.add(var); };
  private:
    long _singleGroundLitCompMinSize;
    BK::Array<BK::GlobAlloc,long,BK::DefaultRelocation<long>,32UL> _comp;
    long _lit[MaxNumOfVars];
    BK::Multiset<ulong,MaxNumOfVars> _reg; 
    long _currLit; 
    ulong _numOfComps;
    long _smallGroundLitComponent;
  }; 
}; 
#endif
