/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_DEMODULATION_HPP
#define FORWARD_DEMODULATION_HPP
#ifndef NO_DEBUG_VIS
# include <iostream>
#endif 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "GList.hpp"
#include "DestructionMode.hpp"
#include "RuntimeError.hpp"
#include "ForwardMatchingIndexTree.hpp"
#include "ForwardMatchingIndex.hpp"
#include "ExpandingMultiset.hpp"
#include "Term.hpp"
#include "Flatterm.hpp"
#include "ForwardMatchingSubstitutionCache.hpp"
#include "Tuple.hpp"
#include "SimplificationOrdering.hpp"
namespace VK
{
  class Clause;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
#  define DEBUG_NAMESPACE "ForwardDemodulation"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class ForwardDemodulation
  {
  public:
    typedef BK::ExpandingMultiset<BK::GlobAlloc,ulong,32UL> SetOfSplittingBranches;
    class Index 
    {
    public:
      typedef BK::Pair<ulong,const Flatterm* const *> VarInstanceLocationPair;
      class Id {};
      typedef BK::GList<BK::GlobAlloc,VarInstanceLocationPair,Id> StoredSubstitution;
      class ConstrRule
      {
      public:
	ConstrRule() : 
	  clause(0), 
	  constr(0),
	  subst(0),
	  lhs(0)
	{
	};
	ConstrRule(Clause* cl,
		   SimplificationOrdering::StoredConstraint* con,
		   StoredSubstitution* sub,
		   Term* t) :
	  clause(cl),
	  subst(sub),
	  lhs(t)
	{
	  CALL("constructor ConstrRule(Clause* cl,...)");
	  if (con) 
	    {
	      constr = SimplificationOrdering::current()->copy(con);
	    }
	  else 
	    constr = 0;
	};
	~ConstrRule() 
	{
	  CALL("desructor ~ConstrRule()");
	  if (BK::DestructionMode::isThorough())
	    {
	      if (constr) 
		{
		  SimplificationOrdering::current()->releaseConstraint(constr); 
		};
	      StoredSubstitution::destroyList(subst);
	    };
	};
	void replace(Clause* cl,
		     SimplificationOrdering::StoredConstraint* con,
		     StoredSubstitution* sub,
		     Term* t)
	{
	  if (constr)
	    SimplificationOrdering::current()->releaseConstraint(constr);
	  StoredSubstitution::destroyList(subst);
	  clause = cl;
	  if (con) 
	    {
	      constr = SimplificationOrdering::current()->copy(con);
	    }
	  else 
	    constr = 0;
	  subst = sub;
	  lhs = t;
	}; 
      public:
	Clause* clause;
	SimplificationOrdering::StoredConstraint* constr;
	StoredSubstitution* subst;
	Term* lhs;  
      private:
	ConstrRule(const ConstrRule& cr);
	ConstrRule& operator=(const ConstrRule&);
      };
      typedef BK::GList<BK::GlobAlloc,ConstrRule,Id> ConstrRuleList; 
    public:
      class LeafInfo 
      {
      public:
	LeafInfo() : _rules(0) {};
	~LeafInfo() 
	{
	  CALL("Index::LeafInfo::destructor ~LeafInfo()");
	  if (BK::DestructionMode::isThorough())
	    ConstrRuleList::destroyList(_rules);
	};
        ConstrRuleList*& rules() { return _rules; };
      private:
	ConstrRuleList* _rules; 
      };
    public:
      Index() : _degree(0L)
      {
	CALL("Index::constructor Index::Index()");
	_specialisedArgComparison = 
	  SimplificationOrdering::current()->reserveSpecialisedComparison();
      };
      ~Index() 
      { 
	CALL("Index::destructor ~Index"); 
	if (BK::DestructionMode::isThorough()) 
	  {
	    ASSERT(_specialisedArgComparison);
	    SimplificationOrdering::current()->releaseSpecialisedComparison(_specialisedArgComparison);
	  };
      }; 
      void SetDegree(long d) { _degree = d; }; 
      void integrate(Clause* cl);
      void remove(Clause* cl);
      FMCodeTree<LeafInfo>::Leaf* Match(const Flatterm* query);
      FMCodeTree<LeafInfo>::Leaf* MatchAgain();
    private:
      void integrate(Clause* cl,
		     Term* t,
		     SimplificationOrdering::StoredConstraint* constr);
      FMCodeTree<LeafInfo>::Leaf* Integrate(Term* t);
      bool Remove(Term* t,FMCodeTree<LeafInfo>::Leaf*& leaf,FMCodeTree<LeafInfo>::Instruction**& branch_to_destroy);
      void remove(Clause* cl,Term* lhs);
      bool 
      appropriateDegreeForGreater(const SimplificationOrdering::SpecialisedComparison::Category& cat) const;
      bool 
      appropriateDegreeForLess(const SimplificationOrdering::SpecialisedComparison::Category& cat) const;
      StoredSubstitution* createStoredSubstitution();
    private:  
      ForwardMatchingIndex<LeafInfo> _matchingIndex;
      long _degree;
      SimplificationOrdering::SpecialisedComparison* _specialisedArgComparison;
    }; 
  public:
    ForwardDemodulation();
    ~ForwardDemodulation();
    bool isLocked() const { return _isLocked; };
    void reset();
    void SetDegree(long d) { _degree = d; index.SetDegree(d); };
    void SetFDOnSplittingBranchesFlag(bool fl) 
    {
      if (fDOnSplittingBranchesFlag && (!fl)) BK::RuntimeError::report("Switching fDOnSplittingBranchesFlag off.");
      fDOnSplittingBranchesFlag = fl; 
    }; 
    void setDemodulateFromSupport(bool fl)
    {
      _demodulateFromSupport = fl;
    };
    void Integrate(Clause* cl);
    void Remove(Clause* cl);
    bool Rewrite(const Flatterm* query,
		 const SetOfSplittingBranches& splittingBranches);
    void abandonCurrentQuery();
    Clause* Rule() const 
    {
      CALL("Rule() const");
      ASSERT(isLocked());
      return rule; 
    };
    const Term& RHSTerm() const 
    { 
      CALL("RHSTerm() const");
      ASSERT(isLocked());
      return _ruleRHSTerm; 
    };
    bool RHSNext() 
    { 
      CALL("RHSNext()");
      ASSERT(isLocked());
      return _ruleRHSInstance.next(); 
    };    
    Term RHSSymbol() 
    { 
      CALL("RHSSymbol()");
      ASSERT(isLocked());
      return _ruleRHSInstance.symbol(); 
    };
    bool RHSSubstPosInSubstitution() const 
    { 
      CALL("RHSSubstPosInSubstitution() const");
      ASSERT(isLocked());
      return  _ruleRHSInstance.substPosInSubstitution(); 
    };
    bool SplittingBranchesUsed() const 
    { 
      CALL("SplittingBranchesUsed() const");
      ASSERT(isLocked());
      return branchesUsed; 
    }; 
    const ForwardMatchingSubstitutionCache& substitutionCache() const
    {
      CALL("substitutionCache() cons");
      ASSERT(isLocked());
      return _substCache;
    };
  private:   
    void lock() { _isLocked = true; };
    void unlock() { _isLocked = false; };
    static 
    Index::ConstrRuleList* 
    removeConstrRule(Clause* cl,
		     Term* t,
		     Index::ConstrRuleList* lst);
    static 
    Index::ConstrRuleList* 
    insertConstrRule(Clause* cl,
		     SimplificationOrdering::StoredConstraint* constr,
		     Index::StoredSubstitution* subst,
		     Term* t,
		     Index::ConstrRuleList* lst);
    bool CanBeRewritingRule(const Clause* cl);   
    bool CheckSplittingBranches(const Clause* cl,
				const SetOfSplittingBranches& splittingBranches);
    void collectSubstCache(const Index::StoredSubstitution* storedSubst);
  private: 
    bool _isLocked;
    Index index;
    Clause* rule;
    Term _ruleRHSTerm;
    ForwardMatchingSubstitutionCache::Instance _ruleRHSInstance;
    long _degree;
    bool fDOnSplittingBranchesFlag;
    bool branchesUsed;
    bool _demodulateFromSupport;
    ForwardMatchingSubstitutionCache _substCache;
    SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* _constrCheck;
    friend class Index;
  }; 
}; 
#endif
