/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_MATCHING_INDEX_HPP
#define FORWARD_MATCHING_INDEX_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Multiset.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
#include "DestructionMode.hpp"
#include "ForwardMatchingIndexTree.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "Flatterm.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_INDEX
#define DEBUG_NAMESPACE "ForwardMatchingIndex<LeafInfo>"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  template <class LeafInfo>
  class ForwardMatchingIndex
  { 
  public: 
    typedef 
    typename VK::FMCodeTree<LeafInfo>::Instruction 
    Instruction;
    typedef 
    typename VK::FMCodeTree<LeafInfo>::Leaf 
    Leaf;
  public:
    ForwardMatchingIndex() : 
      tree(0,"ForwardMatchingIndex<..>::tree"),
      integrator(subst), 
      removal(subst), 
      retrieval()
    {
      CALL("constructor ForwardMatchingIndex()");
      DOP(tree.freeze());  
    };
    ~ForwardMatchingIndex() 
    { 
      CALL("destructor ~ForwardMatchingIndex()"); 
      if (BK::DestructionMode::isThorough()) 
	{
	  for (ulong f = 0; f < tree.size(); f++) 
	    {
	      if (tree[f])
		{ 
		  FMCodeTree<LeafInfo>::Destroy(tree[f]);   
		};
	    };	  
	};
    }; 
    inline void integrationReset(const Term& fun)
    {
      CALL("integrationReset(const Term& fun)");
      ASSERT(fun.isComplex());
      DOP(tree.unfreeze());
      integrator.Reset(&(tree.sub(fun.functor())));
      DOP(tree.freeze());
    };    
    inline void integrationVar(const Term& var) 
    {
      CALL("integrationVar(const Term& var)");
      ASSERT(var.isVariable());
      integrator.PushVar(var);
    }; 
    inline void integrationFunc(const Term& fun)
    {
      CALL("integrationFunc(const Term& fun)");
      ASSERT(fun.isComplex());
      integrator.PushFunc(fun);
    };
    Leaf* integrationGetLeaf(bool& newLeaf)  
    {
      CALL("integrationGetLeaf()");
      return integrator.GetLeaf(newLeaf);
    };   
    BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables>& 
    integrationGetVariables() 
    { 
      return integrator.Vars(); 
    };
    const Flatterm** integrationMapVariable(ulong v) { return integrator.MapVar(v); };
    bool removalReset(const Term& fun)
    {
      CALL("removalReset(const Term& fun)");
      ASSERT(fun.isComplex());
      if (fun.functor() >= tree.size()) return false;
      Instruction** tree_addr = &(tree[fun.functor()]);  
      if (!(*tree_addr)) return false;
      removal.Reset(tree_addr);
      return true;
    };
    bool removalVar(const Term& var) 
    {
      CALL("removalVar(const Term& var)");
      ASSERT(var.isVariable());
      return removal.SkipVar(var);
    };
    bool removalFunc(const Term& fun) 
    {
      CALL("removalFunc(const Term& fun)");
      ASSERT(fun.isComplex());
      return removal.SkipFunc(fun);
    };
    bool removalFinish(Leaf*& leaf,Instruction**& branch_to_destroy)
    {
      CALL("removalGetLeaf(Leaf*& leaf,Instruction**& branch_to_destroy)");
      if (!removal.SkipComps()) return false;
      leaf = removal.GetLeaf();
      ASSERT(leaf->isLeaf());   
      branch_to_destroy = removal.BranchToDestroy();
      return true;
    };
    Leaf* Match(const Flatterm* query);
    Leaf* MatchAgain() 
    { 
      CALL("MatchAgain()"); 
      return retrieval.MatchAgain(); 
    };
  private:
    BK::Array<BK::GlobAlloc,Instruction*,BK::DefaultRelocation<Instruction*>,128UL> tree;
    const Flatterm* subst[VampireKernelConst::MaxTermSize];
    typename FMCodeTree<LeafInfo>::Integrator integrator;
    typename FMCodeTree<LeafInfo>::Removal removal;
    typename FMCodeTree<LeafInfo>::Retrieval retrieval;
  }; 
  template <class LeafInfo>
  inline 
  typename FMCodeTree<LeafInfo>::Leaf* 
  ForwardMatchingIndex<LeafInfo>::Match(const Flatterm* query)
  {
    CALL("Match(const Flatterm* query)");
    ASSERT(query->symbolRef().isComplex());
    ulong funcNum = query->symbolRef().functor();
    if (funcNum >= tree.size()) 
      return 0;
    return retrieval.Match(query + 1,tree[funcNum]);    
  }; 
}; 
#endif
