/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_PROCESSING_HPP
#define FORWARD_PROCESSING_HPP
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "Receiver.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "TemporaryClauseCollector.hpp"
#include "TemporaryClauseQueue.hpp"
#include "ForwardDemodulation.hpp"
#include "Statistics.hpp"
namespace VK
{
  class TemporaryClause;
  class TemporaryClauseCollector;
  class ComponentName; 
  class PropositionalSplitting;
  class NegativeEqualitySplitting;
  class SymbolInfo;
  class ForwardSubsumptionFrontEnd;
  class Clause;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_PROCESSING
#  define DEBUG_NAMESPACE "ForwardProcessing"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class ForwardProcessing
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef TemporaryClause::Extras::Handle ClauseHandle;
    typedef BasicTempLitList::Element LitListElement;
    enum RejectionReason
    {
      None,
      PropositionalTautology,
      EquationalTautology,
      ClauseTooLarge,
      LiteralTooLarge,
      TooManyLiterals,
      TooManyVariables,
      LiteralTooDeep,
      InferenceTooDeep,
      Subsumed,
      ReducedInNonBasicPosition
    };
  public:
    ForwardProcessing();
    ~ForwardProcessing();
    void reset();
    void setShowAnnouncedClauses(ostream* str)
    {
      _internalClauseCollector.setShowAnnouncedClauses(str);
    };
    void setShowCompleteClauses(ostream* str)
    {
      _internalClauseCollector.setShowCompleteClauses(str);
    };
    void setPreprocessingInProgress(bool fl)
    {
      _preprocessingInProgress = fl;
      _internalClauseCollector.setPreprocessingInProgress(fl);
    };
    void setStatistics(Statistics* stat) 
    { 
      _statistics = stat;
      _internalClauseCollector.setStatistics(stat);
    };
    void setEqualityPossible(bool fl) { _equalityPossible = fl; };
    void setForwardSubsumption(ForwardSubsumptionFrontEnd* fs) 
    { 
      _forwardSubsumption = fs; 
    };
    void setForwardSubsumptionResolution(ulong minLength) 
    {
      _forwardSubsumptionResolution = (minLength != 0UL); 
      _minLengthOfSubsumptionResolutionTarget = minLength;
    };
    void setForwardDemodulation(ForwardDemodulation* fd) 
    { 
      _forwardDemodulation = fd; 
    };
    void setOnBranchForwardDemodulation(bool fl) 
    { 
      _onBranchForwardDemodulation = fl; 
    };
    void setDestructiveEqResolution(bool fl) 
    { 
      _destructiveEqResolution = fl; 
    };
    void setPropositionalSplitting(PropositionalSplitting* ps) 
    { 
      _propositionalSplitting = ps; 
    };
    void setNegativeEqualitySplitting(NegativeEqualitySplitting* nes) 
    { 
      _negativeEqualitySplitting = nes; 
    };
    void setCommutativityTableForNormalisation(const SymbolInfo* t) 
    { 
      _commutativityTableForNormalisation = t; 
    };
    void setCommutativityTableForEqTaut(const SymbolInfo* t) 
    { 
      _commutativityTableForEqTaut = t; 
      _weedModuloC = 
	_commutativityTableForEqTaut ||
	_commutativityTableForPropTaut ||
	_commutativityTableForTrivialEqRes ||
	_commutativityTableForIdenticalLitFactoring;
    };
    void setCommutativityTableForPropTaut(const SymbolInfo* t) 
    { 
      _commutativityTableForPropTaut = t;  
      _weedModuloC = 
	_commutativityTableForEqTaut ||
	_commutativityTableForPropTaut ||
	_commutativityTableForTrivialEqRes ||
	_commutativityTableForIdenticalLitFactoring;
    };
    void setCommutativityTableForTrivialEqRes(const SymbolInfo* t) 
    { 
      _commutativityTableForTrivialEqRes = t;  
      _weedModuloC = 
	_commutativityTableForEqTaut ||
	_commutativityTableForPropTaut ||
	_commutativityTableForTrivialEqRes ||
	_commutativityTableForIdenticalLitFactoring;
    };
    void setCommutativityTableForIdenticalLitFactoring(const SymbolInfo* t) 
    { 
      _commutativityTableForIdenticalLitFactoring = t;  
      _weedModuloC = 
	_commutativityTableForEqTaut ||
	_commutativityTableForPropTaut ||
	_commutativityTableForTrivialEqRes ||
	_commutativityTableForIdenticalLitFactoring;
    };
    void setProhibitNonBasicDemodulation(bool fl) { _prohibitNonBasicDemodulation = fl; };
    void setReduceNonBasic(bool fl) { _reduceNonBasic = fl; };
    void setUseInputNonBasicness(bool fl) { _useInputNonBasicness = fl; };
    void setUseRunTimeNonBasicness(bool fl) { _useRunTimeNonBasicness = fl; };
    void setExternalEvaluation(bool fl) { _externalEvaluation = fl; };
    void setClauseSizeLimit(ulong lim) 
    { 
      _clauseSizeLimit = lim;
      if (_clauseSizeLimit > VampireKernelConst::MaxClauseSize)
	_clauseSizeLimit = VampireKernelConst::MaxClauseSize;
      _internalClauseCollector.setSizeLimit(_clauseSizeLimit);
    };
    void setClauseLengthLimit(ulong lim) 
    { 
      _clauseLengthLimit = lim; 
      if (_clauseLengthLimit > VampireKernelConst::MaxlNumOfLits)
	_clauseLengthLimit = VampireKernelConst::MaxlNumOfLits;
      _internalClauseCollector.setLengthLimit(_clauseLengthLimit);
    };
    void setLiteralSizeLimit(ulong lim) 
    { 
      _literalSizeLimit = lim;
      if (_literalSizeLimit > VampireKernelConst::MaxTermSize)
	_literalSizeLimit = VampireKernelConst::MaxTermSize;
      _internalClauseCollector.setLiteralSizeLimit(_literalSizeLimit);
    };
    void setInferenceDepthLimit(ulong lim) { _inferenceDepthLimit = lim; };
    void load(ClauseHandle& cl)
    {
      CALL("load(ClauseHandle& cl)");
      ASSERT(!cl.isNull());
      _internalClauseQueue.collect(cl);
    }; 
    void load(TemporaryClauseQueue& inputQueue)
    {
      CALL("load(TemporaryClauseQueue& inputQueue)");
      if (!inputQueue.isEmpty())
	inputQueue.popTo(_internalClauseQueue);
    };
    bool anotherRetained(ClauseHandle& retained)
    {
      CALL("anotherRetained(ClauseHandle& retained)");
      while (!_internalClauseQueue.isEmpty()) 
	{
	  _internalClauseQueue.popTo(_currentClause);
	  registerNewCurrent();      
	  if (_equalityPossible)
	    {
	      if (processCurrent())
		{
		  ASSERT(!_currentClause.isNull());
		  _currentClause.passTo(retained);
		  return true;
		};
	    }
	  else
	    {
	      if (processCurrentNoEquality())
		{
		  ASSERT(!_currentClause.isNull());
		  _currentClause.passTo(retained);
		  return true;
		};
	    };
	};
      retained.clear();
      return false;
    }; 
    Clause* canBeSubsumed(TemporaryClause& cl);
    Clause* canBeRewrittenByDemodulation(TemporaryClause& cl);
  private:
    class SimplifierReceiver : public BK::Receiver<const Clause*>
    {
    public:
      SimplifierReceiver(ForwardProcessing* fp) :
	_forwardProcessing(fp)
      {
      };
      void connectTo(ForwardProcessing* fp) { _forwardProcessing = fp; };
      void receive(const Clause* const & anc);
    private:
      ForwardProcessing* _forwardProcessing;
    }; 
  private:
    bool processCurrent();
    bool processCurrentNoEquality();
    bool weedCurrent();
    bool subsumeCurrent();
    bool simplifyCurrentByForwardSubsumptionResolution();
    bool normaliseCurrentByCommutativity();
    bool reduceNonBasicInCurrentByCommutativity();
    bool reduceNonBasicInCurrent();
    bool demodulateIntoCurrent();
    bool simplifyCurrentByExternalEvaluation();
    bool simplifyCurrentByDestructiveEqualityResolution();
    bool splitNegativeEqualitiesInCurrent();
    bool applyPropositionalSplittingToCurrent();
    bool normaliseLitByDemodulation(LitHandle& lit,
				    LitHandle& spareLit);
    bool normaliseComplexSubtermByDemodulation(LitHandle& lit,
					       const Flatterm* subterm,
					       LitHandle& spareLit);
    bool rewriteComplexSubtermByDemodulation(LitHandle& lit,
					     const Flatterm* subterm,
					     ulong& modifiedPosition,
					     LitHandle& spareLit);
    bool rewriteTopOfComplexSubtermByDemodulation(LitHandle& lit,
						  const Flatterm* subterm,
						  LitHandle& spareLit);
    bool checkSubstSafetyWRTInputNonBasicness(const Term* complexTerm,
					      bool alreadyInputNonBasic,
					      const ForwardMatchingSubstitutionCache& subt);
    bool simplifyLitByExternalEvaluation(LitHandle& lit,
					 LitHandle& spareLit);
    bool discardCurrentAsTooComplex();
    void registerNewCurrent();
    void registerRejected(const RejectionReason& reason);
    void registerRetained();
  private:
    OPEN_BR_DECL(1);
    bool _preprocessingInProgress;
    bool _equalityPossible;
    ForwardSubsumptionFrontEnd* _forwardSubsumption;
    bool _forwardSubsumptionResolution;
    ulong _minLengthOfSubsumptionResolutionTarget;
    ForwardDemodulation* _forwardDemodulation; 
    bool _onBranchForwardDemodulation;
    bool _destructiveEqResolution;
    PropositionalSplitting* _propositionalSplitting;
    NegativeEqualitySplitting* _negativeEqualitySplitting;
    const SymbolInfo* _commutativityTableForNormalisation;
    const SymbolInfo* _commutativityTableForEqTaut;
    const SymbolInfo* _commutativityTableForPropTaut;
    const SymbolInfo* _commutativityTableForTrivialEqRes;
    const SymbolInfo* _commutativityTableForIdenticalLitFactoring;
    bool _weedModuloC;
    bool _prohibitNonBasicDemodulation;
    bool _reduceNonBasic;
    bool _useInputNonBasicness;
    bool _useRunTimeNonBasicness;
    bool _externalEvaluation;
    ulong _clauseSizeLimit;
    ulong _clauseLengthLimit;
    ulong _literalSizeLimit;
    ulong _inferenceDepthLimit;
    Statistics* _statistics;
    ClauseHandle _currentClause;
    bool _currentHasBeenSimplifiedBySubsumptionResolution;
    bool _currentHasBeenSimplifiedByDemodulation;
    bool _currentHasBeenSimplifiedByOnBranchDemodulation;
    bool _currentHasBeenSimplifiedByDestructiveEqualityResolution;
    bool _currentHasBeenSimplifiedByExternalEvaluation;
    bool _currentHasBeenTransformedByPropositionalSplitting;
    bool _currentHasBeenTransformedByNegativeEqualitySplitting;
    TemporaryClauseQueue _internalClauseQueue;
    TemporaryClauseCollector _internalClauseCollector;
    ForwardDemodulation::SetOfSplittingBranches _splittingBranches;
    SimplifierReceiver _simplifierReceiver;
    CLOSE_BR_DECL(1);
    friend class SimplifierReceiver;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_PROCESSING
#  define DEBUG_NAMESPACE "ForwardProcessing"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline void ForwardProcessing::registerNewCurrent()
  {
    CALL("registerNewCurrent()");
    _currentHasBeenSimplifiedBySubsumptionResolution = false;
    _currentHasBeenSimplifiedByDemodulation = false;
    _currentHasBeenSimplifiedByOnBranchDemodulation = false;
    _currentHasBeenSimplifiedByDestructiveEqualityResolution = false;
    _currentHasBeenSimplifiedByExternalEvaluation = false;
    _currentHasBeenTransformedByPropositionalSplitting = false;
    _currentHasBeenTransformedByNegativeEqualitySplitting = false;
  }; 
  inline
  void ForwardProcessing::registerRejected(const RejectionReason& reason)
  {
    CALL("registerRejected(const RejectionReason& reason)");
    if (_currentHasBeenSimplifiedBySubsumptionResolution)
      ++(_statistics->clauses.generated.simplifiedBySubsumptionResolution);
    if (_currentHasBeenSimplifiedByDemodulation)
      ++(_statistics->clauses.generated.simplifiedByDemodulation);
    if (_currentHasBeenSimplifiedByOnBranchDemodulation)
      ++(_statistics->clauses.generated.simplifiedByOnBranchDemodulation);
    if (_currentHasBeenSimplifiedByDestructiveEqualityResolution)
      ++(_statistics->clauses.generated.simplifiedByDestructiveEqualityResolution);
    if (_currentHasBeenSimplifiedByExternalEvaluation)
      ++(_statistics->clauses.generated.simplifiedByExternalEvaluation);
    if (_currentHasBeenTransformedByPropositionalSplitting)
      ++(_statistics->clauses.generated.transformedByPropositionalSplitting);
    if (_currentHasBeenTransformedByNegativeEqualitySplitting)
      ++(_statistics->clauses.generated.transformedByNegativeEqualitySplitting);
    switch (reason)
      {	
      case None:
	++(_statistics->clauses.generated.discardedForOtherReasons);
	break;
      case PropositionalTautology:
	++(_statistics->clauses.generated.discardedPropositionalTautologies);
	++(_statistics->inferences.simplifying.propositionalTautologyDeletion);
	break;
      case EquationalTautology:
	++(_statistics->clauses.generated.discardedEquationalTautologies);
	++(_statistics->inferences.simplifying.equationalTautologyDeletion);
	break;
      case ClauseTooLarge:
	++(_statistics->clauses.generated.discardedTooLarge);
	break;
      case LiteralTooLarge:
	++(_statistics->clauses.generated.discardedWithTooLargeLiterals);
	break;
      case TooManyLiterals:
	++(_statistics->clauses.generated.discardedWithTooManyLiterals);
	break;
      case TooManyVariables: 
	++(_statistics->clauses.generated.discardedWithTooManyVariables);
	break;
      case LiteralTooDeep:
	++(_statistics->clauses.generated.discardedTooDeep);
	break;
      case InferenceTooDeep:
	++(_statistics->clauses.generated.discardedWithTooDeepInferences);
	break;
      case Subsumed:
	++(_statistics->clauses.generated.subsumed);
	++(_statistics->inferences.simplifying.forwardSubsumption);
	break;
      case ReducedInNonBasicPosition:
	++(_statistics->clauses.generated.reducedInNonBasicPosition);
	++(_statistics->inferences.simplifying.reductionInNonBasicPosition);
	break;
      };
  }; 
  inline void ForwardProcessing::registerRetained()
  {
    CALL("registerRetained()");
    if (_currentHasBeenSimplifiedBySubsumptionResolution)
      ++(_statistics->clauses.generated.simplifiedBySubsumptionResolution);
    if (_currentHasBeenSimplifiedByDemodulation)
      ++(_statistics->clauses.generated.simplifiedByDemodulation);
    if (_currentHasBeenSimplifiedByOnBranchDemodulation)
      ++(_statistics->clauses.generated.simplifiedByOnBranchDemodulation);
    if (_currentHasBeenSimplifiedByDestructiveEqualityResolution)
      ++(_statistics->clauses.generated.simplifiedByDestructiveEqualityResolution);
    if (_currentHasBeenSimplifiedByExternalEvaluation)
      ++(_statistics->clauses.generated.simplifiedByExternalEvaluation);
    if (_currentHasBeenTransformedByPropositionalSplitting)
      ++(_statistics->clauses.generated.transformedByPropositionalSplitting);
    if (_currentHasBeenTransformedByNegativeEqualitySplitting)
      ++(_statistics->clauses.generated.transformedByNegativeEqualitySplitting);
  }; 
}; 
#endif
