/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ForwardSubsumption.hpp"
#include "ForwardSubsumptionCommand.hpp"
#include "GlobalStopFlag.hpp"
#include "Clause.inl"
using namespace BK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION
#define DEBUG_NAMESPACE "ForwardSubsumption"
#endif 
#include "debugMacros.hpp"
using namespace VK;
ForwardSubsumption::ForwardSubsumption() 
  : ForwardSubsumptionCode(subst), 
  backtrack_lit(DOP("ForwardSubsumption::backtrack_lit")),
  backtrack_instr(DOP("ForwardSubsumption::backtrack_instr")),
  backtrack_cursor(DOP("ForwardSubsumption::backtrack_cursor"))
{
  CALL("constructor ForwardSubsumption()");
  DOP(backtrack_lit.freeze());
};
ForwardSubsumption::~ForwardSubsumption() {};
void ForwardSubsumption::reset()
{
  CALL("reset()");
  BK::destroy(*(static_cast<ForwardSubsumptionCode*>(this)));
  BK::init(*(static_cast<ForwardSubsumptionCode*>(this)),
	   const_cast<const Flatterm**>(subst));
}; 
Clause* ForwardSubsumption::Subsume()
{ 
  CALL("ForwardSubsumption::Subsume()");  
  ulong maxLitNum = ForwardSubsumptionCode::maxlNumOfLits();
  DOP(backtrack_lit.unfreeze()); 
  if (!backtrack_lit.tryToExpand(maxLitNum)) 
    return static_cast<Clause*>(0);
  DOP(backtrack_lit.freeze());
#ifdef FS_STAT
  anotherCall();
#endif 
  ForwardSubsumptionCommand* curr_instr = Tree();
  if (!(curr_instr)) return 0; 
  _query.EndOfQuery();
  const Flatterm* cursor = 0;
  Term curr_sym;
  Query::Literal* curr_lit;
  Query::Literal* next_lit;
  backtrack_lit.reset();
  backtrack_instr.reset();
  backtrack_cursor.reset();
  Clause* subsuming_cl;
 check_tag:
#ifdef FS_STAT
  anotherInstr();
#endif 
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::LitHeader :
      curr_lit = _query.FirstLit(curr_instr->HeaderNum());
    check_lit:
      if (curr_lit)
        {
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_lit;
	    }
	  else
	    { 
	      curr_lit->Capture();
	      backtrack_lit.push(curr_lit); 
	      backtrack_instr.push(curr_instr);
	      cursor = curr_lit->Args();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    } ;
        }
      else 
        goto try_fork;
    case ForwardSubsumptionCommand::OrdEq :
      curr_lit = _query.FirstLit(curr_instr->HeaderNum());
    check_ord_eq_lit:
      if (curr_lit)
        { 
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_ord_eq_lit;
	    }
	  else
	    { 
	      curr_lit->Capture();
	      backtrack_lit.push(curr_lit); 
	      backtrack_instr.push(curr_instr);
	      cursor = curr_lit->Arg1();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    };
        }
      else 
        goto try_fork;
    case ForwardSubsumptionCommand::UnordEq : 
      curr_lit = _query.FirstLit(curr_instr->HeaderNum());
    check_unord_eq_lit:
      if (curr_lit)
        {
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_unord_eq_lit;
	    }
	  else
	    {
	      curr_lit->Capture(); 
	      backtrack_lit.push(curr_lit);
	      curr_instr->SwapOn();
	      backtrack_instr.push(curr_instr);
	      cursor = curr_lit->Arg1();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    };
        }
      else 
        goto try_fork; 
    case ForwardSubsumptionCommand::PropLit :
      if (_query.FirstLit(curr_instr->HeaderNum())) 
	{
	  if (curr_instr->Fork()) 
	    {
	      backtrack_instr.push(curr_instr);
	    }; 
	  curr_instr = curr_instr->Next(); 
	  goto check_tag;
	}
      else 
	goto try_fork; 
    case ForwardSubsumptionCommand::SecondEqArg :
      if (cursor->IsBackJump()) 
	{
	  cursor = cursor->after();
	}; 
      curr_instr = curr_instr->Next(); 
      goto check_tag; 
    case ForwardSubsumptionCommand::Func :
      curr_sym = cursor->symbolRef();
      if (curr_sym.isVariable()) goto backtrack;
    compare_func:
      if (curr_sym == curr_instr->func()) 
        {
	  cursor++;
	  curr_instr = curr_instr->Next();
	  goto check_tag; 
        };
      if (curr_sym < curr_instr->func()) goto backtrack;
      curr_instr = curr_instr->Fork();
      if (!curr_instr) goto backtrack;
      if (curr_instr->IsFunc()) 
	{
#ifdef FS_STAT
	  anotherInstr();
#endif
          goto compare_func;
	};
      goto check_tag;
    case ForwardSubsumptionCommand::Var :
      *(curr_instr->var()) = cursor; 
      if (curr_instr->Fork() && !cursor->symbolRef().isVariable())
	{
	  backtrack_instr.push(curr_instr);
	  backtrack_cursor.push(cursor);
	};
      cursor = cursor->after();
      curr_instr = curr_instr->Next();
      goto check_tag;
    case ForwardSubsumptionCommand::CompareNF :
#ifdef FS_STAT
      anotherCompInstr();
#endif
      if ((*(curr_instr->Var1()))->equal(*(curr_instr->Var2())))
	{
#ifdef FS_STAT
	  anotherCompSucc();
#endif
	  curr_instr = curr_instr->Next(); 
	  goto check_tag;     
	}
      else goto backtrack;
    case ForwardSubsumptionCommand::Compare : 
#ifdef FS_STAT
      anotherCompInstr();
#endif
      if ((*(curr_instr->Var1()))->equal(*(curr_instr->Var2())))
	{
#ifdef FS_STAT
	  anotherCompSucc();
#endif
	  if (curr_instr->Fork()) 
	    {
	      backtrack_instr.push(curr_instr);
	      backtrack_cursor.push(cursor);
	    };
	  curr_instr = curr_instr->Next(); 
	  goto check_tag; 
	}
      else goto try_fork;
    case ForwardSubsumptionCommand::Success : 
      subsuming_cl = curr_instr->TheClause();
      if (!subsuming_cl->isBlockedForSubsumption()) return subsuming_cl;
      goto backtrack; 
#ifdef DEBUG_NAMESPACE  
    default: ICP("ICP1"); return 0; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: return 0; 
#endif
#endif  
    }; 
 try_fork:
  curr_instr = curr_instr->Fork();
  if (curr_instr) goto check_tag;
 backtrack: 
  if (backtrack_instr.empty())
    {
      return 0;
    }; 
  curr_instr = backtrack_instr.pop();
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::LitHeader :
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();
      ASSERT(curr_lit); 
      ASSERT(curr_lit->Captured());
      curr_lit->Release();
      curr_lit = curr_lit->Next();       
      goto check_lit;
    case ForwardSubsumptionCommand::OrdEq :
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();
      ASSERT(curr_lit); 
      ASSERT(curr_lit->Captured());
      curr_lit->Release();
      curr_lit = curr_lit->Next();       
      goto check_ord_eq_lit;      
    case ForwardSubsumptionCommand::UnordEq :
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();
      ASSERT(curr_lit);
      ASSERT(curr_lit->Captured());
      if (curr_instr->SwapFlag()) 
	{        
	  next_lit = curr_lit->Next();
	  backtrack_lit.push(curr_lit);
	  curr_instr->SwapOff();
	  backtrack_instr.push(curr_instr);
	  cursor = curr_lit->Arg2();
	  curr_instr = curr_instr->Next();
	  goto check_tag;
	}
      else
	{       
	  curr_lit->Release();        
	  curr_lit = curr_lit->Next(); 
	  goto check_unord_eq_lit;
	}; 
    case ForwardSubsumptionCommand::PropLit :
      curr_instr = curr_instr->Fork();
      goto check_tag; 
    case ForwardSubsumptionCommand::Var :
    case ForwardSubsumptionCommand::Compare :
      curr_instr = curr_instr->Fork();
      cursor = backtrack_cursor.pop();
      goto check_tag;
#ifdef DEBUG_NAMESPACE  
    default: ICP("ICP2"); return 0; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: return 0; 
#endif
#endif 
    };
  ICP("ICP3");
  return 0;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION
#define DEBUG_NAMESPACE "ForwardSubsumption::Query"
#endif 
#include "debugMacros.hpp"
ForwardSubsumption::Query::Query() : 
  _litsWithHeader(static_cast<Query::Literal*>(0),"Query::_litsWithHeader"),
  _posOrdEqHeaderNum(Term::positiveOrdEqHeader().HeaderNum()),
  _negOrdEqHeaderNum(Term::negativeOrdEqHeader().HeaderNum()),
  _presentHeaders(DOP("Query::_presentHeaders"))
{
  CALL("constructor Query()");
  DOP(_litsWithHeader.freeze());
}; 
ForwardSubsumption::Query::~Query() 
{ 
  CALL("destructor ~Query()"); 
};
void ForwardSubsumption::Query::reset()
{
  CALL("reset()");
  _litCache.reset();
  while (_presentHeaders.nonempty())
    {
      _litsWithHeader[_presentHeaders.pop()] = 0;
    }; 
  last_pos_unord_eq = 0;
  last_neg_unord_eq = 0;
};
void ForwardSubsumption::Query::LinkEquations()
{
  CALL("LinkEquations()");
  if (last_pos_unord_eq) 
    last_pos_unord_eq->SetNext(FirstLit(_posOrdEqHeaderNum));
  if (last_neg_unord_eq) 
    last_neg_unord_eq->SetNext(FirstLit(_negOrdEqHeaderNum));
}; 
void ForwardSubsumption::Query::EndOfQuery()
{
  LinkEquations(); 
}; 
bool ForwardSubsumption::Query::PushLit(const Flatterm* lit)
{
  CALL("PushLit(const Flatterm* lit)");
  ASSERT(!lit->symbolRef().IsEquality());
  DOP(_litsWithHeader.unfreeze());
  Literal** lit_lst_ptr = &(_litsWithHeader.sub(lit->symbolRef().HeaderNum()));
  DOP(_litsWithHeader.freeze());
  Literal* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;
  if (*lit_lst_ptr)
    {
      currLit->Set(lit + 1,*lit_lst_ptr);
    }
  else 
    {
      _presentHeaders.pushSafe(lit->symbolRef().HeaderNum());
      currLit->Set(lit + 1,0);
    };
  *lit_lst_ptr = currLit;
  return true;
};   
bool 
ForwardSubsumption::Query::PushUnordEq(Term header,
				       const Flatterm* arg1,
				       const Flatterm* arg2)
{
  CALL("PushUnordEq(Term header,const Flatterm* arg1,const Flatterm* arg2)");
  ASSERT(header.IsUnorderedEq());
  DOP(_litsWithHeader.unfreeze());
  Literal** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
  DOP(_litsWithHeader.freeze());
  Literal* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;
  if (*lit_lst_ptr)
    {
      currLit->Set(arg1,arg2,*lit_lst_ptr);
    }
  else 
    {
      _presentHeaders.pushSafe(header.HeaderNum()); 
      currLit->Set(arg1,arg2,0);
      if (header.Negative()) 
	{ 
	  last_neg_unord_eq = currLit;
	}
      else 
	{
	  last_pos_unord_eq = currLit;
	};
    };
  *lit_lst_ptr = currLit;
  return true;
};
bool 
ForwardSubsumption::Query::PushOrdEq(Term header,
				     const Flatterm* arg1,
				     const Flatterm* arg2)
{
  CALL("PushOrdEq(Term header,const Flatterm* arg1,const Flatterm* arg2)");
  ASSERT(header.IsEquality() && header.IsOrderedEq()); 
  DOP(_litsWithHeader.unfreeze());
  Literal** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
  DOP(_litsWithHeader.freeze());
  Literal* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;
  if (*lit_lst_ptr)
    {
      currLit->Set(arg1,arg2,*lit_lst_ptr);
    }
  else 
    {
      _presentHeaders.pushSafe(header.HeaderNum());
      currLit->Set(arg1,arg2,0);
    };
  *lit_lst_ptr = currLit;
  return true;
}; 
#ifndef NO_DEBUG_VIS
ostream& ForwardSubsumption::Query::output(ostream& str) const
{
  for (ulong p = 0; p < _presentHeaders.size(); p++)
    for (const Literal* lit = _litsWithHeader[_presentHeaders.nth(p)]; lit; lit = lit->Next())
      lit->output(str) << '\n';
  return str;
}; 
#endif
