/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_SUBSUMPTION_CODE_HPP
#define FORWARD_SUBSUMPTION_CODE_HPP
#ifndef NO_DEBUG_VIS
 #include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Multiset.hpp"
#include "EquivalenceRelation.hpp"
#include "ForwardSubsumptionCommand.hpp"
#include "BasicTempLitList.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_CODE
 #define DEBUG_NAMESPACE "ForwardSubsumptionCodeIntegrator"
#endif
#include "debugMacros.hpp"
namespace VK
{
class Term;
class Clause;
class Flatterm;
typedef BK::EquivalenceRelation<VampireKernelConst::MaxClauseSize> VarEquivalence; 
class ForwardSubsumptionCodeIntegrator 
{  
 public:
  ForwardSubsumptionCodeIntegrator(const Flatterm** subst);
  ~ForwardSubsumptionCodeIntegrator();
  void Reset(ForwardSubsumptionCommand** n,Clause* cl) 
    {
      node_addr = n; 
      constr_to_insert.Reset();
      constr_to_shift.Reset();
      constr_passed.Reset();
      constr_integrated.Reset();
      constr_old.Reset();
      DOP(debugLitNum = 0);
      vars.reset();
      fs_var = fs_subst;
    };
  void Reset(ForwardSubsumptionCommand** n,const BasicTempLitList& component) 
    {
      node_addr = n; 
      constr_to_insert.Reset();
      constr_to_shift.Reset();
      constr_passed.Reset();
      constr_integrated.Reset();
      constr_old.Reset();
      DOP(debugLitNum = 0);
      vars.reset();
      fs_var = fs_subst;
    };
  const Flatterm** mapVar(ulong var) const { return first_occ[var]; };
 private:
  void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
  void MoveRight() { node_addr = (*node_addr)->ForkAddr(); };
  ulong VarNum(const Flatterm** v) const { return v - fs_subst; };
  const Flatterm** Var(const ulong& var_num) const { return fs_subst + var_num; };
  inline void MakeShifted(const ulong& v1,const ulong& v2);
  inline void MakeIntegrated(const ulong& v1,const ulong& v2) 
  {
    CALL("MakeIntegrated(const ulong& v1,const ulong& v2)");
    ASSERT(v1 < v2);
    ASSERT(!constr_integrated.Equivalent(v1,v2));
    ASSERT(!constr_passed.Equivalent(v1,v2));
    ASSERT(!constr_old.Equivalent(v1,v2));
    constr_integrated.MakeEquivalent(v1,v2);
  }; 
#ifdef DEBUG_NAMESPACE
  inline void CheckConstrIntegrity();
#endif 
  inline void PassCompNF(ForwardSubsumptionCommand* instr);
  inline bool PassComp(ForwardSubsumptionCommand* instr);
  inline void CopyCompsToShift(ForwardSubsumptionCommand** addr);
  inline void InsertCompsToShift(ForwardSubsumptionCommand** addr);
  inline bool MoveCompsToInsert(ForwardSubsumptionCommand**& addr);
  inline void InsertVar(const Flatterm** v);
 public:
  bool ReadHeader(const Term& header);
  bool ReadPropLit(const Term& header);
  bool ReadUnordEq(const Term& header);
  bool ReadOrdEq(Term header);
  void ReadSecondEqArg();
  bool ReadFunc(const Term& func);
  bool ReadVar();
  bool ReadVar(const ulong& var);
  void ReadCompare(const Flatterm** v1,const Flatterm** v2);
  bool ReadCollectedComps();
  bool ReadLiteral(Term* lit);
  bool ReadLiteral(const Flatterm* lit);
  bool ReadEqLiteral(Term* lit); 
  bool ReadEqLiteral(const Flatterm* lit); 
  bool ReadComplexTerm(Term* t);
  bool ReadComplexTerm(const Flatterm* t);
  bool ReadTerm(const Term& t)
    {
      CALL("ReadTerm(const Term& t)");
      return (t.isVariable()) ? ReadVar(t.var()) : ReadComplexTerm(t.deref());
    };
  bool ReadTerm(const Flatterm* t)
  {
    CALL("ReadTerm(const Flatterm* t)");
    if (t->isVariable()) return ReadVar(t->Var());
    return ReadComplexTerm(t);
  };
  void WriteLiteral(Term* lit);
  void WriteEqLiteral(Term* lit);
  void WriteLiteral(const Flatterm* lit);
  void WriteEqLiteral(const Flatterm* lit);
  void EnterWriteMode();
  void WriteHeader(Term header) 
    { 
      *node_addr = ForwardSubsumptionCommand::NewLitHeader(header);
      MoveDown();
    }; 
  void WritePropLit(Term header)
    {
      *node_addr = ForwardSubsumptionCommand::NewPropLit(header);
      MoveDown();
    }; 
  void WriteUnordEq(Term header)
    {
      *node_addr = ForwardSubsumptionCommand::NewUnordEq(header);
      MoveDown();
    };
  void WriteOrdEq(Term header)
    {
      *node_addr = ForwardSubsumptionCommand::NewOrdEq(header);
      MoveDown();
    };
  void WriteSecondEqArg()
    {
      *node_addr = ForwardSubsumptionCommand::NewSecondEqArg();
      MoveDown();
    }; 
  void WriteFunc(Term func)
    {
      *node_addr = ForwardSubsumptionCommand::NewFunc(func);
      MoveDown();
    }; 
  void InsertFunc(Term func)
    {
      ForwardSubsumptionCommand* tmp = *node_addr;
      *node_addr = ForwardSubsumptionCommand::NewFunc(func);
      (*node_addr)->SetFork(tmp);
      MoveDown();
    };
  void WriteVar(const Flatterm** v)
    {
      *node_addr = ForwardSubsumptionCommand::NewVar(v);
      MoveDown();
    };  
  void WriteVar(ulong var)
    {
      WriteVar(fs_var);
      if (vars.scoreOf(var))
	{
	  WriteCompare(first_occ[var],fs_var);
	}
      else 
	{
	  vars.add(var);
	  first_occ[var] = fs_var;
	}; 
      fs_var++;
    };
  void WriteComplexTerm(Term* t)
    {
      Term::Iterator iter;
      iter.Reset(t);
    write_curr_sym:
      if (iter.CurrentSymIsVar())
	{
	  WriteVar(iter.symbolRef().var());
	}
      else 
	{
	  WriteFunc(iter.symbolRef());
	};
      if (iter.Next()) goto write_curr_sym;
    }; 
  void WriteTerm(Term t)
    {
      if (t.isVariable()) { WriteVar(t.var()); }
      else WriteComplexTerm(t.deref());
    };
  void WriteComplexTerm(const Flatterm* t)
    {
      Flatterm::ConstIterator iter;
      iter.Reset(t);
    write_curr_sym:
      if (iter.CurrPos()->isVariable())
	{
	  WriteVar(iter.CurrPos()->Var());
	}
      else 
	{
	  WriteFunc(iter.CurrPos()->symbolRef());
	};
      iter.NextSym();
      if (iter.NotEndOfTerm()) goto write_curr_sym;
    }; 
  void WriteTerm(const Flatterm* t)
    {
      if (t->isVariable()) 
	{ 
	  WriteVar(t->Var()); 
	}
      else 
	WriteComplexTerm(t);
    };
  void WriteCompare(const Flatterm** v1,const Flatterm** v2)
    {
      CALL("WriteCompare(const Flatterm** v1,const Flatterm** v2)");
      DOP(MakeIntegrated(VarNum(v1),VarNum(v2)));
      *node_addr = ForwardSubsumptionCommand::NewCompareNF(v1,v2);
      MoveDown(); 
    };
  void WriteSuccess(Clause* cl)
    {
      *node_addr = ForwardSubsumptionCommand::NewSuccess(cl);
    };
  void WriteSuccess(void* indObj,ulong numOfLiterals)
  {
    *node_addr = ForwardSubsumptionCommand::NewSuccess(indObj);
    (*node_addr)->SetNumOfDiffVars(vars.numOfDiffElements());
    (*node_addr)->SetNumOfLits(numOfLiterals);
  };
  void InsertSuccess(Clause* cl)
    {
      ForwardSubsumptionCommand* tmp = *node_addr;
      *node_addr = ForwardSubsumptionCommand::NewSuccess(cl);
      (*node_addr)->SetFork(tmp);
    };
  void InsertSuccess(void* indObj,ulong numOfLiterals)
  {
    ForwardSubsumptionCommand* tmp = *node_addr;
    *node_addr = ForwardSubsumptionCommand::NewSuccess(indObj);
    (*node_addr)->SetFork(tmp);
    (*node_addr)->SetNumOfDiffVars(vars.numOfDiffElements());
    (*node_addr)->SetNumOfLits(numOfLiterals);
  };
 public: 
#ifndef NO_DEBUG
  const ForwardSubsumptionCommand* CurrInstr() const { return *node_addr; };  
#endif
#ifndef NO_DEBUG_VIS
  ostream& outputComps(ostream& str); 
#endif
 private:
  ForwardSubsumptionCommand** node_addr;
  VarEquivalence constr_to_insert;
  VarEquivalence constr_to_shift;
  VarEquivalence constr_passed;
  VarEquivalence constr_passed_in_shift; 
  VarEquivalence constr_integrated;
  VarEquivalence constr_old;  
  DOP(ulong debugLitNum);
  const Flatterm** fs_subst;
  const Flatterm** first_occ[VampireKernelConst::MaxNumOfVariables];
  BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> vars; 
  const Flatterm** fs_var;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_CODE
#define DEBUG_NAMESPACE "ForwardSubsumptionCodeRemoval"
#endif
#include "debugMacros.hpp"
namespace VK
{
class ForwardSubsumptionCodeRemoval
{
 private:
  Clause* removed_clause;
  ForwardSubsumptionCommand** node_addr;
  ForwardSubsumptionCommand** branch_to_destroy;
  VarEquivalence constr;
  const Flatterm** fs_subst;
  const Flatterm** first_occ[VampireKernelConst::MaxNumOfVariables];
  BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> vars; 
  const Flatterm** fs_var; 
 private:
  void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
  void MoveRight() { node_addr = (*node_addr)->ForkAddr(); };
  ulong VarNum(const Flatterm** v) const { return v - fs_subst; };
  const Flatterm** Var(ulong var_num) const { return fs_subst + var_num; }; 
 public:
  ForwardSubsumptionCodeRemoval(const Flatterm** subst);
  ~ForwardSubsumptionCodeRemoval();
  void Reset(ForwardSubsumptionCommand** n,Clause* cl) 
    {
      removed_clause = cl;
      node_addr = n; 
      branch_to_destroy = n; 
      constr.Reset();
      vars.reset();
      fs_var = fs_subst;
    };
  bool SkipComplexTerm(Term* t);
  bool SkipLiteral(Term* lit); 
  bool SkipEqLiteral(Term* lit); 
  bool SkipTerm(Term t) 
    {
      return (t.isVariable()) ? SkipVar(t.var()) : SkipComplexTerm(t.deref());
    };
  bool SkipHeader(const Term& header);
  bool SkipPropLit(const Term& header);
  bool SkipUnordEq(const Term& header);
  bool SkipOrdEq(const Term& header);
  bool SkipSecondEqArg();
  bool SkipFunc(Term func);
  bool SkipVar();
  bool SkipVar(ulong var)
    {
      if (SkipVar())
	{
	  if (vars.scoreOf(var))
	    {
	      SkipCompare(first_occ[var],fs_var);
	    } 
	  else 
	    {
	      vars.add(var);
	      first_occ[var] = fs_var;
	    }; 
	  fs_var++;
	  return true;
	}
      else return false;
    }; 
  void SkipCompare(const Flatterm** v1,const Flatterm** v2); 
  bool FindSuccess();
  void DestroyBranch() 
    {
      ForwardSubsumptionCommand* tmp = *branch_to_destroy;
      *branch_to_destroy = tmp->Fork();
      tmp->SetFork(0);
      ForwardSubsumptionCommand::DestroyTree(tmp);
    };
 public: 
#ifndef NO_DEBUG
  ForwardSubsumptionCommand* Node() const { return *node_addr; }; 
#endif
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_CODE
#define DEBUG_NAMESPACE "ForwardSubsumptionCode"
#endif
#include "debugMacros.hpp"
namespace VK
{
class ForwardSubsumptionCode 
{
 public:
  ForwardSubsumptionCode(const Flatterm** sbst) ;
  ~ForwardSubsumptionCode();
  ForwardSubsumptionCommand* Tree() const 
    {
      return tree; 
    }; 
  const Flatterm** Subst() { return subst; };
  const Flatterm** mapVar(ulong var) const { return integrator.mapVar(var); };
  void Integrate(Clause* cl);
  void Integrate(const BasicTempLitList& component,void* indObj);
  void Remove(Clause* cl);
  ulong maxlNumOfLits() const { return _maxlNumOfLits; };
 public: 
#ifndef NO_DEBUG_VIS
  ostream& outputCode(ostream& str) const
    {
      if (!tree) return str << "<<< Empty ForwardSubsumptionCode >>>\n"; 
      return ForwardSubsumptionCommand::outputTree(str,tree,subst)
	<< "<<< End of ForwardSubsumptionCode >>>\n";  
    }; 
#endif
 private:
  ForwardSubsumptionCommand* tree;
  const Flatterm** subst;
  ForwardSubsumptionCodeIntegrator integrator; 
  ForwardSubsumptionCodeRemoval removal; 
  ulong _maxlNumOfLits;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_CODE
#define DEBUG_NAMESPACE "ForwardSubsumptionCodeRemoval"
#endif
#include "debugMacros.hpp"
namespace VK
{
inline bool ForwardSubsumptionCodeRemoval::SkipHeader(const Term& header)
{
  CALL("SkipHeader(const Term& header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
 check_instr:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::LitHeader :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  return true;
	};
    try_fork: 
      if (curr_instr->Fork()) 
        {
	  node_addr = curr_instr->ForkAddr();
	  branch_to_destroy = node_addr;
	  goto check_instr;
        }; 
      return false; 
    case ForwardSubsumptionCommand::CompareNF : 
    case ForwardSubsumptionCommand::Compare :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  goto check_instr; 
	}
      else { goto try_fork; };
    default : goto try_fork;
    };
}; 
inline bool ForwardSubsumptionCodeRemoval::SkipPropLit(const Term& header)
{
  CALL("SkipPropLit(const Term& header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
 check_instr:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::PropLit :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  return true;
	};
    try_fork: 
      if (curr_instr->Fork()) 
        {
	  node_addr = curr_instr->ForkAddr();
	  branch_to_destroy = node_addr;
	  goto check_instr;
        }; 
      return false; 
    case ForwardSubsumptionCommand::CompareNF : 
    case ForwardSubsumptionCommand::Compare :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  goto check_instr; 
	}
      else { goto try_fork; };
    default : goto try_fork;
    };
}; 
inline bool ForwardSubsumptionCodeRemoval::SkipUnordEq(const Term& header)
{
  CALL("SkipUnordEq(const Term& header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
 check_instr:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::UnordEq :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  return true;
	};
    try_fork: 
      if (curr_instr->Fork()) 
        {
	  node_addr = curr_instr->ForkAddr();
	  branch_to_destroy = node_addr;
	  goto check_instr;
        }; 
      return false; 
    case ForwardSubsumptionCommand::CompareNF : 
    case ForwardSubsumptionCommand::Compare :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2); 
      if (constr.Equivalent(var_num1,var_num2))
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  goto check_instr; 
	}
      else { goto try_fork; };
    default : goto try_fork;
    };
}; 
inline bool ForwardSubsumptionCodeRemoval::SkipOrdEq(const Term& header)
{
  CALL("SkipOrdEq(const Term& header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
 check_instr:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::OrdEq :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  return true;
	};
    try_fork: 
      if (curr_instr->Fork()) 
        {
	  node_addr = curr_instr->ForkAddr();
	  branch_to_destroy = node_addr;
	  goto check_instr;
        }; 
      return false; 
    case ForwardSubsumptionCommand::CompareNF : 
    case ForwardSubsumptionCommand::Compare :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  goto check_instr; 
	}
      else { goto try_fork; };
    default : goto try_fork;
    };
}; 
inline bool ForwardSubsumptionCodeRemoval::SkipSecondEqArg()
{
  CALL("SkipSecondEqArg()");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
 check_instr:
  curr_instr = *node_addr;
  ASSERT(!(curr_instr->Fork()));
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::SecondEqArg : 
      node_addr = curr_instr->NextAddr();
      return true;
    case ForwardSubsumptionCommand::CompareNF : 
    case ForwardSubsumptionCommand::Compare :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
	{
	  node_addr = curr_instr->NextAddr();
	  goto check_instr; 
	}
      else return false; 
#ifdef DEBUG_NAMESPACE
    default: ICP("X0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return false;
#endif
#endif 
    };
}; 
inline bool ForwardSubsumptionCodeRemoval::SkipFunc(Term func)
{
  CALL("SkipFunc(Term func)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
  ASSERT(*node_addr);
 check_instr:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::Func :
      if (curr_instr->func() == func)
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  node_addr = curr_instr->NextAddr();
	  return true;
	};
    try_fork: 
      if (curr_instr->Fork()) 
        {
	  node_addr = curr_instr->ForkAddr();
	  branch_to_destroy = node_addr;
	  goto check_instr;
        }; 
      return false;
    case ForwardSubsumptionCommand::CompareNF : 
    case ForwardSubsumptionCommand::Compare :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
	{
	  ASSERT(!(curr_instr->Fork()));
	  node_addr = curr_instr->NextAddr();
	  goto check_instr; 
	}
      else return false;
    default : goto try_fork;  
    };
}; 
inline bool ForwardSubsumptionCodeRemoval::SkipVar()
{
  CALL("SkipVar()");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
  ASSERT(*node_addr);
 check_instr:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::Var :
      if (curr_instr->Fork()) branch_to_destroy = node_addr;
      node_addr = curr_instr->NextAddr();
      return true;
    case ForwardSubsumptionCommand::CompareNF : 
    case ForwardSubsumptionCommand::Compare :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
	{
	  ASSERT(!(curr_instr->Fork()));
	  node_addr = curr_instr->NextAddr();
	  goto check_instr; 
	}
      else return false;
    default : 
      if (curr_instr->Fork()) 
	{
	  node_addr = curr_instr->ForkAddr();
	  branch_to_destroy = node_addr;
	  goto check_instr;
	}; 
      return false;
    };
}; 
inline void ForwardSubsumptionCodeRemoval::SkipCompare(const Flatterm** v1,const Flatterm** v2)
{
  CALL("SkipCompare(const Flatterm** v1,const Flatterm** v2)"); 
  ASSERT(v1 < v2);
  constr.MakeEquivalent(VarNum(v1),VarNum(v2));
}; 
inline bool ForwardSubsumptionCodeRemoval::FindSuccess()
{
  CALL("FindSuccess()");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  ForwardSubsumptionCommand* curr_instr;
  ASSERT(*node_addr);
 check_tag:
  curr_instr = *node_addr;
  if (curr_instr->IsCompare() || curr_instr->IsCompareNF())
    { 
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  MoveDown();
	  goto check_tag;
	};
    try_fork: 
      if (curr_instr->Fork())
        {
	  node_addr = curr_instr->ForkAddr();
	  branch_to_destroy = node_addr;
	  goto check_tag;
        };
      return false;  
    };
  if (curr_instr->IsSuccess())
    {
      if (curr_instr->TheClause() == removed_clause)
	{
	  if (curr_instr->Fork()) branch_to_destroy = node_addr;
	  return true;
	};
    };
  goto try_fork;
}; 
}; 
#endif
