/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ForwardSubsumptionFrontEnd.hpp"
#include "Clause.inl"
#include "TemporaryLiteral.inl"
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_FRONT_END
#  define DEBUG_NAMESPACE "ForwardSubsumptionFrontEnd"
#endif
#include "debugMacros.hpp"
ForwardSubsumptionFrontEnd::ForwardSubsumptionFrontEnd() 
  : _subsumptionResolutionMinLength(0UL) 
{
  unlock();
};
ForwardSubsumptionFrontEnd::~ForwardSubsumptionFrontEnd() {};
void ForwardSubsumptionFrontEnd::reset()
{
  CALL("reset()");
  ForwardSubsumption::reset();
  _subsumptionResolutionMinLength = 0UL; 
  unlock();
};
void ForwardSubsumptionFrontEnd::integrate(Clause* cl)
{
  CALL("integrate(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ForwardSubsumption::Integrate(cl);
  cl->unprotect();
  unlock();
};
void ForwardSubsumptionFrontEnd::remove(Clause* cl)
{
  CALL("remove(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  ForwardSubsumption::Remove(cl);
  cl->unprotect();
  unlock();
};
Clause* 
ForwardSubsumptionFrontEnd::subsume(BasicTempLitList& lits) 
{
  CALL("subsume(BasicTempLitList& lits)");
  ASSERT(!isLocked());
  lock();
  ForwardSubsumption::query().reset(); 
  for (BasicTempLitList::Element* el = lits.first(); el; el = el->next())
    {  
      if (!pushLiteral(el->literal()))
	{
	  unlock();
	  return static_cast<Clause*>(0);
	};
    };
  Clause* subsumingClause = ForwardSubsumption::Subsume();
  if (subsumingClause) subsumingClause->MakeVIP();
  unlock();
  return subsumingClause;
}; 
bool 
ForwardSubsumptionFrontEnd::pushLiteral(const TemporaryLiteral::Extras::Handle& lit)
{
  CALL("pushLiteral(const TemporaryLiteral::Extras::Handle& lit)");
  if (lit->isEquality())
    {       
      lit->linkEqArgs();
      if (lit->isOrderedEquality())
	{
	  return ForwardSubsumption::query().PushOrdEq(lit->header(),lit->arg1(),lit->arg2());
	}
      else 
	return ForwardSubsumption::query().PushUnordEq(lit->header(),lit->arg1(),lit->arg2()); 
    }
  else 
    return ForwardSubsumption::query().PushLit(lit->flatterm()); 
}; 
ulong 
ForwardSubsumptionFrontEnd::simplifyBySubsumptionResolution(BasicTempLitList& lits,
							    OpenClauseBackground& ancestors)
{
  CALL("simplifyBySubsumptionResolution(BasicTempLitList& lits,OpenClauseBackground& ancestors)");
  ASSERT(!isLocked());
  lock();
  if (lits.length() < _subsumptionResolutionMinLength) 
    {
      unlock();
      return 0UL;
    };
  ulong res = 0UL;
  BasicTempLitList::Element* candEl = lits.last();
 try_lit: 
  if (candEl->literal()->isEquality() ||
      candEl->literal()->isAnswerLiteral())
    {
      if (candEl == lits.first()) 
	{
	  unlock();
	  return res;
	};
      candEl = candEl->previous();
      goto try_lit;
    };
  candEl->literal()->flatterm()->symbolRef().flipPolarity();
  query().reset(); 
  for (BasicTempLitList::Element* el = lits.first(); el; el = el->next())
    if (!pushLiteral(el->literal())) 
      {
	unlock();
	return 0UL;
      };
  Clause* simplifier = Subsume();
  if (simplifier)
    {
      simplifier->MakeVIP();
      res++;
      ancestors.pushSimplifier(simplifier,simplifier->inferenceDepth());
      if (simplifier->isSubgoal()) ancestors.markAsSubgoal();
      simplifier->setFlag(Clause::FlagHasChildren);
      BasicTempLitList::Element* tmp = candEl;
      candEl = candEl->next();
      lits.remove(tmp);
      if ((lits.length() < _subsumptionResolutionMinLength) 
	  || (candEl == lits.first()))
	{
	  unlock();
	  return res;
	};
      if (candEl) { candEl = candEl->previous(); }
      else
	candEl = lits.last();
      goto try_lit;  
    }
  else 
    {
      candEl->literal()->flatterm()->symbolRef().flipPolarity();
      if (candEl == lits.first()) 
	{
	  unlock();
	  return res;
	};
      candEl = candEl->previous(); 
      goto try_lit;
    }; 
}; 
