/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef IU_COMMAND_HPP
#define IU_COMMAND_HPP
#ifndef NO_DEBUG_VIS
 #include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "UnifTerm.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_IU_COMMAND
 #define DEBUG_NAMESPACE "IUCommand"
#endif
#include "debugMacros.hpp"
namespace VK
{
class IUCommand
 {
  public: 
   enum Tag
    {
     Start,
     Var,
     VarFirstOcc,
     Func,
     FuncLinear,
     FuncGround,
     Const,
     End
    };
  private: 
   Tag _tag;
   UnifTerm::Symbol sym;
   UnifTerm* term;
   IUCommand* nextCommand;
  public:
   IUCommand() {};
   ~IUCommand() {};
   Tag& tag() { return _tag; };
   const Tag& tag() const { return _tag; };
   const UnifTerm::Symbol& Symbol() const { return sym; };
   UnifTerm::Symbol& Symbol() { return sym; };
   UnifTerm*& Term() { return term; };
   const UnifTerm* const & Term() const 
   { 
     const UnifTerm* const & res = term;
     return res;
   };
   IUCommand*& NextCommand() { return nextCommand; };
   const IUCommand* const & NextCommand() const 
   { 
     const IUCommand* const & res = nextCommand;
     return res;
   };
   IUCommand*& AllVars() 
    {
     CALL("AllVars()");
     ASSERT(tag() == Start);
     return nextCommand; 
    };
   const IUCommand* const & AllVars() const 
    {
     CALL("AllVars()");
     ASSERT(tag() == Start);
     return NextCommand();
    };
   IUCommand*& NextVar() 
    {
     CALL("NextVar()");
     ASSERT(tag() == VarFirstOcc);
     return nextCommand; 
    };
   const IUCommand* const & NextVar() const
    {
     CALL("NextVar()");
     ASSERT(tag() == VarFirstOcc);
     return NextCommand();    
    }; 
   void flipPolarity() 
    {
     CALL("flipPolarity()");
     ASSERT(tag() == Start);
     sym.flipPolarity();
    };
  public: 
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str,const IUCommand* start) const;
    ostream& outputCode(ostream& str) const; 
   #endif
 }; 
}; 
#endif
