/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "GlobalStopFlag.hpp"
#include "IUMachine.hpp"
#include "Unification.hpp"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_IU_MACHINE
#define DEBUG_NAMESPACE "IUMachine"
#endif
#include "debugMacros.hpp"
IUMachine::IUMachine() :
  _codeMemory(DOP("IUMachine::_codeMemory")),
  _dataMemory(DOP("IUMachine::_dataMemory")),
  _backtrackPoints(DOP("IUMachine::_backtrackPoints")),
  _backtrackActions(DOP("IUMachine::_backtrackActions"))
{ 
  DOP(_codeMemory.freeze());
  DOP(_dataMemory.freeze());
  DOP(_backtrackPoints.freeze());
  DOP(_backtrackActions.freeze());
};
IUMachine::~IUMachine() { CALL("destructor ~IUMachine()"); };
inline bool IUMachine::CompleteSearch()
{
  CALL("CompleteSearch()"); 
  IUCommand* com = command;
  bool same_var;
  bool free_var;
  void* bind;
 check_tag:
  ASSERT((com > Code()) && (com < end_of_code));
  ASSERT((com->tag() == IUCommand::End) || (!(proc.CurrNode()->IsLeaf())));
  switch (com->tag())
    {
    case IUCommand::Start : 
      ICP("R0");
      return false;
    case IUCommand::Var :
      bind = com->Symbol().var()->Unref(free_var);
      if (free_var)
	{
	  if (proc.BindWithOccCheck(static_cast<Unifier::Variable*>(bind),same_var))
	    {
              _backtrackPoints.push(com);
              if (same_var) 
               { 
		_backtrackActions.push(AlterBindingWithOCSameVar); 
               }
	      else
		_backtrackActions.push(AlterBindingWithOC); 
	      com++;
	      goto check_tag;  
	    }
	  else { goto backtrack; };
	}
      else
	{
	  ASSERT(UnifTerm::CorrectComplexTerm(static_cast<UnifTerm*>(bind)));
	  if (proc.ExtractUnifiable(static_cast<UnifTerm*>(bind)))
	    {
              _backtrackPoints.push(com);
              _backtrackActions.push(AlterUnifiable);
	      com++;
	      goto check_tag;
	    }
	  else 
	    {
	      goto backtrack; 
	    };
	};
    case IUCommand::VarFirstOcc :
      proc.Bind(com->Symbol().var());
      _backtrackPoints.push(com);
      _backtrackActions.push(AlterBinding);
      com++;
      goto check_tag;  
    case IUCommand::Func :
    func_check_node:
    switch (proc.CurrNode()->tag())
      {
      case UnifIndex::TreeNode::Func : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {
            _backtrackPoints.push(com);
            _backtrackActions.push(TryAlternativeNode);
            com++;
            proc.Below();
            goto check_tag;
	  }
	else
	  {
	  func_try_alternative:
	    proc.Alternative();
	    if (proc.CurrNode()) { goto func_check_node;}
	    else { goto backtrack; };
	  };
      case UnifIndex::TreeNode::Const : goto func_try_alternative;
      case UnifIndex::TreeNode::Var :
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            if (UnifTerm::Occurs(static_cast<Unifier::Variable*>(bind),com->Term()))
	      {
		goto func_try_alternative;
	      }
            else 
	      {
		Unifier::current()->Bind(static_cast<Unifier::Variable*>(bind),com->Term());
                _backtrackPoints.push(com);
                _backtrackActions.push(UnbindLastTryAlternativeNode);
		com = com->NextCommand();
		proc.Below();
		goto check_tag;
	      };
	  }
	else 
	  {
            if (Unification::UnifyComplex(static_cast<UnifTerm*>(bind),com->Term()))
	      {
                _backtrackPoints.push(com);
                _backtrackActions.push(UnbindSectionTryAlternativeNode);
		com = com->NextCommand();
		proc.Below();
		goto check_tag;
	      }
            else { goto func_try_alternative; };
	  };
      case UnifIndex::TreeNode::VarFirstOcc :
	Unifier::current()->Bind(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
        _backtrackPoints.push(com);
        _backtrackActions.push(UnbindLastTryAlternativeNode);
	com = com->NextCommand();
	proc.Below();
	goto check_tag;
      case UnifIndex::TreeNode::Leaf :
	ICP("R1");
	return false;
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("A1");
	return false;    
#endif
      };
    case IUCommand::FuncLinear :
    func_plain_check_node:
    switch (proc.CurrNode()->tag())
      {
      case UnifIndex::TreeNode::Func : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {
            _backtrackPoints.push(com);
            _backtrackActions.push(TryAlternativeNode);
            com++;
            proc.Below();
            goto check_tag;
	  }
	else
	  {
	  func_plain_try_alternative:
	    proc.Alternative();
	    if (proc.CurrNode()) { goto func_plain_check_node;}
	    else { goto backtrack; };
	  };
      case UnifIndex::TreeNode::Const : goto func_plain_try_alternative;
      case UnifIndex::TreeNode::Var :
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            Unifier::current()->Bind(static_cast<Unifier::Variable*>(bind),com->Term());
            _backtrackPoints.push(com);
            _backtrackActions.push(UnbindLastTryAlternativeNode);
            com = com->NextCommand();
            proc.Below();
            goto check_tag;
	  }
	else 
	  {
            if (Unification::UnifyLinear(static_cast<UnifTerm*>(bind),com->Term()))
	      {
                _backtrackPoints.push(com);
                _backtrackActions.push(UnbindSectionTryAlternativeNode);
		com = com->NextCommand();
		proc.Below();
		goto check_tag;
	      }
            else { goto func_plain_try_alternative; }; 
	  };
      case UnifIndex::TreeNode::VarFirstOcc : 
	Unifier::current()->Bind(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
        _backtrackPoints.push(com);
        _backtrackActions.push(UnbindLastTryAlternativeNode);
	com = com->NextCommand();
	proc.Below();
	goto check_tag;
      case UnifIndex::TreeNode::Leaf :
	ICP("S1");
	return false;
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("L1");
	return false;    
#endif
      }; 
    case IUCommand::FuncGround :
    func_ground_check_node:
    switch (proc.CurrNode()->tag())
      {
      case UnifIndex::TreeNode::Func : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {
            _backtrackPoints.push(com);
            _backtrackActions.push(TryAlternativeNode);
            com++;
            proc.Below();
            goto check_tag;
	  }
	else
	  {
	  func_ground_try_alternative:
	    proc.Alternative();
	    if (proc.CurrNode()) { goto func_ground_check_node;}
	    else { goto backtrack; };
	  };
      case UnifIndex::TreeNode::Const : goto func_ground_try_alternative;
      case UnifIndex::TreeNode::Var :
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            Unifier::current()->Bind(static_cast<Unifier::Variable*>(bind),com->Term());
            _backtrackPoints.push(com);
            _backtrackActions.push(UnbindLastTryAlternativeNode);
            com = com->NextCommand();
            proc.Below();
            goto check_tag;
	  }
	else 
	  {
            if (Unification::MatchGround(static_cast<UnifTerm*>(bind),com->Term()))
	      {
                _backtrackPoints.push(com);
                _backtrackActions.push(UnbindSectionTryAlternativeNode);
		com = com->NextCommand();
		proc.Below();
		goto check_tag;
	      }
            else { goto func_ground_try_alternative; }; 
	  };
      case UnifIndex::TreeNode::VarFirstOcc : 
	Unifier::current()->Bind(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
        _backtrackPoints.push(com);
        _backtrackActions.push(UnbindLastTryAlternativeNode);
	com = com->NextCommand();
	proc.Below();
	goto check_tag;
      case UnifIndex::TreeNode::Leaf :
	ICP("R2");
	return false;
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("A2");
	return false; 
#endif
      };
    case IUCommand::Const :
    const_check_node:
    switch (proc.CurrNode()->tag())
      {
      case UnifIndex::TreeNode::Func :
      const_try_alternative:
      proc.Alternative();
      if (proc.CurrNode()) { goto const_check_node;}
      else { goto backtrack; };
      case UnifIndex::TreeNode::Const : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {
            _backtrackPoints.push(com);
            _backtrackActions.push(TryAlternativeNode);
            com++;
            proc.Below();
            goto check_tag;
	  }
	else { goto const_try_alternative; };
      case UnifIndex::TreeNode::Var : 
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            Unifier::current()->Bind(static_cast<Unifier::Variable*>(bind),com->Term());
            _backtrackPoints.push(com);
            _backtrackActions.push(UnbindLastTryAlternativeNode);
            com++;
            proc.Below();
            goto check_tag;
	  }
	else 
	  {
            if (com->Symbol() == (static_cast<UnifTerm*>(bind))->symbol())
	      {
                _backtrackPoints.push(com);
                _backtrackActions.push(TryAlternativeNode);
		com++;
		proc.Below();
		goto check_tag;
	      }
            else { goto const_try_alternative; };
	  };
      case UnifIndex::TreeNode::VarFirstOcc : 
	Unifier::current()->Bind(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
        _backtrackPoints.push(com);
        _backtrackActions.push(UnbindLastTryAlternativeNode);
	com++;
	proc.Below();
	goto check_tag;
      case UnifIndex::TreeNode::Leaf :
	ICP("R3"); 
	return false;
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("A3"); 
	return false;
#endif
      };
    case IUCommand::End :
      ASSERT(proc.CurrNode()->IsLeaf());      
      return true;
#ifdef DEBUG_NAMESPACE
    default :
      ICP("A3"); 
      return false; 
#endif  
    };
 backtrack:
  command = com;
  if (Backtrack()) { com = command; goto check_tag; }
  else { return false; };
}; 
bool IUMachine::FindFirstLeaf(UnifIndex::TreeNode* tree)
{
  command = Code() + 1;
  if (tree)
    {
      _backtrackPoints.push(Code());
      _backtrackActions.push(Failure);
      proc.Start(tree);
      return CompleteSearch();
    }
  else 
    {
      return false; 
    };
}; 
bool IUMachine::FindNextLeaf()
{
  return ((Backtrack()) && (CompleteSearch()));
}; 
inline bool IUMachine::Backtrack()
{
  CALL("Backtrack()");
  IUCommand* com;
  bool same_var;
  Unifier::Variable* var;
  GlobalStopFlag::check();
 rollback:
  com = _backtrackPoints.pop();
  BacktrackAction act = _backtrackActions.pop(); 
  switch (act) 
    {
    case Failure : return false;
    case AlterBindingWithOC : 
      ASSERT(com->Symbol().var()->Binding());
      if (proc.AlterBindingWithOccCheck(Unifier::current()->LastBound(),same_var))
	{
          _backtrackPoints.push(com);
	  if (same_var)
	    {
	      _backtrackActions.push(AlterBindingWithOCSameVar);
	    }
	  else
	    {
	      ASSERT(com->Symbol().var()->Binding());
              _backtrackActions.push(AlterBindingWithOC);
	    };
	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};
    case AlterBindingWithOCSameVar :
      var = static_cast<Unifier::Variable*>(com->Symbol().var()->Unref());
      ASSERT(Unifier::current()->isVariable(static_cast<void*>(var)));
      if (proc.AlterBindingWithOccCheck(var,same_var))
	{
          _backtrackPoints.push(com);
	  if (same_var)
	    {
              _backtrackActions.push(AlterBindingWithOCSameVar);
	    }
	  else
	    {
              _backtrackActions.push(AlterBindingWithOC);
	    };
	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};
    case AlterBinding : 
      ASSERT(com->tag() == IUCommand::VarFirstOcc);
      ASSERT(com->Symbol().var()->Binding());
      if (proc.AlterBinding())
	{
	  ASSERT(com->Symbol().var()->Binding());
          _backtrackPoints.push(com);
          _backtrackActions.push(AlterBinding);
	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};
    case AlterUnifiable :
      if (proc.AlterUnifiable())
	{
          _backtrackPoints.push(com);
          _backtrackActions.push(AlterUnifiable);
	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};
    case TryAlternativeNode :
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
	{
	  goto rollback; 
	};
    case UnbindLastTryAlternativeNode :
      Unifier::current()->UnbindLast();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
	{
	  goto rollback; 
	};
    case UnbindSectionTryAlternativeNode :
      Unifier::current()->UnbindLastSection();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
	{
	  goto rollback; 
	};
    };   
  ICP("R1");
  return false;
}; 
#ifndef NO_DEBUG_VIS
ostream& IUMachine::output(ostream& str) const
{
  str << "CODE:\n";
  return outputCode(str) << "INDEX = " << GetIndex() << '\n';
};
ostream& IUMachine::outputCode(ostream& str) const
{
  const IUCommand* com;
  for (com = Code(); com->tag() != IUCommand::End; com++) 
    com->output(str,Code()) << '\n';
  return com->output(str,Code()) << '\n';
}; 
#endif
