/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef IU_MACHINE_HPP
#define IU_MACHINE_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "IUCommand.hpp"
#include "IUTranslator.hpp"
#include "UnifTerm.hpp"
#include "Unifier.hpp"
#include "UnifIndex.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "ExpandingStack.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_IU_MACHINE
 #define DEBUG_NAMESPACE "IUMachine"
#endif
#include "debugMacros.hpp"
namespace VK
{
class IUMachine
{
 private:
  enum BacktrackAction
  {
    Failure,
    AlterBindingWithOC,
    AlterBindingWithOCSameVar,
    AlterBinding,
    AlterUnifiable,
    TryAlternativeNode,
    UnbindLastTryAlternativeNode,
    UnbindSectionTryAlternativeNode
  };
 private:
  IUTranslator compiler;
  UnifIndex::Retrieval proc;
  BK::Array<BK::GlobAlloc,IUCommand,BK::FastRelocation<IUCommand>,256UL> _codeMemory;
  IUCommand* end_of_code;
  BK::Array<BK::GlobAlloc,UnifTerm,BK::FastRelocation<UnifTerm>,256UL> _dataMemory;
  BK::ExpandingStack<BK::GlobAlloc,IUCommand*,BK::FastRelocation<IUCommand*>,128UL> _backtrackPoints;
  BK::ExpandingStack<BK::GlobAlloc,BacktrackAction,BK::FastRelocation<BacktrackAction>,128UL> _backtrackActions;
  IUCommand* command;
 private: 
  UnifIndex::IndexedObject* candidates;
 public:
  IUMachine();
  ~IUMachine();
  void reset() 
    {
      CALL("reset()");
      candidates = 0;
    };
  void SetIndex(ulong index) 
    {
      CALL("SetIndex(ulong index)");
      proc.SetIndex(index); 
    };
  ulong GetIndex() const { return proc.GetIndex(); };
  IUCommand* Code() { return _codeMemory.memory(); }; 
  const IUCommand* Code() const { return _codeMemory.memory(); }; 
  bool CompileLiteral(UnifTerm* lit)
    {
      CALL("CompileLiteral(UnifTerm* lit)");
      ASSERT(_dataMemory.size() >= _codeMemory.size());
      end_of_code = _codeMemory.memory();
      UnifTerm* dmem = _dataMemory.memory();
      while (!compiler.CompileLiteral(lit,end_of_code,dmem,_codeMemory.size()))
       {
	DOP(_codeMemory.unfreeze());
        DOP(_dataMemory.unfreeze());
        if (!_codeMemory.tryToExpand()) return false;
        if (!_dataMemory.tryToExpand(_codeMemory.size() - 1)) return false;
        DOP(_codeMemory.freeze());
        DOP(_dataMemory.freeze());
        end_of_code = _codeMemory.memory();
        dmem = _dataMemory.memory();
       };
      ASSERT(_dataMemory.size() >= _codeMemory.size());
      return ResetBacktracking();
    };
  bool CompilePseudoLiteral(Term header,UnifTerm* arg)
    {
      CALL("CompilePseudoLiteral(Term header,UnifTerm* arg)");
      ASSERT(_dataMemory.size() >= _codeMemory.size());
      end_of_code = _codeMemory.memory();
      UnifTerm* dmem = _dataMemory.memory();
      while (!compiler.CompilePseudoLiteral(header,arg,end_of_code,dmem,_codeMemory.size()))
       {
	DOP(_codeMemory.unfreeze());
        DOP(_dataMemory.unfreeze());
        if (!_codeMemory.tryToExpand()) return false;
        if (!_dataMemory.tryToExpand(_codeMemory.size() - 1)) return false;
        DOP(_codeMemory.freeze());
        DOP(_dataMemory.freeze());
        end_of_code = _codeMemory.memory();
        dmem = _dataMemory.memory();
       };
      ASSERT(_dataMemory.size() >= _codeMemory.size());
      return ResetBacktracking();
    };
  void flipPolarity()
    {
      Code()->flipPolarity(); 
    };
  bool FindFirstLeaf(UnifIndex::TreeNode* tree);
  bool FindNextLeaf();
  UnifIndex::TreeNode* CurrentLeaf() const { return proc.CurrNode(); };
  void ResetCandidatesFromLeaf() 
    {      
      candidates = CurrentLeaf()->indObjects();
    };
  UnifIndex::IndexedObject* NextCandidate() 
    {
      UnifIndex::IndexedObject* res = candidates;
      if (candidates) { candidates = candidates->Next(); };
      return res;
    };
#ifndef NO_DEBUG_VIS
  ostream& output(ostream& str) const;
  ostream& outputCode(ostream& str) const;
#endif
 private:
  bool ResetBacktracking()
    {
      CALL("ResetBacktracking()");
      if (_codeMemory.size())
       {
        DOP(_backtrackPoints.unfreeze());
        DOP(_backtrackActions.unfreeze());
        if (!_backtrackPoints.tryToExpand(_codeMemory.size() - 1))
	  return false;
        if (!_backtrackActions.tryToExpand(_codeMemory.size() - 1))
	  return false;
        DOP(_backtrackPoints.freeze());
        DOP(_backtrackActions.freeze());
       };
      _backtrackPoints.reset();
      _backtrackActions.reset();
      return true;
    };
  bool Backtrack();
  bool CompleteSearch();
}; 
}; 
#endif
