/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef IU_TRANSLATOR_HPP
#define IU_TRANSLATOR_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "IUCommand.hpp"
#include "GlobAlloc.hpp"
#include "ExpandingStack.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_IU_TRANSLATOR
#  define DEBUG_NAMESPACE "IUTranslator"
#endif
#include "debugMacros.hpp"
namespace VK
{
class IUTranslator
 {
  public: 
   IUTranslator()
   {
    CALL("constructor IUTranslator()");
   };
   ~IUTranslator() { CALL("destructor ~IUTranslator()"); };
   inline bool CompileLiteral(UnifTerm* word,
                              IUCommand*& code_mem,
                              UnifTerm*& data_mem,
                              ulong max_code_size);
   inline bool CompilePseudoLiteral(Term header,
                                    UnifTerm* word,
                                    IUCommand*& code_mem,
                                    UnifTerm*& data_mem,
                                    ulong max_code_size);   
  private:
   Unifier::Variable::Multiset _varMultiset;
   BK::ExpandingStack<BK::GlobAlloc,IUCommand*,BK::FastRelocation<IUCommand*>,128UL> _openBrackets;
   ExpandingBracketCounter _bracketCounter;
 }; 
inline bool IUTranslator::CompileLiteral(UnifTerm* word,
                                          IUCommand*& code_mem,
                                          UnifTerm*& data_mem,
                                          ulong max_code_size)
 {
  CALL("CompileLiteral(UnifTerm*,IUCommand**&,UnifTerm*&,ulong)");
  if (!max_code_size) return false;
  _openBrackets.reset(); 
  UnifTerm::Iter iter;
  _bracketCounter.reset();
  iter.Reset(word);
  IUCommand* var_list = 0; 
  IUCommand* start = code_mem;
  UnifTerm* data = data_mem;
  if (iter.CurrSubterm()->symbol().Func().arity())
   {
    _bracketCounter.openSafe(iter.CurrSubterm()->symbol().Func().arity());
   };
  _openBrackets.pushSafe(code_mem);
  code_mem->tag() = IUCommand::Start;
  code_mem->Symbol() = word->symbol();
  data_mem->symbol() = word->symbol();
  code_mem->Term() = data_mem;
  code_mem++;
  data_mem++;
  max_code_size--;  
  iter.SkipSym(); 
  _varMultiset.reset();
  Unifier::Variable* var;
  IUCommand* last_var = code_mem-1;
  IUCommand* last_var_not_fo = code_mem-1;
  ulong closed = 0UL;
  check_iter:
   if (!max_code_size) return false; 
   if (iter.Empty()) 
    {
     code_mem->tag() = IUCommand::End;
     code_mem++;
     max_code_size--;
     data->SetBracket(data_mem); 
     start->AllVars() = var_list;
     return true;
    };
   var = iter.CurrVar();
   if (var)
    { 
     if (_varMultiset.registerVar(var))
      {
       code_mem->tag() = IUCommand::Var;
       last_var_not_fo = code_mem;
      }
     else 
      {
       code_mem->tag() = IUCommand::VarFirstOcc;
       code_mem->NextVar() = var_list;
       var_list = code_mem;
      };
     code_mem->Symbol().MkVar(var);
     last_var = code_mem;
     data_mem->symbol().MkVar(var);
     data_mem->SetBracket(data_mem + 1);
     closed = _bracketCounter.close();
    }
   else 
    {
     ulong arity = iter.CurrSubterm()->symbol().Func().arity();
     if (arity)
      {
       code_mem->tag() = IUCommand::Func;
       _openBrackets.pushSafe(code_mem);
       _bracketCounter.openSafe(arity);
      }
     else 
      {
       code_mem->tag() = IUCommand::Const;
       closed = _bracketCounter.close();
       data_mem->SetBracket(data_mem + 1);
      };
     code_mem->Symbol() = iter.CurrSubterm()->symbol();
     code_mem->Term() = data_mem;
     data_mem->symbol() = iter.CurrSubterm()->symbol();
    };
   iter.SkipSym();
   code_mem++; 
   data_mem++;
   max_code_size--;  
   while (closed)
    {
     IUCommand* currentOpenBracket = _openBrackets.pop(); 
     currentOpenBracket->NextCommand() = code_mem;    
     unsigned length = code_mem - currentOpenBracket;
     currentOpenBracket->Term()->SetBracket(currentOpenBracket->Term() + length);
     if (last_var < currentOpenBracket) 
      {
       currentOpenBracket->tag() = IUCommand::FuncGround;
      }
     else
      {
       if (last_var_not_fo < currentOpenBracket)
        {
         currentOpenBracket->tag() = IUCommand::FuncLinear;
        };
      };
     closed--;
    };
   goto check_iter;
 }; 
  inline 
  bool 
  IUTranslator::CompilePseudoLiteral(Term header,
				      UnifTerm* word,
				      IUCommand*& code_mem,
				      UnifTerm*& data_mem,
				      ulong max_code_size)
  {
    CALL("CompilePseudoLiteral(Term,UnifTerm*,IUCommand*&,UnifTerm*&)");
    Unifier::Variable* var;   
    UnifTerm::Symbol pr_header(header);      
    IUCommand* var_list = 0; 
    IUCommand* start = code_mem;
    UnifTerm* data = data_mem;
    if (!max_code_size) return false;
    if (word->symbol().isVariable()) 
      {
	var = word->symbol().var();
	bool freeVar;
	void* instance = var->Unref(freeVar);
	if (freeVar)
	  {
	    if (max_code_size < 3UL) return false;
	    var = static_cast<Unifier::Variable*>(instance);
	    code_mem->tag() = IUCommand::Start;       
	    code_mem->Symbol() = pr_header; 
	    data_mem->symbol() = pr_header;        
	    code_mem->Symbol().Func().flipPolarity(); 	
	    code_mem->Term() = data_mem;
	    code_mem++;
	    data_mem++;
	    code_mem->tag() = IUCommand::VarFirstOcc;
	    code_mem->NextVar() = var_list;
	    var_list = code_mem;
	    code_mem->Symbol().MkVar(var); 
	    data_mem->symbol().MkVar(var);
	    data_mem->SetBracket(data_mem + 1);
	    code_mem++; 
	    data_mem++;
	    code_mem->tag() = IUCommand::End;
	    code_mem++;
	    data->SetBracket(data_mem);
	    start->AllVars() = var_list;
	    return true;
	  }
	else
	  {
	    word = static_cast<UnifTerm*>(instance);
	  };
      }; 
    ASSERT(word->symbol().IsFunctor());
    _openBrackets.reset(); 
    UnifTerm::Iter iter;
    _bracketCounter.reset();
    iter.Reset(word);
    if (iter.CurrSubterm()->symbol().Func().arity())
      {
	_bracketCounter.openSafe(iter.CurrSubterm()->symbol().Func().arity());
      };
    _openBrackets.pushSafe(code_mem);
    code_mem->tag() = IUCommand::Start;
    code_mem->Symbol() = pr_header; 
    data_mem->symbol() = pr_header;        
    code_mem->Symbol().Func().flipPolarity(); 
    code_mem->Term() = data_mem;
    code_mem++;
    data_mem++;
    max_code_size--;
    _varMultiset.reset();
    IUCommand* last_var = code_mem-1;
    IUCommand* last_var_not_fo = code_mem-1;
    ulong closed = 0UL;
  check_iter:
    if (!max_code_size) return false; 
    if (iter.Empty()) 
      {
	code_mem->tag() = IUCommand::End;
	code_mem++;
	data->SetBracket(data_mem); 
	start->AllVars() = var_list;
	return true;
      };
    var = iter.CurrVar();
    if (var)
      {
	if (_varMultiset.registerVar(var) > 0)
	  {
	    code_mem->tag() = IUCommand::Var;
	    last_var_not_fo = code_mem;
	  }
	else 
	  {
	    code_mem->tag() = IUCommand::VarFirstOcc;
	    code_mem->NextVar() = var_list;
	    var_list = code_mem;
	  };
	code_mem->Symbol().MkVar(var);
	last_var = code_mem;
	data_mem->symbol().MkVar(var);
	data_mem->SetBracket(data_mem + 1);
	closed = _bracketCounter.close();
      }
    else 
      {
	ulong arity = iter.CurrSubterm()->symbol().Func().arity();
	if (arity)
	  {
	    code_mem->tag() = IUCommand::Func;
	    _openBrackets.pushSafe(code_mem);
	    _bracketCounter.openSafe(arity);
	  }
	else 
	  {
	    code_mem->tag() = IUCommand::Const;
	    closed = _bracketCounter.close();
	    data_mem->SetBracket(data_mem + 1);
	  };
	code_mem->Symbol() = iter.CurrSubterm()->symbol();
	code_mem->Term() = data_mem;
	data_mem->symbol() = iter.CurrSubterm()->symbol();
      };
    iter.SkipSym();
    code_mem++; 
    data_mem++;
    max_code_size--;
    while (closed)
      {
	IUCommand* currentOpenBracket = _openBrackets.pop(); 
	currentOpenBracket->NextCommand() = code_mem;    
	unsigned length = code_mem - currentOpenBracket;
	currentOpenBracket->Term()->SetBracket(currentOpenBracket->Term() + length);
	if (last_var < currentOpenBracket) 
	  {
	    currentOpenBracket->tag() = IUCommand::FuncGround;
	  }
	else
	  {
	    if (last_var_not_fo < currentOpenBracket)
	      {
		currentOpenBracket->tag() = IUCommand::FuncLinear;
	      };
	  };
	closed--;
      };
    goto check_iter;
  }; 
}; 
#endif
