/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "InstanceRetrievalForDemodulation.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INSTANCE_RETRIEVAL_FOR_DEMODULATION
# define DEBUG_NAMESPACE "InstRet"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
void InstRet::integrate(const Term* complexTerm)
{
  CALL("integrate(const Term* complexTerm)");
  _index.integrate(Term(const_cast<Term*>(complexTerm)));
}; 
void InstRet::remove(const Term* complexTerm)
{
  CALL("remove(const Term* complexTerm)");
  _index.remove(Term(const_cast<Term*>(complexTerm)));
}; 
bool InstRet::newQuery(const Term* query)
{
  CALL("newQuery(const Term* query)");
  _noMoreMatches = !_index.compileQuery(Term(const_cast<Term*>(query)));
  DOP(_debugQuery = query);
  return !_noMoreMatches;
}; 
bool InstRet::nextMatch(const Term*& match)
{
  CALL("nextMatch(const Term*& match)");
  if (_noMoreMatches) return false;
  Term matchId;
  bool result = _index.nextMatch(matchId);
  if (result)
    {
      ASSERT(matchId.IsReference());
      match = matchId.deref();
      ASSERT(isValidMatch(match));
      return true;
    };
  _noMoreMatches = true;
  return false;
}; 
#ifdef DEBUG_NAMESPACE
bool InstRet::isValidMatch(const Term* term) const
{
  static Substitution::Instance queryInst;
  static Term::Iterator matchIter; 
  queryInst.reset(_debugQuery,substitution());
  matchIter.Reset(term);
  do
    {
      if (queryInst.symbol() != matchIter.symbolRef()) return false;
      matchIter.Next();
    }
  while (queryInst.next());
  return true;
}; 
#endif
