/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "LiteralProperties.inl"
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_PROPERTIES
#  define DEBUG_NAMESPACE "LiteralProperties"
#endif
#include "debugMacros.hpp"
void LiteralProperties::inheritFrom(const LiteralProperties& prop)
{
  CALL("inheritFrom(const LiteralProperties& prop)");
  make(prop.content());
}; 
void LiteralProperties::inheritFrom(const LiteralProperties& prop,
				    const SLModifier& slmod)
{
  CALL("inheritFrom(const LiteralProperties& prop,const SLModifier& slmod)");  
  make(prop.content());
  if (prop.isDefault()) return;
  setSelectionLayerModifier(slmod);
}; 
void LiteralProperties::factoriseWith(const LiteralProperties& prop,bool symmetryUsed)
{
  CALL("factoriseWith(const LiteralProperties& prop,bool symmetryUsed)");
  if (_data == prop._data) return;
  if (isDefault()) 
    {
      bool selected = isSelected() || prop.isSelected();
      _data = prop._data;
      setFlagSelected(selected);
      return;
    };
  if (prop.isDefault())
    {
      setFlagSelected(isSelected() || prop.isSelected());
      return;
    };
  if (selectionLayer() > prop.selectionLayer())
    setSelectionLayer(prop.selectionLayer());
  if (deploymentMode() < prop.deploymentMode())
    setDeploymentMode(prop.deploymentMode());
  if (equalityMode() < prop.equalityMode())
    setEqualityMode(prop.equalityMode());
  switch (prop.equalityOrientationMode())
    {
    case DefaultEqualityOrientation:
      break;
    case GivenEqualityOrientation:
      if (equalityOrientationMode() == DefaultEqualityOrientation ||
	  equalityOrientationMode() == ProhibitEqualityOrientation)
	if (symmetryUsed)
	  {
	    setEqualityOrientationMode(ReverseGivenEqualityOrientation);
	  }
	else
	  setEqualityOrientationMode(GivenEqualityOrientation);
      break;
    case ReverseGivenEqualityOrientation:
      if (equalityOrientationMode() == DefaultEqualityOrientation ||
	  equalityOrientationMode() == ProhibitEqualityOrientation)
	if (symmetryUsed)
	  {
	    setEqualityOrientationMode(GivenEqualityOrientation);
	  }
	else
	  setEqualityOrientationMode(ReverseGivenEqualityOrientation);
      break;
    case ProhibitEqualityOrientation:
      if (equalityOrientationMode() == DefaultEqualityOrientation)
	setEqualityOrientationMode(ProhibitEqualityOrientation);
      break;
    }; 
  if (selectionLayerModifier() < prop.selectionLayerModifier())
    setSelectionLayerModifier(prop.selectionLayerModifier());
  if (prop.isSelected())
    setFlagSelected(true);
  if (compareSelectabilityForFactoring(prop) == BK::Less)
    {
      setSelectabilityCategory(prop.selectabilityCategory());
    };
}; 
ostream& LiteralProperties::output(ostream& str) const
{
  CALL("output(ostream& str) const");
  if (isDefault()) return str;
  ulong numOfNondefaultEntries = 0UL;
  if (selectionLayer()) ++numOfNondefaultEntries;
  if (deploymentMode() != DefaultDeploymentMode) 
    ++numOfNondefaultEntries;
  if (equalityMode() != DefaultEqualityMode) 
    ++numOfNondefaultEntries;
  if (equalityOrientationMode() != DefaultEqualityOrientation) 
    ++numOfNondefaultEntries;
  if (selectionLayerModifier() != DefaultSLModifier)
    ++numOfNondefaultEntries;
  if (selectabilityCategory() != 
      DefaultSelectability) 
    ++numOfNondefaultEntries;
  if (selectionLayer())
    {
      str << "sl(" << selectionLayer() << ")";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
  if (deploymentMode() != DefaultDeploymentMode) 
    {
      if (deploymentMode() == RuleHead)
	{
	  str << "rh";
	}
      else
	str << "sq";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
  if (equalityMode() != DefaultEqualityMode) 
    {
      if (equalityMode() == UseEqForResolution)
	{
	  str << "addres";
	}
      else
	str << "resonly";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
  if (equalityOrientationMode() != DefaultEqualityOrientation) 
    {
      if (equalityOrientationMode() == GivenEqualityOrientation)
	{
	  str << "given_orient";
	}
      else if (equalityOrientationMode() == ReverseGivenEqualityOrientation)
	{
	  str << "rev_given_orient";
	}
      else
	str << "no_orient";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
  if (selectionLayerModifier() != DefaultSLModifier)
    {
      if (selectionLayerModifier() == AccelerateSelection)
	{
	  str << "accsel";
	}
      else
	str << "postp";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
  if (selectabilityCategory() != DefaultSelectability) 
    {
      if (selectabilityCategory() == LowSelectability)
	{
	  str << "lo";
	}
      else
	str << "hi";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
  return str;
}; 
