/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LITERAL_SELECTION_HPP
#define LITERAL_SELECTION_HPP
#include "jargon.hpp"
#include "RuntimeError.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "StandardLiteralOrdering.hpp"
#include "CustomLiteralSelection.hpp"
class SymbolInfo;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "LiteralSelection"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LiteralSelection
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef BasicTempLitList LitList;
    typedef BasicTempLitList::Element LitListElement;
    typedef TempLitListExtras::StdCharacteristics LitListStatistics;
  public:
    LiteralSelection();
    ~LiteralSelection() {};
    void setFunNum(long num);
    void setMainLiteralOrdering(StandardLiteralOrdering* ord);
    void setSplittingFlag(bool fl) 
    { 
      CALL("setSplittingFlag(bool fl)");
      _splittingIsEnabled = (_splittingIsEnabled || fl);
    };
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    void setStepResolutionLiteralSelection(bool fl)
    {
      _stepResolutionLiteralSelectionFlag = fl;    
    };
#endif
    void 
    select(LitList& lits,LitList& selectedLits,
	   LitList& unselectedLits,bool& allMaximalSelected) const;
    void 
    selectForSupport(LitList& lits,
		     LitList& selectedLits,
		     LitList& unselectedLits) const;
  private:
    class ComplexLiteralCharacteristic1 : 
      public CustomLiteralSelection::ComplexLiteralCharacteristic
    {
    public:
      ComplexLiteralCharacteristic1() {};
      ~ComplexLiteralCharacteristic1();
      void computeFor(TemporaryLiteral& lit);
      BK::Comparison 
      compareWith(CustomLiteralSelection::ComplexLiteralCharacteristic* c);
    private:
      bool _positive;
      ulong _numOfDifferentVars;
      ulong _numOfVarOccurences;
    }; 
    class ComplexLiteralOrdering1 : public LiteralOrdering
    {
    public:
      ComplexLiteralOrdering1(ulong resistanceDepthCoeff,
			      float resistanceConstCoeff,
			      float resistanceVarCoeff) :
	_resistanceDepthCoeff(resistanceDepthCoeff),
	_resistanceConstCoeff(resistanceConstCoeff),
	_resistanceVarCoeff(resistanceVarCoeff)
      {
	_isStratified = true;
      };
      ~ComplexLiteralOrdering1();
      BK::Comparison compare(TemporaryLiteral& lit1,
			     TemporaryLiteral& lit2) const;
      BK::Comparison compareLayers(const TemporaryLiteral& lit1,
				   const TemporaryLiteral& lit2) const;
    private:
      ComplexLiteralOrdering1() {};
    private:
      ulong _resistanceDepthCoeff;
      float _resistanceConstCoeff;
      float _resistanceVarCoeff;
    }; 
  private:
    void 
    selectWithNonblockingSplitting(LitList& lits,LitList& selectedLits,
				   LitList& unselectedLits,bool& allMaximalSelected,
	                           LitListStatistics& stat) const;
    inline 
    static
    void 
    unselectNonselectable(LitList& lits,
			  LitList& unselectedLits,
			  bool& allRemainingShouldBeSelected);
    inline
    static
    BK::Comparison
    compareSelectability(const LiteralProperties& prop1,
			 const LiteralProperties& prop2);
    inline 
    static
    void 
    mark(LitList& selectedLits,LitList& unselectedLits);
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    static 
    inline
    bool 
    isNextStepLiteral(const TemporaryLiteral& lit);
    static
    inline
    bool 
    isCurrentStepLiteral(const TemporaryLiteral& lit);
    void 
    inline
    unselectCurrentStateLiterals(LitList& lits,
				 LitList& selectedLits,
				 LitList& unselectedLits) const;
#endif
  private:
    StandardLiteralOrdering* _mainLiteralOrdering;
    bool _splittingIsEnabled;
    CustomLiteralSelection* _customSelection; 
    CustomLiteralSelection::OneOfEachPolarityFilter _oneOfEachPolarityFilter;
    CustomLiteralSelection::MaxValFilter _oneOfMaxSizeFilter;
    CustomLiteralSelection::MaxValFilter _allOfMaxSizeFilter;
    CustomLiteralSelection::MinValFilter _oneOfMinSizeFilter;
    CustomLiteralSelection::MinValFilter _allOfMinSizeFilter;
    CustomLiteralSelection::AllPOrAllFilter _allNegEqOrAllFilter;
    CustomLiteralSelection::AllPOrAllFilter _oneNegEqOrAnyoneFilter;
    CustomLiteralSelection::AllPOrAllFilter _allExceptPosEqOrAllFilter;
    CustomLiteralSelection::CompositionFilter _oneBigNegEqOrOneBigFilter;
    CustomLiteralSelection::CompositionFilter _oneSmallPreferablyNegEqFilter;
    CustomLiteralSelection::MaxValFilter _allWithMaxNumOfDiffVarsFilter;
    CustomLiteralSelection::InverseVarDepthProfile _inverseVarDepthProfile1;
    CustomLiteralSelection::InverseVarDepthProfile _inverseVarDepthProfile2;
    CustomLiteralSelection::MaxComplexValFilter _allWithMostShallowVarsFilter;
    CustomLiteralSelection::CompositionFilter _oneOfMaxSizeMaxNongroundFilter;
    ComplexLiteralCharacteristic1 _complexCharacteristic1_1;
    ComplexLiteralCharacteristic1 _complexCharacteristic1_2;
    CustomLiteralSelection::CompositionFilter _complexFilter1;
     CustomLiteralSelection::CompositionFilter _complexFilter2;
    CustomLiteralSelection::MaxComplexValFilter _complexFilter3;
    CustomLiteralSelection::CompositionFilter _complexFilter4;
    CustomLiteralSelection::MaxComplexValFilter _complexFilter5;
    CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability _maxSelectionAcceptability1;
    CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability _maxSelectionAcceptability2;
    ComplexLiteralOrdering1 _complexLiteralOrdering1_1;
    ComplexLiteralOrdering1 _complexLiteralOrdering1_2;
    ComplexLiteralOrdering1 _complexLiteralOrdering1_3;
    TotalLiteralSelection _totalSelection;
    MaximalLiteralSelection _maximalSelection;
    HyperLiteralSelection _hyperSelection;
    MaxWithCorrectionLiteralSelection _maxWithCorrectionSelection;
    ArbitraryLiteralSelection _arbitrarySelection;
    IncompleteBidirectionalLiteralSelection _incompleteBidirectSelection;
    BestOnlyLiteralSelection _bestOnlySelection;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    bool _stepResolutionLiteralSelectionFlag;
#endif
  }; 
}; 
#endif
