/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LOCAL_UNIFICATION_HPP
#define LOCAL_UNIFICATION_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "Stack.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
#  define DEBUG_NAMESPACE "LocalUnification"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LocalUnification
  {
  public:
    class UnifierState {};
    class TermTraversal;
  public:
    LocalUnification();
    ~LocalUnification();
    void reset();
    bool unify(const Term* complexTerm1,const Term* complexTerm2);
    bool unify(const Term& term1,const Term& term2);
    const UnifierState* getUnifierState() const 
    { 
      return 
	static_cast<const UnifierState*>(static_cast<const void*>(_instantiated.end())); 
    };
    void rollUnifierBackTo(const UnifierState* notch);
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
#endif
  private:
    bool unify1(const Term* complexTerm1,const Term* complexTerm2);
    bool unify1(const Term& term1,const Term& term2);
    bool occursIn(const Term& var,const Term& term) const;
    void instantiate(ulong varNum,const Term* instance);
  private:
    const Term* _unifier[VampireKernelConst::MaxNumOfVariables];
    BK::Stack<const Term**,VampireKernelConst::MaxNumOfVariables> _instantiated;
    friend class TermTraversal;
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline
  ostream& operator<<(ostream& str,const VK::LocalUnification& unifier)
  {
    return unifier.output(str);
  };
};
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "LocalUnification::TermTraversal"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class LocalUnification::TermTraversal
  {
  public:
    TermTraversal();
    TermTraversal(const Term& term,const LocalUnification* unifier);
    ~TermTraversal();
    void reset(const Term& term,const LocalUnification* unifier);
    operator bool() const { return _term != 0; };
    void next();
    void after();
    const Term& symbol() const { return (_term->isVariable()) ? (*_term) : (*_term->deref()); };
    const Term& term() const { return *_term; };
  private:
    void canonizeVariableInstance();
  private:
    const Term* const * _unifier;
    Term _dummyTerm[2];
    BK::Stack<const Term*,VampireKernelConst::MaxTermSize> _backtrackPoints; 
    const Term* _term;
    const Term* _after;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "LocalUnification::TermTraversal"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  void LocalUnification::TermTraversal::next()
  {
    CALL("next()");
    ASSERT(_term);
    if (_term->isVariable() || (!_term->deref()->arity())) 
      {
	after(); 
      }
    else 
      {
	if (!_after->isNotArg()) _backtrackPoints.push(_after);
	_term = _term->deref()->Args();
	_after = _term->Next();
	canonizeVariableInstance();
      };
  }; 
  inline
  void LocalUnification::TermTraversal::after()
  {
    CALL("after()");
    ASSERT(_term);
    if (_after->isNotArg())
      {
	if (_backtrackPoints)
	  {
	    _term = _backtrackPoints.pop();	  
	  }
	else 
	  {
	    _term = 0;
	    return;
	  };
      }
    else
      {
	_term = _after;
      };
    _after = _term->Next();
    canonizeVariableInstance();
  }; 
  inline
  void LocalUnification::TermTraversal::canonizeVariableInstance()
  {
    CALL("canonizeVariableInstance()");
    while ((_term->isVariable()) && _unifier[_term->var()])
      {
	_term = _unifier[_term->var()];
      };  
  }; 
}; 
#endif
