/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Master.hpp"
#include "Clause.hpp"
#include "ClauseBackground.hpp"
#include "PagedPassiveClauseStorage.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#  define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
void Master::init(ComponentName* componentNaming)
{
  CALL("init(ComponentName* componentNaming)");
  BK::init(options);
  _memoryReleaseInProgress = false;
  BK::init(_timer);
  _continuation = Continuation_Init;
  _terminated = false;
  _proof = 0;
  BK::init(_parameters);
  BK::init(_statistics);
  BK::init(_unifier);
  BK::init(_passive);
  BK::init(_pagedPassive);
  BK::init(_reservedPassive,Clause::MainSetReservedPassive,"_reservedPassive");
  BK::init(_active,Clause::MainSetActive,"_active");
  BK::init(_history,Clause::MainSetHistory,"_history");
  BK::init(_answers,Clause::MainSetAnswers,"_answers");
  BK::init(_trash,Clause::MainSetTrash,"_trash");
#ifdef DEBUG_NAMESPACE
  BK::init(_queueForBackSimplification,"_queueForBackSimplification");
  BK::init(_queueForIntegrationIntoBackwardDemodIndex,"_queueForIntegrationIntoBackwardDemodIndex");
  BK::init(_queueForIntegrationIntoBSIndex,"_queueForIntegrationIntoBSIndex");
#else
  BK::init(_queueForBackSimplification);
  BK::init(_queueForIntegrationIntoBackwardDemodIndex); 
  BK::init(_queueForIntegrationIntoBSIndex);   
#endif
  BK::init(_infDispatcher);
  BK::init(_forwardSubsumption);
  _BSIndex.init();
  _backSubsumption.init(&_BSIndex,&_clauseNumIndex);
  BK::init(_backFolding,&_BSIndex,&_clauseNumIndex,&_mainClauseCollector);
  BK::init(_forwardDemodulation);
  BK::init(_backDemodulation);
  BK::init(_termSharing);
  BK::init(_literalSharing,&_termSharing);
  BK::init(_litSelection);
  BK::init(_standardLiteralOrdering);
  BK::init(_clauseAssembler,&_literalSharing,&_litSelection);
#ifdef DEBUG_NAMESPACE
  BK::init(_newClauseQueue,"Master::_newClauseQueue");
  BK::init(_inputClauseCollector,"Master::_inputClauseCollector");
  BK::init(_reanimatedClauseCollector,"Master::_reanimatedClauseCollector");
  BK::init(_mainClauseCollector,"Master::_mainClauseCollector");
#else
  BK::init(_newClauseQueue);
  BK::init(_inputClauseCollector);
  BK::init(_reanimatedClauseCollector);
  BK::init(_mainClauseCollector);
#endif
  BK::init(_newInputClause);
  BK::init(_retainedClause);
  BK::init(_reanimatedClause);
  BK::init(_propositionalSplitting,componentNaming);
  BK::init(_negativeEqualitySplitting,componentNaming);
  BK::init(_forwardProcessing);
  BK::init(_forwardProcessingForSupport);
  BK::init(_clauseNumIndex);
  _sharedReanimatedClause = 0;
  linkObjects();
  setStrategy();
} 
void Master::destroy()
{
  CALL("destroy()");
  if (DestructionMode::isThorough())
    {	
      ASSERT(!_queueForBackSimplification.isLocked());
      ASSERT(!_queueForIntegrationIntoBackwardDemodIndex.isLocked());
      ASSERT(!_queueForIntegrationIntoBSIndex.isLocked());
      if (_infDispatcher.isLocked()) _infDispatcher.abandonCurrentQuery();
      if (_backDemodulation.isLocked()) _backDemodulation.abandonCurrentQuery();
      ASSERT(!_forwardDemodulation.isLocked());
      ASSERT(!_backSubsumption.isLocked());
      ASSERT(!_backFolding.isLocked());
      ASSERT(!_BSIndex.isLocked());
      ASSERT(!_forwardSubsumption.isLocked());
      ASSERT(!_clauseNumIndex.isLocked());
      ASSERT(!_clauseAssembler.sharingIsLocked());
      switch (_parameters.main_alg)
	{
	case 0: destroyPassive0(); break;
	case 3: 
	  ASSERT(_passive.isEmpty());
	  _pagedPassive.reset(); 
	  break;
	default: ICP("ICP0"); break; 
	} 
      Clause* cl;
      while (_reservedPassive.Nonempty())
	{
	  cl = _reservedPassive.Top();
	  enqueueForDisposal(cl);
	}
      _reservedPassive.Purge(); 
      while (_active.Nonempty())
	{
	  cl = _active.Top();
	  enqueueForDisposal(cl);
	}
      _active.Purge(); 
      while (_answers.Nonempty())
	{
	  cl = _answers.Top();
	  enqueueForDisposal(cl);
	}
      _answers.Purge(); 
      if (_sharedReanimatedClause)
	{
	  ASSERT(_sharedReanimatedClause->mainSet() == Clause::MainSetFree);
	  enqueueForDisposal(_sharedReanimatedClause);
	}
      disposeObsolete();
      _history.DestroyAllClauses();  
    }
  BK::destroy(_clauseNumIndex);
  BK::destroy(_forwardProcessingForSupport);
  BK::destroy(_forwardProcessing);
  BK::destroy(_negativeEqualitySplitting);
  BK::destroy(_propositionalSplitting);
  BK::destroy(_reanimatedClause);
  BK::destroy(_retainedClause);
  BK::destroy(_newInputClause);
  BK::destroy(_mainClauseCollector);
  BK::destroy(_reanimatedClauseCollector);
  BK::destroy(_inputClauseCollector);
  BK::destroy(_newClauseQueue);
  BK::destroy(_clauseAssembler);
  BK::destroy(_standardLiteralOrdering);
  BK::destroy(_litSelection);
  BK::destroy(_literalSharing);
  BK::destroy(_termSharing);
  BK::destroy(_backDemodulation);
  BK::destroy(_forwardDemodulation);
  _backSubsumption.destroy();
  BK::destroy(_backFolding);
  _BSIndex.destroy();
  BK::destroy(_forwardSubsumption);
  BK::destroy(_infDispatcher);
  BK::destroy(_queueForIntegrationIntoBSIndex);
  BK::destroy(_queueForIntegrationIntoBackwardDemodIndex);
  BK::destroy(_queueForBackSimplification);
  BK::destroy(_trash);
  BK::destroy(_answers);
  BK::destroy(_history);
  BK::destroy(_active);
  BK::destroy(_reservedPassive);
  BK::destroy(_pagedPassive);
  BK::destroy(_passive);
  BK::destroy(_unifier);
  BK::destroy(_parameters);
  BK::destroy(_statistics);
  BK::destroy(_timer);
  BK::destroy(options);
  BK_CORRUPT(*this);
} 
void Master::reset() 
{
  CALL("reset()");
  _unifier.reset();
  Unifier::setCurrent(&_unifier);
  switch (_parameters.main_alg)
    {
    case 0: destroyPassive0(); break;
    case 3: 
      ASSERT(_passive.isEmpty());
      _pagedPassive.reset();
      break;
    default: ICP("ICP0"); break; 
    } 
  Clause* cl;
  while (_reservedPassive.Nonempty())
    {
      cl = _reservedPassive.Top();
      enqueueForDisposal(cl);
    }
  _reservedPassive.Purge(); 
  while (_active.Nonempty())
    {
      cl = _active.Top();
      enqueueForDisposal(cl);
    }
  _active.Purge(); 
  while (_answers.Nonempty())
    {
      cl = _answers.Top();
      enqueueForDisposal(cl);
    }
  _answers.Purge(); 
  if (_sharedReanimatedClause)
    {
      ASSERT(_sharedReanimatedClause->mainSet() == Clause::MainSetFree);
      enqueueForDisposal(_sharedReanimatedClause);
    }
  disposeObsolete();
  _history.DestroyAllClauses();  
  _queueForBackSimplification.reset();
  _queueForIntegrationIntoBackwardDemodIndex.reset();
  _queueForIntegrationIntoBSIndex.reset();
  _infDispatcher.reset();
  _forwardSubsumption.reset();
  _BSIndex.reset();
  _backSubsumption.reset();
  _backFolding.reset();
  _forwardDemodulation.reset();
  _backDemodulation.reset();
  _termSharing.reset();
  _literalSharing.reset();
  _clauseAssembler.reset();
  _newClauseQueue.reset();
  _inputClauseCollector.reset();
  _reanimatedClauseCollector.reset();
  _mainClauseCollector.reset();
  _newInputClause.clear();
  _retainedClause.clear();
  _reanimatedClause.clear();
  _propositionalSplitting.reset();
  _negativeEqualitySplitting.reset();
  _forwardProcessing.reset();
  _forwardProcessingForSupport.reset();
  _clauseNumIndex.reset();
  options.reset(); 
  resetStatistics();
  _statistics.reset();
  _sharedReanimatedClause = 0;
  linkObjects();
  setStrategy();
} 
void Master::linkObjects()
{
  CALL("linkObjects()");
  _inputClauseCollector.setOutputQueue(&_newClauseQueue);
  _inputClauseCollector.setStatistics(&_statistics);
  _reanimatedClauseCollector.setOutputQueue(0);
  _reanimatedClauseCollector.setStatistics(&_statistics);
  _mainClauseCollector.setOutputQueue(&_newClauseQueue);
  _mainClauseCollector.setStatistics(&_statistics);
  _forwardProcessing.setStatistics(&_statistics);
  _forwardProcessingForSupport.setStatistics(&_statistics);
  _infDispatcher.setStatistics(&_statistics);
  _infDispatcher.setClauseCollector(&_mainClauseCollector);
  _infDispatcher.setForwardProcessing(&_forwardProcessing);
  _backDemodulation.setClauseCollector(&_mainClauseCollector);
} 
void Master::endOfInput()
{
  CALL("endOfInput()");
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());
  ASSERT(_queueForIntegrationIntoBSIndex.empty());
  ASSERT(_backFolding.queueIsEmpty());
  disposeObsolete(); 
  ASSERT(_trash.Empty());
} 
void Master::mainLoop0()
{
  CALL("mainLoop0()");
  Clause* current;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_MainLoop0_ProcessNewInference:
      cancelTermination();
      goto process_new_inference;
    case Continuation_MainLoop0_BackSimplify:
      cancelTermination();
      goto back_simplify;
    case Continuation_MainLoop0_BackFold:
      cancelTermination();
      goto back_fold;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::mainLoop0()");
    }
  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset(); 
  _timer.start();
  _timerIsOn = true;
  _externalTickFlag = false;
  _startMemory = (float)GlobAlloc::occupiedByObjects(); 
 select_current:
  if (_active.Total() >= _parameters.maxActive) 
    { 
      terminate(MaxActive); 
      _timer.stop();
      _timerIsOn = false;
      return; 
    } 
  current = select0(); 
  if (!current) 
    { 
      terminate(EmptyPassive);
      _timer.stop();
      _timerIsOn = false;
      return; 
    }
  if (_parameters.orphan_murder
      && (!current->IsVIP())
      && (current->IsOrphan()))
    {      
      ++_statistics.clauses.retained.passiveOrphansMurdered;
      disposeClause(current);  
      estimateResources();
      goto select_current;
    } 
  makeActive0(current); 
  if (_parameters.outputStreamForPassiveProfiles &&
      (_active.Total() % _parameters.showPassiveProfileInterval == 0))
    {
      _passive.outputProfile(*_parameters.outputStreamForPassiveProfiles);
    }
  eliminateAllocationBufferDeficit0();  
  _infDispatcher.newQuery(current);
 next_inference:
  if (_infDispatcher.nextInference())
    {
    process_new_inference:
      processCollected0();
      if (_proof) 
	{ 
	  terminate(Proof); 
	  _timer.stop();
	  _timerIsOn = false;
	  _continuation = Continuation_MainLoop0_ProcessNewInference;
	  return; 
	}
      _continuation = Continuation_Init;
      if (_passive.Total() >= _parameters.maxPassive) 
	{ 
	  terminate(MaxPassive);  
	  _infDispatcher.abandonCurrentQuery();
	  _timer.stop();
	  _timerIsOn = false;
	  return; 
	}
      internalTick0();
    back_simplify:
      backSimplify();
      if (_terminated) 
	{ 
	  _timer.stop();
	  _timerIsOn = false;
	  _continuation = Continuation_MainLoop0_BackSimplify; 
	  return; 
	}
      _continuation = Continuation_Init;
      integrateIntoBackDemIndexFromQueue();
    back_fold:
      backFold();
      if (_terminated) 
	{ 
	  _timer.stop();
	  _timerIsOn = false;
	  _continuation = Continuation_MainLoop0_BackFold; 
	  return; 
	}
      _continuation = Continuation_Init;
      integrateIntoBSIndexFromQueue();
      if (!_queueForBackSimplification.empty()) 
	goto back_simplify;
      goto next_inference;
    }
  _infDispatcher.abandonCurrentQuery();
  internalTick0(); 
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());
  ASSERT(_queueForIntegrationIntoBSIndex.empty());
  ASSERT(_backFolding.queueIsEmpty());
  disposeObsolete(); 
  ASSERT(_trash.Empty());
  estimateResources();
  if ((_parameters.lrs == 2) || (_parameters.lrs == 3))
    {
      recycleUnreachablePassive0();
    }  
  goto select_current; 
} 
void Master::backSimplify()
{
  CALL("backSimplify()");  
  Clause* backSimplified;
  bool someActiveSimplified;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause0_BackSimplify:
    case Continuation_MainLoop0_BackSimplify:
      someActiveSimplified = true;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::backSimplify()");
    }
  ASSERT(_parameters.backward_demodulation > 0 || 
	 _queueForBackSimplification.empty());
  someActiveSimplified = false;
  bool onBranch;
  if (_parameters.backward_demodulation > 0)
    {
    next_simplifier:
      if (_queueForBackSimplification.dequeue(_currentSimplifier))
	{ 
	  ASSERT(_currentSimplifier->flag(Clause::FlagInQueueForBackSimplification));
	  _currentSimplifier->clearFlag(Clause::FlagInQueueForBackSimplification);
	  if (_backDemodulation.query(_currentSimplifier))
	    { 
	    next_back_demodulated:
	      if (_backDemodulation.nextSimplified(onBranch))
		{      
		  _currentSimplifier->MakeVIP();
		  if (_backDemodulation.collectInference())
		    {  
		      _currentSimplifier->setFlag(Clause::FlagHasChildren);
		      backSimplified = _backDemodulation.simplified();
		      someActiveSimplified = someActiveSimplified 
			|| (backSimplified->mainSet() == Clause::MainSetActive); 
		      if (onBranch)
			{
			  ++_statistics.inferences.simplifying.onBranchBackwardDemodulation;
			  if (backSimplified->mainSet() == Clause::MainSetActive)
			    {
			      ++_statistics.clauses.retained.activeSimplifiedByOnBranchDemodulation;
			    }
			  else
			    {
			      ++_statistics.clauses.retained.passiveSimplifiedByOnBranchDemodulation;
			    }
			}
		      else
			{
			  ++_statistics.inferences.simplifying.backwardDemodulation;
			  if (backSimplified->mainSet() == Clause::MainSetActive)
			    {
			      ++_statistics.clauses.retained.activeSimplifiedByDemodulation;
			    }
			  else
			    {
			      ++_statistics.clauses.retained.passiveSimplifiedByDemodulation;
			    }
			}
		      backSimplified->setFlag(Clause::FlagHasChildren);
		      enqueueForDisposal(backSimplified);        
		    process_new_inference:
		      processCollected0(); 
		      if (_proof) 
			{ 
			  terminate(Proof); 
			  return; 
			}
		      _continuation = Continuation_Init;
		      if (_passive.Total() >= _parameters.maxPassive) 
			{ 
			  terminate(MaxPassive); 
			  _backDemodulation.abandonCurrentQuery();
			  return; 
			} 
		    }
		  internalTick0();
		  goto next_back_demodulated;
		} 
	      _backDemodulation.abandonCurrentQuery();
	    } 
	  goto next_simplifier;
	} 
      if (_parameters.orphan_murder && someActiveSimplified) killActiveOrphans0();
    } 
} 
void Master::backFold()
{
  CALL("backFold()");  
  Clause* backFolded;
  bool someActiveFolded;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause0_BackFold:
    case Continuation_MainLoop0_BackFold:
      someActiveFolded = true;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::backFold()");
    }
  ASSERT(_parameters.back_fold_in_spl || _backFolding.queueIsEmpty());
  someActiveFolded = false;
  if (_parameters.back_fold_in_spl)
    {
    next_definition:
      if (_backFolding.nextDefinition())
	{
	fold_next:
	  if (_backFolding.foldNext())
	    {
	      if (_backFolding.collectResult())
		{
		  backFolded = _backFolding.folded();
		  someActiveFolded = someActiveFolded 
		    || (backFolded->mainSet() == Clause::MainSetActive); 
		  ++_statistics.inferences.simplifying.backwardFolding;
		  if (backFolded->mainSet() == Clause::MainSetActive)
		    {
		      ++_statistics.clauses.retained.activeBackFolded;
		    }
		  else
		    {
		      ++_statistics.clauses.retained.passiveBackFolded;
		    }
		  backFolded->setFlag(Clause::FlagHasChildren);
		  enqueueForDisposal(backFolded);        
		process_new_inference:
		  processCollected0(); 
		  if (_proof) 
		    { 
		      terminate(Proof); 
		      return; 
		    }
		  _continuation = Continuation_Init;
		  if (_passive.Total() >= _parameters.maxPassive) 
		    { 
		      terminate(MaxPassive); 
		      _backDemodulation.abandonCurrentQuery();
		      return; 
		    }
		} 
	      internalTick0();
	      goto fold_next;
	    } 
	  _backFolding.abandonCurrentDefinition();
	  goto next_definition;
	} 
      if (_parameters.orphan_murder && someActiveFolded) killActiveOrphans0();
    } 
  ASSERT_IN(_backFolding.queueIsEmpty(),"END");
} 
void Master::inputClause0(Clause* cl)
{
  CALL("inputClause0(Clause* cl)");
  if (cl->isSupportClause())
    {
      inputSupportClause0(cl);
    }
  else
    inputQueryClause0(cl);
} 
void Master::inputSupportClause0(Clause* cl)
{
  CALL("inputSupportClause0(Clause* cl)");
  ASSERT(cl->isSupportClause());
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputSupportClause0_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputSupportClause0(Clause* cl)");
    }
  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;
  if (_inputClauseCollector.loadClause(cl,
				       ClauseBackground::Prepro,
				       cl))
    {
      _inputClauseCollector.markAsSupport();
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{
	  ASSERT(!_newInputClause.isNull());
	  _forwardProcessingForSupport.load(_newInputClause);
	process_collected_input:
	  while (_forwardProcessingForSupport.anotherRetained(_retainedClause))
	    {
	      ++_statistics.clauses.retained.total;
	      sharedRetainedClause = 
		_clauseAssembler.assembleShared(*_retainedClause);
	      if (!sharedRetainedClause)
		{
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  ASSERT(sharedRetainedClause);
		  sharedRetainedClause->setMainSet(Clause::MainSetNew);
		  if (sharedRetainedClause->isAnswer()) 
		    {
		      rememberAnswer0(sharedRetainedClause);
		      _proof = sharedRetainedClause;
		      terminate(Proof); 
		      _timer.stop();
		      _timerIsOn = false;
		      _continuation = Continuation_InputSupportClause0_ProcessCollectedInput;
		      return;
		    }
		  sharedRetainedClause->SetNumber(_clauseNumberGenerator->generate());      
		  if (_parameters.useClauseNumIndex) 
		    {
		      sharedRetainedClause->protect();
		      _clauseNumIndex.insert(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		    }
		  if (_parameters.showPassive) _parameters.showPassive(sharedRetainedClause);
		  integratePassive(sharedRetainedClause);
		  makeActive0(sharedRetainedClause);
		}
	    } 
	  _continuation = Continuation_Init;
	  integrateIntoBackDemIndexFromQueue(); 
	} 
    } 
} 
void Master::inputQueryClause0(Clause* cl)
{
  CALL("inputQueryClause0(Clause* cl)");
  ASSERT(!cl->isSupportClause());
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      ASSERT_IN(_backFolding.queueIsEmpty(),"Continuation_Init");
      break;
    case Continuation_InputQueryClause0_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    case Continuation_InputQueryClause0_BackSimplify:
      cancelTermination();
      goto back_simplify;
    case Continuation_InputQueryClause0_BackFold:
      cancelTermination();
      goto back_fold;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputQueryClause0(Clause* cl)");
    }
  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;
  if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro,cl))
    {
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{	  
	  ASSERT(!_newInputClause.isNull());
	  _passive.maxPriorityOn();
	  _forwardProcessing.load(_newInputClause);
	process_collected_input:
	  while (_forwardProcessing.anotherRetained(_retainedClause))
	    {
	      ++_statistics.clauses.retained.total;
	      sharedRetainedClause = 
		_clauseAssembler.assembleShared(*_retainedClause);
	      if (!sharedRetainedClause)
		{
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  sharedRetainedClause->setMainSet(Clause::MainSetNew);
		  if (sharedRetainedClause->isAnswer()) 
		    {
		      rememberAnswer0(sharedRetainedClause);
		      _proof = sharedRetainedClause;
		      terminate(Proof); 
		      _timer.stop();
		      _timerIsOn = false;
		      _continuation = Continuation_InputQueryClause0_ProcessCollectedInput;
		      return;
		    }
		  makePassive0(sharedRetainedClause); 
		  if (_parameters.backward_subsumption) backSubsumeBy(sharedRetainedClause);
		  if (_parameters.backward_demodulation > 0 &&
		      eligibleAsBackwardSimplifier(sharedRetainedClause))
		    {
		      ASSERT(!sharedRetainedClause->flag(Clause::FlagInQueueForBackSimplification));
		      sharedRetainedClause->protect();
		      _queueForBackSimplification.enqueue(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		      sharedRetainedClause->setFlag(Clause::FlagInQueueForBackSimplification);
		    } 
		  integratePassive(sharedRetainedClause);
		}
	    } 
	  _continuation = Continuation_Init;
	  _passive.maxPriorityOff(); 
	back_simplify:
	  backSimplify();
	  if (_terminated)
	    {
	      _timer.stop();
	      _timerIsOn = false;
	      _continuation = Continuation_InputQueryClause0_BackSimplify;
	      return; 
	    }
	  ASSERT(_queueForBackSimplification.empty());
	  _continuation = Continuation_Init;
	  integrateIntoBackDemIndexFromQueue();
	back_fold:
	  backFold();
	  if (_terminated)
	    {
	      _timer.stop();
	      _timerIsOn = false;
	      _continuation = Continuation_InputQueryClause0_BackFold;
	      return; 
	    }
	  ASSERT(_backFolding.queueIsEmpty());
	  _continuation = Continuation_Init;
	  integrateIntoBSIndexFromQueue();
	  if (!_queueForBackSimplification.empty()) 
	    goto back_simplify;	
	} 
      ASSERT_IN(_backFolding.queueIsEmpty(),"END");
    } 
  _timer.stop();
  _timerIsOn = false;
} 
void Master::processCollected0()
{
  CALL("processCollected0()");
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause0_BackSimplify:
    case Continuation_InputQueryClause0_BackFold:
    case Continuation_MainLoop0_ProcessNewInference:
    case Continuation_MainLoop0_BackSimplify:
    case Continuation_MainLoop0_BackFold:
      goto process_retained;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::processCollected0()");
    }
  while (!_newClauseQueue.isEmpty())
    {
      _forwardProcessing.load(_newClauseQueue);
    process_retained:  
      while (_forwardProcessing.anotherRetained(_retainedClause))
	{
	  ++_statistics.clauses.retained.total;
	  sharedRetainedClause = 
	    _clauseAssembler.assembleShared(*_retainedClause);
	  if (!sharedRetainedClause)
	    {
	      ++_statistics.clauses.retained.passiveRecycled;
	    }
	  else
	    {
	      sharedRetainedClause->setMainSet(Clause::MainSetNew);
	      if (sharedRetainedClause->isAnswer()) 
		{
		  rememberAnswer0(sharedRetainedClause);
		  _proof = sharedRetainedClause; 
		  return; 
		} 
	      _continuation = Continuation_Init;
	      makePassive0(sharedRetainedClause);
	      if (sharedRetainedClause->weight() > _parameters.maxWeight)
		{        
		  enqueueForDisposal(sharedRetainedClause);
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  if (_parameters.backward_subsumption) 
		    if (backSubsumeBy(sharedRetainedClause))
		      if (_parameters.orphan_murder) 
			killActiveOrphans0();
		  if (_parameters.backward_demodulation > 0 &&
		      eligibleAsBackwardSimplifier(sharedRetainedClause))
		    {
		      ASSERT(!sharedRetainedClause->flag(Clause::FlagInQueueForBackSimplification));
		      sharedRetainedClause->protect();
		      _queueForBackSimplification.enqueue(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		      sharedRetainedClause->setFlag(Clause::FlagInQueueForBackSimplification);
		    }
		  integratePassive(sharedRetainedClause);
		  eliminateAllocationBufferDeficit0();
		}
	    }
	} 
    } 
} 
Clause* Master::select0()
{
  CALL("select0()");
  Clause* selected;
 try_next_passive:
  selected = _passive.Next();
  if (!selected)
    {
      _parameters.lrs = 0;
      updateWeightLimit(VampireKernelConst::MaxClauseSize);
      loadReservedClauses();    
      selected = _passive.Next();
      if (selected) 
	{
	  ASSERT(selected->mainSet() == Clause::MainSetPassive);
	  selected->setMainSet(Clause::MainSetFree);
	}  
      return selected;
    }
  if (selected->practicallyMinResidualSize() > _parameters.maxWeight)
    { 
      _reservedPassive.Insert(selected); 
      goto try_next_passive;         
    }
  if (selected)     
    {
      ASSERT(selected->mainSet() == Clause::MainSetPassive);
      selected->setMainSet(Clause::MainSetFree);
    } 
  return selected; 
} 
void Master::makeActive0(Clause* cl)
{
  CALL("makeActive0(Clause* cl)");
  ++(_statistics.clauses.retained.selected);
  cl->setFlag(Clause::FlagHasChildren); 
  _active.Insert(cl);
  _infDispatcher.makeActive(cl);
  if (_parameters.showActive) _parameters.showActive(cl); 
} 
void Master::makePassive0(Clause* cl)
{
  CALL("makePassive0(Clause* cl)");
  cl->SetNumber(_clauseNumberGenerator->generate());
  cl->setMainSet(Clause::MainSetPassive);
  _passive.Insert(cl);
  if (_parameters.useClauseNumIndex) {
    cl->protect();
    _clauseNumIndex.insert(cl);
    cl->unprotect();
  }
  if (_parameters.showPassive) {
    _parameters.showPassive(cl);
  }
} 
void Master::rememberAnswer0(Clause* cl)
{
  CALL("rememberAnswer0(Clause* cl)");
  cl->SetNumber(_clauseNumberGenerator->generate());
  _answers.Insert(cl);
  if (_parameters.useClauseNumIndex) 
    {
      cl->protect();
      _clauseNumIndex.insert(cl);
      cl->unprotect();
    }
  if (_parameters.showPassive) _parameters.showPassive(cl);
} 
void Master::integratePassive(Clause* cl)
{
  CALL("integratePassive(Clause* cl)");
  if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
  if (_parameters.backward_subsumption ||
      _parameters.back_fold_in_spl) 
    {
      if (_BSIndex.isLocked())
	{
	  cl->protect();
	  _queueForIntegrationIntoBSIndex.enqueue(cl);
	  cl->unprotect();
	  cl->setFlag(Clause::FlagInQueueForIntegrationIntoBSIndex);
	}
      else
	integrateIntoBackwardSubsumptionIndex(cl);
    }
  if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
  if (_parameters.backward_demodulation > 0) 
    {
      if (_backDemodulation.isLocked())
	{
	  cl->protect();
	  _queueForIntegrationIntoBackwardDemodIndex.enqueue(cl);
	  cl->unprotect();
	  cl->setFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
	}
      else 
	{
	  cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
	  integrateIntoBackDemodulationIndex(cl);
	}
    }
} 
void Master::integrateIntoBackDemIndexFromQueue()
{
  CALL("integrateIntoBackDemIndexFromQueue()");
  ASSERT(!_backDemodulation.isLocked());
  Clause* cl;
  while (_queueForIntegrationIntoBackwardDemodIndex.dequeue(cl))
    { 
      ASSERT(cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex));
      ASSERT(cl->mainSet() == Clause::MainSetPassive);
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
      integrateIntoBackDemodulationIndex(cl);
    }
} 
void Master::integrateIntoBSIndexFromQueue()
{
  CALL("integrateIntoBSIndexFromQueue()");
  ASSERT(!_BSIndex.isLocked());
  Clause* cl;
  while (_queueForIntegrationIntoBSIndex.dequeue(cl))
    { 
      ASSERT(cl->flag(Clause::FlagInQueueForIntegrationIntoBSIndex));
      ASSERT(cl->mainSet() == Clause::MainSetPassive);
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBSIndex);
      integrateIntoBackwardSubsumptionIndex(cl);
    }
} 
void Master::removeFromIndices(Clause* cl)
{
  CALL("removeFromIndices(Clause* cl)");
  if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl);
  if (cl->flag(Clause::FlagInForwardDemodulationIndex)) removeFromForwardDemodulationIndex(cl);
  if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) removeFromBackwardSubsumptionIndex(cl);
  if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) removeFromForwardSubsumptionIndex(cl);
} 
bool Master::canBeEnqueuedForDisposal(Clause* cl)
{
  CALL("canBeEnqueuedForDisposal(Clause* cl)");
  if (cl->mainSet() == Clause::MainSetTrash) return true;
  if (cl->flag(Clause::FlagInQueueForBackSimplification) &&
      _queueForBackSimplification.isLocked())
    return false;
  if (cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex) &&
      _queueForIntegrationIntoBackwardDemodIndex.isLocked())
    return false;
  if (cl->flag(Clause::FlagInQueueForIntegrationIntoBSIndex) &&
      _queueForIntegrationIntoBSIndex.isLocked())
    return false;
  return true;
} 
void Master::enqueueForDisposal(Clause* cl)
{
  CALL("enqueueForDisposal(Clause* cl)"); 
  ASSERT(canBeEnqueuedForDisposal(cl));
  if (cl->mainSet() == Clause::MainSetTrash) return;
  switch (cl->mainSet()) 
    {
    case Clause::MainSetFree: break;
    case Clause::MainSetNew:
      ICP("ICP0");
      break;
    case Clause::MainSetPassive: 
      _passive.Remove(cl); 
      cl->setMainSet(Clause::MainSetFree);
      break;
    case Clause::MainSetReservedPassive: _reservedPassive.Remove(cl); break;    
    case Clause::MainSetActive: _active.Remove(cl); break;
    case Clause::MainSetAnswers: _answers.Remove(cl); break;
#ifdef DEBUG_NAMESPACE
    default: 
      ICP("ICP1"); 
      return;
#else
    default: 
      RuntimeError::report("Wrong clause affiliation.");
#endif
    }
  ASSERT(cl->mainSet() == Clause::MainSetFree);
  if (cl->flag(Clause::FlagInQueueForBackSimplification))
    {
      ASSERT(!_queueForBackSimplification.isLocked());
      _queueForBackSimplification.removeFirst(cl); 
      cl->clearFlag(Clause::FlagInQueueForBackSimplification);
    }
  if (cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex))
    {
      ASSERT(!_queueForIntegrationIntoBackwardDemodIndex.isLocked());
      _queueForIntegrationIntoBackwardDemodIndex.removeFirst(cl); 
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
    }
  if (cl->flag(Clause::FlagInQueueForIntegrationIntoBSIndex))
    {
      ASSERT(!_queueForIntegrationIntoBSIndex.isLocked());
      _queueForIntegrationIntoBSIndex.removeFirst(cl); 
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBSIndex);
    }
  bool canBeDestroyed = !cl->isProtected();
  if (!_infDispatcher.isLocked())
    {
      if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
    }
  else
    {
      if (cl->flag(Clause::FlagInInferenceDispatcherAsActive))
	canBeDestroyed = false;
    }
  if (!_backDemodulation.isLocked())
    {
      if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl); 
    }
  else
    if (cl->flag(Clause::FlagInBackDemodulationIndex)) canBeDestroyed = false;
  if (!_forwardDemodulation.isLocked())
    {
      if (cl->flag(Clause::FlagInForwardDemodulationIndex)) 
	removeFromForwardDemodulationIndex(cl);
    }
  else
    if (cl->flag(Clause::FlagInForwardDemodulationIndex)) 
      canBeDestroyed = false;
  if (!_BSIndex.isLocked() && !_clauseNumIndex.isLocked())
    {
      if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) 
	removeFromBackwardSubsumptionIndex(cl);
    }
  else 
    if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) 
      canBeDestroyed = false;
  if (!_forwardSubsumption.isLocked())
    {
      if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) 
	removeFromForwardSubsumptionIndex(cl);
    }
  else
    if (cl->flag(Clause::FlagInForwardSubsumptionIndex))
      canBeDestroyed = false;
  if (cl->flag(Clause::FlagHasChildren)) canBeDestroyed = false;
  if (cl->flag(Clause::FlagInClauseNumIndex) &&
      _clauseNumIndex.isLocked())
   canBeDestroyed = false;
  bool shared = cl->flag(Clause::FlagLiteralsShared);
  if (shared && _clauseAssembler.sharingIsLocked())
   canBeDestroyed = false;
  if (canBeDestroyed)
    {
      if (shared)
	{
	  _clauseAssembler.removeFromSharing(cl);
	  cl->clearFlag(Clause::FlagLiteralsShared);
	}
      if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
      if (shared) { cl->Destroy(); } else cl->DestroyUnshared();
    }
  else 
    {
      _trash.Insert(cl);
    }
} 
void Master::disposeObsolete()
{
  CALL("disposeObsolete()"); 
  Clause* disposedClause;
  while (_trash.Nonempty())
    { 
      disposedClause = _trash.Pop();
      ASSERT(disposedClause->mainSet() == Clause::MainSetFree);
      disposeClause(disposedClause);
    }
} 
void Master::disposeClause(Clause* cl)
{
  CALL("disposeClause(Clause* cl)");
  ASSERT(cl->mainSet() == Clause::MainSetFree);
  if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
  removeFromIndices(cl);
  ASSERT(!cl->flag(Clause::FlagInQueueForBackSimplification));
  ASSERT(!cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex));
  if (cl->flag(Clause::FlagHasChildren))
    {
      _history.Insert(cl); 
    }
  else 
    {
      ASSERT(!cl->isProtected());
      bool shared = cl->flag(Clause::FlagLiteralsShared);
      if (shared) 
	{
	  _clauseAssembler.removeFromSharing(cl);
	  cl->clearFlag(Clause::FlagLiteralsShared);
	}
      if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
      if (shared) { cl->Destroy(); } else cl->DestroyUnshared();
    }
} 
void Master::loadReservedClauses()
{
  CALL("loadReservedClauses()");
  Clause* cl;
  while (!_reservedPassive.Empty())
    {
      cl = _reservedPassive.Pop();
      ASSERT(cl->mainSet() == Clause::MainSetFree); 
      cl->setMainSet(Clause::MainSetPassive);
      _passive.Insert(cl); 
    }
} 
bool Master::backSubsumeBy(Clause* cl)
{
  CALL("backSubsumeBy(Clause* cl)");
  if (cl->isNonSubsuming()) return false;
  Clause* subsumedClause;
  bool usedSubsumed = false;
  if (_backSubsumption.query(cl))
    {
      while (_backSubsumption.subsumeNext())
	{ 
	  ++_statistics.inferences.simplifying.backwardSubsumption;
	  cl->MakeVIP();
	  subsumedClause = _backSubsumption.subsumed();
	  ASSERT(subsumedClause != cl);
	  ASSERT(subsumedClause->isPersistent()); 
	  usedSubsumed = 
	    usedSubsumed || 
	    (subsumedClause->mainSet() == Clause::MainSetActive);     
	  if (usedSubsumed)
	    {
	      ++_statistics.clauses.retained.activeSubsumed;
	    }
	  else
	    {
	      ++_statistics.clauses.retained.passiveSubsumed;
	    }
	  enqueueForDisposal(subsumedClause);     
	}
      _backSubsumption.abandonCurrentQuery();
    }
  return (bool)usedSubsumed;
} 
void Master::estimateResources1()
{
  CALL("estimateResources1()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();
  bool noTimeCheck = 
    used_time < 
    (_parameters.simulatedTimeLimit * 
     _parameters.first_time_check);
  bool noMemoryCheck = 
    used_memory < _parameters.memoryLimit * _parameters.first_mem_check;
  if (noTimeCheck && noMemoryCheck) return;
  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.simulatedTimeLimit   - used_time)*moved_from_passive)/used_time);
  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable1(_parameters.reachable);
  updateWeightLimit(_passive.MaxAcceptedWeight());
  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
} 
void Master::estimateResources2()
{
  CALL("estimateResources2()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();
  bool noTimeCheck = used_time < _parameters.simulatedTimeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;
  if (noTimeCheck && noMemoryCheck) return;
  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.simulatedTimeLimit - used_time)*moved_from_passive)/used_time);
  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable2(_parameters.reachable);
  updateWeightLimit(_passive.MaxAcceptedWeight());
  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
} 
void Master::estimateResources3()
{
  CALL("estimateResources3()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();
  bool noTimeCheck = 
    used_time < _parameters.simulatedTimeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;
  if (noTimeCheck && noMemoryCheck) return;
  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.simulatedTimeLimit - used_time)*moved_from_passive)/used_time);
  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable3(_parameters.reachable);
  updateWeightLimit(_passive.MaxAcceptedWeight());
  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
} 
void Master::estimateResources4()
{
  CALL("estimateResources4()");
  ASSERT(_parameters.lrs == 4);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();
  bool noTimeCheck = 
    used_time < _parameters.simulatedTimeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;
  if (noTimeCheck && noMemoryCheck) return;
  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    (LONG_MAX/2) 
    : 
    ((long)(((_parameters.simulatedTimeLimit - used_time)*moved_from_passive)/used_time));
  long reachableByMemory = (noMemoryCheck)? 
    (LONG_MAX/2)
    :
    ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  long currentlyUnreachable = _passive.Size() - _parameters.reachable;
  long newWeightLimit = _parameters.maxWeight;
  if (currentlyUnreachable <= 0)
    { 
      if ((currentlyUnreachable <= _parameters.unreachable) && 
	  (newWeightLimit < ((long)VampireKernelConst::MaxClauseSize)))
	{ 
	  ++newWeightLimit;
	  if (newWeightLimit > _parameters.maxMaxWeight) 
	    {
	      newWeightLimit = _parameters.maxMaxWeight;
	    }
	}
    }
  else 
    { 
      if (currentlyUnreachable >= _parameters.unreachable)
	{
	  if (newWeightLimit > _passive.MaxPresentWeight()) 
	    {
	      newWeightLimit = _passive.MaxPresentWeight(); 
	    }       
	  else 
	    if (newWeightLimit > _passive.MinPresentWeight()) 
	      { 
		--newWeightLimit;
	      }
	}
    }
  _parameters.unreachable = currentlyUnreachable;
  updateWeightLimit(newWeightLimit);
  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
} 
void Master::estimateResources5()
{
  CALL("estimateResources5()");
  ASSERT(_parameters.lrs == 5);
  float used_memory = usedMemory();
  float used_time = _timer.elapsedDeciseconds();
  bool noTimeCheck = 
    used_time < _parameters.simulatedTimeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;
  if (noTimeCheck && noMemoryCheck) return;
  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    (LONG_MAX/2) 
    : 
    ((long)(((_parameters.simulatedTimeLimit - used_time)*moved_from_passive)/used_time));
  long reachableByMemory = (noMemoryCheck)? 
    (LONG_MAX/2)
    :
    ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  long currentlyUnreachable = _passive.Size() - _parameters.reachable;
  long newWeightLimit = _parameters.maxWeight;
  if (currentlyUnreachable <= 0)
    { 
      if ((currentlyUnreachable <= _parameters.unreachable) && 
	  (newWeightLimit < ((long)VampireKernelConst::MaxClauseSize)))
	{ 
	  ++newWeightLimit;
	  if (newWeightLimit > _parameters.maxMaxWeight) 
	    {
	      newWeightLimit = _parameters.maxMaxWeight;
	    }
	}
    }
  else 
    { 
      if (currentlyUnreachable >= _parameters.unreachable)
	{
	  if (newWeightLimit > _passive.MaxPresentWeight()) 
	    {
	      newWeightLimit = _passive.MaxPresentWeight(); 
	    }       
	  else 
	    if (newWeightLimit >= _passive.MinPresentWeight())
	      { 
		--newWeightLimit;
	      }
	}
    }
  _parameters.unreachable = currentlyUnreachable;
  updateWeightLimit(newWeightLimit);
  if (_parameters.showUsagePace) 
    _parameters.showUsagePace(moved_from_passive/used_time);
} 
void Master::updateWeightLimit(long newLimit)
{
  CALL("updateWeightLimit(long newLimit)");
  _parameters.previousMaxWeight = _parameters.maxWeight;
  _parameters.maxWeight = newLimit;
  if (_parameters.maxWeight > _parameters.maxMaxWeight) 
    {
      _parameters.maxWeight = _parameters.maxMaxWeight;
    }
  _infDispatcher.setClauseSizeLimit(_parameters.maxWeight);
  _mainClauseCollector.setSizeLimit(_parameters.maxWeight);
  _forwardProcessing.setClauseSizeLimit(_parameters.maxWeight); 
  _clauseAssembler.setMaxWeight(_parameters.maxWeight); 
  if (_parameters.showWeightChange && 
      (_parameters.previousMaxWeight != _parameters.maxWeight))
    _parameters.showWeightChange(_parameters.previousMaxWeight,
				 _parameters.maxWeight); 
} 
bool Master::completeStrategyUsed()
{
  return !(_parameters.incompleteAPriori
	   || _statistics.completenessViolated()
	   || _parameters.chaining_for_transitive
	   || (_parameters.selection >= 1000)
	   || (_parameters.selection <= -1000)
           || (!Signature::current()->symbolPropertiesPreserveCompleteness()));
} 
Clause* Master::goodActiveCandidateForDisposal()
{
  CALL("goodActiveCandidateForDisposal()");
  ClauseSet::Iter iter(_active);
  Clause* cl;
  do
    {
      cl = iter.clause();
      if (!cl) return static_cast<Clause*>(0);
      iter.next();
    }
  while (cl->isInputOrPreprocessedInputClause() ||
	 !canBeEnqueuedForDisposal(cl));
  Clause* currentCandidate = cl;
  ulong numOfTrials = 100UL;
  ulong numOfImprovements = 5UL;
  while (!iter.isEnd())
    {
      cl = iter.clause();
      iter.next();
      if (!cl->isInputOrPreprocessedInputClause() &&
	  canBeEnqueuedForDisposal(cl))
	{
	  if (cl->weight() > currentCandidate->weight())
	    {
	      currentCandidate = cl;
	      --numOfImprovements;
	      if (!numOfImprovements) return currentCandidate;
	    }
	}
      --numOfTrials;
      if (!numOfTrials) return currentCandidate; 
    }
  return currentCandidate; 
} 
Clause* Master::goodPassiveCandidateForDisposal0()
{
  CALL("goodPassiveCandidateForDisposal0()");
  Clause* cl = _passive.TheWorstUnblockedClause();
  while (cl &&
	 (cl->isInputOrPreprocessedInputClause() ||
	  !canBeEnqueuedForDisposal(cl)))
    {
      cl = _passive.theWorstUnblockedClauseAfter(cl);
    }  
  return cl; 
} 
void Master::recycleUnreachablePassive0()
{
  CALL("recycleUnreachablePassive0()");
  Clause* cl;
 recycle_next:
  if (_passive.UnreachableKept() <= (_passive.CurrentlyKept()/5)) 
    return; 
  cl = _passive.TheMostUnreachableClause();    
  if (!cl) return;  
  ASSERT(cl->mainSet() == Clause::MainSetPassive);
  _passive.Cancel(cl);
  cl->setMainSet(Clause::MainSetFree);
  enqueueForDisposal(cl);
  ++_statistics.clauses.retained.passiveRecycled;
  goto recycle_next;
} 
bool Master::releaseALotOfMemory0()
{
  CALL("releaseALotOfMemory0()");
  bool res = false;
  ulong numOfClausesToRecycle =
    _passive.Size() / 10;
  Clause* cl;
  while (numOfClausesToRecycle &&
	 ((cl = goodPassiveCandidateForDisposal0())))
    {
      _passive.Cancel(cl);
      cl->setMainSet(Clause::MainSetFree);
      enqueueForDisposal(cl);
      ++_statistics.clauses.retained.passiveRecycled;
      --numOfClausesToRecycle;
      res = true;  
      BK::GlobalStopFlag::check();
    }
#ifdef CHANGE_WEIGHT_LIMIT_AFTER_MAJOR_RECYCLING
  if (res) {
    _parameters.maxMaxWeight = _passive.MaxPresentWeight();
    updateWeightLimit(_passive.MaxPresentWeight());
  }
#endif
#ifdef DONT_RECYCLE_ACTIVE
  numOfClausesToRecycle = _active.Size() / 10;
  while (numOfClausesToRecycle) {
    Clause* cand = goodActiveCandidateForDisposal();
    if (!cand) return res;
    --numOfClausesToRecycle;
    res = true;
    enqueueForDisposal(cand);
    ++_statistics.clauses.retained.activeRecycled;
    BK::GlobalStopFlag::check();
  }
#endif
  return res;
} 
void Master::killActiveOrphans0()
{
  CALL("killActiveOrphans0()");
  ASSERT(!_active.mainIteratorIsEngaged());
  _active.resetMainIterator();
  for (Clause* cl = _active.Next(); cl; cl = _active.Next())
    if ((cl->isPersistent())
	&& (!cl->IsVIP())
	&& (cl->IsOrphan()))
      {
	enqueueForDisposal(cl);
	++_statistics.clauses.retained.activeOrphansMurdered;
      }
} 
void Master::eliminateAllocationBufferDeficit0()
{
  CALL("eliminateAllocationBufferDeficit0()");
  if (BK::GlobAlloc::activeBufferDeficit()) {
    long disposalBound = 
      (long)(BK::GlobAlloc::occupiedByObjects() - 
	     (_parameters.allocation_buffer_price * 
	      BK::GlobAlloc::activeBufferDeficit()));
    do {
      Clause* cl = _passive.TheWorstUnblockedClause();    
      if (!cl) {
	return;
      }
      ASSERT(cl->mainSet() == Clause::MainSetPassive);
      ++_statistics.clauses.retained.passiveRecycled;
      _passive.Cancel(cl);
      cl->setMainSet(Clause::MainSetFree);
      enqueueForDisposal(cl);
    }
    while (BK::GlobAlloc::activeBufferDeficit() && 
	   BK::GlobAlloc::occupiedByObjects() > disposalBound);
    if (BK::GlobAlloc::activeBufferDeficit()) {
      BK::GlobAlloc::annulBufferDeficit();
    }
  }
} 
void Master::destroyPassive0() 
{ 
  CALL("destroyPassive0()");
  _passive.DestroyAllClauses();
  _passive.ResetIterator();
  for (Clause* cl = _passive.IterNext(); cl; cl = _passive.IterNext())
    enqueueForDisposal(cl);
} 
void Master::initParameters()
{
  CALL("Master::initParameters()");
  _parameters.incompleteAPriori = options.incompleteAPriori;
  _parameters.main_alg = options.main_alg;
  _parameters.binary_resolution = options.binary_resolution;
  _parameters.no_inferences_between_rules = options.no_inferences_between_rules;
  _parameters.splitting = options.splitting; 
  _parameters.small_basic_comp_in_hyp_splitting = options.small_basic_comp_in_hyp_splitting;
  _parameters.fold_single_comp_in_splitting = options.fold_single_comp_in_splitting;
  _parameters.single_ground_lit_comp_min_size = options.single_ground_lit_comp_min_size;
  _parameters.reuse_names_for_negated_ground_lits_in_splitting = options.reuse_names_for_negated_ground_lits_in_splitting;
  _parameters.basic_comp_in_spl_is_max_wrt = options.basic_comp_in_spl_is_max_wrt;
  _parameters.back_fold_in_spl = 
    options.back_fold_in_spl && options.splitting;
  _parameters.neg_eq_splitting = options.neg_eq_splitting;  
  _parameters.neg_eq_spl_var_count_limit = options.neg_eq_spl_var_count_limit;  
  _parameters.paramodulation = options.paramodulation;  
  _parameters.paramodulate_from_variables = options.paramodulate_from_variables;
  _parameters.sim_back_sup = options.sim_back_sup;
  _parameters.forw_sup_no_dup_red = options.forw_sup_no_dup_red;
  _parameters.back_sup_no_dup_red = options.back_sup_no_dup_red;
  _parameters.prohibit_non_basic_superposition = options.prohibit_non_basic_superposition;
  _parameters.prohibit_non_basic_fd = options.prohibit_non_basic_fd;
  _parameters.prohibit_non_basic_bd = options.prohibit_non_basic_bd;
  _parameters.reduce_non_basic = options.reduce_non_basic;
  _parameters.input_non_basicness = options.input_non_basicness;
  _parameters.run_time_non_basicness = options.run_time_non_basicness;
  _parameters.selection = options.selection;
  _parameters.inherited_negative_selection = options.inherited_negative_selection; 
  _parameters.external_evaluation = options.external_evaluation;
  _parameters.handle_equality = options.handle_equality;
  _parameters.forward_subsumption = options.forward_subsumption;
  _parameters.simplify_by_fs = options.simplify_by_fs;
  _parameters.main_premise_subsumption_expiry =
    (options.forward_subsumption) ? 
    options.main_premise_subsumption * (options.timeLimit/100)
    :
    0.0;
  _parameters.side_premise_subsumption_expiry =
    (options.forward_subsumption) ?  
    options.side_premise_subsumption * (options.timeLimit/100)
    :
    0.0;
  _parameters.main_selected_lit_demodulation_expiry = 
    (options.forward_demodulation) ?
    options.main_selected_lit_demodulation * (options.timeLimit/100)
    :
    0.0;
  _parameters.side_selected_lit_demodulation_expiry = 
    (options.forward_demodulation) ?
    options.side_selected_lit_demodulation * (options.timeLimit/100)
    :
    0.0;
  _parameters.forward_demodulation = options.forward_demodulation;
  _parameters.fd_on_splitting_branches = options.fd_on_splitting_branches;
  _parameters.normalize_by_commutativity = options.normalize_by_commutativity;
  _parameters.eq_taut_mod_C = options.eq_taut_mod_C;
  _parameters.prop_taut_mod_C = options.prop_taut_mod_C;
  _parameters.trivial_eq_res_mod_C = options.trivial_eq_res_mod_C;
  _parameters.identical_lit_factoring_mod_C = options.identical_lit_factoring_mod_C;
  _parameters.chaining_for_transitive = options.chaining_for_transitive;
  _parameters.chaining_through_variables = options.chaining_through_variables;
  _parameters.backward_demodulation = options.backward_demodulation;
  _parameters.bd_on_splitting_branches = options.bd_on_splitting_branches;
  _parameters.simplify_by_eq_res = options.simplify_by_eq_res;
  _parameters.backward_subsumption = options.backward_subsumption;
  _parameters.orphan_murder = options.orphan_murder;
  _parameters.lrs = options.lrs;
  _parameters.orient_pos_equalities_in_support = options.orient_pos_equalities_in_support;
  _parameters.orient_neg_equalities_in_support = options.orient_neg_equalities_in_support;
  _parameters.answer_predicate_table = options.answer_predicate_table;
  _parameters.answer_predicate_table_size = options.answer_predicate_table_size;
  _parameters.age_weight_ratio_numerator = options.age_weight_ratio_numerator;
  _parameters.age_weight_ratio_denominator = options.age_weight_ratio_denominator;
  _parameters.best_good_weight_ratio_numerator = options.best_good_weight_ratio_numerator;
  _parameters.best_good_weight_ratio_denominator = options.best_good_weight_ratio_denominator;
  _parameters.timeLimit = options.timeLimit;
  _parameters.simulatedTimeLimit = options.simulatedTimeLimit; 
  _parameters.memoryLimit = options.memoryLimit;
  _parameters.allocation_buffer_price = options.allocation_buffer_price;
  _parameters.maxActive = options.maxActive;
  _parameters.maxPassive = options.maxPassive;
  _parameters.maxWeight = (options.maxWeight <= ((long)VampireKernelConst::MaxClauseSize)) ? options.maxWeight : VampireKernelConst::MaxClauseSize;
  _parameters.previousMaxWeight = options.maxWeight;
  _parameters.maxMaxWeight = options.maxWeight;
  _parameters.maxInferenceDepth = options.maxInferenceDepth;
  _parameters.first_time_check = options.first_time_check; 
  _parameters.first_mem_check = options.first_mem_check;   
  _parameters.reachable = LONG_MAX;
  _parameters.unreachable = (-1)*(LONG_MAX/2);     
  _parameters.outputStreamForGenerated = options.outputStreamForGenerated;
  _parameters.showPassive = options.showPassive;
  _parameters.showActive = options.showActive;
  _parameters.showLitProp = options.showLitProp;
  _parameters.outputStreamForPassiveProfiles = options.outputStreamForPassiveProfiles;
  _parameters.showPassiveProfileInterval = options.showPassiveProfileInterval;
  _parameters.showWeightChange = options.showWeightChange;
  _parameters.showUsagePace = options.showUsagePace;
  _parameters.useClauseNumIndex = 
    options.backward_subsumption || options.back_fold_in_spl;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _parameters.step_resolution_literal_selection = options.step_resolution_literal_selection;
#endif
} 
void Master::passParametersToObjects()
{
  CALL("passParametersToObjects()");
  _passive.setAgeWeightRatio(_parameters.age_weight_ratio_numerator,
			     _parameters.age_weight_ratio_denominator);
  _passive.setBestGoodWeightRatio(_parameters.best_good_weight_ratio_numerator,
				  _parameters.best_good_weight_ratio_denominator);
  _pagedPassive.setAgeWeightRatio(_parameters.age_weight_ratio_numerator,
				  _parameters.age_weight_ratio_denominator);
  _pagedPassive.setBestGoodWeightRatio(_parameters.best_good_weight_ratio_numerator,
				       _parameters.best_good_weight_ratio_denominator);
  TemporaryLiteral::Context::getCurrent()->setInversePolarityForSelection(_parameters.selection < 0);
  TemporaryLiteral::Context::getCurrent()->setChainingTable(_parameters.chaining_for_transitive);
  TemporaryLiteral::Context::getCurrent()->setShowLiteralProperties(_parameters.showLitProp);
  _litSelection.setSplittingFlag(_parameters.splitting);
  _litSelection.setFunNum(_parameters.selection);
  _litSelection.setMainLiteralOrdering(&_standardLiteralOrdering);
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _litSelection.setStepResolutionLiteralSelection(_parameters.step_resolution_literal_selection);
#endif
  _clauseAssembler.setInheritNegativeSelection(_parameters.inherited_negative_selection);
  _clauseAssembler.setEqualityIsPossibleFlag(_parameters.handle_equality);
  _clauseAssembler.setMaxMaxWeight(_parameters.maxMaxWeight);
  _clauseAssembler.setMaxWeight(_parameters.maxWeight);
  _inputClauseCollector.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
  _inputClauseCollector.setShowCompleteClauses(_parameters.outputStreamForGenerated);  
  _inputClauseCollector.setSizeLimit(_parameters.maxWeight);
  _reanimatedClauseCollector.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
  _reanimatedClauseCollector.setShowCompleteClauses(_parameters.outputStreamForGenerated);
  _reanimatedClauseCollector.setSizeLimit(VampireKernelConst::MaxClauseSize);
  _mainClauseCollector.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
  _mainClauseCollector.setShowCompleteClauses(_parameters.outputStreamForGenerated);  
  _mainClauseCollector.setSizeLimit(_parameters.maxWeight);
  _mainClauseCollector.setMarkSubstitutionPositions(_parameters.run_time_non_basicness);
  if (_parameters.splitting) {
    _propositionalSplitting.setSmallBasicComponentInHyper(_parameters.small_basic_comp_in_hyp_splitting);
    _propositionalSplitting.setFoldSingleComponents(_parameters.fold_single_comp_in_splitting);
    _propositionalSplitting.setSingleGroundLitCompMinSize(_parameters.single_ground_lit_comp_min_size);
    _propositionalSplitting.setReuseNamesForNegations(_parameters.reuse_names_for_negated_ground_lits_in_splitting);
    _propositionalSplitting.setCompMakerForBasicComp(_parameters.basic_comp_in_spl_is_max_wrt);
    _backFolding.setFoldForSplitting(_parameters.back_fold_in_spl);
    _backFolding.setFoldSingleComponentsForPropSpl(_parameters.fold_single_comp_in_splitting);
    _backFolding.setReusePropSplNamesForNegations(_parameters.reuse_names_for_negated_ground_lits_in_splitting);
  } 
  if (_parameters.neg_eq_splitting > 0L) {
    _negativeEqualitySplitting.setMinimalEligibleArgSize(_parameters.neg_eq_splitting + 1);
    _negativeEqualitySplitting.setMaximalEligibleVarCount(_parameters.neg_eq_spl_var_count_limit);
  }
  else {
    _negativeEqualitySplitting.setMinimalEligibleArgSize(ULONG_MAX);
  }
  _forwardProcessing.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
  _forwardProcessing.setShowCompleteClauses(_parameters.outputStreamForGenerated);  
  _forwardProcessing.setEqualityPossible(_parameters.handle_equality);
  if (_parameters.forward_subsumption) {
    _forwardProcessing.setForwardSubsumption(&_forwardSubsumption);
  }
  else {
    _forwardProcessing.setForwardSubsumption(0);
  }
  _forwardProcessing.setForwardSubsumptionResolution(_parameters.simplify_by_fs);
  if (_parameters.forward_demodulation > 0) {
    _forwardProcessing.setForwardDemodulation(&_forwardDemodulation);
    _forwardProcessing.setOnBranchForwardDemodulation(_parameters.fd_on_splitting_branches);
  }
  else {
    _forwardProcessing.setForwardDemodulation(0);
  }
  _forwardProcessing.setDestructiveEqResolution(_parameters.simplify_by_eq_res); 
  if (_parameters.splitting) {
    _forwardProcessing.setPropositionalSplitting(&_propositionalSplitting);
  }
  else {
    _forwardProcessing.setPropositionalSplitting(0);
  }
  if (_parameters.neg_eq_splitting > 0L) {
    _forwardProcessing.setNegativeEqualitySplitting(&_negativeEqualitySplitting);
  }
  else {
    _forwardProcessing.setNegativeEqualitySplitting(0);
  }
  _forwardProcessing.setCommutativityTableForNormalisation(_parameters.normalize_by_commutativity);
  _forwardProcessing.setCommutativityTableForEqTaut(_parameters.eq_taut_mod_C);
  _forwardProcessing.setCommutativityTableForPropTaut(_parameters.prop_taut_mod_C);
  _forwardProcessing.setCommutativityTableForTrivialEqRes(_parameters.trivial_eq_res_mod_C);
  _forwardProcessing.setCommutativityTableForIdenticalLitFactoring(_parameters.identical_lit_factoring_mod_C);
  _forwardProcessing.setProhibitNonBasicDemodulation(_parameters.prohibit_non_basic_fd ||
						     _parameters.reduce_non_basic);
  _forwardProcessing.setReduceNonBasic(_parameters.reduce_non_basic);
  _forwardProcessing.setUseInputNonBasicness(_parameters.input_non_basicness);
  _forwardProcessing.setUseRunTimeNonBasicness(_parameters.run_time_non_basicness);
  _forwardProcessing.setClauseSizeLimit(_parameters.maxWeight);
  _forwardProcessing.setInferenceDepthLimit(_parameters.maxInferenceDepth);
  _forwardProcessing.setExternalEvaluation(_parameters.external_evaluation);
  _forwardProcessingForSupport.setShowAnnouncedClauses(_parameters.outputStreamForGenerated);
  _forwardProcessingForSupport.setShowCompleteClauses(_parameters.outputStreamForGenerated);  
  _forwardProcessingForSupport.setEqualityPossible(_parameters.handle_equality);
  _forwardProcessingForSupport.setForwardSubsumption(0);
  _forwardProcessingForSupport.setForwardSubsumptionResolution(0UL);
  _forwardProcessingForSupport.setForwardDemodulation(0);
  _forwardProcessingForSupport.setDestructiveEqResolution(false); 
  _forwardProcessingForSupport.setPropositionalSplitting(0);
  _forwardProcessingForSupport.setNegativeEqualitySplitting(0);
  _forwardProcessingForSupport.setCommutativityTableForNormalisation(0);
  _forwardProcessingForSupport.setClauseSizeLimit(_parameters.maxWeight);
  _forwardProcessingForSupport.setInferenceDepthLimit(_parameters.maxInferenceDepth);
  _forwardProcessingForSupport.setExternalEvaluation(_parameters.external_evaluation);
  _forwardSubsumption.setSubsumptionResolutionMinLength(_parameters.simplify_by_fs);
  _forwardDemodulation.SetDegree(_parameters.forward_demodulation);
  _forwardDemodulation.SetFDOnSplittingBranchesFlag(_parameters.fd_on_splitting_branches);
  _forwardDemodulation.setDemodulateFromSupport(_parameters.orient_pos_equalities_in_support);
  _backDemodulation.setClauseCollector(&_mainClauseCollector);
  _backDemodulation.setDegree(_parameters.backward_demodulation);
  _backDemodulation.setBDOnSplittingBranchesFlag(_parameters.bd_on_splitting_branches);
  _backDemodulation.setDemodulateFromSupport(_parameters.orient_pos_equalities_in_support);
  _backDemodulation.setProhibitDemodulationIntoInputNonBasic(_parameters.input_non_basicness &&
							     (_parameters.prohibit_non_basic_bd ||
							      _parameters.reduce_non_basic));
  _backDemodulation.setCheckSubstSafetyWRTInputNonBasicness(_parameters.input_non_basicness);
  _infDispatcher.setClauseCollector(&_mainClauseCollector);
  _infDispatcher.setForwardProcessing(&_forwardProcessing);
  _infDispatcher.binResFlag = _parameters.binary_resolution;
  _infDispatcher.setNoInfBetweenRulesFlag(_parameters.no_inferences_between_rules);
  _infDispatcher.setParamodulationFlag(_parameters.paramodulation);
  _infDispatcher.setParamodulateFromVariablesFlag(_parameters.paramodulate_from_variables);
  _infDispatcher.setSimBackSup(_parameters.sim_back_sup);
  _infDispatcher.setForwSupNoDuplicateRedexes(_parameters.forw_sup_no_dup_red);
  _infDispatcher.setBackSupNoDuplicateRedexes(_parameters.back_sup_no_dup_red);
  _infDispatcher.setUseInputNonBasicness(_parameters.input_non_basicness &&
					 _parameters.prohibit_non_basic_superposition);
  _infDispatcher.setChainingTable(_parameters.chaining_for_transitive);
  _infDispatcher.setChainingThroughVariablesFlag(_parameters.chaining_through_variables);
  _infDispatcher.setMainPremiseSubsumption(_parameters.main_premise_subsumption_expiry > 0.001);
  _infDispatcher.setSidePremiseSubsumption(_parameters.side_premise_subsumption_expiry > 0.001);
  _infDispatcher.setMainSelectedLitDemodulation(_parameters.main_selected_lit_demodulation_expiry > 0.001);
  _infDispatcher.setSideSelectedLitDemodulation(_parameters.side_selected_lit_demodulation_expiry > 0.001);
  _infDispatcher.setOrientEqualitiesInSupport(_parameters.orient_pos_equalities_in_support,
					      _parameters.orient_pos_equalities_in_support);
  _infDispatcher.setClauseSizeLimit(_parameters.maxWeight);
} 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#  define DEBUG_NAMESPACE "Master::Options"
#endif                  
#include "debugMacros.hpp"
void Master::Options::reset()
{
  incompleteAPriori = false;
  main_alg = 0;
  binary_resolution = true;
  no_inferences_between_rules = false;
  splitting = false;  
  small_basic_comp_in_hyp_splitting = false;
  fold_single_comp_in_splitting = false; 
  neg_eq_splitting = 0L; 
  neg_eq_spl_var_count_limit = 0L; 
  paramodulation = false;
  paramodulate_from_variables = true;
  sim_back_sup = 0;
  forw_sup_no_dup_red = false;
  back_sup_no_dup_red = false;
  prohibit_non_basic_superposition = false;
  prohibit_non_basic_fd = true;
  prohibit_non_basic_bd = true;
  reduce_non_basic = false;
  input_non_basicness = false;
  run_time_non_basicness = false;
  selection = 1;
  inherited_negative_selection = false;
  external_evaluation = false;
  handle_equality = false;
  forward_subsumption = true;
  simplify_by_fs = 0;
  main_premise_subsumption = 0;
  side_premise_subsumption = 0;
  main_selected_lit_demodulation = 0;
  side_selected_lit_demodulation = 0;
  forward_demodulation = 0;
  fd_on_splitting_branches = false;
  normalize_by_commutativity = 0;
  eq_taut_mod_C = 0;
  prop_taut_mod_C = 0;
  trivial_eq_res_mod_C = 0;
  identical_lit_factoring_mod_C = 0;
  chaining_for_transitive = 0;
  chaining_through_variables = true;
  backward_demodulation = 0;
  bd_on_splitting_branches = false;
  simplify_by_eq_res = false;
  backward_subsumption = false;
  orphan_murder = false;
  lrs = 0;
  orient_pos_equalities_in_support = true;
  orient_neg_equalities_in_support = true;
  answer_predicate_table = 0;
  answer_predicate_table_size = 0UL;
  age_weight_ratio_numerator = 1;
  age_weight_ratio_denominator = 0;
  best_good_weight_ratio_numerator = 1;
  best_good_weight_ratio_denominator = 0;
  timeLimit = LONG_MAX/10;
  memoryLimit = (float)LONG_MAX;
  allocation_buffer_price = 1.0;
  maxActive = LONG_MAX;
  maxPassive = LONG_MAX;
  maxWeight = VampireKernelConst::MaxClauseSize;
  maxInferenceDepth = LONG_MAX;
  simulatedTimeLimit = LONG_MAX/10;
  first_time_check = 1.0;
  first_mem_check = 1.0;
  outputStreamForGenerated = 0;
  showPassive = 0;
  showActive = 0;
  showLitProp = false;
  outputStreamForPassiveProfiles = 0;
  showPassiveProfileInterval = LONG_MAX;
  showWeightChange = 0;
  showUsagePace = 0;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  step_resolution_literal_selection = false;
#endif
} 
