/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef MASTER_HPP
#define MASTER_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "GlobalStopFlag.hpp"
#include "GlobalClock.hpp"
#include "RuntimeError.hpp"
#include "ClauseSet.hpp"
#include "PassiveClauseSet.hpp"
#include "TermSharing.hpp"
#include "LiteralSharing.hpp"
#include "ClauseAssembler.hpp"
#include "ForwardSubsumptionFrontEnd.hpp"
#include "ComponentName.hpp"
#include "InferenceDispatcher.hpp"
#include "BSIndexFrontEnd.hpp"
#include "BackwardSubsumption.hpp"
#include "BackwardFolding.hpp"
#include "ForwardDemodulation.hpp"
#include "BackDemodulation.hpp"
#include "ClauseNumIndex.hpp"
#include "FIFO.hpp"
#include "PagedPassiveClauseStorage.hpp"
#include "LiteralSelection.hpp"
#include "StandardLiteralOrdering.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TemporaryClauseCollector.hpp"
#include "TemporaryClauseQueue.hpp"
#include "ForwardProcessing.hpp"
#include "PropositionalSplitting.hpp"
#include "NegativeEqualitySplitting.hpp"
#include "Statistics.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#  define DEBUG_NAMESPACE "Master"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Clause;
  class SymbolInfo;
  class Master
  {
  public:
    enum TerminationReason 
    { 
      UnknownTR,
      Proof, 
      EmptyPassive, 
      MaxPassive, 
      MaxActive 
    }; 
    class Options
    {
    public:
      Options() { reset(); };
      ~Options() {};
      void reset(); 
      bool incompleteAPriori;
      long main_alg;
      bool binary_resolution;
      bool no_inferences_between_rules;
      bool splitting;
      bool small_basic_comp_in_hyp_splitting;
      bool fold_single_comp_in_splitting;
      long single_ground_lit_comp_min_size;
      bool reuse_names_for_negated_ground_lits_in_splitting;
      TempLitListExtras::ComparisonMaker* basic_comp_in_spl_is_max_wrt;
      bool back_fold_in_spl;
      long neg_eq_splitting;   
      long neg_eq_spl_var_count_limit;   
      bool paramodulation; 
      bool paramodulate_from_variables;
      long sim_back_sup;
      bool forw_sup_no_dup_red;
      bool back_sup_no_dup_red;
      bool prohibit_non_basic_superposition;
      bool prohibit_non_basic_fd;
      bool prohibit_non_basic_bd;
      bool reduce_non_basic;
      bool input_non_basicness;
      bool run_time_non_basicness;
      long selection;
      bool inherited_negative_selection;
      bool external_evaluation;
      bool handle_equality;
      bool forward_subsumption;
      long simplify_by_fs;
      long main_premise_subsumption;
      long side_premise_subsumption;
      long main_selected_lit_demodulation;
      long side_selected_lit_demodulation;
      long forward_demodulation;
      bool fd_on_splitting_branches;
      const SymbolInfo* normalize_by_commutativity;
      const SymbolInfo* eq_taut_mod_C;
      const SymbolInfo* prop_taut_mod_C;
      const SymbolInfo* trivial_eq_res_mod_C;
      const SymbolInfo* identical_lit_factoring_mod_C;
      const SymbolInfo* chaining_for_transitive;
      bool chaining_through_variables;
      long backward_demodulation;
      bool bd_on_splitting_branches;
      bool simplify_by_eq_res;
      bool backward_subsumption;
      bool orphan_murder;
      long lrs;
      bool orient_pos_equalities_in_support;
      bool orient_neg_equalities_in_support;
      const bool* answer_predicate_table;
      ulong answer_predicate_table_size;
      long age_weight_ratio_numerator;
      long age_weight_ratio_denominator;
      long best_good_weight_ratio_numerator;
      long best_good_weight_ratio_denominator;
      long timeLimit; 
      long simulatedTimeLimit; 
      float memoryLimit;
      float allocation_buffer_price; 
      long maxActive;
      long maxPassive;
      long maxWeight;
      long maxInferenceDepth;
      float first_time_check;
      float first_mem_check;
      ostream* outputStreamForGenerated;
      void (*showPassive) (const Clause*);        
      void (*showActive) (const Clause*); 
      bool showLitProp;
      ostream* outputStreamForPassiveProfiles;
      long showPassiveProfileInterval;
      void (*showWeightChange) (long,long);
      void (*showUsagePace) (float);
#ifdef VKERNEL_FOR_STEP_RESOLUTION
      bool step_resolution_literal_selection;
#endif
    }; 
  public:
    void init(ComponentName* componentNaming);
    void destroy();
    void reset();
    void resetStatistics() { _statistics.reset(); };
    void setStrategy()
    { 
      CALL("setStrategy()");
      initParameters();
      passParametersToObjects();
    };
    void setClauseNumberGenerator(Clause::NewNumber* ng) 
    {
      _clauseNumberGenerator = ng;
    };
    void endOfInput();
    void setPreprocessingInProgress(bool fl)
    {
      _inputClauseCollector.setPreprocessingInProgress(fl);
      _reanimatedClauseCollector.setPreprocessingInProgress(fl);
      _mainClauseCollector.setPreprocessingInProgress(fl);
      _forwardProcessing.setPreprocessingInProgress(fl);
      _forwardProcessingForSupport.setPreprocessingInProgress(fl);
    };
    const BK::GlobalClock::Timer& timer() const { return _timer; }; 
    bool terminated() const { return _terminated; };
    TerminationReason terminationReason() const { return _terminationReason; };  
    const Clause* proof() const { return _proof; };
    void mainLoop() 
    { 
      setPreprocessingInProgress(false);
      switch (_parameters.main_alg)
	{
	case 0: mainLoop0(); break;
	case 3: mainLoop3(); break; 
	default:
	  BK::RuntimeError::report("Wrong --main_alg value.");
	};     
    };
    void inputClause(Clause* cl) 
    { 
      setPreprocessingInProgress(true);
      switch (_parameters.main_alg)
	{
	case 0: inputClause0(cl); break;
	case 3: inputClause3(cl); break;
	default: 
	  BK::RuntimeError::report("Wrong --main_alg value.");
	};
    };
    BK::Receiver<VK::Clause*>* nameIntroReceiver() 
    {
      return _backFolding.asNameIntroReceiver();
    };
    void externalTick()
    {
      _externalTickFlag = true;
    };
    void suspend()
    {
      if (_timerIsOn) _timer.stop();
    };
    void resume()
    {
      if (_timerIsOn) _timer.start();
    };
    bool releaseMemory(ulong size) 
    {
      CALL("releaseMemory(ulong size)");
      if (_memoryReleaseInProgress) return false;
      _memoryReleaseInProgress = true;
      bool res;
      switch (_parameters.main_alg)
	{
	case 0: res = false; break;
	case 3: res = releaseMemory3(size); break;
	default: ICP("ICP0"); res = false;
	};     
      _memoryReleaseInProgress = false;
      return res;
    };
    bool releaseALotOfMemory() 
    {
      CALL("releaseALotOfMemory()");
      if (_memoryReleaseInProgress) return false;
      _memoryReleaseInProgress = true;
      bool res;
      switch (_parameters.main_alg)
	{
	case 0: res = releaseALotOfMemory0(); break;
	case 3: res = releaseALotOfMemory3(); break;
	default: ICP("ICP0"); res = false;  
	};    
      _memoryReleaseInProgress = false;
      return res;    
    };
    void passiveResetIter() { _passive.ResetIterator(); };
    Clause* nextPassive() { return _passive.IterNext(); };
    void historyResetIter() { _history.resetMainIterator(); };
    Clause* nextHistory() { return _history.Next(); };
    void terminate(TerminationReason tr) 
    {
      _terminated = true; 
      _terminationReason = tr; 
    };
    void cancelTermination()
    {
      CALL("cancelTermination()");
      ASSERT(_terminated);
      ASSERT(_proof);
      _terminated = false;
      _proof = 0;
      ASSERT(!_timerIsOn);
      _timer.start();
      _timerIsOn = true;
    };
    const Statistics& statistics() const { return _statistics; };
    bool completeStrategyUsed();
    ulong numOfCurrentlyPassive() const 
    { 
      return 
	_passive.CurrentlyKept() + 
	_pagedPassive.numberOfCurrentlyKeptClauses() + 
	_reservedPassive.Size();
    };
    ulong numOfReachablePassive() const 
    {
      return _parameters.reachable;
    };
    ulong numOfCurrentlyActive() const 
    {
      return  _active.Size();
    };
    const ClauseSet& activeClauses() const { return _active; };
  public:
    Options options;
  private:
    class Parameters
    {
    public:
      Parameters() {};
      ~Parameters() {};
      bool incompleteAPriori;
      long main_alg;
      bool binary_resolution;
      bool no_inferences_between_rules;
      bool splitting;  
      bool small_basic_comp_in_hyp_splitting;
      bool fold_single_comp_in_splitting;
      long single_ground_lit_comp_min_size;
      bool reuse_names_for_negated_ground_lits_in_splitting;
      TempLitListExtras::ComparisonMaker* basic_comp_in_spl_is_max_wrt;
      bool back_fold_in_spl;
      long neg_eq_splitting;
      long neg_eq_spl_var_count_limit;
      bool paramodulation;
      bool paramodulate_from_variables;
      long sim_back_sup;
      bool forw_sup_no_dup_red;
      bool back_sup_no_dup_red;
      bool prohibit_non_basic_superposition;
      bool prohibit_non_basic_fd;
      bool prohibit_non_basic_bd;
      bool reduce_non_basic;
      bool input_non_basicness;
      bool run_time_non_basicness;
      long selection;
      bool inherited_negative_selection;
      bool external_evaluation;
      bool handle_equality;
      bool forward_subsumption;
      long simplify_by_fs;
      float main_premise_subsumption_expiry; 
      float side_premise_subsumption_expiry; 
      float main_selected_lit_demodulation_expiry; 
      float side_selected_lit_demodulation_expiry; 
      long forward_demodulation;
      bool fd_on_splitting_branches;
      const SymbolInfo* normalize_by_commutativity;
      const SymbolInfo* eq_taut_mod_C;
      const SymbolInfo* prop_taut_mod_C;
      const SymbolInfo* trivial_eq_res_mod_C;
      const SymbolInfo* identical_lit_factoring_mod_C;
      const SymbolInfo* chaining_for_transitive;
      bool chaining_through_variables;
      long backward_demodulation;
      bool bd_on_splitting_branches;
      bool simplify_by_eq_res;
      bool backward_subsumption;
      bool orphan_murder;
      long lrs; 
      bool orient_pos_equalities_in_support;
      bool orient_neg_equalities_in_support;
      const bool* answer_predicate_table; 
      ulong answer_predicate_table_size;
      long age_weight_ratio_numerator;
      long age_weight_ratio_denominator;
      long best_good_weight_ratio_numerator;
      long best_good_weight_ratio_denominator;
      long timeLimit; 
      long simulatedTimeLimit; 
      float memoryLimit;
      float allocation_buffer_price; 
      long maxActive;
      long maxPassive; 
      long maxWeight;
      long previousMaxWeight;
      long maxMaxWeight;
      long maxInferenceDepth;
      float first_time_check;
      float first_mem_check; 
      long reachable;
      long unreachable;
      ostream* outputStreamForGenerated;
      void (*showPassive) (const Clause*);     
      void (*showActive) (const Clause*);
      bool showLitProp;
      ostream* outputStreamForPassiveProfiles;
      long showPassiveProfileInterval;
      void (*showWeightChange) (long,long);
      void (*showUsagePace) (float); 
      bool useClauseNumIndex;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
      bool step_resolution_literal_selection;
#endif
    }; 
    enum Continuation
    {
      Continuation_Init = 0UL,
      Continuation_InputSupportClause0_ProcessCollectedInput = 1UL,
      Continuation_InputQueryClause0_ProcessCollectedInput = 2UL,
      Continuation_InputQueryClause0_BackSimplify = 3UL,
      Continuation_InputQueryClause0_BackFold = 4UL,
      Continuation_MainLoop0_ProcessNewInference = 5UL,
      Continuation_MainLoop0_BackSimplify = 6UL,
      Continuation_MainLoop0_BackFold = 7UL,
      Continuation_InputSupportClause3_ProcessCollectedInput = 8UL,
      Continuation_InputQueryClause3_ProcessCollectedInput = 9UL,
      Continuation_InputQueryClause3_BackFold = 10UL,
      Continuation_MainLoop3_SelectCurrent = 11UL,
      Continuation_MainLoop3_BackSimplify = 12UL, 
      Continuation_MainLoop3_BackFold = 13UL, 
      Continuation_MainLoop3_ProcessNewInference = 14UL
    };
  private:
    void linkObjects();
    void initParameters();            
    void passParametersToObjects();
    void inputClause0(Clause* cl);
    void inputSupportClause0(Clause* cl);
    void inputQueryClause0(Clause* cl);
    void inputClause3(Clause* cl);
    void inputQueryClause3(Clause* cl);
    void inputSupportClause3(Clause* cl);
    void mainLoop0();
    void mainLoop3();
    Clause* reanimate3(PagedPassiveClauseStorage::StoredClause& cl); 
    void processCollected0();
    void processCollected3();  
    Clause* select0();
    void makeActive0(Clause* cl);
    void makeActive3(Clause* cl);
    void makePassive0(Clause* cl);
    void rememberAnswer0(Clause* cl);
    void integratePassive(Clause* cl);
    void integrateIntoBackDemIndexFromQueue();
    void integrateIntoBSIndexFromQueue();
    bool eligibleAsBackwardSimplifier(const Clause* cl)
    {
      return _backDemodulation.eligibleAsQuery(cl);
    };
    bool canBeEnqueuedForDisposal(Clause* cl);
    void enqueueForDisposal(Clause* cl);
    void disposeObsolete();
    void disposeClause(Clause* cl);
    void eliminateAllocationBufferDeficit0();
    void eliminateAllocationBufferDeficit3();
    bool releaseMemory3(ulong size); 
    bool releaseALotOfMemory0(); 
    bool releaseALotOfMemory3(); 
    Clause* goodActiveCandidateForDisposal(); 
    Clause* goodPassiveCandidateForDisposal0(); 
    void recycleUnreachablePassive0();    
    void killActiveOrphans0();
    void killActiveOrphans3();
    void removeFromIndices(Clause* cl);
    void destroyPassive0();
    void loadReservedClauses();
    bool backSubsumeBy(Clause* cl);
    void backSimplify();
    void backFold();
    void backSimplifyBy3(Clause* simplifier);
    void backFold3();
    float usedMemory();
    void estimateResources()
    {
      switch (_parameters.lrs)
	{
	case 0L: 
	  if (_parameters.showUsagePace)
	    {
	      float used_time = _timer.elapsedSeconds();
	      if (used_time < 1.0) used_time = 1.0;      
	      long moved_from_passive = _passive.UsedAndRemoved();    
	      _parameters.showUsagePace(moved_from_passive/used_time);
	    };
	  return;
	case 1L: estimateResources1(); break;
	case 2L: estimateResources2(); break;
	case 3L: estimateResources3(); break;
	case 4L: estimateResources4(); break;
	case 5L: estimateResources5(); break;
	default: estimateResources1(); 
	};
    };
    void estimateResources1(); 
    void estimateResources2();
    void estimateResources3();
    void estimateResources4();
    void estimateResources5();
    void updateWeightLimit(long newLimit);
    void internalTick0();
    void internalTick3();
    void integrateIntoForwardSubsumptionIndex(Clause* cl) 
    { 
      CALL("integrateIntoForwardSubsumptionIndex(Clause* cl)");
      ASSERT(!cl->flag(Clause::FlagInForwardSubsumptionIndex));
      if (!cl->isNonSubsuming())
	{
	  _forwardSubsumption.integrate(cl);
	  cl->setFlag(Clause::FlagInForwardSubsumptionIndex);
	};
    };
    void removeFromForwardSubsumptionIndex(Clause* cl) 
    { 
      CALL("removeFromForwardSubsumptionIndex(Clause* cl)");
      ASSERT(cl->flag(Clause::FlagInForwardSubsumptionIndex));  
      _forwardSubsumption.remove(cl);
      cl->clearFlag(Clause::FlagInForwardSubsumptionIndex);
    };
    void integrateIntoBackwardSubsumptionIndex(Clause* cl)
    {
      CALL("integrateIntoBackwardSubsumptionIndex(Clause* cl)");
      ASSERT(!cl->flag(Clause::FlagInBackwardSubsumptionIndex));
      if (!cl->isNonSubsumable())
	{
	  _BSIndex.integrate(cl);
	  cl->setFlag(Clause::FlagInBackwardSubsumptionIndex);
	};
    };
    void removeFromBackwardSubsumptionIndex(Clause* cl) 
    { 
      CALL("removeFromBackwardSubsumptionIndex(Clause* cl)");
      ASSERT(cl->flag(Clause::FlagInBackwardSubsumptionIndex));  
      _BSIndex.remove(cl);
      cl->clearFlag(Clause::FlagInBackwardSubsumptionIndex);
    };
    void integrateIntoForwardDemodulationIndex(Clause* cl) 
    { 
      CALL("integrateIntoForwardDemodulationIndex(Clause* cl)");
      ASSERT(!cl->flag(Clause::FlagInForwardDemodulationIndex));
      _forwardDemodulation.Integrate(cl);
      cl->setFlag(Clause::FlagInForwardDemodulationIndex);
    };
    void removeFromForwardDemodulationIndex(Clause* cl) 
    { 
      CALL("removeFromForwardDemodulationIndex(Clause* cl)");
      ASSERT(cl->flag(Clause::FlagInForwardDemodulationIndex));  
      _forwardDemodulation.Remove(cl);
      cl->clearFlag(Clause::FlagInForwardDemodulationIndex);
    };
    void integrateIntoBackDemodulationIndex(Clause* cl)
    {
      CALL("integrateIntoBackDemodulationIndex(Clause* cl)");
      ASSERT(!cl->flag(Clause::FlagInBackDemodulationIndex));
      _backDemodulation.integrate(cl);
      cl->setFlag(Clause::FlagInBackDemodulationIndex);
    };    void removeFromBackDemodulationIndex(Clause* cl) 
    { 
      CALL("removeFromBackDemodulationIndex(Clause* cl)");
      ASSERT(cl->flag(Clause::FlagInBackDemodulationIndex));  
      _backDemodulation.remove(cl);
      cl->clearFlag(Clause::FlagInBackDemodulationIndex);
    };
  private:
    bool _memoryReleaseInProgress;
    BK::GlobalClock::Timer _timer;
    Continuation _continuation;
    bool _terminated;
    TerminationReason _terminationReason; 
    Clause* _proof;
    bool _timerIsOn;
    bool _externalTickFlag;
    Parameters _parameters;
    Statistics _statistics;
    Unifier _unifier;
    PassiveClauseSet _passive;
    PagedPassiveClauseStorage _pagedPassive;
    ClauseSet _reservedPassive;
    ClauseSet _active;
    ClauseSet _history;
    ClauseSet _answers;
    ClauseSet _trash;
    BK::FIFO<BK::GlobAlloc,Clause*> _queueForBackSimplification;
    BK::FIFO<BK::GlobAlloc,Clause*> _queueForIntegrationIntoBackwardDemodIndex;
    BK::FIFO<BK::GlobAlloc,Clause*> _queueForIntegrationIntoBSIndex;
    InferenceDispatcher _infDispatcher;
    ForwardSubsumptionFrontEnd _forwardSubsumption;
    BSIndexFrontEnd _BSIndex;
    BackwardSubsumption _backSubsumption;
    BackwardFolding _backFolding;
    ForwardDemodulation _forwardDemodulation;
    BackDemodulation _backDemodulation;
    TermSharing _termSharing;
    LiteralSharing _literalSharing;
    LiteralSelection _litSelection;
    StandardLiteralOrdering _standardLiteralOrdering;
    ClauseAssembler _clauseAssembler; 
    TemporaryClauseQueue _newClauseQueue;
    TemporaryClauseCollector _inputClauseCollector;
    TemporaryClauseCollector _reanimatedClauseCollector;
    TemporaryClauseCollector _mainClauseCollector;
    BK::ObjectPool<TemporaryClause>::EntryHandle _newInputClause;
    BK::ObjectPool<TemporaryClause>::EntryHandle _retainedClause;
    BK::ObjectPool<TemporaryClause>::EntryHandle _reanimatedClause;
    PropositionalSplitting _propositionalSplitting;
    NegativeEqualitySplitting _negativeEqualitySplitting;
    ForwardProcessing _forwardProcessing;
    ForwardProcessing _forwardProcessingForSupport;
    float _startMemory; 
    ClauseNumIndex _clauseNumIndex; 
    Clause::NewNumber* _clauseNumberGenerator; 
    Clause* _sharedReanimatedClause;
    Clause* _currentSimplifier;
  }; 
  inline float Master::usedMemory()
  { 
    float res = ((float)BK::GlobAlloc::occupiedByObjects()) - _startMemory;
    if (res <= 0) return 0.0001f;
    return res;
  }; 
  inline void Master::internalTick0()
  {
    CALL("internalTick0()");
    BK::GlobalStopFlag::check();
    if (_externalTickFlag)
      {
	estimateResources();
	if (_parameters.main_premise_subsumption_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setMainPremiseSubsumption(false);
	if (_parameters.side_premise_subsumption_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setSidePremiseSubsumption(false);
	if (_parameters.main_selected_lit_demodulation_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setMainSelectedLitDemodulation(false);
	if (_parameters.side_selected_lit_demodulation_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setSideSelectedLitDemodulation(false);
	_externalTickFlag = false;      
      };
  }; 
  inline void Master::internalTick3()
  {
    CALL("internalTick3()");
    BK::GlobalStopFlag::check();
    if (_externalTickFlag)
      {	
	if (_parameters.main_premise_subsumption_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setMainPremiseSubsumption(false);
	if (_parameters.side_premise_subsumption_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setSidePremiseSubsumption(false);
	if (_parameters.main_selected_lit_demodulation_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setMainSelectedLitDemodulation(false);
	if (_parameters.side_selected_lit_demodulation_expiry <= _timer.elapsedDeciseconds())
	  _infDispatcher.setSideSelectedLitDemodulation(false);
	_externalTickFlag = false;      
      };
  }; 
}; 
#endif
