/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PATH_INDEX_WITH_OPEN_PATHS_HPP
#define PATH_INDEX_WITH_OPEN_PATHS_HPP 
#include "jargon.hpp"
#include "GlobAlloc.hpp"
#include "VampireKernelDebugFlags.hpp" 
#include "VampireKernelConst.hpp" 
#include "Stack.hpp"
#include "Term.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEX_WITH_OPEN_PATHS
#  define DEBUG_NAMESPACE "PathIndexWithOpenPaths<IndObj,ConstIndObj>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class IndObj,class ConstIndObj>
  class PathIndexWithOpenPathsDestruction;
  template <class IndObj,class ConstIndObj>
  class PathIndexWithOpenPaths
  {
  public:
    class Integrator;
    class Retrieval;
    class Removal;
  private:
    class OccNode;
    class FuncNode
    {
    public:
      FuncNode(const Term& f) : _func(f), _altern(0) 
      {
	CALL("FuncNode::(constructor FuncNode(const Term& f))");
	ASSERT(_func.isComplex());
	ulong arity = f.arity();
	ASSERT(arity);
	OccNode** deeperPos = deeper();
	for (ulong argn = 0; argn < arity; argn++) deeperPos[argn] = 0;  
      }; 
      FuncNode(const Term& c,bool) 
	: _func(c), _altern(0) 
      {
	CALL("FuncNode::constructor FuncNode(const Term& c)");
	ASSERT(_func.isComplex());
	ASSERT(!(_func.arity()));
	BK::init(indObj());
      };
      ~FuncNode() {};
      void* operator new(size_t,ulong arity) 
      { 
	CALL("FuncNode::operator new(size_t,ulong arity)");
	ASSERT(arity);
	return BK::GlobAlloc::allocate(sizeof(FuncNode) + arity*sizeof(OccNode*)); 
      }; 
      void* operator new(size_t) 
      { 
	CALL("FuncNode::operator new(size_t)");
	return BK::GlobAlloc::allocate(sizeof(FuncNode) + sizeof(ConstIndObj));  
      }; 
      void operator delete(void* obj)
      {
	CALL("FuncNode::operator delete(void* obj)");
	ulong arity = (static_cast<FuncNode*>(obj))->func().arity();
	if (arity)
	  {
	    BK::GlobAlloc::deallocate(obj,sizeof(FuncNode) + arity*sizeof(OccNode*));
	  }
	else
	  {
	    BK::GlobAlloc::deallocate(obj,sizeof(FuncNode) + sizeof(ConstIndObj));
	  };
      }; 
      void operator delete(void* obj,ulong arity)
      {
	CALL("FuncNode::operator delete(void* obj,ulong arity)");
	if (arity)
	  {
	    BK::GlobAlloc::deallocate(obj,sizeof(FuncNode) + arity*sizeof(OccNode*));
	  }
	else
	  {
	    BK::GlobAlloc::deallocate(obj,sizeof(FuncNode) + sizeof(ConstIndObj));
	  };
      }; 
#ifndef NO_DEBUG
      static ulong funcMinSize() { return sizeof(FuncNode) + sizeof(OccNode*); };
      static ulong funcMaxSize() { return sizeof(FuncNode) + Term::maxArity()*sizeof(OccNode*); };
      static ulong constMinSize() { return sizeof(FuncNode) + sizeof(ConstIndObj); };
      static ulong constMaxSize() { return sizeof(FuncNode) + sizeof(ConstIndObj); };
#endif    
      const Term& func() const { return _func; }; 
      FuncNode*& altern() { return _altern; };
      const FuncNode* const & altern() const { return _altern; };
      ConstIndObj& indObj() 
      {
	CALL("FuncNode::indObj()");
	ASSERT(isConstant());
	return *(static_cast<ConstIndObj*>(static_cast<void*>(this + 1)));
      };
      const ConstIndObj& indObj() const 
      {
	CALL("FuncNode::indObj() const");
	ASSERT(isConstant());
	return *(static_cast<const ConstIndObj*>(static_cast<const void*>(this + 1)));
      };
      ConstIndObj* indObjAddr() 
      {
	CALL("FuncNode::indObjAddr()");
	ASSERT(isConstant());
	return static_cast<ConstIndObj*>(static_cast<void*>(this + 1));
      };
      const ConstIndObj* indObjAddr() const 
      {
	CALL("FuncNode::indObjAddr() const");
	ASSERT(isConstant());
	return static_cast<const ConstIndObj*>(static_cast<const void*>(this + 1));
      };
      OccNode** deeper() 
      {
	CALL("FuncNode::deeper()");
	ASSERT(isNonconst());
	return static_cast<OccNode**>(static_cast<void*>(this + 1));
      };   
      const OccNode* const * deeper() const 
      {
	CALL("FuncNode::deeper() const");
	ASSERT(isNonconst());
	return static_cast<const OccNode* const *>(static_cast<const void*>(this + 1));
      };
      ulong arity() const { return func().arity(); };
      bool isConstant() const { return !(func().arity()); };
      bool isNonconst() const { return func().arity() != 0UL; };
      bool obsolete() const 
      {
	if (isConstant()) return !(indObj());
	for (ulong a = 0; a < arity(); a++) if (deeper()[a]) return false; 
	return true;  
      };   
    private:
      Term _func;
      FuncNode* _altern;
    }; 
    class OccNode
    { 
    public:
      OccNode() : _complex(0) 
      {
	CALL("OccNode::constructor OccNode()");
	BK::init(indObj());
      };
      ~OccNode() {};
      void* operator new(size_t)
      {
	CALL("operator new(size_t)");
	return BK::GlobAlloc::allocate(sizeof(OccNode));
      }; 
      void operator delete(void* obj)
      { 
	CALL("operator delete(void* obj)");
	BK::GlobAlloc::deallocate(obj,sizeof(OccNode)); 
      }; 
      IndObj& indObj() { return *(static_cast<IndObj*>(static_cast<void*>(_indObj))); }; 
      const IndObj& indObj() const { return *(static_cast<const IndObj*>(static_cast<const void*>(_indObj))); };
      IndObj* indObjAddr() { return static_cast<IndObj*>(static_cast<void*>(_indObj)); };  
      const IndObj* indObjAddr() const { return static_cast<const IndObj*>(static_cast<const void*>(_indObj)); };
      FuncNode*& complex() { return _complex; }; 
      const FuncNode* const & complex() const { return _complex; }; 
    private:
      char _indObj[sizeof(IndObj)]; 
      FuncNode* _complex;
    }; 
  public:
    class Trie
    {
    public:
      Trie() : _content(0) {};
      ~Trie() {};
      operator bool() const { return (bool)_content; };
    private:
      OccNode** contentAddr() { return &_content; };   
    private:
      OccNode* _content;
      friend class Integrator;
      friend class Retrieval;
      friend class Removal;
      friend class PathIndexWithOpenPathsDestruction<IndObj,ConstIndObj>;
    }; 
    class Integrator
    {
    public:
      Integrator()
      {
      };
      ~Integrator() {};
      void reset(Trie& tree)
      { 
	_occAddr = tree.contentAddr();
	createOccNodeIfNeeded();
	_backtrackFuncNodes.reset();
	_backtrackArgNums.reset();
      }; 
      void nonconstFunc(const Term& f)
      {
	CALL("nonconstFunc(const Term& f)");
	ASSERT(f.isComplex());
	ulong arity = f.arity();
	ASSERT(arity);
	ASSERT(*_occAddr);
	FuncNode** nodeAddr = &((*_occAddr)->complex());
	FuncNode* node = *nodeAddr;
	while (node)
	  {
	    if (node->func() == f) 
	      { 
		_occAddr = node->deeper(); 
		createOccNodeIfNeeded();      
		if (arity > 1)
		  { 
		    _backtrackFuncNodes.push(node);
		    _backtrackArgNums.push(1);
		  };
		return;
	      };  
	    nodeAddr = &(node->altern());
	    node = *nodeAddr;
	  };
	node = new (arity) FuncNode(f);
	*nodeAddr = node;
	_occAddr = node->deeper();
	*_occAddr = new OccNode();        
	if (arity > 1)
	  { 
	    _backtrackFuncNodes.push(node);
	    _backtrackArgNums.push(1);
	  };
      }; 
      void constant(const Term& c)
      {
	CALL("constant(const Term& c)");   
	ASSERT(c.isComplex());
	ASSERT(!c.arity());
	ASSERT(*_occAddr);
	FuncNode** nodeAddr = &((*_occAddr)->complex());
	_lastConstNode = *nodeAddr;
	while (_lastConstNode)
	  {
	    if (_lastConstNode->func() == c) 
	      { 
		backtrack();
		return;
	      };  
	    nodeAddr = &(_lastConstNode->altern());
	    _lastConstNode = *nodeAddr;
	  };
	_lastConstNode = new FuncNode(c,true);
	*nodeAddr = _lastConstNode;
	backtrack();
      };  
      void variable()
      {
	CALL("variable()");
	ASSERT(*_occAddr);
	backtrack();  
      }; 
      IndObj& indexedObj() const { return (*_occAddr)->indObj(); };
      ConstIndObj& constIndObj() const { return _lastConstNode->indObj(); };  
    private:
      void createOccNodeIfNeeded() const
      {
	if (!(*_occAddr)) 
	  {
	    *_occAddr = new OccNode();
	  };   	
      };
      void backtrack()
      {
	CALL("backtrack()");
	if (_backtrackFuncNodes.nonempty()) 
	  { 
	    FuncNode* node = _backtrackFuncNodes.pop();    
	    ulong argn = _backtrackArgNums.pop();
	    _occAddr = node->deeper() + argn;
	    createOccNodeIfNeeded();
	    argn++;
	    ASSERT(argn <= node->arity());
	    if (argn < node->arity())
	      {
		_backtrackFuncNodes.push(node);    
		_backtrackArgNums.push(argn);
	      };  
	  };
      }; 
    private:
      OccNode** _occAddr;
      FuncNode* _lastConstNode;
      BK::Stack<FuncNode*,VampireKernelConst::MaxTermDepth> _backtrackFuncNodes;
      BK::Stack<ulong,VampireKernelConst::MaxTermDepth> _backtrackArgNums;
    }; 
    class Retrieval
    {
    public:
      Retrieval()  {};
      ~Retrieval() {};
      bool reset(Trie& tree)
      { 
	_occAddr = tree.contentAddr();
	_backtrackFuncNodes.reset();
	_backtrackArgNums.reset();
	DOP(debugEndOfTerm = false);
	return (*_occAddr) != 0;
      }; 
      bool nonconstFunc(const Term& f)
      {
	CALL("nonconstFunc(const Term& f)");
	ASSERT(!debugEndOfTerm);
	if (!(*_occAddr)) return false;
	ulong arity = f.arity();
	ASSERT(arity);
	FuncNode* node = (*_occAddr)->complex();
	while (node)
	  {
	    if (node->func() == f) 
	      { 
		_occAddr = node->deeper();     
		if (arity > 1)
		  { 
		    _backtrackFuncNodes.push(node);
		    _backtrackArgNums.push(1);
		  };
		return true;
	      };  
	    node = node->altern();
	  };
	return false;
      }; 
      bool constant(const Term& c)
      {
	CALL("constant(const Term& c)"); 
	ASSERT(!debugEndOfTerm);   
	ASSERT(!c.arity());
	if (!(*_occAddr)) return false;
	_lastConstNode = (*_occAddr)->complex();
	while (_lastConstNode)
	  {
	    if (_lastConstNode->func() == c) 
	      { 
		backtrack();
		return true;
	      }  
	    else _lastConstNode = _lastConstNode->altern();
	  };
	return false;
      }; 
      void after() { CALL("Retrieval::after()"); ASSERT(!debugEndOfTerm); backtrack(); };
      IndObj& indexedObj() const { return (*_occAddr)->indObj(); };
      ConstIndObj& constIndObj() const { return _lastConstNode->indObj(); };  
      IndObj* indexedObjAddr() const { return (*_occAddr)->indObjAddr(); };
      ConstIndObj* constIndObjAddr() const { return _lastConstNode->indObjAddr(); };  
    private:
      void backtrack()
      {
	CALL("backtrack()");
	if (_backtrackFuncNodes.nonempty()) 
	  { 
	    FuncNode* node = _backtrackFuncNodes.pop();    
	    ulong argn = _backtrackArgNums.pop();
	    _occAddr = node->deeper() + argn;
	    argn++;
	    ASSERT(argn <= node->arity());
	    if (argn < node->arity())
	      {
		_backtrackFuncNodes.push(node);    
		_backtrackArgNums.push(argn);
	      };  
	  }
	else 
	  {
	    DOP(debugEndOfTerm = true);
	  };
      }; 
    private:
      OccNode** _occAddr;
      FuncNode* _lastConstNode;
      BK::Stack<FuncNode*,VampireKernelConst::MaxTermDepth> _backtrackFuncNodes;
      BK::Stack<ulong,VampireKernelConst::MaxTermDepth> _backtrackArgNums;
#ifdef DEBUG_NAMESPACE
      bool debugEndOfTerm;  
#endif
    }; 
    class Removal
    {
    public:
      Removal()
	: _backtrackFuncNodes(),
	  _backtrackArgNums(),
	  _candidatesForRecycling()
      {
      };
      ~Removal() {};
      void reset(Trie& tree)
      { 
	_occAddr = tree.contentAddr();
	_backtrackFuncNodes.reset();
	_backtrackArgNums.reset();
	_candidatesForRecycling.reset();
	DOP(debugEndOfTerm = false);
      }; 
      bool nonconstFunc(const Term& f)
      {
	CALL("nonconstFunc(const Term& f)");
	ASSERT(!debugEndOfTerm);
	if (!(*_occAddr)) return false;
	ulong arity = f.arity();
	ASSERT(arity);
	_candForRecycling = &((*_occAddr)->complex());
	FuncNode* node = (*_occAddr)->complex();
	while (node)
	  {
	    if (node->func() == f) 
	      { 
		_occAddr = node->deeper();     
		if (arity > 1)
		  { 
		    _backtrackFuncNodes.push(node);
		    _backtrackArgNums.push(1);
		  };
		ASSERT(*_candForRecycling);
		_candidatesForRecycling.push(_candForRecycling);
		return true;
	      };  
	    _candForRecycling = &(node->altern());
	    node = node->altern();
	  };
	return false;
      }; 
      bool constant(const Term& c)
      {
	CALL("constant(const Term& c)"); 
	ASSERT(!debugEndOfTerm);   
	ASSERT(!c.arity());
	if (!(*_occAddr)) return false;
	_candForRecycling = &((*_occAddr)->complex());
	_lastConstNode = (*_occAddr)->complex();
	while (_lastConstNode)
	  {
	    if (_lastConstNode->func() == c) 
	      { 
		backtrack();
		ASSERT(*_candForRecycling);
		_candidatesForRecycling.push(_candForRecycling);
		return true;
	      }  
	    else 
	      {
		_candForRecycling = &(_lastConstNode->altern());
		_lastConstNode = _lastConstNode->altern();
	      };
	  };
	return false;
      }; 
      bool occurences() { return (*_occAddr) != 0; }; 
      void after() { CALL("Removal::after()"); ASSERT(!debugEndOfTerm); backtrack(); }; 
      IndObj& indexedObj() const 
      { 
	CALL("Removal::indexedObj() const"); 
	ASSERT(*_occAddr);  
	return (*_occAddr)->indObj(); 
      };
      ConstIndObj& constIndObj() const 
      { 
	CALL("Removal::constIndObj() const"); 
	ASSERT(_lastConstNode); 
	return _lastConstNode->indObj(); 
      };  
      void recycleObsoleteNodes()
      {
	CALL("Removal::recycleObsoleteNodes()"); 
	while (_candidatesForRecycling.nonempty()) recycleObsolete(*(_candidatesForRecycling.pop()));
      };
    private:
      void backtrack()
      {
	CALL("Removal::backtrack()");
	if (_backtrackFuncNodes.nonempty()) 
	  { 
	    FuncNode* node = _backtrackFuncNodes.pop();    
	    ulong argn = _backtrackArgNums.pop();
	    _occAddr = node->deeper() + argn;
	    argn++;
	    ASSERT(argn <= node->arity());
	    if (argn < node->arity())
	      {
		_backtrackFuncNodes.push(node);    
		_backtrackArgNums.push(argn);
	      };  
	  }
	else 
	  {
	    DOP(debugEndOfTerm = true);
	  };
      }; 
    private:
      OccNode** _occAddr;
      FuncNode* _lastConstNode;
      FuncNode** _candForRecycling;
      BK::Stack<FuncNode*,VampireKernelConst::MaxTermDepth> _backtrackFuncNodes;
      BK::Stack<ulong,VampireKernelConst::MaxTermDepth> _backtrackArgNums;
      BK::Stack<FuncNode**,VampireKernelConst::MaxTermSize> _candidatesForRecycling;
#ifdef DEBUG_NAMESPACE
      bool debugEndOfTerm;  
#endif
    }; 
  private: 
    static void destroyTree(FuncNode*& tree);
    static void destroyTree(OccNode*& tree);
    static void recycleObsolete(FuncNode*& tree);
    static void recycleObsolete(OccNode*& tree);
    friend class Removal;
    friend class PathIndexWithOpenPathsDestruction<IndObj,ConstIndObj>;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEX_WITH_OPEN_PATHS
#define DEBUG_NAMESPACE "PathIndexWithOpenPathsDestruction<IndObj,ConstIndObj>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class IndObj,class ConstIndObj>
  class PathIndexWithOpenPathsDestruction
  {
  public:
    typedef 
    typename PathIndexWithOpenPaths<IndObj,ConstIndObj>::Trie
    Trie;
  public:
    PathIndexWithOpenPathsDestruction() {};
    ~PathIndexWithOpenPathsDestruction() {};
    bool reset(Trie& tree,bool& constant);
    void recycle() 
    { 
      PathIndexWithOpenPaths<IndObj,ConstIndObj>::destroyTree(*_tree); 
    };
    bool next(bool& constant);
    IndObj& indexedObj() const 
    { 
      CALL("Removal::indexedObj() const"); 
      ASSERT(*_occAddr);  
      return (*_occAddr)->indObj(); 
    };
    ConstIndObj& constIndObj() const 
    { 
      CALL("Removal::constIndObj() const"); 
      ASSERT(_lastConstNode); 
      return _lastConstNode->indObj(); 
    }; 
  private:
    typedef
    typename PathIndexWithOpenPaths<IndObj,ConstIndObj>::OccNode 
    OccNode;
    typedef
    typename PathIndexWithOpenPaths<IndObj,ConstIndObj>::FuncNode
    FuncNode;
  private:
    OccNode** _occAddr;
    OccNode** _tree;
    FuncNode* _lastConstNode;
    BK::Stack<FuncNode*,VampireKernelConst::MaxTermDepth> _backtrackFuncNodes;
    BK::Stack<ulong,VampireKernelConst::MaxTermDepth> _backtrackArgNums;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEX_WITH_OPEN_PATHS
#define DEBUG_NAMESPACE "PathIndexWithOpenPaths<IndObj,ConstIndObj>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class IndObj,class ConstIndObj>
  inline void PathIndexWithOpenPaths<IndObj,ConstIndObj>::destroyTree(FuncNode*& tree)
  {   
    CALL("destroyTree(FuncNode*& tree)");
    if (tree)
      {
	destroyTree(tree->altern());
	if (tree->isNonconst())
	  {             
	    for (ulong a = 0; a < tree->arity(); a++) 
	      {
		if (tree->deeper()[a])
		  {
		    destroyTree(tree->deeper()[a]);
		  }; 
	      };
	  }
	else
	  {       
	  };
	delete tree;
	tree = 0;
      };
  }; 
  template <class IndObj,class ConstIndObj>
  inline void PathIndexWithOpenPaths<IndObj,ConstIndObj>::destroyTree(OccNode*& tree)
  {
    CALL("destroyTree(OccNode*& tree)");
    if (tree)   
      {
	destroyTree(tree->complex());     
	delete tree;
	tree = 0;
      };
  }; 
  template <class IndObj,class ConstIndObj>
  inline void PathIndexWithOpenPaths<IndObj,ConstIndObj>::recycleObsolete(FuncNode*& tree)
  {
    CALL("recycleObsolete(FuncNode*& tree)");
    if (tree) 
      {
	bool obsolete = true;
	if (tree->isConstant()) { obsolete = !tree->indObj(); }
	else
	  {
	    for (ulong a = 0; a < tree->arity(); a++) 
	      { 
		recycleObsolete(tree->deeper()[a]);
		if (tree->deeper()[a]) obsolete = false; 
	      };
	  };
	if (obsolete) 
	  { 
	    FuncNode* tmp = tree;
	    tree = tmp->altern();
	    delete tmp;
	  }; 
      };   
  }; 
  template <class IndObj,class ConstIndObj>
  inline void PathIndexWithOpenPaths<IndObj,ConstIndObj>::recycleObsolete(OccNode*& tree)
  {
    CALL("recycleObsolete(OccNode*& tree)");
    if (tree && (!tree->indObj()) && (!tree->complex()))
      {
	delete tree; 
	tree = 0;
      };                  
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEX_WITH_OPEN_PATHS
#define DEBUG_NAMESPACE "PathIndexWithOpenPathsDestruction<IndObj,ConstIndObj>"
#endif
#include "debugMacros.hpp"
namespace VK
{
  template <class IndObj,class ConstIndObj>
  inline 
  bool 
  PathIndexWithOpenPathsDestruction<IndObj,ConstIndObj>::reset(Trie& tree,bool& constant) 
  { 
    CALL("reset(Trie& tree,bool& constant)");
    _occAddr = tree.contentAddr();
    _tree = _occAddr;
    constant = false;  
    _lastConstNode = 0;
    return (*_occAddr) != 0;
  }; 
  template <class IndObj,class ConstIndObj>
  inline 
  bool 
  PathIndexWithOpenPathsDestruction<IndObj,ConstIndObj>::next(bool& constant)
  {
    CALL("next(bool& constant)");  
    FuncNode* nextFunc;
    if (_lastConstNode)  
      {
	ASSERT(_lastConstNode->isConstant());
	nextFunc = _lastConstNode->altern();
      check_func_node:
	if (nextFunc)
	  { 
	    if (nextFunc->isConstant()) 
	      {
		_lastConstNode = nextFunc;
		constant = true;
		return true;
	      }
	    else
	      {
		for (ulong a = 0; a < nextFunc->arity(); a++)
		  if (nextFunc->deeper()[a])
		    {
		      if (a + 1 < nextFunc->arity())  
			{ 
			  _backtrackFuncNodes.push(nextFunc);
			  _backtrackArgNums.push(a + 1);                
			}
		      else
			if (nextFunc->altern()) 
			  {
			    _backtrackFuncNodes.push(nextFunc->altern());
			    _backtrackArgNums.push(0);
			  };
		      _occAddr = nextFunc->deeper() + a;
		      _lastConstNode = 0;
		      constant = false;
		      return true; 
		    };
		nextFunc = nextFunc->altern();
		goto check_func_node;   
	      };
	  }      
	else 
	  {
	    if (_backtrackFuncNodes.empty()) return false;     
	    nextFunc = _backtrackFuncNodes.pop();  
	    if (nextFunc->isConstant())
	      {
		_backtrackArgNums.pop();
		_lastConstNode = nextFunc;
		constant = true;
		return true;
	      };
	    for (ulong a = _backtrackArgNums.pop(); a < nextFunc->arity(); a++) 
	      if (nextFunc->deeper()[a])
		{
		  if (a + 1 < nextFunc->arity())  
		    { 
		      _backtrackFuncNodes.push(nextFunc);
		      _backtrackArgNums.push(a + 1);                
		    }
		  else
		    if (nextFunc->altern()) 
		      {
			_backtrackFuncNodes.push(nextFunc->altern());
			_backtrackArgNums.push(0);
		      };
		  _occAddr = nextFunc->deeper() + a;
		  _lastConstNode = 0;
		  constant = false;
		  return true;
		};
	    nextFunc = nextFunc->altern();
	    goto check_func_node;
	  };       
      }
    else 
      {
	nextFunc = (*_occAddr)->complex();
	goto check_func_node;
      };
  }; 
}; 
#endif
