/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "PathIndexingForInstanceRetrieval.hpp"
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEXING_FOR_INSTANCE_RETRIEVAL
#  define DEBUG_NAMESPACE "PathIndexingForInstanceRetrieval"
#endif
#include "debugMacros.hpp"
PathIndexingForInstanceRetrieval::PathIndexingForInstanceRetrieval() : 
  _intermediateQueryRepresentation(&_abstractMachine)
{
};
PathIndexingForInstanceRetrieval::~PathIndexingForInstanceRetrieval() 
{
};
bool PathIndexingForInstanceRetrieval::integrate(const Term& integratedTermId)
{
  CALL("integrate(const Term& integratedTermId)");
  _integratedTermTraversal.reset(integratedTermId);
  _integrator.reset(_trie);
  bool newNode;
  ulong arity;
 check_state:
  switch (_integratedTermTraversal.state())
    { 
    case Term::Traversal::Func:
      arity = _integratedTermTraversal.symbol().arity();
      if (arity)
	{
	  _integrator.nonconstFunc(_integratedTermTraversal.symbol());
	  NonmaxPathRelation::Node* node = 
	    _integrator.indexedObj().insert(integratedTermId,newNode);
	  if (!newNode) return false;
	  node->value().reset(_integratedTermTraversal.term());             
	}
      else 
	{
	  _integrator.constant(_integratedTermTraversal.symbol());
	  if (!_integrator.constIndObj().insert(integratedTermId)) return false;
	};
      _integratedTermTraversal.next();
      goto check_state;
    case Term::Traversal::Var:
      _integrator.variable();
      _integratedTermTraversal.next();
      goto check_state;
    case Term::Traversal::End: 
      return true; 
#ifdef DEBUG_NAMESPACE 
    default: ICP("ICP0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default: 
      BK::RuntimeError::report("Wrong term traversal state in PathIndexingForInstanceRetrieval<..>::integrate(..)");
      return false;
#endif
#endif       
    };
} 
bool PathIndexingForInstanceRetrieval::remove(const Term& removedTermId)
{
  CALL("remove(const Term& removedTermId)");
  _removedTermTraversal.reset(removedTermId);
  _removal.reset(_trie);
  bool somethingRemoved = false;
  ulong arity;
 check_state:
  switch (_removedTermTraversal.state())
    { 
    case Term::Traversal::Func:
      arity = _removedTermTraversal.symbol().arity();
      if (arity)
	{
	  if (_removal.nonconstFunc(_removedTermTraversal.symbol()))
	    {
	      somethingRemoved = _removal.indexedObj().remove(removedTermId) || somethingRemoved;
	      _removedTermTraversal.next();
	    }
	  else
	    {
	      _removal.after();
	      _removedTermTraversal.after();
	    };
	}
      else 
	{
	  if (_removal.constant(_removedTermTraversal.symbol()))
	    {
	      somethingRemoved = _removal.constIndObj().remove(removedTermId) || somethingRemoved;
	    }
	  else
	    {
	      _removal.after();
	    };
	  _removedTermTraversal.next();
	};
      goto check_state;
    case Term::Traversal::Var:
      _removal.after();
      _removedTermTraversal.next();
      goto check_state;
    case Term::Traversal::End: 
      _removal.recycleObsoleteNodes();     
      return somethingRemoved;
#ifdef DEBUG_NAMESPACE 
    default: ICP("ICP0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default: 
      BK::RuntimeError::report("Wrong term traversal state in PathIndexingForInstanceRetrieval<..>::remove(..)");
      return false;
#endif
#endif       
    };
} 
bool PathIndexingForInstanceRetrieval::compileQuery(const Term& query)
{
  CALL("compileQuery(const Term& query)");
#ifndef NO_DEBUG
  _debugQueryVariables.reset();
#endif
  _queryTraversal.reset(query);
  _pathRelationRetrieval.reset(_trie);
  _intermediateQueryRepresentation.reset();
  ulong arity;
 check_state: 
  switch (_queryTraversal.state())
    {
    case Term::Traversal::Func:
      arity = _queryTraversal.symbol().arity();
      if (arity) 
	{
	  if (!_pathRelationRetrieval.nonconstFunc(_queryTraversal.symbol())) return false;
	  _intermediateQueryRepresentation.compound(arity,_pathRelationRetrieval.indObjAddr());
	}
      else 
	{     
	  if (!_pathRelationRetrieval.constant(_queryTraversal.symbol())) return false;
	  _intermediateQueryRepresentation.constant(_pathRelationRetrieval.constIndObjAddr());
	};
      _queryTraversal.next(); 
      goto check_state;
    case Term::Traversal::Var:
#ifndef NO_DEBUG
      _debugQueryVariables.add1(_queryTraversal.symbol().var());
#endif
      _intermediateQueryRepresentation.variable(_queryTraversal.symbol().var());
      _pathRelationRetrieval.after();
      _queryTraversal.next(); 
      goto check_state;
    case Term::Traversal::End: 
      _intermediateQueryRepresentation.endOfQuery();
      _intermediateQueryRepresentation.generateCode();
#ifdef INSTANCE_RETRIEVAL_INDEX_USE_WEIGHT_FILTERS
      _abstractMachine.queryWeight = Term::weight(query);
#endif
#ifdef INSTANCE_RETRIEVAL_INDEX_USE_SIGNATURE_FILTERS
      _abstractMachine.querySignatureFilter = Term::signatureFilter(query);
#endif
      return true;
#ifdef DEBUG_NAMESPACE 
    default: ICP("ICP0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default: 
      BK::RuntimeError::report("Wrong term traversal state in PathIndexingForInstanceRetrieval<..>::compileQuery(..)");
      return false;
#endif
#endif    
    };
} 
bool PathIndexingForInstanceRetrieval::nextMatch(Term& matchId)
{
  CALL("nextMatch(Term& matchId)");
  return _abstractMachine.nextMatch(matchId);
}
#ifndef NO_DEBUG_VIS
ostream& PathIndexingForInstanceRetrieval::outputSubst(ostream& str) const
{
  for (const ulong* v = _debugQueryVariables.begin();
       v != _debugQueryVariables.end();
       ++v)
    {
      str << "[X" << *v << " -> " << subst(*v) << "] "; 
    };
  return str;
} 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEXING_FOR_INSTANCE_RETRIEVAL
#  define DEBUG_NAMESPACE "PathIndexingForInstanceRetrieval::AbstractMachine"
#endif
#include "debugMacros.hpp"
PathIndexingForInstanceRetrieval::AbstractMachine::AbstractMachine() 
{
  DOP(debugEndOfCode = code + (2*VampireKernelConst::MaxTermSize + 1));
};
PathIndexingForInstanceRetrieval::AbstractMachine::~AbstractMachine() {};
bool PathIndexingForInstanceRetrieval::AbstractMachine::nextMatch(Term& matchId)
{
  CALL("nextMatch(Term& matchId)");
  Instruction* instr = code;
  Term termId;
  switch (instr->tag) 
    { 
    case NonmaxEqualiseTermId:
      if (firstRetrievalCall) 
	{
	  firstRetrievalCall = false;
	}
      else  
	{
	  if (!instr->nonmaxPathRelationSurfer->next()) return false;
	};
      ASSERT_IN(instr->nonmaxPathRelationSurfer->currNode(),"A100");  
      termId = instr->nonmaxPathRelationSurfer->currNode()->key();
      break;    
    case MaxEqualiseTermId:
      if (firstRetrievalCall) 
	{
	  firstRetrievalCall = false;
	}
      else  
	{
	  if (!instr->maxPathRelationSurfer->next()) return false;
	};
      ASSERT(instr->maxPathRelationSurfer->currNode());
      termId = instr->maxPathRelationSurfer->currNode()->key();
      break;
#ifdef DEBUG_NAMESPACE 
    default: ICP("ICP0"); return false; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: 
      BK::RuntimeError::report("Wrong tag in PathIndexingForInstanceRetrieval<..>::AbstractMachine::nextMatch(Term& matchId)"); 
      return false;
#endif   
#endif
    };
 next_instr:
  instr++;
 perform_instr:
  ASSERT(instr < debugEndOfCode);
  switch (instr->tag)
    {
    case NonmaxEqualiseTermId:
      if (instr->nonmaxPathRelationSurfer->find(termId)) goto next_instr; 
      if (instr->nonmaxPathRelationSurfer->currNode())
	{
	  termId = instr->nonmaxPathRelationSurfer->currNode()->key();
	  instr = code;
	  goto perform_instr;
	}
      else 
	{
	  return false; 
	};   
    case MaxEqualiseTermId:
      if (instr->maxPathRelationSurfer->find(termId)) goto next_instr; 
      if (instr->maxPathRelationSurfer->currNode())
	{
	  termId = instr->maxPathRelationSurfer->currNode()->key();
	  instr = code;
	  goto perform_instr;
	}
      else 
	{
	  return false;
	};
    case GetToTermId: 
      instr->nonmaxPathRelationSurfer->getTo(termId);
      goto next_instr;
    case LoadSubst:
      *(instr->subst) = *((*(instr->nonmaxPathCurrNode))->value().nth(instr->argNum));    
      goto next_instr;
    case Compare: 
      if ((*(instr->subst)) != (*(*(instr->nonmaxPathCurrNode))->value().nth(instr->argNum))) 
	goto backtrack;
      goto next_instr;
    case Success:
      matchId = termId;     
      return true;
    case GetToTermIdAndLoadSubst:
      instr->nonmaxPathRelationSurfer->getTo(termId); 
      *(instr->subst) = *((*(instr->nonmaxPathCurrNode))->value().nth(instr->argNum)); 
      goto next_instr;
    case GetToTermIdAndCompare:
      COP("GetToTermIdAndCompare");
      instr->nonmaxPathRelationSurfer->getTo(termId);      
      if ((*(instr->subst)) != (*(*(instr->nonmaxPathCurrNode))->value().nth(instr->argNum))) 
	goto backtrack;
      goto next_instr;
#ifdef DEBUG_NAMESPACE 
    default: ICP("ICP1"); return false; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: 
      BK::RuntimeError::report("Wrong tag in PathIndexingForInstanceRetrieval<..>::AbstractMachine::nextMatch(Term& matchId)"); 
      return false;
#endif   
#endif
    };
  ICP("ICP2");
 backtrack:
  BK::GlobalStopFlag::check();
  instr = code;
  switch (instr->tag)
    {
    case NonmaxEqualiseTermId:
      if (instr->nonmaxPathRelationSurfer->next()) 
	{
	  ASSERT_IN(instr->nonmaxPathRelationSurfer->currNode(),"A200");
	  termId = instr->nonmaxPathRelationSurfer->currNode()->key();
	  goto next_instr;
	}
      else 
	{
	  return false;
	};
    case MaxEqualiseTermId:
      if (instr->maxPathRelationSurfer->next())
	{
	  ASSERT(instr->maxPathRelationSurfer->currNode());
	  termId = instr->maxPathRelationSurfer->currNode()->key();
#ifdef INSTANCE_RETRIEVAL_INDEX_USE_FILTERS
	  while (INSTANCE_RETRIEVAL_INDEX_FILTER_OUT(termId))
	    {
	      if (!instr->maxPathRelationSurfer->next()) return false;
	      termId = instr->maxPathRelationSurfer->currNode()->key();
	    };
#endif
	  goto next_instr;
	}
      else 
	{
	  return false;
	};
#ifdef DEBUG_NAMESPACE 
    default: ICP("ICP3"); return false; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: 
      BK::RuntimeError::report("Wrong tag in PathIndexingForInstanceRetrieval<..>::AbstractMachine::nextMatch(Term& matchId)"); 
      return false;
#endif   
#endif
    };
  ICP("END");
  return false;
}; 
#ifndef NO_DEBUG_VIS
inline ostream& PathIndexingForInstanceRetrieval::AbstractMachine::outputCode(ostream& str) const
{
  CALL("outputCode(ostream& str) const");
  const Instruction* instr = code;
  while (instr->tag != Success) 
    {
      instr->output(str,this,substitution) << "\n";
      instr++;
    }; 
  instr->output(str,this,substitution) << "\n"; 
  return str;
} 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEXING_FOR_INSTANCE_RETRIEVAL
#  define DEBUG_NAMESPACE "PathIndexingForInstanceRetrieval::AbstractMachine::Instruction"
#endif
#include "debugMacros.hpp"
#ifndef NO_DEBUG_VIS
ostream& PathIndexingForInstanceRetrieval::AbstractMachine::Instruction::output(ostream& str,const AbstractMachine* machine,const Term* substitution) const
{
  CALL("output(ostream& str,const AbstractMachine* machine,const Term* substitution) const");
  str << (this - machine->code) << " ";
  switch (tag) 
    {
    case NonmaxEqualiseTermId:
      str << "NonmaxEqualiseTermId"
	  << " surf " << machine->surferNum(nonmaxPathRelationSurfer);
      break;
    case MaxEqualiseTermId:
      str << "MaxEqualiseTermId"
	  << " surf " << machine->surferNum(maxPathRelationSurfer);
      break;
    case GetToTermId:
      str << "GetToTermId"
	  << " surf " << machine->surferNum(nonmaxPathRelationSurfer);
      break;
    case LoadSubst:
      str << "LoadSubst"
	  << " surf " << machine->surferNum(nonmaxPathCurrNode)
	  << " arg " << argNum
	  << " var " << (subst - substitution);
      break;
    case Compare: 
      str << "Compare"
	  << " surf " << machine->surferNum(nonmaxPathCurrNode)
	  << " arg " << argNum
	  << " var " << (subst - substitution);
      break;
    case Success:
      str << "Success";
      break;
    case GetToTermIdAndLoadSubst:
      str << "GetToTermIdAndLoadSubst"
	  << " surf1 " << machine->surferNum(nonmaxPathRelationSurfer)
	  << " surf2 " << machine->surferNum(nonmaxPathCurrNode) 
	  << " arg " << argNum
	  << " var " << (subst - substitution);
      break;
    case GetToTermIdAndCompare:
      str << "GetToTermIdAndCompare"
	  << " surf1 " << machine->surferNum(nonmaxPathRelationSurfer)
	  << " surf2 " << machine->surferNum(nonmaxPathCurrNode) 
	  << " arg " << argNum
	  << " var " << (subst - substitution);
      break;
    default: str << "UnknownInstructionType"; 
    }; 
  return str;
} 
ulong PathIndexingForInstanceRetrieval::AbstractMachine::surferNum(const NonmaxPathRelation::Node* const * currNdAddr) const
{
  for (ulong res = 0UL; res < VampireKernelConst::MaxTermSize; res++)
    if (nonmaxPathRelationSurfers[res].currNodeAddr() == currNdAddr) return res;
  BK::RuntimeError::report("Surfer not found in  PathIndexingForInstanceRetrieval<..>::AbstractMachine::surferNum(const NonmaxPathRelation::Node* const * currNdAddr) const");
  return 0UL; 
} 
inline ulong PathIndexingForInstanceRetrieval::AbstractMachine::surferNum(const NonmaxPathRelation::Surfer* surf) const
{
  return surf - nonmaxPathRelationSurfers;
} 
inline ulong PathIndexingForInstanceRetrieval::AbstractMachine::surferNum(const MaxPathRelation::Surfer* surf) const
{
  return surf - maxPathRelationSurfers; 
} 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEXING_FOR_INSTANCE_RETRIEVAL
#  define DEBUG_NAMESPACE "PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation"
#endif
#include "debugMacros.hpp"
PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::IntermediateQueryRepresentation(AbstractMachine* abstractMachine) 
  : _abstractMachine(abstractMachine)
{
  _positions[0].tag = Position::Start;
  _positions[0].arity = 1UL;
  _positions[0].numOfVarArgs = 0UL;
}; 
PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::~IntermediateQueryRepresentation() {};
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::reset()
{
  CALL("reset()");
  _nextPosition = _positions + 1;
  _currentHoles = 1UL;
  _currentNest = _positions;
  _backtrackHoles.reset();
  _backtrackNonmaxPositions.reset();
  _primaryNestSorting.reset();
  _secondaryNests.reset();
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::compound(ulong arity,const NonmaxPathRelation* pathRelation)
{
  CALL("compound(const Term& subterm,ulong arity,const NonmaxPathRelation* pathRelation)");
  _nextPosition->tag = Position::Compound;
  _nextPosition->nonmaxPathRelation = pathRelation;
  _nextPosition->nestPosition = _currentNest;
  _nextPosition->arity = arity;
  _nextPosition->numOfVarArgs = 0UL;
  _nextPosition->nextVarArg = 0;
  _nextPosition->hasNonvarArg = false;
  _currentNest->hasNonvarArg = true;
  ASSERT(_currentHoles);
  _nextPosition->argNum = _currentNest->arity - _currentHoles; 
  _currentHoles--;
  if (_currentHoles) 
    {
      _backtrackHoles.push(_currentHoles);
      _backtrackNonmaxPositions.push(_currentNest);   
    }
  else 
    if (_currentNest->numOfVarArgs)
      {
	_secondaryNests.push(_currentNest);
      };
  _currentHoles = arity;
  _currentNest = _nextPosition;
  _nextPosition++;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::constant(const MaxPathRelation* pathRelation)
{
  CALL("constant(const MaxPathRelation* pathRelation)");
  _nextPosition->tag = Position::Constant;
  _nextPosition->maxPathRelation = pathRelation;
  _nextPosition->nestPosition = _currentNest; 
  _currentNest->hasNonvarArg = true;
  ASSERT(_currentHoles);
  _nextPosition->argNum = _currentNest->arity - _currentHoles;
  _currentHoles--;
  if (!_currentHoles)
    {
      if (_currentNest->numOfVarArgs)
	{
	  _secondaryNests.push(_currentNest);
	};
      if (_backtrackHoles)
	{
	  _currentHoles = _backtrackHoles.pop();
	  ASSERT(_currentHoles);
	  _currentNest = _backtrackNonmaxPositions.pop();
	};
    };
  _primaryNestSorting.push(PositionPtr(_nextPosition));
  _nextPosition++;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::variable(ulong var)
{
  CALL("variable(ulong var)");
  _substLoaded[var] = false;
  _availableVarPosition[var] = 0;
  _nextPosition->tag = Position::Variable;
  _nextPosition->nestPosition = _currentNest;
  _nextPosition->variable = var;
  _currentNest->numOfVarArgs++;
  _nextPosition->nextVarArg = _currentNest->nextVarArg; 
  _currentNest->nextVarArg = _nextPosition;
  ASSERT(_currentHoles);
  _nextPosition->argNum = _currentNest->arity - _currentHoles; 
  _currentHoles--;
  if (!_currentHoles)
    {
      if (_currentNest->hasNonvarArg) 
	{
	  _secondaryNests.push(_currentNest);
	}
      else 
	{
	  _primaryNestSorting.push(PositionPtr(_currentNest));
	}; 
      if (_backtrackHoles)
	{
	  _currentHoles = _backtrackHoles.pop();
	  ASSERT(_currentHoles);
	  _currentNest = _backtrackNonmaxPositions.pop();
	};
    };
  _nextPosition++;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::endOfQuery()
{
  CALL("endOfQuery()");
  ASSERT(!_currentHoles);
  ASSERT(!_backtrackHoles);
  _nextPosition->tag = Position::End;
  _numOfPrimaryPositions = _primaryNestSorting.size();
  for (ulong i = 0; i < _numOfPrimaryPositions; i++)
    _primaryPositions[_primaryNestSorting.num()[i]] = _primaryNestSorting.elem()[i].ptr;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::generateCode()
{
  CALL("generateCode()");
  _nextInstr = _abstractMachine->code; 
  generatePrimaryCode(); 
  generateSecondaryCode(); 
  generateCodeForSingletons();
  _nextInstr->tag = AbstractMachine::Success;
  _abstractMachine->firstRetrievalCall = true;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::generatePrimaryCode()
{
  CALL("generatePrimaryCode()");
  _singletonVars.reset();
  for (ulong i = 0; i < _numOfPrimaryPositions; i++)
    {
      Position* primaryNest = _primaryPositions[i];
      if (primaryNest->tag == Position::Constant)
	{  
	  writeMaxEqualiseTermId(primaryNest);    
	}
      else
	{ 
	  writeNonmaxEqualiseTermId(primaryNest);
	  Position* newVarPos;
	  for (newVarPos = primaryNest->nextVarArg; newVarPos; newVarPos = newVarPos->nextVarArg)
	    {
	      ulong newVar = newVarPos->variable;        
	      if (_substLoaded[newVar])
		{
		  writeCompare(newVarPos);
		}
	      else
		{
		  newVarPos->mainInstruction = 0;
		};
	    };
	  for (newVarPos = primaryNest->nextVarArg; newVarPos; newVarPos = newVarPos->nextVarArg)
	    {
	      if (!newVarPos->mainInstruction)
		{
		  ulong newVar = newVarPos->variable;
		  ASSERT(!_substLoaded[newVar]);
		  Position* availableVarPos = _availableVarPosition[newVar];
		  if (availableVarPos)
		    {
		      ASSERT(_singletonVars.scoreOf(newVar));
		      writeLoadSubst(availableVarPos);
		      _availableVarPosition[newVar] = 0;
		      _singletonVars.removeAll(newVar);
		      _substLoaded[newVar] = true;
		      writeCompare(newVarPos);
		      for (Position* newVarPos1 = newVarPos->nextVarArg; newVarPos1; newVarPos1 = newVarPos1->nextVarArg)
			{
			  if (newVarPos1->variable == newVar)
			    {
			      writeCompare(newVarPos1);
			    };
			};
		    }
		  else
		    {
		      _availableVarPosition[newVar] = newVarPos;
		      _singletonVars.add(newVar);
		    };
		};
	    };
	};
    };
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::generateSecondaryCode()
{
  CALL("generateSecondaryCode()");
  Position* varPos;
  _lastCompare = _nextInstr - 1;
  while (_secondaryNests.nonempty()) 
    {
      Position** endOfNests = _secondaryNests.end();
      Position** bestChoice = _secondaryNests.begin();
      ulong maxAvailComp = 0UL;
      ulong maxImmediateComp = 0UL;
      _internalVars.reset();
      for (varPos = (*bestChoice)->nextVarArg; varPos; varPos = varPos->nextVarArg)
	{
	  ulong var = varPos->variable;
	  if (_substLoaded[var]) 
	    {
	      maxImmediateComp++;
	    }
	  else
	    { 
	      varPos->mainInstruction = 0;
	      if (_availableVarPosition[var] || _internalVars.scoreOf(var)) 
		{ 
		  maxAvailComp++;
		}
	      else 
		{
		  _internalVars.add(var);
		};
	    };       
	}; 
      for (Position** nst = _secondaryNests.begin() + 1; nst != endOfNests; nst++)
	{
	  ulong immediateComp = 0UL;
	  ulong availComp = 0UL;
	  _internalVars.reset(); 
	  for (varPos = (*nst)->nextVarArg; varPos; varPos = varPos->nextVarArg)
	    {
	      ulong var = varPos->variable;
	      if (_substLoaded[var]) 
		{
		  immediateComp++;
		}
	      else
		{ 
		  varPos->mainInstruction = 0;
		  if (_availableVarPosition[var] || _internalVars.scoreOf(var)) 
		    { 
		      availComp++;
		    }
		  else          
		    _internalVars.add(var);
		};   
	    }; 
	  if ((immediateComp > maxImmediateComp) 
	      || ((immediateComp == maxImmediateComp)
		  && ((availComp > maxAvailComp)
		      || ((availComp == maxAvailComp)
			  && ((*nst)->numOfVarArgs > (*bestChoice)->numOfVarArgs)))))  
	    {
	      bestChoice = nst;
	      maxImmediateComp = immediateComp;
	      maxAvailComp = availComp;
	    };
	}; 
      Position* secondaryNest = *bestChoice;  
      *bestChoice = _secondaryNests.pop();  
      writeGetToTermId(secondaryNest);
      if (maxImmediateComp)
	{
	  for (varPos = secondaryNest->nextVarArg; varPos; varPos = varPos->nextVarArg)
	    {
	      ulong var = varPos->variable;
	      if (_substLoaded[var]) 
		{
		  if ((_nextInstr - 1)->tag == AbstractMachine::GetToTermId)
		    {
		      superimposeCompare(_nextInstr - 1,varPos);                      
		    }
		  else
		    {
		      writeCompare(varPos); 
		    };
		};
	    };
	};
      for (varPos = secondaryNest->nextVarArg; varPos; varPos = varPos->nextVarArg)
	{
	  if (!varPos->mainInstruction) 
	    {
	      ulong var = varPos->variable;	 
	      ASSERT(!_substLoaded[var]);
	      Position* availableVarPos = _availableVarPosition[var];
	      if (availableVarPos)
		{
		  ASSERT(availableVarPos->nestPosition->mainInstruction);
		  AbstractMachine::Instruction* nestInstruction = 
		    availableVarPos->nestPosition->mainInstruction; 
		  if ((nestInstruction > _lastCompare)    
		      && (nestInstruction->tag == AbstractMachine::GetToTermId))
		    {
		      superimposeLoadSubst(nestInstruction,availableVarPos);          
		      _singletonVars.removeAll(var);
		      _substLoaded[var] = true;
		      _availableVarPosition[var] = 0;
		      if ((_nextInstr - 1)->tag == AbstractMachine::GetToTermId)
			{
			  superimposeCompare(_nextInstr - 1,varPos);
			}
		      else
			{
			  writeCompare(varPos);
			};
		    }
		  else
		    {
		      if ((_nextInstr - 1)->tag == AbstractMachine::GetToTermId)
			{
			  superimposeLoadSubst(_nextInstr - 1,varPos);          
			}
		      else
			{  
			  writeLoadSubst(varPos);
			};
		      _singletonVars.removeAll(var);
		      _substLoaded[var] = true;
		      _availableVarPosition[var] = 0;
		      writeCompare(availableVarPos);
		    };
		  for (Position* varPos1 = varPos->nextVarArg; varPos1; varPos1 = varPos1->nextVarArg)
		    {
		      if (varPos1->variable == var)
			{
			  writeCompare(varPos1);
			};
		    };
		}
	      else
		{
		  _availableVarPosition[var] = varPos;
		  _singletonVars.add(var);
		};
	    };
	}; 
    }; 
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::generateCodeForSingletons()
{
  CALL("generateCodeForSingletons()");
  ulong numOfSingletonVars = _singletonVars.numOfDiffElements();
  for (ulong n = 0UL; n < numOfSingletonVars; n++) 
    {
      ulong var = _singletonVars.nth(n);
      ASSERT(!_substLoaded[var]);
      Position* availableVarPos = _availableVarPosition[var];
      ASSERT(availableVarPos);
      ASSERT(availableVarPos->nestPosition->mainInstruction);
      AbstractMachine::Instruction* nestInstruction = 
	availableVarPos->nestPosition->mainInstruction; 
      if ((nestInstruction > _lastCompare)    
	  && (nestInstruction->tag == AbstractMachine::GetToTermId))
	{
	  superimposeLoadSubst(nestInstruction,availableVarPos);
	}
      else 
	{
	  writeLoadSubst(availableVarPos);
	};
    };
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::writeNonmaxEqualiseTermId(Position* primaryNest)
{
  CALL("writeNonmaxEqualiseTermId(Position* primaryNest)");
  ASSERT(primaryNest->tag == Position::Compound);
  ASSERT(!primaryNest->hasNonvarArg);
  _nextInstr->tag = AbstractMachine::NonmaxEqualiseTermId;    
  NonmaxPathRelation::Surfer* nextNonmaxPathRelationSurfer = 
    _abstractMachine->nonmaxPathRelationSurfers + (primaryNest - _positions);
  nextNonmaxPathRelationSurfer->reset(*(primaryNest->nonmaxPathRelation));  
  _nextInstr->nonmaxPathRelationSurfer = nextNonmaxPathRelationSurfer; 
  primaryNest->mainInstruction = _nextInstr;
  _nextInstr++;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::writeMaxEqualiseTermId(Position* primaryNest)
{
  CALL("writeMaxEqualiseTermId(Position* primaryNest)");
  ASSERT(primaryNest->tag == Position::Constant);
  _nextInstr->tag = AbstractMachine::MaxEqualiseTermId;
  MaxPathRelation::Surfer* nextMaxPathRelationSurfer = 
    _abstractMachine->maxPathRelationSurfers + (primaryNest - _positions);
  nextMaxPathRelationSurfer->reset(*(primaryNest->maxPathRelation));
  _nextInstr->maxPathRelationSurfer = nextMaxPathRelationSurfer;
  _nextInstr++;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::writeGetToTermId(Position* secondaryNest)
{
  CALL("writeGetToTermId(Position* secondaryNest)");
  _nextInstr->tag = AbstractMachine::GetToTermId;    
  NonmaxPathRelation::Surfer* nextNonmaxPathRelationSurfer = 
    _abstractMachine->nonmaxPathRelationSurfers + (secondaryNest - _positions);
  nextNonmaxPathRelationSurfer->reset(*(secondaryNest->nonmaxPathRelation));  
  _nextInstr->nonmaxPathRelationSurfer = nextNonmaxPathRelationSurfer; 
  secondaryNest->mainInstruction = _nextInstr;
  _nextInstr++; 
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::writeCompare(Position* varPos)
{
  CALL("writeCompare(Position* varPos)");
  ASSERT(varPos->nestPosition->tag == Position::Compound);
  ASSERT(varPos->tag == Position::Variable); 
  _nextInstr->tag = AbstractMachine::Compare;
  varPos->mainInstruction = _nextInstr;
  _nextInstr->nonmaxPathCurrNode = 
    _abstractMachine->nonmaxPathRelationSurfers[varPos->nestPosition - _positions].currNodeAddr(); 
  _nextInstr->argNum = varPos->argNum;
  _nextInstr->subst = _abstractMachine->substitution + varPos->variable; 
  _nextInstr++;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::writeLoadSubst(Position* varPos)
{
  CALL("writeLoadSubst(Position* varPos)");
  ASSERT(varPos->nestPosition->tag == Position::Compound);
  ASSERT(varPos->tag == Position::Variable); 
  _nextInstr->tag = AbstractMachine::LoadSubst;
  varPos->mainInstruction = _nextInstr;
  _nextInstr->nonmaxPathCurrNode = 
    _abstractMachine->nonmaxPathRelationSurfers[varPos->nestPosition - _positions].currNodeAddr(); 
  _nextInstr->argNum = varPos->argNum;
  _nextInstr->subst = _abstractMachine->substitution + varPos->variable; 
  _nextInstr++;
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::superimposeCompare(AbstractMachine::Instruction* instr,
											   Position* varPos)
{
  CALL("superimposeCompare(AbstractMachine::Instruction* instr,Position* varPos)");
  ASSERT(instr->tag == AbstractMachine::GetToTermId);
  instr->tag = AbstractMachine::GetToTermIdAndCompare;
  _lastCompare = instr;
  varPos->mainInstruction = instr; 
  instr->nonmaxPathCurrNode = 
    _abstractMachine->nonmaxPathRelationSurfers[varPos->nestPosition - _positions].currNodeAddr();
  instr->argNum = varPos->argNum;
  instr->subst = _abstractMachine->substitution + varPos->variable;    
}; 
void PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::superimposeLoadSubst(AbstractMachine::Instruction* instr,
											     Position* varPos)
{
  CALL("superimposeLoadSubst(AbstractMachine::Instruction* instr,Position* varPos)");
  ASSERT(instr->tag == AbstractMachine::GetToTermId);
  instr->tag = AbstractMachine::GetToTermIdAndLoadSubst;
  _lastCompare = instr;
  varPos->mainInstruction = instr; 
  instr->nonmaxPathCurrNode = 
    _abstractMachine->nonmaxPathRelationSurfers[varPos->nestPosition - _positions].currNodeAddr();
  instr->argNum = varPos->argNum;
  instr->subst = _abstractMachine->substitution + varPos->variable;  
}; 
#ifndef NO_DEBUG_VIS
ostream& PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::output(ostream& str)
{
  for (const Position* p = _positions; p < _nextPosition; p++)
    p->output(str,_positions) << "\n";
  str << "\nPRIM: { ";
  for (ulong i = 0; i < _numOfPrimaryPositions; i++)
    {
      str << (_primaryPositions[i] - _positions) 
	  << "(" << PositionPtr(_primaryPositions[i]).key() << ") ";
    };
  str << "}";
  str << "\nSEC: { ";
  for (Position** p = _secondaryNests.begin(); p != _secondaryNests.end(); p++)
    { 
      str << ((*p) - _positions) << " ";  
    };
  str << "}";
  return str << "\n";
} 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEXING_FOR_INSTANCE_RETRIEVAL
#  define DEBUG_NAMESPACE "PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::Position"
#endif
#include "debugMacros.hpp"
#ifndef NO_DEBUG_VIS
ostream& PathIndexingForInstanceRetrieval::IntermediateQueryRepresentation::Position::output(ostream& str,const Position* base) const
{
  str << (this - base) << " ";
  switch (tag) 
    {
    case Start: str << "Start" << " arity(" << arity << ")"; break;
    case Compound: 
      str << "Compound" 
	  << " arg(" << argNum << ")"
	  << " nest(" << (nestPosition - base) << ")"
	  << " pr(" << (ulong)nonmaxPathRelation << ")"
	  << " arity(" << arity << ")"
	  << " numOfVarArgs(" << numOfVarArgs << ")";  
      if (hasNonvarArg) str << " hasNonvarArg"; 
      if (nextVarArg)
	{
	  str << " varArgs{";
	  for (Position* pos = nextVarArg; pos; pos = pos->nextVarArg)
	    {
	      str << (pos - base);
	      if (pos->nextVarArg) str << ",";
	    };
	  str << "}";
	};
      break;   
    case Constant: 
      str << "Constant" 
	  << " arg(" << argNum << ")"
	  << " nest(" << (nestPosition - base) << ")"
	  << " pr(" << (ulong)maxPathRelation << ")"; 
      break;  
    case Variable: 
      str << "Variable" 
	  << " arg(" << argNum << ")"
	  << " nest(" << (nestPosition - base) << ")"
	  << " #" << variable;   
      break;
    case End: str << "End"; break;
    default: str << "Unknown position tag"; break;
    };
  return str;
} 
#endif
