/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "PropositionalSplitting.hpp"
#include "TemporaryLiteral.inl"
#include "Clause.inl"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROPOSITIONAL_SPLITTING
#  define DEBUG_NAMESPACE "PropositionalSplitting"
#endif
#include "debugMacros.hpp"
PropositionalSplitting::PropositionalSplitting(ComponentName* componentNaming) : _smallBasicComponentInHyper(false),
  _foldSingleComponents(false),
  _componentNaming(componentNaming),
  _someNamesExist(false),
  _numLitMap(DOP("Splitting::_numLitMap")),
  _components(DOP("Splitting::_components")),
  _registeredCompNums(DOP("Splitting::_registeredCompNums")),
  _componentNameHeader(DOP("Splitting::_componentNameHeader")),
  _newCompName(DOP("Splitting::_newCompName")),
  _maximalNumOfLiteralsSoFar(0UL)
{    
};
PropositionalSplitting::~PropositionalSplitting()
{
  CALL("destructor ~PropositionalSplitting()");
}; 
void PropositionalSplitting::reset()
{
  CALL("reset()");
  _smallBasicComponentInHyper = false;
  _foldSingleComponents = false;
  for (ulong i = 0UL; i < _components.size(); ++i)
    _components[i].reset();
}; 
bool 
PropositionalSplitting::splitWithHyperParallel(TemporaryClause& cl,
					       TemporaryClauseCollector* newClauseCollector,
					       ulong& numOfComponents,
					       ulong& numOfFoldings,
					       ulong& numOfNewNames)
{
  CALL("splitWithHyperParallel(..)");
  if (!adjustSizes(cl.literals()->length())) return false;
  ASSERT(!cl.literals()->isEmpty());
  DOP(ulong debugNumOfLits = cl.literals()->length());
  ulong n;
  Term::Polarity basicCompPol = Signature::NegativePolarity;
  Term::Polarity nonBasicCompPol = Signature::PositivePolarity;
  LiteralProperties prop; 
  _divisionIntoComponents.reset();
  for (LitListElement* el = cl.literals()->first(); el; el = el->next())
    {
      if (el->literal()->header().arity() == 0UL)
	{
	}
      else 
	{
	  bool groundLit = !(el->literal()->numOfDifferentVariables());
	  if (groundLit)
	    {
	      _numLitMap[_divisionIntoComponents.nextLit(true,el->literal()->size())] = el;
	    }
	  else
	    {
	      _numLitMap[_divisionIntoComponents.nextLit(false,0UL)] = el;
	      const ulong* beginVar;
	      const ulong* endVar;
	      el->literal()->variables(beginVar,endVar);
	      for (const ulong* v = beginVar; v != endVar; ++v)
		_divisionIntoComponents.nextVar(*v);
	    };
	};  
    }; 
  if (_divisionIntoComponents.numOfComponents() == 0)
    return false;
  bool foldingOnly = false;
  if (_divisionIntoComponents.numOfComponents() == 1)
    {
      if (!_foldSingleComponents) return false;
      if (!_someNamesExist) return false;
      foldingOnly = true;
    };
  numOfComponents = _divisionIntoComponents.numOfComponents();
  numOfFoldings = numOfComponents - 1;
  numOfNewNames = 0UL;
  _registeredCompNums.reset();      
  ulong litNum;
  ulong compNum;
  while (_divisionIntoComponents.litCompPair(litNum,compNum))
    {      
      if (!_registeredCompNums.scoreOf(compNum))
	{
	  _registeredCompNums.add(compNum);
	  _components[compNum].reset();    
	};
      cl.literals()->passTo(_numLitMap[litNum],_components[compNum]);
      DOP(_numLitMap[litNum] = 0);
    };
  Term header;
  LitHandle newLiteral;
  ulong numOfNewComponents = 0UL;
  ulong numOfNamedComponents = 0UL;
  for (n = 0UL; n < _registeredCompNums.numOfDiffElements(); ++n)
    {
      Clause* nameIntroClause;
      compNum = _registeredCompNums.nth(n);
      bool isNameForNegation;
      _newCompName[compNum] =
	!findName(_components[compNum],
		  _componentNameHeader[compNum],
		  isNameForNegation,
		  nameIntroClause);
    check_if_new:
      if (_newCompName[compNum])
	{
	  ++numOfNewComponents;
	}
      else 
	{
	  if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(newLiteral))
	    {
	      for (ulong m = 0; 
		   m < _registeredCompNums.numOfDiffElements(); 
		   ++m)
		cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(m)]);
	      if (numOfNamedComponents)
		{
		  numOfComponents = numOfNamedComponents + 1;
		  numOfFoldings = numOfNamedComponents;
		  numOfNewNames = 0UL;
		  ASSERT(debugNumOfLits >= cl.literals()->length());
		  ASSERT(_someNamesExist);
		  return true;
		}
	      else
		{
		  ASSERT(debugNumOfLits == cl.literals()->length());
		  return false;
		};
	    };
	  ASSERT(!isNameForNegation || _reuseNamesForNegations); 
	  ASSERT(!isNameForNegation || 
		 TempLitListExtras::isGroundSingleton(_components[compNum]));
	  if (isNameForNegation && 
	      !nameIntroClause->isNameIntroWithAllPartsDeployed())
	    {
	      if (!Signature::current()->canCreateNSkolemSymbols(1UL) ||
		  Signature::current()->nextSymbolNumber() >= Term::maxFunctor())
		{
		  _newCompName[compNum] = true;
		  goto check_if_new;
		}
	      else
		{
		  ulong compNamePredNum = 
		    _componentNameHeader[compNum].functor();
		  ASSERT(compNamePredNum <= Term::maxFunctor());
		  if (collectNameIntroPart((Term::Functor)(compNamePredNum),
					   basicCompPol,
					   _components[compNum],
					   newClauseCollector,
					   nameIntroClause,
					   cl))
		    {
		      ASSERT(_components[compNum].isEmpty());
		      nameIntroClause->markAsNameIntroWithAllPartsDeployed();
		    }
		  else
		    {
		      _components[compNum].reset();
		    };
		};
	    };
	  newLiteral->pushPropLit(_componentNameHeader[compNum],prop);
	  cl.background().pushSimplifier(nameIntroClause,0L);
	  cl.background().UsedRule(ClauseBackground::Folding);
	  cl.literals()->collect(newLiteral);
	  numOfNamedComponents++;
	  _components[compNum].reset();
	};      
    }; 
  if (!Signature::current()->canCreateNSkolemSymbols(numOfNewComponents) ||
      Signature::current()->nextSymbolNumber() + numOfNewComponents - 1 > Term::maxFunctor() ||
      foldingOnly)
    {
      for (n = 0; n < _registeredCompNums.numOfDiffElements(); ++n)
	if (_newCompName[_registeredCompNums.nth(n)])
	  cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(n)]);
      ASSERT(debugNumOfLits >= cl.literals()->length());
      if (numOfNamedComponents > 0UL)
	{
	  numOfComponents = numOfNamedComponents + 1;
	  numOfFoldings = numOfNamedComponents;
	  ASSERT(_someNamesExist);
	  return true;
	}
      else
	return false;
    };
  bool basicComponentChosen = false;
  ulong basicComponentNum;
  if (!_smallBasicComponentInHyper) 
    {
      if (_compMakerForBasicComp)
	{
	  _basicComponendFinder.reset();      
	  for (n = 0UL; n < _registeredCompNums.numOfDiffElements(); ++n)
	    {
	      compNum = _registeredCompNums.nth(n);
	      if (_newCompName[compNum]) 
		{
		  _basicComponendFinder.anotherEntry(&(_components[compNum]),
						     compNum);
		};
	    };
	  if (!_basicComponendFinder.isClean())
	    {
	      basicComponentChosen = true;
	      basicComponentNum = _basicComponendFinder.max();
	    };
	}
      else 
	{
	  for (n = 0UL; 
	       (n < _registeredCompNums.numOfDiffElements() && !basicComponentChosen); 
	       ++n)
	    {
	      compNum = _registeredCompNums.nth(n);
	      if (_newCompName[compNum]) 
		{
		  basicComponentChosen = true;
		  basicComponentNum = compNum;
		};
	    };
	};
    }; 
  if (basicComponentChosen)
    {
      _components[basicComponentNum].passAllTo(*(cl.literals()));
      _registeredCompNums.erase(basicComponentNum);
    };
  for (n = 0UL; n < _registeredCompNums.numOfDiffElements(); ++n)
    {
      Clause* nameIntroClause;
      compNum = _registeredCompNums.nth(n);
      if (_newCompName[compNum]) 
	{
	  bool isNameForNegation;
	  _newCompName[compNum] = 
	    !findName(_components[compNum],
		      _componentNameHeader[compNum],
		      isNameForNegation,
		      nameIntroClause);
	  if (!_newCompName[compNum])
	    {
	      DOP(Term debugHeader = _componentNameHeader[compNum]);
	      DOP(Term debugFlippedHeader = debugHeader);
	      DOP(debugFlippedHeader.flipPolarity());
	      ASSERT(TempLitListExtras::containsHeader(*(cl.literals()),debugHeader) ||
		     (_reuseNamesForNegations && TempLitListExtras::containsHeader(*(cl.literals()),debugFlippedHeader)));
	      ++numOfNamedComponents;
	      --numOfNewComponents;
	      _components[compNum].reset();
	    };
	};
      if (_newCompName[compNum]) 
	{
	  if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(newLiteral))
	    {
	      for (n = 0; n < _registeredCompNums.numOfDiffElements(); ++n)
		if (_newCompName[_registeredCompNums.nth(n)])
		  cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(n)]);
	      if (numOfNamedComponents > 0UL)
		{
		  numOfComponents = numOfNamedComponents + 1;
		  numOfFoldings = numOfNamedComponents;
		  ASSERT(debugNumOfLits >= cl.literals()->length());
		  ASSERT(_someNamesExist);
		  return true;
		}
	      else
		{
		  ASSERT(debugNumOfLits == cl.literals()->length());
		  return false;
		};
	    };
	  ulong compNamePredNum;
	  Signature::current()->createNewSkolemPredicateNumber(0UL,
							       compNamePredNum);
	  ASSERT(compNamePredNum <= Term::maxFunctor());
	  ++numOfNewNames;
	  ++numOfNamedComponents;
	  header.MakeComplex((Term::Arity)0,
			     (Term::Functor)(compNamePredNum),
			     basicCompPol);
	  header.MakeSplitting();
	  Signature::current()->assignUniqueSymbolPrecedence(header.functor());
	  Signature::current()->assignDefaultHeaderRank(header.polarity(),
							header.functor()); 
	  newLiteral->pushPropLit(header,prop);
	  nameIntroClause = 
	    _componentNaming->registerNameIntro(_components[compNum],
						*newLiteral);
	  cl.literals()->collect(newLiteral);
	  cl.background().pushSimplifier(nameIntroClause,0L);
	  cl.background().UsedRule(ClauseBackground::Folding);
	  if (collectNameIntroPart((Term::Functor)(compNamePredNum),
				   nonBasicCompPol,
				   _components[compNum],
				   newClauseCollector,
				   nameIntroClause,
				   cl))
	    {
	      ASSERT(_components[compNum].isEmpty());
	    }
	  else
	    {
	      _components[compNum].reset();
	    };
	}; 
    }; 
  _someNamesExist = true;
  return true;
}; 
bool 
PropositionalSplitting::collectNameIntroPart(Term::Functor branchNum,
						Term::Polarity branchPolarity,
						BasicTempLitList& component,
						TemporaryClauseCollector* newClauseCollector,
						Clause* nameIntroClause,
						const TemporaryClause& originalClause)
{
  CALL("collectNameIntroPart(..)");
  LiteralProperties prop; 
  Term header((Term::Arity)0,branchNum,branchPolarity);
  header.MakeSplitting();  
  Signature::current()->assignDefaultHeaderRank(branchPolarity,
						branchNum);
  newClauseCollector->openNewClause(ClauseBackground::Clausification,nameIntroClause);
  newClauseCollector->addRule(ClauseBackground::PartOfNameIntroForPropSplitting);
  newClauseCollector->announceNewClause();
  if (!newClauseCollector->pushPropLiteral(header,prop)) 
    return false;
  if (!newClauseCollector->collectLiteralList(component))
    return false;
  return newClauseCollector->finalisePropNameIntroPart(originalClause);
}; 
inline
bool 
PropositionalSplitting::findName(BasicTempLitList& component,
				 Term& nameHeader,
				 bool& isNameForNegation,
				 Clause*& nameIntroClause)
{
  CALL("findName(..)");
  bool res = _componentNaming->findName(component,nameIntroClause);
  if (res)
    {
      nameHeader = *(nameIntroClause->LitList()->firstLit());
      ASSERT(nameHeader.IsSplitting());
      isNameForNegation = false;
      return true;
    }
  else
    if (_reuseNamesForNegations && 
	TempLitListExtras::isGroundSingleton(component))
      {
	component.first()->literal()->flipPolarity();
	res = _componentNaming->findName(component,nameIntroClause);
	component.first()->literal()->flipPolarity();
	if (res)
	  {
	    nameHeader = *(nameIntroClause->LitList()->firstLit());
	    nameHeader.flipPolarity();
	    ASSERT(nameHeader.IsSplitting());
	    isNameForNegation = true;
	    return true;
	  };
      };
  return false;
}; 
inline
bool 
PropositionalSplitting::adjustSizes(ulong numOfLiterals)
{
  CALL("adjustSizes(ulong numOfLiterals)");
  if (numOfLiterals > _maximalNumOfLiteralsSoFar)
    {
      if (!_divisionIntoComponents.adjustSizes(numOfLiterals))
	return false;
      DOP(_numLitMap.unfreeze());
      if (!_numLitMap.tryToExpand(numOfLiterals)) return false;
      DOP(_numLitMap.freeze());
      DOP(_components.unfreeze());  
      if (!_components.tryToExpand(numOfLiterals)) return false;
      DOP(_components.freeze());
      DOP(_registeredCompNums.unfreeze());
      if (!_registeredCompNums.tryToExpand(numOfLiterals)) return false;
      DOP(_registeredCompNums.freeze());
      DOP(_componentNameHeader.unfreeze());
      if (!_componentNameHeader.tryToExpand(numOfLiterals)) return false;
      DOP(_componentNameHeader.freeze());   
      DOP(_newCompName.unfreeze());
      if (!_newCompName.tryToExpand(numOfLiterals)) return false;
      DOP(_newCompName.freeze());
      _maximalNumOfLiteralsSoFar = numOfLiterals;
    }; 
  return true;
}; 
