/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PROPOSITIONAL_SPLITTING_H
#define PROPOSITIONAL_SPLITTING_H
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "ExpandingMultiset.hpp"
#include "OneMaxFinder.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ComponentName.hpp"
#include "DivIntoComp.hpp"
namespace VK
{
  class TemporaryClause;
  class TemporaryClauseCollector;
  class ComponentName;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROPOSITIONAL_SPLITTING
#  define DEBUG_NAMESPACE "PropositionalSplitting"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class PropositionalSplitting
  {
  public:
    typedef BasicTempLitList::Element LitListElement;
    typedef TemporaryLiteral::Extras::Handle LitHandle; 
  public:
    PropositionalSplitting(ComponentName* componentNaming);
    ~PropositionalSplitting();
    void reset();
    void setSmallBasicComponentInHyper(bool fl) {
      _smallBasicComponentInHyper = fl; };
    void setSingleGroundLitCompMinSize(long s) {
      _divisionIntoComponents.setSingleGroundLitCompMinSize(s);
    };
    void setFoldSingleComponents(bool fl) {
    _foldSingleComponents = fl; };
    void setReuseNamesForNegations(bool fl) {
      _reuseNamesForNegations =  fl; };
    void 
    setCompMakerForBasicComp(TempLitListExtras::ComparisonMaker* comp)
    {
      _compMakerForBasicComp = comp;
      _basicComponendFinder.switchTo(comp);
    };
    bool split(TemporaryClause& cl,
	       TemporaryClauseCollector* newClauseCollector,
	       ulong& numOfComponents,
	       ulong& numOfFoldings,
	       ulong& numOfNewNames)
    {
      return splitWithHyperParallel(cl,
				    newClauseCollector,
				    numOfComponents,
				    numOfFoldings,
				    numOfNewNames);
    };
  private:
    PropositionalSplitting() {};
    bool splitWithHyperParallel(TemporaryClause& cl,
				TemporaryClauseCollector* newClauseCollector,
				ulong& numOfComponents,
				ulong& numOfFoldings,
				ulong& numOfNewNames);
    bool 
    collectNameIntroPart(Term::Functor branchNum,
			 Term::Polarity branchPolarity,
			 BasicTempLitList& component,
			 TemporaryClauseCollector* newClauseCollector,
			 Clause* nameIntroClause,
			 const TemporaryClause& originalClause);
    bool findName(BasicTempLitList& component,
		  Term& nameHeader,
		  bool& isNameForNegation,
		  Clause*& nameIntroClause);
    bool adjustSizes(ulong numOfLiterals);
  private:
    bool _smallBasicComponentInHyper;
    bool _foldSingleComponents;
    bool _reuseNamesForNegations;
    TempLitListExtras::ComparisonMaker* _compMakerForBasicComp;
    ComponentName* _componentNaming;
    bool _someNamesExist;
    DivIntoComp<VampireKernelConst::MaxNumOfVariables> _divisionIntoComponents;
    BK::Array<BK::GlobAlloc,LitListElement*,BK::FastRelocation<LitListElement*>,32UL> _numLitMap; 
    BK::Array<BK::GlobAlloc,BasicTempLitList,BK::CustomRelocation<BasicTempLitList>,32UL> _components;   
    BK::ExpandingMultiset<BK::GlobAlloc,ulong,32UL> _registeredCompNums;
    BK::Array<BK::GlobAlloc,Term,BK::FastRelocation<Term>,32UL> _componentNameHeader; 
    BK::Array<BK::GlobAlloc,bool,BK::FastRelocation<bool>,32UL> _newCompName;
    ulong _maximalNumOfLiteralsSoFar;
    BK::OneMaxFinder<BasicTempLitList*,ulong,TempLitListExtras::ComparisonMaker> _basicComponendFinder;
  }; 
}; 
#endif
