/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "RulePremise.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_RULE_PREMISE
#  define DEBUG_NAMESPACE "RulePremise"
#endif
#include "debugMacros.hpp"
using namespace VK;
RulePremise::RulePremise() : 
  _termMemory(DOP("RulePremise::_termMemory")),
  _selectedLiteral(DOP("RulePremise::_selectedLiteral")),
  _selectedLiteralRotation(DOP("RulePremise::_selectedLiteralRotation")),
  _selectedLiteralProperties(DOP("RulePremise::_selectedLiteralProperties")),
  _litIsMarked(DOP("RulePremise::_litIsMarked"))
{
  CALL("constructor RulePremise()"); 
  DOP(_termMemory.freeze());
  DOP(_selectedLiteral.freeze());
  DOP(_selectedLiteralRotation.freeze());
  DOP(_selectedLiteralProperties.freeze());
  DOP(_litIsMarked.freeze());
}; 
RulePremise::~RulePremise() { CALL("destructor ~RulePremise()"); };
bool RulePremise::load(ulong ind,Clause* c)
{
  CALL("load(ulong ind,Clause* c)");
  ASSERT(!c->isAnswer());
  DOP(_termMemory.unfreeze());
  if (!_termMemory.tryToExpand(c->size() * 2)) return false;
  DOP(_termMemory.freeze());
  _freeTermMemory = _termMemory.memory();
  DOP(_selectedLiteral.unfreeze());
  if (!_selectedLiteral.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_selectedLiteral.freeze()); 
  DOP(_selectedLiteralRotation.unfreeze());
  if (!_selectedLiteralRotation.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_selectedLiteralRotation.freeze()); 
  DOP(_selectedLiteralProperties.unfreeze());
  if (!_selectedLiteralProperties.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_selectedLiteralProperties.freeze());
  DOP(_litIsMarked.unfreeze());
  if (!_litIsMarked.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_litIsMarked.freeze());
  clause = c;
  index = ind;
  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  _lastSelected = _allLiterals->numOfSelectedLits() - 1;
  ulong max_lit = 0; 
  for(LiteralList::Iterator iter(_allLiterals);
      iter.currentElement() != _firstNonselectedLit;
      iter.next())
    {
      _selectedLiteralProperties[max_lit] =
	iter.currentElement()->properties();
      _selectedLiteral[max_lit] = _freeTermMemory;
      _litIsMarked[max_lit] = false;
      if (!iter.currentLiteral()->IsEquality())
	{
	  UnifTerm::StandardTermToUnifTerm(ind,iter.currentLiteral(),_freeTermMemory);
	  _selectedLiteralRotation[max_lit] = 0;
	}
      else
	{
	  Term header = *(iter.currentLiteral());
	  header.MakeUnordered();
	  UnifTerm::binaryStandardTermToUnifTerm(ind,
					     header,
					     *(iter.currentLiteral()->Arg1()),
					     *(iter.currentLiteral()->Arg2()),
					     _freeTermMemory);
	  if (iter.currentLiteral()->IsUnorderedEq() &&
	      iter.currentElement()->properties().useEqForResolution())
	    {
	      _selectedLiteralRotation[max_lit] = _freeTermMemory;
	      UnifTerm::binaryStandardTermToUnifTerm(ind,
						 header,
						 *(iter.currentLiteral()->Arg2()),
						 *(iter.currentLiteral()->Arg1()),
						 _freeTermMemory);
	    }
	  else
	    _selectedLiteralRotation[max_lit] = 0;
	};
      max_lit++;
    };   
  return true;
}; 
bool RulePremise::loadOptimized(ulong ind,Clause* c)
{
  CALL("loadOptimized(ulong ind,Clause* c)");
  ASSERT(!c->isAnswer());
  DOP(_termMemory.unfreeze());
  if (!_termMemory.tryToExpand(c->size() * 2)) return false;
  DOP(_termMemory.freeze());
  _freeTermMemory = _termMemory.memory();
  DOP(_selectedLiteral.unfreeze());
  if (!_selectedLiteral.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_selectedLiteral.freeze()); 
  DOP(_selectedLiteralRotation.unfreeze());
  if (!_selectedLiteralRotation.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_selectedLiteralRotation.freeze()); 
  DOP(_selectedLiteralProperties.unfreeze());
  if (!_selectedLiteralProperties.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_selectedLiteralProperties.freeze());
  DOP(_litIsMarked.unfreeze());
  if (!_litIsMarked.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_litIsMarked.freeze());
  clause = c;
  index = ind;
  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  _lastSelected = _allLiterals->numOfSelectedLits() - 1;
  if (_lastSelected)
    {
      ulong max_lit = 0; 
      for(LiteralList::Iterator iter(_allLiterals);
	  iter.currentElement() != _firstNonselectedLit;
	  iter.next())
	{
	  _selectedLiteralProperties[max_lit] =
	    iter.currentElement()->properties();
	  _selectedLiteral[max_lit] = _freeTermMemory;
	  _litIsMarked[max_lit] = false;
	  if (!iter.currentLiteral()->IsEquality())
	    {
	      UnifTerm::StandardTermToUnifTerm(ind,iter.currentLiteral(),_freeTermMemory);
	      _selectedLiteralRotation[max_lit] = 0;
	    }
	  else
	    {
	      Term header = *(iter.currentLiteral());
	      header.MakeUnordered();
	      UnifTerm::binaryStandardTermToUnifTerm(ind,
						 header,
						 *(iter.currentLiteral()->Arg1()),
						 *(iter.currentLiteral()->Arg2()),
						 _freeTermMemory);
	      if (iter.currentLiteral()->IsUnorderedEq() &&
		  iter.currentElement()->properties().useEqForResolution())
		{
		  _selectedLiteralRotation[max_lit] = _freeTermMemory;
		  UnifTerm::binaryStandardTermToUnifTerm(ind,
						     header,
						     *(iter.currentLiteral()->Arg2()),
						     *(iter.currentLiteral()->Arg1()),
						     _freeTermMemory);
		}
	      else
		_selectedLiteralRotation[max_lit] = 0;
	    };
	  max_lit++;
	};   
    }
  else
    {
      _selectedLiteralProperties[0] =
	_allLiterals->firstElem()->properties();
      _selectedLiteral[0] = 0; 
      _selectedLiteralRotation[0] = 0;
      _litIsMarked[0] = false; 
    };
  return true;
}; 
#ifndef NO_DEBUG_VIS  
ostream& RulePremise::output(ostream& str) const
{
  str <<  "<<<<<< PREMISE >>>>>>>\n"
      << " Clause : " << clause << "\n Index = " << index
      << "\n Maximal literals: \n";
  for (ulong ml = 0; ml <= _lastSelected; ml++)
    {
      str << "  [" << _litIsMarked[ml] << "] ";
      UnifTerm::outputTerm(str,_selectedLiteral[ml]);
      if (_selectedLiteralRotation[ml])
	{
	  str << " {rotation: ";
	  UnifTerm::outputTerm(str,_selectedLiteralRotation[ml]);
	  str << "}";
	};
      str << '\n';
    };
  return str; 
}; 
#endif
