/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef RULE_PREMISE_HPP
#define RULE_PREMISE_HPP
#ifndef NO_DEBUG_VIS 
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "Clause.inl"
#include "UnifTerm.hpp" 
#include "UnifIndex.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
#include "TemporaryClauseCollector.hpp"
#include "LiteralProperties.hpp"
#include "TemporaryLiteral.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_RULE_PREMISE
#  define DEBUG_NAMESPACE "RulePremise"
#endif
#include "debugMacros.hpp"
namespace VK
{
class RulePremise
 {
  public: 
   RulePremise(); 
   ~RulePremise();
   UnifTerm* nthSelectedLit(ulong n) const
     {
       CALL("nthSelectedLit(ulong n) const");
       ASSERT(n <= _lastSelected);
       return _selectedLiteral[n]; 
     };
   UnifTerm* nthSelectedLitRotation(ulong n) const
     {
       CALL("nthSelectedLitRotation(ulong n) const");
       ASSERT(n <= _lastSelected);
       return _selectedLiteralRotation[n]; 
     };
   bool nthSelectedLitEligibleForResolution(ulong n) const
   {
     CALL("nthSelectedLitEligibleForResolution(ulong n) const");
     ASSERT(n <= _lastSelected);
     return 
       !_selectedLiteral[n]->symbol().Func().IsEquality() ||
       _selectedLiteralProperties[n].useEqForResolution();
   };
   Clause* TheClause() const { return clause; };
   ulong lastSelected() { return _lastSelected; };
   ulong Index() { return index; };
   ulong CutLit() { return cut_lit; };
   const LiteralProperties& resolvedLiteralProperties() const 
   {
     return _resolvedLiteralProperties;
   };
   bool load(ulong ind,Clause* c);
   bool loadOptimized(ulong ind,Clause* c);
   void Cut(ulong n) 
    {
     CALL("Cut(ulong n)");
     cut_lit = n;
     _litIsMarked[cut_lit] = true;
     _resolvedLiteralProperties = 
       _selectedLiteralProperties[cut_lit];
    };
   bool load(ulong ind,UnifIndex::IndexedObject* res_pair)
    {
     CALL("load(ulong ind,UnifIndex::IndexedObject* res_pair)");
     if (!load(ind,res_pair->TheClause())) return false;
     Cut((ulong)(res_pair->LitNum()));
     return true;
    };
   bool loadOptimized(ulong ind,UnifIndex::IndexedObject* res_pair)
    {
     CALL("loadOptimized(ulong ind,UnifIndex::IndexedObject* res_pair)");
     if (!loadOptimized(ind,res_pair->TheClause())) return false;
     Cut((ulong)(res_pair->LitNum()));
     return true;
    };
   void Cut(Term* lit) { Cut(clause->NumberOfLiteral(lit)); };
   void UnmarkCut() 
   { 
    CALL("UnmarkCut()");
    _litIsMarked[cut_lit] = false; 
   };
   bool collectUnmarkedLiterals(TemporaryClauseCollector& clauseCollector,
				const LiteralProperties::SLModifier& slmod);
   bool collectAllLiterals(TemporaryClauseCollector& clauseCollector); 
   bool collectSubtermsAvailableForParamodulationInto(TemporaryClauseCollector& clauseCollector,
						      bool collectMarked);
#ifndef NO_DEBUG_VIS  
   ostream& output(ostream& str) const;
#endif
  protected: 
   Clause* clause;
   ulong index;
   BK::Array<BK::GlobAlloc,UnifTerm,BK::FastRelocation<UnifTerm>,256UL> _termMemory;
   UnifTerm* _freeTermMemory;
   BK::Array<BK::GlobAlloc,UnifTerm*,BK::FastRelocation<UnifTerm*>,32UL> _selectedLiteral;
   BK::Array<BK::GlobAlloc,UnifTerm*,BK::FastRelocation<UnifTerm*>,32UL> _selectedLiteralRotation;
   BK::Array<BK::GlobAlloc,LiteralProperties,BK::FastRelocation<LiteralProperties>,32UL> _selectedLiteralProperties;
   LiteralProperties _resolvedLiteralProperties;
   ulong _lastSelected;
   BK::Array<BK::GlobAlloc,bool,BK::FastRelocation<bool>,32UL> _litIsMarked;
   LiteralList* _allLiterals;
   LiteralList::Element* _firstNonselectedLit;
   ulong cut_lit;
   bool nonmarkedExist;
 }; 
inline bool RulePremise::collectUnmarkedLiterals(TemporaryClauseCollector& clauseCollector,
						const LiteralProperties::SLModifier& slmod)
 {
  CALL("collectUnmarkedLiterals(TemporaryClauseCollector& clauseCollector,LiteralProperties::SLModifier& slmod)");
  ulong currMax = 0;
  LiteralProperties prop;
  nonmarkedExist = (_firstNonselectedLit != 0);   
  while (currMax <= _lastSelected)
   {   
    if (!(_litIsMarked[currMax])) 
     {
      nonmarkedExist = true;
      prop.inheritFrom(_selectedLiteralProperties[currMax],slmod);
      if (!clauseCollector.pushLiteralWithUnifier(_selectedLiteral[currMax],prop))
        return false; 
     };
    currMax++;
   };
  for(LiteralList::Iterator iter(_allLiterals,_firstNonselectedLit);
      iter.notEnd();
      iter.next())     
    {
      prop.inheritFrom(iter.currentElement()->properties(),slmod);
      if (!clauseCollector.pushLiteralWithUnifier(index,
						  iter.currentLiteral(),
						  prop)) 
	return false;
    };
  return true;
 }; 
inline bool RulePremise::collectAllLiterals(TemporaryClauseCollector& clauseCollector)
 {
  CALL("collectAllLiterals(TemporaryClauseCollector& clauseCollector)");
  for(LiteralList::Iterator iter(_allLiterals); iter.notEnd(); iter.next())     
    if (!clauseCollector.pushLiteralWithUnifier(index,
						iter.currentLiteral(),
						iter.currentElement()->properties())) 
     return false;
  return true;
 }; 
inline 
bool RulePremise::collectSubtermsAvailableForParamodulationInto(TemporaryClauseCollector& clauseCollector,
							       bool collectMarked)
{
  CALL("collectSubtermsAvailableForParamodulationInto(TemporaryClauseCollector& clauseCollector&...)");
  LiteralProperties prop;
  bool atLeastOneLiteralCollected = false;
  ulong litNum = 0UL;
  bool markedCollected = !collectMarked;
  for(LiteralList::Iterator iter(_allLiterals);
      iter.currentElement() != _firstNonselectedLit;
      iter.next())
    {
      if (!_litIsMarked[litNum] || !markedCollected)
	{
	  markedCollected = _litIsMarked[litNum];
	  Term* lit = iter.currentLiteral();
	  if (lit->IsEquality())
	    {
	      if (lit->IsOrderedEq())
		{
		  if (!clauseCollector.pushPseudoLiteralWithUnifier(index,
								     Term::dummyPositiveHeader1(),
								     *(lit->Args())))
		    return false;
		}
	      else
		if (!clauseCollector.pushLiteralWithUnifier(index,
							    lit,
							    prop))
		  return false;
	      atLeastOneLiteralCollected = true;
	    }
	  else
	    if (lit->arity())
	      {
		if (!clauseCollector.pushLiteralWithUnifier(index,
							    lit,
							    prop))
		  return false;
		atLeastOneLiteralCollected = true;
	      };
	};
      ++litNum;
    };
  return atLeastOneLiteralCollected;
}; 
}; 
#endif
