/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef SIZE_FORMULA_HPP
#define SIZE_FORMULA_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "Comparison.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "UnifTerm.hpp"
#include "Unifier.hpp"
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_SIZE_FORMULA
#  define DEBUG_NAMESPACE "SizeFormula"
#endif
#include "debugMacros.hpp"
namespace VK
{
class SizeFormula
 {
  public:
   SizeFormula() 
    : const_part(0),
      index(0)
   {
     CALL("constructor SizeFormula()");
   };
   ~SizeFormula() { CALL("destructor ~WEIGHT_FORMULA()"); };
   inline void SetIndex(unsigned long ind) 
     { 
       CALL("SetIndex(unsigned long ind)");
       index = ind; 
     };
   inline void Reset()
    {
     _varMultiset.reset();
     const_part = 0;
    };  
   inline void RegVar(const Unifier::Variable* v) { _varMultiset.registerVar(v); };
   inline void RegFunc() { const_part++; };
   inline long Compute() 
    {
     long res = const_part;
     _varMultiset.resetIter();
     for (const Unifier::Variable* var = _varMultiset.nextVar();
	  var;
	  var = _varMultiset.nextVar())
      {
       bool free_var;
       const void* bind = var->Unref(free_var);
       long w = (free_var)? 1 : static_cast<const UnifTerm*>(bind)->size(); 
       res += _varMultiset.score(var) * w;
      };
     return res; 
    };
   inline bool GreaterThan(long x)
    {
     x -= const_part;
     if (x < 0) return true;
     _varMultiset.resetIter();
     for (const Unifier::Variable* var = _varMultiset.nextVar();
	  var;
	  var = _varMultiset.nextVar())
      {
       bool free_var;
       const void* bind = var->Unref(free_var);
       long w = (free_var)? 1 : static_cast<const UnifTerm*>(bind)->size(); 
       x -= _varMultiset.score(var) * w;
       if (x < 0) return true;
      };
     return false; 
    }; 
   inline void Compile(Term* t)
    {
     Term::Iterator iter;
     iter.Reset(t);
     do
      {
       if (iter.CurrentSymIsVar())
        {
         RegVar(Unifier::current()->variable(iter.symbolRef().var(),index));
        }
       else RegFunc();   
      }
     while (iter.Next()); 
    }; 
   inline void Compile(Term t) 
    {
     if (t.isVariable())
      {
       RegVar(Unifier::current()->variable(t.var(),index));
      }
     else Compile(t.deref()); 
    };
   inline void Compile(UnifTerm* t)
    {
     CALL("Compile(UnifTerm* t)");
     if (t->symbol().isVariable()) 
      {
       ASSERT(t->symbol().var()->Index() == index);
       RegVar(t->symbol().var()); 
       return; 
      };
     UnifTerm::SimpleIter iter;
     iter.Reset(t);
     do
      {
       UnifTerm::Symbol sym = iter.CurrPos()->symbol();
       if (sym.isVariable())
        {
         ASSERT(sym.var()->Index() == index);
         RegVar(sym.var());
        }
       else RegFunc(); 
       iter.SkipSym();
      }
     while (!iter.Empty()); 
    }; 
  public: 
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str);
   #endif
  private:
   Unifier::Variable::Multiset _varMultiset;
   long const_part;
   unsigned long index;
 }; 
}; 
#endif
