/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef STANDARD_KB_ORDERING_HPP
#define STANDARD_KB_ORDERING_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Array.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "SimplificationOrdering.hpp"
#include "Flatterm.hpp"
#include "Term.hpp"
#include "UnifTerm.hpp"
#include "WeightPolynomial.hpp"
#include "Comparison.hpp"
#include "GlobAlloc.hpp"
#include "ObjectPoolWithMemory.hpp"
#include "WeightPolynomialSpecialisedComparisonCommand.hpp"
#include "ForwardMatchingSubstitutionCache.hpp"
#include "InstanceRetrievalForDemodulation.hpp"
#include "LocalUnification.hpp"
#include "InstanceWithUnifier.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class StandardKBOrdering : public SimplificationOrdering
  {
  public:
    StandardKBOrdering();
    ~StandardKBOrdering();
    void* operator new(size_t);
    void operator delete(void* obj);
    BK::Comparison compare(const Flatterm* term1,
		       const WeightPolynomial& weight1,
		       const Flatterm* term2,
		       const WeightPolynomial& weight2);
    BK::Comparison compareAtoms(const Flatterm* lit1,
			    const WeightPolynomial& weight1,
			    const Flatterm* lit2,
			    const WeightPolynomial& weight2);
    BK::Comparison compareOnCurrentUnifier(UnifTerm* term1,
				       UnifTerm* term2);
    BK::Comparison compareOnCurrentUnifier(const Term& term1,
				       const Term& term2,
				       ulong variableIndex);
    BK::Comparison compare(const Term* complexTerm1,
		       const Term* complexTerm2,
		       const InstRet::Substitution* subst);
    BK::Comparison compare(const Term& term1,
		       const Term& term2,
		       const ForwardMatchingSubstitutionCache* subst);
    SimplificationOrdering::SpecialisedComparison* reserveSpecialisedComparison();
    void releaseSpecialisedComparison(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintGreater(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintGreaterOrEqual(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintLess(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintLessOrEqual(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint*
    storedConstraintNotEqual(SimplificationOrdering::SpecialisedComparison* sc);
    SimplificationOrdering::StoredConstraint* copy(SimplificationOrdering::StoredConstraint* constr);
    void releaseConstraint(SimplificationOrdering::StoredConstraint* constr);
    SimplificationOrdering::ConstraintCheckOnUnifier* reserveConstraintCheckOnUnifier();
    void releaseConstraintCheckOnUnifier(SimplificationOrdering::ConstraintCheckOnUnifier* ccu);
    SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* 
    reserveConstraintCheckOnForwardMatchingSubst();
    void 
    releaseConstraintCheckOnForwardMatchingSubst(SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* ccfms);
    SimplificationOrdering::CompiledConstraintForUnifiers* reserveCompiledConstraintForUnifiers();
    void releaseCompiledConstraintForUnifiers(SimplificationOrdering::CompiledConstraintForUnifiers* cc);
    SimplificationOrdering::CompiledConstraintForInstanceRetrieval* reserveCompiledConstraintForInstanceRetrieval();
    void releaseCompiledConstraintForInstanceRetrieval(SimplificationOrdering::CompiledConstraintForInstanceRetrieval* cc);
    BK::Comparison compareWithTransposition(const Flatterm* term);
#ifdef NO_DEBUG
  private:
#else
  public:
#endif
    class SpecialisedComparison;
    class ConstraintCheckOnUnifier;
    class ConstraintCheckOnForwardMatchingSubst;
    class StoredConstraint;
    class CompiledConstraintForUnifiers;
    class CompiledConstraintForInstanceRetrieval;
  private:
    typedef WeightPolynomial::SpecialisedComparison::Extras WPSCExtras;
  private:
    BK::Comparison compare(const Flatterm* term1,
		       const Flatterm* term2);
    BK::Comparison compareLexicographically(const Flatterm* term1,
					const Flatterm* term2);
    BK::Comparison compareAtomsLexicographically(const Flatterm* lit1,
					     const Flatterm* lit2);
    BK::Comparison compareLexicographically(const Term* complexTerm1,
					const Term* complexTerm2);
    BK::Comparison compareLexicographically(const Term& term1,
					const Term& term2);
    bool termWithUnifierContainsVariable(UnifTerm* complexTerm,
					 Unifier::Variable* var);
    BK::Comparison compareLexOnCurrentUnifier(UnifTerm* term1,
					      UnifTerm* term2);
    bool termWithUnifierContainsVariable(const Term* complexTerm,
					 ulong variableIndex,
					 Unifier::Variable* var);
    bool termWithUnifierContainsVariable(const Term& term,
					 ulong variableIndex,
					 Unifier::Variable* var);
    BK::Comparison compareOnCurrentUnifier(UnifTerm* complexTerm1,
				       const Term* complexTerm2,
				       ulong variableIndex);
    BK::Comparison compareLexOnCurrentUnifier(UnifTerm* complexTerm1,
					  const Term* complexTerm2,
					  ulong variableIndex);
    BK::Comparison compareLexOnCurrentUnifier(UnifTerm* complexTerm1,
					  const Term& term2,
					  ulong variableIndex);
    BK::Comparison compareLexOnCurrentUnifier(const Term& term1,
					  const Term& term2,
					  ulong variableIndex);
    BK::Comparison compareLexicographically(const Term* complexTerm1,
					const InstRet::Substitution* subst,
					const Term* complexTerm2);
    BK::Comparison compareLexicographically(const Term* complexTerm1,
					const InstRet::Substitution* subst,
					const Term& term2);
    BK::Comparison compare(const Term* complexTerm1,
		       const InstRet::Substitution* subst,
		       const Term& term2);
    BK::Comparison compare(const Term* complexTerm1,const Term* complexTerm2);
    BK::Comparison compare(const Term& term1,const Term& term2);
    BK::Comparison compare(const Term& term1,
		       const Term& term2,
		       const InstRet::Substitution* subst);
    BK::Comparison compareLexicographically(const Term* complexTerm1,
					const Term* complexTerm2,
					const InstRet::Substitution* subst);
    BK::Comparison compare(const Term& term1,
		       const ForwardMatchingSubstitutionCache* subst,
		       const Flatterm* term2);
    BK::Comparison compareLexicographically(const Term* complexTerm1,
					const ForwardMatchingSubstitutionCache* subst,
					const Flatterm* term2);
    BK::Comparison compareLexicographically(const Term& term1,
					const Term& term2,
					const ForwardMatchingSubstitutionCache* subst);
    BK::ObjectPool<WeightPolynomial::SpecialisedComparison>& 
    specialisedWeightComparisonPool() 
    { 
      return _specialisedWeightComparisonPool; 
    };
  private:
    SpecialisedComparison* _specialisedComparisonPool;
    ConstraintCheckOnUnifier* _constraintCheckOnUnifierPool;
    ConstraintCheckOnForwardMatchingSubst* _constraintCheckOnForwardMatchingSubstPool;
    CompiledConstraintForUnifiers* _compiledConstraintOnUnifiersPool;
    CompiledConstraintForInstanceRetrieval* _compiledConstraintForInstanceRetrievalPool;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,WeightPolynomial::SpecialisedComparison> 
    _specialisedWeightComparisonPool;
    WeightPolynomial _weight1;
    WeightPolynomial _weight2;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,UnifTerm::Iter> _prefixIterPool;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,InstanceWithUnifier> _stTermWithSubstPool;
    ulong _numberOfCompiledConstraintsForInstanceRetrieval;
    ulong _numberOfConstraintChecksForInstanceRetrieval;
    ulong _numberOfCompiledConstraintsOnUnifiers;
    ulong _numberOfConstraintChecksOnUnifiers;
    friend class SpecialisedComparison;
    friend class ConstraintCheckOnUnifier;
    friend class ConstraintCheckOnForwardMatchingSubst;
    friend class CompiledConstraintForUnifiers;
    friend class CompiledConstraintForInstanceRetrieval;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::SpecialisedComparison"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class StandardKBOrdering::SpecialisedComparison :
    public SimplificationOrdering::SpecialisedComparison
  {
  public:
    void load(const Term& lhs,const Term& rhs);
    bool lhsVarInstMayOrder();
    bool rhsVarInstMayOrder();
    void specialiseGreater();
    void specialiseGreaterOrEqual();
    void specialiseLess();
    void specialiseLessOrEqual();
    void specialiseNotEqual();
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
#endif
  private:
    SpecialisedComparison(StandardKBOrdering* ord) : 
      _ordering(ord),
      _nextInPool(0),
      _constraintGreater(0),
      _constraintGreaterOrEqual(0),
      _constraintLess(0),
      _constraintLessOrEqual(0),
      _constraintNotEqual(0)
    {
    };
    ~SpecialisedComparison();
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(SpecialisedComparison));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(SpecialisedComparison));
    };
    SpecialisedComparison* getNextInPool() { return _nextInPool; };
    void setNextInPool(StandardKBOrdering::SpecialisedComparison* sc) { _nextInPool = sc; };
    void specialise();
    StandardKBOrdering::StoredConstraint* storedConstraintGreater();
    StandardKBOrdering::StoredConstraint* storedConstraintGreaterOrEqual();
    StandardKBOrdering::StoredConstraint* storedConstraintLess();
    StandardKBOrdering::StoredConstraint* storedConstraintLessOrEqual();
    StandardKBOrdering::StoredConstraint* storedConstraintNotEqual();
  private:
    StandardKBOrdering* _ordering;
    Term _lhs;
    Term _rhs;
    WeightPolynomial _lhsWeight;
    WeightPolynomial _rhsWeight;
    bool _specialised;
    WeightPolynomial::SpecialisedComparison _weightComparison;
    SpecialisedComparison* _nextInPool;
    StandardKBOrdering::StoredConstraint* _constraintGreater;
    StandardKBOrdering::StoredConstraint* _constraintGreaterOrEqual;
    StandardKBOrdering::StoredConstraint* _constraintLess;
    StandardKBOrdering::StoredConstraint* _constraintLessOrEqual;
    StandardKBOrdering::StoredConstraint* _constraintNotEqual;
    friend class StandardKBOrdering;
    friend class StandardKBOrdering::CompiledConstraintForUnifiers;
    friend class StandardKBOrdering::CompiledConstraintForInstanceRetrieval;
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline
  ostream& operator<<(ostream& str,
		      const VK::StandardKBOrdering::SpecialisedComparison& sc)
  {
    return sc.output(str);
  };
  inline
  ostream& operator<<(ostream& str,
		      const VK::StandardKBOrdering::SpecialisedComparison* sc)
  {
    return sc->output(str);
  };
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::ConstraintCheckOnUnifier"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class StandardKBOrdering::ConstraintCheckOnUnifier :
    public SimplificationOrdering::ConstraintCheckOnUnifier
  {
  public:
    bool holds(const SimplificationOrdering::StoredConstraint* constr,
	       ulong variableIndex);
  private:
    ConstraintCheckOnUnifier(StandardKBOrdering* ord) : 
      _ordering(ord),
      _nextInPool(0) 
    {
    };
    ~ConstraintCheckOnUnifier();   
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(ConstraintCheckOnUnifier));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(ConstraintCheckOnUnifier));
    };  
    ConstraintCheckOnUnifier* getNextInPool() { return _nextInPool; };
    void setNextInPool(StandardKBOrdering::ConstraintCheckOnUnifier* ccu) { _nextInPool = ccu; };
    bool holdsGreater(const StoredConstraint* constr);
    bool holdsGreaterOrEqual(const StoredConstraint* constr);
    bool holdsLess(const StoredConstraint* constr);
    bool holdsLessOrEqual(const StoredConstraint* constr);
    bool holdsNotEqual(const StoredConstraint* constr);
    BK::FunctionComparison 
    checkGreaterOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison);
    BK::FunctionComparison 
    checkLessOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison);
    BK::BoolPlus
    checkNotEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison);
    bool lexicorgaphicallyGreater(const Term& lhs,const Term& rhs);
    bool lexicorgaphicallyGreaterOrEqual(const Term& lhs,const Term& rhs);
    bool lexicorgaphicallyLess(const Term& lhs,const Term& rhs);
    bool lexicorgaphicallyLessOrEqual(const Term& lhs,const Term& rhs);
  private:
    ConstraintCheckOnUnifier();
  private:
    StandardKBOrdering* _ordering;
    ConstraintCheckOnUnifier* _nextInPool;
    ulong _currentVariableIndex;
    InstanceWithUnifier _lhsIter;
    InstanceWithUnifier _rhsIter;
    WeightPolynomial _weightPositivePartInstance;
    WeightPolynomial _weightNegativePartInstance;
    friend class StandardKBOrdering;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst :
    public SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst
  {
  public:
    bool holds(const SimplificationOrdering::StoredConstraint* constr,
	       const ForwardMatchingSubstitutionCache* subst);
  private: 
    ConstraintCheckOnForwardMatchingSubst(StandardKBOrdering* ord) : 
      _ordering(ord),
      _nextInPool(0)
    {
    };
    ~ConstraintCheckOnForwardMatchingSubst();
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(ConstraintCheckOnForwardMatchingSubst));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(ConstraintCheckOnForwardMatchingSubst));
    };  
    ConstraintCheckOnForwardMatchingSubst* getNextInPool() { return _nextInPool; };
    void setNextInPool(StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst* ccfms) 
    { 
      _nextInPool = ccfms; 
    };
    bool holdsGreater(const StoredConstraint* constr);
    bool holdsGreaterOrEqual(const StoredConstraint* constr);
    bool holdsLess(const StoredConstraint* constr);
    bool holdsLessOrEqual(const StoredConstraint* constr);
    bool holdsNotEqual(const StoredConstraint* constr);
    BK::FunctionComparison 
    checkGreaterOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison);
    BK::FunctionComparison 
    checkLessOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison);
    BK::BoolPlus
    checkNotEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison);
    bool lexicorgaphicallyGreater(const Term& lhs,const Term& rhs);
    bool lexicorgaphicallyGreaterOrEqual(const Term& lhs,const Term& rhs);
    bool lexicorgaphicallyLess(const Term& lhs,const Term& rhs);
    bool lexicorgaphicallyLessOrEqual(const Term& lhs,const Term& rhs);
  private:
    ConstraintCheckOnForwardMatchingSubst();
  private:
    StandardKBOrdering* _ordering;
    ConstraintCheckOnForwardMatchingSubst* _nextInPool;
    const ForwardMatchingSubstitutionCache* _subst;
    ForwardMatchingSubstitutionCache::Instance _lhsInstance;
    ForwardMatchingSubstitutionCache::Instance _rhsInstance;
    WeightPolynomial _weightPositivePartInstance;
    WeightPolynomial _weightNegativePartInstance;
    friend class StandardKBOrdering;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::StoredConstraint"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class StandardKBOrdering::StoredConstraint
  {
  public:
    enum TargetRelation { Greater, GreaterOrEqual, Less, LessOrEqual, NotEqual };
  public:
    StoredConstraint(TargetRelation tr,
		     const WeightPolynomial::SpecialisedComparison& wc,
		     const Term& lhs,
		     const Term& rhs) :
      _referenceCounter(0L),
      _targetRelation(tr),
      _weightComparison(wc),
      _lhs(lhs),
      _rhs(rhs)
    {
      CALL("constructor StoredConstraint(TargetRelation tr,..)");
    };
    ~StoredConstraint() 
    {
      CALL("destructor ~StoredConstraint()");
      ASSERT(!referenceCounter());
    };
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(StoredConstraint));
    };  
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(StoredConstraint));
    };
    const long& referenceCounter() const { return _referenceCounter; };
    void incReferenceCounter() 
    {
      CALL("incReferenceCounter()");
      ++_referenceCounter; 
    };
    void decReferenceCounter() 
    {
      CALL("decReferenceCounter()");
      --_referenceCounter;
      ASSERT(_referenceCounter >= 0L);
    };
    const TargetRelation& targetRelation() const { return _targetRelation; };
    const WeightPolynomial::SpecialisedComparison::Stored& weightComparison() const
    {
      return _weightComparison;
    };
    const Term& lhs() const { return _lhs; };
    const Term& rhs() const { return _rhs; };
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
      switch (targetRelation())
	{
	case Greater:        str << "Greater        "; break;
	case GreaterOrEqual: str << "GreaterOrEqual "; break;
	case Less:           str << "Less           "; break;
	case LessOrEqual:    str << "LessOrEqual    "; break;
	case NotEqual:       str << "NotEqual       "; break;
	default:             str << "?????????????? ";
	};
      str << lhs() << " vs " << rhs() << " --> " << weightComparison();
      return str;
    };
#endif
  private:
    StoredConstraint();
    StoredConstraint(const StoredConstraint&);
    StoredConstraint& operator=(const StoredConstraint&);
  private:
    long _referenceCounter;
    TargetRelation _targetRelation;
    WeightPolynomial::SpecialisedComparison::Stored _weightComparison;
    Term _lhs;
    Term _rhs;
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline
  ostream& operator<<(ostream& str,
		      const VK::StandardKBOrdering::StoredConstraint& constr)
  {
    return constr.output(str);
  };
  inline
  ostream& operator<<(ostream& str,
		      const VK::StandardKBOrdering::StoredConstraint* constr)
  {
    return constr->output(str);
  };
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::CompiledConstraintForUnifiers"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class StandardKBOrdering::CompiledConstraintForUnifiers :
    public SimplificationOrdering::CompiledConstraintForUnifiers
  {
  public:
    CompiledConstraintForUnifiers(StandardKBOrdering* ord) :
      _ordering(ord),
      _nextInPool(0),
      _weightComparisonCode(DOP("StandardKBOrdering::CompiledConstraintForUnifiers::_weightComparisonCode")),
      _mainCode(DOP("StandardKBOrdering::CompiledConstraintForUnifiers::_mainCode"))
    {
      _weightComparisonCode.expand();
      DOP(_weightComparisonCode.freeze());
      _mainCode.expand();
      DOP(_mainCode.freeze());
    };
    ~CompiledConstraintForUnifiers();  
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(CompiledConstraintForUnifiers));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(CompiledConstraintForUnifiers));
    }; 
    CompiledConstraintForUnifiers* getNextInPool() const { return _nextInPool; };
    void setNextInPool(CompiledConstraintForUnifiers* cc) { _nextInPool = cc; };
    void load(UnifTerm* lhs,UnifTerm* rhs);
    bool lhsVarInstMayOrder();
    bool rhsVarInstMayOrder();
    void compileGreaterOrEqual();
    void compileGreater();
    bool holdsOnCurrentUnifier();
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
#endif
  private:
    class Command
    {
    public:
      enum Tag
      {
	RetSuccess,
	RetFailure,
	CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL,
	IfWCFailRetFail
      };
    public:
      Command() {};
      ~Command() {};
      const Tag& tag() const { return _tag; };
      void setTag(const Tag& t) 
      { 
	_tag = t;
      };
      bool isTerminal() const;
      bool requiresLabel() const;
      UnifTerm* termS() const { return _termS; };
      void setTermS(UnifTerm* t)
      {
	CALL("Command::setTermS(UnifTerm* t)");
	ASSERT(t->symbol().IsFunctor());
	_termS = t;
      };
      UnifTerm* termT() const { return _termT; };
      void setTermT(UnifTerm* t)
      {
	CALL("Command::setTermT(UnifTerm* t)");
	ASSERT(t->symbol().IsFunctor());
	_termT = t;
      };
      Unifier::Variable* varX() const { return _varX; };
      void setVarX(Unifier::Variable* v) 
      { 
	_varX = v; 
	DOP(_varXNum = v->absoluteVarNum());
      };
      Unifier::Variable* varY() const { return _varY; };
      void setVarY(Unifier::Variable* v) 
      { 
	_varY = v; 
	DOP(_varYNum = v->absoluteVarNum());
      };
#ifndef NO_DEBUG
      ulong varXNum() const { return _varXNum; };
      ulong varYNum() const { return _varYNum; };
#endif
      WeightPolynomial::SpecialisedComparison::Command* 
      weightComparisonSubroutine() const 
      {
	return _weightComparisonSubroutine;
      };
      void setWeightComparisonSubroutine(WeightPolynomial::SpecialisedComparison::Command* lab)
      {
	_weightComparisonSubroutine = lab;
      };
      Command* labelL() const { return _labelL; };
      void setLabelL(Command* l) { _labelL = l; };
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str,const Command* base) const;
#endif
    private:
      Tag _tag;       
#ifndef NO_DEBUG
      ulong _varXNum;
      ulong _varYNum;
#endif
      Unifier::Variable* _varX;
      Unifier::Variable* _varY;
      UnifTerm* _termS;
      UnifTerm* _termT;
      WeightPolynomial::SpecialisedComparison::Command* 
      _weightComparisonSubroutine;
      Command* _labelL;
    }; 
  private:
    typedef WeightPolynomial::SpecialisedComparison::Command WeightComparisonCommand;
#ifndef NO_DEBUG
    enum TargetRelation
    {
      GreaterOrEqual,
      Greater
    };
#endif
  private:
    CompiledConstraintForUnifiers();
    BK::Comparison compareLexicographically(UnifTerm* term1,UnifTerm* term2);
    BK::FunctionComparison interpretWeightComparisonCode(WeightComparisonCommand* subroutine);
    ulong currentSpecialisationStrength() const
    {
      return 
	_specialisationStrengthBase + 
	(_ordering->_numberOfConstraintChecksOnUnifiers /
	 _ordering->_numberOfCompiledConstraintsOnUnifiers) /
	_inverseSpecialisationStrengthRate;
    };
    bool compileGreaterOrEqual(UnifTerm* lhs,
			       UnifTerm* rhs,
			       WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
			       Command*& mainCode,
			       const Command* endOfMainCode,
			       WeightComparisonCommand*& weightComparisonCode,
			       const WeightComparisonCommand* endOfWeightComparisonCode,
			       bool& complete,
			       ulong& specialisationStrength);
    bool compileGreaterOrEqualLex(UnifTerm* lhsComplexTerm,
				  UnifTerm* rhsComplexTerm,
				  WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
				  Command*& mainCode,
				  const Command* endOfMainCode,
				  WeightComparisonCommand*& weightComparisonCode,
				  const WeightComparisonCommand* endOfWeightComparisonCode,
				  bool& complete,
				  ulong& specialisationStrength);
  private:
    StandardKBOrdering* _ordering;
    StandardKBOrdering::CompiledConstraintForUnifiers* _nextInPool;
#ifndef NO_DEBUG
    TargetRelation _targetRelation;
#endif
    UnifTerm* _lhs;
    UnifTerm* _rhs;
    UnifTerm::Iter _lhsIter;
    UnifTerm::Iter _rhsIter;
    WeightPolynomial _lhsWeight;
    WeightPolynomial _rhsWeight;
    WeightPolynomial::SpecialisedComparison _topLevelWeightComparison;
    BK::Array<BK::GlobAlloc,
      WeightComparisonCommand,
      BK::FastRelocation<WeightComparisonCommand>,
      64UL> _weightComparisonCode;
    BK::Array<BK::GlobAlloc,Command,BK::FastRelocation<Command>,32UL> _mainCode;
    bool _grnd;
    long _acc;
    WeightPolynomial _posinst;
    WeightPolynomial _neginst;
    WeightPolynomial _weight1;
    WeightPolynomial _weight2;
    enum 
    { 
      _specialisationStrengthBase = 0UL,
      _inverseSpecialisationStrengthRate = 4UL
    };
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::CompiledConstraintForInstanceRetrieval"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class StandardKBOrdering::CompiledConstraintForInstanceRetrieval :
    public SimplificationOrdering::CompiledConstraintForInstanceRetrieval
  {
  public:
    CompiledConstraintForInstanceRetrieval(StandardKBOrdering* ord) :
      _ordering(ord),
      _nextInPool(0),
      _weightComparisonCode(DOP("StandardKBOrdering::CompiledConstraintForInstanceRetrieval::_weightComparisonCode")),
      _mainCode(DOP("StandardKBOrdering::CompiledConstraintForInstanceRetrieval::_mainCode"))
    {
      _weightComparisonCode.expand();
      DOP(_weightComparisonCode.freeze());
      _mainCode.expand();
      DOP(_mainCode.freeze());
    };
    ~CompiledConstraintForInstanceRetrieval();  
    void* operator new(size_t)
    {
      return BK::GlobAlloc::allocate(sizeof(CompiledConstraintForInstanceRetrieval));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(CompiledConstraintForInstanceRetrieval));
    }; 
    CompiledConstraintForInstanceRetrieval* getNextInPool() const { return _nextInPool; };
    void setNextInPool(CompiledConstraintForInstanceRetrieval* cc) { _nextInPool = cc; };
    void loadGreater(const Term* lhs,const Term* rhs);
    void loadNotEqual(const Term* lhs,const Term* rhs);
    bool lhsVarInstMayOrder();
    bool rhsVarInstMayOrder();
    void compile(const InstRet::Substitution* subst);
    bool holds();
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
#endif
  private:
    class Command
    {
    public:
      enum Tag
      {
	RetSuccess,
	RetFailure,
	CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareLexXY_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareLexXT_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	CompareLexSY_IfCmpGrRetSucc_IfCmpNotEqRetFail,
	IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL,
	IfWCFailRetFail
      }; 
    public:
      Command() {};
      ~Command() {};
      const Tag& tag() const { return _tag; };
      void setTag(const Tag& t) 
      { 
	_tag = t;
      };
      bool isTerminal() const;
      bool requiresLabel() const 
      { 
	return _tag == IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL;
      };
      const Term& termS() const { return _termS; };
      void setTermS(const Term& t) 
      {
	CALL("setTermS(const Term& t)");
	ASSERT(t.IsReference());
	_termS = t; 
      };
      const Term& termT() const { return _termT; };
      void setTermT(const Term& t) 
      {
	CALL("setTermT(const Term& t)");
	ASSERT(t.IsReference());
	_termT = t; 
      };
      const Term* varX() const { return _varX; };
      void setVarX(const Term* v) { _varX = v; };
      const Term* varY() const { return _varY; };
      void setVarY(const Term* v) { _varY = v; };
#ifndef NO_DEBUG
      ulong varXNum() const { return _varXNum; };
      void setVarXNum(ulong v) { _varXNum = v; };
      ulong varYNum() const { return _varYNum; };
      void setVarYNum(ulong v) { _varYNum = v; };
#endif
      WeightPolynomial::SpecialisedComparison::Command* 
      weightComparisonSubroutine() const 
      {
	return _weightComparisonSubroutine;
      };
      void setWeightComparisonSubroutine(WeightPolynomial::SpecialisedComparison::Command* lab)
      {
	_weightComparisonSubroutine = lab;
      };
      Command* labelL() const { return _labelL; };
      void setLabelL(Command* l) { _labelL = l; };
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str,const Command* base) const;
      ostream& outputShort(ostream& str,const Command* base) const;
#endif
    private:
      Tag _tag;       
#ifndef NO_DEBUG
      ulong _varXNum;
      ulong _varYNum;
#endif
      const Term* _varX;
      const Term* _varY; 
      Term _termS;
      Term _termT;
      WeightPolynomial::SpecialisedComparison::Command* 
      _weightComparisonSubroutine;
      Command* _labelL;
    }; 
  private:
    typedef WeightPolynomial::SpecialisedComparison::Command WeightComparisonCommand;
    enum TargetRelation
    {
      Greater,
      NotEqual
    };
  private:
    CompiledConstraintForInstanceRetrieval();
    BK::FunctionComparison interpretWeightComparisonCode(WeightComparisonCommand* subroutine);
    bool compileGreaterOrEqual(const Term& lhs,
			       const Term& rhs,
			       WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
			       Command*& mainCode,
			       const Command* endOfMainCode,
			       WeightComparisonCommand*& weightComparisonCode,
			       const WeightComparisonCommand* endOfWeightComparisonCode,
			       bool& complete,
			       ulong& specialisationStrength);
    bool compileGreaterOrEqualLex(const Term& lhs,
				  const Term& rhs,
				  WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
				  Command*& mainCode,
				  const Command* endOfMainCode,
				  WeightComparisonCommand*& weightComparisonCode,
				  const WeightComparisonCommand* endOfWeightComparisonCode,
				  bool& complete,
				  ulong& specialisationStrength);
    ulong currentSpecialisationStrength() const
    {
      return 
	_specialisationStrengthBase + 
	(_ordering->_numberOfConstraintChecksForInstanceRetrieval /
	 _ordering->_numberOfCompiledConstraintsForInstanceRetrieval) /
	_inverseSpecialisationStrengthRate;
    };
  private:
    StandardKBOrdering* _ordering;
    StandardKBOrdering::CompiledConstraintForInstanceRetrieval* _nextInPool;
    TargetRelation _targetRelation;
    const InstRet::Substitution* _subst;
    Term _lhs;
    Term _rhs;
    WeightPolynomial _lhsWeight;
    WeightPolynomial _rhsWeight;
    WeightPolynomial::SpecialisedComparison _topLevelWeightComparison;
    BK::Array<BK::GlobAlloc,
      WeightComparisonCommand,
      BK::FastRelocation<WeightComparisonCommand>,
      64UL> _weightComparisonCode;
    BK::Array<BK::GlobAlloc,Command,BK::FastRelocation<Command>,64UL> _mainCode;
    bool _grnd;
    long _acc;
    WeightPolynomial _posinst;
    WeightPolynomial _neginst;
    LocalUnification _assumedUnifier;
    WeightPolynomial _weight1;
    WeightPolynomial _weight2;
    enum 
    { 
      _specialisationStrengthBase = 0UL,
      _inverseSpecialisationStrengthRate = 4UL
    };
  }; 
}; 
#endif
