/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "StandardLiteralOrdering.hpp"
#include "TemporaryLiteral.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_LITERAL_ORDERING
#  define DEBUG_NAMESPACE "StandardLiteralOrdering"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
StandardLiteralOrdering::StandardLiteralOrdering()
{
  _isStratified = true;
};
StandardLiteralOrdering::~StandardLiteralOrdering()
{
};
Comparison StandardLiteralOrdering::compare(TemporaryLiteral& lit1,
					    TemporaryLiteral& lit2) const
{
  CALL("compare(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  BK::Comparison cmp;
  cmp = 
    Signature::current()->compareHeaderRanks(lit1.polarity(),
					     lit1.header().functor(),
					     lit2.polarity(),
					     lit2.header().functor());
  ASSERT(cmp != Incomparable);
  if (cmp != Equal) return cmp;
  if (lit1.isEquality())
    {
      if (lit2.isEquality())
	{
	  if (lit1.polarity() == lit2.polarity())
	    return compareEqSamePolarity(lit1,lit2);
	  if (lit1.isNegative())
	    {
	      return compareEqNegPos(lit1,lit2);
	    }
	  else 
	    {
	      return inverse(compareEqNegPos(lit2,lit1));
	    };
	}
      else 
	{
	  return Less; 
	};
    }
  else 
    {
      if (lit2.isEquality())
	{
	  return Greater;
	};
    };
  cmp =  
    SimplificationOrdering::current()->compareAtoms(lit1.flatterm(),
						    lit1.weight(),
						    lit2.flatterm(),
						    lit2.weight()); 
  if (cmp == Equal)
    {
      if (lit1.polarity() == lit2.polarity()) return Equal;
      return (lit1.isNegative()) ? Greater : Less;
    };
  return cmp;
}; 
Comparison StandardLiteralOrdering::compareLayers(const TemporaryLiteral& lit1,
						  const TemporaryLiteral& lit2) const
{
  CALL("compareLayers(const TemporaryLiteral& lit1,const TemporaryLiteral& lit2) const");
  ASSERT(isStratified());
  BK::Comparison cmp;
  cmp = 
    Signature::current()->compareHeaderRanks(lit1.polarity(),
					     lit1.header().functor(),
					     lit2.polarity(),
					     lit2.header().functor());
  ASSERT(cmp != Incomparable);
  if (cmp != Equal) return cmp;
  if (lit1.isEquality())
    {
      if (lit2.isEquality())
	{
	  return Equal;
	}
      else 
	{
	  return Less; 
	};
    }
  else 
    {
      if (lit2.isEquality())
	{
	  return Greater;
	};
    };
  return Equal;
}; 
inline
Comparison 
StandardLiteralOrdering::compareEqSamePolarity(TemporaryLiteral& lit1,
					       TemporaryLiteral& lit2) const
{
  CALL("compareEqSamePolarity(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isEquality());
  ASSERT(lit2.isEquality());
  ASSERT(lit1.polarity() == lit2.polarity());
  Comparison acmp;
  if (lit1.isOrderedEquality())
    {
      if (lit2.isOrderedEquality())
	{
	  acmp = compareEqAtomsOO(lit1,lit2);
	}
      else
	{
	  acmp = compareEqAtomsOU(lit1,lit2);
	};
    }
  else
    if (lit2.isOrderedEquality())
      {
	acmp = inverse(compareEqAtomsOU(lit2,lit1));
      }
    else 
      {
	acmp = compareEqAtomsUU(lit1,lit2);
      };
  if (acmp == Equal)
    {     
      if (lit1.polarity() == lit2.polarity()) return Equal;
      return (lit1.isPositive()) ? Less : Equal;
    };
  return acmp;
}; 
Comparison StandardLiteralOrdering::compareEqAtomsOO(TemporaryLiteral& lit1,
					     TemporaryLiteral& lit2) const
{
  CALL("compareEqAtomsOO(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isOrderedEquality());
  ASSERT(lit2.isOrderedEquality());
  Comparison cmp11 = 
    SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
					       lit2.arg1(),lit2.arg1Weight()); 
  if (cmp11 != Equal) return cmp11;
  return 
    SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
					       lit2.arg2(),lit2.arg2Weight()); 
}; 
Comparison StandardLiteralOrdering::compareEqAtomsOU(TemporaryLiteral& lit1,
					             TemporaryLiteral& lit2) const
{
  CALL("compareEqAtomsOU(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isOrderedEquality());
  ASSERT(!lit2.isOrderedEquality());
  switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
						     lit2.arg1(),lit2.arg1Weight()))
    {
    case Less:
      return Less;
    case Equal: 
      return SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							lit2.arg2(),lit2.arg2Weight());	
    case Greater: 
      return SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							lit2.arg2(),lit2.arg2Weight());
    case Incomparable:         
      switch (SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							 lit2.arg1(),lit2.arg1Weight()))
	{
	case Less: 
	  if (SimplificationOrdering::current()->lessOrEqual(lit1.arg1(),lit1.arg1Weight(),
							     lit2.arg2(),lit2.arg2Weight()))
	    {
	      return Less;
	    }
	  else
	    return Incomparable;
	case Equal:
	  return SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							    lit2.arg2(),lit2.arg2Weight());
	case Greater:
	  switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							     lit2.arg2(),lit2.arg2Weight()))
	    {
	    case Less: return Less;
	    case Equal: return Greater;
	    case Greater: return Greater;
	    case Incomparable: 
	      if (SimplificationOrdering::current()->greaterOrEqual(lit1.arg2(),lit1.arg2Weight(),
								    lit2.arg2(),lit2.arg2Weight()))
		{
		  return Greater;
		}
	      else
		return Incomparable;
	    };
	case Incomparable:
	  if (SimplificationOrdering::current()->less(lit1.arg1(),lit1.arg1Weight(),
						      lit2.arg2(),lit2.arg2Weight()))
	    {
	      return Less;
	    }
	  else
	    return Incomparable;
	};
    }; 
  ICP("ICP0");
  return Incomparable;
}; 
Comparison StandardLiteralOrdering::compareEqAtomsUU(TemporaryLiteral& lit1,
						     TemporaryLiteral& lit2) const
{
  CALL("compareEqAtomsUU(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(!lit1.isOrderedEquality());
  ASSERT(!lit2.isOrderedEquality());
  Comparison cmp = SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),lit2.arg1(),lit2.arg1Weight());
  switch (cmp)
    {
    case Greater:
      cmp = SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),lit2.arg2(),lit2.arg2Weight()); 
      switch (cmp)
	{
	case Greater:
	  return Greater;
	case Less:
	  return SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	case Incomparable:
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      return Greater;
	    case Less:
	      return Incomparable;
	    case Incomparable:
	      return Incomparable;
	    case Equal:
	      return Greater;
	    };
	case Equal:
#ifdef DEBUG_NAMESPACE 
	  ICP("DE20"); 
#endif
	  return Incomparable;
	};
    case Less:
      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
      switch (cmp)
	{
	case Greater:
	  return SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	case Less:
	  return Less; 
	case Incomparable:
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      return Incomparable;
	    case Less:
	      return Less;
	    case Incomparable:
	      return Incomparable;
	    case Equal:
	      return Less; 
	    }; 
	case Equal:
#ifdef DEBUG_NAMESPACE 
	  ICP("DE40");
#endif
	  return Incomparable;
	};
    case Incomparable:
      cmp = SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),lit2.arg2(),lit2.arg2Weight()); 
      switch (cmp)
	{
	case Greater:
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      return Greater; 
	    case Less:
	      return Incomparable;
	    case Incomparable:
	      return Incomparable;
	    case Equal:
	      return Greater;
	    };
	case Less:
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight());
	      ASSERT(cmp != Equal);
	      return cmp;
	    case Less:
	      return Less;
	    case Incomparable:
	      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	      if (cmp == Less) return Less;
	      return Incomparable;
	    case Equal:
	      return Less; 
	    }; 
	case Incomparable:
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	      if (cmp == Greater) return Greater;
	      ASSERT(cmp != Equal);
	      return Incomparable;
	    case Less:
	      return Incomparable;
	    case Incomparable:
	      return Incomparable;
	    case Equal:
	      return Incomparable;
	    }; 
	case Equal:
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  return cmp;
	};
    case Equal:
      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
      return cmp;
    };
  ICP("DE100");
  return Incomparable; 
}; 
inline
Comparison 
StandardLiteralOrdering::compareEqNegPos(TemporaryLiteral& lit1,
					 TemporaryLiteral& lit2) const
{
  CALL("compareEqNegPos(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isNegative());
  ASSERT(lit2.isPositive());
  if (lit1.isOrderedEquality())
    {
      if (lit2.isOrderedEquality())
	{
	  return compareEqNegPosOO(lit1,lit2);
	}
      else
	return compareEqNegPosOU(lit1,lit2);
    }
  else 
    {
      if (lit2.isOrderedEquality())
	{
	  return compareEqNegPosUO(lit1,lit2);
	}
      else
	return compareEqNegPosUU(lit1,lit2);
    };
}; 
inline
Comparison 
StandardLiteralOrdering::compareEqNegPosOO(TemporaryLiteral& lit1,
					   TemporaryLiteral& lit2) const
{
  CALL("compareEqNegPosOO(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isNegative());
  ASSERT(lit2.isPositive());
  ASSERT(lit1.isOrderedEquality());
  ASSERT(lit2.isOrderedEquality());
  Comparison cmp11 = 
    SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
					       lit2.arg1(),lit2.arg1Weight()); 
  if (cmp11 != Equal) return cmp11;
  return Greater;
}; 
inline
Comparison 
StandardLiteralOrdering::compareEqNegPosOU(TemporaryLiteral& lit1,
					   TemporaryLiteral& lit2) const
{
  CALL("compareEqNegPosOU(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isNegative());
  ASSERT(lit2.isPositive());
  ASSERT(lit1.isOrderedEquality());
  ASSERT(!lit2.isOrderedEquality());
  switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
						     lit2.arg1(),lit2.arg1Weight()))
    {
    case Less: return Less;
    case Equal:
      switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							 lit2.arg2(),lit2.arg2Weight()))
	{
	case Less: return Less;
	case Equal: return Greater;
	case Greater: return Greater;
	case Incomparable: return Incomparable;
	};
    case Greater: 
      switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							 lit2.arg2(),lit2.arg2Weight()))
	{
	case Less: return Less;
	case Equal: return Greater;
	case Greater: return Greater;;
	case Incomparable: return Incomparable;
	};
    case Incomparable: return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
inline
Comparison 
StandardLiteralOrdering::compareEqNegPosUO(TemporaryLiteral& lit1,
					   TemporaryLiteral& lit2) const
{
  CALL("compareEqNegPosUO(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isNegative());
  ASSERT(lit2.isPositive());
  ASSERT(!lit1.isOrderedEquality());
  ASSERT(lit2.isOrderedEquality());
  switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
						     lit2.arg1(),lit2.arg1Weight()))
    {
    case Less: 
      switch (SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							 lit2.arg1(),lit2.arg1Weight()))
	{
	case Less: return Less;
	case Equal: return Greater; 
	case Greater: return Greater;
	case Incomparable: return Incomparable;
	};
    case Equal: return Greater;
    case Greater: return Greater;
    case Incomparable: return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
inline
Comparison 
StandardLiteralOrdering::compareEqNegPosUU(TemporaryLiteral& lit1,
					   TemporaryLiteral& lit2) const
{
  CALL("compareEqNegPosUO(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isNegative());
  ASSERT(lit2.isPositive());
  ASSERT(!lit1.isOrderedEquality());
  ASSERT(!lit2.isOrderedEquality());
  switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
						     lit2.arg1(),lit2.arg1Weight()))
    {
    case Less: 
      switch (SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							 lit2.arg1(),lit2.arg1Weight()))
	{
	case Less: return Less;
	case Equal: 
	  return Incomparable;
	case Greater:
	  switch (SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							     lit2.arg2(),lit2.arg2Weight()))
	    {
	    case Less: return Less;
	    case Equal: return Greater;
	    case Greater: return Greater;
	    case Incomparable: return Incomparable;
	    };
	case Incomparable: return Incomparable;
	};
    case Equal: 
      switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							 lit2.arg2(),lit2.arg2Weight()))
	{
	case Less: 
	  switch (SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							     lit2.arg2(),lit2.arg2Weight()))
	    {
	    case Less: return Less;
	    case Equal: return Greater;
	    case Greater: return Greater;
	    case Incomparable: return Incomparable;
	    };
	case Equal:
	  return Greater;
	case Greater:
	  return Greater;
	case Incomparable: return Incomparable;
	};
    case Greater: 
      switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							 lit2.arg2(),lit2.arg2Weight()))
	{
	case Less: 
	  switch (SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							     lit2.arg2(),lit2.arg2Weight()))
	    {
	    case Less: return Less;
	    case Equal: return Greater;
	    case Greater: return Greater;
	    case Incomparable: return Incomparable;
	    }; 
	case Equal:
	  return Greater;
	case Greater:
	  return Greater;
	case Incomparable: return Incomparable;
	};
    case Incomparable: return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
