/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef STATISTICS_HPP
#define STATISTICS_HPP
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
using namespace std;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class Statistics
  {
  public:
    class Inferences
    {
    public:
      class Deduction
      {
      public:
	class RuleProfile
	{
	public:
	  RuleProfile() { reset(); };
	  void reset();
	  void absorb(const RuleProfile& stat);
	  bool somethingRegistered() const;
	  bool someNonredundantDiscarded() const;
	  ulong totalNonredundantDiscarded() const;
	  ulong totalNonredundantBranchesDiscarded() const;
	  ostream& outputDetailed(ostream& str) const;
	public:
	  ulong accomplished; 
	  ulong redundantDiscarded; 
	  ulong redundantBranchesDiscarded;
	  ulong discardedBySize;
	  ulong branchesDiscardedBySize;
	  ulong discardedParamodulationsFromVariables;
	  ulong branchesDiscardedWithParamodulationsFromVariables;
	  ulong discardedChainingsThroughVariables;
	  ulong branchesDiscardedWithChainingsThroughVariables;
	  ulong mainPremisesSubsumed;
	  ulong sidePremisesSubsumed;
	  ulong mainPremisesDemodulated;
	  ulong sidePremisesDemodulated;
	  ulong discardedForOtherReasons;
	  ulong branchesDiscardedForOtherReasons;
	}; 
      public:
	Deduction() { reset(); };
	void reset();
	void absorb(const Deduction& stat);
	bool somethingRegistered() const
	{
	  return 
	    resolution.somethingRegistered() ||
	    forwardSuperposition.somethingRegistered() ||
	    backwardSuperposition.somethingRegistered() ||
	    equalityFactoring.somethingRegistered() ||
	    equalityResolution.somethingRegistered() ||
	    forwardChaining.somethingRegistered() ||
	    backwardChaining.somethingRegistered();
	};
	ulong totalAccomplished() const;
	ulong totalRedundantDiscarded() const;
	ulong totalRedundantBranchesDiscarded() const;
	ulong totalNonredundantDiscarded() const;
	ulong totalNonredundantBranchesDiscarded() const;
      public:
	RuleProfile resolution;
	RuleProfile forwardSuperposition;
	RuleProfile backwardSuperposition;
	RuleProfile equalityFactoring;
	RuleProfile equalityResolution;
	RuleProfile forwardChaining;
	RuleProfile backwardChaining;
      }; 
      class Simplifying
      {
      public:
	Simplifying() { reset(); };
	void reset();
	void absorb(const Simplifying& stat);
	ulong total() const;
	ulong totalForward() const;
	ulong totalBackward() const;
      public:
	ulong propositionalTautologyDeletion;
	ulong equationalTautologyDeletion;
	ulong reductionInNonBasicPosition;
	ulong forwardSubsumption;
	ulong backwardSubsumption;
	ulong forwardSubsumptionResolution;
	ulong forwardDemodulation;
	ulong onBranchForwardDemodulation;
	ulong backwardDemodulation;
	ulong onBranchBackwardDemodulation;
	ulong destructiveEqualityResolution;
	ulong externalEvaluation;
	ulong forwardFolding;
	ulong backwardFolding;
      }; 
      class PropositionalSplitting
      {
      public:
	PropositionalSplitting() { reset(); };
	void reset();
	void absorb(const PropositionalSplitting& stat);
	bool somethingRegistered() const;
      public:
	ulong accomplished;
	ulong components;
	ulong introducedComponentNames;
      }; 
      class NegativeEqualitySplitting
      {
      public:
	NegativeEqualitySplitting() { reset(); };
	void reset();
	void absorb(const NegativeEqualitySplitting& stat);
	bool somethingRegistered() const;
      public:
	ulong accomplished;
	ulong argumentsSplitOff;
	ulong introducedNames;
      }; 
    public:
      Inferences() { reset(); };
      void reset();
      void absorb(const Inferences& stat);
      bool someNonredundantDiscarded() const
      {
	return 
	  deduction.totalNonredundantDiscarded() ||
	  deduction.totalNonredundantBranchesDiscarded();
      };
      ostream& outputDetailed(ostream& str) const;
      ostream& outputSummary(ostream& str) const;
    public:
      Deduction deduction;
      Simplifying simplifying;
      PropositionalSplitting propositionalSplitting;
      NegativeEqualitySplitting negativeEqualitySplitting;
    }; 
    class Clauses
    {
      class Generated
      {
      public:
	Generated() { reset(); };
	void reset();
	void absorb(const Generated& stat);
	ulong total() const;
	ulong totalGeneratedByDeduction() const;
	ulong totalGeneratedBySimplification() const;
	ulong totalRedundant() const;
	ulong totalNonredundantDiscarded() const;
	ulong totalSimplified() const;
      public:
	ulong byResolution;
	ulong byForwardSuperposition;
	ulong byBackwardSuperposition;
	ulong byEqualityFactoring;
	ulong byEqualityResolution;
	ulong byForwardChaining;
	ulong byBackwardChaining;
	ulong byBackwardDemodulation;
	ulong byOnBranchBackwardDemodulation;
	ulong byBackwardFolding;
	ulong byPropositionalSplitting;
	ulong byNegativeEqualitySplitting;
	ulong byOther;
	ulong discardedPropositionalTautologies;
	ulong discardedEquationalTautologies;
	ulong reducedInNonBasicPosition;
	ulong subsumed;
	ulong discardedWithTooDeepInferences;
	ulong discardedTooLarge;
	ulong discardedWithTooLargeLiterals;
	ulong discardedWithTooManyLiterals;
	ulong discardedTooDeep;
	ulong discardedWithTooManyVariables;
	ulong discardedForOtherReasons;
	ulong simplifiedBySubsumptionResolution;
	ulong simplifiedByDemodulation;
	ulong simplifiedByOnBranchDemodulation;
	ulong simplifiedByDestructiveEqualityResolution;
	ulong simplifiedByExternalEvaluation;
	ulong transformedByPropositionalSplitting;
	ulong transformedByNegativeEqualitySplitting;
      }; 
      class Retained
      {
      public:
	Retained() { reset(); };
	void reset();
	void absorb(const Retained& stat);
	ulong totalSimplified() const;
	ulong totalNonredundantDiscarded() const
	{
	  return passiveRecycled + activeRecycled;
	};
      public:
	ulong total;
	ulong selected;
	ulong passiveSubsumed;	
	ulong activeSubsumed;
	ulong passiveSimplifiedByDemodulation;
	ulong activeSimplifiedByDemodulation;
	ulong passiveSimplifiedByOnBranchDemodulation;
	ulong activeSimplifiedByOnBranchDemodulation;
	ulong passiveBackFolded;
	ulong activeBackFolded;
	ulong passiveOrphansMurdered;
	ulong activeOrphansMurdered;
	ulong passiveRecycled;
	ulong activeRecycled;
      }; 
    public:
      Clauses() { reset(); };
      void reset();
      void absorb(const Clauses& stat);
      bool someNonredundantDiscarded() const 
      {
	return 
	  generated.totalNonredundantDiscarded() ||
	  retained.totalNonredundantDiscarded();
      };
      ostream& outputDetailed(ostream& str) const;
      ostream& outputSummary(ostream& str) const;
    public:
      Generated generated;
      Retained retained;
    }; 
  public:
    Statistics() { reset(); };
    void reset();
    void absorb(const Statistics& stat);
    bool completenessViolated() const;
    ostream& outputDetailed(ostream& str) const;
    ostream& outputSummary(ostream& str) const;
  public:
    Inferences inferences;
    Clauses clauses;
  }; 
}; 
#endif
