/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "TempLitListExtras.hpp"
#include "SymbolInfo.hpp"
using namespace BK;
using namespace VK;
TempLitListExtras::Context* TempLitListExtras::Context::_current; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras"
#endif
#include "debugMacros.hpp"
long 
TempLitListExtras::computeWeightBase(const BasicTempLitList& lst)
{
  CALL("computeWeightBase(const BasicTempLitList& lst)");
  long res = 0L;
  for (const Element* el = lst.first(); el; el = el->next())
    res += el->literal()->weightBase();
  return res;
}; 
long
TempLitListExtras::size(const BasicTempLitList& lst)
{
  CALL("size(const BasicTempLitList& lst)");
  long res = 0L;
  for (const Element* el = lst.first(); el; el = el->next())
    {
      res += el->literal()->size();
    };
  return res;
}; 
void
TempLitListExtras::computeSizeAndWeightBase(const BasicTempLitList& lst,
					    long& size,
					    long& weightBase)
{
  CALL("computeSizeAndWeightBase(..)");
  size = 0L;
  weightBase = 0L;
  for (const Element* el = lst.first(); el; el = el->next())
    {
      size += el->literal()->size();
      weightBase += el->literal()->weightBase();
    };
}; 
void
TempLitListExtras::computeSizeAndWeightBaseAndNumOfPosEq(const BasicTempLitList& lst,
							 long& size,
							 long& weightBase,
							 long& numOfPosEq)
{
  CALL("computeSizeAndWeightBaseAndNumOfPosEq(..)");
  size = 0L;
  weightBase = 0L;
  numOfPosEq = 0L;
  for (const Element* el = lst.first(); el; el = el->next())
    {
      size += el->literal()->size();
      weightBase += el->literal()->weightBase();
      if (el->literal()->isEquality() &&
	  el->literal()->isPositive())
	++numOfPosEq;
    };
}; 
long 
TempLitListExtras::minimalPosEqLHSSize(const BasicTempLitList& lst)
{
  CALL("minimalPosEqLHSSize(const BasicTempLitList& lst)");
  for (const Element* el = lst.first(); el; el = el->next())
    if (el->literal()->isEquality() && 
	el->literal()->isPositive())
      {
	long res = el->literal()->minPosEqLHSSize();
	for (el = el->next(); el; el = el->next())
	  if (el->literal()->isEquality() && 
	      el->literal()->isPositive())
	    {
	      long tmp = el->literal()->minPosEqLHSSize();
	      if (tmp < res) res = tmp;
	    };
	return res;
      };
    return 0L;
}; 
bool 
TempLitListExtras::instantiationIsPossible(const BasicTempLitList& lst,
					   ulong var,
					   ulong instanceDepth,
					   long instanceSize,
					   bool checkInputNonBasicness)
{
  CALL("instantiationIsPossible(..)");
  ASSERT(instanceDepth);
  for (const Element* el = lst.first(); el; el = el->next())
    if (!el->literal()->instantiationPossible(var,
					      instanceDepth,
					      instanceSize,
					      checkInputNonBasicness))
      return false;
  return true;
}; 
void
TempLitListExtras::passInOrder(BasicTempLitList& src,
			       BasicTempLitList& dest,
			       const VariableWeighting& varWeighting)
{
  CALL("passInOrder(..,const VariableWeighting& varWeighting)");
  DOP(ulong debugDestInitLength = dest.length());
  DOP(ulong debugSrcInitLength = src.length());
  if (src.isEmpty()) return;
  if (src.length() == 1UL) 
    {
      src.popTo(dest);
      return;
    };
  if (src.length() == 2UL)
    {
      Element* fst = src.first();
      Element* snd = src.last();
      if (fst->literal()->compareWithVarWeighting(*(snd->literal()),varWeighting) == 
	  BK::Greater)
	{
	  src.popTo(dest);
	  src.popTo(dest);
	}
      else 
	{
	  src.passTo(fst,dest);
	  src.passTo(snd,dest);
	};
      return;
    };
  LitHandle pivot;
  src.popTo(pivot);
  BasicTempLitList greaterThanPivot(src.elementPool());
  for (Element* el = src.first(); el;  )
    {
      Element* tmp = el->next();
      if (el->literal()->compareWithVarWeighting(*pivot,varWeighting) == 
	  BK::Greater) 
	{
	  src.passTo(el,greaterThanPivot);
	};
      el = tmp;
    };
  passInOrder(src,dest,varWeighting);
  dest.collect(pivot);
  passInOrder(greaterThanPivot,dest,varWeighting);
  ASSERT(src.isEmpty());
  ASSERT(greaterThanPivot.isEmpty());
  ASSERT(dest.length() == debugDestInitLength + debugSrcInitLength);
}; 
void 
TempLitListExtras::renameVariables(const BasicTempLitList& lst,
				   VariableRenaming& varRenaming)
{
  CALL("renameVariables(..)");
  for (const Element* el = lst.last(); el; el = el->previous()) 
    el->literal()->renameVariables(varRenaming);
}; 
void TempLitListExtras::weed(BasicTempLitList& lst,
			     bool& isTautology,
			     bool& isEqTautology)
{
  CALL("weed(BasicTempLitList& lst,bool& isTautology,bool& isEqTautology)");
  for (Element* el1 = lst.first(); el1; )
    {
      Element* next = el1->next();
      LitHandle lit1 = el1->literal();
      if (lit1->isEquality())
	{
	  if (lit1->argsAreEqual())
	    {
	      if (lit1->isPositive())
		{
		  isTautology = true;
		  isEqTautology = true;
		  return;
		}
	      else 
		lst.remove(el1);
	    }
	  else 
	    {
	      lit1->orderEqArgs();
	      for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
		if (el2->literal()->isEquality())
		  {
		    bool eqSymmetryUsed;
		    if (el2->literal()->equalEqLit(*lit1,eqSymmetryUsed))
		      {
			lit1->properties().factoriseWith(el2->literal()->properties(),eqSymmetryUsed);
			lst.remove(el2);
		      }
		    else
		      if (el2->literal()->complementaryEqLit(*lit1))
			{
			  isTautology = true;
			  isEqTautology = false;
			  return;
			};
		  };
	    };	    
	}
      else 
	{
	  if (lit1->header() == Term::builtInTrue())
	    {
	      isTautology = true;
	      isEqTautology = false;
	      return;
	    };
	  if (lit1->header() == Term::builtInFalse())
	    {
	      lst.remove(el1);
	    }
	  else
	    {
	      for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
		if (el2->literal()->equal(*lit1))
		  {
		    lit1->properties().factoriseWith(el2->literal()->properties(),false);
		    lst.remove(el2);
		  }
		else
		  if (el2->literal()->isComplementary(*lit1))
		    {
		      isTautology = true;
		      isEqTautology = false;
		      return;
		    };
	    };
	}; 
      el1 = next;
    }; 
  isTautology = false;
  isEqTautology = false;
}; 
void TempLitListExtras::weed(BasicTempLitList& lst,
			     bool& isTautology,
			     bool& isEqTautology,
			     const SymbolInfo* commutativityTableForEqTaut,
			     const SymbolInfo* commutativityTableForPropTaut,
			     const SymbolInfo* commutativityTableForTrivialEqRes,
			     const SymbolInfo* commutativityTableForIdenticalLitFactoring,
			     BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("weed(..,const SymbolInfo* commutativityTableForEqTaut,..)");
  bool commutativityUsed;
  for (Element* el1 = lst.first(); el1; )
    {
      Element* next = el1->next();
      LitHandle lit1 = el1->literal();
      if (lit1->isEquality())
	{
	  if (lit1->isPositive())
	    {
	      if (commutativityTableForEqTaut)
		{
		  if (lit1->argsAreEqual(commutativityTableForEqTaut,commutativityUsed))
		    {
		      if (commutativityUsed)
			{
			  if ((lit1->arg1()->size() == 3L) &&
			      lit1->arg1()->next()->isVariable() &&
			      lit1->arg1()->next()->next()->isVariable())
			    {
			    }
			  else
			    {
			      lit1->commutativityLawsForArgsAreEqual(commutativityTableForEqTaut,
								     usedCommutativityLaws);	
			      isTautology = true;
			      isEqTautology = true;
			      return;
			    };			  
			}
		      else
			{
			  isTautology = true;
			  isEqTautology = true;
			  return;
			};
		    };
		}
	      else
		if (lit1->argsAreEqual())
		  {
		    isTautology = true;
		    isEqTautology = true;
		    return;
		  };
	      goto scan_for_equal_or_complementary;
	    }
	  else 
	    {
	      bool resolved;
	      if (commutativityTableForTrivialEqRes)
		{
		  resolved = 
		    lit1->argsAreEqual(commutativityTableForTrivialEqRes,commutativityUsed);
		  if (resolved)
		    lit1->commutativityLawsForArgsAreEqual(commutativityTableForTrivialEqRes,
							   usedCommutativityLaws);	
		}
	      else
		resolved = lit1->argsAreEqual();
	      if (resolved)
		{
		  lst.remove(el1);
		}
	      else
		{
		scan_for_equal_or_complementary:
		  for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
		    {
		      if (el2->literal()->isEquality())
			if (el2->literal()->polarity() == lit1->polarity())
			  {
			    bool eqSymmetryUsed;
			    if (equalEq(el2->literal(),
					lit1,
					commutativityTableForIdenticalLitFactoring,
					usedCommutativityLaws,
					eqSymmetryUsed))
			      {
				lit1->properties().factoriseWith(el2->literal()->properties(),eqSymmetryUsed);
				lst.remove(el2);
			      };
			  }
			else 
			  {
			    if (complementaryEq(el2->literal(),
						lit1,
						commutativityTableForPropTaut,
						usedCommutativityLaws))
			      {
				isTautology = true;
				isEqTautology = false;
				return;
			      };
			  };
		    }; 
		  lit1->orderEqArgs();
		};
	    };
	}
      else 
	{
	  if (lit1->header() == Term::builtInTrue())
	    {
	      isTautology = true;
	      isEqTautology = false;
	      return;
	    };
	  if (lit1->header() == Term::builtInFalse())
	    {
	      lst.remove(el1);
	    }
	  else
	    {
	      for (Element* el2 = lst.first(); el2 != el1; el2 = el2->next())
		{
		  if (el2->literal()->polarity() == lit1->polarity())
		    {
		      if (equal(el2->literal(),
				lit1,
				commutativityTableForIdenticalLitFactoring,
				usedCommutativityLaws))
			{
			  lit1->properties().factoriseWith(el2->literal()->properties(),false);
			  lst.remove(el2);
			};
		    }
		  else 
		    {
		      if (complementary(el2->literal(),
					lit1,
					commutativityTableForPropTaut,
					usedCommutativityLaws))
			{
			  isTautology = true;
			  isEqTautology = false;
			  return;
			};
		    };
		}; 
	    };
	}; 
      el1 = next;
    }; 
  isTautology = false;
  isEqTautology = false;
}; 
inline
bool 
TempLitListExtras::complementary(const LitHandle& lit1,
				 const LitHandle& lit2,
				 const SymbolInfo* commutativityTable,
				 BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("complementary(const LitHandle& lit1,...)");
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->isComplementary(*lit2,
				commutativityTable,
				commutativityUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForComplementary(*lit2,
						    commutativityTable, 
						    usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->isComplementary(*lit1);
}; 
inline
bool 
TempLitListExtras::equal(const LitHandle& lit1,
			 const LitHandle& lit2,
			 const SymbolInfo* commutativityTable,
			 BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("equal(const LitHandle& lit1,...)");
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->equal(*lit2,
			commutativityTable,
			commutativityUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForEqual(*lit2,
					    commutativityTable, 
					    usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->equal(*lit2);
}; 
inline
bool 
TempLitListExtras::complementaryEq(const LitHandle& lit1,
				 const LitHandle& lit2,
				 const SymbolInfo* commutativityTable,
				 BK::Receiver<const Clause*>* usedCommutativityLaws)
{
  CALL("complementaryEq(const LitHandle& lit1,...)");
  ASSERT(lit1->isEquality() && lit2->isEquality());
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->complementaryEqLit(*lit2,
				   commutativityTable,
				   commutativityUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForComplementaryEqLit(*lit2,
							 commutativityTable, 
							 usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->complementaryEqLit(*lit1);
}; 
inline
bool 
TempLitListExtras::equalEq(const LitHandle& lit1,
			   const LitHandle& lit2,
			   const SymbolInfo* commutativityTable,
			   BK::Receiver<const Clause*>* usedCommutativityLaws,
			   bool& eqSymmetryUsed)
{
  CALL("equalEq(const LitHandle& lit1,...)");
  ASSERT(lit1->isEquality() && lit2->isEquality());
  if (commutativityTable)
    {
      bool commutativityUsed;
      if (lit1->equalEqLit(*lit2,
			   commutativityTable,
			   commutativityUsed,
			   eqSymmetryUsed))
	{
	  if (commutativityUsed)
	    lit1->commutativityLawsForEqualEqLit(*lit2,
						 commutativityTable, 
						 usedCommutativityLaws);
	  return true;
	}
      else 
	return false;
    }
  else
    return lit1->equalEqLit(*lit2,eqSymmetryUsed);
}; 
ulong
TempLitListExtras::countDifferentVariables(const BasicTempLitList& lst)
{
  CALL("countDifferentVariables(const BasicTempLitList& lst)");
  if (lst.length() == 1UL)
    {
      return lst.first()->literal()->numOfDifferentVariables();
    }
  else
    {      
      static WeightPolynomial::Monomials vars;
      vars.reset();
      collectVariableOccurences(lst,vars);
      return vars.length();
    };
}; 
void 
TempLitListExtras::collectVariableOccurences(const BasicTempLitList& lst,
					     WeightPolynomial::Monomials& vars)
{
  CALL("collectVariableOccurences(..)");  
  for (const BasicTempLitList::Element* el = lst.first(); el; el = el->next())
    vars.add(el->literal()->variableOccurences());
}; 
bool
TempLitListExtras::containsHeader(const BasicTempLitList& lst,
				  const Term& header)
{
  CALL("containsHeader(..)");
  for (const BasicTempLitList::Element* el = lst.first(); el; el = el->next())
    if (el->literal()->header() == header)
      return true;
  return false;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::Size"
#endif
#include "debugMacros.hpp"
TempLitListExtras::Size::Size() {};
TempLitListExtras::Size::~Size() {};
void* 
TempLitListExtras::Size::operator new(size_t)
{
  CALL("operator new(size_t)");
  return Context::getCurrent()->_sizeAlloc.reserveObject();
}; 
void 
TempLitListExtras::Size::operator delete(void* obj)
{
  CALL("operator delete(void*)");
}; 
void TempLitListExtras::Size::computeFor(const BasicTempLitList* const & lst)
{
  CALL("computeFor(const BasicTempLitList* const & lst)");
  _value = 0;
  for (const BasicTempLitList::Element* el = lst->first(); el; el = el->next())
    _value += el->literal()->size();
}; 
BK::Comparison 
TempLitListExtras::Size::compare(ComparableCharacteristic& c)
{
  CALL("compare(ComparableCharacteristic<BasicTempLitList*>& c)");
  return BK::compare(_value,static_cast<Size&>(c)._value);
}; 
BK::Comparison 
TempLitListExtras::Size::inverseCompare(ComparableCharacteristic& c)
{
  CALL("inverseCompare(ComparableCharacteristic<BasicTempLitList*>& c)");
  return BK::inverseCompare(_value,static_cast<Size&>(c)._value);
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::NumOfVarOccurences"
#endif
#include "debugMacros.hpp"
TempLitListExtras::NumOfVarOccurences::NumOfVarOccurences() {};
TempLitListExtras::NumOfVarOccurences::~NumOfVarOccurences() {};
void* 
TempLitListExtras::NumOfVarOccurences::operator new(size_t)
{
  CALL("operator new(size_t)");
  return Context::getCurrent()->_numOfVarOccurencesAlloc.reserveObject();
}; 
void 
TempLitListExtras::NumOfVarOccurences::operator delete(void* obj)
{
  CALL("operator delete(void*)");
}; 
void TempLitListExtras::NumOfVarOccurences::computeFor(const BasicTempLitList* const & lst)
{
  CALL("computeFor(const BasicTempLitList* const & lst)");
  _value = 0;
  for (const BasicTempLitList::Element* el = lst->first(); el; el = el->next())
    _value += el->literal()->numberOfVariableOccurences();
}; 
BK::Comparison 
TempLitListExtras::NumOfVarOccurences::compare(ComparableCharacteristic& c)
{
  CALL("compare(ComparableCharacteristic<BasicTempLitList*>& c)");
  return BK::compare(_value,static_cast<NumOfVarOccurences&>(c)._value);
}; 
BK::Comparison 
TempLitListExtras::NumOfVarOccurences::inverseCompare(ComparableCharacteristic& c)
{
  CALL("inverseCompare(ComparableCharacteristic<BasicTempLitList*>& c)");
  return BK::inverseCompare(_value,static_cast<NumOfVarOccurences&>(c)._value);
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::NumOfLiterals"
#endif
#include "debugMacros.hpp"
TempLitListExtras::NumOfLiterals::NumOfLiterals() {};
TempLitListExtras::NumOfLiterals::~NumOfLiterals() {};
void* 
TempLitListExtras::NumOfLiterals::operator new(size_t)
{
  CALL("operator new(size_t)");
  return Context::getCurrent()->_numOfLiteralsAlloc.reserveObject();
}; 
void 
TempLitListExtras::NumOfLiterals::operator delete(void* obj)
{
  CALL("operator delete(void*)");
}; 
void TempLitListExtras::NumOfLiterals::computeFor(const BasicTempLitList* const & lst)
{
  CALL("computeFor(const BasicTempLitList* const & lst)");
  _value = lst->length();
}; 
BK::Comparison 
TempLitListExtras::NumOfLiterals::compare(ComparableCharacteristic& c)
{
  CALL("compare(ComparableCharacteristic<BasicTempLitList*>& c)");
  return BK::compare(_value,static_cast<NumOfLiterals&>(c)._value);
}; 
BK::Comparison 
TempLitListExtras::NumOfLiterals::inverseCompare(ComparableCharacteristic& c)
{
  CALL("inverseCompare(ComparableCharacteristic<BasicTempLitList*>& c)");
  return BK::inverseCompare(_value,static_cast<NumOfLiterals&>(c)._value);
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::CharacteristicComparisonMaker"
#endif
#include "debugMacros.hpp"
void* 
TempLitListExtras::CharacteristicComparisonMaker::operator new(size_t)
{
  CALL("operator new(size_t)");
  return Context::getCurrent()->_charactCompMakerAlloc.reserveObject();
}; 
void 
TempLitListExtras::CharacteristicComparisonMaker::operator delete(void* obj)
{
  CALL("operator delete(void*)");
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::ComplexComparisonMaker"
#endif
#include "debugMacros.hpp"
void* 
TempLitListExtras::ComplexComparisonMaker::operator new(size_t)
{
  CALL("operator new(size_t)");
  return Context::getCurrent()->_complexCompMakerAlloc.reserveObject();
}; 
void 
TempLitListExtras::ComplexComparisonMaker::operator delete(void* obj)
{
  CALL("operator delete(void*)");
}; 
