/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TEMPORARY_CLAUSE_HPP
#define TEMPORARY_CLAUSE_HPP
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "GlobAlloc.hpp"
#include "ObjectPool.hpp"
#include "ObjectPoolWithMemory.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "ClauseBackground.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause"
#endif
#include "debugMacros.hpp"
namespace VK
{  
  class TemporaryClause
  {
  public:
    typedef BK::ObjectPool<BasicTempLitList>::EntryHandle LitListHandle;
    class Context;
    class Extras;
  public:
    TemporaryClause();
    ~TemporaryClause();
    void hibernate() { _literals.clear(); };
    void reset();
    const LitListHandle& literals() const { return _literals; };
    LitListHandle& literals() { return _literals; };
    const OpenClauseBackground& background() const { return _background; };
    OpenClauseBackground& background() { return _background; };
    void resetSimplificationFlags();
    void resetSimplificationFlagsInPropIntroPart(const TemporaryClause& parentCl);
    bool inputNonBasicPositionsMarked() const { return _inputNonBasicPositionsMarked; };
    void markInputNonBasicPositions();
    bool weeded() const { return _weeded; };
    bool weededModuloCommutativity() const 
    { 
      return _weededModuloCommutativity; 
    };
    bool testedForSubsumption() const { return _testedForSubsumption; };
    bool simplifiedBySubsumptionResolution() const { return _simplifiedBySubsumptionResolution; };
    bool normalisedByCommutativity() const { return _normalisedByCommutativity; };
    bool testedForReductionOfNonBasicByCommutativity() const
    {
      return _testedForReductionOfNonBasicByCommutativity;
    };
    bool normalisedByDemodulation() const { return _normalisedByDemodulation; };
    bool testedForReductionOfNonBasic() const { return _testedForReductionOfNonBasic; };
    bool simplifiedByExternalEvaluation() const { return _simplifiedByExternalEvaluation; };
    bool simplifiedByDestructiveEqualityResolution() const { return _simplifiedByDestructiveEqualityResolution; };
    bool lockedForPropositionalSplitting() const { return _lockedForPropositionalSplitting; };
    bool lockedForNegativeEqualitySplitting() const { return _lockedForNegativeEqualitySplitting; };
    void markAsWeeded() { _weeded = true; };
    void markAsWeededModuloCommutativity() 
    {
      _weeded = true;
      _weededModuloCommutativity = true; 
    };
    void markAsTestedForSubsumption() { _testedForSubsumption = true; };
    void markAsSimplifiedBySubsumptionResolution() { _simplifiedBySubsumptionResolution = true; };
    void markAsNormalisedByCommutativity() { _normalisedByCommutativity = true; };
    void markAsTestedForReductionOfNonBasicByCommutativity()
    {
      _testedForReductionOfNonBasicByCommutativity = true;
    };
    void markAsNormalisedByDemodulation() {  _normalisedByDemodulation = true; };
    void markAsTestedForReductionOfNonBasic() { _testedForReductionOfNonBasic = true; };
    void markAsSimplifiedByExternalEvaluation() { _simplifiedByExternalEvaluation = true; };
    void markAsSimplifiedByDestructiveEqualityResolution() { _simplifiedByDestructiveEqualityResolution = true; };
    void markAsLockedForPropositionalSplitting() { _lockedForPropositionalSplitting = true; };
    void markAsLockedForNegativeEqualitySplitting() { _lockedForNegativeEqualitySplitting = true; };
    void registerIntraLiteralModification()
    {
      CALL("registerIntraLiteralModification()");  
      _background.CancelRule(ClauseBackground::PartOfNameIntroForPropSplitting);
      _inputNonBasicPositionsMarked = false;
      _weeded = false;
      _weededModuloCommutativity = false;
      _testedForSubsumption = false;
      _simplifiedBySubsumptionResolution = false;
      _normalisedByCommutativity = false;
      _normalisedByDemodulation = false;
      _simplifiedByExternalEvaluation = false;
      _simplifiedByDestructiveEqualityResolution = false;
      _lockedForPropositionalSplitting = false;
      _lockedForNegativeEqualitySplitting = false;
    };
    void markAsPropSplittingBase()
    {
      CALL("markAsPropSplittingBase()");
      ASSERT(!isPartOfNameIntroForPropSplitting());
      _weeded = false;
      _testedForSubsumption = false;
      _simplifiedBySubsumptionResolution = false;
      _simplifiedByDestructiveEqualityResolution = false; 
      _lockedForPropositionalSplitting = true;
    };
    bool isAnswer() const
    {
      CALL("isAnswer() const");
      return 
	TempLitListExtras::containsOnlyAnswerLiterals(*_literals);
    };
    bool isNonSubsumable() const
    {
      CALL("isNonSubsumable() const");
      return _background.isNonSubsumable();
    };
    bool isPartOfNameIntroForPropSplitting() const {
      return _background.isPartOfNameIntroForPropSplitting();
    };
    ostream& outputAnnounced(ostream& str) const;
    ostream& outputComplete(ostream& str) const;
  private:
    ulong _universalId;
    LitListHandle _literals;
    OpenClauseBackground _background;
    bool _inputNonBasicPositionsMarked;
    bool _weeded;
    bool _weededModuloCommutativity;
    bool _testedForSubsumption;
    bool _simplifiedBySubsumptionResolution;
    bool _normalisedByCommutativity;
    bool _testedForReductionOfNonBasicByCommutativity;
    bool _normalisedByDemodulation;
    bool _testedForReductionOfNonBasic;
    bool _simplifiedByExternalEvaluation;
    bool _simplifiedByDestructiveEqualityResolution;
    bool _lockedForPropositionalSplitting;
    bool _lockedForNegativeEqualitySplitting;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "<global>"
#endif
#include "debugMacros.hpp"
namespace BK
{
  template <>
  inline
  static
  void hibernate(VK::TemporaryClause& obj)
  {
    CALL("hibernate(VK::TemporaryClause& obj)");
    obj.hibernate();
  };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  void TemporaryClause::resetSimplificationFlags()
  {
    CALL("resetSimplificationFlags()");
    _inputNonBasicPositionsMarked = false;
    _weeded = false;
    _weededModuloCommutativity = false;
    _testedForSubsumption = false;
    _simplifiedBySubsumptionResolution = false;
    _normalisedByCommutativity = false;
    _testedForReductionOfNonBasicByCommutativity = false;
    _normalisedByDemodulation = false;
    _testedForReductionOfNonBasic = false;
    _simplifiedByExternalEvaluation = false;
    _simplifiedByDestructiveEqualityResolution = false;
    _lockedForPropositionalSplitting = 
      isPartOfNameIntroForPropSplitting();
    _lockedForNegativeEqualitySplitting = false;
  }; 
  inline
  void 
  TemporaryClause::resetSimplificationFlagsInPropIntroPart(const TemporaryClause& parentCl)
  {
    CALL("resetSimplificationFlagsInPropIntroPart(..)");
    ASSERT(isPartOfNameIntroForPropSplitting());
    _inputNonBasicPositionsMarked = parentCl._inputNonBasicPositionsMarked;
    _weeded = false;
    _weededModuloCommutativity = parentCl._weededModuloCommutativity;
    _testedForSubsumption = parentCl._testedForSubsumption;
    _simplifiedBySubsumptionResolution = parentCl._simplifiedBySubsumptionResolution;
    _normalisedByCommutativity = parentCl._normalisedByCommutativity;
    _testedForReductionOfNonBasicByCommutativity = parentCl._testedForReductionOfNonBasicByCommutativity;
    _normalisedByDemodulation = parentCl._normalisedByDemodulation;
    _testedForReductionOfNonBasic = parentCl._testedForReductionOfNonBasic;
    _simplifiedByExternalEvaluation = parentCl._simplifiedByExternalEvaluation;
    _simplifiedByDestructiveEqualityResolution = false;
    _lockedForPropositionalSplitting = true;
    _lockedForNegativeEqualitySplitting = parentCl._lockedForNegativeEqualitySplitting;
  }; 
  inline
  void TemporaryClause::markInputNonBasicPositions()
  {
    CALL("markInputNonBasicPositions()");
    if (!_inputNonBasicPositionsMarked)
      {
	for (BasicTempLitList::Element* el = _literals->first();
	     el;
	     el = el->next())
	  el->literal()->markInputNonBasicPositions();
	_inputNonBasicPositionsMarked = true;
      };
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause::Context"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TemporaryClause::Context
  {
  public:
    class AutoActivator
    {
    public:
      AutoActivator(Context* c) : _previous(Context::getCurrent())
      {
	Context::setCurrent(c);
      };
      ~AutoActivator()
      {
	Context::setCurrent(_previous);
      };
    private:
      Context* _previous;
    }; 
  public:
    Context() : _previousUniversalId(0UL) {};
    ~Context() {};
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
    ulong generateUniversalId() 
    {
      ++_previousUniversalId;
      return _previousUniversalId;
    };
    BK::ObjectPool<TemporaryClause>* defaultPool()
    {
      return static_cast<BK::ObjectPool<TemporaryClause>*>(&_defaultPool);
    };
  private:
    ulong _previousUniversalId;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,TemporaryClause> _defaultPool;
    static Context* _current;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause::Extras"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TemporaryClause::Extras
  {    
  public:
    typedef BK::ObjectPool<TemporaryClause> Pool;
    typedef BK::ObjectPool<TemporaryClause>::EntryHandle Handle;
  public:
    static void freshClause(Handle& h)
    {
      CALL("freshClause(Handle& h)");
      ASSERT(Context::getCurrent());
      Context::getCurrent()->defaultPool()->reserveObject(h);
      h->reset();
    };
  }; 
}; 
#endif
