/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "TemporaryLiteral.inl" 
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_LITERAL
#  define DEBUG_NAMESPACE "TemporaryLiteral"
#endif
#include "debugMacros.hpp"
TemporaryLiteral::Context* 
TemporaryLiteral::Context::_current; 
ulong TemporaryLiteral::minPosEqLHSSize() const
{
  CALL("minPosEqLHSSize() const");
  ASSERT(_debugLiteralComplete);
  ASSERT(isEquality() && isPositive());
  if (isOrderedEquality())
    {
      return arg2()->size();
    }
  else 
    {
      long w1 = arg1()->size();
      long w2 = arg2()->size();
      return (w1 > w2) ? w2 : w1;
    }; 
}; 
bool TemporaryLiteral::pushTerm(const Flatterm* term)
{
  CALL("pushTerm(const Flatterm* term)");
  ASSERT(!_debugLiteralComplete);
  Flatterm::ConstIterator iter(term); 
  do
    {
      if (!pushSym(iter.CurrPos()->symbolRef(),
		   iter.CurrPos()->isSubstitutionPosition())) 
	return false;
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());
  return true;
};  
bool TemporaryLiteral::pushBinaryLit(const Term& header,
				     const Flatterm* arg1,
				     const Flatterm* arg2,
				     const LiteralProperties& prop)
{
  CALL("pushBinaryLit(const Term& header,..)");
  ASSERT(header.arity() == 2);
  _openFlatterm.Reset();
  _size = 3UL; 
  _weightCollected = false;
  _variableOccurencesCollected = false;
  _resistanceComputed = false;
  DOP(_debugLiteralComplete = false);
  DOP(_arg1 = 0);
  DOP(_arg2 = 0);
  _openFlatterm.PushNonconstFun(header); 
  _header = header;
  _properties = prop;
  Flatterm::ConstIterator iter(arg1); 
  do
    {
      if (!pushSym(iter.CurrPos()->symbolRef(),false)) return false;
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());
  iter.Reset(arg2);
  do
    {
      if (!pushSym(iter.CurrPos()->symbolRef(),false)) return false;
      iter.NextSym();
    }
  while (iter.NotEndOfTerm()); 
  endOfLiteral();
  return true;
}; 
bool TemporaryLiteral::instantiateVar(TemporaryLiteral& new_lit,
				      ulong var,
				      const Flatterm* subst_term,
				      bool unmarkSubstPositionsAbove)
{
  CALL("instantiateVar(..)");
  Term new_header = header();
  ASSERT(new_header.arity());
  LiteralProperties new_prop(properties());
  if(new_header.IsEquality())
    {
      if (new_header.IsOrderedEq())
	{
	  new_header.MakeUnordered();
	  if (argumentsSwapped()) 
	    new_prop.reverseEqualityOrientationMode();
	};
    };
  new_lit.pushNonpropHeader(new_header,new_prop);
  Flatterm::Iterator iter(flatterm());
  Flatterm::ConstIterator subst_iter;
  iter.NextSym(); 
  while (iter.NotEndOfTerm())
    {
      Term sym = iter.CurrPos()->symbolRef();  
      if (sym.isVariable() && sym.var() == var) 
	{
	  subst_iter.Reset(subst_term);             
	  do
	    {
	      if (!new_lit.pushSym(subst_iter.CurrPos()->symbolRef(),
				   subst_iter.CurrPos()->isSubstitutionPosition())) 
		return false;             
	      subst_iter.NextSym();
	    }
	  while (subst_iter.NotEndOfTerm());
	}
      else 
	if (!new_lit.pushSym(sym,iter.CurrPos()->isSubstitutionPosition())) 
	  return false;
      iter.NextSym();
    };
  new_lit.endOfLiteral();
  if (unmarkSubstPositionsAbove)
    {
      iter.Reset(flatterm());
      Flatterm::Iterator newLitIter(new_lit.flatterm());
      iter.NextSym();
      newLitIter.NextSym();
      while (iter.NotEndOfTerm())
	{
	  ASSERT(newLitIter.NotEndOfTerm());
	  Term sym = iter.CurrPos()->symbolRef();  
	  if (sym.isVariable() && sym.var() == var) 
	    {
	      new_lit.flatterm()->unmarkSubstitutionPositionsAbove(newLitIter.CurrPos());
	      iter.NextSym();
	      newLitIter.after();
	    }
	  else
	    {
	      iter.NextSym();
              newLitIter.NextSym();
	    };
	};
    };
  return true;
}; 
void TemporaryLiteral::renameVariables(VariableRenaming& var_renaming)
{
  CALL("renameVariables(VariableRenaming& var_renaming)");
  if (isEquality())
    {
      arg1()->renameVariables(var_renaming);
      arg2()->renameVariables(var_renaming);
    }
  else 
    flatterm()->renameVariables(var_renaming);
};
BK::Comparison 
TemporaryLiteral::compareWithVarWeighting(TemporaryLiteral& lit,
                                          const VariableWeighting& var_weighting)
{
  CALL("compareWithVarWeighting(TemporaryLiteral& lit,const VariableWeighting& var_weighting)");
  ASSERT(_debugLiteralComplete);
  ASSERT(lit._debugLiteralComplete);
  BK::Comparison hd_cmp = 
    Signature::current()->compareHeaderRanks(header().polarity(),
					     header().functor(),
					     lit.header().polarity(),
					     lit.header().functor());
  if (hd_cmp != BK::Equal) return hd_cmp; 
  if (size() > lit.size()) return BK::Greater; 
  if (size() < lit.size()) return BK::Less; 
  BK::Comparison var_cmp;
  const Flatterm* fst = flatterm() + 1;
  const Flatterm* const end_of_fst = flatterm()->after();
  const Flatterm* snd = lit.flatterm();
 check_end:
  if (fst == end_of_fst) return BK::Incomparable;
  snd++;
  if (fst->symbolRef() != snd->symbolRef())
    {
      if (fst->symbolRef().isVariable()) 
	{
	  if (snd->symbolRef().isVariable()) 
	    {
	      var_cmp = var_weighting.Compare(fst->symbolRef().var(),snd->symbolRef().var());
	      if (var_cmp != BK::Equal) { return var_cmp; };
	    }
	  else 
	    return BK::Less; 
	}
      else 
	{
	  if (snd->symbolRef().isVariable()) return BK::Greater;
	  return fst->symbolRef().comparePrecedence(snd->symbolRef()); 
	}; 
    };
  fst++;
  goto check_end;
}; 
void TemporaryLiteral::collectVarDepthProfile(VarDepthProfile& profile)
{
  CALL("collectVarDepthProfile(VarDepthProfile& profile)");
  profile.reset();    
  if (_weightCollected)
    {
      if (_weight.monomials().isZero()) return;
    }
  else
    if (_variableOccurencesCollected)
      {
	if (_variableOccurences.isZero()) return; 
      };
  Flatterm::Iterator iter(flatterm());
  SmartDepthCounter<VampireKernelConst::MaxTermDepth> depth_counter;
  depth_counter.Reset(1UL); 
  while (iter.NotEndOfTerm())
    {
      depth_counter.NextSym(iter.CurrPos()->symbolRef());  
      if (iter.CurrPos()->isVariable())
	{ 
	  profile.add(depth_counter.Depth());      
	};
      iter.NextSym();
    };
}; 
ostream& TemporaryLiteral::output(ostream& str) const
{
  if (isEquality())
    { 
      str << '(';
      arg1()->outputTerm(str) << ')' << header() << '(';
      arg2()->outputTerm(str);
      str << ')';
    }
  else 
    flatterm()->outputTerm(str);
  if (Context::getCurrent()->_showLiteralProperties &&
      !_properties.isDefault())
    {
      str << " <@[" << _properties << "]";
    };
  return str; 
}; 
float 
TemporaryLiteral::computeResistance(ulong depthCoeff,
				    float constCoeff,
				    float varCoeff) const
{
  CALL("computeResistance(ulong depthCoeff,float constCoeff,float varCoeff) const");
  if (isEquality())
    {
      if (isOrderedEquality())
	{
	  return arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	}
      else 
	{
	  if (isNegative())
	    {
	      if (arg1()->isVariable())
		{
		  return arg2()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
		};
	      if (arg2()->isVariable())
		{
		  return arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
		};
	      float res1 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      float res2 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      return (res1 < res2) ? res1 : res2;
	    }
	  else
	    {
	      float res1 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      float res2 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      return (res1 < res2) ? res1 : res2;
	    };
	};
    }
  else
    return flatterm()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
}; 
