/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TEMPORARY_LITERAL_HPP
#define TEMPORARY_LITERAL_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Comparison.hpp"
#include "DestructionMode.hpp"
#include "Multiset.hpp"
#include "ObjectPool.hpp"
#include "ObjectPoolWithMemory.hpp"
#include "Receiver.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Flatterm.hpp"
#include "Term.hpp"
#include "WeightPolynomial.hpp"
#include "TermWeighting.hpp"
#include "SimplificationOrdering.hpp"
#include "Signature.hpp"
#include "LiteralProperties.hpp"
#include "VariableRenaming.hpp"
#include "VariableWeighting.hpp"
namespace VK
{
  class SymbolInfo;
  class Clause;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_LITERAL
#  define DEBUG_NAMESPACE "TemporaryLiteral"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TemporaryLiteral
  {
  public:
    typedef BK::MultisetWithMinMax<ulong,VampireKernelConst::MaxTermDepth + 1> VarDepthProfile;
  public:
    class Extras;
    class Context;
  public:
    TemporaryLiteral() 
    { 
    };    
    ~TemporaryLiteral() {};
    bool isAnswerLiteral() const { return header().isAnswerHeader(); };
    const Flatterm* flatterm() const 
    {
      CALL("flatterm() const");
      ASSERT(_debugLiteralComplete);
      return _openFlatterm.Term(); 
    };
    Flatterm* flatterm() 
    {
      CALL("flatterm()");
      ASSERT(_debugLiteralComplete);
      return _openFlatterm.Term(); 
    };
    const Term& header() const { return _header; };
    ulong arity() const { return _header.arity(); };
    const LiteralProperties& properties() const { return _properties; };
    LiteralProperties& properties() { return _properties; }
    bool isSelected() const { return _properties.isSelected(); };
    void markAsSelected() { _properties.makeSelected(); };
    void markAsUnselected() { _properties.makeUnselected(); };
    unsigned long depth() { CALL("depth()"); return flatterm()->Depth(); };
    ulong size() const { return _size; };
    ulong weightBase() const 
    {
      return (isAnswerLiteral()) ? 0UL : size(); 
    };
    const WeightPolynomial& weight();
    const WeightPolynomial::Monomials& variableOccurences();
    long numberOfVariableOccurences();
    void variables(const ulong*& begin,const ulong*& end);
    ulong numOfDifferentVariables();
    bool isGround() { CALL("isGround()"); return numOfDifferentVariables() == 0UL; };
    bool containsVariable(ulong var);
    const WeightPolynomial& arg1Weight();
    const WeightPolynomial& arg2Weight();
    bool arg1ContainsVariable(ulong var);
    bool arg2ContainsVariable(ulong var);
    bool arg1IsGround();
    bool arg2IsGround();
    ulong numOfDifferentVariablesInArg1();
    ulong numOfDifferentVariablesInArg2();
    const float& resistance(ulong depthCoeff,float constCoeff,float varCoeff)
    {
      if (!_resistanceComputed ||
	  _resistanceDepthCoeff != depthCoeff ||
	  _resistanceConstCoeff != constCoeff ||
	  _resistanceVarCoeff != varCoeff)
	{
	  _resistanceComputed = true;
	  _resistance = computeResistance(depthCoeff,constCoeff,varCoeff);
	  _resistanceDepthCoeff = depthCoeff;
          _resistanceConstCoeff = constCoeff;
	  _resistanceVarCoeff = varCoeff;
	};
      return _resistance;
    };
    bool isEquality() const { return _header.IsEquality(); };
    bool isOrderedEquality() const { return _header.IsOrderedEq(); };
    bool isUnorderedEquality() const { return _header.IsUnorderedEq(); };
    bool isPreorientedEquality() const {
      return _properties.isPreorientedEq();
    };
    bool isOrientableEquality() const {
      return _properties.isOrientableEq();
    };
    bool argumentsSwapped() const { return _arg1 == rawArg2(); };
    bool isSplittingBranch() const { return _header.IsSplitting(); };
    bool isPositive() const { return _header.Positive(); };
    bool isNegative() const { return _header.Negative(); }; 
    Term::Polarity polarity() const { return _header.polarity(); };
    void flipPolarity() {
      _header.flipPolarity();
      flatterm()->symbolRef().flipPolarity();
    };
    bool isPositiveForSelection() const;
    bool isNegativeForSelection() const;
    const Flatterm* arg1() const;
    Flatterm* arg1(); 
    const Flatterm* arg2() const; 
    Flatterm* arg2();
    const Flatterm* equalityLHS() const;
    const Flatterm* equalityRHS() const;
    void linkEqArgs();
    void makeEqUnordered();
    ulong minPosEqLHSSize() const;
    void pushVar(const ulong& v);
    void pushVar(const ulong& v,bool substPos);
    void pushVar(const Term& v);
    void pushVar(const Term& v,bool substPos);
    void pushConst(const Term& c);
    void pushConst(const Term& c,bool substPos);
    bool pushNonconstFun(const Term& f);
    bool pushNonconstFun(const Term& f,bool substPos);
    bool pushSym(const Term& sym);
    bool pushSym(const Term& sym,bool substPos);
    bool pushTerm(const Flatterm* term); 
    void pushNonpropHeader(const Term& hd,
			   const LiteralProperties& prop);
    void pushPropLit(const Term& lit,
		     const LiteralProperties& prop);
    bool pushPseudoLit(const Term& header,
		       const Flatterm* arg1,
		       const WeightPolynomial::Monomials& varsForPadding,
		       const LiteralProperties& prop);
    bool pushBinaryLit(const Term& header,
		       const Flatterm* arg1,
		       const Flatterm* arg2,
		       const LiteralProperties& prop);
    void endOfLiteral();
    bool argsAreEqual() const;
    bool argsAreEqual(const SymbolInfo* commutativityTable,
		      bool& commutativityUsed) const;
    void commutativityLawsForArgsAreEqual(const SymbolInfo* commutativityTable,
					  BK::Receiver<const Clause*>* laws) const;
    bool equal(const TemporaryLiteral& lit) const;
    bool equal(const TemporaryLiteral& lit,
	       const SymbolInfo* commutativityTable,
	       bool& commutativityUsed) const;
    void commutativityLawsForEqual(const TemporaryLiteral& lit,
				   const SymbolInfo* commutativityTable,
				   BK::Receiver<const Clause*>* laws) const; 
    bool isComplementary(const TemporaryLiteral& lit) const;
    bool isComplementary(const TemporaryLiteral& lit,
			 const SymbolInfo* commutativityTable,
			 bool& commutativityUsed) const;
    void commutativityLawsForComplementary(const TemporaryLiteral& lit,
					   const SymbolInfo* commutativityTable,
					   BK::Receiver<const Clause*>* laws) const;
    bool complementaryEqLit(const TemporaryLiteral& lit) const;
    bool complementaryEqLit(const TemporaryLiteral& lit,
			    const SymbolInfo* commutativityTable,
			    bool& commutativityUsed) const;  
    void commutativityLawsForComplementaryEqLit(const TemporaryLiteral& lit,
						const SymbolInfo* commutativityTable,
						BK::Receiver<const Clause*>* laws) const;
    bool equalEqLit(const TemporaryLiteral& lit,
		    bool& eqSymmetryUsed) const;
    bool equalEqLit(const TemporaryLiteral& lit,
		    const SymbolInfo* commutativityTable,
		    bool& commutativityUsed,
		    bool& eqSymmetryUsed) const;
    void commutativityLawsForEqualEqLit(const TemporaryLiteral& lit,
					const SymbolInfo* commutativityTable,
					BK::Receiver<const Clause*>* laws) const;
    bool instantiationPossible(ulong var,
			       ulong substDepth,
			       long substSize,
			       bool checkInputNonBasicness);
    bool instantiateVar(TemporaryLiteral& new_lit,
			ulong var,
			const Flatterm* subst_term,
			bool unmarkSubstPositionsAbove);
    void renameVariables(VariableRenaming& var_renaming);
    BK::Comparison compareWithVarWeighting(TemporaryLiteral& lit,const VariableWeighting& var_weighting);
    void orderEqArgs();
    void collectVarDepthProfile(VarDepthProfile& profile);
    ulong markInputNonBasicPositions();
    void unmarkAllSubstitutionPositions();
    ostream& output(ostream& str) const;
  private:
    TemporaryLiteral& operator=(const TemporaryLiteral&);
    TemporaryLiteral(const TemporaryLiteral&);
    const Flatterm* rawArg1() const;  
    Flatterm* rawArg1();
    const Flatterm* rawArg2() const; 
    Flatterm* rawArg2();   
    void swapArguments();
    void collectWeight();
    void collectVariableOccurences();
    void collectArg1Weight();
    void collectArg2Weight();
    float computeResistance(ulong depthCoeff,float constCoeff,float varCoeff) const;
  private:
    Term _header;
    Flatterm::OpenWithMemory _openFlatterm;
    LiteralProperties _properties;
    ulong _size;
    bool _weightCollected;
    WeightPolynomial _weight;
    bool _variableOccurencesCollected;
    WeightPolynomial::Monomials _variableOccurences;
    bool _resistanceComputed;
    float _resistance;
    ulong _resistanceDepthCoeff;
    float _resistanceConstCoeff;
    float _resistanceVarCoeff;
    Flatterm* _arg1;
    Flatterm* _arg2;
    WeightPolynomial* _arg1Weight;
    WeightPolynomial* _arg2Weight;
    WeightPolynomial _rawArg1Weight;
    WeightPolynomial _rawArg2Weight;
#ifdef DEBUG_NAMESPACE
    bool _debugLiteralComplete;
#endif
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_LITERAL
#  define DEBUG_NAMESPACE "TemporaryLiteral::Context"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TemporaryLiteral::Context
  {
  public:
    class AutoActivator
      {
      public:
	AutoActivator(Context* c) : _previous(Context::getCurrent())
	{
	  Context::setCurrent(c);
	};
	~AutoActivator()
	{
	  Context::setCurrent(_previous);
	};
      private:
	Context* _previous;
    }; 
  public:
    Context() : 
      _inversePolarityForSelection(false), 
      _chainingTable(0), 
      _showLiteralProperties(false),
      _allPosEqMode(LiteralProperties::DefaultEqualityMode),
      _allNegEqMode(LiteralProperties::DefaultEqualityMode)
    {
    };
    ~Context() {};
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
    void setChainingTable(const SymbolInfo* t) { _chainingTable = t; };
    void setInversePolarityForSelection(bool fl) 
    {
      _inversePolarityForSelection = fl;
    };
    BK::ObjectPool<TemporaryLiteral>* defaultPool()
    {
      return static_cast<BK::ObjectPool<TemporaryLiteral>*>(&_defaultPool);
    };
    void setShowLiteralProperties(bool fl) { _showLiteralProperties = fl; };
    void setAllPosEqMode(const LiteralProperties::EqualityMode& m)
    {
      _allPosEqMode = m;
    };
    void setAllNegEqMode(const LiteralProperties::EqualityMode& m)
    {
      _allNegEqMode = m;
    };
    const LiteralProperties::EqualityMode& allPosEqMode() const
    {
      return _allPosEqMode;
    };
    const LiteralProperties::EqualityMode& allNegEqMode() const
    {
      return _allNegEqMode;
    };
  private:
    bool _inversePolarityForSelection;
    const SymbolInfo* _chainingTable;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,TemporaryLiteral> _defaultPool;
    bool _showLiteralProperties;
    LiteralProperties::EqualityMode _allPosEqMode;
    LiteralProperties::EqualityMode _allNegEqMode;
    static Context* _current;
    friend class TemporaryLiteral;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_LITERAL
#  define DEBUG_NAMESPACE "TemporaryLiteral::Extras"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TemporaryLiteral::Extras
  {
  public:
    typedef BK::ObjectPool<TemporaryLiteral> Pool;
    typedef BK::ObjectPool<TemporaryLiteral>::EntryHandle Handle;
  public:    
    static void freshLiteral(Handle& h)
    {
      CALL("freshLiteral(Handle& h)");
      Context::getCurrent()->defaultPool()->reserveObject(h);
    };
    static bool tryToGetFreshLiteral(Handle& h)
    {
      CALL("tryToGetFreshLiteral(Handle& h)");
      return 
	Context::getCurrent()->defaultPool()->tryToReserveObject(h);
    };
  }; 
}; 
#endif
