/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "TermSharing.hpp" 
#include "Flatterm.hpp"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM_SHARING
#  define DEBUG_NAMESPACE "TermSharing"
#endif
#include "debugMacros.hpp"
TermSharing::TermSharing() 
  : _iter(), 
  _subterms("_subterms"),
  _discTree(static_cast<IndexNode*>(0),"TermSharing::_discTree"),
  _sharedConst(static_cast<Term*>(0),"TermSharing::_sharedConst")
{
  CALL("constructor TermSharing()");
  DOP(_discTree.freeze());
  DOP(_sharedConst.freeze());
}; 
TermSharing::~TermSharing() 
{ 
  CALL("destructor ~TermSharing()"); 
  if (DestructionMode::isThorough())
    {    
      for (ulong f = 0; f < _discTree.size(); f++)
	{
	  DestroyTree(_discTree[f]);
	};
      for (ulong n = 0; n < _sharedConst.size(); n++)
	if (_sharedConst[n])
	  {
	    _sharedConst[n]->DestroyOccurences();
	    delete _sharedConst[n];
	  };
    };
}; 
void TermSharing::reset() 
{
  CALL("reset()");    
  for (ulong f = 0; f < _discTree.size(); f++)
    {
      DestroyTree(_discTree[f]);
      _discTree[f] = 0;
    };
  for (ulong n = 0; n < _sharedConst.size(); n++)
    if (_sharedConst[n])
      _sharedConst[n]->DestroyOccurences();
}; 
inline Term* TermSharing::IntegrateComplexNonconst(Term* t)
{
  CALL("IntegrateComplexNonconst(Term* t)");
  ASSERT(t->arity());
  Term::Iterator iter;
  iter.Reset(t);
  Integrator surf(this);
  IndexNode* leaf;
  Term sym; 
  if (surf.Reset(*t))
    {
    read_mode_next_sym:
      if (iter.Next())
	{
	  sym = iter.symbolRef();
	  if (surf.Read(sym)) { goto read_mode_next_sym; }
	  else
	    {
	      surf.Write(sym);
	      goto write_mode_next_sym;
	    };
	}
      else 
	{
	  leaf = surf.Leaf();
	  ASSERT(leaf);
	  ASSERT(leaf->IsLeaf());
	  leaf->incEntryMultiplicity();
	  return leaf->term();
	}; 
    }
  else
    {
    write_mode_next_sym:
      if (iter.Next())
	{
	  surf.Write(iter.symbolRef());
	  goto write_mode_next_sym;
	}
      else 
	{
	  Term* copy = new (t->arity()) Term(*t);
	  copy->InitOccurences();
	  Term* src = t->Args();
	  Term* targ = copy->Args();
	check_arg:
	  switch (src->typeTag())
	    {
	    case TermVar:
	      *targ = *src;
	      src = src->Next();
	      targ = targ->Next();
	      goto check_arg;
	    case TermRef:
	      if (src->deref()->arity())
		{
		  *targ = *src;
		  _subterms.push(targ);
		}
	      else 
		{
		  targ->makeReference(Const(*(src->deref())));
		};
	      src = src->Next();
	      targ = targ->Next();
	      goto check_arg;
	    case TermComplex:
	      surf.CreateLeaf(copy);
	      return copy;
	    };
	};
    };
  return 0; 
}; 
Term* TermSharing::IntegrateComplex(Term* t)
{
  if (t->arity())
    {
      _subterms.reset();
      Term* result = IntegrateComplexNonconst(t);
      Term* subt;
      while (!_subterms.empty())
	{
	  subt = _subterms.pop();
	  subt->makeReference(IntegrateComplexNonconst(subt->deref()));
	};
      return result;
    }
  else 
    {
      return Const(*t);
    };
}; 
Term* TermSharing::CopyLiteral(Term* lit)
{
  CALL("CopyLiteral(Term* lit)");
  Term* target = new (lit->arity()) Term(*lit);
  lit = lit->Args();
  Term* targ = target->Args();
 copy_arg:
  switch (lit->typeTag())
    {
    case TermVar:
      *targ = *lit;
      lit = lit->Next();
      targ = targ->Next();
      goto copy_arg;
    case TermRef:
      targ->makeReference(IntegrateComplex(lit->deref()));
      lit = lit->Next();
      targ = targ->Next();
      goto copy_arg;
    case TermComplex:
      return target;
    };
  return 0; 
}; 
Term* TermSharing::IntegrateFlatComplex(const Flatterm* t)
{
  CALL("IntegrateFlatComplex(const Flatterm* t)");
  if (t->symbolRef().arity())
    {
      return IntegrateFlatComplexNonconst(t);
    }
  else 
    {
      return Const(t->symbolRef());
    };
}; 
inline 
Term* 
TermSharing::IntegrateFlatComplexNonconst(const Flatterm* t)
{
  CALL("IntegrateFlatComplexNonconst(const Flatterm* t)");
  ASSERT(t->symbolRef().arity());
  Flatterm::ConstIterator iter;
  iter.Reset(t);
  Integrator surf(this);
  IndexNode* leaf;
  Term sym;
  if (surf.Reset(t->symbolRef()))
    {
    read_mode_next_sym:
      iter.NextSym();
      if (iter.EndOfTerm()) 
	{
	  leaf = surf.Leaf();
	  ASSERT(leaf);
	  ASSERT(leaf->IsLeaf());
	  leaf->incEntryMultiplicity();
	  return leaf->term();
	}; 
      sym = iter.CurrPos()->symbolRef();
      if (surf.Read(sym)) { goto read_mode_next_sym; };
      surf.Write(sym);
      goto write_mode_next_sym;
    }
  else
    {
    write_mode_next_sym:
      iter.NextSym();
      if (iter.EndOfTerm()) 
	{
	  Term* copy = new (t->symbolRef().arity()) Term(t->symbolRef());
	  copy->InitOccurences();
	  surf.CreateLeaf(copy);
	  Term* targ = copy->Args();
	  iter.Reset(t);
	  iter.NextSym();
	copy_argument:
	  if (iter.CurrPos()->symbolRef().isVariable())
	    {
	      *targ = iter.CurrPos()->symbolRef();
	    }
	  else 
	    {
	      if (iter.CurrPos()->symbolRef().arity())
		{
		  targ->makeReference(IntegrateFlatComplexNonconst(iter.CurrPos()));
		}
	      else 
		{
		  targ->makeReference(Const(iter.CurrPos()->symbolRef())); 
		}; 
	    }; 
	  iter.after();
	  if (iter.EndOfTerm())
	    {
	      return copy;
	    };
	  targ = targ->Next(); 
	  goto copy_argument;
	};
      surf.Write(iter.CurrPos()->symbolRef());
      goto write_mode_next_sym;
    };
}; 
Term* TermSharing::CopyFlatLiteral(const Flatterm* lit)
{
  CALL("CopyFlatLiteral(const Flatterm* lit)");
  Term* target = new (lit->arity()) Term(lit->symbolRef());
  Term* targ = target->Args();
  Flatterm::ConstIterator iter;
  iter.Reset(lit);
  iter.NextSym();
 copy_arg:
  if (iter.EndOfTerm()) 
    { 
      return target; 
    };
  if (iter.CurrPos()->symbolRef().isVariable())
    {
      *targ = iter.CurrPos()->symbolRef(); 
    }
  else 
    {
      targ->makeReference(IntegrateFlatComplex(iter.CurrPos()));
    }; 
  iter.after();
  targ = targ->Next(); 
  goto copy_arg;
}; 
Term* TermSharing::CopyFlatEq(const Term& header,
				const Flatterm* arg1,
				const Flatterm* arg2)
{
  CALL("CopyFlatEq(const Term& header,const Flatterm* arg1,const Flatterm* arg2)");
  Term* result = new (header.arity()) Term(header);
  Term* rarg = result->Args();
  if (arg1->symbolRef().isVariable()) { *rarg = arg1->symbolRef(); }
  else rarg->makeReference(IntegrateFlatComplex(arg1));
  rarg = rarg->Next();
  if (arg2->symbolRef().isVariable()) { *rarg = arg2->symbolRef(); }
  else rarg->makeReference(IntegrateFlatComplex(arg2));
  return result;
}; 
void TermSharing::DestroyLiteralCopy(Term* t)
{ 
  CALL("DestroyLiteralCopy(Term* t)");
  for (Term* arg = t->Args(); ; arg = arg->Next())
    {
      switch (arg->typeTag())
	{
	case TermVar : break;
	case TermRef : 
	  RemoveComplex(arg->deref());
	  break;
	case TermComplex :
	  delete t;
	  return;
	};
    };
}; 
void TermSharing::RemoveComplex(Term* t)
{
  CALL("RemoveComplex(Term* t)");
  if (t->arity())
    {
      RemoveComplexNonconst(t);
    };
}; 
void TermSharing::RemoveComplexNonconst(Term* t)
{
  CALL("RemoveComplexNonconst(Term* t)");
  ASSERT((t->typeTag() == TermComplex) && (t->arity()));
  Stack<Term*,VampireKernelConst::MaxTermSize> _subterms("_subterms(local)");
  Term* subterm;
  IndexNode* node;     
  IndexNode** branch;
  IndexNode** below_addr; 
  Term::Iterator iter;
  IndexNode* branch_to_delete;
  IndexNode* node_to_delete;
 process_next_term: 
  ASSERT_IN((t->typeTag() == TermComplex) && (t->arity()),"A2"); 
  branch = Tree(*t);
  node = *branch;
  ASSERT_IN(node,"A3");
  below_addr = branch;
  iter.Reset(t);
  iter.Next();
  do
    {
      ASSERT_IN(!(node->IsLeaf()),"A4"); 
      if (node->Alternative())
	{
	  branch = below_addr;
        check_node:
	  if (node->symbol() == iter.symbolRef())
	    {
	      below_addr = node->BelowAddr();
	      node = node->Below();
	    }
	  else 
	    {
	      branch = node->AlternativeAddr();
	      node = node->Alternative();
	      ASSERT(node);
	      goto check_node;
	    };
	}
      else
	{
	  ASSERT(node->symbol() == iter.symbolRef());
	  below_addr = node->BelowAddr();
	  node = node->Below();
	};
    }
  while (iter.Next());
  node->decEntryMutiplicity();
  if (!(node->entryMultuplicity())) 
    {
      subterm = node->term()->Args();
      while (!(subterm->isNotArg()))
	{
	  if ((subterm->IsReference()) && (subterm->deref()->arity()))
	    {
	      _subterms.push(subterm->deref());
	    };
	  subterm = subterm->Next();
	}; 
      delete node->term(); 
      branch_to_delete = *branch;
      *branch = (*branch)->Alternative(); 
      while (!(branch_to_delete->IsLeaf()))
	{
	  node_to_delete = branch_to_delete;
	  branch_to_delete = branch_to_delete->Below();
	  delete node_to_delete;
	};
      delete branch_to_delete; 
    };
  if (!(_subterms.empty()))    
    {
      t = _subterms.pop();
      goto process_next_term;
    };     
}; 
inline void TermSharing::DestroyTree(IndexNode* tree)
{
  CALL("DestroyTree(IndexNode* tree)");
  if (tree)
    {
      if (tree->IsLeaf()) 
	{
	  ASSERT(tree->term()->isComplex());
	  tree->term()->DestroyOccurences();
	  delete tree->term();
	  delete tree; 
	}
      else
	{
	  if (tree->Alternative()) DestroyTree(tree->Alternative());
	  if (tree->Below()) DestroyTree(tree->Below());
	  delete tree;
	};
    };
}; 
TermSharing::IndexNode** TermSharing::Tree(const Term& fun)
{
  CALL("Tree(const Term& fun)");
  ASSERT(fun.arity());
  DOP(_discTree.unfreeze());
  IndexNode** res = &(_discTree.sub(fun.functor()));
  DOP(_discTree.unfreeze());
  return res;
};
Term* TermSharing::Const(const Term& c)
{ 
  CALL("Const(const Term& c)");
  DOP(_sharedConst.unfreeze());
  Term* res = _sharedConst.sub(c.functor());
  DOP(_sharedConst.freeze()); 
  if (!res)
    {
      res = new(0) Term(c);
      res->InitOccurences();
      _sharedConst[c.functor()] = res;
    };
  return res;
};
