/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TERM_SHARING_HPP
#define TERM_SHARING_HPP
#include <climits>
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"                 
#include "Term.hpp"
#include "Stack.hpp"
namespace VK
{
class Flatterm;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM_SHARING
#  define DEBUG_NAMESPACE "TermSharing"
#endif
#include "debugMacros.hpp"
namespace VK
{
class TermSharing
{       
 public:
  class IndexNode;
  class Integrator; 
 public:
  TermSharing();
  ~TermSharing();
  void reset();
  Term* IntegrateComplex(Term* t);
  Term* IntegrateFlatComplex(const Flatterm* t);
  Term* CopyLiteral(Term* source);
  Term* CopyFlatLiteral(const Flatterm* source);
  Term* CopyFlatEq(const Term& header,const Flatterm* arg1,const Flatterm* arg2);
  void DestroyLiteralCopy(Term* t);
  void RemoveComplex(Term* t);
  void RemoveComplexNonconst(Term* t);
 private:
  Term* IntegrateComplexNonconst(Term* t);
  Term* IntegrateFlatComplexNonconst(const Flatterm* t);
  static void DestroyTree(IndexNode* tree);
  IndexNode** Tree(const Term& fun);
  Term* Const(const Term& c);
private:
  Term::Iterator _iter;
  BK::Stack<Term*,VampireKernelConst::MaxTermSize> _subterms;
  BK::Array<BK::GlobAlloc,IndexNode*,BK::FastRelocation<IndexNode*>,512UL> _discTree;
  BK::Array<BK::GlobAlloc,Term*,BK::FastRelocation<Term*>,128UL> _sharedConst;  
  friend class Integrator;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM_SHARING
#  define DEBUG_NAMESPACE "TermSharing::IndexNode"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TermSharing::IndexNode 
  {
  public: 
    IndexNode(const Term& sym) : _tag(sym) 
    {
      CALL("IndexNode(const Term& sym)");
      _below = 0;
      _alternative = 0;
    };
    IndexNode(Term* t) : _tag(t) 
    {
      CALL("IndexNode(Term* t)");
      _entryMultiplicity = 1;
    };
    ~IndexNode() {};
    void* operator new(size_t)
    {
      CALL("operator new(size_t)");
      return BK::GlobAlloc::allocate(sizeof(IndexNode));
    };
    void operator delete(void* obj)
    {
      CALL("operator delete(void* obj)");
      BK::GlobAlloc::deallocate(obj,sizeof(IndexNode));
    };
  public:
    bool IsLeaf() const { return _tag.IsReference(); };
    const Term& symbol() const { 
      CALL("symbol() const");
      ASSERT(!IsLeaf());
      return _tag; 
    };
    Term* term() { return _tag.deref(); }; 
    long entryMultuplicity() { return _entryMultiplicity; };
    void incEntryMultiplicity() { if (_entryMultiplicity < LONG_MAX) { _entryMultiplicity++; }; };
    void decEntryMutiplicity() { if (_entryMultiplicity < LONG_MAX) { _entryMultiplicity--; }; };
    IndexNode* Below() { return _below; };
    IndexNode** BelowAddr() { return &_below; };
    IndexNode* Alternative() { return _alternative; };
    IndexNode** AlternativeAddr() { return &_alternative; };
    void SetBelow(IndexNode* n) { _below = n; };  
    void SetAlternative(IndexNode* n) { _alternative = n; }; 
  private: 
    Term _tag;
    union
    {
      IndexNode* _below;
      long _entryMultiplicity;
    };
    IndexNode* _alternative;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM_SHARING
#  define DEBUG_NAMESPACE "TermSharing::Integrator"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class TermSharing::Integrator
  {
  public:
    Integrator(TermSharing* si) : index(si) {};
    bool Reset(const Term& fun)
    {
      CALL("Reset(const Term& fun)");
      node_addr = index->Tree(fun);
      return (*node_addr) != 0;
    };
    bool Read(const Term& sym)
    {
    check_tag:
      if ((*node_addr)->symbol() == sym)
	{
	  node_addr = (*node_addr)->BelowAddr();
	  return true;
	}
      else 
	{
	  node_addr = (*node_addr)->AlternativeAddr();
	  if (*node_addr) { goto check_tag; }
	  else return false;
	};
    }; 
    void Write(const Term& sym)
    {
      *node_addr = new IndexNode(sym);
      node_addr = (*node_addr)->BelowAddr();
    }; 
    void CreateLeaf(Term* t) { *node_addr = new IndexNode(t); };
    IndexNode* Leaf() { return (*node_addr); }; 
  private:
    TermSharing* index;
    IndexNode** node_addr;
  }; 
}; 
#endif
