/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef TERM_WEIGHTING_HPP
#define TERM_WEIGHTING_HPP
#include <iostream>
#include "WeightPolynomial.hpp"
#include "Flatterm.hpp"
#include "Term.hpp"
#include "InstanceRetrievalForDemodulation.hpp"
namespace VK
{
  class UnifTerm;
  class ForwardMatchingSubstitutionCache;
  class LocalUnification;
  class TermWeighting
    {
    public:
      enum Scheme
      {
	NonuniformConstant,
	Uniform
      };
    public:
      TermWeighting() : _scheme(Uniform) {};
      static TermWeighting* current() { return _current; };
      static void setCurrent(TermWeighting* tw) { _current = tw; };
      virtual ~TermWeighting() = 0;
      Scheme scheme() const { return _scheme; };
      virtual void collectWeight(WeightPolynomial& weight,const Flatterm* term) = 0;
      virtual void collectWeight(WeightPolynomial& weight,long coefficient,const Flatterm* term) = 0;
      virtual 
      long computeMinimalInstanceWeight(const Flatterm* term,
						  bool& containsVariables) = 0;
      virtual 
      long 
      computeWeightIfGroundAndLessOrEqual(long coefficient,
					  const Flatterm* term,
					  long weightLimit) = 0;
      virtual void collectWeight(WeightPolynomial& weight,const Term& term) = 0;
      virtual void collectWeight(WeightPolynomial& weight,long coefficient,const Term& term) = 0;
      virtual void collectWeight(WeightPolynomial& weight,long coefficient,const UnifTerm* term) = 0;
      virtual void collectWeight(WeightPolynomial& weight,const UnifTerm* term) = 0;
      virtual void collectWeight(WeightPolynomial& weight,const Term& term,ulong variableIndex) = 0;
      virtual 
	void 
	collectWeight(WeightPolynomial& weight,
		      long coefficient,
		      const Term* complexTerm,
		      const InstRet::Substitution* subst) = 0;
      virtual 
	void 
	collectWeight(WeightPolynomial& weight,
		      const Term* complexTerm,
		      const InstRet::Substitution* subst) = 0;
      virtual void collectWeight(WeightPolynomial& weight,
				 const Term& term,
				 const ForwardMatchingSubstitutionCache* subst) = 0;
      virtual 
      void
      collectWeight(WeightPolynomial& weight,
		    const Term& term,
		    const LocalUnification* unifier) = 0;
      virtual 
      long 
      computeMinimalInstanceWeight(const UnifTerm* term,
				   bool& containsVariables) = 0;
      virtual 
      long 
      computeWeightIfGroundAndLessOrEqual(long coefficient,
					  const UnifTerm* term,
					  long weightLimit) = 0;
      virtual 
      long 
      computeWeightIfGroundAndLessOrEqual(const UnifTerm* term,
					  long weightLimit) = 0;
      virtual 
      long 
      computeMinimalInstanceWeightIfLessOrEqual(long coefficient,
						const UnifTerm* term,
						long weightLimit,
						bool& ground) = 0;
      virtual 
      long 
      computeMinimalInstanceWeightIfLessOrEqual(const UnifTerm* term,
						long weightLimit,
						bool& ground) = 0;
      virtual 
      long 
      computeMinimalInstanceWeight(const Term* term,
				   bool& containsVariables) = 0;
      virtual 
      long 
      computeWeightIfGroundAndLessOrEqual(long coefficient,
					  const Term* term,
					  long weightLimit) = 0;
      virtual 
      long 
      computeWeightIfGroundAndLessOrEqual(const Term* term,
					  long weightLimit) = 0;
      virtual 
      long 
      computeMinimalInstanceWeightIfLessOrEqual(long coefficient,
						const Term* term,
						long weightLimit,
						bool& ground) = 0;
      virtual 
      long 
      computeMinimalInstanceWeightIfLessOrEqual(const Term* term,
						long weightLimit,
						bool& ground) = 0;
    protected:
      void setScheme(const Scheme& s) { _scheme = s; };
    private:
      Scheme _scheme;
      static TermWeighting* _current;
    }; 
}; 
namespace std
{
inline
ostream& operator<<(ostream& str,const VK::TermWeighting::Scheme& s)
{
  switch (s)
    {
    case VK::TermWeighting::NonuniformConstant:
      str << "nonuniform_constant";
      break;
    case VK::TermWeighting::Uniform:
      str << "uniform";
      break;
    default: str << "unknown_term_weighting_scheme";
    };
  return str;
}; 
};
#endif
