/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "UnifTerm.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_TERM
#  define DEBUG_NAMESPACE "UnifTerm"
#endif
#include "debugMacros.hpp"
using namespace std;
using namespace VK;
#ifndef NO_DEBUG_VIS
ostream& UnifTerm::outputTerm(ostream& str,const UnifTerm* term)
{
  if (term->symbol().isVariable()) return term->symbol().output(str);
  SimpleConstIter iter;
  ExpandingBracketCounter bracketCounter; 
  bracketCounter.reset();    
  iter.Reset(term);
  while (!(iter.Empty()))
    {
      iter.CurrPos()->symbol().output(str);
      if ((iter.CurrPos()->symbol().IsFunctor()) 
	  && (iter.CurrPos()->symbol().Func().arity()))
	{
	  bracketCounter.openSafe(iter.CurrPos()->symbol().Func().arity());
	  str << '(';
	  iter.SkipSym();
	}
      else
	{
	  ulong closed = bracketCounter.close();
	  while (closed) { str << ')';  closed--; };
	  iter.SkipSym();
	  if (!(iter.Empty())) str << ',';
	};
    };       
  return str;
}; 
#endif
void 
VK::UnifTerm::StandardTermToUnifTerm(ulong index,
				     const Term* t,
				     UnifTerm*& memory)
{
  CALL("StandardTermToUnifTerm(ulong index,const Term* t,UnifTerm*& memory)");
  DOP(UnifTerm* debugMemory = memory);
  ASSERT(t->isComplex());
  Term::Iterator t_iter;
  BracketCounter<VampireKernelConst::MaxTermDepth> brack_counter;
  UnifTerm* left_bracket[VampireKernelConst::MaxTermSize];
  UnifTerm** next_left_bracket = left_bracket;
  t_iter.Reset(t);
  Term sym;
  unsigned closed;
  unsigned var_offset = Unifier::variableOffset(index);
 next_sym:
  sym = t_iter.symbolRef();
  if (sym.isVariable())
    {
      memory->symbol().MkVar(Unifier::current()->variableBase(sym.var())+var_offset);
      memory->SetBracket(memory+1);
      closed = brack_counter.Close();    
      while (closed)
	{
	  next_left_bracket--;
	  (*next_left_bracket)->SetBracket(memory+1);
	  closed--;
	};
    }
  else 
    {
      memory->symbol().MkFunc(sym);
      if (sym.arity())
	{
	  brack_counter.Open(sym.arity());
	  *next_left_bracket = memory;
	  next_left_bracket++;
	}
      else 
	{
	  memory->SetBracket(memory+1);
	  unsigned closed = brack_counter.Close();    
	  while (closed)
	    {
	      next_left_bracket--;
	      (*next_left_bracket)->SetBracket(memory+1);
	      closed--;
	    };  
	};
    };
  memory++;
  if (t_iter.Next())
    {
      goto next_sym;
    };
  ASSERT(next_left_bracket == left_bracket); 
  ASSERT(CorrectComplexTerm(debugMemory)); 
}; 
void UnifTerm::binaryStandardTermToUnifTerm(ulong index,
					    const Term& top,
					    const Term& arg1,
					    const Term& arg2,
					    UnifTerm*& memory)
{
  CALL("binaryStandardTermToUnifTerm(..)");
  ASSERT(top.isComplex());
  ASSERT(top.arity() == 2UL);
  UnifTerm* topLocation = memory;
  memory->symbol().MkFunc(top);
  ++memory;
  StandardTermToUnifTerm(index,arg1,memory);
  StandardTermToUnifTerm(index,arg2,memory);
  topLocation->SetBracket(memory);
  ASSERT(CorrectComplexTerm(topLocation)); 
}; 
bool UnifTerm::equal(UnifTerm* term1,UnifTerm* term2)
{
  CALL("equal(UnifTerm* term1,UnifTerm* term2)");
  if (term1->symbol().isVariable())
    {
      bool freeVar1;
      void* instance1 = term1->symbol().var()->Unref(freeVar1);
      if (freeVar1)
	{
	  if (term2->symbol().isVariable())
	    {
	      bool freeVar2;
	      void* instance2 = term2->symbol().var()->Unref(freeVar2);
	      if (freeVar2)
		{
		  return instance1 == instance2;
		};
	    };
	  return false;
	};
      term1 = static_cast<UnifTerm*>(instance1);
    };
  ASSERT(term1->symbol().IsFunctor());
  if (term2->symbol().isVariable())
    {
      bool freeVar2;
      void* instance2 = term2->symbol().var()->Unref(freeVar2);
      if (freeVar2) return false;
      term2 = static_cast<UnifTerm*>(instance2);
    };
  ASSERT(term2->symbol().IsFunctor());
  UnifTerm::CorrectBracket(term1);
  UnifTerm::CorrectBracket(term2);
  return equalComplex(term1,term2);
}; 
bool UnifTerm::equalComplex(UnifTerm* term1,UnifTerm* term2)
{
  CALL("equalComplex(UnifTerm* term1,UnifTerm* term2)");
  ASSERT(CorrectComplexTerm(term1));
  ASSERT(CorrectComplexTerm(term2));
  Iter iter1;
  Iter iter2;
  iter1.Reset(term1);
  iter2.Reset(term2);
  while (!(iter1.Empty()))
    {
      if (iter1.CurrVar())
	{
	  if (iter1.CurrVar() != iter2.CurrVar()) { return false; };
	}
      else
	{
	  if (iter2.CurrVar() || 
	      iter1.CurrSubterm()->symbol() != iter2.CurrSubterm()->symbol())
	    { return false; };
	};
      iter1.SkipSym();
      iter2.SkipSym();
    };
  return true;
}; 
bool UnifTerm::identical(const UnifTerm* term1,const UnifTerm* term2)
{
  CALL("identical(const UnifTerm* term1,const UnifTerm* term2)");
  SimpleConstIter iter1;
  SimpleConstIter iter2;
  iter1.Reset(term1);
  iter2.Reset(term2);
  while (!iter1.Empty())
    {
      if (iter1.CurrPos()->symbol() != iter2.CurrPos()->symbol() ||
	  iter1.CurrPos()->size() != iter2.CurrPos()->size())
	return false;
      iter1.SkipSym();
      iter2.SkipSym();
    };
  return true;
}; 
#ifndef NO_DEBUG
bool UnifTerm::CorrectComplexTerm(const UnifTerm* term)
{
  CALL("CorrectComplexTerm(const UnifTerm* term)");
  if (!term) 
    {
      DMSG("UnifTerm::CorrectComplexTerm(term)> !term\n");
      return false; 
    };
  if (!term->symbol().IsFunctor())
    {
      DMSG("UnifTerm::CorrectComplexTerm(term)> !term->symbol().IsFunctor()\n");
      return false;
    };
  if (!term->Bracket())
    {
      DMSG("UnifTerm::CorrectComplexTerm(term)> !term->Bracket()\n");
      return false;
    };
  if (term->Bracket() <= term) 
    {
      DMSG("UnifTerm::CorrectComplexTerm(term)> term->Bracket() <= term\n");
      return false;
    };
  UnifTerm* end = term->Bracket();
  long holes = term->symbol().Func().arity();
  term++;
  while (term < end)
    {
      holes--;
      if (holes < 0) 
	{
	  DMSG("UnifTerm::CorrectComplexTerm(term)> holes < 0\n"); 
	  return false; 
	};
      if (term->symbol().IsFunctor())
	{
	  holes += term->symbol().Func().arity();
	}
      else
	{
	  if (!Unifier::current()->isVariable(term->symbol().var())) 
	    {
	      DMSG("UnifTerm::CorrectComplexTerm(term)> !Unifier::current()->isVariable(term->symbol().var())\n"); 
	      return false;
	    };
	};
      term++;
    };
  if (holes) 
    {
      DMSG("UnifTerm::CorrectComplexTerm(term)> holes\n"); 
      return false;
    };
  return true;
}; 
#endif
void UnifTerm::CorrectBracket(UnifTerm* term)
{
  if (!(term->Bracket()))
    {
      UnifTerm* curr_pos = term+1; 
      BracketCounter<VampireKernelConst::MaxTermDepth> brack_counter;
      brack_counter.Open(term->symbol().Func().arity());
      do
	{
	  if (curr_pos->symbol().isVariable() || (!(curr_pos->symbol().Func().arity())))
	    {
	      brack_counter.Close();
	    }
	  else { brack_counter.Open(curr_pos->symbol().Func().arity()); };
	  curr_pos++;
	}
      while (brack_counter.IsOpen());
      term->SetBracket(curr_pos);
    };
}; 
void UnifTerm::resetCorruptBrackets(UnifTerm* term,
				    UnifTerm* changed_suffix,
				    UnifTerm* old_end_of_term,
				    UnifTerm* new_end_of_term)
{
  while (term < changed_suffix)
    {
      if (term->Bracket() > changed_suffix)
	{
	  if (term->Bracket() == old_end_of_term)
	    {
	      term->SetBracket(new_end_of_term);
	    }
	  else
	    {
	      term->ResetBracket();
	    };
	};
      term++;
    };
}; 
bool UnifTerm::Occurs(Unifier::Variable* v,UnifTerm* term)
{
  CALL("Occurs(Unifier::Variable* v,UnifTerm* term)");
  DOP(UnifTerm* debug_top_level_term = term);
  static Unifier::Variable::Multiset link_counter;
  link_counter.reset();
#ifdef NO_DEBUG
  static BK::Stack<void*,VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices > 
    _links;    
#else
  static BK::Stack<void*,VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices > 
    _links("UnifTerm::Occurs(..):_links");    
#endif
  _links.reset();
  SimpleIter iter;
  void* link;
 check_top_level:
  iter.Reset(term);
  iter.SkipSym();
  while (!iter.Empty())
    {
      Symbol sym  = iter.CurrPos()->symbol();
      if (sym.isVariable())
	{
	  link = sym.var()->Binding();
	  if (link) 
	    {
	      if (link_counter.registerVar(sym.var()) == 0)
		{
		  _links.push(link);
		};
	    }
	  else 
	    {
	      if (sym.var() == v) 
		{
		  ASSERT(OccursIneff(v,debug_top_level_term));
		  link_counter.reset(); 
		  return true; 
		};
	    };
	};
      iter.SkipSym();
    }; 
 try_next_link:
  if (_links.empty())
    {
      ASSERT(!(OccursIneff(v,debug_top_level_term)));
      link_counter.reset(); 
      return false; 
    };
  link = _links.pop();
 check_link:
  if (Unifier::current()->isVariable(link))
    {
      if (((Unifier::Variable*)link)->Binding()) 
	{
	  if (link_counter.registerVar((Unifier::Variable*)link) == 0)
	    {
	      link = ((Unifier::Variable*)link)->Binding();
	      goto check_link;
	    }
	  else
	    {
	      goto try_next_link;
	    };
	}
      else 
	{
	  if (((Unifier::Variable*)link) == v) 
	    {
	      ASSERT(OccursIneff(v,debug_top_level_term));
	      link_counter.reset(); 
	      return true; 
	    };
	  goto try_next_link;
	};
    }
  else 
    {
      term = (UnifTerm*)link;
      goto check_top_level;
    };
}; 
#ifndef NO_DEBUG
bool UnifTerm::OccursIneff(Unifier::Variable* v,UnifTerm* term) 
{
  CALL("OccursIneff(Unifier::Variable* v,UnifTerm* term)");
  ASSERT(v);
  ASSERT(Unifier::current()->isVariable(v));
  ASSERT(CorrectComplexTerm(term));
  Iter iter;
  iter.Reset(term);
 next_sym:
  iter.SkipSym();
  if (iter.Empty())
    {
      return false;
    }
  else
    {
      if (iter.CurrVar() == v)
	{
	  return true;
	}
      else
	{
	  goto next_sym;
	};
    };
}; 
#endif
