/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef UNIFICATION_HPP
#define UNIFICATION_HPP
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "IUCommand.hpp"
#include "Unifier.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIFICATION
#  define DEBUG_NAMESPACE "Unification"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Unification
  {
  public: 
    static inline bool Unify(UnifTerm* t1,UnifTerm* t2)
    {
      CALL("Unify(UnifTerm* t1,UnifTerm* t2)");
      if (t1->symbol().isVariable())
	{
	  ASSERT(!(t1->symbol().var()->Binding()));
	  if (t2->symbol().isVariable())
	    {
	      ASSERT(!(t2->symbol().var()->Binding()));
	      Unifier::current()->MarkSection();
	      if (t1->symbol() != t2->symbol())
		Unifier::current()->Bind(t1->symbol().var(),static_cast<void*>(t2->symbol().var()));
	      return true;
	    };
	  if (UnifTerm::Occurs(t1->symbol().var(),t2)) return false;
	  Unifier::current()->MarkSection();
	  Unifier::current()->Bind(t1->symbol().var(),static_cast<void*>(t2));
	  return true;
	}
      else
	{
	  if (t2->symbol().isVariable())
	    {
	      ASSERT(!(t2->symbol().var()->Binding()));
	      if (UnifTerm::Occurs(t2->symbol().var(),t1)) return false;
	      Unifier::current()->MarkSection();
	      Unifier::current()->Bind(t2->symbol().var(),static_cast<void*>(t1));
	      return true; 
	    }
	  else 
	    return UnifyComplex(t1,t2);
	};
    }; 
    static inline bool UnifyArbitrary(UnifTerm* t1,UnifTerm* t2)
    {
      CALL("UnifyArbitrary(UnifTerm* t1,UnifTerm* t2)");
      Unifier::Variable* var1;
      Unifier::Variable* var2;
      UnifTerm* complexTerm1;
      UnifTerm* complexTerm2;
      DOP(complexTerm1 = 0);
      DOP(complexTerm2 = 0);
      if (t1->symbol().isVariable())
	{
	  bool freeVar;
	  void* instance = t1->symbol().var()->Unref(freeVar);
	  if (freeVar)
	    {
	      var1 = static_cast<Unifier::Variable*>(instance);
	    }
	  else
	    {
	      var1 = 0;
	      complexTerm1 = static_cast<UnifTerm*>(instance);
	    };
	}
      else
	{
	  var1 = 0;
	  complexTerm1 = t1;
	};
      if (t2->symbol().isVariable())
	{
	  bool freeVar;
	  void* instance = t2->symbol().var()->Unref(freeVar);
	  if (freeVar)
	    {
	      var2 = static_cast<Unifier::Variable*>(instance);
	    }
	  else
	    {
	      var2 = 0;
	      complexTerm2 = static_cast<UnifTerm*>(instance);
	    };
	}
      else
	{
	  var2 = 0;
	  complexTerm2 = t2;
	};
      if (var1)
	{	 
	  if (var2)
	    { 
	      Unifier::current()->MarkSection();
	      if (var1 != var2)
		Unifier::current()->Bind(var1,static_cast<void*>(var2));
	      return true;
	    }
	  else 
	    {
	      if (UnifTerm::Occurs(var1,complexTerm2))
		return false;
	      Unifier::current()->MarkSection();
	      Unifier::current()->Bind(var1,static_cast<void*>(complexTerm2));
	      return true;
	    };
	}
      else
	if (var2)
	  {
	    if (UnifTerm::Occurs(var2,complexTerm1))
	      return false;
	    Unifier::current()->MarkSection();
	    Unifier::current()->Bind(var2,static_cast<void*>(complexTerm1));
	    return true;
	  }
	else 
	  {
	    UnifTerm::CorrectBracket(complexTerm1);
	    UnifTerm::CorrectBracket(complexTerm2);
	    return UnifyComplex(complexTerm1,complexTerm2);  
	  };   
    }; 
    static inline bool UnifyComplex(UnifTerm* t1,UnifTerm* t2)
    {
      CALL("UnifyComplex(UnifTerm* t1,UnifTerm* t2)");
      Unifier::current()->MarkSection();
      if (UnifyComplex_(t1,t2)) { return true; }
      else { Unifier::current()->UnbindLastSection(); return false; };
    }; 
    static inline bool UnifyLinear(UnifTerm* t,UnifTerm* lin_t)
    {
      CALL("UnifyLinear(UnifTerm* t,UnifTerm* lin_t)");
      Unifier::current()->MarkSection();
      if (UnifyLinear_(t,lin_t)) { return true; }
      else { Unifier::current()->UnbindLastSection(); return false; };
    }; 
    static inline bool MatchGround(UnifTerm* t,UnifTerm* gr_t)
    {
      CALL("MatchGround(UnifTerm* const t,UnifTerm* const gr_t)");
      Unifier::current()->MarkSection();
      if (MatchGround_(t,gr_t)) { return true; }
      else { Unifier::current()->UnbindLastSection(); return false; };
    }; 
  private:
    static inline bool UnifyComplex_(UnifTerm* t1,UnifTerm* t2)
    {
      CALL("UnifyComplex_(UnifTerm* t1,UnifTerm* t2)");
      ASSERT(t1);
      ASSERT(t2);
      ASSERT(t1->symbol().IsFunctor());
      ASSERT(t2->symbol().IsFunctor());
      ASSERT(t1->Bracket());
      ASSERT(t2->Bracket());
      ASSERT(t1->Bracket() > t1);
      ASSERT(t2->Bracket() > t2);
      ASSERT(UnifTerm::CorrectComplexTerm(t1));
      ASSERT(UnifTerm::CorrectComplexTerm(t2));
      if (t1->symbol() != t2->symbol()) { return false; };
      UnifTerm::Iter iter1;
      UnifTerm::Iter iter2;
      iter1.Reset(t1);
      iter2.Reset(t2);
      iter1.SkipSym();
      iter2.SkipSym();
    next_sym:
      if (iter1.Empty())
	{
	  ASSERT(iter2.Empty());
	  ASSERT(UnifTerm::equal(t1,t2));
	  return true;
	}     
      else
	{
	  if (iter1.CurrVar())
	    {
	      if (iter2.CurrVar()) 
		{
		  if (iter1.CurrVar() != iter2.CurrVar()) 
		    {
		      Unifier::current()->Bind(iter1.CurrVar(),static_cast<void*>(iter2.CurrVar()));
		    };
		  iter1.SkipSym();
		  iter2.SkipSym();
		  goto next_sym;
		}
	      else
		{
		  UnifTerm::CorrectBracket(iter2.CurrSubterm());
		  ASSERT(UnifTerm::CorrectComplexTerm(iter2.CurrSubterm()));
		  if (UnifTerm::Occurs(iter1.CurrVar(),iter2.CurrSubterm()))
		    {
		      return false;
		    }
		  else
		    {
		      Unifier::current()->Bind(iter1.CurrVar(),static_cast<void*>(iter2.CurrSubterm()));
		      iter1.SkipSym();
		      iter2.SkipTerm();
		      goto next_sym;
		    };
		};
	    }
	  else 
	    {
	      if (iter2.CurrVar())
		{
		  UnifTerm::CorrectBracket(iter1.CurrSubterm());
		  ASSERT(UnifTerm::CorrectComplexTerm(iter1.CurrSubterm())); 
		  if (UnifTerm::Occurs(iter2.CurrVar(),iter1.CurrSubterm()))
		    {
		      return false;
		    }
		  else
		    {
		      Unifier::current()->Bind(iter2.CurrVar(),static_cast<void*>(iter1.CurrSubterm()));
		      iter1.SkipTerm();
		      iter2.SkipSym();
		      goto next_sym;
		    };
		}
	      else 
		{
		  if (iter1.CurrSubterm()->symbol() == iter2.CurrSubterm()->symbol())
		    {
		      iter1.SkipSym();
		      iter2.SkipSym();
		      goto next_sym;
		    }
		  else
		    {
		      return false;
		    };
		};
	    };
	};
    }; 
    static inline bool UnifyLinear_(UnifTerm* t,UnifTerm* lin_t)
    {
      CALL("UnifyLinear_(UnifTerm* t,UnifTerm* lin_t)");   
      if (t->symbol() != lin_t->symbol()) { return false; };
      UnifTerm::Iter iter1;
      UnifTerm::SimpleIter iter2;
      iter1.Reset(t);
      iter2.Reset(lin_t);
      iter1.SkipSym();
      iter2.SkipSym();
    check_iter:
      if (iter2.Empty()) { return true; };
      if (iter1.CurrVar()) 
	{
	  if (iter2.CurrPos()->symbol().isVariable())
	    {
	      Unifier::current()->Bind(iter1.CurrVar(),static_cast<void*>(iter2.CurrPos()->symbol().var()));
	      iter2.SkipSym();
	    }
	  else
	    {
	      Unifier::current()->Bind(iter1.CurrVar(),static_cast<void*>(iter2.CurrPos()));
	      iter2.SkipTerm();
	    };
	  iter1.SkipSym();
	  goto check_iter;
	}
      else 
	{
	  if (iter2.CurrPos()->symbol().isVariable())
	    {
	      UnifTerm::CorrectBracket(iter1.CurrSubterm());
	      Unifier::current()->Bind(iter2.CurrPos()->symbol().var(),iter1.CurrSubterm());
	      iter1.SkipTerm();
	      iter2.SkipSym();
	      goto check_iter;
	    }
	  else 
	    {
	      if (iter1.CurrSubterm()->symbol() != iter2.CurrPos()->symbol())
		{
		  return false;
		};
	      iter1.SkipSym();
	      iter2.SkipSym();
	      goto check_iter;
	    };
	};
    }; 
    static inline bool MatchGround_(UnifTerm* t,UnifTerm* gr_t)
    {
      CALL("MatchGround_(UnifTerm* t,UnifTerm* gr_t)");
      if (t->symbol() != gr_t->symbol()) { return false; };
      Unifier::Variable* var;
      UnifTerm::Iter iter1;
      UnifTerm::SimpleIter iter2;
      iter1.Reset(t);
      iter2.Reset(gr_t);
      iter1.SkipSym();
      iter2.SkipSym();
    next_sym:
      if (iter1.Empty())
	{
	  return true;
	}
      else
	{
	  var = iter1.CurrVar();
	  if (var)
	    {
	      Unifier::current()->Bind(var,static_cast<void*>(iter2.CurrPos()));
	      iter1.SkipSym();
	      iter2.SkipTerm();
	      goto next_sym;
	    }
	  else 
	    {
	      if (iter1.CurrSubterm()->symbol() == iter2.CurrPos()->symbol())
		{
		  iter1.SkipSym();
		  iter2.SkipSym();
		  goto next_sym;
		}
	      else
		{
		  return false;
		};
	    };
	}; 
    }; 
  }; 
}; 
#endif
