/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef UNIFIER_HPP
#define UNIFIER_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Multiset.hpp"
#include "Comparison.hpp"
#include "Stack.hpp"
#include "WeightPolynomial.hpp"
#include "LinearCombination.hpp"
namespace VK
{
  class UnifTerm;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIFIER
#  define DEBUG_NAMESPACE "Unifier"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Unifier
    {
    public:
      class Variable
	{
	public:
	  class Multiset;
	  class RenamingToInt;
	  class Weighting;
	public:
	  Variable() : binding(0) {};  
	  ~Variable() {};
	public:
	  void Bind(void* bind) { binding = bind; };
	  void Unbind() 
	    {
	      CALL("Variable::Unbind()");
	      binding = 0; 
	    };
	  const void* Binding() const { return binding; };
          void* Binding() { return binding; };
	  Variable* Next() { return this+1; };
	  Variable* Previous() { return this-1; };
	  ulong Index() const 
	    { 
	      CALL("Variable::Index() const");
	      return (this - Unifier::current()->firstVariable())/VampireKernelConst::MaxNumOfVariables; 
	    }; 
	  void NumAndIndex(ulong& num,ulong& index) const
	    {
	      CALL("Variable::NumAndIndex(ulong& num,ulong& index) const");
	      ulong pos = this - Unifier::current()->firstVariable();
	      index = pos/VampireKernelConst::MaxNumOfVariables;
	      num = pos%VampireKernelConst::MaxNumOfVariables;
	    };
	  ulong relativeNum() const
	  {
	    CALL("Variable::relativeNum() const");
	    return 
	      (this - Unifier::current()->firstVariable()) % 
	      VampireKernelConst::MaxNumOfVariables;
	  };
	  ulong absoluteVarNum() const
	    {
	      CALL("Variable::absoluteVarNum() const");
	      ASSERT(Unifier::current()->isVariable(static_cast<const void*>(this)));
	      return static_cast<ulong>(this - Unifier::current()->firstVariable());
	    };
	  void* Unref() 
	    {
	      CALL("Variable::Unref()");
	      ASSERT(Unifier::current()->isVariable(static_cast<void*>(this)));
	      void* res = Binding();
	      if (res)
		{
		check_binding:
		  if (Unifier::current()->isVariable(res))
		    {
		      void* tmp = (static_cast<Variable*>(res))->Binding();
		      if (tmp)
			{
			  res = tmp;
			  goto check_binding;
			}
		      else
			{
			  return res; 
			};
		    }
		  else 
		    {
		      return res;
		    };
		}
	      else
		{
		  return this;
		};
	    }; 
	  const void* Unref(bool& free_var) const
	    {
	      CALL("Variable::Unref(bool& free_var const)");
	      ASSERT(Unifier::current()->isVariable(static_cast<const void*>(this)));
	      const void* res = Binding();
	      if (res)
		{
		check_binding:
		  if (Unifier::current()->isVariable(res))
		    {
		      const void* tmp = (static_cast<const Variable*>(res))->Binding();
		      if (tmp)
			{
			  res = tmp;
			  goto check_binding;
			}
		      else
			{
			  free_var = true;
			  return res; 
			};
		    }
		  else 
		    {
		      free_var = false;
		      return res;
		    };
		}
	      else
		{
		  free_var = true;
		  return static_cast<const void*>(this);
		};
	    }; 
	  void* Unref(bool& free_var)
	    {
	      CALL("Variable::Unref(bool& free_var)");
	      ASSERT(Unifier::current()->isVariable(static_cast<void*>(this)));
	      void* res = Binding();
	      if (res)
		{
		check_binding:
		  if (Unifier::current()->isVariable(res))
		    {
		      void* tmp = (static_cast<Variable*>(res))->Binding();
		      if (tmp)
			{
			  res = tmp;
			  goto check_binding;
			}
		      else
			{
			  free_var = true;
			  return res; 
			};
		    }
		  else 
		    {
		      free_var = false;
		      return res;
		    };
		}
	      else
		{
		  free_var = true;
		  return static_cast<void*>(this);
		};
	    }; 
	public: 
	  ostream& output(ostream& str) const;
#ifndef NO_DEBUG_VIS
	  ostream& outputBinding(ostream& str,ostream& (*outputTerm)(ostream& str,const UnifTerm* t)) const;
#endif
	private:
	  void* binding;
	}; 
    public:
      class InitStatic
	{
	public:
	  InitStatic() 
	    {
	      if (!_count)
		{
		  for (ulong i = 0UL; i < VampireKernelConst::MaxNumOfIndices; ++i)
		    {
		      Unifier::_offsetForIndex[i] = i*VampireKernelConst::MaxNumOfVariables;
		    };
		};
	      ++_count;
	    };
	private:
	  static long _count;
	};
    public:
      Unifier();
      ~Unifier();
      void reset() 
	{
	  CALL("reset()");
	  UnbindAll(); 
	};
      const void* state() const 
      {
	return static_cast<const void*>(next);
      };
      static Unifier* current() 
	{
	  return _current; 
	};
      static void setCurrent(Unifier* subst) { _current = subst; };
      Variable* variableBase(ulong v) { return _variableBank + v; };
      static ulong variableOffset(ulong index) 
      {
	CALL("variableOffset(ulong index)");
	return _offsetForIndex[index]; 
      };     
      Variable* variable(ulong v,ulong index) 
	{ 
	  return variableBase(v) + variableOffset(index); 
	};
      Variable* variable(ulong absoluteVarNum)
	{
	  return _variableBank + absoluteVarNum;
	};
      Variable* firstVariable() { return _variableBank; };
      Variable* firstVariable(ulong index) { return firstVariable() + variableOffset(index); };
      Variable* lastVariable(ulong index) 
	{ 
	  return firstVariable(index) + VampireKernelConst::MaxNumOfVariables - 1;
	};
      static ulong variableIndex(ulong varOffset) 
	{
	  return (varOffset/VampireKernelConst::MaxNumOfVariables);
	};
      bool isVariable(const void* ptr) const 
	{ 
	  return ((ptr >= (static_cast<const void*>(_variableBank))) && 
		  (ptr < (static_cast<const void*>(_endOfVariableBank)))); 
	};
      void UnbindAll() 
	{
	  CALL("UnbindAll()");
	  while (next != bound)
	    {
	      next--;
	      (*next)->Unbind();
	    };
	  next_section = section;
	};
#ifndef NO_DEBUG
      bool NoBoundVariables() const 
	{
	  return (next == bound) && (next_section == section);
	};
#endif   
      void Bind(Variable* v,void* val) 
	{
	  CALL("Bind(Variable* v,void* val)");
	  ASSERT(v != (static_cast<Variable*>(val)));
	  v->Bind(val);
	  *next = v;
	  next++;
	};
      Variable* LastBound() const { return *(next-1); };
      void UnbindLast()
	{
	  CALL("UnbindLast()");
	  next--;
	  (*next)->Unbind();
	  ASSERT(SectionsOK());
	};
      void MarkSection() 
	{
	  CALL("MarkSection()");
	  *next_section = next; 
	  next_section++;
	  ASSERT(SectionsOK());
	};
      Variable*** LastMark() const
	{
	  CALL("LastMark()");
	  ASSERT(next_section > section);
	  return next_section-1; 
	};
      void UnbindDownto(Variable*** mark)
	{
	  CALL("UnbindDownto(Variable*** mark)");
	  Variable** m = *mark;
	  next_section = mark;
	  while (next != m)
	    {
	      next--;
	      (*next)->Unbind();
	    };
	  ASSERT(SectionsOK());
	};
      void UnbindLastSection()
	{
	  CALL("UnbindLastSection()");
	  ASSERT(next_section > section);
	  next_section--;
	  Variable** section_mark = *next_section;
	  ASSERT(section_mark >= bound);
	  ASSERT(section_mark <= next);
	  while (next != section_mark)
	    {
	      next--;
	      (*next)->Unbind();
	    };
	  ASSERT(SectionsOK());
	};
      Variable*** MarkState()
	{
	  MarkSection();
	  return LastMark();
	};
    public:
#ifndef NO_DEBUG
      bool SectionsOK() const
	{
	  Variable*** sect = next_section - 1;
	  if (sect >= (static_cast<Variable** const *>(section)))
	    {
	      if ((*sect) > next)
		{
		  return false;
		};
	    };
	  return true;     
	};
      ulong NumOfBindings() const { return (next - static_cast<Variable* const *>(bound)); };
#endif
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str,ostream& (*outputTerm)(ostream& str,const UnifTerm* t)) const;
#endif
    private:
      Variable _variableBank[VampireKernelConst::MaxNumOfIndices*VampireKernelConst::MaxNumOfVariables];
      static ulong _offsetForIndex[VampireKernelConst::MaxNumOfIndices];
      const Variable* _endOfVariableBank;
      Variable* bound[VampireKernelConst::MaxNumOfVariables*VampireKernelConst::MaxNumOfIndices];
      Variable** next;
      Variable** section[VampireKernelConst::MaxNumOfVariables*VampireKernelConst::MaxNumOfIndices];
      Variable*** next_section;
      static Unifier* _current;
      friend class InitStatic;
    }; 
static Unifier::InitStatic unifierInitStatic;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIFIER
#define DEBUG_NAMESPACE "Unifier::Variable::Multiset"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Unifier::Variable::Multiset
  {
  public:
    Multiset() {};
    void reset() { _data.reset(); };
    long registerVar(const Variable* v) { 
      ulong n = v->absoluteVarNum();
      _var[n] = v;
      long res = _data.coefficient(n);
      _data.add1(n);
      return res;
    };
    long registerVar(long coeff,const Variable* v) { 
      ulong n = v->absoluteVarNum();
      _var[n] = v;
      long res = _data.coefficient(n);
      _data.add(coeff,n);
      return res;
    };
    long score(const Variable* v) const { 
      return _data.coefficient(v->absoluteVarNum()); 
    };
    void resetIter() { _iter = _data.begin(); };
    const Variable* nextVar() {
      if (_iter == _data.end()) return 0;
      const Variable* res = _var[*_iter];
      ++_iter;
      return res;
    };
  private:
    BK::LinearCombination<long,
      VampireKernelConst::MaxNumOfIndices * 
    VampireKernelConst::MaxNumOfVariables>
    _data;
    const Variable* _var[VampireKernelConst::MaxNumOfIndices * 
			VampireKernelConst::MaxNumOfVariables];
    const ulong* _iter;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIFIER
#define DEBUG_NAMESPACE "Unifier::Variable::RenamingToInt"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class Unifier::Variable::RenamingToInt
    {
    public: 
      RenamingToInt()
	{
	  for (ulong i = 0; 
	       i< VampireKernelConst::MaxNumOfIndices*VampireKernelConst::MaxNumOfVariables; 
	       i++)
	    {
	      _renaming[i] = 0UL;
	    };
	  next_var = 1;
	  next = renamed;
	};
      void Reset()
	{
	  next_var = 1;
	  while (next != renamed)
	    {
	      next--; 
	      _renaming[(*next)->absoluteVarNum()] = 0UL;
	    };
	};
      ulong Rename(const Unifier::Variable* var)
	{
	  ulong res = _renaming[var->absoluteVarNum()];
	  if (!res) 
	    {
	      res = next_var;
	      _renaming[var->absoluteVarNum()] = res;
	      next_var++;
	      *next = var;
	      next++;
	    };
	  return res;
	};
    public: 
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const;
#endif
    private:
      ulong _renaming[VampireKernelConst::MaxNumOfIndices*VampireKernelConst::MaxNumOfVariables];
      ulong next_var;
      const Unifier::Variable* 
      renamed[VampireKernelConst::MaxNumOfVariables*VampireKernelConst::MaxNumOfIndices]; 
      const Unifier::Variable** next;
    }; 
}; 
#endif 
