/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef VARIABLE_SET_HPP
#define VARIABLE_SET_HPP
#include "jargon.hpp"
#include "Stack.hpp"
#include "VampireKernelConst.hpp"
namespace VK {
  class VariableSet {
  public:
    VariableSet() {
      for (ulong i = 0UL; i < VampireKernelConst::MaxNumOfVariables + 1; ++i)
	_contains[i] = false;
    };
    ~VariableSet() {};
    const ulong* begin() const { return _registered.begin(); };
    const ulong* end() const { return _registered.end(); };
    void reset() {
      while (_registered)
	_contains[_registered.pop()] = false;
    };
    bool contains(const ulong& varNum) const { return _contains[varNum]; };
    void add(const ulong& varNum) {
      if (!_contains[varNum]) 
	{
	  _contains[varNum] = true;
	  _registered.push(varNum);
	};
    };
    bool isSubsetOf(const VariableSet& set) const {
      for (const ulong* el = _registered.begin();
	   el != _registered.end();
	   ++el)
	if (!set.contains(*el)) return false;
      return true;
    };
  private:
    bool _contains[VampireKernelConst::MaxNumOfVariables + 1];
    BK::Stack<ulong,VampireKernelConst::MaxNumOfVariables + 1> _registered;
  };
};
#endif
