/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef VARIABLE_WEIGHTING_HPP
#define VARIABLE_WEIGHTING_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLE_WEIGHTING
#define DEBUG_NAMESPACE "VariableWeighting"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class VariableWeighting
  {
  public:
    VariableWeighting() : registered("registered")
    { 
    };
    void Reset()
    {
      while (!(registered.empty())) { weight[registered.pop()].Reset(); };
    };
    void MoreOccurences(ulong var,ulong occ)
    {
      if (!(weight[var].Nonnull())) { registered.push(var); };
      weight[var].MoreOccurences(occ);
    };
    BK::Comparison Compare(ulong var1,ulong var2) const
    {
      return weight[var1].Compare(weight[var2]);
    };
    void AnotherTermVariables(const WeightPolynomial::Monomials& vars)
    {
      CALL("AnotherTermVariables(const WeightPolynomial::Monomials& vars)");
      const ulong* end = vars.end();
      for (const ulong* v = vars.begin(); v != end; ++v)
	{
	  if (!(weight[*v].Nonnull())) { registered.push(*v); };
	  weight[*v].AnotherTerm();
	  MoreOccurences(*v,vars.coefficient(*v));
	};
    };
  private:
    class Weight
    {
    public:
      Weight() : num_of_occurences(0), different_terms(0) {};
      ~Weight() {};
      void Reset() { num_of_occurences = 0; different_terms = 0; };
      bool Nonnull() { return (num_of_occurences || different_terms); };
      void MoreOccurences(ulong n) { num_of_occurences += n; };
      void AnotherTerm() { different_terms++; };
      BK::Comparison Compare(const Weight& w) const
      {
	if (num_of_occurences == w.num_of_occurences)
	  {
	    if (different_terms == w.different_terms) { return BK::Equal; }
	    else
	      { 
		if (different_terms > w.different_terms) { return BK::Greater; }
		else return BK::Less; 
	      };
	  }
	else
	  {
	    if (num_of_occurences > w.num_of_occurences) { return BK::Greater; }
	    else return BK::Less; 
	  }; 
      };
    private:
      ulong num_of_occurences;
      ulong different_terms; 
    }; 
  private: 
    Weight weight[VampireKernelConst::MaxNumOfVariables];
    BK::Stack<ulong,VampireKernelConst::MaxNumOfVariables> registered;
  }; 
}; 
#endif 
