/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef WEIGHT_POLYNOMIAL_SPECIALISED_COMPARISON_COMMAND_HPP
#define WEIGHT_POLYNOMIAL_SPECIALISED_COMPARISON_COMMAND_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "WeightPolynomial.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
#  define DEBUG_NAMESPACE "WeightPolynomial::SpecialisedComparison::Command"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class WeightPolynomial::SpecialisedComparison::Command
  {
  public:
    enum Tag
    {
      RetFailureToCompare,
      RetAlwaysLess,
      RetAlwaysEqual,
      RetAlwaysGreater,
      RetCanBeLessOrEqual,
      RetCanBeGreaterOrEqual,
      RetVolatileButNearlyAlwaysLess,
      RetVolatileButNearlyAlwaysGreater,
      RetAbsolutelyVolatile,     
      IfXGr1RetFail_IfGrndRetAlwEq_RetFail,
      IfXGr1RetFail_IfNotGrndRetFail,
      IfXGr1RetAlwGr_IfGrndRetAlwEq_RetCanBeGrOrEq,
      IfXGr1RetAlwGr_GrndToGRND,
      IfXGr1RetAlwGr_UpdateGRND,
      IfXGr1RetAlwGr_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq,
      IfNotGrndRetFail_IfXGrWRetFail_IfXLeWRetAlwGr_RetAlwEq,
      IfNotGrndRetFail_IfCXGrWRetFail_IfCXLeWRetAlwGr_RetAlwEq,
      IfNotGrndRetFail_WMinusXToACC_IfACCLe0RetFail,
      IfNotGrndRetFail_WMinusCXToACC_IfACCLe0RetFail,
      IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail,
      IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail,
      IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr,  
      IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr,
      IfXGrWRetAlwGr_IfXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq,
      IfCXGrWRetAlwGr_IfCXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq,
      WMinusXToACC_IfACCLe0RetAlwGr_GrndToGRND,
      WMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND,
      ACCMinusXPlus1ToACC_IfACCLe0RetAlwGr_UpdateGRND,
      ACCMinusCXPlusCToACC_IfACCLe0RetAlwGr_UpdateGRND,
      IfXGrACCPlus1RetAlwGr_IfXLeACCPlus1RetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq,
      IfCXGrACCPlusCRetAlwGr_IfCXLeACCPlusCRetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq,
      XPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL,
      CXPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL,
      XPlusWToPOSINST,
      CXPlusWToPOSINST,
      POSINSTPlusXToPOSINST,
      POSINSTPlusCXToPOSINST,
      POSINSTPlusXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL,
      POSINSTPlusCXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL,
      XMinus1ToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST,
      CXMinusCToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST,
      XMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail,
      CXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail,
      NEGINSTPlusXMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail,
      NEGINSTPlusCXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail,
      NEGINSTPlusXMinus1ToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST,
      NEGINSTPlusCXMinusCToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST,
      IfNotGrndRetFail_IfXMinus1GrPOSINSTRetFail_IfXMinus1LePOSINSTRetAlwGr_RetAlwEq,
      IfNotGrndRetFail_IfCXMinusCGrPOSINSTRetFail_IfCXMinusCLePOSINSTRetAlwGr_RetAlwEq,
      IfNotGrndRetFail_POSINSTPlus1MinusXToACC_IfACCLe0RetFail,
      IfNotGrndRetFail_POSINSTPlusCMinusCXToACC_IfACCLe0RetFail,
      XPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL,
      CXPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL,
      XPlusWToNEGINST,
      CXPlusWToNEGINST,
      NEGINSTPlusXToNEGINST,
      NEGINSTPlusCXToNEGINST,
      NEGINSTPlusXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL,
      NEGINSTPlusCXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL,
      XMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr,
      CXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr,
      POSINSTPlusXMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr,
      POSINSTPlusCXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr,
      POSINSTPlusXMinus1ToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST,
      POSINSTPlusCXMinusCToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST,
      NEGINSTPlus1MinusXToACC_IfACCLe0RetAlwGr_GrndToGRND,
      NEGINSTPlusCMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND
    }; 
  public:
    Command() {};
    ~Command() {};
    const Tag& tag() const { return _tag; };
    void setTag(const Tag& tag) 
    { 
      _tag = tag;
      DOP(_var = ULONG_MAX);
      DOP(_nativeVar = 0);
      DOP(_label = 0);
    };
    const long& coefficient() const 
    {
      CALL("coefficient() const");
      ASSERT(requiresCoefficient());
      return _coefficient; 
    };
    void setCoefficient(const long& coeff) 
    {
      CALL("setCoefficient(const long& coeff)");
      ASSERT(requiresCoefficient());
      ASSERT(coeff > 0);
      ASSERT(coeff != 1);
      _coefficient = coeff; 
    };
    ulong var() const 
    {
      CALL("var() const");
      ASSERT(requiresVar());
      return _var; 
    };
    void setVar(ulong var) 
    {
      CALL("setVar(ulong var)");
      ASSERT(requiresVar());
      _var = var; 
    };
    const void* nativeVar() const
    {
      CALL("nativeVar() const");
      ASSERT(requiresVar());
      ASSERT(_nativeVar);
      return _nativeVar;
    };
    void setNativeVar(const void* v)
    {
      CALL("setNativeVar(const void* v)");
      ASSERT(requiresVar());
      ASSERT(v);
      _nativeVar = v;
    };
    const long& weight() const 
    {
      CALL("weight() const");
      ASSERT(requiresWeight());
      return _weight; 
    };
    void setWeight(const long& w) 
    {
      CALL("setWeight(const long& w)");
      ASSERT(requiresWeight());
      _weight = w; 
    };
    const Command* label() const 
    {
      CALL("label() const");
      ASSERT(requiresLabel());
      return _label; 
    };
    void setLabel(Command* l) 
    {
      CALL("setLabel(Command* l)");
      ASSERT(requiresLabel());
      ASSERT(l > this);
      _label = l; 
    };
    static bool isTerminal(const Tag& tag);
    bool isTerminal() const { return isTerminal(tag()); };
    bool requiresVar() const;
    bool requiresCoefficient() const;
    bool requiresWeight() const;
    bool requiresLabel() const;
#ifndef NO_DEBUG_VIS
    static void setCodeBase(const Command* b) { _codeBase = b; };
    ostream& outputShort(ostream& str) const;
    ostream& output(ostream& str) const 
    {
      return output(str,"");
    };
    ostream& output(ostream& str,const char* indent) const;
    static ostream& outputCode(ostream& str,const Command* code)
    {
      return outputCode(str,"",code);
    };
    static ostream& outputCode(ostream& str,const char* indent,const Command* code);
#endif    
  private:
    Tag _tag;
    long _coefficient;
    ulong _var;
    const void* _nativeVar;
    long _weight;
    Command* _label;
#ifndef NO_DEBUG_VIS
    static const Command* _codeBase;
#endif    
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
#define DEBUG_NAMESPACE "WeightPolynomial::SpecialisedComparison::Command"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  bool WeightPolynomial::SpecialisedComparison::Command::isTerminal(const Tag& tag)
  {
    CALL("isTerminal(const Tag& tag)");
    switch (tag)
      {
      case RetFailureToCompare:
      case RetAlwaysLess:
      case RetAlwaysEqual:
      case RetAlwaysGreater:
      case RetCanBeLessOrEqual:
      case RetCanBeGreaterOrEqual:
      case RetVolatileButNearlyAlwaysLess:
      case RetVolatileButNearlyAlwaysGreater:
      case RetAbsolutelyVolatile:
      case IfXGr1RetFail_IfGrndRetAlwEq_RetFail:
      case IfXGr1RetAlwGr_IfGrndRetAlwEq_RetCanBeGrOrEq:
      case IfXGr1RetAlwGr_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      case IfNotGrndRetFail_IfXGrWRetFail_IfXLeWRetAlwGr_RetAlwEq:
      case IfNotGrndRetFail_IfCXGrWRetFail_IfCXLeWRetAlwGr_RetAlwEq:
      case IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr:
      case IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr:
      case IfXGrWRetAlwGr_IfXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      case IfCXGrWRetAlwGr_IfCXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      case IfXGrACCPlus1RetAlwGr_IfXLeACCPlus1RetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      case IfCXGrACCPlusCRetAlwGr_IfCXLeACCPlusCRetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      case XMinus1ToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST:
      case CXMinusCToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST:
      case NEGINSTPlusXMinus1ToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST:
      case NEGINSTPlusCXMinusCToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST:
      case IfNotGrndRetFail_IfXMinus1GrPOSINSTRetFail_IfXMinus1LePOSINSTRetAlwGr_RetAlwEq:
      case IfNotGrndRetFail_IfCXMinusCGrPOSINSTRetFail_IfCXMinusCLePOSINSTRetAlwGr_RetAlwEq:
      case POSINSTPlusXMinus1ToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST:
      case POSINSTPlusCXMinusCToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST:
	return true;
      default: return false;
      };
  }; 
  inline bool WeightPolynomial::SpecialisedComparison::Command::requiresVar() const
  {
    CALL("requiresVar() const");
    switch (tag())
      {
      case RetFailureToCompare:
      case RetAlwaysLess:
      case RetAlwaysEqual:
      case RetAlwaysGreater:
      case RetCanBeLessOrEqual:
      case RetCanBeGreaterOrEqual:
      case RetVolatileButNearlyAlwaysLess:
      case RetVolatileButNearlyAlwaysGreater:
      case RetAbsolutelyVolatile:  
	return false;
      default: return true;
      };
  }; 
  inline bool WeightPolynomial::SpecialisedComparison::Command::requiresCoefficient() const
  {
    CALL("requiresCoefficient() const");
    switch (tag())
      {
      case IfNotGrndRetFail_IfCXGrWRetFail_IfCXLeWRetAlwGr_RetAlwEq:
      case IfNotGrndRetFail_WMinusCXToACC_IfACCLe0RetFail:
      case IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail:
      case IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr:
      case IfCXGrWRetAlwGr_IfCXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      case WMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      case ACCMinusCXPlusCToACC_IfACCLe0RetAlwGr_UpdateGRND:
      case IfCXGrACCPlusCRetAlwGr_IfCXLeACCPlusCRetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      case CXPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case CXPlusWToPOSINST:
      case POSINSTPlusCXToPOSINST:
      case POSINSTPlusCXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case CXMinusCToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST:
      case CXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      case NEGINSTPlusCXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      case NEGINSTPlusCXMinusCToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST:
      case IfNotGrndRetFail_IfCXMinusCGrPOSINSTRetFail_IfCXMinusCLePOSINSTRetAlwGr_RetAlwEq:
      case IfNotGrndRetFail_POSINSTPlusCMinusCXToACC_IfACCLe0RetFail:
      case CXPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      case CXPlusWToNEGINST:
      case NEGINSTPlusCXToNEGINST:
      case NEGINSTPlusCXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      case CXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      case POSINSTPlusCXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      case POSINSTPlusCXMinusCToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST:
      case NEGINSTPlusCMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND:
	return true;
      default: return false;
      };
  }; 
  inline bool WeightPolynomial::SpecialisedComparison::Command::requiresWeight() const
  {
    CALL("requiresWeight() const");
    switch (tag())
      {
      case IfNotGrndRetFail_IfXGrWRetFail_IfXLeWRetAlwGr_RetAlwEq:
      case IfNotGrndRetFail_IfCXGrWRetFail_IfCXLeWRetAlwGr_RetAlwEq:
      case IfNotGrndRetFail_WMinusXToACC_IfACCLe0RetFail:
      case IfNotGrndRetFail_WMinusCXToACC_IfACCLe0RetFail:
      case IfXGrWRetAlwGr_IfXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      case IfCXGrWRetAlwGr_IfCXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      case WMinusXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      case WMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      case XPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case CXPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case XPlusWToPOSINST:
      case CXPlusWToPOSINST:
      case XPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      case CXPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      case XPlusWToNEGINST:
      case CXPlusWToNEGINST:
	return true;
      default: return false;
      };
  }; 
  inline bool WeightPolynomial::SpecialisedComparison::Command::requiresLabel() const
  {
    CALL("requiresLabel() const");
    switch (tag())
      {
      case XPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case CXPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case POSINSTPlusXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case POSINSTPlusCXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      case XPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      case CXPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      case NEGINSTPlusXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      case NEGINSTPlusCXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
	return true;
      default: return false;
      };
  }; 
}; 
#endif
