/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <cstdio>
#include "Int.hpp"
#include "Term.hpp"
#include "XML.hpp"
using namespace Shell;
bool VarList::equals (const VarList& vs) const
{
   if (isEmpty()) {
    return vs.isEmpty();
  }
  return vs.isNonEmpty() &&
         head() == vs.head() &&
         tail().equals(vs.tail());
} 
XML::Element* Var::toXML () const
{
  XML::EmptyElement* v = new XML::EmptyElement("variable");
  v->addAttribute(new XML::IntegerAttribute("number",_var));
  return v;
} 
string Var::toString () const
{
  static char tmp [20];
  tmp[0] = _var >= 0 ? 'X' : '@';
  sprintf(tmp+1,"%d",abs(_var));
  return string(tmp);
} 
string VarList::toString () const
{
  if (isEmpty()) {
    return "";
  }
  string result = head().toString();
  Iterator<Var> vs(tail());
  while (vs.hasNext()) {
    result += (string)"," + vs.next().toString();
  }
  return result;
} 
VarList::VarList (const TermList& ts)
  : Lst<Var> () 
{
  if (ts.isEmpty()) {
    return;
  }
  VarList ws(ts.tail());
  *this = VarList(ts.head().var(),ws);
} 
