/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Var__
#define __Var__
#include <string>
#include "Lst.hpp"
#include "XML.hpp"
namespace Shell {
class TermList;
class Var
{
public:
  Var () {}
  explicit Var (int v) { _var = v; }
  Var (const Var& v) { _var = v._var; }
  int toInt () const { return _var; }
  static bool lessThan (const Var& v1, const Var& v2) 
    { return v1._var < v2._var; }
  bool operator == (const Var& v) const { return _var == v._var; }
  bool operator != (const Var& v) const { return _var != v._var; }
  XML::Element* toXML () const;
  string toString () const;
  void set (int v) { ASS(v>=0);_var = v; }
private:
  int _var;
};
class VarList :
  public Lst<Var>
{
 public:
  const VarList& tail () const
    { return static_cast<const VarList&>(Lst<Var>::tail()); }
  VarList (Var head, const VarList& tail);
  VarList (const VarList& lst);
  explicit VarList (Var v); 
  explicit VarList (const TermList& ts);
  VarList ();
  bool equals (const VarList&) const;
  string toString () const;
};
class VarListList :
  public Lst<VarList>
{
 public:
  const VarListList& tail () const
    { return static_cast<const VarListList&>(Lst<VarList>::tail()); }
  VarListList (VarList head, const VarListList& tail);
  VarListList (const VarListList& lst);
  explicit VarListList (const VarList& t); 
  VarListList ();
};
inline
VarList::VarList () 
{
}
inline
VarList::VarList (const VarList& lst) 
  :
  Lst<Var> (lst) 
{
}
inline
VarList::VarList (Var v) 
  :
  Lst<Var> (v) 
{
}
inline
VarList::VarList (Var head, const VarList& tail) 
  :
  Lst<Var> (head,tail) 
{
}
inline
VarListList::VarListList () 
{
}
inline
VarListList::VarListList (const VarListList& lst) 
  :
  Lst<VarList> (lst) 
{
}
inline
VarListList::VarListList (VarList head, const VarListList& tail) 
  :
  Lst<VarList> (head,tail) 
{
}
}
#endif 
