/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Bag.hpp"
#include "CASCMode.hpp"
#include "CommandLine.hpp"
#include "BlodKorv/GlobalStopFlag.hpp" 
#include "Environment.hpp"
#include "Options.hpp"
#include "Random.hpp"
#include "Run.hpp"
#include "BlodKorv/RuntimeError.hpp"
#include "SpiderMode.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "VampireKernel/VampireKernel.hpp"
#include "XML.hpp"
#include "XML.hpp"
#include "BlodKorv/Error.hpp"
using namespace Shell;
void runtimeErrorHook ()
{
  TRACER("runtimeErrorHook");
  if (VampireKernel::currentSession()) {
    cout << "% Trying to shut down current kernel session...\n";
    VampireKernel::currentSession()->reportError();
  }
  XML::Element* err = new XML::TextElement("error",
					   "some module detected inconsistency in run time");
  err->addAttribute(new XML::TextAttribute("type","runtime"));
  Run::currentRun()->addToLog(err);
  Run::currentRun()->saveLog();
} 
int main (int argc, char* argv [])
{
  TRACER ("main");
  Random::resetSeed ();
  Run::setSignalHandlers();
  BK::RuntimeError::errorHook = runtimeErrorHook;
  Run* run = 0;
  try {
    CommandLine cl(argc,argv);
    cl.interpret();
    Options& options = *Environment::options(0);
    GlobalAllocator::setMemoryLimitInMegabytes(options.memoryLimit());
    switch (Environment::mode) {
    case MODE_VAMPIRE:
      break;
    case MODE_SPIDER:
      run = new SpiderMode(options);
      break;
    case MODE_CASC:
      run = new CASCMode(options);
      break;
    }
    run->run();
  }
  catch (AllocationFailureException& exception) {
    if (run) {
      run->logResult("allocation failure",UNKNOWN);
    }
    exception.cry(cout);
  }
  catch (VampireTerminationException& exception) {
    if (run) {
      run->logResult(exception.getReason(),exception.getResult());
    }
    exception.cry(cout);
  }
  catch (AssertionViolationException& exception) {
    if (run) {
      run->logResult("assertion_violation",UNKNOWN);
    }
    exception.cry(cout);
  }
  catch (UserErrorException& exception) {
    if (run) {
      run->logResult("user_error",UNKNOWN);
    }
    exception.cry(cout);
  }
  catch (VampireException& exception) {
    if (run) {
      run->logResult("unclassified_exception",UNKNOWN);
    }
    exception.cry(cout);
  }
  catch (VampireKernel::InputError& err) {
    cout << err.description() << "\n";
  }
  catch (const BK::Error& e) {
    cout << e.message() << "\n";
  }
  try {
    delete run;
  }
  catch (AllocationFailureException& exception) {
    exception.cry(cout);
  }
  catch (VampireTerminationException& exception) {
    exception.cry(cout);
  }
  catch (AssertionViolationException& exception) {
    exception.cry(cout);
  }
  catch (UserErrorException& exception) {
    exception.cry(cout);
  }
  catch (VampireException& exception) {
    exception.cry(cout);
  }
  catch (VampireKernel::InputError& err) {
    cout << err.description() << "\n";
  }
  catch (const BK::Error& e) {
    cout << e.message() << "\n";
  }
  catch (BK::GlobalStopFlag::DefaultException& e) {
    cout << "Internal errror\n";
  }
  return EXIT_SUCCESS;
} 
