
module type ElemDB = 
  sig
    type t
    val compare : t -> t -> int
    val assign_fast_key : t -> int -> unit
  end

module type AbstDB =
  sig   
    type elem
    type abstDB 
    val create      : unit -> abstDB 
    val create_name : string -> abstDB
    val mem         : elem -> abstDB -> bool 
    val add         : elem -> abstDB -> abstDB
    val add_ref     : elem -> abstDB ref -> elem
    val remove      : elem -> abstDB -> abstDB 
    val find        : elem -> abstDB -> elem
    val size        : abstDB -> int
    val map         : (elem -> elem)-> abstDB -> abstDB  
    val fold        : (elem -> 'a -> 'a) -> abstDB -> 'a -> 'a
    val iter        : (elem -> unit) -> abstDB -> unit
    val to_string   : (elem -> string) -> string -> abstDB ->string
    val get_name    : abstDB -> string
(*debug*)
    val get_greatest_key : abstDB -> int
  end	
      

module Make(El : ElemDB) =
  struct
    type   elem  = El.t    
    module BasicElem = 
      struct
	type t = El.t
	let compare = El.compare
      end
    module BasicAbstDB =  AbstDB.Make (BasicElem) 
  
    type abstDB = {db : BasicAbstDB.abstDB; greatest_key : int }

    let create () = {db = BasicAbstDB.create (); greatest_key=0}
    let create_name name = {db = BasicAbstDB.create_name name; greatest_key=0}
    let get_name elem_db  = BasicAbstDB.get_name elem_db.db
    let mem elem elem_db  = BasicAbstDB.mem elem elem_db.db
    let find elem elem_db = BasicAbstDB.find elem elem_db.db
    let size elem_db = BasicAbstDB.size elem_db.db   
    let fold f elem_db a = BasicAbstDB.fold f elem_db.db a

    let iter f elem_db = BasicAbstDB.iter f elem_db.db

   (*add_ref is diff from abstDB*)

    let add_ref elem elem_db_ref = 
      try (find elem !elem_db_ref)
      with
	Not_found-> 
	  let new_greatest_key = ((!elem_db_ref).greatest_key + 1) in
	  let ()=(El.assign_fast_key elem new_greatest_key) in          
	  elem_db_ref:={db =(BasicAbstDB.add elem (!elem_db_ref).db); 
			greatest_key = new_greatest_key};
	  elem
	    
(*since add_ref is diff from abstDB, so add also*)
    let add elem elem_db =
      let elem_db_ref = ref elem_db in
      let _= add_ref elem elem_db_ref in
      !elem_db_ref
	
  (*note that remove does not decrease the greatest_key*)   	   
    let remove elem elem_db =   
      {db   = (BasicAbstDB.remove elem elem_db.db); 
       greatest_key = elem_db.greatest_key }
	
    let map f elem_db = 
      {db   = (BasicAbstDB.map f elem_db.db); 
       greatest_key = elem_db.greatest_key}
	
    let to_string el_to_str separator elem_db =
      BasicAbstDB.to_string el_to_str separator elem_db.db

(*debug*)
    let  get_greatest_key elem_db = elem_db.greatest_key

end


