open Lib
type clause  = Clause.clause
type literal = Clause.literal
type term_db = TermDB.termDB
type clause_db = ClauseAssignDB.clauseDB
(*
type  statistics = 
    {num_of_dismatch_blockings : int;
     num_of_non_proper_inst    : int }
*)


let num_of_dismatch_blockings = ref 0 
let num_of_non_proper_inst = ref 0
let num_of_duplicates = ref 0

(* literals l1 l2 are in c1 and c2 *)
(* compl_l1 is ugly *)
let resolution  c1 l1 compl_l1 c_list2 l2 term_db_ref = 
(*  let compl_l1 = Term.compl_lit l1 in*)
  let mgu = Unif.unify_bterms (1,compl_l1) (2,l2) in
  let new_litlist1 = 
    Clause.find_all (fun lit -> not(l1 == lit)) c1 
    in 
  let f rest c2 = 
    let new_litlist2 = 
      Clause.find_all (fun lit -> not(l2 == lit)) c2 in 
    let conclusion = Clause.normalise_blitlist_list 
	term_db_ref mgu [(1,new_litlist1);(2,new_litlist2)] in
    Clause.assign_resolution_history conclusion [c1;c2] [l1;l2];
    conclusion::rest  in 
  List.fold_left f [] c_list2     

(* factors and removes all  repeated l1's in the clause *)
let factoring c l1 l2 term_db_ref =
  if l1==l2 then 
    let new_litlist = 
      l1::(Clause.find_all (fun lit -> not(l1 == lit)) c) in
    let conclusion = Clause.create new_litlist in
    Clause.assign_factoring_history conclusion c [l1;l2];
    conclusion
  else    
    let mgu =  Unif.unify_bterms (1,l1) (1,l2) in
    let new_litlist = 
      Clause.find_all (fun lit -> not(l1 == lit)) c
    in 
    let conclusion = Clause.normalise_b_litlist term_db_ref mgu (1,new_litlist) in
    Clause.assign_factoring_history conclusion c [l1;l2];
    conclusion



(* could be more efficient but messier

(* literals l1 l2 are already CUT  from  c1 and c2 *)
   
   let resolution  c1 l1 c2 l2 term_db_ref = 
   let compl_l1 = Term.compl_lit l1 in
   let mgu = Unif.unify_bterms (1,l1) (2,l2) in
   Clause.normalise_bclause_list  
   term_db_ref mgu [(1,c1);(2,c2)]

(* literals l1 l2 are already CUT from  c *)

let factoring c l1 l2 term_db_ref =
  let mgu =  Unif.unify_bterms (1,l1) (1,l2) in
  Clause.normalise_bclause (1,c) mgu term_db_ref

*)

(************** instantiation *****************)
(****************old version, not correct*********)
(*
let is_not_redundant_inst bound bsubst clause = 
(*  if (SubstBound.is_proper_instantiator bsubst bound)     *)
(*  then proper inst checked first because applies to many clauses with the same lit*)
(*  out_str_debug  "\n-------Constr Check-------\n";*)
  try
    let dismatching = Clause.get_dismatching clause in
    out_str_debug (
    "Inst Clause: "^(Clause.to_string clause)
    ^"  Bound: "^(string_of_int bound)^"\n"      
    ^"Constraint: "^"["^(Dismatching.constr_list_to_string dismatching)^"]"^"\n"
    ^"Subs_to_check: "^(SubstBound.to_string bsubst)^"\n"); 
    if (Dismatching.check_constr_list bound bsubst dismatching)
    then
      (out_str_debug "Constr. is Satisfied \n";
       true)
    else 
      (out_str_debug "Constr. is NOT Satisfied \n";
       false) 
  with Clause.Dismatching_undef -> 
    (*out_str_debug "Constr. is empty";*) true 
(*  else false*)

(* instantiates adding dismatching constr to the parent--clause *)
let instantiate_clause term_db_ref bound bsubst clause =
  let concl_clause = Clause.apply_bsubst term_db_ref bsubst (bound,clause) in
  let new_constr = Dismatching.create_constr term_db_ref bound  bsubst in
  try    
    let dismatching = Clause.get_dismatching clause in
    Clause.assign_dismatching (Dismatching.add_constr dismatching new_constr) clause;
    concl_clause
  with 
    Clause.Dismatching_undef -> 
      let empty_dism = Dismatching.create_constr_list () in
      let new_dism = Dismatching.add_constr empty_dism new_constr in
      Clause.assign_dismatching new_dism clause;
      concl_clause

(* {num_of_dismatch_blockings = ref 0;*)
(*     num_of_non_proper_inst = ref 0}*)

let instantiation term_db_ref c1 l1 compl_l1 c_list2 l2 =
  let mgu = Unif.unify_bterms (1,compl_l1) (2,l2) in
  let conc1 = 
    if (SubstBound.is_proper_instantiator mgu 1) 
    then  
      if (is_not_redundant_inst 1 mgu c1)
      then 
	[(instantiate_clause term_db_ref 1 mgu c1)]
      else 
	(num_of_dismatch_blockings := (!num_of_dismatch_blockings +1); 
	 [])
    else
      (num_of_non_proper_inst :=(!num_of_non_proper_inst +1); 
       [])
  in    
  let conc2 =
    if (SubstBound.is_proper_instantiator mgu 2) then    
      let f rest clause = 
	if (is_not_redundant_inst 2 mgu clause)
	then 
	  (instantiate_clause term_db_ref 2 mgu clause)::rest
	else 
	  (num_of_dismatch_blockings :=(!num_of_dismatch_blockings +1); 
	   rest)
      in
      List.fold_left f [] c_list2
    else
      (num_of_non_proper_inst :=(!num_of_non_proper_inst +1); 
       [])
  in conc1@conc2

*)
(*******************old version end*************)

(*************new version: constr checked on normalized substitutions****)
(***************************************************************************)

let is_not_redundant_inst_norm subst_norm clause =
(*   out_str_debug 
    ("---------Constr Check-------\n");    *)
  try
    let dismatching = Clause.get_dismatching clause in
   (* out_str_debug 
      ("Inst Clause: "^(Clause.to_string clause)
       ^"Constraint: "^"["^(Dismatching.constr_list_to_string dismatching)^"]"^"\n"
       ^"Subs_to_check: "^(Subst.to_string subst_norm)^"\n"); *)
    if (Dismatching.check_constr_norm_list subst_norm dismatching)
      (*(Dismatching.check_constr_feature_list subst_norm dismatching)*)
    then
    ((* out_str_debug "Constr. is Satisfied \n";*)
(* we don't need to add all subt_norm but only vars that occurred in mgu *)
      let new_constr = Dismatching.create_constr_norm subst_norm in
      Clause.assign_dismatching 
	(Dismatching.add_constr dismatching new_constr) clause;
      	(*(Dismatching.add_constr_feature_list dismatching new_constr);*)
      true)
    else
      ((*out_str_debug "Constr. is NOT Satisfied \n";*)
       false) 
  with Clause.Dismatching_undef -> 
  (let new_dismatching =
      (Dismatching.create_constr_list ()) in
  (* let new_dismatching =
     (Dismatching.create_constr_feature_list ()) in*)
   let new_constr = Dismatching.create_constr_norm subst_norm in
    Clause.assign_dismatching 
      (Dismatching.add_constr new_dismatching new_constr) clause;
  (* (Dismatching.add_constr_feature_list new_dismatching new_constr);   
   Clause.assign_dismatching new_dismatching clause;*)
  (*out_str_debug "Constr. is empty\n";*)
    true)

(*out_str_debug "Constr. is empty";*)  
(*  else false*)

  
let dismatching_string clause =   
  try 
   "["^(Dismatching.constr_list_to_string (Clause.get_dismatching clause))^"]"
  with
    Clause.Dismatching_undef -> "[]"


exception Main_concl_redundant

(* instantiation first check duplicates then dismatching constraints*)

let instantiation_norm_dc dismatch_switch term_db_ref clause_db_ref c1 l1 compl_l1 c_list2 l2 =
  let mgu = Unif.unify_bterms (1,compl_l1) (2,l2) in
(*debug*)
(*  out_str_debug ("Main Clause:"^(Clause.to_string c1)^"\n"
		  ^"Constr: "^(dismatching_string c1)^"\n" 
		  ^"Sel Lit: "^(Term.to_string l1)^"\n"); *)
  try 
    let conc1 = 
      if (SubstBound.is_proper_instantiator mgu 1) 
      then  
	let (inst_clause,subst_norm) = 
	  (Clause.apply_bsubst_norm_subst term_db_ref mgu 1 c1)
	in
	if (ClauseAssignDB.mem inst_clause !clause_db_ref)
	then 
	  (num_of_duplicates:=!num_of_duplicates+1; 
(*	   out_str_debug ("Clause is already In DB: "
			  ^(Clause.to_string inst_clause)^"\n");*)
	   raise Main_concl_redundant)
	else
	  if  (not dismatch_switch) || (is_not_redundant_inst_norm subst_norm c1)
	  then 
	    let added_clause = 
	      ClauseAssignDB.add_ref inst_clause clause_db_ref;
	    in	
	    Clause.assign_when_born ((Clause.when_born c1)+1) added_clause;
	    Clause.assign_activity ((Clause.get_activity c1)+1) c1;
	    Clause.add_child c1 added_clause;
	    [added_clause]
	  else 
	    (num_of_dismatch_blockings := (!num_of_dismatch_blockings +1); 
	     raise Main_concl_redundant)
      else
	(num_of_non_proper_inst :=(!num_of_non_proper_inst +1); 
(*       out_str_debug ("Non-proper Inst Main\n");*)
	 [])
    in    
    let conc2 =
    if ((not (Term.get_fun_bool_param Term.in_unif_index l1))& 
      (SubstBound.is_proper_instantiator mgu 2)) then    
      let f rest clause =
(*debug*)
(*	out_str_debug  ("Side Clause:"^(Clause.to_string clause)^"\n"
			^"Constr: "^(dismatching_string clause)^"\n" 
			^"Sel Lit: "^(Term.to_string l2)^"\n");*)
	let (inst_clause,subst_norm) = 
	  Clause.apply_bsubst_norm_subst term_db_ref mgu 2 clause 
	in
	if (ClauseAssignDB.mem inst_clause !clause_db_ref)
	then (num_of_duplicates:=!num_of_duplicates+1; 
(*	      out_str_debug ("Clause is already In DB: "
			     ^(Clause.to_string inst_clause)^"\n");*)
		rest)
	else
	  if (not dismatch_switch) || (is_not_redundant_inst_norm subst_norm clause)
	  then 
	    let added_clause = 
	      ClauseAssignDB.add_ref inst_clause clause_db_ref in
	    Clause.assign_when_born ((Clause.when_born clause)+1) added_clause;
	    Clause.assign_activity ((Clause.get_activity clause)+1) clause;
	    Clause.add_child clause added_clause;
	    added_clause::rest
	  else 
	    (num_of_dismatch_blockings :=(!num_of_dismatch_blockings +1); 
(*	     out_str_debug ("Dismatching \n");*)
	     rest)	  
      in
      List.fold_left f [] c_list2
    else
      (num_of_non_proper_inst :=(!num_of_non_proper_inst +1);
(* debug*)   
     (*  (if (Term.get_fun_bool_param Term.in_unif_index l1) 
       then 	 
	 out_str ("Side is In Unif Index: "^(Term.to_string l1)^"\n")
       else
	 out_str_debug ("Non-proper Inst Side\n")
       );*)
       [])
    in 
    let concl_list = conc1@conc2 in
(*   out_str_debug 
      ("\n Conclusions:\n"^(Clause.clause_list_to_string concl_list)^"\n"
      ^"------------------------------------------------\n");*)
    concl_list
  with 
    Main_concl_redundant -> 
(*      out_str_debug 
	(" ---------Main_concl_redundant ----------\n");*)
      []	      


(* instantiation first check dismatching constraints then duplicates *)

let instantiation_norm_cd term_db_ref clause_db_ref c1 l1 compl_l1 c_list2 l2 =
  let mgu = Unif.unify_bterms (1,compl_l1) (2,l2) in
(*debug*)
  (*out_str_debug ("Main Clause:"^(Clause.to_string c1)^"\n"
		  ^"Constr: "^(dismatching_string c1)^"\n" 
		  ^"Sel Lit: "^(Term.to_string l1)^"\n"); 
 *)
 try 
    let conc1 = 
      if (SubstBound.is_proper_instantiator mgu 1) 
      then  
	let (inst_clause,subst_norm) = 
	  (Clause.apply_bsubst_norm_subst term_db_ref mgu 1 c1)
	in
	if (is_not_redundant_inst_norm subst_norm c1)
	then 
	  if (ClauseAssignDB.mem inst_clause !clause_db_ref)
	  then 
	    (num_of_duplicates:=!num_of_duplicates+1; 
	     (*out_str_debug ("Clause is already In DB: "
	       ^(Clause.to_string inst_clause)^"\n");*)
	     raise Main_concl_redundant)
	  else
	    let added_clause = 
	      ClauseAssignDB.add_ref inst_clause clause_db_ref in
	    Clause.assign_when_born ((Clause.when_born c1)+1) added_clause;
	    [added_clause]	 
	else 
	  (num_of_dismatch_blockings := (!num_of_dismatch_blockings +1); 
	   raise Main_concl_redundant)
      else
	(num_of_non_proper_inst :=(!num_of_non_proper_inst +1); 
	 (*out_str_debug ("Non-proper Inst Main\n");*)
	 [])
    in    
    let conc2 =
    if (SubstBound.is_proper_instantiator mgu 2) then    
      let f rest clause =
(*debug*)
	(*out_str_debug  ("Side Clause:"^(Clause.to_string clause)^"\n"
			^"Constr: "^(dismatching_string clause)^"\n" 
			^"Sel Lit: "^(Term.to_string l2)^"\n"); *)
	let (inst_clause,subst_norm) = 
	  Clause.apply_bsubst_norm_subst term_db_ref mgu 2 clause 
	in
	if (is_not_redundant_inst_norm subst_norm clause)
	then 
	  if (ClauseAssignDB.mem inst_clause !clause_db_ref)
	  then (num_of_duplicates:=!num_of_duplicates+1; 
		(* out_str_debug ("Clause is already In DB: "
		   ^(Clause.to_string inst_clause)^"\n");*)
		rest)
	  else
	    let added_clause = 
	      ClauseAssignDB.add_ref inst_clause clause_db_ref in
	    Clause.assign_when_born ((Clause.when_born clause)+1) added_clause;
	    added_clause::rest	
	else 
	  (num_of_dismatch_blockings :=(!num_of_dismatch_blockings +1); 
	   rest)	  
      in
      List.fold_left f [] c_list2
    else
      (num_of_non_proper_inst :=(!num_of_non_proper_inst +1); 
       (*out_str_debug ("Non-proper Inst Side\n");*)
       [])
    in 
    let concl_list = conc1@conc2 in
    (*out_str_debug 
      ("\n Conclusions:\n"^(Clause.clause_list_to_string concl_list)^"\n"
      ^"------------------------------------------------\n");*)
    concl_list
  with 
    Main_concl_redundant -> 
(*      out_str_debug 
	(" ---------Main_concl_redundant ----------\n");*)
      []	      
	
let instantiation_norm = 
  instantiation_norm_dc
(* instantiation first check dismatching constraints then duplicates *)
(* instantiation_norm_cd *)
