open Lib
  
let proof = false 
    
type clause = Clause.clause
type lit = Term.literal
type term = Term.term
type symbol = Symbol.symbol  
type tmp = bool
(*type prop_var = PropSolver.var *)
type prop_lit = PropSolver.lit


type pass_queue_cmp = clause -> clause -> int

(* below binded by real function *)



module type InstParam =
  sig
    val inst_lit_sel_fun : 
	(lit -> bool) -> clause -> term
    val solver_per_active     : int
    val solver_per_clauses    : int
    val inst_pass_queue1_cmp  :  pass_queue_cmp
    val inst_pass_queue2_cmp  :  pass_queue_cmp
    val inst_pass_queue1_mult : int
    val inst_pass_queue2_mult : int
    val dismatching           : bool
    val eager_unprocessed_to_passive : bool
    val ground_splitting_input : bool
    val ground_splitting_full  : bool
    val learning_loop_flag    : bool
    val learning_start        : int
    val learning_factor       : int
  end
    
module type Instantiation =
  sig
    val start_instantiation : unit -> unit
  end

module Make(InstParam:InstParam)=
  struct       
    exception Unsatisfiable
    exception Satisfiable
    exception DontKnow

    let selection_fun         = InstParam.inst_lit_sel_fun

    let solver_per_active     = InstParam.solver_per_active
    let solver_per_clauses    = InstParam.solver_per_clauses 
                                (*10*solver_per_active   *)
    let inst_pass_queue1_cmp  = InstParam.inst_pass_queue1_cmp 
    let inst_pass_queue2_cmp  = InstParam.inst_pass_queue2_cmp  
    let inst_pass_queue1_mult = InstParam.inst_pass_queue1_mult 
    let inst_pass_queue2_mult = InstParam.inst_pass_queue2_mult 
    let dismatch_switch       = InstParam.dismatching
    let ground_splitting_input = InstParam.ground_splitting_input
    let ground_splitting_full  = InstParam.ground_splitting_full
    let eager_unprocessed_to_passive = InstParam.eager_unprocessed_to_passive
    let learning_loop_flag    = InstParam.learning_loop_flag
    let learning_start        = InstParam.learning_start
    let learning_factor       = InstParam.learning_factor

(***********************selection function*************)

(*
  let selection_fun  = Selection.lit_var_neg_big 
  let ()=add_param_str ("Selection Selection.lit_var_neg_big \n")
*)

(*
  let selection_fun  = Selection.lit_neg_var_big 
  let ()=add_param_str ("Selection Selection.lit_neg_var_big \n")
*)

(*
let selection_fun  = Selection.lit_var_neg_shallow
let ()= add_param_str("Selection Selection.lit_var_neg_sh \n")
*)


(* let selection_fun  = Selection.lit_var_big_neg *)
(* let ()= add_param_str("Selection Selection.lit_var_big_neg\n") *)

(* good!*)
(* let selection_fun  = Selection.lit_var_sh_neg 
   let ()=add_param_str ("Selection Selection.lit_var_sh_neg \n") 
*)
(*
let selection_fun  = Selection.lit_var_big_neg     
let ()=add_param_str ("Selection Selection.lit_var_big_neg\n") 
*)
(*
let selection_fun  = Selection.lit_neg_var_sh
let ()=add_param_str ("Selection Selection.lit_neg_var_sh  \n") 
*)
(*
let selection_fun  = Selection.lit_neg_big_var
let ()=add_param_str ("Selection Selection.lit_neg_big_var  \n") 
*)
(*
let selection_fun  = Selection.lit_pos_var_sh
let ()=add_param_str ("Selection Selection.lit_pos_var_sh  \n") 
*)


(* (* the best so far *) 
   let selection_fun  = Selection.lit_pos_big_var
   let ()=add_param_str ("Selection Selection.lit_pos_big_var  \n") 
*)

(*
let selection_fun  = Selection.lit_big_var_neg 
let ()=add_param_str ("Selection Selection.lit_big_var_neg \n") 
*)

(*
let selection_fun  = Selection.lit_big_var_pos 
let ()=add_param_str ("Selection Selection.lit_big_var_pos \n") 
*)
(*
let selection_fun  = Selection.lit_big_pos_var 
let ()=add_param_str ("Selection Selection.lit_big_pos_var \n") 
*)
(*
let selection_fun  = Selection.lit_big_neg_var 
let ()=add_param_str ("Selection Selection.lit_big_neg_var \n") 
*)


(*let ()=add_param_str("Duplicates removal; Dismatching Constr. with No Features; \n") *)

(*
let ()=add_param_str("Eager_unprocessed_to_passive = "
		     ^string_of_bool (!eager_unprocessed_to_passive)^"\n")
*)

(* let selection_fun  = Selection.lit_neg_gr_shallow *)

(*

(*ugly*)
    let neg_symb       = Parsed_input_to_db.neg_symb
    let bot_symb       = Parsed_input_to_db.bot_symb      
 
*)	

    
    let init_clause_list_ref = Parsed_input_to_db.clause_list_ref
    let clause_db_ref = ref (ClauseAssignDB.create_name "Instantiation_Clauses_DB")
  

    let symbol_db_ref  = Parsed_input_to_db.symbol_db_ref
    let term_db_ref    = Parsed_input_to_db.term_db_ref

    let neg_symb       = Symbol.symb_neg
    let bot_symb       = Symbol.symb_bot
    let bot_term       = Parsed_input_to_db.bot_term      

(* simple passive is just a list and *)
(* only is used in the instantiation_exhaustive_loop*)
(*  *)
    
    let simple_passive_ref = ref []

(* unprocessed is a list of newly generated clauses *)
(* we cannot put them to passive since some truth val of some var *)
(* can be not defined at this stage *)

    let unprocessed_ref    = ref []    

(*for statistics*)

    let num_of_splits      = ref 0
    let num_of_split_atoms = ref 0
    let num_of_instantiation_iter  = ref 0
    let num_of_solver_calls        = ref 0
(*let num_in_passive        = ref 0     *)
    let num_in_simple_passive = ref 0
    let num_in_passive        = ref 0
    let num_in_active         = ref 0     
    let num_moves_active_passive = ref 0 
    let num_of_duplicates = ref 0
    let num_of_learning_restarts = ref 0
    let num_in_unprocessed       = ref 0 
    let num_prop_subs_resolution = ref 0
    let num_child_elim           = ref 0 
    let num_prop_implied         = ref 0 
    let max_clause_activity      = ref 0
    let max_lit_activity         = ref 0
    let num_prop_subsumption     = ref 0
    let subs_limit               = ref 0
    let lit_activity_threshold   = ref 200
    let lit_activity_check_move  = ref 0
    let inst_solver_thershold   = ref 100
(*    let parent_same_sel          = ref 0*)
(*    let simplfy_threshold        = ref 1000*)

    let out_statistics () = 
      out_str 
	("num_of_splits: "
	 ^(string_of_int !num_of_splits)^"\n"
	 ^"num_of_atoms: "
	 ^(string_of_int !num_of_split_atoms)^"\n"
	 ^"num_of_inst_loop_iterations: "
	 ^(string_of_int !num_of_instantiation_iter)^"\n"
	 ^"num_of_clauses_in_db: "
	 ^(string_of_int (ClauseAssignDB.size !clause_db_ref))^"\n"
(*  ^"num_in_simple_passive: "
    ^(string_of_int !num_in_simple_passive)^"\n"*)
	  ^"num_in_passive: "
	  ^(string_of_int !num_in_passive)^"\n"
	  ^"num_in_active: "
	  ^(string_of_int !num_in_active)^"\n"
	  ^"num_in_unprocessed: "
	  ^(string_of_int !num_in_unprocessed)^"\n"
	  ^"num_of_solver_calls: "      
	  ^(string_of_int !num_of_solver_calls)^"\n"
	  ^"num_of_moves_active_passive: "
	  ^(string_of_int !num_moves_active_passive)^"\n"   
	  ^"num_of_dismatch_blockings: " 
	  ^(string_of_int !Inference_rules.num_of_dismatch_blockings)^"\n"
	  ^"num_of_non_proper_inst: "
	  ^(string_of_int !Inference_rules.num_of_non_proper_inst)^"\n"
	  ^"num_of_learning_restarts: "
	  ^(string_of_int !num_of_learning_restarts)^"\n"
	  ^"num_of_simplified_duplicates: "    
	  ^(string_of_int !Inference_rules.num_of_duplicates)^"\n"
	  ^"max_clause_activity: "
	  ^(string_of_int !max_clause_activity)^"\n"
	  ^"num_child_elim: "
	  ^(string_of_int !num_child_elim)^"\n"
	(*  ^"num_prop_implied: "
	  ^(string_of_int !num_prop_implied)^"\n"*)
	  ^"num_prop_subsumption: "
	  ^(string_of_int !num_prop_subsumption)^"\n"
(* ^"num_prop_subs_resolution: "						      ^(string_of_int !num_prop_subs_resolution)^"\n"*)
	  ^"max_lit_activity: "
	  ^(string_of_int !max_lit_activity)^"\n"
	  ^"lit_activity_check_move: "
	  ^(string_of_int !lit_activity_check_move)^"\n"
     (* ^"parent_same_sel: "
      ^(string_of_int !parent_same_sel)^"\n"*)
(*  ^(string_of_int !num_of_duplicates)^"\n"*)
     )
    



    let out_proof s = 
      if proof then out_str s
      else ()
    
(********** complementary literal  *******)
let compl_lit literal = 
  match literal with 
  | Term.Fun(sym,args,_) -> 
      if sym == neg_symb 
      then List.hd (Term.arg_to_list args)  
      else Term.create_fun_term neg_symb [literal]
  |_-> failwith "term: compl_lit it is not a literal" 
 




(******************get term for grounding**************)
  
(* get_term_for_grounding *)
(* gets term as constant with maximal occurences in all clauses *)
   
(* const_ass_list -- assosiation *)
(* list of constants with number of occurences*)
    
type const_ass_list = symbol * (int ref) list 
  
(* finds constant with max value in ass list*)

let rec find_const_max' var_num const_ass_list =
  let (var,num_ref)=var_num in
  match const_ass_list with 
  | h::tl -> let (v_next,num_next_ref) = h in
    if !num_next_ref > !num_ref 
    then find_const_max' h tl
    else find_const_max' var_num tl
  |[] -> var_num
      
let find_const_max const_ass_list =
  match const_ass_list with 
  | h::tl -> find_const_max' h tl
  | [] ->  raise Not_found

(* gets an const_ass_list an returnes new one with added occurences of *)
(* constants according to term *)
let rec get_symb_ass_list_term const_ass_list term = 
  if (Term.is_constant term) 
  then
    try 
      let num_of_occ_ref = List.assq term const_ass_list in
      num_of_occ_ref := !num_of_occ_ref + 1;
      const_ass_list
    with 
      Not_found ->
	(term, ref 1)::const_ass_list
  else			 
    match term with 
    |Term.Fun(symb,t_args,_) -> 
	let f curr_ass_list t = get_symb_ass_list_term curr_ass_list t in
	Term.arg_fold_left f const_ass_list t_args
    |Term.Var(_,_) -> const_ass_list
	  
(* the same as above but for clause *)
let get_symb_ass_list_clause const_ass_list clause =
  List.fold_left 
    get_symb_ass_list_term const_ass_list (Clause.get_literals clause)

let get_term_for_grounding () = 
  let  symb_ass_list =
    List.fold_left get_symb_ass_list_clause [] !init_clause_list_ref in
  try 
    let (max_const,num_ref) = find_const_max symb_ass_list in
    (*out_str_debug ("max constant "^(Term.to_string max_const)
               	   ^" num of occurences "^(string_of_int !num_ref)^"\n");*)
    max_const
  with 
    Not_found -> 
      bot_term


(**********end get grounding term******************)

(********* propositional interpretation ************)

type prop_var_entry = 
    {var_id              : int param;
     prop_var            : prop_lit param; 
     prop_neg_var        : prop_lit param;
(* If  truth_val = Def(Any) the we assume that sel_clauses=[] *)
(* So if we select a lit the we should change Def(Any) *)
    mutable truth_val       : PropSolver.lit_val param; 
(* list of clauses for which selection is based on this var *)
     mutable sel_clauses    : (clause list);
(* used only for output of model*)
     mutable ground_term    : lit param;
     mutable in_solver_sim  : bool;
     mutable pos_activity   : int;
     mutable neg_activity   : int
   } 


let var_entry_to_string solver prop_var_entry = 
  let var_solver_val_str =
    match prop_var_entry.prop_var with
    |Def(prop_var) ->
(* need to check if solver was called at all before *)
	(match prop_var_entry.truth_val with
	|Def(_) -> 
	    PropSolver.lit_val_to_string (PropSolver.lit_val solver prop_var) 
	|_-> "Never checked with Solver" 
	)
    |_-> " Var is undef "
  in
  "var_id: "^(param_to_string string_of_int prop_var_entry.var_id)
  ^"; current truth_val: "
  ^(param_to_string PropSolver.lit_val_to_string prop_var_entry.truth_val)
  ^"; Solver truth val: "^(var_solver_val_str)
  ^"; \n Ground term: "^(param_to_string Term.to_string prop_var_entry.ground_term)
  ^"; Pos Activity: "^(string_of_int prop_var_entry.pos_activity)
  ^"; Neg Activity: "^(string_of_int prop_var_entry.neg_activity)

let var_entry_sel_to_string solver prop_var_entry = 
  try
    let clause_list_str =  
      let get_cl_str rest clause = 
	let sel_lit = Clause.get_inst_sel_lit clause in
	let sel_str = "\n Selected:  "^(Term.to_string sel_lit)^"\n" in
	let clause_str = " In clause: "^(Clause.to_string clause) in
	rest^sel_str^clause_str 
      in			       
      List.fold_left get_cl_str "" prop_var_entry.sel_clauses
    in
    (var_entry_to_string solver prop_var_entry)^clause_list_str
  with
    Clause.Inst_sel_lit_undef -> 
      failwith "var_entry_sel_cl_to_string: Sel_lits_undef "


let empty_var_entry = 
  {
   var_id        = Undef;
   prop_var      = Undef;
   prop_neg_var  = Undef;
   truth_val     = Undef;
   sel_clauses   = [];
   ground_term   = Undef;
   in_solver_sim = false;
   pos_activity  = 0;
   neg_activity  = 0
 }

let create_var_entry solver var_id ground_term = 
  {
   var_id       = Def(var_id);
   prop_var     = Def(PropSolver.create_lit solver PropSolver.Pos var_id); 
   prop_neg_var = Def(PropSolver.create_lit solver PropSolver.Neg var_id); 
   truth_val    = Undef;
   sel_clauses  = [];
   ground_term  = Def(ground_term);
   in_solver_sim = false;
   pos_activity  = 0;
   neg_activity  = 0
 }


let var_table_initsize = 1001

let var_table = TableArray.create var_table_initsize empty_var_entry

let model_to_string_gen entry_to_str =
  let init_str   = "\n ------------Model-----------\n" in
  let final_str = "\n ----------END Model----------\n" in
  let main_str=
    let get_entry_str rest entry = 
      rest ^ "\n---------------------------------\n"
      ^(entry_to_str entry)^"\n" in
    TableArray.fold_left get_entry_str "" var_table  
  in 
  init_str^main_str^final_str


let model_to_string solver  =  model_to_string_gen  (var_entry_to_string solver)

let model_sel_to_string solver =  
  model_to_string_gen  (var_entry_sel_to_string solver)

let get_prop_key_assign solver atom = 
  try Term.get_prop_key atom
  with Term.Prop_key_undef ->
    (let new_key = TableArray.get_next_key var_table in
(* var_id > 0 *)
    let var_id  = TableArray.key_to_int new_key + 1 in
    let var_entry = create_var_entry solver var_id atom in
    TableArray.assign var_table new_key var_entry;
    Term.assign_prop_key new_key atom;
    (if (Term.is_ground atom) 
    then  Term.assign_prop_gr_key new_key atom
    else ());
    new_key)
    

(*  a*)
let get_prop_gr_key_assign solver gr_by term = 
  try Term.get_prop_gr_key term
  with Term.Prop_gr_key_undef ->
    let gr_term = 
      try Term.get_grounding term      
      with
	Term.Term_grounding_undef ->
	  let new_gr_t = Subst.grounding term_db_ref gr_by term in 
	  Term.assign_grounding new_gr_t term;
	  new_gr_t
    in
    try 
      let new_key = Term.get_prop_gr_key gr_term in
      Term.assign_prop_gr_key new_key term;
      new_key
    with Term.Prop_gr_key_undef ->	
      let new_key = TableArray.get_next_key var_table in
(* var_id > 0 *)
      let var_id  = TableArray.key_to_int new_key + 1 in
      let var_entry = create_var_entry solver var_id gr_term in
      TableArray.assign var_table new_key var_entry;
      Term.assign_prop_gr_key new_key term;
      Term.assign_prop_gr_key new_key gr_term;
      Term.assign_prop_key new_key gr_term;
      new_key

(* adds literal without grounding to propositional solver and to var_table *)
let get_prop_lit_assign solver lit = 
  let atom = Term.get_atom lit in 
  let atom_prop_key = get_prop_key_assign solver atom in
  let prop_var_entry = TableArray.get var_table atom_prop_key in
  let def_lit = 
    if (Term.is_neg_lit lit) 
    then      
      prop_var_entry.prop_neg_var 
    else
     prop_var_entry.prop_var 
  in
  match def_lit with 
  |Def(lit) -> lit 
  | _ -> failwith "Insatntiation get_prop_lit: lit is undefind"
  
(*returns prop literal;  assume that prop key is def. *)
let get_prop_lit lit = 
  let atom = Term.get_atom lit in 
  let atom_prop_key = Term.get_prop_key atom in
  let prop_var_entry = TableArray.get var_table atom_prop_key in
  let def_lit = 
    if (Term.is_neg_lit lit) 
    then      
      prop_var_entry.prop_neg_var 
    else
     prop_var_entry.prop_var 
  in
  match def_lit with 
  |Def(lit) -> lit 
  | _ -> failwith "Insatntiation get_prop_lit: lit is undefind"
  

(*returns complementary prop literal;  assume that prop key is def. *)
    let get_prop_compl_lit lit = 
      let atom = Term.get_atom lit in 
      let atom_prop_key = Term.get_prop_key atom in
      let prop_var_entry = TableArray.get var_table atom_prop_key in
      let def_lit = 
	if (Term.is_neg_lit lit) 
	then      
	  prop_var_entry.prop_var 
	else
	  prop_var_entry.prop_neg_var 
      in
      match def_lit with 
      |Def(lit) -> lit 
      | _ -> failwith "Insatntiation get_prop_compl_lit: lit is undefind"
	    

let get_prop_gr_lit lit =
  let atom = Term.get_atom lit in 
  let atom_prop_gr_key = Term.get_prop_gr_key atom in
  let prop_var_entry = TableArray.get var_table atom_prop_gr_key in
  let def_lit = 
    if (Term.is_neg_lit lit) 
    then      
      prop_var_entry.prop_neg_var 
    else
     prop_var_entry.prop_var 
  in
  match def_lit with 
  |Def(lit) -> lit 
  | _ -> failwith "Insatntiation get_prop_gr_lit: lit is undefind"

(* adds literal after grounding to propositional solver and to var_table *)
let get_prop_gr_lit_assign solver gr_by lit =
  let atom = Term.get_atom lit in 
  let atom_prop_gr_key = get_prop_gr_key_assign solver gr_by atom in
  let prop_var_entry = TableArray.get var_table atom_prop_gr_key in
  let def_lit = 
    if (Term.is_neg_lit lit) 
    then      
      prop_var_entry.prop_neg_var 
    else
     prop_var_entry.prop_var 
  in
  match def_lit with 
  |Def(lit) -> lit 
  | _ -> failwith "Insatntiation get_prop_gr_lit_assign: lit is undefind"


let get_prop_gr_var_entry lit = 
  try 
    let atom           = Term.get_atom lit in
    let prop_gr_key       = Term.get_prop_gr_key atom in
    TableArray.get var_table prop_gr_key
  with 
    Term.Prop_gr_key_undef -> failwith "get_prop_gr_var_entry : Term.Prop_key_undef"


let get_prop_var_entry lit = 
  try 
    let atom           = Term.get_atom lit in
    let prop_key       = Term.get_prop_key atom in
    TableArray.get var_table prop_key
  with 
    Term.Prop_key_undef -> failwith "get_prop_var_entry : Term.Prop_key_undef"

let get_truth_var_entry var_entry = 
  match var_entry.truth_val with	
  |Def(truth_val) -> truth_val
  |Undef ->   failwith "get_truth_var_entry: truth_val is Undef\n"

let get_prop_var_var_entry var_entry =
  match var_entry.prop_var with	
  |Def(prop_var) -> prop_var
  |Undef ->   failwith "get_prop_var_var_entry: prop_var is Undef\n"

let get_prop_neg_var_var_entry var_entry =
  match var_entry.prop_neg_var with	
  |Def(prop_neg_var) -> prop_neg_var
  |Undef ->   failwith "get_prop_neg_var_var_entry: prop_var is Undef\n"


	

let get_lit_activity lit = 
(*  out_str ("Lit Act !\n"^(Term.to_string lit));*)
  let var_entry = get_prop_gr_var_entry lit in
  if (Term.is_neg_lit lit) 
  then var_entry.neg_activity
  else var_entry.pos_activity

let lit_activity_compare l1 l2 = 
(*  out_str ("Act Compare !\n"^(Term.to_string l1)^(Term.to_string l2));*)
  compare (get_lit_activity l1) (get_lit_activity l2)


 (*
let selection_fun get_prop_truth clause = 
  let candidate_list = Clause.find_all get_prop_truth clause in 
  let fun_list = [Term.cmp_ground;(compose_12 (~-)lit_activity_compare); 
		  Term.cmp_sign] 
in
  list_find_max_element (lex_combination fun_list)  candidate_list
 
 
let ()=add_param_str ("Real Selection lex [-act;-num_symb]\n")
*)

(* adds both versions of the clauses before and after grounding *)
(* first version is used for simplifications *)
exception PropImplied
let add_clause_to_solver solver_sim solver gr_by clause =
  let lits = Clause.get_literals clause in
  let prop_gr_lit_list = List.map (get_prop_gr_lit_assign solver gr_by) lits in
  let prop_lit_list = List.map (get_prop_lit_assign solver) lits in 
(* Propositional implication  is not compatible with prop_subsumtion:*)
(* an instance of a subsumed clause can imply an instance of the new clause*)
(*  let lits_in_solver_sim =
    let f lit = 
      let var_entry = get_prop_var_entry lit in
      var_entry.in_solver_sim in
    List.find_all f lits in
  let compl_lit_list = List.map get_prop_compl_lit lits_in_solver_sim in 
  match (PropSolver.fast_solve solver_sim compl_lit_list)
  with 
  | PropSolver.FUnsat -> 
      (num_prop_implied:=!num_prop_implied+1; 
(*       out_str ("Prop Implied: "^(Clause.to_string clause)^"\n");*)
       raise PropImplied)
  |PropSolver.FSat | PropSolver.FUnknown -> *)
      (* debug*) 
      (*let str = list_to_string PropSolver.lit_to_string prop_gr_lit_list ";" in
	out_str_debug ("Clause: "^(Clause.to_string clause)^"\n"
	^"Prop Clause:"^str^"\n");*)
      ( PropSolver.add_clause solver prop_gr_lit_list;
	PropSolver.add_clause solver prop_lit_list;
	PropSolver.add_clause solver_sim prop_gr_lit_list;
	PropSolver.add_clause solver_sim prop_lit_list;
	let f lit = 
	  let var_entry = get_prop_var_entry lit in
	  var_entry.in_solver_sim <- true in
	List.iter f lits
)
	

(********* end propositional interpretation ************)

(*************** unprocessed **********************)
let add_clause_to_unprocessed clause = 
  unprocessed_ref:=clause::!unprocessed_ref;
  num_in_unprocessed:=!num_in_unprocessed+1
			     
(*************** simple passive *********************)

exception Passive_Simple_Empty 
let remove_from_simple_passive () = 
  match !simple_passive_ref with 
  | clause::tl -> 
      simple_passive_ref := tl;
      Clause.set_bool_param false Clause.in_passive clause;
      num_in_simple_passive:=!num_in_simple_passive-1;
      clause
  |[]     -> raise Passive_Simple_Empty 
	
let add_to_simple_passive clause = 
  simple_passive_ref := clause::!simple_passive_ref;
  num_in_simple_passive:=!num_in_simple_passive+1;
  Clause.set_bool_param true Clause.in_passive clause; 
  Clause.assign_when_born !num_of_instantiation_iter clause

(*add new clauses*)    
let add_new_clause_to_sp clause = 
  if (Clause.is_empty_clause clause) 
  then
    raise Unsatisfiable
  else 
    if (not (Clause.get_bool_param Clause.is_dead clause))  
    then 
	  add_to_simple_passive clause 
    else ()


(*************** end simple passive *********************)


(********************* New Passive QUEUES *)

module Queue1Elem = 
  struct
    type t = clause
    let compare = inst_pass_queue1_cmp 
  end


module Queue1 = Heap.Functional (Queue1Elem) 
let queue1_ref = ref (Queue1.empty)

module Queue2Elem = 
  struct
    type t = clause
    let compare = inst_pass_queue2_cmp 
  end


module Queue2 = Heap.Functional (Queue2Elem) 
let queue2_ref = ref (Queue2.empty)

type passive_ratio = {mutable queue1 : int; mutable queue2 : int}

(* the algorithm picks from the queue 
   with non zero current_passive_ratio.parameter; if all are 0 
   then it goes to the initial state i.e. 
   assigns passive_ratio to current_passive_ratio *)

let init_passive_ratio () = 
  {queue1 = inst_pass_queue1_mult; queue2 = inst_pass_queue2_mult} 
(* param should be >= 0*)

let current_passive_ratio_ref =  ref (init_passive_ratio ())

(* doesnot work with move_from_active_to_passive since 
 we don't add to passive if clause is in one of the queues! 
let passive_is_empty () =
  ((!queue1_ref  = Queue1.empty) ||
   (!queue2_ref  = Queue2.empty))
*)


(* one queue debug*)
(*
let passive_is_empty () =
  (!queue1_ref  = Queue1.empty)
  *) 

let empty_passive () = 
  queue1_ref := Queue1.empty;
  queue2_ref := Queue2.empty


let passive_is_empty () =
  ((!queue1_ref  = Queue1.empty) &
   (!queue2_ref  = Queue2.empty))

exception Passive_Empty 
let rec remove_from_passive () = 
  if (passive_is_empty ())
  then
(* since we add all clauses to all queues it is enough that one queue is empty*)
    ((*num_in_passive:= 0;*)
     raise Passive_Empty)
  else
 (*   if (!(current_passive_ratio_ref).queue1 > 0)     *)
    if ((!(current_passive_ratio_ref).queue1 > 0) & not (!queue1_ref  = Queue1.empty))
    then 
      let clause =  Queue1.maximum !(queue1_ref) in 
     (* out_str ("Queu1 try:  "^(Clause.to_string clause)^"\n"); *)
      queue1_ref := Queue1.remove !(queue1_ref);
      !(current_passive_ratio_ref).queue1 <- !(current_passive_ratio_ref).queue1 - 1;
      if ((Clause.get_bool_param Clause.in_active clause) || 
          (Clause.get_bool_param Clause.is_dead clause)) 
      then 
	(Clause.set_bool_param false Clause.inst_pass_queue1 clause; 
	 remove_from_passive ())
      else 
	((*out_str_debug ("Queue 1: "^(Clause.to_string clause)^"\n");*)
	  num_in_passive:=!num_in_passive-1;
	 Clause.set_bool_param false Clause.inst_pass_queue1 clause; 
	 clause)
    else 
     (* if (!(current_passive_ratio_ref).queue2 > 0)*)
      if ((!(current_passive_ratio_ref).queue2 > 0) & not (!queue2_ref  = Queue2.empty))
      then 
	let clause =  Queue2.maximum !(queue2_ref) in 
	(*out_str ("Queu2 try:  "^(Clause.to_string clause)^"\n"); *)
	queue2_ref := Queue2.remove !(queue2_ref);
	!(current_passive_ratio_ref).queue2 <- !(current_passive_ratio_ref).queue2 - 1;
	if ((Clause.get_bool_param Clause.in_active clause) ||
           (Clause.get_bool_param Clause.is_dead clause)) 
	then 
	  (Clause.set_bool_param false Clause.inst_pass_queue2 clause; 
	   remove_from_passive ())
	else 
	  ((*out_str_debug ("Queue2: "^(Clause.to_string clause)^"\n");*)
	    num_in_passive:=!num_in_passive-1;
	   Clause.set_bool_param false Clause.inst_pass_queue2 clause; 
	   clause)
      else 
        (* we have (!(current_passive_ratio_ref).queue1 = 0 *)
	(* and !(current_passive_ratio_ref).queue2 = 0) *)
	(current_passive_ratio_ref := init_passive_ratio ();
	 remove_from_passive ())



(* one queue debug*)
(*
let rec remove_from_passive () = 
  if (passive_is_empty ())
  then
(* since we add all clauses to all queues it is enough that one queue is empty*)
    (num_in_passive:= 0;
     raise Passive_Empty)
  else
    (let clause =  Queue1.maximum !(queue1_ref) in 
    queue1_ref := Queue1.remove !(queue1_ref);
    !(current_passive_ratio_ref).queue1 <- !(current_passive_ratio_ref).queue1 - 1;
    if ((Clause.get_bool_param Clause.in_active clause) || 
    not (Clause.get_bool_param Clause.in_passive clause)) 
    then remove_from_passive ()
    else 
      ((*out_str_debug ("Passive Clause: "^(Clause.to_string clause)^"\n");*)
	num_in_passive:=!num_in_passive-1;
       Clause.set_bool_param false Clause.in_passive clause; 
       clause)	
    )
*)

let in_passive clause =   
  ((Clause.get_bool_param Clause.inst_pass_queue1 clause) ||
   (Clause.get_bool_param Clause.inst_pass_queue2 clause))  

let add_to_passive clause = 
  if  (in_passive clause) 
  then (num_in_passive:=!num_in_passive+1)
  else
    ( queue1_ref := 
      Queue1.add clause !(queue1_ref); 
      Clause.set_bool_param true Clause.inst_pass_queue1 clause; 
      queue2_ref := 
	Queue2.add clause !(queue2_ref);
   (*   Clause.set_bool_param true Clause.in_passive clause; *)
      Clause.set_bool_param true Clause.inst_pass_queue2 clause; 
      num_in_passive:=!num_in_passive+1)


(* one queue debug *)
(*let add_to_passive clause = 
  if  (Clause.get_bool_param Clause.in_passive clause) 
  then (num_in_passive:=!num_in_passive+1)
  else
    ( queue1_ref := 
      Queue1.add clause !(queue1_ref); 
      Clause.set_bool_param true Clause.in_passive clause; 
      num_in_passive:=!num_in_passive+1)
  *)
   
(* change empty clause check  to unprocessed*)
let add_new_clause_to_passive clause = 
  if (Clause.is_empty_clause clause) 
  then
    raise Unsatisfiable
  else 
    if (not (Clause.get_bool_param Clause.is_dead clause))  
    then 
	  add_to_passive clause 
    else ()

(********************* End New Passive QUEUES *)





(*************** unification index **********************)

module DTParam = 
  struct let num_of_symb = (SymbolDB.size !symbol_db_ref) end 
module DiscrTreeM = DiscrTree.Make(DTParam)  

(* all clauses with the same literal put together, *)
(*   assoc list with == *)

type unif_index_elem = (lit*(clause list)) list

let (unif_index_ref : (unif_index_elem DiscrTreeM.index) ref ) 
    = ref (DiscrTreeM.create ())
    
(* add to unif index *)
      
let add_to_unif_index main_clause =
  try
    let sel_lit = Clause.get_inst_sel_lit main_clause in
    Term.set_fun_bool_param true  Term.in_unif_index sel_lit;
(*  out_str_debug ("Add to Unif Index: Clause: "^(Clause.to_string main_clause) 
 	 ^"  literal:  "^(Term.to_string sel_lit)^"\n"); *)
    let ind_elem = DiscrTreeM.add_term_path sel_lit unif_index_ref in
    (match !ind_elem with 
    |Elem(old) -> 
	(try
	  let old_clause_list =  List.assq sel_lit old in 
	  let old_with_removed = List.remove_assq sel_lit old in
	  ind_elem := 
	    Elem((sel_lit,(main_clause::old_clause_list))::old_with_removed)
	  with Not_found ->  ind_elem := Elem((sel_lit,[main_clause])::old)
	)	     
    |Empty_Elem   -> 	 
	ind_elem := Elem([(sel_lit,[main_clause])])
    );     
    Clause.set_bool_param true Clause.in_unif_index main_clause
  with
    Clause.Inst_sel_lit_undef -> 
      failwith "add_to_unif_index: clause should have selected literals here"



(************eliminates clause!*****************)

let eliminate_from_unif_index main_clause = 
  try
    let sel_lit = Clause.get_inst_sel_lit main_clause in
    Term.set_fun_bool_param false  Term.in_unif_index sel_lit;
    (*out_str_debug 
      ("Trying to elim from Unif index:"
      ^(Clause.to_string main_clause)
      ^" Literals: "
      ^(Term.to_string sel_lit)^"\n");
     *)
    (try  
      let ind_elem = DiscrTreeM.find sel_lit !unif_index_ref
	(* failwith "discount:  eliminate_from_unif_index lit is in unif_index" *)
      in
      ( match !ind_elem with 
      |Elem(old) ->
	  ( (* old = [(L1,[C_1,..,Cn]),(L2,[C'_1,..,C'n']),..]  
	       old_clause_list = [C_1,..,Cn] corr to sel_lit*)
	    (*  try *) 
	    let old_clause_list  = List.assq sel_lit old in 
            (*  out_str_debug 
		("Elem From Unif index old_cl_list:"
		^(Clause.clause_list_to_string old_clause_list)^"\n"); *)   
	    let old_with_removed = List.remove_assq sel_lit old in
	    
(*remove main_clause*)     
	    let new_clause_list = 
	      List.find_all (fun cl -> not(cl == main_clause)) old_clause_list in
	    (* out_str_debug 
	       ("Elem From Unif index new_cl_list:"
	       ^(Clause.clause_list_to_string new_clause_list)^"\n"); *)
            if new_clause_list = [] 
	    then 
 	      if  old_with_removed = []
		then
		(DiscrTreeM.remove_term_path sel_lit unif_index_ref
                   (*; out_str_debug 
		     ("Elim unif Removed term path"
		     ^(Term.to_string sel_lit)^"\n")*))
	      else 
		  (ind_elem := Elem(old_with_removed)
                      (*; out_str_debug 
			("Elim unif: Old_with_removed")*))
	    else 	    
	      (ind_elem := 
		Elem((sel_lit,new_clause_list)::old_with_removed)
		  (*;out_str_debug 
		    ("Elim unif: Old_with_removed")*) )		
		(*with *)
		(*  Not_found -> ()*)
	   )
    | Empty_Elem -> 
	  failwith "instantiation: eliminate_from_unif_index  \
	    unif index should not contain Empty_Elem"  
       )     
    with
      Not_found -> 
	out_str 
      ("\n Warning: eliminate_from_unif_index: the clause in not in the index!\n ")
  ); 
    Clause.set_bool_param  
      false Clause.in_unif_index main_clause 
  with 
    Clause.Inst_sel_lit_undef -> 
      failwith "eliminate_from_unif_index: Clause.Sel_lits_undef "
	

(* eliminates all clauses indexed by lit from unif_index and returns*)
(* the eliminated clause list   *)

let eliminate_lit_from_unif_index lit = 
  let  ind_elem = DiscrTreeM.remove_term_path_ret lit unif_index_ref in
  match !ind_elem with 
  | Elem(elem) ->
   (* elem = [(L1,[C_1,..,Cn]),(L2,[C'_1,..,C'n']),..] *)  
      (* elem_clause_list = [C_1,..,Cn] corr to sel_lit*)
      let elem_f rest (lit,cl_list) = 
	let add_cl rest' clause = 
	  Clause.set_bool_param  
	    false Clause.in_unif_index clause;
	  Term.set_fun_bool_param false  Term.in_unif_index lit;
	  clause::rest'
	in
	List.fold_left add_cl rest cl_list 
      in 
      List.fold_left elem_f [] elem
  | Empty_Elem -> 
      failwith "instantiation: eliminate_lit_from_unif_index  \
	unif index should not contain Empty_Elem"  


(************** end  unification index **********************)

(****************end simplification*******************)
(*
let dismatching_string clause = 
  try 
   "["^(Dismatching.constr_list_to_string (Clause.get_dismatching clause))^"]"
  with
    Clause.Dismatching_undef -> "[]"
*)

(*************** apply_new_model_sp ************)

let add_to_active clause = 
  if not (Clause.get_bool_param Clause.in_active clause)
   then
    (Clause.set_bool_param true Clause.in_active clause;
     add_to_unif_index clause;
    (* out_str ("Add to Active: "^(Clause.to_string clause));*)
     num_in_active := !num_in_active +1
    )
  else ()

let remove_from_active clause =
  if (Clause.get_bool_param Clause.in_active clause) 
  then 
    (eliminate_from_unif_index clause;
     Clause.set_bool_param false Clause.in_active clause;
     (*out_str ("Remove from Active: "^(Clause.to_string clause));*)
     num_in_active := !num_in_active -1
    )
  else ()

let remove_lit_from_active lit =
  let cl_list = eliminate_lit_from_unif_index lit in
  let set_param clause = 
    (*out_str ("Remove from Active: "^(Clause.to_string clause));*)
    Clause.set_bool_param false Clause.in_active clause;
    num_in_active := !num_in_active -1;
(*    out_str ("Removed from Unif: "^(Clause.to_string clause))*)
  in
  List.iter set_param cl_list;
  cl_list



(* simple passive version old*)
let move_from_active_to_sp clause = 
  remove_from_active clause; 
  add_to_simple_passive clause;
  num_moves_active_passive:=!num_moves_active_passive+1

let move_from_active_to_passive clause = 
  remove_from_active clause; 
 (* add_clause_to_unprocessed clause;*)
(*  out_str ("move_from_active_to_passive: "^(Clause.to_string clause)^"\n");*)
  (if (not (in_passive clause)) then
    (Clause.assign_when_born !num_of_instantiation_iter clause;
     add_to_passive clause)
  else num_in_passive := !num_in_passive+1
  );  
  num_moves_active_passive:=!num_moves_active_passive+1


(*  moves all clauses from univ index which are indexed *)
(* by the same literal *)


let move_lit_from_active_to_passive solver lit = 
    let cl_list = remove_lit_from_active lit in
    let to_pass clause = 
  
 (*   add_clause_to_unprocessed clause;*)
(*    Clause.assign_when_born (!num_of_instantiation_iter+2) clause;*)
(*debug*)
    (*  out_str ("\n Act_to_Pass: "^(Clause.to_string clause)^"\n");
      let sel_lit = Clause.get_inst_sel_lit clause in
      let var_entry   = get_prop_gr_var_entry sel_lit in
      out_str ("Lit: "^(Term.to_string sel_lit)^"\n"
	       ^"Var entry:"^(var_entry_to_string solver var_entry)^"\n");*)
      (if (not (in_passive clause)) then
	(Clause.assign_when_born !num_of_instantiation_iter clause;
	 add_to_passive clause)
      else num_in_passive := !num_in_passive+1
      );
      num_moves_active_passive:=!num_moves_active_passive+1
    in
    List.iter to_pass cl_list

    
(* PropSolver.Any  as false *)	
(* 
   let lit_val_to_bool lit_val = 
   match lit_val with
   |PropSolver.Any   -> false
   |PropSolver.True  -> true 
   |PropSolver.False -> false
*)

(*
let is_true_in_model solver lit = 
  let is_neg         = Term.is_neg_lit lit in    
  let var_entry      = get_prop_var_entry lit in
  let prop_var       = get_prop_var_var_entry var_entry in
  match var_entry.prop_var with	
  |Def(prop_var) -> 
      let var_truth_val = PropSolver.lit_val solver prop_var in
      if
	((lit_val_to_bool var_truth_val) & (not is_neg)) ||  
	((not (lit_val_to_bool var_truth_val)) & is_neg)
      then true
      else false
  |Undef ->  failwith "is_true_in_model : prop_var is Undef\n"
*)


(*Warning both A and neg A can be consitent with the solver *)
(* (if the solver returns Any) *)
(* after grounding*)
let consistent_with_solver solver lit = 
  let var_entry      = get_prop_gr_var_entry lit in
  let prop_var       = get_prop_var_var_entry var_entry in
  let var_truth_val  = PropSolver.lit_val solver prop_var in
  if var_truth_val = PropSolver.Any 
  then true 
  else
    let is_neg = Term.is_neg_lit lit in    
    if
      ((var_truth_val = PropSolver.True)  & (not is_neg)) ||  
      ((var_truth_val = PropSolver.False) & is_neg)
    then true
    else false

(* without grounding*)
let consistent_with_solver_lit solver lit = 
  let var_entry      = get_prop_var_entry lit in
  let prop_var       = get_prop_var_var_entry var_entry in
  let var_truth_val  = PropSolver.lit_val solver prop_var in
  if var_truth_val = PropSolver.Any 
  then true 
  else
    let is_neg = Term.is_neg_lit lit in    
    if
      ((var_truth_val = PropSolver.True)  & (not is_neg)) ||  
      ((var_truth_val = PropSolver.False) & is_neg)
    then true
    else false

(* after grounding *)
let consistent_with_model lit = 
  let var_entry      = get_prop_gr_var_entry lit in
(*    let var_truth_val  = get_truth_var_entry var_entry in*)
  match var_entry.truth_val with
  |Def(var_truth_val) ->
      if (var_truth_val = PropSolver.Any)
      then true 
      else
      let is_neg = Term.is_neg_lit lit in    
      if
	((var_truth_val = PropSolver.True)  & (not is_neg)) ||  
	((var_truth_val = PropSolver.False) & is_neg)
      then true
      else false
  | Undef -> true



let get_lit_activity lit = 
  let var_entry = get_prop_var_entry lit in
  if (Term.is_neg_lit lit) then 
    var_entry.neg_activity
  else
    var_entry.pos_activity

let  consistent_with_model_activity  lit = 
  let activity = get_lit_activity lit in
  if (activity < !max_lit_activity lsr 2) || (activity < !lit_activity_threshold)
  then consistent_with_model lit
  else false
      
let  consistent_with_solver_activity solver lit = 
  let activity = get_lit_activity lit in
    if (activity < !max_lit_activity lsr 2) || (activity < !lit_activity_threshold)
    then consistent_with_solver solver lit
    else false



(* without grounding*)
let consistent_with_model_lit lit = 
  let var_entry      = get_prop_var_entry lit in
  let var_truth_val  = get_truth_var_entry var_entry in
  if (var_truth_val = PropSolver.Any)
  then true 
  else
    let is_neg = Term.is_neg_lit lit in    
    if
      ((var_truth_val = PropSolver.True)  & (not is_neg)) ||  
      ((var_truth_val = PropSolver.False) & is_neg)
    then true
    else false
  
    
(* returns true if it can preserve model (solver is run under entry assumption) *)
(* otherwise return false *)

let preserve_model_solver solver var_entry =
  let prop_var      = get_prop_var_var_entry var_entry in
  let prop_neg_var  = get_prop_neg_var_var_entry var_entry in
  let new_truth_val = PropSolver.lit_val solver prop_var in
(*  out_str ("Var enty:"^(var_entry_to_string solver var_entry)^"\n");
  out_str ("Solver truth val:"^(PropSolver.lit_val_to_string new_truth_val)^"\n");*)
  match var_entry.truth_val with 
  |Def(old_truth_val) ->
      if (old_truth_val = PropSolver.Any) 
      then 
        (* if truth_val = Def(Any) the we assume that sel_clauses=[] *)
	(var_entry.truth_val <- Def(new_truth_val); 
	 true)
      else
	if (old_truth_val = new_truth_val) || (new_truth_val = PropSolver.Any)
	then true
	else 
	  (
	    let assumption = 
	      if  old_truth_val = PropSolver.True 
	      then prop_var
	      else prop_neg_var 
	    in	   
	    (match  (PropSolver.solve_assumptions solver [assumption]) 
	    with 
	    | PropSolver.Sat -> true
	    | PropSolver.Unsat (*| PropSolver.FUnknown*) ->
		(
(*		 var_entry.truth_val <- Def(new_truth_val);
		 let move_cl_from_active_to_passive clause =
		   if  (Clause.get_bool_param Clause.in_active clause) 
		   then 
		     (let sel_lit = Clause.get_inst_sel_lit clause in
		     (* moves all clauses indexed by sel_lit *)
                    move_lit_from_active_to_passive sel_lit)
                   else ()
		 in
		 List.iter move_cl_from_active_to_passive var_entry.sel_clauses;
		 var_entry.sel_clauses <- [];
 *)
(*		 out_str ("Preserve Model: Unsat: "
			  ^"Assum: "^(PropSolver.lit_to_string  assumption)^"\n"
			 ^(var_entry_to_string solver var_entry)^"\n");*)
		   
(*		 (match (PropSolver.solve solver)
		 with PropSolver.Unsat -> raise Unsatisfiable
		 |PropSolver.Sat -> ());*)
                 
		 false)
	    ))

  |Undef -> 
      (var_entry.truth_val <- Def(new_truth_val); 
       true)
(*with Not_found -> failwith "Nor_found here"*)


(* if model value is  defferent from solver then chage the model value *)
let change_model_solver  solver var_entry =
  let prop_var      = get_prop_var_var_entry var_entry in
(*  let prop_neg_var  = get_prop_neg_var_var_entry var_entry in*)
  let new_truth_val = PropSolver.lit_val solver prop_var in
(*  out_str ("Var enty:"^(var_entry_to_string solver var_entry)^"\n");
  out_str ("Solver truth val:"^(PropSolver.lit_val_to_string new_truth_val)^"\n");*)
  match var_entry.truth_val with 
  |Def(old_truth_val) ->
      if (old_truth_val = PropSolver.Any) 
      then 
        (* if truth_val = Def(Any) the we assume that sel_clauses=[] *)
	(var_entry.truth_val <- Def(new_truth_val)) 	
      else
	if (old_truth_val = new_truth_val) || (new_truth_val = PropSolver.Any)
	then ()
	else 
	  (	 
     var_entry.truth_val <- Def(new_truth_val);
(*	   out_str ("change_model_solver: "^
		    (var_entry_to_string solver var_entry)^"\n");*)
	   let move_cl_from_active_to_passive clause =
	     if  (Clause.get_bool_param Clause.in_active clause) 
	     then 
	       (let sel_lit = Clause.get_inst_sel_lit clause in
	     (*  out_str ("Change_model in Active: "^(Clause.to_string clause)^"\n");*)
	       (* moves all clauses indexed by sel_lit *)
	       move_lit_from_active_to_passive solver sel_lit
	       )
	     else ()
	   in
	   List.iter move_cl_from_active_to_passive var_entry.sel_clauses;
	   var_entry.sel_clauses <- []
	    )	  

  |Undef -> 
      var_entry.truth_val <- Def(new_truth_val)

(* auxilary function for below, apply only if lit is consitent with the model!*)
       
let ass_if_consistent lit clause= 
  let var_entry = get_prop_gr_var_entry lit in   
  if (get_truth_var_entry var_entry = PropSolver.Any)
  then 
    (
     var_entry.sel_clauses <- [clause];
     if (Term.is_neg_lit lit)
     then var_entry.truth_val <- Def(PropSolver.False)
     else var_entry.truth_val <- Def(PropSolver.True))
  else    
    var_entry.sel_clauses <- clause::(var_entry.sel_clauses)
(* we assume that the clause is from passive and therefore *)
(* not assigned in to any of var_entry.sel_clauses         *)
(* we try to keep the old selection                        *)
(*
let find_nex_true_lit 
*)

let remove_undef_model solver clause = 
  let remove_undef_lit lit = 
    let var_entry = get_prop_gr_var_entry lit in
    match var_entry.truth_val with 
    |Def(PropSolver.Any)|Undef ->
        (* if truth_val = Def(Any) the we assume that sel_clauses=[] *)
	(let prop_var  = get_prop_var_var_entry var_entry in
	let new_truth_val = PropSolver.lit_val solver prop_var in
	var_entry.truth_val <- Def(new_truth_val))
    |_ -> ()
  in 
  Clause.iter remove_undef_lit clause
	


exception Sel_Changed  
exception Solver_Sel
let rec selection_renew solver_sim solver clause =  
(*  out_str (" selection_renew clause:  "^(Clause.to_string clause)^"\n");*)
(*
  let accord lit = 
    let var_entry     = get_prop_gr_var_entry lit in
   let _= model_accords_solver solver var_entry in () in
  Clause.iter accord clause;          *)
    try
      let sel_lit       = Clause.get_inst_sel_lit clause in    
      let sel_var_entry = get_prop_gr_var_entry sel_lit in
      if (consistent_with_model sel_lit)
      then     
	if (preserve_model_solver solver sel_var_entry)
	then ass_if_consistent sel_lit clause	  
	else    
	  raise Sel_Changed
      else  raise Sel_Changed
    with
      Sel_Changed | Clause.Inst_sel_lit_undef ->
	(
	 try 	   
	  (  
	     remove_undef_model solver clause;
	   (*  out_str ("----------------------------\n");
	     out_str ("Try consist with Model:");
	     let out_entry lit = 
	       let var_entry     = get_prop_gr_var_entry lit in
	       out_str ("\n Lit: "^(Term.to_string lit)^"\n"
			^(var_entry_to_string solver var_entry)^"\n"
			^"Consist_with_Model: "
			^(string_of_bool (consistent_with_model lit))^"\n");
	     in Clause.iter out_entry clause; *)
	     let model_sel_lit = 
	       selection_fun consistent_with_model_activity clause in
(*             out_str ("Consist_model: "^(Term.to_string model_sel_lit)^"\n");*)
	     let model_sel_var_entry  = get_prop_gr_var_entry model_sel_lit in
	       if  (preserve_model_solver solver model_sel_var_entry) then 
		 ((*change_model_solver solver model_sel_var_entry;*)
		  Clause.assign_inst_sel_lit model_sel_lit clause;
		  ass_if_consistent model_sel_lit clause;
(*		  out_str ("preserve model:  "^
			   (var_entry_to_string solver model_sel_var_entry)^"\n");
		  out_str ("----------------------------\n")*)
		 )
	       else 
(* optional change_model*)
		 (change_model_solver solver model_sel_var_entry;
		  raise Solver_Sel)
	   )    
	 with Not_found | Solver_Sel  ->	   
	   try (
(*debug*)
(*	     out_str ("----------------------------\n");
	     out_str ("No consist with Model:");
	     let out_entry lit = 
	       let var_entry     = get_prop_gr_var_entry lit in
	       out_str ("\n Lit: "^(Term.to_string lit)^"\n"
			^(var_entry_to_string solver var_entry)^"\n"
                        ^"Consist_with_Model: "
                        ^(string_of_bool (consistent_with_model lit))^"\n");
	     in Clause.iter out_entry clause;*)

	     let solver_sel_lit = 
	       selection_fun (consistent_with_solver_activity solver) clause in	  
	     let solver_sel_var_entry  = get_prop_gr_var_entry solver_sel_lit in
	   (*  out_str ("Before change\n");*)
	     change_model_solver solver solver_sel_var_entry;
(*	     out_str("Sel_Lit: "^(Term.to_string solver_sel_lit)^"\n"
		     ^"Sel_lit entry: "
		     ^(var_entry_to_string solver solver_sel_var_entry));*)
	     Clause.assign_inst_sel_lit solver_sel_lit clause; 
	     ass_if_consistent solver_sel_lit clause
(*	     out_str ("----------------------------\n");*)
	   )
	   with Not_found -> 
	     match (PropSolver.solve solver) with 
	     |PropSolver.Unsat -> raise Unsatisfiable
	     |PropSolver.Sat   ->
		 let new_solver_sel_lit = 
		   selection_fun (consistent_with_solver solver) clause in	  
		 let new_solver_sel_var_entry  = 
		   get_prop_gr_var_entry new_solver_sel_lit in
		 change_model_solver solver new_solver_sel_var_entry;
(*		 out_str ("Solver select:"^
			  "Sel_Lit: "^(Term.to_string new_solver_sel_lit)^"\n"
			  ^"Sel_lit entry: "
			  ^(var_entry_to_string solver new_solver_sel_var_entry)^"\n");*)
		 Clause.assign_inst_sel_lit new_solver_sel_lit clause; 
		 ass_if_consistent new_solver_sel_lit clause
	)

(*
exception Sel_Unchanged
let apply_new_model solver = 
  let sel_is_changed = ref false in
  let change_entry var_entry =
    let prop_var = get_prop_var_var_entry var_entry in
    match var_entry.truth_val with	
    |Def(_) -> 	
	if (change_model_solver solver var_entry) 
	then ()
	else sel_is_changed:=true
    |Undef ->
	(    (* sel_is_changed:=true; *)
	     let new_truth_val = PropSolver.lit_val solver prop_var in
	     var_entry.truth_val <- Def(new_truth_val)
	    )
  in
  TableArray.iter change_entry var_table;  
  if !sel_is_changed then ()
  else 
    raise Sel_Unchanged
  *)    
      
(*************** end apply_new_model ************)
(************* Simplification *****************) 

(*************** simplification  ***)

	    
exception New_Clause_Simplified 

let simplify_new clause = 
  if (ClauseAssignDB.mem clause !clause_db_ref)
  then  
    (num_of_duplicates:= !num_of_duplicates+1;
     (*out_str_debug ("\n Simplified: "^(Clause.to_string clause)^"\n"); *)  
  raise New_Clause_Simplified)
  else clause


exception Activity_Check
exception Activity_Undef
let lit_activity_check solver lit = 
  try 
(*  let sel_lit = Clause.get_inst_sel_lit clause in*)
(*  let var_entry = get_prop_gr_var_entry sel_lit in*)
    let var_entry = get_prop_gr_var_entry lit in 
    let model_truth_val = 
      match var_entry.truth_val with 
      |Def(PropSolver.True) -> true 
      |Def(PropSolver.False) -> false
      |_ -> raise Activity_Undef 
    in
    if ((model_truth_val = false)
	  & (var_entry.neg_activity > 
	     (var_entry.pos_activity + !lit_activity_threshold+(!max_lit_activity lsr 2))))
    then
      (
       match (PropSolver.solve_assumptions solver  
		[(get_prop_var_var_entry var_entry)])
       with 
       |PropSolver.Unsat -> 
	   ((*match (PropSolver.solve solver)
	   with PropSolver.Unsat -> raise Unsatisfiable
	   |PropSolver.Sat -> ()*) )
       |PropSolver.Sat   -> (lit_activity_check_move:=!lit_activity_check_move+1;
			 (*    out_str ("Act Lit: "^(Term.to_string lit)^"\n"
				      ^"Before Change: "
				      ^(var_entry_to_string solver var_entry)^"\n");*)

			     change_model_solver solver var_entry;
			     
			   (*  out_str ("After Chnage: "
				      ^(var_entry_to_string solver var_entry)^"\n");*)
			     raise Activity_Check)
      )
    else 
      if ((model_truth_val = true)
	    & (var_entry.pos_activity > (var_entry.neg_activity + !lit_activity_threshold+(!max_lit_activity lsr 2))))
      then
	( 
	  match (PropSolver.solve_assumptions solver  
		   [(get_prop_neg_var_var_entry var_entry )])
	 with 
	 |PropSolver.Unsat -> 
	     ((*match (PropSolver.solve solver)
	     with PropSolver.Unsat -> raise Unsatisfiable
	     |PropSolver.Sat -> ()*)) 
(* (out_str ("Act_Check Unsat with assumption: "
				    ^(PropSolver.lit_to_string 
				    (get_prop_neg_var_var_entry var_entry))^"\n"))*)
	 |PropSolver.Sat   -> (lit_activity_check_move:=!lit_activity_check_move+1;
(*			       out_str ("Act Lit: "^(Term.to_string lit)^"\n"
					^"Before Change: "
					^(var_entry_to_string solver var_entry)^"\n");*)
			       
			       change_model_solver solver var_entry;
			       
			      (* out_str ("After Chnage: "
					^(var_entry_to_string solver var_entry)^"\n");*)
			       raise Activity_Check)
	)
      else ()
  with Activity_Undef -> ()
     
let rec eliminate_clause clause = 
  remove_from_active clause;
  Clause.set_bool_param true Clause.is_dead clause;
  num_child_elim:=!num_child_elim+1;
  List.iter eliminate_clause (Clause.get_children clause)

exception Given_clause_simplified 
(* is weaker than prop_subsumption below *)
(*
let prop_subs_resolution solver_sim solver gr_by clause = 
(*  out_str ("\n Given Clause: "^(Clause.to_string clause)^"\n");*)
  let is_simplified = ref false in
(*  let lits = Clause.get_lits clause in*)
  let resolve rest lit = 
    if (consistent_with_solver_lit solver lit) then lit::rest
    else
      (let prop_lit = get_prop_lit lit in
(*      let prop_compl_lit = get_prop_compl_lit solver lit in*)
      let result = (PropSolver.solve_assumptions solver_sim [prop_lit]) in
      match result with 
      |PropSolver.Unsat -> 
	  (* out_str "Unsat Assmpt \n ";*)
	  is_simplified:= true; 
	  num_prop_subs_resolution:=!num_prop_subs_resolution+1;
	  rest
      |PropSolver.Sat -> lit::rest
      ) 
    in 			      
  let new_lit_list = Clause.fold resolve [] clause in 
  if !is_simplified then 
    ( (*out_str "Given Simplified! \n";*)
      eliminate_clause clause;
      if (new_lit_list = []) 
      then raise Unsatisfiable 
      else
	(let new_clause   = Clause.create new_lit_list in
	if (not (ClauseAssignDB.mem new_clause !clause_db_ref))
	then
	  (let added_clause = 
	    ClauseAssignDB.add_ref new_clause clause_db_ref in             
	  (*add_clause_to_solver solver_sim solver gr_by added_clause;*)
	  Clause.assign_when_born !num_of_instantiation_iter  added_clause;
(*	out_str ("\n Clause: "^(Clause.to_string clause)^"\n");
  out_str ("\n Simplified to: "^(Clause.to_string added_clause)^"\n");*)
	 (* add_clause_to_unprocessed added_clause;*)
	  added_clause)
	else
	  raise Given_clause_simplified 
	(*added_clause*))      
     )
  else () (*clause*)
*)


(* needed for prop_subsumption *)
type add_lit = 
    {
     slit           : lit;
     sprop_lit      : PropSolver.lit;
     sprop_compl_lit : PropSolver.lit;
     mutable slabel : bool
   }

exception Subsum_all_tried 
let rec first_non_tried rest list_add_lit =
  match list_add_lit with
  | h::tl -> 
      if (not h.slabel) 
      then (h,(rest@tl))
      else (first_non_tried (h::rest) tl)
  |[] -> raise Subsum_all_tried

let rec prop_subs_lits solver_sim list_add_lit =
  try
    let (lit, to_test_add_lits) = first_non_tried [] list_add_lit in
    lit.slabel<-true;
    let to_test_prop_lits = 
      List.map (function add_lit -> add_lit.sprop_compl_lit) to_test_add_lits
    in  
    match (PropSolver.fast_solve solver_sim to_test_prop_lits) with
    |PropSolver.FUnsat -> 
(*    match (PropSolver.solve_assumptions solver_sim to_test_prop_lits) with
    |PropSolver.Unsat ->  
*)
	( num_prop_subsumption:=!num_prop_subsumption+1;
	 (* out_str("\n Solver_Sim Unsat under:\n");
	  out_str ((list_to_string PropSolver.lit_to_string to_test_prop_lits ";")
		   ^"\n");*)
	  prop_subs_lits solver_sim to_test_add_lits)
    |PropSolver.FSat | PropSolver.FUnknown ->
(*    |PropSolver.Sat ->*)
	    prop_subs_lits solver_sim list_add_lit

    with 
      Subsum_all_tried -> list_add_lit
  

exception Simplified_exists

let  prop_subsumption solver_sim solver gr_by clause = 
  let lits = Clause.get_literals clause in
  let list_add_lit = 
    let f lit = 
      {slit = lit; 
       sprop_lit = (get_prop_lit lit); 
       sprop_compl_lit = (get_prop_compl_lit lit); 
       slabel = false }
    in 
    List.map f lits in
  let new_add_lits = prop_subs_lits solver_sim list_add_lit in
  if (List.length new_add_lits) < (List.length list_add_lit)
  then 
    (eliminate_clause clause;
     num_child_elim:=!num_child_elim-1;
     if (new_add_lits = []) 
     then raise Unsatisfiable 
     else
       (let new_clause   = 
	 Clause.normalise term_db_ref (Clause.create (List.map (function add_lit -> add_lit.slit) new_add_lits)) in
       if (not (ClauseAssignDB.mem new_clause !clause_db_ref))
       then
	 ((try 
	   add_clause_to_solver solver_sim solver gr_by new_clause
	 with PropImplied -> ());
	  let added_clause = 
	    ClauseAssignDB.add_ref new_clause clause_db_ref in             
	  (*(if (PropSolver.fast_solve solver []) = PropSolver.FUnsat  
	  then raise Unsatisfiable);*)
	  (*add_clause_to_solver solver_sim solver gr_by added_clause;*)
	  Clause.assign_when_born !num_of_instantiation_iter  added_clause;
(*	  out_str ("\n Clause: "^(Clause.to_string clause)^"\n");
	  out_str ("\n Simplified to: "^(Clause.to_string added_clause)^"\n");*)
	  (* add_clause_to_unprocessed added_clause;
	     raise Given_clause_simplified *)
	  added_clause)
       else
	 raise Simplified_exists
       )      
    )
  else clause

      
let simplify_given_clause solver_sim solver gr_by clause = 
  let activity = Clause.get_activity clause in
  (if activity > (!max_clause_activity) 
  then max_clause_activity:=activity);
  if ((!num_of_instantiation_iter < !subs_limit) || 
  (activity > (!max_clause_activity)-5 ))
  then
    (*  prop_subs_resolution solver_sim solver gr_by clause *)
    prop_subsumption solver_sim solver gr_by clause 
  else clause

exception Simplified
let simplify_clause_list solver_sim solver gr_by clause_list =
  let simpl_clause rest clause = 
    if (10 >(Clause.when_born clause))
    then
      try 
	let new_clause = prop_subsumption solver_sim solver gr_by clause in
	if (new_clause == clause) then   
	  new_clause::rest
	else 
	  (add_clause_to_unprocessed new_clause; 
	   rest)	
      with 
	  Simplified_exists  -> rest
    else 
      clause::rest
  in List.fold_left simpl_clause [] clause_list
    
let increase_lit_activity solver i lit = 
  try 
    let var_entry = get_prop_gr_var_entry lit in 
    let model_truth_val = 
      match var_entry.truth_val with 
      |Def(PropSolver.True) -> true 
      |Def(PropSolver.False) -> false
      |_ -> raise Activity_Undef 
    in
    if (model_truth_val = false) 
    then 
      (var_entry.neg_activity <- var_entry.neg_activity+i;
       if var_entry.neg_activity > !max_lit_activity 
       then (max_lit_activity:=var_entry.neg_activity
	     (*out_str ("\n Max Lit Act: "^(Term.to_string lit)^"\n"
		      ^(var_entry_to_string solver var_entry))*))
      )
    else 
      (var_entry.pos_activity <- var_entry.pos_activity+i;
       if var_entry.pos_activity > !max_lit_activity 
       then (max_lit_activity:=var_entry.pos_activity
	    (* out_str ("\n Max Lit Act: "^(Term.to_string lit)^"\n"
		      ^(var_entry_to_string solver var_entry))*)
	    )
      )
  with Activity_Undef -> ()


(************************ End Simplification*********************)
  
(************ all_instantiations ********)
let all_instantiations solver_sim solver gr_by main_clause =
  try
    (*we assume that sel in main_clause is checked before *)
    (*on accordance with solver*)
    let sel_lit_tmp = Clause.get_inst_sel_lit main_clause in
    (try 
      (lit_activity_check solver sel_lit_tmp;
       increase_lit_activity solver 1 sel_lit_tmp ) 
    with Activity_Check ->  
      (selection_renew solver_sim solver main_clause));
    let sel_lit = Clause.get_inst_sel_lit main_clause in
    
(*     out_str_debug ("all_instantiations main clause: "
        ^(Clause.to_string main_clause)^"\n");*)
    let compl_sel_lit = compl_lit sel_lit in
    let unif_candidates = 
      DiscrTreeM.unif_candidates !unif_index_ref compl_sel_lit in
    let for_all_candidates (lit,clause_cand_list) =       
(*      out_str_debug ("inst_try cand_list:"^"Sel_lit: "^(Term.to_string  lit)^"\n"
		     ^(Clause.clause_list_to_string clause_cand_list)^"\n"); *)

(*Simplification tun later on *)
      try( 
	lit_activity_check solver lit;

	let clause_list = simplify_clause_list solver_sim solver gr_by clause_cand_list  in
(*      let clause_list = clause_cand_list in*)
      if clause_list != []
      then
	try 
	  (
      (* let var_entry     = get_prop_gr_var_entry lit in
	  if (model_accords_solver solver var_entry)
	  then*)
	    (
	     let conclusion_list = 
	       Inference_rules.instantiation_norm dismatch_switch
		 term_db_ref clause_db_ref main_clause sel_lit compl_sel_lit 
		 clause_list lit in		  
	     increase_lit_activity solver (List.length  conclusion_list) lit;
             let apply_to_concl clause =  
	    (*try *)
(* uncoment this if back to  constraint checking then simplifed *)
	      (* let simplified_clause = simplify_new clause in	 *)
	    let simplified_clause = clause in
(* uncoment this if back to  constrint checking then simplifed *)
	    (* let added_clause = 
		ClauseAssignDB.add_ref clause clause_db_ref in *)          
	    let added_clause = clause in
	(*    try *)
	      add_clause_to_solver solver_sim solver gr_by added_clause;
	      add_clause_to_unprocessed added_clause
	  (*  with PropImplied -> ()*)
	      
	    (* with Clause_Simplified -> ()*)	  	  	    
	   in
	   List.iter apply_to_concl conclusion_list    
	  )
	(*else ()*) (*  model_accords_solver will move all clauses to passive!*)
	)
	with Unif.Unification_failed  -> ()       
      else ()
       )     
      with Activity_Check ->  ()
    in	    
    List.iter for_all_candidates unif_candidates;
  with
    Clause.Inst_sel_lit_undef -> 
      failwith "all_instantiations: clause should have selected literals here"
 
(************ end all_instantiations ********)


(************ all_instantiations_sel with the selection: *) 
(************ lit is sel if it has the least  number     *)
(************ of unif. compl lits in unif index          *)
(*
let all_instantiations_sel solver_sim solver gr_by main_clause =
  try 
    let accord lit = 
      let var_entry     = get_prop_gr_var_entry lit in
      change_model_solver solver var_entry in
    Clause.iter accord main_clause; 
    let sel_cand_lits  = 
      Clause.find_all consistent_with_model main_clause in
(* returns list (lit, unif_cand_list_[] ) *) 
    let lits_unif = 
      let get_unif_cand lit = 
	(lit, (DiscrTreeM.unif_candidates !unif_index_ref (compl_lit lit))) in      
      List.map get_unif_cand sel_cand_lits 
    in
    let comp_cand (l1,unif_list1) (l2,unif_list2) = 
      -(compare (List.length unif_list1) (List.length unif_list2)) in
    let (sel_lit,unif_candidates) = 
      Lib.list_find_max_element comp_cand lits_unif in
    Clause.assign_inst_sel_lit sel_lit main_clause;
    ass_if_consistent sel_lit main_clause;
    let compl_sel_lit = compl_lit sel_lit in
(*old part*)
    let for_all_candidates (lit,clause_list) =       
      (*out_str_debug ("inst_try: "^(Clause.to_string main_clause)*)
      (*^(Clause.clause_list_to_string clause_list)); *) 
      try 
	(let var_entry     = get_prop_gr_var_entry lit in
(*	if (model_accords_solver solver var_entry)
	then*)
	  (
	   let conclusion_list = 
	     Inference_rules.instantiation_norm dismatch_switch
	       term_db_ref clause_db_ref main_clause sel_lit compl_sel_lit 
	       clause_list lit in		  
           let apply_to_concl clause =  
	    (*try *)
(* uncomment this if back to  constraint checking then simplified *)
	      (* let simplified_clause = simplify clause in	 *)
	    let simplified_clause = clause in
(* uncomment this if back to  constraint checking then simplified *)
	    (* let added_clause = 
		ClauseAssignDB.add_ref clause clause_db_ref in *)          
	    let added_clause = clause in
	    add_clause_to_solver solver_sim solver gr_by added_clause;
            add_clause_to_unprocessed added_clause
	    (* with Clause_Simplified -> ()*)	  	  	    
	   in
	   List.iter apply_to_concl conclusion_list    
	  )
(*	else ()*) (*  model_accords_solver will move all clauses to passive!*)
	)
      with Unif.Unification_failed  -> ()       
    in	
    List.iter for_all_candidates unif_candidates
  with
    Clause.Inst_sel_lit_undef -> 
      failwith "all_instantiations: clause should have selected literals here"

*)
(************ end all_instantiations ********)

(**************** instantiation_exhaustive_loop *********************)
(* we exhaustivelly apply instantiations util passive is empty *)
(* then apply prop_solver *)
(*    
let rec instantiation_exhaustive_loop solver_sim solver gr_by =
  let stat_counter = ref 0 in
  while true do  
    (*  out_str_debug 
	("instantiation_exhaustive_loop \n "
	^(string_of_int !num_of_instantiation_iter));*)
    num_of_instantiation_iter := !num_of_instantiation_iter + 1;  
    stat_counter := !stat_counter +1;
(* often output of statistic *)
  (*  (if !stat_counter > 1000 then (stat_counter:=0; out_statistics ()) else ());*)
    try  
    (*  let clause = remove_from_simple_passive () in*)
      let clause = remove_from_passive () in
      if ((Clause.get_bool_param Clause.is_dead clause) ||
      (Clause.get_bool_param Clause.in_active clause))
      then () 
      else
(* sp for simple passive *)
	(selection_renew solver clause;
	 all_instantiations solver_sim solver gr_by clause;
	 add_to_active clause)	
    with
    |Passive_Empty ->
	(if (PropSolver.solve solver) = PropSolver.Unsat 
	then raise Unsatisfiable
	else 
	  try
 	    List.iter add_new_clause_to_passive !unprocessed_ref;
	    unprocessed_ref:=[];
	    apply_new_model solver;
	    num_of_solver_calls := !num_of_solver_calls +1
	   (* out_str_debug (model_sel_to_string ())*)
	  with 
	    Sel_Unchanged -> 
	      if (passive_is_empty ())
	      then raise Satisfiable
	      else())
    |PropSolver.Unsatisfiable -> raise Unsatisfiable
  done

*)

(**************** instantiation loop with each clause added to passive *)
(******************solver called**************)
(* replace here simple passive to passive*)
(*
let rec instantiation_each_loop solver_per_active solver_sim solver gr_by =
  let stat_counter = ref 0 in
  let solver_counter = ref 0 in
  while true do  
  (*  out_str_debug 
      ("instantiation_exhaustive_loop \n "
      ^(string_of_int !num_of_instantiation_iter));*)
    num_of_instantiation_iter := !num_of_instantiation_iter + 1;  
    stat_counter := !stat_counter +1;
    solver_counter:=!solver_counter+1;
(* often output of statistic *)
  (*  (if !stat_counter > 1000 then (stat_counter:=0; out_statistics ()) else ());*)
 
    try  
(* test run solver every loop *)
     (* if (PropSolver.solve solver) = PropSolver.Unsat 
      then raise Unsatisfiable
      else*)       
	(if ((!solver_counter > solver_per_active ) || 
	(passive_is_empty ()))
	then 
	  if (!unprocessed_ref = []) && (passive_is_empty ())
	  then raise Satisfiable
	  else
	    (solver_counter:=0;	 
	     num_of_solver_calls := !num_of_solver_calls +1; 
	   if (PropSolver.solve solver) = PropSolver.Unsat 
	   then raise Unsatisfiable
	   else       
	     (List.iter add_new_clause_to_passive !unprocessed_ref;
	      unprocessed_ref:=[];
	      try
		apply_new_model solver
	      with
		Sel_Unchanged -> 
		  if (passive_is_empty ())
		  then raise Satisfiable
		  else()
	     )	
	  )
	else
	  let clause = remove_from_passive () in       
	if ((Clause.get_bool_param Clause.is_dead clause) ||
	(Clause.get_bool_param Clause.in_active clause))
	then () 
	else	 
	  (selection_renew solver clause;
	   all_instantiations solver_sim solver gr_by clause;
	   add_to_active clause)       
	)
   with	  
    |Passive_Empty -> ()
    |PropSolver.Unsatisfiable -> raise Unsatisfiable
  done
*)
(****************end instantiation_each_loop***************)

(***************instantiation lazy loop***********************)
(*we change model patially and lazily for literals detected having different *)
(* value in the solver vs in the model *)

(*
let rec instantiation_lazy_loop solver_per_active solver gr_by =
  let stat_counter = ref 0 in
  let solver_counter = ref 0 in
  while true do  
  (*  out_str_debug 
      ("instantiation_exhaustive_loop \n "
      ^(string_of_int !num_of_instantiation_iter));*)
    num_of_instantiation_iter := !num_of_instantiation_iter + 1;  
    stat_counter := !stat_counter +1; (* out statistic after some steps*)
    solver_counter:=!solver_counter+1;
(* often output of statistic *)
    (if !stat_counter > 1000 then (stat_counter:=0; out_statistics ()) else ());
    try  
      (if !solver_counter > solver_per_active  
      then       
	(num_of_solver_calls := !num_of_solver_calls +1;
	 if (PropSolver.solve solver) = PropSolver.Unsat 
	 then raise Unsatisfiable
	 else solver_counter:=0));      
      let clause = remove_from_passive () in       
      if 
	((Clause.get_bool_param Clause.is_dead clause) ||
	(Clause.get_bool_param Clause.in_active clause))
      then () 
      else	 
	((*out_str_debug ("Given Clause: "^(Clause.to_string clause)^"\n");*)
	 selection_renew solver clause;
	(* out_str_debug ("Sel in Given: "^ *)
	(*		(Term.to_string (Clause.get_inst_sel_lit clause)^"\n"));*)
	(* out_str_debug (model_sel_to_string solver); *)
	 all_instantiations solver_sim solver gr_by clause;
	 add_to_active clause)       	
    with	  
  |Passive_Empty -> 
	(if (PropSolver.solve solver) = PropSolver.Unsat 
	then raise Unsatisfiable
	else 
	  try
  List.iter add_new_clause_to_passive !unprocessed_ref;
	    unprocessed_ref:=[];
             num_in_unprocessed:=0;
	    apply_new_model solver;
	    num_of_solver_calls := !num_of_solver_calls +1
		(* out_str_debug (model_sel_to_string ())*)
	  with 
	    Sel_Unchanged -> 
	      (if (*!simple_passive_ref =[] *)
		!passive_queue_ref = PassQueue.empty
	      then  ((*out_str_debug (model_sel_to_string solver); *)
		     raise Satisfiable)
	      else())
	)	  
    |PropSolver.Unsatisfiable -> raise Unsatisfiable
  done
*)

(*let ()= add_param_str "all_instantiations_sel\n"*)

(*
let return_solver_state solver = 
  let apply_entry var_entry =
    let prop_var = get_prop_var_var_entry var_entry in
    let prop_neg_var = get_prop_neg_var_var_entry var_entry in
    match var_entry.truth_val with	
    |Def(PropSolver.True) -> 	
	let _= PropSolver.solve_assumptions solver [prop_var] in ()
    |Def(PropSolver.False) ->
	let _= PropSolver.solve_assumptions solver [prop_neg_var] in ()
    |_ -> ()
  in
  TableArray.iter apply_entry var_table
*)


(* auxilary *)
(* nonsence
let rec check_same_sel_desc clause desc_clause = 
  if ((Clause.get_bool_param Clause.in_active desc_clause)
	&((Clause.compare_sel_place clause desc_clause) = 0))
  then 
    (out_str ("parent: "^(Clause.to_string desc_clause)
     ^"Sel: "^(Term.to_string (Clause.get_inst_sel_lit desc_clause))^"\n"); 
     true)
  else
    false
    let parent = Clause.get_parent desc_clause in 
    match parent with 
    |Def(p) -> check_same_sel_desc clause p
    |Undef -> false
    
 
let check_parent_same_sel clause = 
  let parent = Clause.get_parent clause in 
  match parent with 
  |Def(p) -> check_same_sel_desc clause p
  |Undef -> false
*)
    
exception Given_Splitted

(**********************LAZY LOOP BODY*******************************)
let lazy_loop_body 
    split_map_ref solver_counter sover_clause_counter solver_sim solver gr_by =

  num_of_instantiation_iter := !num_of_instantiation_iter + 1;  
  solver_counter:=!solver_counter+1;  
  try  
    (if (!solver_counter > solver_per_active || 
    !num_of_instantiation_iter < !inst_solver_thershold ||
(*     solver_per_new_claues *)
    (ClauseAssignDB.size !clause_db_ref) > 
    (!sover_clause_counter + solver_per_active))
    then       
      (num_of_solver_calls := !num_of_solver_calls +1;
       if ((PropSolver.solve solver) = PropSolver.Unsat ||
       (PropSolver.solve solver_sim) = PropSolver.Unsat)
       then raise Unsatisfiable
       else solver_counter:=0;
	(sover_clause_counter:=(ClauseAssignDB.size !clause_db_ref);
	 if eager_unprocessed_to_passive then
	   (List.iter add_new_clause_to_passive !unprocessed_ref;
	    unprocessed_ref:=[];
	    num_in_unprocessed:=0)
	  else ()
	)
      ));      
    let clause = remove_from_passive () in       
    if 
      ((Clause.get_bool_param Clause.is_dead clause) ||
      (Clause.get_bool_param Clause.in_active clause))
    then () 
    else	 
      ( 
	(*if (PropSolver.solve solver) = PropSolver.Unsat 
	then raise Unsatisfiable
	  else*)
	(try
	  (
	   let new_clause = simplify_given_clause solver_sim solver gr_by clause in
(*	   out_str("\n--------------------------\n");
	   out_str_debug ("\n Given Clause: "^(Clause.to_string clause)^"\n");*)

	   (if  ground_splitting_full 
	   then
	     let split_result = 
	       (Splitting.ground_split_clause 
		  symbol_db_ref term_db_ref split_map_ref new_clause) in
	     if (Splitting.get_num_of_splits split_result)>0 
	     then
	       (eliminate_clause clause;
		let splitted_clauses = Splitting.get_split_list split_result in
		let f new_clause = 
		  let added_clause = 
		    ClauseAssignDB.add_ref new_clause clause_db_ref in             
		  Clause.assign_when_born !num_of_instantiation_iter added_clause;
		  add_clause_to_solver solver_sim solver gr_by added_clause;
		  add_clause_to_unprocessed added_clause;
		(*  out_str ("Splitted_clause: "^(Clause.to_string added_clause)^"\n")*)
		in List.iter f splitted_clauses;	    
		num_of_splits  := !num_of_splits + (Splitting.get_num_of_splits split_result);
		num_of_split_atoms := 
		  !num_of_split_atoms + (Splitting.get_num_of_split_atoms split_result);
		raise Given_Splitted)
	     else ()
	   );
	     
	

(*       out_str_debug ("\n Given after simpl: "^(Clause.to_string new_clause)^"\n");*)

	     selection_renew solver_sim solver new_clause;

(*	    lit_activity_check solver new_clause;    *)

	     
(* let new_clause = simplify_given_clause solver_sim solver gr_by clause in *)



	 (*  out_str ("\n Age Given: "
		    ^(string_of_int(!num_of_instantiation_iter - (Clause.when_born new_clause)))); *)

	(* out_str_debug (model_sel_to_string solver); *)
	

(*	   out_str_debug ("Sel in Given: "^ 
			  (Term.to_string (Clause.get_inst_sel_lit new_clause)^"\n"));*)
(*  out_str("Clauses in DB: "^(string_of_int (ClauseAssignDB.size !clause_db_ref))^"\n");*)
(*Debug*)
  (*     let lits_consist_model = 
	 Clause.find_all consistent_with_model clause in
       out_str_debug ("Act. Clause: "^(Clause.to_string clause)^"\n");
       out_str_debug ("Cons Lits: \n ");
       let out_consist lit = 
	 out_str_debug ((Term.to_string lit)^("\n"))
       in 
       List.iter out_consist lits_consist_model; 
       out_str_debug ("Sel Lit: "
		      ^(Term.to_string (Clause.get_inst_sel_lit clause))^"\n");	   
*)(*End Debug*)
	     
             all_instantiations solver_sim solver gr_by new_clause;
	   
(*     all_instantiations_sel solver gr_by clause;*)	  
	     add_to_active new_clause;
	    )
	  
	  with  Activity_Check ->  add_clause_to_unprocessed clause
	)
	   
       )
	
       	
  with	  
  |Passive_Empty -> 
      (if (PropSolver.solve solver) = PropSolver.Unsat 
      then raise Unsatisfiable
      else 
	(
(* old,  if passive and unprocessed are empty we *)
(* don't need that model accords with the solver, *)
(* it still will be a model for active clauses! *)
 (*  try
 	  List.iter add_new_clause_to_passive !unprocessed_ref;
	  unprocessed_ref:=[];
	  num_in_unprocessed:=0;
	  apply_new_model solver;
	  num_of_solver_calls := !num_of_solver_calls +1
	      (* out_str_debug (model_sel_to_string ())*)
	with 
	  Sel_Unchanged -> 
	    (if (*!simple_passive_ref =[] *)
		(passive_is_empty ()) &&  (!unprocessed_ref=[])
	    then  (*out_str_debug (model_sel_to_string solver); *)
	      raise Satisfiable
	    else())
  *)

	 if !unprocessed_ref=[]
	 then  
	   raise Satisfiable
	 else
	   (List.iter add_new_clause_to_passive !unprocessed_ref;
	    unprocessed_ref:=[];
	    num_in_unprocessed:=0;
	    num_of_solver_calls := !num_of_solver_calls +1)	 
	)
      )
  |PropSolver.Unsatisfiable -> raise Unsatisfiable
      
  |Simplified_exists  -> ()(*(out_str ("\n Simplified_exists\n "))*)
  |Given_Splitted     -> () (*out_str "Given_Splitted\n"*)



(********************** Lazy Loop **********************)

let rec instantiation_lazy_loop 
    split_map_ref  solver_sim solver gr_by =
  let solver_counter = ref 0 in
  let solver_clause_counter = ref 0 in
  let stat_counter   = ref 0 in
  let bound_iter     = ref 0 in 
  out_str !param_str_ref;
  while true do
(* while !bound_iter < 10000 do*)
(*    (if !stat_counter > 1000 then (stat_counter:=0; out_statistics ()) else ());*)
   lazy_loop_body split_map_ref solver_counter solver_clause_counter solver_sim solver gr_by;
    stat_counter := !stat_counter +1;
    bound_iter:=!bound_iter +1
  done  

(********************  Learning  Loop  *****************)

let prop_learning_loop 
    split_map_ref  solver_sim solver gr_by =
  let solver_counter   = ref 0 in
  let solver_clause_counter = ref 0 in
  let learning_bound   = ref learning_start in 
  let learning_factor  = 2 in 
  let learning_counter = ref 0 in
  add_param_str
    (
     "Learning Loop: "
     ^"Start: "^(string_of_int learning_start)
     ^"; Factor: "^(string_of_int learning_factor)
     ^" \n");
  out_str !param_str_ref;
  while true do
    if !learning_counter < !learning_bound 
    then
      (
       lazy_loop_body split_map_ref solver_counter solver_clause_counter solver_sim solver gr_by;
       learning_counter:=!learning_counter+1)
    else
      (learning_bound:=!learning_bound * learning_factor;
       learning_counter:=0;
       num_of_learning_restarts:= !num_of_learning_restarts+1;
(* clear all clause DB, except solver*)
       clause_db_ref :=  
	 (ClauseAssignDB.create_name 
	    ("Instantiation_Clauses_DB: "^(string_of_int !num_of_learning_restarts)));
       
       empty_passive ();
       
(* empty unif index *)
       unif_index_ref  :=  (DiscrTreeM.create ());
(* for all terms set in_unif_index to false  *)
(* change later to terms in unif index only *)
       let f t = 
	 match t with 
	 |Term.Fun(_,_,_) ->  (Term.set_fun_bool_param false  Term.in_unif_index t)
	 |_->() 
       in
       TermDB.iter f  !term_db_ref; 

(* remove sel clauses in var_entries *)
       TableArray.iter (function var_entry -> var_entry.sel_clauses <- []) var_table;
       unprocessed_ref := [];    
       num_in_passive:=0;
       num_in_active:=0;
    (*   out_str("\n Learning Restart\n ");*)
       let add_cl clause =
	 let new_clause   = Clause.create (Clause.get_literals clause) in
(*  let new_clause = Clause.normalise term_db_ref clause in *)
	 let added_clause = 
	   ClauseAssignDB.add_ref new_clause clause_db_ref in             
	 add_clause_to_unprocessed added_clause;
	 Clause.set_bool_param true Clause.in_prop_solver added_clause;
(*debug*)
(*	 out_str ((Clause.to_string added_clause)^"\n"^
		  (string_of_bool (Clause.get_bool_param Clause.in_prop_solver added_clause))^"\n");*)
	 Clause.assign_when_born 0 added_clause in
       List.iter add_cl !init_clause_list_ref	     
	   
   (* out_str_debug ("Learning restart: "^(string_of_int !num_of_restarts)^"\n");*)
   (*      out_statistics ()*)
      ) 
  done


(************************ end lazy loop **********************************)

(*************************Start Instantiation*****************************)

(*(* it's better to simplify before splitting ... add later*)
let simplify_input init_clause_list_ref = 
(* need to add to solver before simplifying *)
    let add_to_prop_solver clause =
      add_clause_to_solver solver_sim solver grounding_term clause
    in List.iter add_to_prop_solver !init_clause_list_ref;
    let subs clause =
    let  new_clause = prop_subsumption clause in
    if ground_splitting_input || ground_splitting_full
    then
      let split_result = 
	(Splitting.ground_split_clause 
	   symbol_db_ref term_db_ref split_map_ref clause) in      
      num_of_splits  := !num_of_splits + (Splitting.get_num_of_splits split_result);
      num_of_split_atoms := 
	!num_of_split_atoms + (Splitting.get_num_of_split_atoms split_result);
      init_clause_list_ref:=Splitting.get_split_list split_result);
*)

(*let remove_false () = 
  let remove_false_clause clause = *)

let start_instantiation  ()= 
  try
(* signals:*)
    let signal_handler signal =
      if 
	(signal = Sys.sigquit ||
	signal = Sys.sigvtalrm ||
	signal = Sys.sigint  ||
	signal = Sys.sigalrm || 	
	signal = Sys.sigterm || 
	signal = Sys.sigtstp ) 
      then
	(out_statistics (); raise Termination_Signal)     
      else failwith "Uknown Signal"
    in  	  
    Sys.set_signal Sys.sigquit (Sys.Signal_handle signal_handler);
    Sys.set_signal Sys.sigint (Sys.Signal_handle signal_handler);
    Sys.set_signal Sys.sigvtalrm (Sys.Signal_handle signal_handler);
(*    Sys.set_signal Sys.sigkill (Sys.Signal_handle signal_handler);*)
    Sys.set_signal Sys.sigalrm  (Sys.Signal_handle signal_handler);
    Sys.set_signal Sys.sigterm (Sys.Signal_handle signal_handler);
    Sys.set_signal Sys.sigtstp (Sys.Signal_handle signal_handler);



 
  (*  lit_sel_fun_ref:=lit_sel_fun;*)

    let grounding_term = get_term_for_grounding  () in
(*    let grounding_term = bot_term in*)
(*  out_str_debug ("Term for grounding: "^(Term.to_string grounding_term)^"\n");*)
    let solver = PropSolver.create_solver () in
(* solver used for simplifications *)
    let solver_sim = PropSolver.create_solver () in
    let equality_axioms = Eq_axioms.axiom_list () in
    init_clause_list_ref:= equality_axioms@(!init_clause_list_ref);  
 (* out_str_debug ("Equality Axioms:\n" 
	      ^(Clause.clause_list_to_string equality_axioms)^"\n"); *)
  (*  out_str_debug 
      ("Init Clauses: \n"
       ^(Clause.clause_list_to_string !init_clause_list_ref));    *)
(* it's better to simplify before splitting ... add later*)
(*    simplify_input init_clause_list_ref;*)
    let split_map_ref = ref (Splitting.create_split_map ()) in
    (if ground_splitting_input || ground_splitting_full
    then
      let split_result = 
	(Splitting.ground_split_clause_list 
	   symbol_db_ref term_db_ref split_map_ref !init_clause_list_ref) in
      
      num_of_splits  := !num_of_splits + (Splitting.get_num_of_splits split_result);
      num_of_split_atoms := 
	!num_of_split_atoms + (Splitting.get_num_of_split_atoms split_result);
      init_clause_list_ref:=Splitting.get_split_list split_result);

(*    out_str_debug 
      ("After Split: \n"
       ^(Clause.clause_list_to_string !init_clause_list_ref));*)
    let add_cl_to_solver_and_unprocessed clause =
      (*try 
	let simplified_clause = simplify_new clause in *)
	(try
	  add_clause_to_solver solver_sim solver grounding_term clause;
	  let added_clause = 
	    ClauseAssignDB.add_ref clause clause_db_ref in             
	  add_clause_to_unprocessed added_clause;
	  Clause.assign_when_born 0 added_clause
	(*out_str_debug ((Clause.to_string added_clause)^"\n")*)
	with PropImplied -> () )
    (*  with New_Clause_Simplified -> ()   *)
    in
    List.iter add_cl_to_solver_and_unprocessed !init_clause_list_ref;
    num_of_solver_calls := !num_of_solver_calls +1;
    if ((PropSolver.solve solver) = PropSolver.Unsat || 
    (PropSolver.solve solver_sim) = PropSolver.Unsat)
    then raise Unsatisfiable
    else 
      (
       (
         (* try
         apply_new_model solver;
       with 
	 Sel_Unchanged ->  
	   if !unprocessed_ref = [] then raise Satisfiable
	   else ()*)
       );
       (
	subs_limit := 100000*(ClauseAssignDB.size !clause_db_ref);
	add_param_str ("Subsumption_limit: "^(string_of_int !subs_limit)^"\n");
	add_param_str  ("Inst_solver_thershold: "
			^(string_of_int !inst_solver_thershold^"\n"));
	add_param_str
	   ("Lazy Loop: "
	    (*^"Solver per New Clauses: "^(string_of_int solver_per_active)*)
            ^"SAT solver per active: "^(string_of_int solver_per_active)^"\n");
	add_param_str
	  ("Lazy Loop: "
            ^"SAT solver per new clauses: "^(string_of_int solver_per_clauses)^"\n");
	add_param_str
	  ("Lit_activity_threshold: "^(string_of_int !lit_activity_threshold)^"\n");
	 (*instantiation_each_loop solver_per_active solver grounding_term*)

	(* *)
	(if learning_loop_flag then
	  prop_learning_loop 
	    split_map_ref solver_sim solver grounding_term
	else
	  instantiation_lazy_loop 
	    split_map_ref  solver_sim solver grounding_term);
	  
(*	       out_str 
		       ("--------Instantiation Exhaustive Loop----------\n");
		 instantiation_exhaustive_loop solver grounding_term*)
(*	
	 prop_learning_loop solver_per_active solver grounding_term
*)
	 )
      )
  with
  |Unsatisfiable |PropSolver.Unsatisfiable -> 
      out_str "PROVED\n";
      out_statistics ()	  	
  |Satisfiable   -> 
      out_str "SATISFIABLE\n";  
      out_statistics ()


 (*out_str (test_sel ())*)


(* 
let start_instantiation () = 
(* get term for grounding*)
  let num_of_symb = SymbolDB.size !symbol_db_ref in
  (if  num_of_symb < !actual_num_of_symb_groups_ref
  then actual_num_of_symb_groups_ref := num_of_symb);  
  partition_symbols !actual_num_of_symb_groups_ref;
    let add_clause clause = 
      add_new_clause_to_passive clause clause
    in
    List.iter add_clause !init_clause_list_ref;
    (* ClauseAssignDB.iter add_caluse !init_clause_db_ref; *)
   (* out_str_debug "initial clauses are added to passive \n";*)
    try  discount_loop () with
    |Unsatisfiable -> 
	out_str "\n PROVED";
	out_statistics ()	  	
    |Satisfiable   -> 
	out_str "\n SATISFIABLE";  
	out_statistics ()

*)

(* tests *)
(*
let test_sel () =  
  let truth_f term = true in
  let sel_lit clause =
    Selection.lit_neg_gr_shallow truth_f clause in
  let to_str rest clause =
    rest^"Clause: "^(Clause.to_string clause)^"\n"
    ^"Sel: "^(Term.to_string (sel_lit clause))^"\n" in
  List.fold_left to_str "" !init_clause_list_ref
*)

end
