open Lib

type term = Term.term

(* follows the interface of compare*)

(* we assume that all useful info is assigned 
  i.e. we have already put term in  term_db*)

(* simple_kbo
   weight of each sybol is 1, 
   precedence is the order in symbol_db
   also use an extension as in our paper *)

(*rewrite since got rid of var_list in terms *)
(* get_var_list is not efficient  *)

(* exception_vars is a list of vars that should be treated as minimal  
   constants, so they should participate,
   t_vars,  s_vars are association list (v,num_of_occureces)
   compare_vars checks  that num of 
   occerences in t of each nonexceptional vars 
   is greater than in s and reterns the list of new exceptional vars*)

exception T_vars_less_than_s_vars
let compare_vars exception_vars t_vars s_vars = 
  let f current_exceptions (v,num_v_in_s) = 
    if List.mem v current_exceptions 
    then current_exceptions
    else 
      try 
	let num_v_in_t = List.assoc v t_vars in 
	if num_v_in_t > num_v_in_s 
	then v::exception_vars
	else 
	  if num_v_in_t < num_v_in_s 
	  then raise T_vars_less_than_s_vars
	  else exception_vars
      with Not_found -> raise T_vars_less_than_s_vars
  in
 List.fold_left f exception_vars s_vars

(* returns cequal if t greater or equal to s and 
   returns cequal+1 if t is strictly greater
   returns cequal-1 if it is not the case*)    

(* uncomment minimal constant when it will be defined*)
let rec general_kbo' 
    get_weight compare_precedece (*minimal_constant*) exception_vars t s = 
  try 
    let new_exception_vars = 
      compare_vars exception_vars 
	(Term.get_var_list t) (Term.get_var_list s) in
    
    let weight_cmp = compare (get_weight t) (get_weight s) in
    if weight_cmp > cequal 
    then cequal+1
    else
      if weight_cmp = cequal 
      then
	(match (t,s) with 
	|(Term.Fun(t_sym,t_args,_),Term.Fun(s_sym,s_args,_)) -> 
	    let sym_cmp = Symbol.compare t_sym s_sym in
	    if sym_cmp > 0 then cequal+1
	    else 
	      if sym_cmp = cequal 
	      then  
		list_compare_lex 
		  (general_kbo' get_weight compare_precedece new_exception_vars) 
		  (Term.arg_to_list t_args) (Term.arg_to_list s_args) 
              else cequal-1 
	|(Term.Fun(_,_,_),Term.Var(_,_)) -> cequal+1 
              (* note that the var on the left occurs in the right *)
	|(Term.Var(_,_), Term.Var(_,_))  -> cequal

	|(Term.Var(_,_), Term.Fun(_,_,_)) -> cequal-1
(* uncomment when minimal constant will be defined

	|(Term.Var(t_v,_), Term.Fun(s_sym,_,_)) ->
            if (Symbol.compare s_sym minimal_constant)=0 
	    then cequal 
	    else cequal-1
*)
	)
      else cequal-1
  with 
    T_vars_less_than_s_vars -> cequal-1 
    
let general_kbo  get_weight compare_precedece t s = 
  general_kbo' get_weight compare_precedece [] t s

let simple_kbo = general_kbo Term.get_num_of_symb Symbol.compare


