
open Lib
open Printf
open Parser_types

let parser_main_fun = Parser_tptp.main
let lexer_token = Lexer_tptp.token

(* when eprover is used, includes are unfolded by eprover*)
(*************** eprover version *************)
let eprover_cmd eprover_path file_name = 
  let base_cmd = eprover_path^"eprover" in
  if (Sys.file_exists base_cmd)
  then base_cmd^" --tstp-format --silent --cnf "^file_name 
  else failwith ("cannot find eprover, please specify an appropriate --eprover_path")

let clausify_channel in_channel =
  let success = ref false in
  Printf.printf "...... Clausification by eprover...";  
  let rec clausify_channel' in_channel out_str_list = 
    try 
      let new_line = input_line in_channel in
(* remove empty lines *)
      if String.length new_line == 0 then 
	clausify_channel' in_channel out_str_list	  
      else
	if new_line = "# CNFization successful!" then       
	(success := true;
	 clausify_channel' in_channel out_str_list)
      else 
	if new_line = "# SZS status: ResourceOut" then 	
	  failwith  "eprover:  # SZS status: ResourceOut" 
	else
  (* remove eprover comments *)
	  if String.get new_line 0 = '#' then 
 	    clausify_channel' in_channel out_str_list
	  else
	    new_line :: (clausify_channel' in_channel out_str_list)
    with
    | End_of_file ->
	List.rev out_str_list
  in
  let clausified_str =  String.concat "\n" (clausify_channel' in_channel []) in
  if !success 
  then 
    (out_str "successful";
     (*(out_str (clausified_str^"\n"));*)
    clausified_str)
  else 
    failwith "fail to clausify with eprover"

let clausify_file eprover_path file_name = 
  let in_channel = 
    Unix.open_process_in (eprover_cmd eprover_path file_name)
  in
  let  out = clausify_channel in_channel in 
  close_in in_channel; 
  out

let parse_lexbuf lexbuf_name lexbuf = 
  let () = init_lexbuf lexbuf in
  let parsed_list=
      try (parser_main_fun lexer_token lexbuf) 
      with  
      |Parsing_fails |_ -> 
          let position = (Lexing.lexeme_end_p lexbuf) in
	  let line_number = position.Lexing.pos_lnum in
	  let fail_str = "Parse error in"^lexbuf_name
	    ^" line: "^(string_of_int line_number) in 
          failwith fail_str
    (*  |_ -> failwith "parse unknown error"*)
  in
  parsed_list

let eprover_clausify_parse eprover_path problem_files =
  let for_one_file rest file_name = 
    let clausified_str =  clausify_file eprover_path file_name in
    let lexbuf = Lexing.from_string clausified_str in
    let clause_list = parse_lexbuf file_name lexbuf in
    clause_list@rest 
  in
  List.fold_left for_one_file [] problem_files

(*************** end eprover version *************)
    
(* parsers without unfolding includes *)
let parse_list file_name_list= 
  let parse_one_file rest file_name = 
    let file = open_in file_name in
    let lexbuf = (Lexing.from_channel file) in
    let ()= init_lexbuf lexbuf in
    let add_parsed_list=
      try (parser_main_fun lexer_token lexbuf) 
      with  
      |Parsing_fails |_ -> 
         let position = (Lexing.lexeme_end_p lexbuf) in
	 let line_number = position.Lexing.pos_lnum in
	  let fail_str = "Parse error in file: "
	    ^file_name^" line: "^(string_of_int line_number) in 
          failwith fail_str
    (*  |_ -> failwith "parse unknown error"*)
   in
    let ()= close_in file in
    add_parsed_list@rest 
  in
  List.fold_left parse_one_file [] file_name_list

(* 
let parsed_input_files = parse_list options.problem_files
*)
 
  
(* circular inputs detected and added once*)
let rec unfold_includes' include_path parsed_list added_files_ref= 
  match parsed_list with
  |Include(file_name,[])::tl ->  
      let full_file_name = include_path^file_name in  
      if List.mem full_file_name !added_files_ref 
      then (unfold_includes' include_path tl added_files_ref)
      else
	(added_files_ref:= full_file_name::!added_files_ref;
      let add_parsed_list = parse_list [full_file_name] in 
 (*         out_str file_name; *)
          unfold_includes' include_path (add_parsed_list@tl) added_files_ref)
	  
  |Include(_,_::_)::tl ->  	  
       failwith "formula selection in include is not supported yet"
  |h::tl-> h::(unfold_includes' include_path tl  added_files_ref)
  |[]->[]
	
let unfold_includes include_path parsed_list = 
  let added_files_ref = ref [] in
  unfold_includes' include_path parsed_list added_files_ref

(*
let parsed_all = unfold_includes parsed_input_files
*)

(* include path ex:  "/home/TPTP-v3.0.1/Problems/"  include('Axioms/SET003-0.ax').*)

let parse_files include_path file_name_list = 
  unfold_includes include_path (parse_list file_name_list) 
   

(*
let all_in_str = parsing_type_to_string  parsed_all
let () = 
  out_str_debug ("Parsed formulas: \n"^all_in_str^"end of parsed formulas \n") 
*)
