open Lib
type symbol
type fast_key 
type stype = 
  |Fun|Pred|Connective|Quantifier

type sproperty = Theory|Split

(* Special Symbols*)
val symb_and        : symbol
val symb_or         : symbol
val symb_impl       : symbol
val symb_forall     : symbol
val symb_exists     : symbol
val symb_neg        : symbol
val symb_true       : symbol
val symb_false      : symbol
val symb_equality   : symbol
val symb_plus       : symbol
val symb_minus      : symbol
val symb_unaryminus : symbol
val symb_bot        : symbol


exception Symbol_fast_key_undef
exception Arity_undef
exception Group_undef

val  create_from_str       : string -> stype -> symbol
val  create_from_str_arity : string -> stype -> int -> symbol 

(* fast key assigned when symbolDB is creating*)
val  assign_fast_key       : symbol -> int -> unit

val assign_group           : symbol -> int -> unit
val assign_is_input        : bool -> symbol  -> unit
val assign_is_skolem       : bool -> symbol  -> unit
val assign_property        : sproperty -> symbol -> unit

val  to_string             : symbol -> string
val  to_prolog             : symbol -> string

val  get_arity             : symbol -> int
val  get_type              : symbol -> stype
val  get_group             : symbol -> int
val  get_is_input          : symbol -> bool
val  get_is_skolem         : symbol -> bool
val  get_property          : symbol -> sproperty param
   
(* compare = compare_fast_key and should not be used before 
   fast_key is assigned i.e. symbolDB build (the same to equal and hash); 
   before that use compare_key *)  


val compare               : symbol -> symbol -> int 
val equal                 : symbol -> symbol -> bool
val compare_key           : symbol -> symbol -> int
val compare_fast_key      : symbol -> symbol -> int
val hash                  : symbol -> int
