type   symbol = Symbol.symbol
      
module SymKey = 
  struct 
    type t       = symbol
    let  compare = Symbol.compare_key
    let  assign_fast_key = Symbol.assign_fast_key
 end
    
module SymbolDBM =  AbstAssignDB.Make (SymKey)


type symbolDB  = 
    {db                       : SymbolDBM.abstDB;
     mutable unused_split_symb_number : int}


let create_name name = 
  let db_ref = ref (SymbolDBM.create_name name) in
  let _=SymbolDBM.add_ref Symbol.symb_and db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_or db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_impl db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_forall db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_exists db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_neg db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_true db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_false db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_equality db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_plus db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_minus db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_unaryminus db_ref in
  let _=SymbolDBM.add_ref Symbol.symb_bot db_ref in


  {db = !db_ref;
   unused_split_symb_number=0;
 }
    
let create ()   = 
  create_name "Symbol_DB"

let get_name db = SymbolDBM.get_name db.db

let mem sym db = SymbolDBM.mem sym db.db 

let remove symb db     =  { db with db = (SymbolDBM.remove symb db.db)}
let find symb db       = SymbolDBM.find symb db.db
let size db            = SymbolDBM.size db.db
let map f db           = { db with db = (SymbolDBM.map f db.db)}
let fold f db a        = SymbolDBM.fold f db.db a
let iter f db          = SymbolDBM.iter f db.db

let add_ref symb db_ref   = 
  let sdb_ref   = ref !db_ref.db in
  let added_sym = SymbolDBM.add_ref symb sdb_ref in
  db_ref:= {!db_ref with db = !sdb_ref};
  added_sym
    
let add sym db  = { db with db = (SymbolDBM.add sym db.db)}

    

let create_new_split_symb symb_db_ref arity = 
  let new_symb_name = ("iProver_split_"
		       ^(string_of_int !symb_db_ref.unused_split_symb_number)) in
  let new_symb = 
    Symbol.create_from_str_arity new_symb_name Symbol.Pred arity in
  Symbol.assign_property Symbol.Split new_symb;
  !symb_db_ref.unused_split_symb_number <- !symb_db_ref.unused_split_symb_number+1;
  add_ref new_symb symb_db_ref 



let to_string symbol_db = 
  SymbolDBM.to_string Symbol.to_string "," symbol_db.db

(*debug*)
let get_greatest_key db = SymbolDBM.get_greatest_key db.db
 
