open Lib

type var      = Var.var
type symbol   = Symbol.symbol
type fast_key = int param
type prop_key = TableArray.key
(*association list for counting the number of occurences of vars in a term*)
type var_list = var num_ass_list

type fun_term_bool_param = int
let in_unif_index    = 0

type args = term list
and term =
  | Fun of symbol * args * fun_info 
  | Var of var * var_info	
and fun_info     = 
    {mutable fun_fast_key  : fast_key;
     mutable num_of_symb : int param; 
     (* mutable var_list   : var_list param; *)
     (* can be very expensive and is used only in resol. for kbo, *)
     (* rewrite it there...*)
     mutable num_of_var   : int param;   (* cheap *)
     mutable fun_grounded : term param; (*term obtained after grounding*)
(* prop. key of term (without grounding) used for simplifiactions *)
     mutable prop_key     : prop_key param;
(* prop. key of term after grounding *)
     mutable prop_gr_key   : prop_key param;
     mutable bool_param    : Bit_vec.bit_vec
   }

and var_info = 
    {mutable var_fast_key : fast_key;
     mutable var_grounded : term param (*term obtained after grounding*)}
      
let empty_fun_info () = 
  {fun_fast_key   = Undef; 
   num_of_symb    = Undef; 
   num_of_var     = Undef; 
   fun_grounded   = Undef; 
   prop_key       = Undef; 
   prop_gr_key    = Undef;  
   bool_param     = Bit_vec.false_vec}

let empty_var_info () = {var_fast_key = Undef; var_grounded = Undef}    
(* easy to reimpliment args as arrays*)
 
let set_fun_bool_param value param term = 
  match term with 
  | Fun(_,_,fun_info) ->
      fun_info.bool_param <- Bit_vec.set value param fun_info.bool_param
  |Var(_,_) -> failwith "term set_fun_bool_param: it's a variable"   

let get_fun_bool_param param term =
  match term with 
  | Fun(_,_,fun_info) ->
      Bit_vec.get param fun_info.bool_param
  |Var(_,_) -> failwith "term get_fun_bool_param: it's a variable"   

type literal = term 

type bound_term = term bind

(* type term = {t : pure_term; info : info} *)  

exception Term_fast_key_undef
exception Term_weight_undef
exception Term_grounding_undef
exception Term_assign_fkey_to_var



let create_var_term var = Var(var,(empty_var_info ()))
let create_var_term_info var var_info = Var(var,var_info) 

exception Num_of_sym_undef

    
let get_num_of_symb term = 
  match term with 
  |Fun(_,_,{num_of_symb = Def(num)}) -> num
  |Var(_,_) -> 1
  |_-> raise Num_of_sym_undef

exception Num_of_var_undef
let get_num_of_var term = 
  match term with 
  |Fun(_,_,{num_of_var = Def(num)}) -> num
  |Var(_,_) -> 1
  |_-> raise Num_of_var_undef

exception Prop_gr_key_undef
let get_prop_gr_key term = 
  match term with 
 |Fun(_,_,{prop_gr_key = Def(key)}) -> key 
 |_-> raise Prop_gr_key_undef 

exception Prop_key_undef
let get_prop_key term = 
  match term with 
 |Fun(_,_,{prop_key = Def(key)}) -> key 
 |_-> raise Prop_key_undef 


let assign_prop_gr_key key term = 
  match term with 
  |Fun(_,_,info) -> info.prop_gr_key <- Def(key)
  |_-> failwith "Prop_gr_key can not be assigned to vars" 

let assign_prop_key key term = 
  match term with 
  |Fun(_,_,info) -> info.prop_key <- Def(key)
  |_-> failwith "Prop_key can not be assigned to vars" 
	

(*  works but got rid of var_list
exception Var_list_is_undef
let get_var_list term = 
  match term with 
  |Fun(_,_,fun_info) -> 
     ( match fun_info.var_list with 
      |Def (vl) -> vl
      |Undef -> raise Var_list_is_undef
     )
  |Var(v,_) -> [(v,1)]

*)
(* not efficient*)
let rec get_var_list term = 
  match term with 
  |Fun(_,args,_) -> 
      let f rest term = append_ass_list (+) (get_var_list term) rest in
      List.fold_left f [] args
  |Var(v,_) -> [(v,1)]



let get_num_of_symb_term_list term_list = 
  let f rest term = rest + (get_num_of_symb term) in
  List.fold_left f 0 term_list
 
let get_num_of_var_term_list term_list = 
  let f rest term = rest + (get_num_of_var term) in
  List.fold_left f 0 term_list
 

let assign_grounding ground_term term = 
   match term with 
  |Fun(_,_,fun_info) -> 
      fun_info.fun_grounded <- Def(ground_term)
  |Var(_,var_info) ->
      var_info.var_grounded <- Def(ground_term)


let get_grounding term = 
   match term with 
  |Fun(_,_,{fun_grounded=Def(grounding)}) -> grounding
  |Var(_,{var_grounded=Def(grounding)}) -> grounding
  |_-> raise Term_grounding_undef


let assign_num_of_symb term = 
  match term with 
  |Fun(_,args,fun_info) ->
      fun_info.num_of_symb <- Def((get_num_of_symb_term_list args)+1)
  |_-> ()

let assign_num_of_var term  = 
  match term with 
  |Fun(_,args,fun_info) ->
      fun_info.num_of_var <- Def((get_num_of_var_term_list args))
  |_-> ()


(* fold_sym folds f for all symbols in the term *)
let rec fold_sym f a term = 
  match term with 
  |Fun(sym,args,fun_info) ->
      List.fold_left (fold_sym f) (f a sym) args 
  |_-> a    

(*
let assign_var_list term = 
   match term with 
    |Fun(_,args,fun_info) ->
      let f rest term = append_ass_list (+) (get_var_list term) rest in
      let var_list = List.fold_left f [] args in
      fun_info.var_list <- Def(var_list)  
   |_ -> ()
*)

(*   all fun info added only when we add term to term db

let create_fun_term symbol term_list = 
  let fun_info = (empty_fun_info ()) in
  let ()=fun_info.num_of_symb <- Def((get_num_of_symb_term_list term_list)) in
  let f rest term = append_ass_list (+) (get_var_list term) rest in
  let var_list = List.fold_left f [] term_list in
  let ()=fun_info.var_list <- Def(var_list) in  
  Fun(symbol,term_list,fun_info)
 *)
   
let create_fun_term symbol term_list = 
  let fun_info = (empty_fun_info ()) in
  Fun(symbol,term_list,fun_info)

let create_fun_term_args  =  create_fun_term


let create_fun_term_info  symbol term_list fun_info = Fun(symbol,term_list,fun_info)

let is_constant t = 
  match t with 
  | Fun(symb,[],_) -> ((Symbol.get_type symb) = Symbol.Fun)
  | _ -> false

let is_ground t = ((get_num_of_var t) = 0)

exception Term_fast_key_is_def
	
let assign_fast_key (t:term) (fkey:int) = 
  match t with 
  | Fun(_,_,fun_info)-> 
      (match fun_info.fun_fast_key  with 
      |Undef ->  fun_info.fun_fast_key <- Def(fkey) 
      |_     -> raise Term_fast_key_is_def
      )
  | Var(_,var_info) -> 
      ( match var_info.var_fast_key with 
      |Undef -> var_info.var_fast_key <- Def(fkey)  
      |_     -> raise Term_fast_key_is_def
       )
	

let assign_all term =  assign_num_of_symb term; assign_num_of_var term 

(* var_list  is used in resolution but not in instantiation 
let assign_all term =  assign_num_of_symb term; assign_var_list term *)



(*raise Term_assign_fkey_to_var*)

let arg_map   = List.map
let arg_map_left = list_map_left

let arg_fold_left = List.fold_left
let arg_fold_right = List.fold_right

let arg_fold_left2 f a l1 l2 = 
  try  List.fold_left2 f a l1 l2  
  with
    Invalid_argument(_) -> 
      failwith "term: fold_left2: argumets are of different length"
	
   
let arg_iter  = List.iter
let arg_iter2 f l1 l2 = 
  try
    List.iter2 f l1 l2
  with  
    Invalid_argument(_)-> 
      failwith "term: iter2: argumets are of different length"
	
let arg_map2 f l1 l2 = 
  try
    List.map2 f l1 l2
  with
    Invalid_argument(_)-> 
      failwith "term: map2: argumets are of different length"
      

let arg_find  = List.find

let arg_find2 = list_find2

(* arg_return_g_if_f2 f g args1 args2 -> first (g h1 h2) such that (f h1 h2) *)
let arg_return_g_if_f2 = list_return_g_if_f2

let arg_find_not_equal = list_find_not_equal

let arg_to_list x = x

let  get_fast_key (t:term) =
  match t with 
  | Fun(_,_,fun_info)->  
      ( 
	match fun_info.fun_fast_key with  
	|Def(fkey) -> fkey 
	|_  -> raise Term_fast_key_undef
       )
  | Var(x,var_info) -> 
      ( 
	match var_info.var_fast_key with  
	|Def(fkey) -> fkey 
	|_  -> raise Term_fast_key_undef
       )
	

(* used in compare_key we assume that var alwyas greater than fun term*)
let compare_fun_var t s =
  match (t,s) with 
  | (Fun(_,_,_),Var(_,_)) -> cless
  | (Var(_,_),Fun(_,_,_)) -> cgreater 
  |_ -> cequal


let rec  compare_key (t1:term)(t2:term)  = 
  match (t1,t2) with 
  |(Fun(symbol1,args1,_),Fun(symbol2,args2,_))->
      let c = Symbol.compare symbol1 symbol2 
      in
      if c  = cequal then
	try (arg_find_not_equal compare_key args1 args2)
	with Not_found -> cequal 
      else c
  | (Var(x,_),Var(y,_)) -> Var.compare x y 
  | _ -> compare_fun_var t1 t2 

let  compare_fast_key (t1:term)(t2:term) = 
 (compare (get_fast_key t1) (get_fast_key t2)) 
 

    
let compare = compare_fast_key


let is_neg_lit literal = 
  match literal with 
  | Fun(sym,_,_) -> 
      if sym == Symbol.symb_neg then true 
      else false
  |_-> false
	
let get_atom literal =
  match literal with 
  | Fun(sym,[t],_) -> 
      if sym == Symbol.symb_neg then t
      else literal  
  |_ -> literal

let is_var term = 
  match term with
  |Var(_,_) -> true
  |_-> false 


(* compare two terms: *)

 
let cmp_ground t1 t2 =
  let gt1 = if (is_ground t1) then 1 else 0 in
  let gt2 = if (is_ground t2) then 1 else 0 in
  Pervasives.compare gt1 gt2

let cmp_num_symb t1 t2 =
  Pervasives.compare (get_num_of_symb t1) (get_num_of_symb t2)

let cmp_num_var t1 t2 =
  Pervasives.compare (get_num_of_var t1) (get_num_of_var t2)

(*pos bigger than neg*)
let cmp_sign t1 t2 =
  let nt1 = if (is_neg_lit t1) then 1 else 0 in
  let nt2 = if (is_neg_lit t2) then 1 else 0 in
  -(Pervasives.compare nt1 nt2)

let is_split t = 
  let atom1 = get_atom t in
  match atom1 with 
  | Fun(symb,_,_)-> 
      if (Symbol.get_property symb) = Def(Symbol.Split)
      then 1
      else -1
  |_-> -1

let cmp_split t1 t2 = 
  Pervasives.compare (is_split t1) (is_split t2)
    

(* use prop_key!
exception Term_porpos_id_not_positive 
exception Term_porpos_id_not_def
let get_propos_id_term t = 
 let id=  
   (match t with 
   | Fun(_,_,{fun_fast_key=Def(id)}) -> id
   | Var(_,{var_fast_key=Def(id)}) -> id
   | _-> raise Term_porpos_id_not_def)
 in 
 if id > 0 then id 
 else raise Term_porpos_id_not_positive 

let get_propos_id lit = 
  if (is_neg_lit lit) 
  then -(get_propos_id_term (get_atom lit))
  else get_propos_id_term lit

let get_propos_gr_id lit = get_propos_id (get_grounding lit) 
*)



(* to string's  *)

let var_int_to_string (v,int) = 
  "("^(Var.to_string v)^","^(string_of_int int)^")" 

let var_list_to_string v_l =
"["^(list_to_string var_int_to_string v_l ",")^"]"
(*  ^(string_of_int(List.length v_l))^"  " *)
 

let rec to_string_term t =   
  match t with
  | Fun(sym,[],_) -> Symbol.to_string sym 
  | Fun(sym,args,_) -> 
       (Symbol.to_string sym)^"("^(list_to_string to_string_term args ",")^")"
  | Var(v,_)     -> Var.to_string v

(*
let to_string_with_var_list t = 
  (to_string_term t)^"-"^(var_list_to_string (get_var_list t))
*)

let to_string_with_num_of_sym t =
 (to_string_term t)^"-num_of_symb -"^(string_of_int (get_num_of_symb t))

(* let to_string = to_string_with_var_list *)

 let to_string = to_string_term  

let term_list_to_string t_list = 
  list_to_string to_string t_list ";"

(*
 let to_string = to_string_with_num_of_symb
*)
(*	
let rec subterm subt  t = 
 if equal sumbt t then True
 else  
   match t with 
   |Fun(_,args,_) -> map_arg subterm args
   |_-> False
*)
