open Lib

type var    = Var.var
type symbol = Symbol.symbol
type args 
type fun_info   
type var_info

(*association list for counting the number of occurences of vars in a term*)
type var_list = var num_ass_list

type term =
   Fun of symbol * args * fun_info
 | Var of var * var_info 

type fun_term_bool_param 
val in_unif_index  : fun_term_bool_param


type literal = term 

type bound_term = term Lib.bind

exception Term_fast_key_undef
exception Term_weight_undef


val create_var_term       : var -> term
val create_var_term_info  : var -> var_info -> term

val create_fun_term       : symbol -> term list -> term
val create_fun_term_args  : symbol -> args -> term
val create_fun_term_info  : symbol -> term list -> fun_info -> term

val get_num_of_symb       : term -> int
val get_num_of_var        : term -> int

val get_var_list          : term -> var_list  

val set_fun_bool_param : bool ->  fun_term_bool_param -> term -> unit
val get_fun_bool_param : fun_term_bool_param -> term -> bool 

(* prop_gr_key for the table prop_var-term *)

type prop_key             = TableArray.key

(* prop. key of term (without grounding) used for simplifiactions *)
exception Prop_key_undef
val get_prop_key          : term -> prop_key
val assign_prop_key       : prop_key -> term -> unit

(* prop. key of term after grounding *)
exception Prop_gr_key_undef
val get_prop_gr_key          : term -> prop_key
val assign_prop_gr_key       : prop_key -> term -> unit
(* *)

exception Term_grounding_undef
val get_grounding         : term -> term

(* use get_prop_key ! *)
(* propositional id of grounding of the literal *)
(* can raise Term_ground_undef*)
(* val get_propos_gr_id : term -> int *)

(*folds f on all symbols in the term*)
val fold_sym : ('a -> symbol -> 'a) -> 'a -> term -> 'a 

(* assign_fast_key is done when building termDB *)
 
val assign_fast_key   : term -> int -> unit
(* only to be used in termDB*)
val assign_num_of_symb : term -> unit
(* first arg is a ground term assigned to the second arg *)
val assign_grounding  : term  -> term -> unit
(* val assign_var_list   : term -> unit *)
val assign_all        : term -> unit

val arg_map          : (term -> term) -> args -> args	
val arg_to_list      : args -> term list
(* explicitly maps from left to right, 
   since order can matter when use imperative features *)
val arg_map_left     : (term -> term) -> args -> args	


val arg_fold_left    : ('a -> term -> 'a)-> 'a -> args -> 'a
val arg_fold_right   : (term -> 'a -> 'a)-> args -> 'a -> 'a 
val arg_fold_left2   : 
    ('a ->  term -> term -> 'a) -> 'a -> args -> args -> 'a

val arg_iter  : (term -> unit) -> args -> unit 
val arg_iter2 : (term -> term -> unit) -> args -> args -> unit


val is_constant : term -> bool
val is_ground : term -> bool
val is_var     : term -> bool 


(* compare = compare_fast_key and should not be used before 
   fast_key is assigned i.e. termDB is build; 
   before that use compare_key *)  

val compare               : term -> term -> int 

(* compare_key impl. as structural equality used for termDB*)
(* variables and variables as terms Var(v,i) can have different fast keys*)
val compare_key           : term -> term -> int
val compare_fast_key      : term -> term -> int

val is_neg_lit            : literal -> bool
val get_atom              : literal -> term


(* compare two terms *)
val cmp_ground   : term -> term -> int 
val cmp_num_symb : term -> term -> int 
val cmp_num_var  : term -> term -> int 
val cmp_sign     : term -> term -> int 
val cmp_split    : term -> term -> int 

 
val to_string : term -> string
(*val subterm          : term -> term -> bool *)
val term_list_to_string : (term list) -> string

(*only for debug*)
val get_fast_key : term -> int
