
type literal       = Term.literal
type symbol        = Symbol.symbol
type literal_list  = literal list
type term_db       = TermDB.termDB
type subst         = Subst.subst
type bound         = int
type bound_subst   = SubstBound.bound_subst
type clause
type bound_clause = clause Lib.bind

(* all boolean param are set false and other Undef*)
val create         : literal_list -> clause
val create_parent  : clause -> literal_list -> clause

(* boolean parameters of te clause*)
type clause_bool_param 

val is_dead                      : clause_bool_param 
val in_clause_db                 : clause_bool_param
val in_active                    : clause_bool_param
val in_unif_index                : clause_bool_param
val in_subset_subsumption_index  : clause_bool_param
val in_subsumption_index         : clause_bool_param
val in_prop_solver               : clause_bool_param
val inst_in_sim_passive          : clause_bool_param 
val inst_pass_queue1             : clause_bool_param 
val inst_pass_queue2             : clause_bool_param 
val res_pass_queue_age           : clause_bool_param 
val res_pass_queue_size          : clause_bool_param 
val res_in_sim_passive           : clause_bool_param 
val res_sel_max                  : clause_bool_param 
val eq_axiom                     : clause_bool_param 
val input_under_eq               : clause_bool_param 
val has_conj_symb                : clause_bool_param 

val set_bool_param : bool ->  clause_bool_param -> clause -> unit
val get_bool_param : clause_bool_param -> clause -> bool 
(* inherit_bool_param param from_c to_c *)
val inherit_bool_param     :  clause_bool_param -> clause -> clause -> unit
(* inherit_bool_param_all  from_c to_c *)
val inherit_bool_param_all :  clause -> clause -> unit

val num_of_symb                 : clause -> int 
val num_of_var                  : clause -> int 

val length                      : clause -> int 

(* can be used in e.g. resolution loops for calculating age*)
val when_born                   : clause -> int

(***************assignments**********)
val assign_when_born         : int -> clause -> unit


val assign_res_sel_lits          : literal_list -> clause -> unit

val assign_inst_sel_lit          : literal -> clause -> unit

(* conjecture distance *)

(* big int  *)
val max_conjecture_dist : int

(* all input conjectors 0 *)
val conjecture_int      : int

(* inherit_conj_dist from_c to_c*)
val inherit_conj_dist : clause -> clause -> unit

(* assign max {int,max_conjecture_dist}*)
val assign_conjecture_distance   : int -> clause -> unit
val cmp_conjecture_distance      : clause -> clause -> int
val get_conjecture_distance      : clause -> int
val cmp_has_conj_symb                : clause -> clause -> int

exception Res_sel_lits_undef
val get_res_sel_lits             : clause -> literal_list
val res_sel_is_def               : clause -> bool

exception Inst_sel_lit_undef
val get_inst_sel_lit            : clause -> literal

val get_parent                  : clause -> clause Lib.param

(* comapares places of two clauses, is used to compare that   *)
(* sel literal in parent corresponds to sel lit in child      *)
(* do not renormalise parents and children!*)
val compare_sel_place           : clause -> clause -> int


type dismatching = Dismatching.constr_list

(*type dismatching = Dismatching.constr_list_feature*)

val assign_dismatching : dismatching -> clause -> unit

exception Dismatching_undef
val get_dismatching : clause -> dismatching

(* history when this clause is obtined by resolution from parents upon_literals*)
 
val assign_resolution_history : clause -> clause list -> literal list -> unit 

(* history when this clause is obtined by factoring from parent upon_literals*)

val assign_factoring_history : clause -> clause -> literal list -> unit

val assign_input_history : clause -> unit

(* first is parent sendon child*)
val add_child : clause -> clause -> unit
val get_children : clause -> clause list

val get_activity : clause -> int
val assign_activity : int -> clause -> unit 


(******)

val assign_all_for_clause_db : clause -> unit

(* only to be used in clauseDB where the fast_key is assigned*)
val assign_fast_key : clause -> int -> unit

(* compare = compare_fast_key and should not be used before 
   fast_key is assigned i.e. clauseDB is build; 
   before that use compare_key the same for equal*)  

val compare  : clause -> clause -> int 
val equal    : clause -> clause -> bool             

(* 
  compare_key impl. as structural equality used for clauseDB 
  we assume that literals in clause are in termDB *)

val compare_key           : clause -> clause -> int
val compare_fast_key      : clause -> clause -> int

(* physical membership  *)
val memq    : literal -> clause -> bool

val exists      : (literal -> bool) -> clause -> bool
val find        : (literal -> bool) -> clause -> literal
val fold        : ('a -> literal -> 'a) -> 'a -> clause -> 'a 
val find_all    : (literal -> bool) -> clause -> literal_list
val partition   : (literal -> bool) -> clause -> literal_list * literal_list
val iter        : (literal -> unit) -> clause -> unit

val get_literals : clause -> literal_list

val is_empty_clause : clause -> bool

val is_ground : clause -> bool

(*Comapre two clauses*)
val cmp_num_var  : clause -> clause -> int
val cmp_num_symb : clause -> clause -> int
val cmp_age      : clause -> clause -> int
val cmp_ground   : clause -> clause -> int

(* folds f on all symbols in the clause *)
val fold_sym : ('a -> symbol -> 'a) -> 'a -> clause -> 'a 


(* returns clause without cut literal, 
  literal should be physically equal to literal in clause, 
  raises Literal_not_found otherwise 
 *)

(*  should use get_literals
val cut_literal : literal -> clause -> clause *)

val apply_bsubst : term_db ref -> bound_subst -> bound_clause -> clause 

val apply_bsubst_norm_subst :  
    term_db ref -> bound_subst -> bound -> clause -> clause * subst

(* during normalisations below obtained literals are added to term_db*)
val normalise_lit_list :  term_db ref -> literal_list ->literal_list
val normalise :  term_db ref -> clause -> clause

type b_litlist = literal_list Lib.bind

val normalise_b_litlist : 
    term_db ref -> bound_subst -> b_litlist -> clause 

val normalise_blitlist_list :
    term_db ref -> bound_subst -> (b_litlist list) -> clause 

 

(*
val normalise_bclause : 
    bound_clause -> bound_subst -> term_db ref -> clause

val normalise_bclause_list : 
    bound_clause list -> bound_subst -> term_db ref -> clause
*)

val to_string : clause -> string
val clause_list_to_string : clause list -> string
val to_string_history : clause -> string
