#!/bin/sh
#------------
# File:    ileancop_bushy.sh
# Version: mptp07_bushy
# Date:    1 July 2007
#------------
# Purpose: Invokes the ileanCoP theorem prover
# Usage:   ileancop_bushy.sh <problem file> [<time limit>]
# Author:  Jens Otten
# Web:     www.leancop.de/ileancop
#------------

#-----------
# Parameters

# set ECLiPSe Prolog path
ECLIPSE=/usr/local/eclipse/bin/i386_linux/eclipse

# set ileanCoP prover path
PROVERPATH=/home/tptp/Systems/ileanCoP---1.2

#----------
# Functions

ileancop()
{
    # Input: $SETTINGS, $TIMELIMIT, Output: $RETURN
    "$ECLIPSE" -b "$PROVERPATH/ileancop_tlimit.pl" \
     -e "leancop_tlimit('$PROVERPATH/ileancop12',$SETTINGS,'$FILE',$TLIMIT)"
    RETURN=$?
    if [ $RETURN -eq 0 ]; then echo " Intuitionistic Theorem for $FILE"; exit 0; fi
}

#-------------
# Main Program

if [ $# -eq 0 -o $# -gt 2 ]; then
    echo "Usage: $0 <problem file> [<time limit>]" >&2
    exit 2
fi

FILE=$1

if [ $# -eq 1 ]; then TIMELIMIT=300
                 else TIMELIMIT=$2
fi

# invoke prover with different settings

SETTINGS="[def,scut,cut,comp(7)]"; TLIMIT=$(( 1 * $TIMELIMIT / 50)); ileancop
if [ $RETURN -eq 1 ]; then echo " Intuitionistic Non-Theorem for $FILE"; exit 1; fi
SETTINGS="[def,conj,scut,cut]"; TLIMIT=$((30 * $TIMELIMIT / 50)); ileancop
SETTINGS="[scut,cut]";          TLIMIT=$((15 * $TIMELIMIT / 50)); ileancop
SETTINGS="[conj,cut]";          TLIMIT=$(( 1 * $TIMELIMIT / 50)); ileancop
SETTINGS="[]";                  TLIMIT=$TIMELIMIT               ; ileancop
if [ $RETURN -eq 1 ]; then echo " Intuitionistic Non-Theorem for $FILE"; exit 1; fi
exit 2
