#!/bin/sh
#------------
# File:    leancop_bushy.sh
# Version: mptp07_bushy
# Date:    1 July 2007
#------------
# Purpose: Invokes the leanCoP theorem prover
# Usage:   leancop_bushy.sh <problem file> [<time limit>]
# Author:  Jens Otten
# Web:     www.leancop.de
#------------

#-----------
# Parameters

# set ECLiPSe Prolog path
ECLIPSE=/usr/local/eclipse/bin/i386_linux/eclipse

# set leanCoP prover path
PROVERPATH=/home/tptp/Systems/leanCoP---2.0

#----------
# Functions

leancop()
{
    # Input: $SETTINGS, $TIMELIMIT, Output: $RETURN
    "$ECLIPSE" -b "$PROVERPATH/leancop_tlimit.pl" \
     -e "leancop_tlimit('$PROVERPATH/leancop20',$SETTINGS,'$FILE',$TLIMIT)"
    RETURN=$?
    if [ $RETURN -eq 0 ]; then echo " SZS status Theorem for $FILE"; exit 0; fi
}

#-------------
# Main Program

if [ $# -eq 0 -o $# -gt 2 ]; then
    echo "Usage: $0 <problem file> [<time limit>]" >&2
    exit 2
fi

FILE=$1

if [ $# -eq 1 ]; then TIMELIMIT=300
                 else TIMELIMIT=$2
fi

# invoke prover with different settings

SETTINGS="[def,scut,cut,comp(7)]";   TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
if [ $RETURN -eq 1 ]; then echo " SZS status CounterSatisfiable for $FILE"; exit 1; fi
SETTINGS="[def,scut,cut]";     TLIMIT=$((25 * $TIMELIMIT / 50)); leancop
SETTINGS="[nodef,scut,cut]";   TLIMIT=$((10 * $TIMELIMIT / 50)); leancop
SETTINGS="[conj,cut]";         TLIMIT=$(( 5 * $TIMELIMIT / 50)); leancop
SETTINGS="[def,scut,cut,reo(23)]";   TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[nodef,scut,cut,reo(29)]"; TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[def,scut,cut,reo(37)]";   TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[conj,cut,reo(41)]";       TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[def,scut,cut,reo(47)]";   TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[]";                 TLIMIT=$TIMELIMIT               ; leancop
if [ $RETURN -eq 1 ]; then echo " SZS status CounterSatisfiable for $FILE"; exit 1; fi
exit 2
