#!/bin/sh
#------------
# File:    leancop_chainy.sh
# Version: mptp07_chainy
# Date:    1 July 2007
#------------
# Purpose: Invokes the leanCoP theorem prover
# Usage:   leancop_chainy.sh <problem file> [<time limit>]
# Author:  Jens Otten
# Web:     www.leancop.de
#------------

#-----------
# Parameters

# set ECLiPSe Prolog path
ECLIPSE=/usr/local/eclipse/bin/i386_linux/eclipse

# set leanCoP prover path
PROVERPATH=/home/tptp/Systems/leanCoP---2.0

#----------
# Functions

leancop()
{
    # Input: $SETTINGS, $TIMELIMIT, Output: $RETURN
    "$ECLIPSE" -b "$PROVERPATH/leancop_tlimit.pl" \
     -e "leancop_tlimit('$PROVERPATH/leancop20',$SETTINGS,'$FILE',$TLIMIT)"
    RETURN=$?
    if [ $RETURN -eq 0 ]; then echo " SZS status Theorem for $FILE"; exit 0; fi
}

#-------------
# Main Program

if [ $# -eq 0 -o $# -gt 2 ]; then
    echo "Usage: $0 <problem file> [<time limit>]" >&2
    exit 2
fi

FILE=$1

if [ $# -eq 1 ]; then TIMELIMIT=300
                 else TIMELIMIT=$2
fi

# invoke prover with different settings

SETTINGS="[nodef,scut,cut,comp(7)]"; TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
if [ $RETURN -eq 1 ]; then echo " SZS status CounterSatisfiable for $FILE"; exit 1; fi
SETTINGS="[nodef,scut,cut]";   TLIMIT=$((30 * $TIMELIMIT / 50)); leancop
SETTINGS="[nodef,conj,cut]";   TLIMIT=$(( 8 * $TIMELIMIT / 50)); leancop
SETTINGS="[conj,cut]";         TLIMIT=$(( 2 * $TIMELIMIT / 50)); leancop
SETTINGS="[nodef,scut,cut,reo( 7)]"; TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[nodef,conj,cut,reo(11)]"; TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[nodef,scut,cut,reo(13)]"; TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[conj,cut,reo(17)]";       TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[nodef,scut,cut,reo(19)]"; TLIMIT=$(( 1 * $TIMELIMIT / 50)); leancop
SETTINGS="[]";                 TLIMIT=$TIMELIMIT               ; leancop
if [ $RETURN -eq 1 ]; then echo " SZS status CounterSatisfiable for $FILE"; exit 1; fi
exit 2
